
#include "fertilityfeature.h"

FertilityFeature::FertilityFeature()
	: Feature()
{
}


FertilityFeature::~FertilityFeature()
{
	//remove derivations
	for(map<const HiddenNode *,SparseArray>::iterator iter = derivations.begin(); iter != derivations.end(); iter++) {
		delete[] iter->second.data;
		delete[] iter->second.index;
	}
}


void FertilityFeature::Init(vector<HiddenNode*> n, double* SFactors, int num, vector<double> & GIZAFert, double GIZASF,int GIZAnumber,int t,int off, int nOfAllF,int gIndex) {
	offset = off;
	gizaIndex = gIndex;
	numberOfAllFeatures = nOfAllF;
	nodes = n;
	ScalingFactors = SFactors;
	number = num;
	type = t;
	//GIZAScalingFactor = GIZASF;
	//GIZAFertility = GIZAFert;
	numberOfGIZAFeatures = GIZAnumber;
	double GIZAProbSum = 0;
	for(int i = 0; i < numberOfGIZAFeatures; i++) {
		GIZAFeatureValue.push_back(exp(GIZAFert[i]));
		GIZAScaledFeatureValue.push_back(exp(GIZAFert[i]*GIZASF));
		GIZAProbSum += GIZAFert[i];
	}
	GIZAFeatureValue.push_back(exp(1 - GIZAProbSum));
	GIZAScaledFeatureValue.push_back(exp((1 - GIZAProbSum)*GIZASF));
	//GIZAFeatureValue.push_back(exp(GIZAProbSum));
	//GIZAScaledFeatureValue.push_back(exp((GIZAProbSum)*GIZASF));

	for(int i = 0; i < nodes.size(); i++) {
		nodes[i]->AddConnection(this);
		messages.insert(make_pair(nodes[i],make_pair(0.5,0.5)));
	}
	assert(GIZAFeatureValue.size() == numberOfGIZAFeatures + 1);
	assert(GIZAScaledFeatureValue.size() == numberOfGIZAFeatures + 1);
}

pair<double,double> FertilityFeature::GetMessage (const HiddenNode * ToNode ) const {
	//return messages[ToNode];
	map<const HiddenNode *,pair<double,double> >::const_iterator iter = messages.find(ToNode);
	if(iter != messages.end()) {
		//cout << "FertilityFeature send message: (" << iter->second.first << "," << iter->second.second << ")" << endl;
		return iter->second;
	}else {
		cerr << "Asking for message to HiddenNode, that not exists" << endl;
		exit(-1);
	}


}

SparseArray FertilityFeature::GetDerivation (const HiddenNode * ToNode) const {
	map<const HiddenNode *,SparseArray >::const_iterator iter = derivations.find(ToNode);
	if(iter != derivations.end()) {
		return iter->second;
	}else {
		cerr << "Asking for derivation to HiddenNode, that not exists" << endl;
		exit(-1);
	} 
}
void FertilityFeature::CalcMessages ( ) {
//	cout << "Number: " << number << endl;
//	cout << "Scaling Factors: " << ScalingFactors[0] << " " << ScalingFactors[1] << " " << ScalingFactors[2] << endl;
	assert(number >= 3);
	int numberOfCalcFeatures = max(number - 2,numberOfGIZAFeatures -1);
	double Prob[numberOfCalcFeatures + 1];
	//double GIZAProb[numberOfGIZAFeatures];
	//all prob, needed for calculation
	double allProb = 1;
	for(int j = 0; j < nodes.size(); j++) {
		pair<double,double> m = nodes[j]->GetMessage(this);
		//cout << "Message form " << j << ": (" << m.first << "," << m.second <<")" << endl;  
		allProb *= (m.first + m.second);
	}
	//cout << "AllProb: " << allProb << endl;


	//calculate message for every node
	for(int i = 0;i < nodes.size(); i++) {
	//cout << "calculate Message to Node: " << i << endl;
		pair<double,double> in = nodes[i]->GetMessage(this);
		pair<double,double> out = make_pair(0,0);

		//init prob
		Prob[0] = 1;
		for(int k = 1; k <= numberOfCalcFeatures; k++) {
			Prob[k] = 0;
		}

		//calc Probabilies
		for(int j = 0; j < nodes.size(); j++) {
			if(i != j) {
				pair<double,double> m = nodes[j]->GetMessage(this);
				for(int k = numberOfCalcFeatures; k > 0;k--) {
					Prob[k] = Prob[k]*m.first + Prob[k-1]*m.second;
				}
				Prob[0] = Prob[0]*m.first;
			}
		}

		//calc message
		double SumProbFirst = 0;
		double SumProbSecond = 0;
		if(numberOfGIZAFeatures == 0) {
			out.first += Prob[0]*ScalingFactors[0];
		}else {
			out.first += Prob[0]*ScalingFactors[0]*GIZAScaledFeatureValue[0];
		}
		//cout << "Prob[0]:" << Prob[0] << " ScalingFactors[0]: " << ScalingFactors[0] << " GIZAScaledFeatureValue[0] " <<GIZAScaledFeatureValue[0] << endl;
		//cout << "out.first: " << out.first << " out.second: " << out.second << endl;
		SumProbFirst += Prob[0];
		for(int k = 1; k <= numberOfCalcFeatures; k++) {
			if(numberOfGIZAFeatures == 0) {
				out.first += Prob[k]*ScalingFactors[min(k,number -1)];
				out.second += Prob[k-1]*ScalingFactors[min(k,number -1)];
			}else {
				out.first += Prob[k]*ScalingFactors[min(k,number -1)]*GIZAScaledFeatureValue[min(k,numberOfGIZAFeatures)];
				out.second += Prob[k-1]*ScalingFactors[min(k,number -1)]*GIZAScaledFeatureValue[min(k,numberOfGIZAFeatures)];		    
			}
			SumProbFirst += Prob[k];
			SumProbSecond += Prob[k - 1];
		//cout << "Prob["<<k<<"]:" << Prob[k] << " ScalingFactors["<< min(k,number -1) <<"]: " << ScalingFactors[min(k,number -1)] << " GIZAScaledFeatureValue["<< min(k,numberOfGIZAFeatures) << "] " <<GIZAScaledFeatureValue[min(k,numberOfGIZAFeatures)] << endl;
			//cout << "out.first: " << out.first << " out.second: " << out.second << endl;
		}
		double restProb = (allProb / (in.first + in.second));
		//cout << "Restprob:" << restProb << " First: " << SumProbFirst << " Second: " << SumProbSecond << endl;
		if(!(restProb - SumProbFirst > -1e-10)) {
			cout << "Diff:" << restProb - SumProbFirst << endl;
			cout << "Restprob:" << restProb << " First: " << SumProbFirst << " Second: " << SumProbSecond << endl;
			cout << "AllProb: " << allProb << endl;
			for(int j = 0; j < nodes.size(); j++) {
				pair<double,double> m = nodes[j]->GetMessage(this);
		//cout << "Message form " << j << ": (" << m.first << "," << m.second <<")" << endl << flush;  
			}
			for(int k = 0; k <= numberOfCalcFeatures; k++) {
				cout << "prob[" << k << "]: " << Prob[k] << endl;
			}
		}
		assert(restProb - SumProbFirst > -1e-10);
		assert(restProb - SumProbSecond > -1e-10);
		if(numberOfGIZAFeatures == 0) {
			out.first += (restProb - SumProbFirst) * ScalingFactors[number -1];
			out.second += (restProb - SumProbSecond) * ScalingFactors[number -1];
		}else {
			out.first += (restProb - SumProbFirst) * ScalingFactors[number -1] * GIZAScaledFeatureValue[numberOfGIZAFeatures];
			out.second += (restProb - SumProbSecond) * ScalingFactors[number -1]* GIZAScaledFeatureValue[numberOfGIZAFeatures];
		}
		//cout << "out.first: " << out.first << " out.second: " << out.second << endl;

		//		cout << "Message befor GIZA: (" << out.first << ";" << out.second << ")" << endl;

		//calc messageGIZA
		/*		if(numberOfGIZAFeatures > 0) {
		double SumGIZAProb = 0;
		SumProbFirst = 0;
		SumProbSecond = 0;
		out.first += Prob[0]*GIZAFertility[0]*GIZAScalingFactor;
		SumProbFirst += Prob[0];
		SumGIZAProb += GIZAFertility[0];
		//cout << "Probs:" << Prob[0];
		for(int k = 1; k < numberOfGIZAFeatures; k++) {
		out.first += Prob[k]*GIZAFertility[k]*GIZAScalingFactor;
		SumProbFirst += Prob[k];
		SumGIZAProb += GIZAFertility[k];
		out.second += Prob[k-1]*GIZAFertility[k]*GIZAScalingFactor;
		SumProbSecond += Prob[k - 1];
		//cout << " " << Prob[k];
	}
		//cout << endl;
		//cout <<"Fertilities: ";
		//for(int k = 0; k < numberOfGIZAFeatures; k++) {
		//  cout << GIZAFertility[k] << " ";
		//}
		
		//cout << endl;
		//cout << "SumGIZAProb: "<<SumProbFirst << endl;
		restProb = (allProb / (in.first + in.second));
		//cout << "Restprob:" << restProb << " First: " << SumProbFirst << " Second: " << SumProbSecond << endl;
		if(!(restProb - SumProbFirst > -1e-10)) {
		cout << "Diff:" << restProb - SumProbFirst << endl;
		cout << "Restprob:" << restProb << " First: " << SumProbFirst << " Second: " << SumProbSecond << endl;
		cout << "AllProb: " << allProb << endl;
		for(int j = 0; j < nodes.size(); j++) {
		pair<double,double> m = nodes[j]->GetMessage(this);
		cout << "Message form " << j << ": (" << m.first << "," << m.second <<")" << endl << flush;  
	}
		for(int k = 0; k < number - 1; k++) {
		cout << "prob[" << k << "]: " << Prob[k] << endl;
	}
	}
		assert(restProb - SumProbFirst > -1e-10);
		assert(restProb - SumProbSecond > -1e-10);
		out.first += (restProb - SumProbFirst) * (1- SumGIZAProb) * GIZAScalingFactor;
		out.second += (restProb - SumProbSecond) * (1 - SumGIZAProb) * GIZAScalingFactor;
	}*/

		//cout << "Message after GIZA: (" << out.first << ";" << out.second << ")" << endl;
		//normalize message
		double sum = out.first + out.second;
		if(isnan(out.first/sum)){
			cout << "Restprob:" << restProb << " First: " << SumProbFirst << " Second: " << SumProbSecond << endl;
			cout << "AllProb: " << allProb << endl;
			for(int j = 0; j < nodes.size(); j++) {
				pair<double,double> m = nodes[j]->GetMessage(this);
				cout << "Message form " << j << ": (" << m.first << "," << m.second <<")" << endl << flush;  
			}
			cout << "First: " << out.first << " Second: " << out.second << " Sum: " << sum << endl;
		}
		out.first /= sum;
		out.second /= sum;
		assert(!isnan(out.first));
		assert(!isnan(out.second));
		//save message
		messages.erase(nodes[i]);
		messages.insert(make_pair(nodes[i],out));
		//cout << "Message generated: (" << out.first << "," << out.second << ")" <<endl;
	}


}


/**
 * @return vector<double>
 */
/*
vector<double> FertilityFeature::GetExpectationValue(){
	double allProb = 1;
	int numberOfCalcFeatures = max(number - 2,numberOfGIZAFeatures -1);
	vector<double> Prob;
	Prob.resize(numberOfCalcFeatures + 1,0);
	Prob[0] = 1;

  //calculate probabilites for the different fertilities by considering the messages
	for(int j = 0; j < nodes.size(); j++) {

    //ignore possible nodes (not a good solution, has to be reconsidered)
		if(nodes[j]->GetGivenValue() != -1) {
			pair<double,double> m = nodes[j]->GetBelief();
      //cout << "Message form " << j << ": (" << m.first << "," << m.second <<")" << endl;  
			allProb *= (m.first + m.second);
			for(int k = numberOfCalcFeatures; k > 0;k--) {
				Prob[k] = Prob[k]*m.first + Prob[k-1]*m.second;
				assert(k < numberOfCalcFeatures + 1);
				assert(k-1 >= 0);
			}
			Prob[0] = Prob[0]*m.first;
      
		}else{
			cout <<"Possible assignment" << endl;
		}
	}

	vector<double> value;
	value.resize(number,0);

	double restProb = allProb;
	for(int k = 0; k < number -1; k++) {
		value[k] = Prob[k];
		restProb -= Prob[k];
		assert(k < numberOfCalcFeatures + 1);
		assert(k >= 0);    
    //cout << "Prob["<<k<< "] = " << Prob[k] << endl;
	}
	value[number-1] = restProb;
  //cout << "restprob: " << restProb << endl;
  //Prob.push_back(restProb);

	if(numberOfGIZAFeatures > 0) {
		double currGIZAFeatureValue = 0;
		restProb = allProb;
		for(int k = 0; k < numberOfGIZAFeatures;k++) {
			currGIZAFeatureValue += Prob[k]*GIZAFeatureValue[k];
			restProb -= Prob[k];
		}
		currGIZAFeatureValue += restProb * GIZAFeatureValue[numberOfGIZAFeatures];
		value.push_back(currGIZAFeatureValue);
	}
	return value;
}
*/

vector<double> FertilityFeature::GetExpectationValue(){
	double allProb = 1;
	int numberOfCalcFeatures = max(number - 2,numberOfGIZAFeatures -1);
	vector<double> Prob;
	Prob.resize(numberOfCalcFeatures + 1,0);
	Prob[0] = 1;

//calculate probabilites for the different fertilities by considering the messages
	for(int j = 0; j < nodes.size(); j++) {

//ignore possible nodes (not a good solution, has to be reconsidered)
		if(nodes[j]->GetGivenValue() != -1) {
			pair<double,double> m = nodes[j]->GetBelief();
//cout << "Message form " << j << ": (" << m.first << "," << m.second <<")" << endl;  
			allProb *= (m.first + m.second);

			for(int k = numberOfCalcFeatures; k > 0;k--) {
				Prob[k] = Prob[k]*m.first + Prob[k-1]*m.second;
				assert(k < numberOfCalcFeatures + 1);
				assert(k-1 >= 0);
			}
			Prob[0] = Prob[0]*m.first;

		}else{
			cout <<"Possible assignment" << endl;
		}
	}

//cacluate probabilite for a fertility of greater or equal number - 1
	double restProb = allProb;
	for(int k = 0; k < number -1; k++) {
		restProb -= Prob[k];
		assert(k < numberOfCalcFeatures + 1);
		assert(k >= 0);    
//cout << "Prob["<<k<< "] = " << Prob[k] << endl;
	}
//cout << "restprob: " << restProb << endl;

	vector<double> value;
	value.resize(number,0);

//find maximal probability
	int max = 0;
	for(int k = 0; k < number - 1; k++) {
		assert(k < numberOfCalcFeatures + 1);
		assert(k >= 0);
		assert(max < numberOfCalcFeatures + 1);
		assert(max >= 0);

		if(Prob[k] > Prob[max]) {
			max = k;
		}
	}

	assert(max < numberOfCalcFeatures + 1);
	assert(max >= 0);
//or is probability for fertility greator or equal number -1 the maximum
	if(restProb > Prob[max]) {
		value[number - 1] = 1;
	}else {
		assert(max < number);
		value[max] = 1;
	}

//detern expectation value of the giza feature
	if(numberOfGIZAFeatures > 0) {
		double currGIZAFeatureValue = 0;
		restProb = allProb;
		for(int k = 0; k < numberOfGIZAFeatures;k++) {
			assert(k < numberOfCalcFeatures + 1);
			assert(k >= 0);

			currGIZAFeatureValue += Prob[k]*GIZAFeatureValue[k];
			restProb -= Prob[k];
		}
		currGIZAFeatureValue += restProb * GIZAFeatureValue[numberOfGIZAFeatures];
		value.push_back(currGIZAFeatureValue);
	}
/*	if(numberOfGIZAFeatures > 0) {
		double currGIZAFeatureValue = 0;
		double maxProb = 0;
		restProb = allProb;
		for(int k = 0; k < numberOfGIZAFeatures;k++) {
			assert(k < numberOfCalcFeatures + 1);
			assert(k >= 0);
			if(maxProb < Prob[k]) {
				currGIZAFeatureValue = Prob[k]*GIZAFeatureValue[k];
				maxProb = Prob[k];
			}
			restProb -= Prob[k];
		}
		if(maxProb < restProb) {
			currGIZAFeatureValue = restProb * GIZAFeatureValue[numberOfGIZAFeatures];
		}
		value.push_back(currGIZAFeatureValue);
	}
*/
	return value;
}

/**
 * @return double
 * @param  featureIndex
 */

vector<double> FertilityFeature::GetExampleValue(){
	vector<double> value;
	value.resize(number,0);
	int fertility = 0;
	for(int j = 0; j < nodes.size(); j++) {
		if(nodes[j]->GetGivenValue() != -1) {
			fertility += nodes[j]->GetGivenValue();
		}
	}
	if(fertility < number) {
		value[fertility] = 1;
	}else {
		value[number - 1] = 1;
	}
	if(numberOfGIZAFeatures > 0) {
		if(fertility < numberOfGIZAFeatures) {
			value.push_back(GIZAFeatureValue[fertility]);
		}else{
			value.push_back(GIZAFeatureValue[numberOfGIZAFeatures]);
		}
	}

	return value;
}

void FertilityFeature::CalcDerivations() {
	assert(number >= 3);
	int numberOfCalcFeatures = max(number - 2,numberOfGIZAFeatures -1);
	double Prob[numberOfCalcFeatures + 1];
	//double GIZAProb[numberOfGIZAFeatures];
	//all prob, needed for calculation
	double allProb = 1;
	for(int j = 0; j < nodes.size(); j++) {
		pair<double,double> m = nodes[j]->GetMessage(this);
		//cout << "Message form " << j << ": (" << m.first << "," << m.second <<")" << endl;  
		allProb *= (m.first + m.second);
	}
	//cout << "AllProb: " << allProb << endl;

	//cout << "Start" << endl;

	//calculate message for every node
	for(int i = 0;i < nodes.size(); i++) {
	//cout << "calculate Message to Node: " << i << endl;
		pair<double,double> in = nodes[i]->GetMessage(this);
		vector<double> a(number + (numberOfGIZAFeatures > 0),0);
		pair<vector<double>,vector<double> > der = make_pair(a,a);
		pair<double,double> out = make_pair(0,0);
		//CalcMessages();
		//pair<double,double> mess = GetMessage(nodes[i]);
		//cout << "Message calc before: (" << mess.first << "," << mess.second << ")" << endl;
		
		//init prob
		Prob[0] = 1;
		for(int k = 1; k <= numberOfCalcFeatures; k++) {
			Prob[k] = 0;
		}
		/*cout << "Giza Features:" << endl;
		for(int j = 0; j <= numberOfGIZAFeatures; j++) {
			cout << j << ": " << log(GIZAFeatureValue[j]) << " " << GIZAFeatureValue[j] << endl;
			
		}*/
		//calc Probabilies
		for(int j = 0; j < nodes.size(); j++) {
			if(i != j) {
				pair<double,double> m = nodes[j]->GetMessage(this);
				for(int k = numberOfCalcFeatures; k > 0;k--) {
					Prob[k] = Prob[k]*m.first + Prob[k-1]*m.second;
				}
				Prob[0] = Prob[0]*m.first;
			}
		}
/*
		for(int k = 0; k <= numberOfCalcFeatures; k++) {
			cout <<"Prob["<<k<<"]: " << Prob[k] << endl; 
		}
*/		
		//calc message
		double SumProbFirst = 0;
		double SumProbSecond = 0;
		if(numberOfGIZAFeatures == 0) {
			der.first[0] += Prob[0]*ScalingFactors[0];
			out.first += Prob[0]*ScalingFactors[0];
		}else {
			der.first[0] += Prob[0]*ScalingFactors[0]*GIZAScaledFeatureValue[0];
			der.first[number] += Prob[0]*ScalingFactors[0]*GIZAScaledFeatureValue[0]* log(GIZAFeatureValue[0]);
			out.first += Prob[0]*ScalingFactors[0]*GIZAScaledFeatureValue[0];
		}
		//cout << "Prob[0]:" << Prob[0] << " ScalingFactors[0]: " << ScalingFactors[0] << " GIZAScaledFeatureValue[0] " <<GIZAScaledFeatureValue[0] << endl;
		//cout << "out.first: " << out.first << " out.second: " << out.second << endl;
		SumProbFirst += Prob[0];
		for(int k = 1; k <= numberOfCalcFeatures; k++) {
			if(numberOfGIZAFeatures == 0) {
				int m = min(k,number -1);
				der.first[m] += Prob[k]*ScalingFactors[m];
				der.second[m] += Prob[k-1]*ScalingFactors[m];
				out.first += Prob[k]*ScalingFactors[m];
				out.second += Prob[k-1]*ScalingFactors[m];
			}else {
				int m = min(k,number -1);
				int n = min(k,numberOfGIZAFeatures);
				der.first[m] += Prob[k]*ScalingFactors[m]*GIZAScaledFeatureValue[n];
				der.first[number] += Prob[k]*ScalingFactors[m]*GIZAScaledFeatureValue[n]* log(GIZAFeatureValue[n]);
				der.second[m] += Prob[k-1]*ScalingFactors[m]*GIZAScaledFeatureValue[n];
				der.second[number] += Prob[k-1]*ScalingFactors[m]*GIZAScaledFeatureValue[n]* log(GIZAFeatureValue[n]);
				out.first += Prob[k]*ScalingFactors[m]*GIZAScaledFeatureValue[n];
				out.second += Prob[k-1]*ScalingFactors[m]*GIZAScaledFeatureValue[n];
			}
			SumProbFirst += Prob[k];
			SumProbSecond += Prob[k - 1];
		//cout << "Prob["<<k<<"]:" << Prob[k] << " ScalingFactors["<< min(k,number -1) <<"]: " << ScalingFactors[min(k,number -1)] << " GIZAScaledFeatureValue["<< min(k,numberOfGIZAFeatures) << "] " <<GIZAScaledFeatureValue[min(k,numberOfGIZAFeatures)] << endl;
		//cout << "out.first: " << out.first << " out.second: " << out.second << endl;
		}
		double restProb = (allProb / (in.first + in.second));
		//cout << "Restprob:" << restProb << " First: " << SumProbFirst << " Second: " << SumProbSecond << endl;
		if(!(restProb - SumProbFirst > -1e-10)) {
			cout << "Diff:" << restProb - SumProbFirst << endl;
			cout << "Restprob:" << restProb << " First: " << SumProbFirst << " Second: " << SumProbSecond << endl;
			cout << "AllProb: " << allProb << endl;
			for(int j = 0; j < nodes.size(); j++) {
				pair<double,double> m = nodes[j]->GetMessage(this);
		//cout << "Message form " << j << ": (" << m.first << "," << m.second <<")" << endl << flush;  
			}
			for(int k = 0; k <= numberOfCalcFeatures; k++) {
				cout << "prob[" << k << "]: " << Prob[k] << endl;
			}
		}
		assert(restProb - SumProbFirst > -1e-10);
		assert(restProb - SumProbSecond > -1e-10);
		if(numberOfGIZAFeatures == 0) {
			der.first[number -1] += (restProb - SumProbFirst) * ScalingFactors[number -1];
			der.second[number -1] += (restProb - SumProbSecond) * ScalingFactors[number -1];
			out.first += (restProb - SumProbFirst) * ScalingFactors[number -1];
			out.second += (restProb - SumProbSecond) * ScalingFactors[number -1];
		}else {
			der.first[number -1] += (restProb - SumProbFirst) * ScalingFactors[number -1] * GIZAScaledFeatureValue[numberOfGIZAFeatures];
			der.first[number] += (restProb - SumProbFirst) * ScalingFactors[number -1] * GIZAScaledFeatureValue[numberOfGIZAFeatures] * log(GIZAFeatureValue[numberOfGIZAFeatures]);
			der.second[number -1] += (restProb - SumProbSecond) * ScalingFactors[number -1]* GIZAScaledFeatureValue[numberOfGIZAFeatures];
			der.second[number] += (restProb - SumProbSecond) * ScalingFactors[number -1]* GIZAScaledFeatureValue[numberOfGIZAFeatures] * log(GIZAFeatureValue[numberOfGIZAFeatures]);

			out.first += (restProb - SumProbFirst) * ScalingFactors[number -1] * GIZAScaledFeatureValue[numberOfGIZAFeatures];
			out.second += (restProb - SumProbSecond) * ScalingFactors[number -1]* GIZAScaledFeatureValue[numberOfGIZAFeatures];
		}
		//normalize message
		double sum = out.first + out.second;
		if(isnan(out.first/sum)){
			cout << "Restprob:" << restProb << " First: " << SumProbFirst << " Second: " << SumProbSecond << endl;
			cout << "AllProb: " << allProb << endl;
			for(int j = 0; j < nodes.size(); j++) {
				pair<double,double> m = nodes[j]->GetMessage(this);
				cout << "Message form " << j << ": (" << m.first << "," << m.second <<")" << endl << flush;  
			}
			cout << "First: " << out.first << " Second: " << out.second << " Sum: " << sum << endl;
		}
		//cout << "Message calc : (" << out.first << "," << out.second<< ")" << endl;
		SparseArray FinalDer;
		FinalDer.size = number + (numberOfGIZAFeatures > 0);
		FinalDer.data = new double[number + (numberOfGIZAFeatures > 0)];
		FinalDer.index = new int[number + (numberOfGIZAFeatures > 0)];
		for(int j = 0; j < number; j ++) {
			//cout << "Der[" << j << "]: (" << derFirst << "," << derSecond << ")" << endl;
			FinalDer.index[j] = offset + j;
			FinalDer.data[j] = der.first[j] * out.second - der.second[j] * out.first;
			FinalDer.data[j] /= sum*sum;
			assert(!isnan(FinalDer.data[j]));
			//cout << "Der[" << j << "]: (" << der.first[j] << "," << der.second[j] <<")" << endl;	
			//cout << "Fert Index " << j  << " : " << FinalDer.index[j] << endl;
		}
		if(numberOfGIZAFeatures != 0) {
			//cout << "Der[" << gizaIndex << "]: (" << derFirst << "," << derSecond <<")" << endl;
			FinalDer.index[number] = gizaIndex; 
			FinalDer.data[number] =  der.first[number]* out.second - der.second[number] * out.first;
			FinalDer.data[number] /= sum*sum;
			assert(!isnan(FinalDer.data[number]));
			//cout << "Der[" << gizaIndex << "]: (" << der.first[gizaIndex] << "," << der.second[gizaIndex] <<")" << endl;
			//cout << "Fert Index " << number  << " : " << FinalDer.index[number] << endl;

		}
		if(derivations.count(nodes[i]) > 0) {
			delete[] derivations[nodes[i]].data;
			delete[] derivations[nodes[i]].index;
			derivations.erase(nodes[i]);			
		}
		derivations.insert(make_pair(nodes[i],FinalDer));
		
		//cout << "Message generated: (" << out.first << "," << out.second << ")" <<endl;
	}
	
}
