
#include "fertilityfeatureset.h"
#include <string.h>

FertilityFeatureSet::FertilityFeatureSet(const char* ParamFile, vector<double> & Factors)
	: FeatureSet(ParamFile,Factors)
{
	useGIZAFertilities = 0;
	GIZASourceScalingFactor = 0;
	GIZATargetScalingFactor = 0;
	numberOfGIZAFertilities = 0;
	paramFile = ParamFile;
	if(ParamFile != NULL) {
		LoadParamFile(ParamFile);
	}
  
	assert(Factors.size() % 2 == 0);
	cout << "Initlize Fertility Feature Set" << endl;
	ScalingFactors = new double[Factors.size()];
	cout << "ScalingFactor: ";
	for(int i = 0; i < Factors.size(); i++) {
		ScalingFactors[i] = exp(Factors[i]);
		cout << ScalingFactors[i] << " "; 
	}
	cout << endl;
	numberOfScalingFactors = Factors.size()/2;
	if(useGIZAFertilities) {
		Factors.push_back(GIZASourceScalingFactor);
		Factors.push_back(GIZATargetScalingFactor);
		InitDefaultGIZAFertilities(); 
	}
}


FertilityFeatureSet::~FertilityFeatureSet()
{
	delete[] ScalingFactors;
}

  /**
 * @param  G
   */
void FertilityFeatureSet::GenerateFeatures (Graph * G , int offset, int numberOfAllFeatures ) {
	SourceLength = G->GetSourceLength();
	TargetLength = G->GetTargetLength();
	sourceFeatures = new FertilityFeature[SourceLength];
	targetFeatures = new FertilityFeature[TargetLength];
	for(int i = 0; i < SourceLength; i++) {
		vector<HiddenNode *> links;
		vector<double> GIZAFert;
		for(int j = 0; j < TargetLength; j++) {
			links.push_back(G->GetNode(i,j));
		}
		if(useGIZAFertilities) {
			const string source = G->GetSourceWord(i);
			map<string, vector<double> >::iterator iter = GIZASourceFertilities.find(source);
			if(iter != GIZASourceFertilities.end()) {
				GIZAFert = iter->second;
			}else {
				cout << "could not find GIZA fert for \"" << source << "\"" << endl;
				GIZAFert = defaultGIZASourceFertilities;
/*			for(int i = 0; i < numberOfGIZAFertilities; i++) {
				cout << GIZAFert[i] << " ";
			}
				cout << endl;
				*/		      //exit(-1);
			}
		    //cout << "Source:" << source;
		    //for(int i = 0; i < GIZAFert.size(); i++){
		    //cout << GIZAFert[i] << "_";
		    ///}
		    //cout << endl;
		}
		assert(numberOfGIZAFertilities <= GIZAFert.size());
		sourceFeatures[i].Init(links,ScalingFactors,numberOfScalingFactors,GIZAFert,GIZASourceScalingFactor,numberOfGIZAFertilities,2,offset,numberOfAllFeatures,offset+2*numberOfScalingFactors);
	}
	for(int i = 0; i < TargetLength; i++) {
		vector<HiddenNode *> links;
		vector<double> GIZAFert;
		for(int j = 0; j < SourceLength; j++) {
			links.push_back(G->GetNode(j,i));
		}
		if(useGIZAFertilities) {
			const string target = G->GetTargetWord(i);
			map<string, vector<double> >::iterator iter = GIZATargetFertilities.find(target);
			if(iter != GIZATargetFertilities.end()) {
				GIZAFert = iter->second;
			}else {
				cout << "could not find GIZA fert for " << target << endl;
				GIZAFert = defaultGIZATargetFertilities;
/*			for(int i = 0; i < numberOfGIZAFertilities; i++) {
				cout << GIZAFert[i] << " ";
			}
				cout << endl;
				*/		      //exit(-1);
			}
		    //cout << "Target:" << target;
		    //for(int i = 0; i < GIZAFert.size(); i++){
		    //cout << GIZAFert[i] << "_";
		    //}
		    //cout << endl;
		}
		targetFeatures[i].Init(links,&ScalingFactors[numberOfScalingFactors],numberOfScalingFactors,GIZAFert,GIZATargetScalingFactor,numberOfGIZAFertilities,3,offset + numberOfScalingFactors,numberOfAllFeatures,offset+2*numberOfScalingFactors+1);
	}

}
void FertilityFeatureSet::RemoveFeatures() {
	delete[] sourceFeatures;
	delete[] targetFeatures;
}
  /**
 * @return FertilitySetIterator
   */
FertilityFeatureSetIterator * FertilityFeatureSet::GetIterator ( ) const {
	return new FertilityFeatureSetIterator(SourceLength,TargetLength,sourceFeatures,targetFeatures);

};

void FertilityFeatureSet::EstimateScalingFactores() {
	EstimatedDerivation.clear();
	vector<double> ExpectValue;
	vector<double> ExampleValue;
	int vectorSize = numberOfScalingFactors * 2;
	if(useGIZAFertilities) {
		vectorSize += 2;
	}
	ExpectValue.resize(vectorSize,0);
	ExampleValue.resize(vectorSize,0);
	for(int i = 0; i < SourceLength; i++) {
		vector<double> TempExpectValue = sourceFeatures[i].GetExpectationValue();
		vector<double> TempExampleValue = sourceFeatures[i].GetExampleValue();
		assert(TempExpectValue.size() *2 == ExpectValue.size());
		assert(TempExpectValue.size() *2 == ExampleValue.size());
		for(int j = 0; j < numberOfScalingFactors; j++) {
			ExpectValue[j] += TempExpectValue[j];
			ExampleValue[j] += TempExampleValue[j];
		}
		if(useGIZAFertilities) {
			ExpectValue[2*numberOfScalingFactors] += TempExpectValue[numberOfScalingFactors];
			ExampleValue[2*numberOfScalingFactors] += TempExampleValue[numberOfScalingFactors];
		}
	}

	for(int i = 0; i < TargetLength; i++) {
		vector<double> TempExpectValue = targetFeatures[i].GetExpectationValue();
		vector<double> TempExampleValue = targetFeatures[i].GetExampleValue();
		assert(TempExpectValue.size() *2 == ExpectValue.size());
		assert(TempExpectValue.size() *2 == ExampleValue.size());
		for(int j = 0; j < numberOfScalingFactors; j++) {
			ExpectValue[numberOfScalingFactors + j] += TempExpectValue[j];
			ExampleValue[numberOfScalingFactors + j] += TempExampleValue[j];
		}
		if(useGIZAFertilities) {
			ExpectValue[2*numberOfScalingFactors + 1] += TempExpectValue[numberOfScalingFactors];
			ExampleValue[2*numberOfScalingFactors + 1] += TempExampleValue[numberOfScalingFactors];
		}
	}

	for(int j = 0; j < vectorSize; j++) {
		double deltaL = ExampleValue[j] - ExpectValue[j];
    //cout << "Calculating Der[" << j << "]: " << ExampleValue[j] << " - " <<  ExpectValue[j] << " = " << deltaL << endl;
		EstimatedDerivation.push_back(deltaL);
	}
}

void FertilityFeatureSet::SetScalingFactors(vector<double> & newScalingFactors) {
	if(useGIZAFertilities) {
		assert(newScalingFactors.size() == 2 * numberOfScalingFactors + 2);
	}else {
		assert(newScalingFactors.size() == 2 * numberOfScalingFactors);
	}
	for(int i = 0; i < 2 * numberOfScalingFactors; i++) {
		ScalingFactors[i] = exp(newScalingFactors[i]);
	}

	if(useGIZAFertilities) {
		GIZASourceScalingFactor = newScalingFactors[2*numberOfScalingFactors];
		GIZATargetScalingFactor = newScalingFactors[2*numberOfScalingFactors + 1];
    //cout << "Set GIZASourceScalingFactor to: " << newScalingFactors[numberOfScalingFactors] << endl;
	}

}

void FertilityFeatureSet::LoadParamFile(const char* ParamFileName) {
	ifstream ParamFile( ParamFileName);

	if (!ParamFile) {
		cerr << "File " << ParamFileName << " could not be opened\n";
		exit(-2);
	}
  
	char *pKey = new char[ 200 ];
	char *pValue = new char[ 200 ];
// TODO: Does not work with LF.Param with empty line at end!!!!!!!!!!!!
	while ( !ParamFile.eof() ) {
		ParamFile >> pKey >> pValue;
		if(strcmp(pKey,"SourceGIZAFertilityFile") == 0) {    
			useGIZAFertilities = 1;
			Utils::LoadGIZAFertilities(pValue,GIZASourceFertilities);
		}else if(strcmp(pKey,"TargetGIZAFertilityFile") == 0) {
			useGIZAFertilities = 1;
			Utils::LoadGIZAFertilities(pValue,GIZATargetFertilities);
		}else if(strcmp(pKey,"GIZASourceScalingFactor") == 0) {
			useGIZAFertilities = 1;
			GIZASourceScalingFactor = atof(pValue);
		}else if(strcmp(pKey,"GIZATargetScalingFactor") == 0) {
			useGIZAFertilities = 1;
			GIZATargetScalingFactor = atof(pValue);
		}else if(strcmp(pKey,"NumberOfGIZAFertilities") == 0) {
			useGIZAFertilities = 1;
			numberOfGIZAFertilities = atoi(pValue);
		}
	}


	assert((useGIZAFertilities == 0) || (GIZASourceFertilities.size() > 0));
	assert((useGIZAFertilities == 0) || (GIZATargetFertilities.size() > 0));
	assert((useGIZAFertilities == 0) || (numberOfGIZAFertilities > 0));

	ParamFile.close();
	delete[] pKey;
	delete[] pValue;    
}



void FertilityFeatureSet::OutputScalingFactors() {
	cout << "FertilityFeatureScalingFactors ";
	for(int i = 0; i < 2 * numberOfScalingFactors; i++) {
		cout << log(ScalingFactors[i]) << "_";
	}
	cout << endl;
	if(useGIZAFertilities) {
		string fileName = string(paramFile) + ".opt";    
		ofstream out(fileName.c_str());
		out << "GIZASourceScalingFactor " << GIZASourceScalingFactor << endl;
		out << "GIZATargetScalingFactor " << GIZATargetScalingFactor << endl;

		out.close();

	}

}

void FertilityFeatureSet::InitDefaultGIZAFertilities() {
	double f1 = 1/sqrt(2*PI);
	for(int i = 0; i < numberOfGIZAFertilities; i++) {
		double f = f1*exp(-0.5*(i-1)*(i-1));
		defaultGIZASourceFertilities.push_back(f);
		defaultGIZATargetFertilities.push_back(f);
	}


}
