#ifndef FERTILITYFEATURESET_H
#define FERTILITYFEATURESET_H

#include <featureset.h>
#include <hiddennode.h>
#include <fertilityfeature.h>
#include <fertilityfeaturesetiterator.h>
#include <iostream>
#include <fstream>
#include <utils.h>

#define PI 3.14159

/**
	@author Jan Niehues <jan@niehues.info>
*/
class FertilityFeatureSet : public FeatureSet
{
private:
	double * ScalingFactors;
	vector<double> EstimatedDerivation;
	int numberOfScalingFactors;
	int SourceLength;
	int TargetLength;
	int useGIZAFertilities;
	const char * paramFile;
	double GIZATargetScalingFactor;
	double GIZASourceScalingFactor;
	vector<double> defaultGIZASourceFertilities;
	vector<double> defaultGIZATargetFertilities;
	map <string,vector<double> > GIZASourceFertilities;
	map <string,vector<double> > GIZATargetFertilities;
	int numberOfGIZAFertilities;
	FertilityFeature *sourceFeatures;
	FertilityFeature *targetFeatures;

	void LoadParamFile(const char* ParamFileName);
	void InitDefaultGIZAFertilities();

public:
    FertilityFeatureSet(const char* ParamFile, vector<double> & ScalingFactores);

    ~FertilityFeatureSet();

  /**
   * @param  G
   */
  void GenerateFeatures (Graph * G  , int offset, int numberOfAllFeatures);
   
  /**
   * @param newScalingFactors
   */
  void SetScalingFactors(vector<double> & newScalingFactors);

  void RemoveFeatures();

  /**
   * @return FertilitySetIterator
   */
  FertilityFeatureSetIterator * GetIterator ( ) const;
  /**
   */
  void EstimateScalingFactores ();

  vector<double> inline GetEstimatedDerivations() const {return EstimatedDerivation;};

  virtual void OutputScalingFactors();

};

#endif
