
#include "fertilitylopfeature.h"

FertilityLOPFeature::FertilityLOPFeature()
 : Feature()
{
	numberOfGIZAFeatures = 0;
}


FertilityLOPFeature::~FertilityLOPFeature()
{
	//remove derivations
	for(map<const HiddenNode *,SparseArray>::iterator iter = derivations.begin(); iter != derivations.end(); iter++) {
		delete[] iter->second.data;
		delete[] iter->second.index;
	}
	delete[] potentialValues;
	delete[] ScaledFeatureValues;
}


void FertilityLOPFeature::Init(vector<HiddenNode*> n, double* SFactors, int num, vector<double> & GIZAFert, double * GIZASF,int GIZAnumber,int t,int off, int nOfAllF,int gIndex,int numberOfExp, double * expWeight) {
	numberOfExperts = numberOfExp;
	expertWeights = expWeight;
	offset = off;
	gizaIndex = gIndex;
	numberOfAllFeatures = nOfAllF;
	nodes = n;
	ScalingFactors = SFactors;
	number = num;
	type = t;
	numberOfGIZAFeatures = GIZAnumber;
	double GIZAProbSum = 0;

	int numberOfPotentials = max(numberOfGIZAFeatures + 1,number);
	ScaledFeatureValues = new double[numberOfPotentials * numberOfExperts];
	
	for(int i = 0; i < numberOfPotentials; i++) {
		for(int j = 0; j < numberOfExperts; j++) {	
			int index = min(i,number - 1);
			ScaledFeatureValues[i*numberOfExperts + j] = ScalingFactors[index*numberOfExperts + j];
		}
	}
	if(numberOfGIZAFeatures > 0) {
		for(int i = 0; i < numberOfGIZAFeatures; i++) {
			for(int j = 0; j < numberOfExperts; j++) {
				ScaledFeatureValues[i*numberOfExperts + j] += GIZAFert[i] * GIZASF[j];
			}
			GIZAProbSum += GIZAFert[i];
		}
		for(int i = numberOfGIZAFeatures; i < numberOfPotentials; i++) {
			for(int j = 0; j < numberOfExperts; j++) {
				ScaledFeatureValues[i*numberOfExperts + j] += (1-GIZAProbSum)* GIZASF[j];
			}
		}
	}

	potentialValues = new double[numberOfPotentials];
	for(int i = 0; i < numberOfPotentials; i++) {
		potentialValues[i] = 0;
		for(int j = 0; j < numberOfExperts; j++) {
			potentialValues[i] += expertWeights[j] * ScaledFeatureValues[i*numberOfExperts + j];
		}
		potentialValues[i] = exp(potentialValues[i]);
	}


	for(int i = 0; i < nodes.size(); i++) {
		nodes[i]->AddConnection(this);
		messages.insert(make_pair(nodes[i],make_pair(0.5,0.5)));
	}



}

pair<double,double> FertilityLOPFeature::GetMessage (const HiddenNode * ToNode ) const {
	//return messages[ToNode];
	map<const HiddenNode *,pair<double,double> >::const_iterator iter = messages.find(ToNode);
	if(iter != messages.end()) {
		//cout << "FertilityFeature send message: (" << iter->second.first << "," << iter->second.second << ")" << endl;
		return iter->second;
	}else {
		cerr << "Asking for message to HiddenNode, that not exists" << endl;
		exit(-1);
	}


}

SparseArray FertilityLOPFeature::GetDerivation (const HiddenNode * ToNode) const {
	map<const HiddenNode *,SparseArray >::const_iterator iter = derivations.find(ToNode);
	if(iter != derivations.end()) {
		return iter->second;
	}else {
		cerr << "Asking for derivation to HiddenNode, that not exists" << endl;
		exit(-1);
	} 
}

void FertilityLOPFeature::CalcMessages ( ) {
	assert(number >= 3);
	int numberOfCalcFeatures = max(number - 2,numberOfGIZAFeatures -1);
	double Prob[numberOfCalcFeatures + 1];

	//all prob, needed for calculation
	double allProb = 1;
	for(int j = 0; j < nodes.size(); j++) {
		pair<double,double> m = nodes[j]->GetMessage(this);
		//cout << "Message form " << j << ": (" << m.first << "," << m.second <<")" << endl;  
		allProb *= (m.first + m.second);
	}


	//calculate message for every node
	for(int i = 0;i < nodes.size(); i++) {
	//cout << "calculate Message to Node: " << i << endl;
		pair<double,double> in = nodes[i]->GetMessage(this);
		pair<double,double> out = make_pair(0,0);

		//init prob
		Prob[0] = 1;
		for(int k = 1; k <= numberOfCalcFeatures; k++) {
			Prob[k] = 0;
		}

		//calc Probabilies
		for(int j = 0; j < nodes.size(); j++) {
			if(i != j) {
				pair<double,double> m = nodes[j]->GetMessage(this);
				for(int k = numberOfCalcFeatures; k > 0;k--) {
					Prob[k] = Prob[k]*m.first + Prob[k-1]*m.second;
				}
				Prob[0] = Prob[0]*m.first;
			}
		}

		//calc message
		double SumProbFirst = 0;
		double SumProbSecond = 0;
		out.first += Prob[0]*potentialValues[0];
		SumProbFirst += Prob[0];

		for(int k = 1; k <= numberOfCalcFeatures; k++) {
			out.first += Prob[k]*potentialValues[k];
			out.second += Prob[k-1]*potentialValues[k];
			SumProbFirst += Prob[k];
			SumProbSecond += Prob[k - 1];
		}

		double restProb = (allProb / (in.first + in.second));
		if(!(restProb - SumProbFirst > -1e-10)) {
			cout << "Diff:" << restProb - SumProbFirst << endl;
			cout << "Restprob:" << restProb << " First: " << SumProbFirst << " Second: " << SumProbSecond << endl;
			cout << "AllProb: " << allProb << endl;
			for(int j = 0; j < nodes.size(); j++) {
				pair<double,double> m = nodes[j]->GetMessage(this);
			}
			for(int k = 0; k <= numberOfCalcFeatures; k++) {
				cout << "prob[" << k << "]: " << Prob[k] << endl;
			}
		}
		assert(restProb - SumProbFirst > -1e-10);
		assert(restProb - SumProbSecond > -1e-10);
		out.first += (restProb - SumProbFirst) * potentialValues[numberOfCalcFeatures + 1];
		out.second += (restProb - SumProbSecond) * potentialValues[numberOfCalcFeatures + 1];

		//normalize message
		double sum = out.first + out.second;
		if(isnan(out.first/sum)){
			cout << "Restprob:" << restProb << " First: " << SumProbFirst << " Second: " << SumProbSecond << endl;
			cout << "AllProb: " << allProb << endl;
			for(int j = 0; j < nodes.size(); j++) {
				pair<double,double> m = nodes[j]->GetMessage(this);
				cout << "Message form " << j << ": (" << m.first << "," << m.second <<")" << endl << flush;  
			}
			cout << "First: " << out.first << " Second: " << out.second << " Sum: " << sum << endl;
		}
		out.first /= sum;
		out.second /= sum;
		assert(!isnan(out.first));
		assert(!isnan(out.second));
		//save message
		messages.erase(nodes[i]);
		messages.insert(make_pair(nodes[i],out));
	}

	
}


/**
 * @return vector<double>
 */
vector<double> FertilityLOPFeature::GetExpectationValue(){
	//cout << "Fertility Expectation Value " << endl;
	double allProb = 1;
	int numberOfCalcFeatures = max(number - 2,numberOfGIZAFeatures -1);
	vector<double> Prob;
	Prob.resize(numberOfCalcFeatures + 1,0);
	Prob[0] = 1;

	//calculate probabilites for the different fertilities by considering the messages
	for(int j = 0; j < nodes.size(); j++) {

		//ignore possible nodes (not a good solution, has to be reconsidered)
		if(nodes[j]->GetGivenValue() != -1) {
			pair<double,double> m = nodes[j]->GetBelief();
			allProb *= (m.first + m.second);

			for(int k = numberOfCalcFeatures; k > 0;k--) {
				Prob[k] = Prob[k]*m.first + Prob[k-1]*m.second;
				assert(k < numberOfCalcFeatures + 1);
				assert(k-1 >= 0);
			}
			Prob[0] = Prob[0]*m.first;

		}else{
			cout <<"Possible assignment" << endl;
		}
	}

	//cacluate probabilite for a fertility of greater or equal number - 1
	double restProb = allProb;
	for(int k = 0; k < number -1; k++) {
		restProb -= Prob[k];
		assert(k < numberOfCalcFeatures + 1);
		assert(k >= 0);    
	}

	vector<double> value;
	value.resize(numberOfExperts,0);

	//calculate derivation
	for(int k = 0; k <= numberOfCalcFeatures; k++) {
		for(int j = 0; j < numberOfExperts; j++) {
			//cout << "value[j] += " << Prob[k] << " * " << ScaledFeatureValues[k*numberOfExperts + j] << endl;
			value[j] += Prob[k]*ScaledFeatureValues[k*numberOfExperts + j];
		}
	}
	for(int j = 0; j < numberOfExperts; j++) {
		//cout << "value[j] += " << restProb << " * " << ScaledFeatureValues[(numberOfCalcFeatures + 1)*numberOfExperts + j] << endl;
		value[j] += restProb*ScaledFeatureValues[(numberOfCalcFeatures + 1)*numberOfExperts + j];
	}
	return value;
}

/**
 * @return double
 * @param  featureIndex
 */

vector<double> FertilityLOPFeature::GetExampleValue(){
	//cout << "Fertility::GetExampleValue " << endl;
	vector<double> value;
	value.resize(numberOfExperts,0);
	int numberOfPotentials = max(numberOfGIZAFeatures + 1,number);
	int fertility = 0;
	for(int j = 0; j < nodes.size(); j++) {
		if(nodes[j]->GetGivenValue() != -1) {
			fertility += nodes[j]->GetGivenValue();
		}
	}
	//cout << "Fertility: " << fertility<<  endl;
	if(fertility < numberOfPotentials) {
		for(int j = 0; j < numberOfExperts; j++) {
			value[j] = ScaledFeatureValues[fertility*numberOfExperts + j];
		}
	}else {
		for(int j = 0; j < numberOfExperts; j++) {
			value[j] = ScaledFeatureValues[(numberOfPotentials - 1)* numberOfExperts + j];
		}
	}
	//cout << "Result: ";
	for(int i = 0; i < numberOfExperts; i++) {
/*		if(value[i] > 100) {
			cout << "Fertility: " << fertility << endl;
			cout << "numberOfPotentials: " << numberOfPotentials << endl;
			cout << "Value:" << value[i] << endl;
		} 
*/	}
	//cout << endl;

	return value;
}

void FertilityLOPFeature::CalcDerivations() {
	assert(number >= 3);
	int numberOfCalcFeatures = max(number - 2,numberOfGIZAFeatures -1);
	double Prob[numberOfCalcFeatures + 1];

	//all prob, needed for calculation
	double allProb = 1;
	for(int j = 0; j < nodes.size(); j++) {
		pair<double,double> m = nodes[j]->GetMessage(this);
		//cout << "Message form " << j << ": (" << m.first << "," << m.second <<")" << endl;  
		allProb *= (m.first + m.second);
	}

	//calculate message for every node
	for(int i = 0;i < nodes.size(); i++) {
		pair<double,double> in = nodes[i]->GetMessage(this);
		vector<double> a(numberOfExperts,0);
		pair<vector<double>,vector<double> > der = make_pair(a,a);
		pair<double,double> out = make_pair(0,0);
		
		//init prob
		Prob[0] = 1;
		for(int k = 1; k <= numberOfCalcFeatures; k++) {
			Prob[k] = 0;
		}

		//calc Probabilies
		for(int j = 0; j < nodes.size(); j++) {
			if(i != j) {
				pair<double,double> m = nodes[j]->GetMessage(this);
				for(int k = numberOfCalcFeatures; k > 0;k--) {
					Prob[k] = Prob[k]*m.first + Prob[k-1]*m.second;
				}
				Prob[0] = Prob[0]*m.first;
			}
		}
		//calc message
		double SumProbFirst = 0;
		double SumProbSecond = 0;
		
		for(int j = 0; j < numberOfExperts; j++) {
			der.first[j] += Prob[0] * potentialValues[0] * ScaledFeatureValues[j];
		}
		out.first += Prob[0]*potentialValues[0];
		SumProbFirst += Prob[0];
		

		for(int k = 1; k <= numberOfCalcFeatures; k++) {
			for(int j = 0; j < numberOfExperts; j++) {
				der.first[j] += Prob[k]*potentialValues[k] * ScaledFeatureValues[k*numberOfExperts + j];
				der.second[j] += Prob[k-1] * potentialValues[k] * ScaledFeatureValues[k*numberOfExperts + j];
			}
			out.first += Prob[k]*potentialValues[k];
			out.second += Prob[k-1]*potentialValues[k];
			SumProbFirst += Prob[k];
			SumProbSecond += Prob[k - 1];
		}

		double restProb = (allProb / (in.first + in.second));
		if(!(restProb - SumProbFirst > -1e-10)) {
			cout << "Diff:" << restProb - SumProbFirst << endl;
			cout << "Restprob:" << restProb << " First: " << SumProbFirst << " Second: " << SumProbSecond << endl;
			cout << "AllProb: " << allProb << endl;
			for(int j = 0; j < nodes.size(); j++) {
				pair<double,double> m = nodes[j]->GetMessage(this);
			}
			for(int k = 0; k <= numberOfCalcFeatures; k++) {
				cout << "prob[" << k << "]: " << Prob[k] << endl;
			}
		}
		assert(restProb - SumProbFirst > -1e-10);
		assert(restProb - SumProbSecond > -1e-10);
		for(int j = 0; j < numberOfExperts; j++) {
			der.first[j] = (restProb - SumProbFirst) * potentialValues[numberOfCalcFeatures + 1] * ScaledFeatureValues[(numberOfCalcFeatures + 1)*numberOfExperts + j];
			der.second[j] = (restProb - SumProbSecond) * potentialValues[numberOfCalcFeatures + 1] * ScaledFeatureValues[(numberOfCalcFeatures + 1)*numberOfExperts + j];
		}
		out.first += (restProb - SumProbFirst) * potentialValues[numberOfCalcFeatures + 1];
		out.second += (restProb - SumProbSecond) * potentialValues[numberOfCalcFeatures + 1];


		//normalize message
		double sum = out.first + out.second;
		if(isnan(out.first/sum)){
			cout << "Restprob:" << restProb << " First: " << SumProbFirst << " Second: " << SumProbSecond << endl;
			cout << "AllProb: " << allProb << endl;
			for(int j = 0; j < nodes.size(); j++) {
				pair<double,double> m = nodes[j]->GetMessage(this);
				cout << "Message form " << j << ": (" << m.first << "," << m.second <<")" << endl << flush;  
			}
			cout << "First: " << out.first << " Second: " << out.second << " Sum: " << sum << endl;
		}

		//calculate final derivations
		SparseArray FinalDer;
		FinalDer.size = numberOfExperts;
		FinalDer.data = new double[FinalDer.size];
		FinalDer.index = new int[FinalDer.size];
		for(int j = 0; j < numberOfExperts; j ++) {
			FinalDer.index[j] = j;
			FinalDer.data[j] = der.first[j] * out.second - der.second[j] * out.first;
			FinalDer.data[j] /= sum*sum;
			assert(!isnan(FinalDer.data[j]));
		}
		if(derivations.count(nodes[i]) > 0) {
			delete[] derivations[nodes[i]].data;
			delete[] derivations[nodes[i]].index;
			derivations.erase(nodes[i]);			
		}
		derivations.insert(make_pair(nodes[i],FinalDer));
		
	}
	
}



