#ifndef FERTILITYLOPFEATURESET_H
#define FERTILITYLOPFEATURESET_H

#include <featureset.h>
#include <fertilitylopfeature.h>
#include <fertilitylopfeaturesetiterator.h>
#include <map>
#include <utils.h>

#define PI 3.14159
/**
	@author Jan Niehues <jan@niehues.info>
*/
class FertilityLOPFeatureSet : public FeatureSet
{
private:
	int numberOfExperts; /**< Number Of Experts of the model. */  
	double * expertWeights; /**< Weights for the different experts */
	double * ScalingFactors;
	vector<double> EstimatedDerivation;
	int numberOfScalingFactors;
	int SourceLength;
	int TargetLength;
	int useGIZAFertilities;
	const char * paramFile;
	double * GIZATargetScalingFactors;
	double * GIZASourceScalingFactors;
	vector<double> defaultGIZASourceFertilities;
	vector<double> defaultGIZATargetFertilities;
	map <string,vector<double> > GIZASourceFertilities;
	map <string,vector<double> > GIZATargetFertilities;
	int numberOfGIZAFertilities;
	FertilityLOPFeature *sourceFeatures;
	FertilityLOPFeature *targetFeatures;

	void LoadParamFile(const char* ParamFileName);
	void InitDefaultGIZAFertilities();



public:
    FertilityLOPFeatureSet(const char* ParamFile, const char * ScalingFactors,int numberOfExpert, double * expertWeights);

    ~FertilityLOPFeatureSet();


  /**
   * @return int
   */
  virtual inline int GetNumberOfScalingFactors() {return 2*(numberOfScalingFactors + useGIZAFertilities);};

  /**
   * @param  G
   */
  void GenerateFeatures (Graph * G  , int offset, int numberOfAllFeatures);
   
  void RemoveFeatures();

  /**
   * @return FertilitySetIterator
   */
  FertilityLOPFeatureSetIterator * GetIterator ( ) const;
  /**
   */
  void EstimateScalingFactores ();

  vector<double> inline GetEstimatedDerivations() const {return EstimatedDerivation;};

};

#endif
