#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <iostream>
#include <sstream>
#include <fstream>
#include <cstdlib>
#include <vector>
#include <map>
#include <cassert>
using namespace std;
struct PhraseItem {
	string source;
	string target;
	vector<string> sourceParts;
	vector<string> targetParts;
	int index;
};
typedef multimap<string,PhraseItem>::iterator PTIterator;

ifstream * LoadFile(const char* File);
void ReadLine(ifstream * stream, vector<string> & words);
void SearchPhrases(vector<string> & sourceSentence, vector<string> & targetSentence, multimap<string,PhraseItem>  & phraseTable
);
void SearchDiscontinousPhrases(vector<string> & sourceSentence, vector<string> & targetSentence, multimap<string,PhraseItem>  & phraseTable
);
void Tokenize(const string& str,vector<string>& tokens, const string& delimiters = " ");
void LoadPhraseTable(const char * fileName,multimap<string,PhraseItem> & phraseTable, int discontinous);
void OutputCounts();
map<pair<string,string>,int> outputTable;

int main(int argc, char *argv[])
{
	if((argc < 4) || (argc > 5)) {
		cout << "Usage: " << argv[0] << " SourceFile TargetFile PhraseTable [Discontinous]" << endl;
		exit(-1);
	}
	char* sourceFile = argv[1];
	char* targetFile = argv[2];
	char* phraseTableFile = argv[3];
	int discontinous = 0;
	if(argc == 5) {
		discontinous = atoi(argv[4]);
	}
	cerr << "Find Possible Phrases" << endl;
	cerr << "SourceFile :          " << sourceFile << endl;
	cerr << "TargetFile :          " << targetFile << endl;
	cerr << "PhraseTable:          " << phraseTableFile << endl;
	cerr << "Discontinous Phrases: " << (discontinous == 0 ? "Off" : "On" )<< endl;

	ifstream * source = LoadFile(sourceFile);
	ifstream * target = LoadFile(targetFile);

	multimap<string,PhraseItem> phraseTable;
	LoadPhraseTable(phraseTableFile, phraseTable,discontinous);

	vector<string> sourceWords;
	vector<string> targetWords;
	int Sent = 0;
	while((!source->eof()) && (!target->eof())) 
	{	
			Sent ++;
			if(Sent%10000 == 0) {
				cout << "Sentence: " << Sent << endl;
			}
			ReadLine(source,sourceWords);
			ReadLine(target,targetWords);
			if(discontinous == 0) {
				SearchPhrases(sourceWords,targetWords,phraseTable);
			}else {
				SearchDiscontinousPhrases(sourceWords,targetWords,phraseTable);
			}
			sourceWords.clear();
			targetWords.clear();
	}


	delete source;
	delete target;

	OutputCounts();
}


ifstream * LoadFile(const char* File) {
	ifstream * stream = new ifstream(File);
	if(!(*stream)) {
		cerr << "ERROR: File " << File <<" could not be open\n";
		exit(-2);
	}
	return stream;
}

void ReadLine(ifstream * stream, vector<string> & words) {
	assert(words.size() == 0);
	string line = "";
	string word = "";
	getline(*stream,line);
	if(line == "") {
		return;
	}
	line.push_back(' ');
	//cout << "Read Line:" << line << endl;
	stringstream sstream(std::stringstream::in|std::stringstream::out);
	sstream << line;
	sstream >> word;
	while (!sstream.eof()) {
		words.push_back(word);
		sstream >> word;
	}	
}


void SearchPhrases(vector<string> & sourceSentence, vector<string> & targetSentence, multimap<string,PhraseItem>  & phraseTable) {
	multimap<string,PhraseItem> targetTable;
	int maxPhraseLength = 4;

	//find matching source phrases
	int sourceLength = sourceSentence.size();
	for(int i = 0; i < sourceLength; i++) {
		string source = "";
		for(int j = 0; (j < maxPhraseLength) && (i + j < sourceLength); j++) {
			if(j != 0) {
				source.push_back(' ');
			}
			source.append(sourceSentence[i+j]);
			//search for phrases starting at this position
			pair<PTIterator,PTIterator> iter = phraseTable.equal_range(source);
			for(PTIterator begin = iter.first; begin != iter.second; begin++) {
				PhraseItem item = begin->second;
				targetTable.insert(make_pair(item.target,item));
			}

		}
	}
	//check if target phrases match also
	int targetLength = targetSentence.size();
	for(int i = 0; i < targetLength; i++) {
		string target = "";
		for(int j = 0; (j < maxPhraseLength) && (i + j < targetLength); j++) {
			if(j != 0) {
				target.push_back(' ');
			}
			target.append(targetSentence[i + j]);
			//check matching target phrases
			pair<PTIterator,PTIterator> iter = targetTable.equal_range(target);
			for(PTIterator begin = iter.first; begin != iter.second; begin++) {
				PhraseItem item = begin->second;
				//phrase matches complete
				outputTable[make_pair(item.source,item.target)]++;
				//cout << "Found item: "<< item.source << " # " << item.target << endl;
			}
		}
	}

}

void SearchDiscontinousPhrases(vector<string> & sourceSentence, vector<string> & targetSentence, multimap<string,PhraseItem>  & phraseTable) {
	multimap<string,PhraseItem> targetTable;
	int maxPhraseLength = 4;
	vector<vector<PhraseItem> > sourceNextPart;
	multimap<string,PhraseItem> sourceContinue;

	//find matching source phrases
	int sourceLength = sourceSentence.size();
	vector<PhraseItem> e;
	sourceNextPart.assign(sourceLength,e);

	for(int i = 0; i < sourceLength; i++) {
		string source = "";
		for(int j = 0; j < sourceNextPart[i].size(); j++) {
			sourceContinue.insert(make_pair(sourceNextPart[i][j].sourceParts[sourceNextPart[i][j].index],sourceNextPart[i][j]));
		}
		for(int j = 0; (j < maxPhraseLength) && (i + j < sourceLength); j++) {
			if(j != 0) {
				source.push_back(' ');
			}
			source.append(sourceSentence[i+j]);
			//search for phrases starting at this position
			pair<PTIterator,PTIterator> iter = phraseTable.equal_range(source);
			for(PTIterator begin = iter.first; begin != iter.second; begin++) {
				PhraseItem item = begin->second;
				if(item.sourceParts.size() == item.index + 1) {
					item.index = 0;
					targetTable.insert(make_pair(item.targetParts[0],item));
				}else if(i + j + 1 < sourceLength) {
					item.index ++;
					sourceNextPart[i + j + 1].push_back(item);
				}
			}
			//search for phrases starting at this position
			iter = sourceContinue.equal_range(source);
			for(PTIterator begin = iter.first; begin != iter.second; begin++) {
				PhraseItem item = begin->second;
				if(item.sourceParts.size() == item.index + 1) {
					item.index = 0;
					targetTable.insert(make_pair(item.targetParts[0],item));
				}else if(i + j + 1 < sourceLength){
					item.index ++;
					sourceNextPart[i + j + 1].push_back(item);
				}
			}

		}
	}

	vector<vector<PhraseItem> > targetNextPart;
	multimap<string,PhraseItem> targetContinue;

	//check if target phrases match also
	int targetLength = targetSentence.size();
	targetNextPart.assign(targetLength,e);
	for(int i = 0; i < targetLength; i++) {
		string target = "";
		for(int j = 0; j < targetNextPart[i].size(); j++) {
			targetContinue.insert(make_pair(targetNextPart[i][j].targetParts[targetNextPart[i][j].index],targetNextPart[i][j]));
		}
		for(int j = 0; (j < maxPhraseLength) && (i + j < targetLength); j++) {
			if(j != 0) {
				target.push_back(' ');
			}
			target.append(targetSentence[i + j]);
			//check matching target phrases
			pair<PTIterator,PTIterator> iter = targetTable.equal_range(target);
			for(PTIterator begin = iter.first; begin != iter.second; begin++) {
				PhraseItem item = begin->second;
				if(item.targetParts.size() == item.index + 1) {
					item.index = 0;
					outputTable[make_pair(item.source,item.target)]++;
				}else if(i + j + 1 < targetLength){
					item.index ++;
					targetNextPart[i + j + 1].push_back(item);
				}
				//phrase matches complete
				//cout << "Found item: "<< item.source << " # " << item.target << endl;
			}
			iter = targetContinue.equal_range(target);
			for(PTIterator begin = iter.first; begin != iter.second; begin++) {
				PhraseItem item = begin->second;
				if(item.targetParts.size() == item.index + 1) {
					item.index = 0;
					outputTable[make_pair(item.source,item.target)]++;
				}else if(i + j + 1 < targetLength){
					item.index ++;
					targetNextPart[i + j + 1].push_back(item);
				}
				//phrase matches complete
				//cout << "Found item: "<< item.source << " # " << item.target << endl;
			}
		}
	}

}

void LoadPhraseTable(const char * fileName,multimap<string,PhraseItem> & phraseTable, int discontinous) {
	const int BufferSize = 1024;
	char Line [BufferSize];
	ifstream file( fileName);

	if (!file) {
		cerr << "PhraseTable " << fileName << " could not be opened\n";
		exit(-2);
	}
	
	while(file.getline(Line,BufferSize)) {
		string line(Line);
		vector<string> elements;
		//cout <<  "TD Line: " << line << endl;
		Tokenize(line,elements,string("#"));
		
		PhraseItem item;

		//Trim spaces
		// Find the first character position after excluding leading blank spaces
		size_t startpos = elements[0].find_first_not_of(" \t"); 
	 	size_t endpos = elements[0].find_last_not_of(" \t"); // Find the first character position from reverse af	
		// if all spaces or empty return an empty string
		if(( string::npos == startpos ) || ( string::npos == endpos)) {
			cout << "Empty PT Entry" << endl;
			exit(1);
		}else {
		        item.source = elements[0].substr( startpos, endpos-startpos+1 );
		}
		startpos = elements[1].find_first_not_of(" \t"); 
	 	endpos = elements[1].find_last_not_of(" \t"); // Find the first character position from reverse af	
		// if all spaces or empty return an empty string
		if(( string::npos == startpos ) || ( string::npos == endpos)) {
			cout << "Empty PT Entry" << endl;
			exit(1);
		}else {
		        item.target = elements[1].substr( startpos, endpos-startpos+1 );
		}
		//cout << "Load: |" << item.source <<"| |" << item.target << "|" << endl;
		if(discontinous == 0) {
			phraseTable.insert(make_pair(item.source,item));
		}else {
			Tokenize(item.source,item.sourceParts,string("|"));
			for(int i = 0; i < item.sourceParts.size(); i++) {
				//Trim spaces
				// Find the first character position after excluding leading blank spaces
				startpos = item.sourceParts[i].find_first_not_of(" \t"); 
			 	endpos = item.sourceParts[i].find_last_not_of(" \t"); // Find the first character position from reverse af	
				// if all spaces or empty return an empty string
				if(( string::npos == startpos ) || ( string::npos == endpos)) {
					cout << "Empty part of phrase" << endl;
					exit(1);
				}else {
				        item.sourceParts[i] = item.sourceParts[i].substr( startpos, endpos-startpos+1 );
				}	
			}
			Tokenize(item.target,item.targetParts,string("|"));
			for(int i = 0; i < item.targetParts.size(); i++) {
				//Trim spaces
				// Find the first character position after excluding leading blank spaces
				startpos = item.targetParts[i].find_first_not_of(" \t"); 
			 	endpos = item.targetParts[i].find_last_not_of(" \t"); // Find the first character position from reverse af	
				// if all spaces or empty return an empty string
				if(( string::npos == startpos ) || ( string::npos == endpos)) {
					cout << "Empty part of phrase" << endl;
					exit(1);
				}else {
				        item.targetParts[i] = item.targetParts[i].substr( startpos, endpos-startpos+1 );
				}	
			}
	
			item.index = 0;
			phraseTable.insert(make_pair(item.sourceParts[0],item));	
		}
		outputTable.insert(make_pair(make_pair(item.source,item.target),0));	
	}

}

void OutputCounts() {
	for(map<pair<string,string>,int>::iterator iter = outputTable.begin(); iter != outputTable.end(); iter ++) {
		cout << iter->first.first << " # " << iter->first.second << " # " << iter->second << endl;
	}
}

void Tokenize(const string& str,
	      vector<string>& tokens,
       const string& delimiters)
{
    // Skip delimiters at beginning.
	string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
	string::size_type pos     = str.find_first_of(delimiters, lastPos);

	while (string::npos != pos || string::npos != lastPos)
	{
        // Found a token, add it to the vector.
		tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
		lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
		pos = str.find_first_of(delimiters, lastPos);
	}
}
