
#include "firstorderfeature.h"

FirstOrderFeature::FirstOrderFeature()
 : Feature()
{
	m1 = make_pair(0.5,0.5);
	m2 = make_pair(0.5,0.5);
	der1.size = 0;
	der2.size = 0;
	
}


FirstOrderFeature::~FirstOrderFeature()
{
	if(der1.size > 0) {
		delete[] der1.data;
		delete[] der1.index;
		delete[] der2.data;
		delete[] der2.index;
	}
}

void FirstOrderFeature::Init(pair<HiddenNode*,HiddenNode*> n, double* SFactors, int num,vector<double> dynScalingFactor, vector<int> dynScalingIndex,int offS, int dynOffSet,int nOfAllF){
	nodes = n;
	number = num;
	offset = offS;
	dynamicOffset = dynOffSet;
	numberOfAllFeatures = nOfAllF;
	//	cout << "Number :" << number << endl;
	ScalingFactors = SFactors;
	nodes.first->AddConnection(this);
	nodes.second->AddConnection(this);
	dynamicScalingFactor = ScalingFactors[2];
	for(int i = 0; i < dynScalingFactor.size();i ++) {
	  dynamicScalingFactor *= exp(dynScalingFactor[i]);
	}
	//cout << " New Value for " <<ScalingFactors[0] << " " << ScalingFactors[1] << " " << dynamicScalingFactor << endl;
	dynamicScalingIndizes = dynScalingIndex;
	if(offset >= 0) {
		der1.size = number + dynamicScalingIndizes.size();
		der2.size = number + dynamicScalingIndizes.size();
		der1.data = new double[der1.size];
		der2.data = new double[der2.size];
		der1.index = new int[der1.size];
		der2.index = new int[der1.size];
	}

}
 /**
   * @return double
   * @param  ToNode
   */
pair<double,double> FirstOrderFeature::GetMessage (const HiddenNode * ToNode ) const{
	if(ToNode == nodes.first) {
		return m1;
	}else if(ToNode == nodes.second){
		return m2;
	}else {
		cerr << "Want message from node, that are not connected" << endl;
		exit(1);
	}
}
SparseArray FirstOrderFeature::GetDerivation (const HiddenNode * ToNode ) const{
	if(ToNode == nodes.first) {
		return der1;
	}else if(ToNode == nodes.second){
		return der2;
	}else {
		cerr << "Want message from node, that are not connected" << endl;
		exit(1);
	}

}

  /**
   */
void FirstOrderFeature::CalcMessages ( ){
	assert(number == 3);
	pair<double,double> in1 = nodes.first->GetMessage(this);
	pair<double,double> in2 = nodes.second->GetMessage(this);
//	cout << "Incoming Message: (" << in1.first << "," << in1.second << ")" << endl;
//	cout << "Incoming Message: (" << in2.first << "," << in2.second << ")" << endl;
	m1.first = in2.first *ScalingFactors[0]+ in2.second * ScalingFactors[1];
	m1.second = in2.first * ScalingFactors[1] + in2.second * dynamicScalingFactor;
	double sum1 = m1.first + m1.second;
	m1.first /= sum1;
	m1.second /= sum1;
	m2.first = in1.first  * ScalingFactors[0]+ in1.second * ScalingFactors[1];
	m2.second = in1.first * ScalingFactors[1] + in1.second * dynamicScalingFactor;
	double sum2 = m2.first + m2.second;
	m2.first /= sum2;
	m2.second /= sum2;
//	cout << "Outgoing Message: (" << m1.first << "," << m1.second << ")" << endl;
//	cout << "Outgoing Message: (" << m2.first << "," << m2.second << ")" << endl;
	assert(!isnan(m1.first));
	assert(!isnan(m1.second));
	assert(!isnan(m2.first));
	assert(!isnan(m2.second));
}
void FirstOrderFeature::CalcDerivations(){
	assert(number == 3);
	pair<double,double> in1 = nodes.first->GetMessage(this);
	pair<double,double> in2 = nodes.second->GetMessage(this);
	//cout << "Incoming Message: (" << in1.first << "," << in1.second << ")" << endl;
	//cout << "Incoming Message: (" << in2.first << "," << in2.second << ")" << endl;
	m1.first = in2.first *ScalingFactors[0]+ in2.second * ScalingFactors[1];
	m1.second = in2.first * ScalingFactors[1] + in2.second * dynamicScalingFactor;
	double sum1 = m1.first + m1.second;
	//cout << "FOF Index:" << offset << " to " << offset + 2 << endl;
	der1.index[0] = offset;
	der1.index[1] = offset + 1;
	der1.index[2] = offset + 2;
	der1.data[0] = ScalingFactors[0] *in2.first *m1.second /(sum1*sum1);
	der1.data[1] = ScalingFactors[1] * (in2.second*m1.second + in2.first*m1.first)/(sum1*sum1);
	der1.data[2] = - dynamicScalingFactor * in2.second * m1.first/(sum1 * sum1);
	for(int i = 0 ; i < dynamicScalingIndizes.size(); i++) {
		der1.index[number + i] = dynamicOffset + dynamicScalingIndizes[i];
		//cout << "FOF Index: " << der1.index[number + i] << endl;
		der1.data[number + i] = der1.data[2];
	}
	m2.first = in1.first  * ScalingFactors[0]+ in1.second * ScalingFactors[1];
	m2.second = in1.first * ScalingFactors[1] + in1.second * dynamicScalingFactor;
	double sum2 = m2.first + m2.second;
	der2.index[0] = offset;
	der2.index[1] = offset + 1;
	der2.index[2] = offset + 2;
	der2.data[0] = ScalingFactors[0] *in1.first *m2.second /(sum2*sum2);
	der2.data[1] = ScalingFactors[1] * (in1.second*m2.second + in1.first*m2.first)/(sum2*sum2);
	der2.data[2] = - dynamicScalingFactor * in1.second * m2.first/(sum2 * sum2);
	for(int i = 0 ; i < dynamicScalingIndizes.size(); i++) {
		der2.index[number + i] = dynamicOffset + dynamicScalingIndizes[i];
		der2.data[number + i] = der2.data[2];
	}
	//cout << "Outgoing Message: (" << m1.first << "," << m1.second << ")" << endl;
	//cout << "Outgoing Message: (" << m2.first << "," << m2.second << ")" << endl;
	/*cout << "Derivation1:" << endl;
	for(int i = 0; i < der1.first.size(); i++) {
		cout << "(" << der1.first[i] << "," <<der1.second[i] << ") ";
	}
	cout << endl;
	cout << "Derivation2:" << endl;
	for(int i = 0; i < der2.first.size(); i++) {
		cout << "(" << der2.first[i] << "," <<der2.second[i] << ") ";
	}
	cout << endl;*/
	assert(!isnan(m1.first));
	assert(!isnan(m1.second));
	assert(!isnan(m2.first));
	assert(!isnan(m2.second));


}


  /**
   * @return vector<double>
   */

vector<double> FirstOrderFeature::GetExpectationValue(){
  vector<double> values;
  pair<double,double> b1 = nodes.first->GetBelief();
  pair<double,double> b2 = nodes.second->GetBelief();
//	cout << "Belief1: (" << b1.first << "," << b1.second << ")" << endl;
//	cout << "Belief2: (" << b2.first << "," << b2.second << ")" << endl;

  if((nodes.first->GetGivenValue() != -1) &&(nodes.second->GetGivenValue() != -1)){
    values.push_back(b1.first * b2.first);
    values.push_back(b1.first * b2.second+b1.second*b2.first);
    values.push_back(b1.second * b2.second);
//	cout << "Values of Expection: (" << values[0] << "," << values[1] << "," << values[2] << ")" << endl;
    //another opimiation methode
    /*int max = 0;
    if(b1.first * b2.first > b1.first * b2.second+b1.second*b2.first) {
      if(b1.first * b2.first > b1.second * b2.second) {
	max = 0;
      }else {
	max = 2;
      }
    }else {
      if(b1.first * b2.second+b1.second*b2.first > b1.second * b2.second) {
	max = 1;
      }else {
	max = 2;
      }
    }
    values.resize(3,0);
    values[max] = 1;*/
  }else {
    values.resize(3,0);
  }
  return values;
}

  /**
   * @return double
   * @param  featureIndex
   */

vector<double> FirstOrderFeature::GetExampleValue(){
  vector<double> values;
  values.resize(3,0);
  if((nodes.first->GetGivenValue() != -1) &&(nodes.second->GetGivenValue() != -1)) {
    int index = nodes.first->GetGivenValue() + nodes.second->GetGivenValue();
    values[index] = 1;
  }
//	cout << "Values of ExampleValue: (" << values[0] << "," << values[1] << "," << values[2] << ")" << endl;

  return values;
}


vector<double> FirstOrderFeature::GetDynamicExpectationValue(){
  vector<double> values;
  pair<double,double> b1 = nodes.first->GetBelief();
  pair<double,double> b2 = nodes.second->GetBelief();
  if((nodes.first->GetGivenValue() != -1) &&(nodes.second->GetGivenValue() != -1)){
    values.resize(dynamicScalingIndizes.size(),b1.second * b2.second);
  }else {
    values.resize(dynamicScalingIndizes.size(),0);
  }

  return values;
}

/**
 * @return double
 */
vector<double> FirstOrderFeature::GetDynamicExampleValue(){
  vector<double> values;
  if((nodes.first->GetGivenValue() != -1) &&(nodes.second->GetGivenValue() != -1)) {
    int index = nodes.first->GetGivenValue() + nodes.second->GetGivenValue();
    if(index == 2) {
      values.resize(dynamicScalingIndizes.size(),1);
    }else{
      values.resize(dynamicScalingIndizes.size(),0);
    }
  }else{
    values.resize(dynamicScalingIndizes.size(),0);    
  }
  return values;
}
/**
 * @return vector<double>
 */
vector<int> FirstOrderFeature::GetDynamicSFIndex(){
  return dynamicScalingIndizes;
}
