#include <string.h>
#include "firstorderfeatureset.h"

FirstOrderFeatureSet::FirstOrderFeatureSet(const char* ParamFile, vector<double> & Factors)
	: FeatureSet(ParamFile,Factors)
{
	usePOSStartFeature = 0;
	usePOSEndFeature = 0;
	usePOSSourceFeature = 0;
	usePOSTargetFeature = 0;
	NumberOfDynamicFeatures = 0;
	assert(Factors.size() % NUMBER_OF_FOF == 0);
	ScalingFactors = new double[Factors.size()];
	cout << "ScalingFactor: ";
	for(int i = 0; i < Factors.size(); i++) {
		ScalingFactors[i] = exp(Factors[i]);
		cout << ScalingFactors[i] << " "; 
	}
	cout << endl;
	numberOfScalingFactors = Factors.size();
	LoadParamFile(ParamFile);
	assert(numberOfScalingFactors == directions.size() * NUMBER_OF_FOF);

  //add addtional Scaling Factors
	if(usePOSStartFeature) {
		for(int i = 0; i < POSStartScalingFactors.size();i++) {
			Factors.push_back(POSStartScalingFactors[i]);
		}
		NumberOfDynamicFeatures += POSStartScalingFactors.size();
	}
	if(usePOSEndFeature) {
		for(int i = 0; i < POSEndScalingFactors.size();i++) {
			Factors.push_back(POSEndScalingFactors[i]);
		}
		NumberOfDynamicFeatures += POSEndScalingFactors.size();
	}
	if(usePOSSourceFeature) {
		for(int i = 0; i < POSSourceScalingFactors.size();i++) {
			Factors.push_back(POSSourceScalingFactors[i]);
		}
		NumberOfDynamicFeatures += POSSourceScalingFactors.size();
	}
	if(usePOSTargetFeature) {
		for(int i = 0; i < POSTargetScalingFactors.size();i++) {
			Factors.push_back(POSTargetScalingFactors[i]);
		}
		NumberOfDynamicFeatures += POSTargetScalingFactors.size();
	}

 

	assert(Factors.size() == numberOfScalingFactors + NumberOfDynamicFeatures);
}


FirstOrderFeatureSet::~FirstOrderFeatureSet()
{
	delete[] ScalingFactors;
}


void FirstOrderFeatureSet::LoadParamFile(const char * paramFileName) {

	ifstream paramFile(paramFileName);  
	
	if (!paramFile) {
		cerr << " ParamFile for Firstorder-features " << paramFileName << " could not be opened\n";
		exit(-2);
	}

	char *pKey = new char[ 200 ];
	char *pValue = new char[ 200 ];

// TODO: Does not work with FOF.Param with empty line at end!!!!!!!!!!!!
	while ( !paramFile.eof() ) {
		paramFile >> pKey >> pValue ;
		if(strcmp(pKey,"DirectionFile") == 0) {
			LoadDirections(pValue);
		}else if(strcmp(pKey, "POSSourceFile") == 0) {
			LoadVoc(pValue,POSSourceVoc);      
		}else if(strcmp(pKey, "POSTargetFile") == 0) {
			LoadVoc(pValue,POSTargetVoc);
		}else if(strcmp(pKey, "POSStartFeatureScalingFactorFile") == 0) {
			usePOSStartFeature = 1;
			POSStartScalingFactorFile = pValue;
		}else if(strcmp(pKey, "POSEndFeatureScalingFactorFile") == 0) {
			usePOSEndFeature = 1;
			POSEndScalingFactorFile = pValue;
		}else if(strcmp(pKey, "POSSourceFeatureScalingFactorFile") == 0) {
			usePOSSourceFeature = 1;
			POSSourceScalingFactorFile = pValue;
		}else if(strcmp(pKey, "POSTargetFeatureScalingFactorFile") == 0) {
			usePOSTargetFeature = 1;
			POSTargetScalingFactorFile = pValue;
		}
	}

	delete[] pKey;
	delete[] pValue;
  //if voc is need, assert that is exists
	assert((usePOSStartFeature == 0) || (POSSourceVoc.size() > 0));
	assert((usePOSStartFeature == 0) || (POSTargetVoc.size() > 0));
	assert((usePOSEndFeature == 0) || (POSSourceVoc.size() > 0));
	assert((usePOSEndFeature == 0) || (POSTargetVoc.size() > 0));
	assert((usePOSSourceFeature == 0) || (POSSourceVoc.size() > 0));
	assert((usePOSTargetFeature == 0) || (POSTargetVoc.size() > 0));

  //assert that Directions file has been loaded
	assert(directions.size() > 0);


	if(usePOSStartFeature) {
		cout << " POS Start Feature " << endl;
		LoadScalingFactorFile(POSStartScalingFactorFile,POSStartScalingFactors,POSSourceVoc.size(),POSTargetVoc.size());
		cout << "Number of POSStartFeature :" << POSSourceVoc.size() * POSTargetVoc.size() * directions.size() << endl;
	}
	if(usePOSEndFeature) {
		cout << " POS End Feature " << endl;
		LoadScalingFactorFile(POSEndScalingFactorFile,POSEndScalingFactors,POSSourceVoc.size(),POSTargetVoc.size());
		cout << "Number of POSEndFeature :" << POSSourceVoc.size() * POSTargetVoc.size() * directions.size() << endl;
	}
	if(usePOSSourceFeature) {
		cout << " POS Source Feature " << endl;
		LoadScalingFactorFile(POSSourceScalingFactorFile,POSSourceScalingFactors,POSSourceVoc.size(),POSSourceVoc.size());
		cout << "Number of POSSourceFeature :" << POSSourceVoc.size() * POSSourceVoc.size() * directions.size() << endl;
	}
	if(usePOSTargetFeature) {
		cout << " POS Target Feature " << endl;
		LoadScalingFactorFile(POSTargetScalingFactorFile,POSTargetScalingFactors,POSTargetVoc.size(),POSTargetVoc.size());
		cout << "Number of POSTargetFeature :" << POSTargetVoc.size() * POSTargetVoc.size() * directions.size() << endl;
	}

}

  /**
 * @param  paramFilename
   */
void FirstOrderFeatureSet::LoadDirections(const char * paramFileName) {
	ifstream paramFile(paramFileName);  
	
	if (!paramFile) {
		cerr << " ParamFile for Firstorder-features " << paramFileName << " could not be opened\n";
		exit(-2);
	}
	char *pKey = new char[ 200 ];
	char *pDir1 = new char[ 200 ];
	char *pDir2 = new char[ 200 ];
	int Dir1;
	int Dir2;
// TODO: Does not work with FOF.Param with empty line at end!!!!!!!!!!!!
	while ( !paramFile.eof() ) {
		paramFile >> pKey >> pDir1 >> pDir2;
		if(strcmp(pKey,"Direction") == 0) {
			Dir1 = atoi(pDir1);
			Dir2 = atoi(pDir2);
			cout << "Establish First Order Feature for Direction: " << Dir1 << "-" << Dir2 << endl;
			directions.push_back(make_pair(Dir1,Dir2));
		}
	}
	delete[] pKey;
	delete[] pDir1;
	delete[] pDir2;
}

  /**
 * @param  G
   */
void FirstOrderFeatureSet::GenerateFeatures (Graph * G, int offset, int numberOfAllFeatures){
	SourceLength = G->GetSourceLength();
	TargetLength = G->GetTargetLength();
	
	assert(features.size() == 0);
	assert(numberOfElements.size() == 0);

	for(int k = 0; k < directions.size();k++){
		numberOfElements.push_back(max((SourceLength-abs(directions[k].first))*(TargetLength-abs(directions[k].second)),0));
		FirstOrderFeature* currentFeatures = new FirstOrderFeature[numberOfElements[k]];
	  //cout << "Direction: " << directions[k].first << " " << directions[k].second << endl;
		for(int i = 0; i < SourceLength - abs(directions[k].first); i++) {
			for(int j = 0; j < TargetLength - abs(directions[k].second); j++) {
	      //cout << "i: " << i << "j: " << j << endl;
				vector<double> dynScalingFactor;
				vector<int> dynScalingIndex;
				int dynOffset = numberOfScalingFactors;
				int i1,i2;
				int j1,j2;
				if(directions[k].first > 0) {
					i1 = i;
					i2 = i + directions[k].first;
				}else {
					i1 = i - directions[k].first;
					i2 = i;
				}
				if(directions[k].second > 0) {
					j1 = j;
					j2 = j + directions[k].second;
				}else {
					j1 = j - directions[k].second;
					j2 = j;
				}
				if(usePOSStartFeature) {
					string SourcePOS = G->GetSourcePOS(i1);
					string TargetPOS = G->GetTargetPOS(j1);
					if(POSSourceVoc.count(SourcePOS) == 0) {
						cout << "Unknown source Tag in FOF: " << SourcePOS << endl;
						SourcePOS = "NN";
			  		}
			  		if(POSTargetVoc.count(TargetPOS) == 0) {
						cout << "Unknown target Tag in FOF: " << TargetPOS << endl;
						TargetPOS = "NN";
			  		}
					assert(POSSourceVoc.count(SourcePOS) != 0);
					assert(POSTargetVoc.count(TargetPOS) != 0);
					int SourcePOSIndex = POSSourceVoc[SourcePOS];
					int TargetPOSIndex = POSTargetVoc[TargetPOS];
					int index = k * (POSTargetVoc.size() * POSSourceVoc.size() )+ SourcePOSIndex * POSTargetVoc.size() + TargetPOSIndex;
					assert(index < POSTargetVoc.size() * POSSourceVoc.size() * directions.size());
		//cout << "Pos: ( "<< i1 << "," <<j1<< ") : " << endl;
		//cout << "Scaling Factor for " << SourcePOS << " and " << TargetPOS << " : " << POSStartScalingFactors[index] << endl;
		//cout << "Index : " << SourcePOSIndex << " , " << TargetPOSIndex << endl;
					dynScalingFactor.push_back(POSStartScalingFactors[index]);
					dynScalingIndex.push_back(dynOffset + index);
					dynOffset += POSTargetVoc.size() * POSSourceVoc.size()*directions.size();
		
				}
				if(usePOSEndFeature) {
					const string SourcePOS = G->GetSourcePOS(i2);
					const string TargetPOS = G->GetTargetPOS(j2);
					assert(POSSourceVoc.count(SourcePOS) != 0);
					assert(POSTargetVoc.count(TargetPOS) != 0);
					int SourcePOSIndex = POSSourceVoc[SourcePOS];
					int TargetPOSIndex = POSTargetVoc[TargetPOS];
					int index = k * (POSTargetVoc.size() * POSSourceVoc.size() )+ SourcePOSIndex * POSTargetVoc.size() + TargetPOSIndex;
					assert(index < POSTargetVoc.size() * POSSourceVoc.size() * directions.size());
		//cout << "Pos: ( "<< i2 << "," <<j2<< ") : " << endl;
		//cout << "Scaling Factor for " << SourcePOS << " and " << TargetPOS << " : " << POSEndScalingFactors[index] << endl;
		//cout << "Index : " << SourcePOSIndex << " , " << TargetPOSIndex << endl;
					dynScalingFactor.push_back(POSEndScalingFactors[index]);
					dynScalingIndex.push_back(dynOffset + index);
					dynOffset += POSTargetVoc.size() * POSSourceVoc.size()*directions.size();
		
				}
				if(usePOSSourceFeature) {
					const string Source1POS = G->GetSourcePOS(i1);
					const string Source2POS = G->GetSourcePOS(i2);
					assert(POSSourceVoc.count(Source1POS) != 0);
					assert(POSSourceVoc.count(Source2POS) != 0);
					int Source1POSIndex = POSSourceVoc[Source1POS];
					int Source2POSIndex = POSSourceVoc[Source2POS];
					int index = k * (POSSourceVoc.size() * POSSourceVoc.size() )+ Source1POSIndex * POSSourceVoc.size() + Source2POSIndex;
					assert(index < POSSourceVoc.size() * POSSourceVoc.size() * directions.size());
		//cout << "Source Pos: ( "<< i1 << "," <<i2<< ") : " << endl;
		//cout << "Scaling Factor for " << Source1POS << " and " << Source2POS << " : " << POSSourceScalingFactors[index] << endl;
		//cout << "Index : " << Source1POSIndex << " , " << Source2POSIndex << endl;
					dynScalingFactor.push_back(POSSourceScalingFactors[index]);
					dynScalingIndex.push_back(dynOffset + index);
					dynOffset += POSSourceVoc.size() * POSSourceVoc.size()*directions.size();
		
				}
				if(usePOSTargetFeature) {
					const string Target1POS = G->GetTargetPOS(j1);
					const string Target2POS = G->GetTargetPOS(j2);
					assert(POSTargetVoc.count(Target1POS) != 0);
					assert(POSTargetVoc.count(Target2POS) != 0);
					int Target1POSIndex = POSTargetVoc[Target1POS];
					int Target2POSIndex = POSTargetVoc[Target2POS];
					int index = k * (POSTargetVoc.size() * POSTargetVoc.size() )+ Target1POSIndex * POSTargetVoc.size() + Target2POSIndex;
					assert(index < POSTargetVoc.size() * POSTargetVoc.size() * directions.size());
		//cout << "Target Pos: ( "<< j1 << "," <<j2<< ") : " << endl;
		//cout << "Scaling Factor for " << Target1POS << " and " << Target2POS << " : " << POSTargetScalingFactors[index] << endl;
		//cout << "Index : " << Target1POSIndex << " , " << Target2POSIndex << endl;
					dynScalingFactor.push_back(POSTargetScalingFactors[index]);
					dynScalingIndex.push_back(dynOffset + index);
					dynOffset += POSTargetVoc.size() * POSTargetVoc.size()*directions.size();
		
				}
	      //int index = i*(TargetLength - directions[k].second) + j;

	      //currentFeatures[index].Init(make_pair(G->GetNode(i,j),G->GetNode(i+directions[k].first,j+directions[k].second)),&ScalingFactors[k*NUMBER_OF_FOF],NUMBER_OF_FOF,dynScalingFactor,dynScalingIndex);
				int index = i*(TargetLength - abs(directions[k].second)) + j;
				assert(index < numberOfElements[k]);
	      //cout << "i: " << i << "j: " << j << "k : " << k << endl;
	      //cout << "i1: " << i1 << " i2: " << i2 << "j1: " << j1 << " j2: " << j2 << endl;
	      //cout << " TargetLength: " <<TargetLength << " directions[k].second: " << directions[k].second << endl;
	      //cout << "Index:" << index << " numberOfElements[k]; " << numberOfElements[k] << endl;
				currentFeatures[index].Init(make_pair(G->GetNode(i1,j1),G->GetNode(i2,j2)),&ScalingFactors[k*NUMBER_OF_FOF],NUMBER_OF_FOF,dynScalingFactor,dynScalingIndex,offset + k*NUMBER_OF_FOF, offset, numberOfAllFeatures);
			}
		}
		features.push_back(currentFeatures);
	}
	assert(features.size() == directions.size());
	
}

void FirstOrderFeatureSet::RemoveFeatures() {
	for(int i = 0; i < directions.size();i++) {
		delete[] features[i];
	}
	features.clear();
	numberOfElements.clear();
}

  /**
 * @return FirstOrderFeatureSetIterator
   */
FirstOrderFeatureSetIterator * FirstOrderFeatureSet::GetIterator ( ) const {
	return new FirstOrderFeatureSetIterator(numberOfElements, features);
}

void FirstOrderFeatureSet::EstimateScalingFactores() {
	EstimatedDerivation.clear();
	vector<double> ExpectValue;
	vector<double> ExampleValue;
	ExpectValue.resize(numberOfScalingFactors + NumberOfDynamicFeatures,0);
	ExampleValue.resize(numberOfScalingFactors + NumberOfDynamicFeatures,0);
	FeatureSetIterator * iter = GetIterator();
	for(int k = 0; k < directions.size(); k++) {
		for(int i = 0; i < numberOfElements[k];i++) {
			vector<double> TempExpectValue = features[k][i].GetExpectationValue();
			vector<double> TempExampleValue = features[k][i].GetExampleValue();
			assert(TempExampleValue.size() == NUMBER_OF_FOF);
			assert(TempExpectValue.size() == NUMBER_OF_FOF);
			for(int j = 0; j < NUMBER_OF_FOF; j++) {
				assert(k*NUMBER_OF_FOF + j < numberOfScalingFactors);
				ExpectValue[k*NUMBER_OF_FOF + j] += TempExpectValue[j];
				ExampleValue[k*NUMBER_OF_FOF + j] += TempExampleValue[j];
			}
			vector<double> TempDynExpectValue = features[k][i].GetDynamicExpectationValue();
			vector<double> TempDynExampleValue = features[k][i].GetDynamicExampleValue();
			vector<int> TempDynIndex = features[k][i].GetDynamicSFIndex();
			assert(TempDynExpectValue.size() == TempDynExampleValue.size());
			assert(TempDynExampleValue.size() == TempDynIndex.size());
			for(int j = 0; j < TempDynExpectValue.size() ; j++) {
				assert(TempDynIndex[j] <  ExpectValue.size());
				ExpectValue[TempDynIndex[j]] += TempDynExpectValue[j];
				ExampleValue[TempDynIndex[j]] += TempDynExampleValue[j];
			}
      
		}
	//cout << "Direction " << k << ": Expected Value(" << ExpectValue[k*NUMBER_OF_FOF] << "," << ExpectValue[k*NUMBER_OF_FOF + 1] << "," << ExpectValue[k*NUMBER_OF_FOF + 2] << ")" << endl;
	//cout << "Direction " << k << ": Example Value(" << ExampleValue[k*NUMBER_OF_FOF] << "," << ExampleValue[k*NUMBER_OF_FOF + 1] << "," << ExampleValue[k*NUMBER_OF_FOF + 2] << ")" << endl;
	}

	for(int j = 0; j < numberOfScalingFactors + NumberOfDynamicFeatures; j++) {
		double deltaL = ExampleValue[j] - ExpectValue[j];
    //cout << "Calculating Der[" << j << "]: " << ExampleValue[j] << " - " <<  ExpectValue[j] << " = " << deltaL << endl;
		EstimatedDerivation.push_back(deltaL);
	}
}

void FirstOrderFeatureSet::SetScalingFactors(vector<double> & newScalingFactors) {
	assert(newScalingFactors.size() == numberOfScalingFactors + NumberOfDynamicFeatures);
	int j = 0;
	for(int i = 0; i < numberOfScalingFactors; i++) {
		ScalingFactors[i] = exp(newScalingFactors[j]);
		j++;
	}
	if(usePOSStartFeature) {
		for(int i = 0; i < POSStartScalingFactors.size();i++) { 
			POSStartScalingFactors[i] = newScalingFactors[j];
			j++;
		}
	}
	if(usePOSEndFeature) {
		for(int i = 0; i < POSEndScalingFactors.size();i++) { 
			POSEndScalingFactors[i] = newScalingFactors[j];
			j++;
		}
	}
	if(usePOSSourceFeature) {
		for(int i = 0; i < POSSourceScalingFactors.size();i++) { 
			POSSourceScalingFactors[i] = newScalingFactors[j];
			j++;
		}
	}
	if(usePOSTargetFeature) {
		for(int i = 0; i < POSTargetScalingFactors.size();i++) { 
			POSTargetScalingFactors[i] = newScalingFactors[j];
			j++;
		}
	}

}

void FirstOrderFeatureSet::LoadScalingFactorFile(string fileName,vector<double> & ScalingFactors,int sourceLength, int targetLength){

	assert(sourceLength > 0);
	assert(targetLength > 0);
	assert(directions.size() > 0);
	ScalingFactors.clear();
	ScalingFactors.resize(sourceLength * targetLength*directions.size(),0);
	ifstream file( fileName.c_str());

	if (!file) {
		cerr << "File " << fileName << " could not be opened\n";
		exit(-2);
	}

	char *pDirection = new char [ 200 ];
	char *pSourceIndex = new char[ 200 ];
	char *pTargetIndex = new char[ 200 ];
	char *pValue = new char[ 200 ];
	while ( !file.eof() ) {
		file >> pDirection >> pSourceIndex >> pTargetIndex >> pValue;
		int direction = atoi(pDirection);
		int sourceIndex = atoi(pSourceIndex);
		int targetIndex = atoi(pTargetIndex);
		assert(sourceIndex < sourceLength);
		assert(targetIndex < targetLength);
		assert(direction < directions.size());
		assert(direction *(targetLength + sourceLength) + sourceIndex * targetLength + targetIndex < sourceLength * targetLength*directions.size());
		ScalingFactors[direction *targetLength * sourceLength + sourceIndex * targetLength + targetIndex] = atof(pValue);
	}
	delete[] pDirection;
	delete[] pSourceIndex;
	delete[] pTargetIndex;
	delete[] pValue;


}

void FirstOrderFeatureSet::LoadVoc(char * fileName,map<string,int> & Voc){
	ifstream file( fileName);

	if (!file) {
		cerr << "File " << fileName << " could not be opened\n";
		exit(-2);
	}

	int i = 0;
	char *pWord = new char[ 200 ];
	while ( !file.eof() ) {
		file >> pWord;
		Voc.insert(make_pair(string(pWord),i));
		i++;
	}
	delete[] pWord;
}

void FirstOrderFeatureSet::OutputScalingFactors() {
	cout << "FirstOrderScalingFactors ";
	for(int i = 0; i < numberOfScalingFactors; i++) {
		cout << log(ScalingFactors[i]) << "_";
	}
	cout << endl;
	if(usePOSStartFeature) {
		string fileName = POSStartScalingFactorFile + ".opt";
		ofstream out(fileName.c_str());
		for(int k = 0; k < directions.size(); k++) {
			for(int i = 0 ; i < POSSourceVoc.size(); i++ ) {
				for(int j = 0; j < POSTargetVoc.size(); j++) {
					int index = k * POSSourceVoc.size() * POSTargetVoc.size() + i*POSTargetVoc.size()+j;
					assert(index < POSSourceVoc.size() * POSTargetVoc.size() * directions.size());
					assert(index >= 0);
					out << k << " " << i << " " << j << " " << POSStartScalingFactors[index] << endl;
				}
			}
		}
		out.close();
	}
	if(usePOSEndFeature) {
		string fileName = POSEndScalingFactorFile + ".opt";
		ofstream out(fileName.c_str());
		for(int k = 0; k < directions.size(); k++) {
			for(int i = 0 ; i < POSSourceVoc.size(); i++ ) {
				for(int j = 0; j < POSTargetVoc.size(); j++) {
					int index = k * POSSourceVoc.size() * POSTargetVoc.size() + i*POSTargetVoc.size()+j;
					assert(index < POSSourceVoc.size() * POSTargetVoc.size() * directions.size());
					assert(index >= 0);
					out << k << " " << i << " " << j << " " << POSEndScalingFactors[index] << endl;
				}
			}
		}
		out.close();
	}
	if(usePOSSourceFeature) {
		string fileName = POSSourceScalingFactorFile + ".opt";
		ofstream out(fileName.c_str());
		for(int k = 0; k < directions.size(); k++) {
			for(int i = 0 ; i < POSSourceVoc.size(); i++ ) {
				for(int j = 0; j < POSSourceVoc.size(); j++) {
					int index = k * POSSourceVoc.size() * POSSourceVoc.size() + i*POSSourceVoc.size()+j;
					assert(index < POSSourceVoc.size() * POSSourceVoc.size() * directions.size());
					assert(index >= 0);
					out << k << " " << i << " " << j << " " << POSSourceScalingFactors[index] << endl;
				}
			}
		}
		out.close();
	}
	if(usePOSTargetFeature) {
		string fileName = POSTargetScalingFactorFile + ".opt";
		ofstream out(fileName.c_str());
		for(int k = 0; k < directions.size(); k++) {
			for(int i = 0 ; i < POSTargetVoc.size(); i++ ) {
				for(int j = 0; j < POSTargetVoc.size(); j++) {
					int index = k * POSTargetVoc.size() * POSTargetVoc.size() + i*POSTargetVoc.size()+j;
					assert(index < POSTargetVoc.size() * POSTargetVoc.size() * directions.size());
					assert(index >= 0);
					out << k << " " << i << " " << j << " " << POSTargetScalingFactors[index] << endl;
				}
			}
		}
		out.close();
	}

}
