#ifndef FIRSTORDERFEATURESET_H
#define FIRSTORDERFEATURESET_H

#include <featureset.h>
#include <hiddennode.h>
#include <firstorderfeature.h>
#include <firstorderfeaturesetiterator.h>
#include <cassert>
#include <math.h>
#include <map>
/**
	@author Jan Niehues <jan@niehues.info>
*/

#define NUMBER_OF_FOF 3

class FirstOrderFeatureSet : public FeatureSet
{
private:
  
	double * ScalingFactors;
	vector<double> EstimatedDerivation;
	int numberOfScalingFactors;
	int NumberOfDynamicFeatures;
	vector<pair<int,int> > directions;
	int SourceLength;
	int TargetLength;
	vector<FirstOrderFeature*> features;
	vector<int> numberOfElements;
	map<string, int> POSSourceVoc;
	map<string, int> POSTargetVoc;
	vector<double> POSStartScalingFactors;
	string POSStartScalingFactorFile;
	int usePOSStartFeature;
	vector<double> POSEndScalingFactors;
	string POSEndScalingFactorFile;
	int usePOSEndFeature;
	vector<double> POSSourceScalingFactors;
	string POSSourceScalingFactorFile;
	int usePOSSourceFeature;
	vector<double> POSTargetScalingFactors;
	string POSTargetScalingFactorFile;
	int usePOSTargetFeature;

  /**
   * @param  ParamFile
   */
    void LoadDirections(const char * ParamFile);
  /**
   * @param  ParamFile
   */
    void LoadParamFile(const char * ParamFile);
    void LoadScalingFactorFile(string fileName,vector<double> & ScalingFactors,int sourceLength, int targetLength);
    void LoadVoc(char * fileName,map<string,int> & Voc);

public:
    FirstOrderFeatureSet(const char* ParamFile, vector<double> & ScalingFactors);

    ~FirstOrderFeatureSet();


  /**
   * @param  G
   */
  void GenerateFeatures (Graph * G , int offset, int numberOfAllFeatures);

  void SetScalingFactors(vector<double> & newScalingFactors);

  void RemoveFeatures();

  /**
   * @return FirstOrderSetIterator
   */
  FirstOrderFeatureSetIterator * GetIterator ( ) const;
  /**
   */
  void EstimateScalingFactores ();

  vector<double> inline GetEstimatedDerivations() const {return EstimatedDerivation;};
  void OutputScalingFactors();
};

#endif
