
#include "firstorderlopfeature.h"

FirstOrderLOPFeature::FirstOrderLOPFeature()
 : Feature()
{
	m1 = make_pair(0.5,0.5);
	m2 = make_pair(0.5,0.5);
	der1.size = 0;
	der2.size = 0;
}


FirstOrderLOPFeature::~FirstOrderLOPFeature()
{
	if(der1.size > 0) {
		delete[] der1.data;
		delete[] der1.index;
		delete[] der2.data;
		delete[] der2.index;
	}
	delete[] potentialValues;
	delete[] scaledFeatureValues;
}

void FirstOrderLOPFeature::Init(pair<HiddenNode*,HiddenNode*> n, double* SFactors, int num,vector<double *> dynScalingFactor, vector<int> dynScalingIndex,int offS, int dynOffSet,int nOfAllF, int numberOfExp, double * expWeight){
	numberOfExperts = numberOfExp;
	expertWeights = expWeight;
	nodes = n;
	number = num;
	offset = offS;
	dynamicOffset = dynOffSet;
	numberOfAllFeatures = nOfAllF;
	ScalingFactors = SFactors;
	nodes.first->AddConnection(this);
	nodes.second->AddConnection(this);

	dynamicScalingIndizes = dynScalingIndex;

	scaledFeatureValues = new double[NUMBER_OF_FOF * numberOfExperts];
	for(int i = 0; i < NUMBER_OF_FOF; i++) {
		for(int j = 0; j < numberOfExperts; j++) {
			assert(i*numberOfExperts + j < NUMBER_OF_FOF * numberOfExperts);
			scaledFeatureValues[i*numberOfExperts + j] = ScalingFactors[i*numberOfExperts + j];
		}
	}
	for(int i = 0; i < dynScalingFactor.size(); i++) {
		for(int j = 0; j < numberOfExperts; j++) {
			assert(2*numberOfExperts + j<NUMBER_OF_FOF * numberOfExperts);
			scaledFeatureValues[2*numberOfExperts + j] += (dynScalingFactor[i])[j];
		}
	}
	potentialValues = new double[NUMBER_OF_FOF];
	for(int i = 0; i < NUMBER_OF_FOF; i++) {
		potentialValues[i] = 0;
		for(int j = 0; j < numberOfExperts; j++) {
			assert(i < NUMBER_OF_FOF);
			assert(j < numberOfExperts);
			assert(i*numberOfExperts + j < NUMBER_OF_FOF * numberOfExperts);
			potentialValues[i] += expertWeights[j] * scaledFeatureValues[i*numberOfExperts + j];
		}
		potentialValues[i] = exp(potentialValues[i]);
	}

}
 /**
   * @return double
   * @param  ToNode
   */
pair<double,double> FirstOrderLOPFeature::GetMessage (const HiddenNode * ToNode ) const{
	if(ToNode == nodes.first) {
		return m1;
	}else if(ToNode == nodes.second){
		return m2;
	}else {
		cerr << "Want message from node, that are not connected" << endl;
		exit(1);
	}
}

SparseArray FirstOrderLOPFeature::GetDerivation (const HiddenNode * ToNode ) const{
	if(ToNode == nodes.first) {
		return der1;
	}else if(ToNode == nodes.second){
		return der2;
	}else {
		cerr << "Want message from node, that are not connected" << endl;
		exit(1);
	}

}

  /**
   */
void FirstOrderLOPFeature::CalcMessages ( ){
	assert(number == 3);
	pair<double,double> in1 = nodes.first->GetMessage(this);
	pair<double,double> in2 = nodes.second->GetMessage(this);
//	cout << "Incoming Message: (" << in1.first << "," << in1.second << ")" << endl;
//	cout << "Incoming Message: (" << in2.first << "," << in2.second << ")" << endl;
	m1.first = in2.first *potentialValues[0]+ in2.second * potentialValues[1];
	m1.second = in2.first * potentialValues[1] + in2.second * potentialValues[2];
	double sum1 = m1.first + m1.second;
	m1.first /= sum1;
	m1.second /= sum1;
	m2.first = in1.first  * potentialValues[0]+ in1.second * potentialValues[1];
	m2.second = in1.first * potentialValues[1] + in1.second * potentialValues[2];
	double sum2 = m2.first + m2.second;
	m2.first /= sum2;
	m2.second /= sum2;
	assert(!isnan(m1.first));
	assert(!isnan(m1.second));
	assert(!isnan(m2.first));
	assert(!isnan(m2.second));
}

void FirstOrderLOPFeature::CalcDerivations(){
	if(der1.size == 0) {
		der1.size = numberOfExperts;
		der2.size = numberOfExperts;
		der1.data = new double[der1.size];
		der2.data = new double[der2.size];
		der1.index = new int[der1.size];
		der2.index = new int[der1.size];
	}

	assert(number == 3);
	pair<double,double> in1 = nodes.first->GetMessage(this);
	pair<double,double> in2 = nodes.second->GetMessage(this);
	//cout << "Incoming Message: (" << in1.first << "," << in1.second << ")" << endl;
	//cout << "Incoming Message: (" << in2.first << "," << in2.second << ")" << endl;
	m1.first = in2.first *potentialValues[0]+ in2.second * potentialValues[1];
	m1.second = in2.first * potentialValues[1] + in2.second * potentialValues[2];
	double sum1 = m1.first + m1.second;
	for(int i = 0; i < numberOfExperts; i++) {
		der1.index[i] = i;
		der1.data[i] = potentialValues[0]*scaledFeatureValues[i] *in2.first *m1.second;
		der1.data[i] += potentialValues[1]*scaledFeatureValues[numberOfExperts + i] * (in2.second*m1.second - in2.first*m1.first);
		der1.data[i] += - potentialValues[2]*scaledFeatureValues[2*numberOfExperts + i] * in2.second * m1.first;
		der1.data[i] /= (sum1 * sum1);
	}
	m2.first = in1.first  * potentialValues[0]+ in1.second * potentialValues[1];
	m2.second = in1.first * potentialValues[1] + in1.second * potentialValues[2];
	double sum2 = m2.first + m2.second;
	for(int i = 0; i < numberOfExperts; i++) {
		der2.index[i] = i;
		der2.data[i] = potentialValues[0]*scaledFeatureValues[i] *in1.first *m2.second;
		der2.data[i] += potentialValues[1]*scaledFeatureValues[numberOfExperts + i] * (in1.second*m2.second - in1.first*m2.first);
		der2.data[i] += - potentialValues[2]*scaledFeatureValues[2*numberOfExperts + i] * in1.second * m2.first;
		der2.data[i] /= (sum2 * sum2);
	}
	assert(!isnan(m1.first));
	assert(!isnan(m1.second));
	assert(!isnan(m2.first));
	assert(!isnan(m2.second));


}


  /**
   * @return vector<double>
   */

vector<double> FirstOrderLOPFeature::GetExpectationValue(){
  vector<double> values;
  pair<double,double> b1 = nodes.first->GetBelief();
  pair<double,double> b2 = nodes.second->GetBelief();

//	cout << "FOFFeature::GetExpectationValue()" << endl;
//	cout << "Belief: (" << b1.first << "," << b1.second << "),(" << b2.first << "," << b2.second << ")" << endl; 

  if((nodes.first->GetGivenValue() != -1) &&(nodes.second->GetGivenValue() != -1)){
	for(int j = 0; j < numberOfExperts; j++) {
//		cout << "ScaledFeaturevalues for " << j << ": ";
//		cout << scaledFeatureValues[j] << " " << scaledFeatureValues[numberOfExperts + j] << " " << scaledFeatureValues[2*numberOfExperts+j] << endl;
		values.push_back(b1.first * b2.first*scaledFeatureValues[j]);
		values[j] += (b1.first * b2.second+b1.second*b2.first)*scaledFeatureValues[numberOfExperts + j];
		values[j] += b1.second * b2.second * scaledFeatureValues[2*numberOfExperts+j];
	}
  }else {
    values.resize(numberOfExperts,0);
  }
/*	cout << "Result: ";
	for(int i = 0; i < numberOfExperts; i++) {
		cout << values[i] << "_"; 
	}
	cout << endl;
*/ 
 return values;
}

  /**
   * @return double
   * @param  featureIndex
   */

vector<double> FirstOrderLOPFeature::GetExampleValue(){
  vector<double> values;
//	cout << "FOF:GetExampleValue()" << endl;
//	cout << "GivenValue: " << nodes.first->GetGivenValue() << nodes.second->GetGivenValue() << endl;
//	cout << "Result: "; 
  if((nodes.first->GetGivenValue() != -1) &&(nodes.second->GetGivenValue() != -1)) {
    int index = nodes.first->GetGivenValue() + nodes.second->GetGivenValue();
	for(int i = 0; i < numberOfExperts; i++) {
	    values.push_back(scaledFeatureValues[index*numberOfExperts + i]);
	}
  }else {
  values.resize(numberOfExperts,0);

  }
  return values;
}


