#include <string.h>
#include "firstorderlopfeatureset.h"

FirstOrderLOPFeatureSet::FirstOrderLOPFeatureSet(const char* ParamFile, const char * ScalingFactorString,int numberOfExp, double * expWeights)
 : FeatureSet()
{
	numberOfExperts = numberOfExp;
	expertWeights = expWeights;
	usePOSStartFeature = 0;
	usePOSEndFeature = 0;
	usePOSSourceFeature = 0;
	usePOSTargetFeature = 0;
	NumberOfDynamicFeatures = 0;

	LoadParamFile(ParamFile);
	numberOfScalingFactors = directions.size() * NUMBER_OF_FOF;

	ScalingFactors = new double[numberOfScalingFactors*numberOfExperts];

	for(int i =0; i < numberOfScalingFactors*numberOfExperts; i++) {
		ScalingFactors[i] =0;
	}
	Utils::Extract2DScalingFactors(ScalingFactorString,ScalingFactors,numberOfScalingFactors,numberOfExperts);


}


FirstOrderLOPFeatureSet::~FirstOrderLOPFeatureSet()
{
	delete[] ScalingFactors;
}


void FirstOrderLOPFeatureSet::LoadParamFile(const char * paramFileName) {

	ifstream paramFile(paramFileName);  
	
	if (!paramFile) {
		cerr << " ParamFile for Firstorder-features " << paramFileName << " could not be opened\n";
		exit(-2);
	}

	char *pKey = new char[ 200 ];
	char *pValue = new char[ 200 ];

// TODO: Does not work with FOF.Param with empty line at end!!!!!!!!!!!!
	while ( !paramFile.eof() ) {
		paramFile >> pKey >> pValue ;
		if(strcmp(pKey,"DirectionFile") == 0) {
			LoadDirections(pValue);
		}else if(strcmp(pKey, "POSSourceFile") == 0) {
			Utils::LoadVoc(pValue,POSSourceVoc);      
		}else if(strcmp(pKey, "POSTargetFile") == 0) {
			Utils::LoadVoc(pValue,POSTargetVoc);
		}else if(strcmp(pKey, "POSStartFeatureScalingFactorFile") == 0) {
			usePOSStartFeature = 1;
			POSStartScalingFactorFile.push_back(string(pValue));
		}else if(strcmp(pKey, "POSEndFeatureScalingFactorFile") == 0) {
			usePOSEndFeature = 1;
			POSEndScalingFactorFile.push_back(string(pValue));
		}else if(strcmp(pKey, "POSSourceFeatureScalingFactorFile") == 0) {
			usePOSSourceFeature = 1;
			POSSourceScalingFactorFile.push_back(string(pValue));
		}else if(strcmp(pKey, "POSTargetFeatureScalingFactorFile") == 0) {
			usePOSTargetFeature = 1;
			POSTargetScalingFactorFile.push_back(string(pValue));
		}
	}

	delete[] pKey;
	delete[] pValue;
  //if voc is need, assert that is exists
	assert((usePOSStartFeature == 0) || (POSSourceVoc.size() > 0));
	assert((usePOSStartFeature == 0) || (POSTargetVoc.size() > 0));
	assert((usePOSEndFeature == 0) || (POSSourceVoc.size() > 0));
	assert((usePOSEndFeature == 0) || (POSTargetVoc.size() > 0));
	assert((usePOSSourceFeature == 0) || (POSSourceVoc.size() > 0));
	assert((usePOSTargetFeature == 0) || (POSTargetVoc.size() > 0));

	assert((usePOSStartFeature == 0) || (POSStartScalingFactorFile.size() == numberOfExperts));
	assert((usePOSEndFeature == 0) || (POSEndScalingFactorFile.size() == numberOfExperts));
	assert((usePOSSourceFeature == 0) || (POSSourceScalingFactorFile.size() == numberOfExperts));
	assert((usePOSTargetFeature == 0) || (POSTargetScalingFactorFile.size() == numberOfExperts));

  //assert that Directions file has been loaded
	assert(directions.size() > 0);


	if(usePOSStartFeature) {
		cout << " POS Start Feature " << endl;
		POSStartScalingFactors = new double[POSSourceVoc.size() * POSTargetVoc.size() * directions.size()* numberOfExperts];
		Utils::LoadFOFScalingFactorFiles(POSStartScalingFactorFile,POSStartScalingFactors,POSSourceVoc.size(),POSTargetVoc.size(),directions.size());
		cout << "Number of POSStartFeature :" << POSSourceVoc.size() * POSTargetVoc.size() * directions.size() << endl;
	}
	if(usePOSEndFeature) {
		cout << " POS End Feature " << endl;
		POSStartScalingFactors = new double[POSSourceVoc.size() * POSTargetVoc.size() * directions.size()* numberOfExperts];
		Utils::LoadFOFScalingFactorFiles(POSEndScalingFactorFile,POSEndScalingFactors,POSSourceVoc.size(),POSTargetVoc.size(),directions.size());
		cout << "Number of POSEndFeature :" << POSSourceVoc.size() * POSTargetVoc.size() * directions.size() << endl;
	}
	if(usePOSSourceFeature) {
		cout << " POS Source Feature " << endl;
		POSStartScalingFactors = new double[POSSourceVoc.size() * POSSourceVoc.size() * directions.size()* numberOfExperts];
		Utils::LoadFOFScalingFactorFiles(POSSourceScalingFactorFile,POSSourceScalingFactors,POSSourceVoc.size(),POSSourceVoc.size(),directions.size());
		cout << "Number of POSSourceFeature :" << POSSourceVoc.size() * POSSourceVoc.size() * directions.size() << endl;
	}
	if(usePOSTargetFeature) {
		cout << " POS Target Feature " << endl;
		POSStartScalingFactors = new double[POSTargetVoc.size() * POSTargetVoc.size() * directions.size()* numberOfExperts];
		Utils::LoadFOFScalingFactorFiles(POSTargetScalingFactorFile,POSTargetScalingFactors,POSTargetVoc.size(),POSTargetVoc.size(),directions.size());
		cout << "Number of POSTargetFeature :" << POSTargetVoc.size() * POSTargetVoc.size() * directions.size() << endl;
	}

}

  /**
 * @param  paramFilename
   */
void FirstOrderLOPFeatureSet::LoadDirections(const char * paramFileName) {
	ifstream paramFile(paramFileName);  
	
	if (!paramFile) {
		cerr << " ParamFile for Firstorder-features " << paramFileName << " could not be opened\n";
		exit(-2);
	}
	char *pKey = new char[ 200 ];
	char *pDir1 = new char[ 200 ];
	char *pDir2 = new char[ 200 ];
	int Dir1;
	int Dir2;
// TODO: Does not work with FOF.Param with empty line at end!!!!!!!!!!!!
	while ( !paramFile.eof() ) {
		paramFile >> pKey >> pDir1 >> pDir2;
		if(strcmp(pKey,"Direction") == 0) {
			Dir1 = atoi(pDir1);
			Dir2 = atoi(pDir2);
			cout << "Establish First Order Feature for Direction: " << Dir1 << "-" << Dir2 << endl;
			directions.push_back(make_pair(Dir1,Dir2));
		}
	}
	delete[] pKey;
	delete[] pDir1;
	delete[] pDir2;
}

  /**
 * @param  G
   */
void FirstOrderLOPFeatureSet::GenerateFeatures (Graph * G, int offset, int numberOfAllFeatures){
	SourceLength = G->GetSourceLength();
	TargetLength = G->GetTargetLength();
	
	assert(features.size() == 0);
	assert(numberOfElements.size() == 0);

	for(int k = 0; k < directions.size();k++){
		numberOfElements.push_back(max((SourceLength-abs(directions[k].first))*(TargetLength-abs(directions[k].second)),0));
		FirstOrderLOPFeature* currentFeatures = new FirstOrderLOPFeature[numberOfElements[k]];
	  //cout << "Direction: " << directions[k].first << " " << directions[k].second << endl;
		for(int i = 0; i < SourceLength - abs(directions[k].first); i++) {
			for(int j = 0; j < TargetLength - abs(directions[k].second); j++) {
	      //cout << "i: " << i << "j: " << j << endl;
				vector<double *> dynScalingFactor;
				vector<int> dynScalingIndex;
				int dynOffset = numberOfScalingFactors;
				int i1,i2;
				int j1,j2;
				if(directions[k].first > 0) {
					i1 = i;
					i2 = i + directions[k].first;
				}else {
					i1 = i - directions[k].first;
					i2 = i;
				}
				if(directions[k].second > 0) {
					j1 = j;
					j2 = j + directions[k].second;
				}else {
					j1 = j - directions[k].second;
					j2 = j;
				}
				if(usePOSStartFeature) {
					const string SourcePOS = G->GetSourcePOS(i1);
					const string TargetPOS = G->GetTargetPOS(j1);
					assert(POSSourceVoc.count(SourcePOS) != 0);
					assert(POSTargetVoc.count(TargetPOS) != 0);
					int SourcePOSIndex = POSSourceVoc[SourcePOS];
					int TargetPOSIndex = POSTargetVoc[TargetPOS];
					int index = k * (POSTargetVoc.size() * POSSourceVoc.size() )+ SourcePOSIndex * POSTargetVoc.size() + TargetPOSIndex;
					assert(index < POSTargetVoc.size() * POSSourceVoc.size() * directions.size());
		//cout << "Pos: ( "<< i1 << "," <<j1<< ") : " << endl;
		//cout << "Scaling Factor for " << SourcePOS << " and " << TargetPOS << " : " << POSStartScalingFactors[index] << endl;
		//cout << "Index : " << SourcePOSIndex << " , " << TargetPOSIndex << endl;
					dynScalingFactor.push_back(&POSStartScalingFactors[index*numberOfExperts]);
					dynScalingIndex.push_back(dynOffset + index);
					dynOffset += POSTargetVoc.size() * POSSourceVoc.size()*directions.size();
		
				}
				if(usePOSEndFeature) {
					const string SourcePOS = G->GetSourcePOS(i2);
					const string TargetPOS = G->GetTargetPOS(j2);
					assert(POSSourceVoc.count(SourcePOS) != 0);
					assert(POSTargetVoc.count(TargetPOS) != 0);
					int SourcePOSIndex = POSSourceVoc[SourcePOS];
					int TargetPOSIndex = POSTargetVoc[TargetPOS];
					int index = k * (POSTargetVoc.size() * POSSourceVoc.size() )+ SourcePOSIndex * POSTargetVoc.size() + TargetPOSIndex;
					assert(index < POSTargetVoc.size() * POSSourceVoc.size() * directions.size());
		//cout << "Pos: ( "<< i2 << "," <<j2<< ") : " << endl;
		//cout << "Scaling Factor for " << SourcePOS << " and " << TargetPOS << " : " << POSEndScalingFactors[index] << endl;
		//cout << "Index : " << SourcePOSIndex << " , " << TargetPOSIndex << endl;
					dynScalingFactor.push_back(&POSEndScalingFactors[index*numberOfExperts]);
					dynScalingIndex.push_back(dynOffset + index);
					dynOffset += POSTargetVoc.size() * POSSourceVoc.size()*directions.size();
		
				}
				if(usePOSSourceFeature) {
					const string Source1POS = G->GetSourcePOS(i1);
					const string Source2POS = G->GetSourcePOS(i2);
					assert(POSSourceVoc.count(Source1POS) != 0);
					assert(POSSourceVoc.count(Source2POS) != 0);
					int Source1POSIndex = POSSourceVoc[Source1POS];
					int Source2POSIndex = POSSourceVoc[Source2POS];
					int index = k * (POSSourceVoc.size() * POSSourceVoc.size() )+ Source1POSIndex * POSSourceVoc.size() + Source2POSIndex;
					assert(index < POSSourceVoc.size() * POSSourceVoc.size() * directions.size());
		//cout << "Source Pos: ( "<< i1 << "," <<i2<< ") : " << endl;
		//cout << "Scaling Factor for " << Source1POS << " and " << Source2POS << " : " << POSSourceScalingFactors[index] << endl;
		//cout << "Index : " << Source1POSIndex << " , " << Source2POSIndex << endl;
					dynScalingFactor.push_back(&POSSourceScalingFactors[index*numberOfExperts]);
					dynScalingIndex.push_back(dynOffset + index);
					dynOffset += POSSourceVoc.size() * POSSourceVoc.size()*directions.size();
		
				}
				if(usePOSTargetFeature) {
					const string Target1POS = G->GetTargetPOS(j1);
					const string Target2POS = G->GetTargetPOS(j2);
					assert(POSTargetVoc.count(Target1POS) != 0);
					assert(POSTargetVoc.count(Target2POS) != 0);
					int Target1POSIndex = POSTargetVoc[Target1POS];
					int Target2POSIndex = POSTargetVoc[Target2POS];
					int index = k * (POSTargetVoc.size() * POSTargetVoc.size() )+ Target1POSIndex * POSTargetVoc.size() + Target2POSIndex;
					assert(index < POSTargetVoc.size() * POSTargetVoc.size() * directions.size());
		//cout << "Target Pos: ( "<< j1 << "," <<j2<< ") : " << endl;
		//cout << "Scaling Factor for " << Target1POS << " and " << Target2POS << " : " << POSTargetScalingFactors[index] << endl;
		//cout << "Index : " << Target1POSIndex << " , " << Target2POSIndex << endl;
					dynScalingFactor.push_back(&POSTargetScalingFactors[index*numberOfExperts]);
					dynScalingIndex.push_back(dynOffset + index);
					dynOffset += POSTargetVoc.size() * POSTargetVoc.size()*directions.size();
		
				}
				int index = i*(TargetLength - abs(directions[k].second)) + j;
				assert(index < numberOfElements[k]);
				currentFeatures[index].Init(make_pair(G->GetNode(i1,j1),G->GetNode(i2,j2)),&ScalingFactors[k*NUMBER_OF_FOF*numberOfExperts],NUMBER_OF_FOF,dynScalingFactor,dynScalingIndex,offset + k*NUMBER_OF_FOF, offset, numberOfAllFeatures,numberOfExperts,expertWeights);
			}
		}
		features.push_back(currentFeatures);
	}
	assert(features.size() == directions.size());
	
}

void FirstOrderLOPFeatureSet::RemoveFeatures() {
	for(int i = 0; i < directions.size();i++) {
		delete[] features[i];
	}
	features.clear();
	numberOfElements.clear();
}

  /**
 * @return FirstOrderFeatureSetIterator
   */
FirstOrderLOPFeatureSetIterator * FirstOrderLOPFeatureSet::GetIterator ( ) const {
	return new FirstOrderLOPFeatureSetIterator(numberOfElements, features);
}

void FirstOrderLOPFeatureSet::EstimateScalingFactores() {
//	cout << "FOF: EstimateScalingFactores" << endl;
	EstimatedDerivation.clear();
	vector<double> ExpectValue;
	vector<double> ExampleValue;
	ExpectValue.resize(numberOfExperts,0);
	ExampleValue.resize(numberOfExperts,0);
	for(int k = 0; k < directions.size(); k++) {
		for(int i = 0; i < numberOfElements[k];i++) {
			vector<double> TempExpectValue = features[k][i].GetExpectationValue();
			vector<double> TempExampleValue = features[k][i].GetExampleValue();
			assert(TempExampleValue.size() == numberOfExperts);
			assert(TempExpectValue.size() == numberOfExperts);
			for(int j = 0; j < numberOfExperts; j++) {
				ExpectValue[j] += TempExpectValue[j];
				ExampleValue[j] += TempExampleValue[j];
			}
		}
	}

	for(int j = 0; j < numberOfExperts; j++) {
		double deltaL = ExampleValue[j] - ExpectValue[j];
//		cout << "Derivation for expert " << j << ": " << deltaL << endl;
		EstimatedDerivation.push_back(deltaL);
	}
}



