#ifndef FIRSTORDERLOPFEATURESET_H
#define FIRSTORDERLOPFEATURESET_H

#include <featureset.h>
#include<firstorderlopfeature.h>
#include <firstorderlopfeaturesetiterator.h>
#include <vector>
#include <map>
#include <utils.h>
/**
	@author Jan Niehues <jan@niehues.info>
*/
class FirstOrderLOPFeatureSet : public FeatureSet
{
private:
  	int numberOfExperts;
	double * expertWeights;
	double * ScalingFactors;
	vector<double> EstimatedDerivation;
	int numberOfScalingFactors;
	int NumberOfDynamicFeatures;
	vector<pair<int,int> > directions;
	int SourceLength;
	int TargetLength;
	vector<FirstOrderLOPFeature*> features;
	vector<int> numberOfElements;
	map<string, int> POSSourceVoc;
	map<string, int> POSTargetVoc;
	double * POSStartScalingFactors;
	vector<string> POSStartScalingFactorFile;
	int usePOSStartFeature;
	double * POSEndScalingFactors;
	vector<string> POSEndScalingFactorFile;
	int usePOSEndFeature;
	double * POSSourceScalingFactors;
	vector<string> POSSourceScalingFactorFile;
	int usePOSSourceFeature;
	double * POSTargetScalingFactors;
	vector<string> POSTargetScalingFactorFile;
	int usePOSTargetFeature;

  /**
   * @param  ParamFile
   */
    void LoadDirections(const char * ParamFile);
  /**
   * @param  ParamFile
   */
    void LoadParamFile(const char * ParamFile);

public:
    FirstOrderLOPFeatureSet(const char* ParamFile, const char * ScalingFactorString,int numberOfExp, double * expWeights);

    ~FirstOrderLOPFeatureSet();

  /**
   * @return int
   */
  virtual inline int GetNumberOfScalingFactors() {return numberOfScalingFactors + NumberOfDynamicFeatures;};

  /**
   * @param  G
   */
  void GenerateFeatures (Graph * G , int offset, int numberOfAllFeatures);

  void RemoveFeatures();

  /**
   * @return FirstOrderSetIterator
   */
  FirstOrderLOPFeatureSetIterator * GetIterator ( ) const;
  /**
   */
  void EstimateScalingFactores ();

  vector<double> inline GetEstimatedDerivations() const {return EstimatedDerivation;};


};

#endif
