
#include "graph.h"

Graph::Graph(int SLength, int TLength,vector<string> SWords,vector<string> TWords,vector<string> SPOS,vector<string> TPOS)
{
	SourceLength = SLength;
	TargetLength = TLength;
	SourceWords = SWords;
	TargetWords = TWords;
	SourcePOS = SPOS;
	TargetPOS = TPOS;
	//create Hidden nodes
	nodes = new HiddenNode[SourceLength*TargetLength];
	SetWords();
	DisplayGraph = 0;
	useTree = 0;
}

Graph::Graph(int SLength, int TLength,vector<string> SWords,vector<string> TWords,vector<string> SPOS,vector<string> TPOS,vector<pair<int,int> > & Alignment,vector<pair<int,int> > & possibleAlignment)
{
	SourceLength = SLength;
	TargetLength = TLength;
	SourceWords = SWords;
	TargetWords = TWords;
	SourcePOS = SPOS;
	TargetPOS = TPOS;
	//create Hidden nodes
	nodes = new HiddenNode[SourceLength*TargetLength];
	SetWords();
	DisplayGraph = 0;
	SetAlignment(Alignment,possibleAlignment);
	useTree = 0;
}
void Graph::SetAlignment(vector<pair<int,int> > & Alignment,vector<pair<int,int> > & possibleAlignment) {
	//cout << "Sure Alignment: ";
	for(int i = 0; i < Alignment.size(); i++) {
		int index = Alignment[i].first*TargetLength + Alignment[i].second;
		//cout << index << " ";
		if(index >= SourceLength*TargetLength) {
			cout << Alignment[i].first << "-" << Alignment[i].second << endl;
			cout << "SourceLength: " << SourceLength << endl;
			cout << "TargetLength:" << TargetLength << endl;
		}
		assert(index >= 0);
		assert(index < SourceLength*TargetLength);
		nodes[index].SetGivenValue(1);
	}
	//cout << endl;
	//cout << "Possible Alignment: ";
	for(int i = 0; i < possibleAlignment.size(); i++) {
		//cout << index << " ";
		int index = possibleAlignment[i].first*TargetLength + possibleAlignment[i].second;
		assert(index < SourceLength*TargetLength);
		nodes[index].SetGivenValue(-1);
	}
	//cout << endl;

}

void Graph::SetWords() {
	for(int i = 0; i < SourceLength; i++) {
		for(int j = 0; j < TargetLength; j++) {
			nodes[i*TargetLength + j].SetWords(SourceWords[i],TargetWords[j]);
		}	

	}

}

Graph::~Graph()
{
	delete[] nodes;
}


void Graph::Inference (int Iterations) {

	assert(featureSets.size() > 0);
	//Init all message of Factored Nodes to same value
	//is done during creation

	for(int i = 0; i < Iterations; i++) {
		//cout << "BP Iteration: " << i << endl;
		//Update messages of hidden nodes
		UpdateHiddenNodes();
		if(DisplayGraph) {
			Display(i);
		}
		//update messages of factoed nodes
		UpdateFactoredNodes();
	}

}

void Graph::Optimize( int Iterations) {
	int storeDisplayGraph = DisplayGraph;
  // do not display the inference during Optimization
	DisplayGraph = 0;
	Inference(Iterations);
	DisplayGraph = storeDisplayGraph;

	for(int i = 0; i < featureSets.size();i++) {
		featureSets[i]->EstimateScalingFactores();
	}

}
void Graph::OutputAlignment(ofstream * AlignmentFile) {
  //cout << "Output Alignment" << endl;
	for(int i = 0; i < SourceLength;i++) {
		for(int j = 0; j < TargetLength; j++) {
			pair<double,double> n = nodes[i*TargetLength + j].GetBelief();
			if(n.first < n.second) {
			//if(n.first < 0.95) {
			  //cout << "(" << i << "," << j << ")" << endl;
				*AlignmentFile << (i+1) << "-" << (j+1) << " ";
			}
		}
	}
	*AlignmentFile << endl;
}

void Graph::UpdateHiddenNodes() {
	for(int i = 0; i < SourceLength;i++) {
		for(int j = 0; j < TargetLength; j++) {
			nodes[i*TargetLength + j].CalcMessages();
		}
	}
}

void Graph::UpdateFactoredNodes() {
	for(int i = 0; i < featureSets.size(); i++) {
		FeatureSetIterator * iter = featureSets[i]->GetIterator();
		//pair<double,double> t = iter->feature()->GetMessage(&nodes[0]);
		//cout << "Message (" << t.first << "," << t.second <<")" << endl;
		for(; !iter->end();iter->next()) {
			Feature *f = iter->feature();
			f->CalcMessages();
		}
		delete iter;
	}
}

void Graph::Display(int time) {
	for(int i = 0; i < SourceLength;i++) {
		for(int j = 0; j < TargetLength; j++) {
			*out << time << " " << i << " " << j << " ";
			nodes[i*TargetLength+j].Display(out);
			*out << endl;
		}
	}
}

double Graph::ComputeLogProbabilityCorrect() {
	double logProb = 0;
	for(int i = 0; i < SourceLength;i++) {
		for(int j = 0; j < TargetLength; j++) {
			int index = nodes[i*TargetLength+j].GetGivenValue();
			if((index != -1) &&(!nodes[i*TargetLength+j].IsStatic())) {
				pair<double,double> b =  nodes[i*TargetLength+j].GetBelief();
				if(index == 0) {
					logProb += log(b.first);
				}else{
					logProb +=log(b.second);
				}
			}
		}
	}
	logProb /= (SourceLength * TargetLength);
	return logProb;
}

double Graph::ComputeLogProbabilityBest() {
	double logProb = 0;
	for(int i = 0; i < SourceLength;i++) {
		for(int j = 0; j < TargetLength; j++) {
			int index = nodes[i*TargetLength+j].GetGivenValue();
			if(index != -1) {
				pair<double,double> b =  nodes[i*TargetLength+j].GetBelief();
				if(b.first > b.second) {
					logProb += log(b.first);
				}else{
					logProb +=log(b.second);
				}
			}
		}
	}
	logProb /= (SourceLength * TargetLength);
	return logProb;
}

void Graph::ComputeCorrectLinks(int & CorrectLink, int & FailedLink, int & CorrectNoLink, int & FailedNoLink) {
	for(int i = 0; i < SourceLength;i++) {
		for(int j = 0; j < TargetLength; j++) {
			pair<double,double> b =  nodes[i*TargetLength+j].GetBelief();
			int index = nodes[i*TargetLength+j].GetGivenValue();
			if(index == 1) {
				if(b.second >= b.first) {
					CorrectLink++;
				}else{
					FailedLink++;
				}
			}else if (index == 0) {
				if(b.first > b.second) {
					CorrectNoLink++;
				}else{
					FailedNoLink++;
				}
			}
		}
	}
}

GraphIterator * Graph::GetIterator ( ) const {
	return new GraphIterator(SourceLength,TargetLength,nodes);
};



/*!
    \fn Graph::LoadPreAlignment()
 */
void Graph::LoadPreAlignment(vector<pair<int,int> > & preAlignments, vector<pair<int,int> > & possiblePreAlignments)
{
	int nodeStatus[SourceLength*TargetLength];
	for(int i = 0; i < SourceLength*TargetLength; i++) {
		nodeStatus[i] = 0;
	}
	for(int i = 0; i < preAlignments.size(); i++) {
		int index = preAlignments[i].first*TargetLength + preAlignments[i].second;
		assert(index >= 0);
		assert(index < SourceLength * TargetLength);
		//cout << "Set Alignment " << preAlignments[i].first << "-" << preAlignments[i].second << endl; 
		nodeStatus[index] = 1;
	}
	for(int i = 0; i < possiblePreAlignments.size(); i++) {
		int index = possiblePreAlignments[i].first*TargetLength + possiblePreAlignments[i].second;
		assert(index >= 0);
		if(index >= SourceLength * TargetLength) {

			cout << "Problem with alignment " << possiblePreAlignments[i].first << "-" << possiblePreAlignments[i].second << "at Index:" << i << endl;
		}
		assert(index < SourceLength * TargetLength);
		//cout << "Set PossibleAlignment " << possiblePreAlignments[i].first << "-" << possiblePreAlignments[i].second << endl;
		nodeStatus[index] = 2;
	}
	for(int i = 0; i < SourceLength*TargetLength; i++) {
		switch(nodeStatus[i]) {
			case 1:
				//cout << "Node " << i << "static 1" << endl;
				nodes[i].SetStaticNode(1);
				break;
			case 0:
				//cout << "Node " << i << "static 0" << endl;
				nodes[i].SetStaticNode(0);
				break;
		}
	}
}
void Graph::SetLinkFeatures(vector<vector<LinkFeatures> > & features) {
	linkFeatures = features;

}
