#ifndef GRAPH_H
#define GRAPH_H


struct LinkFeatures {
	int source;
	int target;
	double value;

};
class Graph;
#include "hiddennode.h"
#include "graphiterator.h"
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <cassert>
#include <featureset.h>
#include <vector>
#include <cmath>
#include <parseTree.h>
using namespace std;
/**
	@author Jan Niehues <jan@niehues.info>
*/

using namespace std;
class Graph{
private:
	int SourceLength;
	int TargetLength;
	vector<string> SourceWords;
	vector<string> TargetWords;
	vector<string> SourcePOS;
	vector<string> TargetPOS;
	int useTree;
	ParseTree SourceTree;
	ParseTree TargetTree;
	vector<FeatureSet*> featureSets;
	HiddenNode* nodes;
	void UpdateHiddenNodes();
	void UpdateFactoredNodes();
	void Display(int time);
	vector< vector<LinkFeatures> > linkFeatures;
	ofstream * out;
	int DisplayGraph;
	void SetAlignment(vector<pair<int,int> > & Alignment,vector<pair<int,int> > & possibleAlignment);
	void SetWords();
public:
	Graph(int SLength, int TLength, vector<string> SWords, vector<string> TWords,vector<string> SPOS,vector<string> TPOS);
	Graph(int SLength, int TLength, vector<string> SWords, vector<string> TWords,vector<string> SPOS,vector<string> TPOS,vector<pair<int,int> > & Alignment,vector<pair<int,int> > & possibleAlignment);
	~Graph();
	void SetLinkFeatures(vector<vector<LinkFeatures> > & features);
	inline vector<vector<LinkFeatures> > & GetLinkFeatures() { return linkFeatures;};
	inline int GetSourceLength() const {return SourceLength;};
	inline int GetTargetLength() const {return TargetLength;};
	inline void SetDisplay(ofstream * O) { out = O; DisplayGraph = 1;};
/**
   * @return HiddenNode
   * @param  SourceIndex
   * @param  TargetIndex
   */
	inline HiddenNode * GetNode(int SourceIndex,int TargetIndex) const {
		assert(SourceIndex*TargetLength + TargetIndex < SourceLength * TargetLength);
		return &nodes[SourceIndex*TargetLength + TargetIndex];
	}
	inline const string GetSourceWord(int index) const {assert(index < SourceLength); return SourceWords[index];};
	inline const string GetTargetWord(int index) const {assert(index < TargetLength);return TargetWords[index];};
	inline const string GetSourcePOS(int index) const {assert(index < SourceLength);if(index >= SourcePOS.size()) return "NN";return SourcePOS[index];};
	inline const string GetTargetPOS(int index) const {assert(index < TargetLength);if(index >= TargetPOS.size()) return "NN";return TargetPOS[index];};
	inline void AddFeatureSet(FeatureSet* set) {featureSets.push_back(set);};
  /**
   * @param Iterations
   */
  void Inference ( int Iterations);
  /**
   * @param Iterations
   * @param learnrate
   */
  void Optimize ( int Iterations);

  void OutputAlignment(ofstream * AlignmentFile);

  double ComputeLogProbabilityBest();
  double ComputeLogProbabilityCorrect();
  void ComputeCorrectLinks(int & CorrectLink, int & FailedLink, int & CorrectNoLink, int & FaildNoLink);

  /**
   * @return FeatureSetIterator
   */
  GraphIterator * GetIterator ( ) const;
    void LoadPreAlignment(vector<pair<int,int> > & preAlignments, vector<pair<int,int> > & possiblePreAlignments);

	inline void LoadTrees(ParseTree sTree, ParseTree tTree) {SourceTree = sTree; TargetTree = tTree;useTree = 1;};

	inline ParseTree & GetSourceTree() {assert(useTree == 1); return SourceTree;}
	inline ParseTree & GetTargetTree() {assert(useTree == 1); return TargetTree;};
  /**
   */


};

#endif
