
#include "hiddennode.h"

HiddenNode::HiddenNode()
{
  givenValue = 0;
	staticNode = 0;
}


HiddenNode::~HiddenNode()
{
}
pair<double,double> HiddenNode::GetMessage(Feature *ToNode) const
{
	if(staticNode) {
		return staticMessage;
	}
	for(int i = 0; i < Connections.size(); i++) {
		if(Connections[i] == ToNode) {
			//cout << "Message from hidden node: (" << messages[i].first << "," << messages[i].second << ")" << endl;
			return messages[i];
		}
	}	
	assert(0);
}
void HiddenNode::CalcMessages()
{
	//cout << "Calc Hidden Node:" << endl;
	if(staticNode) {
		return;
	}
	for(int i = 0; i < Connections.size(); i++) {
		pair<double,double> m = make_pair(1,1);
		for(int j = 0; j < Connections.size();j++) {
			if(i != j) {
				pair <double,double> r = Connections[j]->GetMessage(this);
				//cout << "Got Message: (" << r.first << "," << r.second << ")" << endl;
				m.first *= r.first;
				m.second *= r.second;
			}
		}
		double sum = m.first + m.second;
		m.first /= sum;
		m.second /= sum;
		if(isnan(m.first)) {
		  for(int j = 0; j < Connections.size();j++) {
		    if(i != j) {
		      pair <double,double> r = Connections[j]->GetMessage(this);
		      cout << "Got Message: from type" << Connections[j]->GetType() << " (" << r.first << "," << r.second << ")" << endl;
		    }
		  }
		  cout << " Sum: " << sum << endl << flush;
		}
		assert(!isnan(m.first));
		assert(!isnan(m.second));
		messages[i] = m;
		//cout << "Message to " << i << ": (" << messages[i].first << "," << messages[i].second << ")" << endl;
	}
	//cout << endl;
}

SparseArray HiddenNode::GetDerivation(int size) const
{
	double * der0 = new double[size];
	for(int i = 0; i < size; i++)
		der0[i] = 0;
	if(staticNode) {
		SparseArray r;
		r.size = 0;
		return r;
	}
	double * der1 = new double[size];
	for(int i = 0; i < size; i++)
		der1[i] = 0;
	
	pair<double,double> prod = make_pair(1,1);
	vector<int> usedIndex;
	int existDer[size];
	for(int i = 0; i < size; i++)
		existDer[i] = 0;

  for(int i = 0; i < Connections.size(); i++) {
    pair<double,double> r = Connections[i]->GetMessage(this);
    //cout << " Factor Node: " << i << endl;
    //cout << " Got Message: (" << r.first << "," << r.second << ")" <<endl;
    SparseArray d = Connections[i]->GetDerivation(this);
    //cout << "Input Derivation: " << endl;
    //for(int j = 0; j < d.first.size();j++) {
    //  if((d.first[j] != 0) || (d.second[j] != 0)){
    //	cout << j << ": (" << d.first[j] << "," << d.second[j] << ")" << endl;
    // }
    //}
    /*cout << "Derivation from Factored Node " << i << endl;
    for(int j = 0 ; j < d.first.size(); j++) {
	cout << d.first[j] << " ";
    }
    cout << endl;*/
    //assert(d.first.size() <= size);
    //assert(d.second.size() <= size);
/*    for(int j = 0; j < size; j++) {
      if(j < d.first.size()) {
	der0[j] = der0[j] * r.first + prod.first * d.first[j];
	der1[j] = der1[j] * r.second + prod.second * d.second[j];
      }else {
	//d.first[i] = 0 and d.second[i] = 0;
	der0[j] = der0[j] * r.first;
	der1[j] = der1[j] * r.second;	
      }
    }
*/
	for(int j = 0; j < usedIndex.size(); j++) {
		der0[usedIndex[j]] = der0[usedIndex[j]] * r.first;
		der1[usedIndex[j]] = der1[usedIndex[j]] * r.second;			
	} 
	for(int j = 0; j < d.size; j++) {
		//cout << "Index: "  << d.index[j] << " Size: " << size << endl;
		assert(d.index[j] < size);
		assert(d.index[j] >= 0);
		der0[d.index[j]] += prod.first * d.data[j];
		der1[d.index[j]] += prod.second * (-1 * d.data[j]);
		if(existDer[d.index[j]] == 0) {
			usedIndex.push_back(d.index[j]);
			existDer[d.index[j]] = 1;
		}
	}
   prod.first *= r.first;
    prod.second *= r.second;
  }
  //cout << "Der 1. Step: "<< endl;
  //for(int j = 0; j < der.first.size();j++) {
  //  if((der.first[j] != 0) || (der.second[j] != 0)){
  //    cout << j << ": (" << der.first[j] << "," << der.second[j] << ")" << endl;
  //  }
  //}


	SparseArray result;
	result.size = usedIndex.size();
	result.data = new double[result.size];
	result.index = new int[result.size];

  double denominator = (prod.first + prod.second) * (prod.first + prod.second);
  //cout << "Denominator: " << denominator << endl;
  //cout << "Prod: (" << prod.first << "," << prod.second << ")" << endl;
  for(int j = 0; j < usedIndex.size(); j++) {
	result.index[j] = usedIndex[j];
	result.data[j] = ((der0[usedIndex[j]] * prod.second - der1[usedIndex[j]] * prod.first)/denominator);
    //cout << "Result:" << endl;
    //if((result.first[j]!= 0) ||(result.second[j] != 0)) {
    //	cout << j << ": (" << result.first[j] << "," << result.second[j] << ")" << endl;     
    //}
  }
	delete[] der0;
	delete[] der1;
  return result;
}

pair<double,double> HiddenNode::GetBelief() const
{
	if(staticNode) {
		return staticMessage;
	}
	pair<double,double> b = make_pair(1,1);
	for(int i = 0; i < Connections.size(); i++) {
		pair<double,double> r = Connections[i]->GetMessage(this);
		b.first *= r.first;
		b.second *= r.second;		
	}
	double sum = b.first + b.second;
	b.first /= sum;
	b.second /= sum;
	//cout << "Belief: ("<< b.first << "," << b.second << ")" << endl;
	return b;
}
void HiddenNode::AddConnection(Feature *ToNode)
{
	Connections.push_back(ToNode);
	messages.push_back(make_pair(0,0));
}

void HiddenNode::SetStaticNode(int value) {
staticNode = 1;
switch(value) {
	case 1:
		staticMessage = make_pair(0,1);
		break;
	case 0:
		staticMessage = make_pair(1,0);
		break;
	default:
		cout << "Static Value " << value <<" unknown" << endl;
		exit(1); 
}

}
void HiddenNode::Display(ofstream * out) {
	pair<double,double> t;
	t = GetBelief();
	pair<double,double> messageOut[5];
	pair <double,double> messageIn[5];
	for(int i = 0; i < Connections.size(); i++) {
		int index = Connections[i]->GetType();
		messageOut[index] = GetMessage(Connections[i]);
		messageIn[index] = Connections[i]->GetMessage(this);
	}	
	*out << t.first << " " << t.second << " ";
	for(int i =1; i < 4; i++) {
		*out << messageOut[i].first << " " << messageOut[i].second << " ";
		*out << messageIn[i].first << " " << messageIn[i].second << " ";
	}

}
