#include <string.h>
#include "localfeature.h"

LocalFeature::LocalFeature()
 : Feature()
{
	kernelType = 0;
	number = 0;
	message = make_pair(0.5,0.5);
	derivation.size = 0;
	gamma = 2;
}


LocalFeature::~LocalFeature()
{
	delete[] p;
	delete[] FeatureValue;
	if(derivation.size > 0) {
		delete[] derivation.data;
		delete[] derivation.index;
	}
}
void LocalFeature::Init(HiddenNode * n,double *probability, double* SFactors, int num,vector<double> dynScalingFactor, vector <int> dynScalingIndex,int off, int numberOfAF) {
	node = n;
	p = probability;
	number = num;
	ScalingFactors = SFactors;
	dynamicScalingFactor = dynScalingFactor;
	dynamicScalingIndex = dynScalingIndex;
	offset = off;
	numberOfAllFeatures = numberOfAF;
	node->AddConnection(this);
	FeatureValue = new double[2*num];

	//here we can already calc the messages, because the will always be the same
	PreCalcMessage();
}

void LocalFeature::PreCalcMessage() {
	//init;
	message.first = 0;
	message.second = 0;


	if(kernelType == 0) {
		//basleline features
		for(int i = 0; i < number; i++) {
		  FeatureValue[2*i] = (1-p[i]);
		  FeatureValue[2*i + 1] = p[i]; 
		  message.first += ScalingFactors[2*i] * (1-p[i]);//log((1-p[i]));
		  message.second += ScalingFactors[2*i + 1] *p[i];//*log(p[i]);
		}

		//dynamic scaling Factors
		for(int i = 0; i < dynamicScalingFactor.size(); i++) {
			message.second += dynamicScalingFactor[i];
			//cout << "Use dynamic Feature number: " << dynamicScalingIndex[i] << endl;
		}
	}else if(kernelType == 1) {
		for(int i = 0; i < number; i++) {
			//cout << "Lex Probability "<< i << ":" << p[i];
			//cout << " ScalingFactors:" << ScalingFactors[2*i] << " " << ScalingFactors[2*i+1] << endl;
			FeatureValue[2*i] = (1-p[i]);
			FeatureValue[2*i + 1] = p[i]; 
			message.first += ScalingFactors[2*i] * (1-p[i]);//log((1-p[i]));
			message.second += ScalingFactors[2*i + 1] *p[i];//*log(p[i]);
		}
		//cout << "Current Value:" << message.first << "," << message.second << endl;
		//dynamic scaling Factors
		for(int i = 0; i < dynamicScalingFactor.size(); i++) {
			//cout << "Scaling Factor:" << dynamicScalingFactor[i] << endl;
			message.second += dynamicScalingFactor[i];
			//cout << "Use dynamic Feature number: " << dynamicScalingIndex[i] << endl;
		}
		//cout << "Current Value:" << message.first << "," << message.second << endl;		
		//message.first += 1;
		//message.second += 1;
		kernelValue.first = pow(message.first,gamma -1);
		kernelValue.second = pow(message.second,gamma -1);
		message.first = pow(message.first,gamma);
		message.second = pow(message.second,gamma);
		//cout << "Kernel: " << kernelValue.first  << "," << kernelValue.second << endl;	
		//cout << "Message: " << message.first << "," << message.second << endl;	
	}else {
		cerr << "Not supportet Kernel Type: " << kernelType << endl;
		exit(-1);
	}

	message.first = exp(message.first);
	message.second = exp(message.second);
	double sum = message.first + message.second;
	message.first /= sum;
	message.second /= sum;
	//cout << "Message value:" << message.first << "," << message.second << endl;
	assert(!isnan(message.first));
	assert(!isnan(message.second));

}
  /**
   */
void LocalFeature::CalcMessages ( ) {
	//nothing to da, local features send always the same messages	
}


  /**
   * @return vector<double>
   */
vector<double> LocalFeature::GetExpectationValue(){
	vector<double> values;
	switch(kernelType) {
	case 0:	if(node->GetGivenValue() == -1) {
			values.resize(2*number,0);
			return values;
		}
		for(int i = 0; i < 2* number; i++) {
			if(i % 2 == 0){
				values.push_back(FeatureValue[i]* node->GetBelief().first);
			}else {
				values.push_back(FeatureValue[i]*node->GetBelief().second);
			}
		}
		break;
	case 1:	if(node->GetGivenValue() == -1) {
			values.resize(2*number,0);
			return values;
		}
		for(int i = 0; i < 2* number; i++) {
			if(i % 2 == 0){
				values.push_back(gamma*kernelValue.first*FeatureValue[i]* node->GetBelief().first);
			}else {
				values.push_back(gamma*kernelValue.second*FeatureValue[i]*node->GetBelief().second);
			}
		}
		break;
	default: cerr << "Kernel Type not supportet " << kernelType << endl;
		exit(1);
	}
	return values;
}

  /**
   * @return vector<double>
   */
vector<double> LocalFeature::GetExampleValue(){
  vector<double> values;

	if(kernelType == 0) {
		if(node->GetGivenValue() == -1) {
			values.resize(2*number,0);
			return values;
		}
		for(int i = 0; i < 2* number; i++) {
			if(i% 2 == 0){
				values.push_back(FeatureValue[i] * (1 - node->GetGivenValue()));
			}else {
				values.push_back(FeatureValue[i] * node->GetGivenValue());    
			}
		}
	}else if(kernelType == 1) {
		if(node->GetGivenValue() == -1) {
			values.resize(2*number,0);
			return values;
		}
		for(int i = 0; i < 2* number; i++) {
			if(i% 2 == 0){
				values.push_back(gamma*kernelValue.first*FeatureValue[i] * (1 - node->GetGivenValue()));
			}else {
				values.push_back(gamma*kernelValue.second*FeatureValue[i] * node->GetGivenValue());
			}
		}
	}else {
		cerr << "Kernel Type not supportet " << kernelType;
		exit(-1);
	}		
	return values;
}
/**
 * @return vector<double>
 */
vector<double> LocalFeature::GetDynamicExpectationValue(){
  vector<double> values;
	switch(kernelType) {
		case 0: if(node->GetGivenValue() == -1) {
				values.resize(dynamicScalingFactor.size(),0);
 				return values;
			}
			for(int i = 0; i < dynamicScalingFactor.size(); i++) {
				values.push_back(node->GetBelief().second);
			}
			break;
		case 1:if(node->GetGivenValue() == -1) {
				values.resize(dynamicScalingFactor.size(),0);
 				return values;
			}
			for(int i = 0; i < dynamicScalingFactor.size(); i++) {
				values.push_back(gamma*kernelValue.second*node->GetBelief().second);
			}
			break;
//		default: cerr << "Kernel Type not supported " << kernelType << endl;
//			exit(-1);
	}
	return values;
}

/**
 * @return double
 */
vector<double> LocalFeature::GetDynamicExampleValue(){
	vector<double> values;
	switch(kernelType) {
		case 0: if(node->GetGivenValue() == -1) {
				values.resize(dynamicScalingFactor.size(),0);
				return values;
			}

			for(int i = 0; i < dynamicScalingFactor.size(); i++) {
				values.push_back(node->GetGivenValue());
			}
			break;
		case 1:if(node->GetGivenValue() == -1) {
				values.resize(dynamicScalingFactor.size(),0);
				return values;
			}

			for(int i = 0; i < dynamicScalingFactor.size(); i++) {
				values.push_back(gamma*kernelValue.second*node->GetGivenValue());
			}
			break;
		default: cerr << "Kernel Type not supported " << kernelType << endl;
			exit(-1);
	}
	return values;
  


}
/**
 * @return vector<double>
 */
vector<int> LocalFeature::GetDynamicSFIndex(){
  return dynamicScalingIndex;
}


void LocalFeature::CalcDerivations() {
	derivation.size = 2 * number + dynamicScalingFactor.size();
	derivation.data = new double[derivation.size];
	derivation.index = new int[derivation.size];

	double f = message.first * message.second;
	for(int i = 0; i < number; i++) {
	assert(2*i + 1 < derivation.size);
		derivation.index[2*i] = offset + 2*i;
		derivation.index[2*i+1] = offset + 2*i + 1;
		switch(kernelType) {
		case 0:	derivation.data[2*i] = f * FeatureValue[2*i];
			derivation.data[2*i + 1] = f * -1 * FeatureValue[2*i+1];
			break;
		case 1: derivation.data[2*i] = f * gamma*kernelValue.first*FeatureValue[2*i];
			derivation.data[2*i + 1] = f * -1 * gamma*kernelValue.second*FeatureValue[2*i+1];
			break;
		default: cerr << "Kernel Type not supported " << kernelType << endl;
			exit(-1);
		}
	}
	for(int i = 0; i < dynamicScalingFactor.size(); i++) {
		assert(2*number +i < derivation.size);
		derivation.index[2*number + i] = offset + dynamicScalingIndex[i];
		switch(kernelType) {
		case 0:	derivation.data[2*number + i] = f * -1;
			break;
		case 1:	derivation.data[2*number + i] = f * gamma*kernelValue.second*-1;
			break;
		default: cerr << "Kernel Type not supported " << kernelType << endl;
			exit(-1);
		}
	}
 
}
