#ifndef LOCALFEATURESET_H
#define LOCALFEATURESET_H

#include <fstream>
#include <featureset.h>
#include <localfeature.h>
#include <localfeaturesetiterator.h>
#include <vector>
#include <map>
#include <utility>
/**
	@author Jan Niehues <jan@niehues.info>
*/

class LocalFeatureSet : public FeatureSet
{

private:
	vector<double> EstimatedDerivation;
	map<pair<string,string>,double>* Lexicon;
	vector<int> SourceNormalized;
	vector<int> TargetNormalized;
	map<string, int> POSSourceVoc;
	map<string, int> POSTargetVoc;
	vector<double> POSScalingFactors;
	string POSScalingFactorFile;
	map<string, int> FrequentWordSourceVoc;
	map<string, int> FrequentWordTargetVoc;
	vector<double> FrequentWordScalingFactors;
	string FrequentWordScalingFactorFile;
	int additionalLexFeatures;
	int NumberOfFeatures;
	int NumberOfDynamicFeatures;
	int useRelativePosition;
	int useIdentityFeature;
	int useSimilarityFeature;
	int numberOfLinkFeatures;
        int useTreeFeature; // Wei Chen 2008
	int usePOSFeature;
	int useFrequentWordFeature;
	char * POSFeatureParamFile;
	char * FrequentWordFeatureParamFile;
	LocalFeature *features;
	int SourceLength;
	int TargetLength;
	void InitScalingFactors(vector<double> & Factores);
	double * InitLinkFeatures(vector< vector<LinkFeatures> > & links);
	double * CreateLexMatrix(int & numberOfLexMatrix, Graph * g);
	double* ScalingFactors;
	int numberOfLexicons;
	int additionalFeatures;
	void LoadParamFile(const char* ParamFile);


public:
	LocalFeatureSet(const char* ParamFile, vector<double> & ScalingFactores);

    ~LocalFeatureSet();
  /**
   * Contruct all LocalLOPFeatures needed for the given Graph.
   * Calculated the feature value for every LocalLOPFeature and build the set of features	
   * @param  G Graph for which the features are generate
   * @param  offset Offset of the features used by the set
   * @param numberOfAllFeature Number of all Features used by this LOP-CRF
   */
  virtual void GenerateFeatures (Graph * G  , int offset, int numberOfAllFeatures);

  /**
   * @param newScalingFactors
   */
  virtual void SetScalingFactors(vector<double> & newScalingFactors);

  virtual void RemoveFeatures();

  /**
   * @return FeatureSetIterator
   */
  virtual FeatureSetIterator * GetIterator ( ) const;

  /**
   */
  virtual void EstimateScalingFactores ();

  vector<double> inline GetEstimatedDerivations() const {return EstimatedDerivation;};

  virtual void OutputScalingFactors();
  /**
   * @return int
   */
  virtual inline int GetNumberOfScalingFactors() {return NumberOfFeatures + NumberOfDynamicFeatures;};

};

#endif
