
#include "locallopfeature.h"

LocalLOPFeature::LocalLOPFeature()
 : Feature()
{
	number = 0;
	message = make_pair(0.5,0.5);
	derivation.size = 0;
}


LocalLOPFeature::~LocalLOPFeature()
{
	delete[] FeatureValue;
	delete[] scaledFeatureValue;
	if(derivation.size > 0) {
		delete[] derivation.data;
		delete[] derivation.index;
	}

}

void LocalLOPFeature::Init(HiddenNode * n,double *probability, double* SFactors, int num,vector<double * > dynScalingFactor, vector <int> dynScalingIndex,int off, int numberOfAF,int numberOfExp,double * expWeights) {
	numberOfExperts = numberOfExp;
	expertWeights = expWeights;
	node = n;
	p = probability;
	number = num;
	ScalingFactors = SFactors;
	dynamicScalingFactor = dynScalingFactor;
	dynamicScalingIndex = dynScalingIndex;
	offset = off;
	numberOfAllFeatures = numberOfAF;
	node->AddConnection(this);
	FeatureValue = new double[2*num];

	scaledFeatureValue = new double[2*numberOfExperts];
	for(int i = 0; i < 2*numberOfExperts; i++) {
		scaledFeatureValue[i] = 0;
	}
	//here we can already calc the messages, because the will always be the same
	PreCalcMessage();
}


void LocalLOPFeature::PreCalcMessage() {
	//init;
	message.first = 0;
	message.second = 0;


	//basleline features
	for(int i = 0; i < number; i++) {
	  FeatureValue[2*i] = (1-p[i]);
	  FeatureValue[2*i + 1] = p[i];
	  double SF1 = 0;
	  double SF2 = 0;
	  for(int j = 0; j < numberOfExperts; j++) {
		SF1 += expertWeights[j] * ScalingFactors[2*i*numberOfExperts + j];
		SF2 += expertWeights[j] * ScalingFactors[(2*i+1)*numberOfExperts + j];
		scaledFeatureValue[j] += ScalingFactors[2*i*numberOfExperts + j] * (1-p[i]);
		scaledFeatureValue[numberOfExperts + j] += ScalingFactors[(2*i+1)*numberOfExperts + j] * p[i];
          } 
	  //cout << "SF1: " << SF1 << " SF2: " << SF2 << " p[i]:" << p[i] << endl;
	  message.first += SF1 * (1-p[i]);//log((1-p[i]));
	  message.second += SF2 *p[i];//*log(p[i]);
	}

	//dynamic scaling Factors
	for(int i = 0; i < dynamicScalingFactor.size(); i++) {
		double SF = 0;
		for(int j = 0; j < numberOfExperts; j++) {
			//cout << "expertWeights[j] * dynamicScalingFactor[i][j] = " << expertWeights[j] << " * " << dynamicScalingFactor[i][j] << endl;
			SF += expertWeights[j] * dynamicScalingFactor[i][j];
			scaledFeatureValue[numberOfExperts + j] += dynamicScalingFactor[i][j];
        	} 
		
		message.second += SF;
		//cout << "Use dynamic Feature number: " << dynamicScalingIndex[i] << endl;
	}

	message.first = exp(message.first);
	message.second = exp(message.second);
	double sum = message.first + message.second;
	message.first /= sum;
	message.second /= sum;
	//cout << "Message value:" << message.first << "," << message.second << endl;
	assert(!isnan(message.first));
	assert(!isnan(message.second));

}

  /**
   */
void LocalLOPFeature::CalcMessages ( ) {
	//nothing to da, local features send always the same messages	
}
 
  /**
   * @return vector<double>
   */
vector<double> LocalLOPFeature::GetExpectationValue(){
//	cout << "LocalFeature::GetExpectationValue() " << endl;
	vector<double> values;
	values.resize(numberOfExperts,0);
	if(node->GetGivenValue() == -1) {
		return values;
	}
//	cout << "Belief: " << node->GetBelief().first << "," << node->GetBelief().second << endl;
	for(int j = 0; j < numberOfExperts; j++) {
		values[j] += node->GetBelief().first * scaledFeatureValue[j];
		values[j] += node->GetBelief().second * scaledFeatureValue[numberOfExperts + j];
//		cout << scaledFeatureValue[j] << " und " << scaledFeatureValue[numberOfExperts + j] << "-> " << values[j] << endl;
	}
	return values;
}

  /**
   * @return vector<double>
   */
vector<double> LocalLOPFeature::GetExampleValue(){
//	cout << "LocalFeature::GetExampleValue()" << endl;
  vector<double> values;
  values.resize(numberOfExperts,0);
	if(node->GetGivenValue() == -1) {
		return values;
	}
//	cout << "Given value: " << node->GetGivenValue() << endl;
//	cout << "Result: ";
	for(int j = 0; j < numberOfExperts; j++) {
		values[j] = scaledFeatureValue[node->GetGivenValue() * numberOfExperts + j];
//		cout << values[j] << " " << endl;
	}
	return values;
}


void LocalLOPFeature::CalcDerivations() {
	derivation.size = numberOfExperts;
	derivation.data = new double[derivation.size];
	derivation.index = new int[derivation.size];
	double f = message.first * message.second;
	for(int i = 0; i < numberOfExperts; i++) {
		assert(numberOfExperts ==  derivation.size);
		derivation.index[i] = i;
		derivation.data[i] = f * (scaledFeatureValue[i] - scaledFeatureValue[numberOfExperts + i]);
	}
 
}
