#ifndef LOCALLOPFEATURESET_H
#define LOCALLOPFEATURESET_H

#include <fstream>
#include <locallopfeaturesetiterator.h>
#include <locallopfeature.h>
#include <vector>
#include <map>
#include <utility>
#include <featureset.h>
#include <locallopfeature.h>
#include <utils.h>
/**
	@author Jan Niehues <jan@niehues.info>
*/
class LocalLOPFeatureSet : public FeatureSet
{
private:
	int numberOfExperts; /**< Number Of Experts of the model. */  
	double * expertWeights; /**< Weights for the different experts */
	vector<double> EstimatedDerivation;
	map<pair<string,string>,double>* Lexicon;
	vector<int> SourceNormalized;
	vector<int> TargetNormalized;
	map<string, int> POSSourceVoc;
	map<string, int> POSTargetVoc;
	double * POSScalingFactors;
	vector<string> POSScalingFactorFiles;
	map<string, int> FrequentWordSourceVoc;
	map<string, int> FrequentWordTargetVoc;
	double * FrequentWordScalingFactors;
	vector<string> FrequentWordScalingFactorFiles;
	int additionalLexFeatures;
	int NumberOfFeatures;
	int NumberOfDynamicFeatures;
	int useRelativePosition;
	int useIdentityFeature;
	int useSimilarityFeature;
	int numberOfLinkFeatures;
	int usePOSFeature;
	int useFrequentWordFeature;
	LocalLOPFeature *features;
	int SourceLength;
	int TargetLength;
	double * InitLinkFeatures(vector< vector<LinkFeatures> > & links);
	double * CreateLexMatrix(int & numberOfLexMatrix, Graph * g);
	double* ScalingFactors;
	int numberOfLexicons;
	int additionalFeatures;
	void LoadParamFile(const char* ParamFile);
	void InitScalingFactors(const char * ScalingFactorString);
public:
    LocalLOPFeatureSet(const char* ParamFile, const char * ScalingFactorsString,int numberOfExp,double * expWeights);


    ~LocalLOPFeatureSet();

  /**
   * Contruct all LocalLOPFeatures needed for the given Graph.
   * Calculated the feature value for every LocalLOPFeature and build the set of features	
   * @param  G Graph for which the features are generate
   * @param  offset Offset of the features used by the set
   * @param numberOfAllFeature Number of all Features used by this LOP-CRF
   */
  virtual void GenerateFeatures (Graph * G  , int offset, int numberOfAllFeatures);

  virtual void RemoveFeatures();

  /**
   * @return FeatureSetIterator
   */
  virtual FeatureSetIterator * GetIterator ( ) const;

  /**
   */
  virtual void EstimateScalingFactores ();

  vector<double> inline GetEstimatedDerivations() const {return EstimatedDerivation;};

  /**
   * @return int
   */
  virtual inline int GetNumberOfScalingFactors() {return NumberOfFeatures + NumberOfDynamicFeatures;};

};

#endif
