
#include "lopoptimizeraer.h"

LOPOptimizerAER::LOPOptimizerAER(vector <FeatureSet *> FSets,double* w, char * learnRateString, int numberOfExp)
 : Optimizer(FSets)
{
	FeatureSets = FSets;
	weights = w;
	numberOfExperts = numberOfExp;
	if(learnRateString == NULL) {
		cout << " For optimization the learnRate is needed" << endl;
		exit(-1);
	}
	learnRate = new double[numberOfExperts];
	SetLearnRate(learnRateString,learnRate,numberOfExperts);

	FeatureSets = FSets;

	TruePositiveP = 0;
	FalsePositiveP = 0;
	TruePositiveS = 0;
	FalsePositiveS = 0;
	SLinks = 0;
	PLinks = 0;
	SmoothedTruePositiveP = 0;
	SmoothedFalsePositiveP = 0;
	SmoothedTruePositiveS = 0;
	SmoothedFalsePositiveS = 0;
	deleteCounts = 0;
	FalsePositivePDerivation = new double[numberOfExperts];
	TruePositivePDerivation = new double [numberOfExperts];
	FalsePositiveSDerivation = new double[numberOfExperts];
	TruePositiveSDerivation = new double [numberOfExperts];
	for(int i  = 0; i < numberOfExperts; i++) {
		FalsePositivePDerivation[i] = 0;
		TruePositivePDerivation[i] = 0;
		FalsePositiveSDerivation[i] = 0;
		TruePositiveSDerivation[i] = 0;
	}
	alpha = 20;
	iteration = 0;

	bestIteration = 0;
	bestAER = 1;

	bestWeights = new double[numberOfExperts];

	unconstrainedWeights = new double[numberOfExperts];
	cout << "Init weights: ";
	for(int i = 0; i < numberOfExperts; i++) {
		unconstrainedWeights[i] = weights[i];
		cout << weights[i] << " ";
	}
	cout << endl;


}


LOPOptimizerAER::~LOPOptimizerAER()
{
	delete[] bestWeights;
	delete[] learnRate;
	delete[] unconstrainedWeights;
	delete[] FalsePositivePDerivation;
	delete[] TruePositivePDerivation;
	delete[] FalsePositiveSDerivation;
	delete[] TruePositiveSDerivation;
}

 void LOPOptimizerAER::EvalSentence(Graph * G){
	if(deleteCounts == 1) {
		TruePositiveP = 0;
		FalsePositiveP = 0;
		TruePositiveS = 0;
		FalsePositiveS = 0;
		SLinks = 0;
		PLinks = 0;
		SmoothedTruePositiveP = 0;
		SmoothedFalsePositiveP = 0;
		SmoothedTruePositiveS = 0;
		SmoothedFalsePositiveS = 0;
		deleteCounts = 0;
	}
	//first all features have to calculate their derivations
	for(int i =0; i < FeatureSets.size(); i++) {	
		FeatureSetIterator * iter = FeatureSets[i]->GetIterator();
		for(; !iter->end(); iter->next()) {
			iter->feature()->CalcDerivations();
		}
		delete iter;
	}	

	GraphIterator * iter = G->GetIterator();
	for(;!iter->end();iter->next()) {
		HiddenNode *n = iter->node();
		int gv = n->GetGivenValue();
		pair<double,double> b = n->GetBelief();
		SparseArray derivation = n->GetDerivation(numberOfExperts);
		double s = 1.0/(1+exp(-1*alpha*(b.second - 0.5)));
    //cout <<"b : " << b.second << " s: " << s << endl;
		switch(gv) {
			case 0:
				if(b.first < b.second) {
	//no link in reference link in alignment -> FP
	//cout <<" FP += 1" << endl;
					FalsePositiveP++;
					FalsePositiveS++;
				}
      //and of the smooth version always
      //cout << "SmoothFP +=" << s << endl;
				SmoothedFalsePositiveP += s;
				SmoothedFalsePositiveS += s;	
      //derivation
      //cout << "Derivation+=:" << alpha*s*(1-s)*derivation.first[0] << endl;
				for(int i = 0; i < derivation.size; i++) {
					FalsePositivePDerivation[derivation.index[i]] += alpha*s*(1-s)*(-1)*derivation.data[i];
					FalsePositiveSDerivation[derivation.index[i]] += alpha*s*(1-s)*(-1)*derivation.data[i];
				}
				break;
			case 1:
				SLinks++;
				PLinks++;
				if(b.first < b.second) {
	//both links ->TP
	//cout <<" TP += 1" << endl;
					TruePositiveP++;
					TruePositiveS++;
				}
      //and of the smooth version always
      //cout << "SmoothTP +=" << s << endl;
				SmoothedTruePositiveP += s;
				SmoothedTruePositiveS += s;
      //derivation
      //cout << "Derivation+=:" << alpha*s*(1-s)*derivation.first[0] << endl;
				for(int i = 0; i < derivation.size; i++) {
					TruePositivePDerivation[derivation.index[i]] += alpha*s*(1-s)*(-1)*derivation.data[i];
					TruePositiveSDerivation[derivation.index[i]] += alpha*s*(1-s)*(-1)*derivation.data[i];
				}
				break;
			case -1:
				PLinks++;
				if(b.first < b.second) {
	//cout <<" TPP += 1 FPS += 1" << endl;
	//possbile Link and Alignment ->TPp
					TruePositiveP++;
	//no sure link and alignment-> FPs
					FalsePositiveS++;
				}
      //and of the smooth version always
      //cout << "SmoothTPP/SmoothFPS +=" << s << endl;
				SmoothedTruePositiveP += s;
				SmoothedFalsePositiveS += s;
      //cout << "Derivation+=:" << alpha*s*(1-s)*derivation.first[0] << endl;
				for(int i = 0; i < derivation.size; i++) {
					TruePositivePDerivation[derivation.index[i]] += alpha*s*(1-s)*(-1)*derivation.data[i];
					FalsePositiveSDerivation[derivation.index[i]] += alpha*s*(1-s)*(-1)*derivation.data[i];
				}
      
				break;
     
		}
		if(derivation.size > 0) {
			delete[] derivation.data;
			delete[] derivation.index;
		}
	}
	delete iter;
}

void LOPOptimizerAER::Estimate(){
	//calc softmax derivation
	double * softmaxDerivation = new double[numberOfExperts];
	double sum = 0;
	for(int i = 0; i < numberOfExperts; i++) {
		sum += exp(unconstrainedWeights[i]);
	}
	cout << "Sum: " << sum << endl;
	cout << "Softmax Derivation: ";
	for(int i = 0; i < numberOfExperts; i++) {
		double expValue = exp(unconstrainedWeights[i]);
		softmaxDerivation[i] = (expValue * sum - expValue * expValue) /(sum * sum);
		cout << softmaxDerivation[i] << " ";
	}
	cout << endl;

	double derSum = 0;
	//cout << "Derivations:" << endl;
	for(int i = 0; i < numberOfExperts; i++) {
    //cout << TruePositiveSDerivation[index] << "*(" << SmoothedFalsePositiveS << " + " <<SLinks<< ")";
    //cout << " + " << TruePositivePDerivation[index] << " *( " << SmoothedFalsePositiveP << " + " << SLinks <<" )";
    //cout << " - " << FalsePositiveSDerivation[index] << " * " << SmoothedTruePositiveS;
    //cout << " - " << FalsePositivePDerivation[index] << " * " << SmoothedTruePositiveP << " = ";
    //cout << " / -1*(" << SmoothedTruePositiveS << " + " << SmoothedFalsePositiveS << " + " << SLinks << ")^2" << endl;
		if(learnRate[i] != 0) {
			double der = TruePositiveSDerivation[i] *(SmoothedFalsePositiveS+SLinks);
			der +=TruePositivePDerivation[i]*(SmoothedFalsePositiveP+SLinks);
			der -=FalsePositiveSDerivation[i] *  SmoothedTruePositiveS;
			der -=FalsePositivePDerivation[i] * SmoothedTruePositiveP;
			der /= -1* (SmoothedTruePositiveS+SmoothedFalsePositiveS+SLinks) * (SmoothedTruePositiveS+SmoothedFalsePositiveS+SLinks);
			//cout <<" Derivation[" << i << "][" << j << "](" << index <<") : " << der << endl;
			//weights[i] -= learnRate[i] * der;
			unconstrainedWeights[i] -= learnRate[i]  * der*softmaxDerivation[i];

			//ScalingFactors[i][j] += learnRate[i][j];
			derSum += abs(der);
			//cout << "Learnrate: " << learnRate[i][j] << endl;
			//cout << "ScalingFactor: " << ScalingFactors[i][j] << endl;
		}
	}
	sum = 0;
	for(int i = 0; i < numberOfExperts; i++) {
		sum += exp(unconstrainedWeights[i]);
	}
	cout << "Sum: " << sum << endl;	
	//for(int i = 0; i < numberOfExperts; i++) {
		//weights[i] = exp(unconstrainedWeights[i]) / sum;
		

	//}
	cout << "Iteration : " << iteration << endl;
	cout << (iteration /8) << " " << (iteration % 8) << endl;
	cout << "Unconstrained weights: ";	
	for(int j = 0; j < numberOfExperts;j++) { 
		cout << unconstrainedWeights[j] << "_"; 
	} 
	cout << endl;
	cout << "Weights: ";	
	for(int j = 0; j < numberOfExperts;j++) { 
		cout << weights[j] << "_"; 
	} 

	cout << "Sum of Derivations: " << derSum << endl;
	for(int i = 0; i < numberOfExperts; i++) {
		FalsePositivePDerivation[i] = 0;
		TruePositivePDerivation[i] = 0;
		FalsePositiveSDerivation[i] = 0;
		TruePositiveSDerivation[i] = 0;
	}
	deleteCounts = 1;
}
void LOPOptimizerAER::CheckScalingFactors() {
	double AER = 1.0 - 1.0* (TruePositiveS + TruePositiveP)/(TruePositiveS + FalsePositiveS + SLinks);
	double SmoothedAER = 1.0 - (SmoothedTruePositiveS + SmoothedTruePositiveP)/(SmoothedTruePositiveS + SmoothedFalsePositiveS + SLinks);
	cout << " TruePositive: Sure:" << TruePositiveS << " Possible: " << TruePositiveP << endl;
	cout << " FalsePostive: Sure:" << FalsePositiveS << " Possible: " << FalsePositiveP << endl;
	cout << "Smoothed Version: " << endl;
	cout << " TruePositive: Sure:" << SmoothedTruePositiveS << " Possible: " << SmoothedTruePositiveP << endl;
	cout << " FalsePostive: Sure:" << SmoothedFalsePositiveS << " Possible: " << SmoothedFalsePositiveP << endl;
	cout << "Links Sure:" << SLinks << " Possible: " << PLinks << endl;
	cout << "AER:" << AER << " SmoothedAER:" << SmoothedAER << endl;
  //cout << "Derivations:TPS: " << TruePositiveSDerivation[0] << "TPP" << TruePositivePDerivation[0] << endl;
  //cout << "Derivations:FPS: " << FalsePositiveSDerivation[0] << "FPP" << FalsePositivePDerivation[0] << endl;

	if(AER <= bestAER) {
		bestAER = AER;
		for(int i = 0; i < numberOfExperts; i++) {
			bestWeights[i] = weights[i];
		}
		bestIteration = iteration;

	}
	iteration++;
	deleteCounts = 1;

}
void LOPOptimizerAER::OutputBest(){
	cout << "Best AER: " << bestAER << endl;
	cout << "In iteration: " << bestIteration << endl;
	cout << "ExpertWeights: ";
	for(int i = 0; i < numberOfExperts;i++) {
		cout << bestWeights[i] << "_";
		weights[i] = bestWeights[i];
	}
	cout << endl;
}



