
#include "lopoptimizerall.h"


#include "lopoptimizeraer.h"

LOPOptimizerAll::LOPOptimizerAll(vector <FeatureSet *> FSets,double* w, char * learnRateString, int numberOfExp)
 : Optimizer(FSets)
{
	FeatureSets = FSets;
	weights = w;
	numberOfExperts = numberOfExp;
	if(learnRateString == NULL) {
		cout << " For optimization the learnRate is needed" << endl;
		exit(-1);
	}
	numberOfParts = atoi(learnRateString);

	FeatureSets = FSets;

	TruePositiveP = 0;
	FalsePositiveP = 0;
	TruePositiveS = 0;
	FalsePositiveS = 0;
	SLinks = 0;
	PLinks = 0;

	iteration = 0;

	bestIteration = 0;
	bestAER = 1;

	bestWeights = new double[numberOfExperts];

	parts = new int[numberOfParts];
	for(int i = 0; i < numberOfParts; i++) {
		parts[i] = 0;
	}
	CalcWeights();

}


LOPOptimizerAll::~LOPOptimizerAll()
{
	delete[] bestWeights;
	delete[] parts;
}

 void LOPOptimizerAll::EvalSentence(Graph * G){
	//cout << "Eval Sentence: ";
	if(deleteCounts == 1) {
		TruePositiveP = 0;
		FalsePositiveP = 0;
		TruePositiveS = 0;
		FalsePositiveS = 0;
		SLinks = 0;
		PLinks = 0;
		deleteCounts = 0;
	}

	GraphIterator * iter = G->GetIterator();
	for(;!iter->end();iter->next()) {
		HiddenNode *n = iter->node();
		int gv = n->GetGivenValue();
		pair<double,double> b = n->GetBelief();
		switch(gv) {
			case 0:
				if(b.first < b.second) {
					FalsePositiveP++;
					FalsePositiveS++;
				}
				break;
			case 1:
				SLinks++;
				PLinks++;
				if(b.first < b.second) {
					TruePositiveP++;
					TruePositiveS++;
				}
				break;
			case -1:
				PLinks++;
				if(b.first < b.second) {
					TruePositiveP++;
					FalsePositiveS++;
				}
				break;
     
		}
	}
	delete iter;
	//cout << " SLInks: " << SLinks << endl;
}

void LOPOptimizerAll::Estimate(){
	cout << "Iteration : " << iteration << endl;

	int changed = 0;
	for(int l = numberOfParts - 1; (l >= 0) &&(changed == 0); l--) {
		if(parts[l] < numberOfExperts - 1) {
			parts[l]++;
			for(int m = l+1; m < numberOfParts; m++) {
				parts[m] = parts[l];
			}
			changed = 1;
		}

	}
	cout << "New Parts: ";
	for(int i = 0; i < numberOfParts; i++) {
		cout << parts[i] << " ";
	}
	cout << endl;
	CalcWeights();	
}
void LOPOptimizerAll::CalcWeights() {

	for(int i = 0; i < numberOfExperts; i++) {
		weights[i] = 0;
	}

	for(int i = 0; i < numberOfParts; i++) {
		weights[parts[i]] += 1.0/numberOfParts;
	}
	cout << "Weights: ";
	for(int i = 0; i < numberOfExperts; i++) {
		cout << weights[i] << " ";
	}
	cout <<  endl;
}
void LOPOptimizerAll::CheckScalingFactors() {
	double AER = 1.0 - 1.0* (TruePositiveS + TruePositiveP)/(TruePositiveS + FalsePositiveS + SLinks);
	cout << " TruePositive: Sure:" << TruePositiveS << " Possible: " << TruePositiveP << endl;
	cout << " FalsePostive: Sure:" << FalsePositiveS << " Possible: " << FalsePositiveP << endl;
	cout << "Links Sure:" << SLinks << " Possible: " << PLinks << endl;
	cout << "AER:" << AER << endl;

	if(AER <= bestAER) {
		bestAER = AER;
		for(int i = 0; i < numberOfExperts; i++) {
			bestWeights[i] = weights[i];
		}
		bestIteration = iteration;

	}
	iteration++;
	deleteCounts = 1;

}
void LOPOptimizerAll::OutputBest(){
	cout << "Best AER: " << bestAER << endl;
	cout << "In iteration: " << bestIteration << endl;
	cout << "ExpertWeights: ";
	for(int i = 0; i < numberOfExperts;i++) {
		cout << bestWeights[i] << "_";
		weights[i] = bestWeights[i];
	}
	cout << endl;
}


