
#include "lopoptimizerml.h"

LOPOptimizerML::LOPOptimizerML(vector <FeatureSet *> FSets,double* w, char * learnRateString, int numberOfExp):Optimizer(FSets){
	FeatureSets = FSets;
	weights = w;
	numberOfExperts = numberOfExp;
	if(learnRateString == NULL) {
		cout << " For optimization the learnRate is needed" << endl;
		exit(-1);
	}
	learnRate = new double[numberOfExperts];
	SetLearnRate(learnRateString,learnRate,numberOfExperts);
	
    
	bestCorrectLink = 0;
	bestCorrectNoLink = 0;
	bestFailedLink = INT_MAX;
	bestFailedNoLink = INT_MAX;
	bestLogProbCorrect = -1*DBL_MAX;
	bestLogProbBest = -1*DBL_MIN;
	bestIteration = 0;

	Sent = 0;
	logProbCorrect = 0;
	logProbBest = 0;
	CorrectLink = 0;
	FailedLink = 0;
	CorrectNoLink = 0;
	FailedNoLink = 0;
	iteration = 0;

	//init Estimated Derivations
	EstimatedDerivations = new double[numberOfExperts];
	oldEstimatedDerivations = new double [numberOfExperts];
	lRate = 1;
	for(int i = 0; i < numberOfExperts; i++) {
		EstimatedDerivations[i] = 0;
	}

	bestWeights = new double[numberOfExperts];

	unconstrainedWeights = new double[numberOfExperts];
	cout << "Init weights: ";
	for(int i = 0; i < numberOfExperts; i++) {
		unconstrainedWeights[i] = weights[i];
		cout << weights[i] << " ";
	}
	cout << endl;
}


LOPOptimizerML::~LOPOptimizerML()
{
	delete[] learnRate;
	delete[] EstimatedDerivations;
	delete[] oldEstimatedDerivations;
	delete[] bestWeights;
	delete[] unconstrainedWeights;
}


void LOPOptimizerML::EvalSentence(Graph * G){
	Sent++;
//	cout << "Sentence: " << Sent << endl;
	logProbCorrect += G->ComputeLogProbabilityCorrect();
	logProbBest += G->ComputeLogProbabilityBest();
	G->ComputeCorrectLinks(CorrectLink,FailedLink,CorrectNoLink,FailedNoLink);
	for(int i = 0; i < FeatureSets.size();i++) {
	      //store new Scaling Factors
		vector<double> EstDer = FeatureSets[i]->GetEstimatedDerivations();
		assert(EstDer.size() == numberOfExperts);
//		cout << "EstDer: ";
		for(int j = 0; j < EstDer.size(); j++) {
			EstimatedDerivations[j] += EstDer[j];
//			cout << EstDer[j] << " ";
		}
//		cout << endl;
	}

}
void LOPOptimizerML::Estimate(){
	//calc softmax derivation
	double * softmaxDerivation = new double[numberOfExperts];
	double sum = 0;
	for(int i = 0; i < numberOfExperts; i++) {
		sum += exp(unconstrainedWeights[i]);
	}
	cout << "Sum: " << sum << endl;
	cout << "Softmax Derivation: ";
	for(int i = 0; i < numberOfExperts; i++) {
		double expValue = exp(unconstrainedWeights[i]);
		softmaxDerivation[i] = (expValue * sum - expValue * expValue) /(sum * sum);
		cout << softmaxDerivation[i] << " ";
	}
	cout << endl;
	double derSize = 0;
	int improvedDerivations = 0;
	for(int i = 0; i < numberOfExperts;i++) { 
		if(learnRate[i] != 0) { 
			EstimatedDerivations[i] /= Sent;

			//weights[i] += learnRate[i] * EstimatedDerivations[i]; 

			unconstrainedWeights[i] += learnRate[i]  * EstimatedDerivations[i]*softmaxDerivation[i];
			//cout << "unconstrainedWeight: " << learnRate[i] << " * " << EstimatedDerivations[i] << " * " << softmaxDerivation[i] << endl;

			derSize += abs(EstimatedDerivations[i]);
			if(abs(oldEstimatedDerivations[i]) > abs(EstimatedDerivations[i])) {
				improvedDerivations ++;
			}
			oldEstimatedDerivations[i] = EstimatedDerivations[i];
		} 
	} 

	sum = 0;
	for(int i = 0; i < numberOfExperts; i++) {
		sum += exp(unconstrainedWeights[i]);
	}
	cout << "Sum: " << sum << endl;	
	for(int i = 0; i < numberOfExperts; i++) {
		weights[i] = exp(unconstrainedWeights[i]) / sum;
	}
	cout << "Unconstrained weights: ";	
	for(int j = 0; j < numberOfExperts;j++) { 
		cout << unconstrainedWeights[j] << "_"; 
	} 
	cout << endl;
	cout << "Weights: ";	
	for(int j = 0; j < numberOfExperts;j++) { 
		cout << weights[j] << "_"; 
	} 
		cout << endl; 
	cout << "Derivationsize: " << derSize << " ImprovedDerivations: " << improvedDerivations << endl;

	//reset for new iteration
	Sent = 0;

    //reset Estimated Derivations
	for(int i =0; i < numberOfExperts; i++) {
		EstimatedDerivations[i] = 0;
	}
	iteration ++;

	delete[] softmaxDerivation;
}
void LOPOptimizerML::CheckScalingFactors() {
	logProbCorrect /= Sent;
	logProbBest /= Sent;
	cout << "LogProbCorrect:" << logProbCorrect << " LogProbBest:" << logProbBest << endl;
	cout << "Links: Correct: " << CorrectLink << " Failed: " << FailedLink << endl;
	cout << "NoLinks: Correct: " << CorrectNoLink << " Failed: " << FailedNoLink << endl;
	cout << "All Failed: " << FailedLink + FailedNoLink << endl;
       //if(CorrectLink + CorrectNoLink > bestCorrectLink + bestCorrectNoLink) {
	if(logProbCorrect > bestLogProbCorrect) {
	//save best Scaling Factors
		bestCorrectLink = CorrectLink;
		bestCorrectNoLink = CorrectNoLink;
		bestFailedLink = FailedLink;
		bestFailedNoLink = FailedNoLink;
		bestLogProbCorrect = logProbCorrect;
		bestLogProbBest = logProbBest;
		for(int i = 0; i < numberOfExperts; i++) {
			bestWeights[i] = weights[i];
		}
		bestIteration = iteration;
	}

	//reset for new iteration
	logProbCorrect = 0;
	logProbBest = 0;
	CorrectLink = 0;
	FailedLink = 0;
	CorrectNoLink = 0;
	FailedNoLink = 0;

}

void LOPOptimizerML::OutputBest(){

	cout << "Best Result at Iteration "<< bestIteration <<": " << endl;
	cout << "LogProbCorrect:" << bestLogProbCorrect << " LogProbBest:" << bestLogProbBest << endl;
	cout << "Links: Correct: " << bestCorrectLink << " Failed: " << bestFailedLink << endl;
	cout << "NoLinks: Correct: " << bestCorrectNoLink << " Failed: " << bestFailedNoLink << endl;
    
	cout << "ExpertWeights: ";
	for(int i = 0; i < numberOfExperts;i++) {
		cout << bestWeights[i] << "_";
		weights[i] = bestWeights[i];
	}
	cout << endl;

}
