#include "optimizerfscore.h"
OptimizerFScore::OptimizerFScore(vector <FeatureSet *> FSets,vector < vector<double> > SFactors,char * learnRateFile,double a):Optimizer(FSets){
	alpha = a;
	numberOfAllFeatures = 0;
	FeatureSets = FSets;
	ScalingFactors = SFactors;
	for(int i = 0; i < ScalingFactors.size();i++) {
 		numberOfAllFeatures += ScalingFactors[i].size();
	}
	TruePositiveP = 0;
	FalsePositiveP = 0;
	TruePositiveS = 0;
	FalsePositiveS = 0;
	SLinks = 0;
	PLinks = 0;
	TruePositivePCV = 0;
	FalsePositivePCV = 0;
	TruePositiveSCV = 0;
	FalsePositiveSCV = 0;
	SLinksCV = 0;
	PLinksCV = 0;
	SmoothedTruePositiveP = 0;
	SmoothedFalsePositiveP = 0;
	SmoothedTruePositiveS = 0;
	SmoothedFalsePositiveS = 0;
	deleteCounts = 0;
	FalsePositivePDerivation.resize(numberOfAllFeatures,0);
	TruePositivePDerivation.resize(numberOfAllFeatures,0);
	FalsePositiveSDerivation.resize(numberOfAllFeatures,0);
	TruePositiveSDerivation.resize(numberOfAllFeatures,0);
	alphaSigmuid = 20;
	iteration = 0;
	if(learnRateFile == NULL) {
		cout << " For optimization a learnRate file is needed" << endl;
		exit(-1);
	}
	for(int i = 0; i < ScalingFactors.size(); i++) {
		learnRate.push_back(vector<double>(ScalingFactors[i].size(),0));
      //mean.push_back(vector<double>(ScalingFactors[i].size(),0));
      //sigma.push_back(vector<double>(ScalingFactors[i].size(),10000));
	}
	LoadFactorFile(learnRateFile,learnRate);

	bestIteration = 0;
	bestFScore = 0;
	cout << "Optimize towards FScore with alpha-value: " << alpha << endl;
}

 void OptimizerFScore::EvalSentence(Graph * G){
	if(deleteCounts == 1) {
		TruePositiveP = 0;
		FalsePositiveP = 0;
		TruePositiveS = 0;
		FalsePositiveS = 0;
		SLinks = 0;
		PLinks = 0;
		TruePositivePCV = 0;
		FalsePositivePCV = 0;
		TruePositiveSCV = 0;	
		FalsePositiveSCV = 0;
		SLinksCV = 0;
		PLinksCV = 0;
		SmoothedTruePositiveP = 0;
		SmoothedFalsePositiveP = 0;
		SmoothedTruePositiveS = 0;
		SmoothedFalsePositiveS = 0;
		deleteCounts = 0;
	}
	//first all features have to calculate their derivations
	for(int i =0; i < FeatureSets.size(); i++) {	
		FeatureSetIterator * iter = FeatureSets[i]->GetIterator();
		for(; !iter->end(); iter->next()) {
			iter->feature()->CalcDerivations();
		}
		delete iter;
	}	

	GraphIterator * iter = G->GetIterator();
	for(;!iter->end();iter->next()) {
		HiddenNode *n = iter->node();
		int gv = n->GetGivenValue();
		pair<double,double> b = n->GetBelief();
		SparseArray derivation = n->GetDerivation(numberOfAllFeatures);
		double s = 1.0/(1+exp(-1*alphaSigmuid*(b.second - 0.5)));
    //cout <<"b : " << b.second << " s: " << s << endl;
		switch(gv) {
			case 0:
				if(b.first < b.second) {
	//no link in reference link in alignment -> FP
	//cout <<" FP += 1" << endl;
					FalsePositiveP++;
					FalsePositiveS++;
				}
      //and of the smooth version always
      //cout << "SmoothFP +=" << s << endl;
				SmoothedFalsePositiveP += s;
				SmoothedFalsePositiveS += s;	
      //derivation
      //cout << "Derivation+=:" << alphaSigmuid*s*(1-s)*derivation.first[0] << endl;
				for(int i = 0; i < derivation.size; i++) {
					FalsePositivePDerivation[derivation.index[i]] += alphaSigmuid*s*(1-s)*(-1)*derivation.data[i];
					FalsePositiveSDerivation[derivation.index[i]] += alphaSigmuid*s*(1-s)*(-1)*derivation.data[i];
				}
				break;
			case 1:
				SLinks++;
				PLinks++;
				if(b.first < b.second) {
	//both links ->TP
	//cout <<" TP += 1" << endl;
					TruePositiveP++;
					TruePositiveS++;
				}
      //and of the smooth version always
      //cout << "SmoothTP +=" << s << endl;
				SmoothedTruePositiveP += s;
				SmoothedTruePositiveS += s;
      //derivation
      //cout << "Derivation+=:" << alphaSigmuid*s*(1-s)*derivation.first[0] << endl;
				for(int i = 0; i < derivation.size; i++) {
					TruePositivePDerivation[derivation.index[i]] += alphaSigmuid*s*(1-s)*(-1)*derivation.data[i];
					TruePositiveSDerivation[derivation.index[i]] += alphaSigmuid*s*(1-s)*(-1)*derivation.data[i];
				}
				break;
			case -1:
				PLinks++;
				if(b.first < b.second) {
	//cout <<" TPP += 1 FPS += 1" << endl;
	//possbile Link and Alignment ->TPp
					TruePositiveP++;
	//no sure link and alignment-> FPs
					FalsePositiveS++;
				}
      //and of the smooth version always
      //cout << "SmoothTPP/SmoothFPS +=" << s << endl;
				SmoothedTruePositiveP += s;
				SmoothedFalsePositiveS += s;
      //cout << "Derivation+=:" << alphaSigmuid*s*(1-s)*derivation.first[0] << endl;
				for(int i = 0; i < derivation.size; i++) {
					TruePositivePDerivation[derivation.index[i]] += alphaSigmuid*s*(1-s)*(-1)*derivation.data[i];
					FalsePositiveSDerivation[derivation.index[i]] += alphaSigmuid*s*(1-s)*(-1)*derivation.data[i];
				}
      
				break;
     
		}
		if(derivation.size > 0) {
			delete[] derivation.data;
			delete[] derivation.index;
		}
	}
	delete iter;
}

void OptimizerFScore::Estimate(){
	int index = 0;
	double derSum = 0;
	cout << "New Scaling Factors: " << endl;

	double denominator = alpha *(SmoothedTruePositiveS+SmoothedFalsePositiveS) *SmoothedTruePositiveS;
	denominator += (1 - alpha) * SmoothedTruePositiveP * SLinks;
	//cout << "Denominator:" << denominator << endl;
	for(int i = 0; i < ScalingFactors.size(); i++) {
		for(int j = 0; j < ScalingFactors[i].size();j++) {
			if(learnRate[i][j] != 0) {
				//cout << "TruePositivePDerivation:" << TruePositivePDerivation[index] << " ";
				//cout << "TruePositiveSDerivation:" << TruePositiveSDerivation[index] << " ";
				//cout << "FalsePositiveSDerivation: " << FalsePositiveSDerivation[index] << endl;
				double der1 = TruePositivePDerivation[index] *(SmoothedTruePositiveS);
				der1 +=TruePositiveSDerivation[index]*(SmoothedTruePositiveP);
				double der2 = alpha * (TruePositiveSDerivation[index] + FalsePositiveSDerivation[index])* SmoothedTruePositiveS;
				//cout << "Der2.1: " << der2;
				der2 +=alpha * (SmoothedTruePositiveS + SmoothedFalsePositiveS)* TruePositiveSDerivation[index];
				//cout << " Der2.2: " << der2 << endl;
				der2 += (1-alpha) * TruePositivePDerivation[index]*SLinks;
				double der = der1 * denominator - der2* (SmoothedTruePositiveP * SmoothedTruePositiveS);
				//cout << "Der.1: " << der << endl;
				der /= denominator * denominator;
				//cout << "Der1: " << der1 << " Der2:" << der2 << " Der: " << der << endl;
				//cout <<" Derivation[" << i << "][" << j << "](" << index <<") : " << der << endl;
				ScalingFactors[i][j] += learnRate[i][j] * der;
				//ScalingFactors[i][j] += learnRate[i][j];
				derSum += abs(der);
				if(j < 10)
					cout << ScalingFactors[i][j] << "_";
				//cout << "Learnrate: " << learnRate[i][j] << endl;
				//cout << "ScalingFactor: " << ScalingFactors[i][j] << endl;
			}
			index++;
		}
		cout << endl;
		FeatureSets[i]->SetScalingFactors(ScalingFactors[i]);
	}
	cout << endl;
	cout << "Sum of Derivations: " << derSum << endl;
 FalsePositivePDerivation.assign(numberOfAllFeatures,0);
 TruePositivePDerivation.assign(numberOfAllFeatures,0);
 FalsePositiveSDerivation.assign(numberOfAllFeatures,0);
 TruePositiveSDerivation.assign(numberOfAllFeatures,0);
	deleteCounts = 1;
}
void OptimizerFScore::EvalCVSentence(Graph * G){
	GraphIterator * iter = G->GetIterator();
	for(;!iter->end();iter->next()) {
		HiddenNode *n = iter->node();
		int gv = n->GetGivenValue();
		pair<double,double> b = n->GetBelief();
		//cout <<"b : " << b.second << " s: " << s << endl;
		switch(gv) {
			case 0:
				if(b.first < b.second) {
				//no link in reference link in alignment -> FP
				//cout <<" FP += 1" << endl;
					FalsePositivePCV++;
					FalsePositiveSCV++;
				}
				break;
			case 1:
				SLinksCV++;
				PLinksCV++;
				if(b.first < b.second) {
					//both links ->TP
					//cout <<" TP += 1" << endl;
					TruePositivePCV++;
					TruePositiveSCV++;
				}
				break;
			case -1:
				PLinksCV++;
				if(b.first < b.second) {
					//cout <<" TPP += 1 FPS += 1" << endl;
					//possbile Link and Alignment ->TPp
					TruePositivePCV++;
					//no sure link and alignment-> FPs
					FalsePositiveSCV++;
				}
				break;
     
		}
	}
	delete iter;
}
void OptimizerFScore::CheckScalingFactors() {
	double FScore = (TruePositiveS * TruePositiveP)/(alpha * (TruePositiveS + FalsePositiveS) * TruePositiveS + (1 - alpha) * TruePositiveP * SLinks);
	double SmoothedFScore = (SmoothedTruePositiveS * SmoothedTruePositiveP)/(alpha * (SmoothedTruePositiveS + SmoothedFalsePositiveS) * SmoothedTruePositiveS + (1- alpha) * SmoothedTruePositiveP *  SLinks);
	cout << " TruePositive: Sure:" << TruePositiveS << " Possible: " << TruePositiveP << endl;
	cout << " FalsePostive: Sure:" << FalsePositiveS << " Possible: " << FalsePositiveP << endl;
	cout << "Smoothed Version: " << endl;
	cout << " TruePositive: Sure:" << SmoothedTruePositiveS << " Possible: " << SmoothedTruePositiveP << endl;
	cout << " FalsePostive: Sure:" << SmoothedFalsePositiveS << " Possible: " << SmoothedFalsePositiveP << endl;
	cout << "Links Sure:" << SLinks << " Possible: " << PLinks << endl;
	cout << "FScore:" << FScore << " SmoothedFScore:" << SmoothedFScore << endl;
  //cout << "Derivations:TPS: " << TruePositiveSDerivation[0] << "TPP" << TruePositivePDerivation[0] << endl;
  //cout << "Derivations:FPS: " << FalsePositiveSDerivation[0] << "FPP" << FalsePositivePDerivation[0] << endl;

	if(FScore >= bestFScore) {
		bestFScore = FScore;
		bestScalingFactors = ScalingFactors;
		bestIteration = iteration;

	}
	iteration++;
	deleteCounts = 1;

}

void OptimizerFScore::CheckScalingFactorsWithCV() {
	double FScore = (TruePositiveS * TruePositiveP)/(alpha * (TruePositiveS + FalsePositiveS) * TruePositiveS + (1 - alpha) * TruePositiveP * SLinks);
	double SmoothedFScore =(SmoothedTruePositiveS * SmoothedTruePositiveP)/(alpha * (SmoothedTruePositiveS + SmoothedFalsePositiveS) * SmoothedTruePositiveS + (1- alpha) * SmoothedTruePositiveP *  SLinks);
	double FScoreCV = (TruePositiveSCV * TruePositivePCV)/(alpha * (TruePositiveSCV + FalsePositiveSCV) * TruePositiveSCV + (1 - alpha) * TruePositivePCV * SLinks);
	cout << "Dev: " << endl;
	cout << " TruePositive: Sure:" << TruePositiveS << " Possible: " << TruePositiveP << endl;
	cout << " FalsePostive: Sure:" << FalsePositiveS << " Possible: " << FalsePositiveP << endl;
	cout << "Smoothed Version: " << endl;
	cout << " TruePositive: Sure:" << SmoothedTruePositiveS << " Possible: " << SmoothedTruePositiveP << endl;
	cout << " FalsePostive: Sure:" << SmoothedFalsePositiveS << " Possible: " << SmoothedFalsePositiveP << endl;
	cout << "Links Sure:" << SLinks << " Possible: " << PLinks << endl;
	cout << "FScore:" << FScore << " SmoothedFScore:" << SmoothedFScore << endl;
	cout << "Cross Validation: " << endl;
	cout << " TruePositive: Sure:" << TruePositiveSCV << " Possible: " << TruePositivePCV << endl;
	cout << " FalsePostive: Sure:" << FalsePositiveSCV << " Possible: " << FalsePositivePCV << endl;
	cout << "Links Sure:" << SLinksCV << " Possible: " << PLinksCV << endl;
	cout << "FScore:" << FScoreCV << endl;

	if(FScoreCV >= bestFScore) {
		bestFScore = FScoreCV;
		bestScalingFactors = ScalingFactors;
		bestIteration = iteration;

	}
	iteration++;
	deleteCounts = 1;
}

void OptimizerFScore::OutputBest(){
	cout << "Best FScore: " << bestFScore << endl;
	cout << "In iteration: " << bestIteration << endl;

	for(int i = 0; i < FeatureSets.size(); i ++) {
		FeatureSets[i]->SetScalingFactors(bestScalingFactors[i]);
		cout << "Output Scaling Factors " << i << endl;
		FeatureSets[i]->OutputScalingFactors();
	}

}



