
#include "optimizermap.h"

OptimizerMAP::OptimizerMAP(vector <FeatureSet *> FSets,vector < vector<double> > SFactors,char * learnRateFile, char * sigmaFile):Optimizer(FSets){
	FeatureSets = FSets;
	ScalingFactors = SFactors;
	if(learnRateFile == NULL) {
		cout << " For optimization a learnRate file is needed" << endl;
		exit(-1);
	}

	for(int i = 0; i < ScalingFactors.size(); i++) {
		learnRate.push_back(vector<double>(ScalingFactors[i].size(),0));
		sigma.push_back(vector<double>(ScalingFactors[i].size(),1));
	}
	LoadFactorFile(learnRateFile,learnRate);
	LoadFactorFile(sigmaFile,sigma);
	
    
	bestCorrectLink = 0;
	bestCorrectNoLink = 0;
	bestFailedLink = INT_MAX;
	bestFailedNoLink = INT_MAX;
	bestLogProbCorrect = -1*DBL_MAX;
	bestLogProbBest = -1*DBL_MIN;
	bestIteration = 0;

	Sent = 0;
	logProbCorrect = 0;
	logProbBest = 0;
	CorrectLink = 0;
	FailedLink = 0;
	CorrectNoLink = 0;
	FailedNoLink = 0;
	iteration = 0;
	logProbCorrectCV = 0;
	logProbBestCV = 0;
	CorrectLinkCV = 0;
	FailedLinkCV = 0;
	CorrectNoLinkCV = 0;
	FailedNoLinkCV = 0;

	//init Estimated Derivations
	EstimatedDerivations.resize(ScalingFactors.size());
	for(int i =0; i < EstimatedDerivations.size(); i++) {
		EstimatedDerivations[i].resize(ScalingFactors[i].size(),0);
	}

}


OptimizerMAP::~OptimizerMAP()
{
}


void OptimizerMAP::EvalSentence(Graph * G){
	Sent++;
	logProbCorrect += G->ComputeLogProbabilityCorrect();
	logProbBest += G->ComputeLogProbabilityBest();
	G->ComputeCorrectLinks(CorrectLink,FailedLink,CorrectNoLink,FailedNoLink);
	for(int i = 0; i < FeatureSets.size();i++) {
	      //store new Scaling Factors
		vector<double> EstDer = FeatureSets[i]->GetEstimatedDerivations();
		assert(EstDer.size() <= EstimatedDerivations[i].size());
		for(int j = 0; j < EstDer.size(); j++) {
			EstimatedDerivations[i][j] += EstDer[j];
		}
	}


}
void OptimizerMAP::Estimate(){
	for(int i = 0; i < EstimatedDerivations.size();i++) {
		//cout << "New Opt: ScalingFactors for Featureset " << i << endl;
		for(int j = 0; j < EstimatedDerivations[i].size();j++) {
			if(learnRate[i][j] != 0) {
				//EstimatedDerivations[i][j] /= Sent;
				if(j < 20) {
					cout << "Derivation: " << EstimatedDerivations[i][j];
					cout << "ScalingFactor Derivation: " << ScalingFactors[i][j] / sigma[i][j] << endl;
				}
				EstimatedDerivations[i][j] -= ScalingFactors[i][j] / sigma[i][j];
				ScalingFactors[i][j] += learnRate[i][j] * EstimatedDerivations[i][j];
				if(j < 20) {
					cout << "New Derivation: " << EstimatedDerivations[i][j] << endl;
					cout << "New ScalingFactor: " << ScalingFactors[i][j] << endl;
				}
			}
		}
		for(int j = 0; (j < ScalingFactors[i].size()) && (j < 12);j++) {
			cout << ScalingFactors[i][j] << "_";
		}
		cout << endl;
		FeatureSets[i]->SetScalingFactors(ScalingFactors[i]);
	}


	
	//reset for new iteration
	Sent = 0;

    //reset Estimated Derivations
	for(int i =0; i < EstimatedDerivations.size(); i++) {
      //cout << "ScalingFactors[i].size(): " << ScalingFactors[i].size() << endl;
		EstimatedDerivations[i].assign(ScalingFactors[i].size(),0);
	}
	iteration ++;
}
void OptimizerMAP::EvalCVSentence(Graph * G){
	Sent++;
	logProbCorrectCV += G->ComputeLogProbabilityCorrect();
	logProbBestCV += G->ComputeLogProbabilityBest();
	G->ComputeCorrectLinks(CorrectLinkCV,FailedLinkCV,CorrectNoLinkCV,FailedNoLinkCV);
}
void OptimizerMAP::CheckScalingFactors() {
      
	//logProbCorrect /= Sent;
	//logProbBest /= Sent;
	double MAP = 0;
	int n = 0;
	for(int i = 0; i < ScalingFactors.size();i++) {
		for(int j = 0; j < ScalingFactors[i].size();j++) {
			MAP += ScalingFactors[i][j] * ScalingFactors[i][j] /(2* sigma[i][j]);
			n++;
		}
	}
	//MAP /= n;
	cout << "LogProbCorrect:" << logProbCorrect << " LogProbBest:" << logProbBest << endl;
	cout << "MAP Part: " << MAP << " -> Optimize function value:" << logProbCorrect - MAP<< endl;
	cout << "Links: Correct: " << CorrectLink << " Failed: " << FailedLink << endl;
	cout << "NoLinks: Correct: " << CorrectNoLink << " Failed: " << FailedNoLink << endl;
	cout << "All Failed: " << FailedLink + FailedNoLink << endl;

	logProbCorrect -= MAP;
      
      //if(CorrectLink + CorrectNoLink > bestCorrectLink + bestCorrectNoLink) {
	//if(logProbCorrect > bestLogProbCorrect) {
	//save best Scaling Factors
		bestCorrectLink = CorrectLink;
		bestCorrectNoLink = CorrectNoLink;
		bestFailedLink = FailedLink;
		bestFailedNoLink = FailedNoLink;
		bestLogProbCorrect = logProbCorrect;
		bestLogProbBest = logProbBest;
		bestScalingFactors = ScalingFactors;
		bestIteration = iteration;
	//}
    
	//reset for new iteration
	logProbCorrect = 0;
	logProbBest = 0;
	CorrectLink = 0;
	FailedLink = 0;
	CorrectNoLink = 0;
	FailedNoLink = 0;

}

void OptimizerMAP::CheckScalingFactorsWithCV() {
	//logProbCorrect /= Sent;
	//logProbBest /= Sent;
	double MAP = 0;
	int n = 0;
	for(int i = 0; i < ScalingFactors.size();i++) {
		for(int j = 0; j < ScalingFactors[i].size();j++) {
			MAP += ScalingFactors[i][j] * ScalingFactors[i][j] /(2* sigma[i][j]);
			n++;
		}
	}
	cout << "Dev Set: " << endl;
	cout << "LogProbCorrect:" << logProbCorrect << " LogProbBest:" << logProbBest << endl;
	cout << "MAP Part: " << MAP << " -> Optimize function value:" << logProbCorrect - MAP<< endl;
	cout << "Links: Correct: " << CorrectLink << " Failed: " << FailedLink << endl;
	cout << "NoLinks: Correct: " << CorrectNoLink << " Failed: " << FailedNoLink << endl;
	//logProbCorrectCV /= Sent;
	//logProbBestCV /= Sent;
	cout << "Crossvalidation Set:" << endl;
	cout << "CV:LogProbCorrect:" << logProbCorrectCV << " LogProbBest:" << logProbBestCV << endl;
	cout << "CV:MAP Part: " << MAP << " -> Optimize function value:" << logProbCorrectCV - MAP<< endl;
	cout << "CV:Links: Correct: " << CorrectLinkCV << " Failed: " << FailedLinkCV << endl;
	cout << "CV:NoLinks: Correct: " << CorrectNoLinkCV << " Failed: " << FailedNoLinkCV << endl;
	logProbCorrect -= MAP;
	logProbCorrectCV -= MAP;

      //if(CorrectLink + CorrectNoLink > bestCorrectLink + bestCorrectNoLink) {
	if(logProbCorrectCV > bestLogProbCorrect) {
	//save best Scaling Factors
		bestCorrectLink = CorrectLinkCV;
		bestCorrectNoLink = CorrectNoLinkCV;
		bestFailedLink = FailedLinkCV;
		bestFailedNoLink = FailedNoLinkCV;
		bestLogProbCorrect = logProbCorrectCV;
		bestLogProbBest = logProbBestCV;
		bestScalingFactors = ScalingFactors;
		bestIteration = iteration;
	}

	//reset for new iteration
	logProbCorrect = 0;
	logProbBest = 0;
	CorrectLink = 0;
	FailedLink = 0;
	CorrectNoLink = 0;
	FailedNoLink = 0;
	logProbCorrectCV = 0;
	logProbBestCV = 0;
	CorrectLinkCV = 0;
	FailedLinkCV = 0;
	CorrectNoLinkCV = 0;
	FailedNoLinkCV = 0;

}
void OptimizerMAP::OutputBest(){

	cout << "Best Result at Iteration "<< bestIteration <<": " << endl;
	cout << "LogProbCorrect:" << bestLogProbCorrect << " LogProbBest:" << bestLogProbBest << endl;
	cout << "Links: Correct: " << bestCorrectLink << " Failed: " << bestFailedLink << endl;
	cout << "NoLinks: Correct: " << bestCorrectNoLink << " Failed: " << bestFailedNoLink << endl;
    
	for(int i = 0; i < bestScalingFactors.size();i++) {
		FeatureSets[i]->SetScalingFactors(bestScalingFactors[i]);
		FeatureSets[i]->OutputScalingFactors();
	}

}

