
#include "optimizerml.h"

OptimizerML::OptimizerML(vector <FeatureSet *> FSets,vector < vector<double> > SFactors,char * learnRateFile):Optimizer(FSets){
	FeatureSets = FSets;
	ScalingFactors = SFactors;
	if(learnRateFile == NULL) {
		cout << " For optimization a learnRate file is needed" << endl;
		exit(-1);
	}

	for(int i = 0; i < ScalingFactors.size(); i++) {
		learnRate.push_back(vector<double>(ScalingFactors[i].size(),0));
	}
	LoadFactorFile(learnRateFile,learnRate);
	
    
	bestCorrectLink = 0;
	bestCorrectNoLink = 0;
	bestFailedLink = INT_MAX;
	bestFailedNoLink = INT_MAX;
	bestLogProbCorrect = -1*DBL_MAX;
	bestLogProbBest = -1*DBL_MIN;
	bestIteration = 0;

	Sent = 0;
	logProbCorrect = 0;
	logProbBest = 0;
	CorrectLink = 0;
	FailedLink = 0;
	CorrectNoLink = 0;
	FailedNoLink = 0;
	iteration = 0;
	logProbCorrectCV = 0;
	logProbBestCV = 0;
	CorrectLinkCV = 0;
	FailedLinkCV = 0;
	CorrectNoLinkCV = 0;
	FailedNoLinkCV = 0;

	//init Estimated Derivations
	EstimatedDerivations.resize(ScalingFactors.size());
	oldEstimatedDerivations.resize(ScalingFactors.size());
	for(int i =0; i < EstimatedDerivations.size(); i++) {
		EstimatedDerivations[i].resize(ScalingFactors[i].size(),0);
		oldEstimatedDerivations[i].resize(ScalingFactors[i].size(),0);
	}
	lRate = 1;

}


OptimizerML::~OptimizerML()
{
}


void OptimizerML::EvalSentence(Graph * G){
	Sent++;
	logProbCorrect += G->ComputeLogProbabilityCorrect();
	logProbBest += G->ComputeLogProbabilityBest();
	G->ComputeCorrectLinks(CorrectLink,FailedLink,CorrectNoLink,FailedNoLink);
	for(int i = 0; i < FeatureSets.size();i++) {
	      //store new Scaling Factors
		vector<double> EstDer = FeatureSets[i]->GetEstimatedDerivations();
		assert(EstDer.size() <= EstimatedDerivations[i].size());
		for(int j = 0; j < EstDer.size(); j++) {
			EstimatedDerivations[i][j] += EstDer[j];
		}
	}


}
void OptimizerML::Estimate(){
/*	if((iteration % 20 == 0)&&(iteration != 0)) {
		ScalingFactors = bestScalingFactors;
		for(int i = 0; i < EstimatedDerivations.size();i++) {
			cout << "Set ScalingFactors to best possible " << i << endl;
			for(int j = 0; (j < ScalingFactors[i].size()) && (j < 12);j++) {
				cout << ScalingFactors[i][j] << "_";
			}
			cout << endl;
			FeatureSets[i]->SetScalingFactors(ScalingFactors[i]);
		}
	iteration++;
	//reset for new iteration
	Sent = 0;

    //reset Estimated Derivations
	for(int i =0; i < EstimatedDerivations.size(); i++) {
      //cout << "ScalingFactors[i].size(): " << ScalingFactors[i].size() << endl;
		EstimatedDerivations[i].assign(ScalingFactors[i].size(),0);
	}
	return;
	}


	double maxValue = exp(1.0*-iteration/50);
	double change = 1;
	for(int i = 0; i < EstimatedDerivations.size();i++) {
		for(int j = 0; j < EstimatedDerivations[i].size();j++) {
			if(learnRate[i][j] != 0) {
				EstimatedDerivations[i][j] /= Sent;
				cout << "Derivation: " << EstimatedDerivations[i][j] << endl;
				if(EstimatedDerivations[i][j] * lRate > maxValue) {
	 				if(EstimatedDerivations[i][j] * oldEstimatedDerivations[i][j] < 0) {
						change = -1;
					}else if(change == 1) {
						change = 0;
					}
				}else if(EstimatedDerivations[i][j] * lRate < -1*maxValue) {
	 				if(EstimatedDerivations[i][j] * oldEstimatedDerivations[i][j] < 0) {
						change = -1;
					}else if(change == 1) {
						change = 0;
					}
				}else if((abs(EstimatedDerivations[i][j] * lRate*8) > maxValue) &&(change == 1)) {
					change = 0;

				}
			}
		}
	}
	if(change == -1) {
		lRate /=2;
	}else if((change == 1)&&(lRate < 1)) {
		lRate *= 2;
	}
	cout << "New lRate: " << lRate << endl; 
	cout << "MaxValue: " << maxValue << endl;
	for(int i = 0; i < EstimatedDerivations.size();i++) {
		cout << "New Opt: ScalingFactors for Featureset " << i << endl;
		for(int j = 0; j < EstimatedDerivations[i].size();j++) {
			if(learnRate[i][j] != 0) {
				if(EstimatedDerivations[i][j] * lRate > maxValue) {
					EstimatedDerivations[i][j] = 1/(lRate)*maxValue;
				}else if(EstimatedDerivations[i][j] * lRate < -1*maxValue) {
					EstimatedDerivations[i][j] = -1/(lRate)*maxValue;
				}
					
				ScalingFactors[i][j] += lRate * EstimatedDerivations[i][j];
				oldEstimatedDerivations[i][j] = EstimatedDerivations[i][j];
			}
		}
		for(int j = 0; (j < ScalingFactors[i].size()) && (j < 12);j++) {
			cout << ScalingFactors[i][j] << "_";
		}
		cout << endl;
		FeatureSets[i]->SetScalingFactors(ScalingFactors[i]);
	}
*/
/*	lRate = exp(1.0*-iteration/100);
	cout << "New lRate: " << lRate << endl; 
	for(int i = 0; i < EstimatedDerivations.size();i++) {
		cout << "New Opt: ScalingFactors for Featureset " << i << endl;
		for(int j = 0; j < EstimatedDerivations[i].size();j++) {
			if(learnRate[i][j] != 0) {
				if(EstimatedDerivations[i][j] > 1) {
					EstimatedDerivations[i][j] = 1;
				}else if(EstimatedDerivations[i][j] < -1) {
					EstimatedDerivations[i][j] = -1;
				}
					
				ScalingFactors[i][j] += lRate * EstimatedDerivations[i][j];
				oldEstimatedDerivations[i][j] = EstimatedDerivations[i][j];
			}
		}
		for(int j = 0; (j < ScalingFactors[i].size()) && (j < 12);j++) {
			cout << ScalingFactors[i][j] << "_";
		}
		cout << endl;
		FeatureSets[i]->SetScalingFactors(ScalingFactors[i]);
	}
*/
/*	if((iteration + 1) % 10 == 0) {
		cout << "Change Directions after " << iteration + 1 << " iterations" << endl;
		for(int i = 0; i < EstimatedDerivations.size();i++) {
			cout << "Set:" << i << ": ";
			for(int j = 0; j < EstimatedDerivations[i].size();j++) {
				cout << changedDirections[i][j] << "_";
			}
			cout << endl;
		}
	}
*/	
	double derSize = 0;
	int improvedDerivations = 0;
	cout << "ScalingFactors :"<< endl; 
	for(int i = 0; i < EstimatedDerivations.size();i++) { 
		for(int j = 0; j < EstimatedDerivations[i].size();j++) { 
			if(learnRate[i][j] != 0) { 
				EstimatedDerivations[i][j] /= Sent;
				//cout << "Derivation over all Sentences[" << i << "][" << j << "]:" <<  EstimatedDerivations[i][j] << endl;
				ScalingFactors[i][j] += learnRate[i][j] * EstimatedDerivations[i][j]; 
				//ScalingFactors[i][j] += learnRate[i][j];
				derSize += abs(EstimatedDerivations[i][j]);
				if(abs(oldEstimatedDerivations[i][j]) > abs(EstimatedDerivations[i][j])) {
					improvedDerivations ++;
				}
				oldEstimatedDerivations[i][j] = EstimatedDerivations[i][j];
			} 
		} 
		for(int j = 0; (j < ScalingFactors[i].size());j++) { 
			cout << ScalingFactors[i][j] << "_"; 
		} 
		cout << endl; 
		FeatureSets[i]->SetScalingFactors(ScalingFactors[i]); 
	}
	cout << "Derivationsize: " << derSize << " ImprovedDerivations: " << improvedDerivations << endl;

	//reset for new iteration
	Sent = 0;

    //reset Estimated Derivations
	for(int i =0; i < EstimatedDerivations.size(); i++) {
      //cout << "ScalingFactors[i].size(): " << ScalingFactors[i].size() << endl;
		EstimatedDerivations[i].assign(ScalingFactors[i].size(),0);
	}
	iteration ++;
}
void OptimizerML::EvalCVSentence(Graph * G){
	Sent++;
	logProbCorrectCV += G->ComputeLogProbabilityCorrect();
	logProbBestCV += G->ComputeLogProbabilityBest();
	G->ComputeCorrectLinks(CorrectLinkCV,FailedLinkCV,CorrectNoLinkCV,FailedNoLinkCV);
}
void OptimizerML::CheckScalingFactors() {
	logProbCorrect /= Sent;
	logProbBest /= Sent;
	cout << "LogProbCorrect:" << logProbCorrect << " LogProbBest:" << logProbBest << endl;
	cout << "Links: Correct: " << CorrectLink << " Failed: " << FailedLink << endl;
	cout << "NoLinks: Correct: " << CorrectNoLink << " Failed: " << FailedNoLink << endl;
	cout << "All Failed: " << FailedLink + FailedNoLink << endl;
       //if(CorrectLink + CorrectNoLink > bestCorrectLink + bestCorrectNoLink) {
	if(logProbCorrect > bestLogProbCorrect) {
	//save best Scaling Factors
		bestCorrectLink = CorrectLink;
		bestCorrectNoLink = CorrectNoLink;
		bestFailedLink = FailedLink;
		bestFailedNoLink = FailedNoLink;
		bestLogProbCorrect = logProbCorrect;
		bestLogProbBest = logProbBest;
		bestScalingFactors = ScalingFactors;
		bestIteration = iteration;
	}

	//reset for new iteration
	logProbCorrect = 0;
	logProbBest = 0;
	CorrectLink = 0;
	FailedLink = 0;
	CorrectNoLink = 0;
	FailedNoLink = 0;

}

void OptimizerML::CheckScalingFactorsWithCV() {
	logProbCorrect /= Sent;
	logProbBest /= Sent;
	cout << "Dev Set: " << endl;
	cout << "LogProbCorrect:" << logProbCorrect << " LogProbBest:" << logProbBest << endl;
	cout << "Links: Correct: " << CorrectLink << " Failed: " << FailedLink << endl;
	cout << "NoLinks: Correct: " << CorrectNoLink << " Failed: " << FailedNoLink << endl;
	logProbCorrectCV /= Sent;
	logProbBestCV /= Sent;
	cout << "Crossvalidation Set:" << endl;
	cout << "CV:LogProbCorrect:" << logProbCorrectCV << " LogProbBest:" << logProbBestCV << endl;
	cout << "CV:Links: Correct: " << CorrectLinkCV << " Failed: " << FailedLinkCV << endl;
	cout << "CV:NoLinks: Correct: " << CorrectNoLinkCV << " Failed: " << FailedNoLinkCV << endl;
      
      //if(CorrectLink + CorrectNoLink > bestCorrectLink + bestCorrectNoLink) {
	if(logProbCorrectCV > bestLogProbCorrect) {
	//save best Scaling Factors
		bestCorrectLink = CorrectLinkCV;
		bestCorrectNoLink = CorrectNoLinkCV;
		bestFailedLink = FailedLinkCV;
		bestFailedNoLink = FailedNoLinkCV;
		bestLogProbCorrect = logProbCorrectCV;
		bestLogProbBest = logProbBestCV;
		bestScalingFactors = ScalingFactors;
		bestIteration = iteration;
	}

	//reset for new iteration
	logProbCorrect = 0;
	logProbBest = 0;
	CorrectLink = 0;
	FailedLink = 0;
	CorrectNoLink = 0;
	FailedNoLink = 0;
	logProbCorrectCV = 0;
	logProbBestCV = 0;
	CorrectLinkCV = 0;
	FailedLinkCV = 0;
	CorrectNoLinkCV = 0;
	FailedNoLinkCV = 0;

}
void OptimizerML::OutputBest(){

	cout << "Best Result at Iteration "<< bestIteration <<": " << endl;
	cout << "LogProbCorrect:" << bestLogProbCorrect << " LogProbBest:" << bestLogProbBest << endl;
	cout << "Links: Correct: " << bestCorrectLink << " Failed: " << bestFailedLink << endl;
	cout << "NoLinks: Correct: " << bestCorrectNoLink << " Failed: " << bestFailedNoLink << endl;
    
	for(int i = 0; i < bestScalingFactors.size();i++) {
		FeatureSets[i]->SetScalingFactors(bestScalingFactors[i]);
		FeatureSets[i]->OutputScalingFactors();
	}

}
