
#include "optimizerrecall.h"

OptimizerRecall::OptimizerRecall(vector <FeatureSet *> FSets,vector < vector<double> > SFactors,char * learnRateFile):Optimizer(FSets){
	numberOfAllFeatures = 0;
	FeatureSets = FSets;
	ScalingFactors = SFactors;
	for(int i = 0; i < ScalingFactors.size();i++) {
		numberOfAllFeatures += ScalingFactors[i].size();
	}
	TruePositiveP = 0;
	FalsePositiveP = 0;
	TruePositiveS = 0;
	FalsePositiveS = 0;
	SLinks = 0;
	PLinks = 0;
	SmoothedTruePositiveP = 0;
	SmoothedFalsePositiveP = 0;
	SmoothedTruePositiveS = 0;
	SmoothedFalsePositiveS = 0;
	FalsePositivePDerivation.resize(numberOfAllFeatures,0);
	TruePositivePDerivation.resize(numberOfAllFeatures,0);
	FalsePositiveSDerivation.resize(numberOfAllFeatures,0);
	TruePositiveSDerivation.resize(numberOfAllFeatures,0);
	alpha = 20;
	iteration = 0;
	if(learnRateFile == NULL) {
		cout << " For optimization a learnRate file is needed" << endl;
		exit(-1);
	}
	for(int i = 0; i < ScalingFactors.size(); i++) {
		learnRate.push_back(vector<double>(ScalingFactors[i].size(),0));
      //mean.push_back(vector<double>(ScalingFactors[i].size(),0));
      //sigma.push_back(vector<double>(ScalingFactors[i].size(),10000));
	}
	LoadFactorFile(learnRateFile,learnRate);

	bestIteration = 0;
	bestRecall = 0;
}

void OptimizerRecall::EvalSentence(Graph * G){
	//first all features have to calculate their derivations
	for(int i =0; i < FeatureSets.size(); i++) {	
		FeatureSetIterator * iter = FeatureSets[i]->GetIterator();
		for(; !iter->end(); iter->next()) {
			iter->feature()->CalcDerivations();
		}
		delete iter;
	}	

	GraphIterator * iter = G->GetIterator();
	for(;!iter->end();iter->next()) {
		HiddenNode *n = iter->node();
		int gv = n->GetGivenValue();
		pair<double,double> b = n->GetBelief();
		SparseArray derivation = n->GetDerivation(numberOfAllFeatures);
		double s = 1.0/(1+exp(-1*alpha*(b.second - 0.5)));
    //cout <<"b : " << b.second << " s: " << s << endl;
		switch(gv) {
			case 0:
				if(b.first < b.second) {
	//no link in reference link in alignment -> FP
	//cout <<" FP += 1" << endl;
					FalsePositiveP++;
					FalsePositiveS++;
				}
      //and of the smooth version always
      //cout << "SmoothFP +=" << s << endl;
				SmoothedFalsePositiveP += s;
				SmoothedFalsePositiveS += s;	
      //derivation
      //cout << "Derivation+=:" << alpha*s*(1-s)*derivation.first[0] << endl;
				for(int i = 0; i < derivation.size; i++) {
					FalsePositivePDerivation[derivation.index[i]] += alpha*s*(1-s)*derivation.data[i];
					FalsePositiveSDerivation[derivation.index[i]] += alpha*s*(1-s)*derivation.data[i];
				}
				break;
			case 1:
				SLinks++;
				PLinks++;
				if(b.first < b.second) {
	//both links ->TP
	//cout <<" TP += 1" << endl;
					TruePositiveP++;
					TruePositiveS++;
				}
      //and of the smooth version always
      //cout << "SmoothTP +=" << s << endl;
				SmoothedTruePositiveP += s;
				SmoothedTruePositiveS += s;
      //derivation
      //cout << "Derivation+=:" << alpha*s*(1-s)*derivation.first[0] << endl;
				for(int i = 0; i < derivation.size; i++) {
					TruePositivePDerivation[derivation.index[i]] += alpha*s*(1-s)*derivation.data[i];
					TruePositiveSDerivation[derivation.index[i]] += alpha*s*(1-s)*derivation.data[i];
				}
				break;
			case -1:
				PLinks++;
				if(b.first < b.second) {
	//cout <<" TPP += 1 FPS += 1" << endl;
	//possbile Link and Alignment ->TPp
					TruePositiveP++;
	//no sure link and alignment-> FPs
					FalsePositiveS++;
				}
      //and of the smooth version always
      //cout << "SmoothTPP/SmoothFPS +=" << s << endl;
				SmoothedTruePositiveP += s;
				SmoothedFalsePositiveS += s;
      //cout << "Derivation+=:" << alpha*s*(1-s)*derivation.first[0] << endl;
				for(int i = 0; i < derivation.size; i++) {
					TruePositivePDerivation[derivation.index[i]] += alpha*s*(1-s)*derivation.data[i];
					FalsePositiveSDerivation[derivation.index[i]] += alpha*s*(1-s)*derivation.data[i];
				}
      
				break;
     
		}
		delete[] derivation.data;
		delete[] derivation.index;
	}
	delete iter;
}

void OptimizerRecall::Estimate(){
	double Recall = 1.0*TruePositiveS / SLinks;
	double SmoothedRecall = SmoothedTruePositiveS/SLinks;
	cout << " TruePositive: Sure:" << TruePositiveS << " Possible: " << TruePositiveP << endl;
	cout << " FalsePostive: Sure:" << FalsePositiveS << " Possible: " << FalsePositiveP << endl;
	cout << "Smoothed Version: " << endl;
	cout << " TruePositive: Sure:" << SmoothedTruePositiveS << " Possible: " << SmoothedTruePositiveP << endl;
	cout << " FalsePostive: Sure:" << SmoothedFalsePositiveS << " Possible: " << SmoothedFalsePositiveP << endl;
	cout << "Links Sure:" << SLinks << " Possible: " << PLinks << endl;
	cout << "Recall:" << Recall << " SmoothedRecall:" << SmoothedRecall << endl;
  //cout << "Derivations:TPS: " << TruePositiveSDerivation[0] << "TPP" << TruePositivePDerivation[0] << endl;
  //cout << "Derivations:FPS: " << FalsePositiveSDerivation[0] << "FPP" << FalsePositivePDerivation[0] << endl;


	if(Recall > 0.95) {
		bestRecall = Recall;
		bestScalingFactors = ScalingFactors;
		bestIteration = iteration;
		OutputBest();
		exit(1);
	}

	if(Recall > bestRecall) {
		bestRecall = Recall;
		bestScalingFactors = ScalingFactors;
		bestIteration = iteration;
	}
	int index = 0;
	cout << "Derivations:" << endl;
	for(int i = 0; i < ScalingFactors.size(); i++) {
		for(int j = 0; j < ScalingFactors[i].size();j++) {
    //cout << TruePositiveSDerivation[index] << "*(" << SmoothedFalsePositiveS << " + " <<SLinks<< ")";
    //cout << " + " << TruePositivePDerivation[index] << " *( " << SmoothedFalsePositiveP << " + " << SLinks <<" )";
    //cout << " - " << FalsePositiveSDerivation[index] << " * " << SmoothedTruePositiveS;
    //cout << " - " << FalsePositivePDerivation[index] << " * " << SmoothedTruePositiveP << " = ";
    //cout << " / -1*(" << SmoothedTruePositiveS << " + " << SmoothedFalsePositiveS << " + " << SLinks << ")^2" << endl;
			if(learnRate[i][j] != 0) {
				double der = TruePositiveSDerivation[index] * SLinks;
				der /= SLinks *SLinks;
				cout <<" Derivation[" << i << "]{" << j << "](" << index <<") : " << der << endl;
				cout << der << " ";
				ScalingFactors[i][j] -= learnRate[i][j] * der;
				cout << "Learnrate: " << learnRate[i][j] << endl;
				cout << "ScalingFactor: " << ScalingFactors[i][j] << endl;
			}
			index++;
		}
		FeatureSets[i]->SetScalingFactors(ScalingFactors[i]);
	}
	cout << endl,
 TruePositiveP = 0;
 FalsePositiveP = 0;
 TruePositiveS = 0;
 FalsePositiveS = 0;
 SLinks = 0;
 PLinks = 0;
 SmoothedTruePositiveP = 0;
 SmoothedFalsePositiveP = 0;
 SmoothedTruePositiveS = 0;
 SmoothedFalsePositiveS = 0;
 FalsePositivePDerivation.assign(numberOfAllFeatures,0);
 TruePositivePDerivation.assign(numberOfAllFeatures,0);
 FalsePositiveSDerivation.assign(numberOfAllFeatures,0);
 TruePositiveSDerivation.assign(numberOfAllFeatures,0);
	iteration++;
}
void OptimizerRecall::OutputBest(){
	cout << "Best Recall: " << bestRecall << endl;
	cout << "In iteration: " << bestIteration << endl;

	for(int i = 0; i < FeatureSets.size(); i ++) {
		FeatureSets[i]->SetScalingFactors(bestScalingFactors[i]);
		cout << "Output Scaling Factors " << i << endl;
		FeatureSets[i]->OutputScalingFactors();
	}

}

