#include "parseTree.h"
#include <cassert>
#include <iostream>
using namespace std;

ParseTree::ParseTree(){

	root=new TreeNode;
	Init(root);

}


void ParseTree::Init(TreeNode * t){
	t->parent = NULL;
	t->headChild = NULL;
	t->sister = NULL;
	t->mark = 0;
	t->name = "";
}


void ParseTree::BuildTree(vector<string> tree)

{
	assert(root != NULL);
	root->parent = NULL;
	root->headChild = NULL;
	root->sister = NULL;
	root->name="ROOT0";

	TreeNode * curr = root;  // the current tree node

	for(unsigned i=0; i<tree.size(); i++) {
	//	cout<<tree[i]<<endl;
		if(tree[i] == "("){   // create a new tree node
			TreeNode * tn=new TreeNode;
			Init(tn);
			tn->parent = curr;
			if(curr->headChild == NULL){
				curr->headChild=tn;
			}
			else{
				TreeNode * tail = curr->headChild;
				while(tail->sister!=NULL){
					tail = tail->sister;
				}
				tail->sister = tn;
			}
			curr = tn;
		}
		else if(tree[i] == ")") { // step back to a higher level
			curr = curr->parent;
		}
		else{  // name of the node
			curr->name = tree[i];
		}
	}
}

TreeNode * ParseTree::GetRoot()
{
	return root;
}

int ParseTree::FindDepth(int index)
{
	ClearMark();
	int depth = 0;
	int count = 0;  // count the index of the words
	TreeNode * curr = root->headChild;
	// do a depth first search
	while(curr != root){
		if(curr->mark == 0){ // subtree not traversed yet
			if(curr->headChild != NULL){ // internal node
				curr = curr->headChild;	
				depth++;
			}
			else{  // leaf node
				count++;
				if(count == index){
					cout<<curr->name<<"("<<count<<" "<<depth<<") "<<endl;
					return depth;
				}
				else{
					curr->mark=1;
				}
			}
		}
		else{  // already completed the subtree
			if(curr->sister != NULL) { 
				curr = curr->sister;
			}
			else{ // go back to its parent
				curr = curr->parent;
				curr -> mark=1;
				depth--;
			}
		}
	}
	return 0;	
}

int ParseTree::FindMaxDepth()
{
	ClearMark();
        int depth = 0;
	int max=0;
 
        TreeNode * curr = root->headChild;
        // do a depth first search
        while(curr != root){
                if(curr->mark == 0){ // subtree not traversed yet
                        if(curr->headChild != NULL){ // internal node
                                curr = curr->headChild;
                                depth++;
				if(max<depth){	
					max=depth;
				}
                        }
                        else{  // leaf node
                                curr->mark=1;
                        }
                }
                else{  // already completed the subtree
                        if(curr->sister != NULL) {
                                curr = curr->sister;
                        }
                        else{ // go back to its parent
                                curr = curr->parent;
                                curr -> mark=1;
                                depth--;
                        }
                }
        }
	return max;
}


void ParseTree::ClearMark()
{  // two rounds of clearning up
	TreeNode * curr = root->headChild;
	while(curr != root){
		 if(curr->mark != 2){ // subtree not traversed yet
                        if(curr->headChild != NULL){ // internal node
                                curr = curr->headChild;
                        }
                        else{  // leaf node
                                curr->mark=2;
                        }
                }
                else{  // already completed the subtree
                        if(curr->sister != NULL) {
                                curr = curr->sister;
                        }
                        else{ // go back to its parent
                                curr = curr->parent;
                                curr -> mark=2;
                        }
                }
        }

	curr = root->headChild;
        while(curr != root){
                 if(curr->mark != 0){ // subtree not traversed yet
                        if(curr->headChild != NULL){ // internal node
                                curr = curr->headChild;
                        }
                        else{  // leaf node
                                curr->mark=0;
                        }
                }
                else{  // already completed the subtree
                        if(curr->sister != NULL) {
                                curr = curr->sister;
                        }
                        else{ // go back to its parent
                                curr = curr->parent;
                                curr -> mark=0;
                        }
                }
        }

}

void ParseTree::PrintTree(){	    
	int i=0;   
	ClearMark();
 
	int depth = 0;
        TreeNode * curr = root->headChild;
	if(curr==NULL){
		cout<< "empty tree" <<endl;
		return;
	}
	else{
		cout<<curr->name<<" ";
	}
	
        // do a depth first search
        while(curr != root){
                if(curr->mark == 0){ // subtree not traversed yet
                        if(curr->headChild != NULL){ // internal node
                                curr = curr->headChild;
                                depth++;
				cout<<endl;
				for(i=0;i<depth;i++){
					cout<<" ";
				}
				cout<<"("<<curr->name<<" ";
                        }
                        else{  // leaf node
                                curr->mark=1;
                        }
                }
                else{  // already completed the subtree
			cout<<")";
                        if(curr->sister != NULL) {
                                curr = curr->sister;
				cout<<endl;
				for(i=0;i<depth;i++){
					cout<<" ";
				}
				cout<<"("<<curr->name<<" ";
                        }
                        else{ // go back to its parent
                                curr = curr->parent;
                                curr -> mark=1;
				depth--;
			}
		}
	}
}

