
#include "phrasefeature.h"

PhraseFeature::PhraseFeature()
 : Feature()
{
	numberOfFeatureValues = 0;
	expValueCalculated = 0;
//	oneNodeDerivation.size = 0;
//	zeroNodeDerivation.size = 0;
}


PhraseFeature::~PhraseFeature()
{
/*	if(oneNodeDerivation.size > 0) {
		delete[] zeroNodeDerivation.data;
		delete[] zeroNodeDerivation.index;
		delete[] oneNodeDerivation.data;
		delete[] oneNodeDerivation.index;
	}
*/
	//remove derivations
	for(map<const HiddenNode *,SparseArray>::iterator iter = derivations.begin(); iter != derivations.end(); iter++) {
		delete[] iter->second.data;
		delete[] iter->second.index;
	}
	if(numberOfFeatureValues)
		delete[] featureValues;

}


void PhraseFeature::Init(vector<HiddenNode *> & zeroN, vector<HiddenNode *> & oneN, vector<double> & values, double dynamicValue, double * sFactors, double dynScalingFactor, int dynSFIndex,int off, int dynOff, int nOfAllFeatures) {

	zeroNodes = zeroN;
	//cout << "NumberOfZeroNodes: " << zeroNodes.size() << endl;
	oneNodes = oneN;
	numberOfFeatureValues = values.size();
	featureValues = new double[values.size()];
	for(int i = 0; i < values.size(); i++) {
		featureValues[i] = values[i];
	}
	scalingFactors = sFactors;
	dynamicFeatureValue = dynamicValue;
	dynamicScalingFactor = dynScalingFactor;
	dynamicSFIndex = dynSFIndex;
	numberOfAllFeatures = nOfAllFeatures;
	offset = off;
	dynamicOffset = dynOff;

	featureValueCorrect = 0;
	featureValueFalse = 0;
	for(int i = 0; i < numberOfFeatureValues; i++) {
		featureValueFalse += (1-featureValues[i]) * scalingFactors[2*i];
		featureValueCorrect += featureValues[i] * scalingFactors[2*i+1];
	}
	featureValueCorrect += dynamicValue * dynScalingFactor;
	//cout << "F: " << featureValueCorrect << endl;
	//cout << "Not F: " << featureValueFalse << endl;
	featureValueCorrect = exp(featureValueCorrect);
	//cout << "FeatureValue: " << featureValueCorrect << endl;
	featureValueFalse = exp(featureValueFalse);


	for(int i = 0; i < zeroNodes.size(); i++) {
		zeroNodes[i]->AddConnection(this);
		//messages.insert(make_pair(zeroNodes[i],make_pair(0.5,0.5)));
		messages.insert(make_pair(zeroNodes[i],make_pair(featureValueCorrect/(featureValueCorrect+featureValueFalse),featureValueFalse/(featureValueCorrect + featureValueFalse))));
	}
	for(int i = 0; i < oneNodes.size(); i++) {
		oneNodes[i]->AddConnection(this);
		//messages.insert(make_pair(oneNodes[i],make_pair(0.5,0.5)));
		messages.insert(make_pair(oneNodes[i],make_pair(featureValueFalse/(featureValueCorrect+featureValueFalse),featureValueCorrect/(featureValueCorrect + featureValueFalse))));
	}


	CalcExampleValues();
}


pair<double,double> PhraseFeature::GetMessage (const HiddenNode * ToNode ) const {
	//return messages[ToNode];
	map<const HiddenNode *,pair<double,double> >::const_iterator iter = messages.find(ToNode);
	if(iter != messages.end()) {
		//cout << "PhraseFeature send message: (" << iter->second.first << "," << iter->second.second << ")" << endl;
		return iter->second;
	}else {
		cerr << "Asking for message to HiddenNode, that not exists" << endl;
		exit(-1);
	}


}

void PhraseFeature::CalcMessages() {
	//cout << "CalcMessages():" << endl;
	// some values have to be precalculated
	double allProb = 1;
	double correctProb = 1;
	for(int i = 0; i < zeroNodes.size(); i++) {
		pair<double,double> m = zeroNodes[i]->GetMessage(this);
		//cout << "ZeroNode:" << m.first << endl;
		allProb *= (m.first + m.second);
		correctProb *= m.first;
	
	}	
	for(int i = 0; i < oneNodes.size(); i++) {
		pair<double,double> m = oneNodes[i]->GetMessage(this);
		//cout << "OneNode:" << m.second << endl;
		allProb *= (m.first + m.second);
		correctProb *= m.second;
	}	

	for(int i = 0; i < zeroNodes.size(); i++) {
		pair<double,double> m = zeroNodes[i]->GetMessage(this);
		pair<double,double> message = make_pair(0,0);
		
		// calc message for this node
		if(correctProb > 0) {
			assert(m.first > 0);
			message.first = featureValueCorrect * correctProb /m.first + featureValueFalse * (allProb/(m.first + m.second) - correctProb / m.first);
		}else {
			message.first = featureValueFalse * allProb/(m.first + m.second);
		}
		message.second = featureValueFalse * allProb/(m.first + m.second);
		//normalize message
		double sum = message.first + message.second;
		message.first /= sum;
		message.second /= sum;
		
		if(isnan(message.first)) {
			cout << "CorrectProb: " << correctProb << endl;
			cout << "FeatureValues: " << featureValueFalse << " " << featureValueCorrect << endl;
			cout << "Message In: " << m.first << " " << m.second << endl;
			cout << "Message Out:" << message.first << " " << message.second << endl;
		}

		assert(!isnan(message.first));
		//save message
		messages.erase(zeroNodes[i]);
		messages.insert(make_pair(zeroNodes[i],message));
		
	}
	for(int i = 0; i < oneNodes.size(); i++) {
		pair<double,double> m = oneNodes[i]->GetMessage(this);
		pair<double,double> message = make_pair(0,0);
		
		// calc message for this node
		if(correctProb > 0) {
			assert(m.second > 0);
			message.second = featureValueCorrect * correctProb /m.second + featureValueFalse * (allProb/(m.first + m.second) - correctProb / m.second);
		}else {
			message.second = featureValueFalse * allProb/(m.first + m.second);
		}
		message.first = featureValueFalse * allProb/(m.first + m.second);

		//normalize message
		double sum = message.first + message.second;
		message.first /= sum;
		message.second /= sum;
		//cout << "One Node Out: (" << message.first << "," << message.second << ")" << endl;;
		//cout << " -> (" << message.first << "," << message.second << ")" << endl;
		
		if(isnan(message.first)) {
			cout << "CorrectProb: " << correctProb << endl;
			cout << "FeatureValues: " << featureValueFalse << " " << featureValueCorrect << endl;
			cout << "Message In: " << m.first << " " << m.second << endl;
			cout << "Message Out:" << message.first << " " << message.second << endl;
		}
		assert(!isnan(message.first));
		//save message
		messages.erase(oneNodes[i]);
		messages.insert(make_pair(oneNodes[i],message));
		
	}
	//expectation value has to be recalculated
	expValueCalculated = 0;

}

  /**
   * @return vector<double>
   */
vector<double> PhraseFeature::GetExpectationValue() {
	vector<double> expValues;
	
	int active = 1;
	// some values have to be precalculated
	double correctProb = 1;
	for(int i = 0; i < zeroNodes.size(); i++) {
		pair<double,double> m = zeroNodes[i]->GetBelief();
		active &= (m.first > m.second);
		//cout << "ZeroNode:" << m.first << endl;
		correctProb *= m.first;
	}	
	for(int i = 0; i < oneNodes.size(); i++) {
		pair<double,double> m = oneNodes[i]->GetBelief();
		//cout << "OneNode:" << m.second << endl;
		active &= (m.second > m.first);
		correctProb *= m.second;
	}
	if(0) {
	cout << "Node:";
	for(int i = 0; i < oneNodes.size(); i++) {
		pair<string,string> w = oneNodes[i]->GetWords();
		cout << "(" << w.first << "," << w.second << ") ";
	} 
	if(active) {
		cout << "active" << endl;
	}else {
		cout << "inactive" << endl;
	}
	}	
	//cout << "CorrectProb: " << correctProb << endl;
	for(int i = 0; i < numberOfFeatureValues; i++) {
		expValues.push_back((1 - correctProb) * featureValues[i]);
		expValues.push_back(correctProb * featureValues[i]);
		//cout << featureValues[i] << "->" << expValues[i] << endl;
	}
	dynamicExpecationValue = correctProb * dynamicFeatureValue;
	expValueCalculated = 1;
	return expValues;
}
  /**
   * @return vector<double>
   */
double PhraseFeature::GetDynamicExpectationValue() {
	if(expValueCalculated){
		//cout << "Dyn: " << dynamicFeatureValue << " -> " << dynamicExpecationValue << endl;
		return dynamicExpecationValue;
	}else {
		cout << "PhraseFeature::GetExpecationValue() has to be call first" << endl;
		exit(1);
	}
}
  /**
   * @return vector<double>
   */
int PhraseFeature::GetDynamicSFIndex() {
	return dynamicSFIndex;
}

void PhraseFeature::CalcExampleValues() {
	int active = 1;
	for(int i = 0; active && (i < zeroNodes.size()) ; i++) {
		active &= (zeroNodes[i]->GetGivenValue() == 0);
	}
	for(int i = 0; active && (i < oneNodes.size()) ; i++) {
		active &= (oneNodes[i]->GetGivenValue() == 1);
	}
	if(active) {
		for(int i = 0; i < numberOfFeatureValues; i++) {
			exampleValue.push_back(0);
			exampleValue.push_back(featureValues[i]);
		}
		dynamicExampleValue = dynamicFeatureValue;
	}else {
		for(int i = 0; i < numberOfFeatureValues; i++) {
			exampleValue.push_back(1 - featureValues[i]);
			exampleValue.push_back(0);
		}
		dynamicExampleValue = 0;
	}
	//cout << "Phrase activation: " << active << endl;
}


SparseArray PhraseFeature::GetDerivation (const HiddenNode * ToNode) const {
	map<const HiddenNode *,SparseArray >::const_iterator iter = derivations.find(ToNode);
	if(iter != derivations.end()) {
		//cout << "PhraseFeature send message: (" << iter->second.first << "," << iter->second.second << ")" << endl;
		return iter->second;
	}else {
		cerr << "Asking for message to HiddenNode, that not exists" << endl;
		exit(-1);
	}


}

void PhraseFeature::CalcDerivations() {

// static derivation
/*	oneNodeDerivation.size = 2 * numberOfFeatureValues + (dynamicSFIndex >= 0);
	//cout << "Derivation Size: " << oneNodeDerivation.size << endl;
	oneNodeDerivation.data = new double[oneNodeDerivation.size];
	oneNodeDerivation.index = new int[oneNodeDerivation.size];

	double f = featureValueFalse/(featureValueCorrect+featureValueFalse) * featureValueCorrect/(featureValueCorrect + featureValueFalse);
	for(int i = 0; i < numberOfFeatureValues; i++) {
	assert(2*i + 1 < oneNodeDerivation.size);
		oneNodeDerivation.index[2*i] = offset + 2*i;
		oneNodeDerivation.index[2*i+1] = offset + 2*i + 1;
		oneNodeDerivation.data[2*i] = f * featureValues[2*i];
		oneNodeDerivation.data[2*i + 1] = f * -1 * featureValues[2*i+1];
	}
	if(dynamicSFIndex >= 0) {
		assert(2*numberOfFeatureValues < oneNodeDerivation.size);
		oneNodeDerivation.index[2*numberOfFeatureValues] = dynamicOffset + dynamicSFIndex;
		oneNodeDerivation.data[2*numberOfFeatureValues] = f * -1;
	}

	//copy to zero Node and invert data
	zeroNodeDerivation.size = oneNodeDerivation.size;
	zeroNodeDerivation.data = new double[zeroNodeDerivation.size];
	zeroNodeDerivation.index = new int[zeroNodeDerivation.size];
	for(int i = 0; i < zeroNodeDerivation.size;i++) {
		zeroNodeDerivation.index[i] = oneNodeDerivation.index[i];
		zeroNodeDerivation.data[i] =  -1 * oneNodeDerivation.data[i];		
	}

	//insert into map
	for(int i = 0; i < oneNodes.size(); i++){
		derivations.insert(make_pair(oneNodes[i],oneNodeDerivation));
	}
	for(int i = 0; i < zeroNodes.size(); i++){
		derivations.insert(make_pair(zeroNodes[i],zeroNodeDerivation));
	}
*/ 

// dynmamic derivation

	// some values have to be precalculated
	double allProb = 1;
	double correctProb = 1;
	for(int i = 0; i < zeroNodes.size(); i++) {
		pair<double,double> m = zeroNodes[i]->GetMessage(this);
		//cout << "ZeroNode:" << m.first << endl;
		allProb *= (m.first + m.second);
		correctProb *= m.first;
	}	
	for(int i = 0; i < oneNodes.size(); i++) {
		pair<double,double> m = oneNodes[i]->GetMessage(this);
		//cout << "OneNode:" << m.second << endl;
		allProb *= (m.first + m.second);
		correctProb *= m.second;
	}	

	//cout << "AllProb: " << allProb;
	//cout << " CorrectProb:" << correctProb << endl;

	//for every zero Node
	for(int i = 0; i < zeroNodes.size(); i++) {
		pair<double,double> m = zeroNodes[i]->GetMessage(this);
		SparseArray derivation;
		derivation.size = 2 * numberOfFeatureValues + (dynamicSFIndex >= 0);
		derivation.data = new double[derivation.size];
		derivation.index = new int[derivation.size];

		//individual node precalculation
		double t1 = featureValueCorrect * correctProb /m.first;
		double t2 = featureValueFalse * (allProb/(m.first + m.second) - correctProb / m.first);
		double t3 = featureValueFalse * allProb/(m.first + m.second);
		double m0 = t1 + t2;
		double m1 = t3;
		double quotient = (m0 + m1)*(m0 + m1);
		//derivation calculationen
		for(int j = 0; j < numberOfFeatureValues; j++) {
			assert(2*j + 1 < derivation.size);
			derivation.index[2*j] = offset + 2*j;
			derivation.index[2*j+1] = offset + 2*j + 1;
			derivation.data[2*j] = ((t2*(1-featureValues[j]))*m1 - (t3*(1-featureValues[j]))*m0) / quotient;
			derivation.data[2*j +1] = (t1*featureValues[j])*m1/ quotient;
		}
		if(dynamicSFIndex >= 0) {
			assert(2*numberOfFeatureValues < derivation.size);
			derivation.index[2*numberOfFeatureValues] = dynamicOffset + dynamicSFIndex;
			derivation.data[2*numberOfFeatureValues] = (t1*1)*m1/quotient;;
		}

		//insert into map
		derivations.insert(make_pair(zeroNodes[i],derivation));	
	}
	//for every one Node
	//cout << "OneNodes: " << endl;
	for(int i = 0; i < oneNodes.size(); i++) {
		pair<double,double> m = oneNodes[i]->GetMessage(this);
		//cout << "Message:" << m.first << "," << m.second << endl; 
	
		SparseArray derivation;
		derivation.size = 2 * numberOfFeatureValues + (dynamicSFIndex >= 0);
		derivation.data = new double[derivation.size];
		derivation.index = new int[derivation.size];

		//individual node precalculation
		double t1 = featureValueCorrect * correctProb /m.first;
		double t2 = featureValueFalse * (allProb/(m.first + m.second) - correctProb / m.first);
		double t3 = featureValueFalse * allProb/(m.first + m.second);
		double m1 = t1 + t2;
		double m0 = t3;
		double quotient = (m0 + m1)*(m0 + m1);
		//cout << "featureValueCorrect:" << featureValueCorrect << " False:" << featureValueFalse << endl;
		//cout << "t: " << t1 <<" " << t2 << " "<< t3 << endl;
		//cout << "m: " << m0 << " " << m1 << " Quotient :" << quotient << endl;
	
		//derivation calculationen
		for(int j = 0; j < numberOfFeatureValues; j++) {
			assert(2*j + 1 < derivation.size);
			derivation.index[2*j] = offset + 2*j;
			derivation.index[2*j+1] = offset + 2*j + 1;
			derivation.data[2*j] = ((t3*(1-featureValues[j]))*m1 - (t2*(1-featureValues[j]))*m0) / quotient;
			derivation.data[2*j +1] = - (t1*featureValues[j])*m0/ quotient;
			//cout << "Derivation[" << 2*j<< "]:" << derivation.data[2*j] << " with featureValue: "<< 1-featureValues[j] << endl;
			//cout << "Derivation[" << 2*j + 1<< "]:" << derivation.data[2*j + 1] << " with featureValue: "<< featureValues[j] << endl;
		}
		if(dynamicSFIndex >= 0) {
			assert(2*numberOfFeatureValues < derivation.size);
			derivation.index[2*numberOfFeatureValues] = dynamicOffset + dynamicSFIndex;
			derivation.data[2*numberOfFeatureValues] = -(t1*1)*m0/quotient;;
		}

		//insert into map
		derivations.insert(make_pair(oneNodes[i],derivation));	
	}

}
