#ifndef PHRASEFEATURE_H
#define PHRASEFEATURE_H

#include <feature.h>
#include <map>
/**
	@author Jan Niehues <jan@niehues.info>
*/
class PhraseFeature : public Feature
{
private:

	map<const HiddenNode*,pair<double,double> > messages;

	vector<HiddenNode *> zeroNodes;
	vector<HiddenNode *> oneNodes;
	int numberOfFeatureValues;
	double * featureValues;
	double * scalingFactors;
	double dynamicFeatureValue;
	double dynamicScalingFactor;
	int dynamicSFIndex;
	int numberOfAllFeatures;
	int offset;
	int dynamicOffset;
	double featureValueCorrect;
	double featureValueFalse;

	vector<double> exampleValue;
	double dynamicExampleValue;
	int expValueCalculated;
	double dynamicExpecationValue;

//	SparseArray oneNodeDerivation;
//	SparseArray zeroNodeDerivation;//needed to store pointer to array
	map<const HiddenNode*,SparseArray> derivations;

	void CalcExampleValues();

public:
    PhraseFeature();

    ~PhraseFeature();
     void Init(vector<HiddenNode *> & zeroNodes, vector<HiddenNode *> & oneNodes, vector<double> & values, double dynamicValue, double * ScalingFactors, double dynamicScalingFactors, int dynamicSFIndex, int offset, int dynOff, int numberOfAllFeatures);
	pair<double,double> GetMessage(const HiddenNode * ToNode) const;
	void CalcMessages();


  /**
   * @return vector<double>
   */
  double GetDynamicExpectationValue();

  /**
   * @return double
   */
  inline double GetDynamicExampleValue() {return dynamicExampleValue;};
  /**
   * @return vector<double>
   */
  int GetDynamicSFIndex();


  /**
   * @return vector<double>
   */
  vector<double> GetExpectationValue();

  /**
   * @return double
   */
  inline vector<double> GetExampleValue() {return exampleValue;};


  SparseArray GetDerivation (const HiddenNode * ToNode) const;

void CalcDerivations();

};

#endif
