#include <string.h>
#include "phrasefeatureset.h"

PhraseFeatureSet::PhraseFeatureSet(const char* ParamFile, vector<double> & Factores)
 : FeatureSet(ParamFile,Factores)
{
	assert(ParamFile != NULL);
	numberOfDynamicFeatures = 0;
	numberOfPhraseTables = 0;
	assert(Factores.size() % 2 == 0);
	numberOfFeatures = Factores.size() / 2;

	cout << "ScalingFactors:";
	for(int i = 0;i < Factores.size(); i++) {
		cout << Factores[i] << " ";
	}
	cout << endl;

	InitScalingFactors(Factores);

	LoadParamFile(ParamFile);
	if(numberOfDynamicFeatures) {
		for(int i = 0; i < numberOfDynamicFeatures; i++) {
			Factores.push_back(dynamicScalingFactors[i]);
		}
	}

}


PhraseFeatureSet::~PhraseFeatureSet()
{
	if(numberOfPhraseTables) {
		delete[] phraseTable;
		delete[] featureOffSet;
		delete[] dynamicFeatureOffSet;
	}
	delete[] ScalingFactors;
}

void PhraseFeatureSet::InitScalingFactors(vector<double> & Factors) {
	ScalingFactors = new double[Factors.size()];
	for(int i = 0; i < Factors.size(); i++) {
		ScalingFactors[i] = Factors[i];
	} 
	
}


void PhraseFeatureSet::LoadParamFile(const char* ParamFileName) {

	vector<string> phraseTableFiles;	
	vector<int> numberOfPTFeatures;
	vector<int> numberOfDynamicPTFeatures;


	ifstream ParamFile( ParamFileName);

	if (!ParamFile) {
		cerr << "File " << ParamFileName << " could not be opened\n";
		exit(-2);
	}
  
	char *pKey = new char[ 200 ];
	char *pValue = new char[ 200 ];
// TODO: Does not work with LF.Param with empty line at end!!!!!!!!!!!!
	while ( !ParamFile.eof() ) {
		ParamFile >> pKey >> pValue;
		if(strcmp(pKey,"PhraseTable") == 0) {
			phraseTableFiles.push_back((pValue));
			scalingFactorFiles.push_back("");
			numberOfPTFeatures.push_back(0);
			numberOfDynamicPTFeatures.push_back(0);
			selectedOneNodes.push_back(0);
			exactMatch.push_back(0);
			nonContinous.push_back(0);
		}else if(strcmp(pKey,"ScalingFactorFile") == 0) {
			scalingFactorFiles[scalingFactorFiles.size() -1] = pValue;
		}else if(strcmp(pKey, "NumberOfDynamicFeatures") == 0) {
			numberOfDynamicPTFeatures[numberOfDynamicPTFeatures.size() -1 ] = atoi(pValue);
		}else if(strcmp(pKey, "NumberOfFeatures") == 0) {
			numberOfPTFeatures[numberOfPTFeatures.size() -1 ] = atoi(pValue);
		}else if(strcmp(pKey, "SelectedOneNodes") == 0) {
			selectedOneNodes[selectedOneNodes.size() - 1] = atoi(pValue);
		}else if(strcmp(pKey, "ExactMatch") == 0) {
			exactMatch[exactMatch.size() -1] = atoi(pValue);
		}else if(strcmp(pKey, "NonContinous") == 0) {
			nonContinous[nonContinous.size() -1] = atoi(pValue);
		}



	}

	assert(phraseTableFiles.size() == numberOfPTFeatures.size());
	assert(phraseTableFiles.size() == numberOfDynamicPTFeatures.size());
	assert(phraseTableFiles.size() == scalingFactorFiles.size());
	assert(phraseTableFiles.size() == selectedOneNodes.size());
	assert(phraseTableFiles.size() == exactMatch.size());
	assert(phraseTableFiles.size() == nonContinous.size());

	ParamFile.close();
	delete[] pKey;
	delete[] pValue;

	phraseTable = new multimap<string,PhraseItem>[phraseTableFiles.size()];
	numberOfPhraseTables = phraseTableFiles.size();
	featureOffSet = new int[numberOfPhraseTables + 1];
	dynamicFeatureOffSet = new int[numberOfPhraseTables + 1];

	int fOffSet = 0;
	int dynOffSet = 0;
	for(int i = 0; i < numberOfPhraseTables; i++) {
		LoadPhraseTable(phraseTableFiles[i].c_str(),&phraseTable[i],numberOfPTFeatures[i],numberOfDynamicPTFeatures[i],selectedOneNodes[i],nonContinous[i]);

		assert(phraseTable[i].size() != 0);

		if(scalingFactorFiles[i] != "") {
			cout << "Using additional Phraseindicator features for phrase table "<< i << endl;
			cout << "ScalingFactorFile" << scalingFactorFiles[i] << endl;
			LoadScalingFactorFile(scalingFactorFiles[i],dynamicScalingFactors,dynOffSet,numberOfDynamicPTFeatures[i]);
		}else {
			assert(numberOfDynamicFeatures == 0);	
		}
		featureOffSet[i] = fOffSet;
		dynamicFeatureOffSet[i] = dynOffSet;
		fOffSet += numberOfPTFeatures[i];
		dynOffSet += numberOfDynamicPTFeatures[i];
	}
	numberOfDynamicFeatures = dynOffSet;
	assert(fOffSet == numberOfFeatures);
	featureOffSet[numberOfPhraseTables] = fOffSet;
	dynamicFeatureOffSet[numberOfPhraseTables] = dynOffSet;

}
  /**
   * @param  G
   */
void PhraseFeatureSet::GenerateFeatures (Graph * G  , int offset, int numberOfAllFeatures) {
	int sourceLength = G->GetSourceLength();
	int targetLength = G->GetTargetLength();

/*	cout << "SourceSentence: ";
	for(int i = 0; i < sourceLength; i++)
		cout << G->GetSourceWord(i) << " ";	
	cout << endl;
	
	cout << "TargetSentence: ";
	for(int i = 0; i < targetLength; i++)
		cout << G->GetTargetWord(i) << " ";	
	cout << endl;
*/
	assert(nodes.size() == 0);
	assert(numberOfNodes.size() == 0);
	for(int PTIndex = 0; PTIndex < numberOfPhraseTables;PTIndex ++) {
		vector<PhraseItem> items;
		//cout << "Searching for phrases matching the sentence" << endl;
		if(nonContinous[PTIndex] == 0) {
			SearchPhrases(G,items,&phraseTable[PTIndex]);
		} else {
			SearchDiscontinousPhrases(G,items,&phraseTable[PTIndex]);
		}
		//create feature for every phrase
		nodes.push_back(new PhraseFeature[items.size()]);
		numberOfNodes.push_back(items.size());
		for(int i = 0; i < items.size(); i++) {
/*			cout << "Found Phrase:";
			cout << items[i].source << "#" << items[i].target << endl;
			cout << "Parts: " << items[i].sourceParts.size() << "#" << items[i].targetParts.size() << endl;
			for(int j = 0; j < items[i].sourceParts.size(); j++) {
				cout << items[i].sourceParts[j] << " ";
			}
			cout << " # ";
			for(int j = 0; j < items[i].targetParts.size(); j++) {
				cout << items[i].targetParts[j] << " ";
			}
			cout << endl;
*/		
			//find hidden nodes connected to this feature node
			vector<HiddenNode *> zeroNodes;
			vector<HiddenNode *> oneNodes;
			int * oneNodeIndices;
			if(nonContinous[PTIndex] == 0) {
				oneNodeIndices = new int[items[i].sourceLength * items[i].targetLength ];
			}else {
				oneNodeIndices = new int [sourceLength + targetLength];				
			}
			//cout << "Set one Nodes" << endl;
			if(selectedOneNodes[PTIndex] == 1) {
				assert(nonContinous [PTIndex] == 0);
				for(int j = 0; j < items[i].sourceLength; j++) {
					for(int k = 0; k < items[i].targetLength; k++) {
						oneNodeIndices[j* items[i].targetLength + k] = 0;
					}
				}
				for(int j = 0; j < items[i].Alignments.size(); j++) {
					int sourceIndex = items[i].sourceIndex + items[i].Alignments[j].first;
					int targetIndex = items[i].targetIndex + items[i].Alignments[j].second;
					//cout << "Add OneNode: " << sourceIndex << " " << targetIndex << endl;
					oneNodes.push_back(G->GetNode(sourceIndex,targetIndex));
					oneNodeIndices[items[i].Alignments[j].first*items[i].targetLength + items[i].Alignments[j].second] = 1;
				}
			}else if(nonContinous[PTIndex] == 1){
				for(int j = 0; j < sourceLength + targetLength; j++) {
					oneNodeIndices[j] = 0;
				}
				assert(items[i].sourceParts.size() == items[i].sourceIndices.size());
				for(int j = 0; j < items[i].sourceParts.size(); j++) {
					//cout << "SourcePart:" << j << endl;
					for(int ls = 0; ls < items[i].sourcePartLength[j]; ls++) {
						int sourceIndex = items[i].sourceIndices[j] + ls;
						//cout << "SourceIndex:" << sourceIndex << endl;
						oneNodeIndices[sourceIndex] = 1;
						for(int k = 0; k < items[i].targetParts.size(); k++) {
							for(int lt = 0; lt < items[i].targetPartLength[k]; lt++) {
								int targetIndex = items[i].targetIndices[k] + lt;
								oneNodeIndices[sourceLength + targetIndex] = 1;
								//cout << "TargetIndex:" << targetIndex << endl;
								oneNodes.push_back(G->GetNode(sourceIndex,targetIndex));
								//cout << "Add Link: (" << sourceIndex << "," << targetIndex << ")" << endl;
							}
						}
					}
				}

			}else {
				for(int j = 0; j < items[i].sourceLength; j++) {
					for(int k = 0; k < items[i].targetLength; k++) {
						int sourceIndex = items[i].sourceIndex + j;
						int targetIndex = items[i].targetIndex + k;
						//cout << "Add OneNode: " << sourceIndex << " " << targetIndex << endl;
						oneNodes.push_back(G->GetNode(sourceIndex,targetIndex));
						oneNodeIndices[j * items[i].targetLength + k] = 1;
						
					}
				}
			}
			//cout << " Set Zero Nodes! " << endl;
			if(exactMatch[PTIndex] == 1) {
				if(nonContinous[PTIndex] == 1) {
					for(int j = 0; j < items[i].sourceParts.size(); j++) {
					for(int ls = 0; ls < items[i].sourcePartLength[j]; ls++) {
						int sourceIndex = items[i].sourceIndices[j] + ls;
						//cout << "Looking at Source Index:"  << sourceIndex << endl;
						//indices stored in sorted order
						for(int k = 0; k < targetLength; k++) {
							if(oneNodeIndices[sourceLength + k] == 0) {
								//cout << "Add Zero Node" << sourceIndex << "," << k << endl;
								zeroNodes.push_back(G->GetNode(sourceIndex,k));
							}
						}
					}
					}
					for(int k = 0; k < items[i].targetParts.size(); k++) {
					for(int lt = 0; lt < items[i].targetPartLength[k]; lt++) {
						int targetIndex = items[i].targetIndices[k] + lt;
						//cout << "Looking at Target Index:"  << targetIndex << endl;
						//indices stored in sorted order
						for(int j = 0; j < sourceLength; j++) {
							if(oneNodeIndices[j] == 0) {
								//cout << "Add Zero Node" << j << "," << targetIndex << endl;
								zeroNodes.push_back(G->GetNode(j,targetIndex));
							}
						}
					}
					}
				}else {
					for(int j = 0; j < items[i].sourceLength;j++) {
						int sourceIndex = items[i].sourceIndex + j;
						//cout << "Looking at Source Index:"  << sourceIndex << endl;
						//indices stored in sorted order
						for(int k = 0; k < targetLength; k++) {
							if(k >= items[i].targetIndex + items[i].targetLength) {
								//node has to be zero for this phrase
								//cout << "Zeronode: " << sourceIndex << "," << k << endl;
							zeroNodes.push_back(G->GetNode(sourceIndex,k));
							}else if(k < items[i].targetIndex) {
								//node has to be zero for this phrase
								//cout << "Zeronode: " << sourceIndex << "," << k << endl;
								zeroNodes.push_back(G->GetNode(sourceIndex,k));
							}else{
								//cout << "OneNode: " << sourceIndex << "," << k << endl;
								//one Nodes already done
								if(oneNodeIndices[j * items[i].targetLength + k -items[i].targetIndex] == 0) {
									zeroNodes.push_back(G->GetNode(sourceIndex,k));
								}
							}
						}
					}
					for(int k = 0; k < items[i].targetLength;k++) {
						int targetIndex = items[i].targetIndex + k;
						//cout << "Looking at Target Index:"  << targetIndex << endl;
						//indices stored in sorted order
						for(int j = 0; j < sourceLength; j++) {
							if(j >= items[i].sourceIndex + items[i].targetLength) {
								//node has to be zero for this phrase
								//cout << "Zeronode: " << j << "," << targetIndex << endl;
								zeroNodes.push_back(G->GetNode(j,targetIndex));
							}else if(j < items[i].sourceIndex) {
								//node has to be zero for this phrase
								//cout << "Zeronode: " << j << "," << targetIndex << endl;
								zeroNodes.push_back(G->GetNode(j,targetIndex));
							}
						}
					}
				}
			}
			delete[] oneNodeIndices; 
			
			if(dynamicFeatureOffSet[PTIndex + 1] - dynamicFeatureOffSet[PTIndex] > 0) {
				int dynIndex = dynamicFeatureOffSet[PTIndex] + items[i].dynamicIndex;
				double dynScalingFactor = dynamicScalingFactors[dynIndex];
				nodes[PTIndex][i].Init(zeroNodes,oneNodes,items[i].scores,items[i].dynamicScore,&ScalingFactors[2*featureOffSet[PTIndex]],dynScalingFactor,2* numberOfFeatures + dynIndex,offset + 2* featureOffSet[PTIndex],offset,numberOfAllFeatures);
			}else {
				nodes[PTIndex][i].Init(zeroNodes,oneNodes,items[i].scores,0,&ScalingFactors[2*featureOffSet[PTIndex]],0,-1,offset + 2*featureOffSet[PTIndex],offset,numberOfAllFeatures);
		
			}
		}
	}
}


  /**
   * @param newScalingFactors
   */
void PhraseFeatureSet::SetScalingFactors(vector<double> & newScalingFactors){
  assert(newScalingFactors.size() == numberOfFeatures * 2 + numberOfDynamicFeatures);
  int j = 0;
  for(int i = 0; i < 2*numberOfFeatures; i++) {
    ScalingFactors[i] = newScalingFactors[j];
    j++;
  }
 
	for(int i = 0; i < numberOfDynamicFeatures;i++) { 
		dynamicScalingFactors[i] = newScalingFactors[j];
		j++;
	}

}

void PhraseFeatureSet::RemoveFeatures(){
	while(nodes.size() > 0) {
		delete[] nodes[nodes.size() - 1];
		nodes.pop_back();
		numberOfNodes.pop_back();
	}
}	

  /**
   * @return FeatureSetIterator
   */
FeatureSetIterator * PhraseFeatureSet::GetIterator ( ) const {
	return new PhraseFeatureSetIterator(nodes,numberOfNodes);

}

  /**
   */
void PhraseFeatureSet::EstimateScalingFactores () {
	EstimatedDerivation.clear();
	vector<double> ExpectValue;
	vector<double> ExampleValue;
	ExpectValue.resize(2*numberOfFeatures + numberOfDynamicFeatures,0);
	ExampleValue.resize(2*numberOfFeatures + numberOfDynamicFeatures,0);

	for(int PTIndex = 0; PTIndex < numberOfPhraseTables; PTIndex ++) {
		for(int i = 0; i < numberOfNodes[PTIndex];i++) {
			vector<double> TempExpectValue = nodes[PTIndex][i].GetExpectationValue();
			vector<double> TempExampleValue = nodes[PTIndex][i].GetExampleValue();
			assert(TempExpectValue.size() == 2*(featureOffSet[PTIndex + 1] - featureOffSet[PTIndex]));
			assert(TempExampleValue.size() == 2*(featureOffSet[PTIndex + 1] - featureOffSet[PTIndex]));
			for(int j = 0; j < 2*(featureOffSet[PTIndex + 1] - featureOffSet[PTIndex]); j++) {
    				 //cout << "Get ExpectValue[" << j <<"]:" << TempExpectValue[j] << " New ExampleValue: " << TempExampleValue[j] <<endl;
				ExpectValue[j + 2*featureOffSet[PTIndex]] += TempExpectValue[j];
				ExampleValue[j + 2*featureOffSet[PTIndex]] += TempExampleValue[j];
				//cout << "New ExpectValue[" << j <<"]:" << ExpectValue[j] << " New ExampleValue: " << ExampleValue[j] <<endl;
			}

			if(numberOfDynamicFeatures) {
				double TempDynExpectValue = nodes[PTIndex][i].GetDynamicExpectationValue();
				double TempDynExampleValue = nodes[PTIndex][i].GetDynamicExampleValue();
				int TempDynIndex = dynamicFeatureOffSet[PTIndex] + nodes[PTIndex][i].GetDynamicSFIndex();
				ExpectValue[TempDynIndex] += TempDynExpectValue;
				ExampleValue[TempDynIndex] += TempDynExampleValue;
			}
     			//cout << "New ExpectValue[" << TempDynIndex <<"]:" << ExpectValue[TempDynIndex] << " New ExampleValue: " << ExampleValue[TempDynIndex] <<endl;
  
 		 }
	}
	for(int j = 0; j < 2*numberOfFeatures + numberOfDynamicFeatures; j++) {
		double deltaL = ExampleValue[j] - ExpectValue[j];
		//cout << "Calculating Der[" << j << "]: " << ExampleValue[j] << " - " <<  ExpectValue[j] << " = " << deltaL << endl;
		EstimatedDerivation.push_back(deltaL);
	}
}


void PhraseFeatureSet::OutputScalingFactors(){
	cout << "PhraseFeatureScalingFactors ";
	for(int i = 0; i < 2*numberOfFeatures; i++) {
		cout << ScalingFactors[i] << "_";
	}
	cout << endl;
	for(int PTIndex = 0; PTIndex < numberOfPhraseTables; PTIndex ++) {
		if(dynamicFeatureOffSet[PTIndex + 1] - dynamicFeatureOffSet[PTIndex] > 0) {
			string fileName = scalingFactorFiles[PTIndex] + ".opt";
			ofstream out(fileName.c_str());
			for(int i = dynamicFeatureOffSet[PTIndex] ; i < dynamicFeatureOffSet[PTIndex +1]; i++ ) {
				out << i  -  dynamicFeatureOffSet[PTIndex] << " " << dynamicScalingFactors[i] << endl;
			}
			out.close();
		}
	}
}


void PhraseFeatureSet::LoadScalingFactorFile(string fileName,vector<double> & ScalingFactors,int offset,int size){

	assert(size > 0);
	assert(ScalingFactors.size() == offset);
	for(int i = 0; i < size; i++)
		ScalingFactors.push_back(0);
	ifstream file( fileName.c_str());

	if (!file) {
		cerr << "ScalingFactorFile " << fileName << " could not be opened\n";
		exit(-2);
	}

	char *pIndex = new char[ 200 ];
	char *pValue = new char[ 200 ];
	while ( !file.eof() ) {
		file >> pIndex >> pValue;
		int index = atoi(pIndex);
		assert(index >= 0);
		assert(index < size);
		ScalingFactors[offset + index] = atof(pValue);
	}
	delete[] pIndex;
	delete[] pValue;
  
}
void PhraseFeatureSet::LoadPhraseTable(const char * fileName,multimap<string,PhraseItem> * phraseTable, int numberOfPTFeatures, int numberOfDynamicPTFeatures,int selectedOneNodesForPT,int nonContinous) {
	const int BufferSize = 1024;
	char Line [BufferSize];
	ifstream file( fileName);

	if (!file) {
		cerr << "PhraseTable " << fileName << " could not be opened\n";
		exit(-2);
	}
	
	while(file.getline(Line,BufferSize)) {
		string line(Line);
		vector<string> elements;
		Utils::Tokenize(line,elements,string("#"));
		assert(elements.size() >= numberOfPTFeatures + 2);
		PhraseItem item;

		//remove leading and endling blanks from source string
		vector<string> temp;
		Utils::Tokenize(elements[0],temp);
		item.sourceLength = temp.size();
		item.source = temp[0];
		for(int i = 1; i < temp.size(); i++) {
			item.source.push_back(' ');
			item.source.append(temp[i]);
		}
		temp.clear();

		//remove leading and ending blanks from target string
		Utils::Tokenize(elements[1],temp);
		item.targetLength = temp.size();			
		item.target = temp[0];
		for(int i = 1; i < temp.size(); i++) {
			item.target.push_back(' ');
			item.target.append(temp[i]);
		}

		int i = 0;
		for(; i < numberOfPTFeatures; i++) {
			item.scores.push_back(atof(elements[2+i].c_str()));
		}
		if(numberOfDynamicPTFeatures) {
			vector<string> dynamicScore;
			Utils::Tokenize(elements[2+i],dynamicScore);
			assert(dynamicScore.size() == 2);
			item.dynamicIndex = atoi(dynamicScore[0].c_str());
			assert(item.dynamicIndex > 0);
			assert(item.dynamicIndex < numberOfDynamicPTFeatures);
			item.dynamicScore = atof(dynamicScore[1].c_str());
			i++;
		}
		if(selectedOneNodesForPT) {
			vector<string> links;
			Utils::Tokenize(elements[2+i],links);
			for(int j = 0; j < links.size(); j++) {
				int a;
				int b;
				char c;
				sscanf(links[j].c_str(),"%i%c%i",&a,&c,&b);
				item.Alignments.push_back(make_pair(a-1,b-1));
			}
			i++;
		}
		//cout << "i: " << i << endl;
		//cout << "elementes.size(): " << elements.size() << endl;
		assert(elements.size() == i +2);
		//item.currentWord = 0;
/*		cout << "Found Phrase : " << elements[0] << elements[1];
		for(int j = 0; j < item.Alignments.size(); j++) {
			cout << item.Alignments[j].first <<"-" << item.Alignments[j].second << " ";
		}
		cout << endl;*/
		if(nonContinous == 0) {
			phraseTable->insert(make_pair(item.source,item));
		}else {
			//cout << "Source: " << item.source << endl;
			Utils::Tokenize(item.source,item.sourceParts,string("|"));
			for(int i = 0; i < item.sourceParts.size(); i++) {
				//cout << "Part[" << i << "]: " << item.sourceParts[i] << endl;
				//Trim spaces
				// Find the first character position after excluding leading blank spaces
				size_t startpos = item.sourceParts[i].find_first_not_of(" \t"); 
			 	size_t endpos = item.sourceParts[i].find_last_not_of(" \t"); // Find the first character position from reverse af	
				// if all spaces or empty return an empty string
				if(( string::npos == startpos ) || ( string::npos == endpos)) {
					cout << "Empty part of phrase" << endl;
					exit(1);
				}else {
				        item.sourceParts[i] = item.sourceParts[i].substr( startpos, endpos-startpos+1 );
				}
				temp.clear();	
				Utils::Tokenize(item.sourceParts[i],temp);
				item.sourcePartLength.push_back(temp.size());			
				//cout << "Size: " << temp.size() << endl;
			}
			Utils::Tokenize(item.target,item.targetParts,string("|"));
			for(int i = 0; i < item.targetParts.size(); i++) {
				//Trim spaces
				// Find the first character position after excluding leading blank spaces
				size_t startpos = item.targetParts[i].find_first_not_of(" \t"); 
			 	size_t endpos = item.targetParts[i].find_last_not_of(" \t"); // Find the first character position from reverse af	
				// if all spaces or empty return an empty string
				if(( string::npos == startpos ) || ( string::npos == endpos)) {
					cout << "Empty part of phrase" << endl;
					exit(1);
				}else {
				        item.targetParts[i] = item.targetParts[i].substr( startpos, endpos-startpos+1 );
				}
				temp.clear();	
				Utils::Tokenize(item.targetParts[i],temp);
				item.targetPartLength.push_back(temp.size());			

			}
	
			item.index = 0;
			phraseTable->insert(make_pair(item.sourceParts[0],item));	
		}

		//cout << "Insert item: |" << item.source << "|" << item.target << "|" << endl;
	}
}

void PhraseFeatureSet::SearchPhrases(Graph * G,vector<PhraseItem> & items,multimap<string,PhraseItem> * currentPhraseTable) {
	
	assert(items.size() == 0);
	multimap<string,PhraseItem> targetTable;

	int maxPhraseLength = 4;

	//find matching source phrases
	int sourceLength = G->GetSourceLength();
	for(int i = 0; i < sourceLength; i++) {
		string source = "";
		for(int j = 0; (j < maxPhraseLength) && (i + j < sourceLength); j++) {
			if(j != 0) {
				source.push_back(' ');
			}
			source.append(G->GetSourceWord(i + j));
			//search for phrases starting at this position
			pair<PTIterator,PTIterator> iter = currentPhraseTable->equal_range(source);
			for(PTIterator begin = iter.first; begin != iter.second; begin++) {
				PhraseItem item = begin->second;
				item.sourceIndex = i;
				targetTable.insert(make_pair(item.target,item));
			}

		}
	}

	//check if target phrases match also
	int targetLength = G->GetTargetLength();
	for(int i = 0; i < targetLength; i++) {
		string target = "";
		for(int j = 0; (j < maxPhraseLength) && (i + j < targetLength); j++) {
			if(j != 0) {
				target.push_back(' ');
			}
			target.append(G->GetTargetWord(i + j));

			//check matching target phrases
			pair<PTIterator,PTIterator> iter = targetTable.equal_range(target);
			for(PTIterator begin = iter.first; begin != iter.second; begin++) {
				PhraseItem item = begin->second;
				item.targetIndex = i;
				//phrase matches complete
				items.push_back(item);
				//cout << "Found item: "<< item.source << " # " << item.target << endl;
			}
		}
	}



}

void PhraseFeatureSet::SearchDiscontinousPhrases(Graph * G,vector<PhraseItem> & items,multimap<string,PhraseItem> * currentPhraseTable) {
	assert(items.size() == 0);

	multimap<string,PhraseItem> targetTable;
	int maxPhraseLength = 4;
	vector<vector<PhraseItem> > sourceNextPart;
	multimap<string,PhraseItem> sourceContinue;

	//find matching source phrases
	int sourceLength = G->GetSourceLength();
	vector<PhraseItem> e;
	sourceNextPart.assign(sourceLength,e);

	for(int i = 0; i < sourceLength; i++) {
		string source = "";
		for(int j = 0; j < sourceNextPart[i].size(); j++) {
			sourceContinue.insert(make_pair(sourceNextPart[i][j].sourceParts[sourceNextPart[i][j].index],sourceNextPart[i][j]));
		}
		for(int j = 0; (j < maxPhraseLength) && (i + j < sourceLength); j++) {
			if(j != 0) {
				source.push_back(' ');
			}
			source.append(G->GetSourceWord(i + j));
			//search for phrases starting at this position
			pair<PTIterator,PTIterator> iter = currentPhraseTable->equal_range(source);
			for(PTIterator begin = iter.first; begin != iter.second; begin++) {
				PhraseItem item = begin->second;
				if(item.sourceParts.size() == item.index + 1) {
					item.sourceIndices.push_back(i);
					item.index = 0;
					assert(item.sourceIndices.size() == item.sourceParts.size());
					targetTable.insert(make_pair(item.targetParts[0],item));
				}else if(i + j + 1 < sourceLength) {
					item.sourceIndices.push_back(i);
					item.index ++;
					sourceNextPart[i + j + 1].push_back(item);
				}
			}
			//search for phrases starting at this position
			iter = sourceContinue.equal_range(source);
			for(PTIterator begin = iter.first; begin != iter.second; begin++) {
				PhraseItem item = begin->second;
				if(item.sourceParts.size() == item.index + 1) {
					item.sourceIndices.push_back(i);
					item.index = 0;
					assert(item.sourceIndices.size() == item.sourceParts.size());
					targetTable.insert(make_pair(item.targetParts[0],item));
				}else if(i + j + 1 < sourceLength){
					item.sourceIndices.push_back(i);
					item.index ++;
					sourceNextPart[i + j + 1].push_back(item);
				}
			}

		}
	}

	vector<vector<PhraseItem> > targetNextPart;
	multimap<string,PhraseItem> targetContinue;

	//check if target phrases match also
	int targetLength = G->GetTargetLength();
	targetNextPart.assign(targetLength,e);
	for(int i = 0; i < targetLength; i++) {
		string target = "";
		for(int j = 0; j < targetNextPart[i].size(); j++) {
			targetContinue.insert(make_pair(targetNextPart[i][j].targetParts[targetNextPart[i][j].index],targetNextPart[i][j]));
		}
		for(int j = 0; (j < maxPhraseLength) && (i + j < targetLength); j++) {
			if(j != 0) {
				target.push_back(' ');
			}
			target.append(G->GetTargetWord(i + j));
			//check matching target phrases
			pair<PTIterator,PTIterator> iter = targetTable.equal_range(target);
			for(PTIterator begin = iter.first; begin != iter.second; begin++) {
				PhraseItem item = begin->second;
				item.targetIndices.push_back(i);
				if(item.targetParts.size() == item.index + 1) {
					item.index = 0;
					assert(item.targetIndices.size() == item.targetParts.size());
					items.push_back(item);
				}else if(i + j + 1 < targetLength){
					item.index ++;
					targetNextPart[i + j + 1].push_back(item);
				}
				//phrase matches complete
				//cout << "Found item: "<< item.source << " # " << item.target << endl;
			}
			iter = targetContinue.equal_range(target);
			for(PTIterator begin = iter.first; begin != iter.second; begin++) {
				PhraseItem item = begin->second;
				item.targetIndices.push_back(i);
				if(item.targetParts.size() == item.index + 1) {
					item.index = 0;
					assert(item.targetIndices.size() == item.targetParts.size());
					items.push_back(item);
				}else if(i + j + 1 < targetLength){
					item.index ++;
					targetNextPart[i + j + 1].push_back(item);
				}
				//phrase matches complete
				//cout << "Found item: "<< item.source << " # " << item.target << endl;
			}
		}
	}

}


