#ifndef PHRASEFEATURESET_H
#define PHRASEFEATURESET_H

#include <fstream>
#include <featureset.h>
#include <map>
#include <phrasefeature.h>
#include <phrasefeaturesetiterator.h>
#include <utils.h>
/**
	@author Jan Niehues <jan@niehues.info>
*/


struct PhraseItem {
	string source;
	string target;
	vector<string> sourceParts;
	vector<string> targetParts;
	int index;
	int sourceIndex;
	int targetIndex;
	int sourceLength;
	int targetLength;
	vector<int> sourceIndices;
	vector<int> targetIndices;
	vector<int> sourcePartLength;
	vector<int> targetPartLength;
	vector<double> scores;
	int dynamicIndex;
	double dynamicScore;
	vector<pair<int,int> > Alignments; 

};
typedef multimap<string,PhraseItem>::iterator PTIterator;

class PhraseFeatureSet : public FeatureSet
{
private:
	vector<double> EstimatedDerivation;
	multimap<string,PhraseItem> * phraseTable;
	int numberOfPhraseTables;
	int numberOfFeatures;
	int numberOfDynamicFeatures;
	int * featureOffSet;
	int * dynamicFeatureOffSet;
	vector<int> selectedOneNodes;
	vector<int> exactMatch;
	vector<int> nonContinous;
	vector<double> dynamicScalingFactors;
	double * ScalingFactors;
	vector<PhraseFeature *> nodes;
	vector<int> numberOfNodes;
	vector<string> scalingFactorFiles;

	void LoadParamFile(const char* ParamFile);
	void LoadScalingFactorFile(string fileName,vector<double> & ScalingFactors,int offset,int size);
	void LoadPhraseTable(const char * fileName,multimap<string,PhraseItem> * phraseTable, int numberOfPTFeatures, int numberOfDynamicPTFeatures,int selectedOneNodesForPT,int nonContinous);
	void InitScalingFactors(vector<double> & Factors);
	void SearchPhrases(Graph * G,vector<PhraseItem> & items,multimap<string,PhraseItem> * phraseTable);
	void SearchDiscontinousPhrases(Graph * G,vector<PhraseItem> & items,multimap<string,PhraseItem> * currentPhraseTable);
public:
    PhraseFeatureSet(const char* ParamFile, vector<double> & ScalingFactores);

    ~PhraseFeatureSet();
  /**
   * @param  G
   */
  void GenerateFeatures (Graph * G  , int offset, int numberOfAllFeatures);

  /**
   * @param newScalingFactors
   */
  void SetScalingFactors(vector<double> & newScalingFactors);

  void RemoveFeatures();

  /**
   * @return FeatureSetIterator
   */
  FeatureSetIterator * GetIterator ( ) const;

  /**
   */
  void EstimateScalingFactores ();

  vector<double> inline GetEstimatedDerivations() const {return EstimatedDerivation;};

  virtual void OutputScalingFactors();

};

#endif
