
#include "phraselopfeature.h"

PhraseLOPFeature::PhraseLOPFeature()
 : Feature()
{
	numberOfExperts = 0;
	numberOfFeatureValues = 0;

}


PhraseLOPFeature::~PhraseLOPFeature()
{
	//remove derivations
	for(map<const HiddenNode *,SparseArray>::iterator iter = derivations.begin(); iter != derivations.end(); iter++) {
		delete[] iter->second.data;
		delete[] iter->second.index;
	}

	if(numberOfFeatureValues)
		delete[] scaledFeatureValues;

}



void PhraseLOPFeature::Init(vector<HiddenNode *> & zeroN, vector<HiddenNode *> & oneN, vector<double> & values, double dynamicValue, double * sFactors, double dynScalingFactor, int dynSFIndex,int off, int dynOff, int nOfAllFeatures, int numberOfExp, double * expWeights) {

	numberOfExperts = numberOfExp;
	expertWeights = expWeights;

	zeroNodes = zeroN;
	//cout << "NumberOfZeroNodes: " << zeroNodes.size() << endl;
	oneNodes = oneN;
	numberOfFeatureValues = values.size();
	scalingFactors = sFactors;
	dynamicFeatureValue = dynamicValue;
	dynamicScalingFactor = dynScalingFactor;
	dynamicSFIndex = dynSFIndex;
	numberOfAllFeatures = nOfAllFeatures;
	offset = off;
	dynamicOffset = dynOff;



	scaledFeatureValues = new double[2*numberOfExperts];
	for(int i = 0; i< 2*numberOfExperts; i++) {
		scaledFeatureValues[i] = 0;
	}

	for(int  i = 0; i < numberOfFeatureValues; i++) {
		//cout << "Values [" << i << "] = " << values[i] << endl;
		for(int j = 0; j < numberOfExperts; j++) {
			//cout << "Scaling Factors for expert " << j << ": = ";
			//cout << scalingFactors[2*i*numberOfExperts + j] << " " << scalingFactors[(2*i+1)*numberOfExperts + j] << endl;
			scaledFeatureValues[j] += (1 - values[i]) * scalingFactors[2*i*numberOfExperts + j];
			scaledFeatureValues[numberOfExperts + j] += values[i] * scalingFactors[(2*i+1)*numberOfExperts + j];
		}
	}
	if(dynamicValue != 0) {
		//has to get several dynamic scaling Factors, one for every expert
		assert(0);
		for(int j = 0; j < numberOfExperts;j++) {
			scaledFeatureValues[numberOfExperts + j] += dynamicValue * dynScalingFactor;
		}
	}
/*
	for(int i = 0; i  < numberOfExperts; i++) {
		cout << "ScaledFeatureValues[" << i << " ] = ";
		cout << scaledFeatureValues[i] << " " << scaledFeatureValues[numberOfExperts + i] << endl;
	}
*/	featureValueCorrect = 0;
	featureValueFalse = 0;
	for(int i = 0; i < numberOfExperts; i++) {
		featureValueFalse += expertWeights[i]*scaledFeatureValues[i];
		featureValueCorrect += expertWeights[i] * scaledFeatureValues[numberOfExperts + i];
	}
	featureValueCorrect = exp(featureValueCorrect);
	featureValueFalse = exp(featureValueFalse);

//	cout << "featureValueFalse: " << featureValueFalse << endl;
//	cout << "featureValueCorrect: "<< featureValueCorrect << endl;

	for(int i = 0; i < zeroNodes.size(); i++) {
		zeroNodes[i]->AddConnection(this);
		//messages.insert(make_pair(zeroNodes[i],make_pair(0.5,0.5)));
		messages.insert(make_pair(zeroNodes[i],make_pair(featureValueCorrect/(featureValueCorrect+featureValueFalse),featureValueFalse/(featureValueCorrect + featureValueFalse))));
	}
	for(int i = 0; i < oneNodes.size(); i++) {
		oneNodes[i]->AddConnection(this);
		//messages.insert(make_pair(oneNodes[i],make_pair(0.5,0.5)));
		messages.insert(make_pair(oneNodes[i],make_pair(featureValueFalse/(featureValueCorrect+featureValueFalse),featureValueCorrect/(featureValueCorrect + featureValueFalse))));
	}


	CalcExampleValues();
}


pair<double,double> PhraseLOPFeature::GetMessage (const HiddenNode * ToNode ) const {
	//return messages[ToNode];
	map<const HiddenNode *,pair<double,double> >::const_iterator iter = messages.find(ToNode);
	if(iter != messages.end()) {
		//cout << "PhraseFeature send message: (" << iter->second.first << "," << iter->second.second << ")" << endl;
		return iter->second;
	}else {
		cerr << "Asking for message to HiddenNode, that not exists" << endl;
		exit(-1);
	}


}

void PhraseLOPFeature::CalcMessages() {
	//cout << "CalcMessages():" << endl;
	// some values have to be precalculated
	double allProb = 1;
	double correctProb = 1;
	for(int i = 0; i < zeroNodes.size(); i++) {
		pair<double,double> m = zeroNodes[i]->GetMessage(this);
		//cout << "ZeroNode:" << m.first << endl;
		allProb *= (m.first + m.second);
		correctProb *= m.first;
	}	
	for(int i = 0; i < oneNodes.size(); i++) {
		pair<double,double> m = oneNodes[i]->GetMessage(this);
		//cout << "OneNode:" << m.second << endl;
		allProb *= (m.first + m.second);
		correctProb *= m.second;
	}	

	for(int i = 0; i < zeroNodes.size(); i++) {
		pair<double,double> m = zeroNodes[i]->GetMessage(this);
		pair<double,double> message = make_pair(0,0);
		
		// calc message for this node
		if(correctProb > 0) {
			assert(m.first > 0);
			message.first = featureValueCorrect * correctProb /m.first + featureValueFalse * (allProb/(m.first + m.second) - correctProb / m.first);
		}else {
			message.first = featureValueFalse * allProb/(m.first + m.second);
		}
		message.second = featureValueFalse * allProb/(m.first + m.second);
		//normalize message
		double sum = message.first + message.second;

		if((isnan(message.first/sum))||(isnan(message.second/sum))) {
			cout << "In Message: " << m.first << "," << m.second << endl;
			cout << "Out Message:" << message.first << "," << message.second << endl;
			cout << "Sum: " << sum << endl;
			cout << "Prob: " << correctProb << endl;
			cout << "ScaledFeatureValues: ";
			for(int i = 0; i < 2*numberOfExperts; i++) {
				cout << scaledFeatureValues[i] << " ";
			}
			cout << endl;
			cout << "Scaling Factors: " ;
			for(int i = 0; i < 2*numberOfFeatureValues*numberOfExperts; i++) {
				cout << scalingFactors[i] << " ";
			}
			cout << endl;
		}
		message.first /= sum;
		message.second /= sum;
		assert(!isnan(message.first));
		//save message
		messages.erase(zeroNodes[i]);
		messages.insert(make_pair(zeroNodes[i],message));
		
	}
	for(int i = 0; i < oneNodes.size(); i++) {
		pair<double,double> m = oneNodes[i]->GetMessage(this);
		pair<double,double> message = make_pair(0,0);
		
		// calc message for this node
		if(correctProb > 0) {
			assert(m.second > 0);
			message.second = featureValueCorrect * correctProb /m.second + featureValueFalse * (allProb/(m.first + m.second) - correctProb / m.second);
		}else {
			message.second = featureValueFalse * allProb/(m.first + m.second);
		}
		message.first = featureValueFalse * allProb/(m.first + m.second);

		//normalize message
		double sum = message.first + message.second;
		if((isnan(message.first/sum))||(isnan(message.second/sum))) {
			cout << "In Message: " << m.first << "," << m.second << endl;
			cout << "Out Message:" << message.first << "," << message.second << endl;
			cout << "Sum: " << sum << endl;
			cout << "FeatureValues: " << featureValueCorrect << " " << featureValueFalse << endl;
			cout << "Prob: " << correctProb << endl;
			cout << "ScaledFeatureValues: ";
			for(int i = 0; i < 2*numberOfExperts; i++) {
				cout << scaledFeatureValues[i] << " ";
			}
			cout << endl;
			cout << "Scaling Factors: " ;
			for(int i = 0; i < 2*numberOfFeatureValues*numberOfExperts; i++) {
				cout << scalingFactors[i] << " ";
			}
			cout << endl;
		}
		message.first /= sum;
		message.second /= sum;
		assert(!isnan(message.first));
		//cout << "One Node Out: (" << message.first << "," << message.second << ")" << endl;;
		//cout << " -> (" << message.first << "," << message.second << ")" << endl;
		
		//save message
		messages.erase(oneNodes[i]);
		messages.insert(make_pair(oneNodes[i],message));
		
	}
}

  /**
   * @return vector<double>
   */
vector<double> PhraseLOPFeature::GetExpectationValue() {
	vector<double> expValues;
	
	int active = 1;
	// some values have to be precalculated
	double correctProb = 1;
	for(int i = 0; i < zeroNodes.size(); i++) {
		pair<double,double> m = zeroNodes[i]->GetBelief();
		active &= (m.first > m.second);
		//cout << "ZeroNode:" << m.first << endl;
		correctProb *= m.first;
	}	
	for(int i = 0; i < oneNodes.size(); i++) {
		pair<double,double> m = oneNodes[i]->GetBelief();
		//cout << "OneNode:" << m.second << endl;
		active &= (m.second > m.first);
		correctProb *= m.second;
	}

	//cout << "CorrectProb: " << correctProb << endl;
	for(int i = 0; i < numberOfExperts; i++) {
		expValues.push_back((1 - correctProb) * scaledFeatureValues[i] + correctProb * scaledFeatureValues[numberOfExperts + i]);
	}
	return expValues;
}

void PhraseLOPFeature::CalcExampleValues() {
	int active = 1;
	for(int i = 0; active && (i < zeroNodes.size()) ; i++) {
		active &= (zeroNodes[i]->GetGivenValue() == 0);
	}
	for(int i = 0; active && (i < oneNodes.size()) ; i++) {
		active &= (oneNodes[i]->GetGivenValue() == 1);
	}
	if(active) {
		for(int i = 0; i < numberOfExperts; i++) {
			exampleValue.push_back(scaledFeatureValues[numberOfExperts + i]);
		}
	}else {
		for(int i = 0; i < numberOfExperts; i++) {
			exampleValue.push_back(scaledFeatureValues[i]);
		}
	}
}


SparseArray PhraseLOPFeature::GetDerivation (const HiddenNode * ToNode) const {
	map<const HiddenNode *,SparseArray >::const_iterator iter = derivations.find(ToNode);
	if(iter != derivations.end()) {
		//cout << "PhraseFeature send message: (" << iter->second.first << "," << iter->second.second << ")" << endl;
		return iter->second;
	}else {
		cerr << "Asking for message to HiddenNode, that not exists" << endl;
		exit(-1);
	}


}

void PhraseLOPFeature::CalcDerivations() {

	// dynmamic derivation

	// some values have to be precalculated
	double allProb = 1;
	double correctProb = 1;
	for(int i = 0; i < zeroNodes.size(); i++) {
		pair<double,double> m = zeroNodes[i]->GetMessage(this);
		//cout << "ZeroNode:" << m.first << endl;
		allProb *= (m.first + m.second);
		correctProb *= m.first;
	}	
	for(int i = 0; i < oneNodes.size(); i++) {
		pair<double,double> m = oneNodes[i]->GetMessage(this);
		//cout << "OneNode:" << m.second << endl;
		allProb *= (m.first + m.second);
		correctProb *= m.second;
	}	

	//cout << "AllProb: " << allProb;
	//cout << " CorrectProb:" << correctProb << endl;

	//for every zero Node
	for(int i = 0; i < zeroNodes.size(); i++) {
		pair<double,double> m = zeroNodes[i]->GetMessage(this);
		SparseArray derivation;
		derivation.size = numberOfExperts;
		derivation.data = new double[derivation.size];
		derivation.index = new int[derivation.size];

		//individual node precalculation
		double t1 = featureValueCorrect * correctProb /m.first;
		double t2 = featureValueFalse * (allProb/(m.first + m.second) - correctProb / m.first);
		double t3 = featureValueFalse * allProb/(m.first + m.second);
		double m0 = t1 + t2;
		double m1 = t3;
		double quotient = (m0 + m1)*(m0 + m1);
		//derivation calculationen
		for(int j = 0; j < numberOfExperts; j++) {
			assert(j < derivation.size);
			derivation.index[j] = j;
			derivation.data[j] = (t1*scaledFeatureValues[numberOfExperts + j])*m1;
			derivation.data[j] = ((t2*scaledFeatureValues[j])*m1 - (t3*scaledFeatureValues[j])*m0);
			derivation.data[j] /= quotient;
		}

		//insert into map
		derivations.insert(make_pair(zeroNodes[i],derivation));	
	}
	//for every one Node
	//cout << "OneNodes: " << endl;
	for(int i = 0; i < oneNodes.size(); i++) {
		pair<double,double> m = oneNodes[i]->GetMessage(this);
		//cout << "Message:" << m.first << "," << m.second << endl; 
	
		SparseArray derivation;
		derivation.size = numberOfExperts;
		derivation.data = new double[derivation.size];
		derivation.index = new int[derivation.size];

		//individual node precalculation
		double t1 = featureValueCorrect * correctProb /m.first;
		double t2 = featureValueFalse * (allProb/(m.first + m.second) - correctProb / m.first);
		double t3 = featureValueFalse * allProb/(m.first + m.second);
		double m1 = t1 + t2;
		double m0 = t3;
		double quotient = (m0 + m1)*(m0 + m1);
		//cout << "featureValueCorrect:" << featureValueCorrect << " False:" << featureValueFalse << endl;
		//cout << "t: " << t1 <<" " << t2 << " "<< t3 << endl;
		//cout << "m: " << m0 << " " << m1 << " Quotient :" << quotient << endl;
	
		//derivation calculationen
		for(int j = 0; j < numberOfExperts; j++) {
			assert(j < derivation.size);
			derivation.index[j] = j;
			derivation.data[j] = - (t1*scaledFeatureValues[numberOfExperts + j])*m0;
			derivation.data[j] = ((t3*scaledFeatureValues[j])*m1 - (t2*scaledFeatureValues[j])*m0);
			derivation.data[j] /= quotient;
		}
		//insert into map
		derivations.insert(make_pair(oneNodes[i],derivation));	
	}

}
