#ifndef PHRASELOPFEATURE_H
#define PHRASELOPFEATURE_H

#include <feature.h>
#include <map>

/**
	@author Jan Niehues <jan@niehues.info>
*/
class PhraseLOPFeature : public Feature
{
private:

	int numberOfExperts;
	double * expertWeights;
	map<const HiddenNode*,pair<double,double> > messages;

	vector<HiddenNode *> zeroNodes;
	vector<HiddenNode *> oneNodes;
	int numberOfFeatureValues;
	double * scaledFeatureValues;
	double * scalingFactors;
	double dynamicFeatureValue;
	double dynamicScalingFactor;
	int dynamicSFIndex;
	int numberOfAllFeatures;
	int offset;
	int dynamicOffset;
	double featureValueCorrect;
	double featureValueFalse;

	vector<double> exampleValue;

	map<const HiddenNode*,SparseArray> derivations;

	void CalcExampleValues();


public:
    PhraseLOPFeature();

    ~PhraseLOPFeature();
     void Init(vector<HiddenNode *> & zeroNodes, vector<HiddenNode *> & oneNodes, vector<double> & values, double dynamicValue, double * ScalingFactors, double dynamicScalingFactors, int dynamicSFIndex, int offset, int dynOff, int numberOfAllFeatures, int numberOfExp, double * expWeights);

	pair<double,double> GetMessage(const HiddenNode * ToNode) const;

	void CalcMessages();


  /**
   * @return vector<double>
   */
  vector<double> GetExpectationValue();

  /**
   * @return double
   */
  inline vector<double> GetExampleValue() {return exampleValue;};


  SparseArray GetDerivation (const HiddenNode * ToNode) const;

void CalcDerivations();


};

#endif
