#include <string.h>
#include "phraselopfeatureset.h"

PhraseLOPFeatureSet::PhraseLOPFeatureSet(const char* ParamFile, const char * ScalingFactorString,int numberOfExp, double * expWeights)
 : FeatureSet()
{
	numberOfExperts = numberOfExp;
	expertWeights = expWeights;
	assert(ParamFile != NULL);
	numberOfDynamicFeatures = 0;
	numberOfPhraseTables = 0;
	numberOfFeatures = 0;

	LoadParamFile(ParamFile);


	ScalingFactors = new double[2 * numberOfFeatures *numberOfExperts];
	cout << "NumberOfFeatures:" << numberOfFeatures << endl;
	cout << "NumberOfExperts: " << numberOfExperts << endl;
	Utils::Extract2DScalingFactors(ScalingFactorString,ScalingFactors,2 * numberOfFeatures,numberOfExperts);

/*	cout << "ScalingFactors:";
	for(int i = 0;i < 2*numberOfFeatures*numberOfExperts; i++) {
		cout << ScalingFactors[i] << " ";
	}
	cout << endl;
*/
}


PhraseLOPFeatureSet::~PhraseLOPFeatureSet()
{
	if(numberOfPhraseTables) {
		delete[] phraseTable;
		delete[] featureOffSet;
		delete[] dynamicFeatureOffSet;
	}
	delete[] ScalingFactors;
}


void PhraseLOPFeatureSet::LoadParamFile(const char* ParamFileName) {

	vector<string> phraseTableFiles;	
	vector<int> numberOfPTFeatures;
	vector<int> numberOfDynamicPTFeatures;


	ifstream ParamFile( ParamFileName);

	if (!ParamFile) {
		cerr << "File " << ParamFileName << " could not be opened\n";
		exit(-2);
	}
  
	char *pKey = new char[ 200 ];
	char *pValue = new char[ 200 ];
// TODO: Does not work with LF.Param with empty line at end!!!!!!!!!!!!
	while ( !ParamFile.eof() ) {
		ParamFile >> pKey >> pValue;
		if(strcmp(pKey,"PhraseTable") == 0) {
			phraseTableFiles.push_back((pValue));
			scalingFactorFiles.push_back("");
			numberOfPTFeatures.push_back(0);
			numberOfDynamicPTFeatures.push_back(0);
			selectedOneNodes.push_back(0);
			exactMatch.push_back(0);
		}else if(strcmp(pKey,"ScalingFactorFile") == 0) {
			scalingFactorFiles[scalingFactorFiles.size() -1] = pValue;
		}else if(strcmp(pKey, "NumberOfDynamicFeatures") == 0) {
			numberOfDynamicPTFeatures[numberOfDynamicPTFeatures.size() -1 ] = atoi(pValue);
		}else if(strcmp(pKey, "NumberOfFeatures") == 0) {
			numberOfPTFeatures[numberOfPTFeatures.size() -1 ] = atoi(pValue);
		}else if(strcmp(pKey, "SelectedOneNodes") == 0) {
			selectedOneNodes[selectedOneNodes.size() - 1] = atoi(pValue);
		}else if(strcmp(pKey, "ExactMatch") == 0) {
			exactMatch[exactMatch.size() -1] = atoi(pValue);
		}



	}

	assert(phraseTableFiles.size() == numberOfPTFeatures.size());
	assert(phraseTableFiles.size() == numberOfDynamicPTFeatures.size());
	assert(phraseTableFiles.size() == scalingFactorFiles.size());
	assert(phraseTableFiles.size() == selectedOneNodes.size());
	assert(phraseTableFiles.size() == exactMatch.size());

	ParamFile.close();
	delete[] pKey;
	delete[] pValue;

	phraseTable = new multimap<string,PhraseItem>[phraseTableFiles.size()];
	numberOfPhraseTables = phraseTableFiles.size();
	featureOffSet = new int[numberOfPhraseTables + 1];
	dynamicFeatureOffSet = new int[numberOfPhraseTables + 1];

	int fOffSet = 0;
	int dynOffSet = 0;
	for(int i = 0; i < numberOfPhraseTables; i++) {

		//loading a scaling factor file for every expert not yet implemented
		assert(numberOfDynamicPTFeatures[i] == 0);

		LoadPhraseTable(phraseTableFiles[i].c_str(),&phraseTable[i],numberOfPTFeatures[i],numberOfDynamicPTFeatures[i],selectedOneNodes[i]);
		assert(phraseTable[i].size() != 0);

		if(scalingFactorFiles[i] != "") {
			cout << "Using additional Phraseindicator features for phrase table "<< i << endl;
			cout << "ScalingFactorFile" << scalingFactorFiles[i] << endl;
			LoadScalingFactorFile(scalingFactorFiles[i],dynamicScalingFactors,dynOffSet,numberOfDynamicPTFeatures[i]);
		}else {
			assert(numberOfDynamicFeatures == 0);	
		}
		featureOffSet[i] = fOffSet;
		dynamicFeatureOffSet[i] = dynOffSet;
		fOffSet += numberOfPTFeatures[i];
		dynOffSet += numberOfDynamicPTFeatures[i];
		cout << "numberOfPTFeatures["<< i<< "] = " << numberOfPTFeatures[i] << endl;
		cout << "fOffSet: " << fOffSet  << endl; 
	}
	numberOfDynamicFeatures = dynOffSet;
	numberOfFeatures = fOffSet;
	featureOffSet[numberOfPhraseTables] = fOffSet;
	dynamicFeatureOffSet[numberOfPhraseTables] = dynOffSet;

}
  /**
   * @param  G
   */
void PhraseLOPFeatureSet::GenerateFeatures (Graph * G  , int offset, int numberOfAllFeatures) {
	int sourceLength = G->GetSourceLength();
	int targetLength = G->GetTargetLength();
	assert(nodes.size() == 0);
	assert(numberOfNodes.size() == 0);
	for(int PTIndex = 0; PTIndex < numberOfPhraseTables;PTIndex ++) {
		vector<PhraseItem> items;
		//cout << "Searching for phrases matching the sentence" << endl;
		SearchPhrases(G,items,&phraseTable[PTIndex]);

		//create feature for every phrase
		nodes.push_back(new PhraseLOPFeature[items.size()]);
		numberOfNodes.push_back(items.size());
		for(int i = 0; i < items.size(); i++) {

			//find hidden nodes connected to this feature node
			vector<HiddenNode *> zeroNodes;
			vector<HiddenNode *> oneNodes;
			int oneNodeIndices[items[i].sourceLength][items[i].targetLength];

			if(selectedOneNodes[PTIndex] == 1) {
				for(int j = 0; j < items[i].sourceLength; j++) {
					for(int k = 0; k < items[i].targetLength; k++) {
						oneNodeIndices[j][k] = 0;
					}
				}
				for(int j = 0; j < items[i].Alignments.size(); j++) {
					int sourceIndex = items[i].sourceIndex + items[i].Alignments[j].first;
					int targetIndex = items[i].targetIndex + items[i].Alignments[j].second;
					//cout << "Add OneNode: " << sourceIndex << " " << targetIndex << endl;
					oneNodes.push_back(G->GetNode(sourceIndex,targetIndex));
					oneNodeIndices[items[i].Alignments[j].first][items[i].Alignments[j].second] = 1;
				}
			}else {
				for(int j = 0; j < items[i].sourceLength; j++) {
					for(int k = 0; k < items[i].targetLength; k++) {
						int sourceIndex = items[i].sourceIndex + j;
						int targetIndex = items[i].targetIndex + k;
						//cout << "Add OneNode: " << sourceIndex << " " << targetIndex << endl;
						oneNodes.push_back(G->GetNode(sourceIndex,targetIndex));
						oneNodeIndices[j][k] = 1;
						
					}
				}
			}
			if(exactMatch[PTIndex] == 1) {
				for(int j = 0; j < items[i].sourceLength;j++) {
					int sourceIndex = items[i].sourceIndex + j;
					//cout << "Looking at Source Index:"  << sourceIndex << endl;
					//indices stored in sorted order
					for(int k = 0; k < targetLength; k++) {
						if(k >= items[i].targetIndex + items[i].targetLength) {
							//node has to be zero for this phrase
							//cout << "Zeronode: " << sourceIndex << "," << k << endl;
						zeroNodes.push_back(G->GetNode(sourceIndex,k));
						}else if(k < items[i].targetIndex) {
							//node has to be zero for this phrase
							//cout << "Zeronode: " << sourceIndex << "," << k << endl;
							zeroNodes.push_back(G->GetNode(sourceIndex,k));
						}else{
							//cout << "OneNode: " << sourceIndex << "," << k << endl;
							//one Nodes already done
							if(oneNodeIndices[j][k-items[i].targetIndex] == 0) {
								zeroNodes.push_back(G->GetNode(sourceIndex,k));
							}
						}
					}
				}
				for(int k = 0; k < items[i].targetLength;k++) {
					int targetIndex = items[i].targetIndex + k;
					//cout << "Looking at Target Index:"  << targetIndex << endl;
					//indices stored in sorted order
					for(int j = 0; j < sourceLength; j++) {
						if(j >= items[i].sourceIndex + items[i].targetLength) {
							//node has to be zero for this phrase
							//cout << "Zeronode: " << j << "," << targetIndex << endl;
							zeroNodes.push_back(G->GetNode(j,targetIndex));
						}else if(j < items[i].sourceIndex) {
							//node has to be zero for this phrase
							//cout << "Zeronode: " << j << "," << targetIndex << endl;
							zeroNodes.push_back(G->GetNode(j,targetIndex));
						}
					}
				}
			}
			
			if(dynamicFeatureOffSet[PTIndex + 1] - dynamicFeatureOffSet[PTIndex] > 0) {
				//loading several dynamic Feature Files for every expert not implemented
				assert(0);
				int dynIndex = dynamicFeatureOffSet[PTIndex] + items[i].dynamicIndex;
				double dynScalingFactor = dynamicScalingFactors[dynIndex];
				nodes[PTIndex][i].Init(zeroNodes,oneNodes,items[i].scores,items[i].dynamicScore,&ScalingFactors[2*featureOffSet[PTIndex]*numberOfExperts],dynScalingFactor,2* numberOfFeatures + dynIndex,offset + 2* featureOffSet[PTIndex],offset,numberOfAllFeatures,numberOfExperts,expertWeights);
			}else {
				nodes[PTIndex][i].Init(zeroNodes,oneNodes,items[i].scores,0,&ScalingFactors[2*featureOffSet[PTIndex]*numberOfExperts],0,-1,offset + 2*featureOffSet[PTIndex],offset,numberOfAllFeatures,numberOfExperts, expertWeights);
		
			}
		}
	}
}

void PhraseLOPFeatureSet::RemoveFeatures(){
	while(nodes.size() > 0) {
		delete[] nodes[nodes.size() - 1];
		nodes.pop_back();
		numberOfNodes.pop_back();
	}
}	

  /**
   * @return FeatureSetIterator
   */
FeatureSetIterator * PhraseLOPFeatureSet::GetIterator ( ) const {
	return new PhraseLOPFeatureSetIterator(nodes,numberOfNodes);

}

  /**
   */
void PhraseLOPFeatureSet::EstimateScalingFactores () {
	EstimatedDerivation.clear();
	vector<double> ExpectValue;
	vector<double> ExampleValue;
	ExpectValue.resize(numberOfExperts,0);
	ExampleValue.resize(numberOfExperts,0);

	for(int PTIndex = 0; PTIndex < numberOfPhraseTables; PTIndex ++) {
		for(int i = 0; i < numberOfNodes[PTIndex];i++) {
			vector<double> TempExpectValue = nodes[PTIndex][i].GetExpectationValue();
			vector<double> TempExampleValue = nodes[PTIndex][i].GetExampleValue();
			assert(TempExpectValue.size() == numberOfExperts);
			assert(TempExampleValue.size() == numberOfExperts);
			for(int j = 0; j < numberOfExperts; j++) {
				ExpectValue[j] += TempExpectValue[j];
				ExampleValue[j] += TempExampleValue[j];
			}
 		 }
	}
	for(int j = 0; j < numberOfExperts; j++) {
		double deltaL = ExampleValue[j] - ExpectValue[j];
		//cout << "Calculating Der[" << j << "]: " << ExampleValue[j] << " - " <<  ExpectValue[j] << " = " << deltaL << endl;
		EstimatedDerivation.push_back(deltaL);
	}
}


void PhraseLOPFeatureSet::LoadScalingFactorFile(string fileName,vector<double> & ScalingFactors,int offset,int size){

	assert(size > 0);
	assert(ScalingFactors.size() == offset);
	for(int i = 0; i < size; i++)
		ScalingFactors.push_back(0);
	ifstream file( fileName.c_str());

	if (!file) {
		cerr << "ScalingFactorFile " << fileName << " could not be opened\n";
		exit(-2);
	}

	char *pIndex = new char[ 200 ];
	char *pValue = new char[ 200 ];
	while ( !file.eof() ) {
		file >> pIndex >> pValue;
		int index = atoi(pIndex);
		assert(index >= 0);
		assert(index < size);
		ScalingFactors[offset + index] = atof(pValue);
	}
	delete[] pIndex;
	delete[] pValue;
  
}

void PhraseLOPFeatureSet::SearchPhrases(Graph * G,vector<PhraseItem> & items,multimap<string,PhraseItem> * currentPhraseTable) {
	
	assert(items.size() == 0);
	multimap<string,PhraseItem> targetTable;

	int maxPhraseLength = 4;

	//find matching source phrases
	int sourceLength = G->GetSourceLength();
	for(int i = 0; i < sourceLength; i++) {
		string source = "";
		for(int j = 0; (j < maxPhraseLength) && (i + j < sourceLength); j++) {
			if(j != 0) {
				source.push_back(' ');
			}
			source.append(G->GetSourceWord(i + j));
			//search for phrases starting at this position
			pair<PTIterator,PTIterator> iter = currentPhraseTable->equal_range(source);
			for(PTIterator begin = iter.first; begin != iter.second; begin++) {
				PhraseItem item = begin->second;
				item.sourceIndex = i;
				targetTable.insert(make_pair(item.target,item));
			}

		}
	}

	//check if target phrases match also
	int targetLength = G->GetTargetLength();
	for(int i = 0; i < targetLength; i++) {
		string target = "";
		for(int j = 0; (j < maxPhraseLength) && (i + j < targetLength); j++) {
			if(j != 0) {
				target.push_back(' ');
			}
			target.append(G->GetTargetWord(i + j));

			//check matching target phrases
			pair<PTIterator,PTIterator> iter = targetTable.equal_range(target);
			for(PTIterator begin = iter.first; begin != iter.second; begin++) {
				PhraseItem item = begin->second;
				item.targetIndex = i;
				//phrase matches complete
				items.push_back(item);
				//cout << "Found item: "<< item.source << " # " << item.target << endl;
			}
		}
	}



}

void PhraseLOPFeatureSet::LoadPhraseTable(const char * fileName,multimap<string,PhraseItem> * phraseTable, int numberOfPTFeatures, int numberOfDynamicPTFeatures,int selectedOneNodesForPT) {
	const int BufferSize = 1024;
	char Line [BufferSize];
	ifstream file( fileName);

	if (!file) {
		cerr << "PhraseTable " << fileName << " could not be opened\n";
		exit(-2);
	}
	
	while(file.getline(Line,BufferSize)) {
		string line(Line);
		vector<string> elements;
		Utils::Tokenize(line,elements,string("#"));
		assert(elements.size() >= numberOfPTFeatures + 2);
		PhraseItem item;

		//remove leading and endling blanks from source string
		vector<string> temp;
		Utils::Tokenize(elements[0],temp);
		item.sourceLength = temp.size();
		item.source = temp[0];
		for(int i = 1; i < temp.size(); i++) {
			item.source.push_back(' ');
			item.source.append(temp[i]);
		}
		temp.clear();

		//remove leading and ending blanks from target string
		Utils::Tokenize(elements[1],temp);
		item.targetLength = temp.size();			
		item.target = temp[0];
		for(int i = 1; i < temp.size(); i++) {
			item.target.push_back(' ');
			item.target.append(temp[i]);
		}

		int i = 0;
		for(; i < numberOfPTFeatures; i++) {
			item.scores.push_back(atof(elements[2+i].c_str()));
		}
		if(numberOfDynamicPTFeatures) {
			vector<string> dynamicScore;
			Utils::Tokenize(elements[2+i],dynamicScore);
			assert(dynamicScore.size() == 2);
			item.dynamicIndex = atoi(dynamicScore[0].c_str());
			assert(item.dynamicIndex > 0);
			assert(item.dynamicIndex < numberOfDynamicPTFeatures);
			item.dynamicScore = atof(dynamicScore[1].c_str());
			i++;
		}
		if(selectedOneNodesForPT) {
			vector<string> links;
			Utils::Tokenize(elements[2+i],links);
			for(int j = 0; j < links.size(); j++) {
				int a;
				int b;
				char c;
				sscanf(links[j].c_str(),"%i%c%i",&a,&c,&b);
				item.Alignments.push_back(make_pair(a-1,b-1));
			}
			i++;
		}
		//cout << "i: " << i << endl;
		//cout << "elementes.size(): " << elements.size() << endl;
		assert(elements.size() == i +2);
		//item.currentWord = 0;
/*		cout << "Found Phrase : " << elements[0] << elements[1];
		for(int j = 0; j < item.Alignments.size(); j++) {
			cout << item.Alignments[j].first <<"-" << item.Alignments[j].second << " ";
		}
		cout << endl;*/
		phraseTable->insert(make_pair(item.source,item));
		//cout << "Insert item: |" << item.source << "|" << item.target << "|" << endl;
	}
}





