#ifndef PHRASELOPFEATURESET_H
#define PHRASELOPFEATURESET_H

#include <featureset.h>
#include <phraselopfeature.h>
#include <phraselopfeaturesetiterator.h>
#include <utils.h>
/**
	@author Jan Niehues <jan@niehues.info>
*/
struct PhraseItem {
	string source;
	string target;
	int sourceIndex;
	int targetIndex;
	int sourceLength;
	int targetLength;
	vector<double> scores;
	int dynamicIndex;
	double dynamicScore;
	vector<pair<int,int> > Alignments; 

};
typedef multimap<string,PhraseItem>::iterator PTIterator;



class PhraseLOPFeatureSet : public FeatureSet
{
private:
	int numberOfExperts;
	double * expertWeights;
	vector<double> EstimatedDerivation;
	multimap<string,PhraseItem> * phraseTable;
	int numberOfPhraseTables;
	int numberOfFeatures;
	int numberOfDynamicFeatures;
	int * featureOffSet;
	int * dynamicFeatureOffSet;
	vector<int> selectedOneNodes;
	vector<int> exactMatch;
	vector<double> dynamicScalingFactors;
	double * ScalingFactors;
	vector<PhraseLOPFeature *> nodes;
	vector<int> numberOfNodes;
	vector<string> scalingFactorFiles;

	void LoadParamFile(const char* ParamFile);
	void LoadPhraseTable(const char * FileName, multimap<string,PhraseItem> * phraseTable, int numberOfPTFeatures,int numberOfDynamicPTFeatures,int selectedOneNodesForPT);
	void LoadScalingFactorFile(string fileName,vector<double> & ScalingFactors,int offset,int size);
	void SearchPhrases(Graph * G,vector<PhraseItem> & items,multimap<string,PhraseItem> * phraseTable);

public:
    PhraseLOPFeatureSet(const char* ParamFile, const char * ScalingFactorString,int numberOfExp, double * expWeights);

    ~PhraseLOPFeatureSet();

  /**
   * @param  G
   */
  void GenerateFeatures (Graph * G  , int offset, int numberOfAllFeatures);

  void RemoveFeatures();

  /**
   * @return FeatureSetIterator
   */
  FeatureSetIterator * GetIterator ( ) const;

  /**
   */
  void EstimateScalingFactores ();

  vector<double> inline GetEstimatedDerivations() const {return EstimatedDerivation;};


  virtual inline int GetNumberOfScalingFactors() {return 2* numberOfFeatures + numberOfDynamicFeatures;};
};

#endif
