#include <string.h>
#include "utils.h"

Utils::Utils()
{
}


Utils::~Utils()
{
}

double Utils::CacluateEditDistance(const string Source,const string Target) {
  int sl = Source.length() + 1;
  int tl = Target.length() + 1;
  int distance[sl * tl];

  // init first colum
  for(int j = 0; j < tl;j ++) {
    distance[j] = j;
    //cout << "Distance[" << 0 << "][" << j << "] = " << distance[0*tl + j] << endl;
  } 
  //init first line
  for(int i = 0; i < sl;i ++) {
    distance[i*tl] = i;
    //cout << "Distance[" << i << "][" << 0 << "] = " << distance[i*tl + 0] << endl;
  } 

  //do the rest fieds
 for(int i = 1; i < sl; i++) {
   for(int j = 1; j < tl; j++) {
     int e = min(distance[i*tl + j -1],distance[(i-1)*tl+j]) + 1;
     int f = (Source[i-1] == Target[j-1]) ? distance[(i-1)*tl+j-1] : distance[(i-1)*tl+j-1] +1;
     distance[i*tl + j] = min(e,f);
     //cout << "Distance[" << i << "][" << j << "] = " << distance[i*tl + j] << endl;
   }
 }
 return 1.0*distance[sl*tl -1]/max(Target.length(),Source.length());

}


void Utils::LoadScalingFactorFile(string fileName,vector<double> & ScalingFactors,int sourceLength, int targetLength){

  assert(sourceLength > 0);
  assert(targetLength > 0);
  ScalingFactors.clear();
  ScalingFactors.resize(sourceLength * targetLength,0);
  ifstream file( fileName.c_str());

  if (!file) {
    cerr << "File " << fileName << " could not be opened\n";
    exit(-2);
  }

  char *pSourceIndex = new char[ 200 ];
  char *pTargetIndex = new char[ 200 ];
  char *pValue = new char[ 200 ];
  while ( !file.eof() ) {
    file >> pSourceIndex >> pTargetIndex >> pValue;
    int sourceIndex = atoi(pSourceIndex);
    int targetIndex = atoi(pTargetIndex);
    assert(sourceIndex < sourceLength);
    assert(targetIndex < targetLength);
    ScalingFactors[sourceIndex * targetLength + targetIndex] = atof(pValue);
  }
	delete[] pSourceIndex;
	delete[] pTargetIndex;
	delete[] pValue;
  
}
void Utils::LoadScalingFactorFiles(vector<string> fileName,double * ScalingFactors,int sourceLength, int targetLength){

  assert(sourceLength > 0);
  assert(targetLength > 0);
  assert(fileName.size() > 0);
	for(int i = 0; i < sourceLength * targetLength * fileName.size(); i++) {
		ScalingFactors[i] = 0;
	}
  for(int i = 0; i < fileName.size(); i++) {
  ifstream file( fileName[i].c_str());

  if (!file) {
    cerr << "File " << fileName[i] << " could not be opened\n";
    exit(-2);
  }

  char *pSourceIndex = new char[ 200 ];
  char *pTargetIndex = new char[ 200 ];
  char *pValue = new char[ 200 ];
  while ( !file.eof() ) {
    file >> pSourceIndex >> pTargetIndex >> pValue;
    int sourceIndex = atoi(pSourceIndex);
    int targetIndex = atoi(pTargetIndex);
    assert(sourceIndex < sourceLength);
    assert(targetIndex < targetLength);
    ScalingFactors[(sourceIndex * targetLength + targetIndex)*fileName.size() + i] = atof(pValue);
  }
	delete[] pSourceIndex;
	delete[] pTargetIndex;
	delete[] pValue;
  }
}

void Utils::LoadFOFScalingFactorFiles(vector<string> fileNames,double * ScalingFactors,int sourceLength, int targetLength,int numberOfDirections ){

	assert(sourceLength > 0);
	assert(targetLength > 0);
	assert(numberOfDirections > 0);
	for(int i = 0;  i < sourceLength * targetLength * numberOfDirections * fileNames.size(); i++) {
		ScalingFactors[i] = 0;
	}

	for(int i =0; i < fileNames.size(); i++) {
		ifstream file( fileNames[i].c_str());

		if (!file) {
			cerr << "File " << fileNames[i] << " could not be opened\n";
			exit(-2);
		}

		char *pDirection = new char [ 200 ];
		char *pSourceIndex = new char[ 200 ];
		char *pTargetIndex = new char[ 200 ];
		char *pValue = new char[ 200 ];
		while ( !file.eof() ) {
			file >> pDirection >> pSourceIndex >> pTargetIndex >> pValue;
			int direction = atoi(pDirection);
			int sourceIndex = atoi(pSourceIndex);
			int targetIndex = atoi(pTargetIndex);
			assert(sourceIndex < sourceLength);
			assert(targetIndex < targetLength);
			assert(direction < numberOfDirections);
			assert(direction *(targetLength + sourceLength) + sourceIndex * targetLength + targetIndex < sourceLength * targetLength*numberOfDirections);
			int index = (direction *targetLength * sourceLength + sourceIndex * targetLength + targetIndex) * fileNames.size() + i;
		ScalingFactors[index] = atof(pValue);
		}
		delete[] pDirection;
		delete[] pSourceIndex;
		delete[] pTargetIndex;
		delete[] pValue;
	}

}

void Utils::LoadVoc(char * fileName,map<string,int> & Voc){
   ifstream file( fileName);

  if (!file) {
    cerr << "File " << fileName << " could not be opened\n";
    exit(-2);
  }

  int i = 0;
  char *pWord = new char[ 200 ];
  while ( !file.eof() ) {
    file >> pWord;
    Voc.insert(make_pair(string(pWord),i));
    i++;
  }
	delete[] pWord;
}

void Utils::LoadLex(map<pair<string,string>,double> & Lexicon, const char* fileName, int Direction) {
   ifstream LexFile(fileName);  
	
   if (!LexFile) {
    cerr << "Lexicon " << fileName << " could not be opened\n";
    exit(-2);
  }
  cout << "Load Lexicon " << fileName << endl;
  char *pSource = new char[ 200 ];
  char *pTarget = new char[200];
  char *pValue = new char[ 200 ];
  while ( !LexFile.eof() ) {
    LexFile >> pSource >> pTarget >> pValue;
    if(Direction) {
	double f = atof(pValue);
	Lexicon.insert(make_pair(make_pair(string(pTarget),string(pSource)),f));
	//cout << "Insert: (" << pTarget << "," << pSource << ") = " << f << endl;
    }else {
	double f= atof(pValue);
	Lexicon.insert(make_pair(make_pair(string(pSource),string(pTarget)),f));
	//cout << "Insert: (" << pSource << "," << pTarget << ") = " << f << endl;
    }
  }

  LexFile.close();
  delete[] pSource;
  delete[] pTarget;
  delete[] pValue;
}

void Utils::LoadGIZAFertilities(const char* fileName, map<string, vector<double> > & table ) {
  
	const int bSize = 1024;
	char line[bSize];
	char buffer[bSize];

	ifstream File(fileName);
	if(!File) {
		cout << " GIZA Fertility File : " << fileName << " could not be opend!" << endl;
		exit(0);
	}
  
  //pase lines
	while(File.getline(line, bSize)) {
    //skip comments
		if (line[0] == '#')
			continue;

		strcpy (buffer,line);

    //skip empty lines
		if(strcmp(buffer,"") ==0)
			continue;

		char* items;
		vector <double> probs;
		items = strtok(buffer, " ");
		if(items == NULL)
			continue;
		string word(items);
		items = strtok(NULL, " ");
		while(items != NULL) {
			probs.push_back(atof(items));
			items = strtok(NULL, " ");
		}

		table.insert(make_pair(word,probs));
	}

	cout << " GIZA Fertilities " << fileName << " loaded!" << endl;
}

void Utils::ExtractScalingFactors(const char * string, double * values, int size) {
	char* List = strdup(string);
	char * factores;
	int i = 0;
	factores = strtok (List,"_");
	while ( factores != NULL ){
		assert(i < size);
		values[i] = atof( factores );
	//cout << "Phrase Align Factor " << i << " is: " << PhraseAlignFactors[ i ] << endl;
		factores = strtok( NULL, "_" );
		i++;
	}
	assert(i == size);
	free(List);

}
/** Extract Scaling Factors from string
 *
 * @param ScalingFactorString String containing the scalingfactors
 */

void Utils::Extract2DScalingFactors(const char * s,double * values,int size1, int size2) {
	char* list = strdup(s);
	char * factors;
	vector<string> parts;
	factors = strtok (list,"|");
	while ( factors != NULL ){
		parts.push_back(string(factors));
		factors = strtok( NULL, "|" );
	}
	free(list);
	
	assert(parts.size() == size2);

	for(int i = 0; i < parts.size();i++) {
		list = strdup(parts[i].c_str());
		//cout << list << endl;
		factors = strtok (list,"_");
		int j = 0;
		while ( factors != NULL ){
			//cout << factors;
			//cout << "j " << j << " numberOfLexicons + additionalFeatures: " << numberOfLexicons + additionalFeatures << endl;
			assert(j < size1);
			values[j * (size2) + i] = (double) atof(factors);
			j++;
			factors = strtok( NULL, "_" );
		}
		if(j != size1) {
			cout << "j = " << j << " size1 = " << size1 << endl;
		}
		assert(j == size1);
		free(list);
	}

}

void Utils::Tokenize(const string& str,
	      vector<string>& tokens,
       const string& delimiters)
{
    // Skip delimiters at beginning.
	string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
	string::size_type pos     = str.find_first_of(delimiters, lastPos);

	while (string::npos != pos || string::npos != lastPos)
	{
        // Found a token, add it to the vector.
		tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
		lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
		pos = str.find_first_of(delimiters, lastPos);
	}
}
