#ifndef UTILS_H
#define UTILS_H

/**
	@author Jan Niehues <jan@niehues.info>
*/
#include <vector>
#include <map>
#include <utility>
#include <string>
#include <fstream>
#include <cassert>
#include <cstdlib>
#include <iostream>
#include <string>
using namespace std;
class Utils{
public:
    Utils();

    ~Utils();
	/// Calculates Edit Distance between two words
	/**	Calcuates edit distance between two words using dynamic programming
	*
	*	@param Source Word to calculate edit distance
	*	@param Target 2. word to calculate edit distance
	*/
	static double CacluateEditDistance(const string Source,const string Target);
	/// Load a file containing Scaling Factors
	/** 
	*	@param fileName Filename of the parameter file
	*	@param ScalingFactors Place to store the scaling factors
	*	@param sourceLength Number of different source items
	* 	@param targetLength Number of different target items
	*/
	static void LoadScalingFactorFile(string fileName,vector<double> & ScalingFactors,int sourceLength, int targetLength);
	/// Load several files containing Scaling Factors
	/**     Array has to be allocated before
	*	@param fileName Filename of the parameter file
	*	@param ScalingFactors Place to store the scaling factors
	*	@param sourceLength Number of different source items
	* 	@param targetLength Number of different target items
	*/
	static void LoadScalingFactorFiles(vector<string> fileName,double * ScalingFactors,int sourceLength, int targetLength);
	/// Loads Vocabulary file
	/** 
	*	@param fileName Filename of the vocabulary file
	*	@param Voc map to store the vocabulary (name and index)
	*/
	static void LoadVoc(char * fileName,map<string,int> & Voc);
	/// Loads Lexicon file
	/** 
	*	@param Lexicon Place to store lexicon
	*	@param fileName Name of the lexicon-file
	*	@param Direction 0 = source-target 1 = target-source
	*/

	static void LoadLex(map<pair<string,string>,double> & Lexicon, const char* fileName, int Direction);

	///Load Giza Fertility Table
	/**
	*	@param fileName Filename of giza fertility table
	*	@param table place to store giza fertilities
	*/
	static void LoadGIZAFertilities(const char* fileName , map<string, vector<double> > & table);	

	///Extract double from string
	/**	doubles are written down in string seperated by '_'
	*	@param string Sting containing double
	*	@param values Array to store double
	*	@param size Size of array
	*/
	static void ExtractScalingFactors(const char * string, double * values, int size);

	static void Extract2DScalingFactors(const char * string,double * values,int size1, int size2);
	static void LoadFOFScalingFactorFiles(vector<string> fileNames,double * ScalingFactors,int sourceLength, int targetLength,int numberOfDirections );

	static void Tokenize(const string& str, vector<string>& tokens, const string& delimiters = " ");

};

#endif
