

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define OPTIMIZE_MAP -1
#define OPTIMIZE_ML 1
#define OPTIMIZE_AER 2
#define OPTIMIZE_RECALL 3
#define OPTIMIZE_PRECISION 4
#define OPTIMIZE_AERMAP 5
#define OPTIMIZER_FSCORE 6

#include <string.h>
#include <iostream>
#include <fstream>
#include <featureset.h>
#include <localfeatureset.h>
#include <fertilityfeatureset.h>
#include <firstorderfeatureset.h>
#include <phrasefeatureset.h>
#include <cstdlib>
#include <vector>
#include <climits>
#include <cfloat>
#include <param.hh>
#include <graph.h>
#include <optimizer.h>
#include <optimizermap.h>
#include <optimizerml.h>
#include <optimizeraer.h>
#include <optimizeraermap.h>
#include <optimizerrecall.h>
#include <optimizerprecision.h>
#include <optimizerfscore.h>
#include <parseTree.h>
using namespace std;

ifstream * LoadFile(const char* File);
vector<ifstream *> LoadFileList(char * File);
ofstream * OpenOutputFile(char* File);
void AlignSentence(ifstream *Source, ifstream *Target,ifstream *SourcePOS, ifstream *TargetPOS, ifstream *SourceTree, ifstream *TargetTree, ofstream * output, ifstream * PreAlignment, char * LinkFeatureList, int iterations,int DisplayGraph);
void ReadLine(ifstream * stream, vector<string> & words);
void ReadTree(ifstream * stream, ParseTree & tree);
void ReadAlignmentLine(ifstream * stream, vector<pair<int,int> > & alignemnt,vector<pair<int,int> > & possibleAlignemnt);
void ReadLinkFeatureLine(vector<ifstream *> & streams,vector<vector<LinkFeatures> > & features);
void InitLocalFeature(const char* ParamFile, const char* ScalingFactorList);
void InitFertilityFeature(const char* ParamFile, const char* ScalingFactorString);
void InitFirstOrderFeature(const char * ParamFile, const char* ScalingFactorString);
void InitPhraseFeature(const char * ParamFile, const char* ScalingFactorString);
void InitOptimizer(int mode,char * learnRateFile, char * sigmaFile,double alpha);
void GetScalingFactores(const char* ScalingFactorString, vector<double> & ScalingFactores);
void Optimize(char * Source, char * Target,char * SourcePOS, char * TargetPOS,char * SourceTree, char * TargetTree, char * SourceCV, char * TargetCV,char * SourceCVPOS, char * TargetCVPOS,char * AligmnetCVFile, char * PreAlignmentCVFile,char * CVLinkFeatureList, char * AligmnetFile, char * PreAlignmentFile, char * LinkFeatureList, int OptimizeIterations,int BeliefIterations,int DisplayGraph,int optimize);
void AlignCV(char * Source, char * Target,char * SourcePOS, char * TargetPOS,char * AligmnetFile, char * LinkFeatureList, char * PreAlignmentFile,int Iterations);
//void InitOptimizeFactors(char * Factors);
//void InitMeans(char * means);
//void LoadFactorFile(char * fileName, vector<vector<double> > & values);

CParam             *pParam             = NULL;
Optimizer *pOptimizer = NULL;
vector <FeatureSet*> FeatureSets;
vector < vector<double> > ScalingFactors; // ScalingFactors[i] = ScalingFactors for FeatureSets[i]
int numberOfAllFeatures = 0;
vector <int> numberOfOutputScalingFactors;
//vector <vector<double> > learnRate;
//vector <vector<double> > sigma;
//vector <vector<double> > mean;
int main(int argc, char *argv[])
{
	cout << "          Word Aligner" << endl;
	cout << "--------------------------------" << endl;

	pParam = new CParam;
	int iterations;
	int optimize;
	int OptimizeIterations;
	char* SourceFile = NULL;
	char* TargetFile = NULL;
	char* SourcePOSFile = NULL;
	char* TargetPOSFile = NULL;
	char* SourceTreeFile = NULL;
	char* TargetTreeFile = NULL;
	char* SourceCVFile = NULL;
	char* TargetCVFile = NULL;
	char* SourceCVPOSFile = NULL;
	char* TargetCVPOSFile = NULL;
	char* SourceTestFile = NULL;
	char* TargetTestFile = NULL;
	char* SourceTestPOSFile = NULL;
	char* TargetTestPOSFile = NULL;
	char * SourceTestTreeFile = NULL;
	char * TargetTestTreeFile = NULL;
	char* AlignmentFile = NULL;
	char* AlignmentCVFile = NULL;
	char* PreAlignmentFile = NULL;
	char* PreAlignmentTestFile = NULL;
	char* PreAlignmentCVFile = NULL;
	char* OutputFile = NULL;

	char* LinkFeatureList = NULL;
	char* TestLinkFeatureList = NULL;
	char* CVLinkFeatureList = NULL;

	char* LFParamFile = NULL;
	char* LFScalingFactors = NULL;
	char* FFParamFile = NULL;
	char* FFScalingFactors = NULL;
	char* FOFParamFile = NULL;
	char* FOFScalingFactors = NULL;
	char* PFParamFile = NULL;
	char* PFScalingFactors = NULL;
	char* learnRateFile = NULL;
	char* sigmaFile = NULL;
	char* meanFile = NULL;
	double fScoreAlpha = 0.5;

	int DisplayGraph = 0;

	pParam->AddParamDef( "SourceFile", 'S', &SourceFile, NULL,
			     "List of all Source Sentences",
	False );
	pParam->AddParamDef( "TargetFile", 'T', &TargetFile, NULL,
 			     "List of all Target Sentences",
	False );
	pParam->AddParamDef( "SourcePOSFile", '\0', &SourcePOSFile, NULL,
			     "List of all Source POS Sentences",
	True );
	pParam->AddParamDef( "TargetPOSFile", '\0', &TargetPOSFile, NULL,
			     "List of all Target POS Sentences",
	True );
	pParam->AddParamDef( "SourceTreeFile", '\0', &SourceTreeFile, NULL,
			     "List of all soruce parse trees",
	True );
	pParam->AddParamDef( "TargetTreeFile", '\0', &TargetTreeFile, NULL,
			     "List of all target parse trees",
	True );
	pParam->AddParamDef( "SourceTestFile", '\0', &SourceTestFile, NULL,
			     "List of all Source Test Sentences",
	True );
	pParam->AddParamDef( "TargetTestFile", '\0', &TargetTestFile, NULL,
			     "List of all Target Test Sentences",
	True );
	pParam->AddParamDef( "SourceTestPOSFile", '\0', &SourceTestPOSFile, NULL,
			     "List of all Source Test POS Sentences",
	True );
	pParam->AddParamDef( "TargetTestPOSFile", '\0', &TargetTestPOSFile, NULL,
			     "List of all Target Test POS Sentences",
	True );
	pParam->AddParamDef( "SourceTestTreeFile", '\0', &SourceTestTreeFile, NULL,
			     "List of all source test parse trees",
	True );
	pParam->AddParamDef( "TargetTestTreeFile", '\0', &TargetTestTreeFile, NULL,
			     "List of all target test parse trees",
	True );
	pParam->AddParamDef( "SourceCVFile", '\0', &SourceCVFile, NULL,
			     "List of all Source Crossvalidation Sentences",
	True );
	pParam->AddParamDef( "TargetCVFile", '\0', &TargetCVFile, NULL,
			     "List of all Target Crossvalidation Sentences",
	True );
	pParam->AddParamDef( "SourceCVPOSFile", '\0', &SourceCVPOSFile, NULL,
			     "List of all Source Crossvalidation POS Sentences",
	True );
	pParam->AddParamDef( "TargetCVPOSFile", '\0', &TargetCVPOSFile, NULL,
			     "List of all Target Crossvalidation POS Sentences",
	True );
	pParam->AddParamDef( "AlignmentFile", 'A', &AlignmentFile, NULL,
			     "List of all alignments in pharao-style",
	True );
	pParam->AddParamDef( "AlignmentCVFile", '\0', &AlignmentCVFile, NULL,
			     "List of all alignments in pharao-style for CrossValidation set",
	True );
	pParam->AddParamDef( "PreAlignmentFile", '\0', &PreAlignmentFile, NULL,
			     "File containing Prealignments",
	True );
	pParam->AddParamDef( "PreAlignmentTestFile", '\0', &PreAlignmentTestFile, NULL,
			     "File containing Prealignments for Testset if Optimization",
	True );
	pParam->AddParamDef( "PreAlignmentCVFile", '\0', &PreAlignmentCVFile, NULL,
			     "File containing Prealignments for Crossvaidation",
	True );
	pParam->AddParamDef( "LinkFeatureFileList", '\0', &LinkFeatureList, NULL,
			     "File containing filesnames of links set by other word aligners",
	True );
	pParam->AddParamDef( "TestLinkFeatureFileList", '\0', &TestLinkFeatureList, NULL,
			     "File containing filesnames of links set by other word aligners",
	True );
	pParam->AddParamDef( "CVLinkFeatureFileList", '\0', &CVLinkFeatureList, NULL,
			     "File containing filesnames of links set by other word aligners",
	True );
	pParam->AddParamDef( "OutputFile", '\0', &OutputFile, NULL,
			     "Prints found alignment to this file",
	True );
	pParam->AddParamDef( "LocalFeatureParamFile", 'P', &LFParamFile, NULL,
			     "Parameterfile for the local features",
	False );
	pParam->AddParamDef( "LocalFeatureScalingFactors", 'L', &LFScalingFactors, NULL,
			     "Scaling Factors for the local features",
	False );
	pParam->AddParamDef( "FertilityFeatureScalingFactors", 'F', &FFScalingFactors, NULL,
			     "Scaling Factors for the fertility features",
	True );
	pParam->AddParamDef( "FertilityFeatureParamFile", '\0', &FFParamFile, NULL,
			     "Parameterfile for the fertility features",
	True );
	pParam->AddParamDef( "FirstOrderFeatureParamFile", '\0', &FOFParamFile, NULL,
			     "Parameterfile for the first order features",
	True );
	pParam->AddParamDef( "FirstOrderScalingFactors", 'O', &FOFScalingFactors, NULL,
			     "Scaling Factors for the first order features",
	True );
	pParam->AddParamDef( "PhraseFeatureParamFile", '\0', &PFParamFile, NULL,
			     "Parameterfile for the local features",
	True );
	pParam->AddParamDef( "PhraseFeatureScalingFactors", '\0', &PFScalingFactors, NULL,
			     "Scaling Factors for the local features",
	True );
	pParam->AddParamDef( "Iterations", 'I', &iterations, 10,
			     "Number of Iterations of the Belief Properations algorithm",
	True );
	pParam->AddParamDef( "DisplayGraph", 'D', &DisplayGraph, 0,
			     "File for every Sentence is created containing the Graph at all timesteps",
	True );
	pParam->AddParamDef( "Optimize", 'o', &optimize, 0,
			     "1 = Optimize; 0 = Align Sentences",
	True );
	pParam->AddParamDef( "OptimizeIterations", 'i', &OptimizeIterations, 10,
			     "Iterations of the Optimization",
	True );
	pParam->AddParamDef( "SigmaFile", 's', &sigmaFile, NULL,
			     "File containing sigma for MAP Training",
	True );
	pParam->AddParamDef( "LearnRateFile", '\0', &learnRateFile, NULL,
			     "File containing the learn rates",
	True );
	pParam->AddParamDef( "MeanFile", '\0', &meanFile, NULL,
			     "File containing the means of the features",
	True );
	pParam->AddParamDef( "FScoreAlpha", '\0', &fScoreAlpha, 0.5,
			     "Alpha-Value for F-Score",
	True );


	if (! pParam->Interpret( argv, argc) ) {
		cerr << "\nParameters could not be interpreted correctly\n ... exiting!!\n";
		exit(1); 
	}

	//init lokal features
	InitLocalFeature(LFParamFile,LFScalingFactors);

	if (FFScalingFactors != NULL) {
		InitFertilityFeature(FFParamFile,FFScalingFactors);
	}else {
		cout << "No FertilityFeatures used" << endl;
	}

	if (FOFScalingFactors != NULL) {
		InitFirstOrderFeature(FOFParamFile,FOFScalingFactors);
	}else {
		cout << "No FirstOrderFeatures used" << endl;
	}

	if(PFParamFile != NULL) {
		InitPhraseFeature(PFParamFile,PFScalingFactors);
	}else {
		cout << "No PhraseFeatures used" << endl;
	}

	if(optimize) {
		InitOptimizer(optimize,learnRateFile, sigmaFile,fScoreAlpha);
		if(AlignmentFile == NULL) {
			cout << "For Optimization a alignment file is needed" << endl;
			exit(1);
		}
		Optimize(SourceFile, TargetFile,SourcePOSFile, TargetPOSFile,SourceTreeFile, TargetTreeFile, SourceCVFile, TargetCVFile,SourceCVPOSFile,TargetCVPOSFile,AlignmentCVFile,PreAlignmentCVFile,CVLinkFeatureList, AlignmentFile,PreAlignmentFile,LinkFeatureList,OptimizeIterations,iterations,DisplayGraph,optimize);

		if((SourceTestFile != NULL) && (TargetTestFile != NULL)) {
			if(OutputFile == NULL) {
				cout << "For Alginment an output file needed" << endl;
				exit(1);
			}
			cout << " Algin Test sentences!" << endl;
			ifstream * SourceTest = LoadFile(SourceTestFile);
			ifstream * TargetTest = LoadFile(TargetTestFile);
			ifstream * SourceTestPOS = NULL;
			ifstream * TargetTestPOS = NULL;
			ifstream * SourceTestTree = NULL;
			ifstream * TargetTestTree = NULL;
			ifstream * PreAlignmentTest = NULL;
			if(SourceTestPOSFile != NULL) {
				SourceTestPOS = LoadFile(SourceTestPOSFile);
				TargetTestPOS = LoadFile(TargetTestPOSFile);
			}
			if(SourceTestTreeFile != NULL) {
				SourceTestTree = LoadFile(SourceTestTreeFile);
				TargetTestTree = LoadFile(TargetTestTreeFile);
			}
			if(PreAlignmentTestFile != NULL) {
				PreAlignmentTest = LoadFile(PreAlignmentTestFile);
			}
			ofstream * output = OpenOutputFile(OutputFile);
			AlignSentence(SourceTest, TargetTest,SourceTestPOS, TargetTestPOS,SourceTestTree, TargetTestTree, output,PreAlignmentTest,TestLinkFeatureList,iterations,DisplayGraph);
			delete SourceTest;
			delete TargetTest;
			if(SourceTestPOSFile != NULL) {
				delete SourceTestPOS;
				delete TargetTestPOS;
			}
			if(SourceTestTreeFile != NULL) {
				delete SourceTestTree;
				delete TargetTestTree;
			}
			if(PreAlignmentTestFile != NULL) {
				delete PreAlignmentTest;
			}
			delete output;
			cout << "Test Sentences algined!" << endl;

		}


	}else { 
		if(OutputFile == NULL) {
			cout << "For Alginment an output file needed" << endl;
			exit(1);
		}
		ifstream * Source = LoadFile(SourceFile);
		ifstream * Target = LoadFile(TargetFile);
		ifstream * SourcePOS = NULL;
		ifstream * TargetPOS = NULL;
		ifstream * SourceTree = NULL;
		ifstream * TargetTree = NULL;
		ifstream * PreAlignment = NULL;
		if(SourcePOSFile != NULL) {
			SourcePOS = LoadFile(SourcePOSFile);
			TargetPOS = LoadFile(TargetPOSFile);
		}
		if(SourceTreeFile != NULL) {
			SourceTree = LoadFile(SourceTreeFile);
			TargetTree = LoadFile(TargetTreeFile);
		}
		if(PreAlignmentFile != NULL) {
			PreAlignment = LoadFile(PreAlignmentFile);
		}
		ofstream * output = OpenOutputFile(OutputFile);
		AlignSentence(Source, Target,SourcePOS, TargetPOS,SourceTree, TargetTree, output,PreAlignment, LinkFeatureList, iterations,DisplayGraph);
		delete Source;
		delete Target;
		if(SourcePOSFile != NULL) {
			delete SourcePOS;
			delete TargetPOS;
		}
		if(SourceTreeFile != NULL) {
			delete SourceTree;
			delete TargetTree;
		}
		if(PreAlignmentFile != NULL) {
			delete PreAlignment;
		}
		delete output;
	}


	for(int i = 0; i < FeatureSets.size(); i++) {
		delete FeatureSets[i];
	}
	return EXIT_SUCCESS;
}

ifstream * LoadFile(const char* File) {
	ifstream * stream = new ifstream(File);
	if(!(*stream)) {
		cerr << "ERROR: File " << File <<" could not be open\n";
		exit(-2);
	}
	return stream;
}
vector<ifstream *> LoadFileList(char * File) {

	vector <ifstream *> Files;
	ifstream * stream = new ifstream(File);
	if(!(*stream)) {
		cerr << "ERROR: File list " << File <<" could not be open\n";
		exit(-2);
	}
	string line = "";
	string word = "";
	while(getline(*stream,line)) {
		//cout << "Load File: " << line << endl;
		Files.push_back(LoadFile(line.c_str()));
	}

	return Files;
}

ofstream * OpenOutputFile(char* File) {
	ofstream * stream = new ofstream(File);
	if(!(*stream)) {
		cerr << "ERROR: File " << File <<" could not be open\n";
		exit(-2);
	}
	return stream;
}

void AlignSentence(ifstream * Source, ifstream * Target,ifstream * SourcePOSStream, ifstream * TargetPOSStream,ifstream * SourceTreeStream, ifstream * TargetTreeStream, ofstream * output, ifstream * preAlignmentStream, char * LinkFeatureList, int iterations, int DisplayGraph) {

	vector<string> SourceWords;
	vector<string> TargetWords;
	vector<string> SourcePOS;
	vector<string> TargetPOS;
	ParseTree SourceTree;
	ParseTree TargetTree;
	int Sent = 0;
	ofstream * GraphOut;
	vector<pair<int,int> > preAlignment;
	vector<pair<int,int> > possiblePreAlignment;
	vector<ifstream *> LinkFeatureFiles;
	if(LinkFeatureList != NULL) {
		LinkFeatureFiles = LoadFileList(LinkFeatureList);
	}
	vector<vector<LinkFeatures> > linkFeatures;

	while((!Source->eof()) && (!Target->eof())) 
	{	

		ReadLine(Source,SourceWords);
		ReadLine(Target,TargetWords);
		if(SourcePOSStream != NULL) {
			ReadLine(SourcePOSStream,SourcePOS);
		}
		if(TargetPOSStream != NULL) {
			ReadLine(TargetPOSStream,TargetPOS);
		}
		if(SourceTreeStream != NULL) {
			ReadTree(SourceTreeStream,SourceTree);
		}
		if(TargetTreeStream != NULL) {
			ReadTree(TargetTreeStream,TargetTree);
		}

		if(preAlignmentStream != NULL) {
			ReadAlignmentLine(preAlignmentStream,preAlignment,possiblePreAlignment);	
		}
		if(LinkFeatureFiles.size() != 0) {
			ReadLinkFeatureLine(LinkFeatureFiles,linkFeatures);	
		}

		if((SourceWords.size() > 0) && (TargetWords.size() > 0)) {
			if(DisplayGraph) {
				Sent++;
				char filename[256];
				sprintf(filename, "Sentence%d.gda", Sent);
				//filename += string(Sent);
				GraphOut = new ofstream(filename);
				*GraphOut << SourceWords.size() << " " <<TargetWords.size() << " " << iterations << endl;
				for(int i = 0; i < SourceWords.size(); i++) {
					*GraphOut << SourceWords[i] << " ";
				}
				*GraphOut << endl;
				for(int i = 0; i < TargetWords.size(); i++) {
					*GraphOut << TargetWords[i] << " ";
				}
				*GraphOut << endl;
				*GraphOut << "#Format: time sourceindex targetindex belief belief Lok.messageOut Lok.messageOut Lok.messageIn Lok.messageIn";
				*GraphOut << "SourceFert.messageOut SourceFert.messageOut SourceFert.messageIn SourceFert.messageIn";
				*GraphOut << "TargetFert.messageOut TargetFert.messageOut TargetFert.messageIn TargetFert.messageIn";
				*GraphOut << endl;
			}
			/*cout << "Source: ";
			for(int i = 0; i < SourceWords.size(); i++) {
			cout << i << SourceWords[i] << " ";
		}
			cout << endl;	
			cout << "SourceLength: " << SourceWords.size() << endl;
			cout << "Target: ";
			for(int i = 0; i < TargetWords.size(); i++) {
			cout << i << TargetWords[i] << " ";
		}
			cout << endl;
			cout << "TargetLength: " << TargetWords.size() << endl;
			*/
			//Build Graph
			Graph * G = new Graph(SourceWords.size(), TargetWords.size(),SourceWords,TargetWords,SourcePOS,TargetPOS);

			// store trees in graph
			if(SourceTreeStream != NULL) {
				G->LoadTrees(SourceTree, TargetTree);
			}
			if(preAlignmentStream != NULL) {
				G->LoadPreAlignment(preAlignment,possiblePreAlignment);
			}
			//give LinkFeatures to Graph
			if(linkFeatures.size() != 0) {
				G->SetLinkFeatures(linkFeatures);
			}

			for(int i = 0; i < FeatureSets.size(); i++) {
				FeatureSets[i]->GenerateFeatures(G,-1,0);
				G->AddFeatureSet(FeatureSets[i]);
			}
			if(DisplayGraph) {
				G->SetDisplay(GraphOut);
			}
			G->Inference(iterations);
			G->OutputAlignment(output);
			for(int i = 0; i < FeatureSets.size();i++) {
				FeatureSets[i]->RemoveFeatures();
			}
			SourceWords.clear();
			TargetWords.clear();
			SourcePOS.clear();
			TargetPOS.clear();
			if(preAlignmentStream != NULL) {
				preAlignment.clear();
				possiblePreAlignment.clear();
			}
			//delete linksFeatures for Sentence
			if(linkFeatures.size() != 0) {
				linkFeatures.clear();
			}
			if(DisplayGraph) {
				GraphOut->close();
				delete GraphOut;
			}
			delete G;
		}
	}
	//delete file handlers for link feature files
	if(LinkFeatureList != NULL) {
		for(int i = 0; i < LinkFeatureFiles.size(); i++) {
		delete LinkFeatureFiles[i];
			
		}
	}


}

void Optimize(char * SourceFile, char * TargetFile,char * SourcePOSFile, char * TargetPOSFile, char * SourceTreeFile, char * TargetTreeFile, char * SourceCV, char * TargetCV,char * SourceCVPOS, char * TargetCVPOS,char * AlignmentCVFile, char * preAlignmentCVFile, char * CVLinkFeatureList, char * AlignmentFile,char * preAlignmentFile, char * LinkFeatureList, int OptimizationIterations,int BeliefIterations,int DisplayGraph,int optimize) {

	for(int k = 0; k < OptimizationIterations; k++) {

		ifstream * Source = LoadFile(SourceFile);
		ifstream * Target = LoadFile(TargetFile);
		ifstream * SourcePOSStream = NULL;
		ifstream * TargetPOSStream = NULL;
		ifstream * SourceTreeStream = NULL;
		ifstream * TargetTreeStream = NULL;
		if(SourcePOSFile != NULL) {
			SourcePOSStream = LoadFile(SourcePOSFile);
			TargetPOSStream = LoadFile(TargetPOSFile);
		}
		if(SourceTreeFile != NULL) {
			SourceTreeStream = LoadFile(SourceTreeFile);
			TargetTreeStream = LoadFile(TargetTreeFile);
		}
		ifstream * alignmentStream = LoadFile(AlignmentFile);
		ifstream * preAlignmentStream = NULL;
		if(preAlignmentFile != NULL) {
			preAlignmentStream = LoadFile(preAlignmentFile);
		}
		vector<ifstream *> LinkFeatureFiles;
		if(LinkFeatureList != NULL) {
			LinkFeatureFiles = LoadFileList(LinkFeatureList);
		}
		vector<string> SourceWords;
		vector<string> TargetWords;
		vector<string> SourcePOS;
		vector<string> TargetPOS;
		ParseTree SourceTree;
		ParseTree TargetTree;
		vector<pair<int,int> > alignment;
		vector<pair<int,int> > possibleAlignment;

		vector<pair<int,int> > preAlignment;
		vector<pair<int,int> > possiblePreAlignment;

		vector<vector<LinkFeatures> > linkFeatures;
		cout << "Iteration " << k << endl;

		int Sent = 0;
		ofstream * GraphOut;
		while((!Source->eof()) && (!Target->eof())) 
		{	
			ReadLine(Source,SourceWords);
			ReadLine(Target,TargetWords);
			if(SourcePOSStream != NULL) {
				ReadLine(SourcePOSStream,SourcePOS);
			}
			if(TargetPOSStream != NULL) {
				ReadLine(TargetPOSStream,TargetPOS);
			}
			if(SourceTreeStream != NULL) {
				ReadTree(SourceTreeStream,SourceTree);
			}
			if(TargetTreeStream != NULL) {
				ReadTree(TargetTreeStream,TargetTree);
			}
			ReadAlignmentLine(alignmentStream,alignment,possibleAlignment);

			if(preAlignmentFile != NULL) {
				ReadAlignmentLine(preAlignmentStream,preAlignment,possiblePreAlignment);	
			}
			
			//read lines from linkFeatureFiles
			if(LinkFeatureFiles.size() != 0) {
				ReadLinkFeatureLine(LinkFeatureFiles,linkFeatures);	
			}

			if((SourceWords.size() > 0) && (TargetWords.size() > 0)) {
				Sent++;
				if(DisplayGraph) {
					char filename[256];
					sprintf(filename, "Sentence%d.gda", Sent);
	    //filename += string(Sent);
					GraphOut = new ofstream(filename);
					*GraphOut << SourceWords.size() << " " <<TargetWords.size() << " " << BeliefIterations << endl;
					for(int i = 0; i < SourceWords.size(); i++) {
						*GraphOut << SourceWords[i] << " ";
					}
					*GraphOut << endl;
					for(int i = 0; i < TargetWords.size(); i++) {
						*GraphOut << TargetWords[i] << " ";
					}
					*GraphOut << endl;
					*GraphOut << "#Format: time sourceindex targetindex belief belief Lok.messageOut Lok.messageOut Lok.messageIn Lok.messageIn";
					*GraphOut << "SourceFert.messageOut SourceFert.messageOut SourceFert.messageIn SourceFert.messageIn";
					*GraphOut << "TargetFert.messageOut TargetFert.messageOut TargetFert.messageIn TargetFert.messageIn";
					*GraphOut << endl;
				}

/*				for(int i = 0; i < SourceWords.size(); i++) {
					cout << SourceWords[i] << " ";
				}
				cout << endl;
				for(int i = 0; i < TargetWords.size(); i++) {
					cout << TargetWords[i] << " ";
				}
				cout << endl;
*///				cout << "Sentence: " << Sent << endl << flush;
	  //Build Graph
				Graph * G = new Graph(SourceWords.size(), TargetWords.size(),SourceWords,TargetWords,SourcePOS,TargetPOS,alignment,possibleAlignment);
				// store trees in graph
				if(SourceTreeStream != NULL) {
					G->LoadTrees(SourceTree, TargetTree);
				}

				if(preAlignmentFile != NULL) {
					G->LoadPreAlignment(preAlignment,possiblePreAlignment);
					
				}
					
				//give LinkFeatures to Graph
				if(linkFeatures.size() != 0) {
					G->SetLinkFeatures(linkFeatures);
				}

				int offset = 0;
				for(int i = 0; i < FeatureSets.size(); i++) {
					if(optimize >= 2) {
						FeatureSets[i]->GenerateFeatures(G,offset,numberOfAllFeatures);
						offset += ScalingFactors[i].size();
					}else {
						FeatureSets[i]->GenerateFeatures(G,-1,0);	      
					}
					G->AddFeatureSet(FeatureSets[i]);
				}
				if(DisplayGraph) {
					G->SetDisplay(GraphOut);
				}
				if(optimize >= 2) {
					G->Inference(BeliefIterations);
				}else {
					G->Optimize(BeliefIterations);
				}	    
				pOptimizer->EvalSentence(G);

	  
				for(int i = 0; i < FeatureSets.size(); i++) {
					FeatureSets[i]->RemoveFeatures();
				}
	  
				SourceWords.clear();
				TargetWords.clear();
				SourcePOS.clear();
				TargetPOS.clear();
				alignment.clear();
				possibleAlignment.clear();

				if(preAlignmentFile != NULL) {
					preAlignment.clear();
					possiblePreAlignment.clear();
				}
				//delete linksFeatures for Sentence
				if(linkFeatures.size() != 0) {
					linkFeatures.clear();
				}

				if(DisplayGraph) {
					GraphOut->close();
					delete GraphOut;
				}
				delete G;
			}
		}
    
		if(SourceCV == NULL) {
			pOptimizer->CheckScalingFactors();
		}
		pOptimizer->Estimate();
		delete Source;
		delete Target;
		if(SourcePOSFile != NULL) {
			delete SourcePOSStream;
			delete TargetPOSStream;
		}
		if(SourceTreeFile != NULL) {
			delete SourceTreeStream;
			delete TargetTreeStream;
		}
		delete alignmentStream;
		if(preAlignmentFile != NULL) {
			delete preAlignmentStream;
		}

		//delete file handlers for link feature files
		if(LinkFeatureList != NULL) {
			for(int i = 0; i < LinkFeatureFiles.size(); i++) {
				delete LinkFeatureFiles[i];
				
			}
		}
 		if(SourceCV != NULL) {
			AlignCV(SourceCV,TargetCV,SourceCVPOS,TargetCVPOS,AlignmentCVFile, preAlignmentCVFile,CVLinkFeatureList, BeliefIterations);
		}	
	}

	pOptimizer->OutputBest();

}
void AlignCV(char * SourceFile, char * TargetFile,char * SourcePOSFile, char * TargetPOSFile,char * AlignmentFile, char * preAlignmentFile,char * LinkFeatureList, int Iterations) {
	ifstream * Source = LoadFile(SourceFile);
	ifstream * Target = LoadFile(TargetFile);
	ifstream * SourcePOSStream = NULL;
	ifstream * TargetPOSStream = NULL;
	if(SourcePOSFile != NULL) {
		SourcePOSStream = LoadFile(SourcePOSFile);
		TargetPOSStream = LoadFile(TargetPOSFile);
	}
	ifstream * alignmentStream = LoadFile(AlignmentFile);
	ifstream * preAlignmentStream = NULL;
	if(preAlignmentFile != NULL) {
		preAlignmentStream = LoadFile(preAlignmentFile);
	}
	vector<string> SourceWords;
	vector<string> TargetWords;
	vector<string> SourcePOS;
	vector<string> TargetPOS;
	vector<pair<int,int> > alignment;
	vector<pair<int,int> > possibleAlignment;

	vector<pair<int,int> > preAlignment;
	vector<pair<int,int> > possiblePreAlignment;

	int Sent = 0;
	while((!Source->eof()) && (!Target->eof())) 
	{	
		ReadLine(Source,SourceWords);
		ReadLine(Target,TargetWords);
		if(SourcePOSStream != NULL) {
			ReadLine(SourcePOSStream,SourcePOS);
		}
		if(TargetPOSStream != NULL) {
			ReadLine(TargetPOSStream,TargetPOS);
		}
		ReadAlignmentLine(alignmentStream,alignment,possibleAlignment);

		if(preAlignmentFile != NULL) {
			ReadAlignmentLine(preAlignmentStream,preAlignment,possiblePreAlignment);	
		}
			
		if((SourceWords.size() > 0) && (TargetWords.size() > 0)) {
			Sent++;
			//cout << "Sentence: " << Sent << endl << flush;
			//Build Graph
			Graph * G = new Graph(SourceWords.size(), TargetWords.size(),SourceWords,TargetWords,SourcePOS,TargetPOS,alignment,possibleAlignment);
			if(preAlignmentFile != NULL) {
				G->LoadPreAlignment(preAlignment,possiblePreAlignment);
			}
				
			for(int i = 0; i < FeatureSets.size(); i++) {
				FeatureSets[i]->GenerateFeatures(G,-1,0);	      
				G->AddFeatureSet(FeatureSets[i]);
			}
			G->Inference(Iterations);
			//G->Optimize(Iterations);	
			pOptimizer->EvalCVSentence(G);
	  
			for(int i = 0; i < FeatureSets.size(); i++) {
				FeatureSets[i]->RemoveFeatures();
			}
	  
			SourceWords.clear();
			TargetWords.clear();
			SourcePOS.clear();
			TargetPOS.clear();
			alignment.clear();
			possibleAlignment.clear();

			if(preAlignmentFile != NULL) {
				preAlignment.clear();
				possiblePreAlignment.clear();
			}

			delete G;
		}
	}
	pOptimizer->CheckScalingFactorsWithCV();
}

void ReadLine(ifstream * stream, vector<string> & words) {
	assert(words.size() == 0);
	string line = "";
	string word = "";
	getline(*stream,line);
	if(line == "") {
		return;
	}
	line.push_back(' ');
	//cout << "Read Line:" << line << endl;
	stringstream sstream(stringstream::in|stringstream::out);
	sstream << line;
	sstream >> word;
	while (!sstream.eof()) {
		words.push_back(word);
		sstream >> word;
	}	
}

void ReadTree(ifstream * stream, ParseTree & tree) {  // Wei Chen 2008
        assert(&tree != NULL);

        string line = "";
        string word = "";
        getline(*stream,line);
        if(line == "") {
                return;
        }
        line.push_back(' ');
        stringstream sstream(stringstream::in|stringstream::out);
        sstream << line;
        sstream >> word;

        vector<string> treeString;

        while(!sstream.eof()) {
                treeString.push_back(word);
                sstream >> word;
        }

        tree.BuildTree(treeString);

}
void ReadAlignmentLine(ifstream * stream, vector<pair<int,int> > & alignment, vector<pair<int,int> > & possibleAlignment) {
	assert(alignment.size() == 0);
	assert(alignment.size() == 0);
	string line = "";
	string word = "";
	getline(*stream,line);
	if(line == "") {
		return;
	}
	line.push_back(' ');
	stringstream sstream(stringstream::in|stringstream::out);
	sstream << line;
	sstream >> word;
	while (!sstream.eof()) {
		int a;
		int b;
		char c;
    //cout << "String:" << word << endl;
		sscanf(word.c_str(),"%i%c%i",&a,&c,&b);
		if(c == '-') {
			//cout << "Alignment:" << a << " -> " << b << endl;
			alignment.push_back(make_pair(a - 1,b - 1));
			//cout << "Insert: " << alignment[alignment.size() -1].first << " -> " << alignment[alignment.size() -1].second << endl;
		}else if(c == 'p') {
			//cout << "Possible Alignment:" << a << " -> " << b << endl;
			possibleAlignment.push_back(make_pair(a - 1, b - 1));
			//cout << "Insert: " << possibleAlignment[possibleAlignment.size() -1].first << " -> " << possibleAlignment[possibleAlignment.size() -1].second << " as " <<possibleAlignment.size() -1<< ". Element " << endl;
		}else {
			cout << "Don't now character " << c << endl;
			exit(1);
		}
		sstream >> word;
	}

}
void ReadLinkFeatureLine(vector<ifstream *> & streams,vector<vector<LinkFeatures> > & features) {
	//cout << "New Sentence" << endl;
	assert(features.size() == 0);
	for(int i = 0; i < streams.size(); i++) {
		string line = "";
		string word = "";
		getline(*(streams[i]),line);
		if(line == "") {
			cout << "Empty link feature line" << endl;
			vector<LinkFeatures> links;
			features.push_back(links); 
		}else {
			vector<LinkFeatures> links;
			line.push_back(' ');
			stringstream sstream(stringstream::in|stringstream::out);
			sstream << line;
			sstream >> word;
			while (!sstream.eof()) {
				int a;
				int b;
				float c;
				//cout << "String:" << word << endl;
				sscanf(word.c_str(),"%i-%i#%f",&a,&b,&c);
				//cout << "Link Feature:" << a << " -> " << b << ": "<< c << endl;
				LinkFeatures f;
				f.source = a - 1;
				f.target = b - 1;
				f.value = c;
				links.push_back(f);
				sstream >> word;
			}

			features.push_back(links);
		}
	}
}

void InitLocalFeature(const char* ParamFile, const char* ScalingFactorString) 
{
	cout << "Init local Feature ..." << endl;
	vector<double> ScalingFactores;	
	GetScalingFactores(ScalingFactorString,ScalingFactores);
	numberOfOutputScalingFactors.push_back(ScalingFactores.size());
     //cout << " ParameterSize bevor: " << ScalingFactores.size() << endl;
	FeatureSet *pSet = new LocalFeatureSet(ParamFile,ScalingFactores);
	FeatureSets.push_back(pSet);
     //cout << " ParameterSize after: " << ScalingFactores.size() << endl;
	ScalingFactors.push_back(ScalingFactores);
	numberOfAllFeatures += ScalingFactores.size();
	cout << "Done!" << endl << endl;
}
void InitFertilityFeature(const char* ParamFile, const char* ScalingFactorString) 
{
	cout << "Init fertility Feature ..." << endl;
	vector<double> ScalingFactores;
	GetScalingFactores(ScalingFactorString,ScalingFactores);
	numberOfOutputScalingFactors.push_back(ScalingFactores.size());
     //cout << " Size before: " << ScalingFactores.size() << endl;
	FeatureSet *pSet = new FertilityFeatureSet(ParamFile,ScalingFactores);
     //cout << " Size after: " << ScalingFactores.size() << endl;
     //cout << "Last values: " << ScalingFactores[ScalingFactores.size() - 1] << endl;
	FeatureSets.push_back(pSet);
	ScalingFactors.push_back(ScalingFactores);
	numberOfAllFeatures += ScalingFactores.size();
	cout << "Done!" << endl << endl;
}

void InitFirstOrderFeature(const char* ParamFile, const char* ScalingFactorString) 
{
	cout << "Init first order Feature ..." << endl;
	vector<double> ScalingFactores;
	GetScalingFactores(ScalingFactorString,ScalingFactores);
	numberOfOutputScalingFactors.push_back(ScalingFactores.size());
	FeatureSet *pSet = new FirstOrderFeatureSet(ParamFile,ScalingFactores);
	FeatureSets.push_back(pSet);
	ScalingFactors.push_back(ScalingFactores);
	numberOfAllFeatures += ScalingFactores.size();
     
	cout << "Done!" << endl << endl;
}

void InitPhraseFeature(const char* ParamFile, const char* ScalingFactorString) 
{
	cout << "Init phrase Feature ..." << endl;
	vector<double> ScalingFactores;
	if(ScalingFactorString != NULL)
		GetScalingFactores(ScalingFactorString,ScalingFactores);
	numberOfOutputScalingFactors.push_back(ScalingFactores.size());
	FeatureSet *pSet = new PhraseFeatureSet(ParamFile,ScalingFactores);
	FeatureSets.push_back(pSet);
	ScalingFactors.push_back(ScalingFactores);
	numberOfAllFeatures += ScalingFactores.size();
     
	cout << "Done!" << endl << endl;
}

void GetScalingFactores(const char* ScalingFactorString, vector<double> & ScalingFactores) {
	char* List = strdup(ScalingFactorString);
	char * factores;
	int i = 1;
	factores = strtok (List,"_");
	while ( factores != NULL ){
	//cout << "Token is " << factores << endl << flush;
		ScalingFactores.push_back(atof( factores ));
	//cout << "Phrase Align Factor " << i << " is: " << PhraseAlignFactors[ i ] << endl;
		factores = strtok( NULL, "_" );
		i++;
	}
	free(List);
}

/*
void InitOptimizeFactors(char * Factors) {
  vector<int> ScalingFactores;
  char* List = strdup(Factors);
  char * factores;
  int i = 1;
  factores = strtok (List,"_");
  while ( factores != NULL ){
	//cout << "Token is " << factores << endl << flush;
	ScalingFactores.push_back(atoi( factores ));
	//cout << "Phrase Align Factor " << i << " is: " << PhraseAlignFactors[ i ] << endl;
	factores = strtok( NULL, "_" );
	i++;
}
  OptimizeScalingFactors.push_back(ScalingFactores);
}

void InitMeans(char * m) {
  // first init optimizeScalingfactors, then means for each featureset
  assert(mean.size() == OptimizeScalingFactors.size() - 1);
  if(m == NULL) {
    vector<double> currentMeans;
    int size = OptimizeScalingFactors[OptimizeScalingFactors.size() -1].size();
    currentMeans.resize(size,0);
    //cout << " Size: " << size << endl;
    mean.push_back(currentMeans);
}else {
    vector<double> currentMeans;
    char* List = strdup(m);
    char * factores;
    int i = 1;
    factores = strtok (List,"_");
    while ( factores != NULL ){
      //cout << "Token is " << factores << endl << flush;
      currentMeans.push_back(atof( factores ));
      //cout << "Phrase Align Factor " << i << " is: " << PhraseAlignFactors[ i ] << endl;
      factores = strtok( NULL, "_" );
      i++;
}
    mean.push_back(currentMeans);
}
  assert(mean[mean.size() -1].size() == OptimizeScalingFactors[OptimizeScalingFactors.size() - 1].size());
  cout << "Means for Scalingfactors set "<< mean.size() - 1 << endl;
  for(int i = 0; i < mean[mean.size() -1].size(); i ++) {
    cout <<  mean[mean.size() -1][i] << " ";
} 
  cout << endl;

}
*/

void LoadFactorFile(char * fileName, vector<vector<double> > & values) {
	ifstream file( fileName);

	if (!file) {
		cerr << "ERROR: File " << fileName << " could not be opened\n";
		exit(-2);
	}
  
	char *pFeaturSetIndex = new char[ 200 ];
	char *pBeginFeaturIndex = new char[ 200 ];
	char *pEndFeaturIndex = new char[ 200 ];
	char *pValue = new char[ 200 ];
	while ( !file.eof() ) {
		file >> pFeaturSetIndex >> pBeginFeaturIndex >> pEndFeaturIndex >> pValue;
		int featureSetIndex = atoi(pFeaturSetIndex);
		assert(featureSetIndex >= 0);
		assert(featureSetIndex < values.size());
		int beginFeatureIndex = atoi(pBeginFeaturIndex);
		int endFeatureIndex = atoi(pEndFeaturIndex);
		assert(beginFeatureIndex >= 0);
		assert(beginFeatureIndex <= endFeatureIndex);
		assert(endFeatureIndex < values[featureSetIndex].size());
    
		for(int i = beginFeatureIndex; i <= endFeatureIndex; i++) {
			values[featureSetIndex][i] = atof(pValue);
		}
	}
	delete[] pFeaturSetIndex;
	delete[] pBeginFeaturIndex;
	delete[] pEndFeaturIndex;
	delete[] pValue;

}

void InitOptimizer(int mode,char * learnRateFile, char * sigmaFile,double alpha) {
	switch(mode) {
		case OPTIMIZE_ML:
			pOptimizer = new OptimizerML(FeatureSets,ScalingFactors,learnRateFile);
			break;	
		case OPTIMIZE_MAP:
			pOptimizer = new OptimizerMAP(FeatureSets,ScalingFactors,learnRateFile,sigmaFile);
			break;	
		case OPTIMIZE_AER:
			pOptimizer = new OptimizerAER(FeatureSets,ScalingFactors,learnRateFile);
			break;
		case OPTIMIZE_RECALL:
			pOptimizer = new OptimizerRecall(FeatureSets,ScalingFactors,learnRateFile);
			break;
		case OPTIMIZE_PRECISION:
			pOptimizer = new OptimizerPrecision(FeatureSets,ScalingFactors,learnRateFile);
			break;
		case OPTIMIZE_AERMAP:
			pOptimizer = new OptimizerAERMAP(FeatureSets,ScalingFactors,learnRateFile);
			break;
		case OPTIMIZER_FSCORE:
			pOptimizer = new OptimizerFScore(FeatureSets,ScalingFactors,learnRateFile,alpha);
			break;
		default:
			cout << "Optimization mode: " << mode << " not supproted!" << endl;
			exit(1); 
	}
}
