
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define OPTIMIZE_MAP -1
#define OPTIMIZE_ML 1
#define OPTIMIZE_AER 2
#define OPTIMIZE_RECALL 3
#define OPTIMIZE_PRECISION 4
#define OPTIMIZE_AERMAP 5
#define OPTIMIZER_FSCORE 6
#define OPTIMIZE_ALL 100

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <vector>
#include <climits>
#include <cfloat>

#include <featureset.h>
#include <locallopfeatureset.h>
#include <fertilitylopfeatureset.h>
#include <firstorderlopfeatureset.h>
#include <phraselopfeatureset.h>
#include <param.hh>
#include <graph.h>
#include <optimizer.h>
#include <lopoptimizerml.h>
#include <lopoptimizeraer.h>
#include <lopoptimizerall.h>

using namespace std;

ifstream * LoadFile(const char* File);
vector<ifstream *> LoadFileList(char * File);
ofstream * OpenOutputFile(char* File);
void ReadLine(ifstream * stream, vector<string> & words);
void ReadAlignmentLine(ifstream * stream, vector<pair<int,int> > & alignemnt,vector<pair<int,int> > & possibleAlignemnt);
void ReadLinkFeatureLine(vector<ifstream *> & streams,vector<vector<LinkFeatures> > & features);
void InitExpertWeights(const char * string);

void InitLocalFeature(const char* ParamFile, const char* ScalingFactorList);
void InitFertilityFeature(const char* ParamFile, const char* ScalingFactorString);
void InitFirstOrderFeature(const char * ParamFile, const char* ScalingFactorString);
void InitPhraseFeature(const char * ParamFile, const char* ScalingFactorString);
void InitOptimizer(int mode,char * learnRateFile, double alpha);

void AlignSentence(char *Source, char *Target,char *SourcePOS, char *TargetPOS,char * OutputFile, char * LinkFeatureList, int iterations,int DisplayGraph);
void Optimize(char * Source, char * Target,char * SourcePOS, char * TargetPOS, char * AligmnetFile, char * LinkFeatureList, int OptimizeIterations,int BeliefIterations,int DisplayGraph,int optimize);

CParam             *pParam             = NULL;
Optimizer *pOptimizer = NULL;
vector <FeatureSet*> FeatureSets;


vector <int> ScalingFactorSize; // ScalingFactors[i] = ScalingFactors for FeatureSets[i]
int numberOfAllSF = 0;
//vector <int> numberOfOutputScalingFactors;
int numberOfExperts = 0;
double * expertWeights;
int main(int argc, char *argv[])
{
	cout << "          Word Aligner" << endl;
	cout << "--------------------------------" << endl;

	pParam = new CParam;
	int iterations;
	int optimize;
	int OptimizeIterations;
	char* SourceFile = NULL;
	char* TargetFile = NULL;
	char* SourcePOSFile = NULL;
	char* TargetPOSFile = NULL;
	char* SourceTestFile = NULL;
	char* TargetTestFile = NULL;
	char* SourceTestPOSFile = NULL;
	char* TargetTestPOSFile = NULL;
	char* AlignmentFile = NULL;
	char* OutputFile = NULL;

	char* LinkFeatureList = NULL;
	char* TestLinkFeatureList = NULL;

	char* LFParamFile = NULL;
	char* LFScalingFactors = NULL;
	char* FFParamFile = NULL;
	char* FFScalingFactors = NULL;
	char* FOFParamFile = NULL;
	char* FOFScalingFactors = NULL;
	char* PFParamFile = NULL;
	char* PFScalingFactors = NULL;
	char* learnRateFile = NULL;
	double fScoreAlpha = 0.5;

	int DisplayGraph = 0;

	char * expertWeightString = NULL;

	pParam->AddParamDef( "SourceFile", 'S', &SourceFile, NULL,
			     "List of all Source Sentences",
	False );
	pParam->AddParamDef( "TargetFile", 'T', &TargetFile, NULL,
 			     "List of all Target Sentences",
	False );
	pParam->AddParamDef( "SourcePOSFile", '\0', &SourcePOSFile, NULL,
			     "List of all Source POS Sentences",
	True );
	pParam->AddParamDef( "TargetPOSFile", '\0', &TargetPOSFile, NULL,
			     "List of all Target POS Sentences",
	True );
	pParam->AddParamDef( "SourceTestFile", '\0', &SourceTestFile, NULL,
			     "List of all Source Test Sentences",
	True );
	pParam->AddParamDef( "TargetTestFile", '\0', &TargetTestFile, NULL,
			     "List of all Target Test Sentences",
	True );
	pParam->AddParamDef( "SourceTestPOSFile", '\0', &SourceTestPOSFile, NULL,
			     "List of all Source Test POS Sentences",
	True );
	pParam->AddParamDef( "TargetTestPOSFile", '\0', &TargetTestPOSFile, NULL,
			     "List of all Target Test POS Sentences",
	True );
	pParam->AddParamDef( "AlignmentFile", 'A', &AlignmentFile, NULL,
			     "List of all alignments in pharao-style",
	True );
	pParam->AddParamDef( "LinkFeatureFileList", '\0', &LinkFeatureList, NULL,
			     "File containing filesnames of links set by other word aligners",
	True );
	pParam->AddParamDef( "TestLinkFeatureFileList", '\0', &TestLinkFeatureList, NULL,
			     "File containing filesnames of links set by other word aligners",
	True );
	pParam->AddParamDef( "OutputFile", '\0', &OutputFile, NULL,
			     "Prints found alignment to this file",
	True );
	pParam->AddParamDef( "LocalFeatureParamFile", 'P', &LFParamFile, NULL,
			     "Parameterfile for the local features",
	False );
	pParam->AddParamDef( "LocalFeatureScalingFactors", 'L', &LFScalingFactors, NULL,
			     "Scaling Factors for the local features",
	False );
	pParam->AddParamDef( "FertilityFeatureScalingFactors", 'F', &FFScalingFactors, NULL,
			     "Scaling Factors for the fertility features",
	True );
	pParam->AddParamDef( "FertilityFeatureParamFile", '\0', &FFParamFile, NULL,
			     "Parameterfile for the fertility features",
	True );
	pParam->AddParamDef( "FirstOrderFeatureParamFile", '\0', &FOFParamFile, NULL,
			     "Parameterfile for the first order features",
	True );
	pParam->AddParamDef( "FirstOrderScalingFactors", 'O', &FOFScalingFactors, NULL,
			     "Scaling Factors for the first order features",
	True );
	pParam->AddParamDef( "PhraseFeatureParamFile", '\0', &PFParamFile, NULL,
			     "Parameterfile for the local features",
	True );
	pParam->AddParamDef( "PhraseFeatureScalingFactors", '\0', &PFScalingFactors, NULL,
			     "Scaling Factors for the local features",
	True );
	pParam->AddParamDef( "Iterations", 'I', &iterations, 10,
			     "Number of Iterations of the Belief Properations algorithm",
	True );
	pParam->AddParamDef( "DisplayGraph", 'D', &DisplayGraph, 0,
			     "File for every Sentence is created containing the Graph at all timesteps",
	True );
	pParam->AddParamDef( "Optimize", 'o', &optimize, 0,
			     "1 = Optimize; 0 = Align Sentences",
	True );
	pParam->AddParamDef( "OptimizeIterations", 'i', &OptimizeIterations, 10,
			     "Iterations of the Optimization",
	True );
	pParam->AddParamDef( "LearnRate", '\0', &learnRateFile, NULL,
			     "String containing the learn rates",
	True );
	pParam->AddParamDef( "FScoreAlpha", '\0', &fScoreAlpha, 0.5,
			     "Alpha-Value for F-Score",
	True );

	pParam->AddParamDef( "NumberOfExperts", '\0', &numberOfExperts, 0,
			     "Number of Experts used in LOP",
	False );
	
	pParam->AddParamDef( "ExpertWeights", '\0', &expertWeightString, 0,
			     "Weights for the different experts",
	False );

	if (! pParam->Interpret( argv, argc) ) {
		cerr << "\nParameters could not be interpreted correctly\n ... exiting!!\n";
		exit(1); 
	}

	
	InitExpertWeights(expertWeightString);
	//init lokal features
	InitLocalFeature(LFParamFile,LFScalingFactors);

	if (FFScalingFactors != NULL) {
		InitFertilityFeature(FFParamFile,FFScalingFactors);
	}else {
		cout << "No FertilityFeatures used" << endl;
	}

	if (FOFScalingFactors != NULL) {
		InitFirstOrderFeature(FOFParamFile,FOFScalingFactors);
	}else {
		cout << "No FirstOrderFeatures used" << endl;
	}

	if(PFParamFile != NULL) {
		InitPhraseFeature(PFParamFile,PFScalingFactors);
	}else {
		cout << "No PhraseFeatures used" << endl;
	}


	if(optimize) {
		InitOptimizer(optimize,learnRateFile,fScoreAlpha);
		if(AlignmentFile == NULL) {
			cout << "For Optimization a alignment file is needed" << endl;
			exit(1);
		}
		Optimize(SourceFile, TargetFile,SourcePOSFile, TargetPOSFile,AlignmentFile,LinkFeatureList,OptimizeIterations,iterations,DisplayGraph,optimize);

		if((SourceTestFile != NULL) && (TargetTestFile != NULL)) {
			if(OutputFile == NULL) {
				cout << "For Alginment an output file needed" << endl;
				exit(1);
			}
			cout << " Algin Test sentences!" << endl;
			AlignSentence(SourceTestFile, TargetTestFile,SourceTestPOSFile, TargetTestPOSFile,OutputFile,TestLinkFeatureList,iterations,DisplayGraph);
			cout << "Test Sentences algined!" << endl;

		}


	}else { 
		if(OutputFile == NULL) {
			cout << "For Alginment an output file needed" << endl;
			exit(1);
		}
		AlignSentence(SourceFile, TargetFile,SourcePOSFile, TargetPOSFile,OutputFile, LinkFeatureList, iterations,DisplayGraph);
	}


	for(int i = 0; i < FeatureSets.size(); i++) {
		delete FeatureSets[i];
	}
	delete[] expertWeights;
	return EXIT_SUCCESS;
}


ifstream * LoadFile(const char* File) {
	ifstream * stream = new ifstream(File);
	if(!(*stream)) {
		cerr << "ERROR: File " << File <<" could not be open\n";
		exit(-2);
	}
	return stream;
}
vector<ifstream *> LoadFileList(char * File) {

	vector <ifstream *> Files;
	ifstream * stream = new ifstream(File);
	if(!(*stream)) {
		cerr << "ERROR: File list " << File <<" could not be open\n";
		exit(-2);
	}
	string line = "";
	string word = "";
	while(getline(*stream,line)) {
		//cout << "Load File: " << line << endl;
		Files.push_back(LoadFile(line.c_str()));
	}

	return Files;
}

ofstream * OpenOutputFile(char* File) {
	ofstream * stream = new ofstream(File);
	if(!(*stream)) {
		cerr << "ERROR: File " << File <<" could not be open\n";
		exit(-2);
	}
	return stream;
}

void AlignSentence(char *SourceFile, char *TargetFile,char *SourcePOSFile, char *TargetPOSFile,char * OutputFile, char * LinkFeatureList, int iterations,int DisplayGraph){
	ifstream * Source = LoadFile(SourceFile);
	ifstream * Target = LoadFile(TargetFile);
	ifstream * SourcePOSStream = NULL;
	ifstream * TargetPOSStream = NULL;
	if(SourcePOSFile != NULL) {
		SourcePOSStream = LoadFile(SourcePOSFile);
		TargetPOSStream = LoadFile(TargetPOSFile);
	}
	ofstream * output = OpenOutputFile(OutputFile);

	vector<string> SourceWords;
	vector<string> TargetWords;
	vector<string> SourcePOS;
	vector<string> TargetPOS;
	int Sent = 0;
	ofstream * GraphOut;
	vector<ifstream *> LinkFeatureFiles;
	if(LinkFeatureList != NULL) {
		LinkFeatureFiles = LoadFileList(LinkFeatureList);
	}
	vector<vector<LinkFeatures> > linkFeatures;

	while((!Source->eof()) && (!Target->eof())) 
	{	

		ReadLine(Source,SourceWords);
		ReadLine(Target,TargetWords);
		if(SourcePOSStream != NULL) {
			ReadLine(SourcePOSStream,SourcePOS);
		}
		if(TargetPOSStream != NULL) {
			ReadLine(TargetPOSStream,TargetPOS);
		}

		if(LinkFeatureFiles.size() != 0) {
			ReadLinkFeatureLine(LinkFeatureFiles,linkFeatures);	
		}

		if((SourceWords.size() > 0) && (TargetWords.size() > 0)) {
			if(DisplayGraph) {
				Sent++;
				char filename[256];
				sprintf(filename, "Sentence%d.gda", Sent);
				//filename += string(Sent);
				GraphOut = new ofstream(filename);
				*GraphOut << SourceWords.size() << " " <<TargetWords.size() << " " << iterations << endl;
				for(int i = 0; i < SourceWords.size(); i++) {
					*GraphOut << SourceWords[i] << " ";
				}
				*GraphOut << endl;
				for(int i = 0; i < TargetWords.size(); i++) {
					*GraphOut << TargetWords[i] << " ";
				}
				*GraphOut << endl;
				*GraphOut << "#Format: time sourceindex targetindex belief belief Lok.messageOut Lok.messageOut Lok.messageIn Lok.messageIn";
				*GraphOut << "SourceFert.messageOut SourceFert.messageOut SourceFert.messageIn SourceFert.messageIn";
				*GraphOut << "TargetFert.messageOut TargetFert.messageOut TargetFert.messageIn TargetFert.messageIn";
				*GraphOut << endl;
			}
			/*cout << "Source: ";
			for(int i = 0; i < SourceWords.size(); i++) {
			cout << i << SourceWords[i] << " ";
		}
			cout << endl;	
			cout << "SourceLength: " << SourceWords.size() << endl;
			cout << "Target: ";
			for(int i = 0; i < TargetWords.size(); i++) {
			cout << i << TargetWords[i] << " ";
		}
			cout << endl;
			cout << "TargetLength: " << TargetWords.size() << endl;
			*/
			//Build Graph
			Graph * G = new Graph(SourceWords.size(), TargetWords.size(),SourceWords,TargetWords,SourcePOS,TargetPOS);

			//give LinkFeatures to Graph
			if(linkFeatures.size() != 0) {
				G->SetLinkFeatures(linkFeatures);
			}

			for(int i = 0; i < FeatureSets.size(); i++) {
				FeatureSets[i]->GenerateFeatures(G,-1,0);
				G->AddFeatureSet(FeatureSets[i]);
			}
			if(DisplayGraph) {
				G->SetDisplay(GraphOut);
			}
			G->Inference(iterations);
			G->OutputAlignment(output);
			for(int i = 0; i < FeatureSets.size();i++) {
				FeatureSets[i]->RemoveFeatures();
			}
			SourceWords.clear();
			TargetWords.clear();
			SourcePOS.clear();
			TargetPOS.clear();
			//delete linksFeatures for Sentence
			if(linkFeatures.size() != 0) {
				linkFeatures.clear();
			}
			if(DisplayGraph) {
				GraphOut->close();
				delete GraphOut;
			}
			delete G;
		}
	}
	//delete file handlers for link feature files
	delete Source;
	delete Target;
	if(SourcePOSFile != NULL) {
		delete SourcePOSStream;
		delete TargetPOSStream;
	}
	delete output;

	if(LinkFeatureList != NULL) {
		for(int i = 0; i < LinkFeatureFiles.size(); i++) {
		delete LinkFeatureFiles[i];
			
		}
	}


}

void Optimize(char * SourceFile, char * TargetFile,char * SourcePOSFile, char * TargetPOSFile, char * AlignmentFile, char * LinkFeatureList, int OptimizationIterations,int BeliefIterations,int DisplayGraph,int optimize) {

	for(int k = 0; k < OptimizationIterations; k++) {

		ifstream * Source = LoadFile(SourceFile);
		ifstream * Target = LoadFile(TargetFile);
		ifstream * SourcePOSStream = NULL;
		ifstream * TargetPOSStream = NULL;
		if(SourcePOSFile != NULL) {
			SourcePOSStream = LoadFile(SourcePOSFile);
			TargetPOSStream = LoadFile(TargetPOSFile);
		}
		ifstream * alignmentStream = LoadFile(AlignmentFile);
		ifstream * preAlignmentStream = NULL;
		vector<ifstream *> LinkFeatureFiles;
		if(LinkFeatureList != NULL) {
			LinkFeatureFiles = LoadFileList(LinkFeatureList);
		}
		vector<string> SourceWords;
		vector<string> TargetWords;
		vector<string> SourcePOS;
		vector<string> TargetPOS;
		vector<pair<int,int> > alignment;
		vector<pair<int,int> > possibleAlignment;

		vector<vector<LinkFeatures> > linkFeatures;
		cout << "Iteration " << k << endl;

		int Sent = 0;
		ofstream * GraphOut;
		while((!Source->eof()) && (!Target->eof())) 
		{	
			ReadLine(Source,SourceWords);
			ReadLine(Target,TargetWords);
			if(SourcePOSStream != NULL) {
				ReadLine(SourcePOSStream,SourcePOS);
			}
			if(TargetPOSStream != NULL) {
				ReadLine(TargetPOSStream,TargetPOS);
			}
			ReadAlignmentLine(alignmentStream,alignment,possibleAlignment);

			//read lines from linkFeatureFiles
			if(LinkFeatureFiles.size() != 0) {
				ReadLinkFeatureLine(LinkFeatureFiles,linkFeatures);	
			}

			if((SourceWords.size() > 0) && (TargetWords.size() > 0)) {
				Sent++;
				if(DisplayGraph) {
					char filename[256];
					sprintf(filename, "Sentence%d.gda", Sent);
	    //filename += string(Sent);
					GraphOut = new ofstream(filename);
					*GraphOut << SourceWords.size() << " " <<TargetWords.size() << " " << BeliefIterations << endl;
					for(int i = 0; i < SourceWords.size(); i++) {
						*GraphOut << SourceWords[i] << " ";
					}
					*GraphOut << endl;
					for(int i = 0; i < TargetWords.size(); i++) {
						*GraphOut << TargetWords[i] << " ";
					}
					*GraphOut << endl;
					*GraphOut << "#Format: time sourceindex targetindex belief belief Lok.messageOut Lok.messageOut Lok.messageIn Lok.messageIn";
					*GraphOut << "SourceFert.messageOut SourceFert.messageOut SourceFert.messageIn SourceFert.messageIn";
					*GraphOut << "TargetFert.messageOut TargetFert.messageOut TargetFert.messageIn TargetFert.messageIn";
					*GraphOut << endl;
				}

/*				for(int i = 0; i < SourceWords.size(); i++) {
					cout << SourceWords[i] << " ";
				}
				cout << endl;
				for(int i = 0; i < TargetWords.size(); i++) {
					cout << TargetWords[i] << " ";
				}
				cout << endl;
*/				//cout << "Sentence: " << Sent << endl << flush;
	  //Build Graph
				Graph * G = new Graph(SourceWords.size(), TargetWords.size(),SourceWords,TargetWords,SourcePOS,TargetPOS,alignment,possibleAlignment);

				//give LinkFeatures to Graph
				if(linkFeatures.size() != 0) {
					G->SetLinkFeatures(linkFeatures);
				}

				int offset = 0;
				for(int i = 0; i < FeatureSets.size(); i++) {
					if(optimize >= 2) {
						FeatureSets[i]->GenerateFeatures(G,offset,numberOfAllSF);
						offset += ScalingFactorSize[i];
					}else {
						FeatureSets[i]->GenerateFeatures(G,-1,0);	      
					}
					G->AddFeatureSet(FeatureSets[i]);
				}
				if(DisplayGraph) {
					G->SetDisplay(GraphOut);
				}
				if(optimize >= 2) {
					G->Inference(BeliefIterations);
				}else {
					G->Optimize(BeliefIterations);
				}	    
				pOptimizer->EvalSentence(G);

	  
				for(int i = 0; i < FeatureSets.size(); i++) {
					FeatureSets[i]->RemoveFeatures();
				}
	  
				SourceWords.clear();
				TargetWords.clear();
				SourcePOS.clear();
				TargetPOS.clear();
				alignment.clear();
				possibleAlignment.clear();

				//delete linksFeatures for Sentence
				if(linkFeatures.size() != 0) {
					linkFeatures.clear();
				}

				if(DisplayGraph) {
					GraphOut->close();
					delete GraphOut;
				}
				delete G;
			}
		}
    
		pOptimizer->CheckScalingFactors();
		pOptimizer->Estimate();
		delete Source;
		delete Target;
		if(SourcePOSFile != NULL) {
			delete SourcePOSStream;
			delete TargetPOSStream;
		}
		delete alignmentStream;

		//delete file handlers for link feature files
		if(LinkFeatureList != NULL) {
			for(int i = 0; i < LinkFeatureFiles.size(); i++) {
				delete LinkFeatureFiles[i];
				
			}
		}
	}

	pOptimizer->OutputBest();

}
void AlignCV(char * SourceFile, char * TargetFile,char * SourcePOSFile, char * TargetPOSFile,char * AlignmentFile, char * preAlignmentFile,char * LinkFeatureList, int Iterations) {
	ifstream * Source = LoadFile(SourceFile);
	ifstream * Target = LoadFile(TargetFile);
	ifstream * SourcePOSStream = NULL;
	ifstream * TargetPOSStream = NULL;
	if(SourcePOSFile != NULL) {
		SourcePOSStream = LoadFile(SourcePOSFile);
		TargetPOSStream = LoadFile(TargetPOSFile);
	}
	ifstream * alignmentStream = LoadFile(AlignmentFile);
	ifstream * preAlignmentStream = NULL;
	if(preAlignmentFile != NULL) {
		preAlignmentStream = LoadFile(preAlignmentFile);
	}
	vector<string> SourceWords;
	vector<string> TargetWords;
	vector<string> SourcePOS;
	vector<string> TargetPOS;
	vector<pair<int,int> > alignment;
	vector<pair<int,int> > possibleAlignment;

	vector<pair<int,int> > preAlignment;
	vector<pair<int,int> > possiblePreAlignment;

	int Sent = 0;
	while((!Source->eof()) && (!Target->eof())) 
	{	
		ReadLine(Source,SourceWords);
		ReadLine(Target,TargetWords);
		if(SourcePOSStream != NULL) {
			ReadLine(SourcePOSStream,SourcePOS);
		}
		if(TargetPOSStream != NULL) {
			ReadLine(TargetPOSStream,TargetPOS);
		}
		ReadAlignmentLine(alignmentStream,alignment,possibleAlignment);

		if(preAlignmentFile != NULL) {
			ReadAlignmentLine(preAlignmentStream,preAlignment,possiblePreAlignment);	
		}
			
		if((SourceWords.size() > 0) && (TargetWords.size() > 0)) {
			Sent++;
			//cout << "Sentence: " << Sent << endl << flush;
			//Build Graph
			Graph * G = new Graph(SourceWords.size(), TargetWords.size(),SourceWords,TargetWords,SourcePOS,TargetPOS,alignment,possibleAlignment);
			if(preAlignmentFile != NULL) {
				G->LoadPreAlignment(preAlignment,possiblePreAlignment);
			}
				
			for(int i = 0; i < FeatureSets.size(); i++) {
				FeatureSets[i]->GenerateFeatures(G,-1,0);	      
				G->AddFeatureSet(FeatureSets[i]);
			}
			G->Inference(Iterations);
			//G->Optimize(Iterations);	
			pOptimizer->EvalCVSentence(G);
	  
			for(int i = 0; i < FeatureSets.size(); i++) {
				FeatureSets[i]->RemoveFeatures();
			}
	  
			SourceWords.clear();
			TargetWords.clear();
			SourcePOS.clear();
			TargetPOS.clear();
			alignment.clear();
			possibleAlignment.clear();

			if(preAlignmentFile != NULL) {
				preAlignment.clear();
				possiblePreAlignment.clear();
			}

			delete G;
		}
	}
	pOptimizer->CheckScalingFactorsWithCV();
}

void ReadLine(ifstream * stream, vector<string> & words) {
	assert(words.size() == 0);
	string line = "";
	string word = "";
	getline(*stream,line);
	if(line == "") {
		return;
	}
	line.push_back(' ');
	//cout << "Read Line:" << line << endl;
	stringstream sstream(stringstream::in|stringstream::out);
	sstream << line;
	sstream >> word;
	while (!sstream.eof()) {
		words.push_back(word);
		sstream >> word;
	}	
}
void ReadAlignmentLine(ifstream * stream, vector<pair<int,int> > & alignment, vector<pair<int,int> > & possibleAlignment) {
	assert(alignment.size() == 0);
	assert(alignment.size() == 0);
	string line = "";
	string word = "";
	getline(*stream,line);
	if(line == "") {
		return;
	}
	line.push_back(' ');
	stringstream sstream(stringstream::in|stringstream::out);
	sstream << line;
	sstream >> word;
	while (!sstream.eof()) {
		int a;
		int b;
		char c;
    //cout << "String:" << word << endl;
		sscanf(word.c_str(),"%i%c%i",&a,&c,&b);
		if(c == '-') {
			//cout << "Alignment:" << a << " -> " << b << endl;
			alignment.push_back(make_pair(a - 1,b - 1));
			//cout << "Insert: " << alignment[alignment.size() -1].first << " -> " << alignment[alignment.size() -1].second << endl;
		}else if(c == 'p') {
			//cout << "Possible Alignment:" << a << " -> " << b << endl;
			possibleAlignment.push_back(make_pair(a - 1, b - 1));
			//cout << "Insert: " << possibleAlignment[possibleAlignment.size() -1].first << " -> " << possibleAlignment[possibleAlignment.size() -1].second << " as " <<possibleAlignment.size() -1<< ". Element " << endl;
		}else {
			cout << "Don't now character " << c << endl;
			exit(1);
		}
		sstream >> word;
	}

}
void ReadLinkFeatureLine(vector<ifstream *> & streams,vector<vector<LinkFeatures> > & features) {
	//cout << "New Sentence" << endl;
	assert(features.size() == 0);
	for(int i = 0; i < streams.size(); i++) {
		string line = "";
		string word = "";
		getline(*(streams[i]),line);
		if(line == "") {
			cout << "Empty link feature line" << endl;
			vector<LinkFeatures> links;
			features.push_back(links); 
		}else {
			vector<LinkFeatures> links;
			line.push_back(' ');
			stringstream sstream(stringstream::in|stringstream::out);
			sstream << line;
			sstream >> word;
			while (!sstream.eof()) {
				int a;
				int b;
				float c;
				//cout << "String:" << word << endl;
				sscanf(word.c_str(),"%i-%i#%f",&a,&b,&c);
				//cout << "Link Feature:" << a << " -> " << b << ": "<< c << endl;
				LinkFeatures f;
				f.source = a - 1;
				f.target = b - 1;
				f.value = c;
				links.push_back(f);
				sstream >> word;
			}

			features.push_back(links);
		}
	}
}
void InitExpertWeights(const char * string) {
	expertWeights = new double[numberOfExperts];
	char* List = strdup(string);
	char * factores;
	int i = 0;
	factores = strtok (List,"_");
	while ( factores != NULL ){
		assert(i < numberOfExperts);
		expertWeights[i] = atof( factores );
	//cout << "Phrase Align Factor " << i << " is: " << PhraseAlignFactors[ i ] << endl;
		factores = strtok( NULL, "_" );
		i++;
	}
	free(List);
}

void InitLocalFeature(const char* ParamFile, const char* ScalingFactorString) 
{
	cout << "Init local Feature ..." << endl;
/*	vector<double> ScalingFactores;	
	GetScalingFactores(ScalingFactorString,ScalingFactores);
	numberOfOutputScalingFactors.push_back(ScalingFactores.size());
     //cout << " ParameterSize bevor: " << ScalingFactores.size() << endl;
*/
	FeatureSet *pSet = new LocalLOPFeatureSet(ParamFile,ScalingFactorString,numberOfExperts,expertWeights);
	FeatureSets.push_back(pSet);
	ScalingFactorSize.push_back(pSet->GetNumberOfScalingFactors());
	numberOfAllSF += pSet->GetNumberOfScalingFactors();
	cout << "Local Features successfull initialized" << endl << endl;
}

void InitFertilityFeature(const char* ParamFile, const char* ScalingFactorString) 
{
	cout << "Init fertility Feature ..." << endl;
	FeatureSet *pSet = new FertilityLOPFeatureSet(ParamFile,ScalingFactorString,numberOfExperts,expertWeights);
	FeatureSets.push_back(pSet);
	ScalingFactorSize.push_back(pSet->GetNumberOfScalingFactors());
	numberOfAllSF += pSet->GetNumberOfScalingFactors();
	cout << "Fertility Feature successfull initialized" << endl << endl;
}

void InitFirstOrderFeature(const char* ParamFile, const char* ScalingFactorString) 
{
	cout << "Init first order Feature ..." << endl;
	FeatureSet *pSet = new FirstOrderLOPFeatureSet(ParamFile,ScalingFactorString,numberOfExperts,expertWeights);
	FeatureSets.push_back(pSet);
	ScalingFactorSize.push_back(pSet->GetNumberOfScalingFactors());
	numberOfAllSF += pSet->GetNumberOfScalingFactors();
	cout << "FirstOrder Feature successfull initialized" << endl << endl;
}
void InitPhraseFeature(const char* ParamFile, const char* ScalingFactorString) 
{
	cout << "Init phrase Feature ..." << endl;
	FeatureSet *pSet = new PhraseLOPFeatureSet(ParamFile,ScalingFactorString,numberOfExperts,expertWeights);
	FeatureSets.push_back(pSet);
	ScalingFactorSize.push_back(pSet->GetNumberOfScalingFactors());
	numberOfAllSF += pSet->GetNumberOfScalingFactors();
     
	cout << "Phrase Feature successfull initialized" << endl << endl;
}


void InitOptimizer(int mode,char * learnRateFile, double alpha) {
	switch(mode) {
		case OPTIMIZE_ML:
			pOptimizer = new LOPOptimizerML(FeatureSets,expertWeights,learnRateFile,numberOfExperts);
			break;	
		case OPTIMIZE_AER:
			pOptimizer = new LOPOptimizerAER(FeatureSets,expertWeights,learnRateFile,numberOfExperts);
			break;
		case OPTIMIZE_ALL:
			pOptimizer = new LOPOptimizerAll(FeatureSets,expertWeights,learnRateFile,numberOfExperts);
			break;
/*		case OPTIMIZE_PRECISION:
			pOptimizer = new OptimizerPrecision(FeatureSets,ScalingFactors,learnRateFile);
			break;
		case OPTIMIZE_AERMAP:
			pOptimizer = new OptimizerAERMAP(FeatureSets,ScalingFactors,learnRateFile);
			break;
		case OPTIMIZER_FSCORE:
			pOptimizer = new OptimizerFScore(FeatureSets,ScalingFactors,learnRateFile,alpha);
			break;
*/		default:
			cout << "Optimization mode: " << mode << " not supproted!" << endl;
			exit(1); 
	}
}

