#include <stdlib.h>
/*
 *  EvaluatorTER.cc
 *  Created by Anton D'Auria (apd@andrew.cmu.edu) on 4/9/06.
 *  Original perl implementation by Matthew Snover (snover@cs.umd.edu) 
 *
 */
#include <fstream>
#include <iostream>

#include <EvaluatorTER_MER.hh>
using namespace std;


//====================================================================
//      CEvaluatorTER_MER:: 
//

//CEvaluatorTER_MER::CEvaluatorTER_MER( CVocabulary *pVocabulary, CEvaluatorSharedData *pSharedData )
CEvaluatorTER_MER::CEvaluatorTER_MER()
{
  m_MAX_SHIFT_SIZE = 10;
  m_MAX_SHIFT_DIST = 50;
  m_MAX_BEAM_WIDTH = 10;
  
  m_NUM_SHIFTS_CONSIDERED = 0;
  m_NUM_BEAM_SEARCH_CALLS = 0;
  
  //11 total elements
  m_SCORE_KEY          = 0;
  m_INSERTION_KEY      = 10;
  m_DELETION_KEY       = 1;
  m_SUBSTITUTION_KEY   = 2;
  m_SHIFT_KEY          = 3;
  m_MATCH_KEY          = 4;
  m_HYP_WORD_COUNT_KEY = 5;
  m_REF_WORD_COUNT_KEY = 6;
  m_EDIT_COUNT_KEY     = 7;
  m_REF_WORD_COUNT_KEY = 8;
  m_HYP_SENT_COUNT_KEY = 9;
  
  //m_IN_FIND_BEST_SHIFT = 0;
}



CEvaluatorTER_MER::~CEvaluatorTER_MER()
{
}


//------------------------------------------------------------------
//      CEvaluatorTER_MER:: Evaluate
//

/*
bool
CEvaluatorTER_MER::Evaluate( const char *pFileWithTranslations, 
                          CNumericalArray< float > & Scores )
{
}



bool 
CEvaluatorTER_MER::Evaluate( CArray< CString > &Translations, 
                         CNumericalArray< float > & Scores )
{
}


bool
CEvaluatorTER_MER::Evaluate( CText* pTranslations, 
                         CNumericalArray< float > & Scores )
{
  
  CArray< CArray< CSentence* > >* p_refSentences = m_p_sharedData->GetReferences();
  
  CSentence* p_currHypSentence;
  CSentence* p_currRefSentence;
  
  int hypSentCount = pTranslations->NumberOfSentences();
  int refSentCount;                    // per hyp sentence
 
  Scores[m_INSERTION_KEY]      = 0;                       
  Scores[m_DELETION_KEY]       = 0;                       
  Scores[m_SUBSTITUTION_KEY]   = 0; 
  Scores[m_SHIFT_KEY]          = 0;
  Scores[m_MATCH_KEY]          = 0;
  Scores[m_HYP_WORD_COUNT_KEY] = 0;
  Scores[m_REF_WORD_COUNT_KEY] = 0;
  Scores[m_EDIT_COUNT_KEY]     = 0;
  Scores[m_HYP_SENT_COUNT_KEY] = hypSentCount;
  
  for(int h=0; h < hypSentCount; h++)
  {
    p_currHypSentence = pTranslations->GetSentence(h);
    
    refSentCount = (int)((*p_refSentences)[h].Size());
    
    CEvaluatorTER_MERalignment* p_bestForAllRefAlignment = 0;
    
    for(int r=0; r < refSentCount; r++)
    {
      p_currRefSentence = (*p_refSentences)[h][r];

      // create valid subphrases out of ref sentence
      PhraseToSet_hash_map* p_phraseStartHash 
        = BuildWordMatches(p_currHypSentence, p_currRefSentence);
      
      // Begin min edit distance on given hyp and ref, w/o shifts
      
      Phrase hypPhrase = BuildPhrase(p_currHypSentence);
      Phrase hypPrimePhrase = hypPhrase;
      Phrase refPhrase = BuildPhrase(p_currRefSentence);
      CEvaluatorTER_MERcost costFunc;
      
      CEvaluatorTER_MERalignment* p_curHypRefAlignment
        = ComputeMinEditDistance(hypPhrase, refPhrase, costFunc);

      p_curHypRefAlignment->SetHyp(hypPhrase);
      p_curHypRefAlignment->SetRef(refPhrase);
      p_curHypRefAlignment->SetHypPrime(hypPrimePhrase); //in UMD, it's "aftershift"
      
      float edits = 0;
      int numShifts = 0;
      
      std::vector< CEvaluatorTER_MERshift* >* 
        p_allShiftsVector = new std::vector< CEvaluatorTER_MERshift* >;
            
      bool moreShifts = true;
      while (moreShifts) 
      {
        std::pair< CEvaluatorTER_MERshift*, CEvaluatorTER_MERalignment* >* 
        p_bestShiftAlignmentPair = FindBestShift(hypPrimePhrase, hypPhrase, 
                                                 refPhrase, p_phraseStartHash,
                                                 p_curHypRefAlignment, costFunc);
        if (p_bestShiftAlignmentPair == 0) {
          moreShifts = false;
        } else {
          CEvaluatorTER_MERshift* p_bestShift = p_bestShiftAlignmentPair->first;
          edits += p_bestShift->GetCost();
          p_allShiftsVector->push_back(p_bestShift);
          
          //delete p_curHypRefAlignment;
          p_curHypRefAlignment = p_bestShiftAlignmentPair->second;
          hypPrimePhrase = p_curHypRefAlignment->GetHypPrime();
        }        
      }
       
      p_curHypRefAlignment->SetAllShifts(p_allShiftsVector);
      p_curHypRefAlignment
        ->SetNumEdits(edits+p_curHypRefAlignment->GetNumEdits());
      
      //p_curHypRefAlignment->Print();
      
      if (p_bestForAllRefAlignment == 0) {
        p_bestForAllRefAlignment = p_curHypRefAlignment;
      } else if( p_curHypRefAlignment->ComputeScore() 
                 < p_bestForAllRefAlignment->ComputeScore() ) 
      {
        //delete p_bestForAllRefAlignment->first;
        //delete p_bestForAllRefAlignment->second;
        delete p_bestForAllRefAlignment;
        p_bestForAllRefAlignment = p_curHypRefAlignment;
        
      } else {
        //delete p_curHypRefAlignment->first;
        //delete p_curHypRefAlignment->second;
        delete p_curHypRefAlignment;
      }
      
      //cout << "Beam Calls: " << m_NUM_BEAM_SEARCH_CALLS << endl;
      
    } // ref sent loop
    
    //cout << "BEST ALIGNMENT" << endl; p_bestForAllRefAlignment->Print();
    Scores[m_EDIT_COUNT_KEY] += p_bestForAllRefAlignment->GetNumEdits();
    Scores[m_REF_WORD_COUNT_KEY] += p_bestForAllRefAlignment->GetNumRefWords();
    Scores[m_HYP_WORD_COUNT_KEY] += p_bestForAllRefAlignment->GetNumHypWords();
    
    // delete p_bestForAllRefAlignment
    
  } // hyp sent loop
  
  // save final score
  Scores[m_SCORE_KEY] = Scores[m_EDIT_COUNT_KEY] / Scores[m_REF_WORD_COUNT_KEY];
  
}
*/

bool CEvaluatorTER_MER::Evaluate( int srcSentIndex, vector<CWordIndex>& hypSentence, vector< vector<CWordIndex> >& references, vector< float >& Scores)
{

  //CArray< CArray< CSentence* > >* p_refSentences = m_p_sharedData->GetReferences();
  Scores = vector<float>(11, 0); //so that each of the keys are valid
  Scores[m_INSERTION_KEY]      = 0;                       
  Scores[m_DELETION_KEY]       = 0;                       
  Scores[m_SUBSTITUTION_KEY]   = 0; 
  Scores[m_SHIFT_KEY]          = 0;
  Scores[m_MATCH_KEY]          = 0;
  Scores[m_HYP_WORD_COUNT_KEY] = 0;
  Scores[m_REF_WORD_COUNT_KEY] = 0; //num reference picked
  Scores[m_EDIT_COUNT_KEY]     = 0; //num edits
  Scores[m_HYP_SENT_COUNT_KEY] = 1;

  //CSentence* p_currRefSentence;
  vector<CWordIndex> currRefSentence;
  //int refSentCount = (int)((*p_refSentences)[srcSentIndex].Size());
  int refSentCount = references.size(); //since these are the references for one sentence
  
  CEvaluatorTER_MERalignment* p_bestForAllRefAlignment = 0;
  
  for(int r=0; r < refSentCount; r++)
  {
    //p_currRefSentence = (*p_refSentences)[srcSentIndex][r];
    currRefSentence = references[r];
    
    // create valid subphrases out of ref sentence
    PhraseToSet_hash_map* p_phraseStartHash 
      = BuildWordMatches(hypSentence, currRefSentence);
    
    // Begin min edit distance on given hyp and ref, w/o shifts
    
    Phrase hypPhrase      = BuildPhrase(hypSentence); //basically copies the hypSentence
    Phrase hypPrimePhrase = hypPhrase;
    Phrase refPhrase      = BuildPhrase(currRefSentence);
    CEvaluatorTER_MERcost costFunc;
    
    CEvaluatorTER_MERalignment* p_curHypRefAlignment
      = ComputeMinEditDistance(hypPhrase, refPhrase, costFunc);
    /*
     //-- debug
     p_curHypRefAlignment->Print();  
     //-- debug
     */
    p_curHypRefAlignment->SetHyp(hypPhrase);
    p_curHypRefAlignment->SetRef(refPhrase);
    p_curHypRefAlignment->SetHypPrime(hypPrimePhrase); //in UMD, it's "aftershift"
    
    float edits = 0;
    int numShifts = 0;
    
    
    // this doesn't seem to be necessary! - ad
    //std::vector< CEvaluatorTER_MERshift* >* 
      //p_allShiftsVector = new std::vector< CEvaluatorTER_MERshift* >;
    
    bool moreShifts = true;
    while (moreShifts) 
    {
      std::pair< CEvaluatorTER_MERshift*, CEvaluatorTER_MERalignment* >* 
      p_bestShiftAlignmentPair = FindBestShift(hypPrimePhrase, hypPhrase, 
                                               refPhrase, p_phraseStartHash,
                                               p_curHypRefAlignment, costFunc);
      if (p_bestShiftAlignmentPair == 0) {
        moreShifts = false;
      } else {
        CEvaluatorTER_MERshift* p_bestShift = p_bestShiftAlignmentPair->first;
        edits += p_bestShift->GetCost();
       // p_allShiftsVector->push_back(p_bestShift);
        //std::cerr << "NOW DELETING best Shift: " << p_bestShift << endl;
        delete p_bestShift; // valid if above is true
        
        //Note: p_curHypRefAlignment was already deleted by FindBestShift
        p_curHypRefAlignment = p_bestShiftAlignmentPair->second;
        hypPrimePhrase = p_curHypRefAlignment->GetHypPrime();
        delete p_bestShiftAlignmentPair;
      } 
    }
    
    //p_curHypRefAlignment->SetAllShifts(p_allShiftsVector);
    
    assert( p_curHypRefAlignment );
    
    p_curHypRefAlignment->SetNumEdits(edits + p_curHypRefAlignment->GetNumEdits());
    
    //p_curHypRefAlignment->Print();
    
    if (p_bestForAllRefAlignment == 0) {
      p_bestForAllRefAlignment = p_curHypRefAlignment;
    } else if( p_curHypRefAlignment->ComputeScore() 
               < p_bestForAllRefAlignment->ComputeScore() ) 
    {
      delete p_bestForAllRefAlignment;
      p_bestForAllRefAlignment = p_curHypRefAlignment;
      
    } else {
      delete p_curHypRefAlignment;
    }
    
    //cout << "Beam Calls: " << m_NUM_BEAM_SEARCH_CALLS << endl;
    p_phraseStartHash->clear(); // TODO: Is this necessary?
    delete p_phraseStartHash;
    
    
    // p_allShiftsVector is deleted in TERalignment destructor
    
  } // ref sent loop
  
  assert( p_bestForAllRefAlignment );
  
  //cout << "BEST ALIGNMENT" << endl; p_bestForAllRefAlignment->Print();
  Scores[m_EDIT_COUNT_KEY]     += p_bestForAllRefAlignment->GetNumEdits();
  Scores[m_REF_WORD_COUNT_KEY] += p_bestForAllRefAlignment->GetNumRefWords();
  Scores[m_HYP_WORD_COUNT_KEY] += p_bestForAllRefAlignment->GetNumHypWords();
 
  delete p_bestForAllRefAlignment;
  // save final score in position 0
  Scores[m_SCORE_KEY] = Scores[m_EDIT_COUNT_KEY] / Scores[m_REF_WORD_COUNT_KEY];
}

/*
bool
CEvaluatorTER_MER::Evaluate( const CArray< CNumericalArray< CWordIndex > >
                          *pTanslations, CNumericalArray< float > & Scores )
{
}
*/



//------------------------------------------------------------------
//      CEvaluatorTER_MER:: GetConfidenceInterval
//        
/*
std::pair<float, float>
CEvaluatorTER_MER::GetConfidenceInterval()
{
  
}
*/


//------------------------------------------------------------------
//      CEvaluatorTER_MER:: GetOracleScore
//
// calculate the oracle score for a nbest list; Format: n m word sequence
// oracle scores means: take those translations in nbest lists which optimize the scores
// no matter if read from file, or getting the translations in arrays
//


/*!
\note Does not delete p_NBestList

 */
/*
bool
CEvaluatorTER_MER::GetOracleScore( const char *pFileWithNBestTranslations,
                                CNumericalArray< float > & Scores )
{
    
  CArray< CArray< CSentence* > >* p_NBestList 
    = LoadNBestTranslations(pFileWithNBestTranslations);

  GetOracleScore(p_NBestList, Scores);
  
  
  return true;
  
}
*/

/*
bool
CEvaluatorTER_MER::GetOracleScore( CArray< CString > *pNBestTranslations,
                                CNumericalArray< float > & Scores )
{
}
*/


/*
bool
CEvaluatorTER_MER::GetOracleScore( CText *pNBestTranslations, 
                                CNumericalArray< float > & Scores )
{
}
*/

/*
bool
CEvaluatorTER_MER::GetOracleScore( CArray< CNumericalArray< CWordIndex > > *pNBestTranslations, CNumericalArray< float > & Scores )
{
}
*/

/*
bool
CEvaluatorTER_MER::GetOracleScore( CArray< CArray< CSentence* > >* p_NBestTranslations,
                               CNumericalArray< float >& oracleScores )
{
  assert( p_NBestTranslations );
  
  // Send nbest list to SharedData, which will reset itself if nbest list is new
  m_p_sharedData->SetNBestTranslations(p_NBestTranslations);

  // Prepare to create gain file, if necessary
  std::ofstream gainTableOutFile;    
  if (m_writeGainToFile)
  {
    gainTableOutFile.open( m_gainFileName );
    if( !gainTableOutFile ) {
      std::cerr << "Unable to open gain file \"" << m_gainFileName 
      << "\" for writing! Exiting..."  << endl;
      exit(1);
    }
  }
  
  
  int sourceSentCount = p_NBestTranslations->MaxIndex() + 1;
  int hypSentCount;

  // store baseline scores as a pair of the numerator (edit count) and 
  // denominator (ref word count of best sentence)
  std::vector< CNumericalArray< float > > baselineScoresVector( sourceSentCount );
  //std::vector< CNumericalArray< float > > baselineScoresVector;

  CSentence* p_hypSentence;
  CNumericalArray< float > baselineScores, bestHypScores;

  // init finalScores to which bestHypScores will be added for each source sentence
  oracleScores[m_SCORE_KEY]          = 0;
  oracleScores[m_EDIT_COUNT_KEY]     = 0;
  oracleScores[m_REF_WORD_COUNT_KEY] = 0;
  oracleScores[m_HYP_SENT_COUNT_KEY] = 0;
  oracleScores[m_HYP_WORD_COUNT_KEY] = 0;
  
  // 1. Store edit count and ref word count for every 1best hyp for each source sent
  //    Used to quickly compute TER with a different hyp for each source sent
  //    by subtracting 1-best hyp's counts and adding different hyp's counts 
  //    from TER totals and then dividing new total edit count by new total ref word count
  std::cerr << "computing baseline. src sent: ";
      
  for(int s = 0; s < sourceSentCount; s++)
  {
    std::cerr << s << " " << flush;
    p_hypSentence = (*p_NBestTranslations)[s][0];
    CNumericalArray< float > currHypScores;
    Evaluate(s, p_hypSentence, currHypScores);
    
    baselineScoresVector[s] = currHypScores;  // vector size = # of src sent
    
    // get cumulative TER score
    baselineScores[m_EDIT_COUNT_KEY]     += currHypScores[m_EDIT_COUNT_KEY];
    baselineScores[m_REF_WORD_COUNT_KEY] += currHypScores[m_REF_WORD_COUNT_KEY];
    baselineScores[m_HYP_SENT_COUNT_KEY] += currHypScores[m_HYP_SENT_COUNT_KEY];
    baselineScores[m_HYP_WORD_COUNT_KEY] += currHypScores[m_HYP_WORD_COUNT_KEY];

  }
  
  cout << endl;
  baselineScores[m_SCORE_KEY] 
    = baselineScores[m_EDIT_COUNT_KEY] / baselineScores[m_REF_WORD_COUNT_KEY];
  
  cout << "Starting with Baseline Scores: " << endl;
  ReportScores(baselineScores);
  
  // 2. Get gain for each hyp in nbest list
  float baselineTer = baselineScores[m_SCORE_KEY];

  CNumericalArray< float > baselineHypScores;
  
  for(int s = 0; s < sourceSentCount; s++) // TODO: shouldn't this start with 1?
  {
    float maxHypGain = 0;
    int   bestHypIndex;
  
    hypSentCount = (*p_NBestTranslations)[s].MaxIndex() + 1;
    
    cout << "s: " << s << "\th: " << hypSentCount << flush;
    
    // save baseline in gain file
    if( m_writeGainToFile ) {
      gainTableOutFile << s << " " << 0 << " " << 0 << endl;
    }
    
    
    baselineHypScores = baselineScoresVector[s];
  
    CNumericalArray< float > currHypScores;

    for(int h = 1; h < hypSentCount; h++)
    {
      p_hypSentence = (*p_NBestTranslations)[s][h];
      Evaluate(s, p_hypSentence, currHypScores);
      
      // subtract 1-best score (of hyp 0)
      float baselinePrimeTer 
        = (float)(baselineScores[m_EDIT_COUNT_KEY] - baselineHypScores[m_EDIT_COUNT_KEY] 
                  + currHypScores[m_EDIT_COUNT_KEY]) 
        / (float)(baselineScores[m_REF_WORD_COUNT_KEY] - baselineHypScores[m_REF_WORD_COUNT_KEY] 
                  + currHypScores[m_REF_WORD_COUNT_KEY]);
      
      float currGain =  baselineTer - baselinePrimeTer;  // lower score is better
      
      // save gain to a file here
      if( m_writeGainToFile ) {
        gainTableOutFile << s << " " << h << " " << currGain << endl;
      }
      
      
      if( currGain > maxHypGain )
      {
        maxHypGain = currGain;
        bestHypScores = currHypScores;
        bestHypIndex = h;
      }
      
    } // end hyp loop
    
    if( maxHypGain == 0 ) {
      // baseline is best
      bestHypScores = baselineHypScores;  // TODO: is this necessary?
      bestHypIndex = 0;                   // TODO: initialize w/this value
    }
    
    // add best hyp scores to finalBestScore
    oracleScores[m_EDIT_COUNT_KEY]     += bestHypScores[m_EDIT_COUNT_KEY];
    oracleScores[m_REF_WORD_COUNT_KEY] += bestHypScores[m_REF_WORD_COUNT_KEY];
    oracleScores[m_HYP_SENT_COUNT_KEY] += bestHypScores[m_HYP_SENT_COUNT_KEY];
    oracleScores[m_HYP_WORD_COUNT_KEY] += bestHypScores[m_HYP_WORD_COUNT_KEY];
    cout << "\tpicked h: " << bestHypIndex << "\tgain: " << maxHypGain << endl << flush;
    
  } // end source sent loop
  
  oracleScores[m_SCORE_KEY] = oracleScores[m_EDIT_COUNT_KEY] / oracleScores[m_REF_WORD_COUNT_KEY];
  
  return true;
}
*/
/*
void
CEvaluatorTER_MER::ReportScores( CNumericalArray< float > & Scores )
{
  cout << "TER: " << Scores[m_SCORE_KEY] 
  //<< " INS=" << Scores[m_INSERTION_KEY] 
  //<< " DEL=" << Scores[m_DELETION_KEY] 
  //<< " SUB=" << Scores[m_SUBSTITUTION_KEY] 
  //<< " MATCH=" << Scores[m_MATCH_KEY] 
  << " HypSent=" << Scores[m_HYP_SENT_COUNT_KEY] 
  << " RefWords=" << Scores[m_REF_WORD_COUNT_KEY] 
  << " HypWords=" << Scores[m_HYP_WORD_COUNT_KEY] 
  << " ShiftsConsidered=" << m_NUM_SHIFTS_CONSIDERED
  << endl;
}

*/


//------------------------------------------------------------------
//      CEvaluatorTER_MER:: BuildWordMatches
//

PhraseToSet_hash_map*
CEvaluatorTER_MER::BuildWordMatches(vector<CWordIndex>& hyp, vector<CWordIndex>& ref)
{
  
  // 1. make a set of hyp words for fast word lookup

  //int hypLength = p_hyp->Length();
  //int refLength = p_ref->Length();

  int hypLength = hyp.size();
  int refLength = ref.size();
  
  IntSet hypWordSet;
    
  for(int h=0; h < hypLength; h++) {
    //hypWordSet.insert( (int)p_hyp->GetAt(h+1) );  // CSentence index starts at 1
    hypWordSet.insert( hyp[h] );  // CSentence index starts at 1
  }
  
  // 2. mark which ref words are in hyp by labeling ref word positions with 1
  
  bool refWordInHyp[refLength];
  
  for(int r=0; r < refLength; r++) 
  {
    //if( hypWordSet.count( (int)p_ref->GetAt(r+1) ) != 0 ) {
    if( hypWordSet.count( ref[r] ) != 0 ) {
      refWordInHyp[r] = true;
    } else {
      refWordInHyp[r] = false;
    }
  }
  
  // 3. loop through every ref word
  //         if starting ref word is in hyp and subsequent ref words are in hyp, make phrase
  //         - make phrase by adding <subphrase,set of starting positions>

    
  //{Phrase}->{Set(starting positions)}
  PhraseToSet_hash_map* p_phraseStartHash = new PhraseToSet_hash_map(); 
  
  for(int start = 0; start < refLength; start++)
  {
    
    if(refWordInHyp[start])
    {
      for(int end = start; ( end < refLength 
                             && (end - start <= m_MAX_SHIFT_SIZE)
                             && refWordInHyp[end] ); end++)
        
      {
        Phrase subPhrase = BuildPhrase(ref, start, end+1); // endIndex exclusive

        // does subphrase exist already?
        PhraseToSet_hash_map::iterator i_phraseStartHash 
          = p_phraseStartHash->find(subPhrase);
        
        //cout << "Looking for Phrase: "; PrintPhrase(subPhrase);
        if( i_phraseStartHash != p_phraseStartHash->end() )
        {
          // if found, retrieve set, and add starting point to it
          IntSet startingPosSet = i_phraseStartHash->second;
          startingPosSet.insert(start);
          //cout << " Starting: " << start << endl;
        } else {
          // if not found, make set, add starting point to it.          
          IntSet startingPosSet;
          startingPosSet.insert(start);
          //cout << " Starting " << start << endl;
          p_phraseStartHash->insert( std::pair<Phrase, IntSet>
                                (subPhrase, startingPosSet) );
        }
      } // end ref subphrase loop
    }
  } // end of ref sent word loop
  
  return p_phraseStartHash;
  
}



//------------------------------------------------------------------
//      CEvaluatorTER_MER:: BuildPhrase
//
//        returns Phrase 
//        beginIndex inclusive, endIndex exclusive
//

Phrase
CEvaluatorTER_MER::BuildPhrase(vector<CWordIndex>& sentence, int beginIndex, int endIndex)
{
  Phrase memoryrase;
  
  for (int i=beginIndex; i < endIndex; i++) {
    //memoryrase.push_back( p_sentence->GetAt(i+1) );
    memoryrase.push_back( sentence[i] );
  }
  
  return memoryrase;
}


/*
Phrase CEvaluatorTER_MER::BuildPhrase(Phrase& phrase, int beginIndex, int endIndex)
{
  Phrase subPhrase;
  
  for (int i=beginIndex; i < endIndex; i++) {
    subPhrase.push_back( phrase[i] );
  }
  
  return subPhrase;
}
*/

Phrase
CEvaluatorTER_MER::BuildPhrase(vector<CWordIndex>& sentence)
{
  Phrase sentencePhrase;
  
  for (int i=0; i < sentence.size(); i++) {
    //sentencePhrase.push_back(p_sentence->GetAt(i+1));
    sentencePhrase.push_back(sentence[i]);
  }
  
  return sentencePhrase;
}


//------------------------------------------------------------------
//      CEvaluatorTER_MER:: PerformShift
//

Phrase
CEvaluatorTER_MER::PerformShift(Phrase& origPhrase, CEvaluatorTER_MERshift* p_shift)
{
  return PerformShift(origPhrase, p_shift->GetStart(), p_shift->GetEnd(), 
                      p_shift->GetNewloc());
}

Phrase
CEvaluatorTER_MER::PerformShift(Phrase& origPhrase, int beginIndex, int endIndex,
                            int toIndex)
{
  int origPhraseLength = origPhrase.size();

  // TODO: change all incoherent UMD var names
  int pos = 0;
  Phrase memoryrase = origPhrase;
  if (toIndex == -1) {
    for (int i = beginIndex; i <= endIndex; i++) 
      { memoryrase[pos++] = origPhrase[i]; }
    for (int i = 0; i <= beginIndex-1; i++) 
      { memoryrase[pos++] = origPhrase[i]; }
    for (int i = endIndex+1; i < origPhraseLength;i++) 
      { memoryrase[pos++] = origPhrase[i]; }
  } else if (toIndex < beginIndex) {
    for (int i = 0; i <= toIndex; i++) 
      { memoryrase[pos++] = origPhrase[i]; }
    for (int i = beginIndex; i <= endIndex;i++) 
      { memoryrase[pos++] = origPhrase[i]; }
    for (int i = toIndex+1; i <= beginIndex-1;i++) 
      { memoryrase[pos++] = origPhrase[i]; }
    for (int i = endIndex+1; i < origPhraseLength;i++) 
      { memoryrase[pos++] = origPhrase[i]; }
  } else if (toIndex > endIndex) {
    for (int i = 0; i <= beginIndex-1; i++) 
      { memoryrase[pos++] = origPhrase[i]; }
    for (int i = endIndex+1; i <= toIndex;i++) 
      { memoryrase[pos++] = origPhrase[i]; }	    
    for (int i = beginIndex; i <= endIndex;i++) 
      { memoryrase[pos++] = origPhrase[i]; }
    for (int i = toIndex+1; i < origPhraseLength;i++) 
      { memoryrase[pos++] = origPhrase[i]; }	    
  } else {
    // we are moving inside of ourselves    // UMD comment
    for (int i = 0; i <= beginIndex-1; i++) 
      { memoryrase[pos++] = origPhrase[i]; }
    for (int i = endIndex+1; i <= (endIndex + (toIndex - beginIndex)); i++) 
      { memoryrase[pos++] = origPhrase[i]; }
    for (int i = beginIndex; i <= endIndex;i++) 
      { memoryrase[pos++] = origPhrase[i]; }
    for (int i = (endIndex + (toIndex - beginIndex)+1); i < origPhraseLength;i++) 
      { memoryrase[pos++] = origPhrase[i]; }
  }
  
  m_NUM_SHIFTS_CONSIDERED++;
  
  return memoryrase;
  
}


//------------------------------------------------------------------
//      CEvaluatorTER_MER:: PrintPhrase
//

void
CEvaluatorTER_MER::PrintPhrase(Phrase &p) const
{
  for(int i=0; i < p.size(); i++) {
    cout << p[i] << " ";
  }
  cout << endl;
}

//------------------------------------------------------------------
//      CEvaluatorTER_MER:: BuildValidShifts
//

std::vector< std::vector< CEvaluatorTER_MERshift* > >*
CEvaluatorTER_MER::BuildValidShifts(Phrase& hypPrime, Phrase& ref, 
                                PhraseToSet_hash_map* p_phraseStartHash,
                                vector< bool >& herr, vector< bool >& rerr,
                                vector< int >& ralign,
                                CEvaluatorTER_MERcost& costFunc)
{
  
  //-- Valid shifts, indexed by length, where [1][] is arr of shifts length 1
  std::vector< std::vector< CEvaluatorTER_MERshift* > >* p_validShiftsByLength 
    = new std::vector< std::vector< CEvaluatorTER_MERshift* > >( m_MAX_SHIFT_SIZE );
  
  //-- Store shifts as vector[length of shift][shifts]
  /*
  for (int i=m_MAX_SHIFT_SIZE; i >= 0; i--){
    std::vector< CEvaluatorTER_MERshift* > emptyVector;
    p_validShiftsByLength->push_back(emptyVector);
  }
*/
  
  //-- Build all shifts
  /*
  //-- debug
  cout << "getting ready to build all shifts" << endl;
  //-- end debug
  */
  for (int start = 0; start < hypPrime.size(); start++)
  {
    Phrase subPhrase = BuildPhrase(hypPrime, start, start+1); // endIndex exclusive
    
    //cout << "working on subphrase: "; PrintPhrase(subPhrase);
    
    PhraseToSet_hash_map::iterator i_phraseStartHash 
      = p_phraseStartHash->find(subPhrase);

    
    if (i_phraseStartHash != p_phraseStartHash->end())
    {
      bool validStart = false;

      IntSet startingPosSet = i_phraseStartHash->second;
      
      //cout << "# starting points: " << startingPosSet.size() << endl;
      
      // pick valid starting point that's not the same as in ref
      for (IntSet::iterator i_startingPosSet = startingPosSet.begin();
           i_startingPosSet != startingPosSet.end() && ! validStart;
           i_startingPosSet++)
      {
        int moveTo = *i_startingPosSet;
        
        /*
        //-- debug
        cout << "First moveto check: " << moveTo << endl;
        cout << "RALIGN: ralign["<<moveTo<<"]="<<ralign[moveTo]<<endl;
        //CONDITION FAILS (POSSIBLY BECAUSE OF PATH)
        cout << "if (( " << start << "!=" << ralign[moveTo] << ") && (" << ralign[moveTo] - start <<" <= " << m_MAX_SHIFT_DIST << ") && ((" << start- ralign[moveTo]-1 << ") <=" << m_MAX_SHIFT_DIST << "))" << endl;
        //-- end debug
        */
        
        if ( (start != ralign[moveTo]) 
             && (ralign[moveTo] - start <= m_MAX_SHIFT_DIST) 
             && (start - ralign[moveTo] -1 <= m_MAX_SHIFT_DIST) )
        {
              //cout << "going to break loop 1" << endl;
              validStart = true; // breaks loop 
        } else {
          //cout << "KEY CONDITION FAILS" << endl;
        }
        
      } // end starting pt loop
      
      if (validStart)
      {
        // build phrase
        for (int end = start; (validStart && (end < hypPrime.size()) 
                               && (end < start + m_MAX_SHIFT_SIZE)); end++)
        {
          Phrase candidatePhrase = BuildPhrase(hypPrime, start, end+1);
          /*
          //--debug
          cout << "now candidate phrase (" << start << "," << end+1 <<"): "; 
          PrintPhrase(candidatePhrase);
          //--end debug
           */
          validStart = false;
          
          i_phraseStartHash = p_phraseStartHash->find(candidatePhrase);
          
          if (i_phraseStartHash != p_phraseStartHash->end())
          {
            bool anyHypErrors = false;
            for(int i=0; (i <= end-start) && (! anyHypErrors); i++){
              if (herr[start+i])
                anyHypErrors = true;
            }
            
            if(anyHypErrors == false) {
              validStart = true;
            } else {
              
              startingPosSet = i_phraseStartHash->second;
              
              for(IntSet::iterator i_startingPosSet = startingPosSet.begin();
                  i_startingPosSet != startingPosSet.end(); i_startingPosSet++)
              {
                int moveto = *i_startingPosSet;
                //cout << "second starting pt lookup: " << moveto << endl;
  
                if( ((ralign[moveto] != start) 
                     && ((ralign[moveto] < start) || (ralign[moveto] > end)) 
                     && (ralign[moveto] - start <= m_MAX_SHIFT_DIST) 
                     && ((start - ralign[moveto]) - 1 <= m_MAX_SHIFT_DIST)))
                {
                  validStart = true;
                  
                  bool anyRefErrors = false;
                  
                  for (int i=0; (i <= end - start) && (! anyRefErrors); i++){
                    if (rerr[moveto+i])
                      anyRefErrors = true;
                  }
                  
                  
                  if(anyRefErrors)
                  {
                    for (int roff = 0; roff <= (end - start); roff++)
                    {
                      if ((start != ralign[moveto+roff]) &&
                          ((roff == 0) || 
                           (ralign[moveto+roff] != ralign[moveto]))) 
                      {
                        // new TERshift
                        CEvaluatorTER_MERshift* p_validShift 
                          = new CEvaluatorTER_MERshift(start, end, moveto+roff,
                                                   ralign[moveto+roff]);
                        p_validShift->SetShiftedPhrase(candidatePhrase);
                        p_validShift->SetCost(costFunc.Shift());
                        // CArray< vector< CEvaluatorTER_MERshift > > 
                        (*p_validShiftsByLength)[end-start].push_back(p_validShift);
                      }
                    }
                  }
                  
                  
                }
                
              } // starting points loop
            }
          }
          
        } // end phrase-building loop
      }
    } else {
      //cout << "subphrase not found:"; 
      //PrintPhrase(subPhrase);
    }
    
  } // end hyp word loop
  
  return p_validShiftsByLength;
  
}


//------------------------------------------------------------------
//      CEvaluatorTER_MER:: ComputeMinEditDistance
//
//      returns CEvaluatorTER_MERalignment
//

CEvaluatorTER_MERalignment*
CEvaluatorTER_MER::ComputeMinEditDistance(Phrase& hyp, Phrase& ref,
                                      CEvaluatorTER_MERcost& costFunc)
{
  
  size_t hypWordCount = hyp.size();
  size_t refWordCount = ref.size();
  
  float minEditDistance[hypWordCount+1][refWordCount+1];
  
  //ARV: initialize the backtrack so that we have normal access to all the elements!
  // backtrack to create path
  //CArray< CArray< int > > backtrack;
  vector< vector< int > > backtrack;
  for (int i=0; i<hypWordCount+1; i++){
    vector<int> inside(refWordCount+1, 0);
    backtrack.push_back(inside);
  }

  
  for (int i = 0; i < hypWordCount+1; i++)
    minEditDistance[i][0] = i;
  
  for (int j = 0; j < refWordCount+1; j++)
    minEditDistance[0][j] = j;
  
  // fill rest of matrix with 0s
  for (int i = 1; i < hypWordCount+1; i++)
    for (int j = 1; j < refWordCount+1; j++)
      minEditDistance[i][j] = 0;
  
    
  // start computing min edit distance
  
  CWordIndex refWord;
  CWordIndex hypWord;
    
  for (int i = 1; i <= hypWordCount; i++)
  {
    hypWord = hyp[i-1];
    
    for (int j = 1; j <= refWordCount; j++)
    {
      refWord = ref[j-1];

      if (hypWord == refWord)
      {
        minEditDistance[i][j] = minEditDistance[i-1][j-1];
        backtrack[i][j] = m_MATCH_KEY;
      } else {
        minEditDistance[i][j] = minEditDistance[i-1][j-1] + costFunc.Substitution();
        backtrack[i][j] = m_SUBSTITUTION_KEY;
      }
      
      if (minEditDistance[i][j] > minEditDistance[i-1][j] + costFunc.Insertion()) {
        minEditDistance[i][j] = minEditDistance[i-1][j] + costFunc.Insertion();
        backtrack[i][j] = m_INSERTION_KEY;
      }
      
      if (minEditDistance[i][j] > minEditDistance[i][j-1] + costFunc.Deletion()) {
        minEditDistance[i][j] = minEditDistance[i][j-1] + costFunc.Deletion();
        backtrack[i][j] = m_DELETION_KEY;
      }
      
    } // end hyp word loop
  }// end ref word loop

  //cout << "IN THE MIDDLE, MED: " << minEditDistance[hypWordCount][refWordCount] << endl;
  
  // retrace edit path
  std::deque< unsigned short > medPath;
  
  size_t b_y = hypWordCount;
  size_t b_x = refWordCount;
  
  size_t step;
  
  while (b_x > 0 || b_y > 0)
  {
    step = backtrack[b_y][b_x];
    
    // check if hit edge of matrix
    if (b_x == 0) {
      backtrack[b_y][1] = -1;
      step = m_INSERTION_KEY;
    } else if (b_y == 0) {
      backtrack[1][b_x] = -1;
      step = m_DELETION_KEY;
    } else {
      backtrack[b_y][b_x] = -1;
    }
  
    medPath.push_front(step);
    //Scores[step]++;
    
    if (step == m_MATCH_KEY){
      // move backward diagonally
      b_x--;
      b_y--;
    }
    if (step == m_INSERTION_KEY) {
      // move up
      b_y--;
    }
    if (step == m_SUBSTITUTION_KEY) {
      // move backward diagonally
      b_x--;
      b_y--;
    }
    if (step == m_DELETION_KEY) {
      // move to the left
      b_x--;
    }
  }
  
  CEvaluatorTER_MERalignment* p_medAlignment 
    = new CEvaluatorTER_MERalignment(hyp, ref, medPath, 
                                 minEditDistance[hypWordCount][refWordCount],
                                 refWordCount);
  
  m_NUM_BEAM_SEARCH_CALLS++;
  
  return p_medAlignment;
}

//------------------------------------------------------------------
//      CEvaluatorTER_MER:: FindBestShift
//

std::pair< CEvaluatorTER_MERshift*, CEvaluatorTER_MERalignment* >*
CEvaluatorTER_MER::FindBestShift(Phrase& hypPrime, Phrase& hyp, Phrase& ref,
                             PhraseToSet_hash_map* p_phraseStartHash,
                             CEvaluatorTER_MERalignment* p_hypRefAlignment,
                             CEvaluatorTER_MERcost& costFunc)
{
  //m_IN_FIND_BEST_SHIFT++;
  
  //cout << "Entered FindBestShift()" << endl;
  //cout.flush();
  
  std::pair< CEvaluatorTER_MERshift*, CEvaluatorTER_MERalignment* >* p_bestShiftAlignmentPair = 0;
  
  // mark which hyp and ref words are wrong (from UMD code)
  //CArray< bool > herr;                 // bad UMD naming
  //CArray< bool > rerr;
  //CArray< int > ralign;              // bad UMD naming

  //ARV: code below assumes 0 indexing even though it was based on CArray
  vector< bool > herr;                 // bad UMD naming
  vector< bool > rerr;
  vector< int > ralign;              // bad UMD naming


  FindAlignError(p_hypRefAlignment, herr, rerr, ralign);
    
  std::vector< std::vector< CEvaluatorTER_MERshift* > >* p_validShiftsByLength 
    = BuildValidShifts(hypPrime, ref, p_phraseStartHash, 
                       herr, rerr, ralign, costFunc);
  
  
  
  //cout << " - BuildValidShifts done" <<endl;
  //cout.flush();
  
  CEvaluatorTER_MERalignment* p_bestHypRefAlignment = p_hypRefAlignment;
  CEvaluatorTER_MERshift* p_bestShift = 0;
  
  float curErr = p_hypRefAlignment->GetNumEdits(); // bad UMD naming
  float curBestShiftCost = 0.0;
  
  bool anyGain = false;
  
  for (int length=p_validShiftsByLength->size()-1; length >= 0; length--) 
    // but there's no 0-length shift??
  {
    
    //UMD sez: consider shifts of length i+1
    
    float curfix = curErr - (curBestShiftCost + p_bestHypRefAlignment->GetNumEdits());
    
    float maxfix = (2 * (1 + length)) - costFunc.Shift();
    
    if ((curfix > maxfix) || ((curBestShiftCost != 0) && (curfix == maxfix))) {
      break;
    }
    
    for (int s = (*p_validShiftsByLength)[length].size()-1; s >=0;  s--)
    {
      CEvaluatorTER_MERshift* p_curShift = (*p_validShiftsByLength)[length][s];
      /*
      //-- debug
      cout << "Valid Shift - length: " << length << " ";
      p_curShift->Print();
      //-- end debug
      */
      curfix = curErr - (curBestShiftCost + p_bestHypRefAlignment->GetNumEdits());
      
      if ((curfix > maxfix) || ((curBestShiftCost != 0) && (curfix == maxfix))) {
        break;
      }
      

      Phrase hypPrimePrime = PerformShift(hypPrime, p_curShift);
      CEvaluatorTER_MERalignment* p_curAlignment 
        = ComputeMinEditDistance(hypPrimePrime, ref, costFunc);
      p_curAlignment->SetHyp(hyp);
      p_curAlignment->SetRef(ref);
      p_curAlignment->SetHypPrime(hypPrimePrime);
      
      float gain = ((float)p_bestHypRefAlignment->GetNumEdits() + curBestShiftCost) - ((float)p_curAlignment->GetNumEdits() + p_curShift->GetCost());
      
      if ((gain > 0) || ((curBestShiftCost == 0) && (gain == 0))) 
      {
        //delete p_bestShift;
        delete p_bestHypRefAlignment;
          
        anyGain = true;
        p_bestShift = p_curShift;
        curBestShiftCost = p_curShift->GetCost();
        p_bestHypRefAlignment = p_curAlignment;
      } else {
        //delete p_curShift;
        delete p_curAlignment;
      }
      
    } // end of shifts vector loop
    
    //(*p_validShiftsByLength)[length].clear();
    
  } // end of length vector loop
  
  
  if(anyGain) {
    p_bestShiftAlignmentPair = new std::pair< CEvaluatorTER_MERshift*, CEvaluatorTER_MERalignment* >;
    p_bestShiftAlignmentPair->first = p_bestShift;
    p_bestShiftAlignmentPair->second = p_bestHypRefAlignment;
  }
  
  // clean up shifts container without deleting best shift
  
  std::vector< std::vector< CEvaluatorTER_MERshift* > >::iterator i_lengthVectors;
  std::vector< CEvaluatorTER_MERshift* >::iterator i_shifts;

  for(i_lengthVectors = p_validShiftsByLength->begin();
      i_lengthVectors != p_validShiftsByLength->end();
      i_lengthVectors++) 
  {
    for(i_shifts = (*i_lengthVectors).begin();
        i_shifts != (*i_lengthVectors).end();
        i_shifts++) 
    {
      if( p_bestShiftAlignmentPair == 0 || 
          *i_shifts != p_bestShiftAlignmentPair->first) {
        delete *i_shifts; // value is a pointer to CEvaluatorTER_MERshift
      }
    }
    (*i_lengthVectors).clear();
  }
  (*p_validShiftsByLength).clear();
  delete p_validShiftsByLength;
  
  return p_bestShiftAlignmentPair;
}



//------------------------------------------------------------------
//      CEvaluatorTER_MER:: FindAlignError
//
void
CEvaluatorTER_MER::FindAlignError(CEvaluatorTER_MERalignment* p_alignment,
                              vector< bool >& herr, vector< bool >& rerr,
                              vector< int >& ralign )
{
  
  std::deque< unsigned short > path = p_alignment->GetMedPath();
  
  int hpos = -1;
  int rpos = -1;
  
  
  // TODO: use deque iterator
  
  for (int i=0; i < path.size(); i++)
  {
    size_t step = path[i];

    if(step == m_MATCH_KEY) {
      hpos++;
      rpos++;
      //herr[hpos] = false;
      //rerr[rpos] = false;
      herr.push_back(false);
      rerr.push_back(false);

      //ralign[rpos] = hpos;
      ralign.push_back(hpos);
    } else if (step == m_SUBSTITUTION_KEY) {
      hpos++; 
      rpos++;

      //herr[hpos] = true; 
      //rerr[rpos] = true;
      herr.push_back(true);
      rerr.push_back(true);

      //ralign[rpos] = hpos;
      ralign.push_back(hpos);

    } else if (step == m_INSERTION_KEY) {
      hpos++;
      //herr[hpos] = true; 
      herr.push_back(true);
    } else if (step == m_DELETION_KEY) {
      rpos++; 		
      //rerr[rpos] = true;
      rerr.push_back(true);
      //ralign[rpos] = hpos;
      ralign.push_back(hpos);

    }
  }
  
}



//====================================================================
//      CEvaluatorTER_MERshift:: 
//

CEvaluatorTER_MERshift::CEvaluatorTER_MERshift()
{
  m_start  = 0;
  m_end    = 0;
  m_moveto = 0;
  m_newloc = 0;
  
  m_cost   = 1.0;
}

CEvaluatorTER_MERshift::CEvaluatorTER_MERshift(int start, int end, 
                                       int moveto, int newloc)
{
  m_start  = start;
  m_end    = end;
  m_moveto = moveto;                   
  m_newloc = newloc;
  
  m_cost   = 1.0;
}

CEvaluatorTER_MERshift::~CEvaluatorTER_MERshift()
{
  
}


//------------------------------------------------------------------
//      CEvaluatorTER_MERshift:: Distance
//        

int
CEvaluatorTER_MERshift::Distance() const
{
  if(m_moveto < m_start){
    return m_start - m_moveto;
  } else if (m_moveto > m_end) {
    return m_moveto - m_end;
  } else {
    return m_moveto - m_start;
  }
}


//------------------------------------------------------------------
//      CEvaluatorTER_MERshift:: LeftShift
//        

bool
CEvaluatorTER_MERshift::LeftShift() const
{
  return (m_moveto < m_start);
}


//------------------------------------------------------------------
//      CEvaluatorTER_MERshift:: Size
//        

int
CEvaluatorTER_MERshift::Size() const
{
  return (m_end - m_start) + 1;
}

//------------------------------------------------------------------
//      CEvaluatorTER_MERshift:: Print
//        


void
CEvaluatorTER_MERshift::Print() const
{
  cout << "Start: " << m_start << "\nEnd: " << m_end
  << " Shifted Phrase: ";
  for(int i=0; i < m_shiftedPhrase.size(); i++) {
    cout << " " << m_shiftedPhrase[i];
  }
}

//------------------------------------------------------------------
//      CEvaluatorTER_MERshift:: get/set members
//        

size_t CEvaluatorTER_MERshift::GetStart() const         { return m_start; }
size_t CEvaluatorTER_MERshift::GetEnd() const           { return m_end; }
size_t CEvaluatorTER_MERshift::GetMoveto() const        { return m_moveto; }
size_t CEvaluatorTER_MERshift::GetNewloc() const        { return m_newloc; }
Phrase CEvaluatorTER_MERshift::GetShiftedPhrase() const { return m_shiftedPhrase; }
float CEvaluatorTER_MERshift::GetCost() const          { return m_cost; }

void CEvaluatorTER_MERshift::SetStart(size_t start) 
  { m_start = start; }
void CEvaluatorTER_MERshift::SetEnd(size_t end) 
  { m_end = end; }
void CEvaluatorTER_MERshift::SetMoveto(size_t moveto) 
  { m_moveto = moveto; }
void CEvaluatorTER_MERshift::SetNewloc(size_t newloc) 
  { m_newloc = newloc; }
void CEvaluatorTER_MERshift::SetShiftedPhrase(Phrase& shiftedPhrase) 
  { m_shiftedPhrase = shiftedPhrase; }
void CEvaluatorTER_MERshift::SetCost(float cost) 
  { m_cost = cost; }


//====================================================================
//      CEvaluatorTER_MERalignment:: 
//

CEvaluatorTER_MERalignment::CEvaluatorTER_MERalignment()
{
  m_numEdits    = 0;
  m_numRefWords = 0;
  
  m_p_allShifts = NULL;                   
  
}

CEvaluatorTER_MERalignment::CEvaluatorTER_MERalignment(Phrase &hyp, Phrase &ref,
                                               std::deque< unsigned short > medPath,
                                               float numEdits,
                                               size_t numRefWords)
{
  m_hyp = hyp;
  m_ref = ref;
  
  m_numEdits    = numEdits;
  m_numRefWords = numRefWords;
  
  m_medPath     = medPath;                 // deque< unsigned short >
  
  m_p_allShifts = 0;
}

CEvaluatorTER_MERalignment::~CEvaluatorTER_MERalignment()
{
  if( m_p_allShifts != NULL)
  {
    for(int i = m_p_allShifts->size()-1; i>=0; i--) {
      delete (*m_p_allShifts)[i];
    }
    m_p_allShifts->clear();
    delete m_p_allShifts;
  }
}

//------------------------------------------------------------------
//      CEvaluatorTER_MERalignment:: Score
//

float
CEvaluatorTER_MERalignment::ComputeScore() const
{
  if(m_numRefWords <= 0) {
    return 0.0;
  }
  
  return (float)m_numEdits / (float)m_numRefWords;
}


//------------------------------------------------------------------
//      CEvaluatorTER_MERalignment:: Print
//


void
CEvaluatorTER_MERalignment::Print() const
{
  
  std::cerr << "START OF TER ALIGNMENT ----------------------------------"<<endl;
  
  
  std::cerr << "Ref:";
  for (int i=0; i < m_ref.size(); i++) {
    std::cerr << " " << m_ref[i];
  }
  return;
  cout << "\nHyp:";
  for (int i=0; i < m_hyp.size(); i++) {
    cout << " " << m_hyp[i];
  }
  cout << "\nHypPrime:";
  for (int i=0; i < m_hypPrime.size(); i++) {
    cout << " " << m_hypPrime[i];
  }

  cout << "\nAlignment (MedPath):";
  for (int i=0; i < m_medPath.size(); i++) {
    cout << " " << m_medPath[i];
  }
  
  // TODO: num shifts
  
  cout << "\nScore: " << ComputeScore() 
    << "(" << m_numEdits << "/" << m_numRefWords << ")" << endl;
  
}


//------------------------------------------------------------------
//      CEvaluatorTER_MERalignment:: get/set members
//        

Phrase CEvaluatorTER_MERalignment::GetRef() const                   
  { return m_ref; }
Phrase CEvaluatorTER_MERalignment::GetHyp() const                   
  { return m_hyp; }
float CEvaluatorTER_MERalignment::GetNumEdits() const              
  { return m_numEdits; }
size_t CEvaluatorTER_MERalignment::GetNumRefWords() const
  { return m_numRefWords; }
size_t CEvaluatorTER_MERalignment::GetNumHypWords() const
  { return m_hyp.size(); }
Phrase CEvaluatorTER_MERalignment::GetHypPrime() const              
  { return m_hypPrime; }
std::deque< unsigned short > CEvaluatorTER_MERalignment::GetMedPath() const
  { return m_medPath; }
std::vector< CEvaluatorTER_MERshift* >* CEvaluatorTER_MERalignment::GetAllShifts() const
  { return m_p_allShifts; }

void CEvaluatorTER_MERalignment::SetRef(Phrase ref) 
  { m_ref = ref; }
void CEvaluatorTER_MERalignment::SetHyp(Phrase hyp) 
  { m_hyp = hyp; }
void CEvaluatorTER_MERalignment::SetHypPrime(Phrase afterShift) 
  { m_hypPrime = afterShift; }
void CEvaluatorTER_MERalignment::SetAllShifts(std::vector< CEvaluatorTER_MERshift* >* shifts) 
  { m_p_allShifts = shifts; }
void CEvaluatorTER_MERalignment::SetMedPath(std::deque< unsigned short > medPath)
  { m_medPath = medPath; }
void CEvaluatorTER_MERalignment::SetNumEdits(float edits)
  { m_numEdits = edits; }


//====================================================================
//        CEvaluatorTER_MERcost::
//

CEvaluatorTER_MERcost::CEvaluatorTER_MERcost()
{
  m_matchCost        = 1.0;
  m_substitutionCost = 1.0;
  m_insertionCost    = 1.0;
  m_deletionCost     = 1.0;
  m_shiftCost        = 1.0;
}

//------------------------------------------------------------------
//      CEvaluatorTER_MERcost:: get/set members
//        

float
CEvaluatorTER_MERcost::Match() const        { return m_matchCost; }
float
CEvaluatorTER_MERcost::Substitution() const { return m_substitutionCost; }
float
CEvaluatorTER_MERcost::Insertion() const   { return m_insertionCost; }
float
CEvaluatorTER_MERcost::Deletion() const    { return m_deletionCost; }
float
CEvaluatorTER_MERcost::Shift() const       { return m_shiftCost; }

void CEvaluatorTER_MERcost::SetMatchCost(float cost) 
  { m_matchCost = cost; }
void CEvaluatorTER_MERcost::SetSubstitutionsCost(float cost) 
  { m_substitutionCost = cost; }
void CEvaluatorTER_MERcost::SetInsertionCost(float cost) 
  { m_insertionCost = cost; }
void CEvaluatorTER_MERcost::SetDeletionCost(float cost) 
  { m_deletionCost = cost; }
void CEvaluatorTER_MERcost::SetShiftCost(float cost)
  { m_shiftCost = cost; }



