#ifndef __OPTIMIZE_HH__
#define __OPTIMIZE_HH__

#include <MetricSearch.hh>
#include <MetricRaw.hh>
#include <vector>
#include <paramdef.hh>
#include <param.hh>


//!Parameters that multiply the respective feature
typedef std::vector<double> Lambdas;

//!Feature data for each candidate, as float since they need to be stored in memory
typedef std::vector<float> Features;

//!The optimization metric is initialized against a particular metric
class COptimize{
public:

  //default, no metric
  COptimize();

  //!Initialize an optimizer against a metric
  COptimize(MetricSearch* pMetric){
    m_pMetric = pMetric;
  }

  MetricSearch * Metric() { return m_pMetric; }

  /*!Each optimizer is initialized with different top level params
   * We require just that each optimizer knows what to do for its own params
   */
  virtual void Init(CParam* param)=0;
  //! Find parameters that use argmin to maximize metric score (this version works with doubles for the starting params)
  virtual Lambdas minModelCostMaxMetricScore(const std::vector<Features*>& costs, const std::vector<RawCounts*>&  scores, const std::vector<int>& startEnd, const Lambdas& initialValues, float &FinalScore, int firstXsentences)=0;

protected: 

  //!The error metric that determines the metric score
  MetricSearch* m_pMetric;
};


#endif
