#include <stdlib.h>
#include <stdio.h>
// _AccessCharacterDB.cpp: implementation of the C_AccessCharacterDB class.
//
//////////////////////////////////////////////////////////////////////

#include "_AccessCharacterDB.h"
#include "time.h"
#include "ngramDB_shared.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

C_AccessCharacterDB::C_AccessCharacterDB(const char * dbFileName)
{

	//Initialize DBs for n-gram frequencies
	unsigned int openFlag = DB_RDONLY;	//read-only when accessing the DB

	this->ngramDb = new Db(NULL, 0);

	try{
		ngramDb->open(NULL,
			dbFileName,
			NULL,
			DB_BTREE,
			openFlag,
			0);
	}
	catch (DbException & e){
		cerr<<"Error while opening the ngram DB file "<<dbFileName<<endl;
		cerr<<"Error information: "<<e.get_errno()<<endl;
		exit(e.get_errno());
	}
}

C_AccessCharacterDB::~C_AccessCharacterDB()
{
	try{
		this->ngramDb->close(0);
		delete(this->ngramDb);
	}
	catch (DbException & e){
		cerr<<"Error while closing the DB"<<endl;
		cerr<<"Error information: "<<e.get_errno()<<endl;
		exit(e.get_errno());
	}
}

T_FREQ_INFO C_AccessCharacterDB::getFreq(C_String ngram)
{
	long currentTime;
	T_FREQ_INFO freq;

	//first check the cache
	map< C_String, S_Cached_Ngram_Freq_Info>::iterator iterCache;

	iterCache = this->cache.find(ngram);
	if(iterCache!=this->cache.end()){
		//exist in cache already

		//update the last visited time		
		time(&currentTime);
		iterCache->second.lastTimeVisited = currentTime;
		freq = iterCache->second.freq;
		return freq;
	}
	
	//else, not exist in cache yet
	int ngramLen = ngram.length();
	Dbt key(ngram.toString(), ngramLen);
	key.set_flags(DB_DBT_USERMEM);
	key.set_ulen(ngramLen);

	Dbt value(&freq, Size_Of_T_FREQ_INFO);
	value.set_flags(DB_DBT_USERMEM);
	value.set_ulen(Size_Of_T_FREQ_INFO);
	
	int result = this->ngramDb->get(NULL, &key, &value, 0);

	if(result!=0){
		freq = 0;	//does not exist
	}
	
	//else successful in access the DB
	//add the n-gram into the cache
	S_Cached_Ngram_Freq_Info tmpNode;
	time(&currentTime);

	tmpNode.freq = freq;
	tmpNode.lastTimeVisited = currentTime;

	this->cache.insert(make_pair(ngram, tmpNode));

	return freq;
}

void C_AccessCharacterDB::purgeCache(long freshTime)
{
	cerr<<"Size of cache: "<<this->cache.size()<<" ";
	map< C_String, S_Cached_Ngram_Freq_Info>::iterator iterCache1, iterCache2;

	iterCache1 = this->cache.begin();

	while(iterCache1 != this->cache.end() ){
		iterCache2=iterCache1;
		iterCache2++;

		if(iterCache1->second.lastTimeVisited<freshTime){
			this->cache.erase(iterCache1);
		}

		iterCache1=iterCache2;;
	}

	cerr<<" purged to: "<<this->cache.size()<<" entries\n";
}


