// _CharacterDBLM.h: interface for the C_CharacterDBLM class.
//
//////////////////////////////////////////////////////////////////////
//
// $Rev$
// $LastChangedDate$
//
//
#if !defined(AFX__CHARACTERDBLM_H__937F6551_3672_4026_A3D5_D4F017C53DA8__INCLUDED_)
#define AFX__CHARACTERDBLM_H__937F6551_3672_4026_A3D5_D4F017C53DA8__INCLUDED_

#include "_String.h"	// Added by ClassView
#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "LMInterface.hh"
#include "_AccessCharacterDB.h"
#include "shared.h"
#include "ngramDB_shared.h"

#include <map>

using namespace std;



const double CHAR_DBLM_LOG_0 = -40;



class C_CharacterDBLM : public CLM  
{
public:
	C_String returnSentEndString();
	C_String returnSentStartString();
	CLMSentenceState * NewSentence() const;

	void purgeCache(long lastVisitedTime);
	void StartsAnotherSentenceNow();

	virtual const vector< string > & Vocabulary() const;
	virtual LMWordIndex VocabularySize() const;
	LMWordIndex GetVocabularyIndex(string word);
	virtual LMWordIndex returnSentEndVocId();
	LMWordIndex returnSentStartVocId();

	double LogProb(vector<C_String> & history, C_String nextWord, vector<C_String> & nextStateAsNgram);
	
	void setParam_maxLenOfNgramConsidered(int maxLenOfNgramConsidered);
	void setParam_interpolationStrategy(char interpolationStrategy);
	void setParam_numberOfSentencesProcessedBeforeCachePurge(int sentNumber);
	void setParam_freshnessOfCacheToAvoidPurge(long freshTime);

	C_String getText(IndexType id);
	IndexType getVocId(C_String word);

	C_CharacterDBLM(const char * cfgFileName);
	virtual ~C_CharacterDBLM();

private:

	double calcLogProb(double *freq);
	double calcLogProb_equalWeightedInterpolation(double * freq);
	double calcLogProb_ibmHeuristicInterpolation(double *freq);
	double calcLogProb_maxProbInterpolation(double *freq);

	long freshTime;
	unsigned int numberOfSentencesProcessedBeforeCachedPurge;
	int sentProcessed;

	C_AccessCharacterDB * dbAccess;

	//internal vocabulary
	IndexType nextAvailableId;
	map<C_String, IndexType> text2id;
	map<IndexType, C_String> id2text;

	map<vector<C_String>, S_CachedNgramLogProb> cache;

	double corpusSize;
	int maxLenOfNgramConsidered;

	double CHAR_DBLM_PROB_UNK;
	double CHAR_DBLM_LOG_PROB_UNK;

	char interpolationStrategy;

	C_String sentStartString;
	IndexType sentStartVocId;

	C_String sentEndString;
	IndexType sentEndVocId;
};

#endif // !defined(AFX__CHARACTERDBLM_H__937F6551_3672_4026_A3D5_D4F017C53DA8__INCLUDED_)
