// _CharacterDBLMSentenceState.cpp: implementation of the C_CharacterDBLMSentenceState class.
//
//////////////////////////////////////////////////////////////////////
//
// $Rev$
// $LastChangedDate$
//

#include "_CharacterDBLMSentenceState.h"


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

C_CharacterDBLMSentenceState::C_CharacterDBLMSentenceState(const C_CharacterDBLM * cdblm)
{
	this->cdblm = (C_CharacterDBLM *) cdblm;
	this->sentEndVocId = this->cdblm->returnSentEndVocId();
}

C_CharacterDBLMSentenceState::~C_CharacterDBLMSentenceState()
{

}

void C_CharacterDBLMSentenceState::initialLmState()
{
	//add sentence start
	S_CDBLM_State_Info sentStartNode;
	sentStartNode.ngram.push_back(this->cdblm->returnSentStartString());	//if corpus is indexed correctly position 1 should be <s>

	this->allLMStates.push_back(sentStartNode);
	this->lmStateInfo2Id.insert(make_pair(sentStartNode, 0));

}

void C_CharacterDBLMSentenceState::resetLmStates()
{
	this->allLMStates.clear();
	this->buffer.clear();
	this->lmStateInfo2Id.clear();

}

//start a new sentence now, clear up the sentence LM state
CLMState C_CharacterDBLMSentenceState::BeginOfSentenceState()
{
	//check if we need to purge the cache 
	this->cdblm->StartsAnotherSentenceNow();

	this->resetLmStates();
	this->initialLmState();	

	return 0;
}

/**
* Apply LM over an edge
**/
void C_CharacterDBLMSentenceState::LogProb(CLMTransition & LMTransition)
{
	double totalCost = 0.0;
	CLMState currentState, nextState;

	currentState = LMTransition.OldState;
	int totalWords = LMTransition.Words.size();
	for(int i=0; i<totalWords; i++){
		totalCost+=this->LogProb(currentState, LMTransition.Words[i], nextState);
		currentState = nextState;
	}

	LMTransition.Prob = totalCost;
	LMTransition.NewState = currentState;

}

double C_CharacterDBLMSentenceState::LogProb(CLMState lmstate, const LMWordIndex nextWordId_in, CLMState & nextstate)
{

	IndexType nextWordId = (IndexType) nextWordId_in;

	//first check if we have already seen this before
	map< pair<CLMState, IndexType>, S_Buffered_CDBLM_Info>::iterator iterBuffer;
	iterBuffer = this->buffer.find( make_pair( lmstate, nextWordId) );

	if(iterBuffer==this->buffer.end()){ //we haven't seen this lmstate+word yet
		C_String nextWord = this->cdblm->getText(nextWordId);

		//search for it in the corpus
		S_CDBLM_State_Info lmStateInfo = this->allLMStates[lmstate];
		S_CDBLM_State_Info updatedLmStateInfo;
					
		double logProb = this->cdblm->LogProb(lmStateInfo.ngram, nextWord, updatedLmStateInfo.ngram);
		
		int updatedLmStateId;
		map<S_CDBLM_State_Info, int, lt_CDBLM_StateInfo>::iterator iterLmStateInfo2Id;
		iterLmStateInfo2Id = this->lmStateInfo2Id.find(updatedLmStateInfo);
		if(iterLmStateInfo2Id==this->lmStateInfo2Id.end()){	//this updated lm state does not exist yet
			this->allLMStates.push_back(updatedLmStateInfo);
			updatedLmStateId = this->allLMStates.size()-1;
			this->lmStateInfo2Id.insert(make_pair(updatedLmStateInfo, updatedLmStateId));
		}
		else{
			updatedLmStateId = iterLmStateInfo2Id->second;
		}

		//buffer this
		S_Buffered_CDBLM_Info bufferedLmInfo;
		bufferedLmInfo.logProb = logProb;
		bufferedLmInfo.nextState = updatedLmStateId;

		this->buffer.insert(make_pair( make_pair(lmstate, nextWordId), bufferedLmInfo));

		//updated next state
		nextstate = updatedLmStateId;
		
		return logProb;
	}

	nextstate = iterBuffer->second.nextState;

	return iterBuffer->second.logProb;
}

double C_CharacterDBLMSentenceState::LogProb(CLMState lmstate, const LMWordIndex nextWord )
{
  CLMState NewState;
  return LogProb( lmstate, nextWord, NewState );
}

double C_CharacterDBLMSentenceState::LogProbEnd(CLMState lmstate)
{
	CLMState dummyNextState;
	return this->LogProb(lmstate, this->sentEndVocId, dummyNextState);
}

double C_CharacterDBLMSentenceState::LogProbEnd(CLMTransition & LMTransition)
{
	CLMState dummyNextState;
	return this->LogProb(LMTransition.OldState, this->sentEndVocId, dummyNextState);
}

CLMState C_CharacterDBLMSentenceState::AppendWord( CLMState LMState, const LMWordIndex WordIndex )
{
	CLMState nextState;

	this->LogProb(LMState, WordIndex, nextState);

	return nextState;
}

