// _NgramDBLM.h: interface for the C_NgramDBLM class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX__NGRAMDBLM_H__B7FC1CC1_B203_4B45_8374_B946731AB065__INCLUDED_)
#define AFX__NGRAMDBLM_H__B7FC1CC1_B203_4B45_8374_B946731AB065__INCLUDED_

#include "shared.h"
#include "math.h"

#include "LMInterface.hh"
#include "_AccessNgramDB.h"

using namespace std;

const double DBLM_PROB_UNK = 0.00000000023283064365386962890625; // 1/4G
const double DBLM_LOG_PROB_UNK = log(DBLM_PROB_UNK);
const double DBLM_LOG_0 = -20;



class C_NgramDBLM  : public CLM
{
public:
	void purgeCache(long lastVisitedTime);
	void StartsAnotherSentenceNow();

	virtual LMWordIndex VocabularySize() const;
	virtual LMWordIndex returnSentEndVocId();
	LMWordIndex returnSentStartVocId();
	virtual double LogProb(vector<LMWordIndex> & history, LMWordIndex nextWord, vector<LMWordIndex> & nextStateAsNgram);

	virtual LMWordIndex GetVocabularyIndex(string word);
	virtual const vector< string > & Vocabulary() const;

	CLMSentenceState * NewSentence() const;

	void setParam_maxLenOfNgramConsidered(int maxLenOfNgramConsidered);
	void setParam_interpolationStrategy(char interpolationStrategy);
	void setParam_numberOfSentencesProcessedBeforeCachePurge(int sentNumber);
	void setParam_freshnessOfCacheToAvoidPurge(long freshTime);

	C_NgramDBLM(const char * cfgFileName);	
	virtual ~C_NgramDBLM();
private:
	
	double calcLogProb(double *freq);
	double calcLogProb_equalWeightedInterpolation(double * freq);
	double calcLogProb_ibmHeuristicInterpolation(double *freq);
	double calcLogProb_maxProbInterpolation(double *freq);


	LMWordIndex sentEndVocId;
	LMWordIndex sentStartVocId;
	int maxLenOfNgramConsidered;
	char interpolationStrategy;

	double corpusSize;
	C_AccessNgramDB * dbAccess;

	map< vector<LMWordIndex>, S_CachedNgramLogProb> cache;
	int numberOfSentencesProcessedBeforeCachedPurge;
	long freshTime;
	int sentProcessed;
};

#endif // !defined(AFX__NGRAMDBLM_H__B7FC1CC1_B203_4B45_8374_B946731AB065__INCLUDED_)

// _NgramDBLM.h: interface for the C_NgramDBLM class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX__NGRAMDBLM_H__B7FC1CC1_B203_4B45_8374_B946731AB065__INCLUDED_)
#define AFX__NGRAMDBLM_H__B7FC1CC1_B203_4B45_8374_B946731AB065__INCLUDED_

#include "shared.h"
#include "math.h"

#include "LMInterface.hh"
#include "_AccessNgramDB.h"

using namespace std;

const double DBLM_PROB_UNK = 0.00000000023283064365386962890625; // 1/4G
const double DBLM_LOG_PROB_UNK = log(DBLM_PROB_UNK);
const double DBLM_LOG_0 = -20;

/**
* cached information for n-gram log prob
* key: n-gram history plus next word
* cached: log prob as next word given history
* and the longest matching history
**/
typedef struct s_cachedNgramLogProb{
	double logProb;
	unsigned char longestMatchingHistory;	//the longest matching length when history extended with next word
	unsigned long lastVisitedTimeStamp;
}S_CachedNgramLogProb;

class C_NgramDBLM  : public CLM
{
public:
	void purgeCache(long lastVisitedTime);
	void StartsAnotherSentenceNow();

	virtual LMWordIndex VocabularySize() const;
	virtual LMWordIndex returnSentEndVocId();
	LMWordIndex returnSentStartVocId();
	virtual double LogProb(vector<LMWordIndex> & history, LMWordIndex nextWord, vector<LMWordIndex> & nextStateAsNgram);

	virtual LMWordIndex GetVocabularyIndex(string word);
	virtual const vector< string > & Vocabulary() const;

	CLMSentenceState * NewSentence() const;

	void setParam_maxLenOfNgramConsidered(int maxLenOfNgramConsidered);
	void setParam_interpolationStrategy(char interpolationStrategy);
	void setParam_numberOfSentencesProcessedBeforeCachePurge(int sentNumber);
	void setParam_freshnessOfCacheToAvoidPurge(long freshTime);

	C_NgramDBLM(const char * cfgFileName);	
	virtual ~C_NgramDBLM();
private:
	
	double calcLogProb(double *freq);
	double calcLogProb_equalWeightedInterpolation(double * freq);
	double calcLogProb_ibmHeuristicInterpolation(double *freq);
	double calcLogProb_maxProbInterpolation(double *freq);


	LMWordIndex sentEndVocId;
	LMWordIndex sentStartVocId;
	int maxLenOfNgramConsidered;
	char interpolationStrategy;

	double corpusSize;
	C_AccessNgramDB * dbAccess;

	map< vector<LMWordIndex>, S_CachedNgramLogProb> cache;
	int numberOfSentencesProcessedBeforeCachedPurge;
	long freshTime;
	int sentProcessed;
};

#endif // !defined(AFX__NGRAMDBLM_H__B7FC1CC1_B203_4B45_8374_B946731AB065__INCLUDED_)
