// _NgramDBLMSentenceState.cpp: implementation of the C_NgramDBLMSentenceState class.
//
//////////////////////////////////////////////////////////////////////

#include "_NgramDBLMSentenceState.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

C_NgramDBLMSentenceState::C_NgramDBLMSentenceState(const C_NgramDBLM * dblm)
{
	this->dblm = (C_NgramDBLM *) dblm;
	this->sentEndVocId = this->dblm->returnSentEndVocId();
}

C_NgramDBLMSentenceState::~C_NgramDBLMSentenceState()
{

}

void C_NgramDBLMSentenceState::initialLmState()
{
	//add sentence start
	S_DBLM_State_Info sentStartNode;
	sentStartNode.ngram.push_back(this->dblm->returnSentStartVocId());	//if corpus is indexed correctly position 1 should be <s>

	this->allLMStates.push_back(sentStartNode);
	this->lmStateInfo2Id.insert(make_pair(sentStartNode, 0));

}

void C_NgramDBLMSentenceState::resetLmStates()
{
	this->allLMStates.clear();
	this->buffer.clear();
	this->lmStateInfo2Id.clear();

}

//start a new sentence now, clear up the sentence LM state
CLMState C_NgramDBLMSentenceState::BeginOfSentenceState()
{
	//check if we need to purge the cache 
	this->dblm->StartsAnotherSentenceNow();

	this->resetLmStates();
	this->initialLmState();	

	return 0;
}

/**
* Apply LM over an edge
**/
void C_NgramDBLMSentenceState::LogProb(CLMTransition & LMTransition)
{
	double totalCost = 0.0;
	CLMState currentState, nextState;

	currentState = LMTransition.OldState;
	int totalWords = LMTransition.Words.size();
	for(int i=0; i<totalWords; i++){
		totalCost+=this->LogProb(currentState, LMTransition.Words[i], nextState);
		currentState = nextState;
	}

	LMTransition.Prob = totalCost;
	LMTransition.NewState = currentState;

}

double C_NgramDBLMSentenceState::LogProb(CLMState lmstate, const LMWordIndex nextWord, CLMState & nextstate)
{

	//first check if we have already seen this before
	map< pair<CLMState, IndexType>, S_Buffered_DBLM_Info>::iterator iterBuffer;
	iterBuffer = this->buffer.find( make_pair( lmstate, nextWord) );

	if(iterBuffer==this->buffer.end()){ //we haven't seen this lmstate+word yet
		//search for it in the corpus
		S_DBLM_State_Info lmStateInfo = this->allLMStates[lmstate];
		S_DBLM_State_Info updatedLmStateInfo;
					
		double logProb = this->dblm->LogProb(lmStateInfo.ngram, nextWord, updatedLmStateInfo.ngram);
		
		int updatedLmStateId;
		map<S_DBLM_State_Info, int, lt_DBLM_StateInfo>::iterator iterLmStateInfo2Id;
		iterLmStateInfo2Id = this->lmStateInfo2Id.find(updatedLmStateInfo);
		if(iterLmStateInfo2Id==this->lmStateInfo2Id.end()){	//this updated lm state does not exist yet
			this->allLMStates.push_back(updatedLmStateInfo);
			updatedLmStateId = this->allLMStates.size()-1;
			this->lmStateInfo2Id.insert(make_pair(updatedLmStateInfo, updatedLmStateId));
		}
		else{
			updatedLmStateId = iterLmStateInfo2Id->second;
		}

		//buffer this
		S_Buffered_DBLM_Info bufferedLmInfo;
		bufferedLmInfo.logProb = logProb;
		bufferedLmInfo.nextState = updatedLmStateId;

		this->buffer.insert(make_pair( make_pair(lmstate, nextWord), bufferedLmInfo));

		//updated next state
		nextstate = updatedLmStateId;
		
		return logProb;
	}

	nextstate = iterBuffer->second.nextState;

	return iterBuffer->second.logProb;
}

double C_NgramDBLMSentenceState::LogProb(CLMState lmstate, const LMWordIndex nextWord )
{
  CLMState NewState;
  return LogProb( lmstate, nextWord, NewState );
}

double C_NgramDBLMSentenceState::LogProbEnd(CLMState lmstate)
{
	CLMState dummyNextState;
	return this->LogProb(lmstate, this->sentEndVocId, dummyNextState);
}

double C_NgramDBLMSentenceState::LogProbEnd(CLMTransition & LMTransition)
{
	CLMState dummyNextState;
	return this->LogProb(LMTransition.OldState, this->sentEndVocId, dummyNextState);
}

CLMState C_NgramDBLMSentenceState::AppendWord( CLMState LMState, const LMWordIndex WordIndex )
{
	CLMState nextState;

	this->LogProb(LMState, WordIndex, nextState);

	return nextState;
}

