#include <stdlib.h>

#ifdef HAVE_CONFIG_H
#include <config.h>


#endif

#include <math.h>
#include <fstream>

#include <DistortionModelJump.hh>



CDistortionModelJump::CDistortionModelJump( CVocabulary *pVoc )
  : CDistortionModel( pVoc )
{
  m_pVoc = pVoc;

  m_pLogProbs = NULL;
  m_pLogProbs2 = NULL;
  m_pGlobalDistribution = NULL;
  m_pNormalDistribution = NULL;
}



CDistortionModelJump::~CDistortionModelJump()
{
  cout << "free memory in CDistortionModelJump" << endl;
  delete m_pLogProbs;
  delete m_pLogProbs2;
  delete m_pGlobalDistribution;
  delete m_pNormalDistribution;
  //  delete m_pTotalCounts;
}



void
CDistortionModelJump::Report()
{
  cout << "Report: Distortion model with dependency on one word" << endl;
  cout << "Dependency: " << m_DistortionModelType << endl;
  cout << "Maximum jump width: " << m_MaxJump << endl;
  //  cout << "Size of probability table = " << m_pLogProbs.Size() << endl;
  //CSparseMatrix *m_pLogProbs2;

  //CNumericalArray< float > *m_pNormalDistribution;

  //CNumericalArray< float > *m_pTotalCounts;

  
}




float
CDistortionModelJump::NormalDistribution( float mean, float stdev, float value )
{
	float t1 = pow(stdev * sqrt(2 * M_PI), -1.0);
	float t2 = (0.0 - ((value - mean) * (value - mean)) / (2 * stdev * stdev));
	return (t1 * exp(t2));
}


/*
int
CDistortionModelJump::Mode()
{
	return m_Mode;
}
*/


float
CDistortionModelJump::LogProb( int JumpWidth,
                 CWordIndex FromSourceWord, CWordIndex FromTargetWord,
                 CWordIndex ToSourceWord,   CWordIndex ToTargetWord )
{
  /*
  cout << "CDistortionModelJump::LogProb " << JumpWidth 
       << " " << FromSourceWord << " " << FromTargetWord
       << " " << ToSourceWord   << " " << ToTargetWord << endl;
  */

  switch ( m_DistortionModelType ) {
  case JumpSourceOutbound:
    return LogProb( JumpWidth, FromSourceWord );
    break;

  case JumpSourceInbound:
    return LogProb( JumpWidth, ToSourceWord );
    break;
    
  case JumpTargetOutbound:
    return LogProb( JumpWidth, FromTargetWord );
    break;

  case JumpTargetInbound:
    return LogProb( JumpWidth, ToTargetWord );
    break;

  default:
    cout << "unknow jump model" << endl;
    exit( -7 );
  }


}



float
CDistortionModelJump::LogProb( int Jump, CWordIndex WordIndex )
{
  //cout << "CDistortionModelJump::LogProb " << Jump << " " << WordIndex << endl;

	// if jump value is higher than the value stored, return probability according
  // to normal distribution
  if( (Jump > m_MaxJump) || (Jump < ( 0 - m_MaxJump )) ) {

    //cout << "Jump too large, return default value" << endl;
    return log( 0.001 );

    /*
	  float normVal = NormalDistribution( mu, sigma, (float)Jump );
    
    if( normVal > 0 ) {
      return log( normVal );
    } else {
      return log( 0.0001 );   // sv: return floor value
    }
    */

  }


  //if word is out of voc, return general value
  // sv: was if ( w1 <= WCount->MinCol() || w1 >= WCount->MaxCol() ) {
  // could have been the problem
  if ( ( WordIndex <= m_pLogProbs->MinIndex1() ) || ( WordIndex >= m_pLogProbs->MaxIndex1() ) ) {
	  //return log( m_pNormalDistribution[ Jump ] );
    //cout << "WordIndex " << WordIndex << " outside of range of stored words " << m_pLogProbs->MinIndex1()
    //<< " - " << m_pLogProbs->MaxIndex1() << "   return default value" << endl;
	  return log( 0.001 );
  }


  float LogProb = m_pLogProbs->GetAt( WordIndex, Jump );
  //cout << "found log probability " << LogProb << endl;

  return LogProb;

}



int
CDistortionModelJump::GetMaxJump()
{
  return m_MaxJump;
}



bool
CDistortionModelJump::Load( const char* filename )
{
  cout << "Loading Unigram Distortion Model..." << endl << flush;
  cout << "CDistortionModelJump::Load from " << filename << endl;

  ifstream Stream;
  Stream.open(filename, ios::binary);

  int TotalItems;

  // read the header
  int ModelType;
  Stream >> ModelType;
  Stream >> TotalItems;                       // number of items to load (should not be needed)
  Stream >> m_MaxJump;                        // the largest jump for which we have counts

  switch ( ModelType ) {
  case JumpSourceOutbound:
    cout << "Model type = JumpSourceOutbound" << endl;
    m_DistortionModelType = JumpSourceOutbound;
    break;

  case JumpSourceInbound:
    cout << "Model type = JumpSourceInbound" << endl;
    m_DistortionModelType = JumpSourceInbound;
    break;
    
  case JumpTargetOutbound:
    cout << "Model type = JumpTargetOutbound" << endl;
    m_DistortionModelType = JumpTargetOutbound;
    break;

  case JumpTargetInbound:
    cout << "Model type = JumpTargetInbound" << endl;
    m_DistortionModelType = JumpTargetInbound;
    break;

  default:
    cout << "unknow jump model" << endl;
    exit( -7 );
  }

  cout << "Items      = " << TotalItems << endl;
  cout << "MaxJump    = " << m_MaxJump << endl;


  m_pLogProbs           = new CNumericalMatrix< float >;
  m_pGlobalDistribution = new CNumericalArray< float >;
  m_pGlobalDistribution->AssureBoundaries( -m_MaxJump - 1, m_MaxJump + 1 );
  //m_pTotalCounts        = new CNumericalArray< float >;

  //bool retval = WCount->SetBoundaries(0-maxJump-1, maxJump+1, 0, (totalItems+ pVoc->Size()) /*totalItems*/);
  //WCount->SetDefaultValue(DMUnigram_noVal);
 
  /*
  int jNormTotals[(maxJump*2)+3];
  for(int j = 0; j < (2*maxJump)+3; j++){
    jNormTotals[j] = 0;
  }
  */


  //char* word1 = (char*)malloc(256 * sizeof(char));

  int ItemsLoaded = 0;
  CString Word( "" );

  // read first word
  Stream >> Word;
  while( !Stream.eof() ) {

    //cout << "Loaded " << Word << endl;

    int WordIndex = m_pVoc->GetIndex( Word );
    if( WordIndex == CVocabulary::UNK ) {
      WordIndex = m_pVoc->AddWord( Word );
      //indexToStore = pVoc->GetIndex(wrd.c_str());
      //cout << "is new word, added to vocabulary, index is " << WordIndex << endl;
    }

    double Count;
    for( int Jump = 0 - m_MaxJump - 1; Jump <= m_MaxJump + 1; Jump++ ) {
      Stream >> Count;
      if ( Count < 0 ) cout << "ERROR: " << Count << " is < 0" << endl;
      if ( Count < 0.000001 ) Count = 0.000001;  // this is because we take the log, and Count[ 0 ] = 0;
      //cout << "store " << Jump << " " << WordIndex << " " << m_pVoc->GetWord( WordIndex ) << " " << Count << endl;
      m_pLogProbs->SetAt( WordIndex, Jump, Count );
      (*m_pGlobalDistribution)[ Jump ] = (*m_pGlobalDistribution)[ Jump ] + Count;
      //(*m_pTotalCounts)[ WordIndex ] += Count;
    }
    //cout << endl;

    ItemsLoaded++;
    //if ( (ItemsLoaded % 1000) == 0 ) {
    //cout << ItemsLoaded << " items loaded" << endl;
    //}

    // read next word
    Stream >> Word;

  }

	cout << "done loading" << endl;
  cout << "total jump counts:" << endl;
  for ( int Jump = m_pGlobalDistribution->MinIndex(); Jump <= m_pGlobalDistribution->MaxIndex(); Jump++ ) {
    cout << (*m_pGlobalDistribution)[ Jump ] << " ";
  }
  cout << endl;



  // normalize, smooth and convert into probs
  // we smooth with the global distribution
  // the more often we have seen a word, the more we rely on the specific count
  // lambda = 1.0 / ( m_TotalCount[ w1 ] + 1.0 );
  // p( jump | w ) = ( lambda * Count( jump, w ) / TotalCount( w ) + ( 1 - lambda ) TotalCount( jump ) / TotalCount


  // first, normalize global distribution
  double TotalCounts = m_pGlobalDistribution->Sum();
  cout << "global jump probabilities:" << endl;
  for ( int Jump = m_pGlobalDistribution->MinIndex(); Jump <= m_pGlobalDistribution->MaxIndex(); Jump++ ) {
    (*m_pGlobalDistribution)[ Jump ] /= TotalCounts;
    cout << (*m_pGlobalDistribution)[ Jump ] << " ";
  }
  cout << endl;
  

  for ( int WordIndex = m_pLogProbs->MinIndex1(); WordIndex <= m_pLogProbs->MaxIndex1(); WordIndex++ ) {

    cout << WordIndex << "  Word = " << m_pVoc->GetWord( WordIndex ) << endl;

    CNumericalArray< float > &Row = m_pLogProbs->GetRow( WordIndex );
    //double Sum = Row.Sum();
    double Count = 0.0;
    double TotalCountForCurrentWord = 0.0;
    //double Sum   = 0.0;

    for( int Jump = 0 - m_MaxJump - 1; Jump <= m_MaxJump + 1; Jump++ ) {
      //cout << Row[ Jump ] << " ";
      Count = Row[ Jump ];
      TotalCountForCurrentWord += Count;
      //if ( Count <= 0.000001 ) {
        //Count = 0.000001;
        Row.SetAt( Jump, Count );
        //}
      //Sum += Count;
    }
    //cout << endl;

    double Lambda = 1.0 / ( ( 1.0 - m_SmoothingFactor ) * TotalCountForCurrentWord + 1.0 );
    //cout << "SmoothingFactor = " << m_SmoothingFactor << " Lambda = " << Lambda << " TotalCount = " << TotalCountForCurrentWord << endl;

    for( int Jump = 0 - m_MaxJump - 1; Jump <= m_MaxJump + 1; Jump++ ) {
      Count = ( 1.0 - Lambda ) * Row[ Jump ] / TotalCountForCurrentWord + Lambda * (*m_pGlobalDistribution)[ Jump ];
      //cout << Count << ":";
      Row[ Jump ] = Count;
      //cout << Row[ Jump ] << " ";
    }
    //cout << endl;
    
      
    //    Row.Div( Sum );
    Row.Log();

    //for( int Jump = 0 - m_MaxJump - 1; Jump <= m_MaxJump + 1; Jump++ ) {
    //cout << Row[ Jump ] << " ";
    //}
    //cout << endl;
    
  }

  /*
  // calculate a Gaussian distribution
  jNorms = (double*)malloc(((maxJump*2)+3) * sizeof(double));
  m_Mu = 0.0;
  for ( int j = 0; j < ( 2 * maxJump ) + 3; j++ ) {
    jNorms[ j ] = (double)jNormTotals[ j ] / (double)total;
    if ( j != 0 && j != ( 2 * maxJump + 2 ) ){ 
      m_Mu += jNorms[ j ]; 
    }
  }
  m_Mu = mu / (2 * maxJump + 1);
	m_Sigma = 1.0;
	double sumOfSquares = 0.0;
  //calculate mu and sigma
  for ( int j = 1; j < ( 2 * maxJump ) + 2; j++ ) {
    sumOfSquares += (( jNorms[ j ] - mu ) * ( jNorms[ j ] - mu ) );
  }
  m_Sigma = sqrt( (1.0 / ( (2 * maxJump + 1) * sumOfSquares ) );
  //free(jNormTotals);
  */

  return true;
}

