#ifndef __DistortionModelJump__
#define __DistortionModelJump__


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <DistortionModel.hh>


#define DMUnigram_noVal 0.0
#define DMUnigram_defaultProb .0000002
#define _DMUNIGRAM_LEFT -1
#define _DMUNIGRAM_RIGHT 1
#define _DMUNIGRAM_CENTER 0




class CDistortionModelJump: public CDistortionModel

{
public:
  //! create a distortion model and give it a vocabulary
  CDistortionModelJump( CVocabulary* pVocabulary );
  ~CDistortionModelJump();

  void Report();

  //! load distortion model from file
  bool Load( const char* FileName );

  int GetMaxJump();
  int Mode();

  virtual float LogProb( int JumpWidth,
                         CWordIndex FromSourceWord, CWordIndex FromTargetWord,
                         CWordIndex ToSourceWord,   CWordIndex ToTargetWord );

  //! normal distribution, used for smoothing
  float NormalDistribution( float mean, float stdev, float value );


private:
  void Init();

  //! return the probability for a distorion (jump) and a word index
  float LogProb( int JumpWidth, CWordIndex Word );

  int m_MaxJump;
  //int m_Mode;

  //! pointer to the vocabulary
  // CVocabulary* m_pVoc;  // in the base class

  //double* jNorms; //jump normalizer
  //CSparseMatrix* m_pLogProbs2;  // in base class

  //! a matrix, which stores the log probabilities
  // first dimension is word jump, second dimension is word index
  //CNumericalMatrix< float > *m_pLogProbs;  // in base class

  //! the background model, a normal distribution or an accumulated distribution
  //  i.e. counts summed over all words
  CNumericalArray< float > *m_pNormalDistribution; //jump normalizer


  //! for normal distribution, the average
  float mu;

  //! for normal distribution, the standard deviation
  float sigma;
};





#endif














