#include <stdlib.h>
#include <config.h>
#include <fortify.h>
#include <string>
#include <map>
#include <stdio.h>

#include <CompactNBest.hh>

//ARV 
//Removed to make sure this is not always required
//#include <gsl/gsl_math.h>

//ARV 07/06 
// double* fb = new double[m_numFeatures];
// CNBestItem* oneItem = new CNBestItem();

CTransducerState CNBest::s_NotFound = -1;
CVocabulary* CNBestItem::m_pVoc = NULL;
int CNBestItem::m_numFeatures;

CNBestItem::CNBestItem(){
  m_isCalculated = false;
  for (int i=0; i<NSZ; i++){
    m_numberCorrect[i] = 0;
    m_numberSuggested[i] = 0;
  }
  m_closestRefLength = 0;
  
}

CNBestItem::~CNBestItem(){
  for (FeatureBank::iterator iter=m_FeatureSets.begin(); iter!=m_FeatureSets.end(); ++iter){
    double* data = (*iter);
    delete [] data;
  }
  
}

bool CNBestItem::ReplaceFeatureSet(double* fb){
  //no need to check to see if another feature set exists
  
  if (m_FeatureSets.size()==0){
    double* fbNew = new double[CNBestItem::m_numFeatures];
    for (int i=0; i<CNBestItem::m_numFeatures; i++){
      fbNew[i] = fb[i];
    }    
    m_FeatureSets.push_back(fbNew);
  }else{
    for (int i=0; i<CNBestItem::m_numFeatures; i++){
      m_FeatureSets[0][i] = fb[i];
    }
  }
  return true;

}

bool
CNBestItem::AddFeatureSet(double* fb){
  // go thru the each level, if you see a difference, add it and quit
  //cout<<"Comparing against:"<<m_FeatureSets.size()<<endl;
  
  int numExamined = 0;
  for(FeatureBank::iterator iter = m_FeatureSets.begin(); iter!=m_FeatureSets.end(); ++iter){
    
    numExamined++;
    //cout<<"Comparing:"<<(*iter)[0]<<" -- "<<fb[0]<<endl;
    //cout<<"Comparing:"<<(*iter)[1]<<" -- "<<fb[1]<<endl;
    //cout<<"Comparing:"<<(*iter)[2]<<" -- "<<fb[2]<<endl;
    //cout<<"Comparing:"<<(*iter)[3]<<" -- "<<fb[3]<<endl;
    //cout<<"NumFeatures:"<<m_numFeatures<<endl;
    
    int f=0;
    
    double* dptr = (*iter);

    //if there is a difference detected, we break out
    while ((fabs(dptr[f]- fb[f])<0.001) && (f<m_numFeatures)){
      //cout<<"Inside Loop:"<<f<<endl;
      f++;
    }

    //there is already an exact match in here    
    if (f==m_numFeatures){
      //cout<<"SlowDuplicate("<<numExamined<<"):"<<fb[0]<<" "<<fb[1]<<" "<<fb[2]<<" "<<fb[3]<<" "<<endl;
      return false;
    }      
    
  }

  //cout<<"SlowNew("<<numExamined<<"):"<<fb[0]<<" "<<fb[1]<<" "<<fb[2]<<" "<<fb[3]<<" "<<endl;    
  double* fbNew = new double[CNBestItem::m_numFeatures];
  for (int i=0; i<CNBestItem::m_numFeatures; i++){
    fbNew[i] = fb[i];
  }
  m_FeatureSets.push_back(fbNew);
  return true;

  /*
  double product = 1.0;
  for (int i=0; i<m_numFeatures; i++){
    
    if (fb[i]!=0){
      product *= fb[i];
    }
  }
  */
  //cout<<"Product generated:"<<product<<endl;

  /*
  if (m_uniqueFeatures.count(product)!=0){
        
    for(FeatureBank::iterator iter = m_FeatureSets.begin(); iter!=m_FeatureSets.end(); ++iter){
            
      //cout<<"Comparing:"<<(*iter)[0]<<" -- "<<fb[0]<<endl;
      //cout<<"Comparing:"<<(*iter)[1]<<" -- "<<fb[1]<<endl;
      //cout<<"Comparing:"<<(*iter)[2]<<" -- "<<fb[2]<<endl;
      //cout<<"Comparing:"<<(*iter)[3]<<" -- "<<fb[3]<<endl;
      //cout<<"NumFeatures:"<<m_numFeatures<<endl;
      
      int f=0;
      
      double* dptr = (*iter);
      
      while ((fabs(dptr[f]- fb[f])<0.001) && (f<m_numFeatures)){
	//cout<<"Inside Loop:"<<f<<endl;
	f++;
      }
      
      if (f==m_numFeatures){
	// complete match against another node, just reject it
	//cout<<"True Duplicate Found!"<<endl;
	cout<<"SlowDuplicate:"<<fb[0]<<" "<<fb[1]<<" "<<fb[2]<<" "<<fb[3]<<" "<<endl;
	return false;
      }
    }
    
    cout<<"SlowNew: "<<fb[0]<<" "<<fb[1]<<" "<<fb[2]<<" "<<fb[3]<<" "<<endl;    
    double* fbNew = new double[CNBestItem::m_numFeatures];
    for (int i=0; i<CNBestItem::m_numFeatures; i++){
      fbNew[i] = fb[i];
    }
    m_FeatureSets.push_back(fbNew);
    return false;
  }else{
    
    
    cout<<"FastNew: "<<fb[0]<<" "<<fb[1]<<" "<<fb[2]<<" "<<fb[3]<<" "<<endl;
    double* fbNew = new double[CNBestItem::m_numFeatures];
    for (int i=0; i<CNBestItem::m_numFeatures; i++){
      fbNew[i] = fb[i];
    }
    m_FeatureSets.push_back(fbNew);
    
    m_uniqueFeatures.insert(pair<double, double*>(product,fbNew));
    return true;
  }
  */
}



Counter* CNBestItem::getNumberCorrect(){
  return &m_numberCorrect;
}
Counter* CNBestItem::getNumberSuggested(){
  return &m_numberSuggested;
}
double* CNBestItem::getClosestRefLength(){
  return &m_closestRefLength;
}

void CNBestItem::setNGramData(int n, double correctScore, double suggestScore){
  m_numberCorrect[n] = correctScore;
  m_numberSuggested[n] = suggestScore;
}

void CNBestItem::incrementNGramData(int n, double correctScore, double suggestScore){
  m_numberCorrect[n] += correctScore;
  m_numberSuggested[n] += suggestScore;
}

void CNBestItem::setRefLengthData(double refLength){
  m_closestRefLength = refLength;
}

void CNBestItem::incrementRefLengthData(double refLength){
  m_closestRefLength += refLength;
}



void CNBestItem::setCandidateVector(std::vector<CWordIndex> candidateVector){
  m_candidateVector = candidateVector;
}

string CNBestItem::getCandidateString(){
  stringstream ss(stringstream::in|stringstream::out);

  int numElements = 0;
  for (vector<CWordIndex>::iterator iter = m_candidateVector.begin(); iter!=m_candidateVector.end(); ++iter){
    string str = string(m_pVoc->GetWord((*iter)));
    //cout<<"Converting:"<<str<<":"<<endl;
    ss<<str;
    vector<CWordIndex>::iterator tmpIter = iter;
    if (numElements!=m_candidateVector.size()-1){
      ss<<" ";
    }
    numElements++;
  }
  string retString = string(ss.str());
  return retString;
}

void CNBestItem::AddEdgeList(EdgeList eList){
  m_EdgeSets.Append(eList);
}

EdgeBank& CNBestItem::getEdgeBank(){
  return m_EdgeSets;
}

double CNBestItem::GetConfidenceScore(int pathNumber, CCompactTSequence<double>* wordProbLookup){

  int numEdges = m_EdgeSets[pathNumber].Size();
  //cout<<"numEdges: "<<numEdges<<" ";
  //go thru each edge and print it
  double confidenceScore = 0.0;

  int numSourceWords = 0;
  for (int d=0; d<numEdges; d++){
    CWordInfo* wInfo = m_EdgeSets[pathNumber][d];
    cout<<*wInfo;
    int srcLen = wInfo->SourceStringLength();
    int tgtLen = wInfo->TargetStringLength();
    
    //cout<<"srcLen: "<<srcLen<<" tgtLen: "<<tgtLen<<" ";
    
      CSentence srcSentence = wInfo->SourceWords();
      CSentence tgtSentence = wInfo->TargetWords();


      vector<int> sourceVector;
      for (int sw=srcSentence.MinIndex(); sw<=srcSentence.MaxIndex(); sw++){
	CString srcWord = srcSentence.Word(sw);
	sourceVector.push_back(m_pVoc->GetIndex(srcWord));
      }

      vector<int> targetVector;
      for (int tw=tgtSentence.MinIndex(); tw<=tgtSentence.MaxIndex(); tw++){
	CString tgtWord = tgtSentence.Word(tw);
	targetVector.push_back(m_pVoc->GetIndex(tgtWord));
      }
      CCompactTSequence<int>* sourceGrams = CollectNGrams(sourceVector, sourceVector.size());
      CCompactTSequence<int>* targetGrams = CollectNGrams(targetVector, targetVector.size());      

      double singleEdgeScore = 0.0;
      //now go thru each word in the tgtVoc, if its been aligned in this sentence, give it the
      //prob indicated, otherwise, give it 1-prob.
      singleEdgeScore = GetPhraseConfidence(sourceGrams, targetGrams, wordProbLookup);


      confidenceScore+=singleEdgeScore;
      cout<<" se: "<<singleEdgeScore<<" ";
  }
  return confidenceScore/(double)numEdges;
}

void CNBestItem::PrintOptimizationData(const int sentenceNumber, const int NGramsConsidered,  int& startingIndex, ostream& os){
  
  for (FeatureBank::iterator iter = m_FeatureSets.begin(); iter!= m_FeatureSets.end(); ++iter){
    double* oneSet = (*iter);
    //os<<"Translation:"<<getCandidateString()<<":"<<endl;
    //os<<"SFeatures: "<<sentenceNumber<<" "<<startingIndex<<" ";
    for (int i=0; i<CNBestItem::m_numFeatures; i++){
      //os.precision(10);
      os<<-1.0*oneSet[i]<<" ";
    }
    for (int n=0; n<NGramsConsidered; n++){
      os<<m_numberCorrect[n]<<" "<<m_numberSuggested[n]<<" ";
    }
    os<<m_closestRefLength<<endl;
    startingIndex++;
  }
}

void CNBestItem::PrintOptimizationDataWithSentenceAndSegmentation(const int sentenceNumber, const int NGramsConsidered, int& startingIndex, vector<float>& scalingFactors, fstream& os){



    
    /*
      //only write the translation entries, now we have the write source and target phrases
    int sz = m_candidateVector.size();
    os.write((char*)&(sz), sizeof(int));
    for (int i=0; i<m_candidateVector.size(); i++){
      //cout<<m_candidateVector[i];
      if (i!=m_candidateVector.size()-1){
	//cout<<" ";
      }
      os.write((char*)&(m_candidateVector[i]), sizeof(int));
    }
    */

    //for each full translation (one EdgeSet)
    int numFullPaths = m_EdgeSets.Size();
    os.write((char*)&(numFullPaths), sizeof(int));
    //cout<<"numFullPaths: "<<numFullPaths<<endl;

    for (int es=m_EdgeSets.MinIndex(); es<=m_EdgeSets.MaxIndex(); es++){


      int numEdges = m_EdgeSets[es].Size();
      os.write((char*)&(numEdges), sizeof(int));
      //cout<<"numEdges: "<<numEdges<<" ";
      //go thru each edge and print it
      for (int d=0; d<numEdges; d++){
	CWordInfo* wInfo = m_EdgeSets[es][d];
	//cout<<*wInfo;
	int srcLen = wInfo->SourceStringLength();
	int tgtLen = wInfo->TargetStringLength();
	os.write((char*)&(srcLen), sizeof(int));
	os.write((char*)&(tgtLen), sizeof(int));
	//cout<<"srcLen: "<<srcLen<<" tgtLen: "<<tgtLen<<" ";

	CSentence srcSentence = wInfo->SourceWords();
	CSentence tgtSentence = wInfo->TargetWords();
	for (int i=srcSentence.MinIndex(); i<=srcSentence.MaxIndex(); i++){
	  int src = m_pVoc->GetIndex(srcSentence.Word(i));
	  os.write((char*)&(src), sizeof(int));

	  //cout<<src<<" ";
	}
	//cout<<" # ";

	for (int i=tgtSentence.MinIndex(); i<=tgtSentence.MaxIndex(); i++){
	  int tgt = m_pVoc->GetIndex(tgtSentence.Word(i));
	  os.write((char*)&(tgt), sizeof(int));
	  //cout<<tgt<<" ";
	}

      }
      //cout<<endl;


      double* oneSet = m_FeatureSets[es];
      
      
      //cout<<"SFeatures: ("<<sentenceNumber<<" "<<startingIndex<<") ";
      
      os.write((char*)&(CNBestItem::m_numFeatures), sizeof(int));
      for (int i=0; i<CNBestItem::m_numFeatures; i++){
	//cout.precision(10);
	//cout<<-1.0*oneSet[i]<<" ";
	float mod = -1.0*oneSet[i];
	os.write((char*)&(mod), sizeof(float));
      }
      
      
      
      int nn=NGramsConsidered;
      //cout<<" NGCons: ("<<nn<<") ";
      os.write((char*)&(nn), sizeof(int));
      for (int n=0; n<nn; n++){
	//cout<<m_numberCorrect[n]<<" "<<m_numberSuggested[n]<<" ";
	os.write((char*)&(m_numberCorrect[n]), sizeof(double));
	os.write((char*)&(m_numberSuggested[n]), sizeof(double));
      }
      //cout<<m_closestRefLength<<endl;
      os.write((char*)&(m_closestRefLength), sizeof(double));
      startingIndex++;
      //cout<<endl;
  }

}

void CNBestItem::PrintOptimizationDataWithSentence(const int sentenceNumber, const int NGramsConsidered, int& startingIndex, vector<float>& scalingFactors, fstream& os){
  
  for (FeatureBank::iterator iter = m_FeatureSets.begin(); iter!= m_FeatureSets.end(); ++iter){
    

    double* oneSet = (*iter);
    //cout<<"Translation: ";
    int sz = m_candidateVector.size();
    os.write((char*)&(sz), sizeof(int));
    for (int i=0; i<m_candidateVector.size(); i++){
      //cout<<m_candidateVector[i];
      if (i!=m_candidateVector.size()-1){
	//cout<<" ";
      }
      os.write((char*)&(m_candidateVector[i]), sizeof(int));
    }
    //cout<<endl;
    //cout<<"SFeatures: "<<sentenceNumber<<" "<<startingIndex<<" ";
    
    os.write((char*)&(CNBestItem::m_numFeatures), sizeof(int));
    for (int i=0; i<CNBestItem::m_numFeatures; i++){
      //cout.precision(10);
      //cout<<-1.0*oneSet[i]<<" ";
      float mod = -1.0*oneSet[i];
      os.write((char*)&(mod), sizeof(float));
    }
    

    
    int nn=NGramsConsidered;
    //cout<<" NGCons: "<<nn<<endl;
    os.write((char*)&(nn), sizeof(int));
    for (int n=0; n<nn; n++){
      //cout<<m_numberCorrect[n]<<" "<<m_numberSuggested[n]<<" ";
      os.write((char*)&(m_numberCorrect[n]), sizeof(double));
      os.write((char*)&(m_numberSuggested[n]), sizeof(double));
    }
    //cout<<m_closestRefLength<<endl;
    os.write((char*)&(m_closestRefLength), sizeof(double));
    startingIndex++;
  }
}

vector<double> CNBestItem::GenerateAllSequenceProbs(vector<float>& scaling){

  vector<double> eachPathProb;
  for (FeatureBank::iterator iter = m_FeatureSets.begin(); iter!= m_FeatureSets.end(); ++iter){
    double* oneSet = (*iter);
    double sum = 0.0;
    for (int i=0; i<CNBestItem::m_numFeatures; i++){
      //cout.precision(10);
      //cout<<-1.0*oneSet[i]<<" ";
      float mod = oneSet[i];
      sum+=mod*scaling[i];
    }
    double prob = exp(sum);
    eachPathProb.push_back(prob);
    //cout<<"Sum: "<<sum<<" Prob: "<<prob<<endl;
  }
  return eachPathProb;
}

void CNBestItem::IncrementOccurrencesNGram(CCompactTSequence<double>* wordProbLookup, double prob,   CCompactTSequence<double>* conditionedSum){

  //indentify all pair occurrences
  for (int es=m_EdgeSets.MinIndex(); es<=m_EdgeSets.MaxIndex(); es++){
    
    int numEdges = m_EdgeSets[es].Size();
    //cout<<"numEdges: "<<numEdges<<" ";
    //go thru each edge and print it
    for (int d=0; d<numEdges; d++){
      CWordInfo* wInfo = m_EdgeSets[es][d];
      //cout<<*wInfo<<endl;
      int srcLen = wInfo->SourceStringLength();
      int tgtLen = wInfo->TargetStringLength();

      //cout<<"srcLen: "<<srcLen<<" tgtLen: "<<tgtLen<<" ";
      
      CSentence srcSentence = wInfo->SourceWords();
      CSentence tgtSentence = wInfo->TargetWords();

      vector<int> sourceVector;
      for (int sw=srcSentence.MinIndex(); sw<=srcSentence.MaxIndex(); sw++){
	CString srcWord = srcSentence.Word(sw);
	sourceVector.push_back(m_pVoc->GetIndex(srcWord));
      }

      vector<int> targetVector;
      for (int tw=tgtSentence.MinIndex(); tw<=tgtSentence.MaxIndex(); tw++){
	CString tgtWord = tgtSentence.Word(tw);
	targetVector.push_back(m_pVoc->GetIndex(tgtWord));
      }
      CCompactTSequence<int>* sourceGrams = CollectNGrams(sourceVector, sourceVector.size());
      CCompactTSequence<int>* targetGrams = CollectNGrams(targetVector, targetVector.size());

      
      IncrementAndSum(sourceGrams, targetGrams, wordProbLookup, conditionedSum, prob);
      
      
      sourceGrams->ClearAll();
      targetGrams->ClearAll();
      delete sourceGrams;
      delete targetGrams;

    }
  }  
}

double CNBestItem::GetPhraseConfidence(CCompactTSequence<int>* conditioned, CCompactTSequence<int>* randomVariable, CCompactTSequence<double>* wordProbLookup){

  double eventHappenned = 0;
  double eventSpace = 0;
  for (int t = conditioned->m_TransducerItems.MinIndex(); t<=conditioned->m_TransducerItems.MaxIndex(); t++){
    
    vector<int> targetSequence = conditioned->m_TransducerItems[t]->GetSequence();
    //append the source grams    
    for (int s=randomVariable->m_TransducerItems.MinIndex(); s<=randomVariable->m_TransducerItems.MaxIndex(); s++){
      vector<int> sourceSequence = randomVariable->m_TransducerItems[s]->GetSequence();
      
      vector<int> targetSequenceCopy = targetSequence;
      for (int sw=0; sw<sourceSequence.size(); sw++){
	targetSequenceCopy.push_back(sourceSequence[sw]);
      }
      
      //cout<<"Searching:";
      //for (int t=0; t<targetSequenceCopy.size(); t++){
      //  cout<<" "<<m_pVoc->GetWord(targetSequenceCopy[t]);
      //}
      //cout<<endl;
      
      CCompactTSequenceItem<double>* existing = wordProbLookup->GetCompactTSequence(targetSequenceCopy);
      if (existing!=NULL){
	//cout<<"Found: ";
	//existing->Display(m_pVoc);	
	eventHappenned +=existing->GetValue();
      }
    }
    eventSpace += randomVariable->m_TransducerItems.Size();
  }
  return eventHappenned/eventSpace;
}

void CNBestItem::IncrementAndSum(CCompactTSequence<int>* conditioned, CCompactTSequence<int>* randomVariable, CCompactTSequence<double>* wordProbLookup,  CCompactTSequence<double>* conditionedSum, double prob){

  for (int t = conditioned->m_TransducerItems.MinIndex(); t<=conditioned->m_TransducerItems.MaxIndex(); t++){

    vector<int> targetSequence = conditioned->m_TransducerItems[t]->GetSequence();
    int numSourceGrams = randomVariable->m_TransducerItems.Size();
    
    CCompactTSequenceItem<double>* existingTarget = conditionedSum->GetCompactTSequence(targetSequence);
    if (existingTarget==NULL){
      existingTarget=conditionedSum->InsertPhrasePair(targetSequence, prob*numSourceGrams);
      existingTarget->SetKey(1);
    }else{
      existingTarget->SetValue(existingTarget->GetValue()+prob*numSourceGrams);
      existingTarget->SetKey(existingTarget->GetKey()+1);
    }

    int numTargetWords = targetSequence.size();
    //cout<<"Conditioned:"<<numTargetWords<<endl;
    //existingTarget->Display(m_pVoc);
    //append the source grams    
    
    for (int s=randomVariable->m_TransducerItems.MinIndex(); s<=randomVariable->m_TransducerItems.MaxIndex(); s++){
      vector<int> sourceSequence = randomVariable->m_TransducerItems[s]->GetSequence();
      
      vector<int> targetSequenceCopy = targetSequence;
      for (int sw=0; sw<sourceSequence.size(); sw++){
	targetSequenceCopy.push_back(sourceSequence[sw]);
      }
      
      CCompactTSequenceItem<double>* existing = wordProbLookup->GetCompactTSequence(targetSequenceCopy);
      if (existing==NULL){
	existing = wordProbLookup->InsertPhrasePair(targetSequenceCopy, prob);
	existing->SetKey(numTargetWords);
	//cout<<" new: "<<prob<<endl;
      }else{
	existing->SetValue(existing->GetValue()+prob);
	//cout<<" inc: "<<prob<<endl;
      }
      //existing->Display(m_pVoc);
      
      
    }
  }
}

void CNBestItem::IncrementOccurrences(CCompactTSequence<double>* wordProbLookup, double prob){

  //indentify all pair occurrences
  for (int es=m_EdgeSets.MinIndex(); es<=m_EdgeSets.MaxIndex(); es++){
    
    int numEdges = m_EdgeSets[es].Size();
    //cout<<"numEdges: "<<numEdges<<" ";
    //go thru each edge and print it
    for (int d=0; d<numEdges; d++){
      CWordInfo* wInfo = m_EdgeSets[es][d];
      //cout<<*wInfo;
      int srcLen = wInfo->SourceStringLength();
      int tgtLen = wInfo->TargetStringLength();

      //cout<<"srcLen: "<<srcLen<<" tgtLen: "<<tgtLen<<" ";
      
      CSentence srcSentence = wInfo->SourceWords();
      CSentence tgtSentence = wInfo->TargetWords();

      //UNIGRAM only procedure
      
	 
      for (int sw=srcSentence.MinIndex(); sw<=srcSentence.MaxIndex(); sw++){
	for (int tw=tgtSentence.MinIndex(); tw<=tgtSentence.MaxIndex(); tw++){
	  vector<CWordIndex> check;

	  CString srcWord = srcSentence.Word(sw);
	  CString tgtWord = tgtSentence.Word(tw);
	  //cout<<"Pair: "<<srcWord<<" "<<tgtWord;

	  check.push_back(m_pVoc->GetIndex(srcWord));
	  check.push_back(m_pVoc->GetIndex(tgtWord));

	  CCompactTSequenceItem<double>* existing = wordProbLookup->GetCompactTSequence(check);
	  if (existing==NULL){
	    existing = wordProbLookup->InsertPhrasePair(check, prob);
	    existing->SetKey(1);
	    //cout<<" new: "<<prob<<endl;
	  }else{
	    existing->SetValue(existing->GetValue()+prob);
	    existing->SetKey(existing->GetKey()+1);
	    //cout<<" inc: "<<prob<<endl;
	  }
	}
      }    
  

    }
  }  
}

/*
void CNBestItem::RescoreAllTranslations(CLexicon* pLexicon, CSentence& sourceSentence){
  CRescorerTM_IBM1 rescoreIBM1;
  rescoreIBM1.Init(pLexicon);
  
  CSentence targetSentence;
  for (int i=0; i<m_candidateVector.size(); i++){
    targetSentence.Append(m_candidateVector[i]);
  }
  //sourceSentence.Display(m_pVoc);
  //cout<<endl;
  //targetSentence.Display(m_pVoc);
  //cout<<endl;
  
  double logProbIBM1 = rescoreIBM1.LogProb(sourceSentence, targetSentence);
  //cout<<"logProbIBM1: "<<logProbIBM1<<endl;
  for (FeatureBank::iterator iter=m_FeatureSets.begin(); iter!=m_FeatureSets.end(); ++iter){
    double* data = (*iter);
    data[0] = logProbIBM1;
  }
}
*/

//returns the minCost of the elements within this item
double CNBestItem::GetMinCost(vector<float>& scalingFactors){

  double minCost = MAXBUF;
  for (FeatureBank::iterator iter=m_FeatureSets.begin(); iter!=m_FeatureSets.end(); ++iter){
    double* data = (*iter);
    double cost=0;
    for (int i=0; i<scalingFactors.size(); i++){
      cost += scalingFactors[i]*data[i];
    }
    if (cost<minCost){
      cost = minCost;
    }
  }
  return minCost;
}

FeatureBank* CNBestItem::getFeatureSets(){
  return &m_FeatureSets;
}

bool CNBestItem::isCalculated(){
  return m_isCalculated;
}

void CNBestItem::setCalculated(bool isCalc){
  m_isCalculated = isCalc;
}


CNBest::CNBest(){
  m_pVoc = NULL;
}

CNBest::CNBest(int numFeatures){
  m_pVoc = NULL;
  //m_scalingFactors = scalingFactors;
  m_numFeatures = numFeatures;
  CNBestItem::m_numFeatures = m_numFeatures;
  tmpHolder = new double[CNBestItem::m_numFeatures];
  m_NumberOfStatesSource = 0;
}

CNBest::~CNBest(){
  for ( int i = m_TransducerItems.MinIndex(); i <= m_TransducerItems.MaxIndex(); i++ ) {
    Trace( 5, "delete transducer item " << i );
    delete m_TransducerItems[ i ];
    m_TransducerItems[ i ] = NULL;
  }
  delete [] tmpHolder;
}

bool CNBest::Init(CVocabulary* m_pVoc, vector<float> scalingFactorPtrs){
  this->m_pVoc = m_pVoc;
  CNBestItem::m_pVoc = m_pVoc;
  
  //global float* 's 
  m_globalScalingPtrs = scalingFactorPtrs;
}

bool CNBest::Init( CVocabulary *pInputVocabulary ){
  this->m_pVoc = pInputVocabulary;
  CNBestItem::m_pVoc = m_pVoc;
  

}


// sv 051102 Decoder result should not be decoder spefic
//int CNBest::LoadFromBank(int sentenceNumber, CArray<CDecoderNBResult*>& TranslationBank, CSentence sourceSentence)
int CNBest::LoadFromBank(int sentenceNumber, CArray<CDecoderResult*>& TranslationBank, CSentence sourceSentence)
{

  m_sourceSentence = sourceSentence;
  // assist in adding elements to the source table
  int Pos, NextPos;
  int Length;  // length of source = number of words
  std::string Word;
  CWordIndex WordIndex;
  CTransducerState State, NextState;

  //build up the transducer bank as well
  

  // go thru each translation and initialize the counter data
  int newTranslationsFound = 0;
  for(int i=TranslationBank.MinIndex(); i<=TranslationBank.MaxIndex(); i++){
    //cout<<"Adding:"<<TranslationBank[i]->m_translation<<endl;
    newTranslationsFound += AddOneCandidate(sentenceNumber, TranslationBank[i], true);
    CNBestItem* foundItem = GetNBestItem(string((const char*)TranslationBank[i]->m_translation));
    //cout<<"Restore:"<<foundItem->getCandidateString();
  }//end of translation bank

  return newTranslationsFound;
}



// sv 051102 Decoder result should not be decoder spefic
//int CNBest::LoadFromBankUnique(int sentenceNumber, CArray<CDecoderNBResult*>& TranslationBank, CSentence sourceSentence)
int CNBest::LoadFromBankUnique(int sentenceNumber, CArray<CDecoderResult*>& TranslationBank, CSentence sourceSentence)
{
  m_sourceSentence = sourceSentence;
  // go thru each translation and initialize the counter data
  int newTranslationsFound = 0;
  for(int i=TranslationBank.MinIndex(); i<=TranslationBank.MaxIndex(); i++){
    //cout<<"Adding:"<<TranslationBank[i]->m_translation<<endl;
    newTranslationsFound += AddOneCandidate(sentenceNumber, TranslationBank[i], false);
    //cout<<"Restore:"<<foundItem->getCandidateString();
  }//end of translation bank
  return newTranslationsFound;
}


void CNBest::ClearAll(){

  for (int i=m_TransducerItems.MinIndex(); i<=m_TransducerItems.MaxIndex(); i++){
    CNBestItem* ptr = m_TransducerItems[i];
    delete ptr;
    ptr = NULL;
  }
  m_TransducerItems.Clear();
  m_EmissionTableSource.clear();
  m_TransitionTableSource.clear();
}


// sv 051102 Decoder result should not be decoder spefic
//int CNBest::AddOneCandidate(int sentenceNumber, CDecoderNBResult* result, bool saveDuplicates)
int CNBest::AddOneCandidate(int sentenceNumber, CDecoderResult* result, bool saveDuplicates)
{
  // assist in adding elements to the source table
  m_BuildMode = true;
  int Pos, NextPos;
  int Length;  // length of source = number of words
  std::string Word;
  CWordIndex WordIndex;
  CTransducerState State, NextState;

  //build up the transducer bank as well
  
  
  Length = 0;
  State = 0;
  Pos = 0;
  string sourceWord = string((const char*)(result)->m_translation);
  //cout<<"Source:"<<sourceWord<<":"<<endl;
  
  int newTranslationsFound =0;
    // process the features availble in this candidate
  const double negative = -1.0;
  //double* fb = new double[m_numFeatures];
  
  //ARV 04/25
  //tmpHolder's size is specified when creating the n-best
  //list in TranslateTS
  for (int m=0; m<result->m_features.size(); m++){
    tmpHolder[m] = negative*result->m_features[m];
  }

  //ARV 04/25 removed this to generalize to higher order models
  //tmpHolder[0] = negative*result->m_features[0];
  //tmpHolder[1] = negative*result->m_features[1];
  //tmpHolder[2] = negative*result->m_features[2];
  //tmpHolder[3] = negative*result->m_features[3];


  EdgeList transItems;
  
  //========================================================================
  // Add the transducer perceptron feature packs
  //========================================================================
  for (int i=result->m_edges.MinIndex(); i<=result->m_edges.MaxIndex(); i++){
    CLatticeEdgeTrans* edge = result->m_edges[i];
    //edge->Display();
    CWordInfo* wInfo = edge->WordInfo();
    
    if (wInfo->HasTargetString()){

      transItems.Append(wInfo);
    }

  }


  //cout<<"Scaling Factors Completed:"<<fb[0]<<" "<<fb[1]<<" "<<fb[2]<<" "<<fb[3]<<endl;

  bool addBestOnly = false;

  vector<CWordIndex> candidateVector;
  //insert this element into the transducer source mapping
  do{
    NextPos = sourceWord.find(' ', Pos);
    //cout<<"NextPost:"<<NextPos<<endl;
    Length++;
    if ( NextPos > 0 ) {
      Word = sourceWord.substr(Pos, NextPos-Pos);
      //cout<<"Word:"<<Word<<":"<<endl<<" voc:"<<m_pVoc<<endl;
      WordIndex = m_pVoc->GetIndex( CString( Word.c_str() ) );
      //cout<<"WordIndex:"<<WordIndex<<endl;

      if ( WordIndex == m_pVoc->UNK ) {
	std::cout << std::endl << std::flush;
	std::cout << "WARNING: Adding word:" << Word << ":to the vocab:"<<Word<<endl;
	std::cout << "WARNING: Unknown word:" << Word << ":in NBest List:"<<Word<<endl;
	std::cout << "WARNING: Sentence:" << sourceWord << ":in NBest List:"<<Word<<endl;
	WordIndex = m_pVoc->AddWord(Word);
	cout<<"NewWordIndex:"<<WordIndex<<endl;
      }
      candidateVector.push_back(WordIndex);

      if (NextSource( State, WordIndex, NextState ) == 0) {
	std::cout << "CTransducer::BuildSourceTables: Warning: NextSource() returned 0 (whatever this means)" << std::endl;
      }
      
    }else{
      // NextPos==-1, there is only one word left. Also, this state is a final state, we can attach emission table to it
      int l = sourceWord.size();
      Word = sourceWord.substr( Pos, l );
      //cout<<"Word:"<<Word<<":"<<endl;
      WordIndex = m_pVoc->GetIndex( CString( Word.c_str() ) );
      //cout<<"WordIndex:"<<WordIndex<<endl;

      if ( WordIndex == m_pVoc->UNK ) {
	std::cout << std::endl << std::flush;
	std::cout << "WARNING: Adding word:" << Word << ":to the vocab:"<<Word<<endl;
	std::cout << "ERROR: Unknown word " << Word << " in NBest:" <<Word<<endl;
	std::cout << "ERROR: Sentence:" << sourceWord << ":in NBest List:"<<Word<<endl;	
	WordIndex = m_pVoc->AddWord(Word);
	cout<<"NewWordIndex:"<<WordIndex<<endl;
      }
      candidateVector.push_back(WordIndex);

      // insert
      if ( NextSource( State, WordIndex, NextState ) == 0) {
	std::cout << "CTransducer::BuildSourceTables: Warning: NextSource() returned 0 (whatever this means)" << std::endl;
      } // endif: error (?)
      

      // if there is no data corresponding to this string
      if (m_EmissionTableSource.count(NextState)==0){
	//cout<<"New:"<<sourceWord<<endl;
	CNBestItem* oneItem = new CNBestItem();
	oneItem->setCandidateVector(candidateVector);
	// add feature set will return true if its a new feature vector
	//cout<<"Adding Initial Item"<<endl;
	if (oneItem->AddFeatureSet(tmpHolder)){
	  oneItem->AddEdgeList(transItems);
	}	

	
	m_TransducerItems.Append(oneItem); 
	m_EmissionTableSource[ NextState ] = oneItem;
	

	newTranslationsFound++;
      }else{
	//Duplicate translation
	
	  // add feature set will return true if its a new feature vector
	if (saveDuplicates){
	  if (m_EmissionTableSource[ NextState ]->AddFeatureSet(tmpHolder)){
	    m_EmissionTableSource[ NextState ]->AddEdgeList(transItems);
	  }
	}else{
	  //cout<<"Duplicate Translation in N-Best list"<<endl;
	}
	
      }
      // the entry has been properly inserted
      
    }//end operation in the case of 1 word left
    Pos = NextPos + 1;
    State = NextState;
    
  } while (NextPos >= 0);
  return newTranslationsFound;
}

CNBestItem* CNBest::GetNBestItem(string sourceWord){
  m_BuildMode = false;
    // assist in adding elements to the source table
  int Pos, NextPos;
  int Length;  // length of source = number of words
  std::string Word;
  CWordIndex WordIndex;
  CTransducerState State, NextState;

  //build up the transducer bank as well
  
  
  Length = 0;
  State = 0;
  Pos = 0;
  //finds the item 
  do{
    NextPos = sourceWord.find(' ', Pos);
    Length++;
    if ( NextPos > 0 ) {
      Word = sourceWord.substr(Pos, NextPos-Pos);
      //cout<<"Searching:"<<Word<<endl;
      WordIndex = m_pVoc->GetIndex( CString( Word.c_str() ) );
      //cout<<"WordIndex:"<<WordIndex<<endl;
      if ( WordIndex == m_pVoc->UNK ) {
	std::cout << std::endl << std::flush;
	std::cout << "ERROR: Unknown word:" << Word << ":in NBest List:"<<Word<<endl;
	return NULL;
      }else{
	
	if (NextSource( State, WordIndex, NextState ) == 0) {
	  std::cout << "CTransducer::BuildSourceTables: Warning: NextSource() returned 0 (whatever this means)" << std::endl;
	  return NULL;
	}
      }
      
    }else{
      // NextPos==-1, there is only one word left. Also, this state is a final state, we can attach emission table to it
      int l = sourceWord.size();
      Word = sourceWord.substr( Pos, l );
      WordIndex = m_pVoc->GetIndex( CString( Word.c_str() ) );
      //cout<<"SearchingLast:"<<Word<<endl;
      if ( WordIndex == m_pVoc->UNK ) {
	std::cout << std::endl << std::flush;
	std::cout << "ERROR: Unknown word " << Word << " in NBest:" <<Word<<endl;
	return NULL;
      }else {
	// insert
	if ( NextSource( State, WordIndex, NextState ) == 0) {
	  std::cout << "CTransducer::BuildSourceTables: Warning: NextSource() returned 0 -> unable to find the word" << std::endl;
	} // endif: error (?)
	
	return m_EmissionTableSource[ NextState ];
	// the entry has been properly inserted
      }//word is not unknown
    }//end operation in the case of 1 word left
    Pos = NextPos + 1;
    State = NextState;
    
  } while (NextPos >= 0);
  return NULL;
}

vector<Counter*> CNBest::getNumberCorrectList(){
  vector<Counter*> v;
  for (int i=m_TransducerItems.MinIndex(); i<=m_TransducerItems.MaxIndex(); i++){
    v.push_back(m_TransducerItems[i]->getNumberCorrect());
  }
  return v;
}
vector<Counter*> CNBest::getNumberSuggestedList(){
  vector<Counter*> v;
  for (int i=m_TransducerItems.MinIndex(); i<=m_TransducerItems.MaxIndex(); i++){
    v.push_back(m_TransducerItems[i]->getNumberSuggested());
  }
  return v;
}

vector<double*> CNBest::getReferenceLengthsList(){
  vector<double*> v;
  for (int i=m_TransducerItems.MinIndex(); i<=m_TransducerItems.MaxIndex(); i++){
    v.push_back(m_TransducerItems[i]->getClosestRefLength());
  }
  return v;
}

int CNBest::ConvertCandidateToBloc(int candidateNumber){
  
  int trackNumber =0;
  int blocNumber = 0;
  for (int i=m_TransducerItems.MinIndex(); i<=m_TransducerItems.MaxIndex(); i++){
    CNBestItem* pNBItem = m_TransducerItems[i];
    if (pNBItem->getFeatureSets()->size() + trackNumber > candidateNumber){
      return blocNumber;
    }
    trackNumber +=pNBItem->getFeatureSets()->size();
    blocNumber++;
  }
  return blocNumber;
}

void CNBest::PrintDiagnosticInformation(int sentenceNumber, std::ostream& os){
  os<<"Sentence:"<<sentenceNumber<<" UniqueTranslations:"<<m_TransducerItems.Size()<<" NumTotal:"<<getNumCands()<<endl;
}

// the candidate coming in is in bloc scale
void CNBest::PrintOptimizationData(const int sentenceNumber, const int NGramsConsidered,ostream& osFeats, ostream& osCands ){

  // just increments after each iteration
  int candidateNumber =0;
  osCands<<sentenceNumber<<" "<<getNumCands()<<endl;
  vector<int> proxCands;
  //int proxBloc = ConvertCandidateToBloc(proximateCandidate);
  for (int i=m_TransducerItems.MinIndex(); i<=m_TransducerItems.MaxIndex(); i++){
    int oldCandidate = candidateNumber;
    //os<<"Optimization Block:"<<i<<endl;
    //each transducer item is a unique sentence, so we just lookup the score from the proximity map
    m_TransducerItems[i]->PrintOptimizationData(sentenceNumber, NGramsConsidered, candidateNumber, osFeats);    
    
  }

}

void CNBest::PrintOptimizationDataWithSentence(const int sentenceNumber, const int NGramsConsidered, fstream& osFeats, fstream& osCands){
  // just increments after each iteration
  int candidateNumber =0;
  pair<int, int> candPair(sentenceNumber, getNumCands());
  osCands.write((char*)&candPair, sizeof(pair<int, int>));

  //osCands<<sentenceNumber<<" "<<getNumCands()<<endl;
  
  for (int i=m_TransducerItems.MinIndex(); i<=m_TransducerItems.MaxIndex(); i++){
    int oldCandidate = candidateNumber;
    //each transducer item is a unique sentence, so we just lookup the score from the proximity map
    m_TransducerItems[i]->PrintOptimizationDataWithSentence(sentenceNumber, NGramsConsidered, candidateNumber, m_globalScalingPtrs, osFeats);    
    
  }

}

void CNBest::PrintOptimizationDataWithSentenceAndSegmentation(const int sentenceNumber, const int NGramsConsidered, fstream& osFeats, fstream& osCands){
  // just increments after each iteration
  int candidateNumber =0;
  pair<int, int> candPair(sentenceNumber, getNumCands());
  osCands.write((char*)&candPair, sizeof(pair<int, int>));

  //osCands<<sentenceNumber<<" "<<getNumCands()<<endl;
  
  for (int i=m_TransducerItems.MinIndex(); i<=m_TransducerItems.MaxIndex(); i++){
    int oldCandidate = candidateNumber;
    //each transducer item is a unique sentence, so we just lookup the score from the proximity map
    m_TransducerItems[i]->PrintOptimizationDataWithSentenceAndSegmentation(sentenceNumber, NGramsConsidered, candidateNumber, m_globalScalingPtrs, osFeats);    
    
  }

}

int CNBest::getNumCands(){
  int numCands = 0;
  for (int i=m_TransducerItems.MinIndex(); i<=m_TransducerItems.MaxIndex(); i++){
    numCands += m_TransducerItems[i]->getFeatureSets()->size();
  }
  return numCands;
}


int
CNBest::NextSource( CTransducerState State, CWordIndex Word, 
		   CTransducerState &NextState )
{

  Trace( 1, "CTransducer::NextSource " << State << " " << Word );

  // if in BuildMode we make new entry into transition table
  // otherwise: error

  CStateWord StateWord( State, Word );

  std::map< CStateWord, CTransducerState, CompareStateWord >::const_iterator IterState;
  IterState = m_TransitionTableSource.find( StateWord );

  if ( IterState == m_TransitionTableSource.end() ) {
    if ( m_BuildMode ) {
      m_NumberOfStatesSource++;
      NextState = m_NumberOfStatesSource;
      m_TransitionTableSource[ StateWord ] = NextState;
      Trace( 1, "Insert " << State 
			 << " " << m_pVoc->GetWord( Word ) 
			 << " " << NextState );
	  
	} else {
	  std::cout << "ERROR: in CTransducer::Next" << std::endl
		    << "ERROR: State " << State << " is not in the transition table" << std::endl;
	  NextState = s_NotFound;
	  return 0;
	}
  }else {
    Trace(1,"Found " << m_pVoc->GetWord( IterState->first.m_WordIndex ) 
		   << " leading from state " << IterState->first.m_State 
	  << " to state " << " " << IterState->second);
	NextState = IterState->second;
  }

  return 1;
  
}

void CNBest::BuildWordPosteriors(){

  vector<double> allPathProbs;

  map<double, pair<int, int> > sortedCandidates;
  for (int i=m_TransducerItems.MinIndex(); i<=m_TransducerItems.MaxIndex(); i++){
    //each transducer item is a unique sentence, so we just lookup the score from the proximity map
    CNBestItem* oneUniqeWordSequence = m_TransducerItems[i];
    vector<double> onePathProbs = oneUniqeWordSequence->GenerateAllSequenceProbs(m_globalScalingPtrs);
    int initalSize = allPathProbs.size();
    for (int e=0; e<onePathProbs.size(); e++){
      allPathProbs.push_back(onePathProbs[e]);
      sortedCandidates.insert(pair<double, pair<int, int> >(onePathProbs[e], pair<int, int>(i, e)));
    }
  }

  //optional below
  /*
  double sum = 0.0;
  for (int i=0; i<=allPathProbs.size(); i++){
    sum += allPathProbs[i];
  }
  
  vector<double> finalProb;
  cout<<"FinalProb: ";
  for (int i=0; i<=allPathProbs.size(); i++){
    finalProb.push_back(allPathProbs[i]/sum);
    cout<<finalProb[i]<<" ";
  }
  */
  

  //CCompactTSequence<double> wordProbLookup;
  CCompactTSequence<double> wordProbLookupAlt;
  //wordProbLookup.Init(m_pVoc);
  wordProbLookupAlt.Init(m_pVoc);


  CCompactTSequence<double> conditionedSum;  
  conditionedSum.Init(m_pVoc);


  //accumulate the prob counts
  for (int i=m_TransducerItems.MinIndex(); i<=m_TransducerItems.MaxIndex(); i++){
  //for (int i=0; i<1; i++){
    CNBestItem* oneUniqeWordSequence = m_TransducerItems[i];
    //oneUniqeWordSequence->IncrementOccurrences(&wordProbLookup, finalProb[i]);
    oneUniqeWordSequence->IncrementOccurrencesNGram(&wordProbLookupAlt, allPathProbs[i], &conditionedSum);
  }

  wordProbLookupAlt.DisplayAllItems("EstiAlt:");
  //wordProbLookupAlt.DisplayEmmissionTable();

  conditionedSum.DisplayAllItems("Totals:");

  cout<<"Raw Estimation completed====================="<<endl;
  for (int i=wordProbLookupAlt.m_TransducerItems.MinIndex(); i<=wordProbLookupAlt.m_TransducerItems.MaxIndex(); i++){
    CCompactTSequenceItem<double>* wItem = wordProbLookupAlt.m_TransducerItems[i];
    //wItem->Display(m_pVoc);

    
    vector<int> fullSequence = wItem->GetSequence();
    vector<int> targetOnly = fullSequence;
    targetOnly.resize(wItem->GetKey());
    
    CCompactTSequenceItem<double>* targetSum = conditionedSum.GetCompactTSequence(targetOnly);
    if (targetSum==NULL){
      cerr<<"Unable to lookup value for targetSum"<<endl;
      wItem->Display();
      exit(-1);
    }

    double tgSum = targetSum->GetValue();
    //cout<<"Target:";
    //for (int t=0; t<targetOnly.size(); t++){
    //  cout<<" "<<targetOnly[t];
    //}
    //cout<<" Sum: "<<tgSum<<endl;

    wItem->SetValue(wItem->GetValue()/tgSum);

  }

  wordProbLookupAlt.DisplayAllItems("EstiNorm:");


  conditionedSum.ClearAll();


  //build a list of unique target words that are used in this sentence, represents possible confusions

  /*
  CCompactTSequence<double> targetLookup;
  targetLookup.Init(m_pVoc);

  multimap<int, int> sourceInvertedIndex;
  for (int i=wordProbLookupAlt.m_TransducerItems.MinIndex(); i<=wordProbLookupAlt.m_TransducerItems.MaxIndex(); i++){
    CCompactTSequenceItem<double>* pWordGivenTime = wordProbLookupAlt.m_TransducerItems[i];
    vector<CWordIndex> seq = pWordGivenTime->GetSequence();
    vector<CWordIndex> tgtOnly;
    tgtOnly.push_back(seq[1]);

    CCompactTSequenceItem<double>* existing = targetLookup.GetCompactTSequence(tgtOnly);
    if (existing==NULL){
      existing = targetLookup.InsertPhrasePair(tgtOnly, 1.0);
    }else{
      existing->SetValue(existing->GetValue()+1.0);
    }
    sourceInvertedIndex.insert(pair<int, int>(seq[0], seq[1]));
  }
  targetLookup.DisplayAllItems("Tgt:");
  int numTargetWords = targetLookup.m_TransducerItems.Size();
  //now go thru and assign all the words that were allocated to a particular position
  cout<<"numTargetWords: "<<numTargetWords<<endl;

  
  for (multimap<int, int>::iterator it = sourceInvertedIndex.begin();
       it != sourceInvertedIndex.end();
       ++it){
    cout << "  [" << (*it).first << ", " << (*it).second << "]" << endl;
  }

  */
  //build a multimap that indicates which target words each source word aligns to
  
  
  /*
  CCompactTSequence<double> positionLookup;
  positionLookup.Init(m_pVoc);
  for (int i=wordProbLookupAlt.m_TransducerItems.MinIndex(); i<=wordProbLookupAlt.m_TransducerItems.MaxIndex(); i++){
    CCompactTSequenceItem<double>* pWordGivenTime = wordProbLookupAlt.m_TransducerItems[i];
    vector<CWordIndex> seq = pWordGivenTime->GetSequence();
    vector<CWordIndex> srcOnly;
    srcOnly.push_back(seq[0]);
    CCompactTSequenceItem<double>* existing = positionLookup.GetCompactTSequence(srcOnly);
    if (existing==NULL){
      existing = positionLookup.InsertPhrasePair(srcOnly, pWordGivenTime->GetValue());
      existing->SetKey(1);
    }else{
      existing->SetValue(existing->GetValue() + pWordGivenTime->GetValue());
      existing->SetKey(existing->GetKey()+1);
    }
  }

  for (int i=positionLookup.m_TransducerItems.MinIndex(); i<=positionLookup.m_TransducerItems.MaxIndex(); i++){
    CCompactTSequenceItem<double>* timeSum = positionLookup.m_TransducerItems[i];
    //timeSum->SetValue(timeSum->GetValue()/(double)timeSum->GetKey());
    //timeSum->Display(m_pVoc);
    //cout<<"Increment: "<<
    timeSum->SetValue((timeSum->GetValue()+ (numTargetWords-timeSum->GetKey()))/numTargetWords);
  }

  

  //double timeNorm = positionLookup.L1();
  //positionLookup.DivideBy(timeNorm);
  positionLookup.DisplayAllItems("Time:");

  targetLookup.ClearAll();
  positionLookup.ClearAll();
  */

  //Display Confidence Scores for each path
  

  for (int i=m_TransducerItems.MinIndex(); i<=m_TransducerItems.MaxIndex(); i++){
    //each transducer item is a unique sentence, so we just lookup the score from the proximity map
    CNBestItem* oneUniqeWordSequence = m_TransducerItems[i];
    vector<double> onePathProbs = oneUniqeWordSequence->GenerateAllSequenceProbs(m_globalScalingPtrs);
    int initalSize = allPathProbs.size();
    for (int e=0; e<onePathProbs.size(); e++){
      double cscore = oneUniqeWordSequence->GetConfidenceScore(e, &wordProbLookupAlt);
      cout<<"Confidence: "<<cscore<<endl;
    }
  }
  wordProbLookupAlt.ClearAll();  
  
}

CCompactTSequence<int>* CNBestItem::CollectNGrams(vector<int> source, int maxNSize){


  CCompactTSequence<int>* pPhraseCollection = new CCompactTSequence<int>();
  pPhraseCollection->Init(m_pVoc);

  //list of sizes to match the n-gram collected 
  std::vector< pair<int, int> > gramSizes;

  int size = source.size();

  
  for (int n=1; n<=maxNSize; n++){

    // where to start searching from
    for (int i=0; i<=size-n; i++){

      //string accumulated = "";
      vector<int> sequence;
      for (int j=0; j<n; j++){
	//foundGram.push_back(source[i+j]);
	//accumulated = accumulated + source[i+j];
	sequence.push_back(source[i+j]);
	
	/*
	if (j+1<n){
	  accumulated += " ";
	}
	*/
      }
      pPhraseCollection->InsertPhrasePair(sequence, 1);
    }
  }
  return pPhraseCollection;
}

/*
void CNBest::RescoreAllTranslations(CLexicon* pLexicon){
  for (int i=m_TransducerItems.MinIndex(); i<=m_TransducerItems.MaxIndex(); i++){
    //each transducer item is a unique sentence, so we just lookup the score from the proximity map
    CNBestItem* oneUniqeWordSequence = m_TransducerItems[i];
    oneUniqeWordSequence->RescoreAllTranslations(pLexicon, m_sourceSentence);
  }
}
*/

//use the scaling factors specified here to select the model best index
//based on a call to each item to give its model best index, cost pair
/*
int GetModelBestIndex(vector<float>& scalingFactors){
  for (int i=m_TransducerItems.MinIndex(); i<=m_TransducerItems.MaxIndex(); i++){
    //each transducer item is a unique sentence, so we just lookup the score from the proximity map
    CNBestItem* oneUniqeWordSequence = m_TransducerItems[i];
    oneUniqeWordSequence->RescoreAllTranslations(pLexicon, m_sourceSentence);
  }
}
*/

int CNBest::GetModelBestIndex(vector<float>& scalingFactors){

  //keep track of the transducer the has the lowest cost
  
  int lowestCostIndex =0;
  double lowestCost = MAXBUF;
  
  for (int i=m_TransducerItems.MinIndex(); i<=m_TransducerItems.MaxIndex(); i++){
    //each transducer item is a unique sentence, so we just lookup the score from the proximity map
    CNBestItem* oneUniqeWordSequence = m_TransducerItems[i];
    double cost = oneUniqeWordSequence->GetMinCost(scalingFactors);
    if (cost<lowestCost){
      lowestCost = cost;
      lowestCostIndex = i;
    }
  }
  return lowestCostIndex;
}

