#ifdef HAVE_CONFIG_H
#include <config.h>
#endif



#define NO_TRACE 1
#include <Trace.hh>

#include <Decoder.hh>

long int CDecoder_Hyp::s_NumberOfObjects = 0;
long int CDecoder_BP::s_NumberOfObjects = 0;



CDecoder_BP::CDecoder_BP()
{
  //  cerr << "CDecoder_BP::CDecoder_BP" << endl;
  m_TotalScore = 0.0;

  m_TargetLength = 0;
  m_pHyp  = NULL;
  //  m_pEdge = NULL;
  m_pBP   = NULL;

  s_NumberOfObjects++;
}



CDecoder_BP::CDecoder_BP( int numModels )
{
  cout << " CDecoder_BP::CDecoder_BP( int numModels ) with numModels = " << numModels << endl;
  assert( 0 );

  m_TotalScore = 0.0;

  m_scaledModelScores.AssureBoundaries(0, numModels-1);
  m_scaledModelScores.SetAll(0.0);

  m_TargetLength = 0;
  m_pHyp  = NULL;
  //  m_pEdge = NULL;
  m_pBP   = NULL;

  s_NumberOfObjects++;
}



CDecoder_BP::CDecoder_BP( CDecoder_Hyp *pHyp )
{
  m_TotalScore = pHyp->m_TotalScore;
  m_scaledModelScores = pHyp->m_scaledModelScores;
  
  m_TargetLength = 0;
  m_pHyp  = pHyp;
  //m_pEdge = pHyp->m_pEdge;
  m_pBP   = NULL;

  s_NumberOfObjects++;
}



CDecoder_BP::CDecoder_BP( CDecoder_BP *pBP )
{
  m_TotalScore = pBP->m_TotalScore;
  m_scaledModelScores = pBP->m_scaledModelScores;

  m_TargetLength = pBP->m_TargetLength;
  m_pHyp       = pBP->m_pHyp;
  //m_pEdge      = pBP->m_pEdge;
  m_pBP        = pBP;

  s_NumberOfObjects++;
}



CDecoder_BP::~CDecoder_BP()
{
  s_NumberOfObjects--;
}






//=========================================================================
CDecoder_Hyp::CDecoder_Hyp()
{
  Trace( 3, "CDecoderNB_Hyp Constructor" );

  m_TotalScore = 0.0;
  s_NumberOfObjects++;
}


CDecoder_Hyp::CDecoder_Hyp( int numModels )
{
  Trace( 3, "CDecoderNB_Hyp Constructor" );

  m_scaledModelScores.AssureBoundaries(0, numModels-1);
  m_scaledModelScores.SetAll(0.0);
  m_TotalScore = 0.0;
  s_NumberOfObjects++;
}



CDecoder_Hyp::CDecoder_Hyp( const CDecoder_Hyp &Hyp )
{
  Trace( 3, "CDecoderNB_Hyp Copy Constructor" );

  m_scaledModelScores = Hyp.m_scaledModelScores;
  m_TotalScore = Hyp.m_TotalScore;
  s_NumberOfObjects++;
}


CDecoder_Hyp &
CDecoder_Hyp::operator = ( const CDecoder_Hyp &Hyp )
{
  if ( this == &Hyp ) 
    //-- Kopie auf sich selber kann ignoriert werden
    return *this;

  m_scaledModelScores = Hyp.m_scaledModelScores;
  m_TotalScore = Hyp.m_TotalScore;

  return *this;
}




CDecoder_Hyp::~CDecoder_Hyp()
{
  s_NumberOfObjects--;
}




//=========================================================================
CDecoderResult::CDecoderResult()
{
  cerr << "CDecoderResult::CDecoderResult default constructor" << endl;
  assert( 0 );
}



CDecoderResult::CDecoderResult( CString Translation, CDecoder_BP *pBP, CArray< CLatticeEdgeMS * > Edges)
{
  //cout << "CDecoderResult::CDecoderResult" << endl;
	m_Translation = Translation;
	m_TotalScore = pBP->m_TotalScore;
  m_pBP = pBP;

  //cout << "CDecoderResult::CDecoderResult - TotalScore = " << pBP->m_TotalScore << endl;
  //cout << "CDecoderResult::CDecoderResult - Translation = " << Translation << endl;


	assert( pBP->m_scaledModelScores.MaxIndex() >= pBP->m_scaledModelScores.MinIndex() );
	//cout << "copy features: minindex = " << pBP->m_scaledModelScores.MinIndex()
	// << " maxindex = " << m_pBP->m_scaledModelScores.MaxIndex() << endl;
	for ( int i = pBP->m_scaledModelScores.MinIndex(); i <= pBP->m_scaledModelScores.MaxIndex(); i++ ) {
	  //cout << i << " " << pBP->m_scaledModelScores[ i ] << endl;
	  m_Features.push_back( pBP->m_scaledModelScores[ i ] );
	}

  // we used the scaled features, keep track of this
  m_FeaturesAreScaled = True;

	m_Edges = Edges;
}



CDecoderResult::~CDecoderResult()
{
	m_Features.clear();
}




void
CDecoderResult::ScaleFeatures( vector< float > &ScalingFactors, bool Divide )
{
  if ( !m_FeaturesAreScaled && Divide ) {
    // don't want to divide raw features
    cout << "Features are unscaled, cannot divide" << endl;
    assert( 0 );
  }

  if ( m_FeaturesAreScaled && !Divide ) {
    // don't want to multiply scaled features again
    cout << "Features are scaled, do not scale them again" << endl;
    assert( 0 );
  }


	for ( int i = 0; i < ScalingFactors.size(); i++ ) {
		if ( Divide && ScalingFactors[ i ] != 0.0 ) {
		    m_Features[ i ] = m_Features[ i ] / ScalingFactors[ i ];
		} else {
			m_Features[ i ] = m_Features[ i ] * ScalingFactors[ i ];
		}
	}
}



CDecoderResult&
CDecoderResult::operator= ( const CDecoderResult &DecoderResult )
{
  //cout << "CDecoderResult::operator=" << endl;
  if ( this != &DecoderResult ) {
    this->m_TotalScore  = DecoderResult.m_TotalScore;
    this->m_Translation = DecoderResult.m_Translation;
    this->m_Features    = DecoderResult.m_Features;
    this->m_Edges       = DecoderResult.m_Edges;
  }
  return *this;

}




bool
CDecoderResult::operator< ( const CDecoderResult& arg )
{
	return this->m_TotalScore < arg.m_TotalScore;
}


//Nguyen, Aug 25, 2008, Cohesive decoding
//bool CovEdg::operator< ( CovEdg& b ) {
//	return ((this->cov < b.getcov()) ? true : ( this->cov==b.getcov() &&  ( this->edg < b.getedg() ) ? true : false)  );
//} 


// End of Nguyen

void
CDecoderResult::Display()
{
  cout << m_Translation << endl;

  cout << "TotalScore: " << m_TotalScore << " ";

  for ( int i = 0; i < m_Features.size(); i++ ) {
    cout << m_Features[ i ] << " ";
  }
  cout << endl;
}



void
CDecoderResult::Display( string Tag, int ID, vector< float > &ScalingFactors, CLatticeLM *Lattice )
{
  cout << Tag << " " << ID << " " << m_Translation << endl;

  cout << "TotalScore: " << m_TotalScore << " Scaled(Raw): ";

  for ( int i = 0; i < m_Features.size(); i++ ) {
    cout << m_Features[ i ] << "(" << m_Features[ i ] / ScalingFactors[ i ] << ") ";
  }
  cout << endl;

  cout << "display path: " << endl;
  if ( Lattice != NULL){
    Lattice->DisplayPath( m_Edges );
  }
}


std::pair< float, float > 
CDecoderResult::GetFeaturesMinMax()
{
  float minScore = 10e20;
  float maxScore = -10e20;

  if ( m_Features.empty() ) {
    return std::pair< float, float >( -10e20, 10e20 );
  }
      
  for ( vector< float >::iterator i_feature = m_Features.begin();
       i_feature != m_Features.end();
       i_feature++ )
  {
    if ( *i_feature > maxScore )
      maxScore = *i_feature;
    if ( *i_feature < minScore )
      minScore = *i_feature;
  }
  
  return std::pair< float, float >( minScore, maxScore );
}


void
CDecoderResult::Display( int SrcSentNum, int NBestNum, vector< float > &ScalingFactors, CLatticeLM *Lattice, bool DisplayPath )
{
  // nbest list is stored 0-based, but we want to output results 1-based
  NBestNum++;

  cout << CString( "TRANSLATION_RESULT_HYP    " ) << SrcSentNum << " " << NBestNum << " " << m_Translation << endl;
  cout << CString( "TRANSLATION_RESULT_SCORES " ) << SrcSentNum << " " << NBestNum << " TotalScore: " << m_TotalScore << " Scaled(Raw): ";

  for ( int i = 0; i < m_Features.size(); i++ ) {
    cout << m_Features[ i ] << "(" << m_Features[ i ] / ScalingFactors[ i ] << ") ";
  }
  cout << endl;

  if ( DisplayPath ) {
    cout << CString( "TRANSLATION_RESULT_ALIGN_BEGIN " ) << SrcSentNum << " " << NBestNum << endl;
    CString Tag = CString( "TRANSLATION_RESULT_ALIGN " ) + CString( SrcSentNum ) + CString( " " ) + CString( NBestNum );
    if ( Lattice != NULL ) {
      Lattice->DisplayPath( m_Edges, Tag );
    }
    cout << CString( "TRANSLATION_RESULT_ALIGN_END" ) << endl;
  }
}



void
CDecoderResult::GetTranslation( CSentence *pSentence, CVocabulary *pVoc )
{
  // not validated
  assert( 0 );


  // first clear the sentence
  pSentence->Clear();

  // loop over edges and loop over target words to build the sentence
  for ( int i = m_Edges.MinIndex(); i <= m_Edges.MaxIndex(); i++ ) {
    cout << i << " ";
    if ( m_Edges[ i ]->WordInfo() ) {
      CSentence PhraseTranslation = m_Edges[ i ]->WordInfo()->TargetWords();
      cout << "has wordinfo; target words = "; PhraseTranslation.Display( pVoc );
      cout << endl;
      pSentence->Append( PhraseTranslation );
      cout << "after appending: "; pSentence->Display( pVoc ); 
      cout << endl;
    } else {
      cout << "no wordinfo" << endl;
    }
  }

}



CDecoder::CDecoder()
{
  Trace( 0, "CDecoder::CDecoder" );
  //  m_pTranslationMemory = new CTranslationMemory;
  m_pTranslationMemory = NULL;
  m_pLattice = NULL;
  m_pLatticeEdgeTransPrototype = new CLatticeEdgeTrans;
  m_DisplayLattice = 0;

  m_StartLMScores = 0;
  m_StartDMScores = 1;
  m_StartWCScores = 2;
  m_StartPCScores = 3;
  
  //Nguyen, Aug 4, 2008: Cohesion Model
  m_StartPCScores = 4;
  
  m_StartSLScores = -1;  // currently not used
  m_StartTMScores = 5;
}


CDecoder::~CDecoder()
{
  //  delete m_pTranslationMemory;
  delete m_pLatticeEdgeTransPrototype;
}


/*
void
CDecoder::Init( CTranslationMemory *pTranslationMemory, CLM *pLM )
{
  m_pTranslationMemory = pTranslationMemory;
  m_pLM = pLM;
  //m_pTranslationMemory->RenormalizeTransducer();

  TraceIf( m_pLM == NULL, 0, "No LM given to decoder - that sucks" );
}


void
CDecoder::Init( const char *ParamFile, CVocabulary *pVocabulary )
{
  m_pTranslationMemory->Init( ParamFile, pVocabulary );
  m_pTranslationMemory->RenormalizeTransducer();
}
*/

/** 
 * @brief Find most significant high bit by searching from high to low-
 *        ordered bits.
 * @return 0-based index of most significant high bit in bitset, or -1
 *         when no bits are high.
 *
 * @author Anthony D'Auria (apd@cs.cmu.edu)
 *
 */
int
BitsetHighestOnBit( const bitset<COVERAGE_BIT_VECTOR_SIZE> &bs )
{
  for ( int i = bs.size()-1; i >= 0; --i )
  {
    if ( bs[i] == 1 ) {
      return i;
    }
  }

  return -1;
}

unsigned long
BitsetHash( const bitset<COVERAGE_BIT_VECTOR_SIZE> &bs )
{

  //unsigned long hashValue = 0;
  unsigned long hashValue = 0xAAAAAAAA;

  bool passedMSB = false;
  // start on the most significant end
  for ( int i = bs.size()-1; i >=0; --i )
  {
    if ( bs[i] == 1 ) {
      if ( ! passedMSB ) { passedMSB = true; }
    }

    // compute hash value if we are on or past the most significant bit
    if ( passedMSB == true ) {
      if ( (i & 1) == 0 ) {
        hashValue ^= (hashValue << 7) ^ bs[i] ^ (hashValue >> 3);
      } else {
        hashValue ^= ~( (hashValue << 11) ^ bs[i] ^ (hashValue >> 5) );
      }
    }
  } // bitset element loop

  return hashValue;
}

//Nguyen, Aug 14, 2008
// if InterruptionCheckAlgorithm, only need 1 more feature weight
void
CDecoder::SetStartPointsOfScalingFactorsWithCohesion()
{
  //cout << "CDecoder::SetDistortionModels - Number of loaded models = " << DistortionModels.Size() 
  //     << "  Number of total models " << (DistortionModels.Size() + 1) << endl;

  m_StartLMScores = m_ScalingFactors.MinIndex();

  m_StartDMScores = m_StartLMScores + m_LMs.Size();

  // we use a simple jump model as additional DM for smoothing
  // i.e. we have DistortionModels.Size() + 1 DMs
  m_StartWCScores = m_StartDMScores + 1 + m_DistortionModels.Size();

  // we currently have only 1 WordCount, 1 PhraseCount, 0 SentenceLength
  m_StartPCScores = m_StartWCScores + 1;
  
  //Nguyen, Aug 4, 2008: Cohesion Model  
  m_StartCMScores = m_StartPCScores + 1;
  
  m_StartSLScores = m_StartCMScores + 1;
  m_StartTMScores = m_StartSLScores + 0;  // as we don't use SL models right now

  cout << "Start LMs " << m_StartLMScores << endl
       << "Start DMs " << m_StartDMScores << endl
       << "Start WCs " << m_StartWCScores << endl
       << "Start PCs " << m_StartPCScores << endl
       << "Start CMs " << m_StartCMScores << endl
       << "Start SLs " << m_StartSLScores << endl
       << "Start TMs " << m_StartTMScores << endl;

}
// if InterruptionCheckAlgorithmCombination, need 4 more feature weights for
// 1) Interruption Check, 2) Interruption Count, 3)Noun Count, 4) VerbCount
void
CDecoder::SetStartPointsOfScalingFactorsWithCohesionCombination()
{
  //cout << "CDecoder::SetDistortionModels - Number of loaded models = " << DistortionModels.Size() 
  //     << "  Number of total models " << (DistortionModels.Size() + 1) << endl;

  m_StartLMScores = m_ScalingFactors.MinIndex();

  m_StartDMScores = m_StartLMScores + m_LMs.Size();

  // we use a simple jump model as additional DM for smoothing
  // i.e. we have DistortionModels.Size() + 1 DMs
  m_StartWCScores = m_StartDMScores + 1 + m_DistortionModels.Size();

  // we currently have only 1 WordCount, 1 PhraseCount, 0 SentenceLength
  m_StartPCScores = m_StartWCScores + 1;
  
  //Nguyen, Aug 4, 2008: Cohesion Model  
  m_StartCMScores = m_StartPCScores + 1; //Interruption Check
  m_StartICScores = m_StartCMScores + 1; //Interruption Count
  m_StartNCScores = m_StartICScores + 1; //Interruption Noun Count
  m_StartVCScores = m_StartNCScores + 1; //Interruption Verb Count
  
  m_StartSLScores = m_StartVCScores + 1;
  m_StartTMScores = m_StartSLScores + 0;  // as we don't use SL models right now

  cout << "Start LMs " << m_StartLMScores << endl
       << "Start DMs " << m_StartDMScores << endl
       << "Start WCs " << m_StartWCScores << endl
       << "Start PCs " << m_StartPCScores << endl
       
       << "Start CMs " << m_StartCMScores << endl
       << "Start ICs " << m_StartICScores << endl
       << "Start NCs " << m_StartNCScores << endl
       << "Start VCs " << m_StartVCScores << endl
              
       << "Start SLs " << m_StartSLScores << endl
       << "Start TMs " << m_StartTMScores << endl;

}
///// End of Nguyen /////

void
CDecoder::SetStartPointsOfScalingFactors()
{
  //cout << "CDecoder::SetDistortionModels - Number of loaded models = " << DistortionModels.Size() 
  //     << "  Number of total models " << (DistortionModels.Size() + 1) << endl;

  m_StartLMScores = m_ScalingFactors.MinIndex();

  m_StartDMScores = m_StartLMScores + m_LMs.Size();

  // we use a simple jump model as additional DM for smoothing
  // i.e. we have DistortionModels.Size() + 1 DMs
  m_StartWCScores = m_StartDMScores + 1 + m_DistortionModels.Size();

  // we currently have only 1 WordCount, 1 PhraseCount, 0 SentenceLength
  m_StartPCScores = m_StartWCScores + 1;  
  
  m_StartSLScores = m_StartPCScores + 1;
  m_StartTMScores = m_StartSLScores + 0;  // as we don't use SL models right now

  cout << "Start LMs " << m_StartLMScores << endl
       << "Start DMs " << m_StartDMScores << endl
       << "Start WCs " << m_StartWCScores << endl
       << "Start PCs " << m_StartPCScores << endl
       << "Start SLs " << m_StartSLScores << endl
       << "Start TMs " << m_StartTMScores << endl;

}

//! Get starting point of translation model scaling factors
//! \author apd@cs.cmu.edu 5/7/07
int
CDecoder::GetStartingPointTMScalingFactor() {
  return m_StartTMScores;
}

CString
CDecoder::TranslateSentence( const CSentence *pSentence )
{
  Trace( 0, "CDecoder::TranslateSentence" );
  return m_pTranslationMemory->TranslateSentence( pSentence );
}



CString
CDecoder::TranslateLattice( CLatticeLM *pLattice )
{
  Trace( 0, "CDecoder::TranslateLattice" );
  return m_pTranslationMemory->TranslateLattice( pLattice );
}


CLatticeLM *
CDecoder::SentenceToLattice( const CSentence *pSentence, const CVocabulary *pVocabulary )
{
  Trace( 1, "CDecoder::SentenceToLattice" );
  // Convert sentence into lattice

  // sv: modify for multiple LMs
  //m_pLattice = new CLatticeLM( m_pLatticeEdgeTransPrototype, m_pLM );
  m_pLattice = new CLatticeLM( m_pLatticeEdgeTransPrototype, m_LMs[ m_LMs.MinIndex() ] );

  CArray< CString > Words;
  m_pTranslationMemory->ConvertToLatticeRepresentation( pSentence, pVocabulary, Words );

  //for ( int i = Words.MinIndex(); i <= Words.MaxIndex(); i++ ) {
  //cout << Words[ i ] <<endl;
  //}

  //convert the source sentence into the first layer
  m_pLattice->ConvertFromText( Words );
  //if ( m_DisplayLattice ) {
  //m_pLattice->Display();
  //cout << endl;
  //}
  
  return m_pLattice;
}

