#ifndef Decoder_hh
#define Decoder_hh

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#ifndef COVERAGE_BIT_VECTOR_SIZE
#define COVERAGE_BIT_VECTOR_SIZE 32
#endif


#include <bitset>
using std::bitset;

//#include <Bool.hh>
#include <param.hh>

//Nguyen, Aug 7, 2008: Cohesive Decoding
#include <tokens.hh>


//#include <Lattice.hh>
//#include <LatticeEdgeTrans.hh>
//#include <LatticeLM.hh>

//#include <LMInterface.hh>
//#include <LMwordSA.hh>

#include <TransMem.hh>
#include <DistortionModel.hh>

class CDecoder_Hyp;
class CDecoder_BP;
class CDecoder_Info;
class CDecoder;



// allow multiple models of each kind
// LM model
/*
const int c_LM1 = 0;
const int c_LM2 = c_LM1 + 1;
const int c_LM3 = c_LM2 + 1;
const int c_LM4 = c_LM3 + 1;
const int c_LastLM = c_LM1;

// alignment or distortion model
const int c_AL1 = c_LastLM + 1;
const int c_AL2 = c_AL1 + 1;
const int c_AL3 = c_AL2 + 1;
const int c_AL4 = c_AL3 + 1;
const int c_LastAL = c_AL1;

// sentence length model
const int c_SL1 = c_LastAL + 1;
const int c_SL2 = c_SL1 + 1;
//const int c_SL3=11;
//const int c_SL4=12;
const int c_LastSL = c_SL1;

//counts the number of phrases used in the translation
const int c_PL1 = c_LastSL + 1;
const int c_LastPL = c_PL1;

//previously known as the first model index
const int c_TM1 = c_LastPL + 1;
*/


///////// Nguyen, Agust 25, 2008 - speedup cohesive decoding ///////////

//a look-up table to speed up interruption check algorithm
//key: Coverage and edge; 
//value: interruption score
class CovEdg{
	string cov; //binary coverage string
	string edg; //current phrase: word position
	
	public:
		CovEdg() { cov = ""; edg = ""; } 
		CovEdg( string a, string b ) { cov = a; edg = b; }
		
		string getcov() { return cov; }
		string getedg() { return edg; }
		
		void setcov (string a) { cov = a; }
		void setedg (string b) { edg = b; }		
		
		friend bool operator< ( const CovEdg& a, const CovEdg& b ) {
			return ((a.cov < b.cov) ? true : ( a.cov==b.cov &&  ( a.edg < b.edg )  ? true : false)  );
		}
		
};
/////// End of Nguyen ///////

// ------------------------------------------------------------------------------------------------------
// CDecoder_BP
// ------------------------------------------------------------------------------------------------------

class CDecoder_BP
{
public:
  CDecoder_BP();
  CDecoder_BP( int numModels );	
  CDecoder_BP( CDecoder_Hyp *pHyp );
  CDecoder_BP( CDecoder_BP *pBP );
  ~CDecoder_BP();

  float              m_TotalScore;      // accumulated total score

  CNumericalArray< float > m_scaledModelScores;
  
  int                m_TargetLength;    // true length of translation
  //  CLatticeEdgeMS *m_pEdge;           // pointer to edge which was traversed
  CDecoder_Hyp    *m_pHyp;            // pointer to hypothesis
  CDecoder_BP     *m_pBP;             // pointer to next BackPointer in the list

  static long int s_NumberOfObjects; 
};




// ------------------------------------------------------------------------------------------------------
// CDecoder_Hyp
// ------------------------------------------------------------------------------------------------------

class CDecoder_Hyp
{

public:
  CDecoder_Hyp();
  CDecoder_Hyp( int numModels );
  CDecoder_Hyp( const CDecoder_Hyp &Hyp );
  ~CDecoder_Hyp();
  CDecoder_Hyp & operator = ( const CDecoder_Hyp & );

  friend ostream &operator << (ostream &s, const CDecoder_Hyp & );

  CDecoder_Info *m_pDecoderInfo; // access to decoder info which can be different 
                                   // for different decoder objects
  CNumericalArray< float > m_scaledModelScores;
  float         m_TotalScore;      // accumulated total score

  // how many hyps exist, can be used to detect memory leaks
  static long int s_NumberOfObjects; 

  static bool s_TraceOn;
};




// ------------------------------------------------------------------------------------------------------
// CDecoderResult
// ------------------------------------------------------------------------------------------------------

//Represents the pair of a translation CString
// and a DecoderNB_Hyp object representing the final 
// scores of this translation
class CDecoderResult
{
public:

  CDecoderResult();

  //! create translation result from a backpointer (features are scaled scores)
	CDecoderResult( CString Translation, CDecoder_BP * pBP, CArray< CLatticeEdgeMS * > Edges );
  
  ~CDecoderResult();

  //! multiply or devide the feature values with the scaling factors
  // sv 060612: would be good to know when and where the features are in original or scaled form
  void ScaleFeatures( vector< float > &ScalingFactors, bool Divide );

  //! compare two translation results
  CDecoderResult& operator= ( const CDecoderResult& arg );

  // less then operator to put DecoderResult objects into sorted arrays
  bool operator< ( const CDecoderResult& arg );
  
  //! stream operator to output DecoderResult
  void Display();


  //! return translation as a CSentence object
  void GetTranslation( CSentence *pTranslation, CVocabulary *pVoc );

  //! return translation as a CString object
  const CString & Translation() const;

  //! return the total score
  float TotalScore() const;

  //! return the individual features

  //! return the path


  

  //! the translation as string
  CString m_Translation;

  //! the total score of the translation (scaled?)
  float m_TotalScore;

  //! the individual scores, could be scaled or unscaled
  vector< float > m_Features;

  //! the sequence of edges traversed in creating this translations
  CArray< CLatticeEdgeMS * > m_Edges;

  //! display a translation result
  void Display( string Tag, int ID, vector< float > &ScalingFactors, CLatticeLM * Lattice = NULL );
  void Display( int SrcSentNum, int NBestNum, vector< float > &ScalingFactors, CLatticeLM *Lattice, bool DisplayPath );


  //CArray< pair<double, CTransducerItem *> > m_edges;

  CDecoder_BP * m_pBP;


  
  bool
  MyLessThan(  CDecoderResult *pLeft, CDecoderResult *pRight ) {
    return pLeft->m_TotalScore < pRight->m_TotalScore;
  };
  

  //typedef pointer_to_binary_function< CDecoderResult *, CDecoderResult *, bool > DecoderResultBinaryFunctionType;

  /*
  struct CompareDecoderResultPtr 
  {
    bool operator()( CDecoderResult *pLeft, CDecoderResult *pRight ) {
      return pLeft->m_TotalScore < pRight->m_TotalScore;
    }
  }
  */

  /*!
    Finds smallest and largest feature function scores in m_Features.
   
   \return a pair of floats, where first is the minimum and second is maximum.
   
   \author Anthony D'Auria (apd@andrew.cmu.edu)
   */
  std::pair< float, float > GetFeaturesMinMax();
  
private:
  //! keep track if the features are scaled or not
  bool m_FeaturesAreScaled;

};



struct DecoderResultLess
{
  bool operator()( const CDecoderResult *pResult1, const CDecoderResult *pResult2 ) const
  {
    return pResult1->m_TotalScore < pResult2->m_TotalScore;
  }
};



// ------------------------------------------------------------------------------------------------------
// CDecoder
// ------------------------------------------------------------------------------------------------------

class CDecoder
{
public:
  CDecoder();
  ~CDecoder();
  
  // initialize the decoder 
  // sv 060708: Init functions replaced by set functions
  //void Init( CTranslationMemory *pTranslationMemory, CLM *pLM );
  //void Init( const char *ParamFileName, CVocabulary *pVocabulary );


  void SetVocabulary( CVocabulary *pVocabulary );

  void SetTranslationMemory( CTranslationMemory *pTranslationMemory );

  //void SetLanguageModel( CLM * pLanguageModel );
  void SetLanguageModels( CNumericalArray< CLM *> &LanguageModels );

  void SetDistortionModel( CDistortionModel *pDistortionModel );
  void SetDistortionModels( CNumericalArray< CDistortionModel *> &DistortionModels );

  void SetStartPointsOfScalingFactors();
  
  //Nguyen, Aug 14, 2008
  void SetStartPointsOfScalingFactorsWithCohesion();    
  void SetStartPointsOfScalingFactorsWithCohesionCombination();
  
  CString TranslateSentence( const CSentence *pSentence );
  CString TranslateLattice( CLatticeLM *pLattice );

  CLatticeLM * SentenceToLattice( const CSentence *pSentence, const CVocabulary *pVocabulary );

  //void DisplayPath( CArray< CLatticeEdge* > &Path );
  //void WritePath( ofstream &Stream, CArray< CLatticeEdge* > &Path );

  CLatticeLM *Lattice();

  //void SetDisplayLattice( int DisplayLattice = 1 );
  
  //! Returns starting point of translation model scaling factors
  int GetStartingPointTMScalingFactor();
  
  
    
protected:
  //void InitParams( const char *ParamFileName );
  
  //CString TargetString( CArray< CLatticeEdge* > &Path );
  //CString Translation( CArray< CLatticeEdge* > &Path );

  //void DisplayBestPath(  const CLattice *pLattice );  
  //void DisplayTargetString( const CLattice *pLattice );
  //void DisplayLatticeStatistics( const CLattice *pLattice );

  //void
  //ShortestPath( const CLattice *pLattice, CArray< CString > &Lattice );
  
  // member variables  
  CLatticeLM *m_pLattice;
  CLatticeEdgeTrans *m_pLatticeEdgeTransPrototype;
  
  
  //Nguyen, Aug 7, 2008: Cohesive decoding
  //store sentence dependency tree 
  vector<tokens> m_Dtree;
  vector<int> m_Hypcover;
  map <CovEdg, vector<int> > m_CovEdg;
  int m_InterruptionCheckAlgo;
  
  
  //CLatticeSearch m_LatticeSearch;

  CTranslationMemory *m_pTranslationMemory;

  //CLM *m_pLM;
  //CLMSentenceState *m_pLMSentenceState;

  CNumericalArray< CLM * >              m_LMs;
  CNumericalArray< CLMSentenceState * > m_LMSentenceStates;
  CNumericalArray< CLMState >           m_LMStates;


  CDistortionModel *m_pDistortionModel;
  CNumericalArray< CDistortionModel * > m_DistortionModels;

  //! the scaling factors for the different models
  CNumericalArray< float > m_ScalingFactors;

  // we need to keep track of the number of LMs, DMs to know where to
  // put the models scores into the feature vector
  int m_StartLMScores;
  int m_StartDMScores;
  int m_StartWCScores;
  int m_StartPCScores;
  
  //Nguyen, August 4, 2008: Cohesion Model
  int m_StartCMScores;
  int m_StartICScores;
  int m_StartNCScores;
  int m_StartVCScores;
  
  bool m_Cohesion;
  int m_InterruptionCount;
  int m_InterruptionNounCount;
  int m_InterruptionVerbCount;  
  
  
  int m_StartSLScores;
  int m_StartTMScores;

  //int m_NumOfNBest;

  //char *m_VocFile;
  CVocabulary *m_pVocabulary;
  
  //char *m_WordLM;
  // the name of the language model to be run
  //char *m_LMName;
  //int m_HistoryLength;

  //double m_LMScale;
  //double m_TMScale;

  //int m_LatticeStatistics;
  int m_DisplayLattice;

  //CParam m_Param;

};




//! return translation as a CString object
inline
const CString & 
CDecoderResult::Translation() const
{
  return m_Translation;
}



//! return the total score
inline
float
CDecoderResult::TotalScore() const
{
  return m_TotalScore;
}



// -----------------------------------------------------------------------------
// CDecoder inlines

inline
void
CDecoder::SetVocabulary( CVocabulary *pVocabulary )
{
  m_pVocabulary = pVocabulary;
}



inline
void
CDecoder::SetTranslationMemory( CTranslationMemory *pTranslationMemory )
{
  m_pTranslationMemory = pTranslationMemory;
}


/*
inline
void
CDecoder::SetLanguageModel( CLM * pLanguageModel )
{
  m_pLM = pLanguageModel;
}
*/


inline
void
CDecoder::SetLanguageModels( CNumericalArray< CLM *> &LanguageModels )
{
  cout << "CDecoder::SetLanguageModels - Number of models = " << LanguageModels.Size() << endl;
  m_StartLMScores = 0;
  m_LMs = LanguageModels;
}



//! set the distortion model
inline
void
CDecoder::SetDistortionModel( CDistortionModel *pDistortionModel )
{
  m_pDistortionModel = pDistortionModel;
}






inline
void
CDecoder::SetDistortionModels( CNumericalArray< CDistortionModel *> &DistortionModels )
{
  cout << "CDecoder::SetDistortionModels - Number of loaded models = " << DistortionModels.Size() << endl;
  //     << "  Number of total models " << (DistortionModels.Size() + 1) << endl;
  m_DistortionModels = DistortionModels;
}


unsigned long BitsetHash( const bitset<COVERAGE_BIT_VECTOR_SIZE> &bs );
int BitsetHighestOnBit( const bitset<COVERAGE_BIT_VECTOR_SIZE> &bs );
    

#endif
