#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <time.h>

#include <iostream>

//#define NO_TRACE 1
#undef NO_TRACE

#include <Trace.hh>
#include <Hash.hh>
#include <Array.hh>
#include <DecoderBLM.hh>
#include <OP_Generator.hh>

#include <time.h>
#include <utils.hh>

#include <stdio.h> 
#include <stdlib.h> 

using namespace std;


// to generate a uniq nbest list we first generate a larger non-uniq n-best list
// and then remove entries which have the same word sequence but lower overall score
// c_NumOfNBestMultiplier determines how big the non-uniq n-best list should be
short CDecoderBLM::c_NumOfNBestMultiplier = 5;

int   CDecoderBLM::m_NumberOfInstances = 0;
bool  CDecoderBLM::s_TraceOn = false;

//============================================================
CDecoderBLM::CDecoderBLM()
{
  TraceIf( s_TraceOn, 0, "CDecoderBLM::CDecoderBLM" );
   
  // default, no overlapping phrases
  m_AddOverlappingPhrases = false;

  // default, no reordering
  m_LengthOfReorderingWindow = 1;

  // default, 1-best decoding
  m_NumOfNBest = 1;

  // no per coverage pruning
  m_MaxHypsPerCoverage = 1000000;

  // no rescoring as default
  m_Rescore = 0;

  // virtually no beam size pruning
  m_BeamSize = 10000;

  m_MaxHyps = 1000;

  // for histogram pruning use 100 buckets
  m_PruningBuckets.SetBoundaries( 0, 100 );

  // is LM applied in batch mode
  m_LMBatchMode = false;
  

  //by default we start with 4 model scores
  m_NumberOfModels = 4;

  // future cost estimator
  m_pFutures = NULL;


  // the standard working hypothesis
  m_pNumberOneHypothesis = NewHypothesis();


  m_pDecoderInfo = new CDecoderBLM_Info;
  m_pNumberOneHypothesis->m_pDecoderInfo = m_pDecoderInfo;


  //  m_pNumberOneBestHyp = new CDecoderBLM_Hyp;
  //m_pNumberOneBestHyp->m_pDecoderInfo = m_pDecoderInfo;

  CDecoderBLM::m_NumberOfInstances = 1;

}



CDecoderBLM::CDecoderBLM( const CDecoderBLM &Decoder )
{
  TraceIf( s_TraceOn, 0, "CDecoderBLM::CDecoderBLM" );


  m_ScalingFactors = Decoder.m_ScalingFactors;

  //m_SLStretch = Decoder.m_SLStretch;
  //m_SLScale = Decoder.m_SLScale;

  m_BeamSize = Decoder.m_BeamSize;
  m_MaxHypsPerCoverage = Decoder.m_MaxHypsPerCoverage;
  m_MaxHyps = 1000;

  //cout<<"Coping BeamFactor:"<<m_BeamFactor<<endl;
  m_AddOverlappingPhrases = Decoder.m_AddOverlappingPhrases;

  // paremeters for reordering
  m_LengthOfReorderingWindow = Decoder.m_LengthOfReorderingWindow;

  // default, 1-best decoding
  m_NumOfNBest = Decoder.m_NumOfNBest;

  m_pDecoderInfo = new CDecoderBLM_Info;
  (*m_pDecoderInfo) = (*(Decoder.m_pDecoderInfo));

  // the standard working hypothesis
  //m_pNumberOneHypothesis = new CDecoderBLM_Hyp;
  m_pNumberOneHypothesis = NewHypothesis();
  m_pNumberOneHypothesis->m_pDecoderInfo = m_pDecoderInfo;
  //m_pNumberOneBestHyp = new CDecoderBLM_Hyp;
  //m_pNumberOneBestHyp->m_pDecoderInfo = m_pDecoderInfo;

  // copy translation memory
  m_pTranslationMemory = Decoder.m_pTranslationMemory;

  // sv: modify for multiple LMs
  //m_pLM = Decoder.m_pLM;
  m_LMs = Decoder.m_LMs;
  //m_pLMSentenceState = NULL;  // use own LM Sentence State

  CDecoderBLM::m_NumberOfInstances++;
}



CDecoderBLM::~CDecoderBLM()
{
  Clear();

  //cout<<"Not removing the last numberOneHypothesis in ~DecoderNB"<<endl;

  //if (m_numInstances==1){
  //delete m_pNumberOneHypothesis;
  //delete m_pNumberOneBestHyp;
  //}

  //cout<<"Remove decoder info"<<endl;
  delete m_pDecoderInfo;

  CDecoderBLM::m_NumberOfInstances--;
}



void
CDecoderBLM::InitModels( int NumberOfModels )
{
  cout << "CDecoderBLM::InitModels - Number of models is: " << NumberOfModels << endl << flush;
  assert( 0 );
  m_NumberOfModels = NumberOfModels;
  m_ScalingFactors.AssureBoundaries( 0, NumberOfModels - 1 );
}



/*
void
CDecoderBLM::Init( CTranslationMemory *pTranslationMemory, CLM *pLM )
{
  m_pTranslationMemory = pTranslationMemory;
  m_pLM = pLM;
  //m_pTranslationMemory->RenormalizeTransducer();
}



void
CDecoderBLM::Init( CTranslationMemory *pTranslationMemory, CLM *pLM, CVocabulary *pVocabulary )
{
  m_pTranslationMemory = pTranslationMemory;
  m_pLM = pLM;
  m_pVocabulary = pVocabulary;
  //m_pTranslationMemory->RenormalizeTransducer();
}



void
CDecoderBLM::Init( const char *ParamFile, CVocabulary *pVocabulary )
{
  m_pVocabulary = pVocabulary;
  m_pTranslationMemory->Init( ParamFile, pVocabulary );
  m_pTranslationMemory->RenormalizeTransducer();
  m_pLM = m_pTranslationMemory->GetLM();
}
*/



// takes a HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual >





void
CDecoderBLM::Clear()
{
  HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual >::iterator IterHyps;
  HypContainer< CDecoderBLM_Hyp *, HypHashForPruning, HypEqualForPruning >::iterator IterBestHyps;

  CDecoderBLM_Hyp *pHyp;

  // delete hyps for each coverage
  //vector<RecombinedMap*> allHarvestedAllHyps;

  for ( int Coverage = m_AllHyps.MinIndex(); Coverage <=  m_AllHyps.MaxIndex(); Coverage++ ) {

    HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual > &Hyps = m_AllHyps[ Coverage ];

    int NumHyp = 0;
    for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {
      pHyp = *IterHyps;
      //ARV
      // This section was used to harvest out the addresses of the recombined map
      /*
        vector<RecombinedMap*> harvestedRecombines = HarvestFromHyp(pHyp);
        for (vector<RecombinedMap*>::iterator recIter = harvestedRecombines.begin(); recIter!= harvestedRecombines.end(); ++recIter){

        unsigned int memLoc = (unsigned int)(*recIter);
        cout<<"Harvested: "<<memLoc<<endl;
        //printf("Harvested:%d\n", (*recIter));
        allHarvestedAllHyps.push_back(*recIter);
        }
      */
      m_HypsForDeleting.SetAt( NumHyp, pHyp );
      NumHyp++;
    }
	
    for ( int i = 0; i < NumHyp; i++ ) {
      delete m_HypsForDeleting[ i ];
    }

    Hyps.clear();
	

    // the best hyps used for pruning
    HypContainer< CDecoderBLM_Hyp *, HypHashForPruning, HypEqualForPruning > &BestHyps = m_BestHyps[ Coverage ];

    NumHyp = 0;
	
    for ( IterBestHyps = BestHyps.begin(); IterBestHyps != BestHyps.end(); ++IterBestHyps ) {
      pHyp = *IterBestHyps;
      m_HypsForDeleting.SetAt( NumHyp, pHyp );
      NumHyp++;
    }
	
    for ( int i = 0; i < NumHyp; i++ ) {
      delete m_HypsForDeleting[ i ];
    }

    BestHyps.clear();

  }

  // remove the stuff in the nbest list
  multimap< float, CDecoderBLM_BP * >::iterator IterOverBPs;
  for ( int j = m_NBestHyps.MinIndex(); j <= m_NBestHyps.MaxIndex(); j++ ) {
    multimap< float, CDecoderBLM_BP * > &CurrentBPs = m_NBestHyps[ j ];

    // loop through list of backpointers up to N
    for ( IterOverBPs = CurrentBPs.begin(); IterOverBPs != CurrentBPs.end(); ++IterOverBPs ) {
      CDecoderBLM_BP* pBP = IterOverBPs->second;
      delete pBP;
      // ARV
      // Moving the erase till after its all cleared up
      //CurrentBPs.erase( IterOverBPs );
    }
    CurrentBPs.clear();
  }


  // and the hyps in the final list and the final rescored list
  multimap< float, CDecoderBLM_Hyp * >::iterator Iter = m_FinalHyps.begin();

  int NumHyp = 0;
  for ( Iter = m_FinalHyps.begin(); Iter != m_FinalHyps.end(); ++Iter ) {
    pHyp = Iter->second;
    m_HypsForDeleting.SetAt( NumHyp, pHyp );
    NumHyp++;
  }

  for ( Iter = m_FinalHypsRescored.begin(); Iter != m_FinalHypsRescored.end(); ++Iter ) {
    pHyp = Iter->second;
    m_HypsForDeleting.SetAt( NumHyp, pHyp );
    NumHyp++;
  }

  for ( int i = 0; i < NumHyp; i++ ) {
  	delete m_HypsForDeleting[ i ];
  }

  m_FinalHyps.clear();
  m_FinalHypsRescored.clear();
  
  // remove all the recombined maps that were not deleted when the backpointers were deleted
  for (vector<RecombinedMap*>::iterator recIter = m_RecombinedMapList.begin(); recIter != m_RecombinedMapList.end(); ++recIter){
    RecombinedMap* pRemoveMe = (*recIter);
    //cout<<"ToRemove: "<<(unsigned int)pRemoveMe<<endl;
    delete pRemoveMe;
  }
  m_RecombinedMapList.clear();


  //remove the new initial sentence state created for this sentence
  //by calling NewSentence

  // sv: modify for multiple LMs
  /*
  if ( m_pLMSentenceState != NULL ){
    delete m_pLMSentenceState;
    m_pLMSentenceState = NULL;
  }
  */

  for ( int i = m_LMSentenceStates.MinIndex(); i <= m_LMSentenceStates.MaxIndex(); i++ ) {
    delete m_LMSentenceStates[ i ];
    m_LMSentenceStates[ i ] = NULL;
  }
  m_LMSentenceStates.RemoveAll();

  
  //cout << "Stored final hyps is " << m_FinalHyps.size() << endl;
  //cout << "all deleted" << endl << flush;

}



CString
CDecoderBLM::TranslateSentence( const CSentence *pSentence )
{
  TraceIf( s_TraceOn, 0, "CDecoderBLM::TranslateSentence" );
  return m_pTranslationMemory->TranslateSentence( pSentence );
}



// sv 051102 Decoder result should not be decoder spefic
//CArray<CDecoderBLMResult*>
CArray<CDecoderResult*>
CDecoderBLM::TranslateLattice( CLatticeLM *pLattice )
{
  s_TraceOn = false;
  TraceIf( s_TraceOn, 0, "CDecoderBLM::TranslateLattice" );
  //return m_pTranslationMemory->TranslateLattice( pLattice );


  m_pTranslationMemory->BuildTranslationLattice( pLattice );
  //cout << "Lattice after transducers applied" << endl;

  //pLattice->Display();
  // insert overlapping phrases
  //cout << "overlapping phrases" << endl;

  if ( m_AddOverlappingPhrases ) {
    COverlappingPhraseGenerator *pOPGenerator = new COverlappingPhraseGenerator;
    pOPGenerator->AddOverlappingPhrases( pLattice );
    delete pOPGenerator;
  }

  m_pLattice = pLattice;

  // here we do the new decoding
  // delete all hypotheses from previous round of decoding

  //CArray<CString> retList = ApplyLanguageModel( );
  return ApplyLanguageModel( );
}





void
CDecoderBLM::BuildTranslationLattice( CLatticeLM *pLattice )
{
  TraceIf( 1, 0, "CDecoderBLM::BuildTranslationLattice" );
  m_pTranslationMemory->BuildTranslationLattice( pLattice );

}



CArray< CDecoderResult* >
CDecoderBLM::ApplyLanguageModel( CLatticeLM *pLattice, CLM *pLM )
{
  TraceIf( s_TraceOn, 0, "CDecoderBLM::ApplyLanguageModel" );

  // lattice and LM are needed elsewhere, so assign them to member variable
  m_pLattice = pLattice;
  
  // sv: modify to allow multiple LMs
  //m_pLM = pLM;
  m_LMs[ m_LMs.MinIndex() ] = pLM;
  
  m_Cohesion = false;

  return ApplyLanguageModel();
}



// this is the function currently used in TranslateOptimize and TranslateBatchLM
CArray< CDecoderResult* >
CDecoderBLM::ApplyLanguageModel( CLatticeLM *pLattice )
{
  TraceIf( s_TraceOn, 0, "CDecoderBLM::ApplyLanguageModel" );

  // lattice needed elsewhere, so assign them to member variable
  m_pLattice = pLattice;
  m_Cohesion = false;

  return ApplyLanguageModel();
}

////////////////////////////////////////////////////////////
//Nguyen, Aug 7, 2008: Cohesive Decoding

CArray< CDecoderResult* >
CDecoderBLM::ApplyLanguageModelWithDependencyTree( CLatticeLM *pLattice, vector<tokens>& sentDtree, int algorithm )
{
  TraceIf( s_TraceOn, 0, "CDecoderBLM::ApplyLanguageModelWithDependencyTree" );

  // lattice needed elsewhere, so assign them to member variable
  m_pLattice = pLattice;
  
  m_Dtree    = sentDtree; 
  
  m_Cohesion = true;
  
  m_InterruptionCheckAlgo = algorithm; 
  cout << "Interruption Check Algorithm: " << m_InterruptionCheckAlgo << endl;
  
  string src = getNameFromVector(sentDtree);
  for (int i = 0; i< src.length(); ++i){
  	if ( isspace(src[i]) ) m_Hypcover.push_back(0); 
  }

  return ApplyLanguageModelWithDependencyTree();
}

CArray< CDecoderResult* >
CDecoderBLM::ApplyLanguageModelWithDependencyTree()
{
  s_TraceOn = false;

  TraceIf( s_TraceOn, 0, "CDecoderBLM::ApplyLanguageModelWithDependencyTree" );
  TraceIf( s_TraceOn, 0, "Length of reordering window is " << m_LengthOfReorderingWindow );
  
  cout << "DEBUG - Starting CDecoderBLM::ApplyLanguageModelWithDependencyTree()" << endl; 
  //cout << "DEBUG - getNameFromVector(m_Dtree): " <<  getNameFromVector(m_Dtree) << endl;
    
  // internal statistics
  m_HypsExpanded = 0;
  m_HypsNotExpanded = 0;
  m_HypsOutOfBeam = 0;
  m_HypsOutOfBeamPerCov = 0;
  m_NumNewHyps = 0;
  m_NumBetterHyps = 0;
  m_NumWorseHyps = 0;
  m_NumCoverageCollisions = 0;

  // first convert all string at edges in lattice 
  // into word indices according to LM vocabulary
  GetLMWordIndicesAndTMScores();
  
  // Nguyen, Cohesive decoding, Aug 29, 2008
  // clean map table
  m_CovEdg.clear();
  
  // get the LMSentenceState and the LMState for sentence start
  // LM state
  // sv: modify to allow multiple LMs
  //CLMState LMState = (CLMState) 0;
  //CLMState NewLMState = (CLMState) 0;
  CNumericalArray< CLMState > LMStates;
  CNumericalArray< CLMState > NewLMState;

  /*  
  assert( m_pLM );
  if ( m_pLM ) {
    m_pLMSentenceState = m_pLM->NewSentence();
    LMState = m_pLMSentenceState->BeginOfSentenceState();
    if (s_TraceOn){
      cout << "LMState for begin of sentence is " << LMState <<endl;
    }
  }
  */

  int NumLM = m_LMs.MinIndex();
  while ( NumLM <= m_LMs.MaxIndex()) {
    m_LMSentenceStates[ NumLM ] = m_LMs[ NumLM ]->NewSentence();
    LMStates[ NumLM ] = m_LMSentenceStates[ NumLM ]->BeginOfSentenceState();
    if (s_TraceOn){
      cout << "LMState for begin of sentence is " << LMStates[ NumLM ] <<endl;
    }

    NumLM++;
  }

  // how long is the source sentence?  we need to cover all words
  int Start = m_pLattice->StartNode()->Number();
  int J = m_pLattice->EndNode()->Number();
  if (s_TraceOn){
    cout << "StartNode " << Start << " EndNode " << J << endl;
    cout << "Sentence length = " << J << endl;
  }
  // set boundaries for arrays; Coverage from 0 to sentence length plus 1
  m_AllHyps.SetBoundaries( 0, J - Start );
  m_BestHyps.SetBoundaries( 0, J - Start );
  

  // should we use average path length for speech recognizer lattices?

  // based on sentence length we set maximum of allowed reorderings
  // as 'a b c d' -> 'a c b d' will count as two reorderings, we
  // multiply by extra factor of 2
  m_pDecoderInfo->m_MaxNumReorderings = (int) (J * 2 * m_ReorderingsPerWord);
  
  // array of multi-hash-sets is dynamic, i.e. when array grows, 
  // items are copied to new memory space; this causes addresses 
  // to individual hash-sets to become invalid
  // therefore, loop over entire range and access hash-set
  // this will create empty hash sets and avoid later copying
  // if a previous sentence has been processed, the hyps are
  // still stored, we clear them
  HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual >::iterator IterHyps;
  CDecoderBLM_Hyp *pHyp, *pBestHyp;
  
  for ( int Coverage = 0; Coverage <= J -Start; Coverage++ ) {
    m_AllHyps[ Coverage ].clear();
    m_BestHyps[ Coverage ].clear();
  }

  // we start with an empty hypothesis
  //pHyp = new CDecoderBLM_Hyp;
  pHyp = NewHypothesis();
  pHyp->m_pDecoderInfo = m_pDecoderInfo;
  pHyp->m_LMStates = LMStates;
  //cout << "Initial Hypothesis: " << *pHyp << endl;

  // and place it into the list of all hypotheses with coverage 0
  assert( pHyp->m_Coverage == 0 );
  m_AllHyps[ 0 ].insert( pHyp );
  //cout << "Initial Hypothesis after storing: " << *pHyp << endl;

  // also best hyp for this coverage
  //pBestHyp = new CDecoderBLM_Hyp;
  pBestHyp = NewHypothesis();
  pBestHyp->m_pDecoderInfo = m_pDecoderInfo;
  pBestHyp->m_LMStates = LMStates;
  assert( pBestHyp->m_Coverage == 0 );
  m_BestHyps[ 0 ].insert( pBestHyp );

  // hypotheses are organized according to their coverage
  // hypotheses are expanded till full coverage J is reached
  // i.e. we expand all hyps with Covereage < J
  for ( int Coverage = 0; Coverage < J - Start; Coverage++ ) {

    // display all hyps with current coverage
    //DisplayAllHypsWithCoverage( Coverage );

    if ( m_LMBatchMode ) {
      m_TmpHyps.RemoveAll();
    }
    
    //PrintTime( "Start of  ExpandAllHypsWithCoverage()" );

    // expand all hyps (does skip of the GetLMScore when batch mode)
    this->ExpandAllHypsWithCoverage( Coverage );
    
    //PrintTime( "End of  ExpandAllHypsWithCoverage()" );

    if ( m_LMBatchMode ) {
      // hyps are so far only in a temporary storage
      // apply now the LM in batch mode
      // then insert into the correct bins

      // in batch mode we get all the LM log probs in one sweep
      // this then needs also to do the pruning
      //this->GetLMLogProbsAllHyps( Coverage );
       this->ApplyLM_BatchMode( Coverage );

      // store now those hyps from the temporary storage
      // this does the recombination and pruning
      //StoreHyps_BatchMode()
    }


    //for ( int Coverage = 0; Coverage <= J; Coverage++ ) {
    //cout << "Stored hyps for coverage " << Coverage << " is " 
    //<< m_AllHyps[ Coverage ].size() << endl;
    //}
  }


  // display all hyps with full coverage
  //DisplayAllHypsWithCoverage( J, 1 );

  // do specific sentence end stuff
  this->ExpandAllHypsSentenceEnd();

  this->DisplayDecoderStatistics();


  // reconstruct the n-best translations using the backpointer information
  //CString Translation = BestTranslation();
  TraceIf( s_TraceOn, 0, "\nDisplay n-best hypotheses with backtrace for n-best = " << m_NumOfNBest );  
  TraceIf( 1, 0, "\nNumber of final hypotheses is: " << m_FinalHyps.size() );

  cout << "NBest = " << m_NumOfNBest << endl;
  
  BuildNBestList();
  


  // do we have all the information

  // if you are tracing
  s_TraceOn = false;

  if (s_TraceOn) {
    TraceIf( s_TraceOn, 0, "After BuildNBestList  Final List of Backpointers " );

    multimap< float, CDecoderBLM_BP * > &CurrentBPs = m_NBestHyps[ 0 ];
    multimap< float, CDecoderBLM_BP * >::iterator IterOverBPs;
    int n = 0;
    for ( IterOverBPs = CurrentBPs.begin(); 
          IterOverBPs != CurrentBPs.end();
          ++IterOverBPs ) {
      
      CDecoderBLM_BP *pBP = IterOverBPs->second;
      
      TraceIf( 1, 0, "\nBackpointer " << n << " " << IterOverBPs->first
               << " TotalScore=" << pBP->m_TotalScore );
      
      //cout << endl << "NBest " << n << endl;
      TraceStringAndPath( pBP, n );
      
      n++;
    }
    
    
    cout << "Number of true N-best translations is: " << n << endl;
    
    Trace( 1, "Number of BPs is " << CDecoderBLM_BP::s_NumberOfObjects++ );
  }




  s_TraceOn = false;

  // extract the actual nbest list from the member variable
  int n = 0;

  // sv 051102 Decoder result should not be decoder spefic
  //  CArray<CDecoderBLMResult*> nBestTranslations;
  CArray< CDecoderResult* > NBestTranslations;
  multimap< float, CDecoderBLM_BP * > &CurrentBPs = m_NBestHyps[ 0 ];
  multimap< float, CDecoderBLM_BP * >::iterator IterOverBPs;
 
  hash_set< CString, StringHasher, StringComparer > NBestHashSet;


  // here, we restrict to m_NumOfNBest uniq translations
  for ( IterOverBPs = CurrentBPs.begin(); 
        (IterOverBPs != CurrentBPs.end()) && (n < m_NumOfNBest);
        ++IterOverBPs ) {
    
    CDecoderBLM_BP *pBP = IterOverBPs->second;
    
    TraceIf( false, 0, "\nBackpointerFinal " << n << " " << IterOverBPs->first
             << " TotalScore=" << pBP->m_TotalScore );

    //cout << endl << "NBest " << n << endl;

    pair< CString, CArray< CLatticeEdgeMS * > > StringAndPath = TraceStringAndPath( pBP, n );
    CString OneTranslation = StringAndPath.first;
    CArray< CLatticeEdgeMS * > EdgesOnPath = StringAndPath.second;
    
    // check if this string already exists
    if ( NBestHashSet.count( OneTranslation ) > 0 ) {
      TraceIf( false, 0, "Old entry in nbest list " << OneTranslation );
    } else {
      TraceIf( false, 0, "New entry in nbest list " << OneTranslation );
      NBestHashSet.insert( OneTranslation );

      //cout << pBP->m_TotalScore << endl;
      CDecoderResult *DecoderResult = new CDecoderResult( OneTranslation, pBP, EdgesOnPath );
      NBestTranslations[ n ] = DecoderResult;
      //cout << pBP->m_TotalScore << endl;
      NBestTranslations[ n ]->m_TotalScore = pBP->m_TotalScore;
      //cout << "Totalscore = " << NBestTranslations[ n ]->m_TotalScore << endl;

      n++;
    }

  }

  cout << "Unique translations added to n-best list = " << n
       << "  size of NBestTranslations = " << NBestTranslations.Size() << endl;


  s_TraceOn = false;

  return NBestTranslations;
}

////////////////////////End of Nguyen ///////////////////////////////

CArray< CDecoderResult* >
CDecoderBLM::ApplyLanguageModel()
{
  s_TraceOn = false;

  TraceIf( s_TraceOn, 0, "CDecoderBLM::ApplyLanguageModel" );

  TraceIf( s_TraceOn, 0, "Length of reordering window is " << m_LengthOfReorderingWindow );
  
  // internal statistics
  m_HypsExpanded = 0;
  m_HypsNotExpanded = 0;
  m_HypsOutOfBeam = 0;
  m_HypsOutOfBeamPerCov = 0;
  m_NumNewHyps = 0;
  m_NumBetterHyps = 0;
  m_NumWorseHyps = 0;
  m_NumCoverageCollisions = 0;

  // first convert all string at edges in lattice 
  // into word indices according to LM vocabulary
  GetLMWordIndicesAndTMScores();

  // get the LMSentenceState and the LMState for sentence start
  // LM state
  // sv: modify to allow multiple LMs
  //CLMState LMState = (CLMState) 0;
  //CLMState NewLMState = (CLMState) 0;
  CNumericalArray< CLMState > LMStates;
  CNumericalArray< CLMState > NewLMState;

  /*  
  assert( m_pLM );
  if ( m_pLM ) {
    m_pLMSentenceState = m_pLM->NewSentence();
    LMState = m_pLMSentenceState->BeginOfSentenceState();
    if (s_TraceOn){
      cout << "LMState for begin of sentence is " << LMState <<endl;
    }
  }
  */

  int NumLM = m_LMs.MinIndex();
  while ( NumLM <= m_LMs.MaxIndex()) {
    m_LMSentenceStates[ NumLM ] = m_LMs[ NumLM ]->NewSentence();
    LMStates[ NumLM ] = m_LMSentenceStates[ NumLM ]->BeginOfSentenceState();
    if (s_TraceOn){
      cout << "LMState for begin of sentence is " << LMStates[ NumLM ] <<endl;
    }

    NumLM++;
  }



  // how long is the source sentence?  we need to cover all words
  int Start = m_pLattice->StartNode()->Number();
  int J = m_pLattice->EndNode()->Number();
  if (s_TraceOn){
    cout << "StartNode " << Start << " EndNode " << J << endl;
    cout << "Sentence length = " << J << endl;
  }
  // set boundaries for arrays; Coverage from 0 to sentence length plus 1
  m_AllHyps.SetBoundaries( 0, J - Start );
  m_BestHyps.SetBoundaries( 0, J - Start );
  

  // should we use average path length for speech recognizer lattices?

  // based on sentence length we set maximum of allowed reorderings
  // as 'a b c d' -> 'a c b d' will count as two reorderings, we
  // multiply by extra factor of 2
  m_pDecoderInfo->m_MaxNumReorderings = (int) (J * 2 * m_ReorderingsPerWord);
  
  // array of multi-hash-sets is dynamic, i.e. when array grows, 
  // items are copied to new memory space; this causes addresses 
  // to individual hash-sets to become invalid
  // therefore, loop over entire range and access hash-set
  // this will create empty hash sets and avoid later copying
  // if a previous sentence has been processed, the hyps are
  // still stored, we clear them
  HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual >::iterator IterHyps;
  CDecoderBLM_Hyp *pHyp, *pBestHyp;
  
  for ( int Coverage = 0; Coverage <= J -Start; Coverage++ ) {
    m_AllHyps[ Coverage ].clear();
    m_BestHyps[ Coverage ].clear();
  }


  // we start with an empty hypothesis
  //pHyp = new CDecoderBLM_Hyp;
  pHyp = NewHypothesis();
  pHyp->m_pDecoderInfo = m_pDecoderInfo;
  pHyp->m_LMStates = LMStates;
  //cout << "Initial Hypothesis: " << *pHyp << endl;

  // and place it into the list of all hypotheses with coverage 0
  assert( pHyp->m_Coverage == 0 );
  m_AllHyps[ 0 ].insert( pHyp );
  //cout << "Initial Hypothesis after storing: " << *pHyp << endl;

  // also best hyp for this coverage
  //pBestHyp = new CDecoderBLM_Hyp;
  pBestHyp = NewHypothesis();
  pBestHyp->m_pDecoderInfo = m_pDecoderInfo;
  pBestHyp->m_LMStates = LMStates;
  assert( pBestHyp->m_Coverage == 0 );
  m_BestHyps[ 0 ].insert( pBestHyp );


  // hypotheses are organized according to their coverage
  // hypotheses are expanded till full coverage J is reached
  // i.e. we expand all hyps with Covereage < J
  for ( int Coverage = 0; Coverage < J - Start; Coverage++ ) {

    // display all hyps with current coverage
    //DisplayAllHypsWithCoverage( Coverage );

    if ( m_LMBatchMode ) {
      m_TmpHyps.RemoveAll();
    }

    // expand all hyps (does skip of the GetLMScore when batch mode)
    this->ExpandAllHypsWithCoverage( Coverage );

    if ( m_LMBatchMode ) {
      // hyps are so far only in a temporary storage
      // apply now the LM in batch mode
      // then insert into the correct bins

      // in batch mode we get all the LM log probs in one sweep
      // this then needs also to do the pruning
      //this->GetLMLogProbsAllHyps( Coverage );
       this->ApplyLM_BatchMode( Coverage );

      // store now those hyps from the temporary storage
      // this does the recombination and pruning
      //StoreHyps_BatchMode()
    }


    //for ( int Coverage = 0; Coverage <= J; Coverage++ ) {
    //cout << "Stored hyps for coverage " << Coverage << " is " 
    //<< m_AllHyps[ Coverage ].size() << endl;
    //}
  }


  // display all hyps with full coverage
  //DisplayAllHypsWithCoverage( J, 1 );

  // do specific sentence end stuff
  this->ExpandAllHypsSentenceEnd();

  this->DisplayDecoderStatistics();


  // reconstruct the n-best translations using the backpointer information
  //CString Translation = BestTranslation();
  TraceIf( s_TraceOn, 0, "\nDisplay n-best hypotheses with backtrace for n-best = " << m_NumOfNBest );  
  TraceIf( 1, 0, "\nNumber of final hypotheses is: " << m_FinalHyps.size() );

  cout << "NBest = " << m_NumOfNBest << endl;
  
  BuildNBestList();
  


  // do we have all the information

  // if you are tracing
  s_TraceOn = false;

  if (s_TraceOn) {
    TraceIf( s_TraceOn, 0, "After BuildNBestList  Final List of Backpointers " );

    multimap< float, CDecoderBLM_BP * > &CurrentBPs = m_NBestHyps[ 0 ];
    multimap< float, CDecoderBLM_BP * >::iterator IterOverBPs;
    int n = 0;
    for ( IterOverBPs = CurrentBPs.begin(); 
          IterOverBPs != CurrentBPs.end();
          ++IterOverBPs ) {
      
      CDecoderBLM_BP *pBP = IterOverBPs->second;
      
      TraceIf( 1, 0, "\nBackpointer " << n << " " << IterOverBPs->first
               << " TotalScore=" << pBP->m_TotalScore );
      
      //cout << endl << "NBest " << n << endl;
      TraceStringAndPath( pBP, n );
      
      n++;
    }    
    cout << "Number of true N-best translations is: " << n << endl;
    
    Trace( 1, "Number of BPs is " << CDecoderBLM_BP::s_NumberOfObjects++ );
  }
  
  s_TraceOn = false;

  // extract the actual nbest list from the member variable
  int n = 0;

  // sv 051102 Decoder result should not be decoder spefic
  //  CArray<CDecoderBLMResult*> nBestTranslations;
  CArray< CDecoderResult* > NBestTranslations;
  multimap< float, CDecoderBLM_BP * > &CurrentBPs = m_NBestHyps[ 0 ];
  multimap< float, CDecoderBLM_BP * >::iterator IterOverBPs;
 
  hash_set< CString, StringHasher, StringComparer > NBestHashSet;


  // here, we restrict to m_NumOfNBest uniq translations
  for ( IterOverBPs = CurrentBPs.begin(); 
        (IterOverBPs != CurrentBPs.end()) && (n < m_NumOfNBest);
        ++IterOverBPs ) {
    
    CDecoderBLM_BP *pBP = IterOverBPs->second;
    
    TraceIf( false, 0, "\nBackpointerFinal " << n << " " << IterOverBPs->first
             << " TotalScore=" << pBP->m_TotalScore );

    //cout << endl << "NBest " << n << endl;

    pair< CString, CArray< CLatticeEdgeMS * > > StringAndPath = TraceStringAndPath( pBP, n );
    CString OneTranslation = StringAndPath.first;
    CArray< CLatticeEdgeMS * > EdgesOnPath = StringAndPath.second;
    
    // check if this string already exists
    if ( NBestHashSet.count( OneTranslation ) > 0 ) {
      TraceIf( false, 0, "Old entry in nbest list " << OneTranslation );
    } else {
      TraceIf( false, 0, "New entry in nbest list " << OneTranslation );
      NBestHashSet.insert( OneTranslation );

      //cout << pBP->m_TotalScore << endl;
      CDecoderResult *DecoderResult = new CDecoderResult( OneTranslation, pBP, EdgesOnPath );
      NBestTranslations[ n ] = DecoderResult;
      //cout << pBP->m_TotalScore << endl;
      NBestTranslations[ n ]->m_TotalScore = pBP->m_TotalScore;
      //cout << "Totalscore = " << NBestTranslations[ n ]->m_TotalScore << endl;

      n++;
    }
  }

  cout << "Unique translations added to n-best list = " << n
       << "  size of NBestTranslations = " << NBestTranslations.Size() << endl;


  s_TraceOn = false;

  return NBestTranslations;
}




void
CDecoderBLM::GetLMWordIndicesAndTMScores()
{
  /*
    The LM has its own vocabulary, therefore we get the LM word IDs for the words at the edges
    to avoid repeated string passing and string->int convertion on the LM side.
    For edges carrying source words we just get the LM word ID.
    For edges resulting from transducers or online phrase alignment we distinguish
    1. if label is one of those which should be treated as normal word then we just get the
       LM word ID of that label
    2. if label is a 'normal' label then we go into the wordinfo structure and get the 
       LM word IDs for all the target words (if those don't exist the source words are used)
  */

  cout << "CDecoderBLM::GetLMWordIndicesAndTMScores " << endl;


  CArray< CLatticeEdge * > Edges = m_pLattice->Edges();

  // array to collect words attached to one edge for LM-scoring
  CArray< CString > Words;
  CArray< LMWordIndex > WordIndices;
  CNumericalArray< LMWordIndex > WordIndicesForLMs;

  // first get the words
  CString Word;
  string word;
  CLatticeEdgeMS *pEdge;
  
  for ( int i = Edges.MinIndex(); i <= Edges.MaxIndex(); i++ ) {
    pEdge = (CLatticeEdgeMS *)Edges[ i ];
	
    Word = pEdge->Word();
    word = (const char*)Word;
    CWordInfo *pWordInfo =  ((CLatticeEdgeMS *)pEdge)->WordInfo();
	
    Trace( 2, "WordInfo: " << *pWordInfo );
    Trace( 2, "SourceString: " << pWordInfo->SourceString() );
    Trace( 2, "TargetString: " << pWordInfo->TargetString() );
	

    /*
    // first, we make sure that we have sufficient scores (not needed for source words)
    if ( ( Word[ 0 ] == '@' ) && ( m_ScalingFactors.Size() > c_TM1 + ((CLatticeEdgeMS*)pEdge)->Scores().Size() ) ) {

      // if we have not enough socres, we supply additional scores
      cout << "Mismatch in number of models and scaling factors: ScalingFactorsSize: " << m_ScalingFactors.Size()
           << "  c_TM1 = " << c_TM1 
           << "  and number of TM scores = " << ((CLatticeEdgeMS*)pEdge)->Scores().Size() << endl;
      ((CLatticeEdgeMS*)pEdge)->Display();
      cout << endl;
      
      while ( m_ScalingFactors.Size() > c_TM1 + ((CLatticeEdgeMS*)pEdge)->Scores().Size() ) {
        //cout << ((CLatticeEdgeMS*)pEdge)->Scores().Size() << " " << m_ScalingFactors.Size() << endl;
        ((CLatticeEdgeMS*)pEdge)->Scores().Append( 0.0 );
      }
      
      cout << "Now, we should have sufficient scores: ";
      ((CLatticeEdgeMS*)pEdge)->Display();
      cout << endl;
    }
    */

    // if edge was not generated by a transducer item
    // then set TM score to TMScoreForSourceWord
    if ( Word[ 0 ] != '@' ) {
      pEdge->WordInfo()->Score() = c_TMScoreForSourceWord;
      TraceIf( s_TraceOn, 0, "      Translation Score for source language edge set to : " 
               << c_TMScoreForSourceWord );
    }


    // check if label is one of the special tags, which the LM should treat as normal word
    if ( LabelIsNormalWord( Word ) ) {
      cout << "Treat " << Word << " as normal word" << endl;
      Words.SetAt( 1, Word );
    } else {

      // extract sequence of words from the WordInfo structure
      if ( pWordInfo->HasTargetString() ) {
        pWordInfo->GetTargetWords( Words );
      } else if ( pWordInfo->HasSourceString() ) {
        pWordInfo->GetSourceWords( Words );
      } else {
        Word = pEdge->Word();
        Words.SetAt( 1, Word );
      }
    }

    // convert strings into indices according to te LM vocabulary
    /*
    for ( int t = Words.MinIndex(); t <= Words.MaxIndex(); t++ ) {
      //LMWordIndex WordIndex = m_pLM->GetVocabularyIndex( (string)((const char *)(Words[ t ])) );
      // we have multiple LMs, but assume that all share the same vocabulary
      LMWordIndex WordIndex = m_LMs[ m_LMs.MinIndex() ]->GetVocabularyIndex( (string)((const char *)(Words[ t ])) );
      pEdge->LMWordIndices()[ t ] = WordIndex;
      TraceIf( s_TraceOn, 0, "Word = " << Words[ t ] << "  LM-WordIndex = " << WordIndex );
    }
    */


    // convert strings into indices according to te LM vocabulary
    for ( int t = Words.MinIndex(); t <= Words.MaxIndex(); t++ ) {
      //LMWordIndex WordIndex = m_pLM->GetVocabularyIndex( (string)((const char *)(Words[ t ])) );
      // we have multiple LMs which may have different vocabularies
      for ( int l = m_LMs.MinIndex(); l <= m_LMs.MaxIndex(); l++ ) {
        LMWordIndex WordIndex = m_LMs[ l ]->GetVocabularyIndex( (string)((const char *)(Words[ t ])) );
        CNumericalMatrix< long > &LMWordIndices = pEdge->WordIndicesForLMs();
        LMWordIndices( l, t ) = WordIndex;
        TraceIf( s_TraceOn, 0, "Word = " << Words[ t ] << "  LM-WordIndex = " << WordIndex );
      }
    }
    
  }
  //m_pLattice->DisplayEdges();
  //cout << endl;
}

bool
CDecoderBLM::LabelIsNormalWord( const CString &Word )
{
  for ( int i = m_TagsConsideredAsWords.MinIndex(); i <= m_TagsConsideredAsWords.MaxIndex(); i++ ) {
    if ( Word == m_TagsConsideredAsWords[ i ] ) {
      cout << Word << " is considered as normal word" << endl;
      return True;
    }
  }

  return False;

}



void
CDecoderBLM::ExpandAllHypsWithCoverage( int Coverage )
{
  ExpandAllHypsWithCoverageBeamSizePruning( Coverage );
}





void
CDecoderBLM::ExpandAllHypsWithCoverageBeamWidthPruning( int Coverage )
{
  s_TraceOn = False;

  TraceIf( s_TraceOn, 0, "\n\n  CDecoderBLM::ExpandAllHypsWithCoverage " << Coverage );

  // some variables we need
  CDecoderBLM_Hyp *pHyp, *pBestHyp;
  CLatticeNode *pNode;
  CLatticeEdgeMS *pEdge;


  // how long is the source sentence?  we need to cover all words
  int J = m_pLattice->EndNode()->Number();
  // should we use average path length for speech recognizer lattices?


  // hypotheses are organized according to their coverage
  // hypotheses of coverage c are expanded till coverage J-1
  // is reached.  Expanding those hypotheses will result in 
  // full coverage

  // first we fetch the set of hypotheses which all have coverage c
  // and we need an iterater to access the hyps in this set
  HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual > &Hyps = m_AllHyps[ Coverage ];
  HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual >::iterator IterHyps;


  // for pruning we need the best hypothesis with the current coverage
  HypContainer< CDecoderBLM_Hyp *, HypHashForPruning, HypEqualForPruning > &BestHyps = m_BestHyps[ Coverage ];
  HypContainer< CDecoderBLM_Hyp *, HypHashForPruning, HypEqualForPruning >::iterator IterBestHyps;
  
  //HistogramPruning( Hyps, BestHyps );
  //  TraceIf( s_TraceOn, 0, "After Histogram Pruning:  Number of hyps with coverage " 
  //		   << Coverage << " is " << Hyps.size() );
  

  // sort hyps by score for beamsize pruning
  typedef multimap< float, CDecoderBLM_Hyp * > CSortedHypContainer;

  // associate expansion candidates with their best-hyp-for-pruning
  typedef map< CDecoderBLM_Hyp *, CSortedHypContainer > CBestSimHyp2AllSimHypsMap;

  CBestSimHyp2AllSimHypsMap HypsByBestHyp;
		   
  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {
    pHyp = *IterHyps;

    // only expand if hyp had survived histogram pruning
    if ( pHyp->m_Expand == False ) {
      
      //cout << "Prune hyp (histogram) " << pHyp << endl;
      //	  m_HypsPruned++;
      
    } else {
      
      // pruning: only expand hyp if it is better than BeamFactor times 
      // score of best hypothesis
      
      IterBestHyps = BestHyps.find( pHyp );
      if ( IterBestHyps == BestHyps.end() ) {
        //cout <<"Coverage: "<<Coverage<< "no best hyp found" << endl;
        pBestHyp->m_TotalScore = 10000;
      } else {
        pBestHyp = *IterBestHyps;
        //cout <<"Coverage: "<<Coverage<< "Best Hyp is " << pBestHyp << endl;
        TraceHyp( s_TraceOn, "\n  best hyp to prune against is ", pBestHyp );
      }


      if ( m_pFutures != NULL ) {
        cout << "Using Future Costs "<<endl;
	    
        /*
          CArray< CLatticeEdgeMS * > transEdges = TraceBackWithPath(pHyp).second;
          for (int i=transEdges.MinIndex(); i<=transEdges.MaxIndex(); i++){
          transEdges[i]->Display();
          }
        */
	    
        double totalFutureCost = 0.0;
        CArray<std::pair<int,int> >untranslatedSegments = LeftToTranslate(pHyp);

        for (int i=untranslatedSegments.MinIndex(); i<=untranslatedSegments.MaxIndex(); i++){
          //cout<<"UT: I:"<<untranslatedSegments[i].first<<" J: "<<untranslatedSegments[i].second; 
          double futureCostEstimate = m_pFutures->GetFutureCostEstimate(untranslatedSegments[i].first, untranslatedSegments[i].second);
          totalFutureCost += futureCostEstimate;
        }
        //out<<endl;
	    
        //cout<<"BaseCost: "<<pHyp->m_TotalScore<<" additional: "<<totalFutureCost<<endl;

        //use the future cost along with the baseline cost to hash 
        HypsByBestHyp[pBestHyp].insert(CSortedHypContainer::value_type(pHyp->m_TotalScore+totalFutureCost,pHyp));

      } else {

        // if there is no future cost estimate, we just use the sum
        // assert: we have pBestHyp here
        // associate hyp with its best-hyp-for-pruning
        HypsByBestHyp[ pBestHyp ].insert( CSortedHypContainer::value_type( pHyp->m_TotalScore, pHyp ) );

      }

    } // endif: histogram pruned

  } // andfor: all recombined hyps

  
  // foreach set of recombined hyps associated to their best hyp, expand beamsize best ones
  int rank;
  CSortedHypContainer::iterator IterSortedHyp;
  CSortedHypContainer HypsInBeam;

  for (CBestSimHyp2AllSimHypsMap::iterator IterHypsByBestHyp = HypsByBestHyp.begin();
       IterHypsByBestHyp != HypsByBestHyp.end();
       ++IterHypsByBestHyp) {
    // sorted list of expansion candidates associated with current best hyp for pruning
    CDecoderBLM_Hyp* pBestHyp = (*IterHypsByBestHyp).first;

    if (s_TraceOn){
      cout<<"Coverage: "<<Coverage<<endl;
      cout<<"pBestHyp:  ";
      //display the best hyp
      CArray< CLatticeEdgeMS * > transEdges = TraceBackWithPath(pBestHyp).second;
      for (int i=transEdges.MinIndex(); i<=transEdges.MaxIndex(); i++){
        transEdges[i]->Display();
        cout<<" ";
      }
      cout<<endl<<endl;
    }

    CSortedHypContainer &SimilarHyps = (*IterHypsByBestHyp).second;
    
    if (s_TraceOn){
      //lets look at all hyps in beam first
      
      for (IterSortedHyp  = SimilarHyps.begin(); IterSortedHyp != SimilarHyps.end();  ++IterSortedHyp){
        CDecoderBLM_Hyp* pHyp = (*IterSortedHyp).second;
        //display the hyp
        CArray< CLatticeEdgeMS * > transEdges = TraceBackWithPath(pHyp).second;
        for (int i=transEdges.MinIndex(); i<=transEdges.MaxIndex(); i++){
          transEdges[i]->Display();
          cout<<" ";
        }
        cout<<endl;
        cout<<"FScore: "<<(*IterSortedHyp).first<<" base: "<<pHyp->m_TotalScore<<endl;
        
      }
    }
    

    // only take the beamsize best
    for(IterSortedHyp = SimilarHyps.begin(), rank = 0;
        IterSortedHyp != SimilarHyps.end() && ++rank <= m_BeamSize;
        ++IterSortedHyp) {
      
      HypsInBeam.insert(*IterSortedHyp);
    } // endfor

    m_HypsOutOfBeam += ( SimilarHyps.size() > m_BeamSize ?
                         SimilarHyps.size() - m_BeamSize :
                         0 );
  } // endfor


  // out of the hyps just collected, only take the m_MaxHypsPerCoverage best
  for (IterSortedHyp  = HypsInBeam.begin(), rank = 0;
       IterSortedHyp != HypsInBeam.end() && ++rank <= m_MaxHypsPerCoverage;
       ++IterSortedHyp)
    ExpandHyp((*IterSortedHyp).second);
  
  m_HypsOutOfBeamPerCov += ( HypsInBeam.size() > m_MaxHypsPerCoverage ?
                             HypsInBeam.size() - m_MaxHypsPerCoverage :
                             0 );
  
  s_TraceOn = false;

} // CDecoderBLM::ExpandAllHypsWithCoverage(...)



void
CDecoderBLM::ExpandAllHypsWithCoverageBeamSizePruning( int Coverage )
{
  s_TraceOn = False;

  TraceIf( true, 2, "\n\n  CDecoderBLM::ExpandAllHypsWithCoverageBeamSizePruning " << Coverage );
  
  //cout << "DEBUG - Starting CDecoderBLM::ExpandAllHypsWithCoverageBeamSizePruning" << endl;

  // some variables we need
  CDecoderBLM_Hyp *pHyp, *pBestHyp;
  CLatticeNode *pNode;
  CLatticeEdgeMS *pEdge;


  // how long is the source sentence?  we need to cover all words
  int J = m_pLattice->EndNode()->Number();
  // should we use average path length for speech recognizer lattices?


  // hypotheses are organized according to their coverage
  // hypotheses of coverage c are expanded till coverage J-1
  // is reached.  Expanding those hypotheses will result in 
  // full coverage

  // first we fetch the set of hypotheses which all have coverage c
  // and we need an iterater to access the hyps in this set
  HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual > &Hyps = m_AllHyps[ Coverage ];
  HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual >::iterator IterHyps;


  // for pruning we need the best hypothesis with the current coverage
  HypContainer< CDecoderBLM_Hyp *, HypHashForPruning, HypEqualForPruning > &BestHyps = m_BestHyps[ Coverage ];
  HypContainer< CDecoderBLM_Hyp *, HypHashForPruning, HypEqualForPruning >::iterator IterBestHyps;

  //HistogramPruning( Hyps, BestHyps );
  //  TraceIf( s_TraceOn, 0, "After Histogram Pruning:  Number of hyps with coverage " 
  //		   << Coverage << " is " << Hyps.size() );


  // sort hyps by score for beamsize pruning
  typedef multimap< float, CDecoderBLM_Hyp * > CSortedHypContainer;

  // associate expansion candidates with their best-hyp-for-pruning
  typedef map< CDecoderBLM_Hyp *, CSortedHypContainer > CBestSimHyp2AllSimHypsMap;

  CBestSimHyp2AllSimHypsMap HypsByBestHyp;

  // loop over all hyps
  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {
    pHyp = *IterHyps;
    //cout << "Hyp is " << pHyp << endl;
    TraceHyp( s_TraceOn, 2, "\n  in ExpandAllHypsWithCoverage - expand next hyp ", pHyp );

    // only expand if hyp had survived histogram pruning
    if ( pHyp->m_Expand == False ) {
      
      //cout << "Prune hyp (histogram) " << pHyp << endl;
      //	  m_HypsPruned++;
      
    } else {
      
      // find best hyp for pruning: only expand hyp if it is better than BeamFactor times score of best hypothesis
      IterBestHyps = BestHyps.find( pHyp );
      if ( IterBestHyps == BestHyps.end() ) {
        //cout <<"Coverage: "<<Coverage<< "no best hyp found" << endl;
        pBestHyp->m_TotalScore = 10000;
      } else {
        pBestHyp = *IterBestHyps;
        //cout <<"Coverage: "<<Coverage<< "Best Hyp is " << pBestHyp << endl;
        TraceHyp( s_TraceOn, 0, "\n  best hyp to prune against is ", pBestHyp );
      }


      if ( m_pFutures != NULL ) {
        cout << "Using Future Costs "<<endl;
        /*
          CArray< CLatticeEdgeMS * > transEdges = TraceBackWithPath(pHyp).second;
          for (int i=transEdges.MinIndex(); i<=transEdges.MaxIndex(); i++){
          transEdges[i]->Display();
          }
        */
        double totalFutureCost = 0.0;
        CArray<std::pair<int,int> >untranslatedSegments = LeftToTranslate(pHyp);
        for (int i=untranslatedSegments.MinIndex(); i<=untranslatedSegments.MaxIndex(); i++){
          //cout<<"UT: I:"<<untranslatedSegments[i].first<<" J: "<<untranslatedSegments[i].second; 
          double futureCostEstimate = m_pFutures->GetFutureCostEstimate(untranslatedSegments[i].first, untranslatedSegments[i].second);
          totalFutureCost += futureCostEstimate;
        }
        //out<<endl;
        //cout<<"BaseCost: "<<pHyp->m_TotalScore<<" additional: "<<totalFutureCost<<endl;
        //use the future cost along with the baseline cost to hash 
        HypsByBestHyp[pBestHyp].insert(CSortedHypContainer::value_type(pHyp->m_TotalScore+totalFutureCost,pHyp));

      } else {

        // if there is no future cost estimate, we just use the sum
        // assert: we have pBestHyp here
        // associate hyp with its best-hyp-for-pruning
        HypsByBestHyp[ pBestHyp ].insert( CSortedHypContainer::value_type( pHyp->m_TotalScore, pHyp ) );

      }

    } // endif: histogram pruned

  } // andfor: all recombined hyps

  
  // foreach set of recombined hyps associated to their best hyp, expand beamsize best ones
  int rank;
  CSortedHypContainer::iterator IterSortedHyp;
  CSortedHypContainer HypsInBeam;

  for ( CBestSimHyp2AllSimHypsMap::iterator IterHypsByBestHyp = HypsByBestHyp.begin();
        IterHypsByBestHyp != HypsByBestHyp.end();
        ++IterHypsByBestHyp) {
    // sorted list of expansion candidates associated with current best hyp for pruning
    CDecoderBLM_Hyp* pBestHyp = (*IterHypsByBestHyp).first;

    if (s_TraceOn){
      cout<<"Coverage: "<<Coverage<<endl;
      cout<<"pBestHyp:  ";
      //display the best hyp
      CArray< CLatticeEdgeMS * > transEdges = TraceBackWithPath(pBestHyp).second;
      for (int i=transEdges.MinIndex(); i<=transEdges.MaxIndex(); i++){
        transEdges[i]->Display();
        cout<<" ";
      }
      cout<<endl<<endl;
    }

    CSortedHypContainer &SimilarHyps = (*IterHypsByBestHyp).second;
    
    if (s_TraceOn){
      //lets look at all hyps in beam first
      
      for (IterSortedHyp  = SimilarHyps.begin(); IterSortedHyp != SimilarHyps.end();  ++IterSortedHyp){
        CDecoderBLM_Hyp* pHyp = (*IterSortedHyp).second;
        //display the hyp
        CArray< CLatticeEdgeMS * > transEdges = TraceBackWithPath(pHyp).second;
        for (int i=transEdges.MinIndex(); i<=transEdges.MaxIndex(); i++){
          transEdges[i]->Display();
          cout<<" ";
        }
        cout<<endl;
        cout<<"FScore: "<<(*IterSortedHyp).first<<" base: "<<pHyp->m_TotalScore<<endl;
      }
    }
    

    // only take the beamsize best
    for(IterSortedHyp = SimilarHyps.begin(), rank = 0;
        IterSortedHyp != SimilarHyps.end() && ++rank <= m_BeamSize;
        ++IterSortedHyp) {
      
      HypsInBeam.insert(*IterSortedHyp);
    } // endfor

    m_HypsOutOfBeam += ( SimilarHyps.size() > m_BeamSize ?
                         SimilarHyps.size() - m_BeamSize :
                         0 );
  } // endfor
  
  //cout << "DEBUG - Start expanding hyp ..." << endl;
  
  // out of the hyps just collected, only take the m_MaxHypsPerCoverage best
  for (IterSortedHyp  = HypsInBeam.begin(), rank = 0;
       IterSortedHyp != HypsInBeam.end() && ++rank <= m_MaxHypsPerCoverage;
       ++IterSortedHyp)
    ExpandHyp((*IterSortedHyp).second);  
  
  m_HypsOutOfBeamPerCov += ( HypsInBeam.size() > m_MaxHypsPerCoverage ?
                             HypsInBeam.size() - m_MaxHypsPerCoverage :
                             0 );
  
  s_TraceOn = false;
  
} // CDecoderBLM::ExpandAllHypsWithCoverage(...)


void
CDecoderBLM::ExpandHyp( CDecoderBLM_Hyp *pHyp )
{
  s_TraceOn = False;

  TraceIf( s_TraceOn, 0, "\n    CDecoderBLM::ExpandHyp" );

  // variable we need
  CLatticeNode *pNode;
  CLatticeEdgeMS *pEdge;
 
  TraceIf( s_TraceOn, 1, "    expand hyp " << *pHyp );

  // get the range of words in the source sentence 
  // which can now be translated, this is the active range
  // make sure that reordering stays within sentence
  int J = m_pLattice->EndNode()->Number();
  int j_min = pHyp->m_CoverageStart;
  int j_max = pHyp->m_CoverageStart + m_LengthOfReorderingWindow;
  if ( j_max > J ) 
	j_max = J;

  TraceIf( s_TraceOn, 0, "    Translate words in the range " << j_min << " " << j_max );
	
  // now loop over all partial translations t
  // i.e. over all edges in the translation lattice
  // which are within the active range
  // the end point of the edge has also to be within the maximally
  // allowed reordering window (c_MaxReorderWindowLength)

  // currently we reorder also source words, this could (should?) be avoided 

  for ( int j = j_min; j < j_max; j++ ) {
	// get node
	pNode = m_pLattice->Nodes()[ j ];
	TraceIf( s_TraceOn, 0, "    process node with number " << pNode->Number() );
	

	// expand this hypothesis only if it would not result in too 
	// many reorderings, but always expand over outgoing edges of node 
	// of current hypothesis, ie the node up to which everything has 
	// been translated (start of cov. pattern bitmap); this is the monotone decoding part
	// the initial hyp has no backpointer edge, the number of the previous
	// node is just the first not in the lattice

	if ( (pNode->Number() == pHyp->m_CoverageStart) 
		 || (pHyp->m_NumReorderings < pHyp->m_pDecoderInfo->m_MaxNumReorderings) ) {

	  // loop over all outgoing edges e
	  const CArray< CLatticeEdge * > & Edges = pNode->OutEdges();
	  TraceIf( s_TraceOn, 0, "    Number of outgoing edges is " << Edges.Size() );  
	  
	  for ( int EdgeNum = Edges.MinIndex();
			EdgeNum <= Edges.MaxIndex();
			EdgeNum++ ) {
	  
		pEdge = (CLatticeEdgeMS *)Edges[ EdgeNum ];
		
		if (m_Cohesion){			
			//Nguyen, July 31, 2008: Expand with dependency tree			
			
			this->ExpandHypOverEdgeWithCohesion( pHyp, pEdge );
						
		} else {
			// if edge is long it can move outside of the
			//assert( 0 );
			this->ExpandHypOverEdge( pHyp, pEdge );
		}
	  } // forall edges

	} else {

	  m_HypsNotExpanded++;
	} // if not to many reordering	
  } // for all positions within reordering window

}


//Nguyen, July 31, 2008
//Cohesive Decoding: need to pass previous edge

void
CDecoderBLM::ExpandHypOverEdgeWithCohesion( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge)
{
  s_TraceOn = false;
  
  m_Hypcover.clear();
  string src = getNameFromVector(m_Dtree);
  for (int i = 0; i< src.length(); ++i){
  	if ( isspace(src[i]) ) m_Hypcover.push_back(0); 
  }  
  
  //Nguyen, Aug 4, 2008: debug cohesive decoding
  //cout << endl;
  //cout << "======================" << endl;
  //cout << "DEBUG: Cohesive Decoding: CDecoderBLM::ExpandHypOverEdgeWithCohesion" << endl;
  //cout << "DEBUG: m_Dtree " << getNameFromVector(m_Dtree) << endl << endl;
  
  CArray< CLatticeEdgeMS * > transEdges = TraceBackWithPath(pHyp).second;
  
  for (int i=transEdges.MinIndex(); i<=transEdges.MaxIndex(); i++){
	  //transEdges[i]->Display();	  
	  //cout << endl;
	  int startnode = transEdges[i]-> FromNode()->Number();
	  int endnode   = transEdges[i]-> ToNode()->Number();
	  for (int j = startnode; j < endnode; j++){
	  	m_Hypcover[j] = 1;
	  }
  }
  //cout << "Current Translation: " << TraceBackWithPath(pHyp).first << endl;
  
  //Source coverage vector
//  for(vector<tokens>::size_type i = 0; i < m_Hypcover.size(); ++i){
//  	cout << m_Hypcover[i] << " " ;
//  }
//  cout << endl;
  
  //cout << "     pEdge: coverage from " << pEdge->FromNode()->Number() << " to " <<  pEdge->ToNode()->Number() << endl;  
  
  //prev source phrase
  vector<tokens> fh; 
  if (transEdges.MaxIndex() > -1) {
  	for (int i = transEdges[transEdges.MaxIndex()]->FromNode()->Number(); i < transEdges[transEdges.MaxIndex()]->ToNode()->Number(); i++){
  		fh.push_back(m_Dtree[ i ]);
  	}
  	//cout << "     previou Edge source: " << getNameFromVector(fh) << endl;  	
  }
  
  
  //current source phrase
  vector<tokens> fh1; 
  for (int i = pEdge->FromNode()->Number(); i < pEdge->ToNode()->Number(); i++){
  	fh1.push_back(m_Dtree[ i ]);
  }  
  //cout << "     pEdge source: " << getNameFromVector(fh1) << endl;
  
  TraceIf( s_TraceOn, 0, "\n      CDecoderBLM::ExpandHypOverEdge " << pHyp << " " << pEdge );
  TraceIf( s_TraceOn, 1, "      expand hyp " << *pHyp  );
  TraceIf( s_TraceOn, 1, "      expand over edge " );  
  
  //cout << "expand hyp over edge" << endl;
  //pEdge->Display();

  double TMScore;
  double LMScore;
  double ALScore;

  CLMState LMState = (CLMState) 0;
  CLMState NewLMState = (CLMState) 0;

  CDecoderBLM_Hyp *pBestHyp;

  // internal statistics
  m_HypsExpanded++;

  // test coverage collision for current hypothesis and this edge
  // we copy the current hypothesis into the NumberOneHyp,
  // as the TestAndSetCoverage function will modify the hypothesis
  *m_pNumberOneHypothesis = *pHyp;
  
  TraceIf( s_TraceOn, 1, "      test coverage from " << pEdge->FromNode()->Number() 
           << " to " <<  pEdge->ToNode()->Number() );
  
  
  if ( ! m_pNumberOneHypothesis->TestAndSetCoverage( pEdge->FromNode()->Number(), 
                                                     pEdge->ToNode()->Number() ) ) {

    TraceIf( s_TraceOn, 1, "      Coverage collision detected" );
    m_NumCoverageCollisions++;
    
  } else {

    TraceIf( s_TraceOn, 1, "      No coverage collision" );

    int NewCoverage = m_pNumberOneHypothesis->m_Coverage;
    TraceIf( s_TraceOn, 1, "      the memoryyp has coverage " << NewCoverage );
    TraceIf( s_TraceOn, 2, "      the memoryyp is " << *m_pNumberOneHypothesis );
            
    // length of the generated translation so far
    //cout << "Target " << pEdge->WordInfo()->TargetString()
    // << " | Source " << pEdge->WordInfo()->SourceString()
    // << " | Label  " << pEdge->Word() << endl;
    
    int AdditionalLength = pEdge->WordInfo()->TargetStringLength();
    if ( AdditionalLength == 0 ) {
      AdditionalLength = pEdge->WordInfo()->SourceStringLength();
    }
    if ( AdditionalLength == 0 ) {
      // this has to be an original edge, i.e. just one word
      AdditionalLength = 1;
    }
    //cout << m_pNumberOneHypothesis->m_TargetLength << " " << AdditionalLength << endl;
    m_pNumberOneHypothesis->m_TargetLength += AdditionalLength;
	
    // calculate the cost for the memoryypothesis,
    // the different Score-methods manipulate m_pNumberOneHypothesis

    // cost from the translation model
    //GetTMScore( pEdge );
    GetTMScoreMS((CLatticeEdgeMS*)pEdge);

    // cost for reordering
    //GetALScore( pHyp, pEdge );
    GetDMScores( pHyp, pEdge );

    // word count model: count target wordsq
    GetWCScore( pHyp, pEdge );

    // phrase count model: counts the number of edges used
    GetPCScore( pHyp, pEdge );
        
    //Nguyen, Aug 1, 2008: cohesive coding
    
    //PrintTime( "Start of  GetCMScore()" );
    
    GetCMScore( pHyp, pEdge, fh1, fh );    
        
    //PrintTime( "End of  GetCMScore()" );

    // store backpointer information
    // memoryypothesis is expansion of hypothesis pHyp over edge pEdge
    m_pNumberOneHypothesis->m_pEdge    = pEdge;
    m_pNumberOneHypothesis->m_pPrevHyp = pHyp;


    // cost from the language model, sets new LMState in m_pNumberOneHypothesis
    if ( m_LMBatchMode ) {
      StoreHypothesisTmp();
    }

    // will be in the else branch later
    //GetLMScore( pHyp, pEdge );
    GetLMScores( pHyp, pEdge );

    StoreHypothesis();
    
    //cout << "======================" << endl;    
  }

}
// end of Nguyen


void
CDecoderBLM::ExpandHypOverEdge( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge )
{
  s_TraceOn = false;

  TraceIf( s_TraceOn, 0, "\n      CDecoderBLM::ExpandHypOverEdge " << pHyp << " " << pEdge );

  TraceIf( s_TraceOn, 1, "      expand hyp " << *pHyp  );
  TraceIf( s_TraceOn, 1, "      expand over edge " );
  //cout << "expand hyp over edge" << endl;
  //pEdge->Display();

  double TMScore;
  double LMScore;
  double ALScore;

  CLMState LMState = (CLMState) 0;
  CLMState NewLMState = (CLMState) 0;

  CDecoderBLM_Hyp *pBestHyp;

  // internal statistics
  m_HypsExpanded++;

  // test coverage collision for current hypothesis and this edge
  // we copy the current hypothesis into the NumberOneHyp,
  // as the TestAndSetCoverage function will modify the hypothesis
  *m_pNumberOneHypothesis = *pHyp;
  
  TraceIf( s_TraceOn, 1, "      test coverage from " << pEdge->FromNode()->Number() 
           << " to " <<  pEdge->ToNode()->Number() );
  
  if ( ! m_pNumberOneHypothesis->TestAndSetCoverage( pEdge->FromNode()->Number(), 
                                                     pEdge->ToNode()->Number() ) ) {

    TraceIf( s_TraceOn, 1, "      Coverage collision detected" );
    m_NumCoverageCollisions++;
    
  } else {

    TraceIf( s_TraceOn, 1, "      No coverage collision" );

    int NewCoverage = m_pNumberOneHypothesis->m_Coverage;
    TraceIf( s_TraceOn, 1, "      the memoryyp has coverage " << NewCoverage );
    TraceIf( s_TraceOn, 2, "      the memoryyp is " << *m_pNumberOneHypothesis );
    
    // length of the generated translation so far
    //cout << "Target " << pEdge->WordInfo()->TargetString()
    // << "Source " << pEdge->WordInfo()->SourceString()
    // << "Label  " << pEdge->Word() << endl;
    
    int AdditionalLength = pEdge->WordInfo()->TargetStringLength();
    if ( AdditionalLength == 0 ) {
      AdditionalLength = pEdge->WordInfo()->SourceStringLength();
    }
    if ( AdditionalLength == 0 ) {
      // this has to be an original edge, i.e. just one word
      AdditionalLength = 1;
    }
    //cout << m_pNumberOneHypothesis->m_TargetLength << " " << AdditionalLength << endl;
    m_pNumberOneHypothesis->m_TargetLength += AdditionalLength;
	
    // calculate the cost for the memoryypothesis,
    // the different Score-methods manipulate m_pNumberOneHypothesis

    // cost from the translation model
    //GetTMScore( pEdge );
    GetTMScoreMS((CLatticeEdgeMS*)pEdge);

    // cost for reordering
    //GetALScore( pHyp, pEdge );
    GetDMScores( pHyp, pEdge );

    // word count model: count target words
    GetWCScore( pHyp, pEdge );

    // phrase count model: counts the number of edges used
    GetPCScore( pHyp, pEdge );

    // store backpointer information
    // memoryypothesis is expansion of hypothesis pHyp over edge pEdge
    m_pNumberOneHypothesis->m_pEdge    = pEdge;
    m_pNumberOneHypothesis->m_pPrevHyp = pHyp;


    // cost from the language model, sets new LMState in m_pNumberOneHypothesis
    if ( m_LMBatchMode ) {
      StoreHypothesisTmp();
    }

    // will be in the else branch later
    //GetLMScore( pHyp, pEdge );
    GetLMScores( pHyp, pEdge );

    StoreHypothesis();


  }

}



void
CDecoderBLM::StoreHypothesisTmp()
{
  // for LM batch mode we do not yes have the LM score
  // store hyp in temporary container to add the LM score later
  assert( m_LMBatchMode );

  // copy current NumberOneHyp, as this will be reused
  CDecoderBLM_Hyp *pTmpHyp = NewHypothesis();
	*pTmpHyp = *m_pNumberOneHypothesis;

  // and store it
  m_TmpHyps.Append( pTmpHyp );

  //TraceHyp( true, "Tmp Hyp stored ", pTmpHyp );
  cout << "Tmp Hyp stored " << *pTmpHyp << endl;


}



void
CDecoderBLM::StoreHypothesis()
{
  // old: for single-best decoding
  // store memoryypothesis according to coverage
  // but only if it is the first one of its kind, or has a lower cost
  // if it is new, create a new NumberOneHyp
  // if it is better, copy content, but don't create a memoryyp object
  

  // new: for n-best decoding

  // store memoryypothesis according to coverage

  // first pruning: this is unchanged to single-best decoding
  // if new: in BestHyps, store it
  // if better: overwrite old one
  // if worse: do nothing

  // now storing: different, as we store recombined hyps with best
  // if too bad compared to best for pruing, then prune
  // if new: store NumberOneHyp and create new NumberOneHyp
  // if better: add pointer of stored hyp to RecombinedHyps; 
  //            store NumberOneHyp and create new NumberOneHyp
  // if worse: add pointer of NumberOneHyp to RecombinedHyps;

  s_TraceOn = False;

  TraceIf( s_TraceOn, 0, "=== Store Hypothesis ===" );

  //TraceHyp( true, "hyp with standard LM score ", m_pNumberOneHypothesis );
  //cout << "hyp with standard LM score " << *m_pNumberOneHypothesis << endl;

  CString Translation;

  // has NumberOneHyp DecoderInfo?
  assert( m_pNumberOneHypothesis->m_pDecoderInfo );

  int NewCoverage = m_pNumberOneHypothesis->m_Coverage;
  assert( NewCoverage > 0 );

  CDecoderBLM_Hyp *pStoredHyp, *pBestHypForPruning;

  // for recombination, access to stored hyp
  HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual >::iterator IterFind;

  // for pruning, access to best hyps
  HypContainer< CDecoderBLM_Hyp *, HypHashForPruning, HypEqualForPruning >::iterator IterFindHypForPruning;

  // copy information to NumberOneBestHyp
  //*m_pNumberOneBestHyp = *m_pNumberOneHypothesis;

  IterFindHypForPruning = m_BestHyps[ NewCoverage ].find( m_pNumberOneHypothesis );


  // first make sure that hyp is in best hyps
  // if it is not there, create it
  // if it is there but we have now a better one, replace it

  if ( IterFindHypForPruning == m_BestHyps[ NewCoverage ].end() ) {
	
    // we don't have a best hypothesis so far
    // make memoryypothesis, copy content from NumberOneHyp and store it

    TraceIf( s_TraceOn, 0, "      Prune: Best Hyp is new, store it" );

    //pBestHypForPruning = new CDecoderBLM_Hyp;
    pBestHypForPruning = NewHypothesis();
    *pBestHypForPruning = *m_pNumberOneHypothesis;
    assert( pBestHypForPruning->m_pDecoderInfo );

    assert( NewCoverage > 0 );
    m_BestHyps[ NewCoverage ].insert( pBestHypForPruning );

    //m_pNumberOneBestHyp = new CDecoderBLM_Hyp;
    //m_pNumberOneBestHyp->m_pDecoderInfo = m_pDecoderInfo;

  } else {

    // so, one of that kind exists already, now we can compare
    // first, would the memoryyp be a better best hyp?
    pBestHypForPruning = *IterFindHypForPruning;

    if ( m_pNumberOneHypothesis->m_TotalScore < pBestHypForPruning->m_TotalScore ) {

      TraceHyp( s_TraceOn, "      Prune: Hyp is better than stored one, overwrite ", pBestHypForPruning );
      TraceHyp( s_TraceOn, "      with ",  m_pNumberOneHypothesis );

      // copy content of NumberOne over
      *pBestHypForPruning = *m_pNumberOneHypothesis;

    } else {

      TraceIf( s_TraceOn, 0, "      Prune: Hyp is worse than stored one, don't use it" );

    }

  }


  // here we have the best hypothesis as pBestHyp
  // we compare the memoryypthesis against this best hyp

  // is it good enough or should it be pruned
  //if ( m_pNumberOneHypothesis->m_TotalScore > pBestHypForPruning->m_TotalScore + (((double)m_BeamFactor/100.0)*pBestHypForPruning->m_TotalScore) ) {
  //no pruning at this stage
  if (0){
    TraceIf( s_TraceOn, 0, "      Hyp is too bad, prune it " );
    //m_HypsPruned++;
		
  } else {

    // hyp is good enough to keep it
    // now we need to compare it to stored hyp for recombination
    IterFind = m_AllHyps[ NewCoverage ].find( m_pNumberOneHypothesis );

    // is it a memoryyp?
    if ( IterFind == m_AllHyps[ NewCoverage ].end() ) {
      // yes
      TraceHyp( s_TraceOn, "      Hyp is new ", m_pNumberOneHypothesis );
	  
      // create map for recombined hyps, emtpy at the beginning
      // ARV
      // changed this to use the managed call to NewRecombinedMap
      //m_pNumberOneHypothesis->m_pRecombinedHyps = new multimap< float, CDecoderBLM_Hyp * >;
      m_pNumberOneHypothesis->m_pRecombinedHyps = NewRecombinedMap();

      // we need to store NumberOneHyp and create a new NumberOneHyp object
      assert( NewCoverage > 0 );
      m_AllHyps[ NewCoverage ].insert( m_pNumberOneHypothesis );

      //cout << "TRACEBACK new" << endl;
      //Translation = TraceBack( m_pNumberOneHypothesis );
      //cout << "TRACEBACK new = " << Translation << endl;

      // test if the correct hyp is stored
      IterFind = m_AllHyps[ NewCoverage ].find( m_pNumberOneHypothesis );
      CDecoderBLM_Hyp *pHypForDebugging;
		
      // ARV
      // following code is debugging only
      if ( IterFind == m_AllHyps[ NewCoverage ].end() ) {
        TraceIf( s_TraceOn, 0, "     Dammned! the just inserted hyp can not be found" );
      } else {
        pHypForDebugging = *IterFind;
        TraceHyp( s_TraceOn, "      Stored best Hyp is now ", pHypForDebugging );
         TraceIf( s_TraceOn, 0, " with " << pHypForDebugging->m_pRecombinedHyps->size() << " recombined hyps" );
      }
      // end debugging information
	       

      //m_pNumberOneHypothesis = new CDecoderBLM_Hyp;
      m_pNumberOneHypothesis = NewHypothesis();
      m_pNumberOneHypothesis->m_pDecoderInfo = m_pDecoderInfo;
      m_NumNewHyps++;
	  
    } else {
		
      // it exist already
      // is it better or worse then stored one

      pStoredHyp = *IterFind;

      if ( m_pNumberOneHypothesis->m_TotalScore < pStoredHyp->m_TotalScore ) {
		  
        assert( m_pNumberOneHypothesis->m_Coverage == pStoredHyp->m_Coverage );

        // better

        // we reuse the stored one and update it with the info from NumberOneHyp
        // (insert into map will not overwrite old pointer with new pointer,
        // so either we delete and insert, or modify)
        // therefore: first create memoryyp object and copy info of stored hyp into memoryyp
        //            second overwrite stored hyp with new best (i.e. NumberOneHyp)
        //            third add memoryyp (i.e old best) to the list of recombined hyps

        TraceHyp( s_TraceOn, "      Hyp ", m_pNumberOneHypothesis );
        TraceHyp( s_TraceOn, "      is better than stored one ", pStoredHyp );
        TraceIf( s_TraceOn, 0, " with " << pStoredHyp->m_pRecombinedHyps->size() << " recombined hyps" );

        // first: memoryyp as copy of currently stored
        //CDecoderBLM_Hyp *pStoredHypCopy = new CDecoderBLM_Hyp;
        CDecoderBLM_Hyp *pStoredHypCopy = NewHypothesis();
        *pStoredHypCopy = *pStoredHyp;
        assert( pStoredHypCopy->m_pRecombinedHyps );

        TraceHyp( s_TraceOn, "      copy of stored hyp is ", pStoredHypCopy );

        // second: overwrite stored with better hyp
        *pStoredHyp = *m_pNumberOneHypothesis;
        // need to carry over the pointer from recombined hyps from
        // copy of stored hyp to stored hyp
        pStoredHyp->m_pRecombinedHyps = pStoredHypCopy->m_pRecombinedHyps;
        pStoredHypCopy->m_pRecombinedHyps = NULL;

        TraceHyp( s_TraceOn, "      modified stored hyp is ", pStoredHyp );

        //Translation = TraceBack( m_pNumberOneHypothesis );
        //cout << "TRACEBACK new best = " << Translation << endl;



        // third: add former best hyp to list of recombined hyps
        assert( pStoredHypCopy->m_Coverage > 0 );
        pStoredHyp->m_pRecombinedHyps->insert( pair< float, CDecoderBLM_Hyp * > ( pStoredHypCopy->m_TotalScore, pStoredHypCopy ) );
		
        TraceHyp( s_TraceOn, "      Resulting best Hyp is ", pStoredHyp );
        TraceIf( s_TraceOn, 0, "Number of recombined hyp is now: " << pStoredHyp->m_pRecombinedHyps->size() );
		
        // test if the correct hyp is stored
        IterFind = m_AllHyps[ NewCoverage ].find( m_pNumberOneHypothesis );
        CDecoderBLM_Hyp *pHypForDebugging;
		
        if ( IterFind == m_AllHyps[ NewCoverage ].end() ) {
          TraceIf( s_TraceOn, 0, "     Dammned! the just inserted hyp can not be found" );
        } else {
          pHypForDebugging = *IterFind;
          TraceHyp( s_TraceOn, "      Stored best Hyp is now ",  pHypForDebugging );
          TraceIf( s_TraceOn, 0, " with " << pHypForDebugging->m_pRecombinedHyps->size() << " recombined hyps" );
        }
		
        // we reuse NumberOneHyp
		
        m_NumBetterHyps++;
		
		
      } else {
		

        TraceHyp( s_TraceOn, "      Hyp ", m_pNumberOneHypothesis );
        TraceHyp( s_TraceOn, "      is worse than stored one ", pStoredHyp );

        // we add it to the list of recombined hyps
        assert( m_pNumberOneHypothesis->m_Coverage > 0 );
        pStoredHyp->m_pRecombinedHyps->insert( pair< float, CDecoderBLM_Hyp * >
                                               ( m_pNumberOneHypothesis->m_TotalScore, m_pNumberOneHypothesis ) );
		
        //cout << "TRACEBACK new worse" << endl;
        //Translation = TraceBack( m_pNumberOneHypothesis );
        //cout << "TRACEBACK new worse = " << Translation << endl;


        TraceIf( s_TraceOn, 0, "Number of recombined hyp is now: " << pStoredHyp->m_pRecombinedHyps->size() );
		
        // and create a new NumberOneHyp;
        //m_pNumberOneHypothesis = new CDecoderBLM_Hyp;
        m_pNumberOneHypothesis = NewHypothesis();
        m_pNumberOneHypothesis->m_pDecoderInfo = m_pDecoderInfo;
		
        // but count it as a worse hyp
        m_NumWorseHyps++;
		
      }
    }
    //
    //TraceIf( s_TraceOn, 0, "      Number of hyps with coverage " << pHyp->m_Coverage 
    //		 << " is still " << m_AllHyps[ pHyp->m_Coverage ].size() );
    TraceIf( s_TraceOn, 0, "      Number of hyps with coverage " << NewCoverage 
             << " is now   " << m_AllHyps[ NewCoverage ].size() );
	
  }
  
  s_TraceOn = false;
  
}




void
CDecoderBLM::ExpandAllHypsSentenceEnd()
{
  TraceIf( 0, 0, "CDecoderBLM::ExpandAllHypsSentenceEnd()" );
  cout << "CDecoderBLM::ExpandAllHypsSentenceEnd()" << endl;

  // loop over all hyps with full coverage
  // apply sentence end scores (LM, sentence length model)
  // and sort them according to cost

  // clear the final list, in case it still has the hyps from previous sentence
  m_FinalHyps.clear();

  // first we fetch the set of hypotheses which all have full coverage,
  // i.e. source sentence length
  int J = m_pLattice->EndNode()->Number();
  HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual > &Hyps = m_AllHyps[ J ];
  TraceIf( s_TraceOn, 1, "\n  Number of hyps with full coverage is " << Hyps.size() );
  cout << "Number of hyps with full coverage is " << Hyps.size() << endl;

  // and we need an iterater to access the hyps in this set
  HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual >::iterator IterHyps;

  CDecoderBLM_Hyp *pHyp;

  //ARV 04/25 
  //moving over to general form of initilizing a model and then using it in the loop
  //InitLexicalSLScoreEnd();
  
  
  cout << "DEBUG - CDecoderBLM::ExpandAllHypsSentenceEnd : loop over hyps" << endl;
  // loop over hyps
  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {

    // create a copy of the hyp and point back to that hyp
    // we don't go over an edge at this point
    //pHyp = new CDecoderBLM_Hyp( **IterHyps );
    pHyp = NewHypothesis();
    *pHyp = **IterHyps;

    // no recombined hyps
    pHyp->m_pRecombinedHyps = NULL;
    // and coverage increased by one, as if going over </s>
    pHyp->m_Coverage++;

    pHyp->m_pDecoderInfo = m_pDecoderInfo;
    pHyp->m_pPrevHyp = *IterHyps;
    pHyp->m_pEdge = NULL;
    

    // cost for sentence end from the language model

    /*
    //double LMScore = -m_pLMSentenceState->LogProbEnd( pHyp->m_LMState );
    // sv: modify for multiple LMs
    double LMScore = -m_LMSentenceStates[ m_LMSentenceStates.MinIndex() ]
      ->LogProbEnd( pHyp->m_LMStates[ pHyp->m_LMStates.MinIndex() ] );

    //pHyp->m_scaledModelScores[ c_LM1 ] += m_ScalingFactors[ c_LM1 ] * LMScore;
    //pHyp->m_TotalScore += m_ScalingFactors[ c_LM1 ] * LMScore;
    pHyp->m_scaledModelScores[ m_StartLMScores ] += m_ScalingFactors[ m_StartLMScores ] * LMScore;
    pHyp->m_TotalScore += m_ScalingFactors[ m_StartLMScores ] * LMScore;
 
    TraceIf( s_TraceOn, 1, "\n  LM cost for sentence end " << LMScore );
    */
    
    GetLMScoresEOS( pHyp );

    // cost from sentence length model
    // exp( -l(cJ - I)^2 )
    // two factors: c is general stretching factor: use 1.2 for now
    //              f is flatness factor: use 0.25 for now

    // SL model is currently only a word count model
    // which is already applied at the phrase level

    //float SLScore;
    //int I = pHyp->m_TargetLength;
    
    //ARV 
    // removed this old style in factor of the simple counting word score
    //SLScore = 0.25 * ( m_SLStretch * J - I ) * ( m_SLStretch * J - I );

    //SLScore = -1.0*I;
    //SLScore = 2*J -I;

    //SLScore = I;

    //ARV 04/20
    //pHyp->m_SLScore = m_SLScale*SLScore;
    //pHyp->m_TotalScore += m_SLScale * SLScore;

    //cout << "WC score already in hyp " << pHyp->m_scaledModelScores[c_SL];
    //cout << "  WC score from sentence end " << SLScore << " " << m_ScalingFactors[c_SL]*SLScore << endl;

    //pHyp->m_scaledModelScores[ c_SL1 ] = m_ScalingFactors[ c_SL1 ] * SLScore;
    //pHyp->m_TotalScore += m_ScalingFactors[ c_SL1 ] * SLScore;

    //pHyp->m_scaledModelScores[ m_StartSLScores ] = m_ScalingFactors[ m_StartSLScores ] * SLScore;
    //pHyp->m_TotalScore += m_ScalingFactors[ m_StartSLScores ] * SLScore;


    //double lexScore = GetLexicalSLScoreEnd(pHyp);
    //pHyp->m_scaledModelScores[c_LSL] = m_ScalingFactors[c_LSL]*lexScore;
    //pHyp->m_TotalScore += m_ScalingFactors[c_LSL]*lexScore;


    // now insert into final list, sorting is done by the container
    TraceIf( s_TraceOn, 1, "\n  insert hyp into final list" 
             << pHyp << " "
             << pHyp->m_Coverage << " " << pHyp->m_CoverageStart << " " 
             << pHyp->m_CoveragePattern );
    
    assert(  pHyp->m_Coverage > 0 );
    m_FinalHyps.insert(pair< double, CDecoderBLM_Hyp * >( pHyp->m_TotalScore , pHyp ) );
    
  }

  TraceIf( s_TraceOn, 1, "Number of Hyps in Final List is: " << m_FinalHyps.size() );

}





CString
CDecoderBLM::TraceBack( CDecoderBLM_Hyp *pHyp )
{
  bool OldTraceOn = CDecoderBLM::s_TraceOn;
  CDecoderBLM::s_TraceOn = false;

  //cout << "Trace Back" << endl;
  //cout << *pHyp << endl;
  //cout << pHyp->m_pPrevHyp << endl;

  int Start = m_pLattice->StartNode()->Number();
  int End   = m_pLattice->EndNode()->Number();
  int L = End - Start;

  if ( s_TraceOn ) {
    cout << endl;
    cout << "Total Score " << pHyp->m_TotalScore;
    for ( int i = 0; i < pHyp->m_scaledModelScores.Size(); i++ ) {
      cout << " Mod " << i << " = " << pHyp->m_scaledModelScores[ i ];
    }
    cout<<endl;

    cout << "Reorderings " << pHyp->m_NumReorderings / 2 << endl;
  }

  CArray< CLatticeEdgeMS * > AllEdges;
  CArray< CLatticeEdge * >   AllEdges2;

  while ( pHyp->m_pPrevHyp ) {
    //cout << "Hyp " << pHyp << " PrevHyp " << pHyp->m_pPrevHyp 
    //<< " Edge " << pHyp->m_pEdge << endl;
    if ( !pHyp->m_pEdge ) {
      //cout << "no edge - </s>" << endl;
    } else {
      //pHyp->m_pEdge->Display();
      AllEdges.Append( pHyp->m_pEdge );
    }
    
    //cout << *pHyp->m_pPrevHyp << endl;
    pHyp = pHyp->m_pPrevHyp;
  }


  // first1, display the path that has been taken
  // invert sequence of edges in list
  for ( int i = AllEdges.MaxIndex(); i >= AllEdges.MinIndex(); i-- ) {
    AllEdges2.Append( AllEdges[ i ] );
  }
  //cout << "PartialPath: ";
  //if (s_TraceOn){
  //m_pLattice->DisplayPath( AllEdges2 );
	//}
  
  // now display the clean translation
  CString Translation;
  CWordInfo *pWordInfo;
  CLatticeEdgeMS *pEdge;

  for ( int i = AllEdges.MaxIndex(); i >= AllEdges.MinIndex(); i-- ) {
	pEdge = AllEdges[ i ];
	pWordInfo =  pEdge->WordInfo();

	// extract sequence of words from the WordInfo structure
	if ( pWordInfo->HasTargetString() ) {
	  Translation += pWordInfo->TargetString();
	} else if ( pWordInfo->HasSourceString() ) {
	  Translation += pWordInfo->SourceString();
	} else {
	  Translation += pEdge->Word();
	}

	Translation += " ";
  }

  //  if (s_TraceOn){
  //  cout << "Translation is: " << Translation << endl;
  //  cout << "------------" << endl;
	//}
  CDecoderBLM::s_TraceOn = OldTraceOn;

  return Translation;
}



std::pair<CString, CArray< CLatticeEdgeMS * > >
CDecoderBLM::TraceBackWithPath( CDecoderBLM_Hyp *pHyp )
{
  bool OldTraceOn = CDecoderBLM::s_TraceOn;
  CDecoderBLM::s_TraceOn = false;

  //cout << "Trace Back" << endl;
  //cout << *pHyp << endl;
  //cout << pHyp->m_pPrevHyp << endl;

  int Start = m_pLattice->StartNode()->Number();
  int End   = m_pLattice->EndNode()->Number();
  int L = End - Start;

  if ( s_TraceOn ) {
    cout << endl << "Total Score " << pHyp->m_TotalScore;
    for ( int i = 0; i < pHyp->m_scaledModelScores.Size(); i++ ) {
      cout << " Mod " << i << " = " << pHyp->m_scaledModelScores[ i ];
    }
    cout << endl;

    cout << "Reorderings " << pHyp->m_NumReorderings / 2 << endl;
  }

  CArray< CLatticeEdgeMS * > AllEdges;
  CArray< CLatticeEdgeMS * > AllEdges2;

  while ( pHyp->m_pPrevHyp ) {
    //cout << "Hyp " << pHyp << " PrevHyp " << pHyp->m_pPrevHyp 
    //<< " Edge " << pHyp->m_pEdge << endl;
    if ( !pHyp->m_pEdge ) {
      //cout << "no edge - </s>" << endl;
    } else {
      //pHyp->m_pEdge->Display();
      AllEdges.Append( pHyp->m_pEdge );
    }

    //cout << *pHyp->m_pPrevHyp << endl;
    pHyp = pHyp->m_pPrevHyp;
  }


  // first1, display the path that has been taken
  // invert sequence of edges in list
  for ( int i = AllEdges.MaxIndex(); i >= AllEdges.MinIndex(); i-- ) {
    AllEdges2.Append( AllEdges[ i ] );
  }
  //cout << "PartialPath: ";
  //if (s_TraceOn){
  //m_pLattice->DisplayPath( AllEdges2 );
	//}

  // now display the clean translation
  CString Translation;
  CWordInfo *pWordInfo;
  CLatticeEdgeMS *pEdge;

  for ( int i = AllEdges.MaxIndex(); i >= AllEdges.MinIndex(); i-- ) {
    pEdge = AllEdges[ i ];
    pWordInfo =  pEdge->WordInfo();

    // extract sequence of words from the WordInfo structure
    if ( pWordInfo->HasTargetString() ) {
      Translation += pWordInfo->TargetString();
    } else if ( pWordInfo->HasSourceString() ) {
      Translation += pWordInfo->SourceString();
    } else {
      Translation += pEdge->Word();
    }

    Translation += " ";
  }

  //  if (s_TraceOn){
  //  cout << "Translation is: " << Translation << endl;
  //  cout << "------------" << endl;
	//}
  CDecoderBLM::s_TraceOn = OldTraceOn;
  
  return pair< CString, CArray< CLatticeEdgeMS * > >( Translation, AllEdges2 );
}




void
CDecoderBLM::BuildNBestList()
{
  s_TraceOn = false;

  Trace( 0, "Build N-Best List" );
  //Trace( 0, "Number of BPs is " << CDecoderBLM_BP::s_NumberOfObjects++ );
  
  // iterator to iterate over list of backpointer for one coverage
  multimap< float, CDecoderBLM_BP * >::iterator IterOverBPs;

  CDecoderBLM_BP *pBP, *pBP_New;
  CString Translation;


  // -----------------------------------------------------------------------------------
  // first, we fill NBestHyps( J + 1 ) with top N hyps from 
  // m_FinalHyps.insert(pair< double, CDecoderBLM_Hyp * >( pHyp->m_TotalScore , pHyp ) );
  // -----------------------------------------------------------------------------------

  TraceIf( s_TraceOn, 0, "Insert Final Hyps into BP Table" );
  TraceIf( s_TraceOn, 0, "============================== " );

  multimap< float, CDecoderBLM_Hyp * >::iterator IterOverFinalHyps;
  CDecoderBLM_Hyp *pHyp, *pPrevHyp;
  CDecoderBLM_Hyp *LastHyp = NULL;  // for debugging


  int J = m_AllHyps.MaxIndex() + 1;  // we have added a (virtual) </s> to sentence
  double WorstScore;
  //m_NumOfNBest = 100;

  //m_NBestHyps.SetMinIndex( 0 );
  m_NBestHyps.SetMaxIndex( J );

  
  int n = 1;
  for ( IterOverFinalHyps = m_FinalHyps.begin();
        (IterOverFinalHyps != m_FinalHyps.end()) && (n <= m_NumOfNBest * c_NumOfNBestMultiplier);
        ++IterOverFinalHyps ) {

    pHyp = IterOverFinalHyps->second;
    TraceHyp( s_TraceOn, 0, "Hyp is ", pHyp );

    // create for each hyp a backtrace item and place it in the list
    CDecoderBLM_BP *pBP = new CDecoderBLM_BP( pHyp );

    // hyps at sentence end have to have coverage equal to sentence length
    assert( pBP->m_pHyp->m_Coverage == J );
    // and we don't want the same hyp several times
    assert( pBP->m_pHyp != LastHyp );
    LastHyp = pHyp;

    //Display( pBP, n );
    //Translation = TraceBack( pBP->m_pHyp ) + " # " + TraceForward( pBP );
    //cout << "Final Translation: " << J + 1 << " " << Translation << endl;

    m_NBestHyps[ J ].insert( pair< float, CDecoderBLM_BP * >( pBP->m_TotalScore, pBP ) );
    
    n++;

  }
  

  int Number = m_NBestHyps[ J ].size();
  TraceIf( s_TraceOn, 0, "Number to start with is " << Number );

  // iterotor for looping back over all coverages down to 0
  multimap< float, CDecoderBLM_Hyp * >::iterator IterRecombinedHyps;

  CDecoderBLM_Hyp *pRecombinedHyp;


  // -----------------------------------------------------------------------------------
  // now loop backwards over coverage
  // -----------------------------------------------------------------------------------

  int j = J;
  while( j > 0 ) {

    TraceIf( s_TraceOn, 0, "Process NBest with coverage " << j );
    TraceIf( s_TraceOn, 0, "=========================== " << j );

    // we expand only top m_NumOfNBest
    // sv 060602: expand now c_NumOfNBestMultiplier * m_NumOfNBest, as we later remove dublicate
    n = 1;   // number of expanded

    // the list of backpointers for current coverage
    multimap< float, CDecoderBLM_BP * > &CurrentBPs = m_NBestHyps[ j ];

    TraceIf( s_TraceOn, 0, "Number of Backpointers is " << CurrentBPs.size() );
    TraceIf( s_TraceOn, 0, "Listing all Backpointers" );

    // what is the score of the m_NumOfNBest entry in the current list?
    // this will be used to exit expansion of backpointers;
    // if we have still not enaugh backpointer, make WorstScore very large
    if ( CurrentBPs.size() > m_NumOfNBest * c_NumOfNBestMultiplier ) {
      for ( IterOverBPs = CurrentBPs.begin();
            (IterOverBPs != CurrentBPs.end()) && (n <= m_NumOfNBest * c_NumOfNBestMultiplier);
            ++IterOverBPs ) {
        pBP = IterOverBPs->second;
        WorstScore = pBP->m_TotalScore;

        //Display( pBP, n );
        //Translation = TraceBack( pBP->m_pHyp ) + " # " + TraceForward( pBP );
        //cout << "Intermediate nbest translation: " << j << " " << Translation << endl;

        n++;
      }
    } else {
      WorstScore = 10000000;
      TraceIf( s_TraceOn, 0, "Not enough hyps - worst Score is 1000000" );
    }


    // ---------------------------------------------------------------------------------
    // 2nd level: loop through list of backpointers for current coverage up to NBest
    // ---------------------------------------------------------------------------------
    n = 1;
    for ( IterOverBPs = CurrentBPs.begin();
          (IterOverBPs != CurrentBPs.end()) && (n <= m_NumOfNBest * c_NumOfNBestMultiplier);
          ++IterOverBPs ) {
      pBP = IterOverBPs->second;

      TraceIf( s_TraceOn, 0, "\n\nCurrent BACKPOINTER Number:" << n << " Score:" << IterOverBPs->first << " " 
               << pBP->m_TotalScore << " HypEdge" << pBP->m_pHyp->m_pEdge << " PrevBP:" << pBP->m_pBP );

      pHyp = pBP->m_pHyp;

      // check if hypothesis is correct, i.e. if it has a link to a previous hypothesis
      // here, coverage is > 0, i.e. there has to be such a link
      if ( (pHyp->m_Coverage == 0) || (!pHyp->m_pPrevHyp) ) {

        TraceHyp( 1, 0, "ALARM: rotten hyp ", pHyp );

      } else {

        pPrevHyp = pHyp->m_pPrevHyp;
		
        if (s_TraceOn){
          TraceHyp( s_TraceOn, 1, "      best hyp is ",  pHyp );
          cout << "     translation is: " << TraceBack( pHyp ) << endl;
          TraceHyp( s_TraceOn, 1, "      pred hyp is ", pPrevHyp );
          cout << "     translation prev hyp is: " << TraceBack( pPrevHyp ) << endl;
        }
        // take hyp pointed to, store predecessor hyp into slot with predecessor coverage
        // this is the best one with given set of features, i.e. we trace it back

        int PredecessorCoverage = pPrevHyp->m_Coverage;
		
        // this initialization is to get the score that is in the pBP
        pBP_New = new CDecoderBLM_BP( pBP );
        pBP_New->m_pHyp  = pPrevHyp;           // the hyp from which this one originated
        //		pBP_New->m_pEdge = pPrevHyp->m_pEdge;  // the edger over which the previous hyp was expanded		
        TraceIf( s_TraceOn, 1, "    pBP_New Length was " << pBP_New->m_TargetLength );
        int NewBPTargetExtension = TargetLength( pPrevHyp );
        // ARV
        // add a variable so that we can remove the extension that came from this memoryyp
        //pBP_New->m_TargetLength += TargetLength( pPrevHyp );
        pBP_New->m_TargetLength += NewBPTargetExtension;
        TraceIf( s_TraceOn, 1, "    pBP_New Length is now " << pBP_New->m_TargetLength );
        TraceIf( s_TraceOn, 1, "    Predeccessor Coverage " << PredecessorCoverage );
		
	  
        // -----------------------------------------------------------------------------
        // End Condition:
        // if new coverage = 0, i.e. we are at the sentence start, then
        // the true sentence length is calculated and the correct SLScore
        // is estimated
        // -----------------------------------------------------------------------------
        if ( PredecessorCoverage == 0 ) {
          //UpdateSLScore( pBP_New, n );
          UpdateWCScore( pBP_New, n );
          //Sanjika 04/25
          //UpdateLexicalSLScore(pBP_New, n);
        }


        TraceIf( s_TraceOn, 1, "  Going from Hyp " << pHyp << " back to Hyp " << pBP_New->m_pHyp );
        TraceIf( s_TraceOn, 1, "  Insert into list of BPs at Coverage " << PredecessorCoverage );
        assert( pBP_New->m_pHyp->m_Coverage == PredecessorCoverage );
        // and we don't want the same hyp several times
        assert( pBP_New->m_pHyp != LastHyp );
        LastHyp = pHyp;

        //Display( pBP_New, n );
        //Translation = TraceBack( pBP_New->m_pHyp ) + " # " + TraceForward( pBP_New );
        //cout << "Intermediate best Translation: " << j << " " << Translation << endl;

        // now, store backpointer according to coverage
        // ARV
        // only if it is actually a memoryypothesis
		
        m_NBestHyps[ PredecessorCoverage ].insert
          ( pair< float, CDecoderBLM_BP * >( pBP_New->m_TotalScore, pBP_New ) ); 
  
        n++;


        // -----------------------------------------------------------------------------
        // now we deal with the recombined hyps
        // if predecessor hyp has recombinded hyps we create backpointers for those
        // for those we have to calculate the true final score
        // -----------------------------------------------------------------------------

        // we initialize TrueScore (used to calculate true final score of recombined hyps)
        // to 0, as it will be used to exit backtracing of recombined hyps against WorstScore
        float TrueScore = 0;

        if ( pPrevHyp->m_pRecombinedHyps ) {

          TraceIf( s_TraceOn, 1, "  Number of recombined Hyps is " << pPrevHyp->m_pRecombinedHyps->size() );


          // ---------------------------------------------------------------------------
          // 3rd level: loop over all recombined hypotheses
          // ---------------------------------------------------------------------------
		  
          for ( IterRecombinedHyps =pPrevHyp->m_pRecombinedHyps->begin();
                (IterRecombinedHyps != pPrevHyp->m_pRecombinedHyps->end()) && (TrueScore < WorstScore);
                ++IterRecombinedHyps ) {
		  
            // loop through all recombined hyps
            pRecombinedHyp = IterRecombinedHyps->second;

            // we don't want the same hyp several times
            assert( pRecombinedHyp != LastHyp );
            LastHyp = pRecombinedHyp;

            TraceHyp( s_TraceOn, 1, "      recombined hyp is ", pRecombinedHyp );

            //cout << "     translation rec hyp is: " << TraceBack( pRecombinedHyp )  + " # " + TraceForward( pBP_New ) << endl;
            //cout << "     translation rec prev hyp is: " << TraceBack( pRecombinedHyp->m_pPrevHyp ) << endl;
            // calculate true Nbest list
			
            TrueScore = pBP_New->m_TotalScore + (pRecombinedHyp->m_TotalScore - pPrevHyp->m_TotalScore); 

            TraceIf( s_TraceOn, 1, "      Additional Score is "
                     << pRecombinedHyp->m_TotalScore - pPrevHyp->m_TotalScore
                     << "      True total score is " << TrueScore );

            if ( TrueScore > WorstScore ) {
              TraceIf( s_TraceOn, 0, "      Break: TrueScore > WorstScore " << WorstScore );
            } else {

              // take hyp pointed to, store predecessor hyp into slot with predecessor coverage
              // this is the best one with given set of features, i.e. we trace it back

              //ARV 4/20 Changed to also pass the number of models
			  
              CDecoderBLM_BP *pBP_RHyp = new CDecoderBLM_BP( m_NumberOfModels );
              pBP_RHyp->m_TotalScore = TrueScore; // we propagate the final score inside the backpointer
              // also other scorse
			  
              //not used anywhere
              //pBP_RHyp->m_ACScore = pBP_New->m_ACScore + (pRecombinedHyp->m_ACScore - pPrevHyp->m_ACScore);
			  


              //ARV: 04/20 Generalizing the decoder to handle a set of scores
              /*
              //pBP_RHyp->m_TMScore = pBP_New->m_TMScore + (pRecombinedHyp->m_TMScore - pPrevHyp->m_TMScore);
              //pBP_RHyp->m_LMScore = pBP_New->m_LMScore + (pRecombinedHyp->m_LMScore - pPrevHyp->m_LMScore);
              //pBP_RHyp->m_ALScore = pBP_New->m_ALScore + (pRecombinedHyp->m_ALScore - pPrevHyp->m_ALScore);
              */
              //set each one via the array process
              //this way it does not need to be set by hand

              //make sure the hyp has the size for this
              for ( int m = 0; m < m_NumberOfModels; m++ ) {
                pBP_RHyp->m_scaledModelScores[ m ] = pBP_New->m_scaledModelScores[ m ]
                  + (pRecombinedHyp->m_scaledModelScores[ m ] - pPrevHyp->m_scaledModelScores[ m ] ); 
              }

              //this is just set here now, but it will be updated at the end of the n-best list
              //since we need to track the target length carefully
              //ARV 4/20
              //pBP_RHyp->m_SLScore = pBP_New->m_SLScore;

              TraceIf( s_TraceOn, 1, "    BP till here length " << pBP_New->m_TargetLength );
              pBP_RHyp->m_TargetLength = pBP_New->m_TargetLength + TargetLength( pRecombinedHyp ) - NewBPTargetExtension;
              TraceIf( s_TraceOn, 1, "    TargetLength is now " << pBP_RHyp->m_TargetLength );
              pBP_RHyp->m_pHyp = pRecombinedHyp;           // the hyp from which this one originated
              //			  pBP_RHyp->m_pEdge = pRecombinedHyp->m_pEdge;  // the edger over which the previous hyp was expanded
              pBP_RHyp->m_pBP   = pBP;                      // and connecting the backpointers
	  
              // if new coverage = 0, i.e. we are at the sentence start, then
              // the true sentence length is calculated and the correct SLScore
              // is estimated
              if ( PredecessorCoverage == 0 ) {
                TraceIf( s_TraceOn, 1, " Reached sentence end " );
                //UpdateSLScore( pBP_RHyp, n );
                UpdateWCScore( pBP_RHyp, n );
                //Sanjika 04/25
                //UpdateLexicalSLScore(pBP_RHyp, n);
              }

              TraceIf( s_TraceOn, 0, "  Going from Hyp " << pRecombinedHyp << " back to Hyp " << pBP_New->m_pHyp );
              TraceIf( s_TraceOn, 0, "  Insert into list of BPs at Coverage " << PredecessorCoverage );
              assert( pBP_New->m_pHyp->m_Coverage == PredecessorCoverage );

              // we don't want the same hyp several times
              //assert( pRecombinedHyp != LastHyp );
              //LastHyp = pRecombinedHyp;

              //Display( pBP_RHyp, -1 );
              //Translation = TraceBack( pBP_RHyp->m_pHyp ) + " # " + TraceForward( pBP_RHyp );
              //cout << "Intermediate rec. Translation: " << j << " " << Translation << endl;
              m_NBestHyps[ PredecessorCoverage ].insert
                (pair< float, CDecoderBLM_BP * >( pBP_RHyp->m_TotalScore, pBP_RHyp ) ); 

            } //( TrueScore > WorstScore )
			
          } // for ( IterRecombinedHyps
		  
        } // ( pPrevHyp->m_pRecombinedHyps )
        // now set the length extension for the initial best hypothesis
		
      }

    } // for ( IterOverBPs 

    j--;
  }


  // -----------------------------------------------------------------------------
  // now we need to reconstruct the actual translations
  // -----------------------------------------------------------------------------

  // loop through list of backpointers up to N

  n = 0;

  // if you are tracing
  s_TraceOn = false;

  if (s_TraceOn){
    TraceIf( s_TraceOn, 0, "  Final List of Backpointers " );

    multimap< float, CDecoderBLM_BP * > &CurrentBPs = m_NBestHyps[ 0 ];
    for ( IterOverBPs = CurrentBPs.begin(); 
          (IterOverBPs != CurrentBPs.end()) && (n < m_NumOfNBest * c_NumOfNBestMultiplier);
          ++IterOverBPs ) {
      
      CDecoderBLM_BP *pBP = IterOverBPs->second;
      
      TraceIf( 1, 0, "\nBackpointer " << n << " " << IterOverBPs->first
               << " TotalScore=" << pBP->m_TotalScore );
      
      //cout << endl << "NBest " << n << endl;
      TraceStringAndPath( pBP, n );
      
      n++;
    }
    
    
    cout << "Number of true N-best translations is: " << n << endl;
    
    Trace( 1, "Number of BPs is " << CDecoderBLM_BP::s_NumberOfObjects++ );
  }
  
  s_TraceOn = false;

}




// ARV added to get the translation back
// ARV 04/28
pair<CString, CArray< CLatticeEdgeMS * > >
CDecoderBLM::TraceStringAndPath( CDecoderBLM_BP * pBP, int n )
{
  s_TraceOn = false;
  CString Translation;
  CArray< CLatticeEdgeMS * > AllEdges;

  // length of source sentence to calculate per word scores
  int Start = m_pLattice->StartNode()->Number();
  int End   = m_pLattice->EndNode()->Number();
  int L = End - Start;


  if (s_TraceOn){
    /*
    cout << "Total Score  " << pBP->m_TotalScore
         << "  TMScore "    << pBP->m_scaledModelScores[ c_TM1 ]
         << "  LMScore "    << pBP->m_scaledModelScores[ c_LM1 ]
         << "  ALScore "    << pBP->m_scaledModelScores[ c_AL1 ]
         << "  SLScore "    << pBP->m_scaledModelScores[ c_SL1 ]
         << "  PLScore "    << pBP->m_scaledModelScores[ c_PL1 ]
         << endl;
    
    cout << "Length "     << pBP->m_TargetLength << "  Reordering Jumps " << pBP->m_pHyp->m_NumReorderings << endl;
    */
  }

  //Translation = "";
  //AllEdges.RemoveAll();
  
  // follow the pBP to access the edges and thereby the actual translation

  // backpointer points to initial hypothesis which has no edge
  // there, move to next backpointer
  pBP = pBP->m_pBP;

  while ( pBP->m_pHyp->m_pEdge ) {
	
    // to display the path, collect all edges
    AllEdges.Append( pBP->m_pHyp->m_pEdge );
	
    // access to the partial translations via the WordInfo
    CWordInfo *pWordInfo = pBP->m_pHyp->m_pEdge->WordInfo();
	
    // extract sequence of words from the WordInfo structure
    Translation += " ";
    if ( pWordInfo->HasTargetString() ) {
      Translation += pWordInfo->TargetString();
    } else if ( pWordInfo->HasSourceString() ) {
      Translation += pWordInfo->SourceString();
    } else {
      Translation += pBP->m_pHyp->m_pEdge->Word();
    }
	
    pBP = pBP->m_pBP;
  }
  
  // now we are at the sentence end again, i.e. we can access 
  // the hypothesis to get all interesting scores
  // well, need to go one additional step, to the hypothesis
  // which has also the sentence end LM score and sentence
  // length model score;
  //pHyp = pBP->m_pBP->m_pHyp;
  
  
  if (s_TraceOn){
    cout << "PartialPath " << n << " is: ";
    m_pLattice->DisplayPath( AllEdges );
    cout << "True NBest Translation " << n << " is:" << Translation << endl;
  }
  return pair<CString, CArray< CLatticeEdgeMS * > >(Translation, AllEdges);
}




CString
CDecoderBLM::TraceForward( CDecoderBLM_BP * pBP )
{
  CString Translation;
  CArray< CLatticeEdge * > AllEdges;

  // length of source sentence to calculate per word scores
  int Start = m_pLattice->StartNode()->Number();
  int End   = m_pLattice->EndNode()->Number();
  int L = End - Start;


  // follow the pBP to access the edges and thereby the actual translation
  Translation = "";
  AllEdges.RemoveAll();

  // edges are leading to the hyps the backpointer is pointing to
  // i.e. go to the next backpointer
  if ( pBP->m_pBP ) {

    pBP = pBP->m_pBP;

    while ( pBP->m_pHyp->m_pEdge ) {

      // to display the path, collect all edges
      AllEdges.Append( pBP->m_pHyp->m_pEdge );
	  
      // access to the partial translations via the WordInfo
      CWordInfo *pWordInfo = pBP->m_pHyp->m_pEdge->WordInfo();
	  
      //cout << " | " << pBP->m_pEdge->WordInfo()->TargetString();
	  
      // extract sequence of words from the WordInfo structure
      Translation += " ";
      if ( pWordInfo->HasTargetString() ) {
        Translation += pWordInfo->TargetString();
      } else if ( pWordInfo->HasSourceString() ) {
        Translation += pWordInfo->SourceString();
      } else {
        Translation += pBP->m_pHyp->m_pEdge->Word();
      }
	  
      pBP = pBP->m_pBP;

    }
  }

  return Translation;

}




void
CDecoderBLM::DisplayAllHypsWithCoverage( int Coverage, int N )
{

  CDecoderBLM_Hyp *pHyp;

  // first we fetch the set of hypotheses which all have coverage Coverage
  // and we need an iterater to access the hyps in this set
  HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual > &Hyps = m_AllHyps[ Coverage ];

  // and we need an iterater to access the hyps in this set
  HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual >::iterator IterHyps;


  TraceIf( s_TraceOn, 0, "\n\n  Number of hyps with coverage " 
           << Coverage << " is " << Hyps.size() );
  //TraceIf( s_TraceOn, 0, "Address of Hyps is: " << &Hyps );

  // display only top N hyps
  int n = 0;
  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {
    if ( n < N ) {
      pHyp = *IterHyps;
      cout << pHyp->m_TotalScore << " " << pHyp->m_Coverage << " " 
           << pHyp->m_CoverageStart << " " << pHyp->m_CoveragePattern << endl;
      n++;
    }
  }

}



//handles the multiple score transducer. The first score goes in the traditional c_TM position
//and all other scores go after the first model index. 
void
CDecoderBLM::GetTMScoreMS( CLatticeEdgeMS *pEdge )
{
  //  CString Word;
  double TMScore;

  if ( pEdge->WordInfo()->HasTargetString() ) {
    //cout<<"Set here"<<endl;
    
    /*
      if ( m_ScalingFactors.Size() > c_TM1 + ((CLatticeEdgeMS*)pEdge)->Scores().Size() ) {
      // if we have not enough socres, we supply additional scores
      cout << "Mismatch in number of models and scaling factors: ScalingFactorsSize: " << m_ScalingFactors.Size() << endl;
      ((CLatticeEdgeMS*)pEdge)->Display();
      cout << endl;
      exit ( -11 );
      }
    */

    CNumericalArray<float> PhraseScores = ((CLatticeEdgeMS*)pEdge)->Scores();

    //we are going to put the models in order, the first one starting at c_FirstModelIndex and continuing on down
    int offset = 0;
    for ( int i = PhraseScores.MinIndex(); i <= PhraseScores.MaxIndex(); i++ ) {
      /*      
      m_pNumberOneHypothesis->m_scaledModelScores[ offset + c_TM1 ]
        += m_ScalingFactors[ c_TM1 + offset ] * PhraseScores[ i ];
      
      m_pNumberOneHypothesis->m_TotalScore
        += m_ScalingFactors[ c_TM1 + offset ] * PhraseScores[ i ];
      */

      float Score = m_ScalingFactors[ m_StartTMScores + offset ] * PhraseScores[ i ];
      m_pNumberOneHypothesis->m_scaledModelScores[ m_StartTMScores + offset ] += Score;
      m_pNumberOneHypothesis->m_TotalScore += Score;

      offset++;
    }
    
  } else {

    //ARV 04/28 N.B the source word probabilities for all the true probability modesl should be set to approx -log (0.0) = 10
    //no target string, need to increase the cost of all the translation model factors

    //still start from the second phrase index forward, but dont do the last one, since its a count variable
    //we dont know how many phrase score there are, so we assume consistency with the number of scaling factors
    int offset = 0;
    //for ( int i = c_TM1; i < m_ScalingFactors.Size(); i++ ) {
    for ( int i = m_StartTMScores; i < m_ScalingFactors.Size(); i++ ) {
      /*
      m_pNumberOneHypothesis->m_scaledModelScores[ offset + c_TM1 ]
        += m_ScalingFactors[ c_TM1 + offset ] * c_TMScoreForSourceWord;
      
      m_pNumberOneHypothesis->m_TotalScore
        += m_ScalingFactors[ c_TM1 + offset ] * c_TMScoreForSourceWord;
      */

      float Score = m_ScalingFactors[ m_StartTMScores + offset ] * c_TMScoreForSourceWord;
      m_pNumberOneHypothesis->m_scaledModelScores[ m_StartTMScores + offset ] += Score;
      m_pNumberOneHypothesis->m_TotalScore += Score;

      offset++;
    }
    
    //pEdge->Display();
    
  }
}



/*
void
CDecoderBLM::GetLMScore( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge )
{

  CNumericalArray< CLMState > LMStates = pHyp->m_LMStates;
  CNumericalArray< CLMState > NewLMStates;
  double   LMScore = 0.0;

  // strings have alreay been converted into indices according to the LM vocabulary
  CNumericalArray< long > & WordIndices2 = pEdge->LMWordIndices();

  // now get the LM score by looping over all word indices
  // copy word indices into STL vector, as this is needed by the LM code
  vector< LMWordIndex > x;
  for ( int i = WordIndices2.MinIndex(); i <= WordIndices2.MaxIndex(); i++ ) {
    x.push_back( WordIndices2[ i ] );
  }

  // get probability and new LM state for the entire target phrase
  // sv: modify for multiple LMs
  //LMScore = -m_pLMSentenceState->LogProb( LMState, x, NewLMState ); 
  LMScore = -m_LMSentenceStates[ m_LMSentenceStates.MinIndex() ]
    ->LogProb( LMStates[ LMStates.MinIndex() ], x, NewLMStates[ NewLMStates.MinIndex() ] ); 
  //cout << "One step " << LMScore << " " << NewLMState << endl;
    

  // write information into hypothesis
  //m_pNumberOneHypothesis->m_scaledModelScores[ c_LM1 ] += m_ScalingFactors[ c_LM1 ] * LMScore;
  //m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ c_LM1 ] * LMScore;

  m_pNumberOneHypothesis->m_scaledModelScores[ m_StartLMScores ] += m_ScalingFactors[ m_StartLMScores ] * LMScore;
  m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ m_StartLMScores ] * LMScore;


  m_pNumberOneHypothesis->m_LMStates = NewLMStates;

}
*/



void
CDecoderBLM::GetLMScores( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge )
{

  CNumericalArray< CLMState > LMStates = pHyp->m_LMStates;
  CNumericalArray< CLMState > NewLMStates;
  double   LMScore = 0.0;

  // strings have alreay been converted into indices according to the LM vocabulary
  //CNumericalArray< long > & WordIndices2 = pEdge->LMWordIndices();

  // we loop over all LMs
  int OffsetForLMScore = 0;
  int FirstRow =  pEdge->WordIndicesForLMs().MinIndex1();

  for ( int l = m_LMSentenceStates.MinIndex(); l <= m_LMSentenceStates.MaxIndex(); l++ ) {

    CNumericalArray< long > & WordIndices2 = pEdge->WordIndicesForLMs().GetRow( FirstRow + OffsetForLMScore );

    // now get the LM score by looping over all word indices
    // copy word indices into STL vector, as this is needed by the LM code
    vector< LMWordIndex > x;
    for ( int i = WordIndices2.MinIndex(); i <= WordIndices2.MaxIndex(); i++ ) {
      x.push_back( WordIndices2[ i ] );
    }

    // get probability and new LM state for the entire target phrase
    // sv: modify for multiple LMs
    //LMScore = -m_pLMSentenceState->LogProb( LMState, x, NewLMState ); 

    LMScore = -m_LMSentenceStates[ l ]
      ->LogProb( LMStates[ l ], x, NewLMStates[ l ] );
    //cout << "One step " << LMScore << " " << NewLMState << endl;
    

    // write information into hypothesis
    //m_pNumberOneHypothesis->m_scaledModelScores[ c_LM1 ] += m_ScalingFactors[ c_LM1 ] * LMScore;
    //m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ c_LM1 ] * LMScore;

    m_pNumberOneHypothesis->m_scaledModelScores[ m_StartLMScores + OffsetForLMScore ]
      += m_ScalingFactors[ m_StartLMScores + OffsetForLMScore ] * LMScore;
    m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ m_StartLMScores + OffsetForLMScore ] * LMScore;

    OffsetForLMScore++;
  }

  // and copy over the new LM states
  m_pNumberOneHypothesis->m_LMStates = NewLMStates;

}



void
CDecoderBLM::GetLMScoresEOS( CDecoderBLM_Hyp *pHyp )
{

  CNumericalArray< CLMState > LMStates = pHyp->m_LMStates;
  CNumericalArray< CLMState > NewLMStates;
  double   LMScore = 0.0;

  // get probability and new LM state for the entire target phrase
  // sv: modify for multiple LMs
  //LMScore = -m_pLMSentenceState->LogProb( LMState, x, NewLMState ); 

  int OffsetForLMScore = 0;
  for ( int i = m_LMSentenceStates.MinIndex(); i <= m_LMSentenceStates.MaxIndex(); i++ ) {

    // cost for sentence end from the language model
    LMScore = -m_LMSentenceStates[ i ]->LogProbEnd( pHyp->m_LMStates[ i ] );

    pHyp->m_scaledModelScores[ m_StartLMScores + OffsetForLMScore ]
      += m_ScalingFactors[ m_StartLMScores + OffsetForLMScore ] * LMScore;
    pHyp->m_TotalScore += m_ScalingFactors[ m_StartLMScores  + OffsetForLMScore ] * LMScore;
 
    TraceIf( s_TraceOn, 1, "\n  LM cost for sentence end " << LMScore );
    

    //LMScore = -m_LMSentenceStates[ i ]
    //->LogProb( LMStates[ i ], x, NewLMStates[ i ] );
    //cout << "One step " << LMScore << " " << NewLMState << endl;

    OffsetForLMScore++;
  }

}



void
CDecoderBLM::GetALScore( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge )
{
  bool TraceOn = false;

  // small extra cost for reordering
  // when and how do we know that we are reordering?
  // when last edge and current edge are not sequential
  // we ended in pHyp->m_pEdge->ToNode() and start from pEdge->FromNode()
  // if these two nodes are the same, then monoton decoding
  // otherwise a jump

  // currently, each jumps incures the same small cost
  // a proper position alignment model should be used.

  //cout << endl << "CDecoderBLM::GetALScore" << endl;

  double ALScore;
  int NumberOfPrevNode;


  // the initial hypothesis has no edge, but we know that the node 
  // is the StartNode of the lattice
  /*
    if ( !pHyp->m_pEdge ) {
    NumberOfPrevNode = 0;
    } else {
    NumberOfPrevNode = pHyp->m_pEdge->ToNode()->Number();
    }

  
    if ( pEdge->FromNode()->Number() != NumberOfPrevNode ) {
    m_pNumberOneHypothesis->m_NumReorderings++;
    if ( m_pNumberOneHypothesis->m_NumReorderings > m_pNumberOneHypothesis->s_MaxNumReorderings ) {
	  ALScore = 100;
    } else {
    ALScore = 1.0;
    }
    } else {
    ALScore = 0.0;
    }
  */


  // here we have a simple jump model:
  // Forward: Prev->ToNode - 1 -> Next->FromNode
  // Backward: Prev->ToNode -> Next->FromNode + 1
  // pHyp has pointer to last traversed edge, pEdge is newly travesed edge



  // general distortion models: jump or distance depending on one or two words
  int JumpStart, JumpEnd, Jump;

  // we start from
  if ( !pHyp->m_pEdge ) {
    JumpStart = 0;
  } else {
    JumpStart = pHyp->m_pEdge->ToNode()->Number();
  }
  
  // we jump to (position of word is node number plus 1)
  JumpEnd = pEdge->FromNode()->Number() + 1;

  // the jump width is
  Jump = JumpEnd - JumpStart;
  
  // on the English side we go from number of words generated so far to the next
  int TargetLength = pHyp->m_TargetLength;

  // first get all the words which are involved in a jump
  // ie. start source, start target, end source, end target
  // We get a hypothesis and an edge, the hypothesis is expanded over the edge.
  // This means that we get information from the hyp about where we come from
  // and from the edge about where we go to.
  // StartOfJump is the last word of the edge we traversed last time.
  // This is in the Hyp.Edge.WordInfo source and target words.
  // However, for the very first hyp we do not have an edge, this is therefore a special case.

  // if we don't have the information neccessary for a lexicalized distortion model 
  // the we back-off to a background model

  
  bool UseBackgroundModel = false;  
  CString FromWordSource, FromWordTarget, ToWordSource, ToWordTarget;
  CWordIndex IndexFromWordSource, IndexFromWordTarget, IndexToWordSource, IndexToWordTarget;

  // first, we collect information about the words to the 'left', i.e. from where the jumps starts
  if ( !pHyp->m_pEdge ) {  // we will not find any information about specific words, return unlexicalized DM prob
    ;//cout << "no previous edge available" << endl;
    UseBackgroundModel = true;
  } else {
    //cout << "previous edge ";
    if ( TraceOn ) 
      pHyp->m_pEdge->Display();

    if ( pHyp->m_pEdge->WordInfo()->HasTargetString() ) {
      CSentence &FromWordsSource = pHyp->m_pEdge->WordInfo()->SourceWords();
      CSentence &FromWordsTarget = pHyp->m_pEdge->WordInfo()->TargetWords();
      FromWordSource = FromWordsSource.Word(  FromWordsSource.MaxIndex() );
      FromWordTarget = FromWordsTarget.Word(  FromWordsTarget.MaxIndex() );
      if ( TraceOn ) {
        cout << "From Source words: "; FromWordsSource.Display( m_pVocabulary ); cout << endl;
        cout << "From Target words: "; FromWordsTarget.Display( m_pVocabulary ); cout << endl;
        cout << "From word source " << FromWordSource << " ";
        cout << "From word target " << FromWordTarget << endl;
      }
      //FromWordSource = FromWordsSource[ FromWordsSource.MaxIndex() ];
      //FromWordTarget = FromWordsTarget[ FromWordsTarget.MaxIndex() ];
    } else {
      // use the source word also as target word
      FromWordSource = pHyp->m_pEdge->Word();
      FromWordTarget = pHyp->m_pEdge->Word();
      if ( TraceOn ) {
        cout << "No annotatioin -> is source word" << endl;
        cout << "From Source word: " << pHyp->m_pEdge->Word() << endl;
        cout << "From Target word: " << pHyp->m_pEdge->Word() << endl;
      }
    }
  }


  // now we collect information about the words to the 'right', i.e. where the jump ends
  if ( !pEdge ) {
    ; //cout << "no current edge available";
    UseBackgroundModel = true;
  } else {
    ; //cout << "current edge ";
    if ( TraceOn ) 
      pEdge->Display();
    
    if ( pEdge->WordInfo()->HasTargetString() ) {

      CSentence &ToWordsSource = pEdge->WordInfo()->SourceWords();
      CSentence &ToWordsTarget = pEdge->WordInfo()->TargetWords();

      ToWordSource = ToWordsSource.Word(  ToWordsSource.MinIndex() );
      ToWordTarget = ToWordsTarget.Word(  ToWordsTarget.MinIndex() );

      if ( TraceOn ) {
        cout << "To Source words: "; ToWordsSource.Display( m_pVocabulary ); cout << endl;
        cout << "To Target words: "; ToWordsTarget.Display( m_pVocabulary ); cout << endl;
        cout << "To word source " << ToWordSource << " ";
        cout << "To word target " << ToWordTarget << endl;
      }
      //ToWordSource  = ToWordsSource[ ToWordsSource.MinIndex() ];
      //ToWordTarget  = ToWordsTarget[ ToWordsTarget.MinIndex() ];

    } else {
      // use the source word also as target word
      ToWordSource = pEdge->Word();
      ToWordTarget = pEdge->Word();
      if ( TraceOn ) {
        cout << "No annotatioin -> is source word" << endl;
        cout << "To Source word: " << pEdge->Word() << endl;
        cout << "To Target word: " << pEdge->Word() << endl;
      }
    }

  }

  if ( UseBackgroundModel ) {
    ; //cout << "Use Background Model" << endl;
  } else {
    IndexFromWordSource = m_pVocabulary->GetIndex( FromWordSource );
    IndexFromWordTarget = m_pVocabulary->GetIndex( FromWordTarget );
    IndexToWordSource = m_pVocabulary->GetIndex( ToWordSource );
    IndexToWordTarget = m_pVocabulary->GetIndex( ToWordTarget );
    if ( 0 ) {
      cout << "In GetALScore: Jump is " << endl;
      cout << "from source " << JumpStart        << " " << FromWordSource;
      cout << "  to source " << JumpEnd          << " " << ToWordSource << endl;
      cout << "from target " << TargetLength     << " " << FromWordTarget;
      cout << "  to target " << TargetLength + 1 << " " << ToWordTarget << endl;
    }
    if ( 0 ) {
      cout << "In GetALScore: Jump is ";
      cout << "from ( " << FromWordSource << " : " << FromWordTarget;
      cout << " ) to ( " << ToWordSource << " : " << ToWordTarget << " ) " << endl;
    }
  }
  

  // if we don't have a distortion model loaded, we use the old simple model
  if ( !m_pDistortionModel ) { 
    //ORIGINAL Gaussian scoring 
    ALScore = 0.25 * ( Jump - 1.0 ) * ( Jump - 1.0 );
  } else { 
    // otherwise we call a jump distortion model with all 4 words
    ALScore = -m_pDistortionModel->LogProb( Jump,
                                            IndexFromWordSource, IndexFromWordTarget, 
                                            IndexToWordSource, IndexToWordTarget );
  }


  // test if we have already too many reorderings; if so, set ALScore very high;
  if ( pEdge->FromNode()->Number() != JumpStart ) {
    m_pNumberOneHypothesis->m_NumReorderings++;
	
    if ( m_pNumberOneHypothesis->m_NumReorderings > m_pNumberOneHypothesis->m_pDecoderInfo->m_MaxNumReorderings ) {
      ALScore = 100;
    }
   }
  
  //cout << "Jump is " << Jump << " ALScore " << ALScore << "NumRe:"<<m_pNumberOneHypothesis->m_NumReorderings<< endl;

  //m_pNumberOneHypothesis->m_scaledModelScores[ c_AL1 ] += m_ScalingFactors[ c_AL1 ] * ALScore;
  //m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ c_AL1 ] * ALScore;

  m_pNumberOneHypothesis->m_scaledModelScores[ m_StartDMScores ] += m_ScalingFactors[ m_StartDMScores ] * ALScore;
  m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ m_StartDMScores ] * ALScore;

  TraceOn = false;
}



void
CDecoderBLM::GetDMScores( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge )
{
  bool TraceOn = false;

  /*
    We may have 0, 1, or several DMs loaded.  We always use the simple jump model as DM0.
    If DMs have been loaded these are used as DM1, DM2, etc in addition to the simple jump model.
    Each distortion model can do it's own smooting.
  */


  // When and how do we know that we are reordering?
  // When last edge and current edge are not sequential.
  // We ended in pHyp->m_pEdge->ToNode() and start from pEdge->FromNode().
  // Ff these two nodes are the same, then monoton decoding, otherwise a jump


  //cout << endl << "CDecoderBLM::ApplyDistortionModels" << endl;

  double DMScore;
  int NumberOfPrevNode;

  // general distortion models: jump or distance depending on one or two words
  int JumpStart, JumpEnd, Jump;

  // we start from
  if ( !pHyp->m_pEdge ) {
    JumpStart = 0;
  } else {
    JumpStart = pHyp->m_pEdge->ToNode()->Number();
  }
  
  // we jump to (position of word is node number plus 1)
  JumpEnd = pEdge->FromNode()->Number() + 1;

  // the jump width is
  Jump = JumpEnd - JumpStart;
  

  // we still use the simple old model
  DMScore = m_ScalingFactors[ m_StartDMScores ] * 0.25 * ( Jump - 1.0 ) * ( Jump - 1.0 );

  // test if we have already too many reorderings; if so, set DMScore very high;
  if ( pEdge->FromNode()->Number() != JumpStart ) {
    m_pNumberOneHypothesis->m_NumReorderings++;
    
    if ( m_pNumberOneHypothesis->m_NumReorderings > m_pNumberOneHypothesis->m_pDecoderInfo->m_MaxNumReorderings ) {
      DMScore = 100;
    }
  }
  
  //cout << "Jump is " << Jump << " DMScore background model " << DMScore << "NumRe:"<<m_pNumberOneHypothesis->m_NumReorderings<< endl;

  m_pNumberOneHypothesis->m_scaledModelScores[ m_StartDMScores ] += DMScore;
  m_pNumberOneHypothesis->m_TotalScore += DMScore;



  // now lexicalized and specific models, for which we need to collect the words first

  // on the English side we go from number of words generated so far to the next
  //int TargetLength = pHyp->m_TargetLength;  // not used right now

  // first get all the words which are involved in a jump
  // ie. start source, start target, end source, end target
  // We get a hypothesis and an edge, the hypothesis is expanded over the edge.
  // This means that we get information from the hyp about where we come from
  // and from the edge about where we go to.
  // StartOfJump is the last word of the edge we traversed last time.
  // This is in the Hyp.Edge.WordInfo source and target words.
  // However, for the very first hyp we do not have an edge, this is therefore a special case.

  // if we don't have the information neccessary for a lexicalized distortion model 
  // the we back-off to a background model

  
  bool UseBackgroundModel = false;  
  CString FromWordSource, FromWordTarget, ToWordSource, ToWordTarget;
  CWordIndex IndexFromWordSource, IndexFromWordTarget, IndexToWordSource, IndexToWordTarget;

  // first, we collect information about the words to the 'left', i.e. from where the jumps starts
  if ( !pHyp->m_pEdge ) {  // we will not find any information about specific words, return unlexicalized DM prob
    ;//cout << "no previous edge available" << endl;
    UseBackgroundModel = true;
   } else {
    //cout << "previous edge ";
    if ( TraceOn ) 
      pHyp->m_pEdge->Display();

    if ( pHyp->m_pEdge->WordInfo()->HasTargetString() ) {
      CSentence &FromWordsSource = pHyp->m_pEdge->WordInfo()->SourceWords();
      CSentence &FromWordsTarget = pHyp->m_pEdge->WordInfo()->TargetWords();
      FromWordSource = FromWordsSource.Word(  FromWordsSource.MaxIndex() );
      FromWordTarget = FromWordsTarget.Word(  FromWordsTarget.MaxIndex() );
      if ( TraceOn ) {
        cout << "From Source words: "; FromWordsSource.Display( m_pVocabulary ); cout << endl;
        cout << "From Target words: "; FromWordsTarget.Display( m_pVocabulary ); cout << endl;
        cout << "From word source " << FromWordSource << " ";
        cout << "From word target " << FromWordTarget << endl;
      }
      //FromWordSource = FromWordsSource[ FromWordsSource.MaxIndex() ];
      //FromWordTarget = FromWordsTarget[ FromWordsTarget.MaxIndex() ];
    } else {
      // use the source word also as target word
      FromWordSource = pHyp->m_pEdge->Word();
      FromWordTarget = pHyp->m_pEdge->Word();
      if ( TraceOn ) {
        cout << "No annotatioin -> is source word" << endl;
        cout << "From Source word: " << pHyp->m_pEdge->Word() << endl;
        cout << "From Target word: " << pHyp->m_pEdge->Word() << endl;
      }
    }
  }


  // now we collect information about the words to the 'right', i.e. where the jump ends
  if ( !pEdge ) {
    ; //cout << "no current edge available";
    UseBackgroundModel = true;
  } else {
    ; //cout << "current edge ";
    if ( TraceOn ) 
      pEdge->Display();
    
    if ( pEdge->WordInfo()->HasTargetString() ) {

      CSentence &ToWordsSource = pEdge->WordInfo()->SourceWords();
      CSentence &ToWordsTarget = pEdge->WordInfo()->TargetWords();

      ToWordSource = ToWordsSource.Word(  ToWordsSource.MinIndex() );
      ToWordTarget = ToWordsTarget.Word(  ToWordsTarget.MinIndex() );

      if ( TraceOn ) {
        cout << "To Source words: "; ToWordsSource.Display( m_pVocabulary ); cout << endl;
        cout << "To Target words: "; ToWordsTarget.Display( m_pVocabulary ); cout << endl;
        cout << "To word source " << ToWordSource << " ";
        cout << "To word target " << ToWordTarget << endl;
      }
      //ToWordSource  = ToWordsSource[ ToWordsSource.MinIndex() ];
      //ToWordTarget  = ToWordsTarget[ ToWordsTarget.MinIndex() ];

    } else {
      // use the source word also as target word
      ToWordSource = pEdge->Word();
      ToWordTarget = pEdge->Word();
      if ( TraceOn ) {
        cout << "No annotatioin -> is source word" << endl;
        cout << "To Source word: " << pEdge->Word() << endl;
        cout << "To Target word: " << pEdge->Word() << endl;
      }
    }

  }

  if ( UseBackgroundModel ) {
    //cout << "Use Background Model" << endl;
    IndexFromWordSource = m_pVocabulary->UNK;
    IndexFromWordTarget = m_pVocabulary->UNK;
    IndexToWordSource   = m_pVocabulary->UNK;
    IndexToWordTarget   = m_pVocabulary->UNK;
    FromWordSource      = m_pVocabulary->GetStringForUNK();
    FromWordTarget      = m_pVocabulary->GetStringForUNK();
    ToWordSource        = m_pVocabulary->GetStringForUNK();
    ToWordTarget        = m_pVocabulary->GetStringForUNK();
  } else {
    IndexFromWordSource = m_pVocabulary->GetIndex( FromWordSource );
    IndexFromWordTarget = m_pVocabulary->GetIndex( FromWordTarget );
    IndexToWordSource   = m_pVocabulary->GetIndex( ToWordSource );
    IndexToWordTarget   = m_pVocabulary->GetIndex( ToWordTarget );
  }
  

  if ( 0 ) {
    cout << "In ApplyDistortionModels: Jump is ";
    cout << "from ( " << FromWordSource << " : " << FromWordTarget;
    cout << " ) to ( " << ToWordSource << " : " << ToWordTarget << " ) " << endl;
  }

  // if we have additional DMs
  if ( m_DistortionModels.Size() > 0 ) { 
    int CountDMs = 1;   // we have already used the background model 
    for ( int n = m_DistortionModels.MinIndex(); n <= m_DistortionModels.MaxIndex(); n++ ) {
      // get the DMScore from the next DM
      //cout << "In ApplyDistortionModels: Jump is ";
      //cout << "from ( " << FromWordSource << " : " << FromWordTarget;
      //cout << " ) to ( " << ToWordSource << " : " << ToWordTarget << " ) " << endl;

      DMScore = -m_DistortionModels[ n ]->LogProb( Jump,
                                                   IndexFromWordSource, IndexFromWordTarget, 
                                                   IndexToWordSource, IndexToWordTarget );
      // scale score
      //cout << "DM " << CountDMs << ": Scores = " << DMScore;
      DMScore *= m_ScalingFactors[ m_StartDMScores + CountDMs ];
      //cout << "   ScaledScore = " << DMScore << endl;
      // store in hyp
      m_pNumberOneHypothesis->m_scaledModelScores[ m_StartDMScores + CountDMs ] += DMScore;
      // add it to total score
      m_pNumberOneHypothesis->m_TotalScore += DMScore;

      CountDMs++;  // go to the next DM
    }

  }

  //cout << "Jump is " << Jump << " DMScore " << DMScore << "NumRe:"<<m_pNumberOneHypothesis->m_NumReorderings<< endl;


  TraceOn = false;
}



void
CDecoderBLM::GetWCScore( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge )
{
  //cout << "GetWCScore" << endl;

  // we take the number of generate target words
  int NumTargetWords= TargetLength( pEdge );

  //cout << "GetWCScore " << NumTargetWords << endl;

  //m_pNumberOneHypothesis->m_scaledModelScores[ c_SL1 ] += m_ScalingFactors[ c_SL1 ] * NumTargetWords;
  //m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ c_SL1 ] * NumTargetWords;
  m_pNumberOneHypothesis->m_scaledModelScores[ m_StartWCScores ] += m_ScalingFactors[ m_StartWCScores ] * NumTargetWords;
  m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ m_StartWCScores ] * NumTargetWords;

  //cout << "GetWCScore " << m_pNumberOneHypothesis->m_TargetLength << " " << m_pNumberOneHypothesis->m_scaledModelScores[ c_SL ] << endl;
}



int
CDecoderBLM::TargetLength( CDecoderBLM_Hyp *pHyp )
{
  // length of translation
  int L = 0;
  if ( pHyp->m_pEdge ) {  // at sentence end we have no edge
    CWordInfo *pWordInfo = pHyp->m_pEdge->WordInfo();
    if ( pWordInfo->HasTargetString() ) {
      //cout << "has target string" << endl;
      L = pWordInfo->TargetStringLength();
    } else if ( pWordInfo->HasSourceString() ) {
      L = pWordInfo->SourceStringLength();
      //cout << "has source string" << endl;
    } else {
      //cout << "no source and no target string" << endl;
      L = 1;
    }
  } else {
    //cout << "no edge in get target length" << endl;
  }


  return L;
}



int
CDecoderBLM::TargetLength( CLatticeEdgeMS *pEdge )
{
  // length of translation
  int L = 0;
  if ( pEdge ) {  // at sentence end we have no edge
    CWordInfo *pWordInfo = pEdge->WordInfo();
    if ( pWordInfo->HasTargetString() ) {
      //cout << "has target string" << endl;
      L = pWordInfo->TargetStringLength();
    } else if ( pWordInfo->HasSourceString() ) {
      L = pWordInfo->SourceStringLength();
      //cout << "has source string" << endl;
    } else {
      //cout << "no source and no target string" << endl;
      L = 1;
    }
  } else {
    ;//cout << "no edge in get target length" << endl;
  }


  return L;
}



void
CDecoderBLM::GetPCScore( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge )
{
  //-log(2.718) = -1, but we just use +1 as a count and expect negative model scores
  //m_pNumberOneHypothesis->m_scaledModelScores[ c_PL1 ] += m_ScalingFactors[ c_PL1 ] * (1);
  //m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ c_PL1 ] * (1);
  m_pNumberOneHypothesis->m_scaledModelScores[ m_StartPCScores ] += m_ScalingFactors[ m_StartPCScores ] * (1);
  m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ m_StartPCScores ] * (1);
}

//////////////////////////////////////////////////////////////////////
//Nguyen, Aug 5, 2008: Cohesion Interruption Check
//
// interruption check: boolean version
// input: 
//   -  source-side dependency structure
//   -  previous phrase f_h
//   -  current phrase f_h+1 
//   -  current partial hypothesis: bit vector
//   -  the subtree that f_h+1 made interruption
//   -  algorithm: 1: original Cherry's 08; 
//                 2: check all subtrees
//                 3: original Cherry's 08 with InterruptionCount
//                 4: all substrees with InterruptionCount
//         InterruptionCount: how many words in a subtree havent been translated       
// output: true/false
//      true/1: translating phrase f_h+1 made cohesion violation
//      false/0: otherwise 
//      m_InterruptionCount
     

bool CDecoderBLM::InterruptionCheck (vector<tokens>& v, vector<tokens>& fh, vector<tokens>& fh1, vector<int>& HypCoverage, tokens& r, int algorithm ){
	bool interrupt = false;
	
	//tokens r;	
	tokens rprime;
	
	vector<tokens> check;
	
	m_InterruptionCount = 0;
	m_InterruptionNounCount = 0;
	m_InterruptionVerbCount = 0;
	
	//cout << "DEBUG - CDecoderBLM::InterruptionCheck ; algorithm = " << algorithm << endl; 
	if ( algorithm == 1 ||  algorithm == 3 || algorithm == 5 ){
		
		//left and right most word of fh
		tokens sl, sr;
		
		if(fh.size() > 0) {			
			sl = (*(fh.begin()));
			sr = (*(fh.end() - 1));
		} else {
			return interrupt; 
		}
		
		if(fh.size() > 1) {		
			check.push_back(sl); 
			check.push_back(sr);
		} else { 
			check.push_back(sl); 
		}
		
		//cout << "left-most= "<< sl.name << " | right-most= " << sr.name << endl;
	}
	else if (algorithm == 2 || algorithm == 4){
		vector<int> prevhyp = HypCoverage;
		for(vector<tokens>::size_type i = 0; i < fh1.size(); ++i)
			prevhyp[fh1[i].post] = 0;
		//check all tokens in partial hypothesis
		for(vector<tokens>::size_type i = 0; i < prevhyp.size(); ++i){
			if (prevhyp[i] != 0 ) check.push_back(v[i]);
				
			//cout << "PrevHyp= " << prevhyp[i] << endl;
		}	
	}	
		
//	for (vector<int>::const_iterator it = HypCoverage.begin(); it != HypCoverage.end(); ++it)
//		cout << (*it) << endl;
	
	vector<int> bCount; //make sure IC, NC, VC are not counted many times
	for(int k = 0; k < HypCoverage.size(); ++k){ bCount.push_back(0); } 
	
	for(vector<tokens>::size_type i = 0; i < check.size(); ++i){
		
		//cout << "check = " << check[i].name << endl;				
		
		r.name = "NULL";		
		rprime = check[i];
			
		vector<tokens>::size_type j = 0;		
		while( j < fh1.size() ){
//			cout << "DEBUG-1 r= " << r.name << "  rprime= "<< rprime.name << " rprime.post= " << rprime.post 
//			<< " v[rprime.post]= " << v[rprime.post].name << "  fh1[j]= " << fh1[j].name << endl;
			if ( !insubtree(fh1[j], rprime, v)){
				r = v[rprime.post];
				
				if (r.parentpost == -1) {
					rprime.name = "root";
					rprime.post = r.parentpost;
				} else {
					rprime = v[r.parentpost];
				}				
				
				j = 0; //restart checking 
								
//				cout << "DEBUG0 r= " << r.name << "  rprime= "<< rprime.name << "  fh1[j]= " << fh1[j].name << endl;
			}
			else { j++; }
		}
		
		//cout << "DEBUG1 r= " << r.name << "  rprime= "<< rprime.name << endl; 
		
		if (r.name != "NULL"){
			for(vector<tokens>::size_type k = 0; k < v.size(); ++k){
				//v[k] in subtree rooted at r, and havent been covered yet
				if ( insubtree(v[k], r, v) &&  ( HypCoverage[v[k].post] != 1 )) {
					
					interrupt = true;
					
					//cout << "DEBUG v[k] =  " << v[k].name << " bCount = " << bCount[v[k].post] << endl;
					
					if (algorithm == 3 || algorithm == 4 || algorithm == 5){
						if ( bCount[v[k].post] != 1){
							
							m_InterruptionCount ++;
							
							if ( algorithm == 5 && v[k].pos.substr(0,2) == "VB" ){ m_InterruptionVerbCount ++; }
							
							if ( algorithm == 5 && v[k].pos.substr(0,2) == "NN" ){ m_InterruptionNounCount ++; }
							
							bCount[v[k].post] = 1;
						}
									 
					} else {
						m_InterruptionCount = 1;
						return interrupt;						
					}
				}
				//cout << "DEBUG2 IC = " << m_InterruptionCount << "  VerbCount = " << m_InterruptionVerbCount << "  NounCount = " << m_InterruptionNounCount << endl;
			}			
		}				
	}
			
	return interrupt;
}
  
void
CDecoderBLM::GetCMScore( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge, vector<tokens> fh1, vector<tokens> fh )
{  
  
  //cout << "DEBUG - CDecoderBLM::GetCMScore" << endl;
    
  //subtreeR is the subtree where fh1 make interruption
  tokens subtreeR;

  //Source coverage vector
  string SrcCov;
  for(vector<int>::size_type i = 0; i < m_Hypcover.size(); ++i){
  	SrcCov += int2str(m_Hypcover[i]) ;
  }  
  string CurrPhrase = getPostFromVector(fh1); //current phrase  
  string PrevPhrase = getPostFromVector(fh);  //prev phrase  
  
//  cout << "DEBUG SrcCov = "<< SrcCov << "||  PrePhrase+CurrPhrase = " << PrevPhrase << CurrPhrase  << endl;  
//  
//  m_InterruptionCount = 0;  
//  bool ok1 = InterruptionCheck(m_Dtree, fh, fh1, m_Hypcover, subtreeR, m_InterruptionCheckAlgo);
//  int m_InterruptionCount_NonHash = m_InterruptionCount;     
//  cout << "DEBUG NON-HASH OK1 = " << ok1  << "   m_InterruptionCount = " << m_InterruptionCount_NonHash <<  "  fh= " <<  getNameFromVector(fh) << "   fh1 = " <<  getNameFromVector(fh1) << endl;
  
  
  //cout << "DEBUG GetCMScore  fh= " <<  getNameFromVector(fh) << "   fh1 = " <<  getNameFromVector(fh1) << endl;
  
  m_InterruptionCount = 0;
  bool ok = 0 ; //InterruptionCheck(m_Dtree, fh, fh1, m_Hypcover, subtreeR, m_InterruptionCheckAlgo);   
  
  CovEdg currCovEdg( SrcCov, PrevPhrase+CurrPhrase );
  
  map< CovEdg, vector<int> >::iterator it = m_CovEdg.find(currCovEdg);  
  
  if ( it == m_CovEdg.end() ){
  	
  	ok = InterruptionCheck( m_Dtree, fh, fh1, m_Hypcover, subtreeR, m_InterruptionCheckAlgo );  	
  	
  	vector <int> CohFeats;
  	CohFeats.push_back(m_InterruptionCount);
  	CohFeats.push_back(m_InterruptionVerbCount);
  	CohFeats.push_back(m_InterruptionNounCount);
  	
  	m_CovEdg.insert( pair< CovEdg, vector<int> >( currCovEdg, CohFeats ) );
  	  	  	  	
  }else {
  	
  	vector<int>::size_type i = 0;
  	if ( it->second.size() >  1){
  		m_InterruptionCount = it->second[0];
  		m_InterruptionVerbCount = it->second[1];
  		m_InterruptionNounCount = it->second[2];
  	}{
  		m_InterruptionCount = it->second[0];  		
  	}  	  	
  	
  	if (m_InterruptionCount > 0) { ok = 1;} else { ok = 0; }
  }  
  
  
//  map< CovEdg, int >::iterator it = m_CovEdg.find(currCovEdg);  
//  
//  if ( it == m_CovEdg.end() ){  	
//  	
//  	ok = InterruptionCheck( m_Dtree, fh, fh1, m_Hypcover, subtreeR, m_InterruptionCheckAlgo );  	
//  	m_CovEdg.insert( pair< CovEdg, int >( currCovEdg, m_InterruptionCount ) );
//  	
//  	//cout << "DEBUG not found !!!  value = " << m_InterruptionCount << " Cov = " << SrcCov << " || CurrentPhrase = "<< CurrPhrase << "  ||  PrevPhrase = " << PrevPhrase << endl;  	  	
//  }else {  	
//  	//CovEdg temp = it->first;  	
//  	//string mcov = temp.getcov();
//  	//string medg = temp.getedg();  	
//  	//cout << "DEBUG Found !!! ok = " << ok << " value = " << it->second <<  "  m_InterruptionCount = " <<  m_InterruptionCount << "   key = " << mcov << "||" << medg << endl;
//  	
//  	m_InterruptionCount = it->second;  	
//  	
//  	if (m_InterruptionCount > 0) { ok = 1;} else { ok = 0; }
//  }  
  
//  if (ok1 != ok){
//  	cout << "DEBUG NonHash ok1 = " << ok1 << "  Hash ok = " << ok  << endl;
//  	cout << "DEBUG NonHash count = " <<  m_InterruptionCount_NonHash << "   Hash count = " << m_InterruptionCount <<  "  fh= " <<  getNameFromVector(fh) << "   fh1 = " <<  getNameFromVector(fh1) << endl;
//  	cout << "-----------------" << endl;  
//  }
  
  //if violate cohesion then penalize
  if(ok == 1) { 
  	
  	if (m_InterruptionCheckAlgo == 1 || m_InterruptionCheckAlgo == 2){
  		//if fh1 interrupts source dependency tree then penalize
  		m_pNumberOneHypothesis->m_scaledModelScores[ m_StartCMScores ] += m_ScalingFactors[ m_StartCMScores ] * (1);
  		m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ m_StartCMScores ] * (1);
  		//cout << "[ok=1] Algo= " << m_InterruptionCheckAlgo << " Cohesion score: " << m_pNumberOneHypothesis->m_scaledModelScores[ m_StartCMScores ] << endl;
  	}
  	else if (m_InterruptionCheckAlgo == 3 || m_InterruptionCheckAlgo == 4){ 
  		m_pNumberOneHypothesis->m_scaledModelScores[ m_StartCMScores ] += m_ScalingFactors[ m_StartCMScores ] * m_InterruptionCount * (1) ;  		
  		m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ m_StartCMScores ] * m_InterruptionCount * (1);
  		//cout << "[ok=1] Algo= " << m_InterruptionCheckAlgo << " Cohesion score: " << m_pNumberOneHypothesis->m_scaledModelScores[ m_StartCMScores ] << endl;
  	}
  	else if (m_InterruptionCheckAlgo == 5 ){
  		//update InterruptionCheck weight
  		m_pNumberOneHypothesis->m_scaledModelScores[ m_StartCMScores ] += m_ScalingFactors[ m_StartCMScores ] * (1);
  		m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ m_StartCMScores ] * (1);
  		
  		//update Interruption Count weight
  		m_pNumberOneHypothesis->m_scaledModelScores[ m_StartICScores ] += m_ScalingFactors[ m_StartICScores ] * m_InterruptionCount * (1) ;  		
  		m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ m_StartICScores ] * m_InterruptionCount * (1);
  		
  		//update Interruption Noun Count weight
  		m_pNumberOneHypothesis->m_scaledModelScores[ m_StartNCScores ] += m_ScalingFactors[ m_StartNCScores ] * m_InterruptionNounCount * (1) ;  		
  		m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ m_StartNCScores ] * m_InterruptionNounCount * (1);
  		
  		//update Interruption Verb Count weight
  		m_pNumberOneHypothesis->m_scaledModelScores[ m_StartVCScores ] += m_ScalingFactors[ m_StartVCScores ] * m_InterruptionVerbCount * (1) ;  		
  		m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ m_StartVCScores ] * m_InterruptionVerbCount * (1);
  		
//  		cout << "ok=1; Algo= " << m_InterruptionCheckAlgo 
//  		<< " CM = " << m_pNumberOneHypothesis->m_scaledModelScores[ m_StartCMScores ] 
//  		<< " IC = " << m_pNumberOneHypothesis->m_scaledModelScores[ m_StartICScores ] 
//  		<< " NC = " << m_pNumberOneHypothesis->m_scaledModelScores[ m_StartNCScores ] 
//  		<< " VC = " << m_pNumberOneHypothesis->m_scaledModelScores[ m_StartVCScores ] << endl;
  	}
  }
}
////////////////// End of Nguyen //////////////////////

void
CDecoderBLM::UpdateSLScore( CDecoderBLM_BP *pBP, int &n )
{
  float TrueSLScore;
  int J = m_pLattice->EndNode()->Number();
  int I = pBP->m_TargetLength;
  // ARV
  // changed to reflect the simple couting score
  //TrueSLScore = m_SLScale * 0.25 * ( m_SLStretch * J - I ) * ( m_SLStretch * J - I );

  //TrueSLScore = m_ScalingFactors[ c_SL1 ] * I;
  TrueSLScore = m_ScalingFactors[ m_StartSLScores ] * I;

  //cout <<"Sentence Length " << I << " " << TrueSLScore << endl;

  /*
    s_TraceOn = false;
    TraceIf( s_TraceOn, 0, "    Old SLScore was " << pBP->m_SLScore
    << " New SLScore is " << TrueSLScore
    << " Correct length is " << pBP->m_TargetLength );
  */


  //pBP->m_TotalScore -= pBP->m_scaledModelScores[ c_SL1 ];   // subtract old SLScore  
  //pBP->m_scaledModelScores[ c_SL1 ] = TrueSLScore;
  pBP->m_TotalScore -= pBP->m_scaledModelScores[ m_StartSLScores ];   // subtract old SLScore  
  pBP->m_scaledModelScores[ m_StartSLScores ] = TrueSLScore;

  pBP->m_TotalScore += TrueSLScore;        // add new SLScore
}


void
CDecoderBLM::UpdateWCScore( CDecoderBLM_BP *pBP, int &n )
{
  int WC = pBP->m_TargetLength;
  float TrueWCScore = m_ScalingFactors[ m_StartWCScores ] * WC;

  //cout <<"Sentence Length " << I << " " << TrueSLScore << endl;

  /*
    s_TraceOn = false;
    TraceIf( s_TraceOn, 0, "    Old WCScore was " << pBP->m_WCScore
    << " New WCScore is " << TrueWCScore
    << " Correct length is " << pBP->m_TargetLength );
  */


  //pBP->m_TotalScore -= pBP->m_scaledModelScores[ c_SL1 ];   // subtract old SLScore  
  //pBP->m_scaledModelScores[ c_SL1 ] = TrueSLScore;
  pBP->m_TotalScore -= pBP->m_scaledModelScores[ m_StartWCScores ];   // subtract old WCScore  
  pBP->m_scaledModelScores[ m_StartWCScores ] = TrueWCScore;
  pBP->m_TotalScore += TrueWCScore;        // add new WCScore
}



//Sanjika 04/25
//Lexicalized SL model
/*
//ARV 04/28
//commented out until the lexicalized SL model has been tested and shown to give improvements

void CDecoderBLM::InitLexicalSLScoreEnd(){
CSentence NewSourceSentence;
for(int j=m_pSourceSentence->MinIndex(); j<=m_pSourceSentence->MaxIndex(); j++) {
NewSourceSentence.SetWord( m_pSourceSentence->Word(j,m_pVocabulary),j+1,m_pVocabulary);
}
m_pSLenModel->GenerateTargetLengthProbabilities( &NewSourceSentence );
}
double CDecoderBLM::GetLexicalSLScoreEnd(CDecoderBLM_Hyp* pHyp){
int I = pHyp->m_TargetLength;
double SLenProb = -log(m_pSLenModel->GetPr( I ));
return SLenProb;
}
  

void CDecoderBLM::UpdateLexicalSLScore(CDecoderBLM_BP *pBP, int &n ){
float TrueSLScore;
int I = pBP->m_TargetLength;
double SLenProb = m_pSLenModel->GetPr( I );
TrueSLScore =  m_ScalingFactors[c_LSL] *-log( SLenProb );

//remove contribution that was there before
pBP->m_TotalScore -= pBP->m_scaledModelScores[c_LSL];
//add correct contribution
pBP->m_scaledModelScores[c_LSL] = TrueSLScore;
pBP->m_TotalScore += TrueSLScore;  
  
}
*/





CDecoderBLM_Hyp *
CDecoderBLM::NewHypothesis()
{
  //ARV 4/20 updated to initialize the hyp with the appropriate number of models
  return new CDecoderBLM_Hyp( m_NumberOfModels );
}



RecombinedMap*
CDecoderBLM::NewRecombinedMap()
{
  RecombinedMap* pSavePtr = new RecombinedMap;
  m_RecombinedMapList.push_back(pSavePtr);
  return pSavePtr;
}





// not yet used
void
CDecoderBLM::GetLMLogProbsAllHypsWithCoverage( int Coverage )
{
  s_TraceOn = False;

  TraceIf( s_TraceOn, 0, "\n\n  CDecoderBLM::ExpandAllHypsWithCoverage " << Coverage );

  // variable we need
  CDecoderBLM_Hyp *pHyp, *pBestHyp;
  CLatticeNode *pNode;
  CLatticeEdgeMS *pEdge;



  // how long is the source sentence?  we need to cover all words
  int J = m_pLattice->EndNode()->Number();
  // should we use average path length for speech recognizer lattices?

  // hypotheses are organized according to their coverage
  // hypotheses of coverage c are expanded till coverage J-1
  // is reached.  Expanding those hypotheses will result in 
  // full coverage

  // first we fetch the set of hypotheses which all have coverage c
  // and we need an iterater to access the hyps in this set
  HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual > &Hyps = m_AllHyps[ Coverage ];
  HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual >::iterator IterHyps;


  // for pruning we need the best hypothesis with the current coverage
  HypContainer< CDecoderBLM_Hyp *, HypHashForPruning, HypEqualForPruning > &BestHyps = m_BestHyps[ Coverage ];
  HypContainer< CDecoderBLM_Hyp *, HypHashForPruning, HypEqualForPruning >::iterator IterBestHyps;
  
  //HistogramPruning( Hyps, BestHyps );

  //  TraceIf( s_TraceOn, 0, "After Histogram Pruning:  Number of hyps with coverage " 
  //		   << Coverage << " is " << Hyps.size() );
  
  // sort hyps by score for beamsize pruning
  typedef multimap<float, CDecoderBLM_Hyp*> CSortedHypContainer;
  // associate expansion candidates with their best-hyp-for-pruning
  typedef map<CDecoderBLM_Hyp*,CSortedHypContainer> CBestSimHyp2AllSimHypsMap;

  CBestSimHyp2AllSimHypsMap HypsByBestHyp;
		   
  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {
    pHyp = *IterHyps;
    //cout << "Hyp is " << pHyp << endl;
    TraceHyp( s_TraceOn, "\n  in ExpandAllHypsWithCoverage - expand next hyp ", pHyp );
    
    // only expand if hyp had survived histogram pruning
    if ( pHyp->m_Expand == False ) {
      
      //cout << "Prune hyp (histogram) " << pHyp << endl;
      //	  m_HypsPruned++;
      
    } else {
    
      // pruning: only expand hyp if it is better than BeamFactor times 
      // score of best hypothesis
      
      IterBestHyps = BestHyps.find( pHyp );
      if ( IterBestHyps == BestHyps.end() ) {
        //cout <<"Coverage: "<<Coverage<< "no best hyp found" << endl;
        pBestHyp->m_TotalScore = 10000;
      } else {
        pBestHyp = *IterBestHyps;
        //cout <<"Coverage: "<<Coverage<< "Best Hyp is " << pBestHyp << endl;
        TraceHyp( s_TraceOn, "\n  best hyp to prune against is ", pBestHyp );
      }
      
    } // endif: histogram pruned
  } // andfor: all recombined hyps


  
  // foreach set of recombined hyps associated to their best hyp, expand beamsize best ones
  int rank;
  CSortedHypContainer::iterator IterSortedHyp;
  CSortedHypContainer HypsInBeam;
  for (CBestSimHyp2AllSimHypsMap::iterator IterHypsByBestHyp = HypsByBestHyp.begin();
       IterHypsByBestHyp != HypsByBestHyp.end();
       ++IterHypsByBestHyp) {
    // sorted list of expansion candidates associated with current best hyp for pruning
    CDecoderBLM_Hyp* pBestHyp = (*IterHypsByBestHyp).first;

    if (s_TraceOn){
      cout<<"Coverage: "<<Coverage<<endl;
      cout<<"pBestHyp:  ";
      //display the best hyp
      CArray< CLatticeEdgeMS * > transEdges = TraceBackWithPath(pBestHyp).second;
      for (int i=transEdges.MinIndex(); i<=transEdges.MaxIndex(); i++){
        transEdges[i]->Display();
        cout<<" ";
      }
      cout<<endl<<endl;
    }

    CSortedHypContainer &SimilarHyps = (*IterHypsByBestHyp).second;

    if (s_TraceOn){
      //lets look at all hyps in beam first

      for (IterSortedHyp  = SimilarHyps.begin(); IterSortedHyp != SimilarHyps.end();  ++IterSortedHyp){
        CDecoderBLM_Hyp* pHyp = (*IterSortedHyp).second;
        //display the hyp
        CArray< CLatticeEdgeMS * > transEdges = TraceBackWithPath(pHyp).second;
        for (int i=transEdges.MinIndex(); i<=transEdges.MaxIndex(); i++){
          transEdges[i]->Display();
          cout<<" ";
        }
        cout<<endl;
        cout<<"FScore: "<<(*IterSortedHyp).first<<" base: "<<pHyp->m_TotalScore<<endl;
	
      }
    }


    // only take the beamsize best
    for(IterSortedHyp = SimilarHyps.begin(), rank = 0;
        IterSortedHyp != SimilarHyps.end() && ++rank <= m_BeamSize;
        ++IterSortedHyp) {

      HypsInBeam.insert(*IterSortedHyp);
    } // endfor

    m_HypsOutOfBeam += ( SimilarHyps.size() > m_BeamSize ?
                         SimilarHyps.size() - m_BeamSize :
                         0 );
  } // endfor


  // out of the hyps just collected, only take the m_MaxHypsPerCoverage best
  for (IterSortedHyp  = HypsInBeam.begin(), rank = 0;
       IterSortedHyp != HypsInBeam.end() && ++rank <= m_MaxHypsPerCoverage;
       ++IterSortedHyp)
    ExpandHyp((*IterSortedHyp).second);

  m_HypsOutOfBeamPerCov += ( HypsInBeam.size() > m_MaxHypsPerCoverage ?
                             HypsInBeam.size() - m_MaxHypsPerCoverage :
                             0 );

  s_TraceOn = false;
}

