#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <fortify.h>
#include <iostream>

#define NO_TRACE 1
#include <Trace.hh>

#include <Array.hh>
#include <DecoderBLM.hh>
#include <OP_Generator.hh>
#include <time.h>

using namespace std;



//bool CDecoderNB_Hyp::s_UseCoverage = true;
//bool CDecoderNB_Hyp::s_UseCoveragePattern = true;
//bool CDecoderNB_Hyp::s_UseLMState = true;
//bool CDecoderNB_Hyp::s_UseTargetLength = true;

//bool CDecoderNB_Hyp::s_UseCoverageForPruning = true;
//bool CDecoderNB_Hyp::s_UseCoveragePatternForPruning = true;
//bool CDecoderNB_Hyp::s_UseLMStateForPruning = true;
//bool CDecoderNB_Hyp::s_UseTargetLengthForPruning = false;

//short CDecoderNB_Hyp::s_MaxNumReorderings = 77;

long int CDecoderNB_Hyp::s_NumberOfObjects = 0;
long int CDecoderNB_BP::s_NumberOfObjects = 0;

bool CDecoderNB::s_TraceOn = false;
//ARV 
int CDecoderNB::m_NumberOfInstances = 0;
bool CDecoderNB_Hyp::s_TraceOn = false;


ostream &
operator <<( ostream& st, const CDecoderNB_Hyp &hyp)
{
  st 
	 << "\nTotal Score " << hyp.m_TotalScore
	//	 << "\nACScore       " << hyp.m_ACScore 
	 << "  TMScore " << hyp.m_scaledModelScores[c_TM]
	 << "  LMScore " << hyp.m_scaledModelScores[c_LM]
	 << "  ALScore " << hyp.m_scaledModelScores[c_AL]
	 << "  SLScore " << hyp.m_scaledModelScores[c_SL];
  for (int index=c_SL+1; index<=hyp.m_scaledModelScores.MaxIndex(); index++){
    st<<" Model "<<index<<" "<<hyp.m_scaledModelScores[index];
  }

  st<< "\nCoverage " << hyp.m_Coverage << " " << hyp.m_CoverageStart << " " << hyp.m_CoveragePattern 
	 << "  LMState " << hyp.m_LMState 
	 << "  Target Length " << hyp.m_TargetLength 
	 << "\nEdge " << hyp.m_pEdge 
	 << "  Previous Hyp " << hyp.m_pPrevHyp;
	//	 << "\nNumber Hyps   " << hyp.s_NumberOfObjects;

  return st;
  
}



//ARV initialize the BP with the number of model scores
CDecoderNB_BP::CDecoderNB_BP( int numModels )
{
  m_TotalScore = 0.0;

  /*
  m_ACScore    = 0.0;
  m_TMScore    = 0.0;
  m_LMScore    = 0.0;
  m_ALScore    = 0.0;
  m_SLScore    = 0.0;
  */

  //ARV: 04/20 Generalizing the decoder to handle a set of scores
  m_scaledModelScores.AssureBoundaries(0, numModels-1);
  m_scaledModelScores.SetAll(0.0);

  m_TargetLength = 0;
  m_pHyp  = NULL;
  //  m_pEdge = NULL;
  m_pBP   = NULL;

  s_NumberOfObjects++;
}



CDecoderNB_BP::CDecoderNB_BP( CDecoderNB_Hyp *pHyp )
{
  //  cout << "CDecoderNB_BP::CDecoderNB_BP( CDecoderNB_Hyp *pHyp )" << endl;
  m_TotalScore = pHyp->m_TotalScore;

  /*
  m_ACScore = pHyp->m_ACScore;
  m_TMScore = pHyp->m_TMScore;
  m_LMScore = pHyp->m_LMScore;
  m_ALScore = pHyp->m_ALScore;
  m_SLScore = pHyp->m_SLScore;
  */

  m_scaledModelScores = pHyp->m_scaledModelScores;
  
  //cout << "features: minindex = " << pHyp->m_scaledModelScores.MinIndex()
  //   << " maxindex = " << pHyp->m_scaledModelScores.MaxIndex() << endl;
  for ( int i = m_scaledModelScores.MinIndex(); i <= m_scaledModelScores.MaxIndex(); i++ ) {
	//cout << i << " in pHyp " << pHyp->m_scaledModelScores[i] << " in pBP " << m_scaledModelScores[i] << endl;
  }

  m_TargetLength = 0;
  m_pHyp  = pHyp;
  //m_pEdge = pHyp->m_pEdge;
  m_pBP   = NULL;

  s_NumberOfObjects++;
}




CDecoderNB_BP::CDecoderNB_BP( CDecoderNB_BP *pBP )
{
  m_TotalScore = pBP->m_TotalScore;

  /*
  m_ACScore    = pBP->m_ACScore;
  m_TMScore    = pBP->m_TMScore;
  m_LMScore    = pBP->m_LMScore;
  m_ALScore    = pBP->m_ALScore;
  m_SLScore    = pBP->m_SLScore;
  */

  //ARV: 04/20 Generalizing the decoder to handle a set of scores
  m_scaledModelScores = pBP->m_scaledModelScores;

  m_TargetLength = pBP->m_TargetLength;
  m_pHyp       = pBP->m_pHyp;
  //m_pEdge      = pBP->m_pEdge;
  m_pBP        = pBP;

  s_NumberOfObjects++;
}




CDecoderNB_BP::~CDecoderNB_BP()
{
  s_NumberOfObjects--;
}




CDecoderNB_Hyp::CDecoderNB_Hyp()
{
  Trace( 3, "CDecoderNB_Hyp Constructor" );
  m_LMState = 0;
  m_TotalScore = 0.0;
  m_Coverage = 0;
  m_CoverageStart = 0;
  m_CoveragePattern = 0;
  m_NumReorderings = 0;
  m_TargetLength = 0;
  m_Expand = true;
  m_pDecoderInfo = NULL;
  m_pEdge = NULL;
  m_pPrevHyp = NULL;
  m_pRecombinedHyps = NULL;
  s_NumberOfObjects++;
}



CDecoderNB_Hyp::CDecoderNB_Hyp(int numModels)
{
  Trace( 3, "CDecoderNB_Hyp Constructor" );
  m_LMState = 0;
  
  /*
  m_ACScore = 0.0;
  m_TMScore = 0.0;
  m_LMScore = 0.0;
  m_ALScore = 0.0;
  m_SLScore = 0.0;
  */

  //ARV: 04/20 Generalizing the decoder to handle a set of scores
  //ARV: 04/22
  m_scaledModelScores.AssureBoundaries(0, numModels-1);
  m_scaledModelScores.SetAll(0.0);

  m_TotalScore = 0.0;
  m_Coverage = 0;
  m_CoverageStart = 0;
  m_CoveragePattern = 0;
  m_NumReorderings = 0;
  m_TargetLength = 0;
  m_Expand = true;
  m_pDecoderInfo = NULL;
  m_pEdge = NULL;
  m_pPrevHyp = NULL;
  m_pRecombinedHyps = NULL;
  s_NumberOfObjects++;
}



CDecoderNB_Hyp::CDecoderNB_Hyp( const CDecoderNB_Hyp &Hyp )
{
  Trace( 3, "CDecoderNB_Hyp Copy Constructor" );
  m_LMState = Hyp.m_LMState;

  /*
  m_ACScore = Hyp.m_ACScore;
  m_TMScore = Hyp.m_TMScore;
  m_LMScore = Hyp.m_LMScore;
  m_ALScore = Hyp.m_ALScore;
  m_SLScore = Hyp.m_SLScore;
  */

  //ARV: 04/20 Generalizing the decoder to handle a set of scores
  m_scaledModelScores = Hyp.m_scaledModelScores;

  m_TotalScore = Hyp.m_TotalScore;
  m_Coverage = Hyp.m_Coverage;
  m_CoverageStart = Hyp.m_CoverageStart;
  m_CoveragePattern = Hyp.m_CoveragePattern;
  m_NumReorderings = Hyp.m_NumReorderings;
  m_TargetLength = Hyp.m_TargetLength;
  m_Expand = Hyp.m_Expand;
  m_pDecoderInfo = Hyp.m_pDecoderInfo;
  m_pEdge = Hyp.m_pEdge;
  m_pPrevHyp = Hyp.m_pPrevHyp;
  m_pRecombinedHyps = m_pRecombinedHyps;
  s_NumberOfObjects++;
}


CDecoderNB_Hyp &
CDecoderNB_Hyp::operator = ( const CDecoderNB_Hyp &Hyp )
{
  if ( this == &Hyp ) 
    //-- Kopie auf sich selber kann ignoriert werden
    return *this;

  m_LMState = Hyp.m_LMState;

  
  /*
  m_ACScore = Hyp.m_ACScore;
  m_TMScore = Hyp.m_TMScore;
  m_LMScore = Hyp.m_LMScore;
  m_ALScore = Hyp.m_ALScore;
  m_SLScore = Hyp.m_SLScore;
  */

  
  //ARV: 04/20 Generalizing the decoder to handle a set of scores
  m_scaledModelScores = Hyp.m_scaledModelScores;

  m_TotalScore = Hyp.m_TotalScore;
  m_Coverage = Hyp.m_Coverage;
  m_CoverageStart = Hyp.m_CoverageStart;
  m_CoveragePattern = Hyp.m_CoveragePattern;
  m_NumReorderings = Hyp.m_NumReorderings;
  m_TargetLength = Hyp.m_TargetLength;
  m_Expand = Hyp.m_Expand;
  m_pDecoderInfo = Hyp.m_pDecoderInfo;
  m_pEdge = Hyp.m_pEdge;
  m_pPrevHyp = Hyp.m_pPrevHyp;
  m_pRecombinedHyps = Hyp.m_pRecombinedHyps;

  return *this;
}




CDecoderNB_Hyp::~CDecoderNB_Hyp() {

  // delete recombined Hyps
  //multimap< float, CDecoderNB_Hyp * > *m_pRecombinedHyps;
  multimap< float, CDecoderNB_Hyp * >::iterator Iter;
  CDecoderNB_Hyp *pHyp;

  //cout<<"m_pRecombinedHyps:"<<m_pRecombinedHyps<<endl;
  if ( m_pRecombinedHyps != NULL) {
    //cout<<"deleting:"<<m_pRecombinedHyps<<endl;
    // error generation
    // ARV
    // added a check in here m_pRecombinedHyps != NULL    
	for ( Iter = m_pRecombinedHyps->begin();  Iter != m_pRecombinedHyps->end(); ++Iter ) {
	  //pHyp = Iter->second;
	  //m_pRecombinedHyps->erase( Iter );
        
	  //cout<<"InnerHyp:"<<Iter->second<<endl;
	  if (Iter->second!=NULL){
	    //cout<<"InnerNOTNULL:"<<Iter->second<<endl;
	    delete Iter->second;
	    
	  }
	  
	}
	// ARV
	// screen to take care of a recombination that was deleted further down
	if (m_pRecombinedHyps != NULL){
	  // changed this to delete the map object that stores the recombined hypothesese
	  m_pRecombinedHyps->clear();
	  //delete m_pRecombinedHyps;
	 m_pRecombinedHyps = NULL;
	}
  }else{
    //ARV
    // it is already removed
  }
  s_NumberOfObjects--;
}


bool
CDecoderNB_Hyp::TestAndSetCoverage( int j_min, int j_max )
{
  bool TraceOn = s_TraceOn;
  s_TraceOn = false;

  TraceIf( s_TraceOn, 0, "CDecoderNB_Hyp::TestAndSetCoverage " << j_min << " " << j_max );

  // the parameters are interpreted as the positions between the words
  // i.e. as node numbers in a lattice
  // 2 3 means therefore: test coverage of word running from 
  // node 2 to 3
  // the coverage signature itself is in terms of positions of the words
  // i.e. 2 to 3 is word position 2

  // return value: true if all positions free, false if collision


  TraceIf( s_TraceOn, 0, m_CoverageStart << " " << m_CoveragePattern );

  // test that first position is free, as otherwise the shifting of
  // the coverage pattern failed in the previous step
  assert( !m_CoveragePattern[ 0 ] );

  // make sure expansion is not moving outside of maximally allowed region
  // otherwise we can not store the coverage pattern
  // ToDo: should allow for long expansion without gap
  if ( j_max - m_CoverageStart > c_MaxReorderWindowLength ) {
	TraceIf( true, 0, "j_max - m_CoverageStart <= c_MaxReorderWindowLength" );
	//assert( j_max - m_CoverageStart <= c_MaxReorderWindowLength );
	return false;
  }
  

  // make sure we are starting beyond what has already been completely covered
  if ( j_min < m_CoverageStart ) {
	TraceIf( true, 0, "j_min = " << j_min << " < m_CoverageStart = " << m_CoverageStart );
	assert( j_min >= m_CoverageStart );
	return false;
  }

  // first we check if there is a collision
  // at the same time we set the requested positions
  for ( int j = j_min; j < j_max; j++ ) {
	if ( m_CoveragePattern[ j - m_CoverageStart ] ) {
	  TraceIf( s_TraceOn, 0, "Collision at position " << j );
	  return false;
	} else {
	  TraceIf( s_TraceOn, 0, "set bit at position " << j );
	  m_CoveragePattern[ j - m_CoverageStart ] = true;
	  TraceIf( s_TraceOn, 0, m_CoveragePattern );
	  m_Coverage++;
	}
  }

  // now we check if the lowest bits are set
  // if yes, shift the pattern and increment CoverageStart
  while ( m_CoveragePattern[ 0 ] ) {
	TraceIf( s_TraceOn, 0, "shift start of coverage pattern" );
	m_CoveragePattern >>= 1;
	m_CoverageStart++;
	TraceIf( s_TraceOn, 0, m_CoverageStart << " " << m_CoveragePattern );
  }

  assert( !m_CoveragePattern[ 0 ] );

  s_TraceOn = false;

  return true;
}


//============================================================
CDecoderNB::CDecoderNB()
{
  TraceIf( s_TraceOn, 0, "CDecoderNB::CDecoderNB" );
   


  // ARV
  // I dont know if this is trusted stuff yet

  // default, no overlapping phrases
  m_AddOverlappingPhrases = false;

  // default, no reordering
  m_LengthOfReorderingWindow = 1;

  // default, 1-best decoding
  m_NumOfNBest = 1;

  // no per coverage pruning
  m_MaxHypsPerCoverage = 1000000;

  // no rescoring as default
  m_Rescore = 0;

  // virtually no beam size pruning
  m_BeamSize = 10000;

  m_MaxHyps = 1000;

  // for histogram pruning use 100 buckets
  m_PruningBuckets.SetBoundaries( 0, 100 );


  m_pDecoderInfo = new CDecoderNB_Info;

  // the standard working hypothesis
  // m_pNumberOneHypothesis = new CDecoderNB_Hyp;
  

  //by default we start with 4 model scores
  m_NumberOfModels = 4;

  // future cost estimator
  m_pFutures = NULL;


  m_pNumberOneHypothesis = NewHypothesis();
  m_pNumberOneHypothesis->m_pDecoderInfo = m_pDecoderInfo;
  //  m_pNumberOneBestHyp = new CDecoderNB_Hyp;
  //m_pNumberOneBestHyp->m_pDecoderInfo = m_pDecoderInfo;
  //ARV
  CDecoderNB::m_NumberOfInstances = 1;

}


CDecoderNB::CDecoderNB( const CDecoderNB &Decoder )
{
  TraceIf( s_TraceOn, 0, "CDecoderNB::CDecoderNB" );


  //ARV: 04/20 Generalizing the decoder to handle a set of scores
  m_ScalingFactors = Decoder.m_ScalingFactors;

  //m_SLStretch = Decoder.m_SLStretch;
  //m_SLScale = Decoder.m_SLScale;

  m_BeamSize = Decoder.m_BeamSize;
  m_MaxHypsPerCoverage = Decoder.m_MaxHypsPerCoverage;
  m_MaxHyps = 1000;

  //cout<<"Coping BeamFactor:"<<m_BeamFactor<<endl;
  m_AddOverlappingPhrases = Decoder.m_AddOverlappingPhrases;

  // paremeters for reordering
  m_LengthOfReorderingWindow = Decoder.m_LengthOfReorderingWindow;

  // default, 1-best decoding
  m_NumOfNBest = Decoder.m_NumOfNBest;

  m_pDecoderInfo = new CDecoderNB_Info;
  (*m_pDecoderInfo) = (*(Decoder.m_pDecoderInfo));

  // the standard working hypothesis
  //m_pNumberOneHypothesis = new CDecoderNB_Hyp;
  m_pNumberOneHypothesis = NewHypothesis();
  m_pNumberOneHypothesis->m_pDecoderInfo = m_pDecoderInfo;
  //m_pNumberOneBestHyp = new CDecoderNB_Hyp;
  //m_pNumberOneBestHyp->m_pDecoderInfo = m_pDecoderInfo;

  // copy translation memory
  m_pTranslationMemory = Decoder.m_pTranslationMemory;

  m_pLM = Decoder.m_pLM;
  m_pLMSentenceState = NULL;  // use own LM Sentence State
  //ARV
  CDecoderNB::m_NumberOfInstances++;
}



CDecoderNB::~CDecoderNB()
{
  Clear();

  //ARV
  cout<<"Not removing the last numberOneHypothesis in ~DecoderNB"<<endl;
  CDecoderNB::m_NumberOfInstances--;
  //if (m_numInstances==1){
  //delete m_pNumberOneHypothesis;
    //delete m_pNumberOneBestHyp;
    //}

  cout<<"Remove decoder info"<<endl;
  delete m_pDecoderInfo;

}


void
CDecoderNB::InitModels( int NumberOfModels )
{
  m_NumberOfModels = NumberOfModels;
  m_ScalingFactors.AssureBoundaries( 0, NumberOfModels - 1 );
}



void
CDecoderNB::Init( CTranslationMemory *pTranslationMemory, CLM *pLM )
{
  m_pTranslationMemory = pTranslationMemory;
  m_pLM = pLM;
  //m_pTranslationMemory->RenormalizeTransducer();
}



void
CDecoderNB::Init( CTranslationMemory *pTranslationMemory, CLM *pLM, CVocabulary *pVocabulary )
{
  m_pTranslationMemory = pTranslationMemory;
  m_pLM = pLM;
  m_pVocabulary = pVocabulary;
  //m_pTranslationMemory->RenormalizeTransducer();
}



void
CDecoderNB::Init( const char *ParamFile, CVocabulary *pVocabulary )
{
  m_pVocabulary = pVocabulary;
  m_pTranslationMemory->Init( ParamFile, pVocabulary );
  m_pTranslationMemory->RenormalizeTransducer();
  m_pLM = m_pTranslationMemory->GetLM();
}



// takes a HypContainer< CDecoderNB_Hyp *, HypHash, HypEqual >
vector< RecombinedMap* >
CDecoderNB::HarvestFromHyp(CDecoderNB_Hyp* pOneHyp)
{
  //iterate over all the recombined hypothesis in this list
  multimap< float, CDecoderNB_Hyp * >::iterator Iter;
  vector<RecombinedMap*> retList;

  if (pOneHyp->m_pRecombinedHyps != NULL) {

	for ( Iter = pOneHyp->m_pRecombinedHyps->begin();  Iter != pOneHyp->m_pRecombinedHyps->end(); ++Iter ) {
	  if (Iter->second!=NULL){
	    //cout<<"InnerNOTNULL:"<<Iter->second<<endl;
	    vector<RecombinedMap*> subList = HarvestFromHyp(Iter->second);
	    // add all from the sublist
	    for(vector<RecombinedMap*>::iterator recIter=subList.begin(); recIter!=subList.end(); recIter++){
	      retList.push_back(*recIter);
	    }
	  }// pointer was valid
	  
	}// all children have been dealt with
	//now add this pointer list
	retList.push_back(pOneHyp->m_pRecombinedHyps);
  }else{
    //ARV
    // it is already removed
  }
  return retList;
  
}



void
CDecoderNB::SetLabelsConsideredAsWords( CArray< CString > Labels )
{

}




void
CDecoderNB::Clear()
{
  HypContainer< CDecoderNB_Hyp *, HypHash, HypEqual >::iterator IterHyps;
  HypContainer< CDecoderNB_Hyp *, HypHashForPruning, HypEqualForPruning >::iterator IterBestHyps;

  CDecoderNB_Hyp *pHyp;

  // delete hyps for each coverage
  //vector<RecombinedMap*> allHarvestedAllHyps;

  for ( int Coverage = m_AllHyps.MinIndex(); Coverage <=  m_AllHyps.MaxIndex(); Coverage++ ) {

	HypContainer< CDecoderNB_Hyp *, HypHash, HypEqual > &Hyps = m_AllHyps[ Coverage ];

	int NumHyp = 0;
	for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {
	  pHyp = *IterHyps;
	  //ARV
	  // This section was used to harvest out the addresses of the recombined map
	  /*
	  vector<RecombinedMap*> harvestedRecombines = HarvestFromHyp(pHyp);
	  for (vector<RecombinedMap*>::iterator recIter = harvestedRecombines.begin(); recIter!= harvestedRecombines.end(); ++recIter){

	    unsigned int memLoc = (unsigned int)(*recIter);
	    cout<<"Harvested: "<<memLoc<<endl;
	    //printf("Harvested:%d\n", (*recIter));
	    allHarvestedAllHyps.push_back(*recIter);
	  }
	  */
	  m_HypsForDeleting.SetAt( NumHyp, pHyp );
	  NumHyp++;
	}
	
	for ( int i = 0; i < NumHyp; i++ ) {
	  delete m_HypsForDeleting[ i ];
	}

	Hyps.clear();
	

	// the best hyps used for pruning
	HypContainer< CDecoderNB_Hyp *, HypHashForPruning, HypEqualForPruning > &BestHyps = m_BestHyps[ Coverage ];

	NumHyp = 0;
	
	for ( IterBestHyps = BestHyps.begin(); IterBestHyps != BestHyps.end(); ++IterBestHyps ) {
	  pHyp = *IterBestHyps;
	  m_HypsForDeleting.SetAt( NumHyp, pHyp );
	  NumHyp++;
	}
	
	for ( int i = 0; i < NumHyp; i++ ) {
	  delete m_HypsForDeleting[ i ];
	}

	BestHyps.clear();

  }

  // remove the stuff in the nbest list
  multimap< float, CDecoderNB_BP * >::iterator IterOverBPs;
  for ( int j = m_NBestHyps.MinIndex(); j <= m_NBestHyps.MaxIndex(); j++ ) {
    multimap< float, CDecoderNB_BP * > &CurrentBPs = m_NBestHyps[ j ];

    // loop through list of backpointers up to N
    for ( IterOverBPs = CurrentBPs.begin(); IterOverBPs != CurrentBPs.end(); ++IterOverBPs ) {
      CDecoderNB_BP* pBP = IterOverBPs->second;
      delete pBP;
      // ARV
      // Moving the erase till after its all cleared up
      //CurrentBPs.erase( IterOverBPs );
    }
    CurrentBPs.clear();
  }


  // and the hyps in the final list and the final rescored list
  multimap< float, CDecoderNB_Hyp * >::iterator Iter = m_FinalHyps.begin();

  int NumHyp = 0;
  for ( Iter = m_FinalHyps.begin(); Iter != m_FinalHyps.end(); ++Iter ) {
	pHyp = Iter->second;
	m_HypsForDeleting.SetAt( NumHyp, pHyp );
	NumHyp++;
  }

  for ( Iter = m_FinalHypsRescored.begin(); Iter != m_FinalHypsRescored.end(); ++Iter ) {
	pHyp = Iter->second;
	m_HypsForDeleting.SetAt( NumHyp, pHyp );
	NumHyp++;
  }

  for ( int i = 0; i < NumHyp; i++ ) {
  	delete m_HypsForDeleting[ i ];
  }

  m_FinalHyps.clear();
  m_FinalHypsRescored.clear();
  
  // remove all the recombined maps that were not deleted when the backpointers were deleted
  for (vector<RecombinedMap*>::iterator recIter = m_RecombinedMapList.begin(); recIter != m_RecombinedMapList.end(); ++recIter){
    RecombinedMap* pRemoveMe = (*recIter);
    //cout<<"ToRemove: "<<(unsigned int)pRemoveMe<<endl;
    delete pRemoveMe;
  }
  m_RecombinedMapList.clear();


  //remove the new initial sentence state created for this sentence
  //by calling NewSentence
  if (m_pLMSentenceState!=NULL){
    delete m_pLMSentenceState;
    m_pLMSentenceState=NULL;
  }
  
  //cout << "Stored final hyps is " << m_FinalHyps.size() << endl;
  //cout << "all deleted" << endl << flush;

}



CString
CDecoderNB::TranslateSentence( const CSentence *pSentence )
{
  TraceIf( s_TraceOn, 0, "CDecoderNB::TranslateSentence" );
  return m_pTranslationMemory->TranslateSentence( pSentence );
}



// sv 051102 Decoder result should not be decoder spefic
//CArray<CDecoderNBResult*>
CArray<CDecoderResult*>
CDecoderNB::TranslateLattice( CLatticeLM *pLattice )
{
  s_TraceOn = false;
  TraceIf( s_TraceOn, 0, "CDecoderNB::TranslateLattice" );
  //return m_pTranslationMemory->TranslateLattice( pLattice );


  m_pTranslationMemory->BuildTranslationLattice( pLattice );
  //cout << "Lattice after transducers applied" << endl;

  //pLattice->Display();
  // insert overlapping phrases
  //cout << "overlapping phrases" << endl;
  if ( m_AddOverlappingPhrases ) {
	COverlappingPhraseGenerator *pOPGenerator = new COverlappingPhraseGenerator;
	pOPGenerator->AddOverlappingPhrases( pLattice );
	delete pOPGenerator;
  }

  m_pLattice = pLattice;

  // here we do the new decoding
  // delete all hypotheses from previous round of decoding

  //CArray<CString> retList = ApplyLanguageModel( );
  return ApplyLanguageModel( );
}



void
CDecoderNB::AddOverlappingPhrases( CLattice *pLattice )
{
  TraceIf( s_TraceOn, 0, "CDecoderNB::AddOverlappingPhrases" );
  
  // insert overlapping phrases
  //cout << "overlapping phrases" << endl;
  if ( m_AddOverlappingPhrases ) {
	COverlappingPhraseGenerator *pOPGenerator = new COverlappingPhraseGenerator;
	pOPGenerator->AddOverlappingPhrases( pLattice );
	delete pOPGenerator;
  }

}



void
CDecoderNB::BuildTranslationLattice( CLatticeLM *pLattice )
{
  TraceIf( 1, 0, "CDecoderNB::BuildTranslationLattice" );
  m_pTranslationMemory->BuildTranslationLattice( pLattice );

}



// sv 051102 Decoder result should not be decoder spefic
//CArray<CDecoderNBResult*>
CArray<CDecoderResult*>
CDecoderNB::ApplyLanguageModel( CLatticeLM *pLattice, CLM *pLM )
{
  TraceIf( s_TraceOn, 0, "CDecoderNB::ApplyLanguageModel" );

  // lattice and LM are needed elsewhere, so assign them to member variable
  m_pLattice = pLattice;
  m_pLM = pLM;

  return ApplyLanguageModel();
}


// sv 051102 Decoder result should not be decoder spefic
//CArray<CDecoderNBResult*>
CArray<CDecoderResult*>
CDecoderNB::ApplyLanguageModel( CLatticeLM *pLattice )
{
  TraceIf( s_TraceOn, 0, "CDecoderNB::ApplyLanguageModel" );

  // lattice needed elsewhere, so assign them to member variable
  m_pLattice = pLattice;

  return ApplyLanguageModel();
}



// sv 051102 Decoder result should not be decoder spefic
//CArray<CDecoderNBResult*>
CArray<CDecoderResult*>
CDecoderNB::ApplyLanguageModel()
{
  s_TraceOn = false;
  TraceIf( s_TraceOn, 0, "CDecoderNB::ApplyLanguageModel" );

  TraceIf( s_TraceOn, 0, "Length of reordering window is " << m_LengthOfReorderingWindow );
  


  // internal statistics
  m_HypsExpanded = 0;
  m_HypsNotExpanded = 0;
  m_HypsOutOfBeam = 0;
  m_HypsOutOfBeamPerCov = 0;
  m_NumNewHyps = 0;
  m_NumBetterHyps = 0;
  m_NumWorseHyps = 0;
  m_NumCoverageCollisions = 0;

  // first convert all string at edges in lattice 
  // into word indices according to LM vocabulary
  GetLMWordIndicesAndTMScores();

  // get the LMSentenceState and the LMState for sentence start
  // LM state
  CLMState LMState = (CLMState) 0;
  CLMState NewLMState = (CLMState) 0;
  
  assert( m_pLM );
  if ( m_pLM ) {
    m_pLMSentenceState = m_pLM->NewSentence();
    LMState = m_pLMSentenceState->BeginOfSentenceState();
    if (s_TraceOn){
      cout << "LMState for begin of sentence is " << LMState <<endl;
    }
  }


  // how long is the source sentence?  we need to cover all words
  int Start = m_pLattice->StartNode()->Number();
  int J = m_pLattice->EndNode()->Number();
  if (s_TraceOn){
    cout << "StartNode " << Start << " EndNode " << J << endl;
    cout << "Sentence length = " << J << endl;
  }
  // set boundaries for arrays; Coverage from 0 to sentence length plus 1
  m_AllHyps.SetBoundaries( 0, J - Start );
  m_BestHyps.SetBoundaries( 0, J - Start );
  

  // should we use average path length for speech recognizer lattices?

  // based on sentence length we set maximum of allowed reorderings
  // as 'a b c d' -> 'a c b d' will count as two reorderings, we
  // multiply by extra factor of 2
  m_pDecoderInfo->m_MaxNumReorderings = (int) (J * 2 * m_ReorderingsPerWord);
  
  if (s_TraceOn){
    cout << "MaxNumReorderings " << m_pDecoderInfo->m_MaxNumReorderings / 2 << endl;
    cout << "Length of reordering window is " << m_LengthOfReorderingWindow << endl;
    cout << "TMScale = " << m_ScalingFactors[c_TM] << " LMScale = " << m_ScalingFactors[c_LM] << " ALScale = " << m_ScalingFactors[c_AL] << " SLScale = " << m_ScalingFactors[c_SL] << endl;
    //<< " SLStretch = " << m_SLStretch 
    cout << "Comparing Hyps with"
         << " Coverage " <<  m_pDecoderInfo->m_UseCoverage 
         << " Coverage Pattern " <<  m_pDecoderInfo->m_UseCoveragePattern
         << " LM state " << m_pDecoderInfo->m_UseLMState
         << " Target stentence length " <<  m_pDecoderInfo->m_UseTargetLength << endl;
    cout << "Pruning Hyps with"
         << " Coverage " <<  m_pDecoderInfo->m_UseCoverageForPruning 
         << " Coverage Pattern " <<  m_pDecoderInfo->m_UseCoveragePatternForPruning
         << " LM state " << m_pDecoderInfo->m_UseLMStateForPruning
         << " Target stentence length " <<  m_pDecoderInfo->m_UseTargetLengthForPruning << endl;
    cout << "Beam Size = " << m_BeamSize << endl;
    cout << "Max # of expanded hyps per coverage = " << m_MaxHypsPerCoverage << endl;
  }// end trace output
  
  // array of multi-hash-sets is dynamic, i.e. when array grows, 
  // items are copied to new memory space; this causes addresses 
  // to individual hash-sets to become invalid
  // therefore, loop over entire range and access hash-set
  // this will create empty hash sets and avoid later copying
  // if a previous sentence has been processed, the hyps are
  // still stored, we clear them
  HypContainer< CDecoderNB_Hyp *, HypHash, HypEqual >::iterator IterHyps;
  CDecoderNB_Hyp *pHyp, *pBestHyp;
  
  for ( int Coverage = 0; Coverage <= J -Start; Coverage++ ) {
    m_AllHyps[ Coverage ].clear();
    m_BestHyps[ Coverage ].clear();
  }


  // we start with an empty hypothesis
  //pHyp = new CDecoderNB_Hyp;
  pHyp = NewHypothesis();
  pHyp->m_pDecoderInfo = m_pDecoderInfo;
  pHyp->m_LMState = LMState;
  //cout << "Initial Hypothesis: " << *pHyp << endl;

  // and place it into the list of all hypotheses with coverage 0
  assert( pHyp->m_Coverage == 0 );
  m_AllHyps[ 0 ].insert( pHyp );
  //cout << "Initial Hypothesis after storing: " << *pHyp << endl;

  // also best hyp for this coverage
  //pBestHyp = new CDecoderNB_Hyp;
  pBestHyp = NewHypothesis();
  pBestHyp->m_pDecoderInfo = m_pDecoderInfo;
  pBestHyp->m_LMState = LMState;
  assert( pBestHyp->m_Coverage == 0 );
  m_BestHyps[ 0 ].insert( pBestHyp );

  // hypotheses are organized according to their coverage
  // hypotheses are expanded till full coverage J is reached
  // i.e. we expand all hyps with Covereage < J

  for ( int Coverage = 0; Coverage < J - Start; Coverage++ ) {
    // display all hyps with current coverage
    //DisplayAllHypsWithCoverage( Coverage );

    // in batch mode we get all the LM log probs in one sweep
    // this then needs also to do the pruning
    //this->GetLMLogProbsAllHyps( Coverage );


    // expand all hyps
    this->ExpandAllHypsWithCoverage( Coverage );

    //for ( int Coverage = 0; Coverage <= J; Coverage++ ) {
    //cout << "Stored hyps for coverage " << Coverage << " is " 
    //<< m_AllHyps[ Coverage ].size() << endl;
    //}
  }


  // display all hyps with full coverage
  //DisplayAllHypsWithCoverage( J, 1 );

  // do specific sentence end stuff
  this->ExpandAllHypsSentenceEnd();

  // display some statistics about the decoding
  s_TraceOn = true;
  if (s_TraceOn){
    cout << "Decoding Statistics               sentence / per word" << endl;
    cout << "Hyps expanded                   = " << m_HypsExpanded    << " / " << m_HypsExpanded / J << endl;
    cout << "Reordering exceeded             = " << m_HypsNotExpanded << " / " << m_HypsNotExpanded <<endl;
    cout << "Beam size exceeded              = " << m_HypsOutOfBeam   << " / " << m_HypsOutOfBeam / J<<endl;
    cout << "Beam size per coverage exceeded = " << m_HypsOutOfBeamPerCov << " / " << m_HypsOutOfBeamPerCov / J<<endl;
    cout << "New Recombined Hyps             = " << m_NumNewHyps      << " / " << m_NumNewHyps / J<<endl;
    cout <<"Better Recombined Hyps           = " << m_NumBetterHyps   << " / " << m_NumBetterHyps / J<<endl;
    cout << "Worse Recombined Hyps           = " << m_NumWorseHyps    << " / " << m_NumWorseHyps / J<<endl;
    cout << "CoverageCollisions              = " << m_NumCoverageCollisions << " / " << m_NumCoverageCollisions / J<<endl;
  }
  s_TraceOn = false;
  
  // reconstruct the n-best translations using the backpointer information
  //CString Translation = BestTranslation();
  TraceIf( s_TraceOn, 0, "\nDisplay n-best hypotheses with backtrace for n-best = " << m_NumOfNBest );  
  TraceIf( 1, 0, "\nNumber of final hypotheses is: " << m_FinalHyps.size() );

  cout << "NBest = " << m_NumOfNBest << endl;
  
  //arv 
  // this is the old style n-best list where we look at the n-best
  // hypothesis that are available at the end node
  bool easyNBest = false;
  CString topTranslation;
  if (easyNBest){
    int n = 0;
    CArray<CString> TranslationList;
    for (multimap< float, CDecoderNB_Hyp * >::iterator Iter = m_FinalHyps.begin();
	 Iter != m_FinalHyps.end();

	 ++Iter) {
      /*
      if ( n <= m_NumOfNBest ) {
	pHyp = (*Iter).second;
	cout << "Trace Back hypothesis " << n << endl;
	//TranslationList[n] = TraceBack( pHyp );
	n++;
      }
      */
      if (n==0){
	pHyp = (*Iter).second;
	topTranslation = TraceBack(pHyp);
	topTranslation.TrimLeadingBlanks();
	topTranslation.TrimTrailingBlanks();
	n++;
      }
    }
  }

  // arv
  // build the true n-best list
  //cout<<"Base List Completed:"<<endl;
  //clock_t tv1, tv2;
  //double time;
  //tv1 = clock();
  BuildNBestList();
  
  //tv2 = clock();
  //time = (tv2 - tv1)/(CLOCKS_PER_SEC / (double) 1000.0);
  //cout<<"Time taken to build NBest List:"<<time<<endl;
  

  // extract the actual nbest list from the member variable
  int n = 0;

// sv 051102 Decoder result should not be decoder spefic
//  CArray<CDecoderNBResult*> nBestTranslations;
  CArray<CDecoderResult*> nBestTranslations;
  multimap< float, CDecoderNB_BP * > &CurrentBPs = m_NBestHyps[ 0 ];
  multimap< float, CDecoderNB_BP * >::iterator IterOverBPs;
 
  for ( IterOverBPs = CurrentBPs.begin(); 
	(IterOverBPs != CurrentBPs.end()) && (n < m_NumOfNBest);
	++IterOverBPs ) {
    
    CDecoderNB_BP *pBP = IterOverBPs->second;
    
    TraceIf( s_TraceOn, 0, "\nBackpointerFinal " << n << " " << IterOverBPs->first
	     << " TotalScore=" << pBP->m_TotalScore );
    
    //cout << endl << "NBest " << n << endl;

    //ARV 04/28
    pair<CString, CArray< CLatticeEdgeMS * > > stringAndPath = TraceStringAndPath( pBP, n );
    CString oneTranslation = stringAndPath.first;
    //ARV 04/28
    CArray< CLatticeEdgeMS * > edgesOnPath = stringAndPath.second;

    if (n==0){
      //oneTranslation.TrimTrailingBlanks();
      //oneTranslation.TrimLeadingBlanks();

      /*
      if (strcmp((const char*)topTranslation, (const char*)oneTranslation) != 0){
	cout<<"ComparingNB"<<endl;
	
	cout<<oneTranslation<<endl;
	cout<<topTranslation<<endl;
	cout<<"============"<<endl;	

	cout<<"Different-----------------"<<endl;
      } 
      */
    }

	// sv 051102 Decoder result should not be decoder spefic
    //CDecoderNBResult* tresult = new CDecoderNBResult(oneTranslation, pBP, edgesOnPath);
    CDecoderResult* tresult = new CDecoderResult(oneTranslation, pBP, edgesOnPath);
    nBestTranslations[n] = tresult;
    //nBestTranslations[n] = oneTranslation;
    n++;
  }
  
  // now we do rescoring for all final hyps

  //ARV: 4/20 removed, since rescoring should be done at the n-best list data structure
  
  if ( m_Rescore ) {
	Rescore();
  }
  

  // extract the best translations at the very end
  s_TraceOn = false;

  return nBestTranslations;
}



void
CDecoderNB::GetLMWordIndicesAndTMScores()
{
  /*
    The LM has its own vocabulary, therefore we get the LM word IDs for the words at the edges
    to avoid repeated string passing and string->int convertion on the LM side.
    For edges carrying source words we just get the LM word ID.
    For edges resulting from transducers or online phrase alignment we distinguish
    1. if label is one of those which should be treated as normal word then we just get the
       LM word ID of that label
    2. if label is a 'normal' label then we go into the wordinfo structure and get the 
       LM word IDs for all the target words (if those don't exist the source words are used)
  */

  cout << "CDecoderNB::GetLMWordIndicesAndTMScores " << endl;


  CArray< CLatticeEdge * > Edges = m_pLattice->Edges();

  // array to collect words attached to one edge for LM-scoring
  CArray< CString > Words;
  CArray< LMWordIndex > WordIndices;

  // first get the words
  CString Word;
  string word;
  CLatticeEdgeMS *pEdge;


  for ( int i = Edges.MinIndex(); i <= Edges.MaxIndex(); i++ ) {
    pEdge = (CLatticeEdgeMS *)Edges[ i ];
	
    Word = pEdge->Word();
    word = (const char*)Word;
    CWordInfo *pWordInfo =  ((CLatticeEdgeMS *)pEdge)->WordInfo();
	
    Trace( 2, "WordInfo: " << *pWordInfo );
    Trace( 2, "SourceString: " << pWordInfo->SourceString() );
    Trace( 2, "TargetString: " << pWordInfo->TargetString() );
	

    // first, we make sure that we have sufficient scores (no needed for source words)
    if ( ( Word[ 0 ] == '@' ) && ( m_ScalingFactors.Size() > c_TM + ((CLatticeEdgeMS*)pEdge)->Scores().Size() ) ) {

      // if we have not enough socres, we supply additional scores
      cout << "Mismatch in number of models and scaling factors: ScalingFactorsSize: " << m_ScalingFactors.Size() << endl;
      ((CLatticeEdgeMS*)pEdge)->Display();
      cout << endl;
      
      while ( m_ScalingFactors.Size() > c_TM + ((CLatticeEdgeMS*)pEdge)->Scores().Size() ) {
        //cout << ((CLatticeEdgeMS*)pEdge)->Scores().Size() << " " << m_ScalingFactors.Size() << endl;
        ((CLatticeEdgeMS*)pEdge)->Scores().Append( 0.0 );
      }
      
      cout << "Now, we should have sufficient scores: ";
      ((CLatticeEdgeMS*)pEdge)->Display();
      cout << endl;
    }
        

    // if edge was not generated by a transducer item
    // then set TM score to TMScoreForSourceWord
    if ( Word[ 0 ] != '@' ) {
      pEdge->WordInfo()->Score() = c_TMScoreForSourceWord;
      TraceIf( s_TraceOn, 0, "      Translation Score for source language edge set to : " 
               << c_TMScoreForSourceWord );
    }


    // check if label is one of the special tags, which the LM should treat as normal word
    if ( LabelIsNormalWord( Word ) ) {
      cout << "Treat " << Word << " as normal word" << endl;
      Words.SetAt( 1, Word );
    } else {

      // extract sequence of words from the WordInfo structure
      if ( pWordInfo->HasTargetString() ) {
        pWordInfo->GetTargetWords( Words );
      } else if ( pWordInfo->HasSourceString() ) {
        pWordInfo->GetSourceWords( Words );
      } else {
        Word = pEdge->Word();
        Words.SetAt( 1, Word );
      }
    }

    // convert strings into indices according to te LM vocabulary
    for ( int t = Words.MinIndex(); t <= Words.MaxIndex(); t++ ) {
      LMWordIndex WordIndex = m_pLM->GetVocabularyIndex( (string)((const char *)(Words[ t ])) );
      pEdge->LMWordIndices()[ t ] = WordIndex;
      TraceIf( s_TraceOn, 0, "Word = " << Words[ t ] << "  LM-WordIndex = " << WordIndex );
    }
    
  }


  //m_pLattice->DisplayEdges();
  //cout << endl;
}


void
CDecoderNB::SetTagsConsideredAsWords( const CString &TagsConsideredAsWords, char SplitCharacter )
{
  //cout << "SetTagsConsideredAsWords " << TagsConsideredAsWords << endl;

  TagsConsideredAsWords.Split( SplitCharacter, m_TagsConsideredAsWords );

  for ( int i = m_TagsConsideredAsWords.MinIndex(); i <= m_TagsConsideredAsWords.MaxIndex(); i++ ) {
    m_TagsConsideredAsWords[ i ].GetLength();
    cout << m_TagsConsideredAsWords[ i ] << " is treated as normal word " << endl;
  }
}



bool
CDecoderNB::LabelIsNormalWord( const CString &Word )
{
  for ( int i = m_TagsConsideredAsWords.MinIndex(); i <= m_TagsConsideredAsWords.MaxIndex(); i++ ) {
    if ( Word == m_TagsConsideredAsWords[ i ] ) {
      cout << Word << " is considered as normal word" << endl;
      return True;
    }
  }

  return False;

}



void
CDecoderNB::ExpandAllHypsWithCoverage( int Coverage )
{
  s_TraceOn = False;

  TraceIf( s_TraceOn, 0, "\n\n  CDecoderNB::ExpandAllHypsWithCoverage " << Coverage );

  // variable we need
  CDecoderNB_Hyp *pHyp, *pBestHyp;
  CLatticeNode *pNode;
  CLatticeEdgeMS *pEdge;



  // how long is the source sentence?  we need to cover all words
  int J = m_pLattice->EndNode()->Number();
  // should we use average path length for speech recognizer lattices?

  // hypotheses are organized according to their coverage
  // hypotheses of coverage c are expanded till coverage J-1
  // is reached.  Expanding those hypotheses will result in 
  // full coverage

  // first we fetch the set of hypotheses which all have coverage c
  // and we need an iterater to access the hyps in this set
  HypContainer< CDecoderNB_Hyp *, HypHash, HypEqual > &Hyps = m_AllHyps[ Coverage ];
  HypContainer< CDecoderNB_Hyp *, HypHash, HypEqual >::iterator IterHyps;


  // for pruning we need the best hypothesis with the current coverage
  HypContainer< CDecoderNB_Hyp *, HypHashForPruning, HypEqualForPruning > &BestHyps = m_BestHyps[ Coverage ];
  HypContainer< CDecoderNB_Hyp *, HypHashForPruning, HypEqualForPruning >::iterator IterBestHyps;
  
  //HistogramPruning( Hyps, BestHyps );

  //  TraceIf( s_TraceOn, 0, "After Histogram Pruning:  Number of hyps with coverage " 
  //		   << Coverage << " is " << Hyps.size() );
  
  // sort hyps by score for beamsize pruning
  typedef multimap<float, CDecoderNB_Hyp*> CSortedHypContainer;
  // associate expansion candidates with their best-hyp-for-pruning
  typedef map<CDecoderNB_Hyp*,CSortedHypContainer> CBestSimHyp2AllSimHypsMap;

  CBestSimHyp2AllSimHypsMap HypsByBestHyp;
		   
  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {
	pHyp = *IterHyps;
	//cout << "Hyp is " << pHyp << endl;
	TraceIf( s_TraceOn, 0, "\n  in ExpandAllHypsWithCoverage - expand next hyp " 
			 << pHyp << " "
			 << pHyp->m_Coverage << " " << pHyp->m_CoverageStart 
			 << " " << pHyp->m_CoveragePattern
			 << " " << pHyp->m_LMState
			 << " " << pHyp->m_TargetLength 
			 << " " << pHyp->m_TotalScore );

	// only expand if hyp had survived histogram pruning
	if ( pHyp->m_Expand == False ) {

	  //cout << "Prune hyp (histogram) " << pHyp << endl;
	  //	  m_HypsPruned++;

	} else {

	  // pruning: only expand hyp if it is better than BeamFactor times 
	  // score of best hypothesis
	  
	  IterBestHyps = BestHyps.find( pHyp );
	  if ( IterBestHyps == BestHyps.end() ) {
	    //cout <<"Coverage: "<<Coverage<< "no best hyp found" << endl;
	    pBestHyp->m_TotalScore = 10000;
	  } else {
	    pBestHyp = *IterBestHyps;
	    //cout <<"Coverage: "<<Coverage<< "Best Hyp is " << pBestHyp << endl;
	    TraceIf( s_TraceOn, 0, "\n  best hyp to prune against is " 
		     << pBestHyp << " "
		     << pBestHyp->m_Coverage << " " << pBestHyp->m_CoverageStart 
		     << " " << pBestHyp->m_CoveragePattern
		     << " " << pBestHyp->m_LMState
		     << " " << pBestHyp->m_TargetLength 
		     << " " << pBestHyp->m_TotalScore );
	  }

	  if (m_pFutures!=NULL){
	    //cout<<"Tracing: "<<endl;
	    
	    /*
	    CArray< CLatticeEdgeMS * > transEdges = TraceBackWithPath(pHyp).second;
	    for (int i=transEdges.MinIndex(); i<=transEdges.MaxIndex(); i++){
	      transEdges[i]->Display();
	    }
	    */
	    
	    double totalFutureCost = 0.0;
	    CArray<std::pair<int,int> >untranslatedSegments = LeftToTranslate(pHyp);
	    for (int i=untranslatedSegments.MinIndex(); i<=untranslatedSegments.MaxIndex(); i++){
	      //cout<<"UT: I:"<<untranslatedSegments[i].first<<" J: "<<untranslatedSegments[i].second; 
	      double futureCostEstimate = m_pFutures->GetFutureCostEstimate(untranslatedSegments[i].first, untranslatedSegments[i].second);
	      totalFutureCost += futureCostEstimate;
	    }
	    //out<<endl;
	    
	    //cout<<"BaseCost: "<<pHyp->m_TotalScore<<" additional: "<<totalFutureCost<<endl;

	    //use the future cost along with the baseline cost to hash 
	    HypsByBestHyp[pBestHyp].insert(CSortedHypContainer::value_type(pHyp->m_TotalScore+totalFutureCost,pHyp));
	  }else{

	    // if there is no future cost estimate, we just use the sum
	    // assert: we have pBestHyp here
	    // associate hyp with its best-hyp-for-pruning
	    HypsByBestHyp[pBestHyp].insert(CSortedHypContainer::value_type(pHyp->m_TotalScore,pHyp));

	  }
	} // endif: histogram pruned
  } // andfor: all recombined hyps

  // foreach set of recombined hyps associated to their best hyp, expand beamsize best ones
  int rank;
  CSortedHypContainer::iterator IterSortedHyp;
  CSortedHypContainer HypsInBeam;
  for (CBestSimHyp2AllSimHypsMap::iterator IterHypsByBestHyp = HypsByBestHyp.begin();
       IterHypsByBestHyp != HypsByBestHyp.end();
       ++IterHypsByBestHyp) {
    // sorted list of expansion candidates associated with current best hyp for pruning
    CDecoderNB_Hyp* pBestHyp = (*IterHypsByBestHyp).first;

    if (s_TraceOn){
      cout<<"Coverage: "<<Coverage<<endl;
      cout<<"pBestHyp:  ";
      //display the best hyp
      CArray< CLatticeEdgeMS * > transEdges = TraceBackWithPath(pBestHyp).second;
      for (int i=transEdges.MinIndex(); i<=transEdges.MaxIndex(); i++){
	transEdges[i]->Display();
	cout<<" ";
      }
      cout<<endl<<endl;
    }

    CSortedHypContainer &SimilarHyps = (*IterHypsByBestHyp).second;

    if (s_TraceOn){
      //lets look at all hyps in beam first

      for (IterSortedHyp  = SimilarHyps.begin(); IterSortedHyp != SimilarHyps.end();  ++IterSortedHyp){
	CDecoderNB_Hyp* pHyp = (*IterSortedHyp).second;
	//display the hyp
	CArray< CLatticeEdgeMS * > transEdges = TraceBackWithPath(pHyp).second;
	for (int i=transEdges.MinIndex(); i<=transEdges.MaxIndex(); i++){
	  transEdges[i]->Display();
	  cout<<" ";
	}
	cout<<endl;
	cout<<"FScore: "<<(*IterSortedHyp).first<<" base: "<<pHyp->m_TotalScore<<endl;
	
      }
    }


    // only take the beamsize best
    for(IterSortedHyp = SimilarHyps.begin(), rank = 0;
	IterSortedHyp != SimilarHyps.end() && ++rank <= m_BeamSize;
	++IterSortedHyp) {

      HypsInBeam.insert(*IterSortedHyp);
    } // endfor

    m_HypsOutOfBeam += ( SimilarHyps.size() > m_BeamSize ?
			 SimilarHyps.size() - m_BeamSize :
			 0 );
  } // endfor


  // out of the hyps just collected, only take the m_MaxHypsPerCoverage best
  for (IterSortedHyp  = HypsInBeam.begin(), rank = 0;
       IterSortedHyp != HypsInBeam.end() && ++rank <= m_MaxHypsPerCoverage;
       ++IterSortedHyp)
    ExpandHyp((*IterSortedHyp).second);

  m_HypsOutOfBeamPerCov += ( HypsInBeam.size() > m_MaxHypsPerCoverage ?
			     HypsInBeam.size() - m_MaxHypsPerCoverage :
			     0 );

  s_TraceOn = false;
} // CDecoderNB::ExpandAllHypsWithCoverage(...)

CArray<std::pair<int, int> > CDecoderNB::LeftToTranslate(CDecoderNB_Hyp* pHyp){
  

  //======================================================
  // Determine which words are untranslated in this hyp
  //======================================================
  bitset<COVERAGE_BIT_VECTOR_SIZE> coveragePatternCopy = pHyp->m_CoveragePattern; 
  CNumericalArray<int> UnTranslated;
  CNumericalArray<int> ForwardTranslated;
  int count =0;
  //cout<<"m_CoverageStart: "<<pHyp->m_CoverageStart<<" coveragePatternCopy: "<<coveragePatternCopy<<endl;
  while (coveragePatternCopy.any()){
    //cout<<"cpCopy: "<<coveragePatternCopy<<endl;
    if (!coveragePatternCopy.test(0)){
      //cout<<"count: "<<count<<" Was 0"<<endl;
      UnTranslated.Append(count);
    }else{
      ForwardTranslated.Append(count);
    }
    coveragePatternCopy>>= 1;
    count++;
  }
  
  //the count tracks how far reordering effects are in play, after all the translated words
  //are accounted for, it reprsents the location of sequentially untranslated words
  
  //now recover the index position in the whole sentence
  //CNumericalArray<int> AllUntranslatedPositions;
  
  //represent contigous sequences of un-translated words with i=number of words, j=starting index of this sequence
  //case 123 5
  //lastJ=1, i=1
  //  i=2
 //  i=3
 //  ?4 - No, lastJ=5, i=1
  

  //======================================================
  // Find contiguous sequences of untranslated words
  //======================================================
  CArray<std::pair<int, int> > segments;

  int icon=1;
  int lastJ =-1;
  bool firstSequenceStarted = false;
  for (int w=UnTranslated.MinIndex(); w<=UnTranslated.MaxIndex(); w++){
    //map from the untranslated count to a count with respect to the sentence
    //this wordIndex is just a relative reference to the sentence
    int wordIndex =  m_pSourceSentence->MinIndex()+pHyp->m_CoverageStart+ UnTranslated[w];
    
    if (wordIndex==lastJ+icon && firstSequenceStarted==true){
      //sequence continues
      icon++;
    }else{
      //if we have already traced one sequence before, then lets save its data
      if (firstSequenceStarted){
	//cout<<"lastJ: "<<lastJ+1<<" icon: "<<icon<<endl;
	segments.Append(std::pair<int,int>( icon, lastJ+1));
      }
      lastJ = wordIndex;
      icon=1;
      firstSequenceStarted = true;
    }

    //cout<<"Jumped: "<<m_pVocabulary->GetWord((*m_pSourceSentence)[wordIndex])<<endl;
    //AllUntranslatedPositions.Append(wordIndex);
  }

  if (firstSequenceStarted){
    //cout<<"lastJ: "<<lastJ+1<<" icon: "<<icon<<endl;
    segments.Append(std::pair<int,int>(icon, lastJ+1));
  }
  lastJ = pHyp->m_CoverageStart+count;
  icon = (m_pSourceSentence->MaxIndex()+1)-(pHyp->m_CoverageStart+count);
  //cout<<"final lastJ: "<<lastJ+1<<" icon: "<<icon<<endl;
  segments.Append(std::pair<int,int>(icon, lastJ+1));
  return segments;
      
  //the coverage start indicates the number of contiguously translated words, to find the leftover 
  //untranslated words, we have to go past the future translated words
  /*
  for (int i=pHyp->m_CoverageStart+count; i<=m_pSourceSentence->MaxIndex(); i++){
    AllUntranslatedPositions.Append(i);
  }
  
  //the leftover words

  
  cout<<"UNT: ";
  for (int i=AllUntranslatedPositions.MinIndex(); i<=AllUntranslatedPositions.MaxIndex(); i++){
    cout<<m_pVocabulary->GetWord((*m_pSourceSentence)[AllUntranslatedPositions[i]])<<" ";
  }
  cout<<endl;
  */
  
}

void
CDecoderNB::ExpandHyp( CDecoderNB_Hyp *pHyp )
{
  s_TraceOn = False;

  TraceIf( s_TraceOn, 0, "\n    CDecoderNB::ExpandHyp" );

  // variable we need
  CLatticeNode *pNode;
  CLatticeEdgeMS *pEdge;


  TraceIf( s_TraceOn, 1, "    expand hyp " << *pHyp );

  // get the range of words in the source sentence 
  // which can now be translated, this is the active range
  // make sure that reordering stays within sentence
  int J = m_pLattice->EndNode()->Number();
  int j_min = pHyp->m_CoverageStart;
  int j_max = pHyp->m_CoverageStart + m_LengthOfReorderingWindow;
  if ( j_max > J ) 
	j_max = J;

  TraceIf( s_TraceOn, 0, "    Translate words in the range " << j_min << " " << j_max );
	
  // now loop over all partial translations t
  // i.e. over all edges in the translation lattice
  // which are within the active range
  // the end point of the edge has also to be within the maximally
  // allowed reordering window (c_MaxReorderWindowLength)

  // currently we reorder also source words, this could (should?) be avoided

  for ( int j = j_min; j < j_max; j++ ) {
	// get node
	pNode = m_pLattice->Nodes()[ j ];
	TraceIf( s_TraceOn, 0, "    process node with number " << pNode->Number() );
	

	// expand this hypothesis only if it would not result in too 
	// many reorderings, but always expand over outgoing edges of node 
	// of current hypothesis, ie the node up to which everything has 
	// been translated (start of cov. pattern bitmap); this is the monotone decoding part
	// the initial hyp has no backpointer edge, the number of the previous
	// node is just the first not in the lattice

	if ( (pNode->Number() == pHyp->m_CoverageStart) 
		 || (pHyp->m_NumReorderings < pHyp->m_pDecoderInfo->m_MaxNumReorderings) ) {

	  // loop over all outgoing edges e
	  const CArray< CLatticeEdge * > & Edges = pNode->OutEdges();
	  TraceIf( s_TraceOn, 0, "    Number of outgoing edges is " << Edges.Size() );
	  
	  for ( int EdgeNum = Edges.MinIndex();
			EdgeNum <= Edges.MaxIndex();
			EdgeNum++ ) {
	  
		pEdge = (CLatticeEdgeMS *)Edges[ EdgeNum ];
		
		// if edge is long it can move outside of the 
		//assert( 0 );
		this->ExpandHypOverEdge( pHyp, pEdge );
		
	  } // forall edges

	} else {

	  m_HypsNotExpanded++;
	} // if not to many reordering
	
  } // for all positions within reordering window

}




void
CDecoderNB::ExpandHypOverEdge( CDecoderNB_Hyp *pHyp, CLatticeEdgeMS *pEdge )
{
  s_TraceOn = false;

  TraceIf( s_TraceOn, 0, "\n      CDecoderNB::ExpandHypOverEdge " << pHyp << " " << pEdge );

  TraceIf( s_TraceOn, 1, "      expand hyp " << *pHyp  );
  TraceIf( s_TraceOn, 1, "      expand over edge " );
  //cout << "expand hyp over edge" << endl;
  //pEdge->Display();

  double TMScore;
  double LMScore;
  double ALScore;

  CLMState LMState = (CLMState) 0;
  CLMState NewLMState = (CLMState) 0;

  CDecoderNB_Hyp *pBestHyp;

  // internal statistics
  m_HypsExpanded++;

  // test coverage collision for current hypothesis and this edge
  // we copy the current hypothesis into the NumberOneHyp,
  // as the TestAndSetCoverage function will modify the hypothesis
  *m_pNumberOneHypothesis = *pHyp;
  
  TraceIf( s_TraceOn, 1, "      test coverage from " << pEdge->FromNode()->Number() 
           << " to " <<  pEdge->ToNode()->Number() );
  
  if ( ! m_pNumberOneHypothesis->TestAndSetCoverage( pEdge->FromNode()->Number(), 
                                                     pEdge->ToNode()->Number() ) ) {

    TraceIf( s_TraceOn, 1, "      Coverage collision detected" );
    m_NumCoverageCollisions++;
    
  } else {

    TraceIf( s_TraceOn, 1, "      No coverage collision" );

    int NewCoverage = m_pNumberOneHypothesis->m_Coverage;
    TraceIf( s_TraceOn, 1, "      the memoryyp has coverage " << NewCoverage );
    TraceIf( s_TraceOn, 2, "      the memoryyp is " << *m_pNumberOneHypothesis );
    
    // length of the generated translation so far
    //cout << "Target " << pEdge->WordInfo()->TargetString()
    // << "Source " << pEdge->WordInfo()->SourceString()
    // << "Label  " << pEdge->Word() << endl;
    
    int AdditionalLength = pEdge->WordInfo()->TargetStringLength();
    if ( AdditionalLength == 0 ) {
      AdditionalLength = pEdge->WordInfo()->SourceStringLength();
    }
    if ( AdditionalLength == 0 ) {
      // this has to be an original edge, i.e. just one word
      AdditionalLength = 1;
    }
    //cout << m_pNumberOneHypothesis->m_TargetLength << " " << AdditionalLength << endl;
    m_pNumberOneHypothesis->m_TargetLength += AdditionalLength;
	
    // calculate the cost for the memoryypothesis,
    // the different Score-methods manipulate m_pNumberOneHypothesis

    // cost from the translation model
    //GetTMScore( pEdge );
    GetTMScoreMS((CLatticeEdgeMS*)pEdge);

    // cost from the language model, sets new LMState in m_pNumberOneHypothesis
    GetLMScore( pHyp, pEdge );
    
    // cost for reordering
    GetALScore( pHyp, pEdge );

    // phrase count model: counts the number of edges used
    GetPCScore( pHyp, pEdge );

    // word count model: count target words
    GetWCScore( pHyp, pEdge );

    // store backpointer information
    // memoryypothesis is expansion of hypothesis pHyp over edge pEdge
    m_pNumberOneHypothesis->m_pEdge = pEdge;
    m_pNumberOneHypothesis->m_pPrevHyp = pHyp;

	

    StoreHypothesis();

  }

}



void
CDecoderNB::StoreHypothesis()
{
  // old: for single-best decoding
  // store memoryypothesis according to coverage
  // but only if it is the first one of its kind, or has a lower cost
  // if it is new, create a new NumberOneHyp
  // if it is better, copy content, but don't create a memoryyp object
  

  // new: for n-best decoding

  // store memoryypothesis according to coverage

  // first pruning: this is unchanged to single-best decoding
  // if new: in BestHyps, store it
  // if better: overwrite old one
  // if worse: do nothing

  // now storing: different, as we store recombined hyps with best
  // if too bad compared to best for pruing, then prune
  // if new: store NumberOneHyp and create new NumberOneHyp
  // if better: add pointer of stored hyp to RecombinedHyps; 
  //            store NumberOneHyp and create new NumberOneHyp
  // if worse: add pointer of NumberOneHyp to RecombinedHyps;

  s_TraceOn = False;

  TraceIf( s_TraceOn, 0, "=== Store Hypothesis ===" );


  CString Translation;

  // has NumberOneHyp DecoderInfo?
  assert( m_pNumberOneHypothesis->m_pDecoderInfo );

  int NewCoverage = m_pNumberOneHypothesis->m_Coverage;
  assert( NewCoverage > 0 );

  CDecoderNB_Hyp *pStoredHyp, *pBestHypForPruning;

  // for recombination, access to stored hyp
  HypContainer< CDecoderNB_Hyp *, HypHash, HypEqual >::iterator IterFind;

  // for pruning, access to best hyps
  HypContainer< CDecoderNB_Hyp *, HypHashForPruning, HypEqualForPruning >::iterator IterFindHypForPruning;

  // copy information to NumberOneBestHyp
  //*m_pNumberOneBestHyp = *m_pNumberOneHypothesis;

  IterFindHypForPruning = m_BestHyps[ NewCoverage ].find( m_pNumberOneHypothesis );


  // first make sure that hyp is in best hyps
  // if it is not there, create it
  // if it is there but we have now a better one, replace it

  if ( IterFindHypForPruning == m_BestHyps[ NewCoverage ].end() ) {
	
	// we don't have a best hypothesis so far
	// make memoryypothesis, copy content from NumberOneHyp and store it

	TraceIf( s_TraceOn, 0, "      Prune: Best Hyp is new, store it" );

	//pBestHypForPruning = new CDecoderNB_Hyp;
	pBestHypForPruning = NewHypothesis();
	*pBestHypForPruning = *m_pNumberOneHypothesis;
	assert( pBestHypForPruning->m_pDecoderInfo );

	assert( NewCoverage > 0 );
	m_BestHyps[ NewCoverage ].insert( pBestHypForPruning );

	//m_pNumberOneBestHyp = new CDecoderNB_Hyp;
	//m_pNumberOneBestHyp->m_pDecoderInfo = m_pDecoderInfo;

  } else {

	// so, one of that kind exists already, now we can compare
	// first, would the memoryyp be a better best hyp?
	pBestHypForPruning = *IterFindHypForPruning;

	if ( m_pNumberOneHypothesis->m_TotalScore < pBestHypForPruning->m_TotalScore ) {

	  TraceIf( s_TraceOn, 0, "      Prune: Hyp is better than stored one, overwrite "
			   << pBestHypForPruning << " "
			   << pBestHypForPruning->m_Coverage << " " << pBestHypForPruning->m_CoverageStart 
			   << " " << pBestHypForPruning->m_CoveragePattern 
			   << " " << pBestHypForPruning->m_LMState
			   << " " << pBestHypForPruning->m_TargetLength
			   << " " << pBestHypForPruning->m_TotalScore );
	  
	  TraceIf( s_TraceOn, 0, "      with "
			   << m_pNumberOneHypothesis << " "
			   << m_pNumberOneHypothesis->m_Coverage << " " << m_pNumberOneHypothesis->m_CoverageStart 
			   << " " << m_pNumberOneHypothesis->m_CoveragePattern 
			   << " " << m_pNumberOneHypothesis->m_LMState
			   << " " << m_pNumberOneHypothesis->m_TargetLength
			   << " " << m_pNumberOneHypothesis->m_TotalScore );
	  
	  // copy content of NumberOne over
	  *pBestHypForPruning = *m_pNumberOneHypothesis;

	} else {

	  TraceIf( s_TraceOn, 0, "      Prune: Hyp is worse than stored one, don't use it" );

	}

  }


  // here we have the best hypothesis as pBestHyp
  // we compare the memoryypthesis against this best hyp

  // is it good enough or should it be pruned
  //if ( m_pNumberOneHypothesis->m_TotalScore > pBestHypForPruning->m_TotalScore + (((double)m_BeamFactor/100.0)*pBestHypForPruning->m_TotalScore) ) {
  //no pruning at this stage
  if (0){
	TraceIf( s_TraceOn, 0, "      Hyp is too bad, prune it " );
	//m_HypsPruned++;
		
  } else {

	// hyp is good enough to keep it
	// now we need to compare it to stored hyp for recombination
	IterFind = m_AllHyps[ NewCoverage ].find( m_pNumberOneHypothesis );

	// is it a memoryyp?
	if ( IterFind == m_AllHyps[ NewCoverage ].end() ) {
	  // yes
	  TraceIf( s_TraceOn, 0, "      Hyp is new " 
			   << m_pNumberOneHypothesis << " "
			   << m_pNumberOneHypothesis->m_Coverage << " " << m_pNumberOneHypothesis->m_CoverageStart 
			   << " " << m_pNumberOneHypothesis->m_CoveragePattern 
			   << " " << m_pNumberOneHypothesis->m_LMState
			   << " " << m_pNumberOneHypothesis->m_TargetLength
			   << " " << m_pNumberOneHypothesis->m_TotalScore );
	  
	  // create map for recombined hyps, emtpy at the beginning
	  // ARV
	  // changed this to use the managed call to NewRecombinedMap
	  //m_pNumberOneHypothesis->m_pRecombinedHyps = new multimap< float, CDecoderNB_Hyp * >;
	  m_pNumberOneHypothesis->m_pRecombinedHyps = NewRecombinedMap();

	  // we need to store NumberOneHyp and create a new NumberOneHyp object
	  assert( NewCoverage > 0 );
	  m_AllHyps[ NewCoverage ].insert( m_pNumberOneHypothesis );

	  //cout << "TRACEBACK new" << endl;
	  //Translation = TraceBack( m_pNumberOneHypothesis );
	  //cout << "TRACEBACK new = " << Translation << endl;

	  // test if the correct hyp is stored
	  IterFind = m_AllHyps[ NewCoverage ].find( m_pNumberOneHypothesis );
	  CDecoderNB_Hyp *pHypForDebugging;
		
	  // ARV
	  // following code is debugging only
	  if ( IterFind == m_AllHyps[ NewCoverage ].end() ) {
		TraceIf( s_TraceOn, 0, "     Dammned! the just inserted hyp can not be found" );
	  } else {
		pHypForDebugging = *IterFind;
		TraceIf( s_TraceOn, 0, "      Stored best Hyp is now " 
				 << pHypForDebugging << " "
				 << pHypForDebugging->m_Coverage << " " << pHypForDebugging->m_CoverageStart 
				 << " " << pHypForDebugging->m_CoveragePattern 
				 << " " << pHypForDebugging->m_LMState
				 << " " << pHypForDebugging->m_TargetLength
				 << " " << pHypForDebugging->m_TotalScore 
				 << " with " << pHypForDebugging->m_pRecombinedHyps->size() << " recombined hyps" );
	  }
	  // end debugging information
	       

	  //m_pNumberOneHypothesis = new CDecoderNB_Hyp;
	  m_pNumberOneHypothesis = NewHypothesis();
	  m_pNumberOneHypothesis->m_pDecoderInfo = m_pDecoderInfo;
	  m_NumNewHyps++;
	  
	} else {
		
	  // it exist already
	  // is it better or worse then stored one

	  pStoredHyp = *IterFind;

	  if ( m_pNumberOneHypothesis->m_TotalScore < pStoredHyp->m_TotalScore ) {
		  
		assert( m_pNumberOneHypothesis->m_Coverage == pStoredHyp->m_Coverage );

		// better

		// we reuse the stored one and update it with the info from NumberOneHyp
		// (insert into map will not overwrite old pointer with new pointer,
		// so either we delete and insert, or modify)
		// therefore: first create memoryyp object and copy info of stored hyp into memoryyp
		//            second overwrite stored hyp with new best (i.e. NumberOneHyp)
		//            third add memoryyp (i.e old best) to the list of recombined hyps

		TraceIf( s_TraceOn, 0, "      Hyp " 
				 << m_pNumberOneHypothesis << " "
				 << m_pNumberOneHypothesis->m_Coverage << " " << m_pNumberOneHypothesis->m_CoverageStart 
				 << " " << m_pNumberOneHypothesis->m_CoveragePattern 
				 << " " << m_pNumberOneHypothesis->m_LMState
				 << " " << m_pNumberOneHypothesis->m_TargetLength
				 << " " << m_pNumberOneHypothesis->m_TotalScore );


		TraceIf( s_TraceOn, 0, "      is better than stored one " 
				 << pStoredHyp << " "
				 << pStoredHyp->m_Coverage << " " << pStoredHyp->m_CoverageStart 
				 << " " << pStoredHyp->m_CoveragePattern 
				 << " " << pStoredHyp->m_LMState
				 << " " << pStoredHyp->m_TargetLength
				 << " " << pStoredHyp->m_TotalScore 
				 << " with " << pStoredHyp->m_pRecombinedHyps->size() << " recombined hyps" );

		// first: memoryyp as copy of currently stored
		//CDecoderNB_Hyp *pStoredHypCopy = new CDecoderNB_Hyp;
		CDecoderNB_Hyp *pStoredHypCopy = NewHypothesis();
		*pStoredHypCopy = *pStoredHyp;
		assert( pStoredHypCopy->m_pRecombinedHyps );

		TraceIf( s_TraceOn, 0, "      copy of stored hyp is " 
				 << pStoredHypCopy << " "
				 << pStoredHypCopy->m_Coverage << " " << pStoredHypCopy->m_CoverageStart 
				 << " " << pStoredHypCopy->m_CoveragePattern 
				 << " " << pStoredHypCopy->m_LMState
				 << " " << pStoredHypCopy->m_TargetLength
				 << " " << pStoredHypCopy->m_TotalScore );

		// second: overwrite stored with better hyp
		*pStoredHyp = *m_pNumberOneHypothesis;
		// need to carry over the pointer from recombined hyps from
		// copy of stored hyp to stored hyp
		pStoredHyp->m_pRecombinedHyps = pStoredHypCopy->m_pRecombinedHyps;
		pStoredHypCopy->m_pRecombinedHyps = NULL;

		TraceIf( s_TraceOn, 0, "      modified stored hyp is " 
				 << pStoredHyp << " "
				 << pStoredHyp->m_Coverage << " " << pStoredHyp->m_CoverageStart 
				 << " " << pStoredHyp->m_CoveragePattern 
				 << " " << pStoredHyp->m_LMState
				 << " " << pStoredHyp->m_TargetLength
				 << " " << pStoredHyp->m_TotalScore );
		

		//Translation = TraceBack( m_pNumberOneHypothesis );
		//cout << "TRACEBACK new best = " << Translation << endl;



		// third: add former best hyp to list of recombined hyps
		assert( pStoredHypCopy->m_Coverage > 0 );
		pStoredHyp->m_pRecombinedHyps->insert( pair< float, CDecoderNB_Hyp * > ( pStoredHypCopy->m_TotalScore, pStoredHypCopy ) );
		
		TraceIf( s_TraceOn, 0, "      Resulting best Hyp is " 
				 << pStoredHyp << " "
				 << pStoredHyp->m_Coverage << " " << pStoredHyp->m_CoverageStart 
				 << " " << pStoredHyp->m_CoveragePattern 
				 << " " << pStoredHyp->m_LMState
				 << " " << pStoredHyp->m_TargetLength
				 << " " << pStoredHyp->m_TotalScore );
		
		TraceIf( s_TraceOn, 0, "Number of recombined hyp is now: " << pStoredHyp->m_pRecombinedHyps->size() );
		
		// test if the correct hyp is stored
		IterFind = m_AllHyps[ NewCoverage ].find( m_pNumberOneHypothesis );
		CDecoderNB_Hyp *pHypForDebugging;
		
		if ( IterFind == m_AllHyps[ NewCoverage ].end() ) {
		  TraceIf( s_TraceOn, 0, "     Dammned! the just inserted hyp can not be found" );
		} else {
		  pHypForDebugging = *IterFind;
		  TraceIf( s_TraceOn, 0, "      Stored best Hyp is now " 
				   << pHypForDebugging << " "
				   << pHypForDebugging->m_Coverage << " " << pHypForDebugging->m_CoverageStart 
				   << " " << pHypForDebugging->m_CoveragePattern 
				   << " " << pHypForDebugging->m_LMState
				   << " " << pHypForDebugging->m_TargetLength
				   << " " << pHypForDebugging->m_TotalScore 
				   << " with " << pHypForDebugging->m_pRecombinedHyps->size() << " recombined hyps" );
		}
		
		// we reuse NumberOneHyp
		
		m_NumBetterHyps++;
		
		
	  } else {
		

		TraceIf( s_TraceOn, 0, "      Hyp " 
				 << m_pNumberOneHypothesis << " "
				 << m_pNumberOneHypothesis->m_Coverage << " " << m_pNumberOneHypothesis->m_CoverageStart 
				 << " " << m_pNumberOneHypothesis->m_CoveragePattern 
				 << " " << m_pNumberOneHypothesis->m_LMState
				 << " " << m_pNumberOneHypothesis->m_TargetLength
				 << " " << m_pNumberOneHypothesis->m_TotalScore );


		TraceIf( s_TraceOn, 0, "      is worse than stored one " 
				 << pStoredHyp << " "
				 << pStoredHyp->m_Coverage << " " << pStoredHyp->m_CoverageStart 
				 << " " << pStoredHyp->m_CoveragePattern 
				 << " " << pStoredHyp->m_LMState
				 << " " << pStoredHyp->m_TargetLength
				 << " " << pStoredHyp->m_TotalScore );

		// we add it to the list of recombined hyps
		assert( m_pNumberOneHypothesis->m_Coverage > 0 );
		pStoredHyp->m_pRecombinedHyps->insert
		  ( 
		   pair< float, CDecoderNB_Hyp * > 
		   ( 
			m_pNumberOneHypothesis->m_TotalScore, m_pNumberOneHypothesis )
		   );
		
		//cout << "TRACEBACK new worse" << endl;
		//Translation = TraceBack( m_pNumberOneHypothesis );
		//cout << "TRACEBACK new worse = " << Translation << endl;



		TraceIf( s_TraceOn, 0, "Number of recombined hyp is now: " 
				 << pStoredHyp->m_pRecombinedHyps->size() );
		
		// and create a new NumberOneHyp;
		//m_pNumberOneHypothesis = new CDecoderNB_Hyp;
		m_pNumberOneHypothesis = NewHypothesis();
		m_pNumberOneHypothesis->m_pDecoderInfo = m_pDecoderInfo;
		
		// but count it as a worse hyp
		m_NumWorseHyps++;
		
	  }
	}
	//
	//TraceIf( s_TraceOn, 0, "      Number of hyps with coverage " << pHyp->m_Coverage 
	//		 << " is still " << m_AllHyps[ pHyp->m_Coverage ].size() );
	TraceIf( s_TraceOn, 0, "      Number of hyps with coverage " << NewCoverage 
			 << " is now   " << m_AllHyps[ NewCoverage ].size() );
	
  }
  
  s_TraceOn = false;
  
}




void
CDecoderNB::ExpandAllHypsSentenceEnd()
{
  TraceIf( 1, 0, "CDecoderNB::ExpandAllHypsSentenceEnd()" );

  // loop over all hyps with full coverage
  // apply sentence end scores (LM, sentence length model)
  // and sort them according to cost

  // clear the final list, in case it still has the hyps from previous sentence
  m_FinalHyps.clear();

  // first we fetch the set of hypotheses which all have full coverage,
  // i.e. source sentence length
  int J = m_pLattice->EndNode()->Number();
  HypContainer< CDecoderNB_Hyp *, HypHash, HypEqual > &Hyps = m_AllHyps[ J ];
  TraceIf( s_TraceOn, 1, "\n  Number of hyps with full coverage is " << Hyps.size() );

  // and we need an iterater to access the hyps in this set
  HypContainer< CDecoderNB_Hyp *, HypHash, HypEqual >::iterator IterHyps;

  CDecoderNB_Hyp *pHyp;

  //ARV 04/25 
  //moving over to general form of initilizing a model and then using it in the loop
  //InitLexicalSLScoreEnd();
  
  
  // loop over hyps
  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {

    // create a copy of the hyp and point back to that hyp
    // we don't go over an edge at this point
    //pHyp = new CDecoderNB_Hyp( **IterHyps );
    pHyp = NewHypothesis();
    *pHyp = **IterHyps;

    // no recombined hyps
    pHyp->m_pRecombinedHyps = NULL;
    // and coverage increased by one, as if going over </s>
    pHyp->m_Coverage++;

    pHyp->m_pDecoderInfo = m_pDecoderInfo;
    pHyp->m_pPrevHyp = *IterHyps;
    pHyp->m_pEdge = NULL;
    

    // cost for sentence end from the language model
    double LMScore = -m_pLMSentenceState->LogProbEnd( pHyp->m_LMState );

    //ARV: 04/20
    //pHyp->m_LMScore += m_ScalingFactors[c_LM] * LMScore;
    pHyp->m_scaledModelScores[c_LM] += m_ScalingFactors[c_LM] * LMScore;
 
    pHyp->m_TotalScore += m_ScalingFactors[c_LM] * LMScore;
    TraceIf( s_TraceOn, 1, "\n  LM cost for sentence end " << LMScore );
    
    // cost from sentence length model
    // exp( -l(cJ - I)^2 )
    // two factors: c is general stretching factor: use 1.2 for now
    //              f is flatness factor: use 0.25 for now

    // SL model is currently only a word count model
    // should not be applied at the sentence end but at the  phrase level
    /*
      float SLScore;
      int I = pHyp->m_TargetLength;
      //ARV 
      // removed this old style in factor of the simple counting word score
      //SLScore = 0.25 * ( m_SLStretch * J - I ) * ( m_SLStretch * J - I );

	
      //SLScore = -1.0*I;
      //SLScore = 2*J -I;
      SLScore = I;

      //ARV 04/20
      //pHyp->m_SLScore = m_SLScale*SLScore;
      //pHyp->m_TotalScore += m_SLScale * SLScore;

      pHyp->m_scaledModelScores[c_SL] = m_ScalingFactors[c_SL]*SLScore;
      pHyp->m_TotalScore += m_ScalingFactors[c_SL] * SLScore;
      
    */

    //double lexScore = GetLexicalSLScoreEnd(pHyp);
    //pHyp->m_scaledModelScores[c_LSL] = m_ScalingFactors[c_LSL]*lexScore;
    //pHyp->m_TotalScore += m_ScalingFactors[c_LSL]*lexScore;


    // now insert into final list, sorting is done by the container
    TraceIf( s_TraceOn, 1, "\n  insert hyp into final list" 
             << pHyp << " "
             << pHyp->m_Coverage << " " << pHyp->m_CoverageStart << " " 
             << pHyp->m_CoveragePattern );
    
    assert(  pHyp->m_Coverage > 0 );
    m_FinalHyps.insert(pair< double, CDecoderNB_Hyp * >( pHyp->m_TotalScore , pHyp ) );
    
  }

  TraceIf( s_TraceOn, 1, "Number of Hyps in Final List is: " << m_FinalHyps.size() );

}



void
CDecoderNB::HistogramPruning( HypContainer< CDecoderNB_Hyp *, HypHash, HypEqual > &Hyps,
							  HypContainer< CDecoderNB_Hyp *, HypHashForPruning, HypEqualForPruning > &BestHyps )
{

  // the histogram pruning reads reworking and rethinking
  // sometimes only hyps suvive which lead to collision
  // so no full translation is generated
  
  return;
  /*

  bool s_TraceOn = false;

  //Trace( 0, "Histogram Pruning: " << Hyps.size() << " " << BestHyps.size() );

  // run over all hyps
  // put hyp into bucket according to BucketNumber = (pHyp->Score - pBestHyp->Score) * 100 / BeamFactor
  // i.e. do not select according to absolute score but according to difference to best hyp
  // with the same features

  CDecoderNB_Hyp *pHyp, *pBestHyp;

  HypContainer< CDecoderNB_Hyp *, HypHash, HypEqual >::iterator IterHyps;
  HypContainer< CDecoderNB_Hyp *, HypHashForPruning, HypEqualForPruning >::iterator IterBestHyps;


  // first, find the best score from the best hyps
  float BestScore = 1.0e10;
  for ( IterBestHyps = BestHyps.begin(); IterBestHyps != BestHyps.end(); ++IterBestHyps ) {

	pBestHyp = *IterBestHyps;
	if ( pBestHyp->m_TotalScore < BestScore ) {
	  BestScore = pBestHyp->m_TotalScore;
	}
  }

  //cout << "Best Score = " << BestScore << endl;

//   for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {

// 	pHyp = *IterHyps;
// 	//cout << "Hyp is " << pHyp << endl;
// 	TraceIf( s_TraceOn, 0, "\n  hyp" 
// 			 << pHyp << " "
// 			 << pHyp->m_Coverage << " " << pHyp->m_CoverageStart 
// 			 << " " << pHyp->m_CoveragePattern
// 			 << " " << pHyp->m_LMState
// 			 << " " << pHyp->m_TargetLength 
// 			 << " " << pHyp->m_TotalScore );


// 	// select corresponding best hyp
// 	IterBestHyps = BestHyps.find( pHyp );
// 	pBestHyp = *IterBestHyps;
// 	//cout << "Best Hyp is " << pBestHyp << endl;
// 	TraceIf( s_TraceOn, 0, "\n  best hyp " 
// 			 << pBestHyp << " "
// 			 << pBestHyp->m_Coverage << " " << pBestHyp->m_CoverageStart 
// 			 << " " << pBestHyp->m_CoveragePattern
// 			 << " " << pBestHyp->m_LMState
// 			 << " " << pBestHyp->m_TargetLength 
// 			 << " " << pBestHyp->m_TotalScore );
	
// 	// now the calculation
// 	int BucketNumber = int( ( pHyp->m_TotalScore - pBestHyp->m_TotalScore ) * 100 / m_BeamFactor );
// 	TraceIf( s_TraceOn, 0, "\n  BucketNumber = " << BucketNumber );
// 	if ( BucketNumber < 100 ) {
// 	  m_PruningBuckets.GetRow( BucketNumber ).Append( pHyp );
// 	}
//   }



  //cout << "In Histogram Pruning we have " << Hyps.size() << " hypotheses" << endl;
  int CollectedHyps = 0;

  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {

	pHyp = *IterHyps;
	pHyp->m_Expand = False;

	//cout << "Hyp is " << pHyp << endl;
	TraceIf( s_TraceOn, 2, "\n  hyp" 
			 << pHyp << " "
			 << pHyp->m_Coverage << " " << pHyp->m_CoverageStart 
			 << " " << pHyp->m_CoveragePattern
			 << " " << pHyp->m_LMState
			 << " " << pHyp->m_TargetLength 
			 << " " << pHyp->m_TotalScore );


	// now the calculation
	int BucketNumber = int( ( pHyp->m_TotalScore - BestScore ) * 100 / m_BeamFactor );
	TraceIf( s_TraceOn, 2, "\n  BucketNumber = " << BucketNumber );
	if ( BucketNumber < 100 ) {
	  CollectedHyps++;
	  m_PruningBuckets.GetRow( BucketNumber ).Append( pHyp );
	}
  }

  //cout << "In the buckets we have " << CollectedHyps << " hypotheses" << endl;

  // now collect enough buckets to get the required number of hyps
  //cout << "collect " << m_MaxHyps << " hyps " << endl;
  int Sum = 0;
  int i = 0;
  while ( (Sum < m_MaxHyps) && (i < 100) ) {
	CArray< CDecoderNB_Hyp * > Bucket = m_PruningBuckets.GetRow( i );
	Sum += Bucket.Size();
	//cout << i << " " << Bucket.Size() << " " << Sum << endl;
	
	for ( int j = Bucket.MinIndex(); j <= Bucket.MaxIndex(); j++ ) {
	  Bucket[ j ]->m_Expand = True;
	}

	i++;
  }

  //cout << "Number of Hyp to expand after Histogram Pruning: " << Sum << endl;

  m_PruningBuckets.Clear();

  s_TraceOn = false; */
}




CString
CDecoderNB::TraceBack( CDecoderNB_Hyp *pHyp )
{
  bool OldTraceOn = CDecoderNB::s_TraceOn;
  CDecoderNB::s_TraceOn = false;

  //cout << "Trace Back" << endl;
  //cout << *pHyp << endl;
  //cout << pHyp->m_pPrevHyp << endl;

  int Start = m_pLattice->StartNode()->Number();
  int End   = m_pLattice->EndNode()->Number();
  int L = End - Start;

  if (s_TraceOn){

    //ARV 4/20 old way that required hardcoded models
    /*
  cout << endl
	   << "Total Score " << pHyp->m_TotalScore
	   << "  TMScore " << pHyp->m_TMScore
	   << "  LMScore " << pHyp->m_LMScore
	   << "  ALScore " << pHyp->m_ALScore
	   << "  SLScore " << pHyp->m_SLScore
	   << endl;
  cout << "Per Word    " << pHyp->m_TotalScore / L
	   << "  TMScore " << pHyp->m_TMScore / L
	   << "  LMScore " << pHyp->m_LMScore / L
	   << "  ALScore " << pHyp->m_ALScore / L
	   << "  SLScore " << pHyp->m_SLScore / L
	   << endl;
    */
    cout<<endl<< "Total Score " << pHyp->m_TotalScore;
    for (int i=0; i<pHyp->m_scaledModelScores.Size(); i++){
      cout<<" Mod "<<i<<" = "<<pHyp->m_scaledModelScores[i];
    }
    cout<<endl;


    cout << "Reorderings " << pHyp->m_NumReorderings / 2 
	 << endl;
  }

  CArray< CLatticeEdgeMS * > AllEdges;
  CArray< CLatticeEdge * > AllEdges2;

  while ( pHyp->m_pPrevHyp ) {
	//cout << "Hyp " << pHyp << " PrevHyp " << pHyp->m_pPrevHyp 
	//<< " Edge " << pHyp->m_pEdge << endl;
	if ( !pHyp->m_pEdge ) {
	  //cout << "no edge - </s>" << endl;
	} else {
	  //pHyp->m_pEdge->Display();
	  AllEdges.Append( pHyp->m_pEdge );
	}

	//cout << *pHyp->m_pPrevHyp << endl;
	pHyp = pHyp->m_pPrevHyp;
  }


  // first1, display the path that has been taken
  // invert sequence of edges in list
  for ( int i = AllEdges.MaxIndex(); i >= AllEdges.MinIndex(); i-- ) {
	AllEdges2.Append( AllEdges[ i ] );
  }
  //cout << "PartialPath: ";
  //if (s_TraceOn){
  //m_pLattice->DisplayPath( AllEdges2 );
	//}

  // now display the clean translation
  CString Translation;
  CWordInfo *pWordInfo;
  CLatticeEdgeMS *pEdge;

  for ( int i = AllEdges.MaxIndex(); i >= AllEdges.MinIndex(); i-- ) {
	pEdge = AllEdges[ i ];
	pWordInfo =  pEdge->WordInfo();

	// extract sequence of words from the WordInfo structure
	if ( pWordInfo->HasTargetString() ) {
	  Translation += pWordInfo->TargetString();
	} else if ( pWordInfo->HasSourceString() ) {
	  Translation += pWordInfo->SourceString();
	} else {
	  Translation += pEdge->Word();
	}

	Translation += " ";
  }

  //  if (s_TraceOn){
  //  cout << "Translation is: " << Translation << endl;
  //  cout << "------------" << endl;
	//}
  CDecoderNB::s_TraceOn = OldTraceOn;

  return Translation;
}

std::pair<CString, CArray< CLatticeEdgeMS * > >
CDecoderNB::TraceBackWithPath( CDecoderNB_Hyp *pHyp )
{
  bool OldTraceOn = CDecoderNB::s_TraceOn;
  CDecoderNB::s_TraceOn = false;

  //cout << "Trace Back" << endl;
  //cout << *pHyp << endl;
  //cout << pHyp->m_pPrevHyp << endl;

  int Start = m_pLattice->StartNode()->Number();
  int End   = m_pLattice->EndNode()->Number();
  int L = End - Start;

  if (s_TraceOn){

    //ARV 4/20 old way that required hardcoded models
    /*
  cout << endl
	   << "Total Score " << pHyp->m_TotalScore
	   << "  TMScore " << pHyp->m_TMScore
	   << "  LMScore " << pHyp->m_LMScore
	   << "  ALScore " << pHyp->m_ALScore
	   << "  SLScore " << pHyp->m_SLScore
	   << endl;
  cout << "Per Word    " << pHyp->m_TotalScore / L
	   << "  TMScore " << pHyp->m_TMScore / L
	   << "  LMScore " << pHyp->m_LMScore / L
	   << "  ALScore " << pHyp->m_ALScore / L
	   << "  SLScore " << pHyp->m_SLScore / L
	   << endl;
    */
    cout<<endl<< "Total Score " << pHyp->m_TotalScore;
    for (int i=0; i<pHyp->m_scaledModelScores.Size(); i++){
      cout<<" Mod "<<i<<" = "<<pHyp->m_scaledModelScores[i];
    }
    cout<<endl;


    cout << "Reorderings " << pHyp->m_NumReorderings / 2 
	 << endl;
  }

  CArray< CLatticeEdgeMS * > AllEdges;
  CArray< CLatticeEdgeMS * > AllEdges2;

  while ( pHyp->m_pPrevHyp ) {
	//cout << "Hyp " << pHyp << " PrevHyp " << pHyp->m_pPrevHyp 
	//<< " Edge " << pHyp->m_pEdge << endl;
	if ( !pHyp->m_pEdge ) {
	  //cout << "no edge - </s>" << endl;
	} else {
	  //pHyp->m_pEdge->Display();
	  AllEdges.Append( pHyp->m_pEdge );
	}

	//cout << *pHyp->m_pPrevHyp << endl;
	pHyp = pHyp->m_pPrevHyp;
  }


  // first1, display the path that has been taken
  // invert sequence of edges in list
  for ( int i = AllEdges.MaxIndex(); i >= AllEdges.MinIndex(); i-- ) {
	AllEdges2.Append( AllEdges[ i ] );
  }
  //cout << "PartialPath: ";
  //if (s_TraceOn){
  //m_pLattice->DisplayPath( AllEdges2 );
	//}

  // now display the clean translation
  CString Translation;
  CWordInfo *pWordInfo;
  CLatticeEdgeMS *pEdge;

  for ( int i = AllEdges.MaxIndex(); i >= AllEdges.MinIndex(); i-- ) {
	pEdge = AllEdges[ i ];
	pWordInfo =  pEdge->WordInfo();

	// extract sequence of words from the WordInfo structure
	if ( pWordInfo->HasTargetString() ) {
	  Translation += pWordInfo->TargetString();
	} else if ( pWordInfo->HasSourceString() ) {
	  Translation += pWordInfo->SourceString();
	} else {
	  Translation += pEdge->Word();
	}

	Translation += " ";
  }

  //  if (s_TraceOn){
  //  cout << "Translation is: " << Translation << endl;
  //  cout << "------------" << endl;
	//}
  CDecoderNB::s_TraceOn = OldTraceOn;

  return pair<CString, CArray< CLatticeEdgeMS * > >(Translation, AllEdges2);
}




void
CDecoderNB::BuildNBestList()
{
  Trace( 0, "Build N-Best List" );
  //Trace( 0, "Number of BPs is " << CDecoderNB_BP::s_NumberOfObjects++ );

  s_TraceOn = false;
  
  // iterator to iterate over list of backpointer for one coverage
  multimap< float, CDecoderNB_BP * >::iterator IterOverBPs;

  CDecoderNB_BP *pBP, *pBP_New;
  CString Translation;

  // first clear all entries
  //Trace( 0, "Clear NBestHyps - start" );

  //ARV
  // Moved the code to clear the m_NBestHyps list to the clear function
  // where it CLEARLY belongs.
  /*
  
  for ( int j = m_NBestHyps.MinIndex(); j <= m_NBestHyps.MaxIndex(); j++ ) {
	multimap< float, CDecoderNB_BP * > &CurrentBPs = m_NBestHyps[ j ];

	// loop through list of backpointers up to N
	for ( IterOverBPs = CurrentBPs.begin(); IterOverBPs != CurrentBPs.end(); ++IterOverBPs ) {
	  pBP = IterOverBPs->second;
	  delete pBP;
	  CurrentBPs.erase( IterOverBPs );
	}
  }
  */
  //Trace( 0, "Clear NBestHyps - end" );
  //Trace( 0, "Number of BPs is " << CDecoderNB_BP::s_NumberOfObjects++ );


  // -----------------------------------------------------------------------------------
  // first, we fill NBestHyps( J + 1 ) with top N hyps from 
  // m_FinalHyps.insert(pair< double, CDecoderNB_Hyp * >( pHyp->m_TotalScore , pHyp ) );
  // -----------------------------------------------------------------------------------

  TraceIf( s_TraceOn, 0, "Insert Final Hyps into BP Table" );
  TraceIf( s_TraceOn, 0, "============================== " );

  multimap< float, CDecoderNB_Hyp * >::iterator IterOverFinalHyps;
  CDecoderNB_Hyp *pHyp, *pPrevHyp;
  CDecoderNB_Hyp *LastHyp = NULL;  // for debugging


  int J = m_AllHyps.MaxIndex() + 1;  // we have added a (virtual) </s> to sentence
  double WorstScore;
  //m_NumOfNBest = 100;

  //m_NBestHyps.SetMinIndex( 0 );
  m_NBestHyps.SetMaxIndex( J );

  
  int n = 1;
  for ( IterOverFinalHyps = m_FinalHyps.begin();
		(IterOverFinalHyps != m_FinalHyps.end()) && (n <= m_NumOfNBest);
		++IterOverFinalHyps ) {

	pHyp = IterOverFinalHyps->second;
	TraceHyp( s_TraceOn, 0, "Hyp is ", pHyp );

	// create for each hyp a backtrace item and place it in the list
	CDecoderNB_BP *pBP = new CDecoderNB_BP( pHyp );

	// hyps at sentence end have to have coverage equal to sentence length
	assert( pBP->m_pHyp->m_Coverage == J );
	// and we don't want the same hyp several times
	assert( pBP->m_pHyp != LastHyp );
	LastHyp = pHyp;

	//Display( pBP, n );
	//Translation = TraceBack( pBP->m_pHyp ) + " # " + TraceForward( pBP );
	//cout << "Final Translation: " << J + 1 << " " << Translation << endl;

	m_NBestHyps[ J ].insert( pair< float, CDecoderNB_BP * >( pBP->m_TotalScore, pBP ) );

	n++;

  }
  

  int Number = m_NBestHyps[ J ].size();
  TraceIf( s_TraceOn, 0, "Number to start with is " << Number );

  // iterotor for looping back over all coverages down to 0
  multimap< float, CDecoderNB_Hyp * >::iterator IterRecombinedHyps;

  CDecoderNB_Hyp *pRecombinedHyp;


  // -----------------------------------------------------------------------------------
  // now loop backwards over coverage
  // -----------------------------------------------------------------------------------

  int j = J;
  while( j > 0 ) {

	TraceIf( s_TraceOn, 0, "Process NBest with coverage " << j );
	TraceIf( s_TraceOn, 0, "=========================== " << j );

	// we expand only top m_NumOfNBest
	n = 1;   // number of expanded

	// the list of backpointers for current coverage
	multimap< float, CDecoderNB_BP * > &CurrentBPs = m_NBestHyps[ j ];

	TraceIf( s_TraceOn, 0, "Number of Backpointers is " << CurrentBPs.size() );
	TraceIf( s_TraceOn, 0, "Listing all Backpointers" );

	// what is the score of the m_NumOfNBest entry in the current list?
	// this will be used to exit expansion of backpointers;
	// if we have still not enaugh backpointer, make WorstScore very large
	if ( CurrentBPs.size() > m_NumOfNBest ) {
	  for ( IterOverBPs = CurrentBPs.begin();
			(IterOverBPs != CurrentBPs.end()) && (n <= m_NumOfNBest);
			++IterOverBPs ) {
		pBP = IterOverBPs->second;
		WorstScore = pBP->m_TotalScore;

		//Display( pBP, n );
		//Translation = TraceBack( pBP->m_pHyp ) + " # " + TraceForward( pBP );
		//cout << "Intermediate nbest translation: " << j << " " << Translation << endl;

		n++;
	  }
	} else {
	  WorstScore = 10000000;
	  TraceIf( s_TraceOn, 0, "Not enough hyps - worst Score is 1000000" );
	}
       

	// ---------------------------------------------------------------------------------
	// 2nd level: loop through list of backpointers for current coverage up to NBest
	// ---------------------------------------------------------------------------------
	n = 1;
	for ( IterOverBPs = CurrentBPs.begin();
		  (IterOverBPs != CurrentBPs.end()) && (n <= m_NumOfNBest);
		  ++IterOverBPs ) {
	  pBP = IterOverBPs->second;

	  TraceIf( s_TraceOn, 0, "\n\nCurrent BACKPOINTER Number:" << n << " Score:" << IterOverBPs->first << " " 
			   << pBP->m_TotalScore << " HypEdge" << pBP->m_pHyp->m_pEdge << " PrevBP:" << pBP->m_pBP );

	  pHyp = pBP->m_pHyp;

	  // check if hypothesis is correct, i.e. if it has a link to a previous hypothesis
	  // here, coverage is > 0, i.e. there has to be such a link
	  if ( (pHyp->m_Coverage == 0) || (!pHyp->m_pPrevHyp) ) {

		TraceHyp( 1, 0, "ALARM: rotten hyp ", pHyp );

	  } else {

		pPrevHyp = pHyp->m_pPrevHyp;
		
		if (s_TraceOn){
		  TraceHyp( s_TraceOn, 1, "      best hyp is ",  pHyp );
		  cout << "     translation is: " << TraceBack( pHyp ) << endl;
		  TraceHyp( s_TraceOn, 1, "      pred hyp is ", pPrevHyp );
		  cout << "     translation prev hyp is: " << TraceBack( pPrevHyp ) << endl;
		}
		// take hyp pointed to, store predecessor hyp into slot with predecessor coverage
		// this is the best one with given set of features, i.e. we trace it back

		int PredecessorCoverage = pPrevHyp->m_Coverage;
		
		// this initialization is to get the score that is in the pBP
		pBP_New = new CDecoderNB_BP( pBP );
		pBP_New->m_pHyp  = pPrevHyp;           // the hyp from which this one originated
		//		pBP_New->m_pEdge = pPrevHyp->m_pEdge;  // the edger over which the previous hyp was expanded		
		TraceIf( s_TraceOn, 1, "    pBP_New Length was " << pBP_New->m_TargetLength );
		int NewBPTargetExtension = TargetLength( pPrevHyp );
		// ARV
		// add a variable so that we can remove the extension that came from this memoryyp
		//pBP_New->m_TargetLength += TargetLength( pPrevHyp );
		pBP_New->m_TargetLength += NewBPTargetExtension;
		TraceIf( s_TraceOn, 1, "    pBP_New Length is now " << pBP_New->m_TargetLength );
		TraceIf( s_TraceOn, 1, "    Predeccessor Coverage " << PredecessorCoverage );
		
	  
		// -----------------------------------------------------------------------------
		// End Condition:
		// if new coverage = 0, i.e. we are at the sentence start, then
		// the true sentence length is calculated and the correct SLScore
		// is estimated
		// -----------------------------------------------------------------------------
		if ( PredecessorCoverage == 0 ) {
		  UpdateSLScore( pBP_New, n );
		  //Sanjika 04/25
		  //UpdateLexicalSLScore(pBP_New, n);
		}


		TraceIf( s_TraceOn, 1, "  Going from Hyp " << pHyp << " back to Hyp " << pBP_New->m_pHyp );
		TraceIf( s_TraceOn, 1, "  Insert into list of BPs at Coverage " << PredecessorCoverage );
		assert( pBP_New->m_pHyp->m_Coverage == PredecessorCoverage );
		// and we don't want the same hyp several times
		assert( pBP_New->m_pHyp != LastHyp );
		LastHyp = pHyp;

		//Display( pBP_New, n );
		//Translation = TraceBack( pBP_New->m_pHyp ) + " # " + TraceForward( pBP_New );
		//cout << "Intermediate best Translation: " << j << " " << Translation << endl;

		// now, store backpointer according to coverage
		// ARV
		// only if it is actually a memoryypothesis
		
		m_NBestHyps[ PredecessorCoverage ].insert
		  ( pair< float, CDecoderNB_BP * >( pBP_New->m_TotalScore, pBP_New ) ); 
  
		n++;


		// -----------------------------------------------------------------------------
		// now we deal with the recombined hyps
		// if predecessor hyp has recombinded hyps we create backpointers for those
		// for those we have to calculate the true final score
		// -----------------------------------------------------------------------------

		// we initialize TrueScore (used to calculate true final score of recombined hyps)
		// to 0, as it will be used to exit backtracing of recombined hyps against WorstScore
		float TrueScore = 0;

		if ( pPrevHyp->m_pRecombinedHyps ) {

		  TraceIf( s_TraceOn, 1, "  Number of recombined Hyps is " << pPrevHyp->m_pRecombinedHyps->size() );


		  // ---------------------------------------------------------------------------
		  // 3rd level: loop over all recombined hypotheses
		  // ---------------------------------------------------------------------------
		  
		  
 		  for ( IterRecombinedHyps =pPrevHyp->m_pRecombinedHyps->begin();
				(IterRecombinedHyps != pPrevHyp->m_pRecombinedHyps->end()) && (TrueScore < WorstScore);
				++IterRecombinedHyps ) {
		  
			// loop through all recombined hyps
			pRecombinedHyp = IterRecombinedHyps->second;

			// we don't want the same hyp several times
			assert( pRecombinedHyp != LastHyp );
			LastHyp = pRecombinedHyp;

			TraceHyp( s_TraceOn, 1, "      recombined hyp is ", pRecombinedHyp );

			//cout << "     translation rec hyp is: " << TraceBack( pRecombinedHyp )  + " # " + TraceForward( pBP_New ) << endl;
			//cout << "     translation rec prev hyp is: " << TraceBack( pRecombinedHyp->m_pPrevHyp ) << endl;
			// calculate true Nbest list
			
			TrueScore = pBP_New->m_TotalScore + (pRecombinedHyp->m_TotalScore - pPrevHyp->m_TotalScore); 

			TraceIf( s_TraceOn, 1, "      Additional Score is "
					 << pRecombinedHyp->m_TotalScore - pPrevHyp->m_TotalScore
					 << "      True total score is " << TrueScore );

 			if ( TrueScore > WorstScore ) {
			  TraceIf( s_TraceOn, 0, "      Break: TrueScore > WorstScore " << WorstScore );
			} else {

			  // take hyp pointed to, store predecessor hyp into slot with predecessor coverage
			  // this is the best one with given set of features, i.e. we trace it back

			  //ARV 4/20 Changed to also pass the number of models
			  
			  CDecoderNB_BP *pBP_RHyp = new CDecoderNB_BP( m_NumberOfModels );
			  pBP_RHyp->m_TotalScore = TrueScore; // we propagate the final score inside the backpointer
			  // also other scorse
			  
			  //not used anywhere
			  //pBP_RHyp->m_ACScore = pBP_New->m_ACScore + (pRecombinedHyp->m_ACScore - pPrevHyp->m_ACScore);
			  


			  //ARV: 04/20 Generalizing the decoder to handle a set of scores
			  /*
			    //pBP_RHyp->m_TMScore = pBP_New->m_TMScore + (pRecombinedHyp->m_TMScore - pPrevHyp->m_TMScore);
			    //pBP_RHyp->m_LMScore = pBP_New->m_LMScore + (pRecombinedHyp->m_LMScore - pPrevHyp->m_LMScore);
			    //pBP_RHyp->m_ALScore = pBP_New->m_ALScore + (pRecombinedHyp->m_ALScore - pPrevHyp->m_ALScore);
			  */
			  //set each one via the array process
			  //this way it does not need to be set by hand

			  //make sure the hyp has the size for this
			  for (int m=0; m<m_NumberOfModels; m++){
			    pBP_RHyp->m_scaledModelScores[m] = pBP_New->m_scaledModelScores[m] + (pRecombinedHyp->m_scaledModelScores[m] - pPrevHyp->m_scaledModelScores[m]); 
			  }

			  //this is just set here now, but it will be updated at the end of the n-best list
			  //since we need to track the target length carefully
			  //ARV 4/20
			  //pBP_RHyp->m_SLScore = pBP_New->m_SLScore;

			  TraceIf( s_TraceOn, 1, "    BP till here length " << pBP_New->m_TargetLength );
			  pBP_RHyp->m_TargetLength = pBP_New->m_TargetLength + TargetLength( pRecombinedHyp ) - NewBPTargetExtension;
			  TraceIf( s_TraceOn, 1, "    TargetLength is now " << pBP_RHyp->m_TargetLength );
			  pBP_RHyp->m_pHyp = pRecombinedHyp;           // the hyp from which this one originated
			  //			  pBP_RHyp->m_pEdge = pRecombinedHyp->m_pEdge;  // the edger over which the previous hyp was expanded
			  pBP_RHyp->m_pBP   = pBP;                      // and connecting the backpointers
	  
			  // if new coverage = 0, i.e. we are at the sentence start, then
			  // the true sentence length is calculated and the correct SLScore
			  // is estimated
			  if ( PredecessorCoverage == 0 ) {
	        	    TraceIf( s_TraceOn, 1, " Reached sentence end " );
			    UpdateSLScore( pBP_RHyp, n );
			    //Sanjika 04/25
			    //UpdateLexicalSLScore(pBP_RHyp, n);
			  }

			  TraceIf( s_TraceOn, 0, "  Going from Hyp " << pRecombinedHyp << " back to Hyp " << pBP_New->m_pHyp );
			  TraceIf( s_TraceOn, 0, "  Insert into list of BPs at Coverage " << PredecessorCoverage );
			  assert( pBP_New->m_pHyp->m_Coverage == PredecessorCoverage );

			  // we don't want the same hyp several times
			  //assert( pRecombinedHyp != LastHyp );
			  //LastHyp = pRecombinedHyp;

			  //Display( pBP_RHyp, -1 );
			  //Translation = TraceBack( pBP_RHyp->m_pHyp ) + " # " + TraceForward( pBP_RHyp );
			  //cout << "Intermediate rec. Translation: " << j << " " << Translation << endl;
			  m_NBestHyps[ PredecessorCoverage ].insert
				(pair< float, CDecoderNB_BP * >( pBP_RHyp->m_TotalScore, pBP_RHyp ) ); 

			} //( TrueScore > WorstScore )
			
		  } // for ( IterRecombinedHyps
		  
		} // ( pPrevHyp->m_pRecombinedHyps )
		// now set the length extension for the initial best hypothesis
		
	  }

	} // for ( IterOverBPs 

	j--;
  }

  // now we need to reconstruct the actual translations
  // loop through list of backpointers up to N
  TraceIf( s_TraceOn, 0, "  Final List of Backpointers " );

  n = 0;

  // if you are tracing
  if (s_TraceOn){
    multimap< float, CDecoderNB_BP * > &CurrentBPs = m_NBestHyps[ 0 ];
    for ( IterOverBPs = CurrentBPs.begin(); 
	  (IterOverBPs != CurrentBPs.end()) && (n < m_NumOfNBest);
	  ++IterOverBPs ) {
      
      CDecoderNB_BP *pBP = IterOverBPs->second;
      
      TraceIf( 1, 0, "\nBackpointer " << n << " " << IterOverBPs->first
	       << " TotalScore=" << pBP->m_TotalScore );
      
      //cout << endl << "NBest " << n << endl;
      TraceStringAndPath( pBP, n );
      
      n++;
    }
    
    
    cout << "Number of true N-best translations is: " << n << endl;
    s_TraceOn = false;
    
    Trace( 1, "Number of BPs is " << CDecoderNB_BP::s_NumberOfObjects++ );
    s_TraceOn = false;
  }
}

// ARV added to get the translation back
// ARV 04/28
pair<CString, CArray< CLatticeEdgeMS * > >
CDecoderNB::TraceStringAndPath( CDecoderNB_BP * pBP, int n )
{
  CString Translation;

  //ARV 04/28
  CArray< CLatticeEdgeMS * > AllEdges;

  // length of source sentence to calculate per word scores
  int Start = m_pLattice->StartNode()->Number();
  int End   = m_pLattice->EndNode()->Number();
  int L = End - Start;


  if (s_TraceOn){
  cout << "Total Score  " << pBP->m_TotalScore
       << "  TMScore "    << pBP->m_scaledModelScores[c_TM]
       << "  LMScore "    << pBP->m_scaledModelScores[c_LM]
       << "  ALScore "    << pBP->m_scaledModelScores[c_AL]
       << "  SLScore "    << pBP->m_scaledModelScores[c_SL]
	   << endl;
  cout << "Length "     << pBP->m_TargetLength << "  Reordering Jumps " << pBP->m_pHyp->m_NumReorderings << endl;
  }
  // follow the pBP to access the edges and thereby the actual translation
  Translation = "";
  AllEdges.RemoveAll();
  
  // backpointer points to initial hypothesis which has no edge
  // there, move to next backpointer
  pBP = pBP->m_pBP;

  while ( pBP->m_pHyp->m_pEdge ) {
	
	// to display the path, collect all edges
	AllEdges.Append( pBP->m_pHyp->m_pEdge );
	
	// access to the partial translations via the WordInfo
	CWordInfo *pWordInfo = pBP->m_pHyp->m_pEdge->WordInfo();
	
	// extract sequence of words from the WordInfo structure
	Translation += " ";
	if ( pWordInfo->HasTargetString() ) {
	  Translation += pWordInfo->TargetString();
	} else if ( pWordInfo->HasSourceString() ) {
	  Translation += pWordInfo->SourceString();
	} else {
	  Translation += pBP->m_pHyp->m_pEdge->Word();
	}
	
	pBP = pBP->m_pBP;
  }
  
  // now we are at the sentence end again, i.e. we can access 
  // the hypothesis to get all interesting scores
  // well, need to go one additional step, to the hypothesis
  // which has also the sentence end LM score and sentence
  // length model score;
  //pHyp = pBP->m_pBP->m_pHyp;
  
  
  if (s_TraceOn){
    cout << "PartialPath " << n << " is: ";
    m_pLattice->DisplayPath( AllEdges );
    cout << "True NBest Translation " << n << " is:" << Translation << endl;
  }
  return pair<CString, CArray< CLatticeEdgeMS * > >(Translation, AllEdges);
}




CString
CDecoderNB::TraceForward( CDecoderNB_BP * pBP )
{
  CString Translation;
  CArray< CLatticeEdge * > AllEdges;

  // length of source sentence to calculate per word scores
  int Start = m_pLattice->StartNode()->Number();
  int End   = m_pLattice->EndNode()->Number();
  int L = End - Start;


  // follow the pBP to access the edges and thereby the actual translation
  Translation = "";
  AllEdges.RemoveAll();

  // edges are leading to the hyps the backpointer is pointing to
  // i.e. go to the next backpointer
  if ( pBP->m_pBP ) {

	pBP = pBP->m_pBP;

	while ( pBP->m_pHyp->m_pEdge ) {

	  // to display the path, collect all edges
	  AllEdges.Append( pBP->m_pHyp->m_pEdge );
	  
	  // access to the partial translations via the WordInfo
	  CWordInfo *pWordInfo = pBP->m_pHyp->m_pEdge->WordInfo();
	  
	  //cout << " | " << pBP->m_pEdge->WordInfo()->TargetString();
	  
	  // extract sequence of words from the WordInfo structure
	  Translation += " ";
	  if ( pWordInfo->HasTargetString() ) {
		Translation += pWordInfo->TargetString();
	  } else if ( pWordInfo->HasSourceString() ) {
		Translation += pWordInfo->SourceString();
	  } else {
		Translation += pBP->m_pHyp->m_pEdge->Word();
	  }
	  
	  pBP = pBP->m_pBP;

	}
  }

  return Translation;

}




//this Rescore function is hard coded to do Model I rescoring, therefore it is being 
//commented out. Rescoring should be done on the n-best list data structure 

void CDecoderNB::Rescore( )
{

  s_TraceOn = False;
  TraceIf( s_TraceOn, 0, "CDecoderNB::Rescore" );


  // go over the n-best list, i.e. the backpointers
  // for each create the translation as array of word indices
  // calculate the TM score using the lexicon

  CArray< CWordIndex > Source;
  CArray< CWordIndex > Target;
	
  for ( int j = m_pSourceSentence->MinIndex(); j <= m_pSourceSentence->MaxIndex(); j++ ) {
	CString Word = m_pSourceSentence->Word( j, m_pVocabulary );
	Source.Append( m_pLexicon->SourceVocabulary()->GetIndex( Word ) );
  }

  int n = 0;
  multimap< float, CDecoderNB_BP * > &CurrentBPs = m_NBestHyps[ 0 ];
  // iterator to iterate over list of backpointer for one coverage
  multimap< float, CDecoderNB_BP * >::iterator IterOverBPs;

  CDecoderNB_BP *pBP;
  //CDecoderNB_BP *pRescoredBP;

  /*
  // clear backpointer list for rescored hyps
  for ( IterOverBPs = m_NBestHypsRescored.begin(); IterOverBPs != m_NBestHypsRescored.end(); ++IterOverBPs ) {
	pBP = IterOverBPs->second;
	delete pBP;
	m_NBestHypsRescored.erase( IterOverBPs );
  }
  */


  for ( IterOverBPs = CurrentBPs.begin(); 
		(IterOverBPs != CurrentBPs.end()) && (n < m_NumOfNBest);
		++IterOverBPs ) {

	pBP = IterOverBPs->second;

	TraceIf( 0, 0, "\nBackpointer " << n << " " << IterOverBPs->first
			 << " TotalScore=" << pBP->m_TotalScore );

	//cout << endl << "Rescore " << n << " ";
	//Display( pBP->m_pBP, n );

	// follow the pBP to access the edges and thereby the actual translation
	//Translation = "";
	//AllEdges.RemoveAll();
	Target.RemoveAll();

	// first BP points to initial hyp which has no edge, i.e. start from next
	pBP = pBP->m_pBP;

	while ( pBP->m_pHyp->m_pEdge ) {
	
	  //pBP->m_pHyp->m_pEdge->Display();
	  
	  // access to the partial translations via the WordInfo
	  CWordInfo *pWordInfo = pBP->m_pHyp->m_pEdge->WordInfo();
	  
	  if ( pWordInfo->HasTargetString() ) {
		CSentence &Words = pWordInfo->TargetWords();
		for ( int n = Words.MinIndex(); n <= Words.MaxIndex(); n++ ) {
		  //Translation += pWordInfo->TargetString();
		  //cout << Words.Word( n ) << " T " << m_pLexicon->TargetVocabulary()->GetIndex( Words.Word( n ) ) << endl;
		  Target.Append( m_pLexicon->TargetVocabulary()->GetIndex( Words.Word( n ) ) );
		}
	  } else if ( pWordInfo->HasSourceString() ) {
		CSentence &Words = pWordInfo->SourceWords();
		for ( int n = Words.MinIndex(); n <= Words.MaxIndex(); n++ ) {
		  //Translation += pWordInfo->SourceString();
		  //cout << Words.Word( n ) << " S " << m_pLexicon->TargetVocabulary()->GetIndex( Words.Word( n ) ) << endl;
		  Target.Append( m_pLexicon->SourceVocabulary()->GetIndex( Words.Word( n ) ) );
		}
	  } else {
		//Translation += pBP->m_pHyp->m_pEdge->Word();
		CString Word = pBP->m_pHyp->m_pEdge->Word();
		//cout << Word << " " << m_pLexicon->SourceVocabulary()->GetIndex( Word ) << endl;
		Target.Append( m_pLexicon->SourceVocabulary()->GetIndex( Word ) );
	  }
	
	  pBP = pBP->m_pBP;
	}
	//cout << endl;
  
	//for ( int i = Source.MinIndex(); i <= Source.MaxIndex(); i++ ) {
	//cout << " " << Source[ i ];
	//}
	//cout << endl;

	//for ( int i = Target.MinIndex(); i <= Target.MaxIndex(); i++ ) {
	//cout << " " << Target[ i ];
	//}
	//cout << endl;

	// now get the TM score
	float Prob = 1.0;
	float SumProb = 0.0;
	float LogProb = 0.0;

	if ( m_pLexicon ) {
	  //cout << "Source Length is " << m_pSourceSentence->Length() 
	  //   << "Target Length is " << Target.Size() << endl;

	  for ( int j = Source.MinIndex(); j <= Source.MaxIndex(); j++ ) {
		SumProb = 0.0;
		for ( int i = Target.MinIndex(); i <= Target.MaxIndex(); i++ ) {
		  //cout << m_pLexicon->SourceVocabulary()->GetWord( Source[ j ] ) << " " 
		  //   << m_pLexicon->TargetVocabulary()->GetWord( Target[ i ] ) << " "
		  //   << m_pLexicon->Pr( Source[ j ], Target[ i ] ) << endl;
		  SumProb += m_pLexicon->Pr( Source[ j ], Target[ i ] );
		}
		LogProb += -log( SumProb );
		//cout << SumProb << endl;
	  }
	  
	}
	else {
	  LogProb = 10.0;
	}
	
	//if (s_TraceOn){
	cout << "New TMScore " << LogProb << endl;
	  //}
	  
	pBP->m_scaledModelScores[ pBP->m_scaledModelScores.MaxIndex() ] = LogProb;


	/*
	pBP = IterOverBPs->second;
	pRescoredBP = new CDecoderNB_BP( pBP );

	pRescoredBP->m_TotalScore -= pRescoredBP->m_TMScore;


	pRescoredBP->m_TotalScore += LogProb;
	pRescoredBP->m_TMScore = LogProb;
	pRescoredBP->m_pBP = pBP->m_pBP;

	m_NBestHypsRescored.insert( pair< float, CDecoderNB_BP * >( pRescoredBP->m_TotalScore , pRescoredBP ) );
	*/
	n++;
  }


  /*
  // now we need to reconstruct the actual translations
  // loop through list of backpointers up to N
  TraceIf( s_TraceOn, 0, "  Final List of Backpointers " );

  n = 0;

  for ( IterOverBPs = m_NBestHypsRescored.begin(); 
		IterOverBPs != m_NBestHypsRescored.end();
		++IterOverBPs ) {

	pRescoredBP = IterOverBPs->second;

	TraceIf( 1, 0, "\nBackpointer Rescored " << n << " " << IterOverBPs->first
			 << " TotalScore=" << pRescoredBP->m_TotalScore );

	cout << "Total Score  " << pRescoredBP->m_TotalScore
		 << "  TMScore "    << pRescoredBP->m_TMScore
		 << "  LMScore "    << pRescoredBP->m_LMScore
		 << "  ALScore "    << pRescoredBP->m_ALScore
		 << "  SLScore "    << pRescoredBP->m_SLScore
		 << endl;

	cout << "Rescored NBest " << n << " is: ";
	cout << TraceForward( pRescoredBP ) << endl;

	n++;
  }

  */

}



double
CDecoderNB::Rescore( CDecoderNB_Hyp *pHyp )
{
  int Start = m_pLattice->StartNode()->Number();
  int End   = m_pLattice->EndNode()->Number();
  int L = End - Start;

  bool s_TraceOn = True;
  
  /*
  if (s_TraceOn){
	cout << "Old TMScore " << pHyp->m_TMScore << endl;
  }
  */

  CArray< CLatticeEdge * > Edges;

  // trace back and collect the edges
  while ( pHyp->m_pPrevHyp ) {
	//cout << "Hyp " << pHyp << " PrevHyp " << pHyp->m_pPrevHyp 
	//<< " Edge " << pHyp->m_pEdge << endl;
	if ( pHyp->m_pEdge ) {
	  Edges.Insert( 0, pHyp->m_pEdge );
	}
	pHyp = pHyp->m_pPrevHyp;
  }


  //cout << "Path for Rescoring: ";
  //if (s_TraceOn){
  //m_pLattice->DisplayPath( Edges );
	//}

  // now convert the path into array of word IDs
  CArray< CWordIndex > Target, Words;
  CWordInfo *pWordInfo;
  CLatticeEdgeMS *pEdge;

  for ( int i = Edges.MinIndex(); i <= Edges.MaxIndex(); i++ ) {
	pEdge = (CLatticeEdgeMS *)Edges[ i ];
	pWordInfo =  pEdge->WordInfo();

	// extract sequence of words from the WordInfo structure
	// no WordIndex available, only String (that's bad), so get index here
	if ( pWordInfo->HasTargetString() ) {

	  CSentence &Words = pWordInfo->TargetWords();
	  //cout << "Edge has " << Words.Size() << " target words " 
	  //   << Words << endl;
	  for ( int n = Words.MinIndex(); n <= Words.MaxIndex(); n++ ) {
		//cout << "word index = "  << Words.Word( n )  << endl;
		//cout << "word = " << m_pVocabulary->GetIndex( Words.Word( n ) ) << endl;
		Target.Append( m_pVocabulary->GetIndex( Words.Word( n ) ) );
	  }

	} else if ( pWordInfo->HasSourceString() ) {

	  CSentence &Words = pWordInfo->SourceWords();
	  //cout << "Edge has " << Words.Size() << " target words " 
	  //   << Words << endl;
	  for ( int n = Words.MinIndex(); n <= Words.MaxIndex(); n++ ) {
		//cout << "word index = "  << Words.Word( n )  << endl;
		//cout << "word = " << m_pVocabulary->GetIndex( Words.Word( n ) ) << endl;
		Target.Append( m_pVocabulary->GetIndex( Words.Word( n ) ) );
	  }

	} else {

	  CWordIndex Word = m_pVocabulary->GetIndex( pEdge->WordIndex() );
	  //cout << "edge has word " << Word << endl;
	  Target.Append( Word );
	}

	for ( int n = Words.MinIndex(); n <= Words.MaxIndex(); n++ ) {
	  //cout << "append word " << Words[ n ] << " to array of all target words " << endl;
	  Target.Append( Words[ n ] );
	}
  }

  //  if (s_TraceOn){
  //cout << "Words are: " << Target << endl;

  // now get the TM score
  float Prob = 1.0;
  float SumProb = 0.0;
  float LogProb = 0.0;

  if ( m_pLexicon ) {
	cout << "Source Length is " << m_pSourceSentence->Length() 
		 << "Target Length is " << Target.Size() << endl;

	for ( int j = m_pSourceSentence->MinIndex(); j <= m_pSourceSentence->MaxIndex(); j++ ) {
	  SumProb = 0.0;
	  for ( int i = Target.MinIndex(); i <= Target.MaxIndex(); i++ ) {
		//cout << (*m_pSourceSentence)[ j ] << " " <<  Target[ i ] << " "
		//	   << m_pLexicon->Pr( (*m_pSourceSentence)[ j ], Target[ i ] ) << endl;
		SumProb += m_pLexicon->Pr( (*m_pSourceSentence)[ j ], Target[ i ] );
	  }
	  LogProb += -log( SumProb );
	  //cout << SumProb << endl;
	}

  }
  else {
	LogProb = 10.0;
  }

  if (s_TraceOn){
	cout << "New TMScore " << LogProb << endl;
  }

  return LogProb;

}



void
CDecoderNB::DisplayAllHypsWithCoverage( int Coverage, int N )
{

  CDecoderNB_Hyp *pHyp;

  // first we fetch the set of hypotheses which all have coverage Coverage
  // and we need an iterater to access the hyps in this set
  HypContainer< CDecoderNB_Hyp *, HypHash, HypEqual > &Hyps = m_AllHyps[ Coverage ];

  // and we need an iterater to access the hyps in this set
  HypContainer< CDecoderNB_Hyp *, HypHash, HypEqual >::iterator IterHyps;


  TraceIf( s_TraceOn, 0, "\n\n  Number of hyps with coverage " 
		   << Coverage << " is " << Hyps.size() );
  //TraceIf( s_TraceOn, 0, "Address of Hyps is: " << &Hyps );

  // display only top N hyps
  int n = 0;
  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {
	if ( n < N ) {
	  pHyp = *IterHyps;
	  cout << pHyp->m_TotalScore << " " << pHyp->m_Coverage << " " 
		   << pHyp->m_CoverageStart << " " << pHyp->m_CoveragePattern << endl;
	  n++;
	}
  }

}

//ARV 04/28
//handles the multiple score transducer. The first score goes in the traditional c_TM position
//and all other scores go after the first model index. 
void CDecoderNB::GetTMScoreMS( CLatticeEdgeMS *pEdge )
{
  //  CString Word;
  double TMScore;

  if ( pEdge->WordInfo()->HasTargetString() ) {
    //cout<<"Set here"<<endl;
    
    if ( m_ScalingFactors.Size() > c_TM + ((CLatticeEdgeMS*)pEdge)->Scores().Size() ) {
      // if we have not enough socres, we supply additional scores
      cout << "Mismatch in number of models and scaling factors: ScalingFactorsSize: " << m_ScalingFactors.Size() << endl;
      ((CLatticeEdgeMS*)pEdge)->Display();
      cout << endl;
      exit ( -11 );
    }
    
    CNumericalArray<float> PhraseScores = ((CLatticeEdgeMS*)pEdge)->Scores();

    //we are going to put the models in order, the first one starting at c_FirstModelIndex and continuing on down
    int offset = 0;
    for ( int i = PhraseScores.MinIndex(); i <= PhraseScores.MaxIndex(); i++ ) {
      
      m_pNumberOneHypothesis->m_scaledModelScores[ offset + c_TM ]
        += m_ScalingFactors[ c_TM + offset ] * PhraseScores[ i ];
      
      m_pNumberOneHypothesis->m_TotalScore
        += m_ScalingFactors[ c_TM + offset ] * PhraseScores[ i ];
      
      offset++;
    }
    
  } else {

    //ARV 04/28 N.B the source word probabilities for all the true probability modesl should be set to approx -log (0.0) = 10
    //no target string, need to increase the cost of all the translation model factors

    //still start from the second phrase index forward, but dont do the last one, since its a count variable
    //we dont know how many phrase score there are, so we assume consistency with the number of scaling factors
    int offset = 0;
    for ( int i = c_TM; i < m_ScalingFactors.Size(); i++ ) {

      m_pNumberOneHypothesis->m_scaledModelScores[ offset + c_TM ]
        += m_ScalingFactors[c_TM + offset] * c_TMScoreForSourceWord;
      
      m_pNumberOneHypothesis->m_TotalScore
        += m_ScalingFactors[ c_TM + offset ] * c_TMScoreForSourceWord;
      
      offset++;
    }
    
    //pEdge->Display();
    
  }
}


void
CDecoderNB::GetTMScore( CLatticeEdgeMS *pEdge )
{
  //  CString Word;
  double TMScore;

  // - translation score (i.e. accumulated transducer score
  // for original edges, use penalty
  /*
  Word = pEdge->Word();
  if ( Word[ 0 ] == '@' ) {
	TMScore = pEdge->WordInfo()->Score();
	TraceIf( s_TraceOn, 0, "      Translation Score for labeled edge: " << TMScore );
  } else {
	TMScore = 10;
	TraceIf( s_TraceOn, 0, "      Translation Score for source language edge: " << TMScore );
  }
  return TMScore;
  */

  //TMScore = m_ScalingFactors[c_TM] * pEdge->WordInfo()->Score();

  //its nulll when its over the source word
  if (pEdge->TransducerItem()!=NULL){
    TMScore = (pEdge->TransducerItem()->ScalingFactor())*pEdge->WordInfo()->Score();
  }else{
    TMScore = pEdge->WordInfo()->Score();
  }


  //ARV 4/20
  //m_pNumberOneHypothesis->m_TMScore += (m_ScalingFactors[c_TM] * TMScore);
  //m_pNumberOneHypothesis->m_TotalScore += (m_ScalingFactors[c_TM] * TMScore);
  m_pNumberOneHypothesis->m_scaledModelScores[c_TM] += (m_ScalingFactors[c_TM] * TMScore);
  m_pNumberOneHypothesis->m_TotalScore += (m_ScalingFactors[c_TM] * TMScore);

}



void
CDecoderNB::GetLMScore( CDecoderNB_Hyp *pHyp, CLatticeEdgeMS *pEdge )
{

  CLMState LMState = pHyp->m_LMState;
  CLMState LMStateNew = pHyp->m_LMState;
  CLMState NewLMState;
  CLMState NewLMStateNew;
  double LMScore = 0.0;
  double LMScoreNew = 0.0;


  // strings have alreay been converted into indices according to the LM vocabulary
  // now get the LM score by looping over all word indices

  CNumericalArray< long > & WordIndices2 = pEdge->LMWordIndices();
  LMState = pHyp->m_LMState;

  //pEdge->Display();

  /*
  for ( int i = WordIndices2.MinIndex(); i <= WordIndices2.MaxIndex(); i++ ) {
    LMScore -= m_pLMSentenceState->LogProb( LMState, WordIndices2[ i ] );
    cout << "word index = " << WordIndices2[ i ] << endl;
    TraceIf( s_TraceOn, 3, "Version 2: LM-Score for " << WordIndices2[ i ] << ": " 
             << -m_pLMSentenceState->LogProb( LMState, WordIndices2[ i ] ) 
             << " total LM-Score = " << LMScore );
    
    NewLMState = m_pLMSentenceState->AppendWord( LMState, WordIndices2[ i ] );
    TraceIf( s_TraceOn, 3, "Version 2: New LM-State: " << NewLMState );
    LMState = NewLMState;
    cout << "Old " << LMScore << " " << LMState << endl;

    LMScoreNew -= m_pLMSentenceState->LogProb( LMStateNew, WordIndices2[ i ], NewLMStateNew );
    LMStateNew = NewLMStateNew;
    cout << "New " << LMScoreNew << " " << LMStateNew << endl;
  }
  */

  {
    CLMState LMState = pHyp->m_LMState;
    CLMState NewLMState;
    double LMScore = 0.0;
    vector< LMWordIndex > x;

    for ( int i = WordIndices2.MinIndex(); i <= WordIndices2.MaxIndex(); i++ ) {
      x.push_back( WordIndices2[ i ] );
    }

    LMScore =- m_pLMSentenceState->LogProb( LMState, x, NewLMState ); 
    //cout << "One step " << LMScore << " " << NewLMState << endl;
  }


  //ARV 4-20
  //m_pNumberOneHypothesis->m_LMScore += m_LMScale * LMScore;
  //m_pNumberOneHypothesis->m_TotalScore += m_LMScale * LMScore;

  m_pNumberOneHypothesis->m_scaledModelScores[c_LM] += m_ScalingFactors[c_LM] * LMScore;
  m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[c_LM] * LMScore;


  m_pNumberOneHypothesis->m_LMState = NewLMState;

}



void
CDecoderNB::GetALScore( CDecoderNB_Hyp *pHyp, CLatticeEdgeMS *pEdge )
{
  // small extra cost for reordering
  // when and how do we know that we are reordering?
  // when last edge and current edge are not sequential
  // we ended in pHyp->m_pEdge->ToNode() and start from pEdge->FromNode()
  // if these two nodes are the same, then monoton decoding
  // otherwise a jump

  // currently, each jumps incures the same small cost
  // a proper position alignment model should be used.

  double ALScore;
  int NumberOfPrevNode;


  // the initial hypothesis has no edge, but we know that the node 
  // is the StartNode of the lattice
  /*
  if ( !pHyp->m_pEdge ) {
	NumberOfPrevNode = 0;
  } else {
	NumberOfPrevNode = pHyp->m_pEdge->ToNode()->Number();
  }

  
  if ( pEdge->FromNode()->Number() != NumberOfPrevNode ) {
	m_pNumberOneHypothesis->m_NumReorderings++;
	if ( m_pNumberOneHypothesis->m_NumReorderings > m_pNumberOneHypothesis->s_MaxNumReorderings ) {
	  ALScore = 100;
	} else {
	ALScore = 1.0;
	}
  } else {
	ALScore = 0.0;
  }
  */


  // here we have a simple jump model:
  // Forward: Prev->ToNode - 1 -> Next->FromNode
  // Backward: Prev->ToNode -> Next->FromNode + 1
  // pHyp has pointer to last traversed edge, pEdge is newly travesed edge

  float Jump;
  int JumpStart, JumpEnd;

  // we start from
  if ( !pHyp->m_pEdge ) {
	JumpStart = 0;
  } else {
	JumpStart = pHyp->m_pEdge->ToNode()->Number();
  }
  
  // we jump to
  JumpEnd = pEdge->FromNode()->Number() + 1;

  // the jump with is
  Jump = JumpEnd - JumpStart;
  
  //ORIGINAL Gaussian scoring
  ALScore = 0.25 * ( Jump - 1.0 ) * ( Jump - 1.0 );
  
  //ABS VALUE SCORING
  //ALScore = abs(Jump-1.0);
  

  
  if ( pEdge->FromNode()->Number() != JumpStart ) {
	m_pNumberOneHypothesis->m_NumReorderings++;
	
	//ORIGINAL Gaussian Scoring
	// this is commented out f
	if ( m_pNumberOneHypothesis->m_NumReorderings > m_pNumberOneHypothesis->m_pDecoderInfo->m_MaxNumReorderings ) {
	  ALScore = 100;
	}
	
  }
  
  //cout << "Jump is " << Jump << " ALScore " << ALScore << "NumRe:"<<m_pNumberOneHypothesis->m_NumReorderings<< endl;

  //ARV 4/20
  //m_pNumberOneHypothesis->m_ALScore += m_ALScale * ALScore;
  //m_pNumberOneHypothesis->m_TotalScore += m_ALScale * ALScore;

  m_pNumberOneHypothesis->m_scaledModelScores[c_AL] += m_ScalingFactors[c_AL] * ALScore;
  m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[c_AL] * ALScore;


}



void
CDecoderNB::GetPCScore( CDecoderNB_Hyp *pHyp, CLatticeEdgeMS *pEdge )
{
  //-log(2.718) = -1, but we just use +1 as a count and expect negative model scores
  m_pNumberOneHypothesis->m_scaledModelScores[ c_PL ] += m_ScalingFactors[ c_PL ] * (1);
  m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ c_PL ] * (1);
}



void
CDecoderNB::GetWCScore( CDecoderNB_Hyp *pHyp, CLatticeEdgeMS *pEdge )
{
  //cout << "GetWCScore" << endl;

  // we take the number of generate target words
  int NumTargetWords= TargetLength( pEdge );

  //cout << "GetWCScore " << NumTargetWords << endl;

  m_pNumberOneHypothesis->m_scaledModelScores[ c_SL ] += m_ScalingFactors[ c_SL ] * NumTargetWords;
  m_pNumberOneHypothesis->m_TotalScore += m_ScalingFactors[ c_SL ] * NumTargetWords;

  //cout << "GetWCScore " << m_pNumberOneHypothesis->m_TargetLength << " " << m_pNumberOneHypothesis->m_scaledModelScores[ c_SL ] << endl;
}



int
CDecoderNB::TargetLength( CDecoderNB_Hyp *pHyp )
{
  // length of translation
  int L = 0;
  if ( pHyp->m_pEdge ) {  // at sentence end we have no edge
    CWordInfo *pWordInfo = pHyp->m_pEdge->WordInfo();
    if ( pWordInfo->HasTargetString() ) {
      //cout << "has target string" << endl;
      L = pWordInfo->TargetStringLength();
    } else if ( pWordInfo->HasSourceString() ) {
      L = pWordInfo->SourceStringLength();
      //cout << "has source string" << endl;
    } else {
      //cout << "no source and no target string" << endl;
      L = 1;
    }
  } else {
    //cout << "no edge in get target length" << endl;
  }


  return L;
}



int
CDecoderNB::TargetLength( CLatticeEdgeMS *pEdge )
{
  // length of translation
  int L = 0;
  if ( pEdge ) {  // at sentence end we have no edge
    CWordInfo *pWordInfo = pEdge->WordInfo();
    if ( pWordInfo->HasTargetString() ) {
      //cout << "has target string" << endl;
      L = pWordInfo->TargetStringLength();
    } else if ( pWordInfo->HasSourceString() ) {
      L = pWordInfo->SourceStringLength();
      //cout << "has source string" << endl;
    } else {
      //cout << "no source and no target string" << endl;
      L = 1;
    }
  } else {
    ;//cout << "no edge in get target length" << endl;
  }


  return L;
}



void
CDecoderNB::UpdateSLScore( CDecoderNB_BP *pBP, int &n )
{
  float TrueSLScore;
  int J = m_pLattice->EndNode()->Number();
  int I = pBP->m_TargetLength;
  // ARV
  // changed to reflect the simple couting score
  //TrueSLScore = m_SLScale * 0.25 * ( m_SLStretch * J - I ) * ( m_SLStretch * J - I );

  //TrueSLScore = -1.0*I*m_SLScale;
  //TrueSLScore = m_SLScale*(2*J -I);

  //ARV 04/02
  //TrueSLScore = m_SLScale*I;
  TrueSLScore = m_ScalingFactors[c_SL] * I;

  //ARV
  // We just fix the score now
  /*
  if (TrueSLScore != pBP->m_SLScore){
    cout<<"TrueScore:"<<TrueSLScore<<" OldScore:"<<pBP->m_SLScore<<endl;
    cout<<"Score Mismatch -------------------------"<<endl;
  
  }
  */
  //cout <<"Sentence Length " << I << " " << TrueSLScore << endl;

  s_TraceOn = false;
  TraceIf( s_TraceOn, 0, "    Old SLScore was " << pBP->m_SLScore
		   << " New SLScore is " << TrueSLScore
		   << " Correct length is " << pBP->m_TargetLength );



  //ARV 4/20
  //pBP->m_TotalScore -= pBP->m_SLScore;   // subtract old SLScore  
  //pBP->m_SLScore = TrueSLScore;
  //pBP->m_TotalScore += TrueSLScore;        // add new SLScore

  pBP->m_TotalScore -= pBP->m_scaledModelScores[ c_SL ];   // subtract old SLScore  
  pBP->m_scaledModelScores[ c_SL ] = TrueSLScore;
  pBP->m_TotalScore += TrueSLScore;        // add new SLScore
}



//Sanjika 04/25
//Lexicalized SL model
/*
  //ARV 04/28
  //commented out until the lexicalized SL model has been tested and shown to give improvements

  void CDecoderNB::InitLexicalSLScoreEnd(){
  CSentence NewSourceSentence;
  for(int j=m_pSourceSentence->MinIndex(); j<=m_pSourceSentence->MaxIndex(); j++) {
  NewSourceSentence.SetWord( m_pSourceSentence->Word(j,m_pVocabulary),j+1,m_pVocabulary);
  }
  m_pSLenModel->GenerateTargetLengthProbabilities( &NewSourceSentence );
  }
  double CDecoderNB::GetLexicalSLScoreEnd(CDecoderNB_Hyp* pHyp){
  int I = pHyp->m_TargetLength;
  double SLenProb = -log(m_pSLenModel->GetPr( I ));
  return SLenProb;
  }
  

void CDecoderNB::UpdateLexicalSLScore(CDecoderNB_BP *pBP, int &n ){
  float TrueSLScore;
  int I = pBP->m_TargetLength;
  double SLenProb = m_pSLenModel->GetPr( I );
  TrueSLScore =  m_ScalingFactors[c_LSL] *-log( SLenProb );

  //remove contribution that was there before
  pBP->m_TotalScore -= pBP->m_scaledModelScores[c_LSL];
  //add correct contribution
  pBP->m_scaledModelScores[c_LSL] = TrueSLScore;
  pBP->m_TotalScore += TrueSLScore;  
  
}
*/



void
CDecoderNB::WriteLattice()
{
  CArray< CLatticeEdge * > Edges = m_pLattice->Edges();
  CLatticeEdgeMS *pEdge;

  cout << "BEGIN_LATTICE" << endl;
  cout << "(" << endl;
  for ( int i = Edges.MinIndex(); i <= Edges.MaxIndex(); i++ ) {
    pEdge = (CLatticeEdgeMS *)Edges[ i ];
    //	pEdge->Display();
    //cout << endl;
    if ( (pEdge->Word())[0] == '@' ) {
      double Score = exp( -(pEdge->WordInfo()->Score() ) );
      cout << '(' << pEdge->Startframe() << " " << (pEdge->Endframe() - 1) << " ";
      cout << "|||" << pEdge->WordInfo()->TargetString() << "|||" << " ";
      cout << " " << Score << " ";
      cout << "|||"<< pEdge->WordInfo()->SourceString() << "|||";
      cout << ')';
      cout << endl;
    }
  }
  cout << ")" << endl;
  cout << "END_LATTICE" << endl;

}



void
CDecoderNB::TraceHyp( bool TraceOn, int TraceLevel, CString Comment, CDecoderNB_Hyp *pHyp )
{
  TraceIf( TraceOn, TraceLevel, Comment 
		   << pHyp 
		   << " " << pHyp->m_Coverage 
		   << " " << pHyp->m_CoverageStart 
		   << " " << pHyp->m_CoveragePattern 
		   << " " << pHyp->m_LMState
		   << " " << pHyp->m_TargetLength
		   << " " << pHyp->m_TotalScore );
}




// ---------------------------------------------------------------
// CDecoder
// ---------------------------------------------------------------
void
CDecoderNB::SetVocabulary( CVocabulary *pVocabulary )
{
  m_pVocabulary = pVocabulary;
}



void
CDecoderNB::SetLexicon( const CLexicon *pLexicon )
{
  m_pLexicon = pLexicon;
}



void
CDecoderNB::SetSourceSentence( const CSentence *pSentence )
{
  m_pSourceSentence = pSentence;
}



//Sanjika 04/25
void CDecoderNB::SetSentenceLengthModel( CSentenceLengthModel *pSLM )
{
  m_pSLenModel = pSLM;
}



void
CDecoderNB::SetLengthOfReorderingWindow( int LengthOfReorderingWindow )
{
  if ( LengthOfReorderingWindow > c_MaxReorderWindowLength ) {
	cerr << "Requested length of reordering window larger than possible" << endl;
	cerr << "Restricted to c_MaxReorderWindowLength" << endl;
	LengthOfReorderingWindow = c_MaxReorderWindowLength;
	m_LengthOfReorderingWindow = LengthOfReorderingWindow;
  } else {
	m_LengthOfReorderingWindow = LengthOfReorderingWindow;
  }
  cout<<"RLength: "<<m_LengthOfReorderingWindow<<endl;
}



void
CDecoderNB::SetReorderingsPerWord( float ReorderingsPerWord )
{
  m_ReorderingsPerWord = ReorderingsPerWord;
  cout<<"RPerWord: "<<m_ReorderingsPerWord<<endl;
}



/**
   Set number of best recombined hypothesis within similarity class
   that will be expanded.
 */
void
CDecoderNB::SetBeamSize( unsigned long int BeamSize )
{
  if ( BeamSize < 1.0 ) {
	cerr << "Beam size " << BeamSize << " has to be >= 1.0" << endl;
	exit ( -2 );
  }

  m_BeamSize = BeamSize;
}



/**
   Set maximal number of hypothesis to be expanded per coverage.

   Hypothesis are organized per coverage. When expanding hypothesis for each
   coverage, you can limit the total number of hypothesis expanded per coverage
   here.
 */
void
CDecoderNB::SetMaxHypsPerCoverage( unsigned long int MaxHyps )
{
  if ( MaxHyps < 1 ) {
	cerr << "Max Hyps Per Coverage " << MaxHyps << " has to be >= 1.0" << endl;
	exit ( -2 );
  }

  m_MaxHypsPerCoverage = MaxHyps;
}



void
CDecoderNB::SetMaxHyps( int MaxHyps )
{
  m_MaxHyps = MaxHyps;
}



void
CDecoderNB::SetNumOfNBest( int N )
{
  m_NumOfNBest = N;
  cout << "NBest: " << m_NumOfNBest << endl;
}



/*
 void
CDecoderNB::SetSLStretchFactor( float SLStretch )
{
  m_SLStretch = SLStretch;
}
*/



void
CDecoderNB::SetCompareHypFeatures( bool UseCoverage, bool UseCoveragePattern, 
                                    bool UseLM, bool UseTargetLength )
{
  m_pDecoderInfo->m_UseCoverage = UseCoverage;
  m_pDecoderInfo->m_UseCoveragePattern = UseCoveragePattern;
  m_pDecoderInfo->m_UseLMState = UseLM;
  m_pDecoderInfo->m_UseTargetLength = UseTargetLength;

  cout << "SetCompareHypFeatures " << UseCoverage  << " " << UseCoveragePattern
	   << " " << UseLM << " " << UseTargetLength << endl;
}



void
CDecoderNB::SetPruneHypFeatures( bool UseCoverage, bool UseCoveragePattern, 
                                  bool UseLM, bool UseTargetLength )
{
  m_pDecoderInfo->m_UseCoverageForPruning = UseCoverage;
  m_pDecoderInfo->m_UseCoveragePatternForPruning = UseCoveragePattern;
  m_pDecoderInfo->m_UseLMStateForPruning = UseLM;
  m_pDecoderInfo->m_UseTargetLengthForPruning = UseTargetLength;

  cout << "SetCompareHypFeatures " << UseCoverage  << " " << UseCoveragePattern
	   << " " << UseLM << " " << UseTargetLength << endl;
}



void
CDecoderNB::SetScalingFactors( double TMScale, double LMScale,
                                double ALScale, double SLScale )
{

  //ARV: 04/20 Generalizing the decoder to handle a set of scores
  m_ScalingFactors.AssureBoundaries(0, 3);
  m_ScalingFactors.SetAll(0.0);
  m_ScalingFactors[c_LM] = LMScale;
  m_ScalingFactors[c_AL] = ALScale;
  m_ScalingFactors[c_SL] = SLScale;
  m_ScalingFactors[c_TM] = TMScale;

}



void CDecoderNB::SetScalingFactors( CNumericalArray< float > ScalingFactors)
{
  m_ScalingFactors = ScalingFactors;
  cout<<"Decoder Scaling Factors: ";
  for (int i=ScalingFactors.MinIndex(); i<=ScalingFactors.MaxIndex(); i++){
    cout<<i<<" = "<<m_ScalingFactors[i]<<" ";
  }
  cout<<endl;
  m_NumberOfModels = ScalingFactors.Size();
}



void CDecoderNB::SetScalingFactors( vector< float > ScalingFactors)
{
  m_ScalingFactors.AssureBoundaries(0, ScalingFactors.size()-1);
  cout<<"Decoder Scaling Factors: ";
  for (int i=0; i<ScalingFactors.size(); i++){
    m_ScalingFactors[i] = ScalingFactors[i];
    cout<<i<<" = "<<m_ScalingFactors[i]<<" ";
  }
  cout<<endl;
  m_NumberOfModels = ScalingFactors.size();
}




void CDecoderNB::GetLengthOfReorderingWindow( int& LengthOfReorderingWindow )
{
  LengthOfReorderingWindow = m_LengthOfReorderingWindow;
}



void CDecoderNB::GetReorderingsPerWord( float& ReorderingsPerWord )
{
  ReorderingsPerWord = m_LengthOfReorderingWindow;
}



void CDecoderNB::GetNumOfNBest( int& N )
{
  N = m_NumOfNBest;
}



void
CDecoderNB::GetCompareHypFeatures( bool& UseCoverage, bool& UseCoveragePattern,
                                         bool& UseLM, bool& UseTargetLength )
{
  UseCoverage = m_pDecoderInfo->m_UseCoverage;
  UseCoveragePattern = m_pDecoderInfo->m_UseCoveragePattern;
  UseLM = m_pDecoderInfo->m_UseLMState;
  UseTargetLength = m_pDecoderInfo->m_UseTargetLength;
}



void
CDecoderNB::GetPruneHypFeatures( bool& UseCoverage, bool& UseCoveragePattern,
                                  bool& UseLM, bool& UseTargetLength)
{
  UseCoverage = m_pDecoderInfo->m_UseCoverageForPruning;
  UseCoveragePattern = m_pDecoderInfo->m_UseCoveragePatternForPruning;
  UseLM = m_pDecoderInfo->m_UseLMStateForPruning;
  UseTargetLength = m_pDecoderInfo->m_UseTargetLengthForPruning;
}



void
CDecoderNB::GetBeamSize( unsigned long int& BeamSize)
{
  BeamSize = m_BeamSize;
}



void CDecoderNB::GetMaxHypsPerCoverage( unsigned long int& MaxHypsPerCov)
{
  MaxHypsPerCov = m_MaxHypsPerCoverage;
}




void CDecoderNB::GetScalingFactors( float& TMScale, float& LMScale,
				    float& ALScale, float& SLScale){
  TMScale = m_ScalingFactors[c_TM];
  LMScale = m_ScalingFactors[c_LM];
  ALScale = m_ScalingFactors[c_AL];
  SLScale = m_ScalingFactors[c_SL];
}



void CDecoderNB::GetScalingFactors( CNumericalArray< float >& ScalingFactors)
{
  ScalingFactors = m_ScalingFactors;
}



void CDecoderNB::SetRescore( bool Rescore )
{
  m_Rescore = Rescore;
}



void CDecoderNB::SetRescorer( CRescorer *pRescorer )
{
  m_pRescorer = pRescorer;
}




CDecoderNB_Hyp * CDecoderNB::NewHypothesis()
{
  //ARV 4/20 updated to initialize the hyp with the appropriate number of models
  return new CDecoderNB_Hyp(m_NumberOfModels);
}

RecombinedMap* CDecoderNB::NewRecombinedMap()
{
  RecombinedMap* pSavePtr = new RecombinedMap;
  m_RecombinedMapList.push_back(pSavePtr);
  return pSavePtr;
}



/*
// sv 051102 decoder result needs to be generic
CDecoderNBResult::CDecoderNBResult(){

}



CDecoderNBResult::CDecoderNBResult( CString translation, CDecoderNB_BP * m_pBP, CArray< CLatticeEdgeMS * >  edges )
{
	m_translation = translation;

	m_TotalScore = m_pBP->m_TotalScore;
	for (int i=m_pBP->m_scaledModelScores.MinIndex(); i<=m_pBP->m_scaledModelScores.MaxIndex(); i++){
	  m_features.push_back(m_pBP->m_scaledModelScores[i]);
	}

	m_edges = edges;
}



CDecoderNBResult::~CDecoderNBResult()
{
	m_features.clear();
}



void
CDecoderNBResult::scaleFeatures(vector<double>& ScalingFactors, bool divide)
{

	for (int i=0; i<ScalingFactors.size(); i++){
		if (divide && ScalingFactors[i]!=0.0){
		    m_features[i] = m_features[i]/ScalingFactors[i];
		}else{
			m_features[i] = m_features[i]*ScalingFactors[i];
		}
	}
}



CDecoderNBResult&
CDecoderNBResult::operator= ( const CDecoderNBResult& arg )
{
  if (this != &arg){
	this->m_TotalScore = arg.m_TotalScore;
	this->m_translation = arg.m_translation;
	this->m_features = arg.m_features;
	this->m_edges = arg.m_edges;
  }
  return *this;

}




bool
CDecoderNBResult::operator< ( const CDecoderNBResult& arg )
{
	return this->m_TotalScore < arg.m_TotalScore;

}




void CDecoderNBResult::Display(string tag, int id, vector<double>& ScalingFactors, CLatticeLM* lattice )
{
  cout<<tag<<" "<<id<<" "<<m_translation<<endl;
  cout << "TotalScore: " << m_TotalScore << " Scaled(Raw): ";
  for (int i=0; i<m_features.size(); i++){
    cout<<m_features[i]<<"("<<m_features[i]/ScalingFactors[i]<<") ";
  }
  cout<<endl;
  cout << "display path: " << endl;
  if (lattice!=NULL){
    lattice->DisplayPath(m_edges);
  }
}
*/



void
CDecoderNB::GetLMLogProbsAllHypsWithCoverage( int Coverage )
{
  s_TraceOn = False;

  TraceIf( s_TraceOn, 0, "\n\n  CDecoderNB::ExpandAllHypsWithCoverage " << Coverage );

  // variable we need
  CDecoderNB_Hyp *pHyp, *pBestHyp;
  CLatticeNode *pNode;
  CLatticeEdgeMS *pEdge;



  // how long is the source sentence?  we need to cover all words
  int J = m_pLattice->EndNode()->Number();
  // should we use average path length for speech recognizer lattices?

  // hypotheses are organized according to their coverage
  // hypotheses of coverage c are expanded till coverage J-1
  // is reached.  Expanding those hypotheses will result in 
  // full coverage

  // first we fetch the set of hypotheses which all have coverage c
  // and we need an iterater to access the hyps in this set
  HypContainer< CDecoderNB_Hyp *, HypHash, HypEqual > &Hyps = m_AllHyps[ Coverage ];
  HypContainer< CDecoderNB_Hyp *, HypHash, HypEqual >::iterator IterHyps;


  // for pruning we need the best hypothesis with the current coverage
  HypContainer< CDecoderNB_Hyp *, HypHashForPruning, HypEqualForPruning > &BestHyps = m_BestHyps[ Coverage ];
  HypContainer< CDecoderNB_Hyp *, HypHashForPruning, HypEqualForPruning >::iterator IterBestHyps;
  
  //HistogramPruning( Hyps, BestHyps );

  //  TraceIf( s_TraceOn, 0, "After Histogram Pruning:  Number of hyps with coverage " 
  //		   << Coverage << " is " << Hyps.size() );
  
  // sort hyps by score for beamsize pruning
  typedef multimap<float, CDecoderNB_Hyp*> CSortedHypContainer;
  // associate expansion candidates with their best-hyp-for-pruning
  typedef map<CDecoderNB_Hyp*,CSortedHypContainer> CBestSimHyp2AllSimHypsMap;

  CBestSimHyp2AllSimHypsMap HypsByBestHyp;
		   
  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {
    pHyp = *IterHyps;
    //cout << "Hyp is " << pHyp << endl;
    TraceIf( s_TraceOn, 0, "\n  in ExpandAllHypsWithCoverage - expand next hyp " 
             << pHyp << " "
             << pHyp->m_Coverage << " " << pHyp->m_CoverageStart 
             << " " << pHyp->m_CoveragePattern
             << " " << pHyp->m_LMState
             << " " << pHyp->m_TargetLength 
             << " " << pHyp->m_TotalScore );
    
    // only expand if hyp had survived histogram pruning
    if ( pHyp->m_Expand == False ) {
      
      //cout << "Prune hyp (histogram) " << pHyp << endl;
      //	  m_HypsPruned++;
      
    } else {
    
      // pruning: only expand hyp if it is better than BeamFactor times 
      // score of best hypothesis
      
      IterBestHyps = BestHyps.find( pHyp );
      if ( IterBestHyps == BestHyps.end() ) {
        //cout <<"Coverage: "<<Coverage<< "no best hyp found" << endl;
        pBestHyp->m_TotalScore = 10000;
      } else {
        pBestHyp = *IterBestHyps;
        //cout <<"Coverage: "<<Coverage<< "Best Hyp is " << pBestHyp << endl;
        TraceIf( s_TraceOn, 0, "\n  best hyp to prune against is " 
                 << pBestHyp << " "
                 << pBestHyp->m_Coverage << " " << pBestHyp->m_CoverageStart 
                 << " " << pBestHyp->m_CoveragePattern
                 << " " << pBestHyp->m_LMState
                 << " " << pBestHyp->m_TargetLength 
                 << " " << pBestHyp->m_TotalScore );
      }
      
    } // endif: histogram pruned
  } // andfor: all recombined hyps


  
  // foreach set of recombined hyps associated to their best hyp, expand beamsize best ones
  int rank;
  CSortedHypContainer::iterator IterSortedHyp;
  CSortedHypContainer HypsInBeam;
  for (CBestSimHyp2AllSimHypsMap::iterator IterHypsByBestHyp = HypsByBestHyp.begin();
       IterHypsByBestHyp != HypsByBestHyp.end();
       ++IterHypsByBestHyp) {
    // sorted list of expansion candidates associated with current best hyp for pruning
    CDecoderNB_Hyp* pBestHyp = (*IterHypsByBestHyp).first;

    if (s_TraceOn){
      cout<<"Coverage: "<<Coverage<<endl;
      cout<<"pBestHyp:  ";
      //display the best hyp
      CArray< CLatticeEdgeMS * > transEdges = TraceBackWithPath(pBestHyp).second;
      for (int i=transEdges.MinIndex(); i<=transEdges.MaxIndex(); i++){
        transEdges[i]->Display();
        cout<<" ";
      }
      cout<<endl<<endl;
    }

    CSortedHypContainer &SimilarHyps = (*IterHypsByBestHyp).second;

    if (s_TraceOn){
      //lets look at all hyps in beam first

      for (IterSortedHyp  = SimilarHyps.begin(); IterSortedHyp != SimilarHyps.end();  ++IterSortedHyp){
	CDecoderNB_Hyp* pHyp = (*IterSortedHyp).second;
	//display the hyp
	CArray< CLatticeEdgeMS * > transEdges = TraceBackWithPath(pHyp).second;
	for (int i=transEdges.MinIndex(); i<=transEdges.MaxIndex(); i++){
	  transEdges[i]->Display();
	  cout<<" ";
	}
	cout<<endl;
	cout<<"FScore: "<<(*IterSortedHyp).first<<" base: "<<pHyp->m_TotalScore<<endl;
	
      }
    }


    // only take the beamsize best
    for(IterSortedHyp = SimilarHyps.begin(), rank = 0;
	IterSortedHyp != SimilarHyps.end() && ++rank <= m_BeamSize;
	++IterSortedHyp) {

      HypsInBeam.insert(*IterSortedHyp);
    } // endfor

    m_HypsOutOfBeam += ( SimilarHyps.size() > m_BeamSize ?
			 SimilarHyps.size() - m_BeamSize :
			 0 );
  } // endfor


  // out of the hyps just collected, only take the m_MaxHypsPerCoverage best
  for (IterSortedHyp  = HypsInBeam.begin(), rank = 0;
       IterSortedHyp != HypsInBeam.end() && ++rank <= m_MaxHypsPerCoverage;
       ++IterSortedHyp)
    ExpandHyp((*IterSortedHyp).second);

  m_HypsOutOfBeamPerCov += ( HypsInBeam.size() > m_MaxHypsPerCoverage ?
			     HypsInBeam.size() - m_MaxHypsPerCoverage :
			     0 );

  s_TraceOn = false;
} // CDecoderNB::ExpandAllHypsWithCoverage(...)






