#include <DecoderBLM.hh>



CArray< std::pair< int, int > > CDecoderBLM::LeftToTranslate( CDecoderBLM_Hyp* pHyp )
{
  

  //======================================================
  // Determine which words are untranslated in this hyp
  //======================================================
  bitset<COVERAGE_BIT_VECTOR_SIZE> coveragePatternCopy = pHyp->m_CoveragePattern; 
  CNumericalArray<int> UnTranslated;
  CNumericalArray<int> ForwardTranslated;
  int count =0;
  //cout<<"m_CoverageStart: "<<pHyp->m_CoverageStart<<" coveragePatternCopy: "<<coveragePatternCopy<<endl;
  while (coveragePatternCopy.any()){
    //cout<<"cpCopy: "<<coveragePatternCopy<<endl;
    if (!coveragePatternCopy.test(0)){
      //cout<<"count: "<<count<<" Was 0"<<endl;
      UnTranslated.Append(count);
    }else{
      ForwardTranslated.Append(count);
    }
    coveragePatternCopy>>= 1;
    count++;
  }
  
  //the count tracks how far reordering effects are in play, after all the translated words
  //are accounted for, it reprsents the location of sequentially untranslated words
  
  //now recover the index position in the whole sentence
  //CNumericalArray<int> AllUntranslatedPositions;
  
  //represent contigous sequences of un-translated words with i=number of words, j=starting index of this sequence
  //case 123 5
  //lastJ=1, i=1
  //  i=2
 //  i=3
 //  ?4 - No, lastJ=5, i=1
  

  //======================================================
  // Find contiguous sequences of untranslated words
  //======================================================
  CArray<std::pair<int, int> > segments;

  int icon=1;
  int lastJ =-1;
  bool firstSequenceStarted = false;
  for (int w=UnTranslated.MinIndex(); w<=UnTranslated.MaxIndex(); w++){
    //map from the untranslated count to a count with respect to the sentence
    //this wordIndex is just a relative reference to the sentence
    int wordIndex =  m_pSourceSentence->MinIndex()+pHyp->m_CoverageStart+ UnTranslated[w];
    
    if (wordIndex==lastJ+icon && firstSequenceStarted==true){
      //sequence continues
      icon++;
    }else{
      //if we have already traced one sequence before, then lets save its data
      if (firstSequenceStarted){
	//cout<<"lastJ: "<<lastJ+1<<" icon: "<<icon<<endl;
	segments.Append(std::pair<int,int>( icon, lastJ+1));
      }
      lastJ = wordIndex;
      icon=1;
      firstSequenceStarted = true;
    }

    //cout<<"Jumped: "<<m_pVocabulary->GetWord((*m_pSourceSentence)[wordIndex])<<endl;
    //AllUntranslatedPositions.Append(wordIndex);
  }

  if (firstSequenceStarted){
    //cout<<"lastJ: "<<lastJ+1<<" icon: "<<icon<<endl;
    segments.Append(std::pair<int,int>(icon, lastJ+1));
  }
  lastJ = pHyp->m_CoverageStart+count;
  icon = (m_pSourceSentence->MaxIndex()+1)-(pHyp->m_CoverageStart+count);
  //cout<<"final lastJ: "<<lastJ+1<<" icon: "<<icon<<endl;
  segments.Append(std::pair<int,int>(icon, lastJ+1));
  return segments;
      
  //the coverage start indicates the number of contiguously translated words, to find the leftover 
  //untranslated words, we have to go past the future translated words
  /*
  for (int i=pHyp->m_CoverageStart+count; i<=m_pSourceSentence->MaxIndex(); i++){
    AllUntranslatedPositions.Append(i);
  }
  
  //the leftover words

  
  cout<<"UNT: ";
  for (int i=AllUntranslatedPositions.MinIndex(); i<=AllUntranslatedPositions.MaxIndex(); i++){
    cout<<m_pVocabulary->GetWord((*m_pSourceSentence)[AllUntranslatedPositions[i]])<<" ";
  }
  cout<<endl;
  */
  
}



void
CDecoderBLM::DisplayDecoderSettings()
{

  if (s_TraceOn){
    cout << "MaxNumReorderings " << m_pDecoderInfo->m_MaxNumReorderings / 2 << endl;
    cout << "Length of reordering window is " << m_LengthOfReorderingWindow << endl;
    /*
    cout << "TMScale = "  << m_ScalingFactors[ c_TM1 ]
         << " LMScale = " << m_ScalingFactors[ c_LM1 ]
         << " ALScale = " << m_ScalingFactors[ c_AL1 ]
         << " SLScale = " << m_ScalingFactors[ c_SL1 ]
         << " SLScale = " << m_ScalingFactors[ c_PL1 ] << endl;
    //<< " SLStretch = " << m_SLStretch 
    */
    cout << "Comparing Hyps with"
         << " Coverage " <<  m_pDecoderInfo->m_UseCoverage 
         << " Coverage Pattern " <<  m_pDecoderInfo->m_UseCoveragePattern
         << " LM state " << m_pDecoderInfo->m_UseLMState
         << " Target stentence length " <<  m_pDecoderInfo->m_UseTargetLength << endl;
    cout << "Pruning Hyps with"
         << " Coverage " <<  m_pDecoderInfo->m_UseCoverageForPruning 
         << " Coverage Pattern " <<  m_pDecoderInfo->m_UseCoveragePatternForPruning
         << " LM state " << m_pDecoderInfo->m_UseLMStateForPruning
         << " Target stentence length " <<  m_pDecoderInfo->m_UseTargetLengthForPruning << endl;
    cout << "Beam Size = " << m_BeamSize << endl;
    cout << "Max # of expanded hyps per coverage = " << m_MaxHypsPerCoverage << endl;
  }// end trace output


}




void
CDecoderBLM::DisplayDecoderStatistics()
{

  // display some statistics about the decoding
  s_TraceOn = true;
  if (s_TraceOn){
    int J = m_pLattice->EndNode()->Number();
    cout << "Decoding Statistics               sentence / per word" << endl;
    cout << "Hyps expanded                   = " << m_HypsExpanded    << " / " << m_HypsExpanded / J << endl;
    cout << "Reordering exceeded             = " << m_HypsNotExpanded << " / " << m_HypsNotExpanded <<endl;
    cout << "Beam size exceeded              = " << m_HypsOutOfBeam   << " / " << m_HypsOutOfBeam / J<<endl;
    cout << "Beam size per coverage exceeded = " << m_HypsOutOfBeamPerCov << " / " << m_HypsOutOfBeamPerCov / J<<endl;
    cout << "New Recombined Hyps             = " << m_NumNewHyps      << " / " << m_NumNewHyps / J<<endl;
    cout << "Better Recombined Hyps           = " << m_NumBetterHyps   << " / " << m_NumBetterHyps / J<<endl;
    cout << "Worse Recombined Hyps           = " << m_NumWorseHyps    << " / " << m_NumWorseHyps / J<<endl;
    cout << "CoverageCollisions              = " << m_NumCoverageCollisions << " / " << m_NumCoverageCollisions / J<<endl;
  }
  s_TraceOn = false;

}





// ??? sv 060608: seems not be used
vector< RecombinedMap* >
CDecoderBLM::HarvestFromHyp( CDecoderBLM_Hyp* pOneHyp )
{
  //iterate over all the recombined hypothesis in this list
  multimap< float, CDecoderBLM_Hyp * >::iterator Iter;
  vector<RecombinedMap*> retList;

  if ( pOneHyp->m_pRecombinedHyps != NULL ) {

    for ( Iter = pOneHyp->m_pRecombinedHyps->begin();  Iter != pOneHyp->m_pRecombinedHyps->end(); ++Iter ) {
      if (Iter->second!=NULL){
        //cout<<"InnerNOTNULL:"<<Iter->second<<endl;
        vector< RecombinedMap* > subList = HarvestFromHyp(Iter->second);
        // add all from the sublist
        for(vector<RecombinedMap*>::iterator recIter=subList.begin(); recIter!=subList.end(); recIter++){
          retList.push_back(*recIter);
        }
      }// pointer was valid
      
    }// all children have been dealt with
    //now add this pointer list
    retList.push_back(pOneHyp->m_pRecombinedHyps);
  }else{
    //ARV
    // it is already removed
  }
  return retList;
  
}




void
CDecoderBLM::AddOverlappingPhrases( CLattice *pLattice )
{
  TraceIf( s_TraceOn, 0, "CDecoderBLM::AddOverlappingPhrases" );
  assert( 0 );

  // insert overlapping phrases
  //cout << "overlapping phrases" << endl;
  /*
  if ( m_AddOverlappingPhrases ) {
    COverlappingPhraseGenerator *pOPGenerator = new COverlappingPhraseGenerator;
    pOPGenerator->AddOverlappingPhrases( pLattice );
    delete pOPGenerator;
  }
  */

}



void
CDecoderBLM::HistogramPruning( HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual > &Hyps,
                               HypContainer< CDecoderBLM_Hyp *, HypHashForPruning, HypEqualForPruning > &BestHyps )
{

  // the histogram pruning reads reworking and rethinking
  // sometimes only hyps suvive which lead to collision
  // so no full translation is generated
  
  return;
  /*

  bool s_TraceOn = false;

  //Trace( 0, "Histogram Pruning: " << Hyps.size() << " " << BestHyps.size() );

  // run over all hyps
  // put hyp into bucket according to BucketNumber = (pHyp->Score - pBestHyp->Score) * 100 / BeamFactor
  // i.e. do not select according to absolute score but according to difference to best hyp
  // with the same features

  CDecoderBLM_Hyp *pHyp, *pBestHyp;

  HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual >::iterator IterHyps;
  HypContainer< CDecoderBLM_Hyp *, HypHashForPruning, HypEqualForPruning >::iterator IterBestHyps;


  // first, find the best score from the best hyps
  float BestScore = 1.0e10;
  for ( IterBestHyps = BestHyps.begin(); IterBestHyps != BestHyps.end(); ++IterBestHyps ) {

	pBestHyp = *IterBestHyps;
	if ( pBestHyp->m_TotalScore < BestScore ) {
	  BestScore = pBestHyp->m_TotalScore;
	}
  }

  //cout << "Best Score = " << BestScore << endl;

//   for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {

// 	pHyp = *IterHyps;
// 	//cout << "Hyp is " << pHyp << endl;
// 	TraceIf( s_TraceOn, 0, "\n  hyp" 
// 			 << pHyp << " "
// 			 << pHyp->m_Coverage << " " << pHyp->m_CoverageStart 
// 			 << " " << pHyp->m_CoveragePattern
// 			 << " " << pHyp->m_LMState
// 			 << " " << pHyp->m_TargetLength 
// 			 << " " << pHyp->m_TotalScore );


// 	// select corresponding best hyp
// 	IterBestHyps = BestHyps.find( pHyp );
// 	pBestHyp = *IterBestHyps;
// 	//cout << "Best Hyp is " << pBestHyp << endl;
// 	TraceIf( s_TraceOn, 0, "\n  best hyp " 
// 			 << pBestHyp << " "
// 			 << pBestHyp->m_Coverage << " " << pBestHyp->m_CoverageStart 
// 			 << " " << pBestHyp->m_CoveragePattern
// 			 << " " << pBestHyp->m_LMState
// 			 << " " << pBestHyp->m_TargetLength 
// 			 << " " << pBestHyp->m_TotalScore );
	
// 	// now the calculation
// 	int BucketNumber = int( ( pHyp->m_TotalScore - pBestHyp->m_TotalScore ) * 100 / m_BeamFactor );
// 	TraceIf( s_TraceOn, 0, "\n  BucketNumber = " << BucketNumber );
// 	if ( BucketNumber < 100 ) {
// 	  m_PruningBuckets.GetRow( BucketNumber ).Append( pHyp );
// 	}
//   }



  //cout << "In Histogram Pruning we have " << Hyps.size() << " hypotheses" << endl;
  int CollectedHyps = 0;

  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {

	pHyp = *IterHyps;
	pHyp->m_Expand = False;

	//cout << "Hyp is " << pHyp << endl;
	TraceIf( s_TraceOn, 2, "\n  hyp" 
			 << pHyp << " "
			 << pHyp->m_Coverage << " " << pHyp->m_CoverageStart 
			 << " " << pHyp->m_CoveragePattern
			 << " " << pHyp->m_LMState
			 << " " << pHyp->m_TargetLength 
			 << " " << pHyp->m_TotalScore );


	// now the calculation
	int BucketNumber = int( ( pHyp->m_TotalScore - BestScore ) * 100 / m_BeamFactor );
	TraceIf( s_TraceOn, 2, "\n  BucketNumber = " << BucketNumber );
	if ( BucketNumber < 100 ) {
	  CollectedHyps++;
	  m_PruningBuckets.GetRow( BucketNumber ).Append( pHyp );
	}
  }

  //cout << "In the buckets we have " << CollectedHyps << " hypotheses" << endl;

  // now collect enough buckets to get the required number of hyps
  //cout << "collect " << m_MaxHyps << " hyps " << endl;
  int Sum = 0;
  int i = 0;
  while ( (Sum < m_MaxHyps) && (i < 100) ) {
	CArray< CDecoderBLM_Hyp * > Bucket = m_PruningBuckets.GetRow( i );
	Sum += Bucket.Size();
	//cout << i << " " << Bucket.Size() << " " << Sum << endl;
	
	for ( int j = Bucket.MinIndex(); j <= Bucket.MaxIndex(); j++ ) {
	  Bucket[ j ]->m_Expand = True;
	}

	i++;
  }

  //cout << "Number of Hyp to expand after Histogram Pruning: " << Sum << endl;

  m_PruningBuckets.Clear();

  s_TraceOn = false; */
}




//this Rescore function is hard coded to do Model I rescoring, therefore it is being 
//commented out. Rescoring should be done on the n-best list data structure 

void CDecoderBLM::Rescore( )
{

  s_TraceOn = False;
  TraceIf( s_TraceOn, 0, "CDecoderBLM::Rescore" );


  // go over the n-best list, i.e. the backpointers
  // for each create the translation as array of word indices
  // calculate the TM score using the lexicon

  CArray< CWordIndex > Source;
  CArray< CWordIndex > Target;
	
  for ( int j = m_pSourceSentence->MinIndex(); j <= m_pSourceSentence->MaxIndex(); j++ ) {
    CString Word = m_pSourceSentence->Word( j, m_pVocabulary );
    Source.Append( m_pLexicon->SourceVocabulary()->GetIndex( Word ) );
  }

  int n = 0;
  multimap< float, CDecoderBLM_BP * > &CurrentBPs = m_NBestHyps[ 0 ];
  // iterator to iterate over list of backpointer for one coverage
  multimap< float, CDecoderBLM_BP * >::iterator IterOverBPs;

  CDecoderBLM_BP *pBP;
  //CDecoderBLM_BP *pRescoredBP;

  /*
  // clear backpointer list for rescored hyps
  for ( IterOverBPs = m_NBestHypsRescored.begin(); IterOverBPs != m_NBestHypsRescored.end(); ++IterOverBPs ) {
	pBP = IterOverBPs->second;
	delete pBP;
	m_NBestHypsRescored.erase( IterOverBPs );
  }
  */


  for ( IterOverBPs = CurrentBPs.begin(); 
        (IterOverBPs != CurrentBPs.end()) && (n < m_NumOfNBest * c_NumOfNBestMultiplier);
        ++IterOverBPs ) {

    pBP = IterOverBPs->second;

    TraceIf( 0, 0, "\nBackpointer " << n << " " << IterOverBPs->first
             << " TotalScore=" << pBP->m_TotalScore );

    //cout << endl << "Rescore " << n << " ";
    //Display( pBP->m_pBP, n );

    // follow the pBP to access the edges and thereby the actual translation
    //Translation = "";
    //AllEdges.RemoveAll();
    Target.RemoveAll();

    // first BP points to initial hyp which has no edge, i.e. start from next
    pBP = pBP->m_pBP;

    while ( pBP->m_pHyp->m_pEdge ) {
	
      //pBP->m_pHyp->m_pEdge->Display();
	  
      // access to the partial translations via the WordInfo
      CWordInfo *pWordInfo = pBP->m_pHyp->m_pEdge->WordInfo();
	  
      if ( pWordInfo->HasTargetString() ) {
        CSentence &Words = pWordInfo->TargetWords();
        for ( int n = Words.MinIndex(); n <= Words.MaxIndex(); n++ ) {
          //Translation += pWordInfo->TargetString();
          //cout << Words.Word( n ) << " T " << m_pLexicon->TargetVocabulary()->GetIndex( Words.Word( n ) ) << endl;
          Target.Append( m_pLexicon->TargetVocabulary()->GetIndex( Words.Word( n ) ) );
        }
      } else if ( pWordInfo->HasSourceString() ) {
        CSentence &Words = pWordInfo->SourceWords();
        for ( int n = Words.MinIndex(); n <= Words.MaxIndex(); n++ ) {
          //Translation += pWordInfo->SourceString();
          //cout << Words.Word( n ) << " S " << m_pLexicon->TargetVocabulary()->GetIndex( Words.Word( n ) ) << endl;
          Target.Append( m_pLexicon->SourceVocabulary()->GetIndex( Words.Word( n ) ) );
        }
      } else {
        //Translation += pBP->m_pHyp->m_pEdge->Word();
        CString Word = pBP->m_pHyp->m_pEdge->Word();
        //cout << Word << " " << m_pLexicon->SourceVocabulary()->GetIndex( Word ) << endl;
        Target.Append( m_pLexicon->SourceVocabulary()->GetIndex( Word ) );
      }
	
      pBP = pBP->m_pBP;
    }
    //cout << endl;
    
    //for ( int i = Source.MinIndex(); i <= Source.MaxIndex(); i++ ) {
    //cout << " " << Source[ i ];
    //}
    //cout << endl;

    //for ( int i = Target.MinIndex(); i <= Target.MaxIndex(); i++ ) {
    //cout << " " << Target[ i ];
    //}
    //cout << endl;

    // now get the TM score
    float Prob = 1.0;
    float SumProb = 0.0;
    float LogProb = 0.0;

    if ( m_pLexicon ) {
      //cout << "Source Length is " << m_pSourceSentence->Length() 
      //   << "Target Length is " << Target.Size() << endl;
      
      for ( int j = Source.MinIndex(); j <= Source.MaxIndex(); j++ ) {
        SumProb = 0.0;
        for ( int i = Target.MinIndex(); i <= Target.MaxIndex(); i++ ) {
          //cout << m_pLexicon->SourceVocabulary()->GetWord( Source[ j ] ) << " " 
          //   << m_pLexicon->TargetVocabulary()->GetWord( Target[ i ] ) << " "
          //   << m_pLexicon->Pr( Source[ j ], Target[ i ] ) << endl;
          SumProb += m_pLexicon->Pr( Source[ j ], Target[ i ] );
        }
        LogProb += -log( SumProb );
        //cout << SumProb << endl;
      }
	  
    } else {
      LogProb = 10.0;
    }
    
    //if (s_TraceOn){
    cout << "New TMScore " << LogProb << endl;
	  //}
	  
    pBP->m_scaledModelScores[ pBP->m_scaledModelScores.MaxIndex() ] = LogProb;


    /*
      pBP = IterOverBPs->second;
      pRescoredBP = new CDecoderBLM_BP( pBP );
      
      pRescoredBP->m_TotalScore -= pRescoredBP->m_TMScore;
      
      
      pRescoredBP->m_TotalScore += LogProb;
      pRescoredBP->m_TMScore = LogProb;
      pRescoredBP->m_pBP = pBP->m_pBP;
      
      m_NBestHypsRescored.insert( pair< float, CDecoderBLM_BP * >( pRescoredBP->m_TotalScore , pRescoredBP ) );
    */
    n++;
  }


  /*
  // now we need to reconstruct the actual translations
  // loop through list of backpointers up to N
  TraceIf( s_TraceOn, 0, "  Final List of Backpointers " );

  n = 0;

  for ( IterOverBPs = m_NBestHypsRescored.begin(); 
  IterOverBPs != m_NBestHypsRescored.end();
  ++IterOverBPs ) {

	pRescoredBP = IterOverBPs->second;

	TraceIf( 1, 0, "\nBackpointer Rescored " << n << " " << IterOverBPs->first
  << " TotalScore=" << pRescoredBP->m_TotalScore );

	cout << "Total Score  " << pRescoredBP->m_TotalScore
  << "  TMScore "    << pRescoredBP->m_TMScore
  << "  LMScore "    << pRescoredBP->m_LMScore
  << "  ALScore "    << pRescoredBP->m_ALScore
  << "  SLScore "    << pRescoredBP->m_SLScore
  << endl;

	cout << "Rescored NBest " << n << " is: ";
	cout << TraceForward( pRescoredBP ) << endl;

	n++;
  }

  */

}



double
CDecoderBLM::Rescore( CDecoderBLM_Hyp *pHyp )
{
  int Start = m_pLattice->StartNode()->Number();
  int End   = m_pLattice->EndNode()->Number();
  int L = End - Start;

  bool s_TraceOn = True;
  
  /*
    if (s_TraceOn){
    cout << "Old TMScore " << pHyp->m_TMScore << endl;
    }
  */

  CArray< CLatticeEdge * > Edges;

  // trace back and collect the edges
  while ( pHyp->m_pPrevHyp ) {
    //cout << "Hyp " << pHyp << " PrevHyp " << pHyp->m_pPrevHyp 
    //<< " Edge " << pHyp->m_pEdge << endl;
    if ( pHyp->m_pEdge ) {
      Edges.Insert( 0, pHyp->m_pEdge );
    }
    pHyp = pHyp->m_pPrevHyp;
  }


  //cout << "Path for Rescoring: ";
  //if (s_TraceOn){
  //m_pLattice->DisplayPath( Edges );
	//}

  // now convert the path into array of word IDs
  CArray< CWordIndex > Target, Words;
  CWordInfo *pWordInfo;
  CLatticeEdgeMS *pEdge;

  for ( int i = Edges.MinIndex(); i <= Edges.MaxIndex(); i++ ) {
    pEdge = (CLatticeEdgeMS *)Edges[ i ];
    pWordInfo =  pEdge->WordInfo();

    // extract sequence of words from the WordInfo structure
    // no WordIndex available, only String (that's bad), so get index here
    if ( pWordInfo->HasTargetString() ) {

      CSentence &Words = pWordInfo->TargetWords();
      //cout << "Edge has " << Words.Size() << " target words " 
      //   << Words << endl;
      for ( int n = Words.MinIndex(); n <= Words.MaxIndex(); n++ ) {
        //cout << "word index = "  << Words.Word( n )  << endl;
        //cout << "word = " << m_pVocabulary->GetIndex( Words.Word( n ) ) << endl;
        Target.Append( m_pVocabulary->GetIndex( Words.Word( n ) ) );
      }

    } else if ( pWordInfo->HasSourceString() ) {

      CSentence &Words = pWordInfo->SourceWords();
      //cout << "Edge has " << Words.Size() << " target words " 
      //   << Words << endl;
      for ( int n = Words.MinIndex(); n <= Words.MaxIndex(); n++ ) {
        //cout << "word index = "  << Words.Word( n )  << endl;
        //cout << "word = " << m_pVocabulary->GetIndex( Words.Word( n ) ) << endl;
        Target.Append( m_pVocabulary->GetIndex( Words.Word( n ) ) );
      }

    } else {

      CWordIndex Word = m_pVocabulary->GetIndex( pEdge->WordIndex() );
      //cout << "edge has word " << Word << endl;
      Target.Append( Word );
    }

    for ( int n = Words.MinIndex(); n <= Words.MaxIndex(); n++ ) {
      //cout << "append word " << Words[ n ] << " to array of all target words " << endl;
      Target.Append( Words[ n ] );
    }
  }

  //  if (s_TraceOn){
  //cout << "Words are: " << Target << endl;

  // now get the TM score
  float Prob = 1.0;
  float SumProb = 0.0;
  float LogProb = 0.0;

  if ( m_pLexicon ) {
    cout << "Source Length is " << m_pSourceSentence->Length() 
         << "Target Length is " << Target.Size() << endl;

    for ( int j = m_pSourceSentence->MinIndex(); j <= m_pSourceSentence->MaxIndex(); j++ ) {
      SumProb = 0.0;
      for ( int i = Target.MinIndex(); i <= Target.MaxIndex(); i++ ) {
        //cout << (*m_pSourceSentence)[ j ] << " " <<  Target[ i ] << " "
        //	   << m_pLexicon->Pr( (*m_pSourceSentence)[ j ], Target[ i ] ) << endl;
        SumProb += m_pLexicon->Pr( (*m_pSourceSentence)[ j ], Target[ i ] );
      }
      LogProb += -log( SumProb );
      //cout << SumProb << endl;
    }

  }
  else {
    LogProb = 10.0;
  }

  if (s_TraceOn){
    cout << "New TMScore " << LogProb << endl;
  }

  return LogProb;

}



void
CDecoderBLM::GetTMScore( CLatticeEdgeMS *pEdge )
{
  assert( 0 );


  //  CString Word;
  double TMScore;

  // - translation score (i.e. accumulated transducer score
  // for original edges, use penalty
  /*
    Word = pEdge->Word();
    if ( Word[ 0 ] == '@' ) {
    TMScore = pEdge->WordInfo()->Score();
    TraceIf( s_TraceOn, 0, "      Translation Score for labeled edge: " << TMScore );
    } else {
    TMScore = 10;
    TraceIf( s_TraceOn, 0, "      Translation Score for source language edge: " << TMScore );
    }
    return TMScore;
  */

  //TMScore = m_ScalingFactors[c_TM] * pEdge->WordInfo()->Score();

  //its null when its over the source word
  if (pEdge->TransducerItem()!=NULL){
    TMScore = (pEdge->TransducerItem()->ScalingFactor())*pEdge->WordInfo()->Score();
  }else{
    TMScore = pEdge->WordInfo()->Score();
  }


  //ARV 4/20
  //m_pNumberOneHypothesis->m_TMScore += (m_ScalingFactors[c_TM] * TMScore);
  //m_pNumberOneHypothesis->m_TotalScore += (m_ScalingFactors[c_TM] * TMScore);


  //m_pNumberOneHypothesis->m_scaledModelScores[ c_TM1 ] += (m_ScalingFactors[ c_TM1 ] * TMScore);
  //m_pNumberOneHypothesis->m_TotalScore += (m_ScalingFactors[ c_TM1 ] * TMScore );

}


void
CDecoderBLM::WriteLattice()
{
  CArray< CLatticeEdge * > Edges = m_pLattice->Edges();
  CLatticeEdgeMS *pEdge;

  cout << "BEGIN_LATTICE" << endl;
  cout << "(" << endl;
  for ( int i = Edges.MinIndex(); i <= Edges.MaxIndex(); i++ ) {
    pEdge = (CLatticeEdgeMS *)Edges[ i ];
    //	pEdge->Display();
    //cout << endl;
    if ( (pEdge->Word())[0] == '@' ) {
      double Score = exp( -(pEdge->WordInfo()->Score() ) );
      cout << '(' << pEdge->Startframe() << " " << (pEdge->Endframe() - 1) << " ";
      cout << "|||" << pEdge->WordInfo()->TargetString() << "|||" << " ";
      cout << " " << Score << " ";
      cout << "|||"<< pEdge->WordInfo()->SourceString() << "|||";
      cout << ')';
      cout << endl;
    }
  }
  cout << ")" << endl;
  cout << "END_LATTICE" << endl;

}



void
CDecoderBLM::TraceHyp( bool TraceOn, int TraceLevel, CString Comment, CDecoderBLM_Hyp *pHyp )
{
  TraceIf( TraceOn, TraceLevel, Comment 
           << pHyp 
           << "  " << pHyp->m_Coverage 
           << " " << pHyp->m_CoverageStart 
           << " " << pHyp->m_CoveragePattern 
           << "  " << ( pHyp->m_LMStates[ pHyp->m_LMStates.MinIndex() ] & ( ( (CLMState)1 << 32 ) - 1 ) )
           << "_" << ( pHyp->m_LMStates[ pHyp->m_LMStates.MinIndex() ] >> 32 )
           << "  " << pHyp->m_TargetLength
           << "  " << pHyp->m_TotalScore );
}



void
CDecoderBLM::TraceHyp( bool s_TraceOn, const char * Message, CDecoderBLM_Hyp *pHyp )
{
  
  TraceIf( s_TraceOn, 0, Message 
           << pHyp 
           << "  " << pHyp->m_Coverage 
           << " " << pHyp->m_CoverageStart 
           << " " << pHyp->m_CoveragePattern 
           << "  " << ( pHyp->m_LMStates[ pHyp->m_LMStates.MinIndex() ] & ( ( (CLMState)1 << 32 ) - 1 ) )
           << "_" << ( pHyp->m_LMStates[ pHyp->m_LMStates.MinIndex() ] >> 32 )
           << "  " << pHyp->m_TargetLength
           << "  " << pHyp->m_TotalScore );
}


