#include <DecoderBLM.hh>


// ---------------------------------------------------------------
// CDecoder
// ---------------------------------------------------------------
/*
void
CDecoderBLM::SetVocabulary( CVocabulary *pVocabulary )
{
  m_pVocabulary = pVocabulary;
}
*/


void
CDecoderBLM::SetLexicon( const CLexicon *pLexicon )
{
  m_pLexicon = pLexicon;
}



void
CDecoderBLM::SetSourceSentence( const CSentence *pSentence )
{
  m_pSourceSentence = pSentence;
}



void
CDecoderBLM::SetSentenceLengthModel( CSentenceLengthModel *pSLM )
{
  m_pSLenModel = pSLM;
}



void
CDecoderBLM::SetLengthOfReorderingWindow( int LengthOfReorderingWindow )
{
  if ( LengthOfReorderingWindow > c_MaxReorderWindowLength ) {
    cerr << "Requested length of reordering window larger than possible" << endl;
    cerr << "Restricted to c_MaxReorderWindowLength" << endl;
    LengthOfReorderingWindow = c_MaxReorderWindowLength;
    m_LengthOfReorderingWindow = LengthOfReorderingWindow;
  } else {
    m_LengthOfReorderingWindow = LengthOfReorderingWindow;
  }
  cout<<"RLength: "<<m_LengthOfReorderingWindow<<endl;
}



void
CDecoderBLM::SetReorderingsPerWord( float ReorderingsPerWord )
{
  m_ReorderingsPerWord = ReorderingsPerWord;
  cout<<"RPerWord: "<<m_ReorderingsPerWord<<endl;
}



/**
   Set number of best recombined hypothesis within similarity class
   that will be expanded.
*/
void
CDecoderBLM::SetBeamSize( unsigned long int BeamSize )
{
  if ( BeamSize < 1.0 ) {
    cerr << "Beam size " << BeamSize << " has to be >= 1.0" << endl;
    exit ( -2 );
  }

  m_BeamSize = BeamSize;
}



/**
   Set maximal number of hypothesis to be expanded per coverage.

   Hypothesis are organized per coverage. When expanding hypothesis for each
   coverage, you can limit the total number of hypothesis expanded per coverage
   here.
*/
void
CDecoderBLM::SetMaxHypsPerCoverage( unsigned long int MaxHyps )
{
  if ( MaxHyps < 1 ) {
    cerr << "Max Hyps Per Coverage " << MaxHyps << " has to be >= 1.0" << endl;
    exit ( -2 );
  }

  m_MaxHypsPerCoverage = MaxHyps;
}



void
CDecoderBLM::SetMaxHyps( int MaxHyps )
{
  m_MaxHyps = MaxHyps;
}



void
CDecoderBLM::SetNumOfNBest( int N )
{
  m_NumOfNBest = N;
  cout << "NBest: " << m_NumOfNBest << endl;
}



/*
  void
  CDecoderBLM::SetSLStretchFactor( float SLStretch )
  {
  m_SLStretch = SLStretch;
  }
*/



void
CDecoderBLM::SetCompareHypFeatures( bool UseCoverage, bool UseCoveragePattern, 
                                    bool UseLM, bool UseTargetLength )
{
  m_pDecoderInfo->m_UseCoverage = UseCoverage;
  m_pDecoderInfo->m_UseCoveragePattern = UseCoveragePattern;
  m_pDecoderInfo->m_UseLMState = UseLM;
  m_pDecoderInfo->m_UseTargetLength = UseTargetLength;

  cout << "SetCompareHypFeatures " << UseCoverage  << " " << UseCoveragePattern
       << " " << UseLM << " " << UseTargetLength << endl;
}



void
CDecoderBLM::SetPruneHypFeatures( bool UseCoverage, bool UseCoveragePattern, 
                                  bool UseLM, bool UseTargetLength )
{
  m_pDecoderInfo->m_UseCoverageForPruning = UseCoverage;
  m_pDecoderInfo->m_UseCoveragePatternForPruning = UseCoveragePattern;
  m_pDecoderInfo->m_UseLMStateForPruning = UseLM;
  m_pDecoderInfo->m_UseTargetLengthForPruning = UseTargetLength;

  cout << "SetCompareHypFeatures " << UseCoverage  << " " << UseCoveragePattern
       << " " << UseLM << " " << UseTargetLength << endl;
}



void
CDecoderBLM::SetScalingFactors( double TMScale, double LMScale,
                                double ALScale, double SLScale )
{
  cout << "CDecoderBLM::SetScalingFactors should not be used anymore" << endl << flush;
  assert( 0 );

  //ARV: 04/20 Generalizing the decoder to handle a set of scores
  m_ScalingFactors.AssureBoundaries(0, 3);
  m_ScalingFactors.SetAll(0.0);
  /*
  m_ScalingFactors[ c_LM1 ] = LMScale;
  m_ScalingFactors[ c_AL1 ] = ALScale;
  m_ScalingFactors[ c_SL1 ] = SLScale;
  m_ScalingFactors[ c_TM1 ] = TMScale;
  */
}



void
CDecoderBLM::SetScalingFactors( CNumericalArray< float > ScalingFactors )
{
  cout << "Wrong SetScalingFactors used" << endl;
  assert( 0 );

  m_ScalingFactors = ScalingFactors;
  cout << "Set Decoder Scaling Factors: ";
  for ( int i = ScalingFactors.MinIndex(); i <= ScalingFactors.MaxIndex(); i++ ) {
    cout << i << " = " << m_ScalingFactors[ i ] << " ";
  }
  cout << endl;
  m_NumberOfModels = ScalingFactors.Size();
  
  // m_pTranslationMemory needs scaling factors for lattice pruning
  assert( m_pTranslationMemory );
  m_pTranslationMemory->SetScalingFactors( &m_ScalingFactors );
  
}



void
CDecoderBLM::SetScalingFactors( vector< float > ScalingFactors )
{
  m_ScalingFactors.AssureBoundaries( 0, ScalingFactors.size() - 1 );
  cout << "Set Decoder Scaling Factors: number = " << ScalingFactors.size() << endl;
  
  // Save wc and tm scaling factors for translation edge pruning
  CNumericalArray< float >* pLatticePruneScalingFactors = new CNumericalArray< float >();

  for ( int i = 0; i < ScalingFactors.size(); i++)
  {
    m_ScalingFactors[ i ] = ScalingFactors[ i ];
    
    // save just what's needed for lattice pruning
    if      ( i == m_StartWCScores )
      (*pLatticePruneScalingFactors)[ 0 ] = ScalingFactors[ i ];
    else if ( i == m_StartPCScores )
      (*pLatticePruneScalingFactors)[ 1 ] = ScalingFactors[ i ];
    else if ( i >= m_StartTMScores )
      pLatticePruneScalingFactors->Append( ScalingFactors[ i ] );
    
    cout << i << " = " << m_ScalingFactors[ i ] << " ";
  }
  cout << endl;
  m_NumberOfModels = ScalingFactors.size();
  cout << "Number of models = " << m_NumberOfModels << endl;
  
  // m_pTranslationMemory needs scaling factors for lattice pruning
  assert( m_pTranslationMemory );
  
  m_pTranslationMemory->SetScalingFactors( pLatticePruneScalingFactors ); 
  
}


void
CDecoderBLM::SetPruningMode( int PruningMode )
{
  m_PruningMode = PruningMode;
}


void
CDecoderBLM::GetLengthOfReorderingWindow( int& LengthOfReorderingWindow )
{
  LengthOfReorderingWindow = m_LengthOfReorderingWindow;
}



void
CDecoderBLM::GetReorderingsPerWord( float& ReorderingsPerWord )
{
  ReorderingsPerWord = m_LengthOfReorderingWindow;
}



void
CDecoderBLM::GetNumOfNBest( int& N )
{
  N = m_NumOfNBest;
}



void
CDecoderBLM::GetCompareHypFeatures( bool& UseCoverage, bool& UseCoveragePattern,
                                    bool& UseLM, bool& UseTargetLength )
{
  UseCoverage = m_pDecoderInfo->m_UseCoverage;
  UseCoveragePattern = m_pDecoderInfo->m_UseCoveragePattern;
  UseLM = m_pDecoderInfo->m_UseLMState;
  UseTargetLength = m_pDecoderInfo->m_UseTargetLength;
}



void
CDecoderBLM::GetPruneHypFeatures( bool& UseCoverage, bool& UseCoveragePattern,
                                  bool& UseLM, bool& UseTargetLength)
{
  UseCoverage = m_pDecoderInfo->m_UseCoverageForPruning;
  UseCoveragePattern = m_pDecoderInfo->m_UseCoveragePatternForPruning;
  UseLM = m_pDecoderInfo->m_UseLMStateForPruning;
  UseTargetLength = m_pDecoderInfo->m_UseTargetLengthForPruning;
}



void
CDecoderBLM::GetBeamSize( unsigned long int& BeamSize)
{
  BeamSize = m_BeamSize;
}



void
CDecoderBLM::GetMaxHypsPerCoverage( unsigned long int& MaxHypsPerCov)
{
  MaxHypsPerCov = m_MaxHypsPerCoverage;
}




void
CDecoderBLM::GetScalingFactors( float& TMScale, float& LMScale,
                                float& ALScale, float& SLScale )
{
  cout << "CDecoderBLM::GetScalingFactors should not be used anymore" << endl << flush;\
  assert( 0 );

  /*
  TMScale = m_ScalingFactors[ c_TM1 ];
  LMScale = m_ScalingFactors[ c_LM1 ];
  ALScale = m_ScalingFactors[ c_AL1 ];
  SLScale = m_ScalingFactors[ c_SL1 ];
  */
}



void
CDecoderBLM::GetScalingFactors( CNumericalArray< float >& ScalingFactors)
{
  ScalingFactors = m_ScalingFactors;
}



void
CDecoderBLM::SetRescore( bool Rescore )
{
  m_Rescore = Rescore;
}



void
CDecoderBLM::SetRescorer( CNBestListRescorer *pRescorer )
{
  m_pRescorer = pRescorer;
}



void
CDecoderBLM::SetLabelsConsideredAsWords( CArray< CString > Labels )
{

}



void
CDecoderBLM::SetTagsConsideredAsWords( const CString &TagsConsideredAsWords, char SplitCharacter )
{
  //cout << "SetTagsConsideredAsWords " << TagsConsideredAsWords << endl;

  TagsConsideredAsWords.Split( m_TagsConsideredAsWords, SplitCharacter );

  for ( int i = m_TagsConsideredAsWords.MinIndex(); i <= m_TagsConsideredAsWords.MaxIndex(); i++ ) {
    m_TagsConsideredAsWords[ i ].GetLength();
    cout << m_TagsConsideredAsWords[ i ] << " is treated as normal word " << endl;
  }
}


