#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>

//#define NO_TRACE 1
#undef NO_TRACE

#include <Trace.hh>
#include <Hash.hh>
#include <Array.hh>
#include <DecoderBLM.hh>
#include <OP_Generator.hh>
#include <time.h>

using namespace std;




void
CDecoderBLM::SetLMBatchMode( int BatchMode )
{
  m_LMBatchMode = BatchMode;
}




void
CDecoderBLM::ApplyLM_BatchMode( int Coverage )
{
  cout << "ApplyLM_BatchMode for coverage " << Coverage << endl;
  cout << "Number of hyps = " << m_TmpHyps.Size() << endl;
  // get all the hyps
  // loop over them and generate the LM information (LMState, WordIndexSequence)
  // call the LM
  // update the scores of the hyps


  // loop over all hyps in the temporary storage and create LM translation objects
  CDecoderBLM_Hyp *pHyp;
  vector< CLMTransition > LMTransitions( m_TmpHyps.Size() );
    
  s_TraceOn = true;
  for ( int n = m_TmpHyps.MinIndex(); n <= m_TmpHyps.MaxIndex(); n++ ) {
    pHyp = m_TmpHyps[ n ];
    TraceHyp( false, 0, "in ApplyLM_BatchMode for coverage - create LMTransitions ", pHyp );

    LMTransitions[ n ].OldState = pHyp->m_LMStates[ pHyp->m_LMStates.MinIndex() ];
    //CNumericalArray< LMWordIndex > &LMWordIndices =  pHyp->m_pEdge->LMWordIndices();
    CNumericalMatrix< LMWordIndex > &LMWordIndices =  pHyp->m_pEdge->WordIndicesForLMs();
    cout << "Need to adjust LM functionality" << endl << flush;
    assert( 0 );
    /*
    for ( int i = LMWordIndices.MinIndex(); i <= LMWordIndices.MaxIndex(); i++ ) {
      LMTransitions[ n ].Words.push_back( LMWordIndices[ i ] ); 
    }
    */

    /*
    cout << "Transitions is: "
         << LMTransitions[ n ].OldState << " " 
         << LMTransitions[ n ].NewState << " " 
         << LMTransitions[ n ].Words.size() << " "
         << LMTransitions[ n ].Prob << endl;
    */
  }



  // first we fetch the set of hypotheses which all have coverage c
  // and we need an iterater to access the hyps in this set
  /*
    HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual > &Hyps = m_AllHyps[ Coverage ];
    HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual >::iterator IterHyps;
    CDecoderBLM_Hyp *pHyp;
    
    vector< CLMTransition > LMTransitions( Hyps.size() );
    int n = 0;
    
    for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {
    pHyp = *IterHyps;
    //cout << "Hyp is " << pHyp << endl;
    TraceHyp( s_TraceOn, 0, "\n  in ExpandAllHypsWithCoverage - expand next hyp ", pHyp );
    
    LMTransitions[ n ].OldState = pHyp->m_LMState;
    CNumericalArray< LMWordIndex > &LMWordIndices =  pHyp->m_pEdge->LMWordIndices();
    for ( int i = LMWordIndices.MinIndex(); i <= LMWordIndices.MaxIndex(); i++ ) {
    LMTransitions[ n ].Words.push_back( LMWordIndices[ i ] ); 
    }
    n++;
    }
  */
  
  // write the Probs and new LM states into the hyps
  /*
  cout << "the lm translations before given to the LM" << endl;
  for ( int i = 0; i < LMTransitions.size(); i++ ) {
    cout << i << " " 
         << LMTransitions[ i ].OldState << " " 
         << LMTransitions[ i ].NewState << " " 
         << LMTransitions[ i ].Prob     << endl;
  }
  */

  // no, get the probs from the LM
  cout << "get the probs from the LM" << endl;
  // sv: modify for multiple LMs
  //m_pLMSentenceState->LogProbs( LMTransitions );
  m_LMSentenceStates[ m_LMSentenceStates.MinIndex() ]->LogProbs( LMTransitions );

  cout << "the lm translations after returned from the LM" << endl;
  for ( int i = 0; i < LMTransitions.size(); i++ ) {
    cout << i << " " 
         << LMTransitions[ i ].OldState << " " 
         << LMTransitions[ i ].NewState << " " 
         << LMTransitions[ i ].Prob     << endl;
  }

  s_TraceOn = false;

}


void
CDecoderBLM::PruneBeamSize_BatchLM( int Coverage )
{
  cout << "PruneBeamSize_BatchLM for coverage " << Coverage << endl;

}



void
CDecoderBLM::ExpandAllHypsWithCoverage_BatchLM( int Coverage )
{
  ExpandAllHypsWithCoverage_BeamSizePruning_BatchLM( Coverage );
}




void
CDecoderBLM::ExpandAllHypsWithCoverage_BeamSizePruning_BatchLM( int Coverage )
{

  // to use the LM in batch mode means to collect all LMStates and word sequences
  // and get the LM score for them all in one call to the LM
  // One way to do this is before we expand the hyps with the current coverage
  // So we break things down:
  // - apply LM
  // - prune 
  // - expand



  s_TraceOn = False;

  TraceIf( s_TraceOn, 0, "\n\n  CDecoderBLM::ExpandAllHypsWithCoverage " << Coverage );

  // some variables we need
  CDecoderBLM_Hyp *pHyp, *pBestHyp;
  CLatticeNode *pNode;
  CLatticeEdgeMS *pEdge;


  // how long is the source sentence?  we need to cover all words
  int J = m_pLattice->EndNode()->Number();
  // should we use average path length for speech recognizer lattices?


  // hypotheses are organized according to their coverage
  // hypotheses of coverage c are expanded till coverage J-1
  // is reached.  Expanding those hypotheses will result in 
  // full coverage

  // first we fetch the set of hypotheses which all have coverage c
  // and we need an iterater to access the hyps in this set
  HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual > &Hyps = m_AllHyps[ Coverage ];
  HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual >::iterator IterHyps;


  // for pruning we need the best hypothesis with the current coverage
  HypContainer< CDecoderBLM_Hyp *, HypHashForPruning, HypEqualForPruning > &BestHyps = m_BestHyps[ Coverage ];
  HypContainer< CDecoderBLM_Hyp *, HypHashForPruning, HypEqualForPruning >::iterator IterBestHyps;
  
  //HistogramPruning( Hyps, BestHyps );
  //  TraceIf( s_TraceOn, 0, "After Histogram Pruning:  Number of hyps with coverage " 
  //		   << Coverage << " is " << Hyps.size() );
  

  // sort hyps by score for beamsize pruning
  typedef multimap< float, CDecoderBLM_Hyp * > CSortedHypContainer;

  // associate expansion candidates with their best-hyp-for-pruning
  typedef map< CDecoderBLM_Hyp *, CSortedHypContainer > CBestSimHyp2AllSimHypsMap;

  CBestSimHyp2AllSimHypsMap HypsByBestHyp;
		   
  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {
    pHyp = *IterHyps;
    //cout << "Hyp is " << pHyp << endl;
    TraceIf( s_TraceOn, 0, "\n  in ExpandAllHypsWithCoverage - expand next hyp " 
             << pHyp << " "
             << pHyp->m_Coverage << " " << pHyp->m_CoverageStart 
             << " " << pHyp->m_CoveragePattern
             << " " << pHyp->m_LMStates[ pHyp->m_LMStates.MinIndex() ]
             << " " << pHyp->m_TargetLength 
             << " " << pHyp->m_TotalScore );

    // only expand if hyp had survived histogram pruning
    if ( pHyp->m_Expand == False ) {
      
      //cout << "Prune hyp (histogram) " << pHyp << endl;
      //	  m_HypsPruned++;
      
    } else {
      
      // pruning: only expand hyp if it is better than BeamFactor times 
      // score of best hypothesis
      
      IterBestHyps = BestHyps.find( pHyp );
      if ( IterBestHyps == BestHyps.end() ) {
        //cout <<"Coverage: "<<Coverage<< "no best hyp found" << endl;
        pBestHyp->m_TotalScore = 10000;
      } else {
        pBestHyp = *IterBestHyps;
        //cout <<"Coverage: "<<Coverage<< "Best Hyp is " << pBestHyp << endl;
        TraceIf( s_TraceOn, 0, "\n  best hyp to prune against is " 
                 << pBestHyp << " "
                 << pBestHyp->m_Coverage << " " << pBestHyp->m_CoverageStart 
                 << " " << pBestHyp->m_CoveragePattern
                 << " " << pBestHyp->m_LMStates[ pBestHyp->m_LMStates.MinIndex() ]
                 << " " << pBestHyp->m_TargetLength 
                 << " " << pBestHyp->m_TotalScore );
      }


      if ( m_pFutures != NULL ) {
        cout << "Using Future Costs "<<endl;
	    
        /*
          CArray< CLatticeEdgeMS * > transEdges = TraceBackWithPath(pHyp).second;
          for (int i=transEdges.MinIndex(); i<=transEdges.MaxIndex(); i++){
          transEdges[i]->Display();
          }
        */
	    
        double totalFutureCost = 0.0;
        CArray<std::pair<int,int> >untranslatedSegments = LeftToTranslate(pHyp);

        for (int i=untranslatedSegments.MinIndex(); i<=untranslatedSegments.MaxIndex(); i++){
          //cout<<"UT: I:"<<untranslatedSegments[i].first<<" J: "<<untranslatedSegments[i].second; 
          double futureCostEstimate = m_pFutures->GetFutureCostEstimate(untranslatedSegments[i].first, untranslatedSegments[i].second);
          totalFutureCost += futureCostEstimate;
        }
        //out<<endl;
	    
        //cout<<"BaseCost: "<<pHyp->m_TotalScore<<" additional: "<<totalFutureCost<<endl;

        //use the future cost along with the baseline cost to hash 
        HypsByBestHyp[pBestHyp].insert(CSortedHypContainer::value_type(pHyp->m_TotalScore+totalFutureCost,pHyp));

      } else {

        // if there is no future cost estimate, we just use the sum
        // assert: we have pBestHyp here
        // associate hyp with its best-hyp-for-pruning
        HypsByBestHyp[ pBestHyp ].insert( CSortedHypContainer::value_type( pHyp->m_TotalScore, pHyp ) );

      }

    } // endif: histogram pruned

  } // andfor: all recombined hyps

  
  // foreach set of recombined hyps associated to their best hyp, expand beamsize best ones
  int rank;
  CSortedHypContainer::iterator IterSortedHyp;
  CSortedHypContainer HypsInBeam;

  for (CBestSimHyp2AllSimHypsMap::iterator IterHypsByBestHyp = HypsByBestHyp.begin();
       IterHypsByBestHyp != HypsByBestHyp.end();
       ++IterHypsByBestHyp) {
    // sorted list of expansion candidates associated with current best hyp for pruning
    CDecoderBLM_Hyp* pBestHyp = (*IterHypsByBestHyp).first;

    if (s_TraceOn){
      cout<<"Coverage: "<<Coverage<<endl;
      cout<<"pBestHyp:  ";
      //display the best hyp
      CArray< CLatticeEdgeMS * > transEdges = TraceBackWithPath(pBestHyp).second;
      for (int i=transEdges.MinIndex(); i<=transEdges.MaxIndex(); i++){
        transEdges[i]->Display();
        cout<<" ";
      }
      cout<<endl<<endl;
    }

    CSortedHypContainer &SimilarHyps = (*IterHypsByBestHyp).second;
    
    if (s_TraceOn){
      //lets look at all hyps in beam first
      
      for (IterSortedHyp  = SimilarHyps.begin(); IterSortedHyp != SimilarHyps.end();  ++IterSortedHyp){
        CDecoderBLM_Hyp* pHyp = (*IterSortedHyp).second;
        //display the hyp
        CArray< CLatticeEdgeMS * > transEdges = TraceBackWithPath(pHyp).second;
        for (int i=transEdges.MinIndex(); i<=transEdges.MaxIndex(); i++){
          transEdges[i]->Display();
          cout<<" ";
        }
        cout<<endl;
        cout<<"FScore: "<<(*IterSortedHyp).first<<" base: "<<pHyp->m_TotalScore<<endl;
        
      }
    }
    

    // only take the beamsize best
    for(IterSortedHyp = SimilarHyps.begin(), rank = 0;
        IterSortedHyp != SimilarHyps.end() && ++rank <= m_BeamSize;
        ++IterSortedHyp) {
      
      HypsInBeam.insert(*IterSortedHyp);
    } // endfor

    m_HypsOutOfBeam += ( SimilarHyps.size() > m_BeamSize ?
                         SimilarHyps.size() - m_BeamSize :
                         0 );
  } // endfor


  // out of the hyps just collected, only take the m_MaxHypsPerCoverage best
  for (IterSortedHyp  = HypsInBeam.begin(), rank = 0;
       IterSortedHyp != HypsInBeam.end() && ++rank <= m_MaxHypsPerCoverage;
       ++IterSortedHyp)
    ExpandHyp((*IterSortedHyp).second);
  
  m_HypsOutOfBeamPerCov += ( HypsInBeam.size() > m_MaxHypsPerCoverage ?
                             HypsInBeam.size() - m_MaxHypsPerCoverage :
                             0 );
  
  s_TraceOn = false;

} // CDecoderBLM::ExpandAllHypsWithCoverage(...)


