#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <fortify.h>
#include <iostream>

#define NO_TRACE 1
//#undef NO_TRACE
#include <Trace.hh>

#include <Array.hh>
#include <DecoderGT.hh>
#include <OP_Generator.hh>

//bool CDecoderGT_Hyp::s_UseCoverage = true;
//bool CDecoderGT_Hyp::s_UseCoveragePattern = true;
//bool CDecoderGT_Hyp::s_UseLMState = true;
//bool CDecoderGT_Hyp::s_UseTargetLength = true;

//bool CDecoderGT_Hyp::s_UseCoverageForPruning = true;
//bool CDecoderGT_Hyp::s_UseCoveragePatternForPruning = true;
//bool CDecoderGT_Hyp::s_UseLMStateForPruning = true;
//bool CDecoderGT_Hyp::s_UseTargetLengthForPruning = false;

//short CDecoderGT_Hyp::s_MaxNumReorderings = 77;

long int CDecoderGT_Hyp::s_NumberOfObjects = 0;

bool CDecoderGT::s_TraceOn = false;
//ARV 
int CDecoderGT::m_numInstances = 1;
bool CDecoderGT_Hyp::s_TraceOn = false;


std::ostream &
operator <<( std::ostream& st, const CDecoderGT_Hyp &hyp)
{
  st 
    << "\nTotal Score " << hyp.m_TotalScore
    //	 << "\nACScore       " << hyp.m_ACScore 
    << "  TMScore " << hyp.m_TMScore
    << "  LMScore " << hyp.m_LMScore 
    << "  ALScore " << hyp.m_ALScore
    << "  SLScore " << hyp.m_SLScore
    << "\nCoverage " << hyp.m_Coverage << " " << hyp.m_CoverageStart << " " << hyp.m_CoveragePattern 
    << "  LMState " << hyp.m_LMState 
    << "  Target Length " << hyp.m_TargetLength 
    << "\nEdge " << hyp.m_pEdge 
    << "  Previous Hyp " << hyp.m_pPrevHyp;
  //	 << "\nNumber Hyps   " << hyp.s_NumberOfObjects;
  
  return st;
  
}


bool
CDecoderGT_Hyp::TestAndSetCoverage( int j_min, int j_max )
{
  bool TraceOn = s_TraceOn;
  s_TraceOn = false;

  TraceIf( s_TraceOn, 0, "CDecoderGT_Hyp::TestAndSetCoverage " << j_min << " " << j_max );

  // the parameters are interpreted as the positions between the words
  // i.e. as node numbers in a lattice
  // 2 3 means therefore: test coverage of word running from 
  // node 2 to 3
  // the coverage signature itself is in terms of positions of the words
  // i.e. 2 to 3 is word position 2

  // return value: true if all positions free, false if collision


  TraceIf( s_TraceOn, 0, m_CoverageStart << " " << m_CoveragePattern );

  // test that first position is free, as otherwise the shifting of
  // the coverage pattern failed in the previous step
  assert( !m_CoveragePattern[ 0 ] );

  // make sure expansion is not moving outside of maximally allowed region
  // otherwise we can not store the coverage pattern
  // ToDo: should allow for long expansion without gap
  if ( j_max - m_CoverageStart > c_MaxReorderWindowLength ) {
	TraceIf( true, 0, "j_max - m_CoverageStart <= c_MaxReorderWindowLength" );
	//assert( j_max - m_CoverageStart <= c_MaxReorderWindowLength );
	return false;
  }
  

  // make sure we are starting beyond what has already been completely covered
  if ( j_min < m_CoverageStart ) {
	TraceIf( true, 0, "j_min = " << j_min << " < m_CoverageStart = " << m_CoverageStart );
	assert( j_min >= m_CoverageStart );
	return false;
  }

  // first we check if there is a collision
  // at the same time we set the requested positions
  for ( int j = j_min; j < j_max; j++ ) {
	if ( m_CoveragePattern[ j - m_CoverageStart ] ) {
	  TraceIf( s_TraceOn, 0, "Collision at position " << j );
	  return false;
	} else {
	  TraceIf( s_TraceOn, 0, "set bit at position " << j );
	  m_CoveragePattern[ j - m_CoverageStart ] = true;
	  TraceIf( s_TraceOn, 0, m_CoveragePattern );
	  m_Coverage++;
	}
  }

  // now we check if the lowest bits are set
  // if yes, shift the pattern and increment CoverageStart
  while ( m_CoveragePattern[ 0 ] ) {
	TraceIf( s_TraceOn, 0, "shift start of coverage pattern" );
	m_CoveragePattern >>= 1;
	m_CoverageStart++;
	TraceIf( s_TraceOn, 0, m_CoverageStart << " " << m_CoveragePattern );
  }

  assert( !m_CoveragePattern[ 0 ] );

  s_TraceOn = TraceOn;

  return true;
}




//============================================================
CDecoderGT::CDecoderGT()
{
  TraceIf( s_TraceOn, 0, "CDecoderGT::CDecoderGT" );

  m_TMScale = 1.0;
  m_LMScale = 1.0;
  m_ALScale = 1.0;
  m_SLScale = 1.0;

  m_AddOverlappingPhrases = true;

  // default, no reordering
  m_LengthOfReorderingWindow = 1;

  // default, 1-best decoding
  m_NumOfNBest = 1;

  // default pruning factor
  m_BeamFactor = 10000.0;

  // for histogram pruning use 100 buckets
  m_PruningBuckets.SetBoundaries( 0, 100 );


  m_pDecoderInfo = new CDecoderGT_Info;

  // the standard working hypothesis
  m_pNumberOneHypothesis = new CDecoderGT_Hyp;
  m_pNumberOneHypothesis->m_pDecoderInfo = m_pDecoderInfo;
  m_pNumberOneBestHyp = new CDecoderGT_Hyp;
  m_pNumberOneBestHyp->m_pDecoderInfo = m_pDecoderInfo;
  //ARV
  CDecoderGT::m_numInstances = 1;

}


CDecoderGT::CDecoderGT( const CDecoderGT &Decoder )
{
  TraceIf( s_TraceOn, 0, "CDecoderGT::CDecoderGT" );

  m_TMScale = Decoder.m_TMScale;
  m_LMScale = Decoder.m_LMScale;
  m_ALScale = Decoder.m_ALScale;
  m_SLStretch = Decoder.m_SLStretch;
  m_SLScale = Decoder.m_SLScale;
  m_BeamFactor = Decoder.m_BeamFactor;
  m_MaxHyps = 1000;
  //std::cout<<"Coping BeamFactor:"<<m_BeamFactor<<std::endl;
  m_AddOverlappingPhrases = Decoder.m_AddOverlappingPhrases;

  // paremeters for reordering
  m_LengthOfReorderingWindow = Decoder.m_LengthOfReorderingWindow;

  // default, 1-best decoding
  m_NumOfNBest = Decoder.m_NumOfNBest;

  m_pDecoderInfo = new CDecoderGT_Info;
  (*m_pDecoderInfo) = (*(Decoder.m_pDecoderInfo));

  // the standard working hypothesis
  m_pNumberOneHypothesis = new CDecoderGT_Hyp;
  m_pNumberOneHypothesis->m_pDecoderInfo = m_pDecoderInfo;
  m_pNumberOneBestHyp = new CDecoderGT_Hyp;
  m_pNumberOneBestHyp->m_pDecoderInfo = m_pDecoderInfo;

  // copy translation memory
  m_pTranslationMemory = Decoder.m_pTranslationMemory;

  m_pLM = Decoder.m_pLM;
  m_pLMSentenceState = NULL;  // use own LM Sentence State
  //ARV
  CDecoderGT::m_numInstances++;
}



CDecoderGT::~CDecoderGT()
{
    //ARV
  CDecoderGT::m_numInstances--;
  if (m_numInstances==1){
    delete m_pNumberOneHypothesis;
    delete m_pNumberOneBestHyp;
  }

  delete m_pDecoderInfo;

}



void
CDecoderGT::Init( CTranslationMemory *pTranslationMemory, CLM *pLM )
{
  m_pTranslationMemory = pTranslationMemory;
  m_pLM = pLM;
  //m_pTranslationMemory->RenormalizeTransducer();
}



void
CDecoderGT::Init( const char *ParamFile, CVocabulary *pVocabulary )
{
  m_pTranslationMemory->Init( ParamFile, pVocabulary );
  m_pTranslationMemory->RenormalizeTransducer();
  m_pLM = m_pTranslationMemory->GetLM();
}




void
CDecoderGT::Clear()
{
  HypContainer< CDecoderGT_Hyp *, HypHash, HypEqual >::iterator IterHyps;
  HypContainer< CDecoderGT_Hyp *, HypHashForPruning, HypEqualForPruning >::iterator IterBestHyps;

  CDecoderGT_Hyp *pHyp;

  for ( int Coverage = m_AllHyps.MinIndex(); Coverage <=  m_AllHyps.MaxIndex(); Coverage++ ) {
	//std::cout << "Stored hyps for coverage " << Coverage << " is " << m_AllHyps[ Coverage ].size() << std::endl;
	//std::cout << "delete hyps" << std::endl << flush;

	HypContainer< CDecoderGT_Hyp *, HypHash, HypEqual > &Hyps = m_AllHyps[ Coverage ];

	int NumHyp = 0;
	
	/*
	while (  Hyps.size() ) {
	  IterHyps = Hyps.begin();
	  pHyp = *IterHyps;
	  //std::cout << " hyp " << NumHyps << " with coverage " << Coverage << std::endl << flush;
	  Hyps.erase( IterHyps );
	  delete pHyp;
	}
	*/
	
	
	for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {
	  pHyp = *IterHyps;
	  m_HypsForDeleting.SetAt( NumHyp, pHyp );
	  NumHyp++;
	  //std::cout << pHyp << " " <<  Hyps.size() << std::endl << flush;
	  //Hyps.erase( IterHyps );
	  //delete pHyp;
	}
	
	for ( int i = 0; i < NumHyp; i++ ) {
	//std::cout << *(m_HypsForDeleting[ i ]) << std::endl;
	  delete m_HypsForDeleting[ i ];
	}
	//std::cout << "Stored hyps for coverage " << Coverage << " is " << m_AllHyps[ Coverage ].size() << std::endl;

	Hyps.clear();
	

	//std::cout << "Stored hyps for coverage " << Coverage << " is " << m_AllHyps[ Coverage ].size() << std::endl;


	// the best hyps used for pruning
	HypContainer< CDecoderGT_Hyp *, HypHashForPruning, HypEqualForPruning > &BestHyps = m_BestHyps[ Coverage ];

	NumHyp = 0;
	
	for ( IterBestHyps = BestHyps.begin(); IterBestHyps != BestHyps.end(); ++IterBestHyps ) {
	  pHyp = *IterBestHyps;
	  m_HypsForDeleting.SetAt( NumHyp, pHyp );
	  NumHyp++;
	  //std::cout << pHyp << " " <<  Hyps.size() << std::endl << flush;
	  //Hyps.erase( IterHyps );
	  //delete pHyp;
	}
	
	for ( int i = 0; i < NumHyp; i++ ) {
	//std::cout << *(m_HypsForDeleting[ i ]) << std::endl;
	  delete m_HypsForDeleting[ i ];
	}
	//std::cout << "Stored hyps for coverage " << Coverage << " is " << m_AllHyps[ Coverage ].size() << std::endl;

	BestHyps.clear();

	//std::cout << "Stored hyps for coverage " << Coverage << " is " << m_AllHyps[ Coverage ].size() << std::endl;

  }


  // and the hyps in the final list
  multimap< double, CDecoderGT_Hyp * >::iterator Iter = m_FinalHyps.begin();

  /*
  while (  m_FinalHyps.size() ) {
	Iter = m_FinalHyps.begin();
	pHyp = Iter->second;
	//std::cout << " hyp " << NumHyps << " with coverage " << Coverage << std::endl << flush;
	m_FinalHyps.erase( Iter );
	delete pHyp;
  }
  */


  int NumHyp = 0;
  for ( Iter = m_FinalHyps.begin(); Iter != m_FinalHyps.end(); ++Iter ) {
	pHyp = Iter->second;
	m_HypsForDeleting.SetAt( NumHyp, pHyp );
	NumHyp++;
	//std::cout << " hyp " << NumHyps << " with coverage " << Coverage << std::endl << flush;
	//m_FinalHyps.erase( Iter );
	//delete pHyp;
  }

  for ( int i = 0; i < NumHyp; i++ ) {
	//std::cout << *(m_HypsForDeleting[ i ]) << std::endl;
	delete m_HypsForDeleting[ i ];
  }

  m_FinalHyps.clear();
  
  //std::cout << "Stored final hyps is " << m_FinalHyps.size() << std::endl;
  //std::cout << "all deleted" << std::endl << flush;

}



CString
CDecoderGT::TranslateSentence( const CSentence *pSentence )
{
  TraceIf( s_TraceOn, 0, "CDecoderGT::TranslateSentence" );
  return m_pTranslationMemory->TranslateSentence( pSentence );
}



CArray<CString>
CDecoderGT::TranslateLattice( CLatticeLM *pLattice )
{
  TraceIf( s_TraceOn, 0, "CDecoderGT::TranslateLattice" );
  //return m_pTranslationMemory->TranslateLattice( pLattice );

  m_pTranslationMemory->BuildTranslationLattice( pLattice );
  
  //pLattice->Display();
  // insert overlapping phrases
  //std::cout << "overlapping phrases" << std::endl;
  if ( m_AddOverlappingPhrases ) {
	COverlappingPhraseGenerator *pOPGenerator = new COverlappingPhraseGenerator;
	pOPGenerator->AddOverlappingPhrases( pLattice );
  }


  // here we do the new decoding
  return   ApplyLanguageModel( pLattice );
}



void
CDecoderGT::AddOverlappingPhrases( CLattice *pLattice )
{
  TraceIf( s_TraceOn, 0, "CDecoderGT::AddOverlappingPhrases" );
  
  // insert overlapping phrases
  std::cout << "overlapping phrases" << std::endl;
  if ( m_AddOverlappingPhrases ) {
	COverlappingPhraseGenerator *pOPGenerator = new COverlappingPhraseGenerator;
	pOPGenerator->AddOverlappingPhrases( pLattice );
  }

}



void
CDecoderGT::BuildTranslationLattice( CLatticeLM *pLattice )
{
  TraceIf( s_TraceOn, 0, "CDecoderGT::BuildTranslationLattice" );
  m_pTranslationMemory->BuildTranslationLattice( pLattice );

}



CArray<CString>
CDecoderGT::ApplyLanguageModel( CLatticeLM *pLattice, CLM *pLM )
{
  
  TraceIf( s_TraceOn, 0, "CDecoderGT::ApplyLanguageModel" );

  // lattice and LM are needed elsewhere, so assign them to member variable
  m_pLattice = pLattice;
  m_pLM = pLM;

  return ApplyLanguageModel();
}


CArray<CString>
CDecoderGT::ApplyLanguageModel( CLatticeLM *pLattice )
{
  TraceIf( s_TraceOn, 0, "CDecoderGT::ApplyLanguageModel" );

  // lattice needed elsewhere, so assign them to member variable
  m_pLattice = pLattice;

  return ApplyLanguageModel();
}



CArray<CString>
CDecoderGT::ApplyLanguageModel()
{
  s_TraceOn = false;
  TraceIf( s_TraceOn, 0, "CDecoderGT::ApplyLanguageModel" );

  TraceIf( s_TraceOn, 0, "Length of reordering window is " << m_LengthOfReorderingWindow );
  
  //m_pLattice->Display();
  //WriteLattice();

  // internal statistics
  m_HypsExpanded = 0;
  m_HypsNotExpanded = 0;
  m_HypsPruned = 0;
  m_NumNewHyps = 0;
  m_NumBetterHyps = 0;
  m_NumWorseHyps = 0;
  m_NumCoverageCollisions = 0;

  // first convert all string at edges in lattice 
  // into word indices according to LM vocabulary
  GetLMWordIndicesAndTMScores();

  // get the LMSentenceState and the LMState for sentence start
  // LM state
  CLMState LMState = (CLMState) 0;
  CLMState NewLMState = (CLMState) 0;
  
  if ( m_pLM ) {
    m_pLMSentenceState = m_pLM->NewSentence();
    LMState = m_pLMSentenceState->BeginOfSentenceState();
    //ARV output
    //std::cout << "LMState for begin of sentence is " << LMState <<std::endl;
  }


  // how long is the source sentence?  we need to cover all words
  int Start = m_pLattice->StartNode()->Number();
  int J = m_pLattice->EndNode()->Number();
  //std::cout << "StartNode " << Start << " EndNode " << J << std::endl;
  //std::cout << "Sentence length = " << J << std::endl;

  // should we use average path length for speech recognizer lattices?

  // based on sentence length we set maximum of allowed reorderings
  // as 'a b c d' -> 'a c b d' will count as two reorderings, we
  // multiply by extra factor of 2
  m_pDecoderInfo->m_MaxNumReorderings = (int) (J * 2 * m_ReorderingsPerWord);
  
  //ARV output
  /*
  std::cout << "MaxNumReorderings " << m_pDecoderInfo->m_MaxNumReorderings / 2 << std::endl;
  std::cout << "Length of reordering window is " << m_LengthOfReorderingWindow << std::endl;
  std::cout << "TMScale = "<< m_TMScale<< " LMScale = " << m_LMScale << " ALScale = " << m_ALScale 
	   << " SLStretch = " << m_SLStretch << " SLScale = " << m_SLScale << std::endl;
  std::cout << "Comparing Hyps with"
	   << " Coverage " <<  m_pDecoderInfo->m_UseCoverage 
	   << " Coverage Pattern " <<  m_pDecoderInfo->m_UseCoveragePattern
	   << " LM state " << m_pDecoderInfo->m_UseLMState
	   << " Target stentence length " <<  m_pDecoderInfo->m_UseTargetLength << std::endl;
  std::cout << "Pruning Hyps with"
	   << " Coverage " <<  m_pDecoderInfo->m_UseCoverageForPruning 
	   << " Coverage Pattern " <<  m_pDecoderInfo->m_UseCoveragePatternForPruning
	   << " LM state " << m_pDecoderInfo->m_UseLMStateForPruning
	   << " Target stentence length " <<  m_pDecoderInfo->m_UseTargetLengthForPruning << std::endl;
  std::cout << "Beam Factor = " << m_BeamFactor << std::endl;
  */

  // array of multi-hash-sets is dynamic, i.e. when array grows, 
  // items are copied to new memory space; this causes addresses 
  // to individual hash-sets to become invalid
  // therefore, loop over entire range and access hash-set
  // this will create empty hash sets and avoid later copying
  // if a previous sentence has been processed, the hyps are
  // still stored, we clear them
  HypContainer< CDecoderGT_Hyp *, HypHash, HypEqual >::iterator IterHyps;
  CDecoderGT_Hyp *pHyp, *pBestHyp;

  for ( int Coverage = 0; Coverage <= J - Start; Coverage++ ) {
	m_AllHyps[ Coverage ].clear();
	m_BestHyps[ Coverage ].clear();
  }


  // we start with an empty hypothesis
  pHyp = new CDecoderGT_Hyp;
  pHyp->m_pDecoderInfo = m_pDecoderInfo;
  pHyp->m_LMState = LMState;
  //std::cout << "Initial Hypothesis: " << *pHyp << std::endl;

  // and place it into the list of all hypotheses with coverage 0
  m_AllHyps[ 0 ].insert( pHyp );
  //std::cout << "Initial Hypothesis after storing: " << *pHyp << std::endl;

  // also best hyp for this coverage
  pBestHyp = new CDecoderGT_Hyp;
  pBestHyp->m_pDecoderInfo = m_pDecoderInfo;
  pBestHyp->m_LMState = LMState;
  m_BestHyps[ 0 ].insert( pBestHyp );

  // hypotheses are organized according to their coverage
  // hypotheses are expanded till full coverage J is reached
  // i.e. we expand all hyps with Covereage < J

  for ( int Coverage = 0; Coverage < J - Start; Coverage++ ) {
	// display all hyps with current coverage
	//DisplayAllHypsWithCoverage( Coverage );

	// expand all hyps
	ExpandAllHypsWithCoverage( Coverage );

	//for ( int Coverage = 0; Coverage <= J; Coverage++ ) {
	//std::cout << "Stored hyps for coverage " << Coverage << " is " 
	//<< m_AllHyps[ Coverage ].size() << std::endl;
	//}
  }


  // display all hyps with full coverage
  //DisplayAllHypsWithCoverage( J, 1 );

  // do specific sentence end stuff
  ExpandAllHypsSentenceEnd();

  // reconstruct the n-best translations using the backpointer information
  //CString Translation = BestTranslation();
  TraceIf( s_TraceOn, 0, "\nDisplay n-best hypotheses with backtrace for n-best = " << m_NumOfNBest );  
  TraceIf( 1, 0, "\nNumber of hypotheses: " << m_FinalHyps.size() );
  
  std::cout << "NBest = " << m_NumOfNBest << std::endl;

  int n = 1;
  CArray<CString> translationList;
  for (multimap< double, CDecoderGT_Hyp * >::iterator Iter = m_FinalHyps.begin();
       Iter != m_FinalHyps.end();
       ++Iter) {
	if ( n <= m_NumOfNBest ) {
	  pHyp = (*Iter).second;
	  std::cout << "Trace Back hypothesis " << n << std::endl;
	  translationList[n] = TraceBack( pHyp );
	  n++;
	}
  }


  // let's see how bad the worst hyp is
  // for (multimap< double, CDecoderGT_Hyp * >::iterator Iter = m_FinalHyps.begin();
  //      Iter != m_FinalHyps.end();
  //      ++Iter) {
// 	pHyp = (*Iter).second;
  // }
  // std::cout << "Worst hypothesis" << std::endl;
  // std::cout << *pHyp << std::endl;

  
  //ARV output
  s_TraceOn = false;
  Trace( s_TraceOn, "Hyps expanded in this sentence          = " << m_HypsExpanded );
  Trace( s_TraceOn, "Hyps expanded per word                  = " << m_HypsExpanded / J );
  Trace( s_TraceOn, "Hyps not expanded (reordering exceeded) = " << m_HypsNotExpanded );
  Trace( s_TraceOn, "Hyps not expanded due to pruning        = " << m_HypsPruned );
  Trace( s_TraceOn, "New Hyps in this sentence               = " << m_NumNewHyps );
  Trace( s_TraceOn, "Better Hyps in this sentence            = " << m_NumBetterHyps );
  Trace( s_TraceOn, "Worse Hyps in this sentence             = " << m_NumWorseHyps );
  Trace( s_TraceOn, "Collisions in this sentence             = " << m_NumCoverageCollisions );

  Clear();

  return translationList;
}



void
CDecoderGT::GetLMWordIndicesAndTMScores()
{

  CArray< CLatticeEdge * > Edges = m_pLattice->Edges();

  // array to collect words attached to one edge for LM-scoring
  CArray< CString > Words;
  CArray< LMWordIndex > WordIndices;

  // first get the words
  CString Word;
  string word;
  CLatticeEdgeTrans *pEdge;


  for ( int i = Edges.MinIndex(); i <= Edges.MaxIndex(); i++ ) {
	pEdge = (CLatticeEdgeTrans *)Edges[ i ];
	
	Word = pEdge->Word();
	word = (const char*)(Word);
	CWordInfo *pWordInfo =  ((CLatticeEdgeTrans *)pEdge)->WordInfo();
	
	Trace( 2, "WordInfo: " << *pWordInfo );
	Trace( 2, "SourceString: " << pWordInfo->SourceString() );
	Trace( 2, "TargetString: " << pWordInfo->TargetString() );
	
	// if edge was not generated by a transducer item
	// then set TM score to TMScoreForSourceWord
	if ( Word[ 0 ] != '@' ) {
	  pEdge->WordInfo()->Score() = c_TMScoreForSourceWord;
	  TraceIf( s_TraceOn, 0, "      Translation Score for source language edge set to : " 
			   << c_TMScoreForSourceWord );
	}

	// extract sequence of words from the WordInfo structure
	if ( pWordInfo->HasTargetString() ) {
	  pWordInfo->GetTargetWords( Words );
	} else if ( pWordInfo->HasSourceString() ) {
	  pWordInfo->GetSourceWords( Words );
	} else {
	  Word = pEdge->Word();
	  Words.SetAt( 1, Word );
	}

	// convert string into indices according to the LM vocabulary
	for ( int t = Words.MinIndex(); t <= Words.MaxIndex(); t++ ) {
	  LMWordIndex WordIndex = m_pLM->GetVocabularyIndex( (string)((const char *)(Words[ t ])) );
	  pEdge->LMWordIndices()[ t ] = WordIndex;
	  TraceIf( s_TraceOn, 0, "Word = " << Words[ t ] << "  LM-WordIndex = " << WordIndex );
	}

  }

}




void
CDecoderGT::ExpandAllHypsWithCoverage( int Coverage )
{
  s_TraceOn = false;

  TraceIf( s_TraceOn, 0, "\n\n  CDecoderGT::ExpandAllHypsWithCoverage " << Coverage );

  // variable we need
  CDecoderGT_Hyp *pHyp, *pBestHyp;
  CLatticeNode *pNode;
  CLatticeEdgeTrans *pEdge;



  // how long is the source sentence?  we need to cover all words
  int J = m_pLattice->EndNode()->Number();
  // should we use average path length for speech recognizer lattices?

  // hypotheses are organized according to their coverage
  // hypotheses of coverage c are expanded till coverage J-1
  // is reached.  Expanding those hypotheses will result in 
  // full coverage

  // first we fetch the set of hypotheses which all have coverage c
  // and we need an iterater to access the hyps in this set
  HypContainer< CDecoderGT_Hyp *, HypHash, HypEqual > &Hyps = m_AllHyps[ Coverage ];
  HypContainer< CDecoderGT_Hyp *, HypHash, HypEqual >::iterator IterHyps;


  // for pruning we need the best hypothesis with the current coverage
  HypContainer< CDecoderGT_Hyp *, HypHashForPruning, HypEqualForPruning > &BestHyps = m_BestHyps[ Coverage ];
  HypContainer< CDecoderGT_Hyp *, HypHashForPruning, HypEqualForPruning >::iterator IterBestHyps;
  
  TraceIf( s_TraceOn, 0, "  Number of best hyps with coverage " 
		   << Coverage << " is " << BestHyps.size() );

  for ( IterBestHyps = BestHyps.begin(); IterBestHyps != BestHyps.end(); ++IterBestHyps ) {
 	pBestHyp = *IterBestHyps;
 	TraceIf( s_TraceOn, 0, "  in ExpandAllHypsWithCoverage - existing best hyp " 
 			 << pBestHyp << " "
 			 << pBestHyp->m_Coverage << " " << pBestHyp->m_CoverageStart 
			 << " " << pBestHyp->m_CoveragePattern 
			 << " " << pBestHyp->m_LMState
			 << " " << pBestHyp->m_TargetLength
			 << " " << pBestHyp->m_TotalScore );
  }
  


  TraceIf( s_TraceOn, 0, "  Number of hyps with coverage " 
		   << Coverage << " is " << Hyps.size() );

  // for each hypothesis h with coverage Coverage
  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {
	//std::cout << *IterHyps << std::endl;
 	pHyp = *IterHyps;
 	TraceIf( s_TraceOn, 0, "  in ExpandAllHypsWithCoverage - existing hyp " 
 			 << pHyp << " "
 			 << pHyp->m_Coverage << " " << pHyp->m_CoverageStart 
			 << " " << pHyp->m_CoveragePattern
			 << " " << pHyp->m_LMState
			 << " " << pHyp->m_TargetLength 
			 << " " << pHyp->m_TotalScore );
	//if ( Coverage == 5 ) assert( 0 );
  }

  // ARV removed for demo 
  //HistogramPruning( Hyps, BestHyps );


  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {
	//TraceIf( s_TraceOn, 0, "  Number of hyps with coverage " 
	//<< Coverage << " is " << m_AllHyps[ Coverage ].size() );
	//TraceIf( s_TraceOn, 0, "Address of Hyps is: " << &Hyps );
	//TraceIf( s_TraceOn, 0, "Address of m_AllHyps[ 0 ] is: " << &m_AllHyps[ Coverage ] );

	pHyp = *IterHyps;
	//std::cout << "Hyp is " << pHyp << std::endl;
	TraceIf( s_TraceOn, 0, "\n  in ExpandAllHypsWithCoverage - expand next hyp " 
			 << pHyp << " "
			 << pHyp->m_Coverage << " " << pHyp->m_CoverageStart 
			 << " " << pHyp->m_CoveragePattern
			 << " " << pHyp->m_LMState
			 << " " << pHyp->m_TargetLength 
			 << " " << pHyp->m_TotalScore );

	// only expand if hyp had survived histogram pruning
	if ( pHyp->m_Expand == False ) {

	  //std::cout << "Prune hyp (histogram) " << pHyp << std::endl;
	  m_HypsPruned++;

	} else {

	  // pruning: only expand hyp if it is better than BeamFactor times 
	  // score of best hypothesis
	  IterBestHyps = BestHyps.find( pHyp );
	  if ( IterBestHyps == BestHyps.end() ) {
		//std::cout << "no best hyp found" << std::endl;
		pBestHyp->m_TotalScore = 10000;
	  } else {
		pBestHyp = *IterBestHyps;
		//std::cout << "Best Hyp is " << pBestHyp << std::endl;
		TraceIf( s_TraceOn, 0, "\n  best hyp to prune against is " 
				 << pBestHyp << " "
				 << pBestHyp->m_Coverage << " " << pBestHyp->m_CoverageStart 
				 << " " << pBestHyp->m_CoveragePattern
				 << " " << pBestHyp->m_LMState
				 << " " << pBestHyp->m_TargetLength 
				 << " " << pBestHyp->m_TotalScore );
	  }

	
	  // now the expansion
	  if ( pHyp->m_TotalScore > m_BeamFactor + pBestHyp->m_TotalScore ) {
		//std::cout << pHyp->m_TotalScore << " " << m_BeamFactor << " " << pBestHyp->m_TotalScore <<std::endl;
	    if (s_TraceOn){
	      std::cout << "Prune hyp (beam) " << pHyp << std::endl;
	    }	
	    m_HypsPruned++;
	  } else {
		
		ExpandHyp( pHyp );

		//TraceIf( s_TraceOn, 0, "  Number of hyps with coverage " << Coverage 
		//<< " is " << Hyps.size() );
		//TraceIf( s_TraceOn, 0, "  Number of hyps with coverage " << Coverage 
		//<< " is " << m_AllHyps[ Coverage ].size() );
		//TraceIf( s_TraceOn, 0, "Address of Hyps is: " << &Hyps );
		//	TraceIf( s_TraceOn, 0, "Address of m_AllHyps[ 0 ] is: " << &m_AllHyps[ Coverage ] );
		
		//	assert( 0 );
	  }
	}
  }

  s_TraceOn = false;
}



void
CDecoderGT::ExpandHyp( CDecoderGT_Hyp *pHyp )
{
  TraceIf( s_TraceOn, 0, "\n    CDecoderGT::ExpandHyp" );

  // variable we need
  CLatticeNode *pNode;
  CLatticeEdgeTrans *pEdge;


  TraceIf( s_TraceOn, 1, "    expand hyp " << *pHyp );

  // get the range of words in the source sentence 
  // which can now be translated, this is the active range
  // make sure that reordering stays within sentence
  int J = m_pLattice->EndNode()->Number();
  int j_min = pHyp->m_CoverageStart;
  int j_max = pHyp->m_CoverageStart + m_LengthOfReorderingWindow;
  if ( j_max > J ) 
	j_max = J;

  TraceIf( s_TraceOn, 0, "    Translate words in the range " << j_min << " " << j_max );
	
  // now loop over all partial translations t
  // i.e. over all edges in the translation lattice
  // which are within the active range
  // the end point of the edge has also to be within the maximally
  // allowed reordering window (c_MaxReorderWindowLength)

  // currently we reorder also source words, this could (should?) be avoided

  for ( int j = j_min; j < j_max; j++ ) {
	// get node
	pNode = m_pLattice->Nodes()[ j ];
	TraceIf( s_TraceOn, 0, "    process node with number " << pNode->Number() );
	

	// expand this hypothesis only if it would not result in too 
	// many reorderings, but always expand over outgoing edges of node 
	// of current hypothesis, ie the node up to which everything has 
	// been translated (PrevNode); this is the monotone decoding part
	// the initial hyp has no backpointer edge, the number of the previous
	// node is just the first not in the lattice
	int NumberOfPrevNode;
	if ( pHyp->m_pEdge ) {
	  NumberOfPrevNode= pHyp->m_pEdge->ToNode()->Number();
	} else {
	  NumberOfPrevNode =  m_pLattice->StartNode()->Number();
	}

	if ( (pNode->Number() == NumberOfPrevNode) 
		 || (pHyp->m_NumReorderings < pHyp->m_pDecoderInfo->m_MaxNumReorderings) ) {

	  // loop over all outgoing edges e
	  const CArray< CLatticeEdge * > & Edges = pNode->OutEdges();
	  TraceIf( s_TraceOn, 0, "    Number of outgoing edges is " << Edges.Size() );
	  
	  for ( int EdgeNum = Edges.MinIndex();
			EdgeNum <= Edges.MaxIndex();
			EdgeNum++ ) {
	  
		pEdge = (CLatticeEdgeTrans *)Edges[ EdgeNum ];
		
		// if edge is long it can move outside of the 

		ExpandHypOverEdge( pHyp, pEdge );
		
	  } // forall edges

	} else {

	  m_HypsNotExpanded++;
	} // if not to many reordering
	
  } // for all positions within reordering window

}




void
CDecoderGT::ExpandHypOverEdge( CDecoderGT_Hyp *pHyp, CLatticeEdgeTrans *pEdge )
{
  TraceIf( s_TraceOn, 1, "\n      CDecoderGT::ExpandHypOverEdge " << pHyp << " " << pEdge );

  TraceIf( s_TraceOn, 1, "      expand hyp " << *pHyp  );
  TraceIf( s_TraceOn, 1, "      expand over edge " );
  //pEdge->Display();

  double TMScore;
  double LMScore;
  double ALScore;

  CLMState LMState = (CLMState) 0;
  CLMState NewLMState = (CLMState) 0;

  CDecoderGT_Hyp *pBestHyp;

  bool NeedNewNumberOneHyp = false;
  bool NeedNewNumberOneBestHyp = false;

  // internal statistics
  m_HypsExpanded++;

  // test coverage collision for current hypothesis and this edge
  // we copy the current hypothesis into the NumberOneHyp,
  // as the TestAndSetCoverage function will modify the hypothesis
  *m_pNumberOneHypothesis = *pHyp;
  
  TraceIf( s_TraceOn, 1, "      test coverage from " << pEdge->FromNode()->Number() 
		 << " to " <<  pEdge->ToNode()->Number() );
  
  if ( ! m_pNumberOneHypothesis->TestAndSetCoverage( pEdge->FromNode()->Number(), 
													 pEdge->ToNode()->Number() ) ) {

	TraceIf( s_TraceOn, 1, "      Coverage collision detected" );
	m_NumCoverageCollisions++;

  } else {

	TraceIf( s_TraceOn, 1, "      No coverage collision" );

	int NewCoverage = m_pNumberOneHypothesis->m_Coverage;
	TraceIf( s_TraceOn, 1, "      the memoryyp has coverage " << NewCoverage );
	TraceIf( s_TraceOn, 2, "      the memoryyp is " << *m_pNumberOneHypothesis );

	// length of the generated translation so far
	//std::cout << "Target " << pEdge->WordInfo()->TargetString()
	// << "Source " << pEdge->WordInfo()->SourceString()
	// << "Label  " << pEdge->Word() << std::endl;

	int AdditionalLength = pEdge->WordInfo()->TargetStringLength();
	if ( AdditionalLength == 0 ) {
	  AdditionalLength = pEdge->WordInfo()->SourceStringLength();
	}
	if ( AdditionalLength == 0 ) {
	  // this has to be an original edge, i.e. just one word
	  AdditionalLength = 1;
	}
	//std::cout << m_pNumberOneHypothesis->m_TargetLength << " " << AdditionalLength << std::endl;
	m_pNumberOneHypothesis->m_TargetLength += AdditionalLength;
	
	// calculate the cost for the memoryypothesis,
	// the different Score-methods manipulate m_pNumberOneHypothesis

	// cost from the translation model
	GetTMScore( pEdge );

	// cost from the language model, returns new LMState
	GetLMScore( pHyp, pEdge );

	// cost for reordering
	GetALScore( pHyp, pEdge );

	// store backpointer information
	// memoryypothesis is expansion of hypothesis pHyp over edge pEdge
	m_pNumberOneHypothesis->m_pEdge = pEdge;
	m_pNumberOneHypothesis->m_pPrevHyp = pHyp;

	StoreHypothesis();

  }

}



void
CDecoderGT::StoreHypothesis()
{
  // store memoryypothesis according to coverage
  // but only if it is the first one of its kind, or has a lower cost
  // if it is new, create a new NumberOneHyp
  // if it is better, copy content, but don't create a memoryyp object
  
  int NewCoverage = m_pNumberOneHypothesis->m_Coverage;
  CDecoderGT_Hyp *pHyp, *pBestHyp;

  HypContainer< CDecoderGT_Hyp *, HypHash, HypEqual >::iterator IterFind;
  IterFind = m_AllHyps[ NewCoverage ].find( m_pNumberOneHypothesis );

  // copy information to NumberOneBestHyp
  *m_pNumberOneBestHyp = *m_pNumberOneHypothesis;

  // for pruning, access to best hyps
  HypContainer< CDecoderGT_Hyp *, HypHashForPruning, HypEqualForPruning >::iterator IterFindBest;
  IterFindBest = m_BestHyps[ NewCoverage ].find( m_pNumberOneHypothesis );


  // first make sure that hyp is in best hyps
  // if it is not there, create it
  // if it is there but we have now a better one, replace it

  if ( IterFindBest == m_BestHyps[ NewCoverage ].end() ) {
	
	// we don't have a best hypothesis so far, store it
	TraceIf( s_TraceOn, 0, "      Prune: Best Hyp is new, store it" );
	pBestHyp = m_pNumberOneBestHyp;
	m_BestHyps[ NewCoverage ].insert( pBestHyp );
	m_pNumberOneBestHyp = new CDecoderGT_Hyp;
	m_pNumberOneBestHyp->m_pDecoderInfo = m_pDecoderInfo;

  } else {

	// so, one of that kind exists already, now we can compare
	// first, would the memoryyp be a better best hyp?
	pBestHyp = *IterFindBest;

	if ( m_pNumberOneBestHyp->m_TotalScore < pBestHyp->m_TotalScore ) {

	  TraceIf( s_TraceOn, 0, "      Prune: Hyp is better than stored one, overwrite "
			   << pBestHyp << " "
			   << pBestHyp->m_Coverage << " " << pBestHyp->m_CoverageStart 
			   << " " << pBestHyp->m_CoveragePattern 
			   << " " << pBestHyp->m_LMState
			   << " " << pBestHyp->m_TargetLength
			   << " " << pBestHyp->m_TotalScore );
	  
	  *pBestHyp = *m_pNumberOneBestHyp;

	} else {

	  TraceIf( s_TraceOn, 0, "      Prune: Hyp is worse than stored one, don't use it" );

	}

  }

  // here we have the best hypothesis as pBestHyp
  // we compare the memoryypthesis against this best hyp

  // is it good enough or should it be pruned
  if ( m_pNumberOneHypothesis->m_TotalScore > m_BeamFactor + pBestHyp->m_TotalScore ) {
	  
	TraceIf( s_TraceOn, 0, "      Hyp is too bad, prune it " );
	m_HypsPruned++;
		
  } else {

	// hyp is good enough to keep it
	// it is a memoryyp
	if ( IterFind == m_AllHyps[ NewCoverage ].end() ) {
	  
	  TraceIf( s_TraceOn, 0, "      Hyp is new" 
			   << m_pNumberOneHypothesis << " "
			   << m_pNumberOneHypothesis->m_Coverage << " " << m_pNumberOneHypothesis->m_CoverageStart 
			   << " " << m_pNumberOneHypothesis->m_CoveragePattern 
			   << " " << m_pNumberOneHypothesis->m_LMState
			   << " " << m_pNumberOneHypothesis->m_TargetLength
			   << " " << m_pNumberOneHypothesis->m_TotalScore );

		// we need to store it and create a memoryyp object
		m_AllHyps[ NewCoverage ].insert( m_pNumberOneHypothesis );
		m_pNumberOneHypothesis = new CDecoderGT_Hyp;
		m_pNumberOneHypothesis->m_pDecoderInfo = m_pDecoderInfo;
		m_NumNewHyps++;

	  } else {
		
		// it exist already, but is it better now
		pHyp = *IterFind;

		if ( m_pNumberOneHypothesis->m_TotalScore < pHyp->m_TotalScore ) {

		  TraceIf( s_TraceOn, 0, "      Hyp is better than stored one, overwrite " 
				   << m_pNumberOneHypothesis << " "
				   << m_pNumberOneHypothesis->m_Coverage << " " << m_pNumberOneHypothesis->m_CoverageStart 
				   << " " << m_pNumberOneHypothesis->m_CoveragePattern 
				   << " " << m_pNumberOneHypothesis->m_LMState
				   << " " << m_pNumberOneHypothesis->m_TargetLength
				   << " " << m_pNumberOneHypothesis->m_TotalScore );

		  *pHyp = *m_pNumberOneHypothesis;
		  m_NumBetterHyps++;
		
		} else {
		  
		  TraceIf( s_TraceOn, 0, "      Hyp is worse than stored one, don't use it" );
		  m_NumWorseHyps++;
		  
		}
	  }
	
	TraceIf( s_TraceOn, 1, "      Number of hyps with coverage " << pHyp->m_Coverage 
			 << " is still " << m_AllHyps[ pHyp->m_Coverage ].size() );
	TraceIf( s_TraceOn, 1, "      Number of hyps with coverage " << NewCoverage 
			 << " is now   " << m_AllHyps[ NewCoverage ].size() );
	
  }
  
}




void
CDecoderGT::ExpandAllHypsSentenceEnd()
{
  // loop over all hyps with full coverage
  // apply sentence end scores (LM, sentence length model)
  // and sort them according to cost

  // clear the final list, in case it still has the hyps from previous sentence
  m_FinalHyps.clear();

  // first we fetch the set of hypotheses which all have full coverage,
  // i.e. source sentence length
  int J = m_pLattice->EndNode()->Number();
  HypContainer< CDecoderGT_Hyp *, HypHash, HypEqual > &Hyps = m_AllHyps[ J ];
  TraceIf( s_TraceOn, 1, "\n  Number of hyps with full coverage is " << Hyps.size() );

  // and we need an iterater to access the hyps in this set
  HypContainer< CDecoderGT_Hyp *, HypHash, HypEqual >::iterator IterHyps;

  CDecoderGT_Hyp *pHyp;

  // loop over hyps
  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {

	// create a copy of the hyp and point back to that hyp
	// we don't go over an edge at this point
	pHyp = new CDecoderGT_Hyp( **IterHyps );
	pHyp->m_pDecoderInfo = m_pDecoderInfo;
	pHyp->m_pPrevHyp = *IterHyps;
	pHyp->m_pEdge = NULL;


	// cost for sentence end from the language model
	double LMScore = -m_pLMSentenceState->LogProbEnd( pHyp->m_LMState );
	pHyp->m_LMScore += m_LMScale * LMScore;
	pHyp->m_TotalScore += m_LMScale * LMScore;
	TraceIf( s_TraceOn, 1, "\n  LM cost for sentence end " << LMScore );


	// cost from sentence length model
	// exp( -l(cJ - I)^2 )
	// two factors: c is general stretching factor: use 1.2 for now
        //              f is flatness factor: use 0.25 for now

	float SLScore;

	// SLScore is now a simple negative count of the length
	int I = pHyp->m_TargetLength;
	SLScore = -1.0*I;

	pHyp->m_SLScore = m_SLScale * SLScore;
	pHyp->m_TotalScore += m_SLScale * SLScore;

	// now insert into final list, sorting is done by the container
	TraceIf( s_TraceOn, 1, "\n  insert hyp into final list" 
			 << pHyp << " "
			 << pHyp->m_Coverage << " " << pHyp->m_CoverageStart << " " 
			 << pHyp->m_CoveragePattern );

	m_FinalHyps.insert(pair<double, CDecoderGT_Hyp *>( pHyp->m_TotalScore , pHyp ) );

  }

  TraceIf( s_TraceOn, 1, "Number of Hyps in Final List is: " << m_FinalHyps.size() );

}



void
CDecoderGT::HistogramPruning( HypContainer< CDecoderGT_Hyp *, HypHash, HypEqual > &Hyps,
							  HypContainer< CDecoderGT_Hyp *, HypHashForPruning, HypEqualForPruning > &BestHyps )
{

  bool s_TraceOn = false;

  //Trace( 0, "Histogram Pruning: " << Hyps.size() << " " << BestHyps.size() );

  // run over all hyps
  // put hyp into bucket according to BucketNumber = (pHyp->Score - pBestHyp->Score) * 100 / BeamFactor
  // i.e. do not select according to absolute score but according to difference to best hyp
  // with the same features

  CDecoderGT_Hyp *pHyp, *pBestHyp;

  HypContainer< CDecoderGT_Hyp *, HypHash, HypEqual >::iterator IterHyps;
  HypContainer< CDecoderGT_Hyp *, HypHashForPruning, HypEqualForPruning >::iterator IterBestHyps;


  // first, find the best score from the best hyps
  float BestScore = 1.0e10;
  for ( IterBestHyps = BestHyps.begin(); IterBestHyps != BestHyps.end(); ++IterBestHyps ) {

	pBestHyp = *IterBestHyps;
	if ( pBestHyp->m_TotalScore < BestScore ) {
	  BestScore = pBestHyp->m_TotalScore;
	}
  }

  //std::cout << "Best Score = " << BestScore << std::endl;

  /*
  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {

	pHyp = *IterHyps;
	//std::cout << "Hyp is " << pHyp << std::endl;
	TraceIf( s_TraceOn, 0, "\n  hyp" 
			 << pHyp << " "
			 << pHyp->m_Coverage << " " << pHyp->m_CoverageStart 
			 << " " << pHyp->m_CoveragePattern
			 << " " << pHyp->m_LMState
			 << " " << pHyp->m_TargetLength 
			 << " " << pHyp->m_TotalScore );


	// select corresponding best hyp
	IterBestHyps = BestHyps.find( pHyp );
	pBestHyp = *IterBestHyps;
	//std::cout << "Best Hyp is " << pBestHyp << std::endl;
	TraceIf( s_TraceOn, 0, "\n  best hyp " 
			 << pBestHyp << " "
			 << pBestHyp->m_Coverage << " " << pBestHyp->m_CoverageStart 
			 << " " << pBestHyp->m_CoveragePattern
			 << " " << pBestHyp->m_LMState
			 << " " << pBestHyp->m_TargetLength 
			 << " " << pBestHyp->m_TotalScore );
	
	// now the calculation
	int BucketNumber = int( ( pHyp->m_TotalScore - pBestHyp->m_TotalScore ) * 100 / m_BeamFactor );
	TraceIf( s_TraceOn, 0, "\n  BucketNumber = " << BucketNumber );
	if ( BucketNumber < 100 ) {
	  m_PruningBuckets.GetRow( BucketNumber ).Append( pHyp );
	}
  }
  */


  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {

	pHyp = *IterHyps;
	pHyp->m_Expand = False;

	//std::cout << "Hyp is " << pHyp << std::endl;
	TraceIf( s_TraceOn, 0, "\n  hyp" 
			 << pHyp << " "
			 << pHyp->m_Coverage << " " << pHyp->m_CoverageStart 
			 << " " << pHyp->m_CoveragePattern
			 << " " << pHyp->m_LMState
			 << " " << pHyp->m_TargetLength 
			 << " " << pHyp->m_TotalScore );


	// now the calculation
	int BucketNumber = int( ( pHyp->m_TotalScore - BestScore ) * 100 / m_BeamFactor );
	TraceIf( s_TraceOn, 0, "\n  BucketNumber = " << BucketNumber );
	if ( BucketNumber < 100 ) {
	  m_PruningBuckets.GetRow( BucketNumber ).Append( pHyp );
	}
  }


  // now collect enough buckets to get the required number of hyps
  int Sum = 0;
  int i = 0;
  while ( (Sum < m_MaxHyps) && (i < 100) ) {
	CArray< CDecoderGT_Hyp * > Bucket = m_PruningBuckets.GetRow( i );
	Sum += Bucket.Size();
	//std::cout << i << " " << Bucket.Size() << " " << Sum << std::endl;
	
	for ( int j = Bucket.MinIndex(); j <= Bucket.MaxIndex(); j++ ) {
	  Bucket[ j ]->m_Expand = True;
	}

	i++;
  }

  //std::cout << "Number of Hyp to expand after Histogram Pruning: " << Sum << std::endl;

  m_PruningBuckets.Clear();

}




CString
CDecoderGT::TraceBack( CDecoderGT_Hyp *pHyp )
{
  CDecoderGT::s_TraceOn = true;

  // ARV output
  //std::cout << "Trace Back" << std::endl;
  //std::cout << *pHyp << std::endl;
  //std::cout << pHyp->m_pPrevHyp << std::endl;

  int Start = m_pLattice->StartNode()->Number();
  int End   = m_pLattice->EndNode()->Number();
  int L = End - Start;

  if (s_TraceOn){
    
    //ARV output
    
    std::cout << std::endl
	      << "Total Score " << pHyp->m_TotalScore
	      << "  TMScore " << pHyp->m_TMScore
	      << "  LMScore " << pHyp->m_LMScore
	      << "  ALScore " << pHyp->m_ALScore
	      << "  SLScore " << pHyp->m_SLScore
	      << std::endl;
    
  std::cout << "Per Word    " << pHyp->m_TotalScore / L
	    << "  TMScore " << pHyp->m_TMScore / (L*m_TMScale)
	    << "  LMScore " << pHyp->m_LMScore / (L*m_LMScale)
	    << "  ALScore " << pHyp->m_ALScore / (L*m_ALScale)
	    << "  SLScore " << pHyp->m_SLScore / (m_SLScale)
	   << std::endl;
  /*
    std::cout << "Reorderings " << pHyp->m_NumReorderings / 2 
  << std::endl; */
  }

  CArray< CLatticeEdgeTrans * > AllEdges;
  CArray< CLatticeEdge * > AllEdges2;

  while ( pHyp->m_pPrevHyp ) {
	//std::cout << "Hyp " << pHyp << " PrevHyp " << pHyp->m_pPrevHyp 
	//<< " Edge " << pHyp->m_pEdge << std::endl;
	if ( !pHyp->m_pEdge ) {
	  //std::cout << "no edge - </s>" << std::endl;
	} else {
	  //pHyp->m_pEdge->Display();
	  AllEdges.Append( pHyp->m_pEdge );
	}

	//std::cout << *pHyp->m_pPrevHyp << std::endl;
	pHyp = pHyp->m_pPrevHyp;
  }


  // first, display the path that has been taken
  // invert sequence of edges in list
  for ( int i = AllEdges.MaxIndex(); i >= AllEdges.MinIndex(); i-- ) {
	AllEdges2.Append( AllEdges[ i ] );
  }
  //ARV output
  //std::cout << "PartialPath: ";
  s_TraceOn = false;
  if (s_TraceOn){
    m_pLattice->DisplayPath( AllEdges2 );
  }

  // now display the clean translation
  CString Translation;
  CWordInfo *pWordInfo;
  CLatticeEdgeTrans *pEdge;

  for ( int i = AllEdges.MaxIndex(); i >= AllEdges.MinIndex(); i-- ) {
	pEdge = AllEdges[ i ];
	pWordInfo =  pEdge->WordInfo();

	// extract sequence of words from the WordInfo structure
	if ( pWordInfo->HasTargetString() ) {
	  Translation += pWordInfo->TargetString();
	} else if ( pWordInfo->HasSourceString() ) {
	  Translation += pWordInfo->SourceString();
	} else {
	  Translation += pEdge->Word();
	}

	Translation += " ";
  }

  s_TraceOn = true;
  if (s_TraceOn){
    std::cout << "Translation is: " << Translation << std::endl;
    //std::cout << "------------" << std::endl;
  }
  //ARV removes the unncessary blanks at the end of the translations
  Translation.TrimLeadingBlanks();
  Translation.TrimTrailingBlanks();
  CDecoderGT::s_TraceOn = false;
  return Translation;
}


void
CDecoderGT::DisplayAllHypsWithCoverage( int Coverage, int N )
{

  CDecoderGT_Hyp *pHyp;

  // first we fetch the set of hypotheses which all have coverage Coverage
  // and we need an iterater to access the hyps in this set
  HypContainer< CDecoderGT_Hyp *, HypHash, HypEqual > &Hyps = m_AllHyps[ Coverage ];

  // and we need an iterater to access the hyps in this set
  HypContainer< CDecoderGT_Hyp *, HypHash, HypEqual >::iterator IterHyps;


  TraceIf( s_TraceOn, 0, "\n\n  Number of hyps with coverage " 
		   << Coverage << " is " << Hyps.size() );
  //TraceIf( s_TraceOn, 0, "Address of Hyps is: " << &Hyps );

  // display only top N hyps
  int n = 0;
  for ( IterHyps = Hyps.begin(); IterHyps != Hyps.end(); ++IterHyps ) {
	if ( n < N ) {
	  
	  //ARV output
	  pHyp = *IterHyps;
	  /*
	  std::cout << pHyp->m_TotalScore << " " << pHyp->m_Coverage << " " 
		   << pHyp->m_CoverageStart << " " << pHyp->m_CoveragePattern << std::endl;
	  */
	  n++;
	}
  }

}



void
CDecoderGT::GetTMScore( CLatticeEdgeTrans *pEdge )
{
  //  CString Word;
  double TMScore;

  // - translation score (i.e. accumulated transducer score
  // for original edges, use penalty
  /*
  Word = pEdge->Word();
  if ( Word[ 0 ] == '@' ) {
	TMScore = pEdge->WordInfo()->Score();
	TraceIf( s_TraceOn, 0, "      Translation Score for labeled edge: " << TMScore );
  } else {
	TMScore = 10;
	TraceIf( s_TraceOn, 0, "      Translation Score for source language edge: " << TMScore );
  }
  return TMScore;
  */

  TMScore = pEdge->WordInfo()->Score();
  
  // ARV multiply the scaling factor back in 
  m_pNumberOneHypothesis->m_TMScore += m_TMScale * TMScore;
  m_pNumberOneHypothesis->m_TotalScore += m_TMScale * TMScore;

}



void
CDecoderGT::GetLMScore( CDecoderGT_Hyp *pHyp, CLatticeEdgeTrans *pEdge )
{

  CLMState LMState = pHyp->m_LMState;
  CLMState NewLMState;
  double LMScore = 0.0;


  // strings have alreay been converted into indices according to the LM vocabulary
  // now get the LM score by looping over all word indices
  LMScore = 0.0;
  CNumericalArray< long > & WordIndices2 = pEdge->LMWordIndices();
  LMState = pHyp->m_LMState;

  for ( int x = WordIndices2.MinIndex(); x <= WordIndices2.MaxIndex(); x++ ) {
	LMScore -= m_pLMSentenceState->LogProb( LMState, WordIndices2[ x ] );
	TraceIf( s_TraceOn, 3, "Version 2: LM-Score for " << WordIndices2[ x ] << ": " 
		   << -m_pLMSentenceState->LogProb( LMState, WordIndices2[ x ] ) 
		   << " total LM-Score = " << LMScore );
	
	NewLMState = m_pLMSentenceState->AppendWord( LMState, WordIndices2[ x ] );
	TraceIf( s_TraceOn, 3, "Version 2: New LM-State: " << NewLMState );
	LMState = NewLMState;
  }

  m_pNumberOneHypothesis->m_LMScore += m_LMScale * LMScore;
  m_pNumberOneHypothesis->m_TotalScore += m_LMScale * LMScore;
  m_pNumberOneHypothesis->m_LMState = NewLMState;

}



void
CDecoderGT::GetALScore( CDecoderGT_Hyp *pHyp, CLatticeEdgeTrans *pEdge )
{
  // small extra cost for reordering
  // when and how do we know that we are reordering?
  // when last edge and current edge are not sequential
  // we ended in pHyp->m_pEdge->ToNode() and start from pEdge->FromNode()
  // if these two nodes are the same, then monoton decoding
  // otherwise a jump

  // currently, each jumps incures the same small cost
  // a proper position alignment model should be used.

  double ALScore;
  int NumberOfPrevNode;

  // the initial hypothesis has no edge, but we know that the node 
  // is the StartNode of the lattice
  /*
  if ( !pHyp->m_pEdge ) {
	NumberOfPrevNode = 0;
  } else {
	NumberOfPrevNode = pHyp->m_pEdge->ToNode()->Number();
  }

  
  if ( pEdge->FromNode()->Number() != NumberOfPrevNode ) {
	m_pNumberOneHypothesis->m_NumReorderings++;
	if ( m_pNumberOneHypothesis->m_NumReorderings > m_pNumberOneHypothesis->s_MaxNumReorderings ) {
	  ALScore = 100;
	} else {
	ALScore = 1.0;
	}
  } else {
	ALScore = 0.0;
  }
  */


  // here we have a simple jump model:
  // Forward: Prev->ToNode - 1 -> Next->FromNode
  // Backward: Prev->ToNode -> Next->FromNode + 1
  // pHyp has pointer to last traversed edge, pEdge is newly travesed edge

  float Jump;
  int JumpStart, JumpEnd;

  // we start from
  if ( !pHyp->m_pEdge ) {
	JumpStart = 0;
  } else {
	JumpStart = pHyp->m_pEdge->ToNode()->Number();
  }
  
  // we jump to
  JumpEnd = pEdge->FromNode()->Number() + 1;

  // the jump with is
  Jump = JumpEnd - JumpStart;
  ALScore = 0.25 * ( Jump - 1.0 ) * ( Jump - 1.0 );

  if ( pEdge->FromNode()->Number() != JumpStart ) {
	m_pNumberOneHypothesis->m_NumReorderings++;
	if ( m_pNumberOneHypothesis->m_NumReorderings > m_pNumberOneHypothesis->m_pDecoderInfo->m_MaxNumReorderings ) {
	  ALScore = 100;
	}
  }

  //std::cout << "Jump is " << Jump << " ALScore " << ALScore << std::endl;

  m_pNumberOneHypothesis->m_ALScore += m_ALScale * ALScore;
  m_pNumberOneHypothesis->m_TotalScore += m_ALScale * ALScore;

}



void
CDecoderGT::WriteLattice()
{
  CArray< CLatticeEdge * > Edges = m_pLattice->Edges();
  CLatticeEdgeTrans *pEdge;

  std::cout << "BEGIN_LATTICE" << std::endl;
  std::cout << "(" << std::endl;
  for ( int i = Edges.MinIndex(); i <= Edges.MaxIndex(); i++ ) {
	pEdge = (CLatticeEdgeTrans *)Edges[ i ];
	//	pEdge->Display();
	//std::cout << std::endl;
	if ( (pEdge->Word())[0] == '@' ) {
	  double Score = exp( -(pEdge->WordInfo()->Score() ) );
	  std::cout << '(' << pEdge->Startframe() << " " << (pEdge->Endframe() - 1) << " ";
	  std::cout << "|||" << pEdge->WordInfo()->TargetString() << "|||" << " ";
	  std::cout << " " << Score << " ";
	  std::cout << "|||"<< pEdge->WordInfo()->SourceString() << "|||";
	  std::cout << ')';
	  std::cout << std::endl;
	}
  }
  std::cout << ")" << std::endl;
  std::cout << "END_LATTICE" << std::endl;

}

