#ifndef __FutureEstimator__
#define __FutureEstimator__

#include <Array.hh>
#include <Vocabulary.hh>
#include <Lattice.hh>
#include <LatticeEdgeMS.hh>
#include <SriLM.hh>

class CFutureEstimator{
public:

  //initialize the estimator with information used to calculate src future costs
  CFutureEstimator(CNumericalArray<double>& scalingFactors, CNumericalArray<double>& srcFeatures, SriLM* pLM);
  
  //set up arrays to hold cost structure, then use the lattice to fill in the initial entries, by iterating outgoing
  //edges from each node in the lattice
  void Init(CSentence* pSentence, CLattice* pLattice);
  
  //initialize the src word future cost estimate so that it can be used as a fixed number from this point on
  void SetSrcModels(CNumericalArray<double>& scalingFactors, CNumericalArray<double>& srcFeatures);

  //fill the min cost structure by parsing from the bottom up, similair to CYK+ execept that all the initialization is done
  //and we only need to track the min cost
  void ParseMinCost();


  //returns the cell i,j from the m_minCostStructure
  double GetFutureCostEstimate(int i, int j);

  //checks if the cost provided is lower than the one currently at i,j, if it is, then the cost is replaced
  void CheckMin(double cost, int i, int j);

  //does the dot product between the pEdge->Scores() and m_scalingFactors
  double CostFromEdge(CLatticeEdgeMS* pEdge);

  //print out the min cost cells
  void Display();

protected:
  CSentence* m_pSourceSentence;

  //each cell, i, j refers to translating i words from position index j
  CArray2D<double> m_minCostStructure;

  //scaling factors used in this estimation
  CNumericalArray<double>  m_scalingFactors;

  //default dummy model values for src words
  CNumericalArray<double>  m_srcFeatures;

  //default src word cost, set in Init()
  double m_srcWordCost;

  //language model used to generate LM score
  SriLM* m_pLM;

};
#endif
