#define NO_TRACE 1
//#undef NO_TRACE
#include <Trace.hh>

#include <NBestListRescorer.hh>


CNBestListRescorer::CNBestListRescorer( )
{
}


CNBestListRescorer::~CNBestListRescorer( )
{
}


double 
CNBestListRescorer::Prob( const CSentence &SourceSentence, const CSentence &TargetSentence )
{
}

double 
CNBestListRescorer::LogProb( const CSentence &SourceSentence, const CSentence &TargetSentence )
{
  Trace( 0, "CNBestListRescorer::LogProb" );
}

double 
CNBestListRescorer::NegLogProb( const CSentence &SourceSentence, const CSentence &TargetSentence )
{
  return -LogProb( SourceSentence, TargetSentence );
}



// ---------------------------------------
// CNBestListRescorerTM
// ---------------------------------------

CNBestListRescorerTM::CNBestListRescorerTM( )
{
  m_pLexicon = NULL;
}

CNBestListRescorerTM::~CNBestListRescorerTM( )
{
}


bool
CNBestListRescorerTM::Init( CLexicon *pLexicon )
{
  m_pLexicon = pLexicon;
}



// ---------------------------------------
// CNBestListRescorerTM_IBM1
// ---------------------------------------

CNBestListRescorerTM_IBM1::CNBestListRescorerTM_IBM1( )
{
}

CNBestListRescorerTM_IBM1::~CNBestListRescorerTM_IBM1( )
{
}


double 
CNBestListRescorerTM_IBM1::Prob( const CSentence &SourceSentence, const CSentence &TargetSentence )
{
  return exp( LogProb( SourceSentence, TargetSentence ) );
}



double 
CNBestListRescorerTM_IBM1::LogProb( const CSentence &SourceSentence, const CSentence &TargetSentence )
{
  // NOTE: multiply over the F's (sourcewords) 
  //       and sum over the E's (targetwords) inside each product term 

  Trace( 1, "CNBestListRescorerTM_IBM1::LogProb()" );
  //SourceSentence.Display( m_pLexicon->SourceVocabulary() );
  //std::cout << std::endl;
  //TargetSentence.Display( m_pLexicon->TargetVocabulary() );
  //std::cout << std::endl;

  assert( m_pLexicon );

  float LogP =  m_pLexicon->LogPr( &SourceSentence, &TargetSentence );

  Trace( 1, "Lexicon LogProb = " << LogP );

  // now the position alignment prob
  //  double AlignLogP = log( 1.0 / TargetSentence.Length() ) * SourceSentence.Length();
  //std::cout << "AlignProb = " << (1.0 / TargetSentence.Length() ) 
  //          << " " << log ( 1.0 / TargetSentence.Length() ) << std::endl;
  //Trace( 0, "Alignment LogProb = " << AlignLogP );
  //return ( LogP + AlignLogP );

  return LogP;
}



// ---------------------------------------
// CNBestListRescorerTM_HMM
// ---------------------------------------

CNBestListRescorerTM_HMM::CNBestListRescorerTM_HMM( )
{
}



CNBestListRescorerTM_HMM::~CNBestListRescorerTM_HMM( )
{
}



double
CNBestListRescorerTM_HMM::Prob( const CSentence & SourceSentence,
                       const CSentence & TargetSentence )
{
  return exp( LogProb( SourceSentence, TargetSentence ) );
}



//double
//CNBestListRescorerTM_HMM::LogProb( const CSentence & SourceSentence,
//                          const CSentence & TargetSentence )
//{  
//  return 0.0;
//}



double
CNBestListRescorerTM_HMM::LogProb( const CSentence &SourceSentence, const CSentence &TargetSentence )
{
  CWordIndex We, Wf; 		  // index of word in vocabulary
  TPosition PosWe, PosWf;         // position of word in text
  int i, iprev, j;                  // position of word within the sentences
  double TPenalty;                // -ln t(f|e)
  //double APenalty;                // -ln A(j|i)
  double JPenalty;                // -ln Jump(distance)
  double NewPenalty;              // used in loop over previous column
  double MinPenalty;              // the minimal penalty in loop over previous column
  CMatrix< double > Q;


  // first, calculate the JPenalties; i.e. renormalize and store them 
  // for all positions and all possible jumps from that position
  // have to know the sentence length

  //cout << "CalculatePathes" << endl;
  //cout.flush();

  int I = SourceSentence.Length();
  int J = TargetSentence.Length();

  Trace( 3,  "I: " << I << " J: " << J );
  // if one sentence is empty -> nothing to do
  if ( I == 0 || J == 0 )
    return -9999999.9;

  // init scores
  for ( j = 0; j <= J + 1; j++ ) {
    for ( i = 0; i <= I + 1; i++ ) {
      Q( j, i ) = 1.0e+10;
    }
  }

  //m_AlignProb.PreparePr( m_J, m_I );

  // take first f word, loop over words in e sentence
  // and init Penalty with -ln t(f, e)

  j = SourceSentence.MinIndex();

  Wf = SourceSentence.GetAt( j );

  Trace( 3, "Calculating prob for source word " << SourceSentence.Word( j )
          <<" at index " << Wf );

  for( int i = TargetSentence.MinIndex(); i <= TargetSentence.MaxIndex(); i++){

    We = TargetSentence.GetAt( i );

    TPenalty = -m_pLexicon->LogPr( Wf, We );

    // jump from (0, 0) to (j, i): Current = i, Previous = 0
    JPenalty = -AlignLogPr( i, 0 );

    Q( j, i ) = TPenalty + JPenalty;

    Trace0( 5, "TPenalty " << TPenalty );
    Trace( 5, "   JPenalty " << JPenalty );

  }

  for( int j = SourceSentence.MinIndex() + 1; j <= SourceSentence.MaxIndex(); j++ ){

    for( int i = TargetSentence.MinIndex(); i <= TargetSentence.MaxIndex(); i++){

      TPenalty = -m_pLexicon->LogPr( Wf, We ); //log(m_pT->GetAt(We,Wf));
      Trace0( 5, "TPenalty " << TPenalty );

      // loop over previous column and calculate the minimum of penalty 
      for( iprev = TargetSentence.MinIndex(); i <= TargetSentence.MaxIndex(); i++){

		MinPenalty = Q( j, i );
		Trace( 2, i << " MinPenalty = " << MinPenalty );

        // jump from (jprev, iprev) to (j, i): Current = i, Previous = 0
        JPenalty = -AlignLogPr( i, iprev );

		NewPenalty = Q( j - 1, iprev ) + TPenalty + JPenalty;

		if ( NewPenalty < Q( j, i ) ) {
		  Q( j, i ) = NewPenalty;
		  Trace( 2, i << " NewMinPenalty = " << MinPenalty );
		}
	
		i++;
      }

    } // loop over skipped fs

    // column is done; increment f-index j
    j++;
  }

  // jump to (EOS, EOS)
  j = J + 1;
  i = I + 1;

  //TPenalty = -m_pLexicon->LogPr( Wf, We ); //log(m_pT->GetAt(We,Wf));
  // Trace0( 5, "TPenalty " << TPenalty );
      
  // loop over previous column and calculate the minimum of penalty 
  // at position (j - 1, pre) plus penalty due to jump in alignment
  
  NewPenalty = 0.0;

  //cout << j << " " << i << " " << JPenalty << endl << flush;
  
  //JPenalty = -AlignLogPr( I + 1, I );
  //MinPenalty = Q( j-1, iprev ) + JPenalty;
    
  for ( iprev = I - 1; iprev >= 1; iprev-- ) {
    // only positive jump
	JPenalty = -AlignLogPr( i, iprev );
    NewPenalty = Q( j-1, iprev ) + JPenalty;

    if ( NewPenalty < MinPenalty ) {
      MinPenalty = NewPenalty;
      
      //cout << "penalties: " << TPenalty << " " << JPenalty << " " 
      //<< NewPenalty << " " << MinPenalty << " Pre " << pre << "\n";
    }
  }
  
  // no TPenalty
  Q( J + 1, I + 1 ) = MinPenalty;
  //m_Penalty[j][i] = TPenalty + MinPenalty;
    
}



double
CNBestListRescorerTM_HMM::AlignLogPr( int i, int iprev )
{
  int Jump = i - iprev;
  return 0.25 * ( Jump - 1.0 ) * ( Jump - 1.0 );
}

