//Nguyen, July 15, 2008: 
//Dependency Tree tokens
//
#include "tokens.hh"

using std::istream;
using std::iostream;
using std::fstream;
using std::vector;
using std::string;
using std::ios;

//return tokes value
string getName(const tokens& x) { return x.name; } 

//return position of the token in string
int getPost(const tokens& x) { return x.post; } 

//return parent's position of the token in string
int getParentPost(const tokens& x) { return x.parentpost; } 

//return dependency type
string getDType(const tokens& x ) { return x.dtype; } 

//return token POS tag
string getPOS(const tokens& x) { return x.pos; } 


//Nguyen, Aug 5, 2008: Dependency trees

//string 2 int
int str2int (const std::string& str) {
	std::stringstream ss(str);
	int n;
	ss >> n;
	return n;
}
//int 2 string
string int2str (int n) {
	std::stringstream ss;
	ss << n;
	return ss.str();
}
//return names of vector: a string sequence of tokens 
string getNameFromVector (const std::vector<tokens>& v){
	string s;
	for(vector<tokens>::size_type j = 0; j < v.size(); ++j){
		s += v[j].name + " ";
	}
	return s;
}

//return position of vector: a string of positions of tokens 
string getPostFromVector (const std::vector<tokens>& v){
	string s;
	for(vector<tokens>::size_type j = 0; j < v.size(); ++j){
		s += int2str( v[j].post ) + " ";
	}
	return s;
}


// split line by space charactes; space = " ", "\t", ...
std::vector<string> split(const std::string& s){	
	std::vector <std::string> ret;
	typedef string::size_type string_size;
	string_size i = 0;
	
	while (i != s.size()){
		//ignore leading delims 
		while (i != s.size() && isspace(s[i]))
			++i;
			
		//find end of next segment
		string_size j=i;
		while (j != s.size() && !isspace(s[j]))			
			++j;			
		
		if ( i != j ){
			ret.push_back(s.substr(i, j-i));
			i = j;
		}
	}
	
	return ret;
}

//read CoNLL dependency data from file into a 2D vector of tokens
void readCoNLLformat( const char* filename, std::vector < std::vector<tokens> >& d){
	std::vector<tokens> wv;	
	
	std::ifstream in(filename);
	
	if (in){
		string s;		
		while (getline(in, s)){			
			std::vector <string> v = split (s);						
			if (v.size()>1){				
				tokens w;
				
				//CoNLL dependency format								
				w.name = v[1];
				w.post = str2int(v[0]) - 1; //index from 0
				w.pos  = v[4]; 
				w.dtype= v[7];
				w.parentpost = str2int(v[6]) - 1;//index from 0
								
				wv.push_back(w);													
			} else {
				d.push_back(wv);				
				wv.clear();	
			}			
		}			
	} else {
		std::cerr << "Cannot open file " << filename << std::endl;
	}	
}

// return FALSE/0 if n is not a node of tree T
//        TRUE/1 otherwise
bool insubtree (const tokens& n, const tokens& t, const vector<tokens>& v){	
	
	if (n.post == t.post) { return true; }
	else{
		std::queue<tokens> nodes;
		
		//initialize queue
		for(vector<tokens>::size_type k = 0; k < v.size(); ++k){
			if (v[k].parentpost == t.post){
				nodes.push(v[k]);	
			}
		}
		
		//Breadth-first search
		if ( !nodes.empty() ){			
			while ( !nodes.empty() ){				
				tokens c = nodes.front();				
				if (n.post == c.post ) { return true; }
				else {
					nodes.pop();					
					for(vector<tokens>::size_type k = 0; k < v.size(); ++k){
						if (v[k].parentpost == c.post)
							nodes.push(v[k]);
					}					
				}				
			}
			return false;
		}
		else{ return false; }		
	}
}


//end of Nguyen
