#include <string.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
// _DiscriminativeLM.cpp: implementation of the C_DiscriminativeLM class.
//
// $Rev: 3925 $
// $LastChangedDate: 2007-07-19 12:15:05 -0400 (Thu, 19 Jul 2007) $
//
// DiscriminativeLM class is to load the trained discriminative lm infor
// and provide the decoder the n-gram discriminative log prob
//
//////////////////////////////////////////////////////////////////////

#include "_DiscriminativeLM.h"
#include "_DiscriminativeLMSentenceState.h"

#include <fstream>
#include <iostream>

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

C_DiscriminativeLM::C_DiscriminativeLM(const char * cfgFileName)
{
	fstream cfgFile;
	cfgFile.open(cfgFileName,ios::in);

	if(!cfgFile){
		fprintf(stderr,"Configuration file does not exist! quit!!\n");
		exit(0);
	}

	//-----------------------------------------------------------------------------
	//reading parameters
	char paraName[1024];
	
	char ngramDiscrFileName[1024];
	ngramDiscrFileName[0]=0;

	char idVocFileName[1024];
	idVocFileName[0]=0;

	this->maxN = 0;

	while(!cfgFile.eof()){
		cfgFile>>paraName;

		if(strcmp(paraName,"NGRAM")==0){
			cfgFile>>ngramDiscrFileName;
		}		
		else if(strcmp(paraName,"VOC")==0){
			cfgFile>>idVocFileName;
		}
		else if(strcmp(paraName,"N")==0){
			cfgFile>>this->maxN;
		}

		paraName[0]=0;
	}

	if(strlen(idVocFileName)==0){
		cerr<<"Need to specify vocabulary file for Discr. LM\n";
		cerr<<"Use keyword: VOC in cfg file.\n";
		exit(-1);
	}

	if(strlen(ngramDiscrFileName)==0){
		cerr<<"Need to specify the filename where n-gram discriminative cost is stored.\n";
		cerr<<"Use keyword: NGRAM in cfg file.\n";
		exit(-1);
	}

	if(this->maxN==0){
		cerr<<"Need to specify the order of n for n-gram discriminative language model.\n";
		cerr<<"Use keyword: N in cfg file.\n";
		exit(-1);
	}

	this->idVoc = new C_IDVocabulary(idVocFileName);

	this->sentStartVocId = this->idVoc->returnId(C_String("<s>"));
	this->sentEndVocId = this->idVoc->returnId(C_String("</s>"));

	//load the n-gram discr. cost	
	this->loadNgramDiscrCost(ngramDiscrFileName);
}

C_DiscriminativeLM::~C_DiscriminativeLM()
{

}

void C_DiscriminativeLM::loadNgramDiscrCost(char *filename)
{
	fstream InputFile;
	InputFile.open(filename);

	if(! InputFile){
		cerr<<"Can not open filename for read!\n";
		exit(-1);
	}

	cerr<<"Loading Discriminative Language Model ...\n";

	char aLine[4096];
	char * aToken;
	char delimit[] = " \t\r\n";

	int ngramsReadIn = 0;
	while(! InputFile.eof()){
		InputFile.getline(aLine, MAX_TOKEN_LEN, '\n');
		
		if(strlen(aLine)>0){	//a meaningful word, esp for the last line during reading file
		  ngramsReadIn++;
		  if(ngramsReadIn==10000){
		    cerr<<".";
		    ngramsReadIn=0;
		  }
			vector<C_String> tokensInLine;

			aToken = strtok(aLine, delimit);				
			while( aToken != NULL ) {	
				tokensInLine.push_back(C_String(aToken));
				aToken = strtok( NULL, delimit);
			}

			int sizeOfN=tokensInLine.size()-1;	//assume that there is only one cost for each n-gram type

			vector<IndexType> ngramAsVocId;
			for(int i=0;i<sizeOfN;i++){
				IndexType vocId = this->idVoc->returnId(tokensInLine[i]);
				ngramAsVocId.push_back(vocId);
			}

			C_Ngram ngram(ngramAsVocId);

			float cost = atof(tokensInLine[sizeOfN].toString());

			this->ngramCost.insert(make_pair(ngram, cost));
		}

		//clear content in aLine
		aLine[0]=0;
	}

	cerr<<"\nDiscriminative LM loaded.\n";
}

//if n-gram exist assign cost with the associated cost value and return true
//else return false, set cost to 0
bool C_DiscriminativeLM::returnCost(C_Ngram &ngram, float &cost)
{
	map<C_Ngram, float>::iterator iterNgramCost;

	iterNgramCost = this->ngramCost.find(ngram);

	if(iterNgramCost==this->ngramCost.end()){
		//does not exist
		cost = 0;
		return false;
	}

	//exist
	cost = iterNgramCost->second;
	return true;
}

LMWordIndex C_DiscriminativeLM::returnSentStartVocId()
{
	return this->sentStartVocId;
}

LMWordIndex C_DiscriminativeLM::returnSentEndVocId()
{
	return this->sentEndVocId;
}

CLMSentenceState * C_DiscriminativeLM::NewSentence() const
{
	return new C_DiscriminativeLMSentenceState(this);
}

LMWordIndex C_DiscriminativeLM::GetVocabularyIndex(string word)
{
	return this->idVoc->returnId(C_String((char *) (word.c_str()) ));
}

const vector< string > & C_DiscriminativeLM::Vocabulary() const
{
  //assert( 0 );
  vector< string > EmptyVector;
  return EmptyVector;
}

LMWordIndex C_DiscriminativeLM::VocabularySize() const
{
	return this->idVoc->getSize();
}

int C_DiscriminativeLM::returnMaxN()
{
	return this->maxN;
}
