// _DiscriminativeLMSentenceState.cpp: implementation of the C_DiscriminativeLMSentenceState class.
//
//
// $Rev: 3922 $
// $LastChangedData$
//
//////////////////////////////////////////////////////////////////////

#include "_DiscriminativeLMSentenceState.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

C_DiscriminativeLMSentenceState::C_DiscriminativeLMSentenceState(const C_DiscriminativeLM * discrimLm)
{
	this->discrimLm = (C_DiscriminativeLM *) discrimLm;

	this->sentStartVocId = this->discrimLm->returnSentStartVocId();
	this->sentEndVocId = this->discrimLm->returnSentEndVocId();
	
	this->maxN = this->discrimLm->returnMaxN();
}

C_DiscriminativeLMSentenceState::~C_DiscriminativeLMSentenceState()
{

}

//start a new sentence now, clear up the sentence LM state
CLMState C_DiscriminativeLMSentenceState::BeginOfSentenceState()
{
	this->resetLmStates();
	this->initialLmState();	

	return 0;
}

void C_DiscriminativeLMSentenceState::initialLmState()
{
	//add sentence start
	C_Ngram sentStartNgram;
	sentStartNgram.appendOneWord(this->sentStartVocId);

	this->allLMStates.push_back(sentStartNgram);
	this->lmStateInfo2Id.insert(make_pair(sentStartNgram, 0));
}

void C_DiscriminativeLMSentenceState::resetLmStates()
{
	this->buffer.clear();
	this->allLMStates.clear();
	this->lmStateInfo2Id.clear();
}

//always keep up to maxN-1 words as history
double C_DiscriminativeLMSentenceState::LogProb(CLMState lmstate, const LMWordIndex nextWord, CLMState & nextstate)
{
	//first check if lmstate+nextword has been seen in the buffer already
	map< pair<CLMState, IndexType>, S_BufferedDiscrimLmInfo>::iterator iterBuffer;
	
	iterBuffer = this->buffer.find(make_pair(lmstate, nextWord));

	if(iterBuffer == this->buffer.end()){
		// lmstate + nextword has not been seen yet
		
		//append the next word to current matched n-gram 
		//and try to chop off the first word till find a match
		C_Ngram ngramWithAppendedWord;
		ngramWithAppendedWord = this->allLMStates[lmstate];
		ngramWithAppendedWord.appendOneWord(nextWord);

		if(ngramWithAppendedWord.returnSize()>this->maxN){	//after appending the next word, size is larger than maxN now
			ngramWithAppendedWord.chopFirstWord();
		}

		//updated lm state is ngramWithAppendedWord
		//update lm state first
		map<C_Ngram, int>::iterator iterState2Id;
		iterState2Id = this->lmStateInfo2Id.find(ngramWithAppendedWord);
		if(iterState2Id==this->lmStateInfo2Id.end()){	//lm state for this n-gram has not been created yet
				this->allLMStates.push_back(ngramWithAppendedWord);
				nextstate = this->allLMStates.size()-1;
				
				//add the lm state to id mapping
				this->lmStateInfo2Id.insert(make_pair(ngramWithAppendedWord, nextstate));
		}
		else{
			nextstate = iterState2Id->second;
		}

		//now, get the discriminative lm cost for this n-gram
		
		//It is a research question here
		//Either accumulate n-gram of all order's discriminative cost
		//or use the value from the longest matched n-gram
		//this also depends on how the discrim. lm is trained

		//here we use the sum of all matched sub n-gram's cost 
		bool exist = false;
		float cost = 0;

		while(ngramWithAppendedWord.returnSize()>0){
			//lookup the list of discriminative lm to check if there exist an n-gram with the discriminative lm cost
			
			float subNgramCost;
			this->discrimLm->returnCost(ngramWithAppendedWord, subNgramCost);
			cost+=subNgramCost;	//if the n-gram does not exist then cost will be returned as 0
			ngramWithAppendedWord.chopFirstWord();
		}

		//add the information to buffer and extend it to the next state
		//add the lmstate plus next word to buffer
		S_BufferedDiscrimLmInfo bufferedInfo;
		bufferedInfo.nextState = nextstate;
		bufferedInfo.cost = cost;

		this->buffer.insert(make_pair( make_pair(lmstate, nextWord), bufferedInfo) );

		return cost;
	}

	//else
	nextstate = iterBuffer->second.nextState;
	return iterBuffer->second.cost;
}


void C_DiscriminativeLMSentenceState::LogProb(CLMTransition & LMTransition)
{
	double totalCost = 0.0;
	CLMState currentState, nextState;

	currentState = LMTransition.OldState;
	int totalWords = LMTransition.Words.size();
	for(int i=0; i<totalWords; i++){
		double logProb = this->LogProb(currentState, LMTransition.Words[i], nextState);
		totalCost+= logProb;
		currentState = nextState;
	}

	LMTransition.Prob = totalCost;
	LMTransition.NewState = currentState;

}


double C_DiscriminativeLMSentenceState::LogProb(CLMState lmstate, const LMWordIndex nextWord )
{
  CLMState NewState;
  return LogProb( lmstate, nextWord, NewState );
}

double C_DiscriminativeLMSentenceState::LogProbEnd(CLMState lmstate)
{
	CLMState dummyNextState;
	return this->LogProb(lmstate, this->sentEndVocId, dummyNextState);
}

double C_DiscriminativeLMSentenceState::LogProbEnd(CLMTransition & LMTransition)
{
	CLMState dummyNextState;
	return this->LogProb(LMTransition.OldState, this->sentEndVocId, dummyNextState);
}

CLMState C_DiscriminativeLMSentenceState::AppendWord( CLMState LMState, const LMWordIndex WordIndex )
{
	CLMState nextState;

	this->LogProb(LMState, WordIndex, nextState);

	return nextState;
}
