#include <Evaluator.hh>
#include <fstream>
#include <iostream>


#define MAXstrlen 10000

//====================================================================
//        CEvaluator
//


//! deprecated
CEvaluator::CEvaluator( CVocabulary* pVocabulary )
{
	m_pVocabulary = pVocabulary;

  //m_p_refSentences = 0;

  m_maxNGram = 9;
}

CEvaluator::CEvaluator( CVocabulary* pVocabulary, 
                        CEvaluatorSharedData* p_sharedData )
{
	m_pVocabulary = pVocabulary;
  m_p_sharedData = p_sharedData;

  m_gainFileName = NULL;
  m_writeGainToFile = false;
  m_maxNGram = 9;
  
  m_nbestSearchDepth = 100000;
  
  m_pTextFilter = NULL;        // when NULL, no text filtering is done
  
  // output control flags
  m_verboseOutput = false;
  m_displayOracleBestSent = false;
}


CEvaluator::~CEvaluator()
{
  // does not destroy m_p_sharedData
}


//------------------------------------------------------------------
//      CEvaluator:: LoadReferences
//

bool
CEvaluator::LoadReferences( const char *pFileName, int Format, int LoadOnlySetNumber )
{

  CArray< CArray< CSentence* > >* p_refSentences = new CArray< CArray< CSentence* > >;

  // normalize and tokenize text before it's saved in the voc and sentence
  CTextFilterBase* pTextFilter = m_p_sharedData->GetTextFilter();
  
  m_p_sharedData->SetReferences(p_refSentences);    // deletes previous refSentences
  
	CSentence* pSentence;
	
  Trace( 1, "CEvaluator::LoadReferences from " << pFileName );

	std::ifstream fp; 
	fp.open(pFileName);
	char str [MAXstrlen + 1 ];
	
	if( !fp.good() ) {
		std::cerr << "reference file couldn't load" << std::endl;
		// flush file handler
		return -1;
	}
	
	assert( m_pVocabulary );
	
	Trace(1, "Load Reference Text" << pFileName );

	int hypIndex = 0;
	int refIndex = 0;

	while (!fp.eof()) 
	{
		fp.getline( str, MAXstrlen );
		
		if( !(pSentence = new CSentence) ) {
			std::cerr << "CEvaluator::LoadReferences(): could not allocate CSentence"
			<< std::endl;
			return -1;
		}
		
		if( str[0] == 0 || ( Format > 0 && refIndex == Format) )                   // blank space for each source 
		{
			hypIndex++;
			refIndex = 0;
		} 
    
    if ( str[0] != 0 ) // skip empty line when using Format = 0
    {
			
			pSentence->SetIncrement(32); //CText::m_SentenceLengthIncrement
      
      // normalize and tokenize
      if( pTextFilter != NULL ) { 
  
        pSentence->ReadFromBuffer( ((std::string)
                                    (pTextFilter->Filter(str))).c_str(), 
                                   m_pVocabulary );
        //pSentence->ReadFromBuffer( filteredStr, m_pVocabulary );

      } else {
        pSentence->ReadFromBuffer( str, m_pVocabulary );
      }
      
      if ( LoadOnlySetNumber == -1 ) {
        // loading all references
        (*p_refSentences)[hypIndex][refIndex] = pSentence;
      } else if ( LoadOnlySetNumber == hypIndex+1 ) {
        // loading only references for hyp sentence of index given by LoadOnlySetNumber
        (*p_refSentences)[0][refIndex] = pSentence;
      }
      

			TraceIf( ( hypIndex % 1000 == 0 ), 1, hypIndex );
      
			refIndex++;

		}		
	}
  
	fp.close();
	
  return 1;
}


//------------------------------------------------------------------
//      CEvaluator:: LoadNBestTranslations
//

CArray< CArray< CSentence* > >*
CEvaluator::LoadNBestTranslations (const char* pFileName, int Format)
{
  
  CArray< CArray< CSentence* > >* p_NBestTranslations = new CArray< CArray< CSentence* > >;

  // normalize and tokenize text before it's saved in the voc and sentences
  CTextFilterBase* pTextFilter = m_p_sharedData->GetTextFilter();
  
  m_p_sharedData->SetNBestTranslations(p_NBestTranslations);

  CSentence* pSentence;
  
  Trace( 0, "CEvaluator::LoadNBestTranslations from " << pFileName );
  
  std::ifstream fp; 

  fp.open(pFileName);
  char str [ MAXstrlen + 1 ];
	
	if( !fp.good() ) {
		std::cerr << "nbest translations file couldn't load" << std::endl;
    return 0;
	}
	
	assert( m_pVocabulary );

  Trace(1, "Load NBestList file" << pFileName );
  
  size_t sourceSentIndex = 0;
  size_t hypSentIndex    = 0;
  
  while (!fp.eof()) 
	{
		fp.getline( str, MAXstrlen );
		
		if( !(pSentence = new CSentence) ) {
			std::cerr << "CEvaluator::LoadReferences(): could not allocate CSentence"
			<< std::endl;
			//return -1;
      return 0;
		}
		
		if(str[0] == 0)                    // blank space
		{
			// hypothesis for next source sentence
			sourceSentIndex++;
			hypSentIndex = 0;
			
		} else {
			
			pSentence->SetIncrement(32); //CText::m_SentenceLengthIncrement

      // normalize and tokenize str before it's saved in the voc and sentence
      const char* filteredStr;
      if( pTextFilter != NULL ) { 
        filteredStr = ((std::string)pTextFilter->Filter(str)).c_str();
        pSentence->ReadFromBuffer( filteredStr, m_pVocabulary );
      } else {
        pSentence->ReadFromBuffer( str, m_pVocabulary );
      }
      
      (*p_NBestTranslations)[sourceSentIndex][hypSentIndex] = pSentence;

			TraceIf( ( hypSentIndex % 1000 == 0 ), 1, hypSentIndex );

			hypSentIndex++;

		}		
	}

	fp.close();
	
  return p_NBestTranslations;
  
}


//------------------------------------------------------------------
//      CEvaluator:: GetNGramsFromSentence
//

NGram_hash_map*
CEvaluator::GetNGramsFromSentence( CSentence* p_sentence )
{
  
  NGram_hash_map* p_ngramFreqHash = new NGram_hash_map();
  NGram_hash_map::iterator iter_ngramFreqHash;
  
  int sentSize = p_sentence->Size();

  // #of ngrams (not-unique) max(sentSize, m_maxNGram) * (sentSize-1)
  
  for(int wordPosition = 0; wordPosition < sentSize; wordPosition++)
  {
    
    int nFitInSentence = std::min(m_maxNGram, sentSize - wordPosition );
    
    //--construct ngrams from wordPosition onwards
    //  append each subsequent word to prevNGram 
    //  thereby enlarging prevNGram by 1 word every iteration to make currNGram
    //  look up currNGram in freq hash_map, increment count value when found
    
    CEvaluatorNGram* p_prevNGram = NULL;
    
    for(int n = wordPosition; n < wordPosition+nFitInSentence; n++)
    {
      CEvaluatorNGram* p_currNGram = NULL;
      
      //p_sentence starts with 1
      const CWordIndex* p_currWord = & p_sentence->GetAt(n+1);  
      
      if(p_prevNGram != NULL)
      {
        //--copy constructor with new size and 1 additional word
        p_currNGram = new CEvaluatorNGram(*p_prevNGram, n-wordPosition+1, *p_currWord); 
      } else {
        p_currNGram = new CEvaluatorNGram(*p_currWord);
      }
      
      iter_ngramFreqHash = p_ngramFreqHash->find(p_currNGram);
      
      if(iter_ngramFreqHash != p_ngramFreqHash->end()) { 
        (iter_ngramFreqHash->second)++;
        
        // no use for current instance of ngram
        delete p_currNGram;
        
        // make it point to what's already in the hash
        p_currNGram = iter_ngramFreqHash->first;
        
      } else {
        p_ngramFreqHash->insert( std::pair< CEvaluatorNGram*, int> ( p_currNGram, 1 ) );
      }
      
      p_prevNGram = p_currNGram;
    }
    
  } // word in segment loop
  
  return p_ngramFreqHash;
  
}


//------------------------------------------------------------------
//      CEvaluator:: validateData
//

//--Validate hypothesis sentences against sets of references
//  # of hyp sentences must equal number of reference sets
//    # of reference sentences per set may vary
bool 
CEvaluator::validateData(CText *pTranslations)
{
  // compare # of ref sentences to # of hyp sent
  int numHypSentences = pTranslations->NumberOfSentences();
  
  int numRefSets = (int)((m_p_sharedData->GetReferences())->Size()); 
  
  if( numHypSentences == numRefSets ) {
    return true;
  } else {
    cout << "DATA INVALID: # of hypothesis sentences (" << numHypSentences << ") mismatches # of reference sets (" << numRefSets << ")" << endl;
  }
}

//------------------------------------------------------------------
//      CEvaluator:: Gain File functions
//

void
CEvaluator::SetWriteGainToFile( bool flag )
{
  m_writeGainToFile = flag;
}

void
CEvaluator::SetGainFileName( char* gainFileName )
{
  m_gainFileName = gainFileName;
}

//------------------------------------------------------------------
//      CEvaluator:: TextFilter functions
//
void
CEvaluator::SetTextFilter(CTextFilterBase* pTextFilter)
{
  m_pTextFilter = pTextFilter;
}

CTextFilterBase*
CEvaluator::GetTextFilter()
{
  return m_pTextFilter;
}

//------------------------------------------------------------------
//      CEvaluator:: NBestSearchDepth functions
//

void
CEvaluator::SetNBestSearchDepth( size_t limit )
{
  m_nbestSearchDepth = limit;
}

size_t
CEvaluator::GetNBestSearchDepth()
{
  return m_nbestSearchDepth;
}

//------------------------------------------------------------------
//      CEvaluator:: set output flags
//

void
CEvaluator::SetVerboseOutput( bool verbose )
{
  m_verboseOutput = verbose;  
}

void
CEvaluator::SetDisplayOracleBestSentences( bool display )
{
  m_displayOracleBestSent = display;  
}



//====================================================================
//        CEvaluatorNGram
//

CEvaluatorNGram::CEvaluatorNGram()
{
  m_size      = 0;
  m_signature = 0;
}


CEvaluatorNGram::CEvaluatorNGram ( const CWordIndex word )
{
  m_p_words = new CWordIndex[ 1 ];  
  m_p_words[ 0 ] = word; 
  m_size = 1;
  m_signature = 0;
}


//-- creates a new NGram of size newSize. if smaller than prevNGram, will
//   truncate prevNGram's CWordIndex*. If larger, will fill rest with 0's
CEvaluatorNGram::CEvaluatorNGram( CEvaluatorNGram* p_prevNGram, size_t newSize)
{
  
  if( newSize != 0 ) 
  {
    m_p_words = new CWordIndex[newSize];
    CWordIndex* p_prevWords = p_prevNGram->getWords();
    int prevWordsBoundary = std::min(newSize, p_prevNGram->getSize());
    
    for(int i=0; i < prevWordsBoundary; i++) {
      m_p_words[i] = p_prevWords[i];
    }
    
    if(prevWordsBoundary < newSize) {
      for(int i = prevWordsBoundary; i < newSize; i++){
        m_p_words[i] = 0;
      }
    }
    
  }
  
  m_size = newSize;
  m_signature = 0;
}


//--Creates a new CEvaluatorNGram starting with contents of passed CEvaluatorNGram. 
//  CEvaluatorNGram argument is the last element of new ngram
CEvaluatorNGram::CEvaluatorNGram( CEvaluatorNGram &prevNGram, size_t newSize, const CWordIndex word)
{
  
  m_p_words = new CWordIndex [newSize];
  
  //--retrieve a const CWordIndex reference from sentences
  //  it's const up to this point
  //  m_p_words[newSize-1] = const_cast< CWordIndex* >( & word );
  m_p_words[newSize-1] = word;
  
  CWordIndex* pp_prevWords = prevNGram.getWords();
  
  int prevNGramSize = prevNGram.getSize();
  
  for(int i=0; i < prevNGramSize; i++) {
    m_p_words[i] = pp_prevWords[i];          
  }
  
  m_size = newSize;
  
  m_signature = 0;
}


CEvaluatorNGram::~CEvaluatorNGram()
{
  if(m_size > 0)
    delete [] m_p_words;    // don't delete the actual CWordIndex* 
}


//------------------------------------------------------------------
//      CEvaluatorNGram:: getSignature
//

unsigned int
CEvaluatorNGram::getSignature() const
{
  
  if(m_signature == 0)                     
  {
    //--Assume max size of ngrams is 9. One prime number per ngram element
    //  Note: signature is not guaranteed to be unique, though it almost always
    //        is. This significantly reduces # of collisions when CEvaluatorNGram
    //        is used as a key in a hash table
    
    //int primes[] = { 2, 5, 7, 11, 13, 17, 19, 23, 29, 31 }; // high # of collisions
    
    int primes[] = {8933, 8941, 8951, 8963, 8969, 8971, 8999, 9001, 9007, 9011};

    for( int i = m_size-1; i >= 0; --i ){
      m_signature +=  m_p_words[i] * primes[i];
    }
  }
  
  return m_signature;
}


//-- CEvaluatorNGram Get Members--------------------------------------
size_t 
CEvaluatorNGram::getSize() const {
  return m_size;
}


CWordIndex* 
CEvaluatorNGram::getWords() const {
  return m_p_words;
}

//-- CEvaluatorNGram Display and ToString ----------------------------

void
CEvaluatorNGram::Display() 
{
  for ( int n = 0; n < m_size; n++ ) {
    std::cout << m_p_words[ n ] << " ";
  }

  std::cout << endl;
}

CString
CEvaluatorNGram::ToString(CVocabulary* pVoc) const
{
  std::string str = "";
  for ( int n = 0; n < m_size; n++ ) {
    str += (std::string)(pVoc->GetWord( m_p_words[ n ])) + " ";
  }
  
  CString String = CString(str.c_str());
  String.TrimTrailingBlanks();
  return String;
}

CString
CEvaluatorNGram::ToString() const
{
  
  std::ostringstream os; 

  for ( int n = 0; n < m_size; n++ ) {
    os << m_p_words[ n ] << " ";
  }
  
  CString String = CString(os.str().c_str());
  String.TrimTrailingBlanks();
  return String;
}


//====================================================================
//      CEvaluatorNGramMatchStats
//

CEvaluatorNGramMatchStats::CEvaluatorNGramMatchStats()
{
  matchingNGrams = new unsigned short [9+1];       // m_maxNGramSize+1
  infoGain = new float[9+1];          // for info gain, used by BLEU
  
  for(int n=1; n<=9; n++) {
    matchingNGrams[n] = 0;
    infoGain      [n] = 0;                   
  }
  
  pHypSentence = NULL;
}

CEvaluatorNGramMatchStats::~CEvaluatorNGramMatchStats()
{
  delete [] matchingNGrams;
  delete [] infoGain;
}



//====================================================================
//      CEvaluatorSharedData
//

CEvaluatorSharedData::CEvaluatorSharedData()
{
  m_p_translations      = 0;
  m_p_nBestTranslations = 0;
  m_p_references        = 0;
  m_pp_maxRefHash       = 0;
  m_p_gainHash          = 0;
  
  m_p_hypStats = new CArray< CArray< CEvaluatorNGramMatchStats* > >;
  
  m_toComputeMatchGain       = true;
  m_toComputeMatchCounts     = true;
  m_toComputeNistLengths     = true;
  m_toComputeBleuLengths     = true;
  m_toComputeNistBleuLengths = true;
  
  m_pTextFilter = NULL;
}

CEvaluatorSharedData::~CEvaluatorSharedData() 
{
  ResetTranslations();
  
  ResetStats();
  delete m_p_hypStats;

  ResetReferences();
}


bool
CEvaluatorSharedData::ToComputeMatchCounts() {
  return m_toComputeMatchCounts;
}

bool
CEvaluatorSharedData::ToComputeMatchGain() {
  return m_toComputeMatchGain;
}

bool
CEvaluatorSharedData::ToComputeNistLengths() {
  return m_toComputeNistLengths;
}

bool
CEvaluatorSharedData::ToComputeBleuLengths() {
  return m_toComputeBleuLengths;
}

bool
CEvaluatorSharedData::ToComputeNistBleuLengths() {
  return m_toComputeNistBleuLengths;
}

bool
CEvaluatorSharedData::ToBuildMaxRefHash() {
  return (m_pp_maxRefHash == 0);       // assume references stay the same
}

bool
CEvaluatorSharedData::ToBuildGainHash() {
  return (m_p_gainHash == 0);          // assume references stay the same
}


//------------------------------------------------------------------
//      CEvaluatorSharedData:: SetNBestTranslations

void
CEvaluatorSharedData::SetNBestTranslations( CArray< CArray< CSentence* > >*
                                              p_nBestTranslations)
{
  if( p_nBestTranslations != m_p_nBestTranslations) 
  {
    ResetTranslations();
    m_p_nBestTranslations = p_nBestTranslations;
  }
}


//------------------------------------------------------------------
//      CEvaluatorSharedData:: SetTranslations

void
CEvaluatorSharedData::SetTranslations( CText* p_translations )
{
  if( p_translations != m_p_translations )
  {
    ResetTranslations();
    m_p_translations = p_translations;
  }
}



//------------------------------------------------------------------
//      CEvaluatorSharedData:: ResetTranslations

void
CEvaluatorSharedData::ResetTranslations()
{

  if( m_p_nBestTranslations != 0)
  {
    // delete m_p_nBestTranslations
    for(int i=m_p_nBestTranslations->MinIndex(); 
        i <= m_p_nBestTranslations->MaxIndex(); i++) {
      // what about CSentence*
      (*m_p_nBestTranslations)[i].Clear();
    }
    m_p_nBestTranslations->Clear();
    delete m_p_nBestTranslations;
    m_p_nBestTranslations = 0;
  }
  
  // delete m_p_translations
  if( m_p_translations !=0 ){
    // what about CSentence*
    m_p_translations->Clear();
    delete m_p_translations;
    m_p_translations = 0;
  }
  
  ResetStats();
  
}

void
CEvaluatorSharedData::ResetReferences()
{
  if( m_p_references != 0 ) 
  {
    //CArray< CArray< CSentence* > >* m_p_references
    for ( int s = m_p_references->MinIndex(); 
          s <= m_p_references->MaxIndex();
          s++)
    {
      for ( int r = (*m_p_references)[ s ].MinIndex(); 
            r <= (*m_p_references)[ s ].MaxIndex();
            r++ )
      {
        delete (*m_p_references)[ s ][ r ];
        (*m_p_references)[ s ][ r ] = NULL;
      }
      
      (*m_p_references)[ s ].Clear();
    }
    
    m_p_references->Clear();
    delete m_p_references;
    
    m_p_references = 0;
  }
}


void
CEvaluatorSharedData::ResetStats()
{
  if( m_p_hypStats != 0 ) 
  {
    for(int i=m_p_hypStats->MinIndex(); i<=m_p_hypStats->MaxIndex(); i++) {
      (*m_p_hypStats)[i].Clear();
      //(*m_p_hypStats)[i].RemoveAll();
    }
    //m_p_hypStats->RemoveAll();
    m_p_hypStats->Clear();
    //delete m_p_hypStats;
    //m_p_hypStats = new CArray< CArray< CEvaluatorNGramMatchStats> >;
  }
  
  //-- Stats must be recomputed (but m_p_gainHash and m_pp_maxRefHash may not)
  m_toComputeMatchGain = true;
  m_toComputeMatchCounts = true;
  m_toComputeNistLengths = true;       // TODO: remove
  m_toComputeBleuLengths = true;
  
}

//-- Set Methods------------------------------------------------------
void 
CEvaluatorSharedData::SetReferences(CArray< CArray< CSentence* > >* p_references) 
{
  if(p_references != m_p_references)
  {
    if (m_p_references != 0) 
    {
      // resetGainHash
      NGramToDouble_hash_map* emptyGainHash = 0;
      SetGainHash(emptyGainHash);
      
      // reset maxRefHash
      NGram_hash_map** emptyMaxRefHash = 0;
      SetMaxRefHash(emptyMaxRefHash);
      
      // clear reference container
      ResetReferences();
      
      // all stats must be recomputed for new references
      ResetStats();
    }
    
    m_p_references = p_references;
  }
}

void 
CEvaluatorSharedData::SetMaxRefHash(NGram_hash_map** pp_maxRefHash) {
  if( (m_pp_maxRefHash != 0) && (m_pp_maxRefHash != pp_maxRefHash) ) 
  {
    /*
    // delete reference ngram->max count hash
    for(int i=0; i < m_maxNGram; i++)
    {
      NGram_hash_map* p_currHash = m_pp_maxRefHash[i];
      
      NGram_hash_map::iterator currHash_iter = p_currHash->begin();
      
      while(currHash_iter != p_currHash->end()) {
        delete currHash_iter->first;
        currHash_iter++;
      }
      
      p_currHash->clear();
    }
    delete [] m_pp_maxRefHash;
     */
  }
    
  m_pp_maxRefHash = pp_maxRefHash;
}

void 
CEvaluatorSharedData::SetGainHash(NGramToDouble_hash_map* p_gainHash) {
  if( (m_p_gainHash != 0) && (m_p_gainHash != p_gainHash) ) 
  {
    // delete reference->info gain hash
    NGramToDouble_hash_map::iterator gainHash_iter = m_p_gainHash->begin();
    while(gainHash_iter != m_p_gainHash->end())
    {
      delete gainHash_iter->first;
      gainHash_iter++;
    }
    m_p_gainHash->clear();
    delete m_p_gainHash;  
  }
  
  m_p_gainHash = p_gainHash;
}

void 
CEvaluatorSharedData::SetHypMatchStats(CArray< CArray< CEvaluatorNGramMatchStats* > >*
                                       p_hypStats) 
{
  if( (m_p_hypStats != 0) && (m_p_hypStats != p_hypStats) ) 
  {
    for(int i=m_p_hypStats->MinIndex(); i<=m_p_hypStats->MaxIndex(); i++) {
      (*m_p_hypStats)[i].Clear();
    }
    m_p_hypStats->Clear();
    delete m_p_hypStats;
  }
  m_p_hypStats = p_hypStats;
}

void 
CEvaluatorSharedData::SetToComputeMatchCounts(bool b) {
  m_toComputeMatchCounts = b;
}

void 
CEvaluatorSharedData::SetToComputeMatchGain(bool b) {
  m_toComputeMatchGain = b;
}

void
CEvaluatorSharedData::SetToComputeNistLengths(bool b) {
  m_toComputeNistLengths = b;
}

void
CEvaluatorSharedData::SetToComputeBleuLengths(bool b) {
  m_toComputeBleuLengths = b;
}

void
CEvaluatorSharedData::SetToComputeNistBleuLengths(bool b) {
  m_toComputeNistBleuLengths = b;
}

void
CEvaluatorSharedData::SetTextFilter( CTextFilterBase* pTextFilter ) {
  m_pTextFilter = pTextFilter;
}

//-- Get Methods------------------------------------------------------
CText* 
CEvaluatorSharedData::GetTranslations() {
  return m_p_translations;
}

CArray< CArray< CSentence* > >* 
CEvaluatorSharedData::GetNBestTranslations() {
  return m_p_nBestTranslations;
}

CArray< CArray< CSentence* > >* 
CEvaluatorSharedData::GetReferences() {
  return m_p_references;
}

NGram_hash_map** 
CEvaluatorSharedData::GetMaxRefHash() {
  return m_pp_maxRefHash;
}

NGramToDouble_hash_map* 
CEvaluatorSharedData::GetGainHash() {
  return m_p_gainHash;
}

CArray< CArray< CEvaluatorNGramMatchStats* > >* 
CEvaluatorSharedData::GetHypMatchStats() {
  return m_p_hypStats;
}

CTextFilterBase*
CEvaluatorSharedData::GetTextFilter() {
  return m_pTextFilter;
}









