
//====================================================================
//      CEvaluatorFScore::
//        

#include "EvaluatorFScore.hh"

CEvaluatorFScore::CEvaluatorFScore( CVocabulary *pVocabulary,
                                    CEvaluatorSharedData *pSharedData )
: CEvaluatorBLEU( pVocabulary, pSharedData )
{
  m_precisionScaleFactor = 1;
  m_recallScaleFactor    = 1;
}

CEvaluatorFScore::~CEvaluatorFScore()
{
}


//------------------------------------------------------------------
//      CEvaluatorFScore::Evaluate
bool
CEvaluatorFScore::Evaluate( CArray< CEvaluatorNGramMatchStats* >* p_hypMatchStats,
                          CNumericalArray< float > &Scores )
{
  
  CumulativeScores cumScores(m_maxNGram);                       // track aggregate values
  // don't need n-gram scores, just cumulative, so unlike Bleu, not using array by n-gram size
  int numRefNGrams = 0;
  int numHypNGrams = 0; 
  int numMatchingNGrams = 0;
  
  int sourceSentCount = p_hypMatchStats->Size();   // iterate over p_hypMatchStats
  
  for(int s=0; s < sourceSentCount; s++)
  {
    CEvaluatorNGramMatchStats* p_hypStats = (*p_hypMatchStats)[s];
    
    //-- add hyp length and closest ref length
    cumScores.hypSentLength += p_hypStats->hypSentLength;
    cumScores.closestRefSentLength += p_hypStats->closestRefSentLength;
    
    //-- add # of possible n-grams in hyp and "best" ref (assuming 1 ref only)
    for ( int n=1; n <= m_maxNGram; n++ ) {
      if (p_hypStats->hypSentLength >= n )
        numHypNGrams += p_hypStats->hypSentLength-n+1;
      if ( p_hypStats->closestRefSentLength >= n )
        numRefNGrams += p_hypStats->closestRefSentLength-n+1; // used by Bleu in length penalty
      numMatchingNGrams += p_hypStats->matchingNGrams[n];
    }
  } // end source sent loop
  
  // Scaled Precision: sum of matches / # of possible n-grams in hypothesis  
  float precision = m_precisionScaleFactor * (float)numMatchingNGrams / (float)numHypNGrams;
  
  // Scaled Recall: sum of matches / # of possible n-grams in references
  float recall = m_recallScaleFactor * (float)numMatchingNGrams / (float)numRefNGrams;

  // Combine recall and precision with harmonic mean
  float harmonicMeanPrecisionRecall = ( 2 * precision * recall ) / ( precision + recall );

  Scores[0] = harmonicMeanPrecisionRecall;
  Scores[1] = precision;
  Scores[2] = recall;
  
  return true;
  
}

/**
 * @brief Score one sentence with the given set of references
 *
 *
 */
bool 
CEvaluatorFScore::Evaluate( CSentence* pHypSentence, CArray< CSentence* > &References,
                        CNumericalArray< float >& Scores )
{
  ;
}

//------------------------------------------------------------------
//      CEvaluatorFScore:: ReportScores
//
//! Report recall, precision, and combined, scaled score
void
CEvaluatorFScore::ReportScores( CNumericalArray< float > & Scores )
{
  cout << "F-Score Metric Score " << endl
    << " Harmonic mean: " << Scores[0] 
    << " Precision: " << Scores[1] 
    << " Recall: " << Scores[2];
}

void
CEvaluatorFScore::SetScalingFactors( float precisionScaleFactor, float recallScaleFactor )
{
  m_precisionScaleFactor = precisionScaleFactor;
  m_recallScaleFactor = recallScaleFactor;
}











