#ifndef __EvaluatorFScore__
#define __EvaluatorFScore__

#include "EvaluatorBLEU.hh"
/*!
  F-Score Evaluator
  Computes F-Score for a source and hypothesis. F-Score is the harmonic mean of 
  n-gram precision and recall.
 
 \author Anthony D'Auria (apd@cs.cmu.edu) 7/30/07
 
 */

//====================================================================
//      CEvaluatorFScore
//

class CEvaluatorFScore : public CEvaluatorBLEU
{
public: 
  CEvaluatorFScore( CVocabulary *pVocabulary, CEvaluatorSharedData *pSharedData );
  ~CEvaluatorFScore();
  
  //! Re-implements all evaluate functions because we only need three values
  virtual bool Evaluate(CArray< CEvaluatorNGramMatchStats* >* p_hypMatchStats,
                        CNumericalArray< float > &Scores);
  virtual bool Evaluate( CSentence* pHypSentence, CArray< CSentence* > &References,
                        CNumericalArray< float >& Scores ); 
  //! Print scores to screen (we are printing different scores)
  virtual void ReportScores( CNumericalArray< float > & Scores );
  
  //! Control effect of precision and recall on final score. Default value: 1
  void SetScalingFactors( float precisionScaleFactor, float recallScaleFactor );
  
private:
  //! Multiplicative scale of precision
  float m_precisionScaleFactor;
  //! Multiplicative scale of recall
  float m_recallScaleFactor;
};

#endif
