#include <EvaluatorNIST.hh>
#include <math.h>

//====================================================================
//      CEvaluatorNIST::
//        

CEvaluatorNIST::CEvaluatorNIST( CVocabulary *pVocabulary, 
                                CEvaluatorSharedData *pSharedData )
  : CEvaluatorBLEU( pVocabulary, pSharedData )
{
  m_p_sharedData = pSharedData;
  m_N_PRECISION = 5;
}

//-- Destroy member hash maps (ref ngram->max count, ref ngram->info gain)
CEvaluatorNIST::~CEvaluatorNIST()
{
}


//------------------------------------------------------------------
//      CEvaluatorNIST:: Evaluate
//        

bool
CEvaluatorNIST::Evaluate( const char *pFileWithTranslations, 
                          CNumericalArray< float > & Scores )
{
}



bool 
CEvaluatorNIST::Evaluate( CArray< CString > &Translations, 
                          CNumericalArray< float > & Scores )
{
  
}

bool CEvaluatorNIST::Evaluate( int sourceSentenceIndex, CSentence* pHypSentence, 
                               CNumericalArray< float >& Scores)
{
  
}


// new Evaluate

bool
CEvaluatorNIST::Evaluate( CText *pTranslations,
                          CNumericalArray< float > & Scores )
{
  
  m_p_sharedData->SetTranslations(pTranslations);
  
  //-- check if ngram match stats and gain been calculated
  if( m_p_sharedData->ToBuildGainHash() || m_p_sharedData->ToComputeMatchGain() ) {
    // cout << "NIST.Evaluate(): about to compute NGramMatchInfoStats" << endl;
    ComputeNGramMatchStats(pTranslations);
  }
  
  //-- get precomputed ngram stats
  CArray< CArray< CEvaluatorNGramMatchStats* > >* p_hypMatchStats
    = m_p_sharedData->GetHypMatchStats();
  
  int sourceSentCount = p_hypMatchStats->Size();
  int hypSentCount;
  
  //-- init cumulative scores
  CumulativeNISTScores cumScores( m_maxNGram );

  for (int s=0; s < sourceSentCount; s++)
  {
    // one hyp per source sentence, so [s][0]
    CEvaluatorNGramMatchStats* p_hypStats = (*p_hypMatchStats)[s][0];
    
    cumScores.avgRefSentLength += p_hypStats->avgRefSentLength;
    
    //-- Add # of possible n-grams
    for(int n=1; n<= m_maxNGram; n++){
      cumScores.numNGramsInHyp[n] += p_hypStats->hypSentLength - n + 1;
      cumScores.matchingNGrams[n] += p_hypStats->matchingNGrams[n];
      cumScores.hypInfoGain[n] += p_hypStats->infoGain[n];
    }
    
    //-- Add hyp length
    cumScores.hypSentLength += p_hypStats->hypSentLength; 
    
  } // end source sent loop
    
  //-- Length penalty
  
  double nistLengthPenalty = ComputeLengthPenalty((double)cumScores.hypSentLength 
                                                  / cumScores.avgRefSentLength);

  Scores[10] = nistLengthPenalty;
  Scores[11] = cumScores.hypSentLength;            
  Scores[12] = cumScores.avgRefSentLength;         
  
  //-- add up scores per n

  Scores[0] = 0;                       // not set
  double runningScore = 0;
  
  for(int n=1; n <= m_maxNGram; n++)
  {
    runningScore += ( cumScores.hypInfoGain[n] 
                      / (double)(std::max(cumScores.numNGramsInHyp[n], 1)) );
    
    Scores[n] = runningScore * nistLengthPenalty;
  }
  
  //delete [] cumScores.numNGramsInHyp;
  //  delete [] cumScores.matchingNGrams;
  //  delete [] cumScores.hypInfoGain;
  
  return true;
  
}

/**
 * @brief Score one sentence with the given set of references
 *
 *
 */
bool 
CEvaluatorNIST::Evaluate( CSentence* pHypSentence, CArray< CSentence* > &References,
                        CNumericalArray< float >& Scores )
{
  ;
}

// old Evaluate
/*
bool
CEvaluatorNIST::Evaluate( CText *pTranslations, 
                          CNumericalArray< float > & Scores )
{
  
  bool createPersistentRefHashes = (m_pp_maxRefHash == 0 && m_p_gainHash == 0);
  

  //-- Validate data (# hyp sentences == # ref sent sets)
  //if(createPersistentRefHashes) {
  //  if( validateData(pTranslations) ){
  //    // good data
  //  } else {
  //    // do something about bad data
  //  }
  //}

  
  //-- init cumulative scores
  CumulativeNISTScores cumScores;
  cumScores.numNGramsInHyp = new int[m_maxNGram+1];
  cumScores.matchingNGrams = new int[m_maxNGram+1];
  cumScores.hypInfoGain = new double[m_maxNGram+1];

  for(int n=0; n<= m_maxNGram; n++){
    cumScores.numNGramsInHyp[n] = 0;
    cumScores.matchingNGrams[n] = 0;
    cumScores.hypInfoGain[n] = 0;
  }
  
  cumScores.avgRefSentLength = 0;
  cumScores.hypSentLength = 0;
  
  int totalRefWords = 0;
  
  //-- Build info gain hash from ngrams in all reference sentences
  
  NGram_hash_map* p_refHash;     // used to build gainHash, then deleted

  int hypSentCount = pTranslations->NumberOfSentences();
  
  if(createPersistentRefHashes) {
    p_refHash = new NGram_hash_map(); 
    m_pp_maxRefHash = new NGram_hash_map* [hypSentCount]; 
  }
  
  for(int h=0; h < hypSentCount; h++)
  {
    int refSentCount = (int)m_refSentences[h].Size();
    int refSentLengths = 0;

    if(createPersistentRefHashes) {
      m_pp_maxRefHash[h] = new NGram_hash_map();
    }
    
    for(int r=0; r < refSentCount; r++)
    {
      CSentence* p_currRefSentence = m_refSentences[h][r];
      
      if(createPersistentRefHashes)
      {
        NGram_hash_map* p_currRefHash = GetNGramsFromSentence(p_currRefSentence);
        BuildCountsHashes(h, m_pp_maxRefHash, p_refHash, p_currRefHash);

        p_currRefHash->clear();
        delete p_currRefHash;
      }      
      
      // get avg length
      refSentLengths += p_currRefSentence->Size();
    }
    
    // add up avg ref sent length per hyp sent
    cumScores.avgRefSentLength += (double)refSentLengths / (double)refSentCount;
    totalRefWords += refSentLengths;   //TODO: not add if !createPersistentRef
  }
  
  if(createPersistentRefHashes)
  {
    // use total ref count hash to build ngram info gain hash
    m_p_gainHash = BuildInfoGainHash(p_refHash, totalRefWords);
    
    //cout << " num of unique ref ngrams: " << p_refHash->size() << " num of buckets in hash: " << p_refHash->bucket_count() << endl;
    
    //cout << " num of unique gain ngrams: " << m_p_gainHash->size() << " numb of buckets in hash: " << m_p_gainHash->bucket_count() << endl;
    
    p_refHash->clear();
    delete p_refHash;
  }  

  //-- Score hypothesis sentences
  int hypSentLength;
  
  for(int h=0; h < hypSentCount; h++)
  {
    CSentence* p_currHypSentence = pTranslations->GetSentence(h);
    hypSentLength = p_currHypSentence->Size();
    NGram_hash_map* p_currHypHash = GetNGramsFromSentence(p_currHypSentence);
    
    //-- Add # of possible n-grams in hyp only
    for(int n=1; n<= m_maxNGram; n++){
      cumScores.numNGramsInHyp[n] += hypSentLength-n+1;
    }
    
    //-- Add hyp length
    cumScores.hypSentLength += hypSentLength; // should use numNGramsInHyp[1]    
    
    //-- Add n-gram co-occurrences and info gain
    for(NGram_hash_map::iterator hypHash_iter = p_currHypHash->begin();
        hypHash_iter != p_currHypHash->end(); hypHash_iter++)
    {
      NGram_hash_map::iterator maxRefHash_iter = 
          (m_pp_maxRefHash[h])->find(hypHash_iter->first);
      
      if(maxRefHash_iter != (m_pp_maxRefHash[h])->end() )
      {
        int n = (hypHash_iter->first)->getSize();
        int numMatched = std::min(hypHash_iter->second,
                                  maxRefHash_iter->second);

        cumScores.matchingNGrams[n] += numMatched; // may not be necessary
        
        //-- add info gain for ngram
        NGramToDouble_hash_map::iterator infoGainHash_iter 
          = m_p_gainHash->find(hypHash_iter->first);
        
        if(infoGainHash_iter != m_p_gainHash->end())
        {
          cumScores.hypInfoGain[n] 
            += infoGainHash_iter->second * (double)numMatched;
        
        } else {
          // hyp ngram not found in gainHash
        }
        
      } else {
        // hyp ngram not found in maxRefHash
      }
      
      delete hypHash_iter->first;
      
    } // end hyp sent ngram loop
    
    p_currHypHash->clear();
    delete p_currHypHash;
    
  } // end hyp sent loop
  
  //-- Length penalty
  
  double nistLengthPenalty = ComputeLengthPenalty((double)cumScores.hypSentLength 
                                                  / cumScores.avgRefSentLength);

  Scores[10] = nistLengthPenalty;
  Scores[11] = cumScores.hypSentLength;            
  Scores[12] = cumScores.avgRefSentLength;         
  
  //-- add up scores per n

  Scores[0] = 0;                       // not set
  double runningScore = 0;
  
  for(int n=1; n <= m_maxNGram; n++)
  {
    runningScore += ( cumScores.hypInfoGain[n] 
                      / (double)(std::max(cumScores.numNGramsInHyp[n], 1)) );
    
    Scores[n] = runningScore * nistLengthPenalty;
  }
  
  delete [] cumScores.numNGramsInHyp;
  delete [] cumScores.matchingNGrams;
  delete [] cumScores.hypInfoGain;
  
  return true;
}
*/


bool
CEvaluatorNIST::Evaluate( const CArray< CNumericalArray< CWordIndex > >
                          *pTanslations, CNumericalArray< float > & Scores )
{
}


bool
CEvaluatorNIST::Evaluate( CArray< CEvaluatorNGramMatchStats* >* p_hypMatchStats,
                          CNumericalArray< float > &Scores )
{
  int sourceSentCount = p_hypMatchStats->Size();
  int hypSentCount;
  
  CumulativeNISTScores cumScores(m_maxNGram);   // stores aggregate values
  
  for (int s=0; s < sourceSentCount; s++)
  {
    // one hyp per source sentence, so [s][0]
    CEvaluatorNGramMatchStats* p_hypStats = (*p_hypMatchStats)[s];
    
    cumScores.avgRefSentLength += p_hypStats->avgRefSentLength;
    
    //-- Add # of possible n-grams
    for(int n=1; n<= m_maxNGram; n++){
      cumScores.numNGramsInHyp[n] += p_hypStats->hypSentLength - n + 1;
      cumScores.matchingNGrams[n] += p_hypStats->matchingNGrams[n];
      cumScores.hypInfoGain   [n] += p_hypStats->infoGain[n];
    }
    
    //-- Add hyp length
    cumScores.hypSentLength += p_hypStats->hypSentLength; 
    
  } // end source sent loop
  
  //-- Length penalty
  
  double nistLengthPenalty = ComputeLengthPenalty((double)cumScores.hypSentLength 
                                                  / cumScores.avgRefSentLength);
  
  Scores[10] = nistLengthPenalty;
  Scores[11] = cumScores.hypSentLength;            
  Scores[12] = cumScores.avgRefSentLength;         
  
  //-- add up scores per n
  
  Scores[0] = 0;                       // not set
  double runningScore = 0;
  
  for(int n=1; n <= m_maxNGram; n++)
  {
    runningScore += ( cumScores.hypInfoGain[n] 
                      / (double)(std::max(cumScores.numNGramsInHyp[n], 1)) );
    
    Scores[n] = runningScore * nistLengthPenalty;
  }
   
  return true;
  
}




//------------------------------------------------------------------
//      CEvaluatorNIST:: ComputeNGramMatchStats
//        

bool
CEvaluatorNIST::ComputeNGramMatchStats( CArray< CArray< CSentence* > >* p_NBestTranslations)
{

  // retrieve shared data
  assert(m_p_sharedData);
  CArray< CArray< CSentence* > >* p_refSentences = m_p_sharedData->GetReferences();
  assert(p_refSentences);
  CArray< CArray< CEvaluatorNGramMatchStats* > >* p_hypMatchStats 
    = m_p_sharedData->GetHypMatchStats();
  assert(p_hypMatchStats);
  
  NGram_hash_map**        pp_maxRefHash = m_p_sharedData->GetMaxRefHash();
  NGramToDouble_hash_map* p_gainHash    = m_p_sharedData->GetGainHash();
  NGram_hash_map*         p_refHash;    // used to build gainHash, then deleted
  
  
  // hyp-specific scores
  bool toComputeMatchCounts = m_p_sharedData->ToComputeMatchCounts();
  bool toComputeMatchGain = m_p_sharedData->ToComputeMatchGain();

  // non-hyp-specific reference stats
  bool toBuildMaxRefHash = m_p_sharedData->ToBuildMaxRefHash();
  bool toBuildGainHash = m_p_sharedData->ToBuildMaxRefHash();
  
  int totalRefWords = 0;
  //-- Build info gain hash from ngrams in all reference sentences
  
  int sourceSentCount = p_NBestTranslations->Size();
  int hypSentCount, refSentCount, refSentLengths;
  
  if (toBuildGainHash) {
    p_refHash = new NGram_hash_map(); 
  }
  
  if (toBuildMaxRefHash) {
    pp_maxRefHash = new NGram_hash_map* [sourceSentCount]; 
    m_p_sharedData->SetMaxRefHash(pp_maxRefHash);
  }
  
  cout << "source sent count: " << sourceSentCount << endl;
  
  // create and store match stats per source sentence
  for(int s=0; s < sourceSentCount; s++)
  {
    hypSentCount = (*p_NBestTranslations)[s].Size();

    refSentCount = (int)(*p_refSentences)[s].Size();
    refSentLengths = 0;
    
    if(toBuildMaxRefHash) {
      pp_maxRefHash[s] = new NGram_hash_map();
    }
    
    for(int r=0; r < refSentCount; r++)
    {
      CSentence* p_currRefSentence = (*p_refSentences)[s][r];
      
      if(toBuildMaxRefHash || toBuildGainHash)
      {
        NGram_hash_map* p_currRefHash = GetNGramsFromSentence(p_currRefSentence);
        BuildCountsHashes(s, pp_maxRefHash, p_refHash, p_currRefHash, 
                          toBuildGainHash, toBuildMaxRefHash);
        
        p_currRefHash->clear();
        delete p_currRefHash;
      }      
      
      // get avg length
      refSentLengths += p_currRefSentence->Size();
    }
    //currHypStats.refSentCount = refSentCount;
    //currHypStats.refSentLenthsSum = refSentLengths;
    
                                                 // add up avg ref sent length per hyp sent
    
    //cumScores.avgRefSentLength += (double)refSentLengths / (double)refSentCount;
    totalRefWords += refSentLengths;   //TODO: not add if !createPersistentRef
    
    //1. Save reference sentence length info in all [s][h] match objects
    //   - Create match objects if necessary

    for(int h = 0; h < hypSentCount; h++) 
    {
      if( toComputeMatchCounts ) {

        (*p_hypMatchStats)[s][h] = new CEvaluatorNGramMatchStats();
        // pointer to hyp sentence for oracle best sentence set
        (*p_hypMatchStats)[s][h]->pHypSentence = (*p_NBestTranslations)[s][h];
        (*p_hypMatchStats)[s][h]->hypSentenceIndex = h;
        
      }
      CEvaluatorNGramMatchStats* p_currHypStats = (*p_hypMatchStats)[s][h];
      p_currHypStats->avgRefSentLength = (double)refSentLengths / (double)refSentCount;
      
      p_currHypStats->totalRefWords = refSentLengths; //TODO: not needed
    }
    
  }
  
  if(toBuildGainHash)
  {
    // use total ref count hash to build ngram info gain hash
    p_gainHash = BuildInfoGainHash(p_refHash, totalRefWords);
    m_p_sharedData->SetGainHash(p_gainHash);
    
    /*
     cout << " num of unique ref ngrams: " << p_refHash->size() << " num of buckets in hash: " << p_refHash->bucket_count() << endl;
     
     cout << " num of unique gain ngrams: " << m_p_gainHash->size() << " numb of buckets in hash: " << m_p_gainHash->bucket_count() << endl;
     */
    
    p_refHash->clear();
    delete p_refHash;
  }  
  
  //-- Score hypothesis sentences
  int hypSentLength;
  //int hypSentCount;
  
  for(int s=0; s < sourceSentCount; s++)
  {
    
    hypSentCount = (*p_NBestTranslations)[s].Size();

    cout << "s: " << s << " hyp sent count: " << hypSentCount << endl;

    for(int h=0; h < hypSentCount; h++)
    {
      
      CSentence* p_currHypSentence = (*p_NBestTranslations)[s][h];
      CEvaluatorNGramMatchStats* p_currHypStats = (*p_hypMatchStats)[s][h];
      
      hypSentLength = p_currHypSentence->Size();
      p_currHypStats->hypSentLength = hypSentLength;
      
      NGram_hash_map* p_currHypHash = GetNGramsFromSentence(p_currHypSentence);
      
      //-- Add hyp length
      //cumScores.hypSentLength += hypSentLength; // should use numNGramsInHyp[1]   
      
      //-- Add n-gram co-occurrences and info gain
      for(NGram_hash_map::iterator hypHash_iter = p_currHypHash->begin();
          hypHash_iter != p_currHypHash->end(); hypHash_iter++)
      {
        NGram_hash_map::iterator maxRefHash_iter = 
        (pp_maxRefHash[s])->find(hypHash_iter->first);
        
        if(maxRefHash_iter != (pp_maxRefHash[s])->end() )
        {
          int n = (hypHash_iter->first)->getSize();
          int numMatched = std::min(hypHash_iter->second,
                                    maxRefHash_iter->second);

          if( toComputeMatchCounts ) {
            p_currHypStats->matchingNGrams[n] += numMatched;
          }          

          if( toComputeMatchGain ) 
          {          
            //-- add info gain for ngram
            NGramToDouble_hash_map::iterator infoGainHash_iter 
            = p_gainHash->find(hypHash_iter->first);
            
            if(infoGainHash_iter != p_gainHash->end())
            {
              
              p_currHypStats->infoGain[n] 
              += infoGainHash_iter->second * (double)numMatched;
              
            } else {
              // hyp ngram not found in gainHash
            }
          }
          
        } else {
          // hyp ngram not found in maxRefHash
        }
        
        delete hypHash_iter->first;
      } // end hyp sent ngram loop
      
      p_currHypHash->clear();
      delete p_currHypHash;
    } // end hyp sent loop
  } // end source sent loop

  m_p_sharedData->SetToComputeMatchCounts(false);
  m_p_sharedData->SetToComputeMatchGain(false);
}


// called by Evaluate() which knows if Match or Info need to be computed
bool
CEvaluatorNIST::ComputeNGramMatchStats( CText* pTranslations )
{
  
  // retrieve shared data
  CArray< CArray< CSentence* > >* p_refSentences = m_p_sharedData->GetReferences();
  assert(p_refSentences);
  CArray< CArray< CEvaluatorNGramMatchStats* > >* p_hypMatchStats 
    = m_p_sharedData->GetHypMatchStats();
  assert(p_hypMatchStats);
  
  NGram_hash_map**        pp_maxRefHash = m_p_sharedData->GetMaxRefHash();
  NGramToDouble_hash_map* p_gainHash    = m_p_sharedData->GetGainHash();
  NGram_hash_map*         p_refHash;    // used to build gainHash, then deleted
  
  // hyp-specific scores
  bool toComputeMatchCounts = m_p_sharedData->ToComputeMatchCounts();
  bool toComputeMatchGain = m_p_sharedData->ToComputeMatchGain();

  // non-hyp-specific reference stats
  bool toBuildMaxRefHash = m_p_sharedData->ToBuildMaxRefHash();
  bool toBuildGainHash = m_p_sharedData->ToBuildGainHash();

  int totalRefWords = 0;
  
  //-- Build info gain hash from ngrams in all reference sentences

  int hypSentCount = pTranslations->NumberOfSentences();
  
  if (toBuildGainHash) {
    p_refHash = new NGram_hash_map(); 
  }
  
  if (toBuildMaxRefHash) {
    pp_maxRefHash = new NGram_hash_map* [hypSentCount];
    m_p_sharedData->SetMaxRefHash(pp_maxRefHash);
  }
  
  //-- store match stats and gain per source sentence (n-best 1)
  for(int h=0; h < hypSentCount; h++)
  {
    int refSentCount = (int)((*p_refSentences)[h].Size());
    int refSentLengths = 0;

    if(toBuildMaxRefHash) {
      pp_maxRefHash[h] = new NGram_hash_map();
    }
    
    for(int r=0; r < refSentCount; r++)
    {
      CSentence* p_currRefSentence = (*p_refSentences)[h][r];
      if(toBuildMaxRefHash || toBuildGainHash)
      {
        NGram_hash_map* p_currRefHash = GetNGramsFromSentence(p_currRefSentence);
        BuildCountsHashes(h, pp_maxRefHash, p_refHash, p_currRefHash, 
                          toBuildGainHash, toBuildMaxRefHash);
        
        p_currRefHash->clear();
        delete p_currRefHash;
      }      
      
      // get avg length
      refSentLengths += p_currRefSentence->Size();
    }

    //-- save values needed for gain
    if( toComputeMatchCounts ) {
      // hypMatchStats ptr hasn't been created for [h][0]
      (*p_hypMatchStats)[h][0] = new CEvaluatorNGramMatchStats();  // n-best 1
      m_p_sharedData->SetToComputeMatchCounts(false);
    }

    CEvaluatorNGramMatchStats* p_currHypStats =  (*p_hypMatchStats)[h][0];    
    
    p_currHypStats->avgRefSentLength = (double)refSentLengths / (double)refSentCount;
    p_currHypStats->totalRefWords = refSentLengths;

    totalRefWords += refSentLengths;   //TODO: not add if !toBuildGainHash

  }
  
  if(toBuildGainHash)
  {
    // use total ref count hash to build ngram info gain hash
    p_gainHash = BuildInfoGainHash(p_refHash, totalRefWords);
    m_p_sharedData->SetGainHash(p_gainHash);
    /*
     cout << " num of unique ref ngrams: " << p_refHash->size() << " num of buckets in hash: " << p_refHash->bucket_count() << endl;
     
     cout << " num of unique gain ngrams: " << m_p_gainHash->size() << " numb of buckets in hash: " << m_p_gainHash->bucket_count() << endl;
     */
    
    p_refHash->clear();
    delete p_refHash;
  }  
  
  //-- Score hypothesis sentences
  int hypSentLength;
  
  for(int h=0; h < hypSentCount; h++)
  {
    CSentence* p_currHypSentence = pTranslations->GetSentence(h);
    CEvaluatorNGramMatchStats* p_currHypStats = (*p_hypMatchStats)[h][0];
    
    hypSentLength = p_currHypSentence->Size();
    p_currHypStats->hypSentLength = hypSentLength;
    
    NGram_hash_map* p_currHypHash = GetNGramsFromSentence(p_currHypSentence);
    
    //-- Add hyp length
    //cumScores.hypSentLength += hypSentLength; // should use numNGramsInHyp[1]   
    
    //-- Add n-gram co-occurrences and info gain
    for(NGram_hash_map::iterator hypHash_iter = p_currHypHash->begin();
        hypHash_iter != p_currHypHash->end(); hypHash_iter++)
    {
      
      NGram_hash_map::iterator maxRefHash_iter = 
      (pp_maxRefHash[h])->find(hypHash_iter->first);
      
      if(maxRefHash_iter != (pp_maxRefHash[h])->end() )
      {
        int n = (hypHash_iter->first)->getSize();
        int numMatched = std::min(hypHash_iter->second,
                                  maxRefHash_iter->second);
        
        if( toComputeMatchCounts ){
          p_currHypStats->matchingNGrams[n] += numMatched;
        }

        if( toComputeMatchGain )            
        {
          //-- add info gain for ngram--------------
          NGramToDouble_hash_map::iterator infoGainHash_iter 
          = p_gainHash->find(hypHash_iter->first);
          
          if(infoGainHash_iter != p_gainHash->end())
          {
            
            p_currHypStats->infoGain[n] 
            += infoGainHash_iter->second * (double)numMatched;
            
          } else {
            // hyp ngram not found in gainHash
          }
        }        
      } else {
        // hyp ngram not found in maxRefHash
      }
      
      delete hypHash_iter->first;
      
    } // end hyp sent ngram loop
    
    p_currHypHash->clear();
    delete p_currHypHash;
    
  } // end hyp sent loop
  
  m_p_sharedData->SetToComputeMatchGain(false);
}


//------------------------------------------------------------------
//      CEvaluatorNIST:: ReportScores
//        

bool
CEvaluatorNIST::GetOracleScore( const char *pFileWithNBestTranslations,
                                CNumericalArray< float > & Scores )
{
  cout << "CEvaluatorNIST::GetOracleScore()" << endl;
  
  CArray< CArray< CSentence* > >* p_NBestList 
    = LoadNBestTranslations(pFileWithNBestTranslations);
  
  GetOracleScoreJoy1(p_NBestList, Scores);

  return true;
}



bool
CEvaluatorNIST::GetOracleScore( CArray< CString > *pNBestTranslations,
                                CNumericalArray< float > & Scores )
{
}



bool
CEvaluatorNIST::GetOracleScore( CText *pNBestTranslations, 
                                CNumericalArray< float > & Scores )
{
}


bool
CEvaluatorNIST::GetOracleScore( CArray< CNumericalArray< CWordIndex > > *pNBestTranslations, CNumericalArray< float > & Scores )
{
}


bool
CEvaluatorNIST::GetOracleScore( CArray< CArray< CSentence* > >* p_NBestTranslations,
                                CNumericalArray< float >& Scores )
{
  // p_NBestTranslations is just like m_refSentences
  
  // run eval for every source/hyp
  // call special version of Evaluate
  // create a local version of ScoredSentence
  // 
}


//------------------------------------------------------------------
//      CEvaluatorNIST:: ReportScores
//        

void
CEvaluatorNIST::ReportScores( CNumericalArray< float > & Scores )
{
  cout << "NIST Metric Scores" << endl;
  
  for(int n=1; n<=m_maxNGram; n++){
    cout<<n<<"-gram NIST score: " << Scores[n] << endl;
  }
  
  cout << "NIST Length Penalty=" << Scores[10] << " # words in hypothesis=" 
    << Scores[11] << " avg # words in references=" << Scores[12] << endl;
  
}


//------------------------------------------------------------------
//      CEvaluatorNIST:: BuildCountsHashes
//        
//      Called on every reference sentence and updates 
//        a hash of total ref ngram counts and max ref ngram counts
//      The two hashes could be merged into one, but now their keys are pointers
//        to the same memory. The loss: hash_map overhead.
void
CEvaluatorNIST::BuildCountsHashes(int sourceSent, 
                                  NGram_hash_map** pp_maxRefHash,
                                  NGram_hash_map* p_refHash,
                                  NGram_hash_map* p_currRefHash,
                                  bool buildTotalRefCountHash,
                                  bool buildMaxRefHash)
{

  bool deleteCurrRefHash = true;

  int foundCount = 0;
  int notFoundCount = 0;
  
  int maxFoundCount = 0;
  int maxNotFoundCount = 0;
  
  for(NGram_hash_map::iterator currRefHash_iter = p_currRefHash->begin();
      currRefHash_iter != p_currRefHash->end(); currRefHash_iter++)
  {
    
    if (buildTotalRefCountHash) 
    {
      //--update overall counts-----------
      NGram_hash_map::iterator refHash_iter = p_refHash->find(currRefHash_iter->first);
      
      if(refHash_iter != p_refHash->end())
      {
        refHash_iter->second += currRefHash_iter->second;
        foundCount++;
      } else {
        p_refHash->insert(std::pair< CEvaluatorNGram*, int>
                          (currRefHash_iter->first, currRefHash_iter->second));
        // not ok to delete currRefHash_iter->first (inserted)
        deleteCurrRefHash = false;
        notFoundCount++;
      }
    }
    
    if (buildMaxRefHash) 
    {

      //--update max counts--------------
      NGram_hash_map::iterator maxRefHash_iter
      = (pp_maxRefHash[sourceSent])->find(currRefHash_iter->first);
      
      if( maxRefHash_iter != (pp_maxRefHash[sourceSent])->end() )
      {
        if(currRefHash_iter->second > maxRefHash_iter->second){
          maxRefHash_iter->second = currRefHash_iter->second;
        }
        //ok to delete ngram currRefHash_iter->first
      } else {
        (pp_maxRefHash[sourceSent])->insert( std::pair< CEvaluatorNGram*, int> (currRefHash_iter->first, currRefHash_iter->second) ); 
        
        // not ok to delete ngram currRefHash_iter->first (just inserted it)
        deleteCurrRefHash = false;
        maxNotFoundCount++;
      }
    }
    
    if(deleteCurrRefHash){
      delete currRefHash_iter->first;
    }      
    
  }
/*  
  p_currRefHash->clear();

  delete p_currRefHash;
  */
}


//------------------------------------------------------------------
//      CEvaluatorNIST:: BuildInfoGainHash
//        
//      Return a ngram->double hash_map. 
//      Built from ref ngrams->total counts hash.
//
NGramToDouble_hash_map*
CEvaluatorNIST::BuildInfoGainHash(NGram_hash_map* p_refHash, 
                                  int totalRefWords)
{

  NGramToDouble_hash_map* p_infoGainHash = new NGramToDouble_hash_map();
  
  // iterate through p_refHash
  for(NGram_hash_map::iterator refHash_iter = p_refHash->begin();
      refHash_iter != p_refHash->end(); refHash_iter++)
  {
    
    size_t currNGramSize = (refHash_iter->first)->getSize();
    // calculate info gain
    double gainScore;
    
    if( currNGramSize > 1 )
    {
      // make n-1 gram
      CEvaluatorNGram* p_shorterNGram = new CEvaluatorNGram(refHash_iter->first,
                                                            currNGramSize-1);
      
      // find count of n-1 gram
      NGram_hash_map::iterator shorterNGram_iter = p_refHash->find(p_shorterNGram);
      
      // n-1 gram valid
      gainScore = ((double)refHash_iter->second) / ((double)shorterNGram_iter->second);

      delete p_shorterNGram;
      
    } else {
      // n-1 gram not valid, should be only when n-1gram->getSize() == 0
      gainScore = ((double)refHash_iter->second) / ((double)totalRefWords);
    }
    
    gainScore = -log (gainScore) / log(2.0);
    
    p_infoGainHash->insert(std::pair<CEvaluatorNGram*, double> 
                           (refHash_iter->first, gainScore) );
  }
  
  return p_infoGainHash;

}

//------------------------------------------------------------------
//      CEvaluatorNIST:: ComputeLengthPenalty
//        

double
CEvaluatorNIST::ComputeLengthPenalty(double ratio)
{
  if( ratio >= 1 )
    return 1;
  if( ratio <= 0 )
    return 0;
  
  double ratio_x = 1.5;
  double score_x = 0.5;
  
  double beta = -log(score_x) / log(ratio_x) / log(ratio_x);

  return exp (-beta * log(ratio) * log(ratio) );
}

//---Get Methods---------------------------------------------
/*
CArray< CArray< CEvaluatorNGramMatchStats > >* 
CEvaluatorNIST::GetNGramMatchInfoStats()
{
  return m_p_hypMatchInfoStats;
}

NGram_hash_map**
CEvaluatorNIST::GetReferenceNGramMaxCountHash()
{
//  return m_pp_maxRefHash;
}

NGramToDouble_hash_map*
CEvaluatorNIST::GetGainHash()
{
  return m_p_gainHash;
}

*/

//---Set Methods---------------------------------------------
/*
void
CEvaluatorNIST::SetReferenceNGramMaxCountHash(NGram_hash_map** pp_maxRefHash)
{
  m_pp_maxRefHash = pp_maxRefHash; 
}

void 
CEvaluatorNIST::SetGainHash(NGramToDouble_hash_map* p_gainHash)
{
  m_p_gainHash = p_gainHash;
}

void 
CEvaluatorNIST::SetNGramMatchInfoStats(CArray< CArray< CEvaluatorNGramMatchStats > >* 
                                       p_hypMatchInfoStats)
{
  m_p_hypMatchInfoStats = p_hypMatchInfoStats;
}

*/



