#ifndef __EvaluatorPER__
#define __EvaluatorPER__


#include <Evaluator.hh>



class CEvaluatorPER : public CEvaluator
{
public:
  // construct an evaluator; give the vocabulary which will be used 
  // to map any word to a word ID
  CEvaluatorPER( CVocabulary *pVocabulary );
  ~CEvaluatorPER();

  // load references is in base class
  //bool LoadReferences( const char *pFileName, int Format = 1 );
  
  // overload methods to calculate the score or scores for the translation
  // scores to be returned:  PER, insertions, deletions, substitutions, matches, length translations, length references
  bool Evaluate( const char *pFileWithTranslations, CNumericalArray< float > & Scores );
  bool Evaluate( CArray< CString > &Translations, CNumericalArray< float > & Scores );
  bool Evaluate( const CText *pTranslations, CNumericalArray< float > & Scores );
  bool Evaluate( const CArray< CNumericalArray< CWordIndex > > *pTanslations, CNumericalArray< float > & Scores );
  bool Evaluate( CSentence* pHypSentence, CArray< CSentence* > &References,
                CNumericalArray< float >& Scores );
  // calculate the oracle score for a nbest list; Format: n m word sequence
  // oracle scores means: take those translations in nbest lists which optimize the scores
  // no matter if read from file, or getting the translations in arrays
  bool GetOracleScore( const char *pFileWithNBestTranslations, 
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( CArray< CString > *pNBestTranslations, 
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( CText *pNBestTranslations, 
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( CArray< CNumericalArray< CWordIndex > > *pNBestTranslations,
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( CArray< CArray< CSentence* > >* p_NBestTranslations,
                       CNumericalArray< float >& Scores ) = 0;

  void ReportScores( CNumericalArray< float > & Scores );

protected:
  // references are for n source sentences, m references, in base class
  // CArray< CArray< CNumericalArray< CWordIndex > > > *pReferences;


};





#endif
