//
// WER Evaluation Metric
// Anthony D'Auria <apd@andrew.cmu.edu>
// 
//
//

#ifndef __EvaluatorWER__
#define __EvaluatorWER__

#include <Evaluator.hh>


class CEvaluatorWER : public CEvaluator
{
public:
  // construct an evaluator; give the vocabulary which will be used 
  // to map any word to a word ID
  //  CEvaluatorWER();
  CEvaluatorWER( CVocabulary *pVocabulary, CEvaluatorSharedData *pSharedData );
  ~CEvaluatorWER();

  // load references is in base class
  //bool LoadReferences( const char *pFileName, int Format = 1 );
  
  // overload methods to calculate the score or scores for the translation
  // scores to be returned:  WER, insertions, deletions, substitutions, matches,
	//	length translations, length references
  bool Evaluate( const char *pFileWithTranslations, CNumericalArray< float > & Scores );
  bool Evaluate( CArray< CString > &Translations, CNumericalArray< float > & Scores );
  virtual bool Evaluate( int sourceSentenceIndex, CSentence* pHypSentence, 
                         CNumericalArray< float >& Scores);
	//  bool Evaluate( const CText *pTranslations, CNumericalArray< float > & Scores );
	bool Evaluate( CText *pTranslations, CNumericalArray< float > & Scores );
  bool Evaluate( const CArray< CNumericalArray< CWordIndex > > *pTranslations, CNumericalArray< float > & Scores );
  bool Evaluate( CSentence* pHypSentence, CArray< CSentence* > &References,
                CNumericalArray< float >& Scores );
  // calculate the oracle score for a nbest list; Format: n m word sequence
  // oracle scores means: take those translations in nbest lists which optimize the scores
  // no matter if read from file, or getting the translations in arrays
  bool GetOracleScore( const char *pFileWithNBestTranslations, 
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( CArray< CString > *pNBestTranslations, 
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( CText *pNBestTranslations, 
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( CArray< CNumericalArray< CWordIndex > > *pNBestTranslations, 
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( CArray< CArray< CSentence* > >* p_NBestTranslations,
                       CNumericalArray< float >& Scores );

  std::pair<float, float> GetConfidenceInterval();
  
  void ReportScores( CNumericalArray< float > & Scores );

protected:
  // references are for n source sentences, m references, in base class
  // CArray< CArray< CNumericalArray< CWordIndex > > > *pReferences;

  // PER compares bag of words
  // CArray< CArray< CNumericalArray< CWordIndex > > > *pSortedWords
	float	minimum (float a, float b, float c);

	// used as indices for Scores[] by various Evaluator classes
	int WER;
	int MATCH;
	int SUBSTITUTION;
	int INSERTION;
	int DELETION;
	int HYP_SENTENCE_COUNT;
	int REF_WORD_COUNT;
	int HYP_WORD_COUNT;
  int LEAST_DISTANCE;
  
	
};





#endif
