#include <string.h>
#include <config.h>
//#include <fortify.h>
// $Id: TM_Graph_HMM.cc 1150 2005-04-13 18:18:37Z vogel $
//
// Translation Model CTM_Graph_HMM
// HMM-like word alignment
//


#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include "Trace.hh"
#include "TM_Graph_IBM1.hh"
#include "TM_Graph_HMM.hh"


//static char rcsid_tm_HMM_cc[] = "$Id";


CTM_Graph_HMM::CTM_Graph_HMM()
{
  m_pLatticeE = NULL;
  m_pLatticeF = NULL;

  m_pTransMemE = NULL;
  m_pTransMemF = NULL;
  m_pTransMemNgram = NULL;

  m_pLatticeEdgeTrans = new CLatticeEdgeTrans;

  m_pPathF = new CNumericalArray< CLatticeEdge * >;
  m_pPathE = new CNumericalArray< CLatticeEdge * >;

  m_pAlignment = NULL;

  //cout << rcsid_tm_HMM_cc << endl;
  //m_SmallJumpLimit = 10;     // only small jumps <= abs(SmallJumpLimit) allowed
  m_PenaltySum = 0.0;
  m_TPenaltySum = 0.0;
  m_APenaltySum = 0.0;

  //m_NumOfFWords = 0;
  m_DisplayPath = FALSE;
  m_PrintAlignments = FALSE;
  
  m_HMMAlignForward = TRUE;

  m_pBiTextFileName = NULL;
  m_pGrammarOutFileName = NULL;

}


CTM_Graph_HMM::~CTM_Graph_HMM()
{
  delete m_pTransMemE;
  delete m_pTransMemF;
  delete m_pLatticeEdgeTrans;
  delete m_pPathF;
  delete m_pPathE;
  delete m_pAlignment;
}



CAlignProbRel &
CTM_Graph_HMM::AlignProbs()
{
  return m_AlignProb;
}


void
CTM_Graph_HMM::SetTransMemParamFiles( CString ParamFileF, CString ParamFileE, CString ParamFileNgram )
{
  m_TransMemParamFileF = ParamFileF;
  m_TransMemParamFileE = ParamFileE;
  // akt 08.02
  m_TransMemParamFileNgram = ParamFileNgram;


  if ( m_TransMemParamFileF.GetLength() > 0 ) {
    m_pTransMemE = new CTranslationMemory;
    m_pTransMemF = new CTranslationMemory;
  }
  if (m_TransMemParamFileNgram.GetLength() > 0 ) {
    m_pTransMemNgram = new CTranslationMemory;
  }
}


void
CTM_Graph_HMM::SetHMMTrainingDirection( CString AlignDirection ){
  if( strcmp(AlignDirection, "Forward") == 0 ){
    m_HMMAlignForward = TRUE;
  }
  else{ m_HMMAlignForward = FALSE;}
}


void
CTM_Graph_HMM::SetBiTextFileName( char *pBiTextFileName )
{
  Trace( 1, "CTM_Graph_HMM::SetBiTextFileName " << pBiTextFileName );

  m_pBiTextFileName = pBiTextFileName;
}


/**
   @brief Set the name of the GrammarFile.

   GrammarFile phrases will be written there. If no GrammarFile has been
   provided until training starts, no phrases will be extracted.

   This method does not strcpy the character string but only copies
   the pointer. Hence, do not use a temporary string here and note that
   GrammarFile name will change if you modify the string afterwards.

   \todo Get rid of that behaviour and use std::string here.
 */
void
CTM_Graph_HMM::SetGrammarOutFileName( char *pGrammarOutFileName )
{
  Trace( 1, "CTM_Graph_HMM::SetGrammarOutFileName " << pGrammarOutFileName );

  // TODO: this should be strcpy
  m_pGrammarOutFileName = pGrammarOutFileName;
}


bool
CTM_Graph_HMM::Train()
{
  Trace( 0, "------------------------------" );
  Trace( 0,  " Train Translation Model HMM-Graph" );


  assert( m_pTextF );
  assert( m_pTextE );
  assert( m_pLexicon );

  // init translation memories
  Trace( 1, "init translation memories" );
  if ( m_pTransMemF ) {
    
    // akt
    // These don't ever exist!!! - td's were used in training
    // in Verbmobile but are NOT used this way in current implement.
    // but the vocabularies do, so don't worry...
    //cout << endl << "should have a source vocab now" << endl ;
    m_pTransMemE->Init( m_TransMemParamFileE, m_pTextE->Vocabulary() );
    m_pTransMemF->Init( m_TransMemParamFileF, m_pTextF->Vocabulary() );
  }
  
  // akt 08.02
  // Note: using the Source vocabulary for now since the Ngrams will be 
  // taken over the source side for the default S->T alignment...
  //
  // 09.2002 conditioning on alignment direction now
  if ( m_pTransMemNgram ) {
    
    if( m_HMMAlignForward ){
	//cout << endl << "we think we have a m_pTransMemNgram" << endl;
	m_pTransMemNgram->Init( m_TransMemParamFileNgram, m_pTextF->Vocabulary() );
    }
    else{
   	m_pTransMemNgram->Init( m_TransMemParamFileNgram, m_pTextE->Vocabulary() );   
    }
  }



  // init lexicon with vocabulary
  //m_pLexicon->Init( m_pTextF->Vocabulary(), m_pTextE->Vocabulary() );
  //m_pLexicon->Display();

  // set smoothing parameter
  //m_pLexicon->SetLambdaSmooth( m_LambdaSmooth );
  m_pLexicon->SetLambdaSmooth( 0.1 );

  // do we use cognates, then modify counts
  //m_pLexicon->UpdateCountsForCognates();

  // normalize the translation probabilities
  //Trace( 1, "Renormalize Lexicon" );
  //m_pLexicon->DisplayCounts();
  //m_pLexicon->Renormalize();
  //m_pLexicon->TestProbabilities();
  //m_pLexicon->Display();

  // calculate logarithm
  m_pLexicon->Log();

  //m_AlignProb.Init();
  //m_AlignProb.TestProbabilities();

  //m_AlignProb.DisplayPr();
  //m_AlignProb.PreparePr( 5, 6 );

  //return TRUE;

  if ( m_pBiTextFileName )
    m_BiTextFile.open( m_pBiTextFileName );

  m_BiRulesFile.open( "NewTransducerRules" );

  if ( m_pGrammarOutFileName )
    m_GrammarFile.open( m_pGrammarOutFileName );


  // iterate
  for ( m_CurrentIteration = 0; 
	m_CurrentIteration <= m_Iterations; 
	m_CurrentIteration++ ) {
    //Trace( 1, "\n\n== Iteration " << m_CurrentIteration << " ==\n" );
    cout << "Iteration " << m_CurrentIteration << "\n";

    m_TPenaltySum = 0.0;
    m_APenaltySum = 0.0;
    m_PenaltySum = 0.0;
    m_NumOfFWords = 0;
    m_AlignProb.ClearCounts();

	delete m_pAlignment;
	m_pAlignment = new CAlignment;
	m_pAlignment->SetIncrement( m_pTextF->NumberOfSentences() + 5 );
	m_pAlignment->SetText( m_pTextF, m_pTextE );

    NextIteration();
    // NextIteration calculates perplexity of the model
    // as it was before the current iteration
    //PP = Perplexity();


    m_LexiconPerplexity[m_CurrentIteration] = exp(m_TPenaltySum / m_NumOfFWords);
    m_AlignPerplexity[m_CurrentIteration] = exp(m_APenaltySum / m_NumOfFWords);
    m_TotalPerplexity[m_CurrentIteration] = exp(m_PenaltySum / m_NumOfFWords);

    cout << "Perplexity in Iteration " << m_CurrentIteration << "\n";
    cout << "Lexicon     Jump      Total\n";
    cout << m_LexiconPerplexity[m_CurrentIteration] << "  "  
	 << m_AlignPerplexity[m_CurrentIteration] << "  " 
	 << m_TotalPerplexity[m_CurrentIteration] << "\n";

    cout.flush();



    //CString S1( "HMM.Align_" );
    //CString S2( (int)m_CurrentIteration );
    //CString S3 = S1 + S2;
    //Trace( 0, "Store global alignment in " << (const char *) S3 );
    //StoreAlignment( (const char *)S3 );

  }


  // display all perplexities
  cout.setf(ios::right | ios::fixed);
  //cout.width(6);
  cout.precision(2);
  cout << "Perplexity as Function of Iteration \n\n";
  cout << "Iteration   &  Lexicon  &   Jump   &   Total \\\\ \\hline \n";
  for ( int Iteration = 0; Iteration <= m_Iterations; Iteration++ ) {
    cout << Iteration << " & " 
	 << m_LexiconPerplexity[Iteration] << " & "  
	 << m_AlignPerplexity[Iteration] << " & "  
	 << m_TotalPerplexity[Iteration] << "\n";
  }
  

  m_BiTextFile.close();
  m_BiRulesFile.close();
  m_GrammarFile.close();

  return TRUE;

}


bool
CTM_Graph_HMM::NextIteration()
{
  //cout << "CTM_Graph_HMMnoI::NextIteration \n" << flush;
  Trace(1, "CTM_Graph_HMM::NextIteration ");

  m_pLexicon->ClearCounts();
  if ( m_pTransMemE ) {
    m_pTransMemE->ClearTransducerCounts();
  }

  CWordIndex Se, Sf;
  //TPosition PosSe, PosSf;
  //int SentenceLengthE = 0, SentenceLengthF = 0;

  CString TranslationE, TranslationF;

  // fetch first sentence pair
  Se = m_pTextE->FirstSentence( m_CurrentPosSe );
  Sf = m_pTextF->FirstSentence( m_CurrentPosSf );

  m_S = 0;
  while (Se != m_pTextE->Vocabulary()->EOT && 
	 Sf != m_pTextF->Vocabulary()->EOT)   {
    if (m_S % 1000 == 0) {
      cout << "Sentence pair " << m_S  << endl << flush;
      //      Trace( 1, "Sentence pair " << m_S );
    }

	if ( TraceLevel > 0 ) {
	  cout << "------------------------------" << endl;
	  cout << "Sentence pair " << m_S  << endl << flush;
	}

	m_S++;

    m_pSentenceE = m_pTextE->GetSentence( m_CurrentPosSe );
    m_pSentenceF = m_pTextF->GetSentence( m_CurrentPosSf );
    
    // get length of sentences; if one is zero, skip
    m_I = m_pSentenceE->Size();
    m_J = m_pSentenceF->Size();

	if ( TraceLevel > 0 ) {
	  //cout << m_J << " ";
	  m_pSentenceF->Display( m_pTextF->Vocabulary(), FALSE ); cout << endl;
	  //cout << m_I << " ";
	  m_pSentenceE->Display( m_pTextE->Vocabulary(), FALSE ); cout << endl;
	}

    if ( (m_I > 0) && (m_J > 0) ) {

      // Convert sentence into lattice

      Trace( 1, "Convert sentences into lattices and apply transducer" );

      if ( m_S <= 5 ) {
	m_pSentenceF->Display( m_pTextF->Vocabulary(), FALSE ); cout << endl;
        m_pSentenceE->Display( m_pTextE->Vocabulary(), FALSE ); cout << endl;
      }
      
      m_pLatticeE = CTM_Graph_IBM1::ConvertToLattice( m_pSentenceE, m_pTextE->Vocabulary() );
      m_pLatticeF = CTM_Graph_IBM1::ConvertToLattice( m_pSentenceF, m_pTextF->Vocabulary() );
      
      //if ( m_S <= 5 ) {
      //m_pLatticeF->Display();
      //m_pLatticeE->Display();
      //}
 
      //TranslationF = m_pTransMemF->TranslateLattice( m_pLatticeF );
      //TranslationE = m_pTransMemE->TranslateLattice( m_pLatticeE );

      if ( m_pTransMemF && m_pTransMemE ) {

	for ( int l = m_pTransMemF->FirstLevel(); l <= m_pTransMemF->LastLevel(); l++ ) {
	  Trace( 1, "Apply transducer: source" );
	  m_pTransMemF->SetDirection( 0 );
	  m_pTransMemF->ApplyTransducer( m_pLatticeF, l );
	  //if ( m_S <= 5 ) {
	  //m_pLatticeF->Display();
	  //}
      
	  Trace( 1, "Apply transducer: target" );
	  m_pTransMemF->SetDirection( 1 );
	  m_pTransMemE->ApplyTransducer( m_pLatticeE, l );
	  //if ( m_S <= 5 ) {
	  //m_pLatticeE->Display();
	  //}

	  //PruneLattices();
	  Trace( 0, "Edges in f-lattice: " << m_pLatticeF->NumberOfEdges() );
	  Trace( 0, "Edges in e-lattice: " << m_pLatticeE->NumberOfEdges() );
	  
	  if ( m_pLatticeF->NumberOfEdges() > 10000 ) {
	    m_pLatticeF->Display();
	  }
	  if ( m_pLatticeE->NumberOfEdges() > 10000 ) {
	    m_pLatticeE->Display();
	  }

	  RemoveUnmatchedEdges();
	  
	  Trace( 0, "Edges in f-lattice: " << m_pLatticeF->NumberOfEdges() );
	  Trace( 0, "Edges in e-lattice: " << m_pLatticeE->NumberOfEdges() );

	  //m_pLatticeF->Display();
	  //m_pLatticeE->Display();

	}

      }

      //ClearPenaltiesAndBacktrack();
      CalculatePathes();
      //if ( m_S <= 5 ) {
	//DisplayQ(m_J, m_I);
	//DisplayB(m_J, m_I);
      //}
      
      //TraceBack();
      //if ( m_S <= 50 ) {
      //	DisplayBestPath();
      //}
      
      // for the first sentences - just as a control
      //if (m_DisplayPath && (m_S < 10)) {
      //if ( m_S <= 100 ) {
        //DisplayQ(m_J, m_I);
        //DisplayB(m_J, m_I);
        //DisplayBestPath();
		//}
      
      delete m_pLatticeE;
      delete m_pLatticeF;

    }
      
    // fetch next sentence pair
    Se = m_pTextE->NextSentence(m_CurrentPosSe);
    Sf = m_pTextF->NextSentence(m_CurrentPosSf);
  } // end loop senteces
  
  // do we use cognates, then modify counts
  m_pLexicon->UpdateCountsForCognates();

  // normalize the translation probabilities
  Trace( 1, "Renormalize Lexicon" );
  //m_pLexicon->DisplayCounts();
  m_pLexicon->Renormalize();
  m_pLexicon->TestProbabilities();
  //m_pLexicon->Display();

  //m_pLexicon->SetLambdaSmooth( m_LambdaSmooth );
  m_pLexicon->Log();

  // normalize the alignment probabilities
  Trace( 1, "Renormalize Alignment" );
  m_AlignProb.DisplayCounts();
  m_AlignProb.Renormalize();
  //m_AlignProb.TestProbabilities();
  //m_AlignProb.DisplayProb();

  //m_pTransMemE->DisplayTransducer();
  if ( m_pTransMemE ) {
    m_pTransMemE->RenormalizeTransducer();
    //m_pTransMemE->DisplayTransducer();
    m_pTransMemE->StoreTransducer();
  }

  return TRUE;
}




////////////////////////////////////////////////////////////////////////
// Initalization
////////////////////////////////////////////////////////////////////////

void
CTM_Graph_HMM::InitFields()
{
  //uint i, j;

  // cout << "Init fields\n";

  // for (j = 0; j < m_MaxSLen; j++) {
  //   for (i = 0; i < m_MaxSLen; i++) {
  //     m_Penalty[j][i] = 0.0;
  //     m_BackTrack[j][i] = 0;
  //     m_JCountsPos[j][i] = 0;
  //     m_JCountsNeg[j][i] = 0;
  //   }
  // }
  
}



////////////////////////////////////////////////////////////////////////
// Calculation
////////////////////////////////////////////////////////////////////////



void
CTM_Graph_HMM::CalculatePathes()
{ //CalculatePathes() def

  // aligning two lattices

  CWordIndex We, Wf; 		  // index of word in vocabulary
  //TPosition PosWe, PosWf;         // position of word in text
  int i, j;                  // position of word within the sentences
  double TPenalty;                // -ln t(f|e)
  //double APenalty;                // -ln A(j|i)
  //double NewPenalty;              // used in loop over previous column
  //double MinPenalty;              // the minimal penalty in loop over previous column

  const char *pWordF;
  const char *pWordE;

  Trace( 1, "\nCalculatePathes: " << m_CurrentPosSf << " " << m_CurrentPosSe <<endl );

  //cout << "CalculatePathes" << endl;
  //cout.flush();

  //m_I = m_pTextE->SentenceLength( m_CurrentPosSe );
  //m_J = m_pTextF->SentenceLength( m_CurrentPosSf );
  Trace( 3,  "m_I: " << m_I << " m_J: " << m_J );
  // if one sentence is empty -> nothing to do
  //if ( m_I == 0 || m_J == 0 )
  // return;


  // i.e. loop over all nodes, for each node over all incoming edges
  const CArray< CLatticeNode * > & NodesE = m_pLatticeE->Nodes();
  const CArray< CLatticeNode * > & NodesF = m_pLatticeF->Nodes();
  
  CLatticeNode *pNodeF, *pNodeE, *pNodeF1, *pNodeE1;

  CLatticeEdgeTrans *pEdgeE, *pEdgeF;

  assert( m_J == NodesF.MaxIndex() );
  assert( m_I == NodesE.MaxIndex() );


  // init scores and backtrace fields
  for ( j = NodesF.MaxIndex(); j >= NodesF.MinIndex(); j-- ) {
    for ( i = NodesE.MaxIndex(); i >= NodesE.MinIndex(); i-- ) {
      m_L( j, i ) = 10000;
      m_C( j, i ) = 0;
      m_H( j, i ) = 0;
      m_Q( j, i ) = 1.0e+10;
      m_B( j, i ) = -1;
      m_EdgeF( j, i ) = NULL;
      m_EdgeE( j, i ) = NULL;
    }
  }
  m_L( NodesF.MinIndex(), NodesE.MinIndex() ) = 0;
  m_L( NodesF.MaxIndex() + 1, NodesE.MaxIndex() + 1 ) = 10000;
  m_Q( NodesF.MinIndex(), NodesE.MinIndex() ) = 0.0;
  m_Q( NodesF.MaxIndex() + 1, NodesE.MaxIndex() + 1 ) = 1.0e+10;


  // first, calculate the JPenalties; i.e. renormalize and store them 
  // for all positions and all possible jumps from that position
  // have to know the sentence length
  m_AlignProb.PreparePr( m_J, m_I );
  //m_AlignProb.DisplayProb();

  int j1, iprev;
  int Length_e, Length_f;
  int Level;

  for ( j = NodesF.MinIndex() + 1; j <= NodesF.MaxIndex(); j++ ) {
    pNodeF = NodesF[ j ];
    const CArray< CLatticeEdge * > & EdgesF = pNodeF->InEdges();

    for ( i = NodesE.MinIndex() + 1; i <= NodesE.MaxIndex(); i++ ) {
      pNodeE = NodesE[ i ];
      const CArray< CLatticeEdge * > & EdgesE = pNodeE->InEdges();

      int NumEdgeF, NumEdgeE;

      if ( TraceLevel > 0 ) {
	Trace( 1, "Nodes " << j << " " << i );
	for ( NumEdgeF = EdgesF.MinIndex(); NumEdgeF <= EdgesF.MaxIndex(); NumEdgeF++ ) {
	  pEdgeF = (CLatticeEdgeTrans *)EdgesF[ NumEdgeF ];
	  //Trace0( 0, "f-edge: " );
	  //pEdgeF->Display();
	  //if ( pEdgeF->TransducerItem() )
	  //pEdgeF->TransducerItem()->Display(); 
	}
	
	for ( NumEdgeE = EdgesE.MinIndex(); NumEdgeE <= EdgesE.MaxIndex(); NumEdgeE++ ) {
	  pEdgeE = (CLatticeEdgeTrans *)EdgesE[ NumEdgeE ];
	  //Trace0( 0, "e-edge: " );
	  //pEdgeE->Display();
	  //if ( pEdgeE->TransducerItem() )
	  //pEdgeE->TransducerItem()->Display();
	}
      }

      for ( NumEdgeF = EdgesF.MinIndex(); NumEdgeF <= EdgesF.MaxIndex(); NumEdgeF++ ) {
	pEdgeF = (CLatticeEdgeTrans *)EdgesF[ NumEdgeF ];
	Wf = pEdgeF->WordIndex();

	if ( Wf == 0 ) {
	  Trace(2, "UNK " << pEdgeF->Word() );
	  Wf = m_pTextF->Vocabulary()->GetIndex( pEdgeF->Word() );
	  Trace( 2, "Set index to " << Wf );
	  pEdgeF->SetWordIndex( Wf );
	  Wf = pEdgeF->WordIndex();
	  Trace(2, "now " << m_pTextF->Vocabulary()->GetWord( Wf ) );
	}
	pWordF = m_pTextF->Vocabulary()->GetWord( Wf );

	pNodeF1 = pEdgeF->FromNode();
	j1 = pNodeF1->Number();

	Length_f = pEdgeF->ToNode()->Number() - pEdgeF->FromNode()->Number();

	// level of transducer from which edge was generated;
	if ( pEdgeF->Transducer() ) {
	  Level = pEdgeF->Transducer()->Level();
	} else {
	  Level = 0;
	}


	for ( NumEdgeE = EdgesE.MinIndex(); NumEdgeE <= EdgesE.MaxIndex(); NumEdgeE++ ) {
	  pEdgeE = (CLatticeEdgeTrans *)EdgesE[ NumEdgeE ];
	  //Trace0( 0, "f-edge: " ); pEdgeF->Display();
	  //Trace0( 0, "e-edge: " ); pEdgeE->Display();
	  We = pEdgeE->WordIndex();
	  if ( We == 0 ) {
	    Trace( 2, "UNK " << pEdgeE->Word() );
	    We = m_pTextE->Vocabulary()->GetIndex( pEdgeE->Word() );
	    Trace( 2, "Set index to " << We );
	    pEdgeE->SetWordIndex( We );
	  }
	  pWordE = m_pTextE->Vocabulary()->GetWord( We );
	  pNodeE1 = pEdgeE->FromNode();
	  Length_e = pEdgeE->ToNode()->Number() - pEdgeE->FromNode()->Number();

	  // lexicon probability:
	  // for transducer added edges from same transducer item: TPenalty == log(0.99999)
	  // else: lexicon prob
	  if ( (pEdgeF->TransducerItem() != NULL) && (pEdgeE->TransducerItem() != NULL) ) {

	    // calculation of score from application of transducers may involve
	    // recursive descend
	    TPenalty = CalcTransducerScore( pEdgeF, pEdgeE );
	    if ( TPenalty < 100 ) {
	      ;
	      Trace( 1, "Total Score = " << TPenalty );
	    }

	  } else if ( (pEdgeF->TransducerItem() != NULL) || 
		      (pEdgeE->TransducerItem() != NULL) ) {
	    // only one but not the other generated through transducer
	    TPenalty = 1e10;
	    Trace( 1, "Only one edge from transducer" );
	    Trace( 1, "TPenalty for label = " << TPenalty );

	  } else {
	    // word-to-word, normal lexicon probability
	    //Trace( 0, "Wf - We: " << pEdgeF->Word() << " " << pEdgeE->Word() );
	    TPenalty = -m_pLexicon->LogPr( Wf, We ); //log(m_pT->GetAt(We,Wf));
	    Trace( 1, "TPenalty from lexicon = " << TPenalty );
	  }

	  Trace( 1, "TPenalty = " << TPenalty );

	  // if TPenalty to large, i.e. edges generated from transducers
	  // but not from same item, then do nothing 
	  if ( TPenalty < 1e10 ) {

	    if ( j1 == NodesF.MinIndex() ) {
	      // from start of lattice, i.e. first alignment 
	      iprev = 0;
	      CalcTransition( j1, iprev, j, i, Length_f, Length_e, Level, TPenalty, pEdgeF, pEdgeE );
	    } else {
	      // coming from position within sentence f
	      for ( iprev = NodesE.MinIndex() + 1; iprev <= NodesE.MaxIndex(); iprev++ ) {
		CalcTransition( j1, iprev, j, i, Length_f, Length_e, Level, TPenalty, pEdgeF, pEdgeE );
	      }
	    }

	  }

	}
      }

    }
  }

  // jump to (EOS, EOS)
  j = m_J + 1;
  i = m_I + 1;

  // loop over previous column and calculate the minimum of penalty 
  // at position (j - 1, pre) plus penalty due to jump in alignment
  m_EdgeF( j, i ) = NULL;
  m_EdgeE( j, i ) = NULL;
  
  for ( iprev = NodesE.MinIndex() + 1; iprev <= NodesE.MaxIndex(); iprev++ ) {
    Trace( 2, "At the End of the  Lattice " << j 
	   << " " << i << " " << iprev );

    bool Better = FALSE;
    if ( m_L( j - 1, iprev ) <= m_L( j, i ) ) {
      if ( m_L( j - 1, iprev ) < m_L( j, i ) ) {
	Better = True;
      } else {
	if ( m_C( j - 1, iprev ) >= m_C( j, i ) ) {
	  if ( m_C( j - 1, iprev ) > m_C( j, i ) ) {
	    Better = True;
	  } else {
	    if ( m_H( j - 1, iprev ) >= m_H( j, i ) ) {
	      if ( m_H( j - 1, iprev ) > m_H( j, i ) ) {
		Better = True;
	      } else if ( m_Q( j - 1, iprev ) < m_Q( j, i ) ) {
		Better = True;
	      }
	    }
	  }
	}
      }
    }
  
    if ( Better) {
      m_L( j, i ) = m_L( j - 1, iprev );
      m_C( j, i ) = m_C( j - 1, iprev );
      m_H( j, i ) = m_H( j - 1, iprev );
      m_Q( j, i ) = m_Q( j - 1, iprev );
      m_B( j, i ) = iprev;
    }
    
  }
  
  Trace( 2, "Best path: L= " << m_L( j, i ) << " C= " << m_C( j, i ) 
	 << " H= " << m_H( j, i ) << " Q= " << m_Q( j, i ) );

  
  //DisplayQ(m_J, m_I);
  //DisplayB(m_J, m_I);
  TraceBack();
  //DisplayLabeledBiText();
  if ( m_CurrentIteration == m_Iterations ) {
    //if ( m_pTransMemF ) {
      WriteLabeledBiText();
	  //}
      // Check if GrammarFile is open and good for write
      if (m_GrammarFile.good()) {
	if( m_HMMAlignForward ){ 
	ExtractForwardBilingualPatterns();
	//ExtractBilingualGrammar();
	} else {
	  ExtractReverseBilingualPatterns();
	}
      } // endif: GrammarFile open

	//WriteAlignment();
  }
  
}




// Calculate the transition in the trellis over the edges pEdgeF and pEdgeE
// with a previous jump from iprev
void
CTM_Graph_HMM::CalcTransition( int jprev, int iprev, int j, int i, 
			       int Length_f, int Length_e, int Level, double TPenalty,
			       CLatticeEdgeTrans *pEdgeF, CLatticeEdgeTrans *pEdgeE )
{
  double Q;
  int L, C, H;
  bool Better;
  double JPenalty;                // -ln Jump(distance)

  JPenalty = -m_AlignProb.LogPr( i, iprev, j );
  Trace( 5, "Qprev = " << m_Q( jprev, iprev ) << "   JPenalty " << JPenalty );
  
  // translation model score
  Q = m_Q( jprev, iprev ) + TPenalty + JPenalty;
  
  // length, coverage and cummulative level of edge f
  L = m_L( jprev, iprev ) + 1;
  C = m_C( jprev, iprev ) + Length_e;
  H = m_H( jprev, iprev ) + Level * Length_f;

  Trace( 1, "CTM_Graph_HMM::CalcTransition " << jprev << " " << iprev << " " << j << " " << i );
  //pEdgeF->Display();
  //pEdgeE->Display();
  Trace( 1, "Prev: Length = " << m_L( jprev, iprev ) << " Coverage = " 
	 << m_C( jprev, iprev ) << " Height = " << m_H( jprev, iprev ) );
  Trace( 1, "Now:  Length = " << L << " Coverage = " << C << " Height = " << H );
  
  Better = FALSE;
  if ( L <= m_L( j, i ) ) {
    if ( L < m_L( j, i ) ) {
      Trace0( 1, "Shorter Path  " );
      Better = True;
    } else {
      if ( C >= m_C( j, i ) ) {
	if ( C > m_C( j, i ) ) {
	  Trace0( 1, "Larger Coverage  " );
	  Better = True;
	} else {
	  if ( H >= m_H( j, i ) ) {
	    if ( H > m_H( j, i ) ) {
	      Trace0( 1, "Higher Level  " );
	      Better = True;
	    } else if ( Q < m_Q( j, i ) ) {
	      Trace0( 1, "Better Score  " );
	      Better = True;
	    }
	  }
	}
      }
    }
  }
  
  if ( Better) {
    m_L( j, i ) = L;
    m_C( j, i ) = C;
    m_H( j, i ) = H;
    m_Q( j, i ) = Q;
    m_B( j, i ) = iprev;
    m_EdgeF( j, i ) = pEdgeF;
    m_EdgeE( j, i ) = pEdgeE;
    
    //Trace( 5, i << " NewMinPenalty = " << MinPenalty 
    //	   << " Backpointer = " << i  );
    Trace( 1, "Best path in (" << j << "," << i << "): L= " << m_L( j, i ) << " C= " << m_C( j, i ) 
	   << " H= " << m_H( j, i ) << " Q= " << m_Q( j, i ) );
  }
  
}



double
CTM_Graph_HMM::CalcTransducerScore( CLatticeEdgeTrans *pEdgeF, CLatticeEdgeTrans *pEdgeE )
{
  assert( pEdgeF->TransducerItem() );
  assert( pEdgeE->TransducerItem() );


  Trace( 1, "CTM_Graph_HMM::CalcTransducerScore: " );
  CLatticeEdgeTrans *pSubEdgeF, *pSubEdgeE;

  double Penalty, NextPenalty, SumPenalty;

  if ( TraceLevel > 0 ) {
    pEdgeF->Display();
    pEdgeE->Display();

    if ( pEdgeF->TransducerItem() )
      pEdgeF->TransducerItem()->Display();

    if ( pEdgeE->TransducerItem() )
      pEdgeE->TransducerItem()->Display();
  
  }
  

  // first, check if current edges were generated by same transducer item
  // if not, high score and return
  // if yes, recursive descend

  CString Tf = pEdgeF->Transducer()->Name();
  CString Te = pEdgeE->Transducer()->Name();
  Trace( 1, "Transducers: " << "-" << Tf << "-" << Te );

  int Nf = pEdgeF->TransducerItem()->Number();
  int Ne = pEdgeE->TransducerItem()->Number();
  Trace( 1, "Transducer item: " << Nf << " " << Ne );

  if ( (Tf != Te) || (Nf != Ne) ) {
    Trace( 1, "Not Same transducer and same item: " 
	   << Tf << " " << Nf << " : " << Te << " " << Ne );
    SumPenalty = 1e10;

  } else {

    if ( TraceLevel > 0 ) {
      Trace( 1, "Same transducer and same item: " << Tf << " " << Te );
      Trace0( 1, "e-edge: " ); pEdgeE->Display();
      Trace0( 1, "f-edge: " ); pEdgeF->Display();
      Trace( 1, ">>> TPenalty from transducer = " 
	     << -log( pEdgeE->TransducerItem()->ScoreSourceGivenTarget() ) );
      //Trace( 0, ">>> TPenalty from transducer = " 
      //   << pEdgeF->WordInfo()->Score() );
    }

    // translation score from top-level transducer item
    SumPenalty = -log( pEdgeE->TransducerItem()->ScoreSourceGivenTarget() );
    //SumPenalty = pEdgeF->WordInfo()->Score();


    // collect score from underlying transducer items
    // for each generating edge e find the best matching generating edge f
    // i.e. the one which was generated by the same transducer item
    
    CArray< CLatticeEdgeTrans * > *pGeneratingEdgesF = pEdgeF->GeneratingEdges();
    CArray< CLatticeEdgeTrans * > *pGeneratingEdgesE = pEdgeE->GeneratingEdges();

    // we give a benefit to edges spanning several source edges
    //if ( pGeneratingEdgesF ) {
    //  Trace( 1, "Bonus for spanning " << pGeneratingEdgesF->Size() << " edges" );
    //  SumPenalty -= 5 * ( pGeneratingEdgesF->Size() - 1 );
    //}


    if ( pGeneratingEdgesE && pGeneratingEdgesF ) {

      for ( int i = pGeneratingEdgesE->MinIndex(); i <= pGeneratingEdgesE->MaxIndex(); i++ ) {
	pSubEdgeE = pGeneratingEdgesE->GetAt( i );
	//pSubEdgeE->Display();

	// if this subedge was generated through transducer, find f-edge
	// which was generated by the same transducer item
	if ( pSubEdgeE->TransducerItem() ) {
	  Penalty = 1e10;
	
	  for ( int j = pGeneratingEdgesF->MinIndex(); j <= pGeneratingEdgesF->MaxIndex(); j++ ) {
	    pSubEdgeF = pGeneratingEdgesF->GetAt( j );
	    //pSubEdgeF->Display();

	    if ( pSubEdgeF->TransducerItem() ) {
	      NextPenalty = CalcTransducerScore( pSubEdgeF, pSubEdgeE );
	      if ( NextPenalty < Penalty ) {
		Penalty = NextPenalty;
	      }
	    }
	  }
	  SumPenalty += Penalty;
	}
      }
    }
  }

  return SumPenalty;

}



void
CTM_Graph_HMM::TraceBack()
{
  // find the best path by following the BackTrack information
  // calculate the perplexities
  // update the counts for the seen word-pairs (TCounts) and
  // the seen JCounts
  // update the alignment field, which stores the alignment for
  // the total text

  Trace( 1, "CTM_Graph_HMM::TraceBack" );

  int i, j;
  //int  sj;
  int Previous;
  float StartValue;
  //int i_min = 0;
  CWordIndex Wf, We;

  //CArray< CLatticeEdge* > EdgesF;
  //CArray< CLatticeEdge* > EdgesE;
  m_BestPathF.Clear();
  m_BestPathE.Clear();

  double TPenalty = 0.0;
  double JPenalty = 0.0;

  // clear positions of aligned words e in target sentence
  m_AlignedE.SetAll( 0 );
  m_AlignedPos.Clear();

  // starting point is point ( m_J + 1, m_I + 1 )
  StartValue = m_Q( m_J + 1, m_I + 1 );
  Previous = m_B(  m_J + 1, m_I + 1 );
  Trace( 2, "Previous = " << Previous );
  // /m_AlignProb.IncCount( m_I + 1, Previous, m_J + 1 );

  Trace( 2, "Starting point: " << m_J << " " << Previous );
  Trace( 2, "StartValue = " << StartValue );

  // for each column, find for current best i the best i in the
  // predeccessor column

  j = m_J;
  while ( j >= 1 ) {

    i = Previous;

    SetAlignment( m_CurrentPosSf, j, i );
    Trace( 2, "Aligned position for " << j << " is " << i);
    //m_EdgeF( j, i )->Display();
    //m_EdgeE( j, i )->Display();

    m_AlignedE[ i ] = 1;

    m_BestPathF.Insert( 0, m_EdgeF( j, i ) );
    m_BestPathE.Insert( 0, m_EdgeE( j, i ) );

    // aligned position: i = a(j)
    m_AlignedPos.Insert( 0, i );

    Wf = m_EdgeF( j, i )->WordIndex();
    We = m_EdgeE( j, i )->WordIndex();

    if ( m_EdgeF( j, i )->TransducerItem() != NULL) {
      Wf = m_pTextF->Vocabulary()->GetIndex( m_EdgeF( j, i )->TransducerItem()->Source().c_str() );
      Trace( 1, "Edge with category: " << m_EdgeF( j, i )->TransducerItem()->Source() << " WordIndex = " << Wf );
    } else {
      Wf = m_EdgeF( j, i )->WordIndex();
    }

    if ( m_EdgeE( j, i )->TransducerItem() != NULL) {
      We = m_pTextE->Vocabulary()->GetIndex( m_EdgeE( j, i )->TransducerItem()->Source().c_str() );
      Trace( 1, "Edge with category: " << m_EdgeE( j, i )->TransducerItem()->Source() << " WordIndex = " << We );
    } else {
      We = m_EdgeE( j, i )->WordIndex();
    }


    // report aligned and unaligned edges with category labels
    //const char *pWordF = m_pTextF->Vocabulary()->GetWord( Wf );
    //const char *pWordE = m_pTextE->Vocabulary()->GetWord( We );
    
    //bool F_IsLabel = IsLabel( pWordF );
    //bool E_IsLabel = IsLabel( pWordE );

    //    if ( F_IsLabel && E_IsLabel ) {
    //Trace0( 0, "Aligned Categories: " );
    //if ( m_EdgeF( j, i )->TransducerItem()->Number() == m_EdgeE( j, i )->TransducerItem()->Number() ) {
    //Trace( 0, "same rule: " << m_EdgeF( j, i )->TransducerItem()->Label() 
    //	<< " " << m_EdgeF( j, i )->TransducerItem()->Source() << " " << m_EdgeE( j, i )->TransducerItem()->Source() );
    //} else {
    //Trace( 0, "different rules: " << m_EdgeF( j, i )->TransducerItem()->Label() << m_EdgeE( j, i )->TransducerItem()->Label() 
    //	<< " " << m_EdgeF( j, i )->TransducerItem()->Source() << " " << m_EdgeE( j, i )->TransducerItem()->Source() );
    //}	
    //} else if ( F_IsLabel || E_IsLabel ) {
    //Trace( 0, "Category - Word: " <<m_EdgeF( j, i )->Word() << " " << m_EdgeE( j, i )->Word() );
    //} else {
    //Trace( 0, "Aligned Words: "  << m_EdgeF( j, i )->Word() << " " << m_EdgeE( j, i )->Word() );
    //}

    Trace( 1, m_pTextF->Vocabulary()->GetWord( Wf )
	   << " " << m_pTextE->Vocabulary()->GetWord( We ) );

    // which transducer items were used to create the edges
    Trace0( 1, "Transducer item for f: " );
    if ( m_EdgeF( j, i )->TransducerItem() ) {
      if ( TraceLevel > 0 ) {
	m_EdgeF( j, i )->TransducerItem()->Display();
      }
    } else {
      Trace( 1, "none" );
    }
    Trace0( 1, "Transducer item for e: " ); 
    if ( m_EdgeE( j, i )->TransducerItem() ) {
      if ( TraceLevel > 0 ) {
	m_EdgeE( j, i )->TransducerItem()->Display();
      }
    } else {
      Trace( 1, "none" );
    }

    // if edge was created through transducer, then update transducer count 
    // using the transducer item on the target side
    // else update lexicon count
    if ( m_EdgeE( j, i )->TransducerItem() ) {
      Trace( 1, "Update transducer item counts" );
      UpdateTransducerCounts( m_EdgeE( j, i ) );
    }

    // for perplexity
    TPenalty += -m_pLexicon->LogPr( Wf, We ); //log(m_pT->GetAt(We,Wf));
    
    // update translation countsGraphAlign.sh FUB | more

    m_pLexicon->IncCount( Wf, We );

    // alignment    
    Previous = m_B( j, i );

    JPenalty += -m_AlignProb.LogPr( i, Previous, j );
    m_AlignProb.IncCount( i, Previous, j );

    j = m_EdgeF( j, i )->FromNode()->Number();

    Trace( 2, "Previous = " << Previous );
    Trace( 2, "Next point: " << j - 1 << " " << Previous );
  }

  if ( TraceLevel > 0 ) {
    Trace0( 0, "Source Sentence: " );
    m_pTransMemF->DisplayPath( m_BestPathF );
    Trace( 0, "" );
    Trace0( 0, "Target Sentence: " );
    m_pTransMemF->DisplayPath( m_BestPathE );
    Trace( 0, "" );
  }


  // summation over all penalties
  m_PenaltySum += StartValue;
  m_TPenaltySum += TPenalty;
  m_APenaltySum += JPenalty;

  m_NumOfFWords += m_J;


  //cout <<  m_S << ": " << m_J << " " << m_I << " " 
    //<< m_TPenaltySum / m_NumOfFWords << " " 
      //<< m_APenaltySum / m_NumOfFWords << " " << m_PenaltySum  / m_NumOfFWords << "\n";
  //cout.flush();

}

  
void
CTM_Graph_HMM::UpdateTransducerCounts( CLatticeEdgeTrans *pEdge )
{
  Trace( 1, "CTM_Graph_HMM::UpdateTransducerCounts: " );
  CLatticeEdgeTrans *pSubEdge;

  if ( TraceLevel > 0 ) {
    pEdge->Display();

    if ( pEdge->TransducerItem() )
      pEdge->TransducerItem()->Display();
  }
    
  if ( pEdge->TransducerItem() ) {
    Trace0( 2, "Increment Count for: " ); pEdge->TransducerItem()->Display();
    pEdge->TransducerItem()->Count() += 1;
  }

  CArray< CLatticeEdgeTrans * > *pGeneratingEdges = pEdge->GeneratingEdges();

  if ( pGeneratingEdges ) {
    //Trace( 0, "Generated by: " );
    for ( int i = pGeneratingEdges->MinIndex(); i <= pGeneratingEdges->MaxIndex(); i++ ) {
      pSubEdge = pGeneratingEdges->GetAt( i );
      //pSubEdge->Display();
      // first recursive descend: run over all edges spanned by this
      if ( pSubEdge->TransducerItem() )
	UpdateTransducerCounts( pSubEdge );
    }
  } else {
    Trace( 2, "Terminal edge" );
  }

}


void
CTM_Graph_HMM::DisplayLabeledBiText()
{
  Trace( 1, "CTM_Graph_HMM::DisplayLabeledBiText" );

  int i, j;
  int Previous;

  CWordIndex Wf, We;

  //CArray< CLatticeEdge* > EdgesF;
  //CArray< CLatticeEdge* > EdgesE;
  m_pPathF->Clear();
  m_pPathE->Clear();

  CArray< CLatticeEdge* > EdgesETmp;
  //CLatticeEdge* pEdgeE;


  const CArray< CLatticeNode * > & NodesE = m_pLatticeE->Nodes();

  for ( i = NodesE.MinIndex(); i <= NodesE.MaxIndex(); i++ ) {
    EdgesETmp.SetAt( i, NULL );
  }

  // starting point is point ( m_J + 1, m_I + 1 )
  Previous = m_B(  m_J + 1, m_I + 1 );

  // for each column, find for current best i the best i in the
  // predeccessor column
  j = m_J;
  while ( j >= 1 ) {

    i = Previous;

    //EdgesF.Insert( 0, m_EdgeF( j, i ) );
	m_pPathF->Insert( 0, m_EdgeF( j, i ) );

    // only write associated labels
    Wf = m_EdgeF( j, i )->WordIndex();
    const char *pWordF = m_pTextF->Vocabulary()->GetWord( Wf );
    We = m_EdgeE( j, i )->WordIndex();
    const char *pWordE = m_pTextE->Vocabulary()->GetWord( We );

    if ( IsLabel( pWordF ) && IsLabel( pWordE ) ) {
      EdgesETmp.SetAt( m_EdgeE( j, i )->ToNode()->Number(), m_EdgeE( j, i ) );
      Trace( 1, "Set e at " << m_EdgeE( j, i )->ToNode()->Number() << " "
	     << pWordE );

      // write BiRule
      m_BiRulesFile << pWordF << " # ";
      m_BiRulesFile << m_EdgeF( j, i )->WordInfo()->SourceString() << " # ";
      m_BiRulesFile << m_EdgeE( j, i )->WordInfo()->SourceString() << endl;
    }

    Previous = m_B( j, i );
    j = m_EdgeF( j, i )->FromNode()->Number();
  }

  // collect remaining edges e

  i = m_I;
  while ( i >= 1 ) {
    if ( EdgesETmp[ i ] ) {
      Trace( 1, "Insert edge with label" );
      //EdgesE.Insert( 0, EdgesETmp[ i ] );
      m_pPathE->Insert( 0, EdgesETmp[ i ] );
    } else {
      Trace( 1, "Insert original edge" );
      CArray< CLatticeEdge * > &InEdges =  NodesE[ i ]->InEdges();
      //EdgesE.Insert( 0, InEdges.GetAt( InEdges.MinIndex() ) );
      m_pPathE->Insert( 0, InEdges.GetAt( InEdges.MinIndex() ) );
    }
    //i = EdgesE[ 0 ]->FromNode()->Number();
    i = (*m_pPathE)[ 0 ]->FromNode()->Number();
    Trace( 3, "next i = " << i );
  }


  if ( TraceLevel > 0 ) {
    cout << "Source Sentence: ";
    m_pTransMemF->DisplayPath( *m_pPathF );
    cout << endl;
    cout <<  "Target Sentence: ";
    m_pTransMemF->DisplayPath( *m_pPathE );
    cout << endl;
  }

}



void
CTM_Graph_HMM::WriteLabeledBiText()
{
  // check if file as been opened and is good for read
  if (!m_BiTextFile.good())
    return;

  Trace( 0, "CTM_Graph_HMM::WriteLabeledBiText" );

  int i, j;
  int Previous;

  CWordIndex Wf, We;

  //CArray< CLatticeEdge* > EdgesF;
  //CArray< CLatticeEdge* > EdgesE;
  CNumericalArray< CLatticeEdge* > EdgesETmp;
  m_BestPathF.Clear();
  m_BestPathE.Clear();

  //CLatticeEdge* pEdgeE;


  const CArray< CLatticeNode * > & NodesE = m_pLatticeE->Nodes();

  for ( i = NodesE.MinIndex(); i <= NodesE.MaxIndex(); i++ ) {
    EdgesETmp.SetAt( i, NULL );
  }

  // starting point is point ( m_J + 1, m_I + 1 )
  Previous = m_B(  m_J + 1, m_I + 1 );

  // for each column, find for current best i the best i in the
  // predeccessor column
  j = m_J;
  while ( j >= 1 ) {

    i = Previous;

    m_BestPathF.Insert( 0, m_EdgeF( j, i ) );

    // only write associated labels
    Wf = m_EdgeF( j, i )->WordIndex();
    const char *pWordF = m_pTextF->Vocabulary()->GetWord( Wf );
    We = m_EdgeE( j, i )->WordIndex();
    const char *pWordE = m_pTextE->Vocabulary()->GetWord( We );

    if ( (m_EdgeF( j, i )->TransducerItem() != NULL) && (m_EdgeE( j, i )->TransducerItem() != NULL) ) {
      //if ( IsLabel( pWordF ) && IsLabel( pWordE ) ) {
      EdgesETmp.SetAt( m_EdgeE( j, i )->ToNode()->Number(), m_EdgeE( j, i ) );
      Trace( 1, "Set e at " << m_EdgeE( j, i )->ToNode()->Number() << " "
	     << pWordE );

      // write BiRule
      m_BiRulesFile << pWordF << " # ";
      m_BiRulesFile << m_EdgeF( j, i )->WordInfo()->SourceString() << " # ";
      m_BiRulesFile << m_EdgeE( j, i )->WordInfo()->SourceString() << endl;
    }

    Previous = m_B( j, i );
    j = m_EdgeF( j, i )->FromNode()->Number();
  }

  // collect remaining edges e

  i = m_I;
  while ( i >= 1 ) {
    if ( EdgesETmp[ i ] ) {
      Trace( 1, "Insert edge with label" );
      m_BestPathE.Insert( 0, EdgesETmp[ i ] );
    } else {
      Trace( 1, "Insert original edge" );
      CArray< CLatticeEdge * > &InEdges =  NodesE[ i ]->InEdges();
     m_BestPathE.Insert( 0, InEdges.GetAt( InEdges.MinIndex() ) );
    }
    i = m_BestPathE[ 0 ]->FromNode()->Number();
    Trace( 3, "next i = " << i );
  }


  if ( TraceLevel > 0 ) {
    Trace0( 0, "Source-Sentence: " );
    m_pTransMemF->DisplayPath( m_BestPathF );
    Trace( 0, "" );
    Trace0( 0, "Target-Sentence: " );
    m_pTransMemF->DisplayPath( m_BestPathE );
    Trace( 0, "" );
  }

  m_pTransMemF->WritePath( m_BiTextFile, m_BestPathF );
  m_BiTextFile << " | ";
  m_pTransMemF->WritePath( m_BiTextFile, m_BestPathE );
  m_BiTextFile << "\n" << flush;

}



void 
CTM_Graph_HMM::DisplayQ( int J, int I )
{
  // display matrix that it resembles the normal display on paper
  // i.e.
  //
  //  ^
  // I|
  //  |      o
  //  |     o
  // i|   oo
  //  | oo 
  //  |o
  // 0|--------->
  //  0   j     J
  
  CWordIndex Word;
  const char *pString = NULL;

  int i, j;
  int maxLen = 0;

  // cout << J << " " << I << "\n";

  cout << endl << endl;
  cout << "Q-Matrix" << endl << endl;

  if (I > 0) {
    // get length of longest e word
    for (i = m_pSentenceE->MaxIndex(); i >= m_pSentenceE->MinIndex(); i--) {
      Word = m_pSentenceE->GetAt( i );
      pString = m_pTextE->Vocabulary()->GetWord( Word );
      Trace( 5, pString );
      if (strlen(pString) > (uint) maxLen) 
	maxLen = strlen(pString);
    }

    for (i = I; i >= m_pSentenceE->MinIndex(); i--) {
      cout.width(maxLen);
      Word = m_pSentenceE->GetAt( i );
      pString = m_pTextE->Vocabulary()->GetWord( Word );
      cout.setf( ios::left );
      cout << pString << " |" << flush;
      
      if (J > 0) {
	j = m_pSentenceF->MinIndex();
	//cout.setf(ios::right | ios::fixed);
	cout.setf(ios::right, ios::adjustfield);
	cout.setf(ios::fixed, ios::floatfield);
	cout.precision(2);
	cout.width(5);
	cout.fill(' ');
	//cout << m_Penalty[j][i];
	cout << m_Q( j, i );
	j++;
	for (; j <= J; j++) {
	  //cout.setf(ios::right | ios::fixed);
	  cout.setf(ios::right, ios::adjustfield);
	  cout.setf(ios::fixed, ios::floatfield);
	  cout.precision(2);
	  cout.width(6);
	  cout.fill(' ');
	  //cout << m_Penalty[j][i];
	  cout << m_Q( j, i );
	}
      }
      cout << "\n";
    }
  }

  DisplayFWords( maxLen );
  //Trace( 1, "DisplayQ done" );
}


void 
CTM_Graph_HMM::DisplayB( int J, int I )
{
  // display matrix that it resembles the normal display on paper
  // i.e.
  //
  //  ^
  // I|
  //  |      o
  //  |     o
  // i|   oo
  //  | oo 
  //  |o
  // 0|--------->
  //  0   j     J
  
  CWordIndex Word;
  const char *pString = NULL;

  int i, j;
  int maxLen = 0;

  // cout << J << " " << I << "\n";
  cout << m_B( m_J + 1, m_I + 1 ) << endl;
  if (I > 0) {
    // get length of longest e word
    for (i = m_pSentenceE->MaxIndex(); i >= m_pSentenceE->MinIndex(); i--) {
      Word = m_pSentenceE->GetAt( i );
      pString = m_pTextE->Vocabulary()->GetWord( Word );
      Trace( 5, pString );
      if (strlen(pString) > (uint) maxLen) 
	maxLen = strlen(pString);
    }

    for (i = I; i >= m_pSentenceE->MinIndex(); i--) {
      cout.width(maxLen);
      Word = m_pSentenceE->GetAt( i );
      pString = m_pTextE->Vocabulary()->GetWord( Word );
      cout << pString << " |" << flush;
      
      if (J > 0) {
	j = 1;
	cout.setf(ios::right | ios::fixed);
	cout.precision(2);
	cout.width(4);
	//cout << m_Penalty[j][i];
	cout << m_B( j, i );
	for (j = 2; j <= J; j++) {
	  cout.setf(ios::right | ios::fixed);
	  cout.precision(2);
	  cout.width(6);
	  //cout << m_Penalty[j][i];
	  cout << m_B( j, i );
	}
      }
      cout << "\n";
    }
  }

  DisplayFWords( maxLen );
  //Trace( 1, "DisplayB done" );
}



void 
CTM_Graph_HMM::DisplayBestPath(bool FromTotalAlign)
{
  Trace( 3, "DisplayBestPath " << m_J << " " << m_I );

  int i, j, pre;
  //float StartValue = 1e10;
  //uint i_min = 0;
  int MaxSLen = 150;
  char BestPath[150][150];


  for (j = m_pSentenceF->MinIndex(); j < MaxSLen; j++) 
    for (i = m_pSentenceE->MinIndex(); i < MaxSLen; i++)
      BestPath[j][i] = '.';

  //Trace( 1, "after init BestPath" );

  //m_J = m_pTextF->GetSentenceLength(m_CurrentPosSf);
  //m_I = m_pTextE->GetSentenceLength(m_CurrentPosSe);

  if ( (m_J > 0) && (m_I > 0) ) {
    // find the best path and store points in BestPath
    if (FromTotalAlign) {
      //Trace( 1, "FromTotalAlignment" );
      for (j = m_pSentenceF->MinIndex(); j <= m_pSentenceF->MaxIndex(); j++) {
	//i = GetAlignment(m_CurrentPosSf, j);
	//BestPath[j][i] = 'o';
	;
      }
    } else {
      //Trace( 1, "From Local Alignment" );
      // find the starting point in the last column
      i = m_B( m_J + 1, m_I + 1) % 1000;
      BestPath[ m_J ][ i ] = 'o';
      Trace( 5, m_J << " " << i );
      
      // for each column, find for current best i the best i in the
      // predeccessor column and set it for display
      for ( j = m_J; j > m_pSentenceF->MinIndex(); j--) {
	pre = m_B( j, i ) % 1000;
	BestPath[ j - 1 ][ pre ] = 'o';
	Trace( 5, j - 1 << " " << pre );
	i = pre;
      }
    }
  }

  // now display BestPath
  CWordIndex Word;
  int maxLen=0;
  const char *pString = NULL;

  if (m_I > 0) {
    // get length of longest e word
    for (i = m_I; i >= m_pSentenceE->MinIndex(); i--) {
      Word = m_pSentenceE->GetAt( i );
      pString = m_pTextE->Vocabulary()->GetWord( Word );
      Trace( 5, pString );
      if (strlen( pString ) > (uint) maxLen) 
	maxLen = strlen( pString );
    }

    for ( i = m_I; i >= m_pSentenceE->MinIndex(); i--) {
      cout.width(maxLen);
      Word = m_pSentenceE->GetAt( i );
      pString = m_pTextE->Vocabulary()->GetWord( Word );
      cout << pString << " |   " << flush;
      
      if (m_J > 0) {
	for (j = 1; j <= (int) m_J; j++) {
	  cout.setf(ios::left);
	  cout.width(5);
	  cout << BestPath[ j ][ i ] << " ";
	}
      }
      cout << "\n";
      cout.flush();
    }
  }

  DisplayFWords( maxLen );

  //Trace( 1, "DisplayBestPath done" );
}



void
CTM_Graph_HMM::DisplayFWords( int maxLen )
{
  int MaxWordLength = 40;
  const char *pString = NULL;
  int MaxSLen = 150;
  char WordMatrix[40][150];
  int NoSpace;
  int i, j;
  CWordIndex Word;


  // display line
  for (i = 0; i < maxLen; i++)
    cout << "-";
  cout << "-|";
  if (m_J > 0) {
    for (j = 1; j <= (int) m_J; j++) {
      cout << "------";
    }
  }
  cout << endl;

  // display french words at bottom
  //Trace( 1, "display french words at bottom" );
  char Blanks[150];

  maxLen += 4;

  for(i = 0; i < maxLen; i++)
    Blanks[i] = ' ';
  Blanks[maxLen]='\0';

  for (j = 0; j < MaxSLen; j++)
    for (i=0; i < MaxWordLength; i++)
      WordMatrix[i][j] = ' ';

  if (m_J > 0) {
    for (j = m_pSentenceF->MinIndex(); j <= m_pSentenceF->MaxIndex(); j++) {
      Word = m_pSentenceF->GetAt( j );
      pString = m_pTextF->Vocabulary()->GetWord( Word );
      cout.setf(ios::left);
      i=0;
      while (pString[i]) {
	WordMatrix[i][j] = (char) pString[i];
	i++;
      }
    }
  }

  NoSpace = 1;
  for (i=0; ((i < MaxWordLength) && (NoSpace == 1)); i++) {
    NoSpace = 0;
    cout << Blanks << " ";
    // for diagonal output an additional blank in the next iteration
    Blanks[maxLen] = ' ';
    maxLen++;
    Blanks[maxLen] = '\0';
    if (m_J > 0) {
      for (j = m_pSentenceF->MinIndex(); j <= m_pSentenceF->MaxIndex(); j++) {
	cout << WordMatrix[i][j] << "     ";
	if (WordMatrix[i][j] != ' ')
	  NoSpace = 1;
      }
    }
    cout << "\n";
  }
    
  cout << "\n";
  cout.flush();
}



/*
void
CTM_Graph_HMM::UpdateJCounts(int distance, int SentenceLength)
{ 

  if (distance >= 0) {
    m_JCountsPos[distance][SentenceLength]++;
    m_JCountsPosSum[SentenceLength]++;
  } else {
    m_JCountsNeg[-distance][SentenceLength]++;
    m_JCountsNegSum[SentenceLength]++;
  }
  m_JCountsTotal++;
}


*/



////////////////////////////////////////////////////////////////////////
// Methods to handle total alignment
////////////////////////////////////////////////////////////////////////

bool  
CTM_Graph_HMM::StoreAlignment( const char *pFileName, int Format )
{
  return m_pAlignment->Store( pFileName, Format );
}


bool 
CTM_Graph_HMM::LoadAlignment( const char *pFileName, int Format )
{
  return m_pAlignment->Load( pFileName, Format );
}


void 
CTM_Graph_HMM::SetAlignment( uint SentenceIndex, uint WordIndex, short Value )
{
  m_pAlignment->Set( SentenceIndex, WordIndex, Value );
}


short
CTM_Graph_HMM::GetAlignment( uint SentenceIndex, uint WordIndex )
{
  return 0;
  //  return m_pAlignment->Get( SentenceIndex, WordIndex );
}



bool 
CTM_Graph_HMM::IsLabel( const char *pWord ) const
{
  return ( pWord[ 0 ] == '@' );
}



void
CTM_Graph_HMM::ExtractReverseBilingualPatterns()
{ // akt 09.2002 

  Trace( 0, "CTM_Graph_HMM::ExtractReverseBilingualPatterns" );  

  int PRINT = 0;
  CTransducerItem *pTransducerItem = NULL;
  
  if( m_pTransMemNgram ){ //if m_pTransMemNgram

    // apply the N-gram transducer to the Target side lattice
    //        
    for ( int l = m_pTransMemNgram->FirstLevel(); l<= m_pTransMemNgram->LastLevel(); l++ ) {
      Trace( 1, "Apply transducer: Ngram" );
      m_pTransMemNgram->SetDirection( 0 );
      m_pTransMemNgram->ApplyTransducer( m_pLatticeE, l );
      //m_pLatticeE->Display();
    }
    
    int j1;
    
    m_pPathE->Clear();
    m_pPathF->Clear();
    
    if ( TraceLevel > 0 ) {
      for ( int j = m_BestPathF.MinIndex(); j <= m_BestPathF.MaxIndex(); j++ ) {
	m_BestPathF[ j ]->Display();
      }
      for ( int j = m_BestPathE.MinIndex(); j <= m_BestPathE.MaxIndex(); j++ ) {
	m_BestPathE[ j ]->Display();
      }
      
      for ( int j = m_AlignedPos.MinIndex(); j <= m_AlignedPos.MaxIndex(); j++ ) {
	cout << j << " " << m_AlignedPos[ j ] << endl;
      }
    }
    

    int MaxLength = 50; // extract patterns with length of source up to MaxLength
    for ( int h = 1; h <= MaxLength; h++ ) {
      
      for ( int j = m_AlignedPos.MinIndex(); 
	    j <= m_AlignedPos.MaxIndex() - h;
	    j++ ) {
	
	// look for minimal and maximal aligned postition in e_1 ... e_I
	// for corrent source sentence subsequence
	int min_i = 10000;
	int max_i = -1;
	
	for ( j1 = j; j1 <= j + h; j1++ ) {
	  int i = m_AlignedPos[ j1 ];
	  if ( i < min_i ) {
	    min_i = i;
	  }
	  if ( i > max_i ) {
	    max_i = i;
	  }
	  
	}
	
	// which part of the best path lies within the minimal and maximal
	// aligned positions, i.e. go from nodes to edges
	int Min_i = 10000;
	int Max_i = -1;
	
	for ( int i = m_BestPathE.MinIndex(); 
	      i <= m_BestPathE.MaxIndex();
	      i++ ) {
	  
	  if ( m_BestPathE[ i ]->ToNode()->Number() >= min_i ) {
	    if ( i < Min_i ) {
	      Min_i = i;
	    }
	  }
	  if ( m_BestPathE[ i ]->ToNode()->Number() <= max_i ) {
	    if ( i > Max_i ) {
	      Max_i = i;
	    }
	  }
	}

	// akt 09.2002
	// Decide to print only if the "target" side has an edge generated 
	// by the Ngram transducer

	// check and see if any of the edges in the current target lattice
	// has the appropriate start and end points.
	//cout << "Working on node " << m_BestPathE[Max_i]->ToNode()->Number() << endl;
	//cout << "Looking for a link from node " << m_BestPathE[Min_i]->FromNode()->Number() << endl;
	  
	const CArray< CLatticeEdge * > &Edges = m_pLatticeE->Edges();
	PRINT = 0;
	for ( int ed = Edges.MinIndex(); ed <= Edges.MaxIndex(); ed++ ) { //edges MinIndex..MaxIndex
	  
	  if( Edges[ed]->FromNode()->Number() == m_BestPathE[Min_i]->FromNode()->Number() ){
	    if( Edges[ed]->ToNode()->Number() == m_BestPathE[Max_i]->ToNode()->Number() ){

	      pTransducerItem = ((CLatticeEdgeTrans *)Edges[ed])->TransducerItem();    
	      if( pTransducerItem ){
		//cout << "This edge was in Ngram td" << endl;
		//Edges[ed]->Display();
		PRINT = 1;
	      }
	    }
	  }
	  else{	    
	    //cout << "Found an edge that goes from Node " << 
	    //  Edges[ed]->FromNode()->Number() <<
	    //  " To node " << Edges[ed]->ToNode()->Number() << endl; 
	  }
	}

	
	// write the bilingual pattern to the file
	// but only, if length in target sentence does not exceed
	// the length in the source sentence by 2 words
	//
	// and if the PRINT variable was set above
	if ( ((Max_i - Min_i) < h + 2) && (PRINT > 0)) {
	//if ( ( h - (Max_i - Min_i) < 2)          && (PRINT > 0)) {
	  Trace( 1, "j=" << j << " j+h=" << (j + h) 
		 << " min_i= " << min_i << " max_i=" << max_i ); 
	  Trace( 1, "j=" << j << " j+h=" << (j + h) 
		 << " Min_i= " << Min_i << " Max_i=" << Max_i ); 

	  //int l = Max_i - Min_i;
	  //int H = h;
	  //m_GrammarFile << l << " to " << H <<": ";

	  for ( int i = Min_i; i <= Max_i; i++ ) {
	    CWordIndex We = ((CLatticeEdgeTrans *)m_BestPathE[ i ])->WordIndex();
	    const char *pWordE = m_pTextE->Vocabulary()->GetWord( We );
	    m_GrammarFile <<  pWordE << " ";
	    Trace0( 1, pWordE << " " );
	  }
	  m_GrammarFile << "# ";
	  Trace0( 1, "# " );
	  for ( j1 = j; j1 <= j + h; j1++ ) {
	    CWordIndex Wf = ((CLatticeEdgeTrans *)m_BestPathF[ j1 ])->WordIndex();
	    const char *pWordF = m_pTextF->Vocabulary()->GetWord( Wf );
	    m_GrammarFile << pWordF << " ";
	    Trace0( 1, pWordF << " " );
	  }
	  m_GrammarFile << endl << flush;
	  Trace( 1, "" );
	}
	else{

	  // akt debugging only 
	  //
	  //
	  /*if(  ((Max_i - Min_i) < (h + 2)) ){
	    cout << "Did not print this edge: " << endl;
	    
	    for ( int i = Min_i; i <= Max_i; i++ ) {
	      CWordIndex We = ((CLatticeEdgeTrans *)m_BestPathE[ i ])->WordIndex();
	      const char *pWordE = m_pTextE->Vocabulary()->GetWord( We );
	      cout <<  pWordE << " ";
	    }
	    cout << "# ";
	    for ( j1 = j; j1 <= j + h; j1++ ) {
	      CWordIndex Wf = ((CLatticeEdgeTrans *)m_BestPathF[ j1 ])->WordIndex();
	      const char *pWordF = m_pTextF->Vocabulary()->GetWord( Wf );
	      cout << pWordF << " ";
	    }
	    cout << " # Because of Missing Ngrams" <<endl;
	  }
	  else{
	    cout << "Did not print this edge: " << endl;
	    
	    for ( int i = Min_i; i <= Max_i; i++ ) {
	      CWordIndex We = ((CLatticeEdgeTrans *)m_BestPathE[ i ])->WordIndex();
	      const char *pWordE = m_pTextE->Vocabulary()->GetWord( We );
	      cout <<  pWordE << " ";
	    }
	    cout << "# ";
	    for ( j1 = j; j1 <= j + h; j1++ ) {
	      CWordIndex Wf = ((CLatticeEdgeTrans *)m_BestPathF[ j1 ])->WordIndex();
	      const char *pWordF = m_pTextF->Vocabulary()->GetWord( Wf );
	      cout << pWordF << " ";
	    }
	    cout << " # Because of Length" <<endl;
	    }*/
	}
      } 
    }
  } // **End of Reverse New version
  else { 

    // START old version here: ********************************* //

    Trace( 0, "CTM_Graph_HMM::ExtractReverseBilingualPatterns" );
    
    int j1;
    
    m_pPathE->Clear();
    m_pPathF->Clear();
    
    if ( TraceLevel > 0 ) {
      for ( int j = m_BestPathF.MinIndex(); j <= m_BestPathF.MaxIndex(); j++ ) {
	m_BestPathF[ j ]->Display();
      }
      for ( int j = m_BestPathE.MinIndex(); j <= m_BestPathE.MaxIndex(); j++ ) {
	m_BestPathE[ j ]->Display();
      }
      
      for ( int j = m_AlignedPos.MinIndex(); j <= m_AlignedPos.MaxIndex(); j++ ) {
	cout << j << " " << m_AlignedPos[ j ] << endl;
      }
    }
    
    //return;
    //int MaxLength = 6; // extract patterns with length of source up to MaxLength
    int MaxLength = 4; // extract patterns with length of source up to MaxLength
    
    for ( int h = 1; h <= MaxLength; h++ ) {
      
      for ( int j = m_AlignedPos.MinIndex(); 
	    j <= m_AlignedPos.MaxIndex() - h;
	    j++ ) {
	
	// look for minimal and maximal aligned postition in e_1 ... e_I
	// for corrent source sentence subsequence
	int min_i = 10000;
	int max_i = -1;
	
	for ( j1 = j; j1 <= j + h; j1++ ) {
	  int i = m_AlignedPos[ j1 ];
	  if ( i < min_i ) {
	    min_i = i;
	  }
	  if ( i > max_i ) {
	    max_i = i;
	  }
	  
	}
	
	// which part of the best path lies within the minimal and maximal
	// aligned positions, i.e. go from nodes to edges
	int Min_i = 10000;
	int Max_i = -1;
	
	for ( int i = m_BestPathE.MinIndex(); 
	      i <= m_BestPathE.MaxIndex();
	      i++ ) {
	  
	  if ( m_BestPathE[ i ]->ToNode()->Number() >= min_i ) {
	    if ( i < Min_i ) {
	      Min_i = i;
	    }
	  }
	  if ( m_BestPathE[ i ]->ToNode()->Number() <= max_i ) {
	    if ( i > Max_i ) {
	      Max_i = i;
	    }
	  }
	}
	
	// write the bilingual pattern to the file
	// but only, if length in target sentence does not exceed
	// the length in the source sentence by 2 words
	//
	// akt 09.2002 only change in old version... print out src, then
	// target, since we're in Reverse mode.
	//
	if ( (Max_i - Min_i) < (h + 2) ) {
	  Trace( 1, "j=" << j << " j+h=" << (j + h) 
		 << " min_i= " << min_i << " max_i=" << max_i ); 
	  Trace( 1, "j=" << j << " j+h=" << (j + h) 
		 << " Min_i= " << Min_i << " Max_i=" << Max_i ); 

	  for ( int i = Min_i; i <= Max_i; i++ ) {
	    CWordIndex We = ((CLatticeEdgeTrans *)m_BestPathE[ i ])->WordIndex();
	    const char *pWordE = m_pTextE->Vocabulary()->GetWord( We );
	    m_GrammarFile <<  pWordE << " " ;
	    Trace0( 1, pWordE << " " );
	  }
	  m_GrammarFile << "# ";
	  Trace0( 1, "# " );
	  for ( j1 = j; j1 <= j + h; j1++ ) {
	    CWordIndex Wf = ((CLatticeEdgeTrans *)m_BestPathF[ j1 ])->WordIndex();
	    const char *pWordF = m_pTextF->Vocabulary()->GetWord( Wf );
	    m_GrammarFile << pWordF << " ";
	    Trace0( 1, pWordF << " " );
	  }
	  m_GrammarFile << endl << flush;
	  Trace( 1, "" );
	}
	
      }
    }
  } // **End of old version** //  
}


void
CTM_Graph_HMM::ExtractForwardBilingualPatterns()
{  // akt 09.2002
  // NEW - adding tgt->src training...

  // akt 08.02
  // have the Ngram TransMem (transducer), 
  // have the Source-side sentence lattice... need to
  // augment this lattice with all possible N-gram edges and
  // then read off these new edges to see where there was an N-gram
  // hit and which of these sentences (alignments) to actually
  // write to the GrammarFile...
  // 
  
  CTransducerItem *pTransducerItem = NULL;
  
  if( m_pTransMemNgram ){ //if m_pTransMemNgram
    
    // apply the N-gram transducer to the Source side lattice
    //        
    for ( int l = m_pTransMemNgram->FirstLevel(); l<= m_pTransMemNgram->LastLevel(); l++ ) {
      Trace( 1, "Apply transducer: Ngram" );
      m_pTransMemNgram->SetDirection( 0 );
      m_pTransMemNgram->ApplyTransducer( m_pLatticeF, l );
      //m_pLatticeF->Display();
    }
    
    // examine the new edges of the Source side lattice
    //
    const CArray< CLatticeEdge * > &Edges = m_pLatticeF->Edges();
    
    for ( int ed = Edges.MinIndex(); ed <= Edges.MaxIndex(); ed++ ) { //edges MinIndex..MaxIndex
      //
      // now we have the words from the source side in Edges[ed]
      //Edges[ed]->Display();
      
      // make a transducer item out of the current edge in this lattice...
      //
      pTransducerItem = ((CLatticeEdgeTrans *)Edges[ed])->TransducerItem();    
      
      if (pTransducerItem ) { // if pTransducerItem
	// this edge was created by the Ngram transducer.
	// go over all aligned nodes and determine the min. and 
	// max index on the target side 
	//
	// changed! was: j1 <= Edges[ed], but that was too far...
	int min_i = 10000;
	int max_i = -1;
	for (int j1 = Edges[ed]->FromNode()->Number(); j1 < Edges[ed]->ToNode()->Number(); j1++) { //Edge:FromNode() to Edge:ToNode()
	  
	  int i = m_AlignedPos[j1];
	  if ( i < min_i ){
	    min_i = i;
	  }
	  if ( i > max_i ) {
	    max_i = i;
	  }
	}
	
	// which part of the best path on the target side lies w/in the min_i and
	// max_i points determined above
	int Min_i = 10000;
	int Max_i = -1;
	for ( int i = m_BestPathE.MinIndex(); i <= m_BestPathE.MaxIndex(); i++ ) {
	  
	  // check for endpoints of the aligned (target) region
	  //
	  if ( m_BestPathE[i]->ToNode()->Number() >= min_i ) {
	    if ( i < Min_i ){
	      Min_i = i;
	    }
	  }
	  
	  if ( m_BestPathE[i]->ToNode()->Number() <= max_i ) {
	    if ( i > Max_i ) {
	      Max_i = i;
	    }
	  }
	}
	
	//
	// write the bilingual pattern to the file
	// but only, if length in target sentence does not exceed
	// the length in the source sentence by 2 words
	// NOTE:  since only 1-N (not N-1) are allowed the target
	// side will always be longer so no +/- check is necessary -akt.
	//
	int srcEdgeLength = (Edges[ed]->ToNode()->Number() - Edges[ed]->FromNode()->Number());
	int targetEdgeLength = (Max_i - Min_i);
	
	if ( (targetEdgeLength) < (srcEdgeLength + 2 ) ) {
	  
	  //Trace( 0, "j=" << j << " j+h=" << (j + h) 
	  //	 << " min_i= " << min_i << " max_i=" << max_i ); 
	  //Trace( 0, "j=" << j << " j+h=" << (j + h) 
	  //	 << " Min_i= " << Min_i << " Max_i=" << Max_i ); 
	  
	  // get the words from this source side edge
	  //
	  CString edgeSource = ((CLatticeEdgeTrans *)Edges[ed])->WordInfo()->SourceString();
	  
	  m_GrammarFile << edgeSource << " ";
	  Trace0( 1, edgeSource << " " );
	  m_GrammarFile << "# ";
	  Trace0( 1, "# " );
	  
	  // get the words from the corresponding target side edges
	  //
	  for ( int i = Min_i; i <= Max_i; i++ ) {
	    CWordIndex We = ((CLatticeEdgeTrans *)m_BestPathE[ i ])->WordIndex();
	    const char *pWordE = m_pTextE->Vocabulary()->GetWord( We );
	    m_GrammarFile << pWordE << " ";
	    Trace0( 1, pWordE << " " );
	  }
	  m_GrammarFile << endl << flush ;
	  Trace( 1, "" );
	}
      }// end: if p_transduceritem
    }// end: for int i = Edges.Minindex... Edges.Maxindex    
  } // end: if m_pTransMemNgram
  else { 

    // START old version here: ********************************* //

    Trace( 0, "CTM_Graph_HMM::ExtractForwardBilingualPatterns" );
    
    int j1;
    
    m_pPathE->Clear();
    m_pPathF->Clear();
    
    if ( TraceLevel > 0 ) {
      for ( int j = m_BestPathF.MinIndex(); j <= m_BestPathF.MaxIndex(); j++ ) {
	m_BestPathF[ j ]->Display();
      }
      for ( int j = m_BestPathE.MinIndex(); j <= m_BestPathE.MaxIndex(); j++ ) {
	m_BestPathE[ j ]->Display();
      }
      
      for ( int j = m_AlignedPos.MinIndex(); j <= m_AlignedPos.MaxIndex(); j++ ) {
	cout << j << " " << m_AlignedPos[ j ] << endl;
      }
    }
    
    //return;
    //int MaxLength = 6; // extract patterns with length of source up to MaxLength
    int MaxLength = 4; // extract patterns with length of source up to MaxLength
    
    for ( int h = 1; h <= MaxLength; h++ ) {
      
      for ( int j = m_AlignedPos.MinIndex(); 
	    j <= m_AlignedPos.MaxIndex() - h;
	    j++ ) {
	
	// look for minimal and maximal aligned postition in e_1 ... e_I
	// for corrent source sentence subsequence
	int min_i = 10000;
	int max_i = -1;
	
	for ( j1 = j; j1 <= j + h; j1++ ) {
	  int i = m_AlignedPos[ j1 ];
	  if ( i < min_i ) {
	    min_i = i;
	  }
	  if ( i > max_i ) {
	    max_i = i;
	  }
	  
	}
	
	// which part of the best path lies within the minimal and maximal
	// aligned positions, i.e. go from nodes to edges
	int Min_i = 10000;
	int Max_i = -1;
	
	for ( int i = m_BestPathE.MinIndex(); 
	      i <= m_BestPathE.MaxIndex();
	      i++ ) {
	  
	  if ( m_BestPathE[ i ]->ToNode()->Number() >= min_i ) {
	    if ( i < Min_i ) {
	      Min_i = i;
	    }
	  }
	  if ( m_BestPathE[ i ]->ToNode()->Number() <= max_i ) {
	    if ( i > Max_i ) {
	      Max_i = i;
	    }
	  }
	}
	
	// write the bilingual pattern to the file
	// but only, if length in target sentence does not exceed
	// the length in the source sentence by 2 words
	if ( (Max_i - Min_i) < (h + 2) ) {
	  Trace( 1, "j=" << j << " j+h=" << (j + h) 
		 << " min_i= " << min_i << " max_i=" << max_i ); 
	  Trace( 1, "j=" << j << " j+h=" << (j + h) 
		 << " Min_i= " << Min_i << " Max_i=" << Max_i ); 
	  
	  for ( j1 = j; j1 <= j + h; j1++ ) {
	    CWordIndex Wf = ((CLatticeEdgeTrans *)m_BestPathF[ j1 ])->WordIndex();
	    const char *pWordF = m_pTextF->Vocabulary()->GetWord( Wf );
	    m_GrammarFile << pWordF << " ";
	    Trace0( 1, pWordF << " " );
	  }
	  m_GrammarFile << "# ";
	  Trace0( 1, "# " );
	  for ( int i = Min_i; i <= Max_i; i++ ) {
	    CWordIndex We = ((CLatticeEdgeTrans *)m_BestPathE[ i ])->WordIndex();
	    const char *pWordE = m_pTextE->Vocabulary()->GetWord( We );
	    m_GrammarFile <<  pWordE << " ";
	    Trace0( 1, pWordE << " " );
	  }
	  m_GrammarFile << endl << flush;
	  Trace( 1, "" );
	}
	
      }
    }
  } // **End of old version** //  
}



void
CTM_Graph_HMM::ExtractBilingualGrammar()
{
  Trace( 0, "CTM_Graph_HMM::ExtractBilingualGrammar" );

  int j1;

  if ( TraceLevel > 0 ) {
	for ( int j = m_BestPathF.MinIndex(); j <= m_BestPathF.MaxIndex(); j++ ) {
	  m_BestPathF[ j ]->Display();
	}
	for ( int j = m_BestPathE.MinIndex(); j <= m_BestPathE.MaxIndex(); j++ ) {
	  m_BestPathE[ j ]->Display();
	}
	
	for ( int j = m_AlignedPos.MinIndex(); j <= m_AlignedPos.MaxIndex(); j++ ) {
	  cout << j << " " << m_AlignedPos[ j ] << endl;
	}
  }

  //return;

  for ( int h = 1; h <= 3; h++ ) {
    for ( int j = m_AlignedPos.MinIndex(); j <= m_AlignedPos.MaxIndex() - h; j++ ) {
      int NumLabels = 0;
      int min_i = 1000;
      int max_i = -1;

      for ( j1 = j; j1 <= j + h; j1++ ) {

	if ( IsLabel( ((CLatticeEdgeTrans *)m_BestPathF[ j1 ])->Word() ) ) {
	  NumLabels++;
	}

	int i = m_AlignedPos[ j1 ];
	if ( i < min_i ) {
	  min_i = i;
	}
	if ( i > max_i ) {
	  max_i = i;
	}

      }

      int Min_i = 1000;
      int Max_i = -1;
      for ( int i = m_BestPathE.MinIndex(); i <= m_BestPathE.MaxIndex(); i++ ) {
	if ( m_BestPathE[ i ]->ToNode()->Number() >= min_i ) {
	  if ( i < Min_i ) {
	    Min_i = i;
	  }
	}
	if ( m_BestPathE[ i ]->ToNode()->Number() <= max_i ) {
	  if ( i > Max_i ) {
	    Max_i = i;
	  }
	}
      }

      if ( ((Max_i - Min_i) < (h + 2)) && (NumLabels >= 0.5 * h)) {
		Trace( 1, "j=" << j << " j+h=" << (j + h) << " min_i= " << min_i << " max_i=" << max_i ); 
		Trace( 1, "j=" << j << " j+h=" << (j + h) << " Min_i= " << Min_i << " Max_i=" << Max_i ); 
		for ( j1 = j; j1 <= j + h; j1++ ) {
		  CWordIndex Wf = ((CLatticeEdgeTrans *)m_BestPathF[ j1 ])->WordIndex();
		  const char *pWordF = m_pTextF->Vocabulary()->GetWord( Wf );
		  m_GrammarFile << pWordF << " ";
		  Trace0( 1, pWordF << " " );
		}
		m_GrammarFile << "# ";
		Trace0( 1, "# " );
		for ( int i = Min_i; i <= Max_i; i++ ) {
		  CWordIndex We = ((CLatticeEdgeTrans *)m_BestPathE[ i ])->WordIndex();
		  const char *pWordE = m_pTextE->Vocabulary()->GetWord( We );
		  m_GrammarFile <<  pWordE << " ";
		  Trace0( 1, pWordE << " " );
		}
		m_GrammarFile << endl << flush;
		Trace( 1, "" );
      }
	  
    }
  }
  
}



void
CTM_Graph_HMM::PruneLattices()
{
  Trace( 0, "CTM_Graph_HMM::PruneLattices" );

  const CTransducer *pTransducerE;
  const CTransducer *pTransducerF;
  //  CTransducerItem *pTransducerItemE;
  //CTransducerItem *pTransducerItemF;
  CLatticeEdgeTrans *pEdgeE;
  CLatticeEdgeTrans *pEdgeF;

  const CArray< CLatticeEdge * > & EdgesE = m_pLatticeE->Edges();
  const CArray< CLatticeEdge * > & EdgesF = m_pLatticeF->Edges();

  CNumericalArray< int > FoundE;
  CNumericalArray< int > FoundF;

  FoundE.SetBoundaries( EdgesE.MinIndex(), EdgesE.MaxIndex() );
  FoundF.SetBoundaries( EdgesF.MinIndex(), EdgesF.MaxIndex() );


  for ( int e = EdgesE.MinIndex(); e <= EdgesE.MaxIndex(); e++ ) {
    pEdgeE = (CLatticeEdgeTrans *)EdgesE[ e ];
    pTransducerE = pEdgeE->Transducer();
    //pTransducerItemE = pEdgeE->TransducerItem();

    Trace( 0, "e = " << e );
    pEdgeE->Display();

    if ( pTransducerE ) {
      Trace( 0, "Generated through transducer " );
      CString Te = pEdgeE->Transducer()->Name();
      int Ne = pEdgeE->TransducerItem()->Number();
      for ( int f = EdgesF.MinIndex(); f <= EdgesF.MaxIndex(); f++ ) {
	pEdgeF = (CLatticeEdgeTrans *)EdgesF[ f ];
	pTransducerF = pEdgeF->Transducer();
	//pTransducerItemF = pEdgeF->TransducerItem();
      
	Trace( 0, "f = " << f );
	pEdgeF->Display();

	if ( pTransducerF ) {
	  Trace( 0, "Generated through transducer " );
	  CString Tf = pEdgeF->Transducer()->Name();
	  int Nf = pEdgeF->TransducerItem()->Number();

	  if ( (Te == Tf) && (Ne == Nf) ) {
	    Trace( 0, "Found " << e << " " << f );
	    FoundE[ e ] = 1;
	    FoundF[ f ] = 1;
	  }
	}
      }
    }
  }

  for ( int e = EdgesE.MinIndex(); e <= EdgesE.MaxIndex(); e++ ) {
    pEdgeE = (CLatticeEdgeTrans *)EdgesE[ e ];
    pTransducerE = pEdgeE->Transducer();
    Trace0( 1, FoundE[ e ] );
    if ( pTransducerE && (FoundE[ e ] == 0) ) {
      Trace0( 1, "Delete e-edge " << e << "  " );
      EdgesE[ e ]->Display();
    }
  }

  for ( int f = EdgesF.MinIndex(); f <= EdgesF.MaxIndex(); f++ ) {
    pEdgeF = (CLatticeEdgeTrans *)EdgesF[ f ];
    pTransducerF = pEdgeF->Transducer();
    Trace0( 1, FoundF[ f ] );
    if ( pTransducerF && (FoundF[ f ] == 0) ) {
      Trace0( 1, "Delete f-edge " << f << "  " );
      EdgesF[ f ]->Display();
    }
  }

	
}



void 
CTM_Graph_HMM::RemoveUnmatchedEdges()
{
  // i.e. loop over all nodes, for each node over all incoming edges
  const CArray< CLatticeEdge * > & EdgesE = m_pLatticeE->Edges();
  const CArray< CLatticeEdge * > & EdgesF = m_pLatticeF->Edges();

  CLatticeEdgeTrans *pEdgeE, *pEdgeF;
  CString Te, Tf;
  int Ne, Nf;

  int Matches = 0;

  Trace( 1, "RemoveUnmatchEdges" );
  Trace( 1, "Number of f-edges: " << EdgesF.Size() );
  Trace( 1, "Number of e-edges: " << EdgesE.Size() );

  //cout << "Number of Edges = " 
  //   << CLatticeEdgeTrans::s_NumberOfObjects << endl;

  /*
  Trace( 0, "---------- F-Edges --------------" );
  for ( int f = EdgesF.MinIndex(); f <= EdgesF.MaxIndex(); f++ ) {
    pEdgeF = (CLatticeEdgeTrans *)EdgesF[ f ];
    pEdgeF->Display();
    if ( pEdgeF->Transducer() ) {
      pEdgeF->TransducerItem()->Display();
    }
  }

  Trace( 0, "---------- E-Edges --------------" );
  for ( int e = EdgesE.MinIndex(); e <= EdgesE.MaxIndex(); e++ ) {
    pEdgeE = (CLatticeEdgeTrans *)EdgesE[ e ];
    pEdgeE->Display();
    if ( pEdgeE->Transducer() ) {
      pEdgeE->TransducerItem()->Display();
    }
  }
  */

  for ( int e = EdgesE.MinIndex(); e <= EdgesE.MaxIndex(); e++ ) {
    pEdgeE = (CLatticeEdgeTrans *)EdgesE[ e ];
    Trace( 2, "Edge e: " << e );
 
    // only test for match if edge was created by transducer
    if ( pEdgeE->Transducer() ) {
      Te = pEdgeE->Transducer()->Name();
      Ne = pEdgeE->TransducerItem()->Number();
      
      for ( int f = EdgesF.MinIndex(); f <= EdgesF.MaxIndex(); f++ ) {
	pEdgeF = (CLatticeEdgeTrans *)EdgesF[ f ];
	Trace( 2, "Edge f: " << f );
	
	// only test for match if edge was created by transducer
	if ( pEdgeF->Transducer() ) {

	  // test, if both edges have been created from same transducer item
	  Tf = pEdgeF->Transducer()->Name();
	  Nf = pEdgeF->TransducerItem()->Number();
	  
	  Trace( 2, "Transducers: " << "-" << Tf << "-" << Te );
	  Trace( 2, "Transducer item: " << Nf << " " << Ne );
	  
	  if ( (Tf == Te) && (Nf == Ne) ) {
	    Trace( 2, "they match" );
	    pEdgeE->SetMatchFlag();
	    pEdgeF->SetMatchFlag();
	    Matches++;
	  }
	}
      }
    }
  }

  // which are marked for deletion
  Trace( 1, "Number of Matches " << Matches );
  /*
  Trace( 0, "---------- F-Edges --------------" );
  for ( int f = EdgesF.MinIndex(); f <= EdgesF.MaxIndex(); f++ ) {
    pEdgeF = (CLatticeEdgeTrans *)EdgesF[ f ];
    if ( pEdgeF->Transducer() && !pEdgeF->MatchFlag() ) {
      Trace0( 0, "Remove f edge: "  << f << " " );
      pEdgeF->Display();
    }
  }
  Trace( 0, "---------- E-Edges --------------" );
  for ( int e = EdgesE.MinIndex(); e <= EdgesE.MaxIndex(); e++ ) {
    pEdgeE = (CLatticeEdgeTrans *)EdgesE[ e ];
    if ( pEdgeE->Transducer() && !pEdgeE->MatchFlag() ) {
      Trace0( 0, "Remove e edge: " << e << " " );
      pEdgeE->Display();
    }
  }
  */

  // now remove edges created by transducer but without match
  //Trace( 0, "---------- Remove F-Edges --------------" );
  for ( int f = EdgesF.MaxIndex(); f >= EdgesF.MinIndex(); f-- ) {
    pEdgeF = (CLatticeEdgeTrans *)EdgesF[ f ];
    if ( pEdgeF->Transducer() && !pEdgeF->MatchFlag() ) {
      //Trace0( 0, "Remove f edge: " << f << " " );
      //pEdgeF->Display();
      m_pLatticeF->RemoveEdge( pEdgeF );
    }
  }

  //Trace( 0, "---------- Remove E-Edges --------------" );
  for ( int e = EdgesE.MaxIndex(); e >= EdgesE.MinIndex(); e-- ) {
    pEdgeE = (CLatticeEdgeTrans *)EdgesE[ e ];
    if ( pEdgeE->Transducer() && !pEdgeE->MatchFlag() ) {
      //Trace0( 0, "Remove e edge: " << e << " " );
      //pEdgeE->Display();
      m_pLatticeE->RemoveEdge( pEdgeE );
    }
  }


  Trace( 2, "After removing unmatched" )
  Trace( 2, "Number of f-edges: " << EdgesF.Size() );
  Trace( 2, "Number of e-edges: " << EdgesE.Size() );

  //cout << "Number of Edges = " 
  //     << CLatticeEdgeTrans::s_NumberOfObjects << endl;
}


