#include <config.h>
//#include <fortify.h>
// $Id: TM_Graph_HMM.hh 1150 2005-04-13 18:18:37Z vogel $
//
// Translation Model CTM_Graph_HMM
// HMM-like word alignment for aligning edges from two lattices
//



#ifndef __TM_Graph_HMM__
#define __TM_Graph_HMM__

#include <fstream>

#include <TM.hh>
#include <AlignProbRel.hh>
#include <Alignment.hh>
#include <TransMem.hh>


//static char rcsid_tm_HMM_noI[] = "$Id";


class CTM_Graph_HMM : public CTranslationModel
{
public:
  CTM_Graph_HMM();
  ~CTM_Graph_HMM();

  CAlignProbRel & AlignProbs();

  //void SetSmallJumpLimit (int LargestSmallJump);

  //  void SetTransMemParamFiles( CString ParamFileF, CString ParamFileE, CString ParamFileNgram = NULL);
  void SetTransMemParamFiles( CString ParamFileF, CString ParamFileE, CString ParamFileNgram = NULL);
  void SetTransMemParamFiles( CString ParamFileF, CString ParamFileE);
  void SetBiTextFileName( char *pBiTextFileName );
  void SetGrammarOutFileName ( char *pGrammarOutFileName );
  void SetHMMTrainingDirection( CString AlignDirection );
  bool Train();

  void DisplayQ( int J = 10, int I = 10 );
  void DisplayB( int J = 10, int I = 10 );
  void DisplayBestPath( bool FromTotalAlign = FALSE );

  // ----- Methods for global Alignment ----- 
  void  SetAlignment( uint SIndex, uint WIndex, short Value );
  short GetAlignment( uint SIndex, uint WIndex );

  bool StoreAlignment( const char *pFileName, int Format = 0 );
  bool LoadAlignment ( const char *pFileName, int Format = 0 );


protected:
  bool NextIteration();
  void CalculatePathes();
  void FirstColumn();
  void PreviousColumn( double SkipPenalty, int ColumnOffset );
  double CalcTransducerScore( CLatticeEdgeTrans *pEdgeF, CLatticeEdgeTrans *pEdgeE );
  void CalcTransition( int jprev, int iprev, int j, int i, 
		       int Length_f, int Length_e, int Level, double TPenalty,
		       CLatticeEdgeTrans *pEdgeF, CLatticeEdgeTrans *pEdgeE );
  void TraceBack();
  void UpdateTransducerCounts( CLatticeEdgeTrans *pEdge );

  void DisplayLabeledBiText();
  void WriteLabeledBiText();
  void ExtractBilingualGrammar();
  void ExtractForwardBilingualPatterns();
  void ExtractReverseBilingualPatterns();

  void DisplayFWords( int maxLen );

  void InitFields();

  bool IsLabel( const char *pWord ) const;

  void PruneLattices();
  void RemoveUnmatchedEdges();


protected:
  CLatticeLM *m_pLatticeE;
  CLatticeLM *m_pLatticeF;

  CString m_TransMemParamFileE;
  CString m_TransMemParamFileF;
  CString m_TransMemParamFileNgram;

  CTranslationMemory *m_pTransMemF;
  CTranslationMemory *m_pTransMemE;
  CTranslationMemory *m_pTransMemNgram;
  bool m_HMMAlignForward;

  CLatticeEdgeTrans *m_pLatticeEdgeTrans;



  //int m_SmallJumpLimit;  // only small jumps <= abs(SmallJumpLimit) allowed

  CAlignProbRel m_AlignProb;

  CSentence *m_pSentenceE, *m_pSentenceF;

  // for dynamic programming of alignment
  CMatrix< double > m_Q;   // score from translation model
  CMatrix< int >    m_L;   // length of path over f
  CMatrix< int >    m_C;   // coverage over e
  CMatrix< int >    m_H;   // cummulative level of path from transducer hierarchy

  // to store the index of the previous best position
  CMatrix< int > m_B;

  // the edges over which point in trellis was reached
  CMatrix< CLatticeEdgeTrans * > m_EdgeF;
  CMatrix< CLatticeEdgeTrans * > m_EdgeE;

  CNumericalArray< CLatticeEdge * > m_BestPathF;
  CNumericalArray< CLatticeEdge * > m_BestPathE;

  CNumericalArray< CLatticeEdge * > *m_pPathF;
  CNumericalArray< CLatticeEdge * > *m_pPathE;

  double m_TPenaltySum;
  double m_APenaltySum;
  double m_PenaltySum;

  // perplexity results from different interations
  CArray< double > m_AlignPerplexity;

  // perplexity of sentence pairs
  CArray< float > *m_pAlignmentPerplexity;

  // flags for output
  bool m_DisplayPath;
  bool m_PrintAlignments;

  // output stream for Prolog-style alignments
  ofstream m_AlignmentFile;
  
  // alignment for total text
  CAlignment *m_pAlignment;

  // is edge aligned
  CArray< int > m_AlignedE;
  // i = a(j) for sentence pair
  CArray< int > m_AlignedPos;

  // output of Viterby path
  char *m_pBiTextFileName;

  // Phrase Alignment output File
  char *m_pGrammarOutFileName;

  ofstream m_BiTextFile;
  ofstream m_BiRulesFile;
  ofstream m_GrammarFile;
};


#endif
