#include <stdlib.h>
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

//#include <fortify.h>
#include <iostream>
#include <sstream>
#include <stdio.h>
#include <assert.h>
#include <math.h>

#define NO_TRACE 1
#include <Trace.hh>

#include <TM_Graph_IBM1.hh>

CLatticeEdgeTrans *CTM_Graph_IBM1::m_pLatticeEdgeTrans = new CLatticeEdgeTrans;


CTM_Graph_IBM1::CTM_Graph_IBM1()
{
  //std::cout << "CTM_Graph_IBM1::CTM_Graph_IBM1() " << endl << flush;
  m_SentencePrune = 0;
  m_BadSentenceValues.SetMaxSize( 0 );
  m_BadSentenceValues.SortReverse();

  m_pTransMemE = NULL;
  m_pTransMemF = NULL;

  //m_pLatticeEdgeTrans = new CLatticeEdgeTrans;

  //std::cout << "CTM_Graph_IBM1::CTM_Graph_IBM1() done " << endl << flush;
  
}


CTM_Graph_IBM1::~CTM_Graph_IBM1()
{
  delete m_pTransMemE;
  delete m_pTransMemF;
  //delete m_pLatticeEdgeTrans;
  //clog << "Destructor CTM_Graph_IBM1\n";
  //delete m_pT1;
}


void
CTM_Graph_IBM1::SetTransMemParamFiles( CString ParamFileF, CString ParamFileE )
{
  m_TransMemParamFileF = ParamFileF;
  m_TransMemParamFileE = ParamFileE;

  if ( m_TransMemParamFileF.GetLength() > 0 ) {
    m_pTransMemE = new CTranslationMemory;
    m_pTransMemE->SetDirection( 1 );
    m_pTransMemF = new CTranslationMemory;
    m_pTransMemF->SetDirection( 0 );
  }
}



bool
CTM_Graph_IBM1::Train()
{
 // return TRUE;

  //m_pT->SetDefaultValue(1.0/m_VocSizeF);

  Trace( 1, "------------------------------" );
  Trace( 1,  " Train IBM1" );

  assert( m_pTextF );
  assert( m_pTextE );


  // set array for word boundaries
  m_pIsBoundaryWord->SetBoundaries( 0, m_pTextF->Vocabulary()->Size() );
  // TODO: Use TRACE here
  std::cout << "Boundary words" << endl;
  CWordIndex Wb;
  for ( int n = m_pBoundaryWords->MinIndex();
	n <= m_pBoundaryWords->MaxIndex();
	n++ ) {
      // TODO: Use TRACE here
      std::cout << n << " " << (*m_pBoundaryWords)[ n ] << endl;
      Wb = m_pTextF->Vocabulary()->GetIndex( (*m_pBoundaryWords)[ n ] );
      // TODO: Use TRACE here
      std::cout << "WordIndex = " << Wb << endl;
      m_pIsBoundaryWord->SetAt( Wb, 1 );
  }


  // init translation memories
  Trace( 1, "init translation memories" );
  if ( m_pTransMemF ) {
      m_pTransMemE->Init( m_TransMemParamFileE, m_pTextE->Vocabulary() );
      m_pTransMemE->RenormalizeTransducer();
      m_pTransMemF->Init( m_TransMemParamFileF, m_pTextF->Vocabulary() );
      m_pTransMemF->RenormalizeTransducer();
  }

  // create and init lexicon
  if ( m_pLexicon == NULL ) {
    Trace( 1, "Create new Lexicon" );
    m_pLexicon = new CLexiconForTraining;
    assert( m_pLexicon );
    // init lexicon with vocabulary
    if ( !m_pLexicon->Init( m_pTextF->Vocabulary(), m_pTextE->Vocabulary() ) ) {
	cerr << "main(): Could not init Lexicon. Exiting." << endl << flush;
	exit(1);      
    } // endif: error
  }


  // set smoothing parameter
  m_pLexicon->SetLambdaSmooth( m_LambdaSmooth );
  Trace( 1, "Test Prob: " << m_pLexicon->Pr( 3, 3 ) );

  // set array for sentence perplexities to correct size
  m_pTotalPerplexity->AssureBoundaries( 0, m_pTextF->NumberOfSentences() );

  //m_pLexicon->TestProbabilities();

  // Perplexity after initialization (should be equal to vocabulary size)
  Perplexity();
  //return TRUE;

  // iterate
  for ( m_CurrentIteration = 0; 
	m_CurrentIteration < m_Iterations; 
	m_CurrentIteration++ ) {
    Trace( 0, "\n\n== Iteration " << m_CurrentIteration << " ==\n" );

    if ( m_CurrentIteration >= m_FirstSegmentingIteration ) {
      Trace( 0, "Segment Sentences" );
      SegmentSentences();
      if ( !NextIteration() ) {
	  cerr << "main(): Error occured during iteration " << m_CurrentIteration+1
	       << ". Exiting." << endl << flush;
	  exit(1);
      } // endif:: error
    } else {
      Trace( 0, "Next Iteration" );
      if ( !NextIteration() ) {
	  cerr << "main(): Error occured during iteration " << m_CurrentIteration+1
	       << ". Exiting." << endl << flush;
	  exit(1);
      } // endif:: error
    }
    // NextIteration calculates perplexity of the model
    // as it was before the current iteration
    //Perplexity();

    if ( m_CurrentIteration >= m_FirstPruningIteration ) {
      DisplayBadSentences();
      RemoveBadSentences();
    }


  }

  // calculate perplexity after final iteration
  //double PP;
  Perplexity();

  //std::cout << "Initial perplexity measure: " << PP << "\n";

  //PerplexityForMaximum();

  // set probability for cognates
  //std::cout << "With Cognates" << endl;
  //InitLexicon();
  //PP = Perplexity();

  //PerplexityForMaximum();

  //DisplayBiText();

  //m_pLexicon->DisplayCounts();

  // TODO: maybe use TRACE for the following?

  // display all perplexities
  std::cout.setf(ios::right | ios::fixed);
  //std::cout.width(6);
  std::cout.precision(2);
  std::cout << "Perplexity as Function of Iteration \n\n";
  std::cout << "Iteration   &  Lexicon  &   Jump   &   Total \\\\ \\hline \n";
  for ( int Iteration = 0; Iteration <= m_Iterations; Iteration++ ) {
    std::cout << Iteration << " & " 
	 << m_LexiconPerplexity[Iteration] << " & "  
	 << m_AlignPerplexity[Iteration] << " & "  
	 << m_TotalPerplexity[Iteration] << "\n";
  }
  

  return TRUE;
}



bool
CTM_Graph_IBM1::NextIteration()
{
  Trace( 1, "CTM_Graph_IBM1::NextIteration ");
  Trace( 3, "Test Prob: " << m_pLexicon->Pr( 3, 3 ) );

  // for reestimation
  CWordIndex Se, Sf, We, Wf;
  TPosition PosSe, PosSf;
  double Sum, Sum1;
  // for perplexity
  double PAlignSum = 0.0, PTransSum = 0.0, PTrans = 0.0, PTotal;
  int SentenceLengthE = 0, SentenceLengthF = 0;
  int TotalLengthF = 0;
  double TProb;

  int Position;
  bool IsNew;

  m_pLexicon->ClearCounts();
  m_BadSentenceValues.RemoveAll();
  m_BadSentenceNumbers.RemoveAll();
  if ( m_pTransMemF ) {
    m_pTransMemF->ClearTransducerCounts();
  }
  
  const CSentence *pSentenceE, *pSentenceF;
  
  CLatticeLM *pLatticeE = NULL;
  CLatticeLM *pLatticeF = NULL;
  CLatticeNode *pNodeF;
  
  CString TranslationE, TranslationF;
  
  // fetch first sentence pair
  //Se = m_pTextE->FirstSentence( PosSe );
  //Sf = m_pTextF->FirstSentence( PosSf );
  
  //Trace( 1, "Sf = " << Sf << " Se " << Se );
  
  //pSentenceE = m_pTextE->GetSentence( PosSe );
  //pSentenceF = m_pTextF->GetSentence( PosSf );
  
  //pSentenceF->Display( m_pTextF->Vocabulary() ); std::cout << endl;
  //pSentenceE->Display( m_pTextE->Vocabulary() ); std::cout << endl;
  
  // fetch first sentence pair
  Se = m_pTextE->FirstSentence( PosSe );
  Sf = m_pTextF->FirstSentence( PosSf );
  
  uint s = 0;
  
  while (Se != m_pTextE->Vocabulary()->EOT && 
	 Sf != m_pTextF->Vocabulary()->EOT)     {
    
    if (s % 1000 == 0) {
      //Trace( 1, "Sentence pair " << s );
	// TODO: TRACE or not TRACE? That is the question ...
      std::cout << "Sentence pair " << s << endl << flush;
      //std::cout.flush();
    }
    
    //std::cout << "--------------------------------------" << endl;
    //std::cout << "Sentence pair " << s << endl;
    
    // get length of sentences; if one is zero, skip
    if ( (m_pTextF->SentenceLength( PosSf ) == 0) ||
	 (m_pTextE->SentenceLength( PosSe ) == 0) ) {
      
      Trace( 0, "Sentence length zero in sentence pair " << s );
      if ( m_pTextF->SentenceLength( PosSf ) == 0 ) 
		Trace( 0, "Source sentence empty" );
      if ( m_pTextE->SentenceLength( PosSe ) == 0 ) 
		Trace( 0, "Target sentence empty" );

    } else {

      (*m_pTotalPerplexity)[ s ] = 0.0;

      // fetch sentences
      pSentenceE = m_pTextE->GetSentence( PosSe );
      pSentenceF = m_pTextF->GetSentence( PosSf );
      
      SentenceLengthE = pSentenceE->Size();
      //std::cout << "Sentence length E = " << SentenceLengthE << endl;
      
      // Convert sentence into lattice
      pLatticeE = ConvertToLattice( pSentenceE, m_pTextE->Vocabulary() );
      pLatticeF = ConvertToLattice( pSentenceF, m_pTextF->Vocabulary() );
      
	  //pLatticeF->Display();
	  //pLatticeE->Display();

      //TraceLevel = 0;
      if ( m_pTransMemF ) {
		Trace( 1, "Translation Memory for Source" );
		m_pTransMemF->SetDirection( 0 );
		TranslationF = m_pTransMemF->TranslateLattice( pLatticeF );
		Trace( 1, "Translation Memory for Target" );
		m_pTransMemF->SetDirection( 1 );
		TranslationE = m_pTransMemF->TranslateLattice( pLatticeE );
      }
      //TraceLevel = 1;
      
      //if (1) {

      if ( s < 10 ) {
	  // TODO: Use trace here
		std::cout << "Sentence pair " << s << endl;
		pSentenceF->Display( m_pTextF->Vocabulary() ); std::cout << endl;
		pSentenceE->Display( m_pTextE->Vocabulary() ); std::cout << endl;
		
		std::cout << TranslationF << endl;
		std::cout << TranslationE << endl;
		
		//pLatticeF->Display();
		//pLatticeE->Display();
      }
      
      
      SentenceLengthE = pSentenceE->Size();
      SentenceLengthF = pSentenceF->Size();
      //std::cout << "Sentence length E = " << SentenceLengthE << endl;
      
      TotalLengthF += SentenceLengthF;
      
      // loop over words in sentence of source (first, F) text
      // i.e. loop over all nodes, for each node over all incoming edges
      const CArray< CLatticeEdge * > & EdgesE = pLatticeE->Edges();
      //const CArray< CLatticeEdge * > & EdgesF = pLatticeF->Edges();
      
      CLatticeEdgeTrans *pEdgeE;
      CLatticeEdgeTrans *pEdgeF;
      
      int n, j, i;
        
      const CArray< CLatticeNode * > & NodesF = pLatticeF->Nodes();

      for ( n = NodesF.MinIndex() + 1; n <= NodesF.MaxIndex(); n++ ) {
		pNodeF = NodesF[ n ];
		const CArray< CLatticeEdge * > & EdgesF = pNodeF->InEdges();
		
		for ( j = EdgesF.MinIndex(); j <= EdgesF.MaxIndex(); j++ ) {
		  pEdgeF = (CLatticeEdgeTrans *)EdgesF[ j ];
		  
		  //Trace( 0, "f-edge: "); pEdgeF->Display();
		  
		  // get word index: normal edge, then already known;
		  // for edge with category, use source string
		  if ( pEdgeF->TransducerItem() != NULL) {
			Wf = m_pTextF->Vocabulary()->GetIndex( pEdgeF->TransducerItem()->Source().c_str() );
			Trace( 1, "Edge with category: " << pEdgeF->TransducerItem()->Source() 
				   << " WordIndex = " << Wf );
		  } else {
			Wf = pEdgeF->WordIndex();
		  }
		  
		  //Wf = pEdgeF->WordIndex();
		  
		  if ( Wf == 0 ) {
			Trace( 1, "UNK " << pEdgeF->Word() );
			Wf = m_pTextF->Vocabulary()->GetIndex( pEdgeF->Word() );
			Trace( 1, "Set index to " << Wf );
			pEdgeF->SetWordIndex( Wf );
		  }
		  
		  
		  Sum = 0.0;
		  Sum1 = 0.0;
		  PTrans = 0.0;
		  
		  for ( i = EdgesE.MinIndex(); i <= EdgesE.MaxIndex(); i++ ) {
			
			pEdgeE = (CLatticeEdgeTrans *)EdgesE[ i ];
			//Trace( 0, "e-edge: "); pEdgeE->Display();
			
			// get word index: normal edge, then already known;
			// for edge with category, use source string
			if ( pEdgeE->TransducerItem() != NULL) {
			  We = m_pTextE->Vocabulary()->GetIndex( pEdgeE->TransducerItem()->Source().c_str() );
			  Trace( 2, "Edge with category: " << pEdgeE->TransducerItem()->Source() 
					 << " WordIndex = " << We );
			} else {
			  We = pEdgeE->WordIndex();
			}
			
			if ( We == 0 ) {
			  Trace( 1, "UNK " << pEdgeE->Word() );
			  We = m_pTextE->Vocabulary()->GetIndex( pEdgeE->Word() );
			  Trace( 1, "Set index to " << We );
			  pEdgeE->SetWordIndex( We );
			}
			
			// lexicon probability:
			// for transducer added edges from same transducer item: TPenalty == log(0.99999)
			// else: lexicon prob
			if ( (pEdgeF->TransducerItem() != NULL) && (pEdgeE->TransducerItem() != NULL) ) {
			  
			  CString Tf = pEdgeF->Transducer()->Name();
			  CString Te = pEdgeE->Transducer()->Name();
			  Trace( 2, "Transducers: " << "-" << Tf << "-" << Te );
			  
			  int Nf = pEdgeF->TransducerItem()->Number();
			  int Ne = pEdgeE->TransducerItem()->Number();
			  Trace( 2, "Transducer item: " << Nf << " " << Ne );
			  
			  if ( (Tf == Te) && (Nf == Ne) ) {
				Trace( 2, "Same transducer and same item: " << Tf << " " << Te );
				Trace( 2, "TProb from transducer= " 
					   << pEdgeE->TransducerItem()->ScoreSourceGivenTarget() );
				
				// TPenalty = -log( 0.99999 );
				TProb = pEdgeE->TransducerItem()->ScoreSourceGivenTarget();
				Sum1 += TProb;
			  }
			} else {
			  //Trace( 0, "Wf - We: " << pEdgeF->Word() << " " << pEdgeE->Word() );
			  TProb = m_pLexicon->Pr( Wf, We ); //log(m_pT->GetAt(We,Wf));
			  Sum1 += TProb;
			  Trace( 1, "TProb from lexicon = " << TProb );
			}
			
			
		  }
		  //if ( s < 5 ) {
		  //Trace( 1, "Sum1 = " << Sum1 );
		  //}
		  
		  assert( Sum1 > 0.0 );
		  
		  
		  for ( i = EdgesE.MinIndex(); i <= EdgesE.MaxIndex(); i++ ) {
			
			pEdgeE = (CLatticeEdgeTrans *)EdgesE[ i ];
			
			// get word index: normal edge, then already known;
			// for edge with category, use source string
			if ( pEdgeE->TransducerItem() != NULL) {
			  We = m_pTextE->Vocabulary()->GetIndex( pEdgeE->TransducerItem()->Source().c_str() );
			  Trace( 2, "Edge with category: " << pEdgeE->TransducerItem()->Source() 
					 << " WordIndex = " << We );
			} else {
			  We = pEdgeE->WordIndex();
			}
			
			// We = pEdgeE->WordIndex();
			if ( (We == 0) || (Wf == 0) ) {
			  Trace( 1, "Unknowns " << Wf << " " << We );
			}
			
			// lexicon probability:
			// for transducer added edges from same transducer item: TPenalty == log(0.99999)
			// else: lexicon prob
			if ( (pEdgeF->TransducerItem() != NULL) && (pEdgeE->TransducerItem() != NULL) ) {
			  
			  CString Tf = pEdgeF->Transducer()->Name();
			  CString Te = pEdgeE->Transducer()->Name();
			  Trace( 1, "Transducers: " << "-" << Tf << "-" << Te );
			  
			  int Nf = pEdgeF->TransducerItem()->Number();
			  int Ne = pEdgeE->TransducerItem()->Number();
			  Trace( 1, "Transducer item: " << Nf << " " << Ne );
			  
			  if ( (Tf == Te) && (Nf == Ne) ) {
				Trace( 1, "Same transducer and same item: " << Tf << " " << Te );
				Trace( 1, "TProb from transducer= " 
					   << pEdgeE->TransducerItem()->ScoreSourceGivenTarget() );
				
				// TPenalty = -log( 0.99999 );
				TProb = pEdgeE->TransducerItem()->ScoreSourceGivenTarget();
				PTrans += TProb;
				Trace( 1, "Add count: " << TProb / Sum1 );
				pEdgeE->TransducerItem()->Count() += (TProb / Sum1);
				
			  }
			} else {
			  //Trace( 0, "Wf - We: " << pEdgeF->Word() << " " << pEdgeE->Word() );
			  TProb = m_pLexicon->Pr( Wf, We ); //log(m_pT->GetAt(We,Wf));
			  Trace( 1, "TProb from lexicon = " << TProb );
			  m_pLexicon->AddCount( Wf, We, TProb / Sum1 );
			  PTrans += TProb;
			}
			
		  }
		  
		}
		
		PTransSum += log(PTrans);
		//std::cout << "Sentence length E = " << SentenceLengthE << endl;
		//std::cout << "current AlignSum = " << PAlignSum << endl;
		
		(*m_pTotalPerplexity)[ s ] += log(PTrans);
		
      }
      
      (*m_pTotalPerplexity)[ s ] = 
		exp( -(*m_pTotalPerplexity)[ s ] / (double) SentenceLengthF );
      
      // insert value of sentence into list of bad sentences
      Position = m_BadSentenceValues.Insert( (*m_pTotalPerplexity)[ s ], IsNew );
      if ( IsNew ) {
		m_BadSentenceNumbers.Insert( Position, s );
      } else {
		m_BadSentenceNumbers.SetAt( Position, s );
      }
      
      if ( m_BadSentenceNumbers.MaxIndex() > m_BadSentenceValues.MaxIndex() ) {
		m_BadSentenceNumbers.SetMaxIndex( m_BadSentenceValues.MaxIndex() );
      }
      
      
      //}
      
      delete pLatticeF;
      delete pLatticeE;
      pLatticeF = NULL;
      pLatticeE = NULL;
    
      PAlignSum += SentenceLengthF * log(1.0 / (float) SentenceLengthE);

      /*
      std::cout << m_CurrentIteration << " & " 
	   << TotalLengthF << " & " 
	   << PTransSum / TotalLengthF << " & " 
	   << PAlignSum / TotalLengthF << endl;
      std::cout << m_CurrentIteration << " & " 
	   << TotalLengthF << " & " 
	   << exp(-PTransSum/TotalLengthF) << " & " 
	   << exp(-PAlignSum/TotalLengthF) << endl;
      */
    }

    // fetch next sentence pair
    Se = m_pTextE->NextSentence( PosSe );
    Sf = m_pTextF->NextSentence( PosSf );
    
    s++;
  } // end loop over sentence pairs
  
  //m_pLexicon->Display();
  
  PTransSum = -PTransSum / TotalLengthF;
  PAlignSum = -PAlignSum / TotalLengthF;
  
  // TODO: use TRACE here
  std::cout << "# TransSum = " << PTransSum << endl;
  std::cout << "# PAlignSum = " << PAlignSum << endl;
  std::cout << "# TotalLengthF = " << TotalLengthF << endl;
  
  PTotal = PAlignSum + PTransSum;
  // TODO: use TRACE here?
  std::cout << m_CurrentIteration << " & " << TotalLengthF << " & " << exp(PTransSum) 
       << " & " << exp(PAlignSum) << " & " << exp(PTotal) << endl;
  
  m_LexiconPerplexity[m_CurrentIteration] = exp(PTransSum);
  m_AlignPerplexity[m_CurrentIteration] = exp(PAlignSum);
  m_TotalPerplexity[m_CurrentIteration] = exp(PTotal);
  
  // normalize the translation probabilities
  Trace( 1, "Renormalize" );
  m_pLexicon->Renormalize();
  //m_pLexicon->TestProbabilities();
  
  //m_pLexicon->DisplayCounts();
  
  // TODO: else?
  if ( m_pTransMemF ) {
    //m_pTransMemE->RenormalizeTransducer();
    m_pTransMemF->DisplayTransducer();
    m_pTransMemF->StoreTransducer();
  }
  
  return TRUE;
}



void
CTM_Graph_IBM1::SegmentSentences()
{
  Trace( 0, "CTM_Graph_IBM1::SegmentSentences ");

  // for reestimation
  CWordIndex Se, Sf;
  TPosition PosSe, PosSf;
  // for perplexity
  int SentenceLengthE = 0, SentenceLengthF = 0;
  int TotalLengthF = 0;
  double TotalLogProb = 0.0, LogProb = 0.0;

  const CSentence *pSentenceE, *pSentenceF;
  CSentence *pSegmentE, *pSegmentF;
  
  CString TranslationE, TranslationF;

  int Best_j = -1, Best_i = -1;
  double BestDifference = 10000;

  m_pSegmenter = new CSegmenter;
  assert( m_pSegmenter );

  // fetch first sentence pair
  Se = m_pTextE->FirstSentence( PosSe );
  Sf = m_pTextF->FirstSentence( PosSf );
  
  uint s = 0;
  
  while (Se != m_pTextE->Vocabulary()->EOT && 
	 Sf != m_pTextF->Vocabulary()->EOT)     {
    
    if (s % 1000 == 0) {
	// TODO: why not TRACE here?
      //Trace( 1, "Sentence pair " << s );
      std::cout << "Sentence pair " << s << endl << flush;
      //std::cout.flush();
    }
    
    // TODO: use TRACE here
    std::cout << "--------------------------------------" << endl;
    std::cout << "Sentence pair " << s << endl;
    
    // fetch sentences
    pSentenceE = m_pTextE->GetSentence( PosSe );
    pSentenceF = m_pTextF->GetSentence( PosSf );
    
    SentenceLengthE = pSentenceE->Size();
    SentenceLengthF = pSentenceF->Size();
    //std::cout << "Sentence length E = " << SentenceLengthE << endl;
    
    if ( (SentenceLengthF == 0) || (SentenceLengthE == 0) ) {

      Trace( 0, "Sentence length zero in sentence pair " << s );

    } else {
      // check for boundary words
      //CheckBoundaries( pSentenceF );
      //CheckBoundaries( pSentenceE );

      pSentenceF->Display( m_pTextF->Vocabulary() ); std::cout << endl;
      pSentenceE->Display( m_pTextE->Vocabulary()); std::cout << endl;

      if ( (SentenceLengthF > 4) && (SentenceLengthE > 4) ) {

	// first calculate log( p( f | e ) ) for complete sentences
	LogProb = SentenceLogProb( pSentenceF, pSentenceE );
	Trace( 0, "Sentence LogProb = " << LogProb );

	// now select segmenting points and calculate logprob for both segment pairs
	Best_j = -1, Best_i = -1;
	BestDifference = 10000;

	int i, j;
	CWordIndex e, f;
	// loop over words in sentence
	for ( j = pSentenceF->MinIndex(); j < pSentenceF->MaxIndex(); j++ ) {
	  f = (*pSentenceF)[ j ];
	  
	  for ( i = pSentenceE->MinIndex(); i < pSentenceE->MaxIndex(); i++ ) {
	    e = (*pSentenceE)[ i ];

	    if ( (*m_pIsBoundaryWord)[ f ] + (*m_pIsBoundaryWord)[ e ] < m_RequiredSplitPoints ) {
	      //Trace( 0, (*m_pIsBoundaryWord)[ f ] << " + " << (*m_pIsBoundaryWord)[ e ] 
	      //	     << " < " << m_RequiredSplitPoints );
	      ;
	    } else {
	      //Trace( 0, "Number of equired split points found" );
	      //std::cout << "Boundary word at " << j << " is ";
	      //std::cout << m_pTextF->Vocabulary()->GetWord( f );
	      //std::cout << endl;
	      
	      //std::cout << "Boundary word at " << i << " is ";
	      //std::cout << m_pTextE->Vocabulary()->GetWord( e );
	      //std::cout << endl;
	      
	      Trace( 0, "Segment Points are " << j << " " << i );
	      
	      if ( abs( i - j * SentenceLengthE / SentenceLengthF ) > 5 ) {
		
		Trace( 0, "Segment Points too far apart" );
		
	      } else {
		
		pSegmentF = new CSentence;
		pSegmentE = new CSentence;
		
		for ( int j1 = pSentenceF->MinIndex(); j1 <= j; j1++ ) {
		  (*pSegmentF)[ j1 ] = (*pSentenceF)[ j1 ];
		}
		for ( int i1 = pSentenceE->MinIndex(); i1 <= i; i1++ ) {
		  (*pSegmentE)[ i1 ] = (*pSentenceE)[ i1 ];
		}
		
		// TODO: Use Tracelevel here
		std::cout << "SegmentF 1: ";
		pSegmentF->Display( m_pTextF->Vocabulary() );
		std::cout << endl;
		std::cout << "SegmentE 1: ";
		pSegmentE->Display( m_pTextE->Vocabulary());
		std::cout << endl;
		
		double LogProbSegment1 = SentenceLogProb( pSegmentF, pSegmentE );
		Trace( 0, "Segment1 LogProb = " << LogProbSegment1 );
		
		delete pSegmentF;
		delete pSegmentE;
		
		// now second half
		pSegmentF = new CSentence;
		pSegmentE = new CSentence;
		
		for ( int j1 = j + 1; j1 <= pSentenceF->MaxIndex(); j1++ ) {
		  (*pSegmentF)[ j1 - j ] = (*pSentenceF)[ j1 ];
		}
		for ( int i1 = i + 1; i1 <= pSentenceE->MaxIndex(); i1++ ) {
		  (*pSegmentE)[ i1 - i ] = (*pSentenceE)[ i1 ];
		}
		// TODO: Use Tracelevel here
		std::cout << "SegmentF 2: "; pSegmentF->Display( m_pTextF->Vocabulary() ); std::cout << endl;
		std::cout << "SegmentE 2: "; pSegmentE->Display( m_pTextE->Vocabulary()); std::cout << endl;
		
		double LogProbSegment2 = SentenceLogProb( pSegmentF, pSegmentE );
		Trace( 0, "Segment2 LogProb = " << LogProbSegment2 );
		
		Trace( 0, "Splittpoint: " << s << " " << j << " " << i
		       << "  Unsplitted: " << LogProb 
		       << "  Splitted: " << (LogProbSegment1 + LogProbSegment2)
		       << "  Difference: " << (LogProb - LogProbSegment1 - LogProbSegment2) );
		
		// store best split points in this sentence pair
		if ( LogProb - LogProbSegment1 - LogProbSegment2 < BestDifference ) {
		  BestDifference = LogProb - LogProbSegment1 - LogProbSegment2;
		  Best_j = j;
		    Best_i = i;
		}
		
	      }
	    }
	  }
	}

	if ( BestDifference < 10000 ) {
	    // TODO: Use TRACE here
	  std::cout << "Best Split Points: " << s << " " << Best_j << " " << Best_i << " " 
	       << SentenceLengthF << " " << SentenceLengthE << " "
	       << BestDifference << endl;
	  CSplitPoint SplitPoint( s, Best_j, Best_i, BestDifference );
	  m_pSegmenter->AddSplitPoint( SplitPoint ); 
	}

      }

      TotalLengthF += SentenceLengthF;
      TotalLogProb += LogProb;

      //PAlignSum += SentenceLengthF * log(1.0 / (float) SentenceLengthE);

      std::cout << m_CurrentIteration << " & " 
	   << TotalLengthF << " & " 
	   << TotalLogProb / TotalLengthF  << endl;
      std::cout << m_CurrentIteration << " & " 
	   << TotalLengthF << " & " 
	   << exp(-TotalLogProb/TotalLengthF) << endl;

    }

    // fetch next sentence pair
    Se = m_pTextE->NextSentence( PosSe );
    Sf = m_pTextF->NextSentence( PosSf );
    
    s++;
  } // end loop over sentence pairs
  
  m_pSegmenter->Display();
  m_pSegmenter->SortToValue();
  // we keep only best split points
  m_pSegmenter->Resize( m_NumberOfSplits );
  m_pSegmenter->Display();

  m_pSegmenter->SortToPosition();
  m_pSegmenter->Display();

  // now we split the sentences
  Trace( 0, "Now the actual splitting" );
  CSplitPoint SplitPoint;
  CSentence *pNewSentence, *pOrgSentence;
  const vector< CSplitPoint > SplitPoints = m_pSegmenter->SplitPoints();

  int SplitsDone = 0;

  for ( int n = SplitPoints.size() - 1; n >= 0; --n ) {
    SplitPoint = SplitPoints[ n ];
    std::cout << SplitPoint;

    // split only if score is negative, i.e. gain in LogProbability
    if ( SplitPoint.m_Score < m_SplitThreshold ) {
      // second segment f (storing first segment will overwrite original sentence)
      pOrgSentence = m_pTextF->GetSentence( SplitPoint.m_Sentence );
      pNewSentence = new CSentence;
      for ( int j = SplitPoint.m_j + 1; j <= pOrgSentence->MaxIndex(); j++ ) {
	(*pNewSentence)[ j - SplitPoint.m_j ] = (*pOrgSentence)[ j ];
      }
      pNewSentence->Display( m_pTextF->Vocabulary() );
      std::cout << " # ";
      m_pTextF->InsertSentence( SplitPoint.m_Sentence + 1, pNewSentence );
      
      // second segment e
      pOrgSentence = m_pTextE->GetSentence( SplitPoint.m_Sentence );
      pNewSentence = new CSentence;
      for ( int i = SplitPoint.m_i + 1; i <= pOrgSentence->MaxIndex(); i++ ) {
	(*pNewSentence)[ i - SplitPoint.m_i ] = (*pOrgSentence)[ i ];
      }
      pNewSentence->Display( m_pTextE->Vocabulary() );
      std::cout << endl;
      m_pTextE->InsertSentence( SplitPoint.m_Sentence + 1, pNewSentence );
      
      
      // first segment f
      pOrgSentence = m_pTextF->GetSentence( SplitPoint.m_Sentence );
      pNewSentence = new CSentence;
      for ( int j = pOrgSentence->MinIndex(); j <= SplitPoint.m_j; j++ ) {
	(*pNewSentence)[ j ] = (*pOrgSentence)[ j ];
      }
      pNewSentence->Display( m_pTextF->Vocabulary() );
      std::cout<< " # ";
      m_pTextF->PutSentence( SplitPoint.m_Sentence, *pNewSentence );
      
      // first segment e
      pOrgSentence = m_pTextE->GetSentence( SplitPoint.m_Sentence );
      pNewSentence = new CSentence;
      for ( int i = pOrgSentence->MinIndex(); i <= SplitPoint.m_i; i++ ) {
	(*pNewSentence)[ i ] = (*pOrgSentence)[ i ];
      }
      pNewSentence->Display( m_pTextE->Vocabulary() );
      std::cout << endl;
      m_pTextE->PutSentence( SplitPoint.m_Sentence, *pNewSentence );

      SplitsDone++;
    }
  }

  Trace( 0, "Number of splits in iteration " << m_CurrentIteration 
	  << " actually performed: " << SplitsDone );
  //m_pTextF->Display();
  //m_pTextE->Display();

  delete m_pSegmenter;
  m_pSegmenter = NULL;

  //m_pLexicon->Display();
  
  //PTransSum = -PTransSum / TotalLengthF;
  //PAlignSum = -PAlignSum / TotalLengthF;
  
  //std::cout << "# TransSum = " << PTransSum << endl;
  //std::cout << "# PAlignSum = " << PAlignSum << endl;
  //std::cout << "# TotalLengthF = " << TotalLengthF << endl;
  
  //PTotal = PAlignSum + PTransSum;
  //std::cout << m_CurrentIteration << " & " << TotalLengthF << " & " << exp(PTransSum) 
  //   << " & " << exp(PAlignSum) << " & " << exp(PTotal) << endl;
  
}



double
CTM_Graph_IBM1::Perplexity()
{
  Trace( 0, "CTM_Graph_IBM1::Perplexity()" );
  Trace( 3, "default Prob " << m_pLexicon->Pr(1, 1) << " 1/VocSize " << 1.0/m_VocSizeF);

  CWordIndex Se, Sf, We, Wf;
  TPosition PosSe, PosSf, PosWe, PosWf;
  //double Sum = 0.0;
  double PEta = 0.0, LogPerplexityAlign = 0.0, LogPerplexityTrans = 0.0, LogPerplexity = 0.0;
  int SentenceLengthE = 0, SentenceLengthF = 0;
  int TotalLengthF = 0;

  double TransProbWord, LogTransProbSentence, LogAlignProbSentence;
  double LogTransProbCorpus = 0.0, LogAlignProbCorpus = 0.0;

  const CSentence *pSentenceE = NULL, *pSentenceF = NULL;

  
  // loop over all sentence pair
  Se = m_pTextE->FirstSentence( PosSe );
  Sf = m_pTextF->FirstSentence( PosSf );
  
  int s = 0;
  while (Se != m_pTextE->Vocabulary()->EOT && 
	 Sf != m_pTextF->Vocabulary()->EOT) {

    if (s % 1000 == 0) {
      Trace( 1, "Sentence pair " << s );
      //std::cout << "Sentence pair " << s << "\n";
    }
    
    pSentenceE = m_pTextE->GetSentence( PosSe );
    pSentenceF = m_pTextF->GetSentence( PosSf );

    SentenceLengthE = pSentenceE->Size();
    SentenceLengthF = pSentenceF->Size();

    // get length of sentences; if one is zero, skip
    if ( (SentenceLengthF == 0) || (SentenceLengthE == 0) ) {

      Trace( 0, "Sentence length zero in sentence pair " << s );

    } else {

      TotalLengthF += SentenceLengthF;
      LogTransProbSentence = 0.0;
      LogAlignProbSentence = 0.0;

      // loop over words in sentence of first text
      int j = 0;
      for (Wf = m_pTextF->FirstWord(PosSf, PosWf);
	   Wf != m_pTextF->Vocabulary()->EOS;
	   Wf = m_pTextF->NextWord(PosSf, PosWf) ) {
	j++;
	// loop over words in sentence of second text
	TransProbWord = 0.0;
	for (We = m_pTextE->FirstWord(PosSe, PosWe);
	     We != m_pTextE->Vocabulary()->EOS;
	     We = m_pTextE->NextWord(PosSe, PosWe) ) {
	  
	  Trace(3, "Wf: " << Wf << " We: " << We);
	  TransProbWord += m_pLexicon->Pr(Wf, We);
	}
	//LogTransProbWord = log( TransProbWord );
	LogTransProbSentence += log( TransProbWord );
      }
      assert( j == SentenceLengthF );
      LogAlignProbSentence = SentenceLengthF * log( 1.0 / (float) SentenceLengthE);

      LogTransProbCorpus += LogTransProbSentence;
      LogAlignProbCorpus += LogAlignProbSentence;

      /*
      std::cout << m_CurrentIteration << " & " 
	   << TotalLengthF << " & " 
	   << LogTransProbCorpus / TotalLengthF << " & " 
	   << LogAlignProbCorpus / TotalLengthF << endl;
      std::cout << m_CurrentIteration << " & " 
	   << TotalLengthF << " & " 
	   << exp(-LogTransProbCorpus/TotalLengthF) << " & " 
	   << exp(-LogAlignProbCorpus/TotalLengthF) << endl;
      */
    }

    // fetch next sentence pair
    Se = m_pTextE->NextSentence(PosSe);
    Sf = m_pTextF->NextSentence(PosSf);
    
    s++;
    
  }				// end loop sentences
  
  LogPerplexityTrans = -LogTransProbCorpus / TotalLengthF;
  LogPerplexityAlign = -LogAlignProbCorpus / TotalLengthF;
  
  LogPerplexity = PEta + LogPerplexityAlign + LogPerplexityTrans;
  
//  std::cout << "Perplexity:\n";
//  std::cout << "\tSentence length: " << exp(PEta) << "\n";
//  std::cout << "\tTranslation:     " << exp(PTrans) << "\n";
//  std::cout << "\tAlignment:       " << exp(PAlign) << "\n";
//  std::cout << "\tTotal:           " << exp(Perplexity) << "\n";

  std::cout << m_CurrentIteration << " & " << TotalLengthF << " & " <<  
  exp(LogPerplexityTrans) << " & " << exp(LogPerplexityAlign) << " & " << exp(LogPerplexity) << "\n";

  m_LexiconPerplexity[ m_CurrentIteration ] = exp( LogPerplexityTrans );
  m_AlignPerplexity[ m_CurrentIteration ] = exp( LogPerplexityAlign );
  m_TotalPerplexity[ m_CurrentIteration ] = exp( LogPerplexity );

  return exp( LogPerplexity );

  //  return 0.0;
}





CLatticeLM *
CTM_Graph_IBM1::ConvertToLattice( const CSentence *pSentence, 
				  const CVocabulary *pVoc )
{
  Trace( 3, "Convert To Lattice: ");
  int s;
  CWordIndex w;

  //pSentence->Display( pVoc );

  CLatticeNode *pFromNode;
  CLatticeNode *pToNode;
  CLatticeEdgeTrans *pEdge;

  CLatticeLM *pLattice = NULL;
  pLattice = new CLatticeLM( m_pLatticeEdgeTrans, NULL );
  assert( pLattice );

  s = pSentence->MinIndex() - 1;
  pFromNode = new CLatticeNode( s );
  pLattice->AddNode( pFromNode );


  for ( s = pSentence->MinIndex(); s <= pSentence->MaxIndex(); s++ ) {
    Trace( 5, "Node " << s );
    pToNode = new CLatticeNode( s );
    pLattice->AddNode( pToNode );

    w = pSentence->GetAt( s );
    pEdge = new CLatticeEdgeTrans( pFromNode, pToNode );
    pEdge->SetWordIndex( w );
    pEdge->SetWord( pVoc->GetWord( w ) );
    pLattice->AddEdge( pEdge );

    pFromNode = pToNode;
  }

  return pLattice;
}



double 
CTM_Graph_IBM1::SentenceLogProb( const CSentence *pSentenceF, const CSentence *pSentenceE )
{
  Trace( 1, "CTM_Graph_IBM1::SentenceLogProb" );

  double TProb;
  double TransProbWord = 0.0;
  double LogTransProbSentence = 0.0;
  double LogAlignProbSentence = 0.0;

  CString TranslationE, TranslationF;

  int SentenceLengthE = pSentenceE->Size();
  int SentenceLengthF = pSentenceF->Size();
  //std::cout << "Sentence length E = " << SentenceLengthE << endl;
  assert( SentenceLengthE > 0 );
  assert( SentenceLengthF > 0 );
  
  // Convert sentence into lattice
  CLatticeLM *pLatticeE = ConvertToLattice( pSentenceE, m_pTextE->Vocabulary() );
  CLatticeLM *pLatticeF = ConvertToLattice( pSentenceF, m_pTextF->Vocabulary() );
  
  //TraceLevel = 0;
  if ( m_pTransMemF ) {
    Trace( 1, "Translation Memory for Source" );
    m_pTransMemF->SetDirection( 0 );
    TranslationF = m_pTransMemF->TranslateLattice( pLatticeF );
    Trace( 1, "Translation Memory for Target" );
    m_pTransMemF->SetDirection( 1 );
    TranslationE = m_pTransMemF->TranslateLattice( pLatticeE );
  }
  //TraceLevel = 1;
      
  // loop over words in sentence of source (first, F) text
  // i.e. loop over all nodes, for each node over all incoming edges
  const CArray< CLatticeEdge * > & EdgesE = pLatticeE->Edges();
  //const CArray< CLatticeEdge * > & EdgesF = pLatticeF->Edges();
  
  CLatticeEdgeTrans *pEdgeE;
  CLatticeEdgeTrans *pEdgeF;
  CLatticeNode *pNodeF;

  int n, j, i;
  
  const CArray< CLatticeNode * > & NodesF = pLatticeF->Nodes();
  
  for ( n = NodesF.MinIndex() + 1; n <= NodesF.MaxIndex(); n++ ) {

    pNodeF = NodesF[ n ];
    const CArray< CLatticeEdge * > & EdgesF = pNodeF->InEdges();
    
    for ( j = EdgesF.MinIndex(); j <= EdgesF.MaxIndex(); j++ ) {

      pEdgeF = (CLatticeEdgeTrans *)EdgesF[ j ];
      //Trace( 0, "f-edge: "); pEdgeF->Display();
      
      TransProbWord = 0.0;
	
      for ( i = EdgesE.MinIndex(); i <= EdgesE.MaxIndex(); i++ ) {
	
	pEdgeE = (CLatticeEdgeTrans *)EdgesE[ i ];
	//Trace( 0, "e-edge: "); pEdgeE->Display();

	TProb = Probability( pEdgeF, pEdgeE );
	//std::cout << TProb << endl;
	//assert( 0 );
	TransProbWord += TProb;
	
      }
      
      LogTransProbSentence += log( TransProbWord );
      
    }
  }

  LogAlignProbSentence = SentenceLengthF * log( 1.0 / (float) SentenceLengthE);
    
  //  std::cout << LogTransProbSentence / SentenceLengthF << " & " 
  //     << LogAlignProbSentence / SentenceLengthF << endl;
  //std::cout << exp( LogTransProbSentence / SentenceLengthF ) << " & " 
  //     << exp( LogAlignProbSentence / SentenceLengthF ) << endl;
  std::cout << exp( LogTransProbSentence ) << " * " << exp( LogAlignProbSentence ) << endl;

  delete pLatticeF;
  delete pLatticeE;
  pLatticeF = NULL;
  pLatticeE = NULL;
  

  return LogAlignProbSentence + LogTransProbSentence;
}


double 
CTM_Graph_IBM1::PhraseTranslationLogProb( const CSentence *pSentenceF, const CSentence *pSentenceE )
{
  // use only lexicon probs:  p( f | e ) = Prod_j Sum_i p(f_j | e_i)

  Trace( 1, "CTM_Graph_IBM1::PhraseTranslationLogProb" );

  double TProb;
  double TransProbWord = 0.0;
  double LogTransProbSentence = 0.0;
  double MaxTProb;
  double ViterbiLogTransProbSentence = 0.0;

  CString TranslationE, TranslationF;

  int SentenceLengthE = pSentenceE->Size();
  int SentenceLengthF = pSentenceF->Size();
  //std::cout << "Sentence length E = " << SentenceLengthE << endl;
  assert( SentenceLengthE > 0 );
  assert( SentenceLengthF > 0 );
  
  // Convert sentence into lattice
  CLatticeLM *pLatticeE = ConvertToLattice( pSentenceE, m_pTextE->Vocabulary() );
  CLatticeLM *pLatticeF = ConvertToLattice( pSentenceF, m_pTextF->Vocabulary() );
  
  //TraceLevel = 0;
  if ( m_pTransMemF ) {
    Trace( 1, "Translation Memory for Source" );
    m_pTransMemF->SetDirection( 0 );
    TranslationF = m_pTransMemF->TranslateLattice( pLatticeF );
    Trace( 1, "Translation Memory for Target" );
    m_pTransMemF->SetDirection( 1 );
    TranslationE = m_pTransMemF->TranslateLattice( pLatticeE );
  }
  //TraceLevel = 1;
      
  // loop over words in sentence of source (first, F) text
  // i.e. loop over all nodes, for each node over all incoming edges
  const CArray< CLatticeEdge * > & EdgesE = pLatticeE->Edges();
  //const CArray< CLatticeEdge * > & EdgesF = pLatticeF->Edges();
  
  CLatticeEdgeTrans *pEdgeE;
  CLatticeEdgeTrans *pEdgeF;
  CLatticeNode *pNodeF;

  int n, j, i;
  
  const CArray< CLatticeNode * > & NodesF = pLatticeF->Nodes();
  
  for ( n = NodesF.MinIndex() + 1; n <= NodesF.MaxIndex(); n++ ) {

    pNodeF = NodesF[ n ];
    const CArray< CLatticeEdge * > & EdgesF = pNodeF->InEdges();
    
    for ( j = EdgesF.MinIndex(); j <= EdgesF.MaxIndex(); j++ ) {

      pEdgeF = (CLatticeEdgeTrans *)EdgesF[ j ];
      //Trace( 0, "f-edge: "); pEdgeF->Display();
      
      TransProbWord = 0.0;
	  MaxTProb = 0.0;

      for ( i = EdgesE.MinIndex(); i <= EdgesE.MaxIndex(); i++ ) {
	
		pEdgeE = (CLatticeEdgeTrans *)EdgesE[ i ];
		//Trace( 0, "e-edge: "); pEdgeE->Display();

		TProb = Probability( pEdgeF, pEdgeE );
		TransProbWord += TProb;
		std::cout << TProb << " " << MaxTProb << endl;
		if ( TProb > MaxTProb ) {
		  MaxTProb = TProb;
		}
	
      }
      
      LogTransProbSentence += log( TransProbWord );
      ViterbiLogTransProbSentence += log( MaxTProb );

    }
  }

  std::cout << "TotalProb       " << exp( LogTransProbSentence ) << endl;
  std::cout << "TotalProbV      " << exp( ViterbiLogTransProbSentence ) << endl;

  std::cout << "PerWordProb     " << exp( LogTransProbSentence / SentenceLengthF ) << endl;
  std::cout << "PerWordProbV    " << exp( ViterbiLogTransProbSentence / SentenceLengthF ) << endl;

  std::cout << "PerWordLogProb  " << LogTransProbSentence / SentenceLengthF  << endl;
  std::cout << "PerWordLogProbV " << ViterbiLogTransProbSentence / SentenceLengthF  << endl;

  delete pLatticeF;
  delete pLatticeE;
  pLatticeF = NULL;
  pLatticeE = NULL;
  

  return LogTransProbSentence;
}




double
CTM_Graph_IBM1::Probability( CLatticeEdgeTrans *pEdgeF, CLatticeEdgeTrans *pEdgeE )
{
  double TProb = 0.0;

  // lexicon probability:
  // for transducer added edges from same transducer item: TPenalty == log(0.99999)
  // else: lexicon prob

  CWordIndex Wf = GetWordIndexF( pEdgeF );
  CWordIndex We = GetWordIndexE( pEdgeE );

  if ( (pEdgeF->TransducerItem() != NULL) && (pEdgeE->TransducerItem() != NULL) ) {
    
    CString Tf = pEdgeF->Transducer()->Name();
    CString Te = pEdgeE->Transducer()->Name();
    Trace( 2, "Transducers: " << "-" << Tf << "-" << Te );
    
    int Nf = pEdgeF->TransducerItem()->Number();
    int Ne = pEdgeE->TransducerItem()->Number();
    Trace( 2, "Transducer item: " << Nf << " " << Ne );
    
    if ( (Tf == Te) && (Nf == Ne) ) {
      Trace( 2, "Same transducer and same item: " << Tf << " " << Te );
      Trace( 2, "TProb from transducer= " 
	     << pEdgeE->TransducerItem()->ScoreSourceGivenTarget() );
      
      // TPenalty = -log( 0.99999 );
      TProb = pEdgeE->TransducerItem()->ScoreSourceGivenTarget();
      //Sum1 += TProb;
    }
  } else {
    Trace( 0, "Wf - We: " << Wf << " " << pEdgeF->Word() << " " << We << " " << pEdgeE->Word() );
    TProb = m_pLexicon->Pr( Wf, We ); //log(m_pT->GetAt(We,Wf));
    //Sum1 += TProb;
    Trace( 0, "TProb from lexicon = " << TProb );
  }
  //assert(0);

  return TProb;
}



CWordIndex
CTM_Graph_IBM1::GetWordIndexE( CLatticeEdgeTrans *pEdgeE )
{
  CWordIndex We;

  // get word index: normal edge, then already known;
  // for edge with category, use source string
  if ( pEdgeE->TransducerItem() != NULL) {
    We = m_pTextE->Vocabulary()->GetIndex( pEdgeE->TransducerItem()->Source().c_str() );
    Trace( 2, "Edge with category: " << pEdgeE->TransducerItem()->Source() 
	   << " WordIndex = " << We );
  } else {
    We = pEdgeE->WordIndex();
  }
  
  if ( We == 0 ) {
    Trace( 1, "UNK " << pEdgeE->Word() );
    We = m_pTextE->Vocabulary()->GetIndex( pEdgeE->Word() );
    Trace( 1, "Set index to " << We );
    pEdgeE->SetWordIndex( We );
  }

  return We;
}


CWordIndex
CTM_Graph_IBM1::GetWordIndexF( CLatticeEdgeTrans *pEdgeF )
{
  CWordIndex Wf;

  // get word index: normal edge, then already known;
  // for edge with category, use source string
  if ( pEdgeF->TransducerItem() != NULL) {
    Wf = m_pTextF->Vocabulary()->GetIndex( pEdgeF->TransducerItem()->Source().c_str() );
    Trace( 2, "Edge with category: " << pEdgeF->TransducerItem()->Source() 
	   << " WordIndex = " << Wf );
  } else {
    Wf = pEdgeF->WordIndex();
  }
  
  if ( Wf == 0 ) {
    Trace( 1, "UNK " << pEdgeF->Word() );
    Wf = m_pTextF->Vocabulary()->GetIndex( pEdgeF->Word() );
    Trace( 1, "Set index to " << Wf );
    pEdgeF->SetWordIndex( Wf );
  }

  return Wf;
}


