#include <config.h>

#ifndef __BFLM__
#define __BFLM__

#include <iostream>
#include <fstream>

#include <LMInterface.hh>
#include <option.h>
#include <File.h>
//#include <LHash.h>
#include <Ngram.h>
#include <LM.h>
#include <Vocab.h>
#include <Prob.h>
#include <BloomFilter.h>

using namespace std;

class BFLM;
class BFLMSentenceState;

#define NN 3


class BFLMSentenceState : public CLMSentenceState
{
public:
  BFLMSentenceState( const BFLM* theLM );
  BFLMSentenceState();
  ~BFLMSentenceState();

  //! generate an lmstate for start of sentence <s>
  virtual CLMState BeginOfSentenceState();

  /** returns the negative difference of the param state and
	  the state that results after the merge */
  virtual double LogProb( const CLMState LMState, const LMWordIndex WordIndex );

  //! Get the logarithmic probability of a single word, also return the new state.
  //virtual double LogProb( CLMState LMState, const LMWordIndex WordIndex, CLMState &NewLMState );


  /** caclulate the probability for a sequence of words and return two LM states
	  one for the left side, one for the right side; start of sentence (<s>) is implicitily added */
  virtual double LogProb( vector< string > Words, CLMState& LMStateLeft, CLMState& LMStateRight );

  /** Get the logarithmic probability of extending lmstate by words
	  which led to the creation of lmstate2. Also return the new LM state
	  in lmstate2 */
  virtual double LogProbMerge( CLMState &LMState_Left_Left, CLMState LMState_Left_Right,
                               CLMState LMState_Right_Left, CLMState &LMState_Right_Right );

  //! return log prob for adding sentence end </s>
  virtual double LogProbEnd( CLMState LMState );

  //! generate new lmstate from old lmstate and word
  virtual CLMState AppendWord( CLMState LMState, const LMWordIndex WordIndex );

  //! delete the LM state
  virtual void DeleteState( CLMState LMState );


private:
  BFLM* lm;
  VocabIndex *wids;
};




class BFLM : public CLM
{
  friend class BFLMSentenceState;

public:
int order;

  //! create the LM from a parameter file
  BFLM( char* paramFile, unsigned order = defaultNgramOrder, int size=1000, int numberhashfunctions=3 );

  //! create the LM with a given vocabulary to filter the n-grams during loading
  BFLM( char* File, int Order, const vector< string > &Vocabulary, bool RestrictVocabulary );

  //! destructor
  virtual ~BFLM();


  //! return the order ( = history length + 1 ) of the LM
  unsigned GetOrder() { return order; }



  //! create a new sentence state
  virtual BFLMSentenceState* NewSentence() const;

  //! delete the sentence state
  virtual void DeleteSentence(CLMSentenceState* st) const;



  //! return the size of the vocabulary
  virtual LMWordIndex VocabularySize() const;

  // pure virtual functions that must be implemented, but are currently
  // not used anywhere in the system

  //! return the vocabulary as a vector of strings
  virtual const vector< string > & Vocabulary() const;

  //! compare a vocabulary to the vocabulary given in the parameter file
  virtual int Compare( char ** Vocab, char *Paramfile );

  //! get the LM internal word index for word
  virtual LMWordIndex GetVocabularyIndex( string word ) const;
  virtual LMWordIndex GetVocabularyIndex( string word );



  //ARV 07/11 used for future cost estimation
  virtual double IsolatedLogProb( vector<string> wordSequence );


private:
  //! the SRI LM
  bloom_filter* ngramLM;


  //! the vocabulary
Vocab* vocab;

  // sv 060604: this is never used except for returning it, which means it is always empty
  vector< string > m_vocabVector;

  //! load only n-grams which are covered by the vocabulary
	bool m_RestrictVocabulary;

};





inline const vector<string> & 
BFLM::Vocabulary() const
{
    return m_vocabVector;
}



inline int 
BFLM::Compare(char ** Vocab, char *Paramfile)
{
    return 0;
}



#endif
