#include <stdlib.h>
#include <config.h>
#include <fstream>
#include <BoolBFLM.hh>
//#include <LHash.h>
#include <utils.hh>
#include <BloomFilter.h>

#include "GeneralHashFunctions.h"

using namespace std;


BoolBFLMSentenceState::BoolBFLMSentenceState( const BoolBFLM* theLM )
{
  lm = (BoolBFLM*)theLM;
  wids = new VocabIndex[lm->GetOrder()];
}



BoolBFLMSentenceState::~BoolBFLMSentenceState()
{
  lm = NULL;  
  delete [] wids;
}



inline CLMState 
BoolBFLMSentenceState::BeginOfSentenceState()
{
  unsigned LMorder = lm->GetOrder();
  CLMState ret;
  if (LMorder == (unsigned)4)
  {
    // lmstate, 64bit: |1bit|- 21bit: mostRecentWord -|- 21bit nextRecentWord -|- 21bit thirdRecentWord -|
    // here all three filled with <s>, like the speech people do it.
    ret = ((CLMState)( lm->vocab->ssIndex ) | ((CLMState)( lm->vocab->ssIndex ) << 21) | ((CLMState)( lm->vocab->ssIndex ) << 42) );
  }
  else
  {
    ret = ((CLMState)0 | ((CLMState)( lm->vocab->ssIndex ) << 32) );
  }
  //CLMState ret = (CLMState)CLM_INIT;
  return ret;
}


double
BoolBFLMSentenceState::LogProb( CLMState lmstate, const LMWordIndex word )
{
  //cout << "Word: "<< word << " -- " << lm->vocab->getWord(word)<<endl; 
  
  // we don't use the proper prob for unknow word, but a fixed prob
  if ( word == lm->vocab->unkIndex ){
     //cout << "Unknown word " << word << endl;
    return log10( 1.0 / lm->VocabularySize() );
  }

  unsigned LMorder = lm->GetOrder();
  if (LMorder == (unsigned)4)
  {
      // lmstate, 64bit: |1bit|- 21bit: mostRecentWord -|- 21bit nextRecentWord -|- 21bit thirdRecentWord -|
      VocabIndex mostRecentWord = lmstate >> 42;
      VocabIndex nextRecentWord = (lmstate >> 21) & ( ( (CLMState)1 << 21 ) - 1 );
      VocabIndex thirdRecentWord = lmstate & ( ( (CLMState)1 << 21 ) - 1 );

      wids[ 0 ] = mostRecentWord;
      wids[ 1 ] = nextRecentWord;
      wids[ 2 ] = thirdRecentWord;
      wids[ 3 ] = Vocab_None;
          /*
          cout << "LogP:" << logP << " for " << lm->vocab->getWord(word)
          << " | " << mostRecentWord << "," << nextRecentWord 
          << "," << thirdRecentWord << endl;
          cout << " | " << lm->vocab->getWord(mostRecentWord) << " " <<lm->vocab->getWord(nextRecentWord)
          << " " << lm->vocab->getWord(thirdRecentWord)<<endl;  
          */
  }
  else
  {
      long mostRecentWord = lmstate >> 32;
      long nextRecentWord = lmstate & ( ( (CLMState)1 << 32 ) - 1 );

      wids[ 0 ] = mostRecentWord;
      wids[ 1 ] = nextRecentWord;
      wids[ 2 ] = Vocab_None;
  }
  //cout << "wordprob call" << endl;
  double logP = lm->ngramLM->wordProb( word, wids );
  return logP;
}



double
BoolBFLMSentenceState::LogProb( vector< string > Words, CLMState& LMStateLeft, CLMState& LMStateRight )
{
  //cout << "BoolBFLMSentenceState::LogProb" << endl;
	//cout << "wordprob call" << endl;
  // we start with BeginOfSentence LM states
  CLMState LMState = BeginOfSentenceState();

  double logP = 0.0;
  double SumLogP = 0.0;

  LMWordIndex w;

  // loop over all given words and accumulate the log prob
  for ( int i = 0; i < Words.size(); i++ ) {
    //cout << Words[ i ] << " ";
    w = lm->GetVocabularyIndex( Words[ i ] );
    logP = LogProb( LMState, w );
    SumLogP += logP;

    //	cout << LMState << " " << Words[ i ] << " " << w << " " << logP;

    LMState = AppendWord( LMState, w );
    //	cout << " " << LMState << endl;
    
    // we want to generate a left LMState, encoding the
    // n left-most words in the word sequence
    if ( i + 1 < NN ) {
      LMStateLeft = LMState;
      //	  cout << "new left state = " << LMStateLeft << endl;
    }

  }

  LMStateRight = LMState;
  
  //cout << "Final state: " << LMStateLeft << " " << LMStateRight << endl;

  return SumLogP;
}


/* Is not modified yet for 4-gram LM. */

/**
   Get the logarithmic probability of a extending lmstate by words
   which led to the creation of lmstate2
   update the outer lmstates, which is necessary for word sequences
   shorter than the history length */
double
BoolBFLMSentenceState::LogProbMerge( CLMState &lmstate_left_left, CLMState lmstate_left_right,
								  CLMState lmstate_right_left, CLMState &lmstate_right_right )
{
  //  cout << "MERGE \n";
  //cout << "ll " << lmstate_left_left << " " << " lr " << lmstate_left 
  //<< " rl " << lmstate_right << " rr " << lmstate_right_right << endl;

  long word_ll_1, word_lr_1, word_rl_1, word_rr_1;
  long word_ll_2, word_lr_2, word_rl_2, word_rr_2;

  word_rl_1 = lmstate_right_left & (((CLMState)1 << 32) - 1);
  word_rl_2 = lmstate_right_left >> 32;

  /*
  long ord_ll_1 = lmstate_left_left &(((CLMState)1 << 32) -1);
  long ord_ll_2 = lmstate_left_left >> 32;
  cout << "ll 1 " << lm->vocab->getWord( Word_ll_1 ) << " ll 2 " << lm->vocab->getWord( Word_ll_2 ) << endl;

  long ord_lr_1 = lmstate_left &(((CLMState)1 << 32) -1);
  long ord_lr_2 = lmstate_left >> 32;
  cout << "lr 1 " << lm->vocab->getWord( Word_lr_1 ) << " lr 2 " << lm->vocab->getWord( Word_lr_2 ) << endl;

  long ord_rl_1 = lmstate_right &(((CLMState)1 << 32) -1);
  long ord_rl_2 = lmstate_right >> 32;
  cout << "rl 1 " << lm->vocab->getWord( Word_rl_1 ) << " rl 2 " << lm->vocab->getWord( Word_rl_2 ) << endl;

  long ord_rr_1 = lmstate_right_right &(((CLMState)1 << 32) -1);
  long ord_rr_2 = lmstate_right_right >> 32;
  cout << "rr 1 " << lm->vocab->getWord( Word_rr_1 ) << " rr 2 " << lm->vocab->getWord( Word_rr_2 ) << endl;
  */


  // we recover the history, i.e. the words from the lmstate_right
  // then expand the lmstate_left over these words
  // but first need to subtract the logprobs calculated before

  double LogProbOld = 0.0;
  double LogProbNew = 0.0;

  // first we recalculate the logprobs which were calculated without history
  CLMState lmstate_old = BeginOfSentenceState();

  if ( word_rl_1 == lm->vocab->ssIndex ) {
	// <s> is used to fill the 2 word history, if only one word has been seen
	// to create the lmstate_right 
  } else {
	LogProbOld = LogProb( lmstate_old, word_rl_1 );
	lmstate_old = AppendWord( lmstate_old, word_rl_1 );
  }

  LogProbOld += LogProb( lmstate_old, word_rl_2 );
  lmstate_old = AppendWord( lmstate_old, word_rl_1 );
  //cout << "LogProbOld = " << LogProbOld << endl;


  // now we calculate the logprobs with the correct history
  if ( word_rl_1 == lm->vocab->ssIndex ) {
	// again, skip of <s> used to fill the 2 word history
  } else {
	LogProbNew = LogProb( lmstate_left_right, word_rl_1 );
	lmstate_left_right = AppendWord( lmstate_left_right, word_rl_1 );
  }
  
  LogProbNew += LogProb( lmstate_left_right, word_rl_2 );
  lmstate_left_right = AppendWord( lmstate_left_right, word_rl_2 );
  //cout << "LogProbNew = " << LogProbNew << endl;


  // and we adjust the LM states
  if ((lmstate_right_right & 0x00000000FFFFFFFF)==lm->vocab->ssIndex) {
    //cout << "Right_Right_State angepasst!!!!!!!!!!\n";
    lmstate_right_right = lmstate_left_right;
    
    word_rr_1 = lmstate_right_right &(((CLMState)1 << 32) -1);
    word_rr_2 = lmstate_right_right >> 32;
    /*
    cout << "in merge: \n";
    cout << "rightfirst: " << word_rr_1 << endl;
    cout << "rightsecond: " << word_rr_2 << endl;
    
    cout << "after change\n";
	*/
  }

  if ((lmstate_left_left & 0x00000000FFFFFFFF)==lm->vocab->ssIndex) {
    if ( word_rl_1 == lm->vocab->ssIndex ) {
      lmstate_left_left = AppendWord( lmstate_left_left, word_rl_2 );
    } else {
      lmstate_left_left = AppendWord( lmstate_left_left, word_rl_1 );
    }
    

  }

  /*
  Word_ll_1 = lmstate_left_left &(((CLMState)1 << 32) -1);
  Word_ll_2 = lmstate_left_left >> 32;
  cout << "ll 1 " << lm->vocab->getWord( Word_ll_1 ) << " ll 2 " << lm->vocab->getWord( Word_ll_2 ) << endl;

  Word_lr_1 = lmstate_left &(((CLMState)1 << 32) -1);
  Word_lr_2 = lmstate_left >> 32;
  cout << "lr 1 " << lm->vocab->getWord( Word_lr_1 ) << " lr 2 " << lm->vocab->getWord( Word_lr_2 ) << endl;

  Word_rl_1 = lmstate_right &(((CLMState)1 << 32) -1);
  Word_rl_2 = lmstate_right >> 32;
  cout << "rl 1 " << lm->vocab->getWord( Word_rl_1 ) << " rl 2 " << lm->vocab->getWord( Word_rl_2 ) << endl;

  Word_rr_1 = lmstate_right_right &(((CLMState)1 << 32) -1);
  Word_rr_2 = lmstate_right_right >> 32;
  cout << "rr 1 " << lm->vocab->getWord( Word_rr_1 ) << " rr 2 " << lm->vocab->getWord( Word_rr_2 ) << endl;
  */
  
  // the net gain is new LogProbOld - LogProbNew
  return LogProbNew - LogProbOld;
}



double
BoolBFLMSentenceState::LogProbEnd( CLMState lmstate )
{
  return LogProb( lmstate, lm->vocab->seIndex );
}



CLMState
BoolBFLMSentenceState::AppendWord( CLMState lmstate, const LMWordIndex word )
{
  //cout << "Appending:" << lm->vocab->getWord(word) << endl;
  unsigned LMorder = lm->GetOrder();
  CLMState ret;
  
  if (LMorder == (unsigned)4)
  {
      // we have only space for a 21 bit index here (LMWordIndex is 32, CLMState is 64)
      // this limits the vocabulary size to 2,097,151
      if (word >= (((CLMState)1 << 21) - 1))
      {
          cout << "Vocabulary index too big for 4-gram LM!\n";
          exit(-1);
      }
     
      // lmstate, 64bit: |1bit|- 21bit: mostRecentWord -|- 21bit nextRecentWord -|- 21bit thirdRecentWord -|
      CLMState mostRecentWord = (CLMState)word;
      CLMState wordHistory = lmstate >> 21;  
      ret = ( (CLMState)wordHistory | ((CLMState)mostRecentWord << 42) );
      
      /*
       VocabIndex mostRecentWordT = ret >> 42;
       VocabIndex thirdRecentWordT = ret & ( ( (CLMState)1 << 21 ) - 1 );
       VocabIndex nextRecentWordT = (ret >> 21) & ( ( (CLMState)1 << 21 ) - 1 );
       cout << "Now in history: " << mostRecentWordT << "," << nextRecentWordT 
       << "," <<  thirdRecentWordT <<endl; 
       cout << " => " << lm->vocab->getWord(mostRecentWordT) << " " << lm->vocab->getWord(nextRecentWordT)
       << " " << lm->vocab->getWord(thirdRecentWordT) << endl;
       */
   }
   else // trigram
   {
      CLMState mostRecentWord= (CLMState)word;
      CLMState nextRecentWord = lmstate >> 32;
      
      // 2nd item, then 1st
      ret = ( (CLMState)nextRecentWord | ((CLMState)mostRecentWord << 32) );
   }
  return ret;
}



void
BoolBFLMSentenceState::DeleteState( CLMState lmstate )
{
  // nothing to eo
}



// Param file for BoolBFLM takes the arpa model only
BoolBFLM::BoolBFLM( char* paramFile, unsigned order, int size, int numberhashfunctions )
{
	this->order=order;
	
	//Instantiate Bloomfilter and vocabulary
	ngramLM=new bool_bloom_filter(size*1000000,numberhashfunctions);
	vocab= new Vocab();
	
	//Open file
	ifstream lmin;
	lmin.open(paramFile);
	
	int inserted=0;
	char item[500];
	char delim='\t';
	char* word;
	VocabIndex i;
	int frequency;
	
	LMWordIndex index=0;
	vector<LMWordIndex> indexes;
	

	
	while(!lmin.eof()) {

	if (inserted % 500000==0) { cout << inserted << endl; }

	delim='\t';
	lmin.getline(item,50000,delim);

	//Split string into words -> ngram
	indexes.erase(indexes.begin(), indexes.end());;
	//	cout << indexes.size() << endl;
		
	word=strtok(item," ");
	while (word)
	{
		//Is this word already known in the vocabulary?
		i=vocab->getIndex(word);
		
		//Do we know that word?
		if (i==Vocab_None) 
		{
			//no..assign new index
			i=vocab->addWord(word);
		}
		//push back in vector
		indexes.push_back(i);
		word=strtok(NULL," ");
	}
	
	//Get frequency
	delim='\n';
	lmin.getline(item,50000,delim);
    int frequency=atoi(item);
	//cout << "Frequency: " <<frequency << endl;
	
	ngramLM->insert(indexes);

	//int f=ngramLM->getfrequency(indexes);
	//if (f!=frequency) {cout << f << "-" << frequency << endl; }
	
	inserted++;
	//cout << "Number inserted: " << inserted << endl;
	//cout << "Index size: "<< indexes.size() << endl;
	
	}
	//cout << vocab->getIndex("<s>");
	//cout << vocab->getIndex("</s>");

}




BoolBFLM::BoolBFLM( char* File, int Order, const vector< string > &Vocabulary, bool RestrictVocabulary )
{


}



BoolBFLM::~BoolBFLM()
{
  delete ngramLM;
}



BoolBFLMSentenceState*
BoolBFLM::NewSentence() const
{
  return new BoolBFLMSentenceState(this);
}



void
BoolBFLM::DeleteSentence( CLMSentenceState* st ) const
{
  delete st;
}



LMWordIndex
BoolBFLM::VocabularySize() const
{
  return this->vocab->numWords();
  //return GetVocabularyIndexMap.size();
}



LMWordIndex
BoolBFLM::GetVocabularyIndex( string word ) const
{
  

	VocabIndex i = vocab->getIndex( word.c_str() );
  if ( i == Vocab_None ) {
    //cout<<"unk word:"<<word<<" - "<<vocab->unkIndex<<endl;
    return vocab->unkIndex;
  }
  return i; 
}



LMWordIndex
BoolBFLM::GetVocabularyIndex( string word )
{
//cout << word;
  VocabIndex i = vocab->getIndex( word.c_str() );
  if ( i == Vocab_None ) {
	//cout<<"unk word:"<<word<<" - "<<vocab->unkIndex<<endl;
	return vocab->unkIndex;
  }
  //cout << i << endl;
  return i; 
}



double
BoolBFLM::IsolatedLogProb( vector<string> wordSequence )
{
  vector<LMWordIndex> wiVector;
  //cout<<"Evaluating: ";
  for ( int i = 0; i < wordSequence.size(); i++ ) {
    // cout << wordSequence[ i ] << " ";
    wiVector.push_back( GetVocabularyIndex( wordSequence[ i ] ) );
  }
  //cout<<endl;

  double sumLogP = 0;
  //now the wi vetor is filled with the lm word indices, for each one, create a context
  //of varying length, and query the lm
  for ( int i = 0; i < wiVector.size(); i++ ) {
    LMWordIndex wq = wiVector[i];
    int contextSize = i;

    double logP = 0.0;
    if ( wq == vocab->unkIndex ) {
      cout << "xxUnknown word" << wordSequence[ i ] << " ";
      logP = log10( 1.0 / VocabularySize() );
    }else{
      //context + 1 for the Vocab_None
      int finalSize;
      MAX( finalSize, contextSize + 1, 3 );

      VocabIndex* wids = new VocabIndex[ finalSize ];
      //cout<<"w: "<<wordSequence[i]<<" context: ";
      
      int cIndex =0;
      //needs to be in reverse order, uptil 3
      for ( int c = i-1 ; c >= i-2 && c >= 0; c-- ) {
		wids[ cIndex ] = wiVector[ c ];
		cIndex++;
		//cout<<wordSequence[c]<<" ";
      }
      wids[ cIndex ] = Vocab_None;
      //cout<<"Vocab_None"<<endl;
//cout << "wordprob call" << endl;
      logP = ngramLM->wordProb( wq, wids );
      delete [] wids;
    }
    //cout<<"logP: "<<wordSequence[i]<<" = "<<logP<<endl;
    sumLogP += logP;
  }
  return sumLogP;
}

