#ifndef INCLUDE_BOOL_BLOOM_FILTER_H
#define INCLUDE_BOOL_BLOOM_FILTER_H

#include <iostream>
#include <vector>
#include <sstream>
#include <LMInterface.hh>
#include <Vocab.h>
#include <math.h>

#include "GeneralHashFunctions.h"


const unsigned int bchar_size = 0x08;    // 8 bits in 1 char(unsigned)

const unsigned char bbit_mask[8] = {
                                   0x01, //00000001
                                   0x02, //00000010
                                   0x04, //00000100
                                   0x08, //00001000
                                   0x10, //00010000
                                   0x20, //00100000
                                   0x40, //01000000
                                   0x80  //10000000
                                 };

class bool_bloom_filter
{

	public:

  bool_bloom_filter(unsigned long tbl_size, unsigned int hash_functions)
	{
		//cout << tbl_size << endl;
		table_size = tbl_size;
		tbl_char=table_size*bchar_size;
		number_hash_functions=hash_functions;
		hash_table = new unsigned char[table_size];
		register_default_hash_functions(hash_functions);
		cout << "Boolean BF created with size " << table_size << " byte and " << hash_functions << " hash functions" << endl;
	}

   ~bool_bloom_filter()
   {
      delete[] hash_table;
   }

   void register_default_hash_functions(unsigned int hash_functions)
   {

      if (hash_functions>0) {hash_function.push_back(RSHash  );}
      if (hash_functions>1) {hash_function.push_back(JSHash  );}
      if (hash_functions>2) {hash_function.push_back(PJWHash );}
      if (hash_functions>3) {hash_function.push_back(BKDRHash);}
      if (hash_functions>4) {hash_function.push_back(SDBMHash);}
      if (hash_functions>5) {hash_function.push_back(DJBHash );}
      if (hash_functions>6) {hash_function.push_back(DEKHash );}
      if (hash_functions>7) {hash_function.push_back(APHash  );}
   }
  
	void insert(vector<LMWordIndex> key)
	{
		//insert 1 in bitarray for each hash function
		for(unsigned int i = 0; i < number_hash_functions; i++)
		{
			unsigned int hash = hash_function[i](key) % (tbl_char);
			hash_table[hash / bchar_size] |= bbit_mask[hash % bchar_size];
		}
		//cout << "inserted" << endl;
	}
   
	double wordProb(VocabIndex word, const VocabIndex *context)
	{
		unsigned int clen = Vocab::length(context);
		
		cont.erase(cont.begin(), cont.end());
		
		for (int i=0; i<clen;i++){
		//cont[i]=context[i];
			cont.insert(cont.end(), context[i]);
		}
		cont.insert(cont.end(), word);
		
		if (contains(cont)) { return 1;} else { return 0.0000001;}
		
	    
		
	}
	
	bool contains(vector<LMWordIndex> key)
   {
		//test for each hashfunction
      for(std::size_t i = 0; i < number_hash_functions; i++)
      {
         unsigned int hash = hash_function[i](key) % (tbl_char);
         unsigned int bit  = hash % bchar_size;
         if ((hash_table[hash / bchar_size] & bbit_mask[bit]) != bbit_mask[bit])
         {
            return false;
         }
      }
      return true;
   }
	
	
	private:

	std::vector < HashFunction > hash_function;
	int 									number_hash_functions;
	unsigned char*        		hash_table;
	unsigned long               table_size;	
	unsigned long  				tbl_char;
	vector<LMWordIndex> 	cont;

};

#endif
