#include "GeneralHashFunctions.h"

unsigned int RSHash(vector<LMWordIndex> index)
{
   unsigned int b    = 378551;
   unsigned int a    = 63689;
   unsigned int hash = 0;

   for(unsigned int i = 0; i < index.size(); i++)
   {
      hash = hash * a + index[i];
      a    = a * b;
   }

   return hash;
}


unsigned int JSHash(vector<LMWordIndex> index)
{
   unsigned int hash = 1315423911;

   for(unsigned int i = 0; i < index.size(); i++)
   {
      hash ^= ((hash << 5) + index[i] + (hash >> 2));
   }

   return hash;
}



unsigned int PJWHash(vector<LMWordIndex> index)
{
   unsigned int BitsInUnsignedInt = (unsigned int)(sizeof(unsigned int) * 8);
   unsigned int ThreeQuarters     = (unsigned int)((BitsInUnsignedInt  * 3) / 4);
   unsigned int OneEighth         = (unsigned int)(BitsInUnsignedInt / 8);
   unsigned int HighBits          = (unsigned int)(0xFFFFFFFF) << (BitsInUnsignedInt - OneEighth);
   unsigned int hash              = 0;
   unsigned int test              = 0;

   for(unsigned int i = 0; i < index.size(); i++)
   {
      hash = (hash << OneEighth) + index[i];

      if((test = hash & HighBits)  != 0)
      {
         hash = (( hash ^ (test >> ThreeQuarters)) & (~HighBits));
      }
   }

   return hash;
}

unsigned int ELFHash(vector<LMWordIndex> index)
{
   unsigned int hash = 0;
   unsigned int x    = 0;

   for(unsigned int i = 0; i < index.size(); i++)
   {
      hash = (hash << 4) + index[i];
      if((x = hash & 0xF0000000L) != 0)
      {
         hash ^= (x >> 24);
      }
      hash &= ~x;
   }

   return hash;
}



unsigned int BKDRHash(vector<LMWordIndex> index)
{
   unsigned int seed = 131; // 31 131 1313 13131 131313 etc..
   unsigned int hash = 0;

   for(unsigned int i = 0; i < index.size(); i++)
   {
      hash = (hash * seed) + index[i];
   }

   return hash;
}



unsigned int SDBMHash(vector<LMWordIndex> index)
{
   unsigned int hash = 0;

   for(unsigned int i = 0; i < index.size(); i++)
   {
      hash = index.at(i) + (hash << 6) + (hash << 16) - hash;
   }

   return hash;
}



unsigned int DJBHash(vector<LMWordIndex> index)
{
   unsigned int hash = 5381;

   for(unsigned int i = 0; i < index.size(); i++)
   {
      hash = ((hash << 5) + hash) + index[i];
   }

   return hash;
}



unsigned int DEKHash(vector<LMWordIndex> index)
{
   unsigned int hash = static_cast<unsigned int>(sizeof(index));

   for(unsigned int i = 0; i < index.size(); i++)
   {
      hash = ((hash << 5) ^ (hash >> 27)) ^ index[i];
   }

   return hash;
}



unsigned int BPHash(vector<LMWordIndex> index)
{
   unsigned int hash = 0;
   for(unsigned int i = 0; i < index.size(); i++)
   {
      hash = hash << 7 ^ index[i];
   }

   return hash;
}



unsigned int FNVHash(vector<LMWordIndex> index)
{
   const unsigned int fnv_prime = 0x811C9DC5;
   unsigned int hash = 0;
   for(unsigned int i = 0; i < index.size(); i++)
   {
      hash *= fnv_prime;
      hash ^= index[i];
   }

   return hash;
}



unsigned int APHash(vector<LMWordIndex> index)
{
   unsigned int hash = 0;

   for(unsigned int i = 0; i < index.size(); i++)
   {
      hash ^= ((i & 1) == 0) ? (  (hash <<  7) ^ index[i] ^ (hash >> 3)) :
                               (~((hash << 11) ^ index[i] ^ (hash >> 5)));
   }

   return hash;
}

