#include <config.h>
#include <fortify.h>
#ifndef __PartsofSpeech__
#define __PartsofSpeech__

#include <map>
#include <set>
#include <string>
#include <vector>
#include <algorithm>
#include <strstream>
#include <fstream>
#include <iostream>
#include <cmath>

#include <LMInterface.hh>
#include <suffixsort.hh>
#include <writeSmall.hh>

class CLMposSA;
class CLMposSASentenceState;

typedef vector<LMSuffix>::iterator suffixptr;

class CLMposSASentenceState : public CLMSentenceState {
  friend class CLMposSA;

public:
  ~CLMposSASentenceState();
  CLMState BeginOfSentenceState();
  inline double Prob(const CLMState lmstate, const LMWordIndex word) ;
  inline double LogProb(CLMState lmstate, const LMWordIndex word) ;
  CLMState AppendWord(CLMState lmstate, const LMWordIndex word) ;
  void DeleteState(CLMState lmstate);
  double ProbEnd(const CLMState lmstate);
  double LogProbEnd(CLMState lmstate);
  virtual void AllProb(CLMState lmstate, double p[]) ;
  virtual void AllLogProb(CLMState lmstate, double logp[]) ;
  virtual void NBestProb(CLMState lmstate, unsigned int N, 
                         LMWordIndex wordindices[], double nbestp[]) ;
  virtual void NBestLogProb(CLMState lmstate, unsigned int N, 
                            LMWordIndex wordindices[], 
                            double nbestlogp[]) ;
  string GetPosTag(CLMState lmstate);
protected:
  CLMposSA *lm;
  CLMposSASentenceState(CLMposSA* LMod);

private:
  double p(vector<LMWordIndex>::const_iterator,  // p(c|h)
           vector<LMWordIndex>::const_iterator);
  double posp(vector<LMWordIndex>::const_iterator, // p(w|c,h)
              vector<LMWordIndex>::const_iterator,
              LMWordIndex, LMWordIndex);
  pair<LMWordIndex, double> getposp(vector<LMWordIndex>::const_iterator,
    vector<LMWordIndex>::const_iterator,const LMWordIndex); //best c and posp(c)
  LMWordIndex getpos(CLMState, LMWordIndex); // returns getposp.first

  bool cache;

  LMWordIndex m_VocSize;
  vector<LMSuffix> getdistinct;             

  void calculate_distinct();                // calculates vector getdistinct
  map<pair<LMWordIndex,LMWordIndex>,unsigned int> pospNwc;
  map<LMWordIndex,unsigned int> pospNc;
  LMSuffix realhistorysize, realposition;
} ;

class CLMposSA : public CLM {
  friend class CLMposSASentenceState;

public: 
  CLMposSASentenceState * NewSentence() const ;

  LMWordIndex VocabularySize() const;
  LMWordIndex GetVocabularyIndex(string) ;

  const vector<string> & Vocabulary() const ;

  int Compare(char ** Vocab, char *Paramfile);

  CLMposSA();
  ~CLMposSA();

  void setgrammity(LMSuffix);
  void setposgrammity(LMSuffix);

  void setdiscparam(vector<double>);
  void setdiscparampos(vector<double>);

  void construct(char *corpusfile, char *posfile);
  void read(char *filename);
  void write(char *filename);

private:
  map<string,LMWordIndex> string2pos;            // map pos on LMWordIndex
  vector<string> word2string;   // map words (LMWordIndex) on (readable) strings
  vector<string> pos2string;    // map pos (LMWordIndex) on (readable) strings
   
  vector<LMWordIndex> corpus;
  vector<LMWordIndex> poscorpus;
  vector<LMSuffix> suffixarray;
  vector<LMSuffix> possuffixarray;
  vector<double> lambda;
  vector<double> lambdapos;
  string EOL, UNK;
  LMSuffix grammity, posgrammity;

  char ** Vocab;
  CLMState lmstate;
} ;

#endif
