#include <algorithm>
#include <string.h>
#include <stdlib.h>
#include <config.h>


#include <math.h>
#include <iostream>
#include <sstream>
#include <LMwordSA.hh>



//=============================================================================
// Class CLMposSASentenceState
//
CLMwordSASentenceState::CLMwordSASentenceState(CLMwordSA* LMod) {
  lm = LMod;
  cache = false;
  m_VocSize = lm->VocabularySize();
}

CLMwordSASentenceState::~CLMwordSASentenceState() {}



CLMState
CLMwordSASentenceState::BeginOfSentenceState() {
  suffixsort s1less(&(lm->corpus),0,lm->GetVocabularyIndex(lm->EOL));
  realhistorysize=1;

  return (CLMState) ((*lower_bound(lm->suffixarray.begin(),
     lm->suffixarray.end(), lm->GetVocabularyIndex(lm->EOL),s1less))|
         ((CLMState)realhistorysize << 32));
}



double
CLMwordSASentenceState::LogProb(CLMState lmstate, LMWordIndex word) {
  return log(Prob(lmstate,word));
}



double 
CLMwordSASentenceState::LogProbMerge(CLMState &lmstate_ll, CLMState lmstate_lr,
									 CLMState lmstate_rl, CLMState &lmstate_rr)
{
  return 0.0;
}



double
CLMwordSASentenceState::Prob(const CLMState lmstate, const LMWordIndex word)
{
  vector<LMWordIndex> lmstatevec;

  if(word == lm->GetVocabularyIndex(lm->UNK)) return 1.0/lm->VocabularySize();
  realhistorysize=lmstate>>32;
  LMSuffix realposition=lmstate & (((CLMState)1<<32)-1);

  lmstatevec.resize(realhistorysize+1);
  for(LMSuffix i=0; i<realhistorysize; i++)
    lmstatevec[i]=lm->corpus[realposition+i];
  lmstatevec[realhistorysize]=word;

  vector<LMWordIndex>::iterator start=lmstatevec.begin();
  while((unsigned)distance(start, lmstatevec.end()) > lm->grammity) start++;

  double prob=p(start,lmstatevec.end());

  return prob;
}



double
CLMwordSASentenceState::LogProbEnd(CLMState lmstate)
{
  return log(ProbEnd(lmstate));
}



double
CLMwordSASentenceState::ProbEnd(const CLMState lmstate)
{
  LMWordIndex word=lm->GetVocabularyIndex(lm->EOL);
  return Prob(lmstate,word);
}

CLMState
CLMwordSASentenceState::AppendWord(CLMState lmstate, LMWordIndex word)
{
  if(word == lm->GetVocabularyIndex(lm->UNK)) return 0;
  vector<LMWordIndex> lmstatevec;
  realhistorysize=lmstate>>32;
  LMSuffix realposition=lmstate & (((CLMState)1<<32)-1);
  lmstatevec.resize(realhistorysize+1);

  for(LMSuffix i=0; i<realhistorysize; i++)
    lmstatevec[i]=lm->corpus[realposition+i];
  lmstatevec[realhistorysize]=word;
  vector<LMWordIndex>::iterator start = lmstatevec.begin();
  realhistorysize++;


  while(lmstatevec.size()>lm->grammity) {
    lmstatevec.erase(lmstatevec.begin());
    realhistorysize--;
  }

  pair<suffixptr, suffixptr> ptrs = make_pair(lm->suffixarray.begin(),
                                              lm->suffixarray.end());
  LMSuffix tempdist=0;
  suffixsort s1less(&(lm->corpus),0,lm->GetVocabularyIndex(lm->EOL));
  for(;start!=lmstatevec.end();) {
    s1less.setdist(tempdist);
    ptrs = equal_range(ptrs.first, ptrs.second, *start, s1less);
    if(ptrs.first==ptrs.second) {
      ptrs = make_pair(lm->suffixarray.begin(),lm->suffixarray.end());
      lmstatevec.erase(lmstatevec.begin());
      start=lmstatevec.begin(); tempdist=0; realhistorysize--;
    } else {
      start++; tempdist++;
    }
  }

  CLMState ret=((CLMState)*ptrs.first | ((CLMState)realhistorysize << 32));
  return ret;
}



void
CLMwordSASentenceState::DeleteState(CLMState lmstate)
{}



void
CLMwordSASentenceState::AllProb(CLMState lmstate, double probs[])
{
  vector<LMWordIndex> lmstatevec;
  realhistorysize=lmstate>>32;
  LMSuffix realposition=lmstate&((CLMState)1<<32-1);
  lmstatevec.resize(realhistorysize);
  for(LMSuffix i=0;i<realhistorysize;i++)lmstatevec[i]=lm->corpus[realposition];
  vector<LMWordIndex>::iterator start = lmstatevec.begin(), ende ;

  for(LMWordIndex j=0; j <= m_VocSize; ++j) {

    lmstatevec.push_back(j);
    ende = lmstatevec.end();
    probs[j] = p(start, ende);
    lmstatevec.pop_back();
  }
}



void
CLMwordSASentenceState::AllLogProb(CLMState lmstate, double logp[])
{
  AllProb(lmstate,logp);

  for(LMWordIndex i=0; i <= m_VocSize; ++i)
    logp[i] = log(logp[i]);
}



void
CLMwordSASentenceState::NBestProb(CLMState lmstate, unsigned int N,
								  LMWordIndex wordindices[], double nbestp[])
{
  if (N==0) return;
  double probs[m_VocSize+1];
  AllProb(lmstate,probs);
  if (N > m_VocSize)
    N = m_VocSize;
  unsigned int n=0;
  signed short j;   // j must be signed!
  double tmp_p;
  nbestp[0] = probs[0];
  wordindices[0] = 0;
  for(unsigned int i=1; i<=m_VocSize; ++i) {    // get N best
      tmp_p = probs[i];
      j=n;
      while((j>=0) && (tmp_p > nbestp[j])) {
          if (j < (signed) N-1) {
              nbestp[j+1] = nbestp[j];
              wordindices[j+1] = wordindices[j];
          }
          --j;
      }
      if (j < (signed) N-1) {
          nbestp[j+1] = tmp_p;
          wordindices[j+1] = i;
      }
      if (n < N-1) ++n;
  }
}



void
CLMwordSASentenceState::NBestLogProb(CLMState lmstate, unsigned int N,
									 LMWordIndex wordindices[],
									 double nbestlogp[])
{
  NBestProb(lmstate,N,wordindices,nbestlogp);
  for(unsigned int i=0; i<N; ++i) nbestlogp[i] = log(nbestlogp[i]);
}



double
CLMwordSASentenceState::p(vector<LMWordIndex>::const_iterator start,
                          vector<LMWordIndex>::const_iterator ende)
{
  int tdist=0, tsize = distance(start, ende);
  double d=lm->lambda[tsize-1], W=lm->VocabularySize(), prob=0.0;
  LMSuffix Nh=0, Nhw=0, Nnh=0;
  if(tsize==0) { return (double)1.0/W; } // Zerogram
  suffixsort s1less(&(lm->corpus),0,lm->GetVocabularyIndex(lm->EOL));
  pair<suffixptr, suffixptr> ptrs = make_pair(lm->suffixarray.begin(),
                                              lm->suffixarray.end());
  vector<LMWordIndex>::const_iterator i=start;
  for(i=start; i!=(ende-1); ++i, ++tdist) {
    s1less.setdist(tdist);
    ptrs = equal_range(ptrs.first, ptrs.second, *i, s1less);
  }
  Nh = distance(ptrs.first, ptrs.second); // calculate N(h,.)
  suffixptr a2=ptrs.first, b2=ptrs.second;
  s1less.setdist(tdist);
  ptrs = equal_range(ptrs.first, ptrs.second, *i, s1less);
  Nhw = distance(ptrs.first, ptrs.second); // calculate N(h,w)
  pair<suffixptr,int> ck=make_pair(a2,tsize-1);
  map<pair<suffixptr,int>,LMSuffix>::iterator Nnhptr=lm->cache_Nnh.find(ck);
  if (a2 == b2) { // history not found?
    prob = p(++start,ende);
  } else if (tsize == 1) { // Unigram
    if (Nhw > 0)
      prob = ((double)Nhw) / (double)Nh;
    else
      prob = (double) 1 / W;
  } else { // 2..oo-gram
    if (Nnhptr==lm->cache_Nnh.end()) {
      LMWordIndex lastone=lm->corpus[(*a2)+tsize-1];
      Nnh = 1;
      for(suffixptr blb=a2; blb!=b2; ++blb) { // distinct events
        if(((*blb)+tsize)>lm->corpus.size()) continue;
        else if(lm->corpus[((*blb)+tsize-1)]!=lastone) {
          lastone=lm->corpus[((*blb)+tsize-1)];
          Nnh++;
      } }
      lm->cache_Nnh.insert(make_pair(ck,Nnh));
    } else Nnh = Nnhptr->second;
    if ((double)Nhw>d) prob = ((double)Nhw-d) / (double)Nh;
    prob += d * (double)Nnh * p(++start,ende) / (double)Nh;
  }
  return prob;
}



//=============================================================================
// Class CLMwordSA
//

// Constructor & destructor
CLMwordSA::CLMwordSA()
{ 
  Vocab = NULL;
  UNK = "<UNK>";
  EOL = "</s>";
  LMName="CLMwordSA";
}



CLMwordSA::CLMwordSA(char* PFile)
{
  Vocab = NULL;
  UNK = "<UNK>";
  EOL = "</s>";
  LMName="CLMwordSA";
  read(PFile);
}



CLMwordSA::~CLMwordSA()
{
  delete[] Vocab;
}



// Vocabulary related methods
LMWordIndex
CLMwordSA::VocabularySize() const
{
  return word2string.size();
}



const vector<string> &
CLMwordSA::Vocabulary() const
{
  return word2string;
}



LMWordIndex
CLMwordSA::GetVocabularyIndex(string word)
{
  if (GetVocabularyIndexMap.find(word)==GetVocabularyIndexMap.end())
    return GetVocabularyIndexMap[UNK];
  else
    return GetVocabularyIndexMap[word];
}



// Initianizing LM for search
CLMwordSASentenceState*
CLMwordSA::NewSentence() const
{
  return new CLMwordSASentenceState( (CLMwordSA*) this );
}



//setting parameters & constructing
void 
CLMwordSA::setgrammity(LMSuffix a)
{
  grammity=a;
}



void
CLMwordSA::setdiscparam(vector<double> b)
{
  lambda=b;
}



void
CLMwordSA::construct(char *filename)
{
  ifstream text(filename);
  string s="", word="";
  LMWordIndex actidx=2;
  GetVocabularyIndexMap[UNK]=0; word2string.push_back(UNK);
  GetVocabularyIndexMap[EOL]=1; word2string.push_back(EOL);
  while (getline(text,s)) {
    istringstream zeile(s.c_str());
    corpus.push_back(1);
    while (zeile>>word) {
      if (GetVocabularyIndexMap.find(word)==GetVocabularyIndexMap.end()) {
          GetVocabularyIndexMap[word]=actidx++;
          word2string.push_back(word);
      }
      corpus.push_back(GetVocabularyIndex(word));
    }
  }

  // create suffixarrays
  suffixarray.resize(corpus.size());
  for(LMSuffix i=0; i<corpus.size(); i++) suffixarray[i]=i;
  suffixsort sless(&corpus,0,1);
  stable_sort(suffixarray.begin(),suffixarray.end(),sless);
}

// Reading & writing suffixarray LM

/* format of parameterfile:

   CLMwordSA
   Vocabularyfile    <filename>
   Corpusfile        <filename>
   Suffixarrayfile   <filename>
   Discparamfile     <filename>
   Grammity          <number>

   */

void
CLMwordSA::write(char *PFile)
{
  LMSuffix i;

  string fname(PFile);


  // write parameter file
  ofstream param((fname + ".param").c_str());
  if (!param.good())
    { cerr << "\nCLMwordSA::write:  Error writing file '" << PFile
	   << ".param'" << endl;
      exit(-1);
    }
  param << LMName << endl;
  param << "Vocabularyfile     " << (fname + ".voc") << endl;
  param << "Corpusfile         " << (fname + ".corp") << endl;
  param << "Suffixarrayfile    " << (fname + ".sa") << endl;
  param << "DiscParamfile      " << (fname + ".disc") << endl;
  param << "Grammity           " << grammity << endl;
  param.close();


  // write Vocabulary
  ofstream voc((fname + ".voc").c_str());
  if (!voc.good())
    { cerr << "\nCLMwordSA::write:  Error writing file '" << PFile
	   << ".voc'" << endl;
      exit(-1);
    }
  // write Voc without UNK and </s>
  for (i=2; i<VocabularySize();i++)
    voc << word2string[i] << endl;
  voc.close();


  // write Corpus
  ofstream corpusfile((fname + ".corp").c_str());
  if (!corpusfile.good())
    { cerr << "\nCLMwordSA::write:  Error writing file '" << PFile
	   << ".corp'" << endl;
      exit(-1);
    }

  corpusfile << corpus.size() << endl;
  for (i=0; i<corpus.size(); i++) write32Bit(corpusfile,corpus[i]);
  corpusfile.close();

  //write Suffixarray
  ofstream suffixfile((fname + ".sa").c_str());
  if (!suffixfile.good())
    { cerr << "\nCLMwordSA::write:  Error writing file '" << PFile
	   << ".sa'" << endl;
      exit(-1);
    }
  for (i=0; i<corpus.size(); i++) write32Bit(suffixfile,suffixarray[i]);
  suffixfile.close();

  // write discounting parameter
  ofstream discfile((fname + ".disc").c_str());
  if (!discfile.good())
    { cerr << "\nCLMwordSA::write:  Error writing file '" << PFile
	   << ".disc'" << endl;
      exit(-1);
    }
  for (vector<double>::iterator i=lambda.begin(); i!=lambda.end(); ++i)
    {discfile << *i << " ";}
  discfile << endl;
  discfile.close();

}



void
CLMwordSA::read(char *PFile)
{
  corpus.clear();
  //word2string.clear();
  suffixarray.clear();
  lambda.clear();

  ifstream param(PFile);
  if (!param.good())
    { cerr << "\nCLMwordSA::read:  Parameter file '" << PFile
	   << "' could not be read." << endl;
      exit(-1);
    }

  string row, pname, pvalue;

  char* Discfile ;
  char* Vocfile ;
  char* Suffixfile;
  char* Corpusfile;


  // read parameter file
  getline(param,row);
  if (row != LMName)
    { cerr << "Wrong LM Type!!"; return; }


  // cout << "reading parameter file from " << PFile << endl;

  while (getline(param,row))
    {
      istringstream zeile(row.c_str());
      pname = ""; pvalue = "";
      zeile >> pname ;
      zeile >> pvalue;

      if ((pname != "") && (pvalue != ""))
	if (pname == "Vocabularyfile")
	  {
	    Vocfile = new char[strlen(pvalue.c_str())+1];
	    strcpy(Vocfile, pvalue.c_str());
	  }
	else if (pname == "Corpusfile")
	  {
	    Corpusfile = new char[strlen(pvalue.c_str())+1];
	    strcpy(Corpusfile, pvalue.c_str());
	  }
	else if (pname == "Suffixarrayfile")
	  {
	    Suffixfile = new char[strlen(pvalue.c_str())+1];
	    strcpy(Suffixfile, pvalue.c_str());
	  }
	else if (pname == "DiscParamfile")
	  {
	    Discfile = new char[strlen(pvalue.c_str())+1];
	    strcpy(Discfile, pvalue.c_str());
	  }
	else if (pname == "Grammity")
	  grammity = atoi(pvalue.c_str());
    }


  LoadVoc(Vocfile);

  string s="", word="";
  int corpussize, i;

  // read Corpus
  ifstream Corpusstr(Corpusfile);
  if (!Corpusstr) cerr << "Error reading file " << Corpusfile << endl;

  // get corpussize
  s="";
  getline(Corpusstr,s);
  corpussize = atoi(s.c_str());

  corpus.resize(corpussize);
  for(i=0; i<corpussize; i++) corpus[i]=read32Bit(Corpusstr);

  // read Suffixarray
  ifstream Suffixstr(Suffixfile);
  if (!Suffixstr) cerr << "Error reading file " << Suffixfile << endl;
  suffixarray.resize(corpussize);
  for(i=0; i<corpussize; i++) suffixarray[i]=read32Bit(Suffixstr);

  //--

  ifstream discstr(Discfile);
  if (!discstr) cerr << "Error reading file " << Discfile << endl;
  getline(discstr,s);
  { istringstream zeile(s.c_str());
    while (zeile>>word) {lambda.push_back(atof(word.c_str()));}
  }

  cout << "Suffix Array LM loaded" << endl;
  cout << "Size of corpus:       " << corpus.size() << "  " << endl;
  cout << "Size of suffix array: " << suffixarray.size() << endl;
  cout << "Size of vocabulary:   " << word2string.size() << endl;
  cout << "Discounting factors:  ";
  for ( int i = 0; i < 10; i++ ) {
    cout << lambda[ i ] << " ";
  }
  cout << endl;

  //cout << "disc:         " << lambda.size() << endl;

}



void
CLMwordSA::LoadVoc(char *File)
{
  word2string.clear();

  ifstream Vocstr(File);
  if (!Vocstr.good())
    { cerr << "\nCLMwordSA::LoadVoc:  Vocabulary file '" << File
	   << "' could not be read." << endl;
      exit(-1);
    }

  string s="", word="";
  //  LMWordIndex actidx=2;

  //  GetVocabularyIndexMap[UNK]=0;
  word2string.push_back(UNK);
  //  GetVocabularyIndexMap[EOL]=1;
  word2string.push_back(EOL);

  // read Vocabulary
  while (!Vocstr.eof())
    {
      s = ""; getline(Vocstr, s);
      if (( s != "") && (s[0] != '#'))
	{
	  istringstream zeile(s.c_str());
	  zeile >> word;
	  //	  GetVocabularyIndexMap[word]=actidx++;
	  word2string.push_back(word);
	}
    }

  int VocabSize = word2string.size();
  Vocab = new char* [VocabSize];
  unsigned int idx=0;
  for(vector<string>::const_iterator i= word2string.begin();
      i != word2string.end(); ++i, ++idx) {
    Vocab[idx]=(char *)i->c_str();
    GetVocabularyIndexMap.insert(make_pair(*i, idx));
  }
}



int
CLMwordSA::Compare(char ** Vocab, char *Paramfile)
{
  ifstream param(Paramfile);
  if (!param.good())
    { cerr << "\nCLMwordSA::Compare:  Parameter file '" << Paramfile
	   << "' could not be read." << endl;
      exit(-1);
    }

  string row, pname, pvalue;

  char* Vocfile ;

  // read parameter file
  getline(param,row);
  if (row != LMName)
    { cerr << "Wrong LM Type!!"; return 0; }


  while (getline(param,row))
    {
      istringstream zeile(row.c_str());
      pname = ""; pvalue = "";
      zeile >> pname ;
      zeile >> pvalue;

      if ((pname != "") && (pvalue != ""))
	if (pname == "Vocabularyfile")
	  {
	    Vocfile = new char[strlen(pvalue.c_str())+1];
	    strcpy(Vocfile, pvalue.c_str());
	  }
    }

  LoadVoc(Vocfile);


  // compare Vocabularies

  int j=0;
  for(vector<string>::iterator it = word2string.begin();
      it != word2string.end(); ++it, ++j)
    if (*it != string(Vocab[j]))
      return 0;

  return 1;
}
