#include <config.h>

#ifndef __Suffix__
#define __Suffix__


#include <LMInterface.hh>
#include <suffixsort.hh>
#include <writeSmall.hh>

class CLMwordSA;
class CLMwordSASentenceState;


typedef std::vector<LMSuffix>::iterator suffixptr;


class CLMwordSASentenceState : public CLMSentenceState
{
  friend class CLMwordSA;

public:
  ~CLMwordSASentenceState();
  CLMState BeginOfSentenceState();
  double Prob(const CLMState lmstate, const LMWordIndex word) ;
  inline double LogProb(CLMState lmstate, const LMWordIndex word) ;
  double LogProbMerge(CLMState &lmstate_ll, CLMState lmstate_lr,
							  CLMState lmstate_rl, CLMState &lmstate_rr);
  double ProbEnd(const CLMState lmstate) ;
  inline double LogProbEnd(CLMState lmstate) ;
  CLMState AppendWord(CLMState lmstate, const LMWordIndex word) ;
  void DeleteState(CLMState lmstate);
  virtual void AllProb(CLMState lmstate, double p[]) ;
  virtual void AllLogProb(CLMState lmstate, double logp[]) ;
  virtual void NBestProb(CLMState lmstate, unsigned int N, 
			 LMWordIndex wordindices[], double nbestp[]) ;
  virtual void NBestLogProb(CLMState lmstate, unsigned int N, 
			    LMWordIndex wordindices[], 
			    double nbestlogp[]) ;

protected:
  CLMwordSASentenceState(CLMwordSA* LMod);
  CLMwordSASentenceState(CLMwordSA* LMod, char* cachefile);
  virtual double p(std::vector<LMWordIndex>::const_iterator,
           std::vector<LMWordIndex>::const_iterator);

  CLMwordSA *lm ;
  LMWordIndex m_VocSize;

private:
  bool cache;
  std::vector<LMSuffix> getdistinct;
  map<pair<suffixptr, LMWordIndex>, LMSuffix> getNn0h;

  LMSuffix realhistorysize, realposition;
};



class CLMwordSA : public CLM
{
  friend class CLMwordSASentenceState;

public:
  CLMwordSA();
  CLMwordSA(char* Paramfile); // read data from parameterfile

  ~CLMwordSA();
  CLMwordSASentenceState * NewSentence() const ;

  LMWordIndex VocabularySize() const;
  const std::vector<std::string> & Vocabulary() const;
  LMWordIndex GetVocabularyIndex(std::string);

  int Compare(char ** Vocab, char *Paramfile);

  void setgrammity(LMSuffix a);
  void construct(char *filename);
 
  void write(char *filename);
  void setdiscparam(std::vector<double> b);
private: 
  map<pair<suffixptr,int>,LMSuffix> cache_Nnh; // Cache for distinct events

protected:
  void read(char *filename);
  void LoadVoc(char* filename); // load Vocabulary

  LMSuffix grammity;
  std::vector<std::string> word2string; // map LMWordIndex on (readable) strings
  std::vector<LMWordIndex> corpus;
  std::vector<LMSuffix> suffixarray;
  std::vector<double> lambda;
  std::string EOL, UNK;
  char ** Vocab;
  std::string LMName;
} ;

#endif
