#include <string.h>
#include <stdlib.h>
#include <config.h>
#include <fortify.h>
#include <NistLM.hh>

using namespace std;
// the reference sentence now needs to be built into a trie 
// that will allow for ease of matching
NistLMSentenceState::NistLMSentenceState(NGram* reference){
  m_begin = new PathState(reference);
  m_track = new Tracker();
  (*m_track)[m_begin->toString()] = m_begin;
  m_Lref = reference->size();
}

NistLMSentenceState::~NistLMSentenceState(){
  //CLMSentenceState::~CLMSentenceState();
  for (Tracker::iterator iter = m_track->begin(); iter!=m_track->end(); ++iter){
    PathState* ps = (*iter).second;
    delete ps;
  }
  delete m_track;
}

CLMState NistLMSentenceState::BeginOfSentenceState(){
  return (CLMState)m_begin;
}

double NistLMSentenceState::LogProb(CLMState lmstate, const LMWordIndex word){

  cout<<"LogProb word:"<<word<<endl;

  PathState* path = (PathState*)lmstate;
  double oldScore = path->calculateScore(m_Lref);

  //===========================================
  PathState* after = path->scoreExtension(word);
  string key = after->toString();
  PathState* ret = NULL;
  // if there is another item with the exact same history/path drop this one
  if (m_track->count(key)==0){
    (*m_track)[key] = after;
    ret = after;
  }else{
    delete after;
    ret = (*m_track)[key];
  }
  //==========================================
  double newScore = ret->calculateScore(m_Lref);
  double diff = newScore - oldScore;
  cout<<"New:"<<newScore<<" Old:"<<oldScore<<" Diff:"<<diff<<endl;

  m_bornState = (CLMState)ret;
  return -diff;
}

double NistLMSentenceState::LogProbEnd(CLMState lmstate){
  return 0.0;
  
}

CLMState NistLMSentenceState::AppendWord(CLMState lmstate, const LMWordIndex word){
  cout<<"Appending word:"<<word<<endl;
  return m_bornState;
}

void NistLMSentenceState::DeleteState(CLMState lmstate){
  PathState* path = (PathState*)lmstate;
  string key = path->toString();
  m_track->erase(key);
  delete path;
}



NistLM::NistLM(char* paramFile){
  
  // read the paramfile to extract gain file, refererenceFile
  ifstream params(paramFile);
  // read the name of the gainmap file, reference sentences
  string vocabFile = "";
  string gainFile = "";
  string referenceFile = "";
  
  getline(params, vocabFile);
  getline(params, gainFile);
  getline(params, referenceFile);
  params.close();
  // Initialize the Nist gain map
  cout<<"Loading vocab:"<<vocabFile<<endl;
  loadVocabulary(vocabFile);
  cout<<"Loading gains"<<gainFile<<endl;
  NistScore::loadGains(GetVocabularyIndexMap, gainFile.c_str());
  cout<<"Building references"<<endl;
  // load the references 
  m_refList = new ReferenceList();
  ifstream reference(referenceFile.c_str());
  if (!reference){
    cerr<<"Unable to load reference list:"<<referenceFile<<endl;
    exit(-1);
  }
  string s = "";
  while (getline(reference, s)){
    //cout<<"R:"<<s<<endl;
    m_refList->insert(m_refList->end(), s);
  }
  reference.close();
  
  m_currentSentence = 0;

  for (map<string,LMWordIndex>::iterator iter= GetVocabularyIndexMap.begin(); iter!=GetVocabularyIndexMap.end(); ++iter){
    m_vocabVector.push_back((*iter).first);
  }
}

NistLM::~NistLM(){
  
  m_refList->clear();
  delete m_refList;
  // remove the static gains structure
  NistScore::deleteGains();

}

// vocab starts at 1!
void NistLM::loadVocabulary(string vocabFile){

  ifstream str(vocabFile.c_str());
  char buf[1000];
  
  int count = 1;
  
  if (!str){
    cerr<<"Unable to load vocabulary:"<<vocabFile<<endl;
    exit(-1);
  }

  while ( !str.eof()) {
    str.getline ( buf, sizeof ( buf));
    if (strlen(buf)>0){
      CLM::GetVocabularyIndexMap[string(buf)] = count;
      count++;
      //cout<<count<<endl;
    } 
  }
  str.close();
}

NistLMSentenceState* NistLM::NewSentence() const{


  string currentSentenceString = (*m_refList)[m_currentSentence];
  //cout<<"REFString:"<<currentSentenceString<<" -- "<<m_currentSentence<<endl;
  NGram* reference = buildReference(currentSentenceString);
  //cout<<"REF:"<<TrieNode::printNGram(reference)<<endl;
  NistLMSentenceState* ss = new NistLMSentenceState(reference);
  // delete the ngram
  reference->clear();
  delete reference;
  return ss;
}

void NistLM::DeleteSentence(CLMSentenceState* st) const{
  NistLMSentenceState* ns = (NistLMSentenceState*)st;
  delete ns;
}

LMWordIndex NistLM::VocabularySize() const{
  return GetVocabularyIndexMap.size();
}

NGram* NistLM::buildReference(string s) const{
  NGram* ref = new NGram();
  char* toTok = (char*)s.c_str();
  char* c = strtok(toTok, " ");
  ref->insert(ref->end(), GetVocabularyIndex(string(c)));
  while ((c=strtok(NULL, " "))!=NULL){
    ref->insert(ref->end(), GetVocabularyIndex(string(c)));
  }
  return ref;
}
NGram* NistLM::buildReference(string s) {

  NGram* ref = new NGram();
  char* toTok = (char*)s.c_str();
  char* c = strtok(toTok, " ");
  ref->insert(ref->end(), GetVocabularyIndex(string(c)));
  while ((c=strtok(NULL, " "))!=NULL){
    ref->insert(ref->end(), GetVocabularyIndex(string(c)));
  }
  return ref;
}


