#include <config.h>
#include <fortify.h>
#ifndef __NIST__
#define __NIST__

#include <LMInterface.hh>
#include <TrieNode.hh>

using namespace std;

class NistLM;
class NistLMSentenceState;

typedef map<string, PathState*> Tracker;

class NistLMSentenceState : public CLMSentenceState{

public:
  NistLMSentenceState();
  NistLMSentenceState(NGram* reference);
  ~NistLMSentenceState();
  virtual CLMState BeginOfSentenceState();
  // returns the negative difference of the param state and
  // the state that results after the merge
  virtual double LogProb(const CLMState lmstate, const LMWordIndex word);
  virtual double LogProbEnd(CLMState lmstate);
  virtual CLMState AppendWord(CLMState lmstate, const LMWordIndex word);
  virtual void DeleteState(CLMState lmstate);
  // returns the tracking list size at this node
  inline int getTrackSize(){
    return m_track->size();
  }
  
private:
  // when log prob is called, the new state is already created
  // but just returned in Append Word
  CLMState m_bornState;
  Tracker* m_track;
  PathState* m_begin;
  // length of the reference
  int m_Lref;
};

typedef vector<string> ReferenceList;

// reads the param file to
// load the reference translation
// track which sentence we are using
// build gain statistics for each sentence

class NistLM : public CLM
{
  friend class NistLMSentenceState;
public:

  NistLM(char* paramFile);
  virtual ~NistLM();
  virtual NistLMSentenceState* NewSentence() const;
  virtual void DeleteSentence(CLMSentenceState* st) const;
  virtual LMWordIndex VocabularySize() const;
  // pure virtual functions that must be implemented, but are currently
  // not used anywhere in the system
  virtual const vector<string> & Vocabulary() const{
    return m_vocabVector;
  }
  virtual int Compare(char ** Vocab, char *Paramfile){
    return 0;
  }
  NGram* buildReference(string s);
  NGram* buildReference(string s)const;

private:
  
  // vector of all strings in this language model's vocabulary
  vector<string> m_vocabVector;
  void loadVocabulary(string);
  //builds a reference Ngram from a string
  int m_currentSentence;
  ReferenceList* m_refList;

};

#endif
