#include <stdlib.h>
#include <config.h>

#include <SriLM.hh>
#include <utils.hh>

using namespace std;


SriLMSentenceState::SriLMSentenceState( const SriLM* theLM )
{
  lm = (SriLM*)theLM;
  wids = new VocabIndex[lm->GetOrder()];
}



SriLMSentenceState::~SriLMSentenceState()
{

  // dump all n-grams seen for this sentence
  bool DumpNGrams = false;

  if ( DumpNGrams ) {
    for ( int j = 0; j < m_AllNGrams.size(); j++ ) {
      NGramType NGram = m_AllNGrams[ j ];
      cout << "NGRAM: ";
      for ( int i = 0; i < NGram.size(); i++ ) {
        cout << lm->vocab->getWord( NGram[ i ] ) << " ";
      }
      cout << endl;
    }
  }

  cout << "In this sentence we called the LM " << m_NumberOfCalls << " times "
       << "and created " << m_AllNGrams.size() << " LM States" << endl << endl;

  lm = NULL;  
  delete [] wids;
}



inline CLMState 
SriLMSentenceState::BeginOfSentenceState2()
{
  unsigned LMorder = lm->GetOrder();
  CLMState ret;

  SriLM *MyLM = lm;
  Vocab *pMyVoc = MyLM->vocab;
  cout << pMyVoc->ssIndex;

  if (LMorder >= (unsigned)4) {

    // lmstate, 64bit: |1bit|- 21bit: mostRecentWord -|- 21bit nextRecentWord -|- 21bit thirdRecentWord -|
    // here all three filled with <s>, like the speech people do it.
    ret = ((CLMState)( lm->vocab->ssIndex ) | ((CLMState)( lm->vocab->ssIndex ) << 21)
           | ((CLMState)( lm->vocab->ssIndex ) << 42) );

  } else if ( LMorder == (unsigned)3 ) {

    ret = ((CLMState)0 | ((CLMState)( lm->vocab->ssIndex ) << 32) );

  }

  return ret;
}


inline CLMState 
SriLMSentenceState::BeginOfSentenceState()
{
  CLMState LMState;
  long IndexOfNGram;

  unsigned LMorder = lm->GetOrder();
  cout << "SriLMSentenceState::BeginOfSentenceState - LM Order = " << LMorder << endl;
  cout << "Unk is word " << lm->vocab->unkIsWord << endl;


  m_NumberOfCalls = 0;

  // 3-gram LM (what about 2-gram?)
  if ( LMorder == (unsigned)3 ) {

    LMState = ((CLMState)0 | ((CLMState)( lm->vocab->ssIndex ) << 32) );
    
    //cout << "DEBUG - 3gram" << endl;
    
    return LMState;

  } else if (LMorder == (unsigned)44) { // switch this 4-gram implementation off for now, to allow large vocabularies

    // lmstate, 64bit: |1bit|- 21bit: mostRecentWord -|- 21bit nextRecentWord -|- 21bit thirdRecentWord -|
    // here all three filled with <s>, like the speech people do it.
    LMState = ((CLMState)( lm->vocab->ssIndex ) | ((CLMState)( lm->vocab->ssIndex ) << 21)
               | ((CLMState)( lm->vocab->ssIndex ) << 42) );
               
     //cout << "DEBUG - 4gram" << endl;          

    return LMState;

  } else {

    // and this for 5-grams and higher
    
    //cout << "DEBUG - 5gram and higher" << endl; 

    // reset NGram storage;
    this->m_AllNGrams.clear();
    this->m_NGramToLMState.clear();

    // initialize NGram storage: create two ngrams
    // add <unk> n times, ie create a n-gram = [ <unk> .... <unk> ]
    // add <s> n times, ie create a n-gram = [ <s> .... <s> <s> ]
    NGramType AllUnksNGram;
    NGramType SentenceStartNGram;

    for ( int n = 1; n <= LMorder - 1; n++ ) {
      AllUnksNGram.push_back( lm->vocab->unkIndex );
      SentenceStartNGram.push_back( lm->vocab->ssIndex );
    }

    // store the n-grams in the list of all n-grams
    // and put it also into the NGram-to-LMState map
    this->m_AllNGrams.push_back( AllUnksNGram );
    IndexOfNGram = m_AllNGrams.size() - 1;
    this->m_NGramToLMState[ AllUnksNGram ] = IndexOfNGram;

    this->m_AllNGrams.push_back( SentenceStartNGram );
    IndexOfNGram = m_AllNGrams.size() - 1;
    this->m_NGramToLMState[ SentenceStartNGram ] = IndexOfNGram;

    // the index of the SentenceStartNGram is what we use as LMState
    LMState = IndexOfNGram;
    return LMState;

  }

  // we should not reach this point
  assert( 0 );
  return (CLMState)0;

}


double
SriLMSentenceState::LogProb( CLMState lmstate, const LMWordIndex word )
{
  //cout << "SriLMSentenceState::LogProb( CLMState, LMWordIndex )" << endl;

  //cout << "Word: "<< word << " -- " << lm->vocab->getWord(word)<<endl; 
  
  // we don't use the proper prob for unknow word, but a fixed prob
  if ( word == lm->vocab->unkIndex ) {
     //cout << "Unknown word " << word << endl;
    return log10( 1.0 / lm->VocabularySize() );
  }

  unsigned LMorder = lm->GetOrder();
  if ( LMorder >= (unsigned)4 ) {
      // lmstate, 64bit: |1bit|- 21bit: mostRecentWord -|- 21bit nextRecentWord -|- 21bit thirdRecentWord -|
      VocabIndex mostRecentWord = lmstate >> 42;
      VocabIndex nextRecentWord = (lmstate >> 21) & ( ( (CLMState)1 << 21 ) - 1 );
      VocabIndex thirdRecentWord = lmstate & ( ( (CLMState)1 << 21 ) - 1 );

      wids[ 0 ] = mostRecentWord;
      wids[ 1 ] = nextRecentWord;
      wids[ 2 ] = thirdRecentWord;
      wids[ 3 ] = Vocab_None;
          /*
          cout << "LogP:" << logP << " for " << lm->vocab->getWord(word)
          << " | " << mostRecentWord << "," << nextRecentWord 
          << "," << thirdRecentWord << endl;
          cout << " | " << lm->vocab->getWord(mostRecentWord) << " " <<lm->vocab->getWord(nextRecentWord)
          << " " << lm->vocab->getWord(thirdRecentWord)<<endl;  
          */
  } else {
      long mostRecentWord = lmstate >> 32;
      long nextRecentWord = lmstate & ( ( (CLMState)1 << 32 ) - 1 );

      wids[ 0 ] = mostRecentWord;
      wids[ 1 ] = nextRecentWord;
      wids[ 2 ] = Vocab_None;
  }

  /*    
  cout << "Words given to SriLM: ";
  for ( int i = 0; i <= 2; i++ ) {
    cout << wids[ i ] << " ";
  }
  cout << endl;
  */

  double logP = lm->ngramLM->wordProb( word, wids );
  return logP;
}



double
SriLMSentenceState::LogProb( CLMState LMState, const LMWordIndex NextWord, CLMState & NewLMState )
{
  //cout << "SriLMSentenceState::LogProb( CLMState, LMWordIndex, CLMState )" << endl;
  double LogP = 0.0;

  // keep track how many LM probs we requested 
  m_NumberOfCalls++;


  // if NextWord is an unknow word
  // the NewState is NULL_STATE (use 0 for now)
  // return small default probabilty
 
  if ( !(lm->vocab->unkIsWord) && (NextWord == lm->vocab->unkIndex) ) {
    // we had created a ngram with sequence of <UNK> and stored a position 0
    // i.e. the LM state is 0
    NewLMState = 0;

    // we don't use the proper prob for unknow word, but a fixed prob
    LogP = log10( 1.0 / lm->VocabularySize() );

    CLMState mostRecentWord= (CLMState)NextWord;
    CLMState nextRecentWord = LMState >> 32;
      
    // 2nd item, then 1st
    NewLMState = ( nextRecentWord | (mostRecentWord << 32) );

    //cout << "Unknown word " << NextWord << " LogProp = " << LogP << endl;
    //cout << "Fall back to LMState " << NewLMState << endl << endl;

    return LogP;
  }

  unsigned LMorder = lm->GetOrder();

  if ( LMorder == (unsigned)3 ) {
    // 3-gram LM (what about 2-gram?)
    long MostRecentWord = LMState >> 32;
    long NextRecentWord = LMState & ( ( (CLMState)1 << 32 ) - 1 );

    wids[ 0 ] = MostRecentWord;
    wids[ 1 ] = NextRecentWord;
    wids[ 2 ] = Vocab_None;

    LogP = lm->ngramLM->wordProb( NextWord, wids );

    // for the next LM State we shift the last word and add the current word
    MostRecentWord = NextWord;
    NextRecentWord = LMState >> 32;
      
    // create new LMState from words: 2nd item, then 1st
    NewLMState = ( (CLMState)NextRecentWord | ((CLMState)MostRecentWord << 32) );

    return LogP;

  } else {

    // and this for 4-grams, 5-grams and higher

    // map LMState to NGram
    // add NextWord to NGram to make a new NGram
    // store this new NGram and get (potentially) a new LMState
    // get probability for new NGram

    // first, map LMState to NGram
    NGramType OldNGram = this->m_AllNGrams[ LMState ];
    /*
      cout << "OldLMState: " << LMState << "  ";
      cout << "OldNGram: ";
      for ( int i = 0; i < OldNGram.size(); i++ ) {
      cout << OldNGram[ i ] << " ";
      }
      cout << endl;
    */
    

    // now, we need to copy the word indices over to the array used by the SRI LM
    NGramType::const_iterator IterNGram = OldNGram.begin();
    int n = 0;
    int m = 0;
    for ( int n = OldNGram.size() - 1; n >= 0; n-- ) {
      wids[ m ] = OldNGram[ n ];
      m++;
    }
    wids[ m ] = Vocab_None;

    /*
      cout << "Words given to SriLM: ";
      for ( int i = 0; i < m; i++ ) {
      cout << wids[ i ] << " ";
      }
      cout << " and " << NextWord << " ";
      cout << endl;
    */

    // and get the LM log prob
    LogP = lm->ngramLM->wordProb( NextWord, wids );

    //virtual void *contextID(const VocabIndex *context, unsigned &length)
    unsigned NgramLength;
    long ContextID = (long)lm->ngramLM->contextID( wids, NgramLength );
    //cout << ContextID << " " << NgramLength << endl;

    // sv: NgramLength could be used to change the LMState, which would
    // allow more hyps to be recombined


//Nguyen, Aug 15, 2007: turn off SRI LM warnings 
//
    if ( (LogP < -100) || (LogP > 100) ) {
      cout << "WARNING: LogProb = " << LogP << endl;
      cout << "Words given to SriLM: ";
      for ( int i = 0; i < m; i++ ) {
        cout << wids[ i ] << " ";
      }
      cout << " and " << NextWord << " ";
      cout << endl;
    }

    // create new ngram, by appending word to old ngram and truncating to max length
    NGramType NewNGram = OldNGram;
    /*  
        if ( NextWord == lm->vocab->unkIndex ) {
        // we had created a ngram with sequence of <UNK> and stored a position 0
        // i.e. the LM state is 0
        NewLMState = 0;
        unsigned LMorder = lm->GetOrder();
        for ( int n = 1; n <= LMorder - 1; n++ ) {
        NewNGram[ n ] = 0;
        }
        
        } else {
    */

    // and restrict the length
    unsigned LMorder = lm->GetOrder();
    if ( NewNGram.size() >= LMorder - 1 ) {
      NewNGram.erase( NewNGram.begin() );
    }

    // store the new word
    NewNGram.push_back( NextWord );

    // turn this new ngram into an LMState: check, if it has been seen before,
    // if not, add it as new LMState
    map< NGramType, int, LessThenForNGrams >::iterator NGramToLMStateIterator;
    long IndexOfNewNGram;
    
    NGramToLMStateIterator = this->m_NGramToLMState.find( NewNGram );
    if( NGramToLMStateIterator == this->m_NGramToLMState.end() ) { //we haven't seen this lmstate+word yet
      m_AllNGrams.push_back( NewNGram );
      NewLMState = m_AllNGrams.size() - 1;
      m_NGramToLMState[ NewNGram ] = NewLMState;
      //cout << "new LMState - insert as " << NewLMState << endl;
    } else {
      NewLMState = NGramToLMStateIterator->second;
      //cout << "old LMState " << NewLMState << endl;
    }
    //}
    /*
      cout << "NewLMState: " << NewLMState << "  ";
      cout << "NewNGram: ";
      for ( int i = 0; i < NewNGram.size(); i++ ) {
      cout << NewNGram[ i ] << " ";
      }
      cout << endl;
    */


    // finally, return the log prob
    return LogP;
  }
  
  // we should not reach this point
  assert( 0 );
}



double
SriLMSentenceState::LogProb( vector< string > Words, CLMState& LMStateLeft, CLMState& LMStateRight )
{
  //cout << "SriLMSentenceState::LogProb" << endl;

  // we start with BeginOfSentence LM states
  CLMState LMState = BeginOfSentenceState();

  double logP = 0.0;
  double SumLogP = 0.0;

  LMWordIndex w;

  // loop over all given words and accumulate the log prob
  for ( int i = 0; i < Words.size(); i++ ) {
    //cout << Words[ i ] << " ";
    w = lm->GetVocabularyIndex( Words[ i ] );
    logP = LogProb( LMState, w );
    SumLogP += logP;

    //	cout << LMState << " " << Words[ i ] << " " << w << " " << logP;

    LMState = AppendWord( LMState, w );
    //	cout << " " << LMState << endl;
    
    // we want to generate a left LMState, encoding the
    // n left-most words in the word sequence
    if ( i + 1 < NN ) {
      LMStateLeft = LMState;
      //	  cout << "new left state = " << LMStateLeft << endl;
    }

  }

  LMStateRight = LMState;
  
  //cout << "Final state: " << LMStateLeft << " " << LMStateRight << endl;

  return SumLogP;
}


/* Is not modified yet for 4-gram LM. */

/**
   Get the logarithmic probability of a extending lmstate by words
   which led to the creation of lmstate2
   update the outer lmstates, which is necessary for word sequences
   shorter than the history length */
double
SriLMSentenceState::LogProbMerge( CLMState &lmstate_left_left, CLMState lmstate_left_right,
								  CLMState lmstate_right_left, CLMState &lmstate_right_right )
{
  //  cout << "MERGE \n";
  //cout << "ll " << lmstate_left_left << " " << " lr " << lmstate_left 
  //<< " rl " << lmstate_right << " rr " << lmstate_right_right << endl;

  long word_ll_1, word_lr_1, word_rl_1, word_rr_1;
  long word_ll_2, word_lr_2, word_rl_2, word_rr_2;

  word_rl_1 = lmstate_right_left & (((CLMState)1 << 32) - 1);
  word_rl_2 = lmstate_right_left >> 32;

  /*
  long ord_ll_1 = lmstate_left_left &(((CLMState)1 << 32) -1);
  long ord_ll_2 = lmstate_left_left >> 32;
  cout << "ll 1 " << lm->vocab->getWord( Word_ll_1 ) << " ll 2 " << lm->vocab->getWord( Word_ll_2 ) << endl;

  long ord_lr_1 = lmstate_left &(((CLMState)1 << 32) -1);
  long ord_lr_2 = lmstate_left >> 32;
  cout << "lr 1 " << lm->vocab->getWord( Word_lr_1 ) << " lr 2 " << lm->vocab->getWord( Word_lr_2 ) << endl;

  long ord_rl_1 = lmstate_right &(((CLMState)1 << 32) -1);
  long ord_rl_2 = lmstate_right >> 32;
  cout << "rl 1 " << lm->vocab->getWord( Word_rl_1 ) << " rl 2 " << lm->vocab->getWord( Word_rl_2 ) << endl;

  long ord_rr_1 = lmstate_right_right &(((CLMState)1 << 32) -1);
  long ord_rr_2 = lmstate_right_right >> 32;
  cout << "rr 1 " << lm->vocab->getWord( Word_rr_1 ) << " rr 2 " << lm->vocab->getWord( Word_rr_2 ) << endl;
  */


  // we recover the history, i.e. the words from the lmstate_right
  // then expand the lmstate_left over these words
  // but first need to subtract the logprobs calculated before

  double LogProbOld = 0.0;
  double LogProbNew = 0.0;

  // first we recalculate the logprobs which were calculated without history
  CLMState lmstate_old = BeginOfSentenceState();

  if ( word_rl_1 == lm->vocab->ssIndex ) {
	// <s> is used to fill the 2 word history, if only one word has been seen
	// to create the lmstate_right 
  } else {
	LogProbOld = LogProb( lmstate_old, word_rl_1 );
	lmstate_old = AppendWord( lmstate_old, word_rl_1 );
  }

  LogProbOld += LogProb( lmstate_old, word_rl_2 );
  lmstate_old = AppendWord( lmstate_old, word_rl_1 );
  //cout << "LogProbOld = " << LogProbOld << endl;


  // now we calculate the logprobs with the correct history
  if ( word_rl_1 == lm->vocab->ssIndex ) {
	// again, skip of <s> used to fill the 2 word history
  } else {
	LogProbNew = LogProb( lmstate_left_right, word_rl_1 );
	lmstate_left_right = AppendWord( lmstate_left_right, word_rl_1 );
  }
  
  LogProbNew += LogProb( lmstate_left_right, word_rl_2 );
  lmstate_left_right = AppendWord( lmstate_left_right, word_rl_2 );
  //cout << "LogProbNew = " << LogProbNew << endl;


  // and we adjust the LM states
  if ((lmstate_right_right & 0x00000000FFFFFFFF)==lm->vocab->ssIndex) {
    //cout << "Right_Right_State angepasst!!!!!!!!!!\n";
    lmstate_right_right = lmstate_left_right;
    
    word_rr_1 = lmstate_right_right &(((CLMState)1 << 32) -1);
    word_rr_2 = lmstate_right_right >> 32;
    /*
    cout << "in merge: \n";
    cout << "rightfirst: " << word_rr_1 << endl;
    cout << "rightsecond: " << word_rr_2 << endl;
    
    cout << "after change\n";
	*/
  }

  if ((lmstate_left_left & 0x00000000FFFFFFFF)==lm->vocab->ssIndex) {
    if ( word_rl_1 == lm->vocab->ssIndex ) {
      lmstate_left_left = AppendWord( lmstate_left_left, word_rl_2 );
    } else {
      lmstate_left_left = AppendWord( lmstate_left_left, word_rl_1 );
    }
    

  }

  /*
  Word_ll_1 = lmstate_left_left &(((CLMState)1 << 32) -1);
  Word_ll_2 = lmstate_left_left >> 32;
  cout << "ll 1 " << lm->vocab->getWord( Word_ll_1 ) << " ll 2 " << lm->vocab->getWord( Word_ll_2 ) << endl;

  Word_lr_1 = lmstate_left &(((CLMState)1 << 32) -1);
  Word_lr_2 = lmstate_left >> 32;
  cout << "lr 1 " << lm->vocab->getWord( Word_lr_1 ) << " lr 2 " << lm->vocab->getWord( Word_lr_2 ) << endl;

  Word_rl_1 = lmstate_right &(((CLMState)1 << 32) -1);
  Word_rl_2 = lmstate_right >> 32;
  cout << "rl 1 " << lm->vocab->getWord( Word_rl_1 ) << " rl 2 " << lm->vocab->getWord( Word_rl_2 ) << endl;

  Word_rr_1 = lmstate_right_right &(((CLMState)1 << 32) -1);
  Word_rr_2 = lmstate_right_right >> 32;
  cout << "rr 1 " << lm->vocab->getWord( Word_rr_1 ) << " rr 2 " << lm->vocab->getWord( Word_rr_2 ) << endl;
  */
  
  // the net gain is new LogProbOld - LogProbNew
  return LogProbNew - LogProbOld;
}



double
SriLMSentenceState::LogProbEnd( CLMState lmstate )
{
  //return LogProb( lmstate, lm->vocab->ssIndex );
  //cout << "SriLMSentenceState::LogProbEnd( CLMState lmstate ) with LMState " << lmstate << endl;
  //cout << lm->vocab->seIndex << " " << lm->vocab->getWord( lm->vocab->seIndex ) << endl;

  CLMState DummyState;
  double LogPr =  LogProb( lmstate, lm->vocab->seIndex, DummyState );
  //cout << "LogPr for sentence end = " << LogPr << endl;
  return LogPr;

}



CLMState
SriLMSentenceState::AppendWord( CLMState lmstate, const LMWordIndex word )
{
  //cout << "Appending:" << lm->vocab->getWord(word) << endl;
  unsigned LMorder = lm->GetOrder();
  CLMState ret;
  
  if ( LMorder >= (unsigned)4 ) {
    // we have only space for a 21 bit index here (LMWordIndex is 32, CLMState is 64)
    // this limits the vocabulary size to 2,097,151
    if ( word >= (((CLMState)1 << 21) - 1) ) {
      cout << "Vocabulary index too big for 4-gram LM!\n";
      exit(-1);
    }
     
    // lmstate, 64bit: |1bit|- 21bit: mostRecentWord -|- 21bit nextRecentWord -|- 21bit thirdRecentWord -|
    CLMState mostRecentWord = (CLMState)word;
    CLMState wordHistory = lmstate >> 21;  
    ret = ( (CLMState)wordHistory | ((CLMState)mostRecentWord << 42) );
      
    /*
       VocabIndex mostRecentWordT = ret >> 42;
       VocabIndex thirdRecentWordT = ret & ( ( (CLMState)1 << 21 ) - 1 );
       VocabIndex nextRecentWordT = (ret >> 21) & ( ( (CLMState)1 << 21 ) - 1 );
       cout << "Now in history: " << mostRecentWordT << "," << nextRecentWordT 
       << "," <<  thirdRecentWordT <<endl; 
       cout << " => " << lm->vocab->getWord(mostRecentWordT) << " " << lm->vocab->getWord(nextRecentWordT)
       << " " << lm->vocab->getWord(thirdRecentWordT) << endl;
    */

  } else { // trigram

    CLMState mostRecentWord= (CLMState)word;
    CLMState nextRecentWord = lmstate >> 32;
      
    // 2nd item, then 1st
    ret = ( (CLMState)nextRecentWord | ((CLMState)mostRecentWord << 32) );
  }
  
  return ret;
}



void
SriLMSentenceState::DeleteState( CLMState lmstate )
{
  // nothing to eo
}



// Param file for SriLM takes the arpa model only
SriLM::SriLM( char* paramFile, unsigned order, int LMUnkIsWord )
{
  string lmFile = string( paramFile );
  //string vocFile = lmFile +  ".vocab";
  
  //lmFile += ".srilm.gz";
  //cout<<"VocFile:"<<vocFile<<endl;
  //cout<<"LMFile:"<<lmFile<<endl;

  //File vocFileH(vocFile.c_str(), "r");
  vocab = new Vocab;
  vocab->unkIsWord = LMUnkIsWord;

  //int tokensRead = vocab->read(vocFileH);
  //cout<<"NumTokens:"<<tokensRead<<endl;

  cout << "Loading SRI language model: " << lmFile << endl;
  cout << "Order of language model:    " << order << endl;

  // ngram order
  //unsigned order = defaultNgramOrder;
  ngramLM = new Ngram( *vocab, order );
  File file( lmFile.c_str(), "r" );
  
  if ( !ngramLM->read(file) ) {
    cerr << "ERROR: could not read " << lmFile << endl;
    exit(1);
  }

  /*
  // maps string to LMWordIndex
  VocabIter vocIter = VocabIter(*vocab);
  VocabString oneWord = vocIter.next();
  int count = 0;
  while (oneWord!=0){
    //cout<<oneWord<<"--"<<vocab->getIndex(oneWord)<<endl;
    CLM::GetVocabularyIndexMap[string(oneWord)] = vocab->getIndex(oneWord);
    oneWord = vocIter.next();
    count++;
  }
  cout<<count<<" words loaded from vocab"<<endl;
  m_startIndex = vocab->getIndex("<s>");
  m_endIndex = vocab->getIndex("</s>");
  */
}



SriLM::SriLM( char* FileName, int Order, const CVocabulary &Vocabulary, bool LMRestrictVocabulary, bool LMUnkIsWord )
{
  cout << "Load SRI LM with restricted vocabulary of size " << Vocabulary.Size() << endl;
  //string lmFile = string( paramFile );

  vocab = new Vocab;
  vocab->unkIsWord = LMUnkIsWord;

  CString Word;
  for ( int i = Vocabulary.First(); i <= Vocabulary.Last(); i++ ) {
    Word = Vocabulary.GetWord( i );
    vocab->addWord( Word );
  }
  cout << "LM vocabulary has now size " << vocab->numWords() << endl;

  cout << "Loading SRI language model: " << FileName << endl;
  cout << "Order of language model:    " << Order << endl;

  // ngram order
  //unsigned order = defaultNgramOrder;
  ngramLM = new Ngram( *vocab, Order );
  File LMFile( FileName, "r" );

  if ( !ngramLM->read( LMFile, LMRestrictVocabulary ) ) {
    cerr << "ERROR: could not read " << FileName << endl;
    exit(1);
  }

}



SriLM::~SriLM()
{
  delete ngramLM;
  delete vocab;
}



SriLMSentenceState*
SriLM::NewSentence() const
{
  return new SriLMSentenceState(this);
}



void
SriLM::DeleteSentence( CLMSentenceState* st ) const
{
  delete st;
}



LMWordIndex
SriLM::VocabularySize() const
{
  return this->vocab->numWords();
  //return GetVocabularyIndexMap.size();
}



LMWordIndex
SriLM::GetVocabularyIndex( string word ) const
{
  VocabIndex i = vocab->getIndex( word.c_str() );
  if ( i == Vocab_None ) {
    //cout<<"unk word:"<<word<<" - "<<vocab->unkIndex<<endl;
    return vocab->unkIndex;
  }
  return i; 
}



LMWordIndex
SriLM::GetVocabularyIndex( string word )
{
  VocabIndex i = vocab->getIndex( word.c_str() );
  if ( i == Vocab_None ) {
	//cout<<"unk word:"<<word<<" - "<<vocab->unkIndex<<endl;
	return vocab->unkIndex;
  }
  return i; 
}



double
SriLM::IsolatedLogProb( vector<string> wordSequence )
{
  vector<LMWordIndex> wiVector;
  //cout<<"Evaluating: ";
  for ( int i = 0; i < wordSequence.size(); i++ ) {
    // cout << wordSequence[ i ] << " ";
    wiVector.push_back( GetVocabularyIndex( wordSequence[ i ] ) );
  }
  //cout<<endl;

  double sumLogP = 0;
  //now the wi vetor is filled with the lm word indices, for each one, create a context
  //of varying length, and query the lm
  for ( int i = 0; i < wiVector.size(); i++ ) {
    LMWordIndex wq = wiVector[i];
    int contextSize = i;

    double logP = 0.0;
    if ( wq == vocab->unkIndex ) {
      cout << "xxUnknown word" << wordSequence[ i ] << " ";
      logP = log10( 1.0 / VocabularySize() );
    }else{
      //context + 1 for the Vocab_None
      int finalSize;
      MAX( finalSize, contextSize + 1, 3 );

      VocabIndex* wids = new VocabIndex[ finalSize ];
      //cout<<"w: "<<wordSequence[i]<<" context: ";
      
      int cIndex =0;
      //needs to be in reverse order, uptil 3
      for ( int c = i-1 ; c >= i-2 && c >= 0; c-- ) {
		wids[ cIndex ] = wiVector[ c ];
		cIndex++;
		//cout<<wordSequence[c]<<" ";
      }
      wids[ cIndex ] = Vocab_None;
      //cout<<"Vocab_None"<<endl;

      logP = ngramLM->wordProb( wq, wids );
      delete [] wids;
    }
    //cout<<"logP: "<<wordSequence[i]<<" = "<<logP<<endl;
    sumLogP += logP;
  }
  return sumLogP;
}

