#include <config.h>
#include <fortify.h>
#ifndef __TRIE__
#define __TRIE__

#include <map>
#include <list>
#include <vector>
#include <string>
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <math.h>

#include <LMInterface.hh>

using namespace std;

//typedef int LMWordIndex;
typedef vector<LMWordIndex> NGram;
typedef vector<NGram*> NGramList; 

class TrieNode;
class PathState;

typedef map<LMWordIndex, TrieNode*> Branches;
// associates N with the pair (value, total) for NIST score
typedef map<int, pair<double, double> > NistMap;


// simple utility class to store NIST counts, and a reference to the gain trie
class NistScore{

public:

  NistScore(NistScore& ns);
  //initailizes the nistMap unto n-gram.
  NistScore(int n);
  ~NistScore();
  
  // extends the score to account for all hypothesis in this ngram
  void extend(NGram* nextGram);
  // increments this ngram's gain from the m_gains Trie
  void incrementScore(NGram* history);
  // calculate nist score for this instance given the lenth params
  double calculateScore(int Lref, int Lhyp);
  // loads the gain trie from the fileName presented
  // only needs to be called once on the initial NistScore
  static void loadGains(map<string, LMWordIndex> GetVocabularyIndexMap, const char* fileName);
  // delets the static gains object
  static void deleteGains();
  // converts the nistscore into a printable string
  string toString();
  // static gains structure
  static TrieNode* m_gains;

private:
  NistMap* m_nist;
  int m_nTotal;
};

class TrieNode{

public:
  // construtor builds complete trie from NGram for the given sequence
  TrieNode();
  TrieNode(TrieNode& t);
  TrieNode(NGram* sentence);
  ~TrieNode();

  // debugging string
  string toString();

  // creates a clone of this trie structure
  // generates all N-grams that would occur from history+word
  // for each one found in the clone, the score at that
  // level is decremented, and the NISTScore updated
  // modifies the history to indicate the updated word
  TrieNode* scoreArc(LMWordIndex nextWord, NGram* history, NistScore* score);
  
  void addNGram(NGram* oneNGram, double score);
  // gets the score for this particular ngram, decrements the score at that level
  // returns the score before the decrement
  double getScore(NGram::iterator pathBegin, NGram::iterator pathEnd, double decrement);
  
  //prints the n-gram 
  inline static string printNGram(NGram* gram){
    
    string s("");
    s.erase();
    cout<<"INEval:"<<s<<endl;
    for (NGram::iterator iter = gram->begin(); iter!=gram->end(); ++iter){
      int val = (*iter);
      cout<<"V:"<<val;
      char tmp[1000];
      sprintf(tmp, "%d:",val);
      s += tmp;
    }
    cout<<"S:"<<s<<endl;    
    return s;
  }

private:

  void incrementCount();

  NGramList* collectNGrams(NGram* source, int depth, int n);

  // represents the branching structure
  Branches* m_branch;
  // indicates that this pathway has already been used
  double m_numInstances;
  
};


class PathState{

public:
  
  PathState(NGram* sentence);
  ~PathState();

  // used when comparing two states
  string toString();

  PathState* scoreExtension(LMWordIndex nextWord);
  // calculate the score given a reference length
  double calculateScore(int Lref);

private:
  // private constructor to clone the path state
  PathState(TrieNode* root, NGram* history, NistScore* score, int numDecoded);
  // the root of this sentence's hitlist structure
  TrieNode* m_root;
  // the history of words that have just been seen
  NGram* m_history;
  // the number of source words decoded
  int m_numDecoded;
  // nist score for this path
  NistScore* m_score;
  
};



#endif




