// _SingleCorpusSALM.h: interface for the C_SingleCorpusSALM class.
//
// $Rev: 2296 $
// $LastChangedDate: 2007-06-27 15:09:33 -0400 (Wed, 27 Jun 2007) $
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX__SINGLECORPUSSALM_H__004675FC_A75A_4CED_A1A5_0811F29865A8__INCLUDED_)
#define AFX__SINGLECORPUSSALM_H__004675FC_A75A_4CED_A1A5_0811F29865A8__INCLUDED_

#include "shared.h"
#include "math.h"

#include "LMInterface.hh"

#include "_IDVocabulary.h"
#include "_String.h"
#include "_SuffixArraySearch.h"

const double SALM_PROB_UNK = 0.00000000023283064365386962890625; // 1/4G
const double SALM_LOG_PROB_UNK = log(SALM_PROB_UNK);
const double SALM_LOG_0 = -20;


typedef struct s_cached_SA_access_key{
	TextLenType currentMatchStart;
	unsigned char currentMatchLen;
	IndexType nextWord;
}S_CachedSA_Access_Key;

typedef struct s_cached_SA_access_info{
	TextLenType updatedMatchingStart;
	unsigned char updatedMatchingLen;
	double logProb;
	long lastTimedUsed;
}S_Cached_SA_Access_Info;

struct lt_s_cached_SA_access_key
{
  bool operator()(S_CachedSA_Access_Key a, S_CachedSA_Access_Key b) const{
		if(a.currentMatchStart<b.currentMatchStart){
			return true;
		}

		if(a.currentMatchStart>b.currentMatchStart){
			return false;
		}

		if(a.currentMatchLen<b.currentMatchLen){
			return true;
		}

		if(a.currentMatchLen>b.currentMatchLen){
			return false;
		}

		if(a.nextWord<b.nextWord){
			return true;
		}

		return false;	
	}
};

class C_SingleCorpusSALM : public CLM  
{
public:
	void setParam_maxLenOfNgramConsidered(int maxLenOfNgramConsidered);
	void setParam_interpolationStrategy(char interpolationStrategy);
	void setParam_stemLen(int stemLen);

	virtual LMWordIndex VocabularySize() const;
	virtual LMWordIndex returnSentEndVocId();
	virtual double LogProb(TextLenType currentMatchStart, unsigned char currentMatchLen, IndexType nextWord, TextLenType &updatedMatchingStart, unsigned char &updatedMatchingLen);

	virtual LMWordIndex GetVocabularyIndex(string word);
	virtual const vector< string > & Vocabulary() const;

	CLMSentenceState * NewSentence() const;

	C_SingleCorpusSALM(const char * cfgFileName);
	C_SingleCorpusSALM();
	virtual ~C_SingleCorpusSALM();

protected:
 	C_IDVocabulary * idVoc;
  
private:	
	
	bool applyDiscounting;

	double calcLogProb(double *freq);
	double calcLogProb_equalWeightedInterpolation(double * freq);
	double calcLogProb_ibmHeuristicInterpolation(double *freq);
	
	double corpusSize;
	int maxLenOfNgramConsidered;

	int stemLen;

	char interpolationStrategy;

	LMWordIndex sentEndVocId;

	C_SuffixArraySearch saSearchObj;


};

#endif // !defined(AFX__SINGLECORPUSSALM_H__004675FC_A75A_4CED_A1A5_0811F29865A8__INCLUDED_)
