#include <string.h>
// include configure results
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>

// enable/disable tracing for this module here
//#define NO_TRACE 1
#undef NO_TRACE
#include <Trace.hh>

#include <Lattice.hh>
#include <LatticeEdgeTrans.hh>
#include <LatticeEdgeMS.hh>



void
CLattice::Display() const
{
#ifdef TRACE_ENABLED
  int OldTraceLevel = TraceLevel;
  TraceLevel = 0;
#endif

  cout << "Lattice total: " << endl;
  cout << "Startnode: " << m_Nodes.MinIndex() << endl;
  cout << "Endnode: " << m_Nodes.MaxIndex() << endl;

  for (int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++) {
    if ( m_Nodes[n] ) {
      //cout << *(m_Nodes[n]);
      m_Nodes[n]->Display();
    } else {
      cout << "node without edges" << endl << endl;
    }
  }

  cout << endl << flush;

#ifdef TRACE_ENABLED
  TraceLevel = OldTraceLevel;
#endif

}



void
CLattice::DisplayEdges() const
{
  int e;

  cout << "All Edges:" << endl;
  for ( e = m_Edges.MinIndex(); e <= m_Edges.MaxIndex(); e++ ) {
    cout << e << " "; m_Edges[ e ]->Display();
  }
  cout << endl << flush;
}



void 
CLattice::DisplayStatistics() const
{
  // some statistics of the lattice
  cout << endl;

  int    MyNumberOfEdges     = this->NumberOfEdges();
  double MyAveragePathLength = this->AveragePathLength();

  cout << "Number of Nodes     = " << this->NumberOfNodes() << endl;
  cout << "Number of Edges     = " << MyNumberOfEdges << endl;
  cout << "Number of Pathes    = " << this->NumberOfPathes() << endl;

  cout << "Minimal Path Length = " << this->MinPathLength() << endl;
  cout << "Maximal Path Lenght = " << this->MaxPathLength() << endl;
  cout << "Average Path Length = " << MyAveragePathLength     << endl;

  float Density = (float)MyNumberOfEdges / (float)MyAveragePathLength;

  cout << "Lattice Density     = " << Density << endl;

  //cout.setf(ios::right | ios::fixed);
  //cout.width(10);
  //cout.precision(6);

  cout << endl;
}



int
CLattice::WriteToFile( const CString & Filename )
{
  CLatticeNode *pNode;
  CLatticeEdge *pEdge;

  ofstream outp( (const char * )Filename);
  
  if (!outp) {
    cout << "Textfile could not be opened for write\n";
    exit (-1);
  }

  outp << "BEGIN_LATTICE" << endl;
  for (int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++) {

    pNode = m_Nodes[ n ];

    const CArray<CLatticeEdge *> &OutEdges = pNode->OutEdges();

    for (int o = OutEdges.MinIndex(); o <= OutEdges.MaxIndex(); o++) {
      pEdge = OutEdges[o];
      
      outp << pEdge->FromNode()->Number() << "\t" << pEdge->ToNode()->Number() << "\t";
      pEdge->WriteTo(outp);
      outp << endl;
    }
  }
  outp << "END_LATTICE" << endl;
  return 0;
}

// einstelliger Ausgabeoperator fr Lattices.
/*
ostream& 
CLattice::operator<<( ostream& os )
{
  int OldTraceLevel = TraceLevel;
  TraceLevel = 0;

  cout << "Lattice total: " << endl;

  CLatticeNode *pNode;

  for (int n = m_Nodes.GetMinIndex(); n <= m_Nodes.GetMaxIndex(); n++) {
    cout << *(m_Nodes[n]);
  }

  cout << "Startnode: " << m_Nodes.GetMinIndex() << endl;
  cout << "Endnode: " << m_Nodes.GetMaxIndex() << endl;
  cout << endl << flush;

  TraceLevel = OldTraceLevel;

  return os;

}
*/


// zweistelliger Ausgabeoperator fr Lattices.
/*
ostream&
CLattice::operator<<( ostream& os, CLattice &Lattice )
{
  int OldTraceLevel = TraceLevel;
  TraceLevel = 0;

  cout << "Lattice total: " << endl;

  CLatticeNode *pNode;

  for (int n = m_Nodes.GetMinIndex(); n <= m_Nodes.GetMaxIndex(); n++) {
    cout << *(m_Nodes[n]);
  }

  cout << "Startnode: " << m_Nodes.GetMinIndex() << endl;
  cout << "Endnode: " << m_Nodes.GetMaxIndex() << endl;
  cout << endl << flush;

  TraceLevel = OldTraceLevel;

  return os;

  //  return Lattice.operator<<(os);
}
*/

// --------------------------------------------------
// AddNode
// --------------------------------------------------
void 
CLattice::AddNode( CLatticeNode *pNode )
{
  Trace( 3, "AddNode " << pNode->Number() );
  m_Nodes[ pNode->Number() ] = pNode;
}



// --------------------------------------------------
// AddEdge
// --------------------------------------------------
void
CLattice::AddEdge( CLatticeEdge *pEdge )
{
  // Kante in die Menge aller Kanten einfgen
  Trace( 3, "CLattice::AddEdge - use Append" );
  m_Edges.Append( pEdge );
  //pEdge->operator<<(cout);
  //cout<< endl;
  
  // Kante als ausgehende Kante eintragen
  CLatticeNode *pFromNode = pEdge->FromNode();
  pFromNode->AddOutEdge( pEdge );

  // Kante als eingehende Kante eintragen
  CLatticeNode *pToNode = pEdge->ToNode();
  pToNode->AddInEdge( pEdge );
}


bool
CLattice::AddEdgeUniq( CLatticeEdge *pEdge )
{
  bool Added = FALSE;

  // Kante in die Menge aller Kanten einfgen
  Trace0( 1, "CLattice::AddEdgeUniq - " );
  if ( TraceLevel >= 4 )
    pEdge->Display();

  CLatticeNode *pFromNode = pEdge->FromNode();
  const CArray< CLatticeEdge * > &OutEdges = pFromNode->OutEdges();

  // first check if edge with equal nodes and word exists
  int o = OutEdges.MinIndex();
  while ( (o <= OutEdges.MaxIndex()) && !(*(OutEdges[ o ]) == *pEdge) ) {
    Trace( 2, o << " " << ( *(OutEdges[ o ]) == *pEdge) );
    if ( TraceLevel >= 4 ) {
      OutEdges[ o ]->Display();
      pEdge->Display();
    }
    o++;
  }

  Trace( 4, o );
  if ( o <= OutEdges.MaxIndex() ) {   // if we have found equal one
    Trace0( 1, "Equal " );
    if ( pEdge->TotalScore() < OutEdges[ o ]->TotalScore() ) {    // and if new one is better
      Trace( 1, "and better" );
      RemoveEdge( OutEdges[ o ] );    // get rid of the old one
      AddEdge( pEdge );               // take the new one instead
      Added = TRUE;
    } else {
      Trace( 1, "but worse" );
    }
  } else {
    Trace( 1, "New one" );
    AddEdge( pEdge );                 // it's new anyway, so add it
    Added = TRUE;
  }

  return Added;
}



bool
CLattice::AddEdge( int FromNodeNumber, int ToNodeNumber, CLatticeEdge *pEdge )
{
  //cout << "AddEdge " << FromNodeNumber << " " << ToNodeNumber << " " << pEdge->Word() << endl;

  // first, check if nodes with these numbers exist; if not, add
  CLatticeNode *pFromNode = m_Nodes[ FromNodeNumber ];
  CLatticeNode *pToNode = m_Nodes[ ToNodeNumber ];

  if ( !pFromNode ) {
    pFromNode = new CLatticeNode( FromNodeNumber );
    m_Nodes[ FromNodeNumber ] = pFromNode;
  }

  if ( !pToNode ) {
    pToNode = new CLatticeNode( ToNodeNumber );
    m_Nodes[ ToNodeNumber ] = pToNode;
  }

  //cout << pFromNode->Number() << " " << pToNode->Number() << endl;

  // the edge needs the nodes
  pEdge->SetFromNode( pFromNode );
  pEdge->SetToNode( pToNode );


  // Kante als ausgehende Kante eintragen
  pFromNode = pEdge->FromNode();
  pFromNode->AddOutEdge( pEdge );
  //pFromNode->Display(); cout << endl;

  // Kante als eingehende Kante eintragen
  pToNode = pEdge->ToNode();
  pToNode->AddInEdge( pEdge );
  //pToNode->Display(); cout << endl;

  //pEdge->operator<<(cout);
  //pEdge->Display();
  //cout<< endl;

  // Kante in die Menge aller Kanten einfgen
  //Trace( 3, "CLattice::AddEdge - use Append" );
  m_Edges.Append( pEdge );

  //Display();
  //cout << endl;
}



/**
   @brief Remove edge from lattice.

   The edge will be removed from the "from node"'s out-list, from the "to node"'s in-list
   and furthermore from the lattice's edges list (the latter in linear time ! :-( )

   IMPORTANT: This method actually <i>frees</i> the edge object after removal!
 */
void
CLattice::RemoveEdge( CLatticeEdge *pEdge )
{
  Trace( 1, "CLattice::RemoveEdge" );

  // remove edge entries in FromNode and ToNode
  pEdge->FromNode()->RemoveOutEdge( pEdge );
  pEdge->ToNode()->RemoveInEdge( pEdge );

  // now, remove edge from list of all edges
  // first check if edge with equal nodes and word exists
  int i = m_Edges.MinIndex();
  //while ( (i <= m_Edges.MaxIndex()) && !(*(m_Edges[ i ]) == *pEdge) ) {
  while ( (i <= m_Edges.MaxIndex()) && (m_Edges[ i ] != pEdge) ) {
    //Trace( 4, i << " " << ( *(m_Edges[ i ]) == *pEdge) );
#ifdef TRACE_ENABLED
    if ( TraceLevel >= 4 ) {
      m_Edges[ i ]->Display();
      pEdge->Display();
    }
#endif
    i++;
  }

  Trace( 4, i );
  if ( i <= m_Edges.MaxIndex() ) {   // if we have found equal one
    Trace( 1, "Remove edge at index " << i );
    m_Edges.Remove( i );
  }
  
  delete pEdge;
}


void
CLattice::RemoveEdgesBelowThreshold( float threshold )
{
  
  
  Trace( 2, "CLattice::RemoveEdgesBelowThreshold" );
  
  for(int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++)
  {
    CLatticeNode *pNode = m_Nodes[ n ];
    CLatticeEdge *pBestEdgeBelowThreshold = NULL; // needed if none are above threshold
    const CArray< CLatticeEdge * > &OutEdges = pNode->OutEdges();
    
    bool hasEdgesAboveThreshold = false;
    
    for (int e = OutEdges.MaxIndex(); e >= OutEdges.MinIndex(); e--)
    {
      CLatticeEdge *pEdge = OutEdges[ e ];
      
      if( pEdge->Score() < threshold ) 
      {
        if( ! hasEdgesAboveThreshold )
        {
          if( pBestEdgeBelowThreshold == NULL ) {
            pBestEdgeBelowThreshold = pEdge;
          } else if( pBestEdgeBelowThreshold->Score() < pEdge->Score() ) {
            RemoveEdge( pBestEdgeBelowThreshold );
            pBestEdgeBelowThreshold = pEdge;
          } else {
            // best under threshold is better
            RemoveEdge( pEdge );
          }
        } else { // have above threshold, so don't need any below it
          RemoveEdge(pEdge);
        }
      } else { // keep current edge
        hasEdgesAboveThreshold = true;
      }
      
    } // end of edge loop
    
    // clean up best edge under threshold, if no longer needed
    if(hasEdgesAboveThreshold && pBestEdgeBelowThreshold != NULL) {
      RemoveEdge(pBestEdgeBelowThreshold);
    }
    
  } // end of node loop
  
}


void
CLattice::RemoveSpecialEdgesConnectLattice( int nodeJumpMax )
{
  Trace( 2, "CLattice::RemoveSpecialEdgesConnectLattice" );
  
  //cerr << "lattice before epsilon edge removal: "; Display();

  CString epsilon = "<epsilon>";

  // Prepare for special expansion case
  bool  epsilonOnFirstNode = false;
  float epsilonOnFirstNodeScore = 0.0;
  
  for( int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++ )
  {
    CLatticeNode *pNode = m_Nodes[ n ];
    const CArray< CLatticeEdge * > &OutEdges = pNode->OutEdges();
    
    bool epsilonEdgeFound = false; // assumption: only one epsilon between nodes
    
    for( int eOut = OutEdges.MaxIndex(); 
         eOut >= OutEdges.MinIndex() && !epsilonEdgeFound; eOut-- )
    {
      CLatticeEdgeMS *pOutEdge = (CLatticeEdgeMS*)OutEdges[ eOut ];
      
      // pOutEdge->Display();
      
      if( pOutEdge->Word() == epsilon )
      {
        epsilonEdgeFound = true;
        
        if ( pNode->Number() == 0 ) {
          epsilonOnFirstNode = true;
          epsilonOnFirstNodeScore = pOutEdge->Score();
        }
         
        // duplicate all incoming edges and make their toNodes n+1
        // change their scores to score*epsilon's probability
        
        CArray< CLatticeEdge * > &InEdges = pNode->InEdges();
        //cerr << "About to delete edge "; pOutEdge->Display();
        
        for( int eIn = InEdges.MaxIndex(); eIn >= InEdges.MinIndex(); eIn-- ) 
        {
          CLatticeEdgeMS *pInEdge = (CLatticeEdgeMS*)InEdges[ eIn ];
          
          
          //-- before exanding incoming edges, check if their prob is within reason (using negative logs)
          //   and that it's making a jump under a given threshold
          if( ( pInEdge->Score() + pOutEdge->Score() < 7 ) 
              && ( pInEdge->FromNode()->Number() + pOutEdge->FromNode()->Number() <= nodeJumpMax ) )
          {
            CLatticeEdge *pExpandedEdge = pInEdge->Clone(); // creates in-place duplicate
                                                            //pExpandedEdge->SetFromNode( pInEdge->FromNode() );
            pExpandedEdge->SetToNode( pOutEdge->ToNode() );
            
            if( OutEdges.Size() == 1 ) // epsilon is only edge
            {
              //isolate current node and redirect all incoming edges to next node
              pExpandedEdge->SetScore( pInEdge->Score() );
              RemoveEdge( pInEdge );
            } else {
              pExpandedEdge->SetScore( pInEdge->Score() + pOutEdge->Score() );
            }
            
            AddEdge( pExpandedEdge );
            //cerr << "Added edge "; pExpandedEdge->Display();
          }
        }
        
        RemoveEdge( pOutEdge );
        /*
         cerr << "node is isolated: " << (( (m_Nodes[ n ]->InEdges().Size() == 0) && 
                                            (m_Nodes[ n ]->OutEdges().Size() == 0) )? "YES": "NO") 
         << endl;
         */
        
        // don't delete isolated nodes here
        // array indexing will change to the right and Node->NumberS() won't
      } 
      
    } // end edge loop
    
  } // end node loop
  
  
  
  
  
  
  //cout << "Lattice stats AFTER: "; DisplayStatistics();
  
  // remove isolated nodes (can only be consecutive, previously connected only
  // by <epsilon>, starting with the first lattice node). deleting right to left
  // because deletion changes array indices on right of deleted node
  
  for(int n = m_Nodes.MaxIndex(); n >= m_Nodes.MinIndex(); n--)
  {
    if ( (m_Nodes[ n ]->InEdges().Size() == 0) && 
         (m_Nodes[ n ]->OutEdges().Size() == 0) ) {
      m_Nodes.Remove( n );
      // move right and decrement Node->Number by 1
    }
  } // end node delete loop
  
  //cerr << "lattice after epsilon edge removal: "; Display();
  if ( epsilonOnFirstNode )
  {
    CLatticeNode* pFirstNode = m_Nodes[ 0 ];
    CArray< CLatticeEdge* > OutEdgesFromFirst = pFirstNode->OutEdges();
    for ( int e =  OutEdgesFromFirst.MinIndex(); 
          e <= OutEdgesFromFirst.MaxIndex(); e++ )
    {
      CLatticeNode* pToNode = OutEdgesFromFirst[ e ]->ToNode();
      
      // Duplicate outgoing edges from ToNode
      CArray< CLatticeEdge* > OutEdgesFromSecond = pToNode->OutEdges();
      for ( int eFromSecond =   OutEdgesFromSecond.MinIndex(); 
            eFromSecond <=  OutEdgesFromSecond.MaxIndex(); eFromSecond++ )
      {
        CLatticeEdgeMS* pEdgeToExpand = (CLatticeEdgeMS*)(OutEdgesFromSecond[ eFromSecond ]);
        
        // Will expand edge backwards if prob and jump within thresholds
        if ( ( pEdgeToExpand->Score() + epsilonOnFirstNodeScore < 7 )
             && ( pEdgeToExpand->ToNode()->Number() <= nodeJumpMax ) )
        {
          
          CLatticeEdge* pBackExpandedEdge = pEdgeToExpand->Clone(); // in-place duplicate
          pBackExpandedEdge->SetFromNode( pFirstNode );
          pBackExpandedEdge->SetScore( pEdgeToExpand->Score() + epsilonOnFirstNodeScore );
          
          AddEdge( pBackExpandedEdge );
        }
      } // edges coming out of ToNode of edges coming out of first node
      
    } // edges coming out of first node
  }    
}



bool
CLattice::RemoveSpecialEdges()
{
  // run over all nodes starting from second node
  // for each node run over all outgoing edges
  // if edge is to be removed
  // span all incoming edges to ToNode of edge which should be removed
  // remove outgoing edge
  // if in the end no outgoing edge is left
  // remove all incoming edges

  
  Trace( 2, "CLattice::RemoveSpecialEdges" );

  //Display();

  CLatticeNode *pNode = NULL;
  CLatticeEdge *pOutEdge = NULL;
  int NumberOfOutEdges;
  //int NumberOfInEdges;

  for (int n = m_Nodes.MinIndex() + 1; n <= m_Nodes.MaxIndex(); n++) {
    pNode = m_Nodes[ n ];

    if ( TraceLevel >= 4 )
      pNode->Display();

    Trace( 1, "Number of Node = " << pNode->Number() );
    const CArray< CLatticeEdge * > &OutEdges = pNode->OutEdges();
    Trace( 1, "Number of OutEdges = " << OutEdges.Size() << " " 
	   << OutEdges.MinIndex() << " " << OutEdges.MaxIndex());

    int o = OutEdges.MaxIndex();
    while ( o >= OutEdges.MinIndex() ) {
      Trace( 1, o );
      pOutEdge = OutEdges[ o ];
      if ( pOutEdge->ShouldBeRemoved() ) {
	Trace0( 1, "OutEdge should be removed! "  );
	if ( TraceLevel >= 1 )
	  pOutEdge->Display();
	CopyInEdges( pNode, pOutEdge );
	RemoveEdge( pOutEdge );
      }

      o--;
    }

    // if node has no outgoing edges anymore
    // we can delete the incoming edges as well
    // well, this does not apply to the last edge
    if ( (n < m_Nodes.MaxIndex()) && (OutEdges.Size() == 0) ) {
      RemoveInEdges( pNode );
    }
    
    Trace( 4, "After remove edges" );
    if ( TraceLevel >= 4 )
      pNode->Display();

  }


  // now the outgoing edges of node 1
  bool IsConnected = TRUE;

  pNode = m_Nodes[ m_Nodes.MinIndex() ];
  if ( TraceLevel >= 4 )
    pNode->Display();

  const CArray< CLatticeEdge * > &OutEdges = pNode->OutEdges();
  NumberOfOutEdges = OutEdges.Size();

  int o = OutEdges.MaxIndex();
  while ( o >= OutEdges.MinIndex() ) {
    pOutEdge = OutEdges[ o ];
    if ( TraceLevel >= 4 )
      pOutEdge->ToNode()->Display();

    if ( pOutEdge->ShouldBeRemoved() ) {
      Trace0( 1, "OutEdge of first node should be removed " );
      if ( TraceLevel >= 4 )
	pOutEdge->Display();
      CopyOutEdges( pOutEdge->ToNode(), pOutEdge );

      RemoveEdge( pOutEdge );

      // if node has no incoming edges anymore
      // we can delete the outgoing edges as well
      if ( pOutEdge->ToNode()->InEdges().Size() == 0 ) {
	RemoveOutEdges( pOutEdge->ToNode() );
	IsConnected = FALSE;
      }

    }
    o--;
  }

  if ( TraceLevel >= 4 )
    pNode->Display();

  return IsConnected;
}



void
CLattice::CopyInEdges( CLatticeNode *pNode, CLatticeEdge *pOutEdge )
{

  CLatticeEdge *pInEdge;
  CLatticeEdge *pNewEdge;

  const CArray< CLatticeEdge * > &InEdges = pNode->InEdges();
  //CLatticeNode *pToNode = pOutEdge->ToNode();

  for ( int i = InEdges.MinIndex(); i <= InEdges.MaxIndex(); i++ ) {
    pInEdge = InEdges[ i ];
    pNewEdge = pInEdge->Concatenate( pOutEdge );

    Trace0(4, "New Edge is: " );
    if ( TraceLevel >= 4 )
      pNewEdge->Display();
    
    AddEdgeUniq( pNewEdge );

    if ( TraceLevel >= 4 )
      pNode->Display();
  }

}



void
CLattice::CopyOutEdges( CLatticeNode *pNode, CLatticeEdge *pInEdge )
{
  CLatticeEdge *pOutEdge;
  CLatticeEdge *pNewEdge;

  const CArray< CLatticeEdge * > &OutEdges = pNode->OutEdges();
  //CLatticeNode *pFromNode = pInEdge->FromNode();

  for ( int o = OutEdges.MinIndex(); o <= OutEdges.MaxIndex(); o++ ) {
    pOutEdge = OutEdges[ o ];
    pNewEdge = pOutEdge->Concatenate( pInEdge );

    Trace0(4, "New Edge is: " );
    if ( TraceLevel >= 4 )
      pNewEdge->Display();
    
    AddEdgeUniq( pNewEdge );
  }

}



// --------------------------------------------------
// Clear
// --------------------------------------------------
void
CLattice::Clear()
{
  TRACE( 3, "CLattice::Clear()" );
  Trace( 3, m_Nodes.MinIndex() << " " <<  m_Nodes.MaxIndex() );

  for (int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++) {
    Trace( 5, "delete node " << n );
    delete m_Nodes[ n ];
    m_Nodes[ n ] = NULL;
  }

  for ( int e = m_Edges.MinIndex(); e <= m_Edges.MaxIndex(); e++ ) {
    Trace( 5, "delete edge " << e );
    delete m_Edges[ e ];
    m_Edges[ e ] = NULL;
  }

  m_Nodes.Clear();
  m_Edges.Clear();

  Trace( 5, m_Nodes.MinIndex() << " " <<  m_Nodes.MaxIndex() );
  Trace( 5, m_Edges.MinIndex() << " " <<  m_Edges.MaxIndex() );

}



int
CLattice::ReadFromFile( const CString & Filename )
{
  Clear();

  ifstream inp( (const char *) Filename);

  int Err = 0;
  if( inp.good() ) {
    char buffer[ MaxLineLength ];
    buffer[0] = '\0';

    while( inp.good() && (Err == 0) ) {
      // Zeile einlesen
      inp.getline( buffer, MaxLineLength );
      buffer[ MaxLineLength-1 ] = '\0';
      Trace( 5, buffer );

      // Zeile verarbeiten
      if (m_FileFormat == DEFAULT) {
        Err = ReadLine( buffer );
      } else if (m_FileFormat == ASR) {
        Err = ReadLineASR( buffer );
      } else if (m_FileFormat == ASRMS) {
        Err = ReadLineMS( buffer );
      } else if (m_FileFormat == SRI) {
        Err = ReadLineSRI( buffer );
      } else {
        cout << "invalid file format condition in ReadFromFile" << endl;
      }
      
    }
  } else {
    return 1;
  }

  for ( int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++ ) {
    if ( m_Nodes[ n ] == NULL ) {
      m_Nodes[ n ] = new CLatticeNode( n );
    }
  }

  return Err;
}




int
CLattice::ReadFromBuffer( const CString & WhgBuffer )
{
  const char *Ptr = WhgBuffer;

  // Lattice in Initialzustand versetzen
  Clear();

  int n = 0;
  char LineBuffer[MaxLineLength];
  LineBuffer[0]='\0';

  while(*Ptr) {
    //cout << *Ptr;
    // eine Zeile eingelesen
    if (*Ptr == '\n') {
      *(LineBuffer + n) = '\0';
      n = 0;
      // Zeile verarbeiten
      //cout << endl << LineBuffer << flush;
      if( m_FileFormat == DEFAULT ) {
        ReadLine( LineBuffer );
      } else if( m_FileFormat == ASR ) {
        ReadLineASR( LineBuffer );
      } else if (m_FileFormat == ASRMS) {
        ReadLineMS( LineBuffer );
      } else if( m_FileFormat == SRI ) {
        ReadLineSRI( LineBuffer );
      }
      //i = AddVMLine(LineBuffer);
      //if(i!=0)
      //return i;
    } else {
      *(LineBuffer + n) = *Ptr;
      n++;
    }
    Ptr++;
  }
  Trace(4, "readFromBuffer done ");
  //Trace(4, *this);

  for ( int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++ ) {
    if ( m_Nodes[ n ] == NULL ) {
      m_Nodes[ n ] = new CLatticeNode( n );
    }
  }
  return 0;
}


int
CLattice::ConvertFromText( const CArray< CString > &Words )
{
  // Lattice in Initialzustand versetzen
  Clear();

  int Index;
  int Err;

  for ( Index = Words.MinIndex(); Index <= Words.MaxIndex(); Index++ ) {
    Trace( 2, Words[ Index ] );
    Err = ReadLine( Words[ Index ] );
    if( Err != 0 )
      return Error;
  }

  return 0;
}


int
CLattice::AddEdgesFromText( const CArray< CString > &Words )
{
  int Index;
  int Err;

  for ( Index = Words.MinIndex(); Index <= Words.MaxIndex(); Index++ ) {
    Trace( 2, Words[ Index ] );
    Err = ReadLine( Words[ Index ] );
    if( Err != 0 )
      return Error;
  }

  return 0;
}

int
CLattice::ReadLineSRI( const CString & Buffer )
{
  // Zeile enth"alt Anfangs- und Endknotennummer, sowie Wort
  // wenn Knoten noch nicht existieren, werden sie neu in die 
  // Liste eingefuegt
  // Kante wird erzeugt und in die Kantenliste eingef"ugt
  int oldTraceLevel = TraceLevel;
  //  TraceLevel = 2;
  Trace(2, "ReadLine " << Buffer);
  // consider line only if it starts with align
  // line in the format align x word1 score1 word2 score2
  //    from x to x+1
  if((strlen(Buffer) > 0) &&
     (strstr(Buffer, "name") == NULL) &&
     (strstr(Buffer, "numaligns" ) == NULL) &&
     (strstr(Buffer, "posterior" ) == NULL) &&
     (strstr(Buffer, "<s>"       ) == NULL) &&
     (strstr(Buffer, "</s>"      ) == NULL)  )  
  {
    
    // Zeile mit sinnvollem Inhalt
    unsigned long From, To;
    CLatticeNode *pFromNode = NULL, *pToNode = NULL;
    float sc;                       
    
    // get "align"
    char* BufferCopy = strdup(Buffer);
    char* wordBuffer = strtok(BufferCopy, " ");
    
    // set From and to
    wordBuffer = strtok(NULL, " ");
    sscanf(wordBuffer, "%lu", &From);
    //    From -= 2; // index starts at 2

    From = ( m_Nodes.IsEmpty() ) ? 0 : m_Nodes.MaxIndex();
    
    To    = From + 1;
    
    // create To and From nodes, if necessary
    // Knoten in Liste eintragen
    Trace(2, "ReadLine - search FromNode");
    if ( (pFromNode = m_Nodes[ From ]) == NULL ) {
      pFromNode = new CLatticeNode( From );
      Trace(2, "ReadLine - add FromNode");
      AddNode( pFromNode );
    }
    
    Trace(2, "ReadLine - search ToNode");
    if ( (pToNode = m_Nodes[ To ]) == NULL ) {
      pToNode = new CLatticeNode( To );
      Trace(2, "ReadLine - add ToNode");
      AddNode( pToNode );
    }
    
    Trace( 2, "Nodes added" );
    
    
    
    // get every word score pair in line and then add as edges between From and To
    while (wordBuffer != NULL) 
    {
      wordBuffer = strtok(NULL, " ");
      
      if (wordBuffer == NULL) {
        break;
      }
      
      // get score from string
      char* scoreStr = strtok(NULL, " ");
      sscanf(scoreStr, "%f", &sc);
      
      // check if it's an arc with no output
      if (strstr(wordBuffer, "*DELETE*") != NULL){
        wordBuffer = "<epsilon>";
      }
      
      // create the edge
      CLatticeEdgeMS *pEdge = new CLatticeEdgeMS();
      
      Trace( 2, "Init edge with buffer " << Buffer );
      
      // assume SRI input must be converted to -log. cut off -log values at 0
      if( sc == 1 ) {
        sc = 0;
      } else {
        sc = -log(sc);
      }
      
      pEdge->Init( pFromNode, pToNode, From, To, CString (wordBuffer), 
                   sc, 0, 0, (const char*) Buffer);  // what about Buffer? it has other edges

      pEdge->Scores().Append( sc );
          
      Trace( 2, "Now add edge" );
      
      AddEdge( pEdge );
      Trace( 2, "From: " << From << " To: " << To << "Word: " << wordBuffer 
             << " Score: " << sc);
    }
    
    free(BufferCopy);
  }
  
  //  Display();
  TraceLevel = oldTraceLevel;
  return 0;
  
}


int
CLattice::ReadLineASR( const CString & Buffer )
{
  // ignore <s> and <\s>
  
  // Zeile enth"alt Anfangs- und Endknotennummer, sowie Wort
  // wenn Knoten noch nicht existieren, werden sie neu in die 
  // Liste eingefuegt
  // Kante wird erzeugt und in die Kantenliste eingef"ugt
  
  Trace(2, "ReadLine " << Buffer);
  
  if((strlen(Buffer)>0) && (Buffer[0]!='%') && (Buffer[0]!='#') &&
     (strstr(Buffer,"BEGIN_LATTICE")==NULL) &&
     (strstr(Buffer,"END_LATTICE") == NULL))
  {
    // Zeile mit sinnvollem Inhalt
    unsigned long From, To;
    unsigned long fframe = 0, tframe = 0;
    float sc;                         // A

    CLatticeNode *pFromNode = NULL, *pToNode = NULL;
    char WordBuffer[ MaxLineLength ];
    WordBuffer[ 0 ] = '\0';
    
    //sscanf(Buffer,"%lu%lu%s%lf%lu%lu",&From, &To, WordBuffer, &sc, &fframe, &tframe);
    //sscanf(Buffer,"%lu%lu", &From, &To );
    
    sscanf(Buffer, "%lu %lu %s %f", &From, &To, WordBuffer, &sc);
    //sscanf(Buffer, "%lu %lu %s %f %lu %lu", &From, &To, WordBuffer, &sc, &fframe, &tframe);
    
    if ( strstr( WordBuffer, "~SIL" ) != NULL ) {
      strcpy( WordBuffer, "<epsilon>" );
    }
    
    //Trace( 1, "From: " << From << " To: " << To << " | " << WordBuffer );
    Trace( 2, "From: " << From << " To: " << To << " Score: " << sc);
    
    // Knoten in Liste eintragen
    Trace(2, "ReadLine - search FromNode");
    if ( (pFromNode = m_Nodes[ From ]) == NULL ) {
      pFromNode = new CLatticeNode( From );
      Trace(2, "ReadLine - add FromNode");
      AddNode( pFromNode );
    }
    
    Trace(2, "ReadLine - search ToNode");
    if ( (pToNode = m_Nodes[ To ]) == NULL ) {
      pToNode = new CLatticeNode( To );
      Trace(2, "ReadLine - add ToNode");
      AddNode( pToNode );
    }
    
    Trace( 2, "Nodes added" );
    
    //CLatticeEdge *pEdge = m_pEdgePrototype->Clone();       //apd
    //CLatticeEdgeTrans *pEdge = new CLatticeEdgeTrans(); 
    CLatticeEdgeMS *pEdge = new CLatticeEdgeMS();
    Trace( 2, "Init edge with buffer " << Buffer );
    
    // asr scores are probabilities that need to be converted to -log prob
    if( sc == 1 ) {
      sc = 0;
    } else {
      sc = -log(sc);
    }
    
    pEdge->Init( pFromNode, pToNode, From, To, CString( WordBuffer ),
                 sc, 0, 0, (const char*) Buffer); 
    
    //pEdge->Init( pFromNode, pToNode, From, To, CString (WordBuffer), 
    //             sc, fframe, tframe, (const char*) Buffer); 

    pEdge->Scores().Append( sc );
    
    Trace( 2, "Now add edge" );
    
    AddEdge( pEdge );
    
  }
  
  //Display();
  
  return 0;
  
}

//! Reads ASR edges with multple edge scores
int
CLattice::ReadLineMS( const CString & Buffer )
{
  // ignore <s> and <\s>
  
  // Zeile enth"alt Anfangs- und Endknotennummer, sowie Wort
  // wenn Knoten noch nicht existieren, werden sie neu in die 
  // Liste eingefuegt
  // Kante wird erzeugt und in die Kantenliste eingef"ugt
  
  Trace(2, "ReadLine " << Buffer);
  
  if((strlen(Buffer)>0) && (Buffer[0]!='%') &&
     (strstr(Buffer,"BEGIN_LATTICE")==NULL) &&
     (strstr(Buffer,"END_LATTICE") == NULL))  
  {
    // Zeile mit sinnvollem Inhalt
    unsigned long From, To;
    unsigned long fframe = 0, tframe = 0;
    float sc;                        

    CLatticeNode *pFromNode = NULL, *pToNode = NULL;
    char WordBuffer[ MaxLineLength ];
    WordBuffer[ 0 ] = '\0';
    
    // Store multiple float scores
    char ScoreBuffer[ MaxLineLength ];
    ScoreBuffer[ 0 ] = '\0'; 
    
    sscanf(Buffer, "%lu %lu %s %[0-9 .]", &From, &To, WordBuffer, ScoreBuffer);
    
    cout << "read: '" << From << "' '" << To << "' '" << WordBuffer 
         << "' scoreBuff:'" << ScoreBuffer << "'" << endl;
    
    if ( strstr( WordBuffer, "~SIL" ) != NULL ) {
      strcpy( WordBuffer, "<epsilon>" );
    }

    //Trace( 1, "From: " << From << " To: " << To << " | " << WordBuffer );
    Trace( 2, "From: " << From << " To: " << To << " Score: " << sc);
    
    // Knoten in Liste eintragen
    Trace(2, "ReadLine - search FromNode");
    if ( (pFromNode = m_Nodes[ From ]) == NULL ) {
      pFromNode = new CLatticeNode( From );
      Trace(2, "ReadLine - add FromNode");
      AddNode( pFromNode );
    }
    
    Trace(2, "ReadLine - search ToNode");
    if ( (pToNode = m_Nodes[ To ]) == NULL ) {
      pToNode = new CLatticeNode( To );
      Trace(2, "ReadLine - add ToNode");
      AddNode( pToNode );
    }
    
    Trace( 2, "Nodes added" );
    
    //CLatticeEdge *pEdge = m_pEdgePrototype->Clone();       //apd
    //CLatticeEdgeTrans *pEdge = new CLatticeEdgeTrans(); 
    CLatticeEdgeMS *pEdge = new CLatticeEdgeMS();
    Trace( 2, "Init edge with buffer " << Buffer );
    

    
    pEdge->Init( pFromNode, pToNode, From, To, CString( WordBuffer ), 
                 0, 0, 0, (const char*) Buffer); 
    
    //pEdge->Init( pFromNode, pToNode, From, To, CString (WordBuffer), 
    //             sc, fframe, tframe, (const char*) Buffer); 
    

    // Parse Scores
    
    char* scoreChar;
    scoreChar = strtok( ScoreBuffer, " " );
    while ( scoreChar != NULL )
    {

      // convert scoreChar to float
      sscanf( scoreChar, "%f", &sc );
      
      // asr scores are probabilities that need to be converted to -log prob
      if( sc == 1 ) {
        sc = 0;
      } else {
        sc = -log(sc);
      }
      pEdge->Scores().Append( sc );
      
      cout << "Appended score: '" << sc << "'" << endl;
      
      scoreChar = strtok( NULL, " " );
    }
    
    Trace( 2, "Now add edge" );
    
    AddEdge( pEdge );
    
  }
  
  //Display();
  
  return 0;
  
}


// Dokumentation: siehe lattice.h
int
CLattice::ReadLine( const CString & Buffer )
{
  // Zeile enth"alt Anfangs- und Endknotennummer, sowie Wort
  // wenn Knoten noch nicht existieren, werden sie neu in die 
  // Liste eingefuegt
  // Kante wird erzeugt und in die Kantenliste eingef"ugt

  Trace(2, "ReadLine " << Buffer);

  if( (strlen(Buffer) > 0) && (Buffer[0]!='%')   &&
      (strstr(Buffer,"BEGIN_LATTICE")==NULL) &&
      (strstr(Buffer,"END_LATTICE")==NULL))  {
    // Zeile mit sinnvollem Inhalt
    unsigned long From, To;
    //unsigned long fframe, tframe;
    //float sc;
    //TNodeID fnode = 0, tnode = 0;
    CLatticeNode *pFromNode = NULL, *pToNode = NULL;
    char WordBuffer[ MaxLineLength ];
    WordBuffer[ 0 ] = '\0';

    //sscanf(Buffer,"%lu%lu%s%lf%lu%lu",&From, &To, WordBuffer, &sc, &fframe, &tframe);
    sscanf(Buffer,"%lu%lu", &From, &To );
    //Trace( 1, "From: " << From << " To: " << To << " | " << WordBuffer );
    Trace( 2, "From: " << From << " To: " << To );
    
    // Knoten in Liste eintragen
    Trace(2, "ReadLine - search FromNode");
    if ( (pFromNode = m_Nodes[ From ]) == NULL ) {
      pFromNode = new CLatticeNode( From );
      Trace(2, "ReadLine - add FromNode");
      AddNode( pFromNode );
    }
    
    Trace(2, "ReadLine - search ToNode");
    if ( (pToNode = m_Nodes[ To ]) == NULL ) {
      pToNode = new CLatticeNode( To );
      Trace(2, "ReadLine - add ToNode");
      AddNode( pToNode );
    }
    
    Trace( 2, "Nodes added" );

    //CLatticeEdge *pEdge = new CLatticeEdge( pFromNode, pToNode, WordBuffer );
    CLatticeEdge *pEdge = m_pEdgePrototype->Clone();
    //Trace( 2, "Init edge with buffer " << Buffer );
    pEdge->Init( pFromNode, pToNode, (const char *)Buffer );
    //pEdge->Init( pFromNode, pToNode );
    Trace( 2, "Now add edge" );
    AddEdge( pEdge );
    
  }
  
  //Display();

  return 0;
}

/**
 * @brief Parses given string for file format and sets m_FileFormat.
 *
 * @return 0 on success, <0 if format is not recognized.
 *
 * @author Anthony D'Auria (apd@cs.cmu.edu)
 *
 */
int 
CLattice::SetLatticeFileFormat( char* LatticeFormat )
{
  
  if (strcmp(LatticeFormat, "ASR") == 0) {
    SetLatticeFileFormat(CLattice::ASR);
  } else if (strcmp(LatticeFormat, "SRI") == 0) {
    SetLatticeFileFormat(CLattice::SRI);
  } else if (strcmp(LatticeFormat, "ASR-MS") == 0) {
    SetLatticeFileFormat( CLattice::ASRMS );
  } else if (strcmp(LatticeFormat, "VM") == 0) {        
    // read in same way as ASR
    SetLatticeFileFormat(CLattice::ASR);
  } else {
    // invalid lattice format given
    cerr << "SetLatticeFileFormat(): Invalid lattice format given: '" << LatticeFormat
    << "'. Options are SRI, ASR, ASR-MS, VM. Exiting." << endl << flush;
    return -1;
  }
  
  return 0;
  
}


void 
CLattice::SetLatticeFileFormat(LatticeFileFormat Format)
{
  m_FileFormat = Format;
}

CLattice::LatticeFileFormat
CLattice::GetLatticeFileFormat()
{
  return m_FileFormat;
}


void
CLattice::Traverse()
{
  Trace( 1, "Traverse Lattice" );
  Trace( 1, m_Nodes.MinIndex() << " " <<  m_Nodes.MaxIndex() );

  CLatticeNode *pNode = NULL;
  CLatticeEdge *pEdge = NULL;

  for (int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++) {
    pNode = m_Nodes[ n ];
    Trace( 1, "Number of Node = " << pNode->Number() );
    const CArray< CLatticeEdge * > &OutEdges = pNode->OutEdges();
    Trace( 1, "Number of OutEdges = " << OutEdges.Size() );
    for ( int o = OutEdges.MinIndex(); o <= OutEdges.MaxIndex(); o++ ) {
      pEdge = OutEdges[ o ];
      //Trace( 1, "OutEdge " << o << ": " << pEdge->Word() ) ;
    }

  } 

}


int
CLattice::MaxPathLength() const
{
  Trace( 1, "Lattice::MaxPathLength" );
  Trace( 1, m_Nodes.MinIndex() << " " <<  m_Nodes.MaxIndex() );

  if ( m_Nodes.MaxIndex() < m_Nodes.MinIndex() ) {
    Trace( 0, "MinPathLength = 0 - empty lattice" );
    return 0;
  }

  CArray< int > MaxLength;
  MaxLength.SetBoundaries( m_Nodes.MinIndex(), m_Nodes.MaxIndex() );
  MaxLength.SetAll( 0 );

  CLatticeNode *pNode = NULL;
  CLatticeEdge *pEdge = NULL;
  int PrevNodeNumber;

  for (int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++) {
    pNode = m_Nodes[ n ];
    Trace( 4, "Number of Node = " << pNode->Number() );
    const CArray< CLatticeEdge * > &InEdges = pNode->InEdges();
    Trace( 4, "Number of InEdges = " << InEdges.Size() );
    for ( int i = InEdges.MinIndex(); i <= InEdges.MaxIndex(); i++ ) {
      pEdge = InEdges[ i ];
      //Trace( 1, "OutEdge " << o << ": " << pEdge->Word() ) ;
      PrevNodeNumber = pEdge->FromNode()->Number();

      if ( MaxLength[ PrevNodeNumber ] + 1 > MaxLength[ n ] ) {
	MaxLength[ n ] = MaxLength[ PrevNodeNumber ] + 1;
      }

    }

  }

  Trace( 1, "MaxPathLength = " << MaxLength[ m_Nodes.MaxIndex() ] );
  return MaxLength[ m_Nodes.MaxIndex() ];
}



int
CLattice::MinPathLength() const
{
  Trace( 1, "Lattice::MinPathLength" );
  Trace( 1, m_Nodes.MinIndex() << " " <<  m_Nodes.MaxIndex() );

  if ( m_Nodes.MaxIndex() < m_Nodes.MinIndex() ) {
    Trace( 0, "MaxPathLength = 0 - empty lattice" );
    return 0;
  }

  CArray< int > MinLength;
  MinLength.SetBoundaries( m_Nodes.MinIndex(), m_Nodes.MaxIndex() );
  MinLength.SetAll( 10000 );
  MinLength[ m_Nodes.MinIndex() ] = 0;

  CLatticeNode *pNode = NULL;
  CLatticeEdge *pEdge = NULL;
  int PrevNodeNumber;

  for (int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++) {
    pNode = m_Nodes[ n ];
    Trace( 4, "Number of Node = " << pNode->Number() );
    const CArray< CLatticeEdge * > &InEdges = pNode->InEdges();
    Trace( 4, "Number of InEdges = " << InEdges.Size() );
    for ( int i = InEdges.MinIndex(); i <= InEdges.MaxIndex(); i++ ) {
      pEdge = InEdges[ i ];
      //Trace( 1, "OutEdge " << o << ": " << pEdge->Word() ) ;
      PrevNodeNumber = pEdge->FromNode()->Number();

      if ( MinLength[ PrevNodeNumber ] + 1 < MinLength[ n ] ) {
	MinLength[ n ] = MinLength[ PrevNodeNumber ] + 1;
      }

    }

  }

  Trace( 1, "MinPathLength = " << MinLength[ m_Nodes.MaxIndex() ] );
  return MinLength[ m_Nodes.MaxIndex() ];
}



float
CLattice::AveragePathLength() const
{
  Trace( 1, "Lattice::AveragePathLength" );
  Trace( 2, m_Nodes.MinIndex() << " " <<  m_Nodes.MaxIndex() );

  if ( m_Nodes.MaxIndex() < m_Nodes.MinIndex() ) {
    Trace( 0, "Average path length = 0 - empty lattice" );
    return 0;
  }

  CArray< float > N; // number of pathes
  N.SetBoundaries( m_Nodes.MinIndex(), m_Nodes.MaxIndex() );
  N.SetAll( 0.0 );
  N[ m_Nodes.MinIndex() ] = 1.0;

  CArray< float > Sum;
  Sum.SetBoundaries( m_Nodes.MinIndex(), m_Nodes.MaxIndex() );
  Sum.SetAll( 0.0 );

  CArray< float > Average;
  Average.SetBoundaries( m_Nodes.MinIndex(), m_Nodes.MaxIndex() );
  Average.SetAll( 0.0 );

 
  CLatticeNode *pNode = NULL;
  CLatticeEdge *pEdge = NULL;
  int PrevNodeNumber;

  for (int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++) {
    pNode = m_Nodes[ n ];
    Trace( 2, "Number of Node = " << pNode->Number() );
    const CArray< CLatticeEdge * > &InEdges = pNode->InEdges();
    Trace( 2, "Number of InEdges = " << InEdges.Size() );
    for ( int i = InEdges.MinIndex(); i <= InEdges.MaxIndex(); i++ ) {
      pEdge = InEdges[ i ];
      //Trace( 1, "OutEdge " << o << ": " << pEdge->Word() ) ;
      PrevNodeNumber = pEdge->FromNode()->Number();

      N[ n ] += N[ PrevNodeNumber ];
      Sum[ n ] += ( Sum[ PrevNodeNumber ] + N[ PrevNodeNumber ] );
      
    }

    if ( N[ n ] > 0.0 )
      Average[ n ] = Sum[ n ] / N[ n ];

    Trace(2, "Number = " << N[ n ] << " Sum = " << Sum[ n ] << " Average = " << Average[ n ] );
  }

  Trace( 1, "AveragePathLength = " << Average[ m_Nodes.MaxIndex() ] );
  return Average[ m_Nodes.MaxIndex() ];
}


/**
   @brief Calculate the number of possible pathes through the lattice.

   This function uses a simple DP approach to calculate the correct number of pathes.
   The lattice must be ordered in the sense that for each node, outgoing edges only
   lead to nodes with a strictly higher node number (this also indicates that the
   lattice is loop-free).
 */
long long int
CLattice::NumberOfPathes() const
{
  Trace( 1, "Lattice::NumberOfPathes" );
  Trace( 1, m_Nodes.MinIndex() << " " <<  m_Nodes.MaxIndex() );

  if ( m_Nodes.MaxIndex() < m_Nodes.MinIndex() ) {
    Trace( 0, "Number of pahtes = 0 - empty lattice" );
    return 0;
  }

  CArray< long long int > N;
  N.SetBoundaries( m_Nodes.MinIndex(), m_Nodes.MaxIndex() );
  N.SetAll( 0 );
  N[ m_Nodes.MinIndex() ] = 1;

  CLatticeNode *pNode = NULL;
  CLatticeEdge *pEdge = NULL;
  int PrevNodeNumber;

  for (int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++) {
    pNode = m_Nodes[ n ];
    Trace( 2, "Number of Node = " << pNode->Number() );
    const CArray< CLatticeEdge * > &InEdges = pNode->InEdges();
    //Trace( 1, "Number of InEdges = " << InEdges.Size() );
    for ( int i = InEdges.MinIndex(); i <= InEdges.MaxIndex(); i++ ) {
      pEdge = InEdges[ i ];
      //Trace( 1, "OutEdge " << o << ": " << pEdge->Word() ) ;
      PrevNodeNumber = pEdge->FromNode()->Number();
      N[ n ] += N[ PrevNodeNumber ];
      Trace( 4, "in prev node: " << N[ PrevNodeNumber ] << "  sum = " << N[ n ] );
    }
    Trace( 2, "Number of Pathes = " << N[ n ] );
  }

  Trace( 1, "NumberOfPathes = " << N[ m_Nodes.MaxIndex() ] );
  return N[ m_Nodes.MaxIndex() ];
}



int
CLattice::ShortestPath( CArray< CLatticeEdge * > &ShortestPath ) const
{
  Trace( 1, "Lattice::ShortestPathes" );
  Trace( 3, m_Nodes.MinIndex() << " " <<  m_Nodes.MaxIndex() );

  if ( m_Nodes.MaxIndex() < m_Nodes.MinIndex() ) {
    Trace( 1, "MinPathLength = 0 - empty lattice" );
    return 0;
  }

  CArray< int > MinLength;
  MinLength.SetBoundaries( m_Nodes.MinIndex(), m_Nodes.MaxIndex() );
  MinLength.SetAll( 1000000 );   // should be MAX_INT
  MinLength[ m_Nodes.MinIndex() ] = 0;

  CArray< CLatticeEdge * > BestEdges;
  BestEdges.SetBoundaries( m_Nodes.MinIndex(), m_Nodes.MaxIndex() );
  BestEdges.SetAll( NULL );   // should be MAX_INT

  CLatticeNode *pNode = NULL;
  CLatticeEdge *pEdge = NULL;
  int PrevNodeNumber;

  for (int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++) {
    pNode = m_Nodes[ n ];
    Trace( 3, "Number of Node = " << pNode->Number() );
    const CArray< CLatticeEdge * > &InEdges = pNode->InEdges();
    Trace( 3, "Number of InEdges = " << InEdges.Size() );
    for ( int i = InEdges.MinIndex(); i <= InEdges.MaxIndex(); i++ ) {
      pEdge = InEdges[ i ];
      //Trace( 1, "OutEdge " << o << ": " << pEdge->Word() ) ;
      PrevNodeNumber = pEdge->FromNode()->Number();

      if ( MinLength[ PrevNodeNumber ] + 1 <= MinLength[ n ] ) {
	MinLength[ n ] = MinLength[ PrevNodeNumber ] + 1;
	BestEdges[ n ] = pEdge;
      }

    }

  }

  Trace( 2, "MinPathLength = " << MinLength[ m_Nodes.MaxIndex() ] );

  // now, trace back to collect shortest path
  // we know already the length of that path
  int PathLength = MinLength[ m_Nodes.MaxIndex() ];
  int NodeNumber = m_Nodes.MaxIndex();



  while ( PathLength > 0 ) {
    Trace( 3, "NodeNumber = " << NodeNumber );
    pEdge = BestEdges[ NodeNumber ];
    ShortestPath[ PathLength ] = pEdge;
    NodeNumber = pEdge->FromNode()->Number();
    PathLength--;
  }

  return MinLength[ m_Nodes.MaxIndex() ];
}


int
CLattice::ShortestPathes( int N,
			  multimap< int, CArray< CLatticeEdge* > > &ShortestPathes ) const
{
  Trace( 2, "Lattice::ShortestPath" );
  Trace( 3, m_Nodes.MinIndex() << " " <<  m_Nodes.MaxIndex() );


  if ( m_Nodes.MaxIndex() < m_Nodes.MinIndex() ) {
    Trace( 1, "MinPathLength = 0 - empty lattice" );
    return 0;
  }

  CArray< multimap< int , CBackPointer * > > PartialPathes;
  PartialPathes.SetBoundaries( m_Nodes.MinIndex(), m_Nodes.MaxIndex() );

  // at the first node, cumulative length is zero
  PartialPathes[ m_Nodes.MinIndex() ].insert( pair< int, CBackPointer * >( 0, NULL ) );


  CLatticeNode *pNode = NULL;
  CLatticeEdge *pEdge = NULL;
  int PrevNodeNumber;
  int NewLength;
  //int Position;
  CBackPointer *pBackPointer;

  for (int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++) {
    pNode = m_Nodes[ n ];
    Trace( 3, "Number of Node = " << pNode->Number() );
    const CArray< CLatticeEdge * > &InEdges = pNode->InEdges();
    Trace( 3, "Number of InEdges = " << InEdges.Size() );

    for ( int i = InEdges.MinIndex(); i <= InEdges.MaxIndex(); i++ ) {
      pEdge = InEdges[ i ];
      //Trace( 1, "OutEdge " << o << ": " << pEdge->Word() ) ;
      PrevNodeNumber = pEdge->FromNode()->Number();

      multimap< int , CBackPointer * >::iterator IterBackPtr;
      multimap< int , CBackPointer * >::iterator CurrIter;
      
      for ( IterBackPtr = PartialPathes[ PrevNodeNumber ].begin(); 
	    IterBackPtr != PartialPathes[ PrevNodeNumber ].end(); 
	    ++IterBackPtr ) {
	//NewLength = PartialPathes[ PrevNodeNumber ].GetAt( k ) + 1;
	NewLength = IterBackPtr->first + 1;
	//Position = MinLength[ n ].Insert( NewLength, Inserted );
	pBackPointer = new CBackPointer( pEdge, (*IterBackPtr).second );
	CurrIter = PartialPathes[ n ].insert( pair< int, CBackPointer * >( NewLength, pBackPointer ) ); 
	cout << NewLength << " " << CurrIter->first << endl;
	
      }
      Trace( 0, "Number of partial pathes into node " << n << " = " << PartialPathes[ n ].size() );
    }
    
  }
  

  //  Trace( 2, "MinPathLength = " << MinLength[ m_Nodes.MaxIndex() ] );

  // how many shortest pathes are available
  int NumPathes;
  int NumLastNode = m_Nodes.MaxIndex();

  int CollectedPathes = PartialPathes[ NumLastNode ].size();
  if ( CollectedPathes < N ) {
    NumPathes = CollectedPathes;
    Trace( 0, "Number of shortest pathes is " << NumPathes << "; smaller than " << N );
  } else {
    Trace( 0, "Number of shortest pathes is " << CollectedPathes << "; larger than " << N );
    NumPathes = N;
  }


  // now, trace back to collect shortest pathes
  // we know already the length of the pathes

  int Length;
  multimap< int , CBackPointer * >::iterator CurrIter;
  CArray< CLatticeEdge *> Array;
  
  for ( CurrIter = PartialPathes[ NumLastNode ].begin();
	CurrIter != PartialPathes[ NumLastNode ].end();
	++CurrIter ) {

    Array.RemoveAll();

    Length = CurrIter->first;
    Trace( 0, "Length: " << Length );
    int L = Length;

    pBackPointer = CurrIter->second;
    pEdge = pBackPointer->m_pEdge;
    if ( pEdge ) {
      pEdge->Display();
      Array.SetAt( L, pEdge );
      L--;
    }

    while ( pEdge->FromNode() != StartNode() ) {
      pBackPointer = pBackPointer->m_pPrevBackPointer;
      pEdge = pBackPointer->m_pEdge;
      if ( pEdge ) {
	pEdge->Display();
	Array.SetAt( L, pEdge );
	L--;
      }
    }

    ShortestPathes.insert( pair< int, CArray< CLatticeEdge * > >( Length, Array ) );
  }

  Trace( 0, "Number of shortest pathes = " << ShortestPathes.size() );

  return NumPathes;
}




float
CLattice::BestPath( CArray< CLatticeEdge * > &ShortestPath ) const
{
  Trace( 1, "Lattice::BestPath" );
  Trace( 3, m_Nodes.MinIndex() << " " <<  m_Nodes.MaxIndex() );

  if ( m_Nodes.MaxIndex() < m_Nodes.MinIndex() ) {
    Trace( 1, "MinPathLength = 0 - empty lattice" );
    return 0;
  }

  CArray< float > BestScore;
  BestScore.SetBoundaries( m_Nodes.MinIndex(), m_Nodes.MaxIndex() );
  BestScore.SetAll( 1000000 ); 
  BestScore[ m_Nodes.MinIndex() ] = 0;

  CArray< int > MinLength;
  MinLength.SetBoundaries( m_Nodes.MinIndex(), m_Nodes.MaxIndex() );
  MinLength.SetAll( 1000000 );
  MinLength[ m_Nodes.MinIndex() ] = 0;

  CArray< CLatticeEdge * > BestEdges;
  BestEdges.SetBoundaries( m_Nodes.MinIndex(), m_Nodes.MaxIndex() );
  BestEdges.SetAll( NULL ); 

  CLatticeNode *pNode = NULL;
  CLatticeEdge *pEdge = NULL;
  int PrevNodeNumber;
  float Score;

  for (int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++) {
    pNode = m_Nodes[ n ];
    Trace( 3, "Number of Node = " << pNode->Number() );
    const CArray< CLatticeEdge * > &InEdges = pNode->InEdges();
    Trace( 3, "Number of InEdges = " << InEdges.Size() );
    for ( int i = InEdges.MinIndex(); i <= InEdges.MaxIndex(); i++ ) {
      pEdge = InEdges[ i ];
      //Trace( 1, "OutEdge " << o << ": " << pEdge->Word() ) ;
      PrevNodeNumber = pEdge->FromNode()->Number();
      Score = pEdge->TotalScore();
      Trace( 3, "Score = " << Score );
      if ( BestScore[ PrevNodeNumber ] + Score <= BestScore[ n ] ) {
	BestScore[ n ] = BestScore[ PrevNodeNumber ] + Score;
	MinLength[ n ] = MinLength[ PrevNodeNumber ] + 1;
	BestEdges[ n ] = pEdge;
      }

    }

  }

  Trace( 1, "Best Score = " << BestScore[ m_Nodes.MaxIndex() ] );
  Trace( 1, "Length     = " << MinLength[ m_Nodes.MaxIndex() ] );

  // now, trace back to collect shortest path
  // we know already the length of that path
  int PathLength = MinLength[ m_Nodes.MaxIndex() ];
  int NodeNumber = m_Nodes.MaxIndex();

  while ( PathLength > 0 ) {
    Trace( 3, "NodeNumber = " << NodeNumber );
    pEdge = BestEdges[ NodeNumber ];
    ShortestPath[ PathLength ] = pEdge;
    NodeNumber = pEdge->FromNode()->Number();
    PathLength--;
  }

  return BestScore[ m_Nodes.MaxIndex() ];
}



int
CLattice::BestPathes( int N,
		      multimap< float, CArray< CLatticeEdge* > > &BestPathes ) const
{
  Trace( 1, "Lattice::BestPathes" );
  Trace( 3, m_Nodes.MinIndex() << " " <<  m_Nodes.MaxIndex() );


  if ( m_Nodes.MaxIndex() < m_Nodes.MinIndex() ) {
    Trace( 1, "MinPathLength = 0 - empty lattice" );
    return 0;
  }

  CArray< multimap< float , CBackPointer * > > PartialPathes;
  PartialPathes.SetBoundaries( m_Nodes.MinIndex(), m_Nodes.MaxIndex() );

  // at the first node, cumulative length is zero
  PartialPathes[ m_Nodes.MinIndex() ].insert( pair< float, CBackPointer * >( 0, NULL ) );


  CLatticeNode *pNode = NULL;
  CLatticeEdge *pEdge = NULL;
  int PrevNodeNumber;
  float Score, NewScore;
  CBackPointer *pBackPointer;
  int NumBest;

  for (int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++) {
    pNode = m_Nodes[ n ];
    Trace( 3, "Number of Node = " << pNode->Number() );
    const CArray< CLatticeEdge * > &InEdges = pNode->InEdges();
    Trace( 3, "Number of InEdges = " << InEdges.Size() );

    for ( int i = InEdges.MinIndex(); i <= InEdges.MaxIndex(); i++ ) {
      pEdge = InEdges[ i ];
      //Trace( 1, "OutEdge " << o << ": " << pEdge->Word() ) ;
      PrevNodeNumber = pEdge->FromNode()->Number();

      multimap< float , CBackPointer * >::iterator IterBackPtr;
      multimap< float , CBackPointer * >::iterator CurrIter;
      int NumBest = 1;
      
      for ( IterBackPtr = PartialPathes[ PrevNodeNumber ].begin(); 
	    (IterBackPtr != PartialPathes[ PrevNodeNumber ].end()) && ( NumBest <= N ); 
	    ++IterBackPtr, NumBest++ ) {
	//NewLength = PartialPathes[ PrevNodeNumber ].GetAt( k ) + 1;

	Score = pEdge->Score();
	NewScore = IterBackPtr->first + Score;
	pBackPointer = new CBackPointer( pEdge, (*IterBackPtr).second );
	CurrIter = PartialPathes[ n ].insert( pair< float, CBackPointer * >( NewScore, pBackPointer ) ); 
	//cout << NewScore << " " << CurrIter->first << endl;
      }

      Trace( 1, "NewScore = " << NewScore );

    }

  }

  //Trace( 1, "Best Score = " << BestScore[ m_Nodes.MaxIndex() ] );
  //Trace( 1, "Length     = " << MinLength[ m_Nodes.MaxIndex() ] );

  // how many shortest pathes are available
  int NumPathes;
  int NumLastNode = m_Nodes.MaxIndex();
  
  int CollectedPathes = PartialPathes[ NumLastNode ].size();
  if ( CollectedPathes < N ) {
    NumPathes = CollectedPathes;
    Trace( 2, "Number of best pathes is " << NumPathes << "; smaller than " << N );
  } else {
    Trace( 2, "Number of best pathes is " << CollectedPathes << "; larger than " << N );
    NumPathes = N;
  }
  

  // now, trace back to collect best path
  // we know already the length of that path
  //int Length;
  multimap< float , CBackPointer * >::iterator CurrIter;
  CArray< CLatticeEdge *> Array;
  NumBest = 1;
  
  for ( CurrIter = PartialPathes[ NumLastNode ].begin();
	( CurrIter != PartialPathes[ NumLastNode ].end() ) && ( NumBest <= NumPathes );
	++CurrIter, NumBest++ ) {

    Array.RemoveAll();

    Score = CurrIter->first;
    Trace( 1, "Total Score: " << Score );

    pBackPointer = CurrIter->second;
    pEdge = pBackPointer->m_pEdge;
    if ( pEdge ) {
      //pEdge->Display();
      Array.Insert( 0, pEdge );
    }

    while ( pEdge->FromNode() != StartNode() ) {
      pBackPointer = pBackPointer->m_pPrevBackPointer;
      pEdge = pBackPointer->m_pEdge;
      if ( pEdge ) {
	//pEdge->Display();
	Array.Insert( 0, pEdge );
      }
    }

    BestPathes.insert( pair< float, CArray< CLatticeEdge * > >( Score, Array ) );
  }

  Trace( 1, "Number of best pathes = " << BestPathes.size() );

  return NumPathes;

}

