#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#define NO_TRACE 1
#include <Trace.hh>

#include <LatticeEdgeMS.hh>

// default LM scale: only acoustic score
//double CLatticeEdgeMS::m_LMScale = 1.0;
//long int CLatticeEdgeMS::s_NumberOfObjects = 0;


void 
CLatticeEdgeMS::CopyFrom( const CLatticeEdgeMS *pEdge )
{
  // copy stuff from base class
  ((CLatticeEdgeTrans *)this)->CopyFrom( pEdge );

  // copy own stuff
  m_Scores = pEdge->m_Scores;
}



void 
CLatticeEdgeMS::CopyTo( CLatticeEdgeMS *pEdge ) const
{
  // copy stuff from base class
  // ((CLatticeEdgeTrans *)this)->CopyTo( pEdge ); // removed by apd because CLatticeEdgeTrans->CopyTo uses CLatticeEdge->CopyTo
  
  // added by apd
  pEdge->m_Word = m_Word;
  pEdge->m_Score = m_Score;
  pEdge->m_Startframe = m_Startframe;
  pEdge->m_Endframe = m_Endframe;
  //pNewEdge->m_pWordInfo = new CStatTransInfo( m_pWordInfo );
  //pNewEdge->m_pWordInfo = new CWordInfo( m_pWordInfo );
  pEdge->m_Prosody = m_Prosody;
  
  pEdge->m_ErrorScore = m_ErrorScore;
  pEdge->m_ConcatScore = m_ConcatScore;
  pEdge->m_WordIndex = m_WordIndex;
  //pNewEdge->m_pGeneratingEdges = NULL;  ... already in constructor
  pEdge->m_pTransducer = m_pTransducer;
  pEdge->m_pTransducerItem = m_pTransducerItem;
  pEdge->m_MatchFlag = m_MatchFlag;
  
  
  
  // copy own stuff
  pEdge->m_Scores = m_Scores;
}




CLatticeEdge *
CLatticeEdgeMS::Clone() const
{
  Trace( 4, "CLatticeEdgeMS::Clone()" );

  CLatticeEdgeMS *pNewEdge = new CLatticeEdgeMS( m_pFromNode, m_pToNode );

  CopyTo( pNewEdge );

  /*
  pNewEdge->m_Word = m_Word;
  pNewEdge->m_Scores = m_Scores;
  pNewEdge->m_Score = m_Score;
  pNewEdge->m_Startframe = m_Startframe;
  pNewEdge->m_Endframe = m_Endframe;
  //pNewEdge->m_pWordInfo = new CWordInfo( m_pWordInfo );
  pNewEdge->m_Prosody = m_Prosody;

  pNewEdge->m_ErrorScore = m_ErrorScore;
  pNewEdge->m_ConcatScore = m_ConcatScore;
  pNewEdge->m_WordIndex = m_WordIndex;
  //pNewEdge->m_pGeneratingEdges = NULL;  ... already in constructor
  pNewEdge->m_pTransducer = m_pTransducer;
  pNewEdge->m_pTransducerItem = m_pTransducerItem;
  pNewEdge->m_MatchFlag = m_MatchFlag;
	*/
  return pNewEdge;
}


CLatticeEdge *
CLatticeEdgeMS::Concatenate( const CLatticeEdge *pEdge )
{
  Trace( 4, "CLatticeEdgeMS::Concatenate()" );

  CLatticeEdgeMS *pNewEdge = NULL;
  CLatticeEdgeMS *pOtherEdge = (CLatticeEdgeMS *)pEdge;

  if ( TraceLevel >= 4 ) {
    Display();
    pOtherEdge->Display();
  }

  // have the two edges one node in common
  if ( m_pToNode == pOtherEdge->m_pFromNode ) {        // concat to the right
    Trace( 4, "Span over right edge - left edge is dominant" );
    pNewEdge = new CLatticeEdgeMS( m_pFromNode, pOtherEdge->m_pToNode );
    pNewEdge->m_Word = m_Word;
    pNewEdge->m_Startframe = m_Startframe;
    pNewEdge->m_Endframe = pOtherEdge->m_Endframe;
    pNewEdge->m_Score = m_Score + pOtherEdge->m_Score;
    pNewEdge->m_ErrorScore = m_ErrorScore + pOtherEdge->m_ErrorScore;
    pNewEdge->m_ConcatScore = m_ConcatScore + pOtherEdge->m_ConcatScore;
    //pNewEdge->m_pWordInfo = new CStatTransInfo( m_pWordInfo );
    pNewEdge->m_pWordInfo = new CWordInfo( m_pWordInfo );
    //pNewEdge->m_pWordInfo->SetNumberOfEdges( m_pWordInfo->NumberOfEdges()
    //					  + pOtherEdge->m_pWordInfo->NumberOfEdges() );
    pNewEdge->m_Prosody = m_Prosody;
    // make new array for generating edges and copy from both edges
    //pNewEdge->m_pGeneratingEdges = ...
    pNewEdge->m_pTransducer = m_pTransducer;
    pNewEdge->m_pTransducerItem = m_pTransducerItem;
  } else if ( m_pFromNode == pOtherEdge->m_pToNode ) {      // concat to the left
    Trace( 4, "Span over left edge - right is dominant" );
    pNewEdge = new CLatticeEdgeMS( pOtherEdge->m_pFromNode, m_pToNode );
    pNewEdge->m_Word = m_Word;
    pNewEdge->m_Startframe = pOtherEdge->m_Startframe;
    pNewEdge->m_Endframe = m_Endframe;
    pNewEdge->m_Score = m_Score + pOtherEdge->m_Score;
    pNewEdge->m_ErrorScore = m_ErrorScore + pOtherEdge->m_ErrorScore;
    pNewEdge->m_ConcatScore = m_ConcatScore + pOtherEdge->m_ConcatScore;
    //pNewEdge->m_pWordInfo = new CStatTransInfo( m_pWordInfo );
    pNewEdge->m_pWordInfo = new CWordInfo( m_pWordInfo );
    //pNewEdge->m_pWordInfo->SetNumberOfEdges( m_pWordInfo->NumberOfEdges() 
    //					  + pOtherEdge->m_pWordInfo->NumberOfEdges() );
    pNewEdge->m_Prosody = m_Prosody;
    // make new array for generating edges and copy from both edges
    //pNewEdge->m_pGeneratingEdges = ...
    pNewEdge->m_pTransducer = m_pTransducer;
    pNewEdge->m_pTransducerItem = m_pTransducerItem;
  } else { 
    Trace( 4, "Edges not connected" );
  }


  return pNewEdge;

}

// einstelliger Ausgabeoperator
ostream& 
CLatticeEdgeMS::operator<<( ostream& os ) const
{
  Trace( 0, "CLatticeEdgeMS::operator << ()" );

  int FromNodeNumber, ToNodeNumber;

  if (m_pFromNode)
    FromNodeNumber = m_pFromNode->Number();
  else
    FromNodeNumber = -1;

  if (m_pToNode)
    ToNodeNumber = m_pToNode->Number();
  else
    ToNodeNumber = -1;

  //  cout<<"In here"<<m_Score<<endl;
  return ( os << "edge " //<< this 
		   // << " (" << m_FromNode << '-' << m_ToNode << " | " 
		   << "(" << FromNodeNumber << "-" << ToNodeNumber
		   << ") \"" << m_Word << " " << m_Score << " "
		   << m_Startframe << " " << m_Endframe
		   << "\"" << endl );
}

void
CLatticeEdgeMS::Display() const
{
  Trace( 5, "CLatticeEdgeMS::Display" );

  int FromNodeNumber, ToNodeNumber;

  if (m_pFromNode)
    FromNodeNumber = m_pFromNode->Number();
  else
    FromNodeNumber = -1;

  if (m_pToNode)
    ToNodeNumber = m_pToNode->Number();
  else
    ToNodeNumber = -1;

  cout << "edge " 
       << "(" << FromNodeNumber << "-" << ToNodeNumber
       << ") Word=" << m_Word << " Score=" << m_Score << " StartFrame="
       <<  m_Startframe << " EndFrame=" << m_Endframe << " ";

  CString Prosody;
  m_Prosody.ToString( Prosody );
  //cout << Prosody << " ";
  //m_pWordInfo->WriteTo( cout );
  cout << *m_pWordInfo;
  cout << " " << m_Scores;
  cout << endl << flush;

}



bool
CLatticeEdgeMS::operator==( CLatticeEdge &Edge )
{
  Trace( 0, "CLatticeEdgeMS::operator==" );
  /*
  return ( ( FromNode() == Edge.FromNode() ) &&
	   ( ToNode() == Edge.ToNode() ) &&
	   ( m_Word == ((CLatticeEdgeMS &)Edge).m_Word ) &&
	   ( m_pWordInfo->TargetString() 
	     == ((CLatticeEdgeMS &)Edge).m_pWordInfo->TargetString() ) );
  */

  return (
	  ( FromNode() == Edge.FromNode() ) &&
	  ( ToNode() == Edge.ToNode() ) &&
	  ( m_Word == ((CLatticeEdgeMS &)Edge).m_Word ) &&
	  ( m_pTransducer == ((CLatticeEdgeMS &)Edge).m_pTransducer ) &&
	  ( m_pTransducerItem == ((CLatticeEdgeMS &)Edge).m_pTransducerItem )
	  );
}


bool
CLatticeEdgeMS::operator<( CLatticeEdge &Edge )
{
  Trace( 1, "CLatticeEdgeMS::operator<" );

  return ( (*this == Edge) && (m_Score < ((CLatticeEdgeMS &)Edge).m_Score) );
}




CNumericalArray< float > & 
CLatticeEdgeMS::Scores()
{
  return m_Scores;
}

