#include <config.h>

#ifndef __LatticeEdgeMS__
#define __LatticeEdgeMS__

#include <sstream>

#include <LatticeEdgeTrans.hh>


// ----------------------------------------------------------------
// CLatticeEdgeMS
// ----------------------------------------------------------------


class CLatticeEdgeMS : public CLatticeEdgeTrans
{
public:
  CLatticeEdgeMS();
  CLatticeEdgeMS( CLatticeNode *FromNode, CLatticeNode *ToNode );

  virtual ~CLatticeEdgeMS();
  
  // create a new edge which is a copy of the current edge
  virtual CLatticeEdge * Clone() const;

  // copy the content of this edge to given edge
  void CopyFrom( const CLatticeEdgeMS *pEdge );

  // copy the content of given edge to this edge
  void CopyTo( CLatticeEdgeMS *pEdge ) const;

  // create a new edge which spans this edge and pEdge
  // it is tested if the to edges are adjacent
  virtual CLatticeEdge * Concatenate( const CLatticeEdge *pEdge );

  //virtual void Init( CLatticeNode *f, CLatticeNode *t );
  //virtual void Init( CLatticeNode *f, CLatticeNode *t, const char *c );
  //void Init( CLatticeNode *f, CLatticeNode *t, int FromNode, int ToNode,
  //	     CString Word, double Score, int FromTime, int ToTime, char *Buffer );

  // complete score
  virtual float TotalScore() const;

  // individual scores
  CNumericalArray< float > & Scores();


  virtual std::ostream& operator<<( std::ostream& os ) const;
  virtual void WriteTo( std::ostream& os ) const;

  virtual void Display() const;

  virtual bool operator == ( CLatticeEdge &Edge );
  virtual bool operator < ( CLatticeEdge &Edge );

  //static long int s_NumberOfObjects;

protected:
  // we can deal with any number of scores
  // can be used for anything
  CNumericalArray< float > m_Scores;


};




// ----------------------------------------------------------------
// CLatticeEdgeMS - Inlines
// ----------------------------------------------------------------


inline
CLatticeEdgeMS::CLatticeEdgeMS()
{
  //  Trace( 2, "CLatticeEdgeMS::CLatticeEdgeMS" );
  //s_NumberOfObjects++;
}


inline
CLatticeEdgeMS::CLatticeEdgeMS( CLatticeNode *FromNode, CLatticeNode *ToNode )
: CLatticeEdgeTrans( FromNode, ToNode )
{
  //s_NumberOfObjects++;
}


inline
CLatticeEdgeMS::~CLatticeEdgeMS()
{
}


inline void
CLatticeEdgeMS::WriteTo( ostream& os ) const
{
  os << m_Word << "\t" << m_Score << 
     "\t" << m_Startframe << "\t" << m_Endframe << "\t";
  
  //m_Prosody.WriteTo(os);
  
  if (m_pWordInfo != NULL)
    //m_pWordInfo->WriteTo(os);
    os << m_pWordInfo;
}



// complete score
inline float
CLatticeEdgeMS::TotalScore() const
{ 
//     Trace( 2, "CLatticeEdgeMS::TotalScore = " 
// 	 << ((1.0  - m_LMScale) * m_Score) << " "
// 	 << (m_LMScale * m_pWordInfo->Score() ) );
  return ( ( 1.0 - m_LMScale ) * m_Score 
	   + m_LMScale * m_pWordInfo->Score() );
  //return m_pWordInfo->Score();
}



#endif


