#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#define NO_TRACE 1
#include <Trace.hh>

#include <LatticeEdgeTrans.hh>

// default LM scale: only acoustic score
float CLatticeEdgeTrans::m_LMScale = 1.0;
//long int CLatticeEdgeTrans::s_NumberOfObjects = 0;

void
CLatticeEdgeTrans::Init( CLatticeNode *f, CLatticeNode *t, const char *Buffer )
{
  Trace( 3, "CLatticeEdgeTrans::Init " << Buffer );
  m_pFromNode = f;
  m_pToNode = t;

  char WordBuffer[ MaxLineLength ];
  WordBuffer[ 0 ] = '\0';
  unsigned long From, To;
  unsigned long fframe, tframe;
  float sc;

  sscanf( Buffer,"%lu%lu%s%f%lu%lu", &From, &To, WordBuffer, &sc, &fframe, &tframe );
  Trace( 3, "From: " << From << " To: " << To << "  " << WordBuffer << " " 
	 << sc << " " << fframe << " " << tframe );
  //cout<<"From: " << From << " To: " << To << "  " << WordBuffer << " " 
  //    << sc << " " << fframe << " " << tframe<<endl;

  m_Word = CString( WordBuffer );
  m_Score = sc;

  m_Startframe = fframe;
  m_Endframe = tframe;

  const char *Prosody = strstr( Buffer, "(PR " );
  Trace( 3, Prosody );
  
  if (Prosody != NULL)
    m_Prosody.ReadFromBuffer( Prosody );

  const char *StatTrans = strstr( Buffer, "(ST {");
  
  if (StatTrans != NULL) {
    Trace( 3, "StatTrans: " << StatTrans );
    Trace( 3, "StatTrans: " << StatTrans + 6);
    if (m_pWordInfo == NULL)
      //m_pWordInfo = new CStatTransInfo;
      m_pWordInfo = new CWordInfo;

    //m_pWordInfo->ReadFromBuffer( StatTrans );
    istringstream Stream( StatTrans + 6 );
    Stream >> *m_pWordInfo;
    Trace( 3, "StatTrans: " << *m_pWordInfo );
  }

  m_ErrorScore = 0.0;
  m_ConcatScore = 0.0;

}



void
CLatticeEdgeTrans::Init( CLatticeNode *f, CLatticeNode *t, 
			 int FromNode, int ToNode,
			 CString Word,
			 float Score,
			 int FromTime, int ToTime,
			 const char *Buffer )
{
  Trace( 3, "CLatticeEdgeTrans::Init " << Buffer );
  m_pFromNode = f;
  m_pToNode = t;

  //char WordBuffer[ MaxLineLength ];
  //WordBuffer[ 0 ] = '\0';
  //unsigned long From, To;
  //unsigned long fframe, tframe;
  //double sc;

  //sscanf( Buffer,"%lu%lu%s%lf%lu%lu", &From, &To, WordBuffer, &sc, &fframe, &tframe );

  m_Word = Word;
  m_Score = Score;


  //cout<<"Init:"<<m_Word<<" Score:"<<m_Score<<endl;
  m_Startframe = FromTime;
  m_Endframe = ToTime;

  const char *Prosody = strstr( Buffer, "(PR " );
  Trace( 3, Prosody );
  
  if (Prosody != NULL)
    m_Prosody.ReadFromBuffer( Prosody );

  const char *StatTrans = strstr( Buffer, "(ST {");
  
  if (StatTrans != NULL) {
    Trace( 3, "StatTrans: " << StatTrans );
    Trace( 3, "StatTrans: " << StatTrans + 6);
    if (m_pWordInfo == NULL)
      //m_pWordInfo = new CStatTransInfo;
      m_pWordInfo = new CWordInfo;

    //cout<<"ReadingFromBuffer:"<<endl;
    //m_pWordInfo->ReadFromBuffer( StatTrans );
    istringstream Stream( StatTrans + 6 );
    Stream >> *m_pWordInfo;
    Trace( 3, "StatTrans: " << *m_pWordInfo );
  }

  m_ErrorScore = 0.0;
  m_ConcatScore = 0.0;

}



CLatticeEdge *
CLatticeEdgeTrans::Clone() const
{
  Trace( 4, "CLatticeEdgeTrans::Clone()" );

  CLatticeEdgeTrans *pNewEdge = new CLatticeEdgeTrans( m_pFromNode, m_pToNode );

  pNewEdge->m_Word = m_Word;
  pNewEdge->m_Score = m_Score;
  pNewEdge->m_Startframe = m_Startframe;
  pNewEdge->m_Endframe = m_Endframe;
  //pNewEdge->m_pWordInfo = new CStatTransInfo( m_pWordInfo );
  //pNewEdge->m_pWordInfo = new CWordInfo( m_pWordInfo );
  pNewEdge->m_Prosody = m_Prosody;

  pNewEdge->m_ErrorScore = m_ErrorScore;
  pNewEdge->m_ConcatScore = m_ConcatScore;
  pNewEdge->m_WordIndex = m_WordIndex;
  //pNewEdge->m_pGeneratingEdges = NULL;  ... already in constructor
  pNewEdge->m_pTransducer = m_pTransducer;
  pNewEdge->m_pTransducerItem = m_pTransducerItem;
  pNewEdge->m_MatchFlag = m_MatchFlag;

  return pNewEdge;
}


CLatticeEdge *
CLatticeEdgeTrans::Concatenate( const CLatticeEdge *pEdge )
{
  Trace( 4, "CLatticeEdgeTrans::Concatenate()" );

  CLatticeEdgeTrans *pNewEdge = NULL;
  CLatticeEdgeTrans *pOtherEdge = (CLatticeEdgeTrans *)pEdge;

  if ( TraceLevel >= 4 ) {
    Display();
    pOtherEdge->Display();
  }

  // have the two edges one node in common
  if ( m_pToNode == pOtherEdge->m_pFromNode ) {        // concat to the right
    Trace( 4, "Span over right edge - left edge is dominant" );
    pNewEdge = new CLatticeEdgeTrans( m_pFromNode, pOtherEdge->m_pToNode );
    pNewEdge->m_Word = m_Word;
    pNewEdge->m_Startframe = m_Startframe;
    pNewEdge->m_Endframe = pOtherEdge->m_Endframe;
    pNewEdge->m_Score = m_Score + pOtherEdge->m_Score;
    pNewEdge->m_ErrorScore = m_ErrorScore + pOtherEdge->m_ErrorScore;
    pNewEdge->m_ConcatScore = m_ConcatScore + pOtherEdge->m_ConcatScore;
    //pNewEdge->m_pWordInfo = new CStatTransInfo( m_pWordInfo );
    pNewEdge->m_pWordInfo = new CWordInfo( m_pWordInfo );
    //pNewEdge->m_pWordInfo->SetNumberOfEdges( m_pWordInfo->NumberOfEdges()
    //					  + pOtherEdge->m_pWordInfo->NumberOfEdges() );
    pNewEdge->m_Prosody = m_Prosody;
    // make new array for generating edges and copy from both edges
    //pNewEdge->m_pGeneratingEdges = ...
    pNewEdge->m_pTransducer = m_pTransducer;
    pNewEdge->m_pTransducerItem = m_pTransducerItem;
  } else if ( m_pFromNode == pOtherEdge->m_pToNode ) {      // concat to the left
    Trace( 4, "Span over left edge - right is dominant" );
    pNewEdge = new CLatticeEdgeTrans( pOtherEdge->m_pFromNode, m_pToNode );
    pNewEdge->m_Word = m_Word;
    pNewEdge->m_Startframe = pOtherEdge->m_Startframe;
    pNewEdge->m_Endframe = m_Endframe;
    pNewEdge->m_Score = m_Score + pOtherEdge->m_Score;
    pNewEdge->m_ErrorScore = m_ErrorScore + pOtherEdge->m_ErrorScore;
    pNewEdge->m_ConcatScore = m_ConcatScore + pOtherEdge->m_ConcatScore;
    //pNewEdge->m_pWordInfo = new CStatTransInfo( m_pWordInfo );
    pNewEdge->m_pWordInfo = new CWordInfo( m_pWordInfo );
    //pNewEdge->m_pWordInfo->SetNumberOfEdges( m_pWordInfo->NumberOfEdges() 
    //					  + pOtherEdge->m_pWordInfo->NumberOfEdges() );
    pNewEdge->m_Prosody = m_Prosody;
    // make new array for generating edges and copy from both edges
    //pNewEdge->m_pGeneratingEdges = ...
    pNewEdge->m_pTransducer = m_pTransducer;
    pNewEdge->m_pTransducerItem = m_pTransducerItem;
  } else { 
    Trace( 4, "Edges not connected" );
  }


  return pNewEdge;

}

// einstelliger Ausgabeoperator
ostream& 
CLatticeEdgeTrans::operator<<( ostream& os ) const
{
  Trace( 5, "CLatticeEdgeTrans::operator << ()" );

  int FromNodeNumber, ToNodeNumber;

  if (m_pFromNode)
    FromNodeNumber = m_pFromNode->Number();
  else
    FromNodeNumber = -1;

  if (m_pToNode)
    ToNodeNumber = m_pToNode->Number();
  else
    ToNodeNumber = -1;

  cout<<"In here"<<m_Score<<endl;
  return (os << "edge " //<< this 
	  // << " (" << m_FromNode << '-' << m_ToNode << " | " 
	  << "(" << FromNodeNumber << "-" << ToNodeNumber
	  << ") \"" << m_Word << " " << m_Score << " " <<
	  m_Startframe << " " << m_Endframe
	  << "\"" ) << endl;
}

void
CLatticeEdgeTrans::Display() const
{
  Trace( 5, "CLatticeEdgeTrans::Display" );

  int FromNodeNumber, ToNodeNumber;

  if (m_pFromNode)
    FromNodeNumber = m_pFromNode->Number();
  else
    FromNodeNumber = -1;

  if (m_pToNode)
    ToNodeNumber = m_pToNode->Number();
  else
    ToNodeNumber = -1;

  cout << "edge " 
       << "(" << FromNodeNumber << "-" << ToNodeNumber
       << ") Word=" << m_Word << " Score=" << m_Score << " StartFrame="
       <<  m_Startframe << " EndFrame=" << m_Endframe << " ";

  CString Prosody;
  m_Prosody.ToString( Prosody );
  //cout << Prosody << " ";
  //m_pWordInfo->WriteTo( cout );
  cout << *m_pWordInfo;
  cout << endl << flush;

}



bool
CLatticeEdgeTrans::operator==( CLatticeEdge &Edge )
{
  Trace( 0, "CLatticeEdgeTrans::operator==" );
  /*
  return ( ( FromNode() == Edge.FromNode() ) &&
	   ( ToNode() == Edge.ToNode() ) &&
	   ( m_Word == ((CLatticeEdgeTrans &)Edge).m_Word ) &&
	   ( m_pWordInfo->TargetString() 
	     == ((CLatticeEdgeTrans &)Edge).m_pWordInfo->TargetString() ) );
  */

  return (
	  ( FromNode() == Edge.FromNode() ) &&
	  ( ToNode() == Edge.ToNode() ) &&
	  ( m_Word == ((CLatticeEdgeTrans &)Edge).m_Word ) &&
	  ( m_pTransducer == ((CLatticeEdgeTrans &)Edge).m_pTransducer ) &&
	  ( m_pTransducerItem == ((CLatticeEdgeTrans &)Edge).m_pTransducerItem )
	  );
}


bool
CLatticeEdgeTrans::operator<( CLatticeEdge &Edge )
{
  Trace( 1, "CLatticeEdgeTrans::operator<" );

  return ( (*this == Edge) && (m_Score < ((CLatticeEdgeTrans &)Edge).m_Score) );
}



