#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#define NO_TRACE 1
#include <Trace.hh>

#include <LatticeLM.hh>

long int CBackPointerLM::s_NumberOfObjects;



CBackPointerLM::CBackPointerLM( CLMState LMState, CLatticeEdge *pEdge, 
				CBackPointerLM *pPrevBackPointer )
{
  m_LMStateWord = LMState;
  m_pEdge = pEdge;
  m_pPrevBackPointer = pPrevBackPointer;

  m_SumACScore = 0.0;
  m_SumLMScore = 0.0;
  m_SumTRScore = 0.0;
  m_SumERScore = 0.0;
  m_SumCCScore = 0.0;

  s_NumberOfObjects++;
}


CBackPointerLM::~CBackPointerLM()
{
  s_NumberOfObjects--;
}

 void 
CBackPointerLM::GetScores( float &SumACScore, float &SumLMScore, 
			   float &SumTRScore, float &SumERScore,
			   float &SumCCScore )
{
  SumACScore = m_SumACScore;
  SumLMScore = m_SumLMScore;
  SumTRScore = m_SumTRScore;
  SumERScore = m_SumERScore;
  SumCCScore = m_SumCCScore;
}


 void 
CBackPointerLM::SetScores( float SumACScore, float SumLMScore, 
			   float SumTRScore, float SumERScore,
			   float SumCCScore )
{
  m_SumACScore = SumACScore;
  m_SumLMScore = SumLMScore;
  m_SumTRScore = SumTRScore;
  m_SumERScore = SumERScore;
  m_SumCCScore = SumCCScore;
}


 void 
CBackPointerLM::DisplayScores()
{
  cout << "AC = " << m_SumACScore 
       << " LM = " << m_SumLMScore
       << " TR = " << m_SumTRScore
       << " ER = " << m_SumERScore
       << " CC = " << m_SumCCScore << endl;
}

CLatticeLM::CLatticeLM( CLatticeEdge * pEdgePrototype, CLM * pLM )
  : CLattice( pEdgePrototype )
{
  // Trace( 2, "CLatticeLM::CLatticeLM" );

  m_pLM = pLM;

  // scaling factor for different kinds of scores
  m_ACScale = 0.0;
  m_LMScale = 0.0;
  m_TRScale = 0.0;
  m_ERScale = 0.0;
  m_CCScale = 0.0;

  m_BeamFactor = 0.0;

  m_UseSource = FALSE;
}


 void 
CLatticeLM::SetScalingFactors( float ACScale, float LMScale,
			       float TRScale, float ERScale,
			       float CCScale )
{
  m_ACScale = ACScale; 
  m_LMScale = LMScale;
  m_TRScale = TRScale;
  m_ERScale = ERScale;
  m_CCScale = CCScale;
}



 void 
CLatticeLM::SetBeamFactor( float BeamFactor )
{
  m_BeamFactor = BeamFactor;
}



 void
CLatticeLM::LocalReordering()
{

  // Trace( 0, "CLatticeLM::LocalReordering" );

  Display();

  CLatticeNode *pNode = NULL;
  CLatticeEdgeTrans *pInEdge = NULL;
  CLatticeEdgeTrans *pOutEdge = NULL;

  // run over all nodes
  int n = m_Nodes.MinIndex();
  n++;
  for (; n < m_Nodes.MaxIndex(); n++) {
    pNode = m_Nodes[ n ];
    // Trace( 0, endl << "Number of Node = " << pNode->Number() );

    // run over all incoming edges and all outgoing edges
    const CArray< CLatticeEdge * > &InEdges = pNode->InEdges();
    const CArray< CLatticeEdge * > &OutEdges = pNode->OutEdges();
    // Trace( 0, "Number of InEdges = " << InEdges.Size() );
    // Trace( 0, "Number of InEdges = " << InEdges.Size() );

    for ( int i = InEdges.MinIndex() + 1; i <= InEdges.MaxIndex(); i++ ) {

      pInEdge = (CLatticeEdgeTrans *)InEdges[ i ];
      //Trace( 0, "InEdge " << i << ": " << pInEdge->Word() ) ;
	  pInEdge->Display();

	  for ( int o = OutEdges.MinIndex() + 1; o <= OutEdges.MaxIndex(); o++ ) {

		pOutEdge = (CLatticeEdgeTrans *)OutEdges[ o ];
		//Trace( 0, "OutEdge " << o << ": " << pOutEdge->Word() ) ;
		pOutEdge->Display();
	  
		CLatticeEdgeTrans *pConcatenedEdge = (CLatticeEdgeTrans *)pOutEdge->Concatenate( pInEdge );
		pConcatenedEdge->Display();
		delete pConcatenedEdge;

	  } // loop over in-edges

	}  // loop over out-edges

  }  // loop over nodes

}



 int
CLatticeLM::BestPathes( int N,
			multimap< float, CArray< CLatticeEdge* > > &BestPathes ) const
{
  // Trace( 1, "LatticeLM::BestPathes" );
  if ( !m_pLM )
    // Trace( 1, "No LM available!" );

  // Trace( 2, m_Nodes.MinIndex() << " " <<  m_Nodes.MaxIndex() );


  if ( m_Nodes.MaxIndex() < m_Nodes.MinIndex() ) {
    // Trace( 1, "MinPathLength = 0 - empty lattice" );
    return 0;
  }

  // for statistics of created and pruned hyps
  long PrunedHyps = 0;
  long GeneratedHyps = 0;

  // LM state
  CLMSentenceState *pLMSentenceStateWord = NULL;
  CLMState LMStateWord = (CLMState) 0;
  CLMState NewLMStateWord = (CLMState) 0;
  
  if (m_pLM ) {
    pLMSentenceStateWord = m_pLM->NewSentence();
    LMStateWord = pLMSentenceStateWord->BeginOfSentenceState();
  }


  // array to collect words attached to one edge for LM-scoring
  CArray< CString > Words;
  CArray< LMWordIndex > WordIndices;

  // Trace( 2, "LM-State for Sentence Beginning: " << LMStateWord );

  CBackPointerLM *pBackPointer = new CBackPointerLM( LMStateWord, NULL, NULL );
  CBackPointerLM *pNewBackPointer;

  // for each node we have a list of the scores for the partial pathes
  // running into that node, together with backpointers
  CArray< map< CLMState, multimap< float , CBackPointerLM * > > > AllNBestListsAllNodes;
  AllNBestListsAllNodes.SetBoundaries( m_Nodes.MinIndex(), m_Nodes.MaxIndex() );

  // at the first node, cumulative score is zero
  multimap< float , CBackPointerLM * > NBestListFirstNode;
  NBestListFirstNode.insert( pair< float, CBackPointerLM * >( 0, pBackPointer ) );
  AllNBestListsAllNodes[ m_Nodes.MinIndex() ].insert( pair< CLMState, multimap< float, CBackPointerLM * > >( LMStateWord, NBestListFirstNode ) );


  CLatticeNode *pNode = NULL;
  CLatticeEdge *pEdge = NULL;
  int PrevNodeNumber;
  float 
    AccumulatedScore = 0.0, 
    AcousticScore = 0.0,
    TranslationScore = 0.0,
    ErrorScore = 0.0,
    LMScoreWord = 0.0,
    ConcatScore = 0.0,
    NewScore = 0.0,
    SumACScore = 0.0, 
    SumLMScore = 0.0,
    SumTRScore = 0.0,
    SumERScore = 0.0,
    SumCCScore = 0.0;
  int NumBest;

  string word;

  CString Word;

  // Trace( 0, "Beam factor = " << m_BeamFactor );

  // run over all nodes
  for (int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++) {
    pNode = m_Nodes[ n ];
    // Trace( 2, endl << "Number of Node = " << pNode->Number() );


	// score of the best hypothesis
	float BestScore = 0.0;

    // run over all predecessor nodes, i.e. over all incoming edges
    const CArray< CLatticeEdge * > &InEdges = pNode->InEdges();
    // Trace( 1, "Number of InEdges = " << InEdges.Size() );

    for ( int i = InEdges.MinIndex(); i <= InEdges.MaxIndex(); i++ ) {

      pEdge = InEdges[ i ];
      //Trace( 0, "OutEdge " << o << ": " << pEdge->Word() ) ;
      PrevNodeNumber = pEdge->FromNode()->Number();
      // Trace( 1, "PrevNodeNumber = " << PrevNodeNumber );

      // - acoustic score
      AcousticScore = pEdge->Score();

      // - translation score (i.e. accumulated transducer score
      // for original edges, use penalty
      Word = ((CLatticeEdgeTrans *)pEdge)->Word();
      if ( Word[ 0 ] == '@' ) {
		TranslationScore = ((CLatticeEdgeTrans *)pEdge)->WordInfo()->Score();
		// Trace( 1, "Translation Score for labeled edge: " << TranslationScore );
      } else {
		TranslationScore = 10;
		// Trace( 1, "Translation Score for source language edge: " << TranslationScore );
      }

      // - error score (i.e. when error tolerant matching is allowed)
      ErrorScore = ((CLatticeEdgeTrans *)pEdge)->ErrorScore();

      /*
      // - benefit (i.e. negative score) for edges spanning longer segments
      // of the input
      {
	  CWordInfo *pWordInfo =  ((CLatticeEdgeTrans *)pEdge)->WordInfo();
	  CTransducerItem *pTransducerItem = ((CLatticeEdgeTrans *)pEdge)->TransducerItem();
	  int l = 1;
	  if ( pTransducerItem ) {
	  l = pTransducerItem->TargetSentence()->Size();
	  }
	  ConcatScore = 1 - l;
	  Trace( 1, "Length of " << pWordInfo->TargetString() << " is " << l
	  << " Score is " << ConcatScore );
      }
      */
	  
      // - Concatenation score
      // if edge starts not from the first node, then the translation
      // is a concatenation of segments
      if ( pEdge->FromNode()->Number() != m_Nodes.MinIndex() ) {
		ConcatScore = 1;
      }
	  
      // - LM score for target language
      LMWordIndex WordIndex;
	  
      // collect the words used for LM-scoring
      if ( m_UseSource ) {
		// Trace( 2, "Use Source" );
		Word = ((CLatticeEdgeTrans *)pEdge)->Word();
		word = (const char*)Word;
		Words.Clear();
		Words.SetAt( 1, Word );
		WordIndex = m_pLM->GetVocabularyIndex( word );
		// Trace( 2, "Word = " << word << "  LM-WordIndex = " << WordIndex );
      } else {
		Word = ((CLatticeEdgeTrans *)pEdge)->Word();
		word = (const char*)Word;
		CWordInfo *pWordInfo =  ((CLatticeEdgeTrans *)pEdge)->WordInfo();
		// Trace( 2, "WordInfo: " << *pWordInfo );
		// Trace( 2, "SourceString: " << pWordInfo->SourceString() );
		// Trace( 2, "TargetString: " << pWordInfo->TargetString() );
		
		Words.Clear();
		if ( pWordInfo->HasTargetString() ) {
		  pWordInfo->GetTargetWords( Words );
		} else if ( pWordInfo->HasSourceString() ) {
		  pWordInfo->GetSourceWords( Words );
		} else {
		  Word = ((CLatticeEdgeTrans *)pEdge)->Word();
		  Words.SetAt( 1, Word );
		}
      }
      
      if ( m_pLM ) {
		WordIndices.Clear();
		for ( int t = Words.MinIndex(); t <= Words.MaxIndex(); t++ ) {
		  LMWordIndex WordIndex = m_pLM->GetVocabularyIndex( (string)((const char *)(Words[ t ])) );
		  WordIndices[ t ] = WordIndex;
		  // Trace( 1, "Word = " << Words[ t ] << "  LM-WordIndex = " << WordIndex );
		}
      }

      // run over all LM-states available in the previous node
      // fetch the map( LM-state, n-best-lists )
      map< CLMState, multimap< float , CBackPointerLM * > > & AllNBestListsPrevNode 
		= AllNBestListsAllNodes[ PrevNodeNumber ];
	  
      // Trace( 1, "Number of different LM-States in previous node: " << AllNBestListsPrevNode.size() );
	  
      // and loop over all entries in the map
      map< CLMState, multimap< float , CBackPointerLM * > >::iterator IterLMStatePrevNode;
      for ( IterLMStatePrevNode = AllNBestListsPrevNode.begin(); 
			IterLMStatePrevNode != AllNBestListsPrevNode.end();
			++IterLMStatePrevNode ) {
		
		// fetch LM-state (==history) and associated n-best list (previous node)
		CLMState LMState = IterLMStatePrevNode->first;
		multimap< float, CBackPointerLM * > &NBestListPrevNode = IterLMStatePrevNode->second;
		
		// Trace( 1, "LM-State: " << LMState );
		
		// run over all hypotheses in n-best list of previous node for current LM state
		multimap< float, CBackPointerLM * >::iterator IterNBestListPrevNode;
		multimap< float, CBackPointerLM * >::iterator CurrIter;
		int NumBest = 1;
		
		// Trace( 2, "Number of Hypotheses for current LM-State: " << NBestListPrevNode.size() );
		
		for ( IterNBestListPrevNode = NBestListPrevNode.begin(); 
			  ( IterNBestListPrevNode != NBestListPrevNode.end() ) && ( NumBest <= N ); 
			  ++IterNBestListPrevNode, NumBest++ ) {
		  
		  // Trace( 2, "PrevHypNumber = " << NumBest );
		  pBackPointer = IterNBestListPrevNode->second;
		  //NewLength = PartialPathes[ PrevNodeNumber ].GetAt( k ) + 1;
		  
		  // get all the scores:
		  // - accumulated score in previous node
		  AccumulatedScore = IterNBestListPrevNode->first;
		  pBackPointer->GetScores( SumACScore, SumLMScore, SumTRScore, SumERScore, SumCCScore );
		  
		  // - LM score
		  if ( m_pLM ) {
			LMScoreWord = 0;
			LMStateWord = pBackPointer->m_LMStateWord;
			// Trace( 2, "LMstate from BackPointer: " << LMStateWord );
			
			for ( int x = WordIndices.MinIndex(); x <= WordIndices.MaxIndex(); x++ ) {
			  //if ( WordIndices[ x ] == 0 ) {
			  //LMScoreWord += 20;
			  //Trace( 1, "LM-Score for " << Words[ x ] << ": 20" );
			  //} else {
			  LMScoreWord += -pLMSentenceStateWord->LogProb( LMStateWord, WordIndices[ x ] );
			  // Trace( 1, "LM-Score for " << Words[ x ] << ": " 
// 					 << -pLMSentenceStateWord->LogProb( LMStateWord, WordIndices[ x ] ) 
// 					 << " total LM-Score = " << LMScoreWord );
			  //}
			  
			  NewLMStateWord = pLMSentenceStateWord->AppendWord( LMStateWord, WordIndices[ x ] );
			  // Trace(1, "New LM-State: " << NewLMStateWord );
			  LMStateWord = NewLMStateWord;
			}
		  }
		  
		  // new cummulative score (= weighted sum)
		  SumACScore += m_ACScale * AcousticScore;
		  SumLMScore += m_LMScale * LMScoreWord;
		  SumTRScore += m_TRScale * TranslationScore;
		  SumERScore += m_ERScale * ErrorScore;
		  SumCCScore += m_CCScale * ConcatScore;
		  
		  NewScore = AccumulatedScore 
			+ m_ACScale * AcousticScore
			+ m_LMScale * LMScoreWord
			+ m_TRScale * TranslationScore 
			+ m_ERScale * ErrorScore
			+ m_CCScale * ConcatScore;
		  
// 		  Trace( 1, "Scores: "
// 				 << " AC = " << m_ACScale << "*" << AcousticScore
// 				 << " LM = " << m_LMScale << "*" << LMScoreWord 
// 				 << " TM = " << m_TRScale << "*" << TranslationScore
// 				 << " ER = " << m_ERScale << "*" << ErrorScore
// 				 << " CC = " << m_CCScale << "*" << ConcatScore );
// 		  Trace( 1, "Accumulated Scores: "
// 				 << " AC = " << SumACScore
// 				 << " LM = " << SumLMScore
// 				 << " TM = " << SumTRScore
// 				 << " ER = " << SumERScore
// 				 << " CC = " << SumCCScore );
// 		  // Trace( 1, "Total Scores: "
// 				 << " Old = " << AccumulatedScore 
// 				 << " New = " << NewScore );
		  
		  // Pruning: added 02.05.03 (happy birthday M.)
		  if ( NewScore < m_BeamFactor * BestScore ) {
			GeneratedHyps++;
			PrunedHyps++;
		  } else {

			GeneratedHyps++;
			if ( NewScore > BestScore ) {
			  BestScore = NewScore;
			}

			// create backtrace information for memoryypotheses
			pNewBackPointer = new CBackPointerLM( NewLMStateWord, pEdge, (*IterNBestListPrevNode).second );
		  
			pNewBackPointer->SetScores( SumACScore, SumLMScore, SumTRScore, SumERScore, SumCCScore );
		  
			// now, store memoryypothesis:
			// do we have that LM state already in the NBestList for the current node ?
			// if yes, fetch it, if not, insert it
			map< CLMState, multimap< float , CBackPointerLM * > > & AllNBestListsThisNode
			  = AllNBestListsAllNodes[ n ];
		  
			multimap< float, CBackPointerLM * > & NBestListThisNode = AllNBestListsThisNode[ NewLMStateWord ];
		  
			// insert the hypothesis into the Score-Backpointer-Map
			CurrIter = NBestListThisNode.insert( pair< float, CBackPointerLM * >( NewScore, pNewBackPointer ) ); 
			// Trace( 2, "NewScore " << NewScore << " " << CurrIter->first );
		  } // if ( NewScore < m_BeamFactor * BestScore )

		} // for ( IterNBestListPrevNode = NBestListPrevNode.begin();
	  } // for ( IterLMStatePrevNode = AllNBestListsPrevNode.begin();
	  
      // Trace( 2, "NewScore = " << NewScore );
    } // for ( int i = InEdges.MinIndex(); i <= InEdges.MaxIndex(); i++ )
	
  } // for (int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++)

  
  // add LM scores for sentence end
  // no separation according to LM-state required any more
  // Trace( 1, "Add sentence end scores" );
  
  multimap< float, CBackPointerLM * > FinalNBestList;
  
  map< CLMState, multimap< float , CBackPointerLM * > > & 
    AllNBestListsThisNode = AllNBestListsAllNodes[ m_Nodes.MaxIndex() ];
  
  // and loop over all entries in the map
  map< CLMState, multimap< float , CBackPointerLM * > >::iterator IterLMState;
  for ( IterLMState = AllNBestListsThisNode.begin(); 
		IterLMState != AllNBestListsThisNode.end();
		++IterLMState ) {
    
    // fetch LM-state (==history) and associated n-best list (previous node)
    CLMState LMState = IterLMState->first;
    multimap< float, CBackPointerLM * > &NBestList = IterLMState->second;
    
    // run over all hypotheses in n-best list of previous node for current LM state
    multimap< float, CBackPointerLM * >::iterator IterNBestList;
    
    int NumBest = 1;
    
    for ( IterNBestList = NBestList.begin(); 
		  ( IterNBestList != NBestList.end() ) && ( NumBest <= N ); 
		  ++IterNBestList, NumBest++ ) {
      
      // Trace( 2, "HypNumber = " << NumBest );
      AccumulatedScore = IterNBestList->first;
      pBackPointer = IterNBestList->second;
      //NewLength = PartialPathes[ PrevNodeNumber ].GetAt( k ) + 1;
	  
      pBackPointer->GetScores( SumACScore, SumLMScore, SumTRScore, SumERScore, SumCCScore );
      //pBackPointer->DisplayScores();
	  
      float LMScore = 0;
      if ( m_pLM ) {
		LMScore = -(m_LMScale * pLMSentenceStateWord->LogProbEnd( LMState ));
		// Trace( 1 , "LM for Sentence End = " << LMScore );
		SumLMScore += m_LMScale * LMScore;
      }
      
      //Bug fix, by Joy, joy@cs.cmu.edu
      //Apr 10,
      //Buggy version:
      //NewScore = AccumulatedScore + LMScore;
      NewScore = AccumulatedScore + m_LMScale * LMScore;



      //Added by Joy, joy@cs.cmu.edu
      //To output the TM and LM score for the best path
      //Apr 10, 2002
      pBackPointer->SetScores(  SumACScore, SumLMScore, SumTRScore, SumERScore, SumCCScore );

// 	  Trace( 1, "Accumulated Scores: "
// 			 << " AC = " << SumACScore
// 			 << " LM = " << SumLMScore
// 			 << " TM = " << SumTRScore
// 			 << " ER = " << SumERScore
// 			 << " CC = " << SumCCScore );
// 	  Trace( 1, "Total Scores: "
// 			 << " Old = " << AccumulatedScore 
// 			 << " New = " << NewScore );
	  

      // create backtrace information for memoryypotheses
      //pNewBackPointer = new CBackPointerLM( LMState, pBackPointer->m_pEdge, (*IterNBestList).second );
      
      FinalNBestList.insert( pair< float, CBackPointerLM * >( NewScore, pBackPointer ) ); 
    }
	
  }
  
  
  
  // how many shortest pathes are available
  int NumPathes;
  //int NumLastNode = m_Nodes.MaxIndex();
  
  int CollectedPathes = FinalNBestList.size();
  if ( CollectedPathes < N ) {
    NumPathes = CollectedPathes;
    // Trace( 2, "Number of best pathes is " << NumPathes << "; smaller than " << N );
  } else {
    // Trace( 2, "Number of best pathes is " << CollectedPathes << "; larger than " << N );
    NumPathes = N;
  }
  
  
  // now, trace back to collect best path
  // we know already the length of that path
  multimap< float , CBackPointerLM * >::iterator CurrIter;
  CArray< CLatticeEdge *> Array;
  NumBest = 1;
  float Score;
  
  for ( CurrIter = FinalNBestList.begin();
		( CurrIter != FinalNBestList.end() ) && ( NumBest <= NumPathes );
		++CurrIter, NumBest++ ) {
	
    Array.RemoveAll();
	
    Score = CurrIter->first;
    
	cout << "Generated Hyps " << GeneratedHyps++ << " Prunded Hyps " << PrunedHyps++ << endl;
    //modified by Joy, joy@cs.cmu.edu
    //Apr 10, 2003
    cout << "TotalScore=" << Score;
	
    pBackPointer = CurrIter->second;
    pBackPointer->GetScores( SumACScore, SumLMScore, SumTRScore, SumERScore, SumCCScore );
    cout << " TMscore="<<SumTRScore<<" LMscore="<<SumLMScore<<endl;
	

    pEdge = pBackPointer->Edge();
    if ( pEdge ) {
      if ( TraceLevel >= 2 ) 
	pEdge->Display();
      Array.Insert( 0, pEdge );
    }

    while ( pEdge->FromNode() != StartNode() ) {
      pBackPointer = pBackPointer->PrevBackPointer();
      pEdge = pBackPointer->Edge();
      if ( pEdge ) {
	if ( TraceLevel >= 2 ) 
	  pEdge->Display();
	Array.Insert( 0, pEdge );
      }
    }

    BestPathes.insert( pair< float, CArray< CLatticeEdge * > >( Score, Array ) );
    
    //commented by Joy, joy@cs.cmu.edu, Apr 02, 2003
    //the partial path will be displayed anyway later
    //DisplayPath( Array );
   
  }

  // Trace( 2, "Number of best pathes = " << BestPathes.size() );


  // now delete the backpointers
  int NodeNumber = 0;
  for (int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++) {
    pNode = m_Nodes[ n ];
    NodeNumber = pNode->Number();
    map< CLMState, multimap< float , CBackPointerLM * > > &
      AllNBestLists = AllNBestListsAllNodes[ NodeNumber ];
    // and loop over all entries in the map
    map< CLMState, multimap< float , CBackPointerLM * > >::iterator 
      IterLMState;

    for ( IterLMState = AllNBestLists.begin(); 
	  IterLMState != AllNBestLists.end();
	  ++IterLMState ) {
      
      multimap< float, CBackPointerLM * > &NBestList = IterLMState->second;
      multimap< float, CBackPointerLM * >::iterator IterNBestList;
      
      for ( IterNBestList = NBestList.begin(); 
	    IterNBestList != NBestList.end(); 
	    ++IterNBestList ) {
	
	// Trace( 2, "PrevHypNumber = " << NumBest );
	pBackPointer = IterNBestList->second;
	delete pBackPointer;
	pBackPointer = NULL;
      }
      
    }

  }

  return NumPathes;

}


 void
CLatticeLM::DisplayPath( CArray< CLatticeEdge* > &Path ) const
{
  cout << "CLatticeLM::DisplayPath" << endl;
  CString String, SourceString, TargetString;
  CString Word;
  CLatticeEdgeTrans * pEdgeTrans;
  CWordInfo * pWordInfo;

  // Trace( 1, "Display Path" );


  for ( int i = Path.MinIndex();
	i<= Path.MaxIndex();
	i++ ) {
    if ( i > Path.MinIndex() )
      cout << "\n";
 
    pEdgeTrans = (CLatticeEdgeTrans *)(Path[ i ]);
    cout << pEdgeTrans->Word();

    pWordInfo = pEdgeTrans->WordInfo();
    if ( pWordInfo->HasSourceString() ) {
      cout << " " << *pWordInfo;
    }

  }
  cout << endl;
}



void
CLatticeLM::DisplayPath( CArray< CLatticeEdgeTrans* > &Path ) const
{
  CString String, SourceString, TargetString;
  CString Word;
  CLatticeEdgeTrans * pEdgeTrans;
  CWordInfo * pWordInfo;

  // Trace( 1, "Display Path" );


  for ( int i = Path.MinIndex();
	i<= Path.MaxIndex();
	i++ ) {
    if ( i > Path.MinIndex() )
      cout << "\n";
 
    pEdgeTrans = (CLatticeEdgeTrans *)(Path[ i ]);
    cout << pEdgeTrans->Word();

    pWordInfo = pEdgeTrans->WordInfo();
    if ( pWordInfo->HasSourceString() ) {
      cout << " " << *pWordInfo;
    }

  }
  cout << endl;
}

void CLatticeLM::DisplayPath( CArray< CLatticeEdgeMS* > &Path, const CString &Tag ) const
{
  CString String, SourceString, TargetString;
  CString Word;
  CLatticeEdgeMS * pEdgeTrans;
  CWordInfo * pWordInfo;

  // Trace( 1, "Display Path" );


  for ( int i = Path.MinIndex(); i<= Path.MaxIndex(); i++ ) {
   
    if ( i > Path.MinIndex() )
      cout << "\n";
    
    cout << Tag << " " << i << " ";

    pEdgeTrans = (CLatticeEdgeMS *)(Path[ i ]);
    cout << "FromNode=" << pEdgeTrans->FromNode()->Number() << " ";
    cout << "ToNode=" << pEdgeTrans->ToNode()->Number() << " ";
    
    pWordInfo = pEdgeTrans->WordInfo();
    if ( pWordInfo->HasSourceString() ) {
      cout << pEdgeTrans->Word() << " ";
      cout << *pWordInfo;
      if ( pWordInfo->HasTargetString() ) {
        cout << pEdgeTrans->Scores();
      }
    } else {
      cout << "@SOURCE_WORD { "
           << pEdgeTrans->Word() << " # " << pEdgeTrans->Word() << " # 100 } ";
      //cout << " Score=" << pEdgeTrans->Score() 
      //   << " StartFrame=" << pEdgeTrans->Startframe()
      //   << " EndFrame=" << pEdgeTrans->Endframe();
    }
  }
  cout << endl;
}
