#include <config.h>

#ifndef LATTICELM_H
#define LATTICELM_H

#include <string>

#include <Lattice.hh>
#include <LMInterface.hh>
#include <LatticeEdgeTrans.hh>
#include <LatticeEdgeMS.hh>

//#include "fortify.h"  -- removed by Joy, Sep 17 4:18pm, because fortify.h too old and not competible with memory in the standard lib

class CLatticeLM;

class CBackPointerLM
{
public:
  CBackPointerLM( CLMState LMState, CLatticeEdge *pEdge, CBackPointerLM *pPrevBackPointer );
  ~CBackPointerLM();

  CLatticeEdge *Edge() { return m_pEdge; };
  CBackPointerLM *PrevBackPointer() { return m_pPrevBackPointer; };

  void GetScores( float &SumACScore, float &SumLMScore, 
		  float &SumTRScore, float &SumERScore,
		  float &SumCCScore );
  void SetScores( float SumACScore, float SumLMScore, 
		  float SumTRScore, float SumERScore,
		  float SumCCScore );

  void DisplayScores();

  static long int s_NumberOfObjects;

private:
  friend class CLatticeLM;
  CLMState m_LMStateWord;
  CLatticeEdge *m_pEdge;
  CBackPointerLM *m_pPrevBackPointer;

  // not neccessary, but interesting
  float m_SumACScore;
  float m_SumLMScore;
  float m_SumTRScore;
  float m_SumERScore;
  float m_SumCCScore;
};


class CLatticeLM : public CLattice
{

public:
  // Konstruktor. Erzeugt eine leere Lattice.
  CLatticeLM( CLatticeEdge * pEdgePrototype, CLM * pLM );

  // Destruktor. Lscht enthaltene Knoten und Kanten mit.
  //~CLattice();

  void SetLMScale( float LMScale ) { m_LMScale = LMScale; };
  void SetScalingFactors( float ACScale, float LMScale,
			  float TRScale, float ERScale,
			  float CCScale );

  void SetBeamFactor( float BeamFactor );

  void LocalReordering();

  // collect the N-best pathes through the lattice and their scores
  // return the number (<=N) of found pathes
  int BestPathes( int N,
		  multimap< float, CArray< CLatticeEdge* > > &BestPathes ) const;
  
  void DisplayPath( CArray< CLatticeEdge* > &Path ) const;
  void DisplayPath( CArray< CLatticeEdgeTrans* > &Path ) const;

  //ARV 04/29 support multi score lattice edge
  void DisplayPath( CArray< CLatticeEdgeMS* > &Path, const CString & Tag = "" ) const;

private:
  CLM *m_pLM;

  // scaling factors for the different scores
  float m_ACScale;       // acoustic
  float m_LMScale;       // language model
  float m_TRScale;       // transducer
  float m_ERScale;       // error tolerant match
  float m_CCScale;       // concatenation benefit

  float m_BeamFactor;    // factor for pruning

  // ??? use source words for LM score
  BOOL m_UseSource;
};

#endif

