#include <config.h>

#include <Array.hh>



inline
CLatticeNode::CLatticeNode( int Number )
{
  m_Number = Number;
}


// Selektor fr die logische Knotennummer 
inline int
CLatticeNode::Number()
{
  return m_Number;
}


// Selektor fr die Liste eingehender Kanten 
inline CArray< CLatticeEdge * > &
CLatticeNode::InEdges()
{
  return m_InEdges;
}


// Selektor fr die Liste ausgehender Kanten 
inline const CArray< CLatticeEdge * > & 
CLatticeNode::OutEdges()
{
  return m_OutEdges;
}



// ausgehende Kante hinzufgen 
inline void
CLatticeNode::AddOutEdge( CLatticeEdge *pOutEdge )
{
  //  Trace( 5, "CLatticeNode::AddOutEdge - use Append" );
  //  Trace( 5, m_OutEdges.MinIndex() << " " << m_OutEdges.MaxIndex() );

  m_OutEdges.Append( pOutEdge );
}


// eingehende Kante hinzufgen 
inline void
CLatticeNode::AddInEdge( CLatticeEdge * pInEdge )
{
  //  Trace( 5, "CLatticeNode::AddInEdge - use Append" );
  m_InEdges.Append( pInEdge );
}




inline ostream & 
CLatticeNode::operator<<( ostream & OutStream )
{
  OutStream << "node " << this << ": " << m_Number << endl;
  
  // alle eingehenden Kanten durchgehen 
  OutStream << "  incoming edges" << endl;
  int i;

  for ( i = m_InEdges.MinIndex(); i <= m_InEdges.MaxIndex(); i++) {
    //    OutStream << *(m_InEdges[i]) << endl;
    m_InEdges[i]->operator<<(OutStream);
  }

  // alle ausgehenden Kanten durchgehen 
  OutStream << "  outgoing edges" << endl;

  for ( i = m_OutEdges.MinIndex(); i <= m_OutEdges.MaxIndex(); i++) {
    //m_OutEdges[i]->operator<<(OutStream);
    m_OutEdges[i]->operator<<(OutStream);
  }

  OutStream << endl << flush;

  return OutStream;
}

inline void 
CLatticeNode::Display()
{

  cout << "node " << this << ": " << m_Number << endl;
  
  // alle eingehenden Kanten durchgehen 
  cout << "  " << m_InEdges.Size() << " incoming edges" << endl;
  int i;
  for ( i = m_InEdges.MinIndex(); i <= m_InEdges.MaxIndex(); i++) {
    //    OutStream << *(m_InEdges[i]) << endl;
    m_InEdges[i]->Display();
  }

  // alle ausgehenden Kanten durchgehen 
  cout << "  " << m_OutEdges.Size() << " outgoing edges" << endl;
  for ( i = m_OutEdges.MinIndex(); i <= m_OutEdges.MaxIndex(); i++) {
    //m_OutEdges[i]->operator<<(OutStream);
    m_OutEdges[i]->Display();
  }

  cout << endl << flush;

}


// Zweistelliger Vergleichsoperator < fr Lattice-Knoten.
//  @param a Referenz auf Knoten einer Lattice
//  @param b Referenz auf Knoten einer Lattice
//  @return Das bool'sche Ergebnis des Vergleichs a.Number<b.Number 
//inline bool operator<(LatticeNode& a,LatticeNode& b)
//{ return a.operator<(&b); }

// Zweistelliger Vergleichsoperator > fr Lattice-Knoten
//  @param a Referenz auf Knoten einer Lattice
//  @param b Referenz auf Knoten einer Lattice
//  @return Das bool'sche Ergebnis des Vergleichs a.Number>b.Number 
//inline bool operator>(LatticeNode& a,LatticeNode& b)
//{ return a.operator>(&b); }



// Binary output operator for Lattice-Nodes.
//  @param os output stream
//  @param n  Reference to node to be output.
inline ostream& 
operator<<( ostream & os, CLatticeNode & n )
{
  return n.operator<<( os );
}


