#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#define NO_TRACE 1
#include <Trace.hh>

#include <LatticeSearch.hh>



//ostream &operator << (ostream &s, const CLatticeSearchHyp &PhraseHyp )
//{
//  s << PhraseHyp.m_StartNode << " " << PhraseHyp.m_Cost << " " << PhraseHyp.m_AcuScore;

//  return s;
//}


long int CLatticeSearchHyp::s_NumberOfObjects = 0;



bool
operator==( CLatticeSearchHyp &Hyp1, CLatticeSearchHyp &Hyp2 )
{
  if ( (Hyp1.m_State == Hyp2.m_State) && (Hyp1.m_StartNode == Hyp2.m_StartNode) )
   return true;
  else
    return false;
}



ostream &
operator<<( ostream& st, const CLatticeSearchHyp &hyp )
{
  st << "StartNode = " << hyp.m_StartNode << " TransducerState = " << hyp.m_State 
     << " Cost = " << hyp.m_Cost << " PathLenght = " << hyp.m_WhgPathLength 
     << " Edge = " << hyp.m_pBackPointerToEdge << " PrevHyp = " << hyp.m_BackPointerToEntry
     << endl;
  return st;
  
}



CLatticeSearch::CLatticeSearch()
{
  m_pLattice= NULL;

  m_MaxCost = 0;
  m_ErrorsPerWord= 0.0;
  m_LengthDependentCosts=FALSE;
}



void
CLatticeSearch::SetVocabulary( const CVocabulary *pVocabulary )
{
  m_pVocabulary = pVocabulary;
}



void
CLatticeSearch::SetLattice( CLattice *pLattice )
{
  m_pLattice = pLattice;
}



void
CLatticeSearch::SetMaxCost( double MaxCost )
{
  Trace( 3, "CLatticeSearch::SetMaxCost " << MaxCost );
  m_MaxCost = MaxCost;
  m_LengthDependentCosts = FALSE;
}




void
CLatticeSearch::InitSearch()
{
  TRACE( 1, "CLatticeSearch::InitSearch" );

  // add empty search hypothesis to each node in lattice

  // get a reference to the list of all nodes in the lattice
  const CArray< CLatticeNode * > NodeList = m_pLattice->Nodes();

  // Set boundaries of arrays for search hypothesis
  m_Hyps.Clear();
  m_Hyps.AssureBoundaries( NodeList.MinIndex(), NodeList.MaxIndex() );
  
  // create an empty search hypothesis with start hypothesis from transducer
  CTransducerState StartState = CTransducer::Start();

  CLatticeSearchHyp Hyp;
  Hyp.m_State = StartState;

  // init each node with empty search hypothesis 
  for ( int index = NodeList.MinIndex(); index <= NodeList.MaxIndex(); index++ ) {
    m_Hyps.GetRow( index ).SetIncrement( MAX_ACTIVE_HYPS );
    Hyp.m_StartNode = index;
    m_Hyps.GetRow( index ).Append( Hyp );
  }

}



int
CLatticeSearch::ApplyTransducer( const CTransducerAccess *pTransducerAccess )
{
  Trace( 1, "-------------------------------------------------------" );
  Trace( 1, "CLatticeSearch::ApplyTransducer" << pTransducerAccess->Transducer()->Name() );
  Trace( 1, "-------------------------------------------------------" );

  // run over lattice and match partial paths against the transducer
  // for each node of lattice, start with new empty hypothesis
  // (which means start node of transducer)
  // and expand current hypotheses over all ingoing edges
  // if possible, i.e. there is a transition from old transducer state 
  // to new transducer node with word, then append memoryypothesis
  // to next node

  // list of hypotheses for complete phrases
  CArray< CLatticeSearchHyp > CompleteHyps;
  CompleteHyps.SetIncrement( 1000 );

  // list of edges which have to be inserted into the lattice
  CArray1D< CLatticeEdgeMS* > NewEdgesList;
  NewEdgesList.SetIncrement( 1000 );

  // assign parameter to member variable for access in other methods
  m_pTransducerAccess = pTransducerAccess;

  // get a reference to the list of all nodes in the lattice
  const CArray< CLatticeNode * > NodeList = m_pLattice->Nodes();;
  //assert( NodeList.MaxIndex() < MAX_WHG_NODES );
  
  // for statistics
  m_NumberOfActiveHypothesis      = 0;
  m_NumberOfStartedPhrases        = 0;
  m_NumberOfExpandedPhrases       = 0;
  m_NumberOfNIBMoves              = 0;
  m_NumberOfDeletedHypotheses     = 0;
  
  m_NumberOfPhrasesFromPhraseTree = 0;
  m_NumberOfCompletePhrases       = 0;

  CTransducerState State, NextState;

  CLatticeNode *pNode;
  CLatticeEdgeMS *pEdge;

  CString Word;
  CWordIndex Wf;
  int n;


  // run over all nodes in the lattice
  for ( n = NodeList.MinIndex(); n <= NodeList.MaxIndex(); n++ ) {
    
    Trace(1,  "Node number " << n);
    pNode = NodeList[ n ];
    
    // the edges running into the current node
    const CArray<CLatticeEdge *> &InEdges = pNode->InEdges();
    
    // run over all incoming edges
    for ( int i = InEdges.MinIndex(); i <= InEdges.MaxIndex(); i++ ) {
      pEdge = (CLatticeEdgeMS *)InEdges[ i ];
      Word = ((CLatticeEdgeMS *)pEdge)->Word();
      Wf = m_pVocabulary->GetIndex( Word );
      Trace(1, "  Edge number " << i << " with word " << Word);
      
      // fetch hypothesis list from previous node
      int PrevNode = pEdge->FromNode()->Number();
      CArray< CLatticeSearchHyp > &HypsInPrevNode = m_Hyps.GetRow( PrevNode );

      Trace( 1, "  Number of Hyps in previous Node = " << HypsInPrevNode.Size() );
      // should have at least an empty search hypothesis
      assert( HypsInPrevNode.Size() > 0 );

      if ( TraceLevel > 0 ) {
        for ( int HypIndex = HypsInPrevNode.MinIndex(); 
              HypIndex <= HypsInPrevNode.MaxIndex();
              HypIndex++ ) {
          Trace( 1, "      Hyp " << HypIndex << " : " << HypsInPrevNode[ HypIndex ] );
        }
      }
	  
      // expand all hypotheses from previous node with word at edge
      for ( int HypIndex = HypsInPrevNode.MinIndex(); 
            HypIndex <= HypsInPrevNode.MaxIndex();
            HypIndex++ ) {
		
        State = HypsInPrevNode[ HypIndex ].m_State;
        Trace( 1, "    Previous state = " << State );
        pTransducerAccess->Next( State, Wf, NextState );
        Trace( 1, "    Nextstate = " << NextState );

        if ( NextState >= 0 ) {
          // store memoryypothesis together with backtrace information
          CLatticeSearchHyp NewHyp;
          NewHyp.m_State = NextState;
          NewHyp.m_StartNode = HypsInPrevNode[ HypIndex ].m_StartNode;
          NewHyp.m_pBackPointerToEdge = pEdge;
          NewHyp.m_BackPointerToEntry = HypIndex;
          
          // accumulate multiple scores of *source* edge in partial hypothesis (apd 8/14/06)
          CNumericalArray< float > EdgeScores = ((CLatticeEdgeMS*) pEdge)->Scores();
          for ( int s = EdgeScores.MinIndex(); s <= EdgeScores.MaxIndex(); s++ )
          {
            float prevNodeHypScore;
            if ( HypsInPrevNode[ HypIndex ].m_Scores.MaxIndex() >= s) {
              //cout << "NUMBER OF SCORES IN PREV PARTIAL HYP *MATCHES*!!" << endl;
              prevNodeHypScore = HypsInPrevNode[ HypIndex ].m_Scores[ s ];
            } else { 
              //cout << "NUMBER OF SCORES IN PREV PARTIAL HYP *MIS*MATCHES!!" << endl;
              prevNodeHypScore = 0;
            }
            
            if ( NewHyp.m_Scores.MaxIndex() < s ) {
              // array doesn't have an init value              
              NewHyp.m_Scores.Append( EdgeScores[ s ] + prevNodeHypScore );
            } else {
              NewHyp.m_Scores[ s ] += EdgeScores[ s ] + prevNodeHypScore ;
            }
          }
          
          //NewHyp.m_AcuScore = HypsInPrevNode[ HypIndex ].m_AcuScore + ((CLatticeEdgeTrans *)pEdge)->Score();
          
          Trace( 1, "    New Hyp: " << NewHyp << " stored for node " << n );
          m_Hyps.GetRow( n ).Append( NewHyp );
		  
          // we remember final states to create new edges later on
          if ( pTransducerAccess->IsFinal( NextState ) ) {
            Trace( 1, "    Nextstate = " << NextState << ": Is final" );
            CArray< CTransducerItem *> *pTransducerItems = pTransducerAccess->GetOutput( NextState );
			
            Trace( 1, "    Result: " );
            for ( int i = pTransducerItems->MinIndex(); i <= pTransducerItems->MaxIndex(); i++ ) {
              //pTransducerItems->GetAt( i )->Display();
              Trace( 3, "  Label is: " << pTransducerItems->GetAt( i )->Label() );
              Trace( 3, "  Translation is: " << pTransducerItems->GetAt( i )->Target() );
            }

            delete pTransducerItems;
            CompleteHyps.Append( NewHyp );
          }
        }

      } // loop over hypotheses
    } // loop over incoming edges
    
    m_NumberOfActiveHypothesis += m_Hyps.GetRow( n ).Size();
    
    TRACE(3, " Active Hyps: " <<  m_Hyps.GetRow( n ).Size());
    
    //TRACE(0, "PhraseList");
    //cout << m_Hyps.GetRow(n) << endl;
    
    // for complete phrases ending in current node, create a new edge
    // this is added to a list of new edges which will be inserted
    // into the lattice only after the complete lattice has been traversed
    
    m_NumberOfCompletePhrases = CompleteHyps.Size();
    Trace( 1, "    Number of complete hypotheses = " << CompleteHyps.Size() );

    for ( int i = CompleteHyps.MinIndex(); i <= CompleteHyps.MaxIndex(); i++ ) {
      Trace( 1, "      Hyp " << i << " " << CompleteHyps[ i ] );
      CreateNewEdgesMS( &CompleteHyps[ i ], pNode, NewEdgesList );
    }

    Trace( 1, "    Numer of complete edges = " << NewEdgesList.Size() );

    CompleteHyps.Clear();
    
    
    // DEBUG****************
    //TRACE(0, "");
    
    //cout << "Number of active hypothesis was:   " << m_NumberOfActiveHypothesis << endl;
    //cout << "Number of started hypothesis was:  " << m_NumberOfStartedPhrases << endl;
    //cout << "Number of expanded phrases was:    " << m_NumberOfExpandedPhrases << endl;
    //cout << "Number of moves over NIB edges     " << m_NumberOfNIBMoves << endl;
    //cout << "Number of deleted hypothesis was   " << m_NumberOfDeletedHypotheses << endl;
    
    //assert ( m_NumberOfActiveHypothesis == m_NumberOfStartedPhrases + m_NumberOfExpandedPhrases +
    //m_NumberOfNIBMoves - m_NumberOfDeletedHypotheses);
    
    //m_NumberOfActiveHypothesis  = 0;
    //m_NumberOfStartedPhrases    = 0;
    //m_NumberOfExpandedPhrases   = 0;
    //m_NumberOfNIBMoves          = 0;
    //m_NumberOfDeletedHypotheses = 0;
    
  } // for all nodes
  

  // Add new edges to Whg
  Trace( 1, "Add new edges to Whg" );
  int NumberOfAddedEdges = 0;
  bool EdgeAdded;

  for( int e = NewEdgesList.MinIndex(); e <= NewEdgesList.MaxIndex(); e++ ) {
    //NewEdgesList.GetAt(e)->Display();
    /*
      EdgeAdded = m_pLattice->AddEdgeUniq( NewEdgesList.GetAt( e ) );
      //m_pLattice->AddEdge( NewEdgesList.GetAt( e ) );
      if ( EdgeAdded ) {
      NumberOfAddedEdges++;
      } else {
      delete NewEdgesList.GetAt( e );
      NewEdgesList.SetAt( e, NULL );
      }
    */
    m_pLattice->AddEdge( NewEdgesList.GetAt( e ) );
    NumberOfAddedEdges++;
  }
  if ( TraceLevel > 0 ) 
    m_pLattice->Display();
  

  return NumberOfAddedEdges;
}



//Added by Joy, joy@cs.cmu.edu
// Mar 26, 2003
// Simply apply the transducer on the word level
// do not consider the Append to hierarchical
// this will run faster than the standard version of ApplyTransducer
// since it do not go over all the in-edges of the current node
// just check the hyps from the previous node
int
CLatticeSearch::ApplyTransducerSimple( const CTransducerAccess *pTransducerAccess )
{
  Trace( 1, "-------------------------------------------------------" );
  Trace( 1, "CLatticeSearch::ApplyTransducer" << pTransducerAccess->Transducer()->Name() );
  Trace( 1, "-------------------------------------------------------" );

  // run over lattice and match partial pathes against the transducer
  // for each node of lattice, start with new empty hypothesis
  // (which means start node of transducer)
  // and expand current hypotheses over the ingoing edge with the source word ( edge number 0 )

  // list of hypotheses for complete phrases
  CArray< CLatticeSearchHyp > CompleteHyps;
  CompleteHyps.SetIncrement( 1000 );

  // list of edges which have to be inserted into the lattice
  CArray1D< CLatticeEdgeMS* > NewEdgesList;

  NewEdgesList.SetIncrement( 1000 );

  // assign parameter to member variable for access in other methods
  m_pTransducerAccess = pTransducerAccess;

  // get a reference to the list of all nodes in the lattice
  const CArray< CLatticeNode * > NodeList = m_pLattice->Nodes();;
  //assert( NodeList.MaxIndex() < MAX_WHG_NODES );

  // for statistics
  m_NumberOfActiveHypothesis      = 0;
  m_NumberOfStartedPhrases        = 0;
  m_NumberOfExpandedPhrases       = 0;
  m_NumberOfNIBMoves              = 0;
  m_NumberOfDeletedHypotheses     = 0;

  m_NumberOfPhrasesFromPhraseTree = 0;
  m_NumberOfCompletePhrases       = 0;

  CTransducerState State, NextState;

  CLatticeNode *pNode;
  CLatticeEdgeMS *pEdge;

  CString Word;
  CWordIndex Wf;
  int n;


  // run over all nodes in the lattice
  for ( n = NodeList.MinIndex(); n <= NodeList.MaxIndex(); n++ ) {

    Trace( 1,  "Node number " << n );
    pNode = NodeList[ n ];

    // the edges running into the current node
    const CArray<CLatticeEdge *> &InEdges = pNode->InEdges();

    int i=InEdges.MinIndex();  //only consider the first edge which is the source word
    if( i <=  InEdges.MaxIndex() ){// for the first node, maxIndex is -1


      pEdge = (CLatticeEdgeMS *)InEdges[ i ];

      Word = ((CLatticeEdgeMS *)pEdge)->Word();
      Wf = m_pVocabulary->GetIndex( Word );
      Trace(1, "  Edge number " << i << " with word " << Word);

      // fetch hypothesis list from previous node
      int PrevNode = pEdge->FromNode()->Number();
      CArray< CLatticeSearchHyp > &HypsInPrevNode = m_Hyps.GetRow( PrevNode );

      Trace( 1, "  Number of Hyps in previous Node = " << HypsInPrevNode.Size() );
      // should have at least an empty search hypothesis
      assert( HypsInPrevNode.Size() > 0 );

#ifdef TRACE_ENABLED
      if ( TraceLevel > 0 ) {
        for ( int HypIndex = HypsInPrevNode.MinIndex();
              HypIndex <= HypsInPrevNode.MaxIndex();
              HypIndex++ ) {
          Trace( 1, "      Hyp " << HypIndex << " : " << HypsInPrevNode[ HypIndex ] );
        }
      }
#endif


      // expand all hypotheses from previous node with word at edge
      for ( int HypIndex = HypsInPrevNode.MinIndex();
            HypIndex <= HypsInPrevNode.MaxIndex();
            HypIndex++ ) {
        //for all hyp in prev node

        State = HypsInPrevNode[ HypIndex ].m_State;
        Trace( 1, "    Previous state = " << State );
        pTransducerAccess->Next( State, Wf, NextState );
        Trace( 1, "    Nextstate = " << NextState );

        if ( NextState >= 0 ) {
          // store memoryypothesis together with backtrace information
          CLatticeSearchHyp NewHyp;
          NewHyp.m_State = NextState;
          NewHyp.m_StartNode = HypsInPrevNode[ HypIndex ].m_StartNode;
          NewHyp.m_pBackPointerToEdge = pEdge;
          NewHyp.m_BackPointerToEntry = HypIndex;
          Trace( 1, "    New Hyp: " << NewHyp << " stored for node " << n );
          m_Hyps.GetRow( n ).Append( NewHyp );

          // we remember final states to create new edges later on
          if ( pTransducerAccess->IsFinal( NextState ) ) {
            Trace( 1, "    Nextstate = " << NextState << ": Is final" );
            CArray< CTransducerItem *> *pTransducerItems = pTransducerAccess->GetOutput( NextState );

            Trace( 1, "    Result: " );
            for ( int i = pTransducerItems->MinIndex(); i <= pTransducerItems->MaxIndex(); i++ ) {
              //pTransducerItems->GetAt( i )->Display();
              Trace( 3, "  Label is: " << pTransducerItems->GetAt( i )->Label() );
              Trace( 3, "  Translation is: " << pTransducerItems->GetAt( i )->Target() );
            }

            delete pTransducerItems;
            CompleteHyps.Append( NewHyp );
          }// for if

        }
      }
    }

    m_NumberOfActiveHypothesis += m_Hyps.GetRow( n ).Size();

    TRACE(3, " Active Hyps: " <<  m_Hyps.GetRow( n ).Size());


    m_NumberOfCompletePhrases = CompleteHyps.Size();
    Trace( 1, "    Number of complete hypotheses = " << CompleteHyps.Size() );

    for ( int i = CompleteHyps.MinIndex(); i <= CompleteHyps.MaxIndex(); i++ ) {
      Trace( 1, "      Hyp " << i << " " << CompleteHyps[ i ] );
      CreateNewEdgesMS( &CompleteHyps[ i ], pNode, NewEdgesList );
    }

    Trace( 1, "    Numer of complete edges = " << NewEdgesList.Size() );

    CompleteHyps.Clear();

  } // for all nodes

  // Add new edges to Whg
  Trace( 1, "Add new edges to Whg" );
  int NumberOfAddedEdges = 0;
  bool EdgeAdded;

  for( int e = NewEdgesList.MinIndex(); e <= NewEdgesList.MaxIndex(); e++ ) {
    m_pLattice->AddEdge( NewEdgesList.GetAt( e ) );
    NumberOfAddedEdges++;
  }
#ifdef TRACE_ENABLED
  if ( TraceLevel > 0 )
    m_pLattice->Display();
#endif

  return NumberOfAddedEdges;

}



void
CLatticeSearch::SetMaxCostPerWord(double ErrorsPerWord)
{
  Trace( 1, "CLatticeSearch::SetMaxCostPerWord " << ErrorsPerWord );
  m_ErrorsPerWord = ErrorsPerWord;
  m_LengthDependentCosts = TRUE;
}



void
CLatticeSearch::CreateNewEdgesMS( CLatticeSearchHyp* pPhraseHyp, CLatticeNode *pEndNode, 
                                  CArray< CLatticeEdgeMS * > &NewEdgesList )
{
  Trace( 2, "CLatticeSearch::AddEdge" << *pPhraseHyp );
  // a phrase can have several translations; collect their positions

  CArray< TPosition > SentencePositions;
  //TPosition pos;
  //int index;

  CLatticeEdgeMS *pNewEdge;
  //CLatticeEdgeTrans *pLastEdge;
  
  // to collect edges which are covered by the new edge
  CArray< CLatticeEdgeMS * > *pWhgEdges = new CArray< CLatticeEdgeMS * >;

  CWordInfo *pWordInfo;

  CTransducerItem *pTransducerItem;

  //CString SentF, SentE, Label, Score;

  double NewScore = 0.0;  // add scores of edges of covered path

  //const CSentence *pSentenceF, *pSentenceE, *pLabel, *pScore;

  //CWordIndex l;  // for label
  //CWordIndex S;  // for score

  CLatticeNode *pStartNode;

  //const CVocabulary *pVocF = m_pVocabulary;
  //const CVocabulary *pVocE = m_pVocabulary;

  // collect the edges of the partial path corresponding to current hypothesis
  ReconstructPathMS( pPhraseHyp, pEndNode, pWhgEdges );

  CArray< CTransducerItem *> *pTransducerItems = m_pTransducerAccess->GetOutput( pPhraseHyp->m_State );
	    
  Trace( 2, "Hypothesis: " );
  for ( int i = pTransducerItems->MinIndex(); i <= pTransducerItems->MaxIndex(); i++ ) {
    pTransducerItem = pTransducerItems->GetAt( i );
    //m_pTransducerAccess->Display( pTransducerItem );
    Trace( 2, "  Label is:  " << pTransducerItems->GetAt( i )->Label() );
    Trace( 2, "  Source is: " << pTransducerItems->GetAt( i )->Source() );
    Trace( 2, "  Transl is: " << pTransducerItems->GetAt( i )->Target() );


    // create a new edge
    // where did the path start
    pStartNode = m_pLattice->Nodes()[ pPhraseHyp->m_StartNode ];
    pNewEdge = new CLatticeEdgeMS( pStartNode, pEndNode );
    CString Label = CString( (pTransducerItems->GetAt( i )->Label()).c_str() );
    
    pNewEdge->Init( pStartNode, pEndNode, 
                    pStartNode->Number(), pEndNode->Number(), 
                    Label,
                    pPhraseHyp->m_AcuScore,
                    pPhraseHyp->m_StartNode, pEndNode->Number(),
                    ""
                    );

    // store which transducer rule generated this new edge
    pNewEdge->SetTransducer( m_pTransducerAccess->Transducer() );
    pNewEdge->SetTransducerItem( pTransducerItems->GetAt( i ) );

    // store which edges are covered by this new edge
    // same path by several edges for different translations
    // therefore copies of the array
    
    /*
    //
    //ARV 05/02 
    //dont need to do any of this stuff since this array is only used in TM_Graph_HMM
    CArray< CLatticeEdgeMS * > *pWhgEdgesCopy = new CArray< CLatticeEdgeMS * >;
    (*pWhgEdgesCopy) = (*pWhgEdges);
    Trace( 5, "Generating Edges copied: " << pWhgEdges->Size() << " " << pWhgEdgesCopy->Size() );
      
      
    //this can be done with the pointer to trans Items
    CArray<CLatticeEdgeTrans*> pWhgEdgesCopyTmp;
    for (int i=pWhgEdgesCopy->MinIndex(); i<=pWhgEdgesCopy->MaxIndex(); i++){
    CLatticeEdgeMS* tmp = pWhgEdgesCopy->GetAt(i);
    pWhgEdgesCopyTmp[i] = (CLatticeEdgeTrans*)tmp;
    }
    //pointers to the same edges as before
    pNewEdge->SetGeneratingEdges( &pWhgEdgesCopyTmp );
    */

    // collect words and wordinfos over covered partial path
    // and form it into a wordinfo for the new edge
    // at the same time, collect scores
    std::ostringstream Stream;
    
    // first source string
    for ( int j = pWhgEdges->MinIndex(); j <= pWhgEdges->MaxIndex(); j++ ) {
      Stream << pWhgEdges->GetAt( j )->Word() << " ";
      Stream << *(pWhgEdges->GetAt( j )->WordInfo()) << " ";
    }
    
    // add the target string
    Stream << "# ";
    //Stream << pTransducerItems->GetAt( i )->Target() << " ";
    NewScore = 0;
    WriteTargetStringMS( Stream, m_pTransducerAccess->TargetSentence( pTransducerItem ), pWhgEdges, NewScore );

    Trace( 2, "Score from parts = " << NewScore );



    //ARV 04/29
    //get the basic scores thru the acces, but then finally just set the model scores for the edge
    
    Trace( 1, "Probability for current pattern = " << m_pTransducerAccess->Score( pTransducerItem ) );
    double Prob = m_pTransducerAccess->Score( pTransducerItem );

    double TMScore;
    if ( Prob > 0 ) {
      TMScore = -log( Prob );

    } else {
      TMScore = -10;
    }
    
    
    // ???
    double CCScore = 0.0;
    //double CCScore = -0.1 - 0.1 * pTransducerItems->GetAt( i )->SourceLength();
    //double CCScore = -1;  // benefit for new level

    //Trace( 1, "TM-Score and CC-Score for current pattern = " << TMScore << " " << CCScore );
    
    NewScore += ( TMScore + CCScore );
    Stream << "# " << NewScore << " ";

    // now, convert it into new word info
    // !!! word info or no word info
    pWordInfo = pNewEdge->WordInfo();

    
    std::istringstream IStream;
    IStream.str(Stream.str());
    IStream >> *pWordInfo; 

    //cout<<"NewWordInfo: " << *pWordInfo<<endl;
    
    // set error score
    Trace( 2, "New Edge" );

    // sv 060517: do the log when loading the transducer, not here
    pNewEdge->Scores() = pTransducerItem->m_ModelScores;
    
    
    // add multiple source edge scores here (apd)
    for ( int s = pPhraseHyp->m_Scores.MinIndex(); 
             s <= pPhraseHyp->m_Scores.MaxIndex(); s++)
    {
      pNewEdge->Scores().Append( pPhraseHyp->m_Scores[ s ] );
    }
    
//    pNewEdge->Scores().Append( pPhraseHyp->m_AcuScore );
    
    /*
    //ARV 04/28
    //set the model scores, after taking the negative log of what is in the raw transducer    
    CNumericalArray<float> LogModelScores;
    for ( int m = pTransducerItem->m_ModelScores.MinIndex(); m <= pTransducerItem->m_ModelScores.MaxIndex(); m++ ){
    LogModelScores[ m ] = -log( pTransducerItem->m_ModelScores[ m ] );
    }
    pNewEdge->Scores() = LogModelScores;
    //pNewEdge->Display();
    */

    NewEdgesList.Append( pNewEdge );
  }

  delete pWhgEdges;
  delete pTransducerItems;
 
}



void
CLatticeSearch::CreateNewEdges( CLatticeSearchHyp* pPhraseHyp, CLatticeNode *pEndNode, 
                                CArray< CLatticeEdgeTrans * > &NewEdgesList )
{
  Trace( 2, "CLatticeSearch::AddEdge" << *pPhraseHyp );
  // a phrase can have several translations; collect their positions

  CArray< TPosition > SentencePositions;
  CLatticeEdgeTrans *pNewEdge;
  CLatticeNode *pStartNode;
  
  // to collect edges which are covered by the new edge
  CArray< CLatticeEdgeTrans * > *pWhgEdges = new CArray< CLatticeEdgeTrans * >;

  CWordInfo *pWordInfo;
  CTransducerItem *pTransducerItem;
  double NewScore = 0.0;  // add scores of edges of covered path

  // collect the edges of the partial path corresponding to current hypothesis
  ReconstructPath( pPhraseHyp, pEndNode, pWhgEdges );

  CArray< CTransducerItem *> *pTransducerItems = m_pTransducerAccess->GetOutput( pPhraseHyp->m_State );
	    
  Trace( 2, "Hypothesis: " );
  for ( int i = pTransducerItems->MinIndex(); i <= pTransducerItems->MaxIndex(); i++ ) {
    pTransducerItem = pTransducerItems->GetAt( i );
    //m_pTransducerAccess->Display( pTransducerItem );
    Trace( 2, "  Label is:  " << pTransducerItems->GetAt( i )->Label() );
    Trace( 2, "  Source is: " << pTransducerItems->GetAt( i )->Source() );
    Trace( 2, "  Transl is: " << pTransducerItems->GetAt( i )->Target() );


    // create a new edge
    // where did the path start
    pStartNode = m_pLattice->Nodes()[ pPhraseHyp->m_StartNode ];
    pNewEdge = new CLatticeEdgeTrans( pStartNode, pEndNode );
    CString Label = CString( (pTransducerItems->GetAt( i )->Label()).c_str() );
    
    pNewEdge->Init( pStartNode, pEndNode, 
                    pStartNode->Number(), pEndNode->Number(), 
                    Label,
                    pPhraseHyp->m_AcuScore,
                    pPhraseHyp->m_StartNode, pEndNode->Number(),
                    ""
                    );

    // store which transducer rule generated this new edge
    pNewEdge->SetTransducer( m_pTransducerAccess->Transducer() );
    pNewEdge->SetTransducerItem( pTransducerItems->GetAt( i ) );

    // store which edges are covered by this new edge
    // same path by several edges for different translations
    // therefore copies of the array
    CArray< CLatticeEdgeTrans * > *pWhgEdgesCopy = new CArray< CLatticeEdgeTrans * >;
    (*pWhgEdgesCopy) = (*pWhgEdges);
    Trace( 5, "Generating Edges copied: " << pWhgEdges->Size() << " " << pWhgEdgesCopy->Size() );
    pNewEdge->SetGeneratingEdges( pWhgEdgesCopy );

    // collect words and wordinfos over covered partial path
    // and form it into a wordinfo for the new edge
    // at the same time, collect scores
    std::ostringstream Stream;
    
    // first source string
    for ( int j = pWhgEdges->MinIndex(); j <= pWhgEdges->MaxIndex(); j++ ) {
      Stream << pWhgEdges->GetAt( j )->Word() << " ";
      Stream << *(pWhgEdges->GetAt( j )->WordInfo()) << " ";
    }
    
    // add the target string
    Stream << "# ";
    //Stream << pTransducerItems->GetAt( i )->Target() << " ";
    NewScore = 0;
    WriteTargetString( Stream, m_pTransducerAccess->TargetSentence( pTransducerItem ), pWhgEdges, NewScore );

    Trace( 2, "Score from parts = " << NewScore );


    // add the score
    //Trace( 1, "Probability for current pattern = " << pTransducerItems->GetAt( i )->Score() );
    //double Prob = pTransducerItems->GetAt( i )->Score();
    Trace( 1, "Probability for current pattern = " << m_pTransducerAccess->Score( pTransducerItem ) );
    double Prob = m_pTransducerAccess->Score( pTransducerItem );
    double TMScore;
    if ( Prob > 0 ) {
      TMScore = -log( Prob );

    } else {
      TMScore = -10;
    }
    
    // ???
    double CCScore = 0.0;
    //double CCScore = -0.1 - 0.1 * pTransducerItems->GetAt( i )->SourceLength();
    //double CCScore = -1;  // benefit for new level

    //Trace( 1, "TM-Score and CC-Score for current pattern = " << TMScore << " " << CCScore );
    
    NewScore += ( TMScore + CCScore );
    Stream << "# " << NewScore << " ";

    // now, convert it into new word info
    // !!! word info or no word info
    pWordInfo = pNewEdge->WordInfo();

    
    std::istringstream IStream;
    IStream.str(Stream.str());
    IStream >> *pWordInfo; 

    //cout<<"NewWordInfo: " << *pWordInfo<<endl;
    
    // set error score
    Trace( 2, "New Edge" );
    //pNewEdge->Display();

    NewEdgesList.Append( pNewEdge );
  }

  delete pWhgEdges;
  delete pTransducerItems;

}



void
CLatticeSearch::ReconstructPathMS( const CLatticeSearchHyp *pHyp, CLatticeNode *pEndNode, 
                                   CArray <CLatticeEdgeMS *> *pWhgEdges )
{
  // reconstruct path through lattice according to backtrace information
  // we start with hypothesis *pHyp
  // backtrace information is edge running into node and hypothesis in
  // previous node which was extented with word at incoming edge
  // backtracing terminates with empty (start) hypothesis

  Trace( 2, "CLatticeSearch::ReconstructPath" );

  int PrevNode;
  int PrevEntry;
  //CLatticeSearchHyp PrevHyp;
  CString WordsInWhg;
  CLatticeEdge *pEdge;

  //const CVocabulary *pVocF;
  
  Trace( 3, "Start with Hyp = " << *pHyp );
  PrevEntry = pHyp->m_BackPointerToEntry;

  WordsInWhg = CString( pEndNode->Number() );

  while ( PrevEntry != -1 ) {
    pEdge = pHyp->m_pBackPointerToEdge;
    PrevNode = pEdge->FromNode()->Number();
#ifdef TRACE_ENABLED
    Trace( 3, "Edge = " ); 
    if ( TraceLevel >= 3 )
      pEdge->Display();
#endif
    Trace( 3, "PrevNode = " << PrevNode << " PrevEntry = " << PrevEntry );
    Trace( 3, "Word = " << pEdge-> Word() );
    //Trace0( 3, "Edge = " ); pEdge->Display();
    pWhgEdges->Insert( 0, (CLatticeEdgeMS *)pEdge );
    WordsInWhg = CString( PrevNode ) + CString(" ") + pEdge-> Word() + CString(" ") + WordsInWhg;

    pHyp = &(m_Hyps.GetAt( PrevNode, PrevEntry ));
    Trace( 3, "PrevHyp = " << *pHyp );
    PrevEntry = pHyp->m_BackPointerToEntry;
  }

  TRACE ( 3, "Path through nodes in Whg was: " << WordsInWhg );

  Trace( 3, "CLatticeSearch::ReconstructPath done" );
}


void
CLatticeSearch::ReconstructPath( const CLatticeSearchHyp *pHyp, CLatticeNode *pEndNode, 
                                 CArray <CLatticeEdgeTrans *> *pWhgEdges )
{
  // reconstruct path through lattice according to backtrace information
  // we start with hypothesis *pHyp
  // backtrace information is edge running into node and hypothesis in
  // previous node which was extented with word at incoming edge
  // backtracing terminates with empty (start) hypothesis

  Trace( 2, "CLatticeSearch::ReconstructPath" );

  int PrevNode;
  int PrevEntry;
  //CLatticeSearchHyp PrevHyp;
  CString WordsInWhg;
  CLatticeEdgeTrans *pEdge;

  //const CVocabulary *pVocF;
  
  Trace( 3, "Start with Hyp = " << *pHyp );
  PrevEntry = pHyp->m_BackPointerToEntry;

  WordsInWhg = CString( pEndNode->Number() );

  while (PrevEntry != -1) {
    pEdge = pHyp->m_pBackPointerToEdge;
    PrevNode = pEdge->FromNode()->Number();
#ifdef TRACE_ENABLED
    Trace( 3, "Edge = " ); 
    if ( TraceLevel >= 3 )
      pEdge->Display();
#endif
    Trace( 3, "PrevNode = " << PrevNode << " PrevEntry = " << PrevEntry );
    Trace( 3, "Word = " << pEdge-> Word() );
    //Trace0( 3, "Edge = " ); pEdge->Display();
    pWhgEdges->Insert(0, pEdge);
    WordsInWhg = CString( PrevNode ) + CString(" ") + pEdge-> Word() + CString(" ") + WordsInWhg;

    pHyp = &(m_Hyps.GetAt( PrevNode, PrevEntry ));
    Trace( 3, "PrevHyp = " << *pHyp );
    PrevEntry = pHyp->m_BackPointerToEntry;
  }

  TRACE (3, "Path through nodes in Whg was: " << WordsInWhg);

  Trace( 3, "CLatticeSearch::ReconstructPath done" );
}


void 
CLatticeSearch::WriteTargetStringMS( std::ostringstream & Stream, 
                                     CSentence *pSentenceE, 
                                     CArray< CLatticeEdgeMS * > *pWhgEdges,
                                     double &NewScore )
{
  int e, e1, i;
  CWordIndex WordE;
  CWordInfo * pWordInfoF;
  int CountE, CountF;
  CString SentE;
  CString SentF;
 
  for ( e = pSentenceE->MinIndex(); e <= pSentenceE->MaxIndex(); e++ ) {
    WordE = pSentenceE->GetAt( e );
    SentE = m_pVocabulary->GetWord( WordE );
    
    TRACE( 1, "Handle LabelE: " << SentE );
    // look for the corresponding label in source sentence
    if ( SentE[ 0 ] == '@' ) {
      TRACE( 1, "SentE[ 0 ] == '@'" );
      
      // how often have we seen that label before
      CountE = 0;
      e1 = e - 1;
      while ( e1 >= pSentenceE->MinIndex() ) {
        if ( WordE == pSentenceE->GetAt( e1 ) ) {
          CountE++;
        }
        e1--;
      }
      TRACE( 1, "Label with count " << CountE );
      
      CountF = 0;
      
      // run over path which makes the source sentence
      i = pWhgEdges->MinIndex();
      while ( ( i <= pWhgEdges->MaxIndex() ) && CountF <= CountE ) {
        SentF = pWhgEdges->GetAt( i )->Word();
	
        if ( SentF == SentE ) {
          if ( CountF == CountE ) {
            TRACE( 1, "Label " << SentF << " found with count " << CountF );
            pWordInfoF = pWhgEdges->GetAt( i )->WordInfo();
	    
            if ( pWordInfoF ) {
              TRACE( 1, "WordInfo = " << *pWordInfoF );
              NewScore += pWordInfoF->Score();
              TRACE( 1, "New Score = " << NewScore );
	      
              //Stream << *pWordInfoF->TargetString << " ";
              CSentence &S = pWordInfoF->TargetWords();
              if ( TraceLevel > 0 ) {
                cout << "... ";
                S.Display( m_pVocabulary ); cout << " ";
              }
              S.Write( Stream, m_pVocabulary );
              Stream << " ";
            }
            Trace( 2, "No Word Info" );
          }
          CountF++;
	  
        } else {
          Stream << " ";
        }
	
        i++;
      }
    } else {
      //Trace( 0, "No Label" );
      Stream << SentE << " ";
    }
  }

}

void 
CLatticeSearch::WriteTargetString( std::ostringstream & Stream, 
                                   CSentence *pSentenceE, 
                                   CArray< CLatticeEdgeTrans * > *pWhgEdges,
                                   double &NewScore )
{
  int e, e1, i;
  CWordIndex WordE;
  CWordInfo * pWordInfoF;
  int CountE, CountF;
  CString SentE;
  CString SentF;
 
  for ( e = pSentenceE->MinIndex(); e <= pSentenceE->MaxIndex(); e++ ) {
    WordE = pSentenceE->GetAt( e );
    SentE = m_pVocabulary->GetWord( WordE );
    
    TRACE( 1, "Handle LabelE: " << SentE );
    // look for the corresponding label in source sentence
    if ( SentE[ 0 ] == '@' ) {
      TRACE( 1, "SentE[ 0 ] == '@'" );
      
      // how often have we seen that label before
      CountE = 0;
      e1 = e - 1;
      while ( e1 >= pSentenceE->MinIndex() ) {
        if ( WordE == pSentenceE->GetAt( e1 ) ) {
          CountE++;
        }
        e1--;
      }
      TRACE( 1, "Label with count " << CountE );
      
      CountF = 0;
      
      // run over path which makes the source sentence
      i = pWhgEdges->MinIndex();
      while ( ( i <= pWhgEdges->MaxIndex() ) && CountF <= CountE ) {
        SentF = pWhgEdges->GetAt( i )->Word();
	
        if ( SentF == SentE ) {
          if ( CountF == CountE ) {
            TRACE( 1, "Label " << SentF << " found with count " << CountF );
            pWordInfoF = pWhgEdges->GetAt( i )->WordInfo();
	    
            if ( pWordInfoF ) {
              TRACE( 1, "WordInfo = " << *pWordInfoF );
              NewScore += pWordInfoF->Score();
              TRACE( 1, "New Score = " << NewScore );
	      
              //Stream << *pWordInfoF->TargetString << " ";
              CSentence &S = pWordInfoF->TargetWords();
              if ( TraceLevel > 0 ) {
                cout << "... ";
                S.Display( m_pVocabulary ); cout << " ";
              }
              S.Write( Stream, m_pVocabulary );
              Stream << " ";
            }
            Trace( 2, "No Word Info" );
          }
          CountF++;
	  
        } else {
          Stream << " ";
        }
	
        i++;
      }
    } else {
      //Trace( 0, "No Label" );
      Stream << SentE << " ";
    }
  }

}

