#include <config.h>

#include <stdlib.h>

// Larger LineLength Required for SRI-format Confusion Networks
#define MaxLineLength 8192

//long int CLattice::s_NumberOfObjects = 0;
//long int CLatticeNode::s_NumberOfObjects = 0;
//long int CLatticeEdge::s_NumberOfObjects = 0;


class CBackPointer
{
public:
  CBackPointer( CLatticeEdge *pEdge, CBackPointer *pPrevBackPointer );

  CLatticeEdge *Edge() { return m_pEdge; };
  CBackPointer *PrevBackPointer() { return m_pPrevBackPointer; };

private:
  friend class CLattice;
  CLatticeEdge *m_pEdge;
  CBackPointer *m_pPrevBackPointer;
};


inline
CBackPointer::CBackPointer( CLatticeEdge *pEdge, CBackPointer *pPrevBackPointer )
{
  m_pEdge = pEdge;
  m_pPrevBackPointer = pPrevBackPointer;
}



inline void
CLatticeEdge::SetFromNode( CLatticeNode *pFromNode )
{
  m_pFromNode = pFromNode;
}


inline void
CLatticeEdge::SetToNode( CLatticeNode *pToNode )
{
  m_pToNode = pToNode;
}



inline void
CLatticeEdge::SetWord( const CString &Word )
{
  m_Word = Word;
}




inline CLattice::CLattice( CLatticeEdge * pEdgePrototype )
{
  // make a copy of the prototype, will be deleted in the destructor
  m_pEdgePrototype = pEdgePrototype->Clone();

  //m_FirstNode = NULL;
  //m_LastNode = NULL;

  // Nodes und Edges sind Pointer-Arrays, sollen deshalb mit
  // 0 initialisiert werden
  m_Nodes.InitToZero();
  m_Edges.InitToZero();

  m_Nodes.SetIncrement( 100 );
  m_Edges.SetIncrement( 1000 );
  
  m_FileFormat = DEFAULT;
}



inline CLattice::~CLattice()
{
  Clear();
  delete m_pEdgePrototype;
}



inline void
CLattice::RemoveInEdges( CLatticeNode *pNode )
{

  CLatticeEdge *pInEdge;

  const CArray< CLatticeEdge * > &InEdges = pNode->InEdges();

  for ( int i = InEdges.MaxIndex(); i >= InEdges.MinIndex(); i-- ) {
    pInEdge = InEdges[ i ];
    RemoveEdge( pInEdge );
  }

}



inline void
CLattice::RemoveOutEdges( CLatticeNode *pNode )
{

  CLatticeEdge *pOutEdge;

  const CArray< CLatticeEdge * > &OutEdges = pNode->OutEdges();

  for ( int o = OutEdges.MaxIndex(); o >= OutEdges.MinIndex(); o-- ) {
    pOutEdge = OutEdges[ o ];
    RemoveEdge( pOutEdge );
  }

}





inline int
CLattice::NumberOfNodes() const
{
  return m_Nodes.Size();
}



inline int
CLattice::NumberOfIsolatedNodes() const
{
  int N = 0;

  for (int n = m_Nodes.MinIndex(); n <= m_Nodes.MaxIndex(); n++) {
    if ( (m_Nodes[ n ]->InEdges().Size() == 0) && 
	 (m_Nodes[ n ]->OutEdges().Size() == 0) ) {
      N++;
    }
  }

  return N;
}



inline int
CLattice::NumberOfEdges() const
{
  return m_Edges.Size();
}



