#ifndef Prosody_hh
#define Prosody_hh

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <stdio.h>
#include <math.h>
#include <string.h>

#include <String.hh>
#include <iostream>

// ----------------------------------------------------------------
class CProsody
{
public:
  CProsody();
  CProsody( const CProsody &Prosody );

  CProsody & operator = ( const CProsody &Prosody );

  void ReadFromBuffer( const char *Buffer );
  virtual void WriteTo( std::ostream& os );
  void ToString( CString& String ) const;

  double GetBoundary2() { return m_Boundary2;}
  double LogNoBoundary2();

  void GetBoundaries( double &Boundary0, double &Boundary1, double &Boundary2, 
		      double &Boundary3, double &Boundary4 );

  void SetBoundaries( double Boundary0, double Boundary1, double Boundary2, 
		      double Boundar3, double Boundary4 );

  void GetMode( double &Mode0, double &Mode1, double &Mode2 );
  void SetMode( double Mode0, double Mode1, double Mode2 );

  double GetMode2();


private:
  void Init();

  double m_Accent0, m_Accent1;
  double m_Boundary0, m_Boundary1, m_Boundary2, m_Boundary3, m_Boundary4;
  double m_Mode0, m_Mode1, m_Mode2;

};



// ----------------------------------------------------------------
// CProsody - Inlines
// ----------------------------------------------------------------

inline
CProsody::CProsody()
{
  Init();
}



inline
CProsody::CProsody( const CProsody &Prosody )
{
  *this = Prosody;
}


inline CProsody &
CProsody::operator = ( const CProsody &Prosody )
{
  if ( this == &Prosody ) 
    //-- Kopie auf sich selber kann ignoriert werden
    return *this;

  m_Accent0   = Prosody.m_Accent0; 
  m_Accent1   = Prosody.m_Accent1;
  m_Boundary0 = Prosody.m_Boundary0;
  m_Boundary1 = Prosody.m_Boundary1;
  m_Boundary2 = Prosody.m_Boundary2; 
  m_Boundary3 = Prosody.m_Boundary3;
  m_Boundary4 = Prosody.m_Boundary4;
  m_Mode0     = Prosody.m_Mode0;
  m_Mode1     = Prosody.m_Mode1;
  m_Mode2     = Prosody.m_Mode2;

  return *this;
}



inline void
CProsody::Init()
{
  m_Accent0 = 1.0; 
  m_Accent1 = 0.0;
  m_Boundary0 = 1.0;
  m_Boundary1 = 0.0;
  m_Boundary2 = 0.0; 
  m_Boundary3 = 0.0;
  m_Boundary4 = 0.0;
  m_Mode0 = 1.0;
  m_Mode1 = 0.0;
  m_Mode2 = 0.0;  
}





inline double
CProsody::LogNoBoundary2()
{ 
  if ( m_Boundary2 >= 1.0 )
    return log( 0.99999 );
  else
    return log( 1.0 - m_Boundary2 );
}


inline void
CProsody::GetBoundaries( double &Boundary0, double &Boundary1, double &Boundary2,
			 double &Boundary3, double &Boundary4 )
{
  // make shure we have no 0.0 values
  Boundary0 = m_Boundary0;
  Boundary1 = m_Boundary1;
  Boundary2 = m_Boundary2;
  Boundary3 = m_Boundary3;
}



inline void
CProsody::SetBoundaries( double Boundary0, double Boundary1,  double Boundary2,
			 double Boundary3, double Boundary4 )
{
  // make shure we have no 0.0 values
  if ( Boundary0 == 0.0 )
    m_Boundary0 = 1.0e-10;
  else
    m_Boundary0 = Boundary0;

  if ( Boundary1 == 0.0 )
    m_Boundary1 = 1.0e-10;
  else
    m_Boundary1 = Boundary1;

  if ( Boundary2 == 0.0 )
    m_Boundary2 = 1.0e-10;
  else
    m_Boundary2 = Boundary2;

  if ( Boundary3 == 0.0 )
    m_Boundary3 = 1.0e-10;
  else
    m_Boundary3 = Boundary3;

}



inline void
CProsody::GetMode( double &Mode0, double &Mode1, double &Mode2 )
{
  Mode0 = m_Mode0;
  Mode1 = m_Mode1;
  Mode2 = m_Mode2;
}



inline void
CProsody::SetMode( double Mode0, double Mode1, double Mode2 )
{
  m_Mode0 = Mode0;
  m_Mode1 = Mode1;
  m_Mode2 = Mode2;
}


inline double
CProsody::GetMode2()
{
  return m_Mode2;
}


inline void
CProsody::ReadFromBuffer( const char *Buffer )
{
  // p is of the form: (PR (A  0.28 0.28) (G  0.28 0.99 0.28 0.99) (M  0.44 0.44 0.44))

  //  Trace( 1, "CProsody::ReadFromBuffer " << Buffer );

  if( ( strncmp( Buffer, "(PR ", 4 ) == 0 ) 
      && ( strstr( Buffer, "(S ") != NULL ) 
      && ( strstr( Buffer, "(F ") != NULL )
      && ( strstr( Buffer, "(A ") != NULL ) ) {
    // Prosodie scheint ok zu sein
    char buffer[ 4096 ];
    char *po;
    
    // Accent
    strcpy( buffer, strstr( Buffer, "(A " ) );
    po = buffer;
    while ( *po != ')' ) {
    po++;
    }
    po++;
    *po = '\0';
    //Trace( 1, "Accent: " << buffer );
    sscanf( buffer, "(A%lf%lf)" , &m_Accent0, &m_Accent1 );
    
    // Mode
    strcpy( buffer, strstr( Buffer, "(F " ) );
    po = buffer;
    while ( *po != ')' ) {
      po++;
    }
    po++;
    *po = '\0';
    //Trace( 1, "Mode: " << buffer );
    sscanf( buffer, "(F%lf%lf%lf)", &m_Mode0, &m_Mode1, &m_Mode2 );
    
    // Boundaries
    strcpy( buffer, strstr( Buffer,"(S " ) );
    po = buffer;
    while ( *po != ')' ) {
      po++;
    }
    po++;
    *po = '\0';
    //Trace( 1, "Boundaries: " << buffer );
    sscanf( buffer,"(S%lf%lf%lf%lf%lf)",&m_Boundary0, &m_Boundary1, &m_Boundary2, &m_Boundary3, &m_Boundary4 );
  }

  //  Trace(1, "Boundaries: " << m_Boundary0 << " " << m_Boundary1 << " " 
  //	<< m_Boundary2 << " " << m_Boundary3 << " " << m_Boundary4 );

  //Trace(1, "Prosody Found " << m_Accent0  << " " << m_Accent1 << " " 
  //<< m_Mode0 << " " << m_Mode1 << " " << m_Mode2 << " "
  //<< m_Boundary0 << " " << m_Boundary1 << " " << m_Boundary2 << " " << m_Boundary3 );
}


inline void
CProsody::WriteTo( std::ostream& os )
{
  // p is of the form: (PR (A  0.28 0.28) (G  0.28 0.99 0.28 0.99) (M  0.44 0.44 0.44))

  os <<  "(PR (A " << m_Accent0  << " " << m_Accent1 << ") (G " 
     << m_Boundary0 << " " << m_Boundary1 << " " << m_Boundary2 << " " << m_Boundary3 << ") (M "
	<< m_Mode0 << " " << m_Mode1 << " " << m_Mode2 << "))";

}



inline void
CProsody::ToString( CString& String ) const
{
  // p is of the form: (PR (A  0.28 0.28) (G  0.28 0.99 0.28 0.99) (M  0.44 0.44 0.44))

  String += CString( "(PR (A " ) + CString( m_Accent0, "%3.2f" ) 
    + CString( " " ) + CString( m_Accent1, "%3.2f" ) 
    + CString( ") (G " ) + CString( m_Boundary0, "%3.2f" ) 
    + CString( " " ) + CString( m_Boundary1, "%3.2f" )
    + CString( " " ) + CString( m_Boundary2, "%3.2f" ) 
    + CString( " " ) + CString( m_Boundary3, "%3.2f" )
    + CString( ") (M " ) + CString( m_Mode0, "%3.2f" ) 
    + CString( " " ) + CString( m_Mode1, "%3.2f" )
    + CString( " " ) + CString( m_Mode2, "%3.2f" ) 
    + CString( "))" );
}


#endif

