#include <SentenceToLattice.hh>
#include <LatticeEdgeMS.hh>


// needs to be changed to use only CLattice

CLatticeLM *
CSentenceToLattice::CreateLatticeFromSentence( const CSentence *pSentence, const CVocabulary *pVoc, TLatticeType LatticeType )
{
  if ( LatticeType == Flat ) 
    return CreateLatticeFromSentenceFlat( pSentence, pVoc );
  else
    return CreateLatticeFromSentenceFull( pSentence, pVoc );

}



CLatticeLM *
CSentenceToLattice::CreateLatticeFromSentenceFlat( const CSentence *pSentence, const CVocabulary *pVoc )
{
  //cout << "CSentenceToLattice::CreateLatticeFromSentenceFlat"  << endl;


  //CLatticeEdgeMS *pLatticeEdge = new CLatticeEdgeMS;
  //CLattice *pLattice = new CLattice( pLatticeEdge );
  CWordInfo      *pWordInfo;
  CLatticeEdgeMS *pLatticeEdge;
  CLatticeLM     *pLattice = new CLatticeLM( new CLatticeEdgeMS, NULL );
  CArray< CString > Words;

  int NodeNumber = 0;
  int StartNode;
    
  // loop over all words in the sentence and create edges
  for ( int j = pSentence->MinIndex(); j <= pSentence->MaxIndex(); j++ ) {
    //cout << "Word in Sentence at j is: " << pSentence->Word( j, pVoc );
    //cout << endl;
    
    pWordInfo = pSentence->WordInfo( j );
    
    pLatticeEdge = new CLatticeEdgeMS;
    
    // init the edge with word
    CString Word = pSentence->Word( j, pVoc );
    //cout << Word << endl;
    pLatticeEdge->SetWord( Word );
    //cout << "Word in edge is: " << pLatticeEdge->Word() << endl;
    
    if ( pWordInfo ) {
      CWordInfo *pNewWordInfo = new CWordInfo( pWordInfo );
      pLatticeEdge->SetWordInfo( pNewWordInfo );
    }
    //cout << "Add edge from StartNode " << StartNode << " to EndNode " << NodeNumber << endl;
    pLattice->AddEdge( NodeNumber, NodeNumber + 1, pLatticeEdge );
    
    NodeNumber++;
    
    //cout << endl;
  }

  return pLattice;

}




CLatticeLM *
CSentenceToLattice::CreateLatticeFromSentenceFull( const CSentence *pSentence, const CVocabulary *pVoc )
{
  //cout << "CSentenceToLattice::CreateLatticeFromSentence"  << endl;


  //CLatticeEdgeMS *pLatticeEdge = new CLatticeEdgeMS;
  //CLattice *pLattice = new CLattice( pLatticeEdge );
  CWordInfo      *pWordInfo;
  CLatticeEdgeMS *pLatticeEdge;
  CLatticeLM     *pLattice = new CLatticeLM( new CLatticeEdgeMS, NULL );
  CArray< CString > Words;

  int NodeNumber = 0;
  int StartNode;
    
  // loop over all words in the sentence and create edges
  for ( int j = pSentence->MinIndex(); j <= pSentence->MaxIndex(); j++ ) {
    //cout << "Word in Sentence at j is: " << pSentence->Word( j, pVoc );
    //cout << endl;
    
    pWordInfo = pSentence->WordInfo( j );
    
    if ( pWordInfo ) {

      //cout << "WordInfo is: " << *(pWordInfo) << endl;

      // if we have several source words, create an edge for each of them
      //CSentence &SourceWords = pWordInfo->SourceWords();
      //for ( int i = SourceWords.MinIndex(); i <= SourceWords.MaxIndex(); i++ ) {
      //cout << "SourceWords " << SourceWords[ i ] << endl;

      // need to store the start node
      StartNode = NodeNumber;

      if ( pWordInfo->HasSourceString() ) {

        // create edges for underlying words, ie. those in the annotation
        pWordInfo->GetSourceWords( Words );
        for ( int k = Words.MinIndex(); k <= Words.MaxIndex(); k++ ) {
          //cout << "Source Words is: " << Words[ k ] << endl;

          // create lattice edge
          pLatticeEdge = new CLatticeEdgeMS;
          pLatticeEdge->SetWord( Words[ k ] );
          //cout << "Add edge from StartNode " << StartNode << " to EndNode " << NodeNumber << endl;
          pLattice->AddEdge( NodeNumber, NodeNumber + 1, pLatticeEdge );
          NodeNumber++;
        }

        // now we create the spanning edge
        //cout << "spanning edge" << endl;

        pLatticeEdge = new CLatticeEdgeMS;
        
        // init the edge with word
        CString Word = pSentence->Word( j, pVoc );
        //cout << Word << endl;
        pLatticeEdge->SetWord( Word );
        //cout << "Word in edge is: " << pLatticeEdge->Word() << endl;
    
        if ( pWordInfo ) {
          CWordInfo *pNewWordInfo = new CWordInfo( pWordInfo );
          pLatticeEdge->SetWordInfo( pNewWordInfo );
        }
        //cout << "Add edge from StartNode " << StartNode << " to EndNode " << NodeNumber << endl;
        pLattice->AddEdge( StartNode, NodeNumber, pLatticeEdge );
      
      }

    } else {
      
      //cout << "no word info given" << endl;

      // just the source word
      pLatticeEdge = new CLatticeEdgeMS;
      
      // init the edge with word
      CString Word = pSentence->Word( j, pVoc );
      pLatticeEdge->SetWord( Word );
      
      //cout << "Add edge from StartNode " << StartNode << " to EndNode " << NodeNumber << endl;
      pLattice->AddEdge( NodeNumber, NodeNumber + 1, pLatticeEdge );
      
      NodeNumber++;
    }
    
    //cout << endl;
  }

  return pLattice;


}


