#ifndef __SentenceToLattice__
#define __SentenceToLattice__


#include <LatticeLM.hh>

/**
   Convert a sentence into a lattice.
   In a sentence a word can have annotations like: this is a @NUM { number # Zahl # 1.0 } .
   We want to be able to convert this into a lattice which has an edge for 'number' and
   also an edge (between the same nodes) for '@NUM'.  This is full conversion.  The alternative
   is flat conversation, where only an edge for the Tag is generated, which has then also
   the annotation, but no edges are generated for the proper words.

   Bug: Currently only one level of annotation is read, even so the annotation structure is recursive.
   Bug: Currently a CLatticeLM is generated, as this is needed elsewhere in the decoder, but 
        should be just a CLattice.
*/


class CSentenceToLattice
{
public:
  typedef enum { Full, Flat } TLatticeType;

  CLatticeLM * CreateLatticeFromSentence( const CSentence *pSentence, const CVocabulary *pVoc, TLatticeType LatticeType = Full );

private:

  CLatticeLM * CreateLatticeFromSentenceFull( const CSentence *pSentence, const CVocabulary *pVoc );
  CLatticeLM * CreateLatticeFromSentenceFlat( const CSentence *pSentence, const CVocabulary *pVoc );

};




#endif
