#define NO_TRACE 1
#include <Trace.hh>

#include <PhraseAlign.hh>
#include <Text.hh>

using namespace std;

CPhraseAligner::CPhraseAligner()
{
  m_pLexiconSourceTarget = NULL;
  m_pLexiconTargetSource = NULL;

  m_AlignBidirectional = false;

  m_StretchFactor = 1.0;
  m_FactorLonger = 1.5;
  m_FactorShorter = 0.66;

}


CPhraseAligner::~CPhraseAligner()
{
}


void
CPhraseAligner::Init( CLexicon *pLexiconSourceTarget, CLexicon *pLexiconTargetSource )
{
  m_pLexiconSourceTarget = pLexiconSourceTarget;
  m_pLexiconTargetSource = pLexiconTargetSource;
}



void
CPhraseAligner::Renormalize()
{
  double SumP;
  int i, j;

  // renormalize for source to target direction
  for ( j = 1; j <= m_J; j++ ) {
    SumP = 0.0;
    for ( i = 1; i <= m_I; i++ ) {
      SumP += m_ProbsSourceTarget[ j ][ i ];
    }
    for ( i = 1; i <= m_I; i++ ) {
      m_ProbsSourceTarget[ j ][ i ] = m_ProbsSourceTarget[ j ][ i ] / SumP;
    }
  }

  // renormalize for target to source direction
  // i.e. for each target word sum probs over source words 
  if ( m_AlignBidirectional ) {
    for ( i = 1; i <= m_I; i++ ) {
      SumP = 0.0;
      for ( j = 1; j <= m_J; j++ ) {
        SumP += m_ProbsTargetSource[ j ][ i ];
      }
      for ( j = 1; j <= m_J; j++ ) {
        m_ProbsTargetSource[ j ][ i ] = m_ProbsTargetSource[ j ][ i ] / SumP;
      }
    }
  }

/*
  for ( i = m_I; i >= 1; i-- ) {
    for ( j = 1; j <= m_J; j++ ) {
      cout.width(8);
      cout.precision(2);
      cout << m_ProbsSourceTarget[ j ][ i ] << "  ";
    }
    cout << endl;
  }
  cout << endl;

  for ( i = m_I; i >= 1; i-- ) {
    for ( j = 1; j <= m_J; j++ ) {
      cout.width(8);
      cout.precision(2);
      cout << m_ProbsTargetSource[ j ][ i ] << "  ";
    }
    cout << endl;
  }
*/
}

bool
CPhraseAligner::AlignPhrase( 
                      CSentence *pSourceSentence, CSentence *pTargetSentence,
                      int StartSource, int EndSource,
                      CNumericalArray< int > & Min_i, CNumericalArray< int > & Max_i,
                      CNumericalArray< double > & Scores  )
{
  assert( 0 );
}



// =======================================================================
CPhraseAlignerIBM1::CPhraseAlignerIBM1()
{
  // currently, nothing to be done here
}


CPhraseAlignerIBM1::~CPhraseAlignerIBM1()
{
  // currently, nothing to be done here
}


bool
CPhraseAlignerIBM1::AlignPhrase( 
                      CSentence *pSourceSentence, CSentence *pTargetSentence,
                      int StartSource, int EndSource,
                      CNumericalArray< int > & Min_i, CNumericalArray< int > & Max_i,
                      CNumericalArray< double > & Scores  )
{

  // use IMB1 alignment model, i.e. only lexicon to
  // find i1 and i2 on target side such that the alignment
  // probability is highest when aligning f_j1 ... f_j2 to
  // e_i1 ... e_i2, and aligning outside f_j to outside e_i
  // i.e. caculate argmax_i1_i2 { Prod_j1...j2 Sum_i1...i2 p(f_j|e_i)
  // x Prod_j!=j1...j2 Sum_i!=i1...i2 p(f_j|e_i) }


  //cout << "AlignPhrase " << StartSource << " " << EndSource << endl;

  // some local variable
  CWordIndex f, e;
  int i, j, k, min_i, max_i;

  // in case there are old results in the arrays
  Min_i.Clear();
  Max_i.Clear();
  Scores.Clear();

  // length of the sentences
  m_J = pSourceSentence->Length();
  m_I = pTargetSentence->Length();
  //cout << "Length source " << m_J << " length target " << m_I << endl;

  // if sentences are longer than we can store probabilities, we skip it
  if ( (m_J > MAX_SENTENCE_LENGTH) && (m_I > MAX_SENTENCE_LENGTH) ) {
    return false;
  }

  // first, get all the probabilities
  //CMatrix< float > m_Probs;
  for ( j = 1; j <= m_J; j++ ) {
    for ( i = 1; i <= m_I; i++ ) {
      f = pSourceSentence->GetAt( j );
      e = pTargetSentence->GetAt( i );
      m_ProbsSourceTarget[ j ][ i ] = m_pLexiconSourceTarget->Pr( f, e );
	  //cout << "j = " << j << " i = " << i 
	  //	 << " f = " << f << " e = " << e << " p = " << m_pLexiconSourceTarget->Pr( f, e )
	  //   << " stored as " << m_ProbsSourceTarget[ j ][ i ] << endl;
    }
  }

  // also for target to source alignemt
  if ( m_AlignBidirectional && m_pLexiconTargetSource) {
    for ( j = 1; j <= m_J; j++ ) {
      for ( i = 1; i <= m_I; i++ ) {
        f = pSourceSentence->GetAt( j );
        e = pTargetSentence->GetAt( i );
        m_ProbsTargetSource[ j ][ i ] = m_pLexiconTargetSource->Pr( e, f );
		//cout << "j = " << j << " i = " << i 
		// << " f = " << f << " e = " << e << " p = " << m_pLexiconSourceTarget->Pr( f, e )
		// << " stored as " << m_ProbsSourceTarget[ j ][ i ] << endl;
      }
    }
  }

  //cout << "matrices initialized" << endl;

  // renormalize probabilities for the current sentence pair
  //if ( m_Renormalize ) {
  Renormalize();
  //}
  //cout << "and renormalized" << endl;

  /*
  for ( i = m_I; i >= 1; i-- ) {
    cout << i;
	for ( j = 1; j <= m_J; j++ ) {
	  	  cout << " " << m_ProbsSourceTarget[ j ][ i ];
    }
	cout << endl;
  }
  */

  // find 'center' of target phrase for source phrase
  // if target phrase is at sentence beginning or sentence end
  // the center might be off.  Therefore iteratate over smaller
  // and smaller target sentence areas
  int n = 0;
  double P, SumP, BestP;
  int Best_i;
  int SumBest_i = 0;

  int i1 = 1;
  int i2 = m_I;
  float i_mid;

  //cout << "Source phrase positions: " << StartSource << " " << EndSource << endl;

  do {

	SumP = 0.0;
	SumBest_i = 0;

	for ( j = StartSource; j <= EndSource; j++ ) {
	  BestP = 0.0;
	  for ( i = i1; i <= i2; i++ ) {
		n++;
		P = m_ProbsSourceTarget[ j ][ i ];
		SumP += i * P;
		//cout << j << " " << i << " " << P << " " << BestP << endl;
		if ( P > BestP ) {
		  BestP = P;
		  Best_i = i;
		  //cout << "new best i " << Best_i << endl;
		}
	  }
	  //cout << "j = " << j << " best i = " << Best_i << endl;
	  SumBest_i += Best_i;
	}

	i_mid = SumBest_i / (EndSource - StartSource + 1);
	//cout << "SumBest_i = " << SumBest_i << " divide by source phrase length = " << (EndSource - StartSource + 1) << endl;
	//cout << "i1 = " << i1 << " i2 = " << i2 << " i_mid = " << i_mid << endl;

	if ( i_mid > ((i2 + i1 + 1 ) / 2) ) {
	  //cout << " i_mid > ((i2 + i1 + 1 ) / 2) = " << (i2 + i1 + 1 ) / 2 << " i1++" << endl;
	  i1++;
	} else {
	  //cout << " i_mid <= ((i2 + i1 + 1 ) / 2) = " << (i2 + i1 + 1 ) / 2 << " i2--" << endl;
	  i2--;
	}

  } while ( (i2 - i1 ) > (EndSource - StartSource + 1) * 1.5 );

  //cout << "Center of target phrase is " 
//	<< SumBest_i 
//	<< " " << StartSource  << " " << i_mid << endl;


  // don't want too short or too long translations
  // on average the target length should be source length times stretch factor
  // where stretch factor is the ratio target corpus lenght to source corpus length
  // then target should not be longer then this average length time a elongation factor
  // and not shorter then the average length time a shortening factor
  // we allow for and addition +/-2 to have e.g. 1->3 alignemnts
  int MinTargetLength = (int)(-2 + ( EndSource - StartSource + 1 ) * m_StretchFactor * m_FactorShorter);
  int MaxTargetLength = (int)(2 + ( EndSource - StartSource + 1 ) * m_StretchFactor * m_FactorLonger);
  if ( MinTargetLength < 1 )
    MinTargetLength = 1;

  //cout << "MinTargetLength = " << MinTargetLength << " MaxTargetLength = " << MaxTargetLength <<endl;


  CSortedArray< double > Probs;
  // sort smallest first
  //Probs.SortReverse();


  // expand the target phrase around the center
  double LogP, LogP1, LogP2;
  bool WasNew;
  int mid = int( i_mid + 0.5 );

  for ( min_i = mid - MaxTargetLength; min_i <= mid + 1; min_i++ ) {
    for ( max_i = mid - 1; max_i <= mid + MaxTargetLength; max_i++ ) {

	  //cout << "min_i = " << min_i << " max_i = " << max_i << " m_I = " << m_I << endl;

	  // currently we do not assume that we have <sos> and <eos>
      if ( (min_i >= 1) && (max_i <= m_I) 
           && ((max_i - min_i + 1) >= MinTargetLength)
		   && ((max_i - min_i + 1) <= MaxTargetLength) ) {

        LogP1 = CalcAlignLogProbSourceTarget( StartSource, EndSource, min_i, max_i ) ;

        // now, we calculate the log prob for the phrase
        // dividing by the sentence length and multiplying by the phrase length
        LogP1 /= m_J;
        LogP1 *= ( EndSource - StartSource + 1 );

        if ( m_AlignBidirectional && m_pLexiconTargetSource ) {
          LogP2 = CalcAlignLogProbTargetSource( StartSource, EndSource, min_i, max_i );

          // now, we calculate the log prob for the phrase
          // dividing by the sentence length and multiplying by the phrase length
          LogP1 /= m_I;
          LogP1 *= ( max_i - min_i + 1 );

          LogP = ( 1 - m_Lambda ) * LogP1 + m_Lambda * LogP2;
        } else {
          LogP2 = -1.0;
          LogP = LogP1;
        }

        //cout << "LogP f->e = " << LogP1 << "  LogP e->f = " << LogP2
	//	   << " LogP bi = " << LogP << endl;

        k = Probs.Insert( LogP, WasNew );
		if ( WasNew ) {
		  Min_i.Insert( k, min_i );
		  Max_i.Insert( k, max_i );
		  Scores.Insert( k, LogP );
		}

        //cout << "P( " << min_i << ", " << max_i << " ) = " << LogP;
        //cout << "  inserted at " << k << endl;
      } else {
        ; //cout << "skip" << endl;
      }
    }
  }

  return true;
}



float
CPhraseAlignerIBM1::CalcAlignLogProbSourceTarget( int j1, int j2, int i1, int i2 ) 
{
  int i, j;
  double P;
  double LogP;
  double TotalLogP = 0.0;

  //cout << "Calc alignment log prob for " << j1 << " " << j2 << " " << i1 << " " << i2 << endl;
  //cout << m_J << " " << m_I << endl;

  for ( j = 1; j <= m_J; j++ ) {

	//cout << " j = " << j << " i = " << i << endl;

    P = 0.0;
    if ( (j >= j1) && (j <= j2) ) {
      // within phrase pair
      for ( i = i1; i <= i2; i++ ) {
        P += m_ProbsSourceTarget[ j ][ i ];
         //cout << j << " " << i << " " << P << endl;
      }
    } else {
      // outside of phrase pair
      for ( i = 1; i <= m_I; i++ ) {
        if ( (i < i1) || (i > i2) ) {
          P += m_ProbsSourceTarget[ j ][ i ];
          //cout << j << " " << i << " " << P << endl;
        }
      }
    }        

    LogP = -log( P );
    TotalLogP += LogP;
    //cout << "j = " << j << " P = " << P << " " << " LogP = " << LogP 
    //     << " TotalLogP = " << TotalLogP << endl;
  }

  return TotalLogP;
}



float
CPhraseAlignerIBM1::CalcAlignLogProbTargetSource( int j1, int j2, int i1, int i2 ) 
{
  //cout << "Calc reverse alignment log prob for " << j1 << " " << j2 << " " << i1 << " " << i2 << endl;
  //cout << m_J << " " << m_I << endl;

  int i, j;
  double P;
  float LogP;
  float TotalLogP = 0.0;

  for ( int i = 1; i <= m_I; i++ ) {

    P = 0.0;
    if ( (i >= i1) && (i <= i2) ) {
      // within phrase pair
      for ( j = j1; j <= j2; j++ ) {
        P += m_ProbsTargetSource[ j ][ i ];
        //cout << j << " " << i << " " << P << endl;
      }
    } else {
      // outside of phrase pair
      for ( j = 1; j <= m_J; j++ ) {
        if ( (j < j1) || (j > j2) ) {
          P += m_ProbsTargetSource[ j ][ i ];
          //cout << j << " " << i << " " << P << endl;
        }
      }
    }

    LogP = -log( P );
    TotalLogP += LogP;
    //cout << "i = " << i << " P = " << P << " " << " LogP = " << LogP 
    //     << " TotalLogP = " << TotalLogP << endl;
  }

  return TotalLogP;
}



// =======================================================================
CPhraseAlignerHMM::CPhraseAlignerHMM()
{
  // currently, nothing to be done here
}


CPhraseAlignerHMM::~CPhraseAlignerHMM()
{
  // currently, nothing to be done here
}



bool
CPhraseAlignerHMM::AlignPhrase( 
                      CSentence *pSourceSentence, CSentence *pTargetSentence,
                      int StartSource, int EndSource,
                      CNumericalArray< int > & Min_i, CNumericalArray< int > & Max_i,
                      CNumericalArray< double > & Scores  )
{
  assert ( 0 );
  return false;

}

float
CPhraseAlignerHMM::CalcAlignLogProbSourceTarget( int j1, int j2, int i1, int i2 ) 
{
  assert ( 0 );
  return 0.0;

}


float
CPhraseAlignerHMM::CalcAlignLogProbTargetSource( int j1, int j2, int i1, int i2 ) 
{
  return 0.0;

}
