#include <string.h>
#include <stdlib.h>
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// switch on/off tracing for this module here
#undef NO_TRACE
#define NO_TRACE 1
#include <Trace.hh>

#include <iostream>
using std::ostream;


#include <PhraseAlignMS.hh>
#include <Text.hh>

using namespace std;



CPhraseAlignItemMS::CPhraseAlignItemMS()
{
  Init();
}


void
CPhraseAlignItemMS::Init()
{
  m_SentenceNumber        = -1;
  m_StartPosInSource      = -1;
  m_EndPosInSource        = -1;
  m_StartPosInTarget      = -1;
  m_EndPosInTarget        = -1;

  m_Keep                  = false;

  m_TranslationCount      = -1;

  m_SourcePhrase          = "";
  m_TargetPhrase          = "";

  m_Scores[ 0 ]           = 0.0;
  m_ScaledScores[ 0 ]     = 0.0;
}



/*
CNumericalArray< float >
CPhraseAlignItemMS::Scores() const
{
  CNumericalArray< float > Scores;
  for ( int i = 1; i <= m_Scores.MaxIndex(); i++ ) {
    Scores[ i ] = m_Scores[ i ];
  }
  return Scores;
}
*/


CPhraseAlignItemMS & 
CPhraseAlignItemMS::operator=( const CPhraseAlignItemMS & Item )
{
  if ( this == &Item ) 
    //-- Kopie auf sich selber kann ignoriert werden
    return *this;

  m_SentenceNumber        = Item.m_SentenceNumber;
  m_StartPosInSource      = Item.m_StartPosInSource;
  m_EndPosInSource        = Item.m_EndPosInSource;
  m_StartPosInTarget      = Item.m_StartPosInTarget;
  m_EndPosInTarget        = Item.m_EndPosInTarget;

  m_SourceWordIDs         = Item.m_SourceWordIDs;
  m_SourcePhrase          = Item.m_SourcePhrase;

  m_TargetWordIDs         = Item.m_TargetWordIDs;
  m_TargetPhrase          = Item.m_TargetPhrase;

  m_Keep                  = Item.m_Keep;

  m_Scores                = Item.m_Scores;
  m_CombinedScores        = Item.m_CombinedScores;
  m_ScaledScores          = Item.m_ScaledScores;
  m_CombinedScaledScores  = Item.m_CombinedScaledScores;

  m_TranslationCount      = Item.m_TranslationCount;

}


bool
CPhraseAlignItemMS::operator==( const CPhraseAlignItemMS &Item ) const
{
  //Trace( 0, "CPhraseAlignItemMS::operator==" );
  //cout << "compare " << m_TargetPhrase << " " << m_Score
  //	   << " with " << Item.m_TargetPhrase << " " << Item.m_Score << endl;

  if ( m_TargetWordIDs == Item.m_TargetWordIDs ) {
    //cout << "true" << endl;
    return true;
  } else {
    //cout << "false" << endl;
    return false;
  }
}


bool
CPhraseAlignItemMS::operator<( const CPhraseAlignItemMS &Item ) const
{
  //Trace( 0, "CPhraseAlignItemMS::operator<" );
  //cout << "compare " << m_TargetPhrase << " " << m_Score
  //	   << " with " << Item.m_TargetPhrase << " " << Item.m_Score << endl;

  if ( m_TargetWordIDs < Item.m_TargetWordIDs ) {
    //cout << "true" << endl;
    return true;
  } else {
    //cout << "false" << endl;
    return false;
  }
}


bool
CPhraseAlignItemMS::operator>( const CPhraseAlignItemMS &Item ) const
{
  //Trace( 0, "CPhraseAlignItemMS::operator>" );
  //cout << "compare " << m_TargetPhrase << " " << m_Score  
  //   << " with " << Item.m_TargetPhrase << " " << Item.m_Score << endl;
 
  if ( m_TargetWordIDs > Item.m_TargetWordIDs ) {
    //cout << "true" << endl;
    return true;
  } else {
    //cout << "false" << endl;
    return false;
  }
}


ostream&
operator<<(ostream& OutStream, const CPhraseAlignItemMS& PhraseAlignItem )
{
  OutStream << "Position = " << PhraseAlignItem.m_SentenceNumber << " " 
            << PhraseAlignItem.m_StartPosInSource << " " << PhraseAlignItem.m_EndPosInSource << " "
            << PhraseAlignItem.m_StartPosInTarget << " " << PhraseAlignItem.m_EndPosInTarget << " "
            << endl;

  int i;
  OutStream << "Raw scores    = ";
  i = PhraseAlignItem.m_Scores.MinIndex();
  while ( i <= PhraseAlignItem.m_Scores.MaxIndex() ) {
    OutStream << PhraseAlignItem.m_Scores[ i ] << " :: ";
    i++;
  }
  OutStream << endl;

  OutStream << "Scaled scores = ";
  i = PhraseAlignItem.m_ScaledScores.MinIndex();
  while ( i <= PhraseAlignItem.m_ScaledScores.MaxIndex() ) {
    OutStream << PhraseAlignItem.m_ScaledScores[ i ] << " :: ";
    i++;
  }
  OutStream << endl;

  OutStream << "PhrasePair " << PhraseAlignItem.m_SourcePhrase 
            << " # " << PhraseAlignItem.m_TargetPhrase << endl << flush;

  return OutStream;
}





//============================================================
CPhraseAlignerMS::CPhraseAlignerMS()
{
  m_pLexiconSourceTarget = NULL;
  m_pLexiconTargetSource = NULL;

  m_AlignBidirectional = false;
  m_UseEmptyWord = false;
  //m_FindCenterWithSum = false;

  m_StretchFactor = 1.0;

  m_FactorLonger = 1.3;
  m_FactorShorter = 1.0/m_FactorLonger;

  m_AddConstLonger = 2.0;
  m_AddConstShorter = -2.0;

  m_PruningStrategy = PRUNE_MULTIPLICATIVE;

  m_BeamFactor      = 1.0;
  m_MaxTranslations = 10;

  m_numberOfScores = 6;
  m_numberOfCombinedScores = 3;

}



CPhraseAlignerMS::~CPhraseAlignerMS()
{
  if ( m_pCalc != NULL )
    delete m_pCalc;
}



void
CPhraseAlignerMS::Init( CLexicon *pLexiconSourceTarget, CLexicon *pLexiconTargetSource )
{
  Init( pLexiconSourceTarget, pLexiconTargetSource, new CPhraseAlignProbCalcIBM1 );
}



void
CPhraseAlignerMS::Init( CLexicon *pLexiconSourceTarget, CLexicon *pLexiconTargetSource, CPhraseAlignProbCalc *pCalc )
{
  Trace( 2, "CPhraseAlignerMS::Init" );
  m_pLexiconSourceTarget = pLexiconSourceTarget;
  m_pLexiconTargetSource = pLexiconTargetSource;

  if ( m_pLexiconTargetSource ) {

    Trace( 2, "CPhraseAlignerMS::Init with two lexicons" );
    m_AlignBidirectional = true;
  } else {
    Trace( 2, "CPhraseAlignerMS::Init with one lexicons" );
    m_AlignBidirectional = false;
  }

  m_pCalc = pCalc;
}



/* deprecated
void 
CPhraseAlignerMS::SetAlignBidirectional( bool AlignBidirectional )
{
  //  assert( 0 );
  //cout << "CPhraseAlignerMS::SetAlignBidirectional" << AlignBidirectional << endl;
  Trace( 2, "CPhraseAlignerMS::SetAlignBidirectional " << AlignBidirectional );
  m_AlignBidirectional = AlignBidirectional;
  if ( m_AlignBidirectional ) {
    Trace( 2, "Align bidirectional" );
  } else {
    Trace( 2, "Align unidirectional" );
  }
}
*/



void 
CPhraseAlignerMS::SetUseEmptyWord( bool UseEmptyWord )
{
  m_UseEmptyWord = UseEmptyWord;
  m_pCalc->SetUseEmptyWord( UseEmptyWord );
}



/* deprecated
void 
CPhraseAlignerMS::SetFindCenterWithSum( bool UseSum )
{
  m_FindCenterWithSum = UseSum;
}
*/



void 
CPhraseAlignerMS::SetFindCenterMethod( int PhraseAlignFindCenterMethod )
{
  m_FindCenterMethod = PhraseAlignFindCenterMethod;
}



void 
CPhraseAlignerMS::SetPruningStrategy( int PruningStrategy )
{
  m_PruningStrategy = PruningStrategy;
}



void 
CPhraseAlignerMS::SetStretchFactors( float StretchFactor, float FactorShorter, float FactorLonger )
{
  m_StretchFactor = StretchFactor;
  m_FactorShorter = FactorShorter;
  m_FactorLonger  = FactorLonger;
}



void 
CPhraseAlignerMS::SetStretchFactors( float StretchFactor, float FactorShorter, float FactorLonger,
                                     float AddConstShorter, float AddConstLonger )
{
  m_StretchFactor   = StretchFactor;
  m_FactorShorter   = FactorShorter;
  m_FactorLonger    = FactorLonger;
  m_AddConstShorter = AddConstShorter;
  m_AddConstLonger  = AddConstLonger;

  
  Trace( 2,  "Stretch Factor = " << m_StretchFactor
         << " Factor Shorter = " << m_FactorShorter
         << " Factor Longer = " << m_FactorLonger
         << " Additive Cosntant Shorter = " << m_AddConstShorter
         << " Additive Constant Longer = " << m_AddConstLonger );
}


/* deprecated
void
CPhraseAlignerMS::SetInterpolationFactor( float Lambda )
{
  m_Lambda = Lambda;
}
*/



CNumericalArray<float> &
CPhraseAlignerMS::GetScalingFactors()
{
  return m_ScalingFactors;
}



void
CPhraseAlignerMS::SetScalingFactors( CNumericalArray<float> ScalingFactors )
{
  m_ScalingFactors = ScalingFactors;

  /*
    cout<<"Alignment Scaling Factors: ";
    for ( int i = ScalingFactors.MinIndex(); i <= ScalingFactors.MaxIndex(); i++ ) {
    cout << i << " = " << m_ScalingFactors[ i ] << " ";
    }
    cout << endl;
  */
}



void
CPhraseAlignerMS::SetBeam( float BeamFactor, int MaxTranslations )
{
  m_BeamFactor = BeamFactor;
  m_MaxTranslations = MaxTranslations;

  Trace( 3, "CPhraseAlignerMS: BeamFactor = " << m_BeamFactor << " MaxTranslations = " << m_MaxTranslations );

}



bool 
CPhraseAlignerMS::AlignPhrase( CSentence *pSourceSentence, CSentence *pTargetSentence,
                               int StartPosInSource, int EndPosInSource,
                               CSortedArray< CPhraseAlignItemMS > & PhraseAlignItems )
{

  //cout << "CPhraseAlignerMS::AlignPhrase" << endl;


#if DEBUG
  int DisplayAlignItem = 0;
  int DisplaySourcePhrase = 0;
#endif

  Trace( 1, "CPhraseAlignerMSIBM1::AlignPhrase" );

  // use IMB1 alignment model, i.e. only lexicon to
  // find i1 and i2 on target side such that the alignment
  // probability is highest when aligning f_j1 ... f_j2 to
  // e_i1 ... e_i2, and aligning outside f_j to outside e_i
  // i.e. caculate argmax_i1_i2 { Prod_j1...j2 Sum_i1...i2 p(f_j|e_i)
  // x Prod_j!=j1...j2 Sum_i!=i1...i2 p(f_j|e_i) }

  if ( !m_pLexiconSourceTarget && !m_pLexiconTargetSource ) {
    cerr << "no lexicon provided to calculate alignment" << endl;
    exit( -3);
  }

  if ( m_AlignBidirectional && (!m_pLexiconSourceTarget || !m_pLexiconTargetSource) ) {
    cerr << "calculation of alignment in both directions requested, but not both lexicons provided" << endl;
    exit( -3);
  }


  // some local variable
  CWordIndex f, e;
  int i, j, k, min_i, max_i;
  int LengthSourcePhrase = EndPosInSource - StartPosInSource + 1;

  //cout << "AlignPhrase " << StartPosInSource << " " << EndPosInSource << endl;
#if DEBUG
  if ( DisplaySourcePhrase ) {
    for ( j = StartPosInSource; j <= EndPosInSource; j++ ) {
      f = pSourceSentence->GetAt( j );
      cout << m_pLexiconSourceTarget->SourceVocabulary()->GetWord( f ) << " ";
    }
    cout << endl;
  }
#endif
  
  m_pSourceSentence = pSourceSentence;
  m_pTargetSentence = pTargetSentence;

  /*
    for ( j = StartPosInSource; j <= EndPosInSource; j++ ) {
    f = pSourceSentence->GetAt( j );
    cout << m_pLexiconSourceTarget->SourceVocabulary()->GetWord( f ) << " ";
    }
    cout << endl;
    cout << endl; cout  << " in: " << endl;
  */

  //pSourceSentence->Display( m_pLexiconSourceTarget->SourceVocabulary() ); cout << endl;
  //pTargetSentence->Display( m_pLexiconSourceTarget->TargetVocabulary() ); cout << endl;


  // length of the sentences
  m_J = pSourceSentence->Length();
  m_I = pTargetSentence->Length();
  //cout << "Length source " << m_J << " length target " << m_I << endl;

  // if sentences are longer than we can store probabilities, we skip it
  if ( (m_J >= MAX_SENTENCE_LENGTH) || (m_I >= MAX_SENTENCE_LENGTH) ) {
    return false;
  }

  // first, get all the probabilities and put them into static matrix
  //get Source->Target Probabilities
  GetProbabilities( m_pSourceSentence, m_pTargetSentence, m_pLexiconSourceTarget, &(m_ProbsSourceTarget[0][0]) );
  // renormalize probabilities for the current sentence pair
  Renormalize( m_J, m_I, &(m_ProbsSourceTarget[0][0]),&(m_RenProbsSourceTarget[0][0]) );

  //get Target->Source Probabilities
  if( m_AlignBidirectional ) {
    GetProbabilities( m_pTargetSentence, m_pSourceSentence, m_pLexiconTargetSource, &(m_ProbsTargetSource[0][0]) );
    // renormalize probabilities for the current sentence pair
    Renormalize( m_I,m_J, &(m_ProbsTargetSource[0][0]), &(m_RenProbsTargetSource[0][0]) );
  }

  //DisplayProbs( &(m_RenProbsSourceTarget[0][0]), &(m_RenProbsTargetSource[0][0]) );

  // find 'center' of target phrase for source phrase
  float i_mid = -1.0;
  /*
    if ( m_FindCenterWithSum ) 
    // i_mid = FindCenterSum( StartPosInSource, EndPosInSource );
    i_mid = FindCenterBlock( StartPosInSource, EndPosInSource );
    else
    i_mid = FindCenterMax( StartPosInSource, EndPosInSource );
  */

  switch ( m_FindCenterMethod ) {
  case FINDCENTER_MAX:
    i_mid = FindCenterMax( StartPosInSource, EndPosInSource );
    break;
  case FINDCENTER_SUM:
    i_mid = FindCenterSum( StartPosInSource, EndPosInSource );
    break;
  case FINDCENTER_BLOCK:
  default:
    i_mid = FindCenterBlock( StartPosInSource, EndPosInSource );
  }

  // don't want too short or too long translations
  // on average the target length should be source length times stretch factor
  // where stretch factor is the ratio target corpus lenght to source corpus length
  // then target should not be longer then this average length time a elongation factor
  // and not shorter then the average length time a shortening factor
  // we allow for and addition +/-2 to have e.g. 1->3 alignemnts
  int MinTargetLength = MinTargetPhraseLength( LengthSourcePhrase );
  int MaxTargetLength = MaxTargetPhraseLength( LengthSourcePhrase );

  if ( MinTargetLength < 1 )
    MinTargetLength = 1;

  /*
    cout << "StretchFactor = " << m_StretchFactor
    << "  FactorShorter = " << m_FactorShorter
    << "  FactorLonger = " << m_FactorLonger
    << "  AddConstShorter = " << m_AddConstShorter
    << "  AddConstLonger = " << m_AddConstLonger << endl;


    cout << "LengthSourcePhrase = " << LengthSourcePhrase 
    << "  MinTargetLength = " << (m_AddConstShorter + LengthSourcePhrase * m_StretchFactor * m_FactorShorter) 
    << " -> " << MinTargetLength
    << "  MaxTargetLength = " << (m_AddConstLonger + LengthSourcePhrase * m_StretchFactor * m_FactorLonger) 
    << " -> " << MaxTargetLength
    << endl;
  */


  int mid = int( i_mid + 0.5 );
  /*
    cout << endl;
    cout << "Phrases positions: " << StartPosInSource << " " << EndPosInSource;
    cout << " Approximate center = " << i_mid << " rounded to " << mid << endl;  
    cout << "LengthSourcePhrase " << LengthSourcePhrase 
    << " StretchFactor " << m_StretchFactor
    << " FactorShorter " << m_FactorShorter
    << " FactorLonger " << m_FactorLonger << endl;
    cout << "MinTargetLength = " << MinTargetLength 
    << " MaxTargetLength = " << MaxTargetLength <<endl;
  */

  // in case there are old results in the arrays
  PhraseAlignItems.Clear();

  // expand the target phrase around the center
  CSortedArray< float > Probs;
  // sort smallest first
  //Probs.SortReverse();
  CNumericalArray< float > AllLogP1;
  CNumericalArray< float > AllLogP2;

  float LogP, LogP1, LogP2;
  float BestLogP = 10000.0, BestLogP1 = 10000.0, BestLogP2 = 10000.0;

  bool WasNew;

  int n = 0;  // counter for translations


  // here, we loop over the two boundaries
  for ( min_i = mid - MaxTargetLength; min_i <= mid + 1; min_i++ ) {
    for ( max_i = mid - 1; max_i <= mid + MaxTargetLength; max_i++ ) {

      // now brute force
      //cout << "brute force phrase alignment" << endl;
      //for ( min_i = pSourceSentence->MinIndex(); min_i <=  pSourceSentence->MaxIndex(); min_i++ ) {
      //for ( max_i = min_i + MinTargetLength - 1; max_i <= min_i + MaxTargetLength; max_i++ ) {

      // currently we do not assume that we have <sos> and <eos>
      if ( (min_i >= 1) && (max_i <= m_I) 
           && ((max_i - min_i + 1) >= MinTargetLength)
           && ((max_i - min_i + 1) <= MaxTargetLength) ) {
		
        //cout << "min_i = " << min_i << " max_i = " << max_i << " m_I = " << m_I << endl;

        n++;
        CPhraseAlignItemMS AlignItem;
        // all scores are stored in CNumericalArray objects, i.e. they are initialized to 0

        // we store the position in the corpus
        AlignItem.m_SentenceNumber   = -1;  // we don't know the sentence number here
        AlignItem.m_StartPosInSource = StartPosInSource;
        AlignItem.m_EndPosInSource   = EndPosInSource;
        AlignItem.m_StartPosInTarget = min_i;
        AlignItem.m_EndPosInTarget   = max_i;

        // get the target phrase: word IDs and string
        for ( int i = min_i; i <= max_i; i++ ) {
          //cout << TextE.GetWord( SentenceID - 1, i ) << " ";
          e = m_pTargetSentence->GetAt( i );
          AlignItem.m_TargetWordIDs.Append( e );
          AlignItem.m_TargetPhrase += CString( " " ) + m_pLexiconSourceTarget->TargetVocabulary()->GetWord( e );
        }

        if ( m_pLexiconSourceTarget ) {

          // now, we calculate the log prob for the phrase from total sentence alignment,
          // dividing by the sentence length or the phrase length, to have per word costs (for pruning)
          // later, i.e. after pruning and before we insert into lattice, we need to multiply again by source phrase length
          LogP = m_pCalc->CalcSentAlignLogProb( StartPosInSource, EndPosInSource, min_i, max_i, m_J, m_I, 
                                                &(m_RenProbsSourceTarget[0][0]) );
          //no done already in ClacSentAlignLogProb
          //LogP /= m_J;
          AlignItem.m_Scores[1] = LogP;
          //AlignItem.m_ScaledLogProbSentAlignFE = m_ScalingFactors[1] * LogP;
		  
          // probability of just aligning the phrase, using renormalized word probabilities
          LogP = m_pCalc->CalcPhraseLogProb( StartPosInSource, EndPosInSource, min_i, max_i, 
                                             &(m_RenProbsSourceTarget[0][0]) );
          //LogP /= ( EndSource - StartSource + 1 ) * LenghtSourcePhrase;
          //LogP /= ( EndPosInSource - StartPosInSource + 1 );
          AlignItem.m_Scores[3] = LogP;
          //AlignItem.m_ScaledLogProbPhraseAlignFE = m_ScalingFactors[3] * LogP;
		  
          // phrase translation probability, using word probabilities without renormalization
          LogP = m_pCalc->CalcPhraseLogProb( StartPosInSource, EndPosInSource, min_i, max_i,
                                             &(m_ProbsSourceTarget[0][0]) );
          //LogP /= ( EndPosInSource - StartPosInSource + 1 );
          AlignItem.m_Scores[5] = LogP;
          //AlignItem.m_ScaledLogProbPhraseFE = m_ScalingFactors[5] * LogP;
		  
        }
		
        if ( m_pLexiconTargetSource ) {
		  
          // probability of aligning phrase, using renormalized word probabilities
          // divide by target phrase length and multiply by target phrase length
          LogP = m_pCalc->CalcSentAlignLogProb( min_i, max_i, StartPosInSource, EndPosInSource,m_I, m_J,
                                                &(m_RenProbsTargetSource[0][0]) );
          //now already done in ClacSentAlignLogProb
          //LogP /= m_I;
          //LogP *= LengthSourcePhrase;
          AlignItem.m_Scores[2] = LogP;
          //AlignItem.m_ScaledLogProbSentAlignEF = m_ScalingFactors[2] * LogP;
		  
          // now, we calculate the log prob for the phrase from total sentence alignment,
          // dividing by the sentence length, multiplied by source phrase length
          LogP = m_pCalc->CalcPhraseLogProb(min_i, max_i, StartPosInSource, EndPosInSource,
                                            &(m_RenProbsTargetSource[0][0]) );
          //cout << "LogP " << LogP << " TargetLength " << ( max_i - min_i + 1 ) 
          //   << " SourceLength " << LengthSourcePhrase << endl;
          //LogP /= ( max_i - min_i + 1 );
          //LogP *= LengthSourcePhrase;
          AlignItem.m_Scores[4] = LogP;
          //AlignItem.m_ScaledLogProbPhraseAlignEF = m_ScalingFactors[4] * LogP;
		  
          // phrase translation probability, using word probabilities without renormalization
          LogP = m_pCalc->CalcPhraseLogProb(min_i, max_i, StartPosInSource, EndPosInSource,
                                            &(m_ProbsTargetSource[0][0]) );
          //LogP /= ( max_i - min_i + 1 );
          //LogP *= LengthSourcePhrase;
          AlignItem.m_Scores[6] = LogP;
          //AlignItem.m_ScaledLogProbPhraseEF = m_ScalingFactors[6] * LogP;
		  
        }
		
        AddScaledAndCombinedScores( AlignItem );

#if DEBUG
        if ( DisplayAlignItem ) {
          cout << "Testing target phrase: " << AlignItem.m_TargetPhrase <<endl;
          cout << AlignItem << endl;
		  
          for ( int ii = min_i; ii <= max_i; ii++ ) {
            e = pTargetSentence->GetAt( ii );
            cout << m_pLexiconSourceTarget->TargetVocabulary()->GetWord( e ) << " ";
          }
		  
          cout << endl;
        }
#endif
		
        // 	cout << "Scores: ";
        // 	for(int i = 1; i <= 6; i++) {
        // 	  cout << AlignItem.m_Scores[i] << " ";
        // 	}
        // 	cout << endl;
        // 	cout << " Scaled Scores: ";
        // 	for(int i = 0; i <= 6; i++) {
        // 	  cout << AlignItem.m_ScaledScores[i] << " ";
        // 	}
        // 	cout << endl;

        // store now in list of all good translations for this source phrase
        int Location;
        bool IsNew;
        if ( PhraseAlignItems.Find( AlignItem, Location ) ) {
          //cout << " translation was found" << endl;
          if ( AlignItem.Score() < PhraseAlignItems.GetAt( Location ).Score() ) {
            // already stored, but current score is lower, i.e. better
            // therefore overwrite it
            //cout << " translation was found, but new score is smaller, i.e. overwrite " << endl;
            PhraseAlignItems.Insert( AlignItem, IsNew );
            //}
            //} else {
            //cout << " but new score is larger, i.e. don't use " << endl;
          }
        } else {
          //cout << " translation was not found - insert" << endl;
          PhraseAlignItems.Insert( AlignItem, IsNew );
        }
		
		
        //cout << "P( " << min_i << ", " << max_i << " ) = " << LogP;
        //cout << "  inserted at " << k << endl;
      } else {
        ; //cout << "skip" << endl;
      }
    }
  }
  
  /*
    if ( DEBUG ) {
    cout << endl;
    cout << "Best translation: ";
	
    cout << Min_i[ Min_i.MinIndex() ] 
    << " " << Max_i[ Max_i.MinIndex() ] 
    << " " << Scores[ Scores.MinIndex() ] << endl;
	
    for ( j = StartSource; j <= EndSource; j++ ) {
	  f = pSourceSentence->GetAt( j );
	  cout << m_pLexiconSourceTarget->SourceVocabulary()->GetWord( f ) << " ";
    }
	
    for ( i = Min_i[ Min_i.MinIndex() ]; i <= Max_i[ Max_i.MinIndex() ]; i++ ) {
	  e = pTargetSentence->GetAt( i );
	  cout << m_pLexiconSourceTarget->TargetVocabulary()->GetWord( e ) << " ";
    }
	
    cout << endl << endl;
    }
  */

  // and now some pruning
  //Trace( 0, "Number of found translations " << Probs.MinIndex() << " ... " << Probs.MaxIndex() );
  //cout << "BestLogP1 = " << BestLogP1 << "  BestLogP2 = " << BestLogP2 << "  BestLogP = " << BestLogP << endl;


  for ( int n = Probs.MinIndex(); n <= Probs.MaxIndex(); n++ ) {

    /*
      if ( DEBUG  ) {
      cout << Min_i[ n ] << " " << Max_i[ n ] << " " << AllLogP1[ n ] << " " << AllLogP2[ n ]
      //		 << " " << Probs[ n ]
      << " " << Scores[ n ] << "   ";
	  
      for ( j = StartSource; j <= EndSource; j++ ) {
      f = pSourceSentence->GetAt( j );
      cout << m_pLexiconSourceTarget->SourceVocabulary()->GetWord( f ) << " ";
      }
      cout << " # ";
      for ( i = Min_i[ n ]; i <= Max_i[ n ]; i++ ) {
      e = pTargetSentence->GetAt( i );
      cout << m_pLexiconSourceTarget->TargetVocabulary()->GetWord( e ) << " ";
      }
      }
    */

    //cout << "BeamFactor = " << m_BeamFactor << endl;

    /*
    // we keep all translations which are within beam to best scores
    // plus also all with small probs 
    // (the BestLogPx can be 0.0, i.e. we would prune everything)

    float SmallFloorProb = 0.01;

    if ( (AllLogP1[ n ] > m_BeamFactor * (BestLogP1 + SmallFloorProb)) ) cout << " (pruned on P1)" << endl;
    if ( (AllLogP2[ n ] > m_BeamFactor * (BestLogP2 + SmallFloorProb)) ) cout << " (pruned on P2)" << endl;
   	if ( (Scores[ n ]   > m_BeamFactor * (BestLogP  + SmallFloorProb)) ) cout << " (pruned on P)" << endl;


    if ( (AllLogP1[ n ] > m_BeamFactor * (BestLogP1+ SmallFloorProb))
    || (AllLogP2[ n ] > m_BeamFactor * (BestLogP2+ SmallFloorProb))
    || (Scores[ n ] > m_BeamFactor * (BestLogP+ SmallFloorProb)) ) {

	  if ( DEBUG ) {
		cout << "    <<< will be pruned >>>";
	  }
	  Scores[ n ] = 10000.0;
    } else {
		cout << "    <<< will NOT be pruned >>>";
    }

    if ( DEBUG ) {
	  cout << endl;
    }
    */

  }

  
  /*
    if ( Min_i.MaxIndex() - Min_i.MinIndex() + 1 > m_MaxTranslations ) {
    Min_i.SetBoundaries( Min_i.MinIndex(), Min_i.MinIndex() + m_MaxTranslations - 1 );
    Max_i.SetBoundaries( Max_i.MinIndex(), Max_i.MinIndex() + m_MaxTranslations - 1 );
    Scores.SetBoundaries( Scores.MinIndex(), Scores.MinIndex() + m_MaxTranslations - 1 );
    }
  */

  //Trace( 0, "Number of returned translations " << Min_i.MinIndex() << " ... " << Min_i.MaxIndex() );

  //pSourceSentence->Display( m_pLexiconSourceTarget->SourceVocabulary() );
  //cout << endl;
  //pTargetSentence->Display( m_pLexiconSourceTarget->SourceVocabulary() );
  //cout << endl;

  //cout << endl;
  return true;
}



void
CPhraseAlignerMS::AddScaledAndCombinedScores( CPhraseAlignItemMS & AlignItem )
{
  //calculating scaled scores
  for ( int i = 1; i <= m_numberOfScores; i++ ) {
    AlignItem.m_ScaledScores[i] = m_ScalingFactors[i] * AlignItem.m_Scores[i];
  }	
		
  // now, we combine these scores according to the scaling factors
  AlignItem.m_Scores[ 0 ] = 0.0;
  for ( int i = 1; i <= m_numberOfScores; i++ ) {
    AlignItem.m_Scores[0] += AlignItem.m_Scores[i];
  }
		
  //Combining the the same type of scores from both directions
  AlignItem.m_CombinedScores[1] = AlignItem.m_Scores[1] + AlignItem.m_Scores[2];
  AlignItem.m_CombinedScores[2] = AlignItem.m_Scores[3] + AlignItem.m_Scores[4];
  AlignItem.m_CombinedScores[3] = AlignItem.m_Scores[5] + AlignItem.m_Scores[6];		

  // and the scaled ones
  AlignItem.m_ScaledScores[ 0 ] = 0.0;
  for ( int i = 1; i <= m_numberOfScores; i++ ) {
    AlignItem.m_ScaledScores[0] +=AlignItem.m_ScaledScores[i];
  }		
		
  //Combining the the same type of scores from both directions
  AlignItem.m_CombinedScaledScores[1] = AlignItem.m_ScaledScores[1] + AlignItem.m_ScaledScores[2];
  AlignItem.m_CombinedScaledScores[2] = AlignItem.m_ScaledScores[3] + AlignItem.m_ScaledScores[4];
  AlignItem.m_CombinedScaledScores[3] = AlignItem.m_ScaledScores[5] + AlignItem.m_ScaledScores[6];

}



bool
CPhraseAlignerMS::FindBestPhrasePair( CSentence *pSourceSentence, CSentence *pTargetSentence,
                                      int MinSourcePhraseLength, int MaxSourcePhraseLength,
                                      CPhraseAlignItemMS & BestPhraseAlignItem )
{
  //cout << "CPhraseAlignerMS::FindBestPhrasePair" << endl << flush;

  // calculate for all source phrases between MinPhraseLength and MaxPhraseLength
  // the alignment and keep track of the best results, this is then returned.
  // i.e. for each sentence pair we get one phrase pair
  // we can evaluate based on combined score, or look into perplexity gain,
  // as we do in sentence splitting (could be total or per word)

  m_pSourceSentence = pSourceSentence;
  m_pTargetSentence = pTargetSentence;

  // first we need to fill the probability matrices
  // length of the sentences
  m_J = pSourceSentence->Length();
  m_I = pTargetSentence->Length();
  //cout << "Length source " << m_J << " length target " << m_I << endl;

  // if sentences are longer than we can store probabilities, we skip it
  if ( (m_J >= MAX_SENTENCE_LENGTH) || (m_I >= MAX_SENTENCE_LENGTH) ) {
    cout << "WARNING: Sentence is too long, we skip it" << endl;
    return false;
  }

  // first, get all the probabilities and put them into static matrix
  // get Source->Target Probabilities
  GetProbabilities( m_pSourceSentence, m_pTargetSentence, m_pLexiconSourceTarget, &(m_ProbsSourceTarget[ 0 ][ 0 ]) );
  // renormalize probabilities for the current sentence pair
  Renormalize( m_J, m_I, &(m_ProbsSourceTarget[ 0 ][ 0 ]), &(m_RenProbsSourceTarget[ 0 ][ 0 ]) );

  //get Target->Source Probabilities
  if( m_AlignBidirectional ) {
    GetProbabilities( m_pTargetSentence, m_pSourceSentence, m_pLexiconTargetSource, &(m_ProbsTargetSource[ 0 ][ 0 ]) );
    // renormalize probabilities for the current sentence pair
    Renormalize( m_I, m_J, &(m_ProbsTargetSource[ 0 ][ 0 ]), &(m_RenProbsTargetSource[ 0 ][ 0 ]) );
  }

  int StartPosInSource, EndPosInSource;
  int StartPosInTarget, EndPosInTarget;
  float LogProbFE, LogProbEF, LogProbFE2, LogProbEF2, LogProbFE3, LogProbEF3;
  float BestLogProb, GlobalBestLogProb = 999999.99;
  CPhraseAlignItemMS BestAlignItem;
  CPhraseAlignItemMS GlobalBestAlignItem;

  // now we loop over all possible source phrases (within length restrictions)
  // and search for best target phrase
  for ( int SourcePhraseLength = MinSourcePhraseLength; SourcePhraseLength <= MaxSourcePhraseLength; SourcePhraseLength++ ) {
    //cout << "Source Length is now " << SourcePhraseLength << endl << flush;

    // loop over all source phrases of given length
    for ( StartPosInSource = pSourceSentence->MinIndex();
          StartPosInSource <= pSourceSentence->MaxIndex() - SourcePhraseLength + 1;
          StartPosInSource++ ) {

      EndPosInSource = StartPosInSource + SourcePhraseLength - 1;
      //cout << "  StartPosInSource = " << StartPosInSource << " EndPosInSource " << EndPosInSource << endl;

      // calculate possible target phrase lengths
      // don't want too short or too long translations
      // on average the target length should be source length times stretch factor
      // where stretch factor is the ratio of target corpus lenght to source corpus length
      // then target should not be longer then this average length time a elongation factor
      // and not shorter then the average length time a shortening factor
      // we allow for and addition +/-2 to have e.g. 1->3 alignemnts
      int LengthSourcePhrase = EndPosInSource - StartPosInSource + 1;
      int MinTargetLength = MinTargetPhraseLength( LengthSourcePhrase );
      int MaxTargetLength = MaxTargetPhraseLength( LengthSourcePhrase );

      if ( MinTargetLength < 1 )
        MinTargetLength = 1;

      if ( MaxTargetLength >= m_I )
        MaxTargetLength = m_I;

      //cout << "  MinTargetPhraseLength " << MinTargetPhraseLength << " MaxTargetPhraseLength " << MaxTargetPhraseLength << endl;

      // initiallize BestLogProb to large number, for testing we get the best align item for each source phrase
      BestLogProb = 999999.99;

      // loop over all possible target lengths
      for ( int TargetPhraseLength = MinTargetLength; TargetPhraseLength <= MaxTargetLength; TargetPhraseLength++ ) {
        //cout << "   Target Phrase Length is now " << TargetPhraseLength << endl;

        // loop over all target phrases of given length
        for ( StartPosInTarget = pTargetSentence->MinIndex();
              StartPosInTarget <= pTargetSentence->MaxIndex() - TargetPhraseLength + 1;
              StartPosInTarget++ ) {

          EndPosInTarget = StartPosInTarget + TargetPhraseLength - 1;
          //cout << "      StartPosInTarget = " << StartPosInTarget << " EndPosInTarget " << EndPosInTarget << endl;

          LogProbFE = m_pCalc->CalcSentAlignLogProb( StartPosInSource, EndPosInSource, StartPosInTarget, EndPosInTarget,
                                                     m_J, m_I, &(m_RenProbsSourceTarget[ 0 ][ 0 ]) );
          //cout << "        LogProg f->e = " << LogProbFE << endl;
		  
          LogProbFE2 = m_pCalc->CalcPhraseLogProb( StartPosInSource, EndPosInSource, StartPosInTarget, EndPosInTarget, 
                                                   &(m_RenProbsSourceTarget[ 0 ][ 0 ]) );
          //AlignItem.m_ScaledLogProbPhraseAlignFE = m_ScalingFactors[3] * LogP;
		  
          LogProbFE3 = m_pCalc->CalcPhraseLogProb( StartPosInSource, EndPosInSource, StartPosInTarget, EndPosInTarget,
                                                   &(m_ProbsSourceTarget[ 0 ][ 0 ]) );
          //AlignItem.m_ScaledLogProbPhraseFE = m_ScalingFactors[5] * LogP;


          LogProbEF = m_pCalc->CalcSentAlignLogProb( StartPosInTarget, EndPosInTarget, StartPosInSource, EndPosInSource,
                                                     m_I, m_J, &(m_RenProbsTargetSource[ 0 ][ 0 ]) );
          //cout << "        LogProg e->f = " << LogProbEF << endl;

          LogProbEF2 = m_pCalc->CalcPhraseLogProb( StartPosInTarget, EndPosInTarget, StartPosInSource, EndPosInSource, 
                                                   &(m_RenProbsTargetSource[ 0 ][ 0 ]) );
          //AlignItem.m_ScaledLogProbPhraseAlignFE = m_ScalingFactors[3] * LogP;
		  
          LogProbEF3 = m_pCalc->CalcPhraseLogProb( StartPosInTarget, EndPosInTarget, StartPosInSource, EndPosInSource,
                                                   &(m_ProbsTargetSource[ 0 ][ 0 ]) );
          //AlignItem.m_ScaledLogProbPhraseFE = m_ScalingFactors[5] * LogP;

          // store best target phrase
          if ( LogProbFE + LogProbEF < BestLogProb ) {
            BestAlignItem.m_Scores[ 1 ] = LogProbFE;
            BestAlignItem.m_Scores[ 2 ] = LogProbEF;
            BestAlignItem.m_Scores[ 3 ] = LogProbFE2;
            BestAlignItem.m_Scores[ 4 ] = LogProbEF2;
            BestAlignItem.m_Scores[ 5 ] = LogProbFE3;
            BestAlignItem.m_Scores[ 6 ] = LogProbEF3;

            AddScaledAndCombinedScores( BestAlignItem );

            BestAlignItem.m_SentenceNumber = 0;
            BestAlignItem.m_StartPosInSource = StartPosInSource;
            BestAlignItem.m_EndPosInSource   = EndPosInSource;
            BestAlignItem.m_StartPosInTarget = StartPosInTarget;
            BestAlignItem.m_EndPosInTarget   = EndPosInTarget;

            // store the best scores so far
            BestLogProb = BestAlignItem.m_ScaledScores[ 0 ];
          }

        }  // loop over all target phrases of given length
      } // loop over all possible target phrase lengths
	  
      if ( 0 ) {
        //cout << "Testing target phrase: " << BestAlignItem.m_TargetPhrase <<endl;
        //cout << BestAlignItem << endl;
        cout << BestLogProb << " # ";
        for ( int ii = BestAlignItem.m_StartPosInSource; ii <= BestAlignItem.m_EndPosInSource; ii++ ) {
          CWordIndex f = pSourceSentence->GetAt( ii );
          cout << m_pLexiconSourceTarget->SourceVocabulary()->GetWord( f ) << " ";
        }
        cout << " # ";
        for ( int ii = BestAlignItem.m_StartPosInTarget; ii <= BestAlignItem.m_EndPosInTarget; ii++ ) {
          CWordIndex e = pTargetSentence->GetAt( ii );
          cout << m_pLexiconSourceTarget->TargetVocabulary()->GetWord( e ) << " ";
        }
        cout << endl;
      }
		  
      if ( BestLogProb < GlobalBestLogProb ) {
        GlobalBestLogProb = BestLogProb;
        GlobalBestAlignItem = BestAlignItem;
      }

    }  // loop over all source phrases of given length
  } // loop over all possible source phrase lengths 


  if ( 0 ) {
    cout << endl;
    cout << "Best alignment in sentence pair: " << endl;

    //cout << "Testing target phrase: " << BestAlignItem.m_TargetPhrase <<endl;
    //cout << BestAlignItem << endl;
    cout << GlobalBestLogProb << " # ";
    for ( int ii = GlobalBestAlignItem.m_StartPosInSource; ii <= GlobalBestAlignItem.m_EndPosInSource; ii++ ) {
      CWordIndex f = pSourceSentence->GetAt( ii );
      //cout << m_pLexiconSourceTarget->SourceVocabulary()->GetWord( f ) << " ";
    }
    cout << " # ";
    for ( int ii = GlobalBestAlignItem.m_StartPosInTarget; ii <= GlobalBestAlignItem.m_EndPosInTarget; ii++ ) {
      CWordIndex e = pTargetSentence->GetAt( ii );
      //cout << m_pLexiconSourceTarget->TargetVocabulary()->GetWord( e ) << " ";
    }
    cout << endl  << endl;
  }
  
  //cout << endl;

  BestPhraseAlignItem = GlobalBestAlignItem;


  /*
  cout << "Get source and target phrase for " << BestPhraseAlignItem.m_StartPosInSource
       << " " << BestPhraseAlignItem.m_EndPosInSource
       << " " << BestPhraseAlignItem.m_StartPosInTarget
       << " " <<  BestPhraseAlignItem.m_EndPosInTarget
       << endl;
  */

  // get the source phrase: word IDs and string
  for ( int j = BestPhraseAlignItem.m_StartPosInSource; j <= BestPhraseAlignItem.m_EndPosInSource; j++ ) {
    CWordIndex f = m_pSourceSentence->GetAt( j );
    BestPhraseAlignItem.m_SourceWordIDs.Append( f );
    BestPhraseAlignItem.m_SourcePhrase += CString( " " ) + m_pLexiconSourceTarget->SourceVocabulary()->GetWord( f );
  }


  // get the target phrase: word IDs and string
  for ( int i = BestPhraseAlignItem.m_StartPosInTarget; i <= BestPhraseAlignItem.m_EndPosInTarget; i++ ) {
    CWordIndex e = m_pTargetSentence->GetAt( i );
    BestPhraseAlignItem.m_TargetWordIDs.Append( e );
    BestPhraseAlignItem.m_TargetPhrase += CString( " " ) + m_pLexiconSourceTarget->TargetVocabulary()->GetWord( e );
  }

  cout << "done with finding best phrase pair" << endl << flush;

  return true;
}



bool
CPhraseAlignerMS::FindAllPhrasePair( CSentence *pSourceSentence, CSentence *pTargetSentence,
                                     int MinSourcePhraseLength, int MaxSourcePhraseLength,
                                     CArray< CPhraseAlignItemMS > & PhraseAlignItems )
{
  //cout << "CPhraseAlignerMS::FindAllPhrasePair" << endl << flush;

  // calculate for all source phrases between MinPhraseLength and MaxPhraseLength
  // the alignment and keep track of the best results for each source phrase, this is then returned.


  m_pSourceSentence = pSourceSentence;
  m_pTargetSentence = pTargetSentence;

  // first we need to fill the probability matrices
  // length of the sentences
  m_J = pSourceSentence->Length();
  m_I = pTargetSentence->Length();
  //cout << "Length source " << m_J << " length target " << m_I << endl;

  // if sentences are longer than we can store probabilities, we skip it
  if ( (m_J >= MAX_SENTENCE_LENGTH) || (m_I >= MAX_SENTENCE_LENGTH) ) {
    cout << "WARNING: Sentence is too long, we skip it" << endl;
    return false;
  }

  // first, get all the probabilities and put them into static matrix
  // get Source->Target Probabilities
  GetProbabilities( m_pSourceSentence, m_pTargetSentence, m_pLexiconSourceTarget, &(m_ProbsSourceTarget[ 0 ][ 0 ]) );
  // renormalize probabilities for the current sentence pair
  Renormalize( m_J, m_I, &(m_ProbsSourceTarget[ 0 ][ 0 ]), &(m_RenProbsSourceTarget[ 0 ][ 0 ]) );

  //get Target->Source Probabilities
  if( m_AlignBidirectional ) {
    GetProbabilities( m_pTargetSentence, m_pSourceSentence, m_pLexiconTargetSource, &(m_ProbsTargetSource[ 0 ][ 0 ]) );
    // renormalize probabilities for the current sentence pair
    Renormalize( m_I, m_J, &(m_ProbsTargetSource[ 0 ][ 0 ]), &(m_RenProbsTargetSource[ 0 ][ 0 ]) );
  }

  int StartPosInSource, EndPosInSource;
  int StartPosInTarget, EndPosInTarget;
  float LogProbFE, LogProbEF, LogProbFE2, LogProbEF2, LogProbFE3, LogProbEF3;
  float BestLogProb = 999999.99;
  CPhraseAlignItemMS BestAlignItem;

  // now we loop over all possible source phrases (within length restrictions)
  // and search for best target phrase
  for ( int SourcePhraseLength = MinSourcePhraseLength; SourcePhraseLength <= MaxSourcePhraseLength; SourcePhraseLength++ ) {
    //cout << "Source Length is now " << SourcePhraseLength << endl << flush;

    // loop over all source phrases of given length
    for ( StartPosInSource = pSourceSentence->MinIndex();
          StartPosInSource <= pSourceSentence->MaxIndex() - SourcePhraseLength + 1;
          StartPosInSource++ ) {

      EndPosInSource = StartPosInSource + SourcePhraseLength - 1;
      //cout << "  StartPosInSource = " << StartPosInSource << " EndPosInSource " << EndPosInSource << endl;

      // calculate possible target phrase lengths
      // don't want too short or too long translations
      // on average the target length should be source length times stretch factor
      // where stretch factor is the ratio of target corpus lenght to source corpus length
      // then target should not be longer then this average length time a elongation factor
      // and not shorter then the average length time a shortening factor
      // we allow for and addition +/-2 to have e.g. 1->3 alignemnts
      int LengthSourcePhrase = EndPosInSource - StartPosInSource + 1;
      int MinTargetLength = MinTargetPhraseLength( LengthSourcePhrase );
      int MaxTargetLength = MaxTargetPhraseLength( LengthSourcePhrase );

      if ( MinTargetLength < 1 )
        MinTargetLength = 1;

      if ( MaxTargetLength >= m_I )
        MaxTargetLength = m_I;

      //cout << "  MinTargetPhraseLength " << MinTargetPhraseLength << " MaxTargetPhraseLength " << MaxTargetPhraseLength << endl;

      // initiallize BestLogProb to large number, for testing we get the best align item for each source phrase
      BestLogProb = 999999.99;

      // loop over all possible target lengths
      for ( int TargetPhraseLength = MinTargetLength; TargetPhraseLength <= MaxTargetLength; TargetPhraseLength++ ) {
        //cout << "   Target Phrase Length is now " << TargetPhraseLength << endl;

        // loop over all target phrases of given length
        for ( StartPosInTarget = pTargetSentence->MinIndex();
              StartPosInTarget <= pTargetSentence->MaxIndex() - TargetPhraseLength + 1;
              StartPosInTarget++ ) {

          EndPosInTarget = StartPosInTarget + TargetPhraseLength - 1;
          //cout << "      StartPosInTarget = " << StartPosInTarget << " EndPosInTarget " << EndPosInTarget << endl;

          // align item to store results for one possible alignment
          CPhraseAlignItemMS AlignItem;

          float LogProb = CalcAllAlignLogProbs( StartPosInSource, EndPosInSource, StartPosInTarget, EndPosInTarget, AlignItem );

          // store best target phrase
          if ( LogProb < BestLogProb ) {
            BestAlignItem = AlignItem;
            BestLogProb = LogProb;
          }

        }  // loop over all target phrases of given length
      } // loop over all possible target phrase lengths
	  

      AddWordIDsAndStrings( BestAlignItem );

      //cout << "Best Align Item:" << endl;
      //cout << BestAlignItem << endl;
      /*
      // get the source phrase: word IDs and string
      for ( int j = BestAlignItem.m_StartPosInSource; j <= BestAlignItem.m_EndPosInSource; j++ ) {
      CWordIndex f = m_pSourceSentence->GetAt( j );
      BestAlignItem.m_SourceWordIDs.Append( f );
      BestAlignItem.m_SourcePhrase += CString( " " ) + m_pLexiconSourceTarget->SourceVocabulary()->GetWord( f );
      }

      // get the target phrase: word IDs and string
      for ( int i = BestAlignItem.m_StartPosInTarget; i <= BestAlignItem.m_EndPosInTarget; i++ ) {
      CWordIndex e = m_pTargetSentence->GetAt( i );
      BestAlignItem.m_TargetWordIDs.Append( e );
      BestAlignItem.m_TargetPhrase += CString( " " ) + m_pLexiconSourceTarget->TargetVocabulary()->GetWord( e );
      }
	  
      cout << BestAlignItem << endl;
      */

      if ( 0 ) {
        //cout << "Testing target phrase: " << BestAlignItem.m_TargetPhrase <<endl;
        cout << BestAlignItem << endl;
        cout << BestAlignItem.ScaledScore() << " # ";
        for ( int ii = BestAlignItem.m_StartPosInSource; ii <= BestAlignItem.m_EndPosInSource; ii++ ) {
          CWordIndex f = m_pSourceSentence->GetAt( ii );
          cout << m_pLexiconSourceTarget->SourceVocabulary()->GetWord( f ) << " ";
        }
        cout << " # ";
        for ( int ii = BestAlignItem.m_StartPosInTarget; ii <= BestAlignItem.m_EndPosInTarget; ii++ ) {
          CWordIndex e = m_pTargetSentence->GetAt( ii );
          cout << m_pLexiconSourceTarget->TargetVocabulary()->GetWord( e ) << " ";
        }
        cout << endl;
      }
		  
      // and add this phrase pair to the list of all phrase pairs
      PhraseAlignItems.Append( BestAlignItem );
		  
    }  // loop over all source phrases of given length
  } // loop over all possible source phrase lengths 


  //cout << "done with finding all phrase pair" << endl << flush;

  return true;
}



bool
CPhraseAlignerMS::FindSomePhrasePairs( CSentence *pSourceSentence, CSentence *pTargetSentence,
                                       CArray< CPhraseAlignItemMS > & PhraseAlignItems )
{
  //cout << "CPhraseAlignerMS::FindSomePhrasePair" << endl << flush;

  // calculate for all source phrases between MinPhraseLength and MaxPhraseLength
  // the alignment and keep track of the best results, this is then returned.
  // i.e. for each sentence pair we get one phrase pair
  // we can evaluate based on combined score, or look into perplexity gain,
  // as we do in sentence splitting (could be total or per word)

  m_pSourceSentence = pSourceSentence;
  m_pTargetSentence = pTargetSentence;

  // first we need to fill the probability matrices
  // length of the sentences
  m_J = pSourceSentence->Length();
  m_I = pTargetSentence->Length();
  //cout << "Length source " << m_J << " length target " << m_I << endl;

  // if sentences are longer than we can store probabilities, we skip it
  if ( (m_J >= MAX_SENTENCE_LENGTH) || (m_I >= MAX_SENTENCE_LENGTH) ) {
    cout << "WARNING: Sentence is too long, we skip it" << endl;
    return false;
  }

  // first, get all the probabilities and put them into static matrix
  // get Source->Target Probabilities
  GetProbabilities( m_pSourceSentence, m_pTargetSentence, m_pLexiconSourceTarget, &(m_ProbsSourceTarget[ 0 ][ 0 ]) );
  // renormalize probabilities for the current sentence pair
  Renormalize( m_J, m_I, &(m_ProbsSourceTarget[ 0 ][ 0 ]), &(m_RenProbsSourceTarget[ 0 ][ 0 ]) );

  //get Target->Source Probabilities
  if( m_AlignBidirectional ) {
    GetProbabilities( m_pTargetSentence, m_pSourceSentence, m_pLexiconTargetSource, &(m_ProbsTargetSource[ 0 ][ 0 ]) );
    // renormalize probabilities for the current sentence pair
    Renormalize( m_I, m_J, &(m_ProbsTargetSource[ 0 ][ 0 ]), &(m_RenProbsTargetSource[ 0 ][ 0 ]) );
  }

  int StartPosInSource, EndPosInSource;
  int StartPosInTarget, EndPosInTarget;
  float LogProbFE, LogProbEF, LogProbFE2, LogProbEF2, LogProbFE3, LogProbEF3;
  float BestLogProb = 999999.99;
  CPhraseAlignItemMS AlignItem;
  CPhraseAlignItemMS BestAlignItem;


  // first we get the frequency information from the suffixarray
  

  for ( int n = PhraseAlignItems.MinIndex(); n <= PhraseAlignItems.MaxIndex(); n++ ) {

    AlignItem = PhraseAlignItems[ n ];
    //cout << "Align for " << AlignItem.m_StartPosInSource << " " << AlignItem.m_EndPosInSource << endl;
    //cout << AlignItem << endl;

    StartPosInSource = m_pSourceSentence->MinIndex() + AlignItem.m_StartPosInSource;
    EndPosInSource   = m_pSourceSentence->MinIndex() + AlignItem.m_EndPosInSource;

    // calculate possible target phrase lengths
    // don't want too short or too long translations
    // on average the target length should be source length times stretch factor
    // where stretch factor is the ratio of target corpus lenght to source corpus length
    // then target should not be longer then this average length time a elongation factor
    // and not shorter then the average length time a shortening factor
    // we allow for and addition +/-2 to have e.g. 1->3 alignemnts
    int LengthSourcePhrase = EndPosInSource - StartPosInSource + 1;
    int MinTargetLength = MinTargetPhraseLength( LengthSourcePhrase );
    int MaxTargetLength = MaxTargetPhraseLength( LengthSourcePhrase );
    
    if ( MinTargetLength < 1 )
      MinTargetLength = 1;

    if ( MaxTargetLength >= m_I )
      MaxTargetLength = m_I;


    // calculate minimum and maximum target boundaries
    
    //cout << "  MinTargetPhraseLength " << MinTargetPhraseLength << " MaxTargetPhraseLength " << MaxTargetPhraseLength << endl;

    // initiallize BestLogProb to large number, for testing we get the best align item for each source phrase
    BestLogProb = 999999.99;

    // loop over all possible target lengths
    for ( int TargetPhraseLength = MinTargetLength; TargetPhraseLength <= MaxTargetLength; TargetPhraseLength++ ) {
      //cout << "   Target Phrase Length is now " << TargetPhraseLength << endl;

      // loop over all target phrases of given length
      for ( StartPosInTarget = pTargetSentence->MinIndex();
            StartPosInTarget <= pTargetSentence->MaxIndex() - TargetPhraseLength + 1;
            StartPosInTarget++ ) {

        EndPosInTarget = StartPosInTarget + TargetPhraseLength - 1;
        //cout << "      StartPosInTarget = " << StartPosInTarget << " EndPosInTarget " << EndPosInTarget << endl;

        // align item to store results for one possible alignment
        //CPhraseAlignItemMS AlignItem;

        float LogProb = CalcAllAlignLogProbs( StartPosInSource, EndPosInSource, StartPosInTarget, EndPosInTarget, AlignItem );
        //cout << AlignItem << endl;

        // store best target phrase
        if ( LogProb < BestLogProb ) {
          BestAlignItem = AlignItem;
          BestLogProb = LogProb;
        }

      }  // loop over all target phrases of given length
    } // loop over all possible target phrase lengths
	  
    PhraseAlignItems[ n ] = BestAlignItem;

    AddWordIDsAndStrings( PhraseAlignItems[ n ] );

    //cout << "Best Align Item:" << endl;
    //cout << BestAlignItem << endl;
    //cout << "@BestPhrasePair # " << PhraseAlignItems[ n ].SourcePhrase() << " # " << PhraseAlignItems[ n ].TargetPhrase() << endl;

    //float LogProb = CalcAllAlignLogProbs( StartPosInSource, EndPosInSource, StartPosInTarget, EndPosInTarget, AlignItem );
    
  }

  /*
    // loop over all source phrases of given length
    for ( StartPosInSource = pSourceSentence->MinIndex();
          StartPosInSource <= pSourceSentence->MaxIndex() - SourcePhraseLength + 1;
          StartPosInSource++ ) {

      EndPosInSource = StartPosInSource + SourcePhraseLength - 1;
      //cout << "  StartPosInSource = " << StartPosInSource << " EndPosInSource " << EndPosInSource << endl;

      // calculate possible target phrase lengths
      // don't want too short or too long translations
      // on average the target length should be source length times stretch factor
      // where stretch factor is the ratio of target corpus lenght to source corpus length
      // then target should not be longer then this average length time a elongation factor
      // and not shorter then the average length time a shortening factor
      // we allow for and addition +/-2 to have e.g. 1->3 alignemnts
      int LengthSourcePhrase = EndPosInSource - StartPosInSource + 1;
      int MinTargetLength = MinTargetPhraseLength( LengthSourcePhrase );
      int MaxTargetLength = MaxTargetPhraseLength( LengthSourcePhrase );

      if ( MinTargetLength < 1 )
        MinTargetLength = 1;

      if ( MaxTargetLength >= m_I )
        MaxTargetLength = m_I;

      //cout << "  MinTargetPhraseLength " << MinTargetPhraseLength << " MaxTargetPhraseLength " << MaxTargetPhraseLength << endl;

      // initiallize BestLogProb to large number, for testing we get the best align item for each source phrase
      BestLogProb = 999999.99;

      // loop over all possible target lengths
      for ( int TargetPhraseLength = MinTargetLength; TargetPhraseLength <= MaxTargetLength; TargetPhraseLength++ ) {
        //cout << "   Target Phrase Length is now " << TargetPhraseLength << endl;

        // loop over all target phrases of given length
        for ( StartPosInTarget = pTargetSentence->MinIndex();
              StartPosInTarget <= pTargetSentence->MaxIndex() - TargetPhraseLength + 1;
              StartPosInTarget++ ) {

          EndPosInTarget = StartPosInTarget + TargetPhraseLength - 1;
          //cout << "      StartPosInTarget = " << StartPosInTarget << " EndPosInTarget " << EndPosInTarget << endl;

          // align item to store results for one possible alignment
          CPhraseAlignItemMS AlignItem;

          float LogProb = CalcAllAlignLogProbs( StartPosInSource, EndPosInSource, StartPosInTarget, EndPosInTarget, AlignItem );

          // store best target phrase
          if ( LogProb < BestLogProb ) {
            BestAlignItem = AlignItem;
            BestLogProb = LogProb;
          }

        }  // loop over all target phrases of given length
      } // loop over all possible target phrase lengths
	  

      AddWordIDsAndStrings( BestAlignItem );

      //cout << "Best Align Item:" << endl;
      //cout << BestAlignItem << endl;
      / *
      // get the source phrase: word IDs and string
      for ( int j = BestAlignItem.m_StartPosInSource; j <= BestAlignItem.m_EndPosInSource; j++ ) {
      CWordIndex f = m_pSourceSentence->GetAt( j );
      BestAlignItem.m_SourceWordIDs.Append( f );
      BestAlignItem.m_SourcePhrase += CString( " " ) + m_pLexiconSourceTarget->SourceVocabulary()->GetWord( f );
      }

      // get the target phrase: word IDs and string
      for ( int i = BestAlignItem.m_StartPosInTarget; i <= BestAlignItem.m_EndPosInTarget; i++ ) {
      CWordIndex e = m_pTargetSentence->GetAt( i );
      BestAlignItem.m_TargetWordIDs.Append( e );
      BestAlignItem.m_TargetPhrase += CString( " " ) + m_pLexiconSourceTarget->TargetVocabulary()->GetWord( e );
      }
	  
      cout << BestAlignItem << endl;
      * /

      if ( 0 ) {
        //cout << "Testing target phrase: " << BestAlignItem.m_TargetPhrase <<endl;
        cout << BestAlignItem << endl;
        cout << BestAlignItem.ScaledScore() << " # ";
        for ( int ii = BestAlignItem.m_StartPosInSource; ii <= BestAlignItem.m_EndPosInSource; ii++ ) {
          CWordIndex f = m_pSourceSentence->GetAt( ii );
          cout << m_pLexiconSourceTarget->SourceVocabulary()->GetWord( f ) << " ";
        }
        cout << " # ";
        for ( int ii = BestAlignItem.m_StartPosInTarget; ii <= BestAlignItem.m_EndPosInTarget; ii++ ) {
          CWordIndex e = m_pTargetSentence->GetAt( ii );
          cout << m_pLexiconSourceTarget->TargetVocabulary()->GetWord( e ) << " ";
        }
        cout << endl;
      }
		  
      // and add this phrase pair to the list of all phrase pairs
      PhraseAlignItems.Append( BestAlignItem );
		  
    }  // loop over all source phrases of given length
  } // loop over all possible source phrase lengths 


*/


  //cout << "done with finding all phrase pair" << endl << flush;

  return true;
}




bool
CPhraseAlignerMS::FindSomePhrasePairsMultiTrans( CSentence *pSourceSentence, CSentence *pTargetSentence,
                                                 CArray< CPhraseAlignItemMS > & PhraseAlignItems )
{
  //cout << "CPhraseAlignerMS::FindSomePhrasePairsMultiTrans" << endl << flush;

  // the same as FindSomePhrasePairs, but now not only the best translation
  // but multiple translations.

  

  // calculate for all source phrases between MinPhraseLength and MaxPhraseLength
  // the alignment and keep track of the best results, this is then returned.
  // i.e. for each sentence pair we get one phrase pair
  // we can evaluate based on combined score, or look into perplexity gain,
  // as we do in sentence splitting (could be total or per word)

  // we move the incoming phrase align items into a temporary array
  // all found (best and close to best) items are then written into PhraseAlignItems
  // this way we avoid that the best item occurs twice in the final list
  // without checking if an additional item is different from the best one

  CArray< CPhraseAlignItemMS > OriginalItems = PhraseAlignItems;
  PhraseAlignItems.RemoveAll();


  m_pSourceSentence = pSourceSentence;
  m_pTargetSentence = pTargetSentence;

  // first we need to fill the probability matrices
  // length of the sentences
  m_J = pSourceSentence->Length();
  m_I = pTargetSentence->Length();
  //cout << "Length source " << m_J << " length target " << m_I << endl;

  // if sentences are longer than we can store probabilities, we skip it
  if ( (m_J >= MAX_SENTENCE_LENGTH) || (m_I >= MAX_SENTENCE_LENGTH) ) {
    cout << "WARNING: Sentence is too long, we skip it" << endl;
    return false;
  }

  // first, get all the probabilities and put them into static matrix
  // get Source->Target Probabilities
  GetProbabilities( m_pSourceSentence, m_pTargetSentence, m_pLexiconSourceTarget, &(m_ProbsSourceTarget[ 0 ][ 0 ]) );
  // renormalize probabilities for the current sentence pair
  Renormalize( m_J, m_I, &(m_ProbsSourceTarget[ 0 ][ 0 ]), &(m_RenProbsSourceTarget[ 0 ][ 0 ]) );

  //get Target->Source Probabilities
  if( m_AlignBidirectional ) {
    GetProbabilities( m_pTargetSentence, m_pSourceSentence, m_pLexiconTargetSource, &(m_ProbsTargetSource[ 0 ][ 0 ]) );
    // renormalize probabilities for the current sentence pair
    Renormalize( m_I, m_J, &(m_ProbsTargetSource[ 0 ][ 0 ]), &(m_RenProbsTargetSource[ 0 ][ 0 ]) );
  }

  int StartPosInSource, EndPosInSource;
  int StartPosInTarget, EndPosInTarget;
  float LogProbFE, LogProbEF, LogProbFE2, LogProbEF2, LogProbFE3, LogProbEF3;
  float BestLogProb = 999999.99;
  CPhraseAlignItemMS AlignItem;
  CPhraseAlignItemMS BestAlignItem;
  CArray< CPhraseAlignItemMS > AdditionalAlignItems;


  int MinIndex = OriginalItems.MinIndex();
  int MaxIndex = OriginalItems.MaxIndex();

  for ( int n = MinIndex; n <= MaxIndex; n++ ) {

    AlignItem = OriginalItems[ n ];
    //cout << "Align for " << AlignItem.m_StartPosInSource << " " << AlignItem.m_EndPosInSource << endl;
    //cout << AlignItem << endl;

    AdditionalAlignItems.RemoveAll();

    StartPosInSource = m_pSourceSentence->MinIndex() + AlignItem.m_StartPosInSource;
    EndPosInSource   = m_pSourceSentence->MinIndex() + AlignItem.m_EndPosInSource;

    // calculate possible target phrase lengths
    // don't want too short or too long translations
    // on average the target length should be source length times stretch factor
    // where stretch factor is the ratio of target corpus lenght to source corpus length
    // then target should not be longer then this average length time a elongation factor
    // and not shorter then the average length time a shortening factor
    // we allow for and addition +/-2 to have e.g. 1->3 alignemnts
    int LengthSourcePhrase = EndPosInSource - StartPosInSource + 1;
    int MinTargetLength = MinTargetPhraseLength( LengthSourcePhrase );
    int MaxTargetLength = MaxTargetPhraseLength( LengthSourcePhrase );
    
    if ( MinTargetLength < 1 )
      MinTargetLength = 1;

    if ( MaxTargetLength >= m_I )
      MaxTargetLength = m_I;

    //cout << "  MinTargetPhraseLength " << MinTargetPhraseLength << " MaxTargetPhraseLength " << MaxTargetPhraseLength << endl;

    // initiallize BestLogProb to large number, for testing we get the best align item for each source phrase
    BestLogProb = 999999.99;

    // loop over all possible target lengths
    for ( int TargetPhraseLength = MinTargetLength; TargetPhraseLength <= MaxTargetLength; TargetPhraseLength++ ) {
      //cout << "   Target Phrase Length is now " << TargetPhraseLength << endl;

      // loop over all target phrases of given length
      for ( StartPosInTarget = pTargetSentence->MinIndex();
            StartPosInTarget <= pTargetSentence->MaxIndex() - TargetPhraseLength + 1;
            StartPosInTarget++ ) {

        EndPosInTarget = StartPosInTarget + TargetPhraseLength - 1;
        //cout << "      StartPosInTarget = " << StartPosInTarget << " EndPosInTarget " << EndPosInTarget << endl;

        float LogProb = CalcAllAlignLogProbs( StartPosInSource, EndPosInSource, StartPosInTarget, EndPosInTarget, AlignItem );
        //cout << AlignItem << endl;

        // store best target phrase
        if ( LogProb < BestLogProb ) {
          BestAlignItem = AlignItem;
          BestLogProb = LogProb;
        }

        // but also store the other ones
        AdditionalAlignItems.Append( AlignItem );

      } // loop over all target phrases of given length
    } // loop over all possible target phrase lengths
	  
    // 
    OriginalItems[ n ] = BestAlignItem;
    AddWordIDsAndStrings( OriginalItems[ n ] );

    //cout << "Best Align Item:" << endl;
    //cout << OriginalItems[ n ];
    //cout << "Scaled score = " << OriginalItems[ n ].m_ScaledScores[ 0 ] << endl << endl;
    //cout << "@BestPhrasePair # " << PhraseAlignItems[ n ].SourcePhrase() << " # " << PhraseAlignItems[ n ].TargetPhrase() << endl;


    // now add all target phrases which are within beam width to the best one
    // also restrict to specified number of translations

    for ( int k = AdditionalAlignItems.MinIndex(); k <= AdditionalAlignItems.MaxIndex(); k++ ) {

      // keep only items close to the best item
      if ( ItemIsGoodEnough( AdditionalAlignItems[ k ] ) && ItemWithinBeamOfBestItem( AdditionalAlignItems[ k ], OriginalItems[ n ] ) ) {

        AddWordIDsAndStrings( AdditionalAlignItems[ k ] );
        //cout << "additional translation within beam " << AdditionalAlignItems[ k ];
        //cout << AdditionalAlignItems[ k ].m_ScaledScores[ 0 ] << " " << m_BeamFactor 
        //   << " " << OriginalItems[ n ].m_ScaledScores[ 0 ] << endl << endl;

        // add it to the phrase align item list which is returned
        PhraseAlignItems.Append( AdditionalAlignItems[ k ] );
      } // if within beam

    } // loop over all additional translations
	  
  } // loop over phrase align items

}



bool
CPhraseAlignerMS::ItemIsGoodEnough( CPhraseAlignItemMS &AlignItem )
{
  // here, we use the unscaled scores

  float m_MinimalRequiredScore = 10.0;

  switch( m_PruningStrategy ) {
  case PRUNE_MULTIPLICATIVE:
    //cout << "Pruning multiplicative" << endl;
    if ( AlignItem.m_Scores[ 0 ] <= m_MinimalRequiredScore ) {
      return true;
    }
    break;

  case PRUNE_ADDITIVE:
    //cout << "Pruning additive" << endl;
    if ( AlignItem.m_Scores[ 0 ] <= m_MinimalRequiredScore ) {
      return true;
    }
    break;

  case PRUNE_ADDITIVE_PERWORD:
    {
      //cout << "Pruning additive per word" << endl;
      int NumWords = AlignItem.m_EndPosInSource - AlignItem.m_StartPosInSource + 1;
      if ( AlignItem.m_Scores[ 0 ] <= NumWords  * m_MinimalRequiredScore ) {
        return true;
      }
    }
    break;

  }


  return false;

}


bool
CPhraseAlignerMS::ItemWithinBeamOfBestItem( CPhraseAlignItemMS &AlignItem, CPhraseAlignItemMS &BestItems )
{

  switch( m_PruningStrategy ) {
  case PRUNE_MULTIPLICATIVE:
    //cout << "Pruning multiplicative" << endl;
    if ( AlignItem.m_ScaledScores[ 0 ] <= m_BeamFactor * BestItems.m_ScaledScores[ 0 ] ) {
      return true;
    }
    break;

  case PRUNE_ADDITIVE:
    //cout << "Pruning additive" << endl;
    if ( AlignItem.m_ScaledScores[ 0 ] <= m_BeamFactor + BestItems.m_ScaledScores[ 0 ] ) {
      return true;
    }
    break;

  case PRUNE_ADDITIVE_PERWORD:
    {
      //cout << "Pruning additive per word" << endl;
      int NumWords = AlignItem.m_EndPosInSource - AlignItem.m_StartPosInSource + 1;
      if ( AlignItem.m_ScaledScores[ 0 ] <= ( NumWords  * m_BeamFactor ) + BestItems.m_ScaledScores[ 0 ] ) {
        return true;
      }
    }
    break;

  }

          

  return false;

}



float
CPhraseAlignerMS::CalcAllAlignLogProbs( int StartPosInSource, int EndPosInSource,
                                        int StartPosInTarget, int EndPosInTarget,
                                        CPhraseAlignItemMS &AlignItem )
{
  AlignItem.m_StartPosInSource = StartPosInSource;
  AlignItem.m_EndPosInSource   = EndPosInSource;
  AlignItem.m_StartPosInTarget = StartPosInTarget;
  AlignItem.m_EndPosInTarget   = EndPosInTarget;

  return CalcAllAlignLogProbs( AlignItem );
}



float
CPhraseAlignerMS::CalcAllAlignLogProbs( CPhraseAlignItemMS &AlignItem )
{
  //cout << "CalcAllAlignLogProbs for " << AlignItem << endl;

  float LogProbFE = m_pCalc->CalcSentAlignLogProb( AlignItem.m_StartPosInSource, AlignItem.m_EndPosInSource,
                                                   AlignItem.m_StartPosInTarget, AlignItem.m_EndPosInTarget,
                                                   m_J, m_I, &(m_RenProbsSourceTarget[ 0 ][ 0 ]) );
  //cout << "        LogProg f->e = " << LogProbFE << endl;
		  
  float LogProbFE2 = m_pCalc->CalcPhraseLogProb( AlignItem.m_StartPosInSource, AlignItem.m_EndPosInSource,
                                                 AlignItem.m_StartPosInTarget, AlignItem.m_EndPosInTarget, 
                                                 &(m_RenProbsSourceTarget[ 0 ][ 0 ]) );
		  
  float LogProbFE3 = m_pCalc->CalcPhraseLogProb( AlignItem.m_StartPosInSource, AlignItem.m_EndPosInSource,
                                                 AlignItem.m_StartPosInTarget, AlignItem.m_EndPosInTarget,
                                                 &(m_ProbsSourceTarget[ 0 ][ 0 ]) );


  float LogProbEF = m_pCalc->CalcSentAlignLogProb( AlignItem.m_StartPosInTarget, AlignItem.m_EndPosInTarget,
                                                   AlignItem.m_StartPosInSource, AlignItem.m_EndPosInSource,
                                                   m_I, m_J, &(m_RenProbsTargetSource[ 0 ][ 0 ]) );
  //cout << "        LogProg e->f = " << LogProbEF << endl;

  float LogProbEF2 = m_pCalc->CalcPhraseLogProb( AlignItem.m_StartPosInTarget, AlignItem.m_EndPosInTarget,
                                                 AlignItem.m_StartPosInSource, AlignItem.m_EndPosInSource, 
                                                 &(m_RenProbsTargetSource[ 0 ][ 0 ]) );
		  
  float LogProbEF3 = m_pCalc->CalcPhraseLogProb( AlignItem.m_StartPosInTarget, AlignItem.m_EndPosInTarget,
                                                 AlignItem.m_StartPosInSource, AlignItem.m_EndPosInSource,
                                                 &(m_ProbsTargetSource[ 0 ][ 0 ]) );


  AlignItem.m_Scores[ 1 ] = LogProbFE;
  AlignItem.m_Scores[ 2 ] = LogProbEF;
  AlignItem.m_Scores[ 3 ] = LogProbFE2;
  AlignItem.m_Scores[ 4 ] = LogProbEF2;
  AlignItem.m_Scores[ 5 ] = LogProbFE3;
  AlignItem.m_Scores[ 6 ] = LogProbEF3;

  AddScaledAndCombinedScores( AlignItem );

  return AlignItem.m_ScaledScores[ 0 ];
}


void
CPhraseAlignerMS::AddWordIDsAndStrings( CPhraseAlignItemMS &AlignItem )
{
  // get the source phrase: word IDs and string

  // reset the strings
  AlignItem.m_SourcePhrase = "";
  AlignItem.m_TargetPhrase = "";
  AlignItem.m_SourceWordIDs.RemoveAll();
  AlignItem.m_TargetWordIDs.RemoveAll();

  for ( int j = AlignItem.m_StartPosInSource; j <= AlignItem.m_EndPosInSource; j++ ) {
    CWordIndex f = m_pSourceSentence->GetAt( j );
    AlignItem.m_SourceWordIDs.Append( f );
    AlignItem.m_SourcePhrase += CString( " " ) + m_pLexiconSourceTarget->SourceVocabulary()->GetWord( f );
  }

  // get the target phrase: word IDs and string
  for ( int i = AlignItem.m_StartPosInTarget; i <= AlignItem.m_EndPosInTarget; i++ ) {
    CWordIndex e = m_pTargetSentence->GetAt( i );
    AlignItem.m_TargetWordIDs.Append( e );
    AlignItem.m_TargetPhrase += CString( " " ) + m_pLexiconSourceTarget->TargetVocabulary()->GetWord( e );
  }

  if ( 0 ) {
    //cout << "Testing target phrase: " << BestAlignItem.m_TargetPhrase <<endl;
    cout << AlignItem.m_ScaledScores[ 0 ] << " # ";
    for ( int ii = AlignItem.m_StartPosInSource; ii <= AlignItem.m_EndPosInSource; ii++ ) {
      CWordIndex f = m_pSourceSentence->GetAt( ii );
      cout << m_pLexiconSourceTarget->SourceVocabulary()->GetWord( f ) << " ";
    }
    cout << " # ";
    for ( int ii = AlignItem.m_StartPosInTarget; ii <= AlignItem.m_EndPosInTarget; ii++ ) {
      CWordIndex e = m_pTargetSentence->GetAt( ii );
      cout << m_pLexiconSourceTarget->TargetVocabulary()->GetWord( e ) << " ";
    }
    cout << endl;

    cout << "---" << endl;
    cout << "in CalcAllAlignLogProbs" << endl;
    cout << AlignItem << endl;
    cout << "---" << endl;
  }

}



void
CPhraseAlignerMS::GetProbabilities( CSentence *pSourceSentence, CSentence *pTargetSentence,
                                    CLexicon *pLexicon, float *Probs )
{
  Trace( 2, "CPhraseAlignerMS::GetProbabilities" );

  CWordIndex f, e;
  int i, j;

  // when using empty word, then we use p( f | 0 ) for F->E or p ( e | 0 ) for E->F
  if ( m_UseEmptyWord ) {

    // for F->E use empty word p( f | 0 )
    e = CVocabulary::EMPTYWORD;
    int IndexE0 = pTargetSentence->MinIndex() - 1;
    for ( int j = pSourceSentence->MinIndex(); j <= pSourceSentence->MaxIndex(); j++ ) {
      f = (*pSourceSentence)[ j ];
      Probs[j * MAX_SENTENCE_LENGTH + IndexE0] = pLexicon->Pr( f, e );
      //Probs[j * MAX_SENTENCE_LENGTH + IndexE0] = ((float)(j+1))/(IndexE0 + 100);
      //Trace( 0, Wf << " " << We << " p = " << m_ProbsFE[ j ][ IndexE0 ] );
    }
    
  } // end if ( m_UseEmptyWord )



  //Save Probabilies in Matrix
  for ( j = 1; j <= pSourceSentence->Length(); j++ ) {
    for ( i = 1; i <= pTargetSentence->Length(); i++ ) {
      Trace( 2, "j = " << j << " i = " << i ); 
      f = pSourceSentence->GetAt( j );
      e = pTargetSentence->GetAt( i );
      Probs[j* MAX_SENTENCE_LENGTH + i] = pLexicon->Pr( f, e );
      //Probs[j* MAX_SENTENCE_LENGTH + i] = ((float)(j+1))/(i+1);     
      Trace( 2,  "j = " << j << " i = " << i 
             << " f = " << f << " e = " << e << " p = " << pLexicon->Pr( f, e )
             << " stored as " << Probs[ j * MAX_SENTENCE_LENGTH + i] );
    }
  }

  //cout << "matrices initialized" << endl;
}



void
CPhraseAlignerMS::Renormalize(int SourceLength, int TargetLength,
                              float *Probs, 
                              float *RenProbs)
{
  float SumP;
  int i, j;
  int min_i;

  //Setting probs for emty word
  if ( m_UseEmptyWord ) {
    min_i = 0;
    Probs[0] =  0;
  } else {
    min_i = 1;
  }


  // renormalize
  for ( j = 1; j <= SourceLength; j++ ) {
    SumP = 0.0;
    for ( i = min_i; i <= TargetLength; i++ ) {
      SumP += Probs[j* MAX_SENTENCE_LENGTH + i];
    }
    for ( i = min_i; i <= TargetLength; i++ ) {
      RenProbs[j* MAX_SENTENCE_LENGTH + i] = Probs[j* MAX_SENTENCE_LENGTH + i] / SumP;
    }
  }
}



float
CPhraseAlignerMS::FindCenterMax( int StartSource, int EndSource )
{
  // find 'center' of target phrase for source phrase
  // if target phrase is at sentence beginning or sentence end
  // the center might be off.  Therefore iteratate over smaller
  // and smaller target sentence areas
  int n = 0;
  float P, BestP;
  int Best_i;
  int SumBest_i = 0;
  int i, j;

  int i1 = 1;
  int i2 = m_I;
  float i_mid;

  if ( m_UseEmptyWord )
    i1 = 0;

  //cout << "Find center max - Source phrase positions: " << StartSource << " " << EndSource << endl;

  do {

    SumBest_i = 0;

    for ( j = StartSource; j <= EndSource; j++ ) {
      BestP = 0.0;
      for ( i = i1; i <= i2; i++ ) {
        n++;
        P = m_RenProbsSourceTarget[ j ][ i ];
        //cout << j << " " << i << " " << P << " " << BestP << endl;
        if ( P > BestP ) {
          BestP = P;
          Best_i = i;
          //cout << "new best i " << Best_i << endl;
        }
      }
      //cout << "j = " << j << " best i = " << Best_i << endl;
      SumBest_i += Best_i;
    }

    i_mid = SumBest_i / (EndSource - StartSource + 1);
    //cout << "SumBest_i = " << SumBest_i << " divide by source phrase length = " << (EndSource - StartSource + 1) << endl;
    //cout << "i1 = " << i1 << " i2 = " << i2 << " i_mid = " << i_mid << endl;

    if ( i_mid > ((i2 + i1 + 1 ) / 2) ) {
      //cout << " i_mid > ((i2 + i1 + 1 ) / 2) = " << (i2 + i1 + 1 ) / 2 << " i1++" << endl;
      i1++;
    } else {
      //cout << " i_mid <= ((i2 + i1 + 1 ) / 2) = " << (i2 + i1 + 1 ) / 2 << " i2--" << endl;
      i2--;
    }

  } while ( (i2 - i1 ) > (EndSource - StartSource + 1) * 1.5 );


  //cout << "Center of target phrase is " << i_mid 
  //   << " for StartSource " << StartSource << endl;


  return i_mid;
}


float
CPhraseAlignerMS::FindCenterSum( int StartSource, int EndSource )
{
  // find 'center' of target phrase for source phrase
  // if target phrase is at sentence beginning or sentence end
  // the center might be off.  Therefore iteratate over smaller
  // and smaller target sentence areas
  int n = 0;
  float P, SumP,MV,SumMV, SumSV;
  int i, j;

  int i1 = 1;
  int i2 = m_I;
  float i_mid;

  if ( m_UseEmptyWord )
    i1 = 0;

  cout << "Find Center Sum - Source phrase positions: " << StartSource << " " << EndSource << endl;

  do {

    SumMV = 0.0;
    SumSV = 0.0;
	
    for ( j = StartSource; j <= EndSource; j++ ) {
	  
      SumP = 0.0;
      MV = 0.0;
      for ( i = i1; i <= i2; i++ ) {
        n++;
        P = m_RenProbsSourceTarget[ j ][ i ];
        MV += i * P;
        SumP += P;
        //cout << j << " " << i << " " << P << " " << BestP << endl;
		
      }
      cout << "MV: " << MV << " SumP: " << SumP << " MV/SumP: " << MV << endl;
      //SumMV += MV / SumP;
      SumMV += MV;
      SumSV += MV/SumP;
      //cout << "j = " << j << " best i = " << Best_i << endl;
    }

    i_mid = SumMV / (EndSource - StartSource + 1);
    cout << "I_mid: " << i_mid << " " << SumSV / (EndSource - StartSource + 1) << endl;
    //cout << "SumBest_i = " << SumBest_i << " divide by source phrase length = " 
    // << (EndSource - StartSource + 1) << endl;
    cout << "i1 = " << i1 << " i2 = " << i2 << " i_mid = " << i_mid << endl;

    if ( i_mid > ((i2 + i1 + 1 ) / 2) ) {
      //cout << " i_mid > ((i2 + i1 + 1 ) / 2) = " << (i2 + i1 + 1 ) / 2 << " i1++" << endl;
      i1++;
    } else {
      //cout << " i_mid <= ((i2 + i1 + 1 ) / 2) = " << (i2 + i1 + 1 ) / 2 << " i2--" << endl;
      i2--;
    }

  } while ( (i2 - i1 ) > (EndSource - StartSource + 1) * 1.5 );


  cout << "Center of target phrase is " << i_mid 
       << " for StartSource " << StartSource << endl;


  return i_mid;
}


float
CPhraseAlignerMS::FindCenterBlock( int StartSource, int EndSource )
{
  // find 'center' of target phrase for source phrase
  // assume target phrase length is source phrase length +- 1
  // calculate p( f | e ) for all (i = 1...I - l, i2 = i1 + l )

  int n = 0;
  float LogP, BestLogP = 999999.9;
  int Best_i1 = 0, Best_i2 = 0;
  int i, j;

  int i1 = 1;
  int i2 = m_I;
  float i_mid = (int) ( (m_I + 1) / 2 );

  int l = EndSource - StartSource;

  // we assume that the target length is about the same as the source length

  //cout << "Find center block - Source phrase positions: " << StartSource << " " << EndSource << " l = " << l << endl;

  // make sure that we do the calculation when phrase length = sentence length, i.e. l = m_I

  int min_l = l - 1;
  int max_l = l + 1;

  // the length of the target phrase can not be shorter than 1 and not longer then the target sentence
  if ( min_l < 1 ) min_l = 1;
  if ( max_l > m_I ) max_l = m_I;
  if ( min_l > max_l ) min_l = max_l;

  //cout << "min_l = " << min_l << " max_l " << max_l << endl;
  /* (min_l < 1 ) ? l=1 : l=min_l; */
  
  for ( l = min_l; l <= max_l; l++ ) {
    for ( i1 = 1; i1 <= m_I - l + 1; i1++ ) {
    
      i2 = i1 + l - 1;

      assert( i1 >= 1);
      /*
      if ( i1 < 1 ) {
        i1 = 1;
      }
       */
      
      //assert( i1 <= m_I ); 
      if ( i1 > m_I ) {
        cout << "i1=" << i1 << " m_I="<<m_I<<" l="<<l<<" min_l="<< min_l<<endl;
        assert( i1 <= m_I );
        i1 = m_I;
      }

      assert( i2 <= m_I );
      /*
      if ( i2 > m_I ) {
        i2 = m_I;
      }
       */
      
      LogP = 
        m_pCalc->CalcPhraseLogProb( StartSource, EndSource, i1, i2, &(m_RenProbsSourceTarget[0][0]) )
        + m_pCalc->CalcPhraseLogProb( i1, i2, StartSource, EndSource, &(m_RenProbsTargetSource[0][0]) );
      
      //cout << i1 << " " << i2 << " LogP = " << LogP << endl;
      if ( LogP < BestLogP ) {
        BestLogP = LogP;
        Best_i1 = i1;
      Best_i2 = i2;
      }
    }


  }

  i_mid = (float)(Best_i1 + Best_i2) / 2.0;

  //cout << "Center of target phrase according to FindCenterBlock is " << i_mid 
  //   << " for StartSource " << StartSource << endl;


  return i_mid;
}



int CPhraseAlignerMS::GetNumberOfScores()
{
  return m_numberOfScores;
}


int CPhraseAlignerMS::GetNumberOfCombinedScores()
{
  return m_numberOfCombinedScores;
}




void CPhraseAlignerMS::DisplayProbs(float* SourceTarget, float* TargetSource)
{
  int maxLen = GetMaxWordLength();

  cout << "Probabilities: (Source-Target)" << endl;
  for(int i = m_I; i >= 1; i--) {
    cout.width(maxLen);
    cout.setf(ios::left);
    cout << m_pLexiconSourceTarget->TargetVocabulary()->GetWord(m_pTargetSentence->GetAt( i )) << " |" <<flush;
    
    int j;
    if(m_UseEmptyWord)
      j = 0;
    else
      j = 1;
      
    for(; j <= m_J; j++) {
      cout.setf(ios::right | ios::fixed);
      cout.precision(3);
      cout.width(8);
      cout << SourceTarget[j * MAX_SENTENCE_LENGTH + i];
    }
    cout << endl; 
  }
  if(m_UseEmptyWord) {
    cout.width(maxLen);
    cout.setf(ios::left);
    cout << "<e>" << " |" <<flush;
    
    int j;
    if(m_UseEmptyWord)
      j = 0;
    else
      j = 1;
      
    for(; j <= m_J; j++) {
      cout.setf(ios::right | ios::fixed);
      cout.precision(3);
      cout.width(8);
      cout << SourceTarget[j * MAX_SENTENCE_LENGTH + 0];
    }
    cout << endl; 


  }

  DisplaySourceSentence(maxLen);
  
  cout << "Probabilities: (Target-Source)" << endl;
  for(int i = m_I; i >= 1; i--) {
    cout.width(maxLen);
    cout.setf(ios::left);
    cout << m_pLexiconSourceTarget->TargetVocabulary()->GetWord(m_pTargetSentence->GetAt( i )) << " |" << flush;
    
    int j;
    if(m_UseEmptyWord)
      j = 0;
    else
      j = 1;
      
    for(; j <= m_J; j++) {
      cout.setf(ios::right | ios::fixed);
      cout.precision(3);
      cout.width(8);
      cout << TargetSource[i * MAX_SENTENCE_LENGTH + j];
    }
    cout << endl; 
  }
  if(m_UseEmptyWord) {
    cout.width(maxLen);
    cout.setf(ios::left);
    cout << "<e>" << " |" <<flush;
    
    int j;
    if(m_UseEmptyWord)
      j = 0;
    else
      j = 1;
      
    for(; j <= m_J; j++) {
      cout.setf(ios::right | ios::fixed);
      cout.precision(3);
      cout.width(8);
      cout << TargetSource[0 * MAX_SENTENCE_LENGTH + j];
    }
    cout << endl; 


  }

  DisplaySourceSentence(maxLen);
  
}



int CPhraseAlignerMS::GetMaxWordLength()
{
  int i;
  int maxLen = 3;
  const char* pString;
  // get length of longest e word
  for (i = m_I; i >= 1; i--) {
    pString = m_pLexiconSourceTarget->TargetVocabulary()->GetWord(m_pTargetSentence->GetAt( i ));
    Trace( 5, pString );
    if (strlen(pString) > (uint) maxLen) 
      maxLen = strlen(pString);
  }
  return maxLen;
}


void CPhraseAlignerMS::DisplaySourceSentence(int maxLen)
{
  int MaxWordLength = 40;
  const char *pString = NULL;
  int MaxSLen = 150;
  char WordMatrix[40][150];
  int NoSpace;
  int i, j;
  CWordIndex Word;


  // display line
  for (i = 0; i < maxLen; i++)
    cout << "-";
  cout << "-|";
  if (m_J > 0) {
    for (j = 0; j <=  m_J; j++) {
      cout << "--------";
    }
  }
  cout << endl;

  // display french words at bottom
  //Trace( 1, "display french words at bottom" );
  char Blanks[150];

  maxLen += 4;

  for(i = 0; i < maxLen; i++)
    Blanks[i] = ' ';
  Blanks[maxLen]='\0';

  for (j = 0; j < MaxSLen; j++)
    for (i=0; i < MaxWordLength; i++)
      WordMatrix[i][j] = ' ';

  if (m_J > 0) {
    pString = "<e>";
    i = 0;
    while (pString[i]){
      WordMatrix[i][j] = (char) pString[i];
      i++;
    }

    for (j = 1; j <= m_J; j++) {
      pString = m_pLexiconSourceTarget->SourceVocabulary()->GetWord(m_pSourceSentence->GetAt( j ));
      cout.setf(ios::left);
      i=0;
      while (pString[i]) {
        WordMatrix[i][j] = (char) pString[i];
        i++;
      }
    }
  }

  NoSpace = 1;
  for (i=0; ((i < MaxWordLength) && (NoSpace == 1)); i++) {
    NoSpace = 0;
    cout << Blanks << " ";
    // for diagonal output an additional blank in the next iteration
    Blanks[maxLen] = ' ';
    maxLen++;
    Blanks[maxLen] = '\0';
    if (m_J > 0) {
      for (j = 0; j <= m_J; j++) {
        cout << WordMatrix[i][j] << "       ";
        if (WordMatrix[i][j] != ' ')
          NoSpace = 1;
      }
    }
    cout << "\n";
  }
    
  cout << "\n";
  cout.flush();

}



int CPhraseAlignerMS::MinTargetPhraseLength( int LengthSourcePhrase )
{
  return (int)(0.5 + m_AddConstShorter + LengthSourcePhrase * m_StretchFactor * m_FactorShorter);
}


int CPhraseAlignerMS::MaxTargetPhraseLength( int LengthSourcePhrase )
{
  return (int)(0.5 + m_AddConstLonger + LengthSourcePhrase * m_StretchFactor * m_FactorLonger);
}

