#include <stdlib.h>
#include <PhraseAlignMSTag.hh>

// switch on/off tracing for this module here
#undef NO_TRACE
//#define NO_TRACE 1


#include <iostream>
using std::ostream;

using namespace std;



//============================================================
CPhraseAlignerMSTag::CPhraseAlignerMSTag()
{

  m_pMappingSource = NULL;
  m_pMappingTarget = NULL;

  m_pTagLexiconSourceTarget = NULL;
  m_pTagLexiconTargetSource = NULL;

  m_numberOfScores = 12;
  m_numberOfCombinedScores = 3; 
  
}


CPhraseAlignerMSTag::~CPhraseAlignerMSTag()
{
}


void
CPhraseAlignerMSTag::Init(CLexicon *pLexiconSourceTarget, CLexicon *pLexiconTargetSource, CPhraseAlignProbCalc *pCalc,
			      CMapWordToTag *pMappingSource,CMapWordToTag *pMappingTarget , 
			      CLexicon *pTagLexiconSourceTarget, CLexicon *pTagLexiconTargetSource)
{
  Trace( 2, "CPhraseAlignerMS::Init" );
  
  //Initialising Lexica
  m_pLexiconSourceTarget = pLexiconSourceTarget;
  m_pLexiconTargetSource = pLexiconTargetSource;
  m_pTagLexiconSourceTarget = pTagLexiconSourceTarget;
  m_pTagLexiconTargetSource = pTagLexiconTargetSource;

  //Setting Mapping
  m_pMappingSource = pMappingSource;
  m_pMappingTarget = pMappingTarget;

  //Setting Bidirectional
  if ( m_pLexiconTargetSource ) {

    Trace( 2, "CPhraseAlignerMS::Init with two lexicons" );
    m_AlignBidirectional = true;
  } else {
	Trace( 2, "CPhraseAlignerMS::Init with one lexicons" );
    m_AlignBidirectional = false;
  }

  m_pCalc = pCalc;
}

bool 
CPhraseAlignerMSTag::AlignPhrase( CSentence *pSourceSentence, CSentence *pTargetSentence,
				   int StartPosInSource, int EndPosInSource,
				   CSortedArray< CPhraseAlignItemMS > & PhraseAlignItems )
{
  CWordIndex f, e;
  int i, j, k, min_i, max_i;
  int LengthSourcePhrase = EndPosInSource - StartPosInSource + 1;


#if DEBUG
  int DisplaySourcePhrase = 0;
  int DisplaySentence = 0;
  int DisplayProb = 0;
  int DisplayAlignItem = 0;



  if(DisplaySourcePhrase || DisplaySentence || DisplayProb || DisplayAlignItem)
    cout << "Starting AlignPhrase" << endl;
#endif


  Trace( 1, "CPhraseAlignerMSIBM1::AlignPhrase" );

  // use IMB1 alignment model, i.e. only lexicon to
  // find i1 and i2 on target side such that the alignment
  // probability is highest when aligning f_j1 ... f_j2 to
  // e_i1 ... e_i2, and aligning outside f_j to outside e_i
  // i.e. caculate argmax_i1_i2 { Prod_j1...j2 Sum_i1...i2 p(f_j|e_i)
  // x Prod_j!=j1...j2 Sum_i!=i1...i2 p(f_j|e_i) }

  if ( !m_pLexiconSourceTarget && !m_pLexiconTargetSource ) {
	cerr << "no lexicon provided to calculate alignment" << endl;
	exit( -3);
  }

  if ( !m_pTagLexiconSourceTarget && !m_pTagLexiconTargetSource ) {
	cerr << "no Tag lexicon provided to calculate alignment" << endl;
	exit( -3);
  }
  if ( !m_pMappingSource || !m_pMappingTarget ) {
	cerr << "no Mapping Word->Tag provided to calculate alignment" << endl;
	exit( -3);
  }
  if ( m_AlignBidirectional && (!m_pLexiconSourceTarget || !m_pLexiconTargetSource) ) {
	cerr << "calculation of alignment in both directions requested, but not both lexicons provided" << endl;
	exit( -3);
  }

#if DEBUG
 
  if ( DisplaySourcePhrase ) {
    cout << "Source Phrase: ";
    for ( j = StartPosInSource; j <= EndPosInSource; j++ ) {
      f = pSourceSentence->GetAt( j );
      cout << m_pLexiconSourceTarget->SourceVocabulary()->GetWord( f ) << " ";
    }
    cout << endl;
  }
#endif
    
  m_pSourceSentence = pSourceSentence;
  m_pTargetSentence = pTargetSentence;
 

  // length of the sentences
  m_J = pSourceSentence->Length();
  m_I = pTargetSentence->Length();


  // if sentences are longer than we can store probabilities, we skip it
  if ( (m_J >= MAX_SENTENCE_LENGTH) || (m_I >= MAX_SENTENCE_LENGTH) ) {
    return false;
  }


  //build Tag-Sentence
  // 1. Source Sentence
  CWordIndex tag;
  m_pSourceTagSentence = new CSentence;
  for(int j = 1; j <= m_J; j++) {
    e = pSourceSentence->GetAt( j );
    tag = m_pMappingSource->getTag(e);
    (*m_pSourceTagSentence)[j] = tag;
    //cout << " Tag for word " << m_pLexiconSourceTarget->SourceVocabulary()->GetWord( e ) << " is " << m_pTagLexiconSourceTarget->SourceVocabulary()->GetWord( tag ) << endl;
  }

  // 2. Target Sentence
  m_pTargetTagSentence = new CSentence;
  for(int i = 1; i <= m_I; i++) {
    f = pTargetSentence->GetAt( i );
    tag = m_pMappingTarget->getTag(f);
    (*m_pTargetTagSentence)[i] = tag;
    //cout << "Tag for word " << m_pLexiconSourceTarget->TargetVocabulary()->GetWord( f ) << " is " << m_pTagLexiconSourceTarget->TargetVocabulary()->GetWord( tag ) << endl;
  }
  

  // first, get all the probabilities and put them into static matrix
  
  //get Source->Target Probabilities
  GetProbabilities(m_pSourceSentence, m_pTargetSentence, m_pLexiconSourceTarget, &(m_ProbsSourceTarget[0][0]));
  // renormalize probabilities for the current sentence pair
  Renormalize(m_J, m_I, &(m_ProbsSourceTarget[0][0]),&(m_RenProbsSourceTarget[0][0]));
  
  //get Target->Source Probabilities
  if(m_AlignBidirectional) {
    GetProbabilities(m_pTargetSentence, m_pSourceSentence, m_pLexiconTargetSource, &(m_ProbsTargetSource[0][0]));
    // renormalize probabilities for the current sentence pair
    Renormalize(m_I,m_J,&(m_ProbsTargetSource[0][0]), &(m_RenProbsTargetSource[0][0]));
  }

  //get Source->Target Tag Probabilities
  GetProbabilities(m_pSourceTagSentence, m_pTargetTagSentence, m_pTagLexiconSourceTarget,&( m_ProbsTagSourceTarget[0][0]));
  // renormalize probabilities for the current sentence pair
  Renormalize(m_J, m_I, &(m_ProbsTagSourceTarget[0][0]),&(m_RenProbsTagSourceTarget[0][0]));
  
  //get Target->Source Probabilities
  if(m_AlignBidirectional) {
    GetProbabilities(m_pTargetTagSentence, m_pSourceTagSentence, m_pTagLexiconTargetSource, &(m_ProbsTagTargetSource[0][0]));
    // renormalize probabilities for the current sentence pair
    Renormalize(m_I,m_J, &(m_ProbsTagTargetSource[0][0]), &(m_RenProbsTagTargetSource[0][0]));
  }

#if DEBUG
  if(DisplaySentence) {
    cout << "Source Sentence: ";
    m_pSourceSentence->Display(m_pLexiconSourceTarget->SourceVocabulary());
    cout << endl;
    cout << "Target Sentence: ";
    m_pTargetSentence->Display(m_pLexiconSourceTarget->TargetVocabulary());
    cout << endl;  
    cout << "Source Tag Sentence: ";
    m_pSourceTagSentence->Display(m_pTagLexiconSourceTarget->SourceVocabulary());
    cout << endl;
    cout << "Target Tag Sentence: ";
    m_pTargetTagSentence->Display(m_pTagLexiconSourceTarget->TargetVocabulary());
    cout << endl;
  }



  if(DisplayProb){
    cout << "Word Probabilitis" << endl;
    DisplayProbs(&(m_ProbsSourceTarget[0][0]), &(m_ProbsTargetSource[0][0]));
    cout <<endl << "Tag Probabilities" << endl;
    DisplayProbs(&(m_ProbsTagSourceTarget[0][0]), &(m_ProbsTagTargetSource[0][0]));
  }
 
#endif

  //###############################
  //#                             #
  //#     Calcuate Scores         #
  //#                             #
  //###############################




  // find 'center' of target phrase for source phrase
  float i_mid = FindCenterSum( StartPosInSource, EndPosInSource );
 
  // don't want too short or too long translations
  // on average the target length should be source length times stretch factor
  // where stretch factor is the ratio target corpus lenght to source corpus length
  // then target should not be longer then this average length time a elongation factor
  // and not shorter then the average length time a shortening factor
  // we allow for and addition +/-2 to have e.g. 1->3 alignemnts
  int MinTargetLength = (int)(-1.5 + LengthSourcePhrase * m_StretchFactor * m_FactorShorter);
  int MaxTargetLength = (int)(2.5 + LengthSourcePhrase * m_StretchFactor * m_FactorLonger);
  if ( MinTargetLength < 1 )
    MinTargetLength = 1;

  /*
  cout << endl;
  cout << "Phrases positions " << StartPosInSource << " " << EndPosInSource << endl;

  cout << "LengthSourcePhrase " << LengthSourcePhrase 
	   << " StretchFactor " << m_StretchFactor
	   << " FactorShorter " << m_FactorShorter
	   << " FactorLonger " << m_FactorLonger << endl;

  cout << " MinTargetLength = " << MinTargetLength 
	   << " MaxTargetLength = " << MaxTargetLength <<endl;
  */

  // in case there are old results in the arrays
  PhraseAlignItems.Clear();

  // expand the target phrase around the center
  CSortedArray< double > Probs;
  // sort smallest first
  //Probs.SortReverse();
  CNumericalArray< double > AllLogP1;
  CNumericalArray< double > AllLogP2;

  double LogP, LogP1, LogP2;
  double BestLogP = 10000.0, BestLogP1 = 10000.0, BestLogP2 = 10000.0;

  bool WasNew;
  int mid = int( i_mid + 0.5 );

  int n = 0;  // counter for translations
  
  // here, we loop over the two boundaries
  for ( min_i = mid - MaxTargetLength; min_i <= mid + 1; min_i++ ) {
    for ( max_i = mid - 1; max_i <= mid + MaxTargetLength; max_i++ ) {

      // currently we do not assume that we have <sos> and <eos>
      if ( (min_i >= 1) && (max_i <= m_I) 
           && ((max_i - min_i + 1) >= MinTargetLength)
	   && ((max_i - min_i + 1) <= MaxTargetLength) ) {

	n++;
	CPhraseAlignItemMS AlignItem;

	// get the target phrase: word IDs and string
        for ( int i = min_i; i <= max_i; i++ ) {
          //cout << TextE.GetWord( SentenceID - 1, i ) << " ";
	  e = m_pTargetSentence->GetAt( i );
	  AlignItem.m_TargetWordIDs.Append( e );
	  AlignItem.m_TargetPhrase += CString( " " ) + m_pLexiconSourceTarget->TargetVocabulary()->GetWord( e );
        }

	// init all scores
	for(int i = 0; i <= m_numberOfScores; i++) {
	  AlignItem.m_Scores[i] = 0.0;
	}
	for(int i = 0; i <= m_numberOfCombinedScores; i++) {
	  AlignItem.m_CombinedScores[i] = 0.0;
	}
	for(int i = 0; i <= m_numberOfScores; i++) {
	  AlignItem.m_ScaledScores[i] = 0.0;
	}
	for(int i = 0; i <= m_numberOfCombinedScores; i++) {
	  AlignItem.m_CombinedScaledScores[i] = 0.0;
	}



        if ( m_pLexiconSourceTarget ) {

	  // now, we calculate the log prob for the phrase from total sentence alignment,
	  // dividing by the sentence length or the phrase length, to have per word costs (for pruning)
	  // later, i.e. after pruning and before we insert into lattice, we need to multiply again by source phrase length
	  LogP = m_pCalc->CalcSentAlignLogProb(StartPosInSource, EndPosInSource, min_i, max_i,m_J,m_I,&(m_RenProbsSourceTarget[0][0]));
	  //LogP /= m_J * LengthSourcePhrase;
	  //LogP /= m_J;
	  AlignItem.m_Scores[1] = LogP;
	  //AlignItem.m_ScaledScores[1] = m_Lambda1FE * LogP;

	  // probability of just aligning the phrase, using renormalized word probabilities
	  LogP = m_pCalc->CalcPhraseLogProb( StartPosInSource, EndPosInSource, min_i, max_i,&(m_RenProbsSourceTarget[0][0]));
	  //LogP /= ( EndPosInSource - StartPosInSource + 1 );
	  AlignItem.m_Scores[3] = LogP;
	  //AlignItem.m_ScaledScores[3] = m_Lambda2FE * LogP;

	  // phrase translation probability, using word probabilities without renormalization
	  LogP = m_pCalc->CalcPhraseLogProb( StartPosInSource, EndPosInSource, min_i, max_i,&(m_ProbsSourceTarget[0][0]));
	  //LogP /= ( EndPosInSource - StartPosInSource + 1 );
	  AlignItem.m_Scores[5] = LogP;
	  //AlignItem.m_ScaledScores[5] = m_Lambda3FE * LogP;

	  // now, we calculate the log prob of tags for the phrase from total sentence alignment,
	  // dividing by the sentence length or the phrase length, to have per word costs (for pruning)
	  // later, i.e. after pruning and before we insert into lattice, we need to multiply again by source phrase length
	  LogP = m_pCalc->CalcSentAlignLogProb(StartPosInSource, EndPosInSource, min_i, max_i,m_J,m_I,&(m_RenProbsTagSourceTarget[0][0]) );
	  //LogP /= m_J * LengthSourcePhrase;
	  //LogP /= m_J;
	  AlignItem.m_Scores[7] = LogP;
	  //AlignItem.m_ScaledScores[7] = m_Lambda1FETag * LogP;

	  // probability of just aligning the phrase, using renormalized word probabilities
	  LogP = m_pCalc->CalcPhraseLogProb( StartPosInSource, EndPosInSource, min_i, max_i,&(m_RenProbsTagSourceTarget[0][0]));
	  //LogP /= ( EndPosInSource - StartPosInSource + 1 );
	  AlignItem.m_Scores[9] = LogP;
	  //AlignItem.m_ScaledScores[9] = m_Lambda2FETag * LogP;

	  // phrase translation probability, using word probabilities without renormalization
	  LogP = m_pCalc->CalcPhraseLogProb( StartPosInSource, EndPosInSource, min_i, max_i, &(m_ProbsTagSourceTarget[0][0]));
	  //LogP /= ( EndPosInSource - StartPosInSource + 1 );
	  AlignItem.m_Scores[11] = LogP;
	  //AlignItem.m_ScaledScores[11] = m_Lambda3FETag * LogP;

	}

        if ( m_pLexiconTargetSource ) {

	  // probability of aligning phrase, using renormalized word probabilities
	  // divide by target phrase length and multiply by target phrase length
          LogP = m_pCalc->CalcSentAlignLogProb( min_i, max_i, StartPosInSource, EndPosInSource,m_I, m_J, &(m_RenProbsTargetSource[0][0]));
          //LogP /= m_I;
	  //LogP *= LengthSourcePhrase;
	  AlignItem.m_Scores[2] = LogP;
	  //AlignItem.m_ScaledScores[2] = m_Lambda1EF * LogP;

	  // now, we calculate the log prob for the phrase from total sentence alignment,
	  // dividing by the sentence length, multiplied by source phrase length
	  LogP = m_pCalc->CalcPhraseLogProb(min_i, max_i, StartPosInSource, EndPosInSource,&(m_RenProbsTargetSource[0][0]));
	  //cout << "LogP " << LogP << " TargetLength " << ( max_i - min_i + 1 ) 
	  //   << " SourceLength " << LengthSourcePhrase << endl;
	  //LogP /= ( max_i - min_i + 1 );
	  //LogP *= LengthSourcePhrase;
	  AlignItem.m_Scores[4] = LogP;
	  //AlignItem.m_ScaledScores[4] = m_Lambda2EF * LogP;
		  
	  // phrase translation probability, using word probabilities without renormalization
	  LogP = m_pCalc->CalcPhraseLogProb(min_i, max_i, StartPosInSource, EndPosInSource,&(m_ProbsTargetSource[0][0]));
	  //LogP /= ( max_i - min_i + 1 );
	  //LogP *= LengthSourcePhrase;
	  AlignItem.m_Scores[6] = LogP;
	  //AlignItem.m_ScaledScores[6] = m_Lambda3EF * LogP;

	  // probability of aligning phrase, using renormalized word probabilities
	  // divide by target phrase length and multiply by target phrase length
          LogP = m_pCalc->CalcSentAlignLogProb( min_i, max_i, StartPosInSource, EndPosInSource,m_I, m_J, &(m_RenProbsTagTargetSource[0][0]));
          //LogP /= m_I;
	  //LogP *= LengthSourcePhrase;
	  AlignItem.m_Scores[8] = LogP;
	  //AlignItem.m_ScaledScores[8] = m_Lambda1EF * LogP;

	  // now, we calculate the log prob for the phrase from total sentence alignment,
	  // dividing by the sentence length, multiplied by source phrase length
	  LogP = m_pCalc->CalcPhraseLogProb(min_i, max_i, StartPosInSource, EndPosInSource,&(m_RenProbsTagTargetSource[0][0]));
	  //cout << "LogP " << LogP << " TargetLength " << ( max_i - min_i + 1 ) 
	  //   << " SourceLength " << LengthSourcePhrase << endl;
	  //LogP /= ( max_i - min_i + 1 );
	  //LogP *= LengthSourcePhrase;
	  AlignItem.m_Scores[10] = LogP;
	  //AlignItem.m_ScaledScores[10] = m_Lambda2EF * LogP;
		  
	  // phrase translation probability, using word probabilities without renormalization
	  LogP = m_pCalc->CalcPhraseLogProb(min_i, max_i, StartPosInSource, EndPosInSource,&(m_ProbsTagTargetSource[0][0]));
	  //LogP /= ( max_i - min_i + 1 );
	  //LogP *= LengthSourcePhrase;
	  AlignItem.m_Scores[12] = LogP;
	  //AlignItem.m_ScaledScores[12] = m_Lambda3EF * LogP;
	  
        }

	//calculating scaled scores
	for(int i = 1; i <= m_numberOfScores; i++) {
	  AlignItem.m_ScaledScores[i] =m_ScalingFactors[i] * AlignItem.m_Scores[i];
	}


	// now, we combine these scores according to the scaling factors
	for(int i = 1; i <= m_numberOfScores; i++) {
	  AlignItem.m_Scores[0] +=AlignItem.m_Scores[i];
	}	

	
	//Combining the the same type of scores from both direction
	AlignItem.m_CombinedScores[1] =AlignItem.m_Scores[1] + AlignItem.m_Scores[2];
	AlignItem.m_CombinedScores[2] =AlignItem.m_Scores[3] + AlignItem.m_Scores[4];
	AlignItem.m_CombinedScores[3] =AlignItem.m_Scores[5] + AlignItem.m_Scores[6];

	// and the scaled ones
	for(int i = 1; i <= m_numberOfScores; i++) {
	  AlignItem.m_ScaledScores[0] +=AlignItem.m_ScaledScores[i];
	}

	//Combining the the same type of scores from both directions
	AlignItem.m_CombinedScaledScores[1] =AlignItem.m_ScaledScores[1] + AlignItem.m_ScaledScores[2];
	AlignItem.m_CombinedScaledScores[2] =AlignItem.m_ScaledScores[3] + AlignItem.m_ScaledScores[4];
	AlignItem.m_CombinedScaledScores[3] =AlignItem.m_ScaledScores[5] + AlignItem.m_ScaledScores[6];


#if DEBUG
	if ( DisplayAlignItem ) {
	  cout << "Testing target phrase: " << AlignItem.m_TargetPhrase <<endl;
	  cout << AlignItem << endl;	  
	  cout << endl;
	}
#endif

// 	cout << "Scores: ";
// 	for(int i = 1; i <= 6; i++) {
// 	  cout << AlignItem.m_Scores[i] << " ";
// 	}
// 	cout << endl;
// 	cout << " Scaled Scores: ";
// 	for(int i = 0; i <= 6; i++) {
// 	  cout << AlignItem.m_ScaledScores[i] << " ";
// 	}
// 	cout << endl;
	
// 	cout << "Scalingfactors: ";
// 	for(int i = 1; i <= 12; i++) {
// 	  cout << m_ScalingFactors[i] << " ";
// 	}
// 	cout << AlignItem;

	// store now in list of all good translations for this source phrase
	int Location;
	bool IsNew;
	if ( PhraseAlignItems.Find( AlignItem, Location ) ) {
	  //cout << " translation was found" << endl;
	  if ( AlignItem.Score() < PhraseAlignItems.GetAt( Location ).Score() ) {
	    // already stored, but current score is lower, i.e. better
	    // therefore overwrite it
	    //cout << " translation was found, but new score is smaller, i.e. overwrite " << endl;
	    PhraseAlignItems.Insert( AlignItem, IsNew );
	    //}
	    //} else {
	    //cout << " but new score is larger, i.e. don't use " << endl;
	  }
	} else {
	  //cout << " translation was not found - insert" << endl;
	  PhraseAlignItems.Insert( AlignItem, IsNew );
	}


        //cout << "P( " << min_i << ", " << max_i << " ) = " << LogP;
        //cout << "  inserted at " << k << endl;
      } else {
        ; //cout << "skip" << endl;
      }
    }
  }

  /*  
  if ( DEBUG ) {
	cout << endl;
	cout << "Best translation: ";
	
	cout << Min_i[ Min_i.MinIndex() ] 
		 << " " << Max_i[ Max_i.MinIndex() ] 
		 << " " << Scores[ Scores.MinIndex() ] << endl;
	
	for ( j = StartSource; j <= EndSource; j++ ) {
	  f = pSourceSentence->GetAt( j );
	  cout << m_pLexiconSourceTarget->SourceVocabulary()->GetWord( f ) << " ";
	}
	
	for ( i = Min_i[ Min_i.MinIndex() ]; i <= Max_i[ Max_i.MinIndex() ]; i++ ) {
	  e = pTargetSentence->GetAt( i );
	  cout << m_pLexiconSourceTarget->TargetVocabulary()->GetWord( e ) << " ";
	}
	
	cout << endl << endl;
  }
  
  */
  // and now some pruning
  //Trace( 0, "Number of found translations " << Probs.MinIndex() << " ... " << Probs.MaxIndex() );
  //cout << "BestLogP1 = " << BestLogP1 << "  BestLogP2 = " << BestLogP2 << "  BestLogP = " << BestLogP << endl;


  for ( int n = Probs.MinIndex(); n <= Probs.MaxIndex(); n++ ) {

	/*
	if ( DEBUG  ) {
	  cout << Min_i[ n ] << " " << Max_i[ n ] << " " << AllLogP1[ n ] << " " << AllLogP2[ n ]
		//		 << " " << Probs[ n ]
		   << " " << Scores[ n ] << "   ";
	  
	  for ( j = StartSource; j <= EndSource; j++ ) {
		f = pSourceSentence->GetAt( j );
		cout << m_pLexiconSourceTarget->SourceVocabulary()->GetWord( f ) << " ";
	  }
	  cout << " # ";
	  for ( i = Min_i[ n ]; i <= Max_i[ n ]; i++ ) {
		e = pTargetSentence->GetAt( i );
		cout << m_pLexiconSourceTarget->TargetVocabulary()->GetWord( e ) << " ";
	  }
	}
	*/

	//cout << "BeamFactor = " << m_BeamFactor << endl;

	/*
	// we keep all translations which are within beam to best scores
	// plus also all with small probs 
	// (the BestLogPx can be 0.0, i.e. we would prune everything)

	double SmallFloorProb = 0.01;

	if ( (AllLogP1[ n ] > m_BeamFactor * (BestLogP1 + SmallFloorProb)) ) cout << " (pruned on P1)" << endl;
	if ( (AllLogP2[ n ] > m_BeamFactor * (BestLogP2 + SmallFloorProb)) ) cout << " (pruned on P2)" << endl;
   	if ( (Scores[ n ]   > m_BeamFactor * (BestLogP  + SmallFloorProb)) ) cout << " (pruned on P)" << endl;


	if ( (AllLogP1[ n ] > m_BeamFactor * (BestLogP1+ SmallFloorProb))
		 || (AllLogP2[ n ] > m_BeamFactor * (BestLogP2+ SmallFloorProb))
		 || (Scores[ n ] > m_BeamFactor * (BestLogP+ SmallFloorProb)) ) {

	  if ( DEBUG ) {
		cout << "    <<< will be pruned >>>";
	  }
	  Scores[ n ] = 10000.0;
	} else {
		cout << "    <<< will NOT be pruned >>>";
	}

	if ( DEBUG ) {
	  cout << endl;
	}
	*/

  }


/*
  if ( Min_i.MaxIndex() - Min_i.MinIndex() + 1 > m_MaxTranslations ) {
	Min_i.SetBoundaries( Min_i.MinIndex(), Min_i.MinIndex() + m_MaxTranslations - 1 );
	Max_i.SetBoundaries( Max_i.MinIndex(), Max_i.MinIndex() + m_MaxTranslations - 1 );
	Scores.SetBoundaries( Scores.MinIndex(), Scores.MinIndex() + m_MaxTranslations - 1 );
  }
*/

  //Trace( 0, "Number of returned translations " << Min_i.MinIndex() << " ... " << Min_i.MaxIndex() );

  //pSourceSentence->Display( m_pLexiconSourceTarget->SourceVocabulary() );
  //cout << endl;
  //pTargetSentence->Display( m_pLexiconSourceTarget->SourceVocabulary() );
  //cout << endl;

  //cout << endl;

  delete m_pSourceTagSentence;
  delete m_pTargetTagSentence;
  return true;
}



//-------------------------------------------------------------------

CPhraseAlignerMSTag222::CPhraseAlignerMSTag222() {
  m_numberOfScores = 8;
  m_numberOfCombinedScores = 2;

}

CPhraseAlignerMSTag222::~CPhraseAlignerMSTag222() {


}


bool CPhraseAlignerMSTag222::AlignPhrase( CSentence *pSourceSentence, CSentence *pTargetSentence,
					 int StartPosInSource, int EndPosInSource,
					 CSortedArray< CPhraseAlignItemMS > & PhraseAlignItems ){


  CWordIndex f, e;
  int i, j, k, min_i, max_i;
  int LengthSourcePhrase = EndPosInSource - StartPosInSource + 1;

#if DEBUG
  int DisplaySourcePhrase = 0;
  int DisplaySentence = 0;
  int DisplayProb = 0;
  int DisplayAlignItem = 0;
  int DisplayForwardAlgo = 0;

  if(DisplaySourcePhrase || DisplaySentence || DisplayProb || DisplayAlignItem)
    cout << "Starting AlignPhrase" << endl;
#endif

  Trace( 1, "CPhraseAlignerMSTag222::AlignPhrase" );


  //###############################
  //#                             #
  //#          Init               #
  //#                             #
  //###############################

  if ( !m_pLexiconSourceTarget && !m_pLexiconTargetSource ) {
	cerr << "no lexicon provided to calculate alignment" << endl;
	exit( -3);
  }

  if ( !m_pTagLexiconSourceTarget && !m_pTagLexiconTargetSource ) {
	cerr << "no Tag lexicon provided to calculate alignment" << endl;
	exit( -3);
  }
  if ( !m_pMappingSource || !m_pMappingTarget ) {
	cerr << "no Mapping Word->Tag provided to calculate alignment" << endl;
	exit( -3);
  }
  if ( m_AlignBidirectional && (!m_pLexiconSourceTarget || !m_pLexiconTargetSource) ) {
	cerr << "calculation of alignment in both directions requested, but not both lexicons provided" << endl;
	exit( -3);
  }

#if DEBUG
  if ( DisplaySourcePhrase ) {
    cout << "Source Phrase: ";
    for ( j = StartPosInSource; j <= EndPosInSource; j++ ) {
      f = pSourceSentence->GetAt( j );
      cout << m_pLexiconSourceTarget->SourceVocabulary()->GetWord( f ) << " ";
    }
    cout << endl;
  }
#endif    

  m_pSourceSentence = pSourceSentence;
  m_pTargetSentence = pTargetSentence;
 

  // length of the sentences
  m_J = pSourceSentence->Length();
  m_I = pTargetSentence->Length();


  // if sentences are longer than we can store probabilities, we skip it
  if ( (m_J >= MAX_SENTENCE_LENGTH) || (m_I >= MAX_SENTENCE_LENGTH) ) {
    return false;
  }


#if DEBUG
  if(DisplayForwardAlgo) {
    m_pCalc->m_DisplayForwardAlgo = 2;
    m_pCalc->SetSentence(pSourceSentence,pTargetSentence, m_pLexiconSourceTarget->SourceVocabulary());    
  }

#endif 

  //build Tag-Sentence
  // 1. Source Sentence
  CWordIndex tag;
  m_pSourceTagSentence = new CSentence;
  for(int j = 1; j <= m_J; j++) {
    e = pSourceSentence->GetAt( j );
    tag = m_pMappingSource->getTag(e);
    (*m_pSourceTagSentence)[j] = tag;
    //cout << " Tag for word " << m_pLexiconSourceTarget->SourceVocabulary()->GetWord( e ) << " is " << m_pTagLexiconSourceTarget->SourceVocabulary()->GetWord( tag ) << endl;
  }

  // 2. Target Sentence
  m_pTargetTagSentence = new CSentence;
  for(int i = 1; i <= m_I; i++) {
    f = pTargetSentence->GetAt( i );
    tag = m_pMappingTarget->getTag(f);
    (*m_pTargetTagSentence)[i] = tag;
    //cout << "Tag for word " << m_pLexiconSourceTarget->TargetVocabulary()->GetWord( f ) << " is " << m_pTagLexiconSourceTarget->TargetVocabulary()->GetWord( tag ) << endl;
  }
  

  //get Source->Target Probabilities
  GetProbabilities(m_pSourceSentence, m_pTargetSentence, m_pLexiconSourceTarget, &(m_ProbsSourceTarget[0][0]));
  // renormalize probabilities for the current sentence pair (needed for FindCentr
  Renormalize(m_J, m_I, &(m_ProbsSourceTarget[0][0]),&(m_RenProbsSourceTarget[0][0]));

  
  //get Target->Source Probabilities
  if(m_AlignBidirectional) {
    GetProbabilities(m_pTargetSentence, m_pSourceSentence, m_pLexiconTargetSource, &(m_ProbsTargetSource[0][0]));

 }

  //get Source->Target Tag Probabilities
  GetProbabilities(m_pSourceTagSentence, m_pTargetTagSentence, m_pTagLexiconSourceTarget,&( m_ProbsTagSourceTarget[0][0]));

  
  //get Target->Source Probabilities
  if(m_AlignBidirectional) {
    GetProbabilities(m_pTargetTagSentence, m_pSourceTagSentence, m_pTagLexiconTargetSource, &(m_ProbsTagTargetSource[0][0]));

  }

#if DEBUG
  if(DisplaySentence) {
    cout << "Source Sentence: ";
    m_pSourceSentence->Display(m_pLexiconSourceTarget->SourceVocabulary());
    cout << endl;
    cout << "Target Sentence: ";
    m_pTargetSentence->Display(m_pLexiconSourceTarget->TargetVocabulary());
    cout << endl;  
    cout << "Source Tag Sentence: ";
    m_pSourceTagSentence->Display(m_pTagLexiconSourceTarget->SourceVocabulary());
    cout << endl;
    cout << "Target Tag Sentence: ";
    m_pTargetTagSentence->Display(m_pTagLexiconSourceTarget->TargetVocabulary());
    cout << endl;
  }


  if(DisplayProb){
    cout << "Word Probabilitis" << endl;
    DisplayProbs(&(m_ProbsSourceTarget[0][0]), &(m_ProbsTargetSource[0][0]));
    cout <<endl << "Tag Probabilities" << endl;
    DisplayProbs(&(m_ProbsTagSourceTarget[0][0]), &(m_ProbsTagTargetSource[0][0]));
  }
 
#endif

  //###############################
  //#                             #
  //#     Calcuate Scores         #
  //#                             #
  //###############################




  // find 'center' of target phrase for source phrase
  float i_mid = FindCenterSum( StartPosInSource, EndPosInSource );
 
  // don't want too short or too long translations
  // on average the target length should be source length times stretch factor
  // where stretch factor is the ratio target corpus lenght to source corpus length
  // then target should not be longer then this average length time a elongation factor
  // and not shorter then the average length time a shortening factor
  // we allow for and addition +/-2 to have e.g. 1->3 alignemnts
  int MinTargetLength = (int)(-1.5 + LengthSourcePhrase * m_StretchFactor * m_FactorShorter);
  int MaxTargetLength = (int)(2.5 + LengthSourcePhrase * m_StretchFactor * m_FactorLonger);
  if ( MinTargetLength < 1 )
    MinTargetLength = 1;

 
  // in case there are old results in the arrays
  PhraseAlignItems.Clear();

  // expand the target phrase around the center
  CSortedArray< double > Probs;
  // sort smallest first
  //Probs.SortReverse();
  CNumericalArray< double > AllLogP1;
  CNumericalArray< double > AllLogP2;

  double LogP, LogP1, LogP2;
  double BestLogP = 10000.0, BestLogP1 = 10000.0, BestLogP2 = 10000.0;

  bool WasNew;
  int mid = int( i_mid + 0.5 );

  int n = 0;  // counter for translations
  
  // here, we loop over the two boundaries
  for ( min_i = mid - MaxTargetLength; min_i <= mid + 1; min_i++ ) {
    for ( max_i = mid - 1; max_i <= mid + MaxTargetLength; max_i++ ) {

      // currently we do not assume that we have <sos> and <eos>
      if ( (min_i >= 1) && (max_i <= m_I) 
           && ((max_i - min_i + 1) >= MinTargetLength)
	   && ((max_i - min_i + 1) <= MaxTargetLength) ) {

	n++;
	CPhraseAlignItemMS AlignItem;

	// get the target phrase: word IDs and string
        for ( int i = min_i; i <= max_i; i++ ) {
          //cout << TextE.GetWord( SentenceID - 1, i ) << " ";
	  e = m_pTargetSentence->GetAt( i );
	  AlignItem.m_TargetWordIDs.Append( e );
	  AlignItem.m_TargetPhrase += CString( " " ) + m_pLexiconSourceTarget->TargetVocabulary()->GetWord( e );
        }

	// init all scores
	for(int i = 0; i <= m_numberOfScores; i++) {
	  AlignItem.m_Scores[i] = 0.0;
	}
	for(int i = 0; i <= m_numberOfCombinedScores; i++) {
	  AlignItem.m_CombinedScores[i] = 0.0;
	}
	for(int i = 0; i <= m_numberOfScores; i++) {
	  AlignItem.m_ScaledScores[i] = 0.0;
	}
	for(int i = 0; i <= m_numberOfCombinedScores; i++) {
	  AlignItem.m_CombinedScaledScores[i] = 0.0;
	}



        if ( m_pLexiconSourceTarget ) {

#if DEBUG
	  if(DisplayForwardAlgo){
	    m_pCalc->m_Direction = true;
	    m_pCalc->m_DisplayForwardAlgo = 0;
	    cout << "SourceTarget SentProb Words: " << endl;
	  }
#endif
	  // now, we calculate the log prob for the phrase from total sentence alignment,
	  // dividing by the sentence length or the phrase length, to have per word costs (for pruning)
	  // later, i.e. after pruning and before we insert into lattice, we need to multiply again by source phrase length
	  LogP = m_pCalc->CalcSentAlignLogProb(StartPosInSource, EndPosInSource, min_i, max_i,m_J,m_I,&(m_ProbsSourceTarget[0][0]));
	  //LogP /= m_J * LengthSourcePhrase;
	  //LogP /= m_J;
	  AlignItem.m_Scores[1] = LogP;
	  //AlignItem.m_ScaledScores[1] = m_Lambda1FE * LogP;
#if DEBUG
	  if(DisplayForwardAlgo) {
	    m_pCalc->m_Direction = true;
	    m_pCalc->m_DisplayForwardAlgo = 0;
	    cout << "SourceTarget PhraseProb Words:" << endl;
	  }
#endif

	  // probability of just aligning the phrase, using renormalized word probabilities
	  LogP = m_pCalc->CalcPhraseLogProb( StartPosInSource, EndPosInSource, min_i, max_i,&(m_ProbsSourceTarget[0][0]));
	  //LogP /= ( EndPosInSource - StartPosInSource + 1 );
	  AlignItem.m_Scores[3] = LogP;
	  //AlignItem.m_ScaledScores[3] = m_Lambda2FE * LogP;

#if DEBUG
	  if(DisplayForwardAlgo) {
	    m_pCalc->m_Direction = true;
	    m_pCalc->m_DisplayForwardAlgo = 0;
	    cout << "SourceTarget SentProb Tags: " << endl;
	  }
#endif
	  // now, we calculate the log prob of tags for the phrase from total sentence alignment,
	  // dividing by the sentence length or the phrase length, to have per word costs (for pruning)
	  // later, i.e. after pruning and before we insert into lattice, we need to multiply again by source phrase length
	  LogP = m_pCalc->CalcSentAlignLogProb(StartPosInSource, EndPosInSource, min_i, max_i,m_J,m_I,&(m_ProbsTagSourceTarget[0][0]) );
	  //LogP /= m_J * LengthSourcePhrase;
	  //LogP /= m_J;
	  AlignItem.m_Scores[5] = LogP;
	  //AlignItem.m_ScaledScores[7] = m_Lambda1FETag * LogP;

#if DEBUG
	  if(DisplayForwardAlgo) {
	    m_pCalc->m_Direction = true;
	    cout << "SourceTarget PhraseProb Tags: " << endl;
	  }

#endif


	  // probability of just aligning the phrase, using renormalized word probabilities
	  LogP = m_pCalc->CalcPhraseLogProb( StartPosInSource, EndPosInSource, min_i, max_i,&(m_ProbsTagSourceTarget[0][0]));
	  //LogP /= ( EndPosInSource - StartPosInSource + 1 );
	  AlignItem.m_Scores[7] = LogP;
	  //AlignItem.m_ScaledScores[9] = m_Lambda2FETag * LogP;
	}

        if ( m_pLexiconTargetSource ) {
#if DEBUG
	  if(DisplayForwardAlgo) {
	    m_pCalc->m_Direction = false;
	    m_pCalc->m_DisplayForwardAlgo = 0;
	    cout << "TargetSource SentProb Words: " << endl;
	  }
#endif

	  // probability of aligning phrase, using renormalized word probabilities
	  // divide by target phrase length and multiply by target phrase length
          LogP = m_pCalc->CalcSentAlignLogProb( min_i, max_i, StartPosInSource, EndPosInSource,m_I, m_J, &(m_ProbsTargetSource[0][0]));
          //LogP /= m_I;
	  //LogP *= LengthSourcePhrase;
	  AlignItem.m_Scores[2] = LogP;
	  //AlignItem.m_ScaledScores[2] = m_Lambda1EF * LogP;
#if DEBUG
	  if(DisplayForwardAlgo) {
	    m_pCalc->m_Direction = false;
	    m_pCalc->m_DisplayForwardAlgo = 1;
	    cout << "TargetSource PhraseProb Words: " << endl;
	  }
#endif

	  // now, we calculate the log prob for the phrase from total sentence alignment,
	  // dividing by the sentence length, multiplied by source phrase length
	  LogP = m_pCalc->CalcPhraseLogProb(min_i, max_i, StartPosInSource, EndPosInSource,&(m_ProbsTargetSource[0][0]));
	  //cout << "LogP " << LogP << " TargetLength " << ( max_i - min_i + 1 ) 
	  //   << " SourceLength " << LengthSourcePhrase << endl;
	  //LogP /= ( max_i - min_i + 1 );
	  //LogP *= LengthSourcePhrase;
	  AlignItem.m_Scores[4] = LogP;
	  //AlignItem.m_ScaledScores[4] = m_Lambda2EF * LogP;
#if DEBUG
	  if(DisplayForwardAlgo) {
	    m_pCalc->m_Direction = false;
	    m_pCalc->m_DisplayForwardAlgo = 0;
	    cout << "TargetSource SentProb Tags: " << endl;
	  }
#endif
		  
	  // probability of aligning phrase, using renormalized word probabilities
	  // divide by target phrase length and multiply by target phrase length
          LogP = m_pCalc->CalcSentAlignLogProb( min_i, max_i, StartPosInSource, EndPosInSource,m_I, m_J, &(m_ProbsTagTargetSource[0][0]));
          //LogP /= m_I;
	  //LogP *= LengthSourcePhrase;
	  AlignItem.m_Scores[6] = LogP;
	  //AlignItem.m_ScaledScores[8] = m_Lambda1EF * LogP;
#if DEBUG
	  if(DisplayForwardAlgo) {
	    m_pCalc->m_Direction = false;
	    m_pCalc->m_DisplayForwardAlgo = 0;
	    cout << "TargetSource PhraseProb Tags: " << endl;
	  }
#endif

	  // now, we calculate the log prob for the phrase from total sentence alignment,
	  // dividing by the sentence length, multiplied by source phrase length
	  LogP = m_pCalc->CalcPhraseLogProb(min_i, max_i, StartPosInSource, EndPosInSource,&(m_ProbsTagTargetSource[0][0]));
	  //cout << "LogP " << LogP << " TargetLength " << ( max_i - min_i + 1 ) 
	  //   << " SourceLength " << LengthSourcePhrase << endl;
	  //LogP /= ( max_i - min_i + 1 );
	  //LogP *= LengthSourcePhrase;
	  AlignItem.m_Scores[8] = LogP;
	  //AlignItem.m_ScaledScores[10] = m_Lambda2EF * LogP;
        }

	//calculating scaled scores
	for(int i = 1; i <= m_numberOfScores; i++) {
	  AlignItem.m_ScaledScores[i] =m_ScalingFactors[i] * AlignItem.m_Scores[i];
	}


	// now, we combine these scores according to the scaling factors
	for(int i = 1; i <= m_numberOfScores; i++) {
	  AlignItem.m_Scores[0] +=AlignItem.m_Scores[i];
	}	

	
	//Combining the the same type of scores from both direction
	AlignItem.m_CombinedScores[1] =AlignItem.m_Scores[1] + AlignItem.m_Scores[2];
	AlignItem.m_CombinedScores[2] =AlignItem.m_Scores[3] + AlignItem.m_Scores[4];

	// and the scaled ones
	for(int i = 1; i <= m_numberOfScores; i++) {
	  AlignItem.m_ScaledScores[0] +=AlignItem.m_ScaledScores[i];
	}

	//Combining the the same type of scores from both directions
	AlignItem.m_CombinedScaledScores[1] =AlignItem.m_ScaledScores[1] + AlignItem.m_ScaledScores[2];
	AlignItem.m_CombinedScaledScores[2] =AlignItem.m_ScaledScores[3] + AlignItem.m_ScaledScores[4];

#if DEBUG
	if ( DisplayAlignItem ) {
	  cout << "Testing target phrase: " << AlignItem.m_TargetPhrase <<endl;
	  cout << AlignItem << endl;	  
	  cout << endl;
	}
#endif

	// store now in list of all good translations for this source phrase
	int Location;
	bool IsNew;
	if ( PhraseAlignItems.Find( AlignItem, Location ) ) {
	  //cout << " translation was found" << endl;
	  if ( AlignItem.Score() < PhraseAlignItems.GetAt( Location ).Score() ) {
	    // already stored, but current score is lower, i.e. better
	    // therefore overwrite it
	    //cout << " translation was found, but new score is smaller, i.e. overwrite " << endl;
	    PhraseAlignItems.Insert( AlignItem, IsNew );
	    //}
	    //} else {
	    //cout << " but new score is larger, i.e. don't use " << endl;
	  }
	} else {
	  //cout << " translation was not found - insert" << endl;
	  PhraseAlignItems.Insert( AlignItem, IsNew );
	}


        //cout << "P( " << min_i << ", " << max_i << " ) = " << LogP;
        //cout << "  inserted at " << k << endl;
      } else {
        ; //cout << "skip" << endl;
      }
    }
  }

   delete m_pSourceTagSentence;
  delete m_pTargetTagSentence;
  return true;




}

