#include <string.h>
#include <stdlib.h>
#include <PhraseAlignProbCalc.hh>

using namespace std;

CPhraseAlignProbCalc::CPhraseAlignProbCalc()
{
  m_UseEmptyWord = false;
}



CPhraseAlignProbCalc::~CPhraseAlignProbCalc()
{
}



void 
CPhraseAlignProbCalc::SetUseEmptyWord( bool UseEmptyWord )
{
  m_UseEmptyWord = UseEmptyWord;
}



float
CPhraseAlignProbCalc::CalcSentAlignLogProb( int j1, int j2, int i1, int i2, 
                                            int SourceLength, int TargetLength,
                                            float *Probs ) 
{
  return 0.0;
}



float
CPhraseAlignProbCalc::CalcPhraseLogProb( int j1, int j2, int i1, int i2, float *Probs )
{
  return 0.0;
}



#if DEBUG
void CPhraseAlignProbCalc::SetSentence(CSentence *pSourceSentence,CSentence *pTargetSentence,const CVocabulary *pVoc)
{
}
#endif
//##########################################################################



CPhraseAlignProbCalcIBM1::CPhraseAlignProbCalcIBM1()
{
  m_UseEmptyWord = false;
}



CPhraseAlignProbCalcIBM1::~CPhraseAlignProbCalcIBM1()
{
}



float
CPhraseAlignProbCalcIBM1::CalcSentAlignLogProb( int j1, int j2, int i1, int i2,
                                                int SourceLength, int TargetLength,
                                                float *Probs )
{
  int i, j;
  double P;
  double LogP;
  double TotalLogP = 0.0;


  for ( j = 1; j <= SourceLength; j++ ) {

    P = 0.0;
    if ( m_UseEmptyWord )
      P += Probs[j* MAX_SENTENCE_LENGTH + 0];
    // Adding Probabilities of Source word
    //Case1: Source Word is part of Phrase -> Add all Probabilities of words in TargetPhrase
    if ( (j >= j1) && (j <= j2) ) {
      for ( i = i1; i <= i2; i++ ) {
        P += Probs[j* MAX_SENTENCE_LENGTH + i];
      }
      //       if(m_UseEmptyWord)
      // 	P /= i2 -i1 + 2;
      //       else
      // 	P /= i2 -i1 + 1;

    } else {
      //Case2: Source Word is not part of Phrase -> Add all Probabilities of words not in TargetPhrase
      for ( i = 1; i <= TargetLength; i++ ) {
        if ( (i < i1) || (i > i2) ) {
          P += Probs[j* MAX_SENTENCE_LENGTH + i];
        }
      }
      //      if(m_UseEmptyWord)
      // 	P /= TargetLength - (i2 -i1 + 1) + 1;
      //       else
      // 	P /= TargetLength - (i2 - i1 + 1);
    }


    
    // there can be situations where the source phrase is aligned to the
    // entire target sentence, i.e. there is no outside source aligned to
    // outside target, and therefore the probability is caclulated as 0.0
    // set P = 1.0
    //cout << "P = " << P << endl;
    if ( P == 0.0 ) {
      //	  cout << "P < 1.0e-10" << endl;
      P = 1.0;
    }

    // sometimes the sum of the probs is > 1.0; restrict to 1.0
    if ( P > 1.0 ) {
      P = 1.0;
    }

    LogP = -log( P );

    // Sum of LogProbs of all Source Words
    TotalLogP += LogP;
  }
  
  return TotalLogP/SourceLength;
}



float
CPhraseAlignProbCalcIBM1::CalcPhraseLogProb( int j1, int j2, int i1, int i2, float *Probs )
{
  int i, j;
  double P;
  double LogP;
  double TotalLogP = 0.0;

  //For all Source Words in Phrase
  for ( j = j1; j <= j2; j++ ) {

    P = 0.0;
    if ( m_UseEmptyWord )
      P += Probs[j* MAX_SENTENCE_LENGTH + 0];
    
    //Add all Source->Target Probabilities for TargetWords in Phrase
    for ( i = i1; i <= i2; i++ ) {
      P += Probs[j* MAX_SENTENCE_LENGTH + i];
    }
    //     if(m_UseEmptyWord)
    //       P /= i2 -i1 + 2;
    //     else
    //       P /= i2 - i1 + 1;
    // sometimes the sum of the probs is > 1.0; restrict to 1.0
    if ( P > 1.0 ) {
      P = 1.0;
    }

    LogP = -log( P );

    //Add LogProbs for all SourceWords
    TotalLogP += LogP;
  }

  return TotalLogP/(j2-j1+1);
}



//###########################################################################
CPhraseAlignProbCalcHMM::CPhraseAlignProbCalcHMM()
{
  m_UseEmptyWord = false;
  m_JumpProbFilled = -1;

#if DEBUG
  m_DisplayForwardAlgo = 0;
#endif
}



CPhraseAlignProbCalcHMM::~CPhraseAlignProbCalcHMM()
{
}



float
CPhraseAlignProbCalcHMM::CalcSentAlignLogProb( int j1, int j2, int i1, int i2,
                                               int SourceLength, int TargetLength,
                                               float *Probs ) 
{
  int i, j;
  int maxP;
  double P;
  double LogP;
  double SumJP;
  double PP [SourceLength + 1][TargetLength + 1];
  int ToEmptyWordFrom[SourceLength + 1];
  ToEmptyWordFrom[0] = 0;

  if(!m_UseEmptyWord){
    cout << "For HMM empty word is needed! " << endl;
    exit(-1);
  }

  //calculate Jump Probabilities, if necessary
  if(m_JumpProbFilled < TargetLength)
    CalcJumpProb(TargetLength);

  for(j = 0; j <= SourceLength; j++)
    for(i = 0; i <= TargetLength; i++) 
      PP[j][i] = 0.0;


  //going over all columes
  for ( j = 1; j <= SourceLength; j++ ) {
    
#if DEBUG
    if(m_DisplayForwardAlgo == 2)
      cout << "Calculating Column " << j << endl;
#endif

    P = 0.0;
    SumJP = 0.0;
    //1. Calculate Probability for SourceWord translatet to emtpy Word
    
    //case1: first word in sentence
    if(j == 1) {
      P = 1.0;
      SumJP = 1.0;
      maxP = 0;
      //case 2: last Source Position was in Phrase -> add only over Probs from Phrase
    }else if((j-1 >= j1) && (j-1 <= j2)) {
      //Jumping cost to Empty word all the same
      SumJP += 1;
      maxP = 0;
      P += 1 * PP[(j - 1)] [0];
      for(int is = i1; is <= i2; is++){
        SumJP += 1;
        P += 1*PP[(j - 1)][is];
        if(PP[(j - 1)][is] > PP[j-1][maxP]) {
          //is is entry with highest prob to go to empty word
          maxP = is;
        }
      }
      SumJP = TargetLength + 1;
    
      //last Source Position was outside Phrase -> add only over Probs outside Phrase
    }else { 
      maxP = 0;
      for(int is = 0; is <= TargetLength; is++) {
        if( (is < i1) || (is > i2) ) {
          SumJP += 1;
          P += 1*PP[ (j - 1) ][ is ];
          if(PP[(j - 1)][is] > PP[j-1][maxP]) {
            //is is entry with highest prob to go to empty word
            maxP = is;
          }
        }
      }
      SumJP = TargetLength + 1;
      
    }
    
    if(maxP != 0)
      ToEmptyWordFrom[j] = maxP;
    else {
      int k = j - 1;
      for(; k > 1 && ToEmptyWordFrom[k] == 0; k-- ){
      }
      ToEmptyWordFrom[j] = ToEmptyWordFrom[k];
    }

    PP[j][0] = P / SumJP * Probs[j* MAX_SENTENCE_LENGTH + 0];

#if DEBUG
    if(m_DisplayForwardAlgo == 2) {
      cout << "Line 0" << endl;
      cout << "ToEmptyWordFrom[" << j << "] = " << ToEmptyWordFrom[j] << endl;
      cout << " P : " << P << " SumJP " << SumJP << endl;
      cout << "Probs: " << Probs[j* MAX_SENTENCE_LENGTH + 0] << endl;
      cout << " => PP[ " << j << " ][0] = " << PP[j][0] << endl;
    }

#endif
 
    // Filling whole colum line by line
    //case1: Source Word is part of Phrase -> Calculation only PP for words in Phrase
    if ( (j >= j1) && (j <= j2) ) {
      for ( i = i1; i <= i2; i++ ) {
        P = 0.0;
        SumJP = 0.0;
        //case a: first word in sentence
        if(j == 1) {
          //P = 1.0;
          //SumJP = 1.0;
          P = JumpCost(0,i);
          SumJP = JumpCost(0,1);
	
          //case b: last Source Position was in Phrase -> add only over Probs from Phrase
        }else if((j-1 >= j1) && (j-1 <= j2)) {
          //cout << "JumpCost (" <<ToEmptyWordFrom[j - 1] <<" , " << i  << ") = " << JumpCost(ToEmptyWordFrom[j - 1],i) << endl;
          SumJP += JumpCost(ToEmptyWordFrom[j - 1],i);
          P += JumpCost(ToEmptyWordFrom[j - 1],i)*PP[(j - 1)][0];
          for(int is = 1; is <= TargetLength; is++){
            if((is >= i1) && (is <= i2)) {
              SumJP += JumpCost(is,i);
              P += JumpCost(is,i)*PP[(j - 1)][is];
            }else {
              SumJP += JumpCost(is,i);
            }
          }

          //case c:last Source Positon was outside Phrase -> add only over Probs outside Phrase
        }else { 
          //cout << "JumpCost (" <<ToEmptyWordFrom[j - 1] <<" , " << i  << ") = " << JumpCost(ToEmptyWordFrom[j - 1],i) << endl;
          SumJP += JumpCost(ToEmptyWordFrom[j - 1],i);
          P += JumpCost(ToEmptyWordFrom[j - 1],i)*PP[(j - 1)][0];
          for(int is = 1; is <= TargetLength; is++) {
            if( (is < i1) || (is > i2) ) {
              SumJP +=JumpCost(is, i);
              P += JumpCost(is,i)*PP[(j - 1)][is];
            }else {
              SumJP += JumpCost(is,i);
            }
          }
        }
      

        PP[j][i] = P / SumJP * Probs[j* MAX_SENTENCE_LENGTH + i];
        //        PP[j][i] = P * Probs[j* MAX_SENTENCE_LENGTH + i];

#if DEBUG
        if(m_DisplayForwardAlgo == 2) {
          cout << "Line " << i << " (word if part of phrase)" << endl;
          cout << " P : " << P << " SumJP " << SumJP << endl;
          // 	  cout << " P : " << P << endl;
          cout << "Probs: " << Probs[j* MAX_SENTENCE_LENGTH + i] << endl;
          cout << " => PP[ " << j << " ][ " << i  << " ] = " << PP[j][i] << endl;
        }

#endif

      }

    } else {
      //Case2: Source Word is not part of Phrase -> Add all Probabilities of words not in TargetPhrase
      for ( i = 1; i <= TargetLength; i++ ) {
        if ( (i < i1) || (i > i2) ) {
          P = 0.0;
          SumJP = 0.0;
          //case a: first word in sentence
          if(j == 1) {
            P = JumpCost(0,i);
            SumJP = JumpCost(0,1);
            
            //case b: last Source Position was in Phrase -> add only over Probs from Phrase
          }else if((j-1 >= j1) && (j-1 <= j2)) {
            //cout << "JumpCost (" <<ToEmptyWordFrom[j - 1] <<" , " << i  << ") = " << JumpCost(ToEmptyWordFrom[j - 1],i) << endl;
            SumJP += JumpCost(ToEmptyWordFrom[j - 1],i);
            P += JumpCost(ToEmptyWordFrom[j - 1],i)*PP[(j - 1)][0];
            for(int is = 1; is <= TargetLength; is++){
              if((is >= i1) && (is <= i2)) {
                //cout << "JumpCost (" << is <<" , " << i  << ") = " << JumpCost(is,i) << endl;
                SumJP += JumpCost(is,i);
                P += JumpCost(is,i)*PP[(j - 1)][is];
              }else {
                SumJP += JumpCost(is,i);
              }
            }

            //case c: last Source Position was outside Phrase -> add only over Probs outside Phrase
          }else { 
            SumJP += JumpCost(ToEmptyWordFrom[j - 1],i);
            P += JumpCost(ToEmptyWordFrom[j - 1],i)*PP[(j - 1)][0];
            for(int is = 1; is <= TargetLength; is++) {
              if( (is < i1) || (is > i2) ) {
                //cout << "JumpCost (" <<is <<" , " << i  << ") = " << JumpCost(is,i) << endl;
                SumJP += JumpCost(is, i);
                P += JumpCost(is,i)*PP[(j - 1)][is];
              }else {
                SumJP += JumpCost(is,i);
              }
            }
          }
   

          PP[j][i] = P / SumJP * Probs[j* MAX_SENTENCE_LENGTH + i];
          // 	  PP[j][i] = P * Probs[j* MAX_SENTENCE_LENGTH + i];


#if DEBUG
          if(m_DisplayForwardAlgo == 2) {
            cout << "Line " << i << " (word if not part of phrase)" << endl;
            cout << " P : " << P << " SumJP " << SumJP << endl;
            // 	  cout << " P : " << P << endl;
            cout << "Probs: " << Probs[j* MAX_SENTENCE_LENGTH + i] << endl;
            cout << " => PP[ " << j << " ][ " << i  << " ] = " << PP[j][i] << endl;
          }

#endif

        }
      }
    }
    
  }

 
  P = 0.0;
  SumJP = 0.0;
  SumJP += JumpCost(ToEmptyWordFrom[SourceLength],TargetLength + 1);
  P += JumpCost(ToEmptyWordFrom[SourceLength],TargetLength + 1) * PP[SourceLength][0];
 
  //case a: last Source Word was in Phrase
  if(j2 == SourceLength) {
    for(i = 1;i <= TargetLength; i++) {
      if((i >=i1) &&(i <= i2)) {
        SumJP += JumpCost(i,TargetLength + 1);
        P += JumpCost(i, TargetLength + 1)*PP[SourceLength][i];
      }else {
        SumJP +=JumpCost(i,TargetLength + 1);
      }
    }
    //case b: last Source Word was not in Phrase  
  }else {
    for(i = 1; i <= TargetLength; i++) {
      if((i < i1) || (i > i2)) {
        SumJP += JumpCost(i,TargetLength + 1);
        P += JumpCost(i, TargetLength + 1)*PP[SourceLength][i];
      }else {
        SumJP += JumpCost(i,TargetLength + 1);
      }
    }

  }

#if DEBUG
  if(m_DisplayForwardAlgo == 2) {
    cout << "Calcuation of last Field: " << endl;
    cout << " P : " << P << " SumJP " << SumJP << endl;    
    cout << " P : " << P << endl;    
  }
#endif 

  P /=SumJP; 
  LogP = -log( P );



#if DEBUG
  if(m_DisplayForwardAlgo) {
    cout << "Table : " << endl;
    DisplaySentProb(&(PP[0][0]),j1,j2,i1,i2);
    cout << "-> P = " << P << endl;
    cout << "=> return -logP/" << SourceLength << " = " << LogP/SourceLength << endl;
  }
#endif
  
  return LogP/SourceLength;

}



float
CPhraseAlignProbCalcHMM::CalcPhraseLogProb( int j1, int j2, int i1, int i2, float *Probs )
{
  int i, j;
  double P;
  //   double SumJP;
  int maxP;
  double LogP;
  int SrcPhraseLength = j2 - j1 + 1;
  int TrgPhraseLength = i2 - i1 + 1;
  double PP [SrcPhraseLength + 1][TrgPhraseLength + 1];
  int ToEmptyWordFrom[SrcPhraseLength + 1];
  ToEmptyWordFrom[0] = 0;


  if(!m_UseEmptyWord){
    cout << "For HMM empty word is needed! " << endl;
    exit(-1);
  }

  if(m_JumpProbFilled < TrgPhraseLength)
    CalcJumpProb(TrgPhraseLength);


  for(j = 0; j <= SrcPhraseLength; j++)
    for(i = 0; i <= TrgPhraseLength; i++) 
      PP[j][i] = 0.0;

    
  //going over all columes
  for ( j = 1; j <= SrcPhraseLength; j++ ) {

#if DEBUG
    if(m_DisplayForwardAlgo == 2)
      cout << "Calculating Column " << j << endl;
#endif


    //     P = 0.0;
    //     SumJP = 0.0;
    //     //1. Calculate Probability for SourceWord translatet to emtpy Word
    
    //     //case1: first word in sentence
    //     if(j == 1) {
    //       P = 1.0;
    //       SumJP = 1.0;
    //       maxP = 0;

    //     }else {
    //       SumJP += 1;
    //       maxP = 0;
    //       P += 1 * PP[j - 1][0];
    //       for(int is = 1; is <= TrgPhraseLength; is++) {
    // 	SumJP += 1;
    // 	P += 1*PP[ (j - 1) ][ is ];
    // 	if(PP[(j - 1)][is] > PP[j-1][maxP]) {
    // 	  //is is entry with highest prob to go to empty word
    // 	  maxP = is;
    // 	}
	
    //       }
    //     }

    //     if(maxP != 0)
    //       ToEmptyWordFrom[j] = maxP;
    //     else {
    //       int k = j - 1;
    //       for(; k > 1 && ToEmptyWordFrom[k] == 0; k-- ){
    //       }
    //       ToEmptyWordFrom[j] = ToEmptyWordFrom[k];
    //     }

    //     PP[j][0] = P / SumJP * Probs[(j + j1 -1)* MAX_SENTENCE_LENGTH + 0];
    //     PP[j][0] = PP[j][0] / 1000;

#if DEBUG
    //     if(m_DisplayForwardAlgo == 2) {
    //       cout << "Line 0" << endl;
    //       cout << "ToEmptyWordFrom[" << j << "] = " << ToEmptyWordFrom[j] << endl;
    //       cout << " P : " << P << " SumJP " << SumJP << endl;
    //       cout << "Probs: " << Probs[(j + j1 - 1)* MAX_SENTENCE_LENGTH + 0] << endl;
    //       cout << " => PP[ " << j << " ][0] = " << PP[j][0] << endl;
    //     }

#endif

    
    //Filling whole colum line by line
    for ( i = 1; i <=  TrgPhraseLength; i++ ) {
      P=0.0;
      //       SumJP = 0.0;
      //case a: first word in sentence
      if(j == 1) {
        P = JumpCost(0,i);
        // 	SumJP = JumpCost(0,1);
	
      }else {
        // 	SumJP += JumpCost(ToEmptyWordFrom[j - 1],i);
        // 	P+= JumpCost(ToEmptyWordFrom[j-1],i) * PP[(j-1)][0];
        for(int is = 1; is <= TrgPhraseLength;is++) {
          // 	  SumJP += JumpCost(is,i);
          P+=JumpCost(is,i)*PP[(j-1)][is];
        }


      }
      
      //       PP[j][i] = P/SumJP * Probs[(j+ j1 -1)* MAX_SENTENCE_LENGTH + (i+ i1 - 1)];
      PP[j][i] = P * Probs[(j+ j1 -1)* MAX_SENTENCE_LENGTH + (i+ i1 - 1)];

#if DEBUG
      if(m_DisplayForwardAlgo == 2) {
        cout << "Line " << i << endl;
        // 	  cout << " P : " << P << " SumJP " << SumJP << endl;
        cout << " P : " << P << endl;
        cout << "Probs: " << Probs[(j + j1 - 1)* MAX_SENTENCE_LENGTH + (i + i1 - 1)] << endl;
        cout << " => PP[ " << j << " ][ " << i  << " ] = " << PP[j][i] << endl;
      }

#endif

    
    }
  }

  P = 0.0;
  //   SumJP = 0.0;
  //   SumJP += JumpCost(ToEmptyWordFrom[SrcPhraseLength],TrgPhraseLength + 1);
  //   P += JumpCost(ToEmptyWordFrom[SrcPhraseLength],TrgPhraseLength + 1) * PP[SrcPhraseLength][0];

  for(i = 1;i <= TrgPhraseLength; i++) {
    //     SumJP += JumpCost(i,TrgPhraseLength + 1);
    P += JumpCost(i, TrgPhraseLength + 1) * PP[SrcPhraseLength][i];
  }

#if DEBUG
  if(m_DisplayForwardAlgo == 2) {
    cout << "Calcuation of last Field: " << endl;
    //     cout << " P : " << P << " SumJP " << SumJP << endl;    
    cout << " P : " << P << endl;    
  }
#endif 

  //   P /=SumJP;

  LogP = -log( P );

#if DEBUG
  if(m_DisplayForwardAlgo) {
    cout << "Table : " << endl;
    DisplayPhraseProb(&(PP[0][0]),SrcPhraseLength,TrgPhraseLength,j1, i1);
    cout << "-> P = " << P << endl;
    cout << "=> return -logP/" << SrcPhraseLength << " = " << LogP/SrcPhraseLength << endl;
  }
#endif

  return LogP/SrcPhraseLength;

}



void
CPhraseAlignProbCalcHMM::CalcJumpProb( int FillTo )
{
  double var = 1;
  if(m_JumpProbFilled < 0){
    m_JumpProb[0] = 1.0/sqrt(2*PI)*exp( 1 /(-2.0*var));
    m_JumpProbNeg[0] = m_JumpProb[0];
    cout << " Filled m_JumpProb[ 0 ] with " << m_JumpProb[0] << endl;
    m_JumpProbFilled = 0;
  }
    

  for(int i = m_JumpProbFilled + 1; i <= FillTo; i++) {
    m_JumpProb[i] = 1.0/sqrt(2*PI)*exp(pow(i - 1,2.0)/(-2.0 * var));
    m_JumpProbNeg[i] = 1.0/sqrt(2*PI)*exp(pow(i + 1,2.0)/(-2.0 * var));
    //m_JumpProb[i] = 1.0/(pow(i - 1,4.0) + 1);
    //m_JumpProbNeg[i] = 1.0/(pow(i + 1,4.0) + 1);
    cout << " Filled m_JumpProb[" << i <<"] with " << m_JumpProb[i] << endl;
    cout << " Filled m_JumpProbNeg[" << i <<"] with " << m_JumpProbNeg[i] << endl;
  }
  m_JumpProbFilled = FillTo;

}



#if DEBUG
void 
CPhraseAlignProbCalcHMM::DisplaySentProb( double *Probs,int j1,int j2,int i1,int i2 )
{
  int maxLen = GetMaxWordLength();

  cout << "Probabilities:" << endl;
  for(int i = m_I; i >= 1; i--) {
    cout.width(maxLen);
    cout.setf(ios::left);
    cout << m_pVoc->GetWord(m_pTargetSentence->GetAt( i )) << " |" <<flush;
    
    int j;
    if(m_UseEmptyWord)
      j = 0;
    else
      j = 1;
      
    for(; j <= m_J; j++) {
      cout.setf(ios::right | ios::fixed);
      cout.precision(3);
      cout.width(8);
      if(m_Direction)
        cout << Probs[j * (m_I + 1) + i] * 1000;
      else
        cout << Probs[i * (m_J + 1) + j] * 1000;
      if(m_Direction) {
        if((j == j1 -1) ||(j == j2)) 
          cout << "|";
      }else{
        if((j == i1 -1) ||(j == i2))
          cout << "|";
      }
    }
    
    cout << endl;
    
    if(m_Direction){
      if((i == i2 + 1) ||(i == i1)) {
        for(int k = 1; k <= maxLen; k++)
          cout << "-";
        cout << "-|";
        for(int k = 0; k <= m_J; k++)
          cout << "--------";
        cout << "--"; // For | when j1 or j2
        cout << endl;
      }
    }else{
      if((i == j2 + 1) ||(i == j1)) {
        for(int k = 1; k <= maxLen; k++)
          cout << "-";
        cout << "-|";
        for(int k = 0; k <= m_J; k++)
          cout << "--------";
        cout << "--"; // For | when j1 or j2
        cout << endl;
      }

    }
	

  }
  

  if(m_UseEmptyWord) {
    cout.width(maxLen);
    cout.setf(ios::left);
    cout << "<e>" << " |" <<flush;
    
    int j;
    if(m_UseEmptyWord)
      j = 0;
    else
      j = 1;
      
    for(; j <= m_J; j++) {
      cout.setf(ios::right | ios::fixed);
      cout.precision(3);
      cout.width(8);
      if(m_Direction)
        cout << Probs[j * (m_I + 1) + 0] * 1000;
      else
        cout << Probs[0 * (m_J + 1) + j] * 1000;
      
      if(m_Direction){
        if((j == j1 -1) ||(j == j2))
          cout << "|";
      }else{
        if((j == i1 -1) ||(j == i2))
          cout << "|";
      }

    }
    cout << endl; 


  }

  DisplaySourceSentence(maxLen);



}



void 
CPhraseAlignProbCalcHMM::DisplayPhraseProb( double *Probs,int SrcPhraseLength, int TrgPhraseLength, int j1,int i1 )
{
  int maxLen = GetMaxWordLength();

  cout << "Probabilities:" << endl;
  for(int i = TrgPhraseLength; i >= 1; i--) {
    cout.width(maxLen);
    cout.setf(ios::left);
    if(m_Direction)
      cout << m_pVoc->GetWord(m_pTargetSentence->GetAt( i + i1 - 1)) << " |" <<flush;
    else
      cout << m_pVoc->GetWord(m_pSourceSentence->GetAt( i + i1 - 1)) << " |" <<flush;

    int j;
    if(m_UseEmptyWord)
      j = 0;
    else
      j = 1;
      
    for(; j <= SrcPhraseLength; j++) {
      cout.setf(ios::right | ios::fixed);
      cout.precision(3);
      cout.width(8);
      cout << Probs[j * (TrgPhraseLength + 1) + i] * 1000;
    }
    cout << endl;

  }
  

  if(m_UseEmptyWord) {
    cout.width(maxLen);
    cout.setf(ios::left);
    cout << "<e>" << " |" <<flush;
    
    int j;
    if(m_UseEmptyWord)
      j = 0;
    else
      j = 1;
      
    for(; j <= SrcPhraseLength; j++) {
      cout.setf(ios::right | ios::fixed);
      cout.precision(3);
      cout.width(8);
      cout << Probs[j * (TrgPhraseLength + 1) + 0] * 1000;
    }
    cout << endl; 
  }

  DisplaySourcePhrase(maxLen,SrcPhraseLength,j1);

}



int
CPhraseAlignProbCalcHMM::GetMaxWordLength()
{
  int i;
  int maxLen = 3;
  const char* pString;
  // get length of longest e word
  for (i = m_I; i >= 1; i--) {
    pString = m_pVoc->GetWord(m_pTargetSentence->GetAt( i ));
    Trace( 5, pString );
    if (strlen(pString) > (uint) maxLen) 
      maxLen = strlen(pString);
  }
  return maxLen;
}



void
CPhraseAlignProbCalcHMM::DisplaySourceSentence( int maxLen )
{
  int MaxWordLength = 40;
  const char *pString = NULL;
  int MaxSLen = 150;
  char WordMatrix[40][150];
  int NoSpace;
  int i, j;
  CWordIndex Word;


  // display line
  for (i = 0; i < maxLen; i++)
    cout << "-";
  cout << "-|";
  if (m_J > 0) {
    for (j = 0; j <=  m_J; j++) {
      cout << "--------";
    }
  }
  cout << endl;

  // display french words at bottom
  //Trace( 1, "display french words at bottom" );
  char Blanks[150];

  maxLen += 4;

  for(i = 0; i < maxLen; i++)
    Blanks[i] = ' ';
  Blanks[maxLen]='\0';

  for (j = 0; j < MaxSLen; j++)
    for (i=0; i < MaxWordLength; i++)
      WordMatrix[i][j] = ' ';

  if (m_J > 0) {
    pString = "<e>";
    i = 0;
    while (pString[i]){
      WordMatrix[i][j] = (char) pString[i];
      i++;
    }

    for (j = 1; j <= m_J; j++) {
      pString = m_pVoc->GetWord(m_pSourceSentence->GetAt( j ));
      cout.setf(ios::left);
      i=0;
      while (pString[i]) {
        WordMatrix[i][j] = (char) pString[i];
        i++;
      }
    }
  }

  NoSpace = 1;
  for (i=0; ((i < MaxWordLength) && (NoSpace == 1)); i++) {
    NoSpace = 0;
    cout << Blanks << " ";
    // for diagonal output an additional blank in the next iteration
    Blanks[maxLen] = ' ';
    maxLen++;
    Blanks[maxLen] = '\0';
    if (m_J > 0) {
      for (j = 0; j <= m_J; j++) {
        cout << WordMatrix[i][j] << "       ";
        if (WordMatrix[i][j] != ' ')
          NoSpace = 1;
      }
    }
    cout << "\n";
  }
    
  cout << "\n";
  cout.flush();

}


void
CPhraseAlignProbCalcHMM::DisplaySourcePhrase( int maxLen, int SrcPhraseLength, int j1 )
{
  int MaxWordLength = 40;
  const char *pString = NULL;
  int MaxSLen = 150;
  char WordMatrix[40][150];
  int NoSpace;
  int i, j;
  CWordIndex Word;


  // display line
  for (i = 0; i < maxLen; i++)
    cout << "-";
  cout << "-|";
  if (m_J > 0) {
    for (j = 0; j <=  m_J; j++) {
      cout << "--------";
    }
  }
  cout << endl;

  // display french words at bottom
  //Trace( 1, "display french words at bottom" );
  char Blanks[150];

  maxLen += 4;

  for(i = 0; i < maxLen; i++)
    Blanks[i] = ' ';
  Blanks[maxLen]='\0';

  for (j = 0; j < MaxSLen; j++)
    for (i=0; i < MaxWordLength; i++)
      WordMatrix[i][j] = ' ';

  if (m_J > 0) {
    pString = "<e>";
    i = 0;
    while (pString[i]){
      WordMatrix[i][j] = (char) pString[i];
      i++;
    }

    for (j = 1; j <= SrcPhraseLength; j++) {
      if(m_Direction)
        pString = m_pVoc->GetWord(m_pSourceSentence->GetAt( j + (j1 -1)));
      else
        pString = m_pVoc->GetWord(m_pTargetSentence->GetAt( j + (j1 -1)));
      cout.setf(ios::left);
      i=0;
      while (pString[i]) {
        WordMatrix[i][j] = (char) pString[i];
        i++;
      }
    }
  }

  NoSpace = 1;
  for (i=0; ((i < MaxWordLength) && (NoSpace == 1)); i++) {
    NoSpace = 0;
    cout << Blanks << " ";
    // for diagonal output an additional blank in the next iteration
    Blanks[maxLen] = ' ';
    maxLen++;
    Blanks[maxLen] = '\0';
    if (m_J > 0) {
      for (j = 0; j <= SrcPhraseLength; j++) {
        cout << WordMatrix[i][j] << "       ";
        if (WordMatrix[i][j] != ' ')
          NoSpace = 1;
      }
    }
    cout << "\n";
  }
    
  cout << "\n";
  cout.flush();

}



void
CPhraseAlignProbCalcHMM::SetSentence( CSentence *pSourceSentence, CSentence *pTargetSentence, const CVocabulary *pVoc )
{
  m_pSourceSentence = pSourceSentence;
  m_J = pSourceSentence->Length();
  m_pTargetSentence = pTargetSentence;
  m_I = pTargetSentence->Length();
  m_pVoc = pVoc;

}

#endif



//###########################################################################
CPhraseAlignProbCalcViterbi::CPhraseAlignProbCalcViterbi()
{
  m_UseEmptyWord = false;
  m_JumpProbFilled = -1;

#if DEBUG
  m_DisplayForwardAlgo = 0;
#endif
}



CPhraseAlignProbCalcViterbi::~CPhraseAlignProbCalcViterbi()
{
}



float
CPhraseAlignProbCalcViterbi::CalcSentAlignLogProb( int j1, int j2, int i1, int i2,
                                                   int SourceLength, int TargetLength,
                                                   float *Probs ) 
{
  int i, j;
  int maxP;
  double P;
  double LogP;
  double PP [SourceLength + 1][TargetLength + 1];
  int ToEmptyWordFrom[SourceLength + 1];
  ToEmptyWordFrom[0] = 0;

  if(!m_UseEmptyWord){
    cout << "For HMM empty word is needed! " << endl;
    exit(-1);
  }

  //calculate Jump Probabilities, if necessary
  if(m_JumpProbFilled < TargetLength)
    CalcJumpProb(TargetLength);

  for(j = 0; j <= SourceLength; j++)
    for(i = 0; i <= TargetLength; i++) 
      PP[j][i] = 0.0;


  //going over all columes
  for ( j = 1; j <= SourceLength; j++ ) {
    
#if DEBUG
    if(m_DisplayForwardAlgo == 2)
      cout << "Calculating Column " << j << endl;
#endif

    P = 0.0;
    //1. Calculate Probability for SourceWord translatet to emtpy Word
    
    //case1: first word in sentence
    if(j == 1) {
      P = 1.0;
      maxP = 0;
      //case 2: last Source Position was in Phrase -> look only at Probs from Phrase
    }else if((j-1 >= j1) && (j-1 <= j2)) {
      //Jumping cost to Empty word all the same
      maxP = 0;
      P = 1 * PP[(j - 1)] [0];
      for(int is = i1; is <= i2; is++){
        if(PP[(j - 1)][is] > P) {
          //is is entry with highest prob to go to empty word
          maxP = is;
          P = 1*PP[(j - 1)][is];
        }
      }
    
      //last Source Position was outside Phrase -> lokk only at Probs outside Phrase
    }else { 
      maxP = 0;
      P = 1 * PP[(j - 1)] [0];
      for(int is = 1; is <= TargetLength; is++) {
        if( (is < i1) || (is > i2) ) {
          if(PP[(j - 1)][is] > P) {
            //is is entry with highest prob to go to empty word
            maxP = is;
            P = 1*PP[ (j - 1) ][ is ];
          }
        }
      }
    }
    
    if(maxP != 0)
      ToEmptyWordFrom[j] = maxP;
    else {
      int k = j - 1;
      for(; k > 1 && ToEmptyWordFrom[k] == 0; k-- ){
      }
      ToEmptyWordFrom[j] = ToEmptyWordFrom[k];
    }

    PP[j][0] = P * Probs[j* MAX_SENTENCE_LENGTH + 0];

#if DEBUG
    if(m_DisplayForwardAlgo == 2) {
      cout << "Line 0" << endl;
      cout << "ToEmptyWordFrom[" << j << "] = " << ToEmptyWordFrom[j] << endl;
      cout << " P : " << P << endl;
      cout << "Probs: " << Probs[j* MAX_SENTENCE_LENGTH + 0] << endl;
      cout << " => PP[ " << j << " ][0] = " << PP[j][0] << endl;
    }

#endif
 
    // Filling whole colum line by line
    //case1: Source Word is part of Phrase -> Calculation only PP for words in Phrase
    if ( (j >= j1) && (j <= j2) ) {
      for ( i = i1; i <= i2; i++ ) {
        P = 0.0;
        //case a: first word in sentence
        if(j == 1) {
          P = 1.0;
          //case b: last Source Position was in Phrase -> look only at Probs from Phrase
        }else if((j-1 >= j1) && (j-1 <= j2)) {
          P = JumpCost(ToEmptyWordFrom[j - 1],i)*PP[(j - 1)][0];
          for(int is = i1; is <= i2; is++){
            if(JumpCost(is,i)*PP[(j - 1)][is] > P)
              P = JumpCost(is,i)*PP[(j - 1)][is];
          }

          //case c:last Source Positon was outside Phrase -> look only at Probs outside Phrase
        }else { 
          P = JumpCost(ToEmptyWordFrom[j - 1],i)*PP[(j - 1)][0];
          for(int is = 1; is <= TargetLength; is++) {
            if( (is < i1) || (is > i2) ) {
              if(JumpCost(is,i)*PP[(j - 1)][is] > P)
                P = JumpCost(is,i)*PP[(j - 1)][is];
            }
          }
        }
      

        PP[j][i] = P * Probs[j* MAX_SENTENCE_LENGTH + i];

#if DEBUG
        if(m_DisplayForwardAlgo == 2) {
          cout << "Line " << i << " (word if part of phrase)" << endl;
          cout << " P : " << P << endl;
          cout << "Probs: " << Probs[j* MAX_SENTENCE_LENGTH + i] << endl;
          cout << " => PP[ " << j << " ][ " << i  << " ] = " << PP[j][i] << endl;
        }

#endif

      }

    } else {
      //Case2: Source Word is not part of Phrase -> Add all Probabilities of words not in TargetPhrase
      for ( i = 1; i <= TargetLength; i++ ) {
        if ( (i < i1) || (i > i2) ) {
          P = 0.0;
          //case a: first word in sentence
          if(j == 1) {
            P = 1.0;
	
            //case b: last Source Position was in Phrase -> add only over Probs from Phrase
          }else if((j-1 >= j1) && (j-1 <= j2)) {
            P = JumpCost(ToEmptyWordFrom[j - 1],i)*PP[(j - 1)][0];
            for(int is = i1; is <= i2; is++){
              if(JumpCost(is,i)*PP[(j - 1)][is] > P)
                P = JumpCost(is,i)*PP[(j - 1)][is];
            }

            //case c: last Source Position was outside Phrase -> add only over Probs outside Phrase
          }else { 
            P = JumpCost(ToEmptyWordFrom[j - 1],i)*PP[(j - 1)][0];
            for(int is = 1; is <= TargetLength; is++) {
              if( (is < i1) || (is > i2) ) {
                if(JumpCost(is,i)*PP[(j - 1)][is] > P)
                  P = JumpCost(is,i)*PP[(j - 1)][is];
              }
            }
          }
   

          PP[j][i] = P * Probs[j* MAX_SENTENCE_LENGTH + i];

#if DEBUG
          if(m_DisplayForwardAlgo == 2) {
            cout << "Line " << i << " (word if not part of phrase)" << endl;
            cout << " P : " << P << endl;
            cout << "Probs: " << Probs[j* MAX_SENTENCE_LENGTH + i] << endl;
            cout << " => PP[ " << j << " ][ " << i  << " ] = " << PP[j][i] << endl;
          }

#endif

        }
      }
    }
    
  }

 
  P = 0.0;
  P = JumpCost(ToEmptyWordFrom[SourceLength],TargetLength + 1) * PP[SourceLength][0];
 
  //case a: last Source Word was in Phrase
  if(j2 == SourceLength) {
    for(i = i1;i <= i2; i++) {
      if(JumpCost(i, TargetLength + 1)*PP[SourceLength][i] > P)
        P = JumpCost(i, TargetLength + 1)*PP[SourceLength][i];
    }
    //case b: last Source Word was not in Phrase  
  }else {
    for(i = 1; i <= TargetLength; i++) {
      if((i < i1) || (i > i2)) {
        if(JumpCost(i, TargetLength + 1)*PP[SourceLength][i] > P)
          P = JumpCost(i, TargetLength + 1)*PP[SourceLength][i];
      }
    }

  }

#if DEBUG
  if(m_DisplayForwardAlgo == 2) {
    cout << "Calcuation of last Field: " << endl;
    cout << " P : " << P << endl;    
  }
#endif 

  double JumpCorrection = CalcJumpCorrection(SourceLength,TargetLength);
#if DEBUG
  if(m_DisplayForwardAlgo == 2) {
    cout << "JumpCorrection: (*1000)" << JumpCorrection * 1000<< endl;
    cout << "SourceLength: " << SourceLength << " TargetLength: " << TargetLength << endl;
  }
#endif
  P /= JumpCorrection;
  LogP = -log( P );



#if DEBUG
  if(m_DisplayForwardAlgo) {
    cout << "Table : " << endl;
    DisplaySentProb(&(PP[0][0]),j1,j2,i1,i2);
    cout << "-> P = " << P << endl;
    cout << "=> return -logP/" << SourceLength << " = " << LogP/SourceLength << endl;
  }
#endif
  
  return LogP/SourceLength;

}



float
CPhraseAlignProbCalcViterbi::CalcPhraseLogProb( int j1, int j2, int i1, int i2, float *Probs )
{
  int i, j;
  double P;
  int maxP;
  double LogP;
  int SrcPhraseLength = j2 - j1 + 1;
  int TrgPhraseLength = i2 - i1 + 1;
  double PP [SrcPhraseLength + 1][TrgPhraseLength + 1];
  int ToEmptyWordFrom[SrcPhraseLength + 1];
  ToEmptyWordFrom[0] = 0;


  if(!m_UseEmptyWord){
    cout << "For HMM empty word is needed! " << endl;
    exit(-1);
  }

  if(m_JumpProbFilled < TrgPhraseLength)
    CalcJumpProb(TrgPhraseLength);


  for(j = 0; j <= SrcPhraseLength; j++)
    for(i = 0; i <= TrgPhraseLength; i++) 
      PP[j][i] = 0.0;

    
  //going over all columes
  for ( j = 1; j <= SrcPhraseLength; j++ ) {

#if DEBUG
    if(m_DisplayForwardAlgo == 2)
      cout << "Calculating Column " << j << endl;
#endif


    P = 0.0;
    //1. Calculate Probability for SourceWord translatet to emtpy Word
    
    //case1: first word in sentence
    if(j == 1) {
      P = 1.0;
      maxP = 0;

    }else {
      maxP = 0;
      P = 1 * PP[j - 1][0];
      for(int is = 1; is <= TrgPhraseLength; is++) {
        if(PP[(j - 1)][is] > P) {
          //is is entry with highest prob to go to empty word
          maxP = is;
          P = 1*PP[ (j - 1) ][ is ];
        }
	
      }
    }

    if(maxP != 0)
      ToEmptyWordFrom[j] = maxP;
    else {
      int k = j - 1;
      for(; k > 1 && ToEmptyWordFrom[k] == 0; k-- ){
      }
      ToEmptyWordFrom[j] = ToEmptyWordFrom[k];
    }

    PP[j][0] = P * Probs[(j + j1 -1)* MAX_SENTENCE_LENGTH + 0];

#if DEBUG
    if(m_DisplayForwardAlgo == 2) {
      cout << "Line 0" << endl;
      cout << "ToEmptyWordFrom[" << j << "] = " << ToEmptyWordFrom[j] << endl;
      cout << " P : " << P << endl;
      cout << "Probs: " << Probs[(j + j1 - 1)* MAX_SENTENCE_LENGTH + 0] << endl;
      cout << " => PP[ " << j << " ][0] = " << PP[j][0] << endl;
    }

#endif

    
    //Filling whole colum line by line
    for ( i = 1; i <=  TrgPhraseLength; i++ ) {
      P=0.0;
      //case a: first word in sentence
      if(j == 1) {
        P = 1.0;
      }else {
        P = JumpCost(ToEmptyWordFrom[j-1],i) * PP[(j-1)][0];
        for(int is = 1; is <= TrgPhraseLength;is++) {
          if(JumpCost(is,i)*PP[(j-1)][is] > P)
            P = JumpCost(is,i)*PP[(j-1)][is];
        }


      }
      
      PP[j][i] = P * Probs[(j+ j1 -1)* MAX_SENTENCE_LENGTH + (i+ i1 - 1)];

#if DEBUG
      if(m_DisplayForwardAlgo == 2) {
        cout << "Line " << i << endl;
        cout << " P : " << P << endl;
        cout << "Probs: " << Probs[(j + j1 - 1)* MAX_SENTENCE_LENGTH + (i + i1 - 1)] << endl;
        cout << " => PP[ " << j << " ][ " << i  << " ] = " << PP[j][i] << endl;
      }

#endif

    
    }
  }

  P = 0.0;
  P = JumpCost(ToEmptyWordFrom[SrcPhraseLength],TrgPhraseLength + 1) * PP[SrcPhraseLength][0];

  for(i = 1;i <= TrgPhraseLength; i++) {
    if(JumpCost(i, TrgPhraseLength + 1) * PP[SrcPhraseLength][i] > P)
      P = JumpCost(i, TrgPhraseLength + 1) * PP[SrcPhraseLength][i];
  }

#if DEBUG
  if(m_DisplayForwardAlgo == 2) {
    cout << "Calcuation of last Field: " << endl;
    cout << " P : " << P << endl;    
  }
#endif 

  double JumpCorrection = CalcJumpCorrection(SrcPhraseLength,TrgPhraseLength);
#if DEBUG
  if(m_DisplayForwardAlgo == 2) {
    cout << "JumpCorrection: (*1000)" << JumpCorrection * 1000<< endl;
    cout << "SourceLength: " << SrcPhraseLength << " TargetLength: " << TrgPhraseLength << endl;
  }
#endif
  P /= JumpCorrection;

  LogP = -log( P );

#if DEBUG
  if(m_DisplayForwardAlgo) {
    cout << "Table : " << endl;
    DisplayPhraseProb(&(PP[0][0]),SrcPhraseLength,TrgPhraseLength,j1, i1);
    cout << "-> P = " << P << endl;
    cout << "=> return -logP/" << SrcPhraseLength << " = " << LogP/SrcPhraseLength << endl;
  }
#endif

  return LogP/SrcPhraseLength;

}



double
CPhraseAlignProbCalcViterbi::CalcJumpCorrection( int SourceLength, int TargetLength )
{
  double MaxProb = 1.0;
  //case a: Source is longer or equal than Target
  if(SourceLength >= TargetLength) {
    // -> TargetLength time a 1 Step jumps
    for(int i = 1; i <= TargetLength; i++)
      MaxProb *= JumpCost(i,i+1);
    for(int i = 1; i <= SourceLength - TargetLength; i++)
      MaxProb *= JumpCost(i,i);
  }else{
    for(int i = 1; i <= SourceLength; i++)
      MaxProb *= JumpCost(i,i+1);
    //MaxProb *= JumpCost(SourceLength,TargetLength + 1);
  }
  return MaxProb;

}

