#ifndef __PHRASE_ALIGN_PROB_CALC2_
#define __PHRASE_ALIGN_PROB_CALC2_

#include <math.h>
#include <iostream>
#include <string.h>
#ifndef MAX_SENTENCE_LENGTH
#define MAX_SENTENCE_LENGTH 201
#endif
#define PI 3.14159265

#define DEBUG 1
#if DEBUG
#include <Text.hh>
#endif



class CPhraseAlignProbCalc
{
public:
  CPhraseAlignProbCalc();
  ~CPhraseAlignProbCalc();

  // use empty word
  void SetUseEmptyWord( bool UseEmptyWord );

  // calcs Probability for Source-Target Source Pair
  //Probs is an array of size MAX_SENTENCE_LENGTH * MAX_SENTENCE_LENGTH;
  virtual float CalcSentAlignLogProb( int j1, int j2, int i1, int i2,
                                      int SourceLength, int TargetLength,
                                      float *Probs );

  // calcs Probability for Source-Target Source Pair
  //Probs is an array of size MAX_SENTENCE_LENGTH * MAX_SENTENCE_LENGTH;
  virtual float CalcPhraseLogProb( int j1, int j2, int i1, int i2, float *Probs );

#if DEBUG
  int m_DisplayForwardAlgo;
  bool m_Direction;
  virtual void SetSentence( CSentence *pSourceSentence, CSentence *pTargetSentence,
                            const CVocabulary *pVoc );
#endif

protected:
  bool m_UseEmptyWord;

};




class CPhraseAlignProbCalcIBM1:public CPhraseAlignProbCalc
{
public:
  CPhraseAlignProbCalcIBM1();
  ~CPhraseAlignProbCalcIBM1();

  // calcs Probability for Source-Target Source Pair
  //Probs is an array of size MAX_SENTENCE_LENGTH * MAX_SENTENCE_LENGTH;
  float CalcSentAlignLogProb( int j1, int j2, int i1, int i2, 
                              int SourceLength, int TargetLength,
                              float *Probs );

  // calcs Probability for Source-Target Source Pair
  //Probs is an array of size MAX_SENTENCE_LENGTH * MAX_SENTENCE_LENGTH;
  float CalcPhraseLogProb( int j1, int j2, int i1, int i2, float *Probs );

};




class CPhraseAlignProbCalcHMM:public CPhraseAlignProbCalc
{
public:
  CPhraseAlignProbCalcHMM();
  ~CPhraseAlignProbCalcHMM();


  // calcs Probability for Source-Target Source Pair
  //Probs is an array of size MAX_SENTENCE_LENGTH * MAX_SENTENCE_LENGTH;
  float CalcSentAlignLogProb( int j1, int j2, int i1, int i2,
                              int SourceLength, int TargetLength,
                              float *Probs );


  // calcs Probability for Source-Target Source Pair
  //Probs is an array of size MAX_SENTENCE_LENGTH * MAX_SENTENCE_LENGTH;
  float CalcPhraseLogProb( int j1, int j2, int i1, int i2, float *Probs );

  //calculates JumpProbabilities
  // 0 is empty word
  double JumpCost( int from, int to );


#if DEBUG
  void SetSentence( CSentence *pSourceSentence, CSentence *pTargetSentence,
                    const CVocabulary *pVoc );
#endif

protected:
#if DEBUG
  void DisplayPhraseProb( double *Probs, int j1, int j2, int i1, int i2 );
  void DisplaySentProb( double *Probs, int j1, int j2, int i1, int i2 );
  int GetMaxWordLength();
  void DisplaySourceSentence( int maxLen );
  void DisplaySourcePhrase( int maxLen, int SrcPhraseLength, int j1 );
  
  int m_J;
  int m_I;
  CSentence *m_pSourceSentence;
  CSentence *m_pTargetSentence;
  const CVocabulary *m_pVoc;
  

#endif
  void CalcJumpProb( int FillTo );


  double m_JumpProb[ MAX_SENTENCE_LENGTH ];
  double m_JumpProbNeg[ MAX_SENTENCE_LENGTH ];
  int m_JumpProbFilled;

};




class CPhraseAlignProbCalcViterbi:public CPhraseAlignProbCalcHMM
{
public:
  CPhraseAlignProbCalcViterbi();
  ~CPhraseAlignProbCalcViterbi();


  // calcs Probability for Source-Target Source Pair
  //Probs is an array of size MAX_SENTENCE_LENGTH * MAX_SENTENCE_LENGTH;
  float CalcSentAlignLogProb( int j1, int j2, int i1, int i2,
                              int SourceLength, int TargetLength,
                              float *Probs );

  // calcs Probability for Source-Target Source Pair
  //Probs is an array of size MAX_SENTENCE_LENGTH * MAX_SENTENCE_LENGTH;
  float CalcPhraseLogProb( int j1, int j2, int i1, int i2, float *Probs );

private:
  double CalcJumpCorrection( int SourceLength, int TargetLength );

};



inline double
CPhraseAlignProbCalcHMM::JumpCost( int from, int to )
{
  int jump = to - from;
  //return 1.0/((jump - 1)*(jump - 1) + 1);
  if(jump >= 0)
    return m_JumpProb[ jump ];
  else
    return m_JumpProbNeg[ (-1) * jump ];
}


#endif
