  virtual bool FindBestPhrasePair( CSentence *pSourceSentence, CSentence *pTargetSentence,
								   int MinPhraseLength, int MaxPhraseLength,
								   CPhraseAlignItemMS & PhraseAlignItems );
  

bool
CPhraseAlignerMS::FindBestPhrasePair( CSentence *pSourceSentence, CSentence *pTargetSentence,
									  int MinPhraseLength, int MaxPhraseLength,
									  CPhraseAlignItemMS & PhraseAlignItems )
{
  // we do for all source phrases between MinPhraseLength and MaxPhraseLength
  // the alignment and keep track of the best results, this is then returned.
  // i.e. for each sentence pair we get one phrase pair
  // we can evaluate based on combined score, or look into perplexity gain,
  // as we do in sentence splitting (could be total or per word)

  // first we need to fill the probability matrices
  // length of the sentences
  m_J = pSourceSentence->Length();
  m_I = pTargetSentence->Length();
  //cout << "Length source " << m_J << " length target " << m_I << endl;

  // if sentences are longer than we can store probabilities, we skip it
  if ( (m_J >= MAX_SENTENCE_LENGTH) || (m_I >= MAX_SENTENCE_LENGTH) ) {
	cout << "WARNING: Sentence is too long, we skip it" << endl;
    return false;
  }

  // first, get all the probabilities and put them into static matrix
  // get Source->Target Probabilities
  GetProbabilities( m_pSourceSentence, m_pTargetSentence, m_pLexiconSourceTarget, &(m_ProbsSourceTarget[ 0 ][ 0 ]) );
  // renormalize probabilities for the current sentence pair
  Renormalize( m_J, m_I, &(m_ProbsSourceTarget[ 0 ][ 0 ]), &(m_RenProbsSourceTarget[ 0 ][ 0 ]) );

  //get Target->Source Probabilities
  if( m_AlignBidirectional ) {
    GetProbabilities( m_pTargetSentence, m_pSourceSentence, m_pLexiconTargetSource, &(m_ProbsTargetSource[ 0 ][ 0 ]) );
    // renormalize probabilities for the current sentence pair
    Renormalize( m_I, m_J, &(m_ProbsTargetSource[ 0 ][ 0 ]), &(m_RenProbsTargetSource[ 0 ][ 0 ]) );
  }

  int StartPosSource;
  int EndPosSource;

  // now we loop over all possible source phrases (within length restrictions)
  // and search for best target phrase
  for ( int Length = MinPhraseLength; Length <= MaxPhraseLength; Length++ ) {
	cout << "Length is now " << Length << endl;

	for ( StartPosSource = pSourceSentence->MinIndex();
		  StartPosSource < pSourceSentence->MaxIndex() - Length;
		  StartPosSource++ ) {

	  EndPosSource = StartPosSource + Length;
	  cout << "  StartPosSource = " << StartPosSource << " EndPosSource " << EndPosSource << endl;

	  // loop over all possible target lengths
	  // don't want too short or too long translations
	  // on average the target length should be source length times stretch factor
	  // where stretch factor is the ratio target corpus lenght to source corpus length
	  // then target should not be longer then this average length time a elongation factor
	  // and not shorter then the average length time a shortening factor
	  // we allow for and addition +/-2 to have e.g. 1->3 alignemnts
	  //	  int MinTargetLength = (int)(-1.5 + LengthSourcePhrase * m_StretchFactor * m_FactorShorter);
	  //int MaxTargetLength = (int)(2.5 + LengthSourcePhrase * m_StretchFactor * m_FactorLonger);
	  //if ( MinTargetLength < 1 )
	  //	MinTargetLength = 1;

  
	}
  }
	  
}

 

