#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sstream>
#include <set>


// switch on/off tracing for this module here
#undef NO_TRACE
#define NO_TRACE 1
#include <Trace.hh>

#include <PhrasePairStats.hh>

using namespace std;




CPhrasePairStats::CPhrasePairStats()
{
  Trace( 1, "CPhrasePairStats::CPhrasePairStats" );

  m_pSourceSA = NULL;
  m_pTargetSA = NULL;
  m_pVoc = NULL;
}



CPhrasePairStats::~CPhrasePairStats()
{
}


void
CPhrasePairStats::Init( C_SuffixArraySearch *pSourceSA, C_SuffixArraySearch *pTargetSA, CVocabulary *pVoc )
{
  m_pSourceSA = pSourceSA;
  m_pTargetSA = pTargetSA;
  m_pVoc = pVoc;
}




CPhrasePairCounts
CPhrasePairStats::GetPhrasePairCounts( const CSentence *pSourcePhrase, const CSentence *pTargetPhrase )
{
  CPhrasePairCounts PhrasePairCounts;

  GetPhrasePairCounts( pSourcePhrase, pTargetPhrase,
                       PhrasePairCounts.SourcePhraseCount, PhrasePairCounts.TargetPhraseCount, PhrasePairCounts.PhrasePairCount );

  return PhrasePairCounts;
}



//! get for a given phrase pair the frequency counts of souce and target phrase and the co-occurrence count
CPhrasePairCounts
CPhrasePairStats::GetPhrasePairCounts( const CNumericalArray< CWordIndex > *pSourcePhrase,
                                       const CNumericalArray< CWordIndex > *pTargetPhrase )
{
  return GetPhrasePairCounts( (const CSentence *)pSourcePhrase, (const CSentence *)pTargetPhrase );
}




void
CPhrasePairStats::GetPhrasePairCounts( const CSentence *pSourcePhrase, const CSentence *pTargetPhrase,
                                       int& FreqSourcePhrase, int& FreqTargetPhrase, int& FreqPhrasePair )
{
  //cout << "GetPhrasePairCounts" << endl;

  CString SourcePhrase = pSourcePhrase->AsString( m_pVoc );
  CString TargetPhrase = pTargetPhrase->AsString( m_pVoc );

  int SourcePhraseLength = pSourcePhrase->Length();
  int TargetPhraseLength = pTargetPhrase->Length();


  vector< S_phraseLocationElement > SourcePhraseLocations;
  vector< S_phraseLocationElement > TargetPhraseLocations;
  //vector< S_locationElement > SourcePhraseLocations;
  //vector< S_locationElement > TargetPhraseLocations;

  //  cout << "Frequency source phrase " << SourcePhrase << " = " << m_pSourceSA->frequencyOfPhrase( SourcePhrase ) << endl;
  //  cout << "Frequency target phrase " << TargetPhrase << " = " << m_pTargetSA->frequencyOfPhrase( TargetPhrase ) << endl;
  set< int, less< int> > SourceSentNum;
  set< int, less< int> > TargetSentNum;
  set< int, less< int> > CoocSentNum;

  // set min phrase length as to not search for embedded phrases
  m_pSourceSA->setParam_shortestUnitToReport( SourcePhraseLength );
  m_pTargetSA->setParam_shortestUnitToReport( TargetPhraseLength );

  SourcePhraseLocations = m_pSourceSA->findPhrasesInASentence( SourcePhrase );
  TargetPhraseLocations = m_pTargetSA->findPhrasesInASentence( TargetPhrase );

  for ( int n = 0; n < SourcePhraseLocations.size(); n++ ) {
    //cout <<  SourcePhraseLocations[ n ].sentId << endl;
    SourceSentNum.insert( SourcePhraseLocations[ n ].sentIdInCorpus );
  }
  //cout << endl;
  for ( int n = 0; n < TargetPhraseLocations.size(); n++ ) {
    //cout << TargetPhraseLocations[ n ].sentId << endl;
    TargetSentNum.insert( TargetPhraseLocations[ n ].sentIdInCorpus );
  }

  //cout << "Number of source phrases " << SourceSentNum.size() << endl;
  //cout << "Number of target phrases " << TargetSentNum.size() << endl;

  //  CoocSentNum = 
  insert_iterator< set < int, less< int > > > Insert_Iterator( CoocSentNum, CoocSentNum.begin() );
  
  set_intersection( SourceSentNum.begin(), SourceSentNum.end(),
                    TargetSentNum.begin(), TargetSentNum.end(),
                    Insert_Iterator );    

  //cout << "Number of phrase pairs " << CoocSentNum.size() << endl;

  FreqSourcePhrase = SourceSentNum.size();
  FreqTargetPhrase = TargetSentNum.size();
  FreqPhrasePair   = CoocSentNum.size();

  //cout << endl;
}
