#ifndef __PhrasePairStats__
#define __PhrasePairStats__


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <Hash.hh>
#include <_SuffixArraySearch.h>
#include <NumericalArray.hh>
#include <Text.hh>



//! source phrase, target phrase and co-occurence counts put together into one object
class CPhrasePairCounts
{
public:
  int SourcePhraseCount;
  int TargetPhraseCount;
  int PhrasePairCount;
};




/**
CPhrasePairStats returns phrase pair counts: source phrase count, target phrase count, co-occurence count.
The co-occurence count does not check if the source and target phrase would align to each other under
some alignment model.  It only counts how often source and target phrase co-occur in paralell sentences.
*/

class CPhrasePairStats
{
public:
  CPhrasePairStats();
  ~CPhrasePairStats();

  //! give the source and target side suffix arrays, which are used to get the counts and the sentence numbers
  void Init( C_SuffixArraySearch *pSourceSA, C_SuffixArraySearch *pTargetSA, CVocabulary *pVoc );

  //! get for a given phrase pair the frequency counts of souce and target phrase and the co-occurrence count
  CPhrasePairCounts GetPhrasePairCounts( const CSentence *pSourcePhrase, const CSentence *pTargetPhrase );

  //! get for a given phrase pair the frequency counts of souce and target phrase and the co-occurrence count
  CPhrasePairCounts GetPhrasePairCounts( const CNumericalArray< CWordIndex > *pSourcePhrase,
                                         const CNumericalArray< CWordIndex > *pTargetPhrase );

  //! same as before, just different way to call the function
  void GetPhrasePairCounts( const CSentence *pSourcePhrase, const CSentence *pTargetPhrase,
                            int& FreqSourcePhrase, int& FreqTargetPhrase, int& FreqPhrasePair );


private:
  //! suffix array for source side of the bilingual corpus
  C_SuffixArraySearch *m_pSourceSA;

  //! suffix array for target side of the bilingual corpus
  C_SuffixArraySearch *m_pTargetSA;

  //! the vocabulary to convert word index into string, as suffix array uses different word indices
  CVocabulary *m_pVoc;
};



#endif
