#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <fortify.h>

// switch on/off tracing for this module here
// #undef NO_TRACE
#define NO_TRACE 1
#include <Trace.hh>

#include <strstream>
#include <PhraseSearch.hh>

using namespace std;



CPhraseSearchItem::CPhraseSearchItem()
{
  m_Score = 0.0;
}


CNumericalArray< CWordIndex > & 
CPhraseSearchItem::TargetWordIDs()
{
  return m_TargetWordIDs;
}


double 
CPhraseSearchItem::Score() const
{
  return m_Score;
}


void
CPhraseSearchItem::SetScore( double Score )
{
  m_Score = Score;
}


double 
CPhraseSearchItem::LogProb() const
{
  return m_LogProb;
}


void
CPhraseSearchItem::SetLogProb( double LogProb )
{
  m_LogProb = LogProb;
}


CString
CPhraseSearchItem::TargetPhrase() const
{
  return m_TargetPhrase;
}


void
CPhraseSearchItem::SetTargetPhrase( CString TargetPhrase )
{
  m_TargetPhrase = TargetPhrase;
}



bool
CPhraseSearchItem::operator==( const CPhraseSearchItem &Item ) const
{
  //Trace( 0, "CPhraseSearchItem::operator==" );
  //cout << "compare " << m_TargetPhrase << " " << m_Score
  //	   << " with " << Item.m_TargetPhrase << " " << Item.m_Score << endl;

  if ( m_TargetWordIDs == Item.m_TargetWordIDs ) {
	//cout << "true" << endl;
	return true;
  } else {
	//cout << "false" << endl;
	return false;
  }
}


bool
CPhraseSearchItem::operator<( const CPhraseSearchItem &Item ) const
{
  //Trace( 0, "CPhraseSearchItem::operator<" );
  //cout << "compare " << m_TargetPhrase << " " << m_Score
  //	   << " with " << Item.m_TargetPhrase << " " << Item.m_Score << endl;

  if ( m_TargetWordIDs < Item.m_TargetWordIDs ) {
	//cout << "true" << endl;
	return true;
  } else {
	//cout << "false" << endl;
	return false;
  }
}

bool
CPhraseSearchItem::operator>( const CPhraseSearchItem &Item ) const
{
  //Trace( 0, "CPhraseSearchItem::operator>" );
  //cout << "compare " << m_TargetPhrase << " " << m_Score  
  //   << " with " << Item.m_TargetPhrase << " " << Item.m_Score << endl;
 
  if ( m_TargetWordIDs > Item.m_TargetWordIDs ) {
	//cout << "true" << endl;
	return true;
  } else {
	//cout << "false" << endl;
	return false;
  }
}



void
CPhraseSearch::SearchAlign( CSentence *pSentence, CLattice *pLattice )
{
  // pSentence is the sentence for which we want to find phrase translations
  // pLattice is the corresponding lattice, into which we insert the found 
  // phrase translation pairs as new edges

  // 1. Step: find occurences of all n-grams from the sentence in the corpus
  // 2. Step: for each occurence of a source phrase find one or more translations
  //          in this case use maximum lexical probability for j1..j2 -> i1..i2
  // 3. Step: merge translations from different occurences of the same 
  //          source n-gram into one sorted list and discard bad candidates

  int SentenceID, PosInSentence, StartPosPhrase, EndPosPhrase;

  // deal with one source phrase at a time
  // typically has many translations
  // need to sort them according to translation score
  // and generated edges only for the best ones
  CArray< CWordIndex > SourceWordIDs;
  CArray< CArray< CWordIndex > * > TargetWordIDs;
  CSortedArray< CPhraseSearchItem > AllTranslations;

  CString SourceNgram;
  CString TargetNgram;  // the new one
  //CArray< CString > TargetNgrams;  // those seen before

  //map< double, CArray< CWordIndex > > SortedTargetWordIDs;
  // which keep each translation only once, but with best score
  //hash_map< char *, double > TranslationsWithScores;

  CPhraseSearchItem TranslationWithScore;

  CLatticeNode *pFromNode;
  CLatticeNode *pToNode;

  double BestScore;  // score for best traslation for one source phrase

#ifdef TRACE_ENABLED
  pSentence->Display( m_pVoc );
  cout << endl;
#endif
  // convert sentence into char*;
  ostrstream StrStream;
  pSentence->Write( StrStream, m_pVoc );
  StrStream << '\0';
  //cout << StrStream.str() << endl;
  //StrStream << "\0";
  //cout << StrStream.str() << endl;
  //cout << "Number of words in CPhraseSearch::Search is " << pSentence->Length() << endl;

  // init the phrase aligner with the lexicon
  m_pPhraseAligner = new CPhraseAlignerIBM1;

  if ( m_pReverseLexicon ) {
    m_pPhraseAligner->Init( m_pLexicon, m_pReverseLexicon );
    m_pPhraseAligner->SetInterpolationFactor( m_LambdaLexicon );
    m_pPhraseAligner->SetAlignBidirectional( true );
  } else {
    m_pPhraseAligner->Init( m_pLexicon, NULL );
  }

  // 1. Step: find locations of phrases from sentence in corpus
  vector< S_phraseLocationElement > results;
  results = m_pSA->findPhrasesInASentence( StrStream.str() );

  // 2. Step: find translations
  // for one source phrase store start and end position of
  // target phrases and the translation scores in the following arrays
  CNumericalArray< int > Min_i;
  CNumericalArray< int > Max_i;
  CNumericalArray< double > Scores;

/*
  for ( int n = 0; n < results.size(); n++ ) {
    StartPosPhrase = results[ n ].srcPosStart;
    EndPosPhrase = results[ n ].srcPosEnd ;
    SentenceID = results[ n ].sentId;
    PosInSentence = results[n].posInSent;

    cout << StartPosPhrase << " "
         << EndPosPhrase << " "
         << SentenceID << " "
         << PosInSentence << endl;
  }

*/
  for ( int n = 0; n < results.size(); n++ ) {
    StartPosPhrase = results[ n ].srcPosStart;
    EndPosPhrase = results[ n ].srcPosEnd ;
    SentenceID = results[ n ].sentId;
    PosInSentence = results[n].posInSent;

/*
    cout << endl << n <<  " Found in Corpus: " 
	 << SentenceID << " "
	 << StartPosPhrase << " "
	 << EndPosPhrase << " "
	 << PosInSentence << endl << endl;
*/

    bool IsLast = ( n == results.size() - 1 );

    bool NextIsNew = ( ( n + 1 < results.size() )
                       && ( (StartPosPhrase != results[ n + 1 ].srcPosStart) 
                            || (EndPosPhrase != results[ n + 1 ].srcPosEnd)
                          )
                     );

    bool IsNewSource = ( (n == 0)
                         || (StartPosPhrase != results[ n - 1 ].srcPosStart)
                         || (EndPosPhrase   != results[ n - 1 ].srcPosEnd) );

    bool CreateEdges = ( IsLast || NextIsNew );

/*
    cout << "NextIsNew: " << NextIsNew
         << " IsLast: " << IsLast
         << " IsNew: " << IsNewSource
         << " Create Edges: " << CreateEdges << endl;
*/
    if ( IsNewSource ) {
      // collect new one
      //cout << endl << "NEW SOURCE PHRASE" << endl;
      SourceNgram = "";
      SourceWordIDs.Clear();
      for ( int k = StartPosPhrase - 1; k <= EndPosPhrase - 1; k++ ) {
        //cout << j << " " << i << " " << TextE.GetWord( SentenceID - 1, i ) << endl;
        SourceWordIDs.Append( pSentence->GetAt( k ) );   // sentence is 0-based
        SourceNgram += (CString)" " + pSentence->Word( k , m_pVoc );
      }

      // initialize best score for translations for new scourse phrase
      BestScore = 10000;

    }
    //cout << "Source phrase: " << SourceNgram << endl;

    CSentence *pSourceSentence = m_pSourceText->GetSentence( SentenceID - 1 );
    //pSourceSentence->Display( m_pVoc ); cout << endl;
    //cout << "Lenght sentence is: " << pS->Length() << endl;
    CSentence *pTargetSentence = m_pTargetText->GetSentence( SentenceID - 1 );
    //pTargetSentence->Display( m_pVoc ); cout << endl;
	
    // now we want to find the best translations for the source phrase
    // in the target sentence by doing an IBM1 style alignment
    // there will be several target phrases for each source phrase
    // sorted according to alignment score

    int StartSource = PosInSentence;
    int EndSource   = PosInSentence + ( EndPosPhrase - StartPosPhrase );

    //cout << "call phrase aligner" << endl;

    m_pPhraseAligner->AlignPhrase( pSourceSentence, pTargetSentence,
                                   StartSource, EndSource,
                                   Min_i, Max_i, Scores  );

/*
    cout << "Found results: " << endl;
    for ( int m = Min_i.MinIndex(); m <= Min_i.MaxIndex(); m++ ) {
      cout << Min_i[ m ] << " " << Max_i[ m ] << " " << Scores[ m ] << endl;
    }
*/

    double LogProb;
    bool IsNew;

    for ( int m = Min_i.MinIndex(); m <= Min_i.MaxIndex(); m++ ) {

      //cout << "Deal with translation candidate " << m << " score = " << Scores[ m ] << endl;
      //cout << "Current best score = " << BestScore << endl;

      // some initial pruning: keep only reasonable good translations
      if ( Scores[ m ] < BestScore + 3.0 ) {

        //cout << "score is within beam" << endl;

        if ( Scores[ m ] < BestScore ) {
          BestScore = Scores[ m ];
          //cout << "new best score = " << BestScore << endl;
        }

        TargetNgram = "";
        CNumericalArray< CWordIndex > &TargetWordIDs = TranslationWithScore.TargetWordIDs();
        TargetWordIDs.Clear();

        for ( int i = Min_i[ m ]; i <= Max_i[ m ]; i++ ) {
          //cout << TextE.GetWord( SentenceID - 1, i ) << " ";
          TargetWordIDs.Append( m_pTargetText->GetWordAt( SentenceID - 1, i ) );
          TargetNgram += (CString)" " + m_pVoc->GetWord( m_pTargetText->GetWordAt( SentenceID - 1, i ) );
        }
        TranslationWithScore.SetTargetPhrase( TargetNgram );
        LogProb = PhraseTranslationLogProb( SourceWordIDs, TargetWordIDs, m_pLexicon );
        TranslationWithScore.SetScore( Scores[ m ] );
        TranslationWithScore.SetLogProb( LogProb );

        //cout << Scores[ m ] << " " << LogProb << " " << TargetNgram << endl;
		
        // store now in list of all good translations for this source phrase
        int Location;
        if ( AllTranslations.Find( TranslationWithScore, Location ) ) {
          //cout << " translation was found" << endl;
          if ( TranslationWithScore.Score() < AllTranslations.GetAt( Location ).Score() ) {
            // alread stored, but current score is lower, i.e. better
            // therefore overwrite it
            //cout << " translation was found, but new score is smaller, i.e. overwrite " << endl;
            AllTranslations.Insert( TranslationWithScore, IsNew );
            //if ( IsNew ) { 
            //  cout << " new translation inserted " << endl;
            //}
          //} else {
            //cout << " but new score is larger, i.e. don't use " << endl;
          }
        } else {
          //if ( TranslationWithScore.Score() < 
          //cout << " translation was not found - insert" << endl;
          AllTranslations.Insert( TranslationWithScore, IsNew );
        }
      }
    }


    if ( CreateEdges ) {
      // create edges for the best translations for the last source phrase
      // need to retrieve position of former source phrase
      int StartPos = results[ n ].srcPosStart - 1;
      int EndPos = results[ n ].srcPosEnd;

      for ( int i = AllTranslations.MinIndex(); i <= AllTranslations.MaxIndex(); i++ ) {
        //cout << AllTranslations.GetAt( i ).Score() << " " 
        //     << AllTranslations.GetAt( i ).LogProb() << " " 
        //     << AllTranslations.GetAt( i ).TargetPhrase() << endl;
  
        CPhraseSearchItem PhraseItem = AllTranslations.GetAt( i );
        CNumericalArray< CWordIndex > &TargetWordIDs = PhraseItem.TargetWordIDs();
        TargetNgram = "";
        for ( int j = TargetWordIDs.MinIndex(); j <= TargetWordIDs.MaxIndex(); j++ ) {
          TargetNgram += (CString)" " + m_pVoc->GetWord( TargetWordIDs[ j ] );
        }
        //cout << TargetNgram << endl;

        //CPhraseSearchItem PhraseItem = AllTranslations.GetAt( i );

        //cout << "n=" << n << " add Edge for " << StartPos << " " << EndPos << endl;

        AddEdge( pLattice, StartPos, EndPos, 
        AllTranslations.GetAt( i ).Score(), 
        SourceWordIDs, PhraseItem.TargetWordIDs() );
      }

      // forget the translations of the former source phrase
      AllTranslations.Clear();

    } // if ( CreateEdges )


        //for ( int i = AllTranslations.MinIndex(); i <= AllTranslations.MaxIndex(); i++ ) {
        //cout << AllTranslations.GetAt( i ).Score() << " " << AllTranslations.GetAt( i ).TargetPhrase() << endl;
        //CPhraseSearchItem PhraseItem = AllTranslations.GetAt( i );
        //CNumericalArray< CWordIndex > &TargetWordIDs = PhraseItem.TargetWordIDs();
        //TargetNgram = "";
        //for ( int j = TargetWordIDs.MinIndex(); j <= TargetWordIDs.MaxIndex(); j++ ) {
          //TargetNgram += (CString)" " + m_pVoc->GetWord( TargetWordIDs[ j ] );
        //}
        //cout << TargetNgram << endl;
        //}

        //AddEdge( pLattice, StartPosPhrase - 1, EndPosPhrase, 
        // Scores[ n ], SourceWordIDs, TargetWordIDs.GetRow( n ) );
  }
  delete m_pPhraseAligner;
}

void
CPhraseSearch::SearchAlignTraining(const int SentenceNumber, CSentence *pSentence, CLattice *pLattice )
{
  // pSentence is the sentence for which we want to find phrase translations
  // pLattice is the corresponding lattice, into which we insert the found 
  // phrase translation pairs as new edges

  // 1. Step: find occurences of all n-grams from the sentence in the corpus
  // 2. Step: for each occurence of a source phrase find one or more translations
  //          in this case use maximum lexical probability for j1..j2 -> i1..i2
  // 3. Step: merge translations from different occurences of the same 
  //          source n-gram into one sorted list and discard bad candidates

  int SentenceID, PosInSentence, StartPosPhrase, EndPosPhrase;

  // deal with one source phrase at a time
  // typically has many translations
  // need to sort them according to translation score
  // and generated edges only for the best ones
  CArray< CWordIndex > SourceWordIDs;
  CArray< CArray< CWordIndex > * > TargetWordIDs;
  CSortedArray< CPhraseSearchItem > AllTranslations;

  CString SourceNgram;
  CString TargetNgram;  // the new one
  //CArray< CString > TargetNgrams;  // those seen before

  //map< double, CArray< CWordIndex > > SortedTargetWordIDs;
  // which keep each translation only once, but with best score
  //hash_map< char *, double > TranslationsWithScores;

  CPhraseSearchItem TranslationWithScore;

  CLatticeNode *pFromNode;
  CLatticeNode *pToNode;

  double BestScore;  // score for best traslation for one source phrase

#ifdef TRACE_ENABLED
  pSentence->Display( m_pVoc );
  cout << endl;
#endif
  // convert sentence into char*;
  ostrstream StrStream;
  pSentence->Write( StrStream, m_pVoc );
  StrStream << '\0';
  //cout << StrStream.str() << endl;
  //StrStream << "\0";
  //cout << StrStream.str() << endl;
  //cout << "Number of words in CPhraseSearch::Search is " << pSentence->Length() << endl;

  // init the phrase aligner with the lexicon
  m_pPhraseAligner = new CPhraseAlignerIBM1;

  if ( m_pReverseLexicon ) {
    m_pPhraseAligner->Init( m_pLexicon, m_pReverseLexicon );
    m_pPhraseAligner->SetInterpolationFactor( m_LambdaLexicon );
    m_pPhraseAligner->SetAlignBidirectional( true );
  } else {
    m_pPhraseAligner->Init( m_pLexicon, NULL );
  }

  // 1. Step: find locations of phrases from sentence in corpus
  vector< S_phraseLocationElement > results;
  results = m_pSA->findPhrasesInASentence( StrStream.str() );


  //ARV
  //Training mode options
  //pruned list (with phrases from the same sentence in the SA removed)
  vector< S_phraseLocationElement > resultsPruned;
  for ( int n = 0; n < results.size(); n++ ) {
    SentenceID = results[ n ].sentId;
    if (SentenceID!=SentenceNumber){
      resultsPruned.push_back(results[n]);
      //debug
      //bool IsNew;
      //usedSentences.Insert(SentenceID, IsNew);
    }
  }
  cout<<"Pre pruning results.size() = "<<results.size()<<endl;
  results.clear();
  results = resultsPruned;
  cout<<"Post pruning results.size() = "<<results.size()<<endl;
  



  // 2. Step: find translations
  // for one source phrase store start and end position of
  // target phrases and the translation scores in the following arrays
  CNumericalArray< int > Min_i;
  CNumericalArray< int > Max_i;
  CNumericalArray< double > Scores;

/*
  for ( int n = 0; n < results.size(); n++ ) {
    StartPosPhrase = results[ n ].srcPosStart;
    EndPosPhrase = results[ n ].srcPosEnd ;
    SentenceID = results[ n ].sentId;
    PosInSentence = results[n].posInSent;

    cout << StartPosPhrase << " "
         << EndPosPhrase << " "
         << SentenceID << " "
         << PosInSentence << endl;
  }

*/
  for ( int n = 0; n < results.size(); n++ ) {
    StartPosPhrase = results[ n ].srcPosStart;
    EndPosPhrase = results[ n ].srcPosEnd ;
    SentenceID = results[ n ].sentId;
    PosInSentence = results[n].posInSent;

/*
    cout << endl << n <<  " Found in Corpus: " 
	 << SentenceID << " "
	 << StartPosPhrase << " "
	 << EndPosPhrase << " "
	 << PosInSentence << endl << endl;
*/

    bool IsLast = ( n == results.size() - 1 );

    bool NextIsNew = ( ( n + 1 < results.size() )
                       && ( (StartPosPhrase != results[ n + 1 ].srcPosStart) 
                            || (EndPosPhrase != results[ n + 1 ].srcPosEnd)
                          )
                     );

    bool IsNewSource = ( (n == 0)
                         || (StartPosPhrase != results[ n - 1 ].srcPosStart)
                         || (EndPosPhrase   != results[ n - 1 ].srcPosEnd) );

    bool CreateEdges = ( IsLast || NextIsNew );

/*
    cout << "NextIsNew: " << NextIsNew
         << " IsLast: " << IsLast
         << " IsNew: " << IsNewSource
         << " Create Edges: " << CreateEdges << endl;
*/
    if ( IsNewSource ) {
      // collect new one
      //cout << endl << "NEW SOURCE PHRASE" << endl;
      SourceNgram = "";
      SourceWordIDs.Clear();
      for ( int k = StartPosPhrase - 1; k <= EndPosPhrase - 1; k++ ) {
        //cout << j << " " << i << " " << TextE.GetWord( SentenceID - 1, i ) << endl;
        SourceWordIDs.Append( pSentence->GetAt( k ) );   // sentence is 0-based
        SourceNgram += (CString)" " + pSentence->Word( k , m_pVoc );
      }

      // initialize best score for translations for new scourse phrase
      BestScore = 10000;

    }
    //cout << "Source phrase: " << SourceNgram << endl;

    CSentence *pSourceSentence = m_pSourceText->GetSentence( SentenceID - 1 );
    //pSourceSentence->Display( m_pVoc ); cout << endl;
    //cout << "Lenght sentence is: " << pS->Length() << endl;
    CSentence *pTargetSentence = m_pTargetText->GetSentence( SentenceID - 1 );
    //pTargetSentence->Display( m_pVoc ); cout << endl;
	
    // now we want to find the best translations for the source phrase
    // in the target sentence by doing an IBM1 style alignment
    // there will be several target phrases for each source phrase
    // sorted according to alignment score

    int StartSource = PosInSentence;
    int EndSource   = PosInSentence + ( EndPosPhrase - StartPosPhrase );

    //cout << "call phrase aligner" << endl;

    m_pPhraseAligner->AlignPhrase( pSourceSentence, pTargetSentence,
                                   StartSource, EndSource,
                                   Min_i, Max_i, Scores  );

/*
    cout << "Found results: " << endl;
    for ( int m = Min_i.MinIndex(); m <= Min_i.MaxIndex(); m++ ) {
      cout << Min_i[ m ] << " " << Max_i[ m ] << " " << Scores[ m ] << endl;
    }
*/

    double LogProb;
    bool IsNew;

    for ( int m = Min_i.MinIndex(); m <= Min_i.MaxIndex(); m++ ) {

      //cout << "Deal with translation candidate " << m << " score = " << Scores[ m ] << endl;
      //cout << "Current best score = " << BestScore << endl;

      // some initial pruning: keep only reasonable good translations
      if ( Scores[ m ] < BestScore + 3.0 ) {

        //cout << "score is within beam" << endl;

        if ( Scores[ m ] < BestScore ) {
          BestScore = Scores[ m ];
          //cout << "new best score = " << BestScore << endl;
        }

        TargetNgram = "";
        CNumericalArray< CWordIndex > &TargetWordIDs = TranslationWithScore.TargetWordIDs();
        TargetWordIDs.Clear();

        for ( int i = Min_i[ m ]; i <= Max_i[ m ]; i++ ) {
          //cout << TextE.GetWord( SentenceID - 1, i ) << " ";
          TargetWordIDs.Append( m_pTargetText->GetWordAt( SentenceID - 1, i ) );
          TargetNgram += (CString)" " + m_pVoc->GetWord( m_pTargetText->GetWordAt( SentenceID - 1, i ) );
        }
        TranslationWithScore.SetTargetPhrase( TargetNgram );
        LogProb = PhraseTranslationLogProb( SourceWordIDs, TargetWordIDs, m_pLexicon );
        TranslationWithScore.SetScore( Scores[ m ] );
        TranslationWithScore.SetLogProb( LogProb );

        //cout << Scores[ m ] << " " << LogProb << " " << TargetNgram << endl;
		
        // store now in list of all good translations for this source phrase
        int Location;
        if ( AllTranslations.Find( TranslationWithScore, Location ) ) {
          //cout << " translation was found" << endl;
          if ( TranslationWithScore.Score() < AllTranslations.GetAt( Location ).Score() ) {
            // alread stored, but current score is lower, i.e. better
            // therefore overwrite it
            //cout << " translation was found, but new score is smaller, i.e. overwrite " << endl;
            AllTranslations.Insert( TranslationWithScore, IsNew );
            //if ( IsNew ) { 
            //  cout << " new translation inserted " << endl;
            //}
          //} else {
            //cout << " but new score is larger, i.e. don't use " << endl;
          }
        } else {
          //if ( TranslationWithScore.Score() < 
          //cout << " translation was not found - insert" << endl;
          AllTranslations.Insert( TranslationWithScore, IsNew );
        }
      }
    }


    if ( CreateEdges ) {
      // create edges for the best translations for the last source phrase
      // need to retrieve position of former source phrase
      int StartPos = results[ n ].srcPosStart - 1;
      int EndPos = results[ n ].srcPosEnd;

      for ( int i = AllTranslations.MinIndex(); i <= AllTranslations.MaxIndex(); i++ ) {
        //cout << AllTranslations.GetAt( i ).Score() << " " 
        //     << AllTranslations.GetAt( i ).LogProb() << " " 
        //     << AllTranslations.GetAt( i ).TargetPhrase() << endl;
  
        CPhraseSearchItem PhraseItem = AllTranslations.GetAt( i );
        CNumericalArray< CWordIndex > &TargetWordIDs = PhraseItem.TargetWordIDs();
        TargetNgram = "";
        for ( int j = TargetWordIDs.MinIndex(); j <= TargetWordIDs.MaxIndex(); j++ ) {
          TargetNgram += (CString)" " + m_pVoc->GetWord( TargetWordIDs[ j ] );
        }
        //cout << TargetNgram << endl;

        //CPhraseSearchItem PhraseItem = AllTranslations.GetAt( i );

        //cout << "n=" << n << " add Edge for " << StartPos << " " << EndPos << endl;

        AddEdge( pLattice, StartPos, EndPos, 
        AllTranslations.GetAt( i ).Score(), 
        SourceWordIDs, PhraseItem.TargetWordIDs() );
      }

      // forget the translations of the former source phrase
      AllTranslations.Clear();

    } // if ( CreateEdges )


        //for ( int i = AllTranslations.MinIndex(); i <= AllTranslations.MaxIndex(); i++ ) {
        //cout << AllTranslations.GetAt( i ).Score() << " " << AllTranslations.GetAt( i ).TargetPhrase() << endl;
        //CPhraseSearchItem PhraseItem = AllTranslations.GetAt( i );
        //CNumericalArray< CWordIndex > &TargetWordIDs = PhraseItem.TargetWordIDs();
        //TargetNgram = "";
        //for ( int j = TargetWordIDs.MinIndex(); j <= TargetWordIDs.MaxIndex(); j++ ) {
          //TargetNgram += (CString)" " + m_pVoc->GetWord( TargetWordIDs[ j ] );
        //}
        //cout << TargetNgram << endl;
        //}

        //AddEdge( pLattice, StartPosPhrase - 1, EndPosPhrase, 
        // Scores[ n ], SourceWordIDs, TargetWordIDs.GetRow( n ) );
  }
  delete m_pPhraseAligner;
}


void
CPhraseSearch::FindTranslations( CSentence *pSentence, CMatrix< CWordIndex > &Translations )
{
  // pSentence is the sentence for which we want to find phrase translations
  // pLattice is the corresponding lattice, into which we insert the found 
  // phrase translation pairs as new edges

  // 1. Step: find occurences of all n-grams from the sentence in the corpus
  // 2. Step: for each occurence of a source phrase find one or more translations
  //          in this case use maximum lexical probability for j1..j2 -> i1..i2
  // 3. Step: merge translations from different occurences of the same 
  //          source n-gram into one sorted list and discard bad candidates

  int SentenceID, PosInSentence, StartPosPhrase, EndPosPhrase;

  // deal with one source phrase at a time
  // typically has many translations
  // need to sort them according to translation score
  // and generated edges only for the best ones
  CArray< CWordIndex > SourceWordIDs;
  CArray< CArray< CWordIndex > * > TargetWordIDs;
  CSortedArray< CPhraseSearchItem > AllTranslations;

  CString SourceNgram;
  CString TargetNgram;  // the new one
  //CArray< CString > TargetNgrams;  // those seen before

  //map< double, CArray< CWordIndex > > SortedTargetWordIDs;
  // which keep each translation only once, but with best score
  //hash_map< char *, double > TranslationsWithScores;

  CPhraseSearchItem TranslationWithScore;

  CLatticeNode *pFromNode;
  CLatticeNode *pToNode;

  double BestScore;  // score for best traslation for one source phrase

  // convert sentence into char*;
  pSentence->Display( m_pVoc );
  cout << endl;
  ostrstream StrStream;
  pSentence->Write( StrStream, m_pVoc );
  StrStream << '\0';
  //cout << StrStream.str() << endl;
  //StrStream << "\0";
  //cout << StrStream.str() << endl;
  //cout << "Number of words in CPhraseSearch::Search is " << pSentence->Length() << endl;

  // init the phrase aligner with the lexicon
  m_pPhraseAligner = new CPhraseAlignerIBM1;

  if ( m_pReverseLexicon ) {
    m_pPhraseAligner->Init( m_pLexicon, m_pReverseLexicon );
    m_pPhraseAligner->SetInterpolationFactor( m_LambdaLexicon );
    m_pPhraseAligner->SetAlignBidirectional( true );
  } else {
    m_pPhraseAligner->Init( m_pLexicon, NULL );
  }

  // 1. Step: find locations of phrases from sentence in corpus
  vector< S_phraseLocationElement > results;
  m_pSA->setParameters( 1000000, 100, pSentence->Length() );
  results = m_pSA->findPhrasesInASentence( StrStream.str() );

  // 2. Step: find translations
  // for one source phrase store start and end position of
  // target phrases and the translation scores in the following arrays
  CNumericalArray< int > Min_i;
  CNumericalArray< int > Max_i;
  CNumericalArray< double > Scores;

/*
  for ( int n = 0; n < results.size(); n++ ) {
    StartPosPhrase = results[ n ].srcPosStart;
    EndPosPhrase = results[ n ].srcPosEnd ;
    SentenceID = results[ n ].sentId;
    PosInSentence = results[n].posInSent;

    cout << StartPosPhrase << " "
         << EndPosPhrase << " "
         << SentenceID << " "
         << PosInSentence << endl;
  }

*/
  for ( int n = 0; n < results.size(); n++ ) {
    StartPosPhrase = results[ n ].srcPosStart;
    EndPosPhrase = results[ n ].srcPosEnd ;
    SentenceID = results[ n ].sentId;
    PosInSentence = results[n].posInSent;

/*
    cout << endl << n <<  " Found in Corpus: " 
	 << SentenceID << " "
	 << StartPosPhrase << " "
	 << EndPosPhrase << " "
	 << PosInSentence << endl << endl;
*/

    bool IsLast = ( n == results.size() - 1 );

    bool NextIsNew = ( ( n + 1 < results.size() )
                       && ( (StartPosPhrase != results[ n + 1 ].srcPosStart) 
                            || (EndPosPhrase != results[ n + 1 ].srcPosEnd)
                          )
                     );

    bool IsNewSource = ( (n == 0)
                         || (StartPosPhrase != results[ n - 1 ].srcPosStart)
                         || (EndPosPhrase   != results[ n - 1 ].srcPosEnd) );

    bool CreateEdges = ( IsLast || NextIsNew );

/*
    cout << "NextIsNew: " << NextIsNew
         << " IsLast: " << IsLast
         << " IsNew: " << IsNewSource
         << " Create Edges: " << CreateEdges << endl;
*/
    if ( IsNewSource ) {
      // collect new one
      //cout << endl << "NEW SOURCE PHRASE" << endl;
      SourceNgram = "";
      SourceWordIDs.Clear();
      for ( int k = StartPosPhrase - 1; k <= EndPosPhrase - 1; k++ ) {
        //cout << j << " " << i << " " << TextE.GetWord( SentenceID - 1, i ) << endl;
        SourceWordIDs.Append( pSentence->GetAt( k ) );   // sentence is 0-based
        SourceNgram += (CString)" " + pSentence->Word( k , m_pVoc );
      }

      // initialize best score for translations for new scourse phrase
      BestScore = 10000;

    }
    //cout << "Source phrase: " << SourceNgram << endl;

    CSentence *pSourceSentence = m_pSourceText->GetSentence( SentenceID - 1 );
    //pSourceSentence->Display( m_pVoc ); cout << endl;
    //cout << "Lenght sentence is: " << pS->Length() << endl;
    CSentence *pTargetSentence = m_pTargetText->GetSentence( SentenceID - 1 );
    //pTargetSentence->Display( m_pVoc ); cout << endl;
	
    // now we want to find the best translations for the source phrase
    // in the target sentence by doing an IBM1 style alignment
    // there will be several target phrases for each source phrase
    // sorted according to alignment score

    int StartSource = PosInSentence;
    int EndSource   = PosInSentence + ( EndPosPhrase - StartPosPhrase );

    //cout << "call phrase aligner" << endl;

    m_pPhraseAligner->AlignPhrase( pSourceSentence, pTargetSentence,
                                   StartSource, EndSource,
                                   Min_i, Max_i, Scores  );

/*
    cout << "Found results: " << endl;
    for ( int m = Min_i.MinIndex(); m <= Min_i.MaxIndex(); m++ ) {
      cout << Min_i[ m ] << " " << Max_i[ m ] << " " << Scores[ m ] << endl;
    }
*/

    double LogProb;
    bool IsNew;

    for ( int m = Min_i.MinIndex(); m <= Min_i.MaxIndex(); m++ ) {

      //cout << "Deal with translation candidate " << m << " score = " << Scores[ m ] << endl;
      //cout << "Current best score = " << BestScore << endl;

      // some initial pruning: keep only reasonable good translations
      if ( Scores[ m ] < BestScore + 3.0 ) {

        //cout << "score is within beam" << endl;

        if ( Scores[ m ] < BestScore ) {
          BestScore = Scores[ m ];
          //cout << "new best score = " << BestScore << endl;
        }

        TargetNgram = "";
        CNumericalArray< CWordIndex > &TargetWordIDs = TranslationWithScore.TargetWordIDs();
        TargetWordIDs.Clear();

        for ( int i = Min_i[ m ]; i <= Max_i[ m ]; i++ ) {
          //cout << TextE.GetWord( SentenceID - 1, i ) << " ";
          TargetWordIDs.Append( m_pTargetText->GetWordAt( SentenceID - 1, i ) );
          TargetNgram += (CString)" " + m_pVoc->GetWord( m_pTargetText->GetWordAt( SentenceID - 1, i ) );
        }
        TranslationWithScore.SetTargetPhrase( TargetNgram );
        LogProb = PhraseTranslationLogProb( SourceWordIDs, TargetWordIDs, m_pLexicon );
        TranslationWithScore.SetScore( Scores[ m ] );
        TranslationWithScore.SetLogProb( LogProb );

        //cout << Scores[ m ] << " " << LogProb << " " << TargetNgram << endl;
		
        // store now in list of all good translations for this source phrase
        int Location;
        if ( AllTranslations.Find( TranslationWithScore, Location ) ) {
          //cout << " translation was found" << endl;
          if ( TranslationWithScore.Score() < AllTranslations.GetAt( Location ).Score() ) {
            // alread stored, but current score is lower, i.e. better
            // therefore overwrite it
            //cout << " translation was found, but new score is smaller, i.e. overwrite " << endl;
            AllTranslations.Insert( TranslationWithScore, IsNew );
            //if ( IsNew ) { 
            //  cout << " new translation inserted " << endl;
            //}
          //} else {
            //cout << " but new score is larger, i.e. don't use " << endl;
          }
        } else {
          //if ( TranslationWithScore.Score() < 
          //cout << " translation was not found - insert" << endl;
          AllTranslations.Insert( TranslationWithScore, IsNew );
        }
      }
    }


    if ( CreateEdges ) {
	  cout << "Create Translations" << endl;

      // create edges for the best translations for the last source phrase
      // need to retrieve position of former source phrase
      int StartPos = results[ n ].srcPosStart - 1;
      int EndPos = results[ n ].srcPosEnd;

	  // find best score for pruning;
	  double BestScore = 10000.0;
      for ( int i = AllTranslations.MinIndex(); i <= AllTranslations.MaxIndex(); i++ ) {
		if ( BestScore > AllTranslations.GetAt( i ).LogProb() ){
		  BestScore = AllTranslations.GetAt( i ).LogProb();
		}
	  }

	  int m = 0;
      for ( int i = AllTranslations.MinIndex(); i <= AllTranslations.MaxIndex(); i++ ) {
        cout << AllTranslations.GetAt( i ).Score() << " " 
             << AllTranslations.GetAt( i ).LogProb() << " " 
             << AllTranslations.GetAt( i ).TargetPhrase() << endl;
  
		if ( AllTranslations.GetAt( i ).LogProb() < BestScore + 3.0 ) {
		  cout << "keep, store at " << m << endl;
		  CPhraseSearchItem PhraseItem = AllTranslations.GetAt( i );
		  CNumericalArray< CWordIndex > &TargetWordIDs = PhraseItem.TargetWordIDs();
		  //TargetNgram = "";
		  for ( int j = TargetWordIDs.MinIndex(); j <= TargetWordIDs.MaxIndex(); j++ ) {
		    TargetNgram += (CString)" " + m_pVoc->GetWord( TargetWordIDs[ j ] );
			Translations.SetAt( m, j, TargetWordIDs[ j ] );
		  }
		  m++;
		}
        //cout << TargetNgram << endl;

        //CPhraseSearchItem PhraseItem = AllTranslations.GetAt( i );

        //cout << "n=" << n << " add Edge for " << StartPos << " " << EndPos << endl;

        //AddEdge( pLattice, StartPos, EndPos, 
        //AllTranslations.GetAt( i ).Score(), 
        //SourceWordIDs, PhraseItem.TargetWordIDs() );
      }

      // forget the translations of the former source phrase
      AllTranslations.Clear();

    } // if ( CreateEdges )


        //for ( int i = AllTranslations.MinIndex(); i <= AllTranslations.MaxIndex(); i++ ) {
        //cout << AllTranslations.GetAt( i ).Score() << " " << AllTranslations.GetAt( i ).TargetPhrase() << endl;
        //CPhraseSearchItem PhraseItem = AllTranslations.GetAt( i );
        //CNumericalArray< CWordIndex > &TargetWordIDs = PhraseItem.TargetWordIDs();
        //TargetNgram = "";
        //for ( int j = TargetWordIDs.MinIndex(); j <= TargetWordIDs.MaxIndex(); j++ ) {
          //TargetNgram += (CString)" " + m_pVoc->GetWord( TargetWordIDs[ j ] );
        //}
        //cout << TargetNgram << endl;
        //}

        //AddEdge( pLattice, StartPosPhrase - 1, EndPosPhrase, 
        // Scores[ n ], SourceWordIDs, TargetWordIDs.GetRow( n ) );
  }
}



void
CPhraseSearch::Search( CSentence *pSentence, CLatticeLM *pLattice )
{

  // convert sentence into char*;
  pSentence->Display( m_pVoc );
  cout << endl;
  ostrstream StrStream;
  pSentence->Write( StrStream, m_pVoc );
  StrStream << '\0';
  cout << StrStream.str() << endl;
  //StrStream << "\0";
  //cout << StrStream.str() << endl;
  cout << "Number of words in CPhraseSearch::Search is " << pSentence->Length() << endl;

  vector< S_phraseLocationElement > results;
  results = m_pSA->findPhrasesInASentence( StrStream.str() );

  //  return;

  // output of the list of ngram occurences

  /*
  if(results.size()==0){
	cout<<"Nothing found!"<<endl;
  } else {
	for ( int n = 0; n < results.size(); n++ ) {
	  cout << results[ n ].srcPosStart << " "
		   << results[ n ].srcPosEnd << " "
		   << results[ n ].sentId << " "
		   << results[ n ].posInSent << endl;
	}
  }
  */

  int SentenceID, PosInSentence, StartPosPhrase, EndPosPhrase;
  CArray< CWordIndex > SourceWordIDs;
  CMatrix< CWordIndex > TargetWordIDs;

  // to detect dublicates
  CString SourceNgram;
  CString TargetNgram;  // the new one
  CArray< CString > TargetNgrams;  // those seen before

  CLatticeNode *pFromNode;
  CLatticeNode *pToNode;

  int m = 0;  // to count uniq source ngrams
  int l;
  int t = 0;  // to count uniq target ngrams


  // loop over ngram occurences
  for ( int n = 0; n < results.size(); n++ ) {
	StartPosPhrase = results[ n ].srcPosStart;
	EndPosPhrase = results[ n ].srcPosEnd ;
	SentenceID = results[ n ].sentId;
	PosInSentence = results[n].posInSent;

	cout << endl << StartPosPhrase << " "
		 << EndPosPhrase << " "
		 << SentenceID << " "
		 << PosInSentence << endl;

        CSentence *pS = m_pSourceText->GetSentence( SentenceID - 1 );
        pS->Display( m_pVoc ); cout << endl;
        //cout << "Lenght sentence is: " << pS->Length() << endl;
        CSentence *pT = m_pTargetText->GetSentence( SentenceID - 1 );
        pT->Display( m_pVoc ); cout << endl;

	// first the source phrase: is it a new one or can we reuse from the previous loop
	if ( (n == 0) 
		 || (StartPosPhrase != results[ n - 1 ].srcPosStart + 1) 
		 || (EndPosPhrase   != results[ n - 1 ].srcPosEnd + 1)
		 ) {

	  //cout << "New source ngram" << endl;
	  
	  SourceWordIDs.Clear();
	  TargetWordIDs.Clear();
	  SourceNgram = "";
	  TargetNgrams.Clear();
	  t = 0;

	  // find nodes in lattice
	  //cout << "FromNode = " << StartPosPhrase
	  //	 << "  ToNode = " << EndPosPhrase << endl;
	  pFromNode = pLattice->Node( StartPosPhrase );
	  pToNode   = pLattice->Node( EndPosPhrase );
	  //pFromNode->Display();
	  //pToNode->Display();

	  m++;
	  for ( int k = StartPosPhrase; k <= EndPosPhrase; k++ ) {
		//cout << j << " " << i << " " << TextE.GetWord( SentenceID - 1, i ) << endl;
		SourceWordIDs.Append( pSentence->GetAt( k ) );   // sentence is 0-based
		SourceNgram += (CString)" " + pSentence->Word( k , m_pVoc );
	  }
	  cout << "Source phrase: " << SourceNgram << endl;

	} // is it a new source phrase


	// now find the target phrase, using the alignment
	int min_i = 1000;
	int max_i = 0;
	int i;
	int j = PosInSentence;
	float LogProb;

	CArray< short > SentenceAlignment = m_pAlignment->GetSimpleAlignment( SentenceID - 1);
        //cout << "Length alignment is: " << SentenceAlignment.Size() << endl;
        cout << SentenceAlignment << endl;

	for ( int k = 0; k <= EndPosPhrase - StartPosPhrase; k++ ) {
	  //	  i = m_pAlignment->Get( SentenceID - 1, j - 1 );
	  i = SentenceAlignment[ PosInSentence + k ];
	  cout << PosInSentence + k << " " << i << " " 
               <<  m_pVoc->GetWord( m_pTargetText->GetWordAt( SentenceID - 1, i ) ) << endl;

	  if ( i < min_i ) {
		min_i = i;
	  }
	  if ( i > max_i ) {
		max_i = i;
	  }
	  j++;

	}

        cout << "align to " << min_i << " " << max_i << endl;

	// if there is a mismatch in phrase length, we don't keep it
	if ( 
		( ( max_i - min_i + 1 ) > ( 1.5 * ( EndPosPhrase - StartPosPhrase + 1 ) + 1 ) )
		|| ( ( EndPosPhrase - StartPosPhrase + 1 ) > ( 1.5 * ( max_i - min_i + 1 ) + 1 ) ) ) {
	  cout << "Phrase lenght mismatch: " << EndPosPhrase - StartPosPhrase + 1 
               << " " << max_i - min_i + 1 << endl;
	} else {

	  // collect the words for the target phrase
	  
	  TargetNgram = "";
	  TargetWordIDs.GetRow( m ).Clear();
	  
	  for ( i = min_i; i <= max_i; i++ ) {
		//cout << TextE.GetWord( SentenceID - 1, i ) << " ";
		TargetWordIDs.GetRow( m ).Append( m_pTargetText->GetWordAt( SentenceID - 1, i ) );
		TargetNgram += (CString)" " + m_pVoc->GetWord( m_pTargetText->GetWordAt( SentenceID - 1, i ) );
	  }
	  
	  cout << "TargetNgram " << TargetNgram << endl;
	  
	  // is it new, then add it to list of target phrases and create lattice edge
	  int p = TargetNgrams.MinIndex();
	  bool NotFound = True;
	  while ( NotFound && (p <= TargetNgrams.MaxIndex()) ) {
		//cout << TargetNgrams[ p ] << " " << TargetNgram << endl;
		if ( TargetNgrams[ p ] == TargetNgram ) {
		  NotFound = false;
		}
		p++;
	  }
	  
	  if ( NotFound ) {
		//cout << "this is a new target ngram " << t << endl;
		TargetNgrams[ t ] = TargetNgram;
		
		CLatticeEdgeTrans *pEdge = new CLatticeEdgeTrans( pFromNode, pToNode );
		CWordInfo *pWordInfo = pEdge->WordInfo();
		
		for ( int j = SourceWordIDs.MinIndex(); j <= SourceWordIDs.MaxIndex(); j++ ) {
		  //cout << m_pVoc->GetWord( SourceWordIDs[ j ] ) << " ";
		  pWordInfo->AddSourceWord( m_pVoc->GetWord( SourceWordIDs[ j ] ) );
		}
		//cout << " # ";
		
		for ( int i = TargetWordIDs.GetRow( m ).MinIndex(); i <= TargetWordIDs.GetRow( m ).MaxIndex(); i++ ) {
		  //cout << m_pVoc->GetWord( TargetWordIDs.GetRow( m ).GetAt( i ) ) << " ";
		  pWordInfo->AddTargetWord( m_pVoc->GetWord( TargetWordIDs.GetRow( m ).GetAt( i ) ) );
		}
		//cout << " # ";
		
		
		if ( m_pLexicon ) {
		  LogProb = PhraseTranslationLogProb( SourceWordIDs, TargetWordIDs.GetRow( m ), m_pLexicon );
		} else {
		  LogProb = 0.0;
		}

		//cout << "LogProb: " << LogProb << endl;
		pWordInfo->Score() = LogProb;
		
		cout << *pWordInfo << endl;
		
		// now the edge
		pEdge->SetWord( "@PESA" );
		pEdge->Display();
		pLattice->AddEdgeUniq( pEdge );
		
		t++;
	  }
	}

	//TargetWordIDs.Clear();

	//cout << StartPosPhrase << "-" << EndPosPhrase << ": " << SentenceID << ", " << PosInSentence << " ";
	//for ( int j = StartPosPhrase; j <= EndPosPhrase; j++ ) {
	//cout << SourceWords[ j ] << " ";
	//}
	//cout << endl;
	//TextF.DisplaySentence( SentenceID - 1); cout << endl;
	//TextE.DisplaySentence( SentenceID - 1); cout << endl;
	//Alignment.DisplaySimple( SentenceID - 1 );

	//	for ( i = min_i; i <= max_i; i++ ) {
	  //cout << TextE.GetWord( SentenceID - 1, i ) << " ";
	  //TargetWordIDs[ m ].Append( TextE.GetWordAt( SentenceID - 1, i ) );
	//}
	//cout << endl;

  }

  pLattice->Display();

  /*
  while( !cin.eof() ) {
	vector< S_phraseLocationElement > results;
	cin.getline( tmpString, 1000, '\n' );
	results = SA.findPhrasesInASentence( tmpString );

	if(results.size()==0){
	  cout<<"Nothing found!"<<endl;
	}

	for(int i=0;i<results.size();i++){
	  cout<<results[i].srcPosStart<<"--"<<results[i].srcPosEnd<<"-->"<<results[i].sentId<<" "<<results[i].posInSent<<endl;
	}
	cout<<endl;
  }
  */


}


void
CPhraseSearch::Search1( CSentence *pSentence, CLatticeLM *pLattice )
{

  // sentence is given to suffix array based phrase finder
  // which returns position and the length of matches
  // these are sorted according to length and start positions of source phrase

  // for each match we get the source phrase (just for debugging)
  // and then we find one or several translations
  // for each translation pair we calculate the translation probability
  // generate a new edge and insert it into the lattice


  // convert sentence into char*;
  pSentence->Display( m_pVoc );
  cout << endl;
  ostrstream StrStream;
  pSentence->Write( StrStream, m_pVoc );
  StrStream << '\0';
  //cout << StrStream.str() << endl;
  //StrStream << "\0";
  //cout << StrStream.str() << endl;
  cout << "Number of words in CPhraseSearch::Search is " << pSentence->Length() << endl;

  vector< S_phraseLocationElement > results;
  results = m_pSA->findPhrasesInASentence( StrStream.str() );

  //  return;

  // output of the list of ngram occurences

  
  if(results.size()==0){
	cout<<"Nothing found!"<<endl;
  } else {
	for ( int n = 0; n < results.size(); n++ ) {
	  cout << results[ n ].srcPosStart << " "
		   << results[ n ].srcPosEnd << " "
		   << results[ n ].sentId << " "
		   << results[ n ].posInSent << endl;
	}
  }
  

  //return;

  int SentenceID, PosInSentence, StartPosPhrase, EndPosPhrase;
  CArray< CWordIndex > SourceWordIDs;
  CMatrix< CWordIndex > TargetWordIDs;

  // to detect dublicates
  CString SourceNgram;

  CLatticeNode *pFromNode;
  CLatticeNode *pToNode;

  cout << "get the source phrases from the corpus" << endl;


  int m = 0;  // to count uniq source ngrams
  int l;


  // loop over ngram occurences
  for ( int n = 0; n < results.size(); n++ ) {

	cout << results[ n ].srcPosStart << " "
		 << results[ n ].srcPosEnd << " "
		 << results[ n ].sentId << " "
		 << results[ n ].posInSent << endl;

	StartPosPhrase = results[ n ].srcPosStart;
	EndPosPhrase = results[ n ].srcPosEnd;
	SentenceID = results[ n ].sentId;
	PosInSentence = results[n].posInSent;

	// first the source phrase: is it a new one or can we reuse from the previous loop
	if ( (n == 0) 
		 || (StartPosPhrase != results[ n - 1 ].srcPosStart) 
		 || (EndPosPhrase   != results[ n - 1 ].srcPosEnd)
		 ) {

	  cout << "New source ngram" << endl;
	  
	  SourceWordIDs.Clear();
	  TargetWordIDs.Clear();
	  SourceNgram = "";
	  //TargetNgrams.Clear();
	  //t = 0;

	  // find nodes in lattice
	  cout << "FromNode = " << StartPosPhrase - 1
	  	 << "  ToNode = " << EndPosPhrase << endl;
	  pFromNode = pLattice->Node( StartPosPhrase - 1 );
	  pToNode   = pLattice->Node( EndPosPhrase );
	  //pFromNode->Display();
	  //pToNode->Display();

	  m++;
	  for ( int k = StartPosPhrase; k <= EndPosPhrase; k++ ) {
		cout << pSentence->Word( k - 1, m_pVoc ) << endl;
		SourceWordIDs.Append( pSentence->GetAt( k - 1) );   // sentence is 0-based
		SourceNgram += (CString)" " + pSentence->Word( k - 1 );
	  }
	  cout << SourceNgram << endl;

	} // is it a new source phrase

	FindTranslation( SentenceID, PosInSentence, TargetWordIDs );
  }
}


void
CPhraseSearch::FindTranslation( int SentenceID, int PosInSentence, CMatrix< CWordIndex > TargetWordIDs )
{

  /*
  CString TargetNgram;  // the new one
  CArray< CString > TargetNgrams;  // those seen before
  int t = 0;  // to count target ngrams


  // now find the target phrase, using the alignment
  int min_i = 1000;
  int max_i = 0;
  int i;
  short int j = PosInSentence;
  CArray< short > SentenceAlignment = m_pAlignment->GetSimpleAlignment( SentenceID - 1 );

  for ( int k = StartPosPhrase; k <= EndPosPhrase; k++ ) {
	i = SentenceAlignment[ k ];
	cout << j << " " << i << " " << m_pTargetText->GetWordAt( SentenceID - 1, i ) << endl;
	
	if ( i < min_i ) {
	  min_i = i;
	}
	if ( i > max_i ) {
	  max_i = i;
	}
	j++;
	
  }

  // if there is a mismatch in phrase length, we don't keep it
  if ( 
	  ( ( max_i - min_i ) > ( 1.5 * ( EndPosPhrase - StartPosPhrase ) + 1 ) )
	  || ( ( EndPosPhrase - StartPosPhrase ) > ( 1.2 * ( max_i - min_i ) + 1 ) ) ) {
	;
  } else {
	
	// collect the words for the target phrase
	
	TargetNgram = "";
	TargetWordIDs.GetRow( m ).Clear();
	
	for ( i = min_i; i <= max_i; i++ ) {
	  //cout << TextE.GetWord( SentenceID - 1, i ) << " ";
	  TargetWordIDs.GetRow( m ).Append( m_pTargetText->GetWordAt( SentenceID - 1, i ) );
	  TargetNgram += (CString)" " + m_pTargetText->GetWordAt( SentenceID - 1, i );
	}
	  
	//cout << "TargetNgram " << TargetNgram << endl;
	  
	// is it new, then add it to list of target phrases and create lattice edge
	int p = TargetNgrams.MinIndex();
	bool NotFound = True;
	while ( NotFound && (p <= TargetNgrams.MaxIndex()) ) {
	  //cout << TargetNgrams[ p ] << " " << TargetNgram << endl;
	  if ( TargetNgrams[ p ] == TargetNgram ) {
		NotFound = false;
	  }
	  p++;
	}
	  
	if ( NotFound ) {
	  //cout << "this is a new target ngram " << t << endl;
	  TargetNgrams[ t ] = TargetNgram;
	  
	  CLatticeEdgeTrans *pEdge = new CLatticeEdgeTrans( pFromNode, pToNode );
	  CWordInfo *pWordInfo = pEdge->WordInfo();
	  
	  for ( int j = SourceWordIDs.MinIndex(); j <= SourceWordIDs.MaxIndex(); j++ ) {
		//cout << m_pVoc->GetWord( SourceWordIDs[ j ] ) << " ";
		pWordInfo->AddSourceWord( m_pVoc->GetWord( SourceWordIDs[ j ] ) );
	  }
	  //cout << " # ";
	  
	  for ( int i = TargetWordIDs.GetRow( m ).MinIndex(); i <= TargetWordIDs.GetRow( m ).MaxIndex(); i++ ) {
		//cout << m_pVoc->GetWord( TargetWordIDs.GetRow( m ).GetAt( i ) ) << " ";
		pWordInfo->AddTargetWord( m_pVoc->GetWord( TargetWordIDs.GetRow( m ).GetAt( i ) ) );
	  }
	  //cout << " # ";
	  
	  
	  float LogProb = PhraseTranslationLogProb( SourceWordIDs, TargetWordIDs.GetRow( m ), m_pLexicon );
	  //cout << "LogProb: " << LogProb << endl;
	  pWordInfo->Score() = LogProb;
	  
	  cout << *pWordInfo << endl;
	  
	  // now the edge
	  //pEdge->SetWord( "@HMM_Phrase" );
	  //pEdge->Display();
	  //pLattice->AddEdge( pEdge );
	  
	  t++;
	}
  }

  */  
  //TargetWordIDs.Clear();
  
  //cout << StartPosPhrase << "-" << EndPosPhrase << ": " << SentenceID << ", " << PosInSentence << " ";
  //for ( int j = StartPosPhrase; j <= EndPosPhrase; j++ ) {
  //cout << SourceWords[ j ] << " ";
  //}
  //cout << endl;
  //TextF.DisplaySentence( SentenceID - 1); cout << endl;
  //TextE.DisplaySentence( SentenceID - 1); cout << endl;
  //Alignment.DisplaySimple( SentenceID - 1 );
  
  //	for ( i = min_i; i <= max_i; i++ ) {
  //cout << TextE.GetWord( SentenceID - 1, i ) << " ";
  //TargetWordIDs[ m ].Append( TextE.GetWordAt( SentenceID - 1, i ) );
  //}
  //cout << endl;
  


  /*
  while( !cin.eof() ) {
  vector< S_phraseLocationElement > results;
  cin.getline( tmpString, 1000, '\n' );
  results = SA.findPhrasesInASentence( tmpString );
  
  if(results.size()==0){
  cout<<"Nothing found!"<<endl;
  }
  
  for(int i=0;i<results.size();i++){
  cout<<results[i].srcPosStart<<"--"<<results[i].srcPosEnd<<"-->"<<results[i].sentId<<" "<<results[i].posInSent<<endl;
  }
  cout<<endl;
  }
*/


}



float
CPhraseSearch::PhraseTranslationLogProb( const CArray< CWordIndex > &Source, 
										 const CArray< CWordIndex > &Target, 
										 const CLexicon *pLexicon )
{
  float Prob = 1.0;
  float SumProb = 0.0;
  float LogProb = 0.0;

  //cout << "PhraseTranslationLogProb " <<  Source.Size() << " " << Target.Size() << endl;
  for ( int j = Source.MinIndex(); j <= Source.MaxIndex(); j++ ) {
	SumProb = 0.0;
	for ( int i = Target.MinIndex(); i <= Target.MaxIndex(); i++ ) {
	  //cout << Source[ j ] << " " <<  Target[ i ] << " "
	  //	   << Lexicon.Pr( Source[ j ], Target[ i ] ) << endl;
	  SumProb += pLexicon->Pr( Source[ j ], Target[ i ] );
	}
	Prob *= SumProb;
	//cout << SumProb << endl;
  }
  //cout << Prob << " " << -log( Prob ) << endl;
  return -log( Prob );
}



void
CPhraseSearch::AddEdge( CLattice *pLattice, int From, int To, double LogProb,
						const CArray< CWordIndex > &SourceWordIDs,
						const CArray< CWordIndex > &TargetWordIDs )
{
  CLatticeNode *pFromNode = pLattice->Node( From );
  CLatticeNode *pToNode   = pLattice->Node( To );

  CLatticeEdgeTrans *pEdge = new CLatticeEdgeTrans( pFromNode, pToNode );
  CWordInfo *pWordInfo = pEdge->WordInfo();
		
  for ( int j = SourceWordIDs.MinIndex(); j <= SourceWordIDs.MaxIndex(); j++ ) {
	//cout << m_pVoc->GetWord( SourceWordIDs[ j ] ) << " ";
	pWordInfo->AddSourceWord( m_pVoc->GetWord( SourceWordIDs[ j ] ) );
  }
  //cout << " # ";
  
  for ( int i = TargetWordIDs.MinIndex(); i <= TargetWordIDs.MaxIndex(); i++ ) {
	//cout << m_pVoc->GetWord( TargetWordIDs.GetRow( m ).GetAt( i ) ) << " ";
	pWordInfo->AddTargetWord( m_pVoc->GetWord( TargetWordIDs[ i ] ) );
  }
  //cout << " # ";
  
  
//  if ( m_pLexicon ) {
//LogProb = PhraseTranslationLogProb( SourceWordIDs, TargetWordIDs.GetRow( m ), m_pLexicon );
//} else {
//LogProb = 0.0;
//}
  
  //cout << "LogProb: " << LogProb << endl;
  pWordInfo->Score() = LogProb;
  
  //cout << *pWordInfo << endl;
  
  // now the edge
  pEdge->SetWord( "@PESA" );
  //pEdge->Display();
  pLattice->AddEdge( pEdge );
}
