#ifndef __PhraseSearch__
#define __PhraseSearch__

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <fortify.h>

#include <LatticeEdgeTrans.hh>
#include <Lattice.hh>
#include <LatticeLM.hh>

#include <Text.hh>
#include <Alignment.hh>
#include <Lexicon.hh>
#include <NumericalArray.hh>
#include <_SuffixArray.h>

#include <PhraseAlign.hh>

#ifdef HAVE_EXT_HASH_MAP
#include <ext/hash_map>
// Hey, what's that?! Seems to be overkill --db
namespace sgi {
  using __gnu_cxx::hash;
  using __gnu_cxx::hash_map;
}
#define SGI_HASH_NAMESPACE __gnu_cxx
#else
#include <hash_map>

namespace sgi {
     using std::hash;
     using std::hash_map;
}
#define SGI_HASH_NAMESPACE std
#endif

/*
struct eqstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) == 0;
  }
};
*/


class CPhraseSearchItem
{
public:
  CPhraseSearchItem();
  
  CNumericalArray< CWordIndex > & TargetWordIDs();
  double Score() const;
  void SetScore( double Score );
  double LogProb() const;
  void SetLogProb( double LogProb );
  CString TargetPhrase() const;
  void SetTargetPhrase( CString TargetPhrase );

  bool operator==( const CPhraseSearchItem & Item ) const;
  bool operator<( const CPhraseSearchItem & Item ) const;
  bool operator>( const CPhraseSearchItem & Item ) const;

private:
  CNumericalArray< CWordIndex > m_TargetWordIDs;
  CString m_TargetPhrase;
  double m_Score;
  double m_LogProb;
};

//####################################
//#
//#       usage: 1. construction Object
//#              2. call setPhraseAligner
//#              3. call init to Initialize
//#
//####################################

class CPhraseSearch
{
public:
  CPhraseSearch();
  ~CPhraseSearch();

  void Init( C_SuffixArray *pSA, CText *pTargetText, CVocabulary *pVoc, CAlignment *pAlignment, CLexicon *pLexicon );
  void Init( C_SuffixArray *pSA, CText *pSourceText, CText *pTargetText, CVocabulary *pVoc,
             CAlignment *pAlignment, CLexicon *pLexicon );
  void Init( C_SuffixArray *pSA, CText *pSourceText, CText *pTargetText, CVocabulary *pVoc,
             CAlignment *pAlignment, CLexicon *pLexicon, CLexicon *pReverseLexicon, double LambdaLexicon );

  void SearchAlign( CSentence *pSentence, CLattice *pLattice );
  
  //phrase search in training mode, used when translating a training
  //set, and the SA matches this same data
  void SearchAlignTraining( const int SentenceNumber, CSentence *pSentence, CLattice *pLattice );

  void FindTranslations( CSentence *pSentence, CMatrix< CWordIndex > &Translations );

  void Search( CSentence *pSentence, CLatticeLM *pLattice );
  void Search1( CSentence *pSentence, CLatticeLM *pLattice );


private:

  void FindTranslation( int SentenceID, int PosInSentence, CMatrix< CWordIndex > TargetWordIDs );


  float PhraseTranslationLogProb( const CArray< CWordIndex > &Source, 
								  const CArray< CWordIndex > &Target, 
								  const CLexicon *pLexicon );

  void AddEdge( CLattice *pLattice, int From, int To, double LogProb,
				const CArray< CWordIndex > &SourceWordIDs,
				const CArray< CWordIndex > &TargetWordIDs );

  C_SuffixArray *m_pSA;
  CText *m_pSourceText;
  CText *m_pTargetText;
  CVocabulary *m_pVoc;
  CAlignment *m_pAlignment;
  CLexicon *m_pLexicon;
  CLexicon *m_pReverseLexicon;
  double m_LambdaLexicon;

  CPhraseAlignerIBM1 *m_pPhraseAligner;
};


inline
CPhraseSearch::CPhraseSearch()
{
  m_pSA = NULL;
  m_pTargetText = NULL;
  m_pVoc = NULL;
  m_pAlignment = NULL;
  m_pLexicon = NULL;
  m_pReverseLexicon = NULL;
  m_LambdaLexicon = 0.0;
}


inline
CPhraseSearch::~CPhraseSearch()
{
}


inline void
CPhraseSearch::Init( C_SuffixArray *pSA, CText *pTargetText, CVocabulary *pVoc, CAlignment *pAlignment, CLexicon *pLexicon )
{
  m_pSA = pSA;
  m_pTargetText = pTargetText;
  m_pVoc = pVoc;
  m_pAlignment = pAlignment;
  m_pLexicon = pLexicon;
}


inline void
CPhraseSearch::Init( C_SuffixArray *pSA, CText *pSourceText, CText *pTargetText, CVocabulary *pVoc, CAlignment *pAlignment, CLexicon *pLexicon )
{
  m_pSA = pSA;
  m_pSourceText = pSourceText;
  m_pTargetText = pTargetText;
  m_pVoc = pVoc;
  m_pAlignment = pAlignment;
  m_pLexicon = pLexicon;
}

inline void
CPhraseSearch::Init( C_SuffixArray *pSA, CText *pSourceText, CText *pTargetText, CVocabulary *pVoc,
                     CAlignment *pAlignment, CLexicon *pLexicon, CLexicon *pReverseLexicon, double LambdaLexicon )
{
  m_pSA = pSA;
  m_pSourceText = pSourceText;
  m_pTargetText = pTargetText;
  m_pVoc = pVoc;
  m_pAlignment = pAlignment;
  m_pLexicon = pLexicon;
  m_pReverseLexicon = pReverseLexicon;
  m_LambdaLexicon = LambdaLexicon;
}



#endif
