#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <sstream>
#include <map>

// switch on/off tracing for this module here
#undef NO_TRACE
//#define NO_TRACE 1
#include <Trace.hh>

#include <Array.hh>
#include <PhraseSearchMS.hh>

using namespace std;



CPhraseSearchMS::CPhraseSearchMS()
{
  Trace( 1, "CPhraseSearchMS::CPhraseSearchMS" );

  m_pSA = NULL;
  m_pTargetText = NULL;
  m_pVoc = NULL;
  m_pAlignment = NULL;
  m_pLexicon = NULL;
  m_pReverseLexicon = NULL;
  m_pPhrasePairStats = NULL;


  m_LambdaLexicon = 0.0;

  m_UseEmptyWord = false;

  m_BeamFactor = 0.3;
  m_MaxTranslations = 10;

  m_pExistingTranslations = new CMatrix< short >;

  m_pPhraseTranslationTable = new CMatrix< CSortedArray< CPhraseAlignItemMS > >;
//  m_pPhraseAligner = new CPhraseAlignerMSIBM1;

  m_pSentenceInfo = NULL;

}



CPhraseSearchMS::~CPhraseSearchMS()
{
  delete m_pExistingTranslations;
//  delete m_pPhraseAligner;
}


void
CPhraseSearchMS::Init( C_SuffixArraySearch *pSA, CText *pTargetText, CVocabulary *pVoc,
                       CAlignment *pAlignment, CLexicon *pLexicon )
{
  m_pSA = pSA;
  m_pTargetText = pTargetText;
  m_pVoc = pVoc;
  m_pAlignment = pAlignment;
  m_pLexicon = pLexicon;
}



void
CPhraseSearchMS::Init( C_SuffixArraySearch *pSA, CText *pSourceText, CText *pTargetText, CVocabulary *pVoc,
                       CAlignment *pAlignment, CLexicon *pLexicon )
{
  m_pSA = pSA;
  m_pSourceText = pSourceText;
  m_pTargetText = pTargetText;
  m_pVoc = pVoc;
  m_pAlignment = pAlignment;
  m_pLexicon = pLexicon;
  m_pReverseLexicon = NULL;

  if(!m_pPhraseAligner) {
    cout << "First set up Phrase Aligner by calling setPhraseAligner" << endl;
    return;
  }

  //m_pPhraseAligner->Init( m_pLexicon, NULL );
  //m_pPhraseAligner->SetAlignBidirectional( false );

}



void
CPhraseSearchMS::Init( C_SuffixArraySearch *pSA, CText *pSourceText, CText *pTargetText, CVocabulary *pVoc,
                       CAlignment *pAlignment, CLexicon *pLexicon, CLexicon *pReverseLexicon, double LambdaLexicon )
{
  
  m_pSA = pSA;
  m_pSourceText = pSourceText;
  m_pTargetText = pTargetText;
  m_pVoc = pVoc;
  m_pAlignment = pAlignment;
  m_pLexicon = pLexicon;
  m_pReverseLexicon = pReverseLexicon;
  m_LambdaLexicon = LambdaLexicon;
  
  if(!m_pPhraseAligner) {
    cout << "First set up Phrase Aligner by calling setPhraseAligner" << endl;
    return;
  }

  if ( m_pReverseLexicon) {

    Trace( 0, "CPhraseSearchMS::Init with two lexicons" );
    //m_pPhraseAligner->Init( m_pLexicon, m_pReverseLexicon );
    //m_pPhraseAligner->SetInterpolationFactor( m_LambdaLexicon );
    //m_pPhraseAligner->SetAlignBidirectional( true );
  } else {
    Trace( 0, "CPhraseSearchMS::Init with one lexicons" );
    //m_pPhraseAligner->Init( m_pLexicon, NULL );
  }
}



void
CPhraseSearchMS::LoadSentenceInfo( const char *pFileName )
{
  if ( m_pSentenceInfo == NULL ) {
    m_pSentenceInfo = new CSentenceInfo;
  }

  m_pSentenceInfo->LoadLogProbPerWord( pFileName );

}



void
CPhraseSearchMS::SetPhrasePairStats( CPhrasePairStats *pPhrasePairStats )
{
  m_pPhrasePairStats = pPhrasePairStats;
}



void
CPhraseSearchMS::SetPhraseAligner( CPhraseAlignerMS *pPhraseAligner )
{
  m_pPhraseAligner = pPhraseAligner;
}



void
CPhraseSearchMS::SetUseEmptyWord( bool UseEmptyWord )
{
  m_UseEmptyWord = UseEmptyWord;
}


void
CPhraseSearchMS::SetMaxAlignments( int MaxAlignments )
{
  m_MaxAlignments = MaxAlignments;
}


void
CPhraseSearchMS::SetBeam( float BeamFactor, int MaxTranslations )
{
  m_BeamFactor = BeamFactor;
  m_MaxTranslations = MaxTranslations;

  //m_pPhraseAligner->SetBeam( BeamFactor, MaxTranslations );

  Trace( 0, "BeamFactor = " << m_BeamFactor << " MaxTranslations = " << m_MaxTranslations );
}



CMatrix< short > *
CPhraseSearchMS::GetExistingTranslationsTable()
{
  return m_pExistingTranslations;
}




void
CPhraseSearchMS::FindTranslations( CSentence *pSentence, CLattice *pLattice, int ForbiddenSentence )
{
  m_ForbiddenSentence = ForbiddenSentence;
  FindTranslations( pSentence, pLattice );
}



void
CPhraseSearchMS::FindTranslations( CSentence *pSentence, CLattice *pLattice )
{
  Trace( 0, "CPhraseSearchMS::FindTranslations" );
  //cout << "CPhraseSearchMS::FindTranslations" << endl;

  // pSentence is the sentence for which we want to find phrase translations
  // pLattice is the corresponding lattice, into which we insert the found 
  // phrase translation pairs as new edges

  // 1. Step: find occurences of all n-grams from the sentence in the corpus
  // 2. Step: for each occurence of a source phrase find one or more translations
  //          in this case use maximum lexical probability for j1..j2 -> i1..i2
  // 3. Step: merge translations from different occurences of the same 
  //          source n-gram into one sorted list and discard bad candidates


  m_pSentence = pSentence;
  m_pLattice = pLattice;

  m_pPhraseAligner->SetBeam( m_BeamFactor, m_MaxTranslations );

  Trace( 0, "BeamFactor = " << m_BeamFactor << " MaxTranslations = " << m_MaxTranslations );


  // 1. Step: find locations of phrases from sentence in corpus
  // convert sentence into char*;
  //#ifdef TRACE_ENABLED
  //pSentence->Display( m_pVoc );
  //cout << endl;
  //#endif
  ostringstream StrStream;
  pSentence->Write( StrStream, m_pVoc );
  StrStream << '\0';
  //cout << StrStream.str() << endl;
  
  string tmp = StrStream.str();
  //cout << "String for SA: " << tmp << endl;
  
  vector< S_phraseLocationElement > PhraseLocations;
  vector< S_phraseLocationElement > BestPhraseLocations;
  
  BestPhraseLocations = m_pSA->findPhrasesInASentence( (char *)tmp.c_str() );
  
  //DisplayPhraseLocations( PhraseLocations );

  //cout << "Find best Locations" << endl;
  //FindBestLocations( PhraseLocations, BestPhraseLocations );

  // Display the phrase locations
  //cout << PhraseLocations.size() << " phrase locations returned from suffix array search" << endl;
  //DisplayPhraseLocations( BestPhraseLocations );

  
  // 2. Step: find translations
  // for one source phrase store start and end position of
  // target phrases and the translation scores in the following arrays
  for ( int n = 0; n < BestPhraseLocations.size(); n++ ) {
    //DisplayPhraseLocation( BestPhraseLocation[ n ] );

    // search phrase translations only for segments for which
    // we don't have translations from transducer
    short StartPosPhrase = BestPhraseLocations[ n ].posStartInSrcSent;  // FromNode is before first word
    short EndPosPhrase   = BestPhraseLocations[ n ].posEndInSrcSent;
    long  SentenceID     = BestPhraseLocations[ n ].sentIdInCorpus;
    short PosInSentence  = BestPhraseLocations[ n ].posInSentInCorpus;

    //cout << endl << " Found in Corpus: " << SentenceID << " " << StartPosPhrase << " "
    //   << EndPosPhrase << " " << PosInSentence << endl << endl;
    //cout << BestPhraseLocations[ n ] << endl;

    // we skip over phrases for which we have already translations from an
    // existing phrase table
    // Note 1: FromNode is before first word, therefore StartPosPhrase - 1
    // Note 2: we ask for > 1 as we always have the source word
    if ( m_pExistingTranslations->GetAt( StartPosPhrase - 1, EndPosPhrase ) > 1 ) {
      ;
      //cout << "Translation for edge " << StartPosPhrase - 1 << " " << EndPosPhrase << " exists " << endl;
    } else {
      //cout << "Edge " << StartPosPhrase << " " << EndPosPhrase << " needs translation " << endl;
      AlignPhrase(  BestPhraseLocations[ n ] );
	  }
    
    //CSortedArray< CPhraseAlignItemMS > &AllPhraseTranslations = (*m_pPhraseTranslationTable)( StartPosPhrase, EndPosPhrase );
    //cout << "Number of phrase translations = " << AllPhraseTranslations.Size() << endl;

    /*
      if ( PhraseLocations[ n ].sentId != m_ForbiddenSentence ) {
      AlignPhrase(  PhraseLocations[ n ] );
      }
    */
  }

  // add for newly aligned phrase pairs the relative phrase frequencies p( SourcePhrase | TargetPhrase )
  // and p( TargetPhrase | SourcePhrase )
  // for now we approximate this based on co-occurrence counts
  //  AddRelativeFrequencies();
  
  // create edges for the found translations and add them to the lattice
  CreateEdges();

  // cout << "Find translations done" << endl;

}



void
CPhraseSearchMS::FindBestLocations( const vector< S_phraseLocationElement > &PhraseLocations,
                                    vector< S_phraseLocationElement > &BestPhraseLocations )
{
  cout << "CPhraseSearchMS::FindBestLocations" << endl;

  // for each source phrase store the locations sorted according to their alignment cost per word
  CMatrix< map< float, S_phraseLocationElement > > Locations;
  S_phraseLocationElement PhraseLocation;
  float AlignmentCostPerWord;

  for ( int n = 0; n < PhraseLocations.size(); n++ ) {
	S_phraseLocationElement PhraseLocation = PhraseLocations[ n ];
    int StartPosPhrase = PhraseLocation.posStartInSrcSent;
    int EndPosPhrase = PhraseLocation.posEndInSrcSent ;
    int SentenceID = PhraseLocation.sentIdInCorpus;
    int PosInSentence = PhraseLocation.posInSentInCorpus;
    /*
      cout << StartPosPhrase << " "
      << EndPosPhrase << " "
      << SentenceID << " "
      << PosInSentence << endl;
    */
    float AlignmentCostPerWord = m_pSentenceInfo->LogProbPerWord( SentenceID );
    //cout << "store in locations" << endl;
    Locations( StartPosPhrase, EndPosPhrase)[ AlignmentCostPerWord ] = PhraseLocation;
  }

  cout << "now, write back the best locations" << endl;

  // now, we write back the best locations
  for ( int StartPosPhrase = Locations.MinIndex1(); StartPosPhrase <= Locations.MaxIndex1(); StartPosPhrase++ ) {

    CArray< map< float, S_phraseLocationElement > > StartLocations = Locations.GetRow( StartPosPhrase );

    for ( int EndPosPhrase = StartLocations.MinIndex(); EndPosPhrase <= StartLocations.MaxIndex(); EndPosPhrase++ ) {
	  
      cout << "from " << StartPosPhrase << " to " << EndPosPhrase << endl;

      const map< float, S_phraseLocationElement > & PhraseLocations = StartLocations[ EndPosPhrase ];
      map< float, S_phraseLocationElement >::const_iterator Iter;
      int Count = 0;

      for ( Iter = PhraseLocations.begin(); Iter != PhraseLocations.end(); ++Iter ) {
        Count++;
        if ( Count <= m_MaxAlignments ) {
          AlignmentCostPerWord = Iter->first;
          PhraseLocation = Iter->second;
          cout << AlignmentCostPerWord << " " << PhraseLocation.posStartInSrcSent << " "
               << PhraseLocation.posEndInSrcSent << " "
               << PhraseLocation.sentIdInCorpus << endl;

          BestPhraseLocations.push_back( PhraseLocation );
        }
      }
	  
    }
  }

}


/*
  AlignPhrase find one or several target phrases for a given source phrase
  in a given sentence pair
*/

void
CPhraseSearchMS::AlignPhrase( S_phraseLocationElement &PhraseLocation )
{
  Trace( 1, "CPhraseSearchMS::AlignPhrase" );
  //cout <<  "CPhraseSearchMS::AlignPhrase" << endl;

  int SentenceID, PosInSentence, StartPosPhrase, EndPosPhrase;

  // deal with one source phrase at a time
  // typically has many translations
  // need to sort them according to translation score
  // and generated edges only for the best ones
  CArray< CWordIndex > SourceWordIDs;
  CArray< CArray< CWordIndex > * > TargetWordIDs;
  CSortedArray< CPhraseAlignItemMS > PhraseAlignItems;

  CString SourceNgram;
  CString TargetNgram;  // the new one
  //CArray< CString > TargetNgrams;  // those seen before

  //map< double, CArray< CWordIndex > > SortedTargetWordIDs;
  // which keep each translation only once, but with best score
  //hash_map< char *, double > TranslationsWithScores;

  //CPhraseAlignItemMS TranslationWithScore;

  CLatticeNode *pFromNode;
  CLatticeNode *pToNode;

  //double BestScore;  // score for best translation for one source phrase
  //double BestLogProb;

  StartPosPhrase = PhraseLocation.posStartInSrcSent;
  EndPosPhrase   = PhraseLocation.posEndInSrcSent ;
  SentenceID     = PhraseLocation.sentIdInCorpus;
  PosInSentence  = PhraseLocation.posInSentInCorpus;

  //cout << endl << " Found in Corpus: " << SentenceID << " " << StartPosPhrase << " "
  //   << EndPosPhrase << " " << PosInSentence << endl << endl;

  
  // collect source phrase
  SourceNgram = "";
  SourceWordIDs.Clear();
  for ( int k = StartPosPhrase; k <= EndPosPhrase; k++ ) {
    SourceWordIDs.Append( m_pSentence->GetAt( k ) );   // sentence is 0-based
    SourceNgram += CString( " " ) + m_pSentence->Word( k , m_pVoc );
  }
  
  // initialize best score for translations for new scourse phrase
  //BestScore = 10000.0;
  //BestLogProb = 10000.0;
  
  //cout << "Source phrase: " << SourceNgram << endl;
  
  // ??? why do we need SentenceID - 1
  CSentence *pSourceSentence = m_pSourceText->GetSentence( SentenceID - 1 );
  CSentence *pTargetSentence = m_pTargetText->GetSentence( SentenceID - 1 );
  
  if ( pTargetSentence->Size() == 0 ) {
    cerr << "PhraseSearchMS:AlignPhrase() - Target Sentence is Emtpy!" << endl;
    return;
  }



  //pSourceSentence->Display( m_pVoc ); cout << endl;
  //pTargetSentence->Display( m_pVoc ); cout << endl;
  
  // now we want to find the best translations for the source phrase
  // in the target sentence by doing an IBM1 style alignment
  // there will be several target phrases for each source phrase
  // sorted according to alignment score
  
  int StartSource = PosInSentence;
  int EndSource   = PosInSentence + ( EndPosPhrase - StartPosPhrase );
  
  //cout << "call phrase aligner" << endl;

  m_pPhraseAligner->SetUseEmptyWord( m_UseEmptyWord );
  m_pPhraseAligner->SetBeam( m_BeamFactor, m_MaxTranslations );
  // what about the other settings for PhraseAlign?

  // here the actual alignment, result is one or several target phrases, 
  // encoded in the PhraseAlignItems
  m_pPhraseAligner->AlignPhrase( pSourceSentence, pTargetSentence,
                                 StartSource, EndSource,
                                 PhraseAlignItems );

  /*
    cout << "Found PhraseAlignItem: (Min Index : " << PhraseAlignItems.MinIndex() << " ) ( Max index :"
    << PhraseAlignItems.MaxIndex() << ") " << endl;
    //cout << "Found results: " << endl;
    for ( int m = PhraseAlignItems.MinIndex(); m <= PhraseAlignItems.MaxIndex(); m++ ) {
    const CPhraseAlignItemMS &P = PhraseAlignItems.GetAt( m );
    cout << m << " " << P << endl;
    }
  */

  // loop over all results and find the best one, which will then be used in pruning
  /*
    for ( int m = PhraseAlignItems.MinIndex(); m <= PhraseAlignItems.MaxIndex(); m++ ) {
    const CPhraseAlignItemMS &P = PhraseAlignItems.GetAt( m );
    if ( P.Score() < BestScore ) {
    BestScore = P.Score();
    }
    }
  */


  // we store them in PhraseTranslationTable (sorting is not according to score but according to target phrase
  // i.e. we if we get the some target phrase again from a different sentence pair, then this matches and
  // we can compare the two translations)
  //cout << "add these results to the list of all phrase pairs for " << StartPosPhrase << " " << EndPosPhrase << endl;
  CSortedArray< CPhraseAlignItemMS > &AllPhraseTranslations = (*m_pPhraseTranslationTable)( StartPosPhrase, EndPosPhrase );
  int i = 0;
  int j = 0;
  for ( int m = PhraseAlignItems.MinIndex(); m <= PhraseAlignItems.MaxIndex(); m++ ) {

    const CPhraseAlignItemMS &AlignItem = PhraseAlignItems.GetAt( m );
    
    int Location;
    bool IsNew;
    if ( AllPhraseTranslations.Find( AlignItem, Location ) ) {

      // now we distinguish between to modes of operation
      // 1. store the best score
      // 2. average the scores from different occurrences

      if ( m_AccumulationMode = SUM ) {
        // add scores to previous scores, keep track of how often this translation has been seen
        // to divide later
        // not yet implemented
        ;

      } else {  // we take the MAX, this is the original mode
        //cout << '"' << AlignItem.TargetPhrase() << '"' << "   was found in global list" << endl;
        if ( AlignItem.ScaledScore() < AllPhraseTranslations.GetAt( Location ).ScaledScore() ) {
          // already stored, but current score is lower, i.e. better, therefore overwrite it
          //cout << '"' << AlignItem.TargetPhrase() << '"' << "   was found in global list, but new score " << AlignItem.ScaledScore()
          //    << " is smaller than old score " << AllPhraseTranslations.GetAt( Location ).ScaledScore()  << "  - overwrite " << endl;
          AllPhraseTranslations.Insert( AlignItem, IsNew );
          i++;
        } else {
          //cout << '"' << AlignItem.TargetPhrase() << '"' << "   was found in global list, but new score " << AlignItem.Score()
          // << " is larger than old score " << AllPhraseTranslations.GetAt( Location ).Score()  << "  - don't use " << endl;
        }

      }


    } else { // this is a new translation for the give source phrase, i.e. store it
      //cout << '"' << AlignItem.TargetPhrase() << '"' << "   was not found in global list - insert" << endl;
      AllPhraseTranslations.Insert( AlignItem, IsNew );
      j++;
    }
  }
  
  //cout << j << " new items inserted in TranslationTable"<< endl;
  //cout << i << " items inserted because of better score in TranslationTable"<< endl;
  

}



void
CPhraseSearchMS::AddRelativeFrequencies()
{
  cout << "Add reletive frequencies" << endl;

  CNumericalArray< CWordIndex > SourceWordIDs;
  CNumericalArray< CWordIndex > TargetWordIDs;
  

  for ( int StartPos = m_pPhraseTranslationTable->MinIndex1();
        StartPos <= m_pPhraseTranslationTable->MaxIndex1();
        StartPos++ ) {

    for ( int EndPos = m_pPhraseTranslationTable->GetRow( StartPos ).MinIndex(); 
          EndPos <= m_pPhraseTranslationTable->GetRow( StartPos ).MaxIndex();
          EndPos++ ) {
      //Trace( 0, "Number of entries in " << StartPos << " " << EndPos 
      //  << " = " << m_pPhraseTranslationTable->GetAt( StartPos, EndPos ).Size() );

      // get the source phrase
      for ( int k = StartPos; k <= EndPos; k++ ) {
        //cout << j << " " << i << " " << TextE.GetWord( SentenceID - 1, i ) << endl;
        SourceWordIDs.Append( m_pSentence->GetAt( k ) );   // sentence is 0-based
      }
      
      // now the target phrases
      CSortedArray< CPhraseAlignItemMS > &AllPhraseTranslations = (*m_pPhraseTranslationTable)( StartPos, EndPos );
      //cout << "Number of phrase translations = " << AllPhraseTranslations.Size() << endl;

      for ( int m = AllPhraseTranslations.MinIndex(); m <= AllPhraseTranslations.MaxIndex(); m++ ) {
        const CPhraseAlignItemMS &AlignItem = AllPhraseTranslations.GetAt( m );
      
        TargetWordIDs = ((CPhraseAlignItemMS)AlignItem).TargetWordIDs();

        // get the phrase counts
        CPhrasePairCounts PhrasePairCounts = m_pPhrasePairStats->GetPhrasePairCounts( &SourceWordIDs, &TargetWordIDs );

      }

    }

  }


}




void
CPhraseSearchMS::CreateEdges()
{
  //cout << "CreateEdges" << endl;

  Trace( 1, "Create Edges with BeamFactor = " << m_BeamFactor << " and MaxTranslations = " << m_MaxTranslations );

  Trace( 1, "Number of entries in PhraseTranslationTable" );
  Trace( 1, " " << m_pPhraseTranslationTable->MinIndex1() << " " << m_pPhraseTranslationTable->MaxIndex1() );

  
  for ( int i = m_pPhraseTranslationTable->MinIndex1(); i <= m_pPhraseTranslationTable->MaxIndex1(); i++ ) {
    for ( int j = m_pPhraseTranslationTable->GetRow( i ).MinIndex(); j <= m_pPhraseTranslationTable->GetRow( i ).MaxIndex(); j++ ) {
      Trace( 1, "Number of entries in " << i << " " << j << " = " << m_pPhraseTranslationTable->GetAt( i, j).Size() );
    }
  }

    
  // for any From-To pair we collect the translations into a sorted array
  // sorted according to overall score
  CSortedArray< CPhraseAlignItemMS > AllTranslations;

  // pruning is done using all individual scores, therefore collect the best scores
  CPhraseAlignItemMS ItemWithBestScores;

  // Pruning:
  // keep overall best hypothesis, ie the one with the best combined score
  // remove hypotheses if any one of its scores is larger than corresponding best score plus beam width

  double BestScore;

  CString TargetNgram;  // the new one
  CArray< CWordIndex > SourceWordIDs;
  int StartPos;
  int EndPos;
  int LengthSourcePhrase;

  // a map for sorting translations according to score
  multimap< double, int > SortedTranslations;

  // create edges for the best translations for the last source phrase
  // need to retrieve position of former source phrase
  //int StartPos = results[ n ].srcPosStart - 1;
  //int EndPos = results[ n ].srcPosEnd;

  //  for ( int StartPos = m_pSentence->MinIndex(); StartPos <= m_pSentence->MaxIndex(); StartPos++ ) {
  //for ( int EndPos = StartPos; EndPos <= m_pSentence->MaxIndex(); EndPos++ ) {

  for ( int StartPos = m_pPhraseTranslationTable->MinIndex1();
        StartPos <= m_pPhraseTranslationTable->MaxIndex1();
        StartPos++ ) {

    for ( int EndPos = m_pPhraseTranslationTable->GetRow( StartPos ).MinIndex(); 
          EndPos <= m_pPhraseTranslationTable->GetRow( StartPos ).MaxIndex();
          EndPos++ ) {
      //Trace( 0, "Number of entries in " << StartPos << " " << EndPos 
      //  << " = " << m_pPhraseTranslationTable->GetAt( StartPos, EndPos ).Size() );

      CSortedArray< CPhraseAlignItemMS > &AllPhraseTranslations = (*m_pPhraseTranslationTable)( StartPos, EndPos );
      //cout << "Number of phrase translations = " << AllPhraseTranslations.Size() << endl;


      // again, start node number is word position - 1
      Trace( 2, "Insert edges for nodes " << StartPos - 1 << " - " << EndPos );

      LengthSourcePhrase = EndPos - StartPos + 1;
      
      BestScore = 10000.0;

      // initialize item for storing best scores; set large initial scores

      for ( int i = 0; i <= m_pPhraseAligner->GetNumberOfScores(); i++ ) {
        ItemWithBestScores.m_ScaledScores[ i ] = 9999.0;
      }

      for ( int i = 0; i <= m_pPhraseAligner->GetNumberOfCombinedScores(); i++ ) {
        ItemWithBestScores.m_CombinedScaledScores[ i ] = 9999.0;
      }
      

      if ( AllPhraseTranslations.Size() > 0 ) {
        //Trace( 0, AllPhraseTranslations.Size() << " Translations for " << StartPos << " to " << EndPos << " " << SourceNgram );
        // collect source phrase
        CString SourceNgram = "";
        SourceWordIDs.Clear();
        
        //cout << "get source from " << StartPos << " to " << EndPos << " in: ";
        //m_pSentence->Display( m_pVoc );
        //cout << endl;
        
        for ( int k = StartPos; k <= EndPos; k++ ) {
          //cout << j << " " << i << " " << TextE.GetWord( SentenceID - 1, i ) << endl;
          SourceWordIDs.Append( m_pSentence->GetAt( k ) );   // sentence is 0-based
          SourceNgram += CString( " " ) + m_pSentence->Word( k , m_pVoc );
        }
      
      
        //CSortedArray< CPhraseAlignItemMS > &AllPhraseTranslations = (*m_pPhraseTranslationTable)( StartPos, EndPos );
        
        //Trace( 0, "BeamFactor = " << m_BeamFactor );
        
        SortedTranslations.clear();
        Trace( 2, "insert them into sorted list" );
        
        for ( int m = AllPhraseTranslations.MinIndex(); m <= AllPhraseTranslations.MaxIndex(); m++ ) {
          const CPhraseAlignItemMS &AlignItem = AllPhraseTranslations.GetAt( m );
          //cout << "Deal with: " << AlignItem << endl;
          
          SortedTranslations.insert( pair< double, int >( AlignItem.ScaledScore(), m ) );
          Trace( 2, "now " << SortedTranslations.size() << " items in sorted list" );
          
          
          // update best scores, for pruning
          for(int i = 0; i <= AlignItem.m_ScaledScores.MaxIndex(); i++) {
            if (AlignItem.m_ScaledScores[i] < ItemWithBestScores.m_ScaledScores[i] ) {
              ItemWithBestScores.m_ScaledScores[i] = AlignItem.m_ScaledScores[i];
            }
          }
          for(int i = AlignItem.m_CombinedScaledScores.MinIndex(); i <= AlignItem.m_CombinedScaledScores.MaxIndex(); i++) {
            if (AlignItem.m_CombinedScaledScores[i] < ItemWithBestScores.m_CombinedScaledScores[i] ) {
              ItemWithBestScores.m_CombinedScaledScores[i] = AlignItem.m_CombinedScaledScores[i];
            }
          }
          
          
          if ( AlignItem.ScaledScore() < BestScore ) {
            Trace( 2, "New best score " << AlignItem.Score() );
            BestScore = AlignItem.ScaledScore();
          }
          
        }
        //cout << endl;
        
        // output sorted list
        Trace( 1, "translations sorted according to score" );
        multimap< double, int >::iterator Iter;
        bool tmpflag = false;
        
        // first the best scores
        //cout << ItemWithBestScores << endl;
        //cout << "-------------------------------------------------------------" << endl;
        
        for ( Iter = SortedTranslations.begin(); 
              Iter != SortedTranslations.end();
              ++Iter ) {
          
          int n = Iter->second;
          const CPhraseAlignItemMS &AlignItem = AllPhraseTranslations.GetAt( n );
          //cout << AlignItem << endl;
          
          bool IsGoodTranslation = true;
          
          // take best translation
          if ( Iter == SortedTranslations.begin() ) {
            //cout << "Best Translation: " << AlignItem << endl;
            ; // keep it
          } else {

            // take additional translation only when it is close enough to the best one
            
            if  ( ( AlignItem.m_ScaledScores[0] > ItemWithBestScores.m_ScaledScores[0] * m_BeamFactor) ) {
              Trace( 2, "Don't use, because TotalScore " << AlignItem.m_ScaledScores[0] 
                     << " > BestScore " << ItemWithBestScores.m_ScaledScores[0]
                     << " * m_BeamFactor " << m_BeamFactor );
              IsGoodTranslation = false;
            }
            
            /*
              if (  AlignItem.m_LogProbSentAlignFE > ItemWithBestScores.m_LogProbSentAlignFE + m_BeamFactor ) {
              Trace( 0, "Don't use, because LogProbSentAlignFE > BestScore + m_BeamFactor" );
              IsGoodTranslation = false;
              }
              
              if (  AlignItem.m_LogProbSentAlignEF > ItemWithBestScores.m_LogProbSentAlignEF + m_BeamFactor ) {
              Trace( 0, "Don't use, because LogProbSentAlignEF > BestScore + m_BeamFactor" );
              IsGoodTranslation = false;
              }
              
              if (  AlignItem.m_LogProbPhraseAlignFE > ItemWithBestScores.m_LogProbPhraseAlignFE + m_BeamFactor ) {
              Trace( 0, "Don't use, because LogProbPhraseAlignFE > BestScore + m_BeamFactor" );
              IsGoodTranslation = false;
              }
              
              if (  AlignItem.m_LogProbPhraseAlignEF > ItemWithBestScores.m_LogProbPhraseAlignEF + m_BeamFactor ) {
              Trace( 0, "Don't use, because LogProbPhraseAlignEF > BestScore + m_BeamFactor" );
              IsGoodTranslation = false;
              }
            */
            
            /*
              if (  AlignItem.m_LogProbPhraseFE > ItemWithBestScores.m_LogProbPhraseFE + m_BeamFactor + m_BeamFactor ) {
              Trace( 2, "Don't use, because LogProbPhraseFE > BestScore + m_BeamFactor" );
              IsGoodTranslation = false;
              }
              
              if (  AlignItem.m_LogProbPhraseEF > ItemWithBestScores.m_LogProbPhraseEF + m_BeamFactor ) {
              Trace( 2, "Don't use, because LogProbPhraseEF > BestScore + m_BeamFactor" );
              IsGoodTranslation = false;
              }
            */
            
            /*
            // take only translation when best translation is reasonably good
            if ( BestScore > ( LengthSourcePhrase + 1) * 2 * m_BeamFactor ) {
            Trace( 2, "Don't use, because even best translation is too bad" );
            IsGoodTranslation = false;
            } else {

            // take additional translation only when it is close enough to the best one
            if  ( ( AlignItem.m_LogProb > ItemWithBestScores.m_LogProb + m_BeamFactor) ) {
            Trace( 2, "Don't use, because AlignItem.Score() > BestScore + m_BeamFactor" );
            IsGoodTranslation = false;
            }
            
            else {
            // take only translations which have have good Phrase translation score
            // as this is the one used in the decoder
            if  ( ( AlignItem.LogProbPhraseFE() > ( LengthSourcePhrase + 1) * m_BeamFactor ) ) {
            Trace( 2, "Don't use, because AlignItem.LogProbPhraseFE() > ( LengthSourcePhrase + 1) * m_BeamFactor" );
            IsGoodTranslation = false;
            }
            }
            */


          }
	  
          if ( IsGoodTranslation ) {
            //cout << Iter->first << " : " << AlignItem << endl;
            
            //CreateEdge( AlignItem );
            CArray< CWordIndex > TargetWordIDs = ((CPhraseAlignItemMS)AlignItem).TargetWordIDs();
            
            /*
            //
            cout << "@PESA-MS: # ";
            //cout << "Edge for: ";
            for ( int j = SourceWordIDs.MinIndex(); j <= SourceWordIDs.MaxIndex(); j++ ) {
            cout << m_pVoc->GetWord( SourceWordIDs[ j ] ) << " ";
            //pWordInfo->AddSourceWord( m_pVoc->GetWord( SourceWordIDs[ j ] ) );
            }
            cout << " # ";
            
            for ( int i = TargetWordIDs.MinIndex(); i <= TargetWordIDs.MaxIndex(); i++ ) {
            cout << m_pVoc->GetWord( TargetWordIDs[ i ] ) << " ";
            //pWordInfo->AddTargetWord( m_pVoc->GetWord( TargetWordIDs[ i ] ) );
            }
            cout << " # ";
            
            const CNumericalArray< double > & Scores = AlignItem.Scores();
            for ( int i = Scores.MinIndex(); i <= Scores.MaxIndex(); i++ ) {
            cout << exp( -1.0 * Scores[ i ] ) << " ";
            }
            cout << endl;
            */
            //cout << AlignItem << endl;
            //
            
            AddEdge( m_pLattice, StartPos - 1, EndPos, AlignItem.Score(), 
                     AlignItem.Scores(), SourceWordIDs, TargetWordIDs );
            
          } else {
            if ( tmpflag ) {
              cout << "====== prune =======" << endl;
              tmpflag = false;
            }
            //cout << Iter->first << " : " << AlignItem << endl;
          }
          
        }		
      }
      
      // we are done with those translations, remove them
      AllPhraseTranslations.Clear();
    }
  }

}




float
CPhraseSearchMS::PhraseTranslationLogProb( const CArray< CWordIndex > &Source, 
					   const CArray< CWordIndex > &Target, 
					   const CLexicon *pLexicon )
{
  float Prob = 1.0;
  float SumProb = 0.0;
  float LogProb = 0.0;

  //cout << "PhraseTranslationLogProb " <<  Source.Size() << " " << Target.Size() << endl;
  for ( int j = Source.MinIndex(); j <= Source.MaxIndex(); j++ ) {
    SumProb = 0.0;
    for ( int i = Target.MinIndex(); i <= Target.MaxIndex(); i++ ) {
      //cout << Source[ j ] << " " <<  Target[ i ] << " "
      //	   << Lexicon.Pr( Source[ j ], Target[ i ] ) << endl;
      SumProb += pLexicon->Pr( Source[ j ], Target[ i ] );
    }
    Prob *= SumProb;
    //cout << SumProb << endl;
  }

  //cout << Prob << " " << -log( Prob ) << endl;
  return -log( Prob );
}


/*
void
CPhraseSearchMS::AddEdge( CLattice *pLattice, int From, int To, double LogProb,
			  const CArray< CWordIndex > &SourceWordIDs,
			  const CArray< CWordIndex > &TargetWordIDs )
{
  CLatticeNode *pFromNode = pLattice->Node( From );
  CLatticeNode *pToNode   = pLattice->Node( To );

  CLatticeEdgeTrans *pEdge = new CLatticeEdgeTrans( pFromNode, pToNode );
  CWordInfo *pWordInfo = pEdge->WordInfo();
		
  for ( int j = SourceWordIDs.MinIndex(); j <= SourceWordIDs.MaxIndex(); j++ ) {
	//cout << m_pVoc->GetWord( SourceWordIDs[ j ] ) << " ";
	pWordInfo->AddSourceWord( m_pVoc->GetWord( SourceWordIDs[ j ] ) );
  }
  //cout << " # ";
  
  for ( int i = TargetWordIDs.MinIndex(); i <= TargetWordIDs.MaxIndex(); i++ ) {
	//cout << m_pVoc->GetWord( TargetWordIDs[ i ] ) << " ";
	pWordInfo->AddTargetWord( m_pVoc->GetWord( TargetWordIDs[ i ] ) );
  }
  //cout << " # ";
  
  
//  if ( m_pLexicon ) {
//LogProb = PhraseTranslationLogProb( SourceWordIDs, TargetWordIDs.GetRow( m ), m_pLexicon );
//} else {
//LogProb = 0.0;
//}
  
  //cout << "LogProb: " << LogProb << endl;
  pWordInfo->Score() = LogProb;
  
  //cout << *pWordInfo << endl;
  
  // now the edge
  pEdge->SetWord( "@PESA" );
  //pEdge->Display();
  pLattice->AddEdge( pEdge );
}
*/


void
CPhraseSearchMS::AddEdge( CLattice *pLattice, int From, int To,
                          double LogProb, CNumericalArray< float > Scores,
                          const CArray< CWordIndex > &SourceWordIDs,
                          const CArray< CWordIndex > &TargetWordIDs )
{
  CLatticeNode *pFromNode = pLattice->Node( From );
  CLatticeNode *pToNode   = pLattice->Node( To );

  CLatticeEdgeMS *pEdge = new CLatticeEdgeMS( pFromNode, pToNode );
  CWordInfo *pWordInfo = pEdge->WordInfo();
		
  for ( int j = SourceWordIDs.MinIndex(); j <= SourceWordIDs.MaxIndex(); j++ ) {
    //cout << m_pVoc->GetWord( SourceWordIDs[ j ] ) << " ";
    pWordInfo->AddSourceWord( m_pVoc->GetWord( SourceWordIDs[ j ] ) );
  }
  //cout << " # ";
  
  for ( int i = TargetWordIDs.MinIndex(); i <= TargetWordIDs.MaxIndex(); i++ ) {
    //cout << m_pVoc->GetWord( TargetWordIDs[ i ] ) << " ";
    pWordInfo->AddTargetWord( m_pVoc->GetWord( TargetWordIDs[ i ] ) );
  }
  //cout << " # ";
  
  
//  if ( m_pLexicon ) {
//LogProb = PhraseTranslationLogProb( SourceWordIDs, TargetWordIDs.GetRow( m ), m_pLexicon );
//} else {
//LogProb = 0.0;
//}
  
  //cout << "LogProb: " << LogProb << endl;
  pWordInfo->Score() = LogProb;


  //cout << "Scores = " << Scores << endl;
  // until now the scores are scores per word
  // in the decoder we want to have the scores for the entire phrase,
  // i.e. we multiply by source phrase length = To - From
  int SourcePhraseLength = To - From;
  CNumericalArray< float > PhraseScores;
  //for ( int i = Scores.MinIndex(); i <= Scores.MaxIndex(); i++ ) {
  for ( int i = 1; i <= Scores.MaxIndex(); i++ ) {
    PhraseScores[ i ] = Scores[ i ] * SourcePhraseLength;
  }

  pEdge->Scores() = PhraseScores;
  
  //cout << *pWordInfo << endl;
  
  // now the edge
  pEdge->SetWord( "@PESA-MS" );
  //cout << "MultiScore Edge: ";
  //pEdge->Display();
  pLattice->AddEdge( pEdge );
}




void
CPhraseSearchMS::DisplayPhraseLocation(  const S_phraseLocationElement &PhraseLocation ) const
{
  // Display the phrase locations
  //cout << PhraseLocations.size() << " phrase locations returned from suffix array search" << endl;

  int StartPosPhrase = PhraseLocation.posStartInSrcSent;
  int EndPosPhrase   = PhraseLocation.posEndInSrcSent ;
  int SentenceID     = PhraseLocation.sentIdInCorpus;
  int PosInSentence  = PhraseLocation.posInSentInCorpus;
  
  cout << StartPosPhrase << " "
       << EndPosPhrase << " "
       << SentenceID << " "
       << PosInSentence << endl;
}


  
void
CPhraseSearchMS::DisplayPhraseLocations(  const vector< S_phraseLocationElement > &PhraseLocations ) const
{
  // Display the phrase locations
  //cout << PhraseLocations.size() << " phrase locations returned from suffix array search" << endl;

  for ( int n = 0; n < PhraseLocations.size(); n++ ) {
    DisplayPhraseLocation( PhraseLocations[ n ] );
  }

}
  
