#include <config.h>

#ifndef __SegmentingInfo__
#define __SegmentingInfo__

#include <vector>
#include <functional>

#include <iostream>


using namespace std;


// Nn hierachical phrase alignment we detect the phrase pairs, which 
// give the best sentence splitting in a sentence pair
// CSegmentInfo is used to store the information for such a phrase pair
// This is essentially a light-weight version of CPhraseAlignItem, as
// we don't need the actual phrases here.

class CSegmentingInfo
{
public:

  CSegmentingInfo();
  CSegmentingInfo( const CSegmentingInfo &SegmentingInfo );
  CSegmentingInfo( int SentenceNumber,  
				   short StartPosInSource, short EndPosInSource, 
				   short StartPosInTarget, short EndPosInTarget,
				   float Score );
  
  CSegmentingInfo &operator = ( const CSegmentingInfo &SegmentingInfo );

  friend ostream& operator << ( ostream& os, const CSegmentingInfo &SegmentingInfo );
  //friend bool operator < ( const CSegmentingInfo &, const CSegmentingInfo & );
  //friend bool CompareValue( const CSegmentingInfo &, const CSegmentingInfo & );

  //protected:

  // in which sentence is the phrase pair
  long m_SentenceNumber;

  // start and end positions of source and target phrase
  short m_StartPosInSource;
  short m_EndPosInSource;
  short m_StartPosInTarget;
  short m_EndPosInTarget;

  // the scores associated with this phrase pair
  // alignment cost can be calculated in different ways:
  // - per word alignment cost from PESA alignment
  //   - using some or all scores there
  //   - using unweighted or weighted scores
  // - perlexity gain for sentence, i.e. alignment cost for
  //   sentence pair with and without segmentation
  // score is calculated in CPhraseAlignMS, i.e. check there what
  // actually is done
  // Here we only assume that Scores is a cost, i.e. lower is better
  float m_Score;

  friend class CSegmenter;
};


//bool operator < ( const CSegmentingInfo &, const CSegmentingInfo & );

// sort SegmentingInfos by their score, lower score is better
struct CompareSegmentingInfoByValue : public binary_function< CSegmentingInfo, CSegmentingInfo, bool >
{
  bool operator()( const CSegmentingInfo &Sp1, const CSegmentingInfo &Sp2 ) 
    { 
      return Sp1.m_Score < Sp2.m_Score;
    }
};


// sort SegmentingInfo by there sentence number to support sequential processing
// were we may need to read from and write to files
struct CompareSegmentingInfoByPosition : public binary_function< CSegmentingInfo, CSegmentingInfo, bool >
{
  bool operator()( const CSegmentingInfo &Sp1, const CSegmentingInfo &Sp2 )
  {
    return Sp1.m_SentenceNumber < Sp2.m_SentenceNumber;
  }
};



class CSegmentingInfoList
{
public:

  CSegmentingInfoList();
  ~CSegmentingInfoList();

  const vector< CSegmentingInfo > & SegmentingInfoList();

  void AddSegmentingInfo( CSegmentingInfo &SegmentingInfo );
  void Resize( int NewSize );

  void SortToValue();
  void SortToPosition();

  void Display();


private:
  vector< CSegmentingInfo > m_SegmentingInfoList;
};

  

#endif
