/*
 * Array.h --
 *	Extensible array class
 *
 * Copyright (c) 1995,1997 SRI International.  All Rights Reserved.
 *
 * @(#)$Header$
 *
 */

#ifndef _Array_h_
#define _Array_h_

#include <assert.h>
#include <Types.h>
#include "MemStats.h"

template <class DataT>
class Array
{
public:
    Array(int base = 0, unsigned int size = 0)
	: _base(base), _size(size), _data(0), alloc_size(0)
	{ if (size > 0) { alloc(size-1); } };
    ~Array() { delete [] _data; } ;

    DataT &operator[](int index)
    	{ int offset = index - _base; assert(offset >= 0);
	  if (offset >= _size) {
	    _size = offset + 1;
	    if (offset >= alloc_size) { alloc(offset); }
	  }
	  return _data[offset];
		}

	Array<DataT>& operator = (const Array<DataT> &other) {

		if (&other == this) {
			return *this;
		}

		delete [] _data;

		_base = other._base;
		_size = other._size;
		alloc_size = other.alloc_size;

		_data = new DataT[alloc_size];
		assert(_data != 0);

#ifdef ZERO_INITIALIZE
		memset(_data, 0, alloc_size * sizeof(DataT));
#endif

		for (unsigned i = 0; i < alloc_size; i++) {
			_data[i] = other._data[i];
		}

		return *this;
	}

    DataT *data() const { return _data - _base; }
    int base() const { return _base; }
    unsigned int size() const { return _size; }

	void memStats(MemStats &stats) const {
		    stats.total += _size * sizeof(_data[0]);
    stats.wasted += (alloc_size - _size) * sizeof(_data[0]);
	}

private:
    unsigned int _size;		/* used size */
    int _base;
    DataT *_data;
    unsigned int alloc_size;	/* allocated size */
	void alloc(unsigned int size) {
		    unsigned int newSize = size + 1 + alloc_size/2;
    DataT *newData = new DataT[newSize];
    assert(newData != 0);

#ifdef ZERO_INITIALIZE
    memset(newData, 0, newSize * sizeof(DataT));
#endif

    for (unsigned i = 0; i < alloc_size; i++) {
	newData[i] = _data[i];
    }

    delete [] _data;

    _data = newData;
    alloc_size = newSize;
	}
};

#endif /* _Array_h_ */
