#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <fortify.h>

#define NO_TRACE 1
#include <Trace.hh>

#include <ServerMode.hh>

ServerMode::ServerMode(void* thread_params){
  ThreadParams* my_params = (ThreadParams*)thread_params;
  pDecoder = my_params->pDecoder;
  pTrueDecoder = my_params->pTrueDecoder;

  //for online phrase search
  pPhraseSearcher=my_params->pPhraseSearcher;
  pLatticeRescorer=my_params->pLatticeRescorer;
  //lexica for scoring
  pLexicon=my_params->pLexicon;
  pReverseLexicon=my_params->pReverseLexicon;  

  pVoc = my_params->pVoc;
  ServerSocket = my_params->ServerSocket;
  ClientSocket = my_params->ClientSocket;
  ServerTimeOut = my_params->ServerTimeOut;
}

// we only get pointers to all these objects, so they are left to 
// be deleted outside of this object
ServerMode::~ServerMode(){

}
// handles threading in the server mode for decoding
CString ServerMode::ServerModeDecoding(const char* Request){


  //std::cout << "Client connection accepted" << std::endl;
  
  //char Buffer[BUFFERSIZE];
  // Client doesn't need the server socket anymore
  //std::cout<<"Child process..."<<getpid()<<endl;
  //std::cout<<"Waiting on input..."<<getpid()<<endl;
  //ClientSocket->SetTimeOut(*ServerTimeOut);
  //ClientSocket->ReadLine(Buffer, BUFFERSIZE);
  //std::cout<<"Received:"<<Buffer<<"--"<<endl;
    
    // processes the server request
  CString response = ProcessServerRequest(Request);
    
  //  ClientSocket->Write(Buffer);
  // ClientSocket->Close();
    
  return response;
}
//sets the translation parameters based on the current codeString
// the codeString must explicity define every parameter in the same order 
// that the call to GetTranslationParams returns them in
CString ServerMode::SetTranslationParams(const char* codeString){
  // ARV introduced get methods to be used when administering the decoding server
  int LengthOfReorderingWindow;
  float ReorderingsPerWord;
  int N;
  bool CompareUseCoverage;
  bool CompareUseCoveragePattern;
  bool CompareUseLM;
  bool CompareUseTargetLength;

  bool PruneUseCoverage;
  bool PruneUseCoveragePattern;
  bool PruneUseLM;
  bool PruneUseTargetLength;

  long unsigned int BeamFactor;
  float SLStretch;
  float TMScale;
  float LMScale;
  float ALScale;
  float SLScale;

  string str(codeString);
  std::istringstream stream(str);

  // name of the param, unused
  string paramName;
  stream>>paramName>>LengthOfReorderingWindow
	>>paramName>>ReorderingsPerWord
	>>paramName>>N
	>>paramName>>CompareUseCoverage
	>>paramName>>CompareUseCoveragePattern
	>>paramName>>CompareUseLM
	>>paramName>>CompareUseTargetLength
	>>paramName>>PruneUseCoverage
	>>paramName>>PruneUseCoveragePattern
	>>paramName>>PruneUseLM
	>>paramName>>PruneUseTargetLength
	>>paramName>>BeamFactor
	>>paramName>>SLStretch
	>>paramName>>TMScale
	>>paramName>>LMScale
	>>paramName>>ALScale
	>>paramName>>SLScale;

  pDecoder->SetLengthOfReorderingWindow( LengthOfReorderingWindow );
  pDecoder->SetReorderingsPerWord( ReorderingsPerWord );
  pDecoder->SetNumOfNBest( N );
  pDecoder->SetCompareHypFeatures( CompareUseCoverage, CompareUseCoveragePattern,
			 CompareUseLM, CompareUseTargetLength);
  pDecoder->SetPruneHypFeatures( PruneUseCoverage , PruneUseCoveragePattern,
		        PruneUseLM, PruneUseTargetLength);
  
  pDecoder->SetBeamSize( BeamFactor);

  //pDecoder->SetSLStretchFactor( SLStretch );
  
  pDecoder->SetScalingFactors( TMScale, LMScale,
		     ALScale, SLScale);

  return "SUCCESS\n";
}
// returns the translation parameters for the given decoder
CString ServerMode::GetTranslationParams(){
    // ARV introduced get methods to be used when administering the decoding server
  int LengthOfReorderingWindow;
  float ReorderingsPerWord;
  int N;
  bool CompareUseCoverage;
  bool CompareUseCoveragePattern;
  bool CompareUseLM;
  bool CompareUseTargetLength;

  bool PruneUseCoverage;
  bool PruneUseCoveragePattern;
  bool PruneUseLM;
  bool PruneUseTargetLength;

  long unsigned int BeamFactor;
  float SLStretch;
  float TMScale;
  float LMScale;
  float ALScale;
  float SLScale;

  pDecoder->GetLengthOfReorderingWindow( LengthOfReorderingWindow );
  pDecoder->GetReorderingsPerWord( ReorderingsPerWord );
  pDecoder->GetNumOfNBest( N );
  pDecoder->GetCompareHypFeatures( CompareUseCoverage, CompareUseCoveragePattern,
			 CompareUseLM, CompareUseTargetLength );
  pDecoder->GetPruneHypFeatures( PruneUseCoverage, PruneUseCoveragePattern,
		        PruneUseLM, PruneUseTargetLength);
  
  pDecoder->GetBeamSize( BeamFactor);

  //pDecoder->GetSLStretchFactor( SLStretch );
  
  pDecoder->GetScalingFactors( TMScale, LMScale,
		     ALScale, SLScale);

  std::stringstream stream;
  //std::sstream stream;
  stream<<"PARAMS:"<<"LengthOfReorderingWindow "<<LengthOfReorderingWindow<<" "\
	<<"ReorderingsPerWord "<<ReorderingsPerWord<<" "
	<<"N "<<N<<" "
	<<"CompareUseCoverage "<<CompareUseCoverage<<" "
	<<"CompareUseCoveragePattern "<<CompareUseCoveragePattern<<" "
	<<"CompareUseLM "<<CompareUseLM<<" "
	<<"CompareUseTargetLength "<<CompareUseTargetLength<<" "
	<<"PruneUseCoverage "<<PruneUseCoverage<<" "
	<<"PruneUseCoveragePattern "<<PruneUseCoveragePattern<<" "
	<<"PruneUseLM "<<PruneUseLM<<" "
	<<"PruneUseTargetLength "<<PruneUseTargetLength<<" "
	<<"BeamFactor "<<BeamFactor<<" "
	<<"SLStretch "<<SLStretch<<" "
	<<"TMScale "<<TMScale<<" "
	<<"LMScale "<<LMScale<<" "
	<<"ALScale "<<ALScale<<" "
	<<"SLScale "<<SLScale<<" ";

  string retString = stream.str();
  return CString(retString.c_str());
  
}

// runs a translation on a sentence object
CString ServerMode::GetTranslation(const char* toTranslate){

  //DeEncoder mydec;

  //char* resultBuffer = new char[BUFFERSIZE];
  //char* tmpVersion = (char*)toTranslate;
  //cout<<"InitialString:"<<tmpVersion<<endl;
  //mydec.Mem_ID_and_Transliterate(tmpVersion, resultBuffer);
  //CString resultCString = CString(resultBuffer);
  //resultCString.TrimLeadingBlanks();
  //resultCString.TrimTrailingBlanks();
  

  //cout<<"EncodingResult:"<<resultBuffer<<endl;
  /*
  CString testString = CString(resultBuffer);
  testString.TrimLeadingBlanks();
  testString.TrimTrailingBlanks();

  if (testString.GetLength()==0){
    return testString;
  }
  */


  //Load the initial darwish sentence
  CSentence* pSentence = new CSentence();
  //const char* tmpToTranslate = (char*)resultCString;
  int numWords = pSentence->ReadFromBuffer(toTranslate, pVoc);

  //introduce components for online phrase search here
  CLatticeLM *pLattice = pDecoder->SentenceToLattice( pSentence, pVoc);

  pDecoder->BuildTranslationLattice( pLattice );
  pPhraseSearcher->SearchAlign( pSentence, pLattice);
  pLatticeRescorer->Rescore( pLattice, pLexicon );
  pDecoder->SetSourceSentence( pSentence);
  // sv 051102 result should not be decoder specific
  //CArray<CDecoderNBResult*> TranslationBank = pDecoder->ApplyLanguageModel(pLattice);
  CArray<CDecoderResult*> TranslationBank = pDecoder->ApplyLanguageModel(pLattice);
  CString Translation = TranslationBank[TranslationBank.MinIndex()]->m_translation;


  //CLatticeLM *pLattice = pDecoder->SentenceToLattice( pSentence, pVoc );
  //CArray<CDecoderNBResult*> TranslationBank = pDecoder->TranslateLattice( pLattice );
  //CString Translation = TranslationBank[TranslationBank.MinIndex()]->m_translation;

  // OLD CString Translation = pDecoder->TranslateLattice( pLattice );


  cout << "LowerCase Translation is:" << Translation << endl << endl << flush;
  delete pSentence;
  pDecoder->Clear();
  TranslationBank.Clear();
  delete pLattice;
  

  //delete resultBuffer;

  // true case it if we have the true casing object
  if (pTrueDecoder){
    char* intermediate = (char*)((const char*)Translation);
    pSentence = new CSentence();
    pSentence->ReadFromBuffer(intermediate, pVoc);
    cout<<"About to re-translate"<<endl;
    pSentence->Display(pVoc);
    pLattice = pTrueDecoder->SentenceToLattice(pSentence, pVoc);  
    TranslationBank = pTrueDecoder->TranslateLattice(pLattice);
    CString FinalTranslation = TranslationBank[TranslationBank.MinIndex()]->m_translation;
    //CString FinalTranslation = pTrueDecoder->TranslateLattice(pLattice);
    delete pSentence;
    
    return FinalTranslation;
  }else{
    return Translation;
  }
}

// processes server requests, either for translation or administration
CString ServerMode::ProcessServerRequest(const char* serverRequestConst){
  string str(serverRequestConst);
  
  int transValue = str.find("TRANSLATE");
  int adminValue = str.find("ADMIN");
  int getParamsValue = str.find("GETPARAMS");
  int setParamsValue = str.find("SETPARAMS");

  int getPingValue = str.find("PING");
  if (getPingValue == 0){
    return "PONG";
  }

  // if its a parameter request
  if (getParamsValue == 0){
    CString getResponse = GetTranslationParams();
    return getResponse;
  }

  // if its a translation request
  if (transValue == 0){
    

    // check the string for format errors
    int endValue = str.find("\n");
    if (endValue != str.length()-1){
      return "Invalid Prototocol - Too many line breaks";
    }
    /*
    // count the number of occurrences of
    int colValue = str.find(":");
    if (colValue==-1){
      return "Invalid Prototocol - No translation received";
    }
    colValue = str.find(":", colValue+1);
    if (colValue==-1){
      return "Invalid Prototocol - No translation parameters received";
    }
    colValue = str.find(":", colValue+1);
    if (colValue!=-1){
      return "Invalid Prototocol - Reserved character : found in translation";
    }
    */

    //CText TextF;
    //TextF.Allocate( 50000 );
    //TextF.SetVocabulary( pVoc );
    // separate the request from the protocol
    char* serverRequest = (char*)serverRequestConst; 
    char* reqName = strtok(serverRequest, "|||");
    char* transText = strtok(NULL, "|||");

    char* paramsText = strtok(NULL, "|||");
    if (paramsText!=NULL){
      cout<<"PARAMS:"<<paramsText<<endl;
      SetTranslationParams(paramsText);
    }
    char* finalText = strtok(transText,"\n");
    cout<<"Converting:"<<finalText<<"--"<<endl;
    

    // TextF.LoadAsciiLine(transText);    
    // run the translation
    CString response = GetTranslation(finalText);
    cout<<"TranslatedAs:"<<response<<"---"<<endl;
    return response;
  }
  
  if (adminValue == 0 ){
    return "NO ADMIN YET";
  }

  return "INVALID REQUEST\n";

}










