#ifndef __SERVERMODE_H_DEFINE__
#define __SERVERMODE_H_DEFINE__

#include <config.h>
#include <fortify.h>
#include <sstream>
#include <msocket.h>
#include <DecoderNB.hh>
#include <PhraseSearch.hh>
#include <LatticeRescorer.hh>

#define BUFFERSIZE 10000

typedef struct{
  CDecoderNB* pDecoder;
  CDecoderNB* pTrueDecoder;
  //added for online phrase search
  CPhraseSearch *pPhraseSearcher;
  CLatticeRescorer *pLatticeRescorer;
  //lexica for scoring
  CLexicon* pLexicon;
  CLexicon *pReverseLexicon;  

  CVocabulary* pVoc;
  MSocket* ServerSocket;
  MSocket* ClientSocket;
  int* ServerTimeOut;
}ThreadParams;

// Operates the decoding process as a server
class ServerMode{

public:

  ServerMode(void* thread_params);
  ~ServerMode();

  // handles threading in the server mode for decoding
  CString ServerModeDecoding(const char* Request);
  //sets the translation parameters based on the current codeString
  // the codeString must explicity define every parameter in the same order 
  // that the call to GetTranslationParams returns them in
  CString SetTranslationParams(const char* codeString);
  // returns the translation parameters for the given decoder
  CString GetTranslationParams();
  CString GetTranslation(const char* toTranslate);
  // runs a translation on a given textF object
  //CString GetTranslation(CText TextF, CDecoderGT *pDecoder);
  // processes server requests, either for translation or administration
  CString ProcessServerRequest(const char* serverRequest);    

private:
  CDecoderNB* pDecoder;
  CDecoderNB* pTrueDecoder;
  //ARV for online phrase search
  CPhraseSearch *pPhraseSearcher;
  CLatticeRescorer *pLatticeRescorer;
  //lexica for scoring
  CLexicon* pLexicon;
  CLexicon *pReverseLexicon;  

  CVocabulary* pVoc;
  MSocket* ServerSocket;
  MSocket* ClientSocket;
  int* ServerTimeOut;
};

#endif
