#ifndef _MSOCKET_H_DEFINE
#define _MSOCKET_H_DEFINE

// Needed by MSocket class:
#include <ctype.h> // for isdigit
#include <sys/types.h> // for connect(), bind(), accept(),
#include <sys/socket.h> // for connect(), listen(), bind() accept(),
#include <netinet/in.h> // for ntons(), htonl(), etc...
#include <arpa/inet.h> // for inet_addre()
#include <unistd.h> // for close(), read(), write()
#include <netdb.h> // for gethostbyname
#include <sys/time.h> // for timeval struct, and select()
#include <string.h> // for memset()
#include <map> // for STL map class (Reference counting)

// The MSocket class
class MSocket
{
public:
MSocket( int nFamily=AF_INET, int nType=SOCK_STREAM, int nProtocol=0 );
MSocket( const MSocket& msSrc ); // Copy Constructor
~MSocket();
operator int() { return( m_nSocket ); }
int Close();

// Socket operations (Client)
int Connect( struct sockaddr * sapSrvrAdrs, int nLenAddr ) const;
int Connect( const char* szHostAddr, short sPort  ) const;

// Socket operation (Server)
int Bind( struct sockaddr * sapSrvrAdrs, int nLenAddr ) const;
int Bind( int nPort, int nFamily=AF_INET, int nAddr=INADDR_ANY) const;
int Listen( int nBackLog=5 ) const;
int Server( int nPort ) const; // Helper Function
int Accept( MSocket& msrRet, struct sockaddr* sapClientAddr=0, int* npLen=0 ) const;

// Atributes for normal- or exception-mode error reporting
static bool m_bThrowException;
static bool SetExceptions( bool bMode );

// Operators
MSocket& operator=( MSocket& msrSrc );
MSocket& operator=( int Socket );

// I/O functions
int SetTimeOut( int MicroSeconds );
int ReadLine( char *cpDest, int nSize, char Term='\n' ) const;
int WriteLine( const char* cpSrc, char Term='\n' ) const;

// Helper functions
void ThrowErrMsg( const char* cpMsg, int nCode=0 ) const;
void ThrowErrMsg( const char* cpMsg, const char* cpInfo ) const { ThrowErrMsg( cpMsg, (int) cpInfo); }
static bool WasConTermErr( const char* ErrMsg );
protected:
int m_nSocket, m_nTimeOut;

int TimedOut() const;

// Support for reference counting, using STL map class
static std::map<int,int> RefCount;
void IncRef()
{ if( RefCount.find( m_nSocket )!=RefCount.end() ) RefCount[m_nSocket]++; else RefCount[m_nSocket]=1; }
void DecRef()
{ if( RefCount.find( m_nSocket )!=RefCount.end() ) RefCount[m_nSocket]--; }
bool GetRef()
{ return( RefCount.find(m_nSocket)!=RefCount.end()? RefCount[m_nSocket] : 0 ); }

};

// Some prototypes for over-ridden operators, to make class look like a cout/cin object
MSocket& operator<<( MSocket& msrOut, const char* cpStr );
MSocket& operator>>( MSocket& msrIn, char* cpStr );

#endif
	   
