#include <assert.h>
#include <math.h>
#include <string.h>
#include <iostream>
using namespace std;

#include "Trace.hh"

#ifndef __Array_hh__
#error "Do not include this file!"
#error "It is included via Array.hh"
#endif



//--------------------------------------------------------------------
//       CArray1D inlines
//

//--------------------------------------------------------------------
//       CArray1D::CArray1D
//
template<class T>
inline
CArray1D<T>::CArray1D()
{
  //-Trace( 5, "CArray1D<T>::CArray1D()" );
  Init();
}


//--------------------------------------------------------------------
//       CArray1D::~CArray1D
//
template<class T>
inline
CArray1D<T>::~CArray1D()
{
  Clear();
}


//--------------------------------------------------------------------
//       CArray1D::CArray1D - Copy Constructor
//
template<class T>
inline
CArray1D<T>::CArray1D( const CArray1D<T> &Array )
{
  Init();
  *this = Array;
}


//--------------------------------------------------------------------
//       CArray1D::GetMinIndex
//
template<class T>
inline int
CArray1D<T>::MinIndex() const
{
  return m_MinIndex;
}


//--------------------------------------------------------------------
//       CArray1D::GetMaxIndex
//
template<class T>
inline int
CArray1D<T>::MaxIndex() const
{
  return m_MaxIndex;
}


//--------------------------------------------------------------------
//       CArray1D::SetMaxIndex
//
template<class T>
inline bool
CArray1D<T>::SetMaxIndex( int MaxIndex)
{
  return SetBoundaries( m_MinIndex, MaxIndex );
}


//--------------------------------------------------------------------
//       CArray1D::SetIncrement
//
template<class T>
inline void
CArray1D<T>::SetIncrement(int Increment)
{
  if (Increment < 1)
    Increment = 1;
  m_Increment = Increment;
}


//--------------------------------------------------------------------
//       CArray1D::MakeStatic
//
template<class T>
inline void
CArray1D<T>::MakeStatic()
{
  m_IsStatic = true;
}


//--------------------------------------------------------------------
//       CArray1D::MakeDynamic
//
template<class T>
inline void
CArray1D<T>::MakeDynamic()
{
  m_IsStatic = false;
}


//--------------------------------------------------------------------
//       CArray1D::InitToZero
//
template<class T>
inline void
CArray1D<T>::InitToZero()
{
  m_InitToZero = true;
}


//inline template<class T> CArray1D &
//CArray1D::operator+=( CContainerItem *arg )
//{
//  Append( arg );
//  return *this;
//}

//inline template<class T> CArray1D &
//CArray1D::operator<<( CContainerItem *arg )
//{
//  Append( arg );
//  return *this;
//}


//--------------------------------------------------------------------
//       CArray1D::IsValidIndex
//
template<class T>
inline bool 
CArray1D<T>::IsValidIndex( int Index ) const
{
  return (Index >= m_MinIndex) && (Index <= m_MaxIndex);
} // IsValidIndex


//--------------------------------------------------------------------
//       CArray1D::IsValid
//
template<class T>
inline bool
CArray1D<T>::IsValid() const
{
  return m_MaxIndex >= m_MinIndex;   // there has to be at least one element
} // isValid


//--------------------------------------------------------------------
//       CArray1D::Size
//
template<class T>
inline size_t
CArray1D<T>::Size() const
{
  if ( !IsValid() ) {
    return 0;
  }
  return m_MaxIndex - m_MinIndex + 1;
} // GetSize



//--------------------------------------------------------------------
//       CArray1D::IsEmpty
//
template<class T>
inline bool
CArray1D<T>::IsEmpty() const
{
  if ( !IsValid() ) {
    return true;
  }
  return m_MaxIndex < m_MinIndex;
} // GetSize



//--------------------------------------------------------------------
//       CArray1D non inlines
//

//--------------------------------------------------------------------
//       CArray1D :: operator =
// 
template<class T>
CArray1D<T> &
CArray1D<T>::operator = ( const CArray1D<T> &Array )
{
  //-Trace( 5, "CArray1D<T>::operator = " );

  //Array.Display();

  if ( this == &Array ) 
    // copy onto itself can be skipped
    return *this;

  if ( !Array.IsValid() ) {
    Invalidate();
    return *this;
  }
   
  // remove stored elements
  Clear();

  // set boundaries, i.e. allocate correct memory
  if ( !AssureBoundaries( Array.MinIndex(), Array.MaxIndex() ) ) 
	// if not enough memory, return
    return *this;

  // copy all elements
  for ( int i = Array.MinIndex(); i <= Array.MaxIndex(); i++ )
    SetAt( i, Array[i] );

  return *this;
} // op =

  
//--------------------------------------------------------------------
//       CArray1D :: operator const char *
// 
/*
template<class T>
CArray1D<T>::operator const char* () const
{
  serBuf = "[\n";
  if ( IsValid() )
    for ( int i = 0; i < count; i++ ) {
      serBuf = (const char*) *p[i];
      serBuf += " ";
      serBuf += "\n";
    }
  else
    serBuf = invalidStr;

  serBuf += "]";
  return serBuf;
} // op const char *
*/


//--------------------------------------------------------------------
//       CArray1D :: toStream
//
/* 
bool
CArray1D<T>::ToStream( ostream &ostr ) const
{
  ostr << "[" << endl;
  if ( IsValid() )
    for ( int i = 0; i < count; i++ )
      ostr << *p[i] << endl;

  else
    ostr << invalidStr;

  ostr << "]" << endl;
  return ostr.good();
} // toStream
*/


//--------------------------------------------------------------------
//       CArray1D:: operator []
//
template<class T>
T&
CArray1D<T>::operator[] ( int Index )
{
  //AssureSize( Index + 1 );
  if (IsValidIndex( Index ) == false)
    AssureBoundaries( Index, Index );

  return m_pElements[ Index ];
} // op []


//--------------------------------------------------------------------
//       CArray1D :: operator []
// 
template<class T>
const T&
CArray1D<T>::operator[] ( int Index ) const
{
  // if object is invalid, throw assertion
  if ( !IsValidIndex( Index ) ) {
    //return *noDataPtr;
    cerr << "Array[] invalid Index " << Index << endl;
    assert(0);
  }

  return m_pElements[ Index ];
} // op []


//--------------------------------------------------------------------
//       CArray1D:: append
//
template<class T>
bool
CArray1D<T>::Append( const T& Element )
{
  if ( !AssureBoundaries( m_MinIndex, m_MaxIndex + 1 ) ) {
    return false;
  }
  // here, we have incremented MaxIndex 
  return SetAt( m_MaxIndex, Element );
} // append


//--------------------------------------------------------------------
//       CArray1D:: Insert
//
template<class T>
bool
CArray1D<T>::Insert( int Index, const T & Element )
{
  int NewMinIndex = m_MinIndex;
  int NewMaxIndex = m_MaxIndex + 1;

  // adjust boundaries
  if (Index < m_MinIndex) {
    NewMinIndex = Index;
  } else { 
    if (Index > m_MaxIndex) {
      NewMaxIndex = Index;
    }
  }

  // store old max index, as it is needed later on, but modified
  // by AssureBoundaries
  int OldMaxIndex = m_MaxIndex;

  if ( !AssureBoundaries( NewMinIndex, NewMaxIndex ) )
    return false;

  // move all elements starting from Index by one position
  int i;
  for ( i = OldMaxIndex; i >= Index; i-- )
    m_pElements[i + 1] = m_pElements[i];

  // set new element at position Index
  m_pElements[Index] = Element;

  return true;
} // Insert



//--------------------------------------------------------------------
//       CArray1D:: remove
//

template<class T>
bool
CArray1D<T>::Remove( int Index, int Number )
{
  if ( !IsValid() || !IsValidIndex( Index ) )
     return false;

  // move all elements at position Index + Number to Index
  int i;
  for ( i = Index + Number; i <= m_MaxIndex; i++ ) {
    //-Trace( 1, i );
    m_pElements[ i - Number ] = m_pElements[ i ];
  }

  m_MaxIndex -= Number;

  return true;
} // remove


//--------------------------------------------------------------------
//       CArray1D:: GetAt
//
template<class T>
const T &
CArray1D<T>::GetAt( int Index ) const
{
  if ( !IsValidIndex( Index ) ) 
    //return noDataPtr;
    assert(0);

  return m_pElements[ Index ];
} // GetAt


//--------------------------------------------------------------------
//       CArray1D:: SetAt
//
template<class T>
bool
CArray1D<T>::SetAt( int Index, const T &Element )
{
  if ( !AssureBoundaries( Index, Index ) )
    return false;

  m_pElements[Index] = Element;

  return true;
}  // SetAt


//--------------------------------------------------------------------
//       CArray1D:: SetAll
//
template<class T>
bool
CArray1D<T>::SetAll( const T &Element )
{
  if ( !IsValid() ) 
    return false;

  for (int i = m_MinIndex; i <= m_MaxIndex; i++) {
    m_pElements[i] = Element;
  }

  return true;
}  // SetAll


//--------------------------------------------------------------------
//       CArray1D:: getUsedMem
//
/*
long
CArray1D<T>::GetUsedMem() const
{
  if ( !IsValid() ) 
    return 0;

  long mem = count * sizeof( CContainerItem ) + sizeof( CArray1D );

  for ( int i=0; i<count; i++ )
    mem+= p[i]-> GetUsedMem();

  return mem;
} // getUsedMem
*/


//--------------------------------------------------------------------
//       CArray1D:: SetBoundaries
//
template<class T>
bool
CArray1D<T>::SetBoundaries( int MinIndex, int MaxIndex )
{
  if (MinIndex > MaxIndex) {
    int tmp = MinIndex;
    MinIndex = MaxIndex;
    MaxIndex = tmp;
  }

  int NewMinIndex = MinIndex;
  int NewMaxIndex = MaxIndex;

  //-Trace( 6, "Required Boundaries are: " << NewMinIndex << " " << NewMaxIndex );
  //-Trace( 6, "Allocated is: " << m_MinAlloc << " " << m_MaxAlloc );


  int NewMinAlloc;
  int NewMaxAlloc;
  int CurrentSize = Size();

  if ( CurrentSize == 0 ) {
	//cerr << "First Allocation" << endl;
	NewMinAlloc = NewMinIndex;
	NewMaxAlloc = NewMaxIndex;
  } else {

	if ( NewMinIndex < m_MinAlloc ) {
	  NewMinAlloc = NewMinIndex - (int)(CurrentSize * 0.1 + 1);
	} else {
	  NewMinAlloc = m_MinAlloc;
	}
	
	if ( NewMaxIndex > m_MaxAlloc ) {
	  NewMaxAlloc = NewMaxIndex + (int)(CurrentSize * 0.1 + 1);
	} else {
	  NewMaxAlloc = m_MaxAlloc;
	}
  }

  //cerr << "Required Boundaries: " << NewMinIndex << " " << NewMaxIndex << endl;
  //cerr << "To Allocate is now:  " << NewMinAlloc << " " << NewMaxAlloc << endl;
  //cerr << "increment was: " << (int)(CurrentSize * 0.1 + 1) << endl;


  //-Trace( 6, "To Allocate is now: " << NewMinAlloc << " " << NewMaxAlloc );

  if ( (NewMinAlloc > NewMinIndex) || (NewMaxAlloc < NewMaxIndex) ) {
    //-Trace( 6, "Required Boundaries are: " 
	//-   << NewMinIndex << " " << NewMaxIndex );
    //-Trace( 6, "To Allocate is now: " 
	//-   << NewMinAlloc << " " << NewMaxAlloc );
    
    assert( NewMinAlloc <= NewMinIndex );
    assert( NewMaxAlloc >= NewMaxIndex );
  }
    
  // get new memory
  T *pNewData = new T[NewMaxAlloc - NewMinAlloc + 1];
    
  if ( !pNewData ) {
    Trace( 0, "CArray1D: cannot allocate %i elements\n"
	  << NewMaxIndex - NewMinIndex + 1 );
    Invalidate();
    return false;
  }
    
  // for 'numerical' array, initialize to 0
  if (m_InitToZero) {
    memset( pNewData, 0, (NewMaxAlloc - NewMinAlloc + 1) * sizeof(T) );
  }

  // set pointer to element with index 0
  T *pNewZero = pNewData - NewMinAlloc;
    
  // which elements need to be copied? only in the overlapping area
  int CopyMinIndex = NewMinIndex > m_MinIndex ? NewMinIndex : m_MinIndex;
  int CopyMaxIndex = NewMaxIndex < m_MaxIndex ? NewMaxIndex : m_MaxIndex;

  int i;
  for ( i = CopyMinIndex; i <= CopyMaxIndex; i++ ) {
    //Trace0(5, "Copy " << i << " " << m_pElements[i]);
    //pNewZero[i] = m_pElements[i];
    pNewData[i - NewMinAlloc] = m_pElements[i];
    //Trace(5, "    " << (i - NewMinAlloc) << " " << pNewData[i - NewMinAlloc]
    //	  << "  " << pNewZero[i]);
    //for ( int j = NewMinAlloc; j <= NewMaxAlloc; j++ ) {
    //  Trace(5, "New " << j << " " << pNewZero[j]);
    //}
  }    
    
  // delete old memory
  delete [] m_pData;

  // set pointers to new memory
  m_pData = pNewData;
  m_pElements = pNewZero;

  m_MinAlloc = NewMinAlloc;
  m_MaxAlloc = NewMaxAlloc;
  
  //-Trace( 6, "Allocated is now: " << m_MinAlloc << " " << m_MaxAlloc );
  //Trace(3, "At Index 0: " << m_pElements[0]);
  
  m_MinIndex = NewMinIndex;
  m_MaxIndex = NewMaxIndex;

  //-Trace( 6, "Boundaries are now: " << m_MinIndex << " " << m_MaxIndex );

  return true;
}


//--------------------------------------------------------------------
//       CArray1D:: AssureBoundaries
//
template<class T>
bool
CArray1D<T>::AssureBoundaries( int MinIndex, int MaxIndex )
{
  //-Trace( 5, "AssureBoundaries " << MinIndex << " " << MaxIndex );

  if (MinIndex > MaxIndex) {
    int tmp = MinIndex;
    MinIndex = MaxIndex;
    MaxIndex = tmp;
  }

  //if ( !IsValid() )
  //  return FALSE;

  // in case boundaries are already large enough
  if ( (MinIndex >= m_MinIndex) && (MaxIndex <= m_MaxIndex) )
    return true;

  int NewMinIndex;
  int NewMaxIndex;
  
  if ( m_MaxIndex < m_MinIndex ) {
    NewMinIndex = MinIndex;
    NewMaxIndex = MaxIndex;
  } else {
    NewMinIndex = MinIndex < m_MinIndex ? MinIndex : m_MinIndex;
    NewMaxIndex = MaxIndex > m_MaxIndex ? MaxIndex : m_MaxIndex;
  }

  //-Trace( 6, "Required Boundaries are: " << NewMinIndex << " " << NewMaxIndex );
  //-Trace( 6, "Allocated is: " << m_MinAlloc << " " << m_MaxAlloc );

  // in case allocated memory is sufficient
  if ( (NewMinIndex < m_MinAlloc) || (NewMaxIndex > m_MaxAlloc) ) {
  
    //-Trace( 6, "Boundaries not large enough - have to reallocate" );
    
    int NewMinAlloc;
    int NewMaxAlloc;
	int CurrentSize = Size();

	if ( CurrentSize == 0 ) {
	  //cerr << "First Allocation" << endl;
	  NewMinAlloc = NewMinIndex;
	  NewMaxAlloc = NewMaxIndex;
	} else {

	  if ( NewMinIndex < m_MinAlloc ) {
		NewMinAlloc = NewMinIndex - (int)(CurrentSize * 0.1 + 1);
	  } else {
		NewMinAlloc = m_MinAlloc;
	  }

	  if ( NewMaxIndex > m_MaxAlloc ) {
		NewMaxAlloc = NewMaxIndex + (int)(CurrentSize * 0.1 + 1);
	  } else {
		NewMaxAlloc = m_MaxAlloc;
	  }
	}

	//cerr << "Required Boundaries: " << NewMinIndex << " " << NewMaxIndex << endl;
	//cerr << "To Allocate is now:  " << NewMinAlloc << " " << NewMaxAlloc << endl;
	//cerr << "increment was: " << (int)(CurrentSize * 0.1 + 1) << endl;
	
    //-Trace( 6, "Required Boundaries are: " 
	//-   << NewMinIndex << " " << NewMaxIndex );
    //-Trace( 6, "To Allocate is now: " 
	//-   << NewMinAlloc << " " << NewMaxAlloc );

    if ( (NewMinAlloc > NewMinIndex) || (NewMaxAlloc < NewMaxIndex) ) {
	  // something went wrong in the calculation
      cerr << "Required Boundaries are: " << NewMinIndex << " " << NewMaxIndex << endl;
	  cerr << "To Allocate is now: " << NewMinAlloc << " " << NewMaxAlloc << endl;
      
      assert( NewMinAlloc <= NewMinIndex );
      assert( NewMaxAlloc >= NewMaxIndex );
    }

    // if lower boundary is changed all the elements have to be shifted to the right
    // int Displacement = m_MinAlloc - NewMinAlloc;

    // get new memory
    T *pNewData = new T[ NewMaxAlloc - NewMinAlloc + 1 ];
    
    if ( !pNewData ) {
      cerr <<  "CArray1D: can not allocate memory for " << NewMaxIndex - NewMinIndex + 1 << " elements" << endl;
      Invalidate();
      return false;
    }
    
	// for 'numerical' array initialize to 0
    if (m_InitToZero) {
      memset( pNewData, 0, (NewMaxAlloc - NewMinAlloc + 1) * sizeof(T) );
    }

	// set pointer to element at index 0
    T *pNewZero = pNewData - NewMinAlloc;
    
	// copy elements
    int i;
    for ( i = m_MinIndex; i <= m_MaxIndex; i++ ) {
      //-Trace0( 8, "Copy " << i << " " );
      //Trace0(5, "Copy " << i << " " << m_pElements[i]);
      //pNewZero[i] = m_pElements[i];
      pNewData[i - NewMinAlloc] = m_pElements[i];
      //Trace(5, "    " << (i - NewMinAlloc) << " " << pNewData[i - NewMinAlloc]
      //    << "  " << pNewZero[i]);
      //for ( int j = NewMinAlloc; j <= NewMaxAlloc; j++ ) {
	//Trace(5, "New " << j << " " << pNewZero[j]);
      //}
    
    }
    
	// delete old memory
    delete [] m_pData;

	// set pointer to new memory
    m_pData = pNewData;
    m_pElements = pNewZero;
    
    m_MinAlloc = NewMinAlloc;
    m_MaxAlloc = NewMaxAlloc;

    //-Trace( 6, "Allocated is now: " << m_MinAlloc << " " << m_MaxAlloc );
    //Trace(5, "At Index 0: " << m_pElements[0]);
  }

  m_MinIndex = NewMinIndex;
  m_MaxIndex = NewMaxIndex;

  //-Trace( 6, "Boundaries are now: " << m_MinIndex << " " << m_MaxIndex );

  return true;
} // AssureBoundaries




//--------------------------------------------------------------------
//       CArray1D:: Invalidate
//
template<class T>
void
CArray1D<T>::Invalidate()
{
  if ( !IsValid() )
    return;

  Clear();
  //count = -1;
} // Invalidate


//--------------------------------------------------------------------
//       CArray1D:: Init
//
template<class T>
void
CArray1D<T>::Init()
{
  m_MinIndex = 0;
  m_MaxIndex = -1;     // MaxIndex < MinIndex -> invalid array
  m_MinAlloc = 0;
  m_MaxAlloc = -1;
  m_pData = NULL;      // no memory allocated so far
  m_pElements = NULL;  // no element with index 0
  m_Increment = 8;
  m_IsStatic = false;  // default is dynamic array
  m_InitToZero = false;
} // Init


//--------------------------------------------------------------------
//       CArray1D:: Clear
//
template<class T>
void
CArray1D<T>::Clear()
{
  //cout << "CArray1D<T>::Clear()" << endl;

  delete [] m_pData;
  //Init();
  m_MinIndex = 0;
  m_MaxIndex = -1;     // MaxIndex < MinIndex -> invalic array
  m_MinAlloc = 0;
  m_MaxAlloc = -1;
  m_pData = NULL;      // no memory allocated
  m_pElements = NULL;  // no element with index 0
} // Clear


//--------------------------------------------------------------------
//       CArray1D:: RemoveAll
//
template<class T>
void
CArray1D<T>::RemoveAll()
{
  m_MinIndex = 0;
  m_MaxIndex = -1;
  m_MinAlloc = 0;
  m_MaxAlloc = -1;
  m_pElements = m_pData;
} // RemoveAll



//--------------------------------------------------------------------
//       CArray1D:: operator <<
//
#ifndef __NO_STREAMS__

template <class T>
ostream &
operator << ( ostream &OutStream, const CArray1D<T> &Array )
{
  int i;
  
  OutStream << "[";
  if( Array.Size() != 0 ) 
    OutStream << Array[Array.MinIndex()];
  for( i = Array.MinIndex() + 1; i <= Array.MaxIndex(); i++ ) {
    OutStream << ", " << Array[i];
  } // for cont
  OutStream << "]";
  return OutStream;
} // operator <<

#endif


#ifndef __NO_STREAMS__
template<class T>
void
CArray1D<T>::Display() const
{
  for ( int j = m_MinIndex; j <= m_MaxIndex; j++ ) {
    cout << "Array[ " << j << " ] = " << m_pElements[j] << endl;
  }
  cout << flush;
}
#endif


template<class T>
void
CArray1D<T>::Debug() const
{
  cout << "MinIndex: " << m_MinIndex << endl;
  cout << "MaxIndex: " << m_MaxIndex << endl;
  cout << "MinAlloc: " << m_MinAlloc << endl;
  cout << "MaxAlloc: " << m_MaxAlloc << endl;
  cout << "Increment: " << m_Increment << endl;
  cout << "IsStatic: " << (m_IsStatic ? "True" : "False") << endl;

  //  for ( int j = m_MinIndex; j <= m_MaxIndex; j++ ) {
  //    cout << "Array[ " << j << " ] = " << m_pElements[j] << endl;
  //}
  cout << flush;
}



//////////////////////////////////////////////////////////////////////////
//
// CArray2D
//
//////////////////////////////////////////////////////////////////////////

//--------------------------------------------------------------------
//       CArray2D inlines
//

//--------------------------------------------------------------------
//       CArray2D::CArray2D
//
template<class T>
inline
CArray2D<T>::CArray2D()
{
  Init();
}


//--------------------------------------------------------------------
//       CArray2D::~CArray2D
//
template<class T>
inline
CArray2D<T>::~CArray2D()
{
  Clear();
}


//--------------------------------------------------------------------
//       CArray2D::CArray2D - Copy Constructor
//
template<class T>
inline
CArray2D<T>::CArray2D( const CArray2D<T> &Array )
{
  Init();
  *this = Array;
}


//--------------------------------------------------------------------
//       CArray2D::GetMinIndex1
//
template<class T>
inline int
CArray2D<T>::MinIndex1() const
{
  //return m_MinIndex;
  return m_Data.MinIndex();
}

//--------------------------------------------------------------------
//       CArray2D::GetMaxIndex1
//
template<class T>
inline int
CArray2D<T>::MaxIndex1() const
{
  //return m_MaxIndex;
  return m_Data.MaxIndex();
}


//--------------------------------------------------------------------
//       CArray2D::SetIncrement
//
template<class T>
inline void
CArray2D<T>::SetIncrement(int Increment)
{
  //if (Increment < 1)
  //  Increment = 1;
  //m_Increment = Increment;
  m_Data.SetIncrement( Increment );
}


//--------------------------------------------------------------------
//       CArray2D::MakeStatic
//
template<class T>
inline void
CArray2D<T>::MakeStatic()
{
  //m_IsStatic = true;
  m_Data.MakeStatic();
}


//--------------------------------------------------------------------
//       CArray2D::MakeDynamic
//
template<class T>
inline void
CArray2D<T>::MakeDynamic()
{
  //m_IsStatic = false;
  m_Data.MakeDynamic();
}


//--------------------------------------------------------------------
//       CArray2D::GetRow
//
template<class T>
inline CArray1D<T> &
CArray2D<T>::GetRow( int Index )
{
  //-Trace(5, "GetRow");
  return m_Data[ Index ];
}



//inline template<class T> CArray1D &
//CArray1D::operator+=( CContainerItem *arg )
//{
//  Append( arg );
//  return *this;
//}

//inline template<class T> CArray1D &
//CArray1D::operator<<( CContainerItem *arg )
//{
//  Append( arg );
//  return *this;
//}


//--------------------------------------------------------------------
//       CArray2D::IsValidIndex
//
template<class T>
inline bool 
CArray2D<T>::IsValidIndex( int Index1, int Index2 ) const
{
  //if (Index1 < m_MinIndex || Index1 > m_MaxIndex)
  //  return FALSE;

  //return (*m_pData)[Index1].IsValidIndex(Index2);

  if ( !m_Data.IsValidIndex( Index1 ) )
    return false;

  return m_Data[ Index1 ].IsValidIndex( Index2 );
} // IsValidIndex


//--------------------------------------------------------------------
//       CArray2D::IsValid
//
template<class T>
inline bool
CArray2D<T>::IsValid() const
{
  return m_Data.IsValid();
} // isValid


//--------------------------------------------------------------------
//       CArray2D::GetSize
//
//template<class T>
//inline int
//CArray2D<T>::GetSize() const
//{
//  if ( !IsValid() ) {
//    return 0;
//  }
//  return m_MaxIndex - m_MinIndex + 1;
//} // GetSize



//--------------------------------------------------------------------
//       CArray2D non inlines
// 

//--------------------------------------------------------------------
//       CArray2D :: operator =
// 
template<class T>
CArray2D<T> &
CArray2D<T>::operator = ( const CArray2D<T> &Array )
{
  if ( this == &Array ) 
    return *this;

  if ( !Array.IsValid() ) {
    Invalidate();
    return *this;
  }
   
  Clear();

  m_Data = Array.m_Data;

  return *this;
} // op =

  
//--------------------------------------------------------------------
//       CArray1D :: operator const char *
// 
/*
template<class T>
CArray1D<T>::operator const char* () const
{
  serBuf = "[\n";
  if ( IsValid() )
    for ( int i = 0; i < count; i++ ) {
      serBuf = (const char*) *p[i];
      serBuf += " ";
      serBuf += "\n";
    }
  else
    serBuf = invalidStr;

  serBuf += "]";
  return serBuf;
} // op const char *
*/


//--------------------------------------------------------------------
//       CArray1D :: tostream
//
/* 
bool
CArray1D<T>::ToStream( ostream &ostr ) const
{
  ostr << "[" << endl;
  if ( IsValid() )
    for ( int i = 0; i < count; i++ )
      ostr << *p[i] << endl;

  else
    ostr << invalidStr;

  ostr << "]" << endl;
  return ostr.good();
} // toStream
*/


//--------------------------------------------------------------------
//       CArray2D:: operator []
//
template<class T>
T&
CArray2D<T>::operator() ( int Index1, int Index2 )
{
  /*
  Trace(5, "operator[] " << Index1 << " " << Index2);
  //AssureSize( Index + 1 );
  if (IsValidIndex( Index1 ) == false)
    AssureBoundaries( Index1 );

  if (IsValidIndex( Index1, Index2 ) == false)
    AssureBoundaries( Index1, Index2 );
  return *(m_pElements[ Index1 ])[ Index2 ];
  */
  return m_Data[Index1][Index2];
} // op []


//--------------------------------------------------------------------
//       CArray2D :: operator []
// 

template<class T>
const T&
CArray2D<T>::operator() ( int Index1, int Index2 ) const
{
  return m_Data[Index1][Index2];
} // op []


//--------------------------------------------------------------------
//       CArray1D:: append
//
/*
template<class T>
bool
CArray1D<T>::Append( T& arg )
{
  if ( !IsValid() )
    return FALSE;

  int insertPos = count;
  if ( !AssureSize( insertPos+1 ) )
    return FALSE;

  p[ insertPos ] = arg;
  return TRUE;
} // append
*/

//--------------------------------------------------------------------
//       CArray1D:: remove
//
/*
CContainerItem *
CArray1D<T>::Remove( int index )
{
  if ( !IsValid() || !IsValidIndex( index ) )
     return new NoData;

  CContainerItem *q = p[ index ];
  p[ index ] = new NoData;
  return q;

} // remove
*/

//--------------------------------------------------------------------
//       CArray2D:: GetAt
//
template<class T>
const T &
CArray2D<T>::GetAt( int Index1, int Index2 ) const
{
  /*
  if ( !IsValidIndex( Index1, Index2 ) ) 
    //return noDataPtr;
    assert(0);

  return *(m_pElements[ Index1 ])[ Index2 ];
  */

  return m_Data[ Index1 ].GetAt( Index2 );
} // GetAt


//--------------------------------------------------------------------
//       CArray2D:: SetAt
//

template<class T>
bool
CArray2D<T>::SetAt( int Index1, int Index2, const T &Element )
{
  /*
  if ( !IsValidIndex(  int Index1, int Index2 ) ) 
    return FALSE;

  //delete m_pData[Index];
  *(m_pElements[ Index1 ])[ Index2 ] = Element;

  return TRUE;
  */

  return m_Data[ Index1 ].SetAt( Index2, Element );
}  // SetAt


//--------------------------------------------------------------------
//       CArray2D:: SetAll
//

template<class T>
bool
CArray2D<T>::SetAll( const T &Element )
{
  if ( !IsValid() ) 
    return false;

  bool Result = true;

  for (int i = m_Data.MinIndex(); i <= m_Data.MaxIndex(); i++)
    Result = Result && m_Data[i].SetAll( Element );

  return Result;
}  // SetAll


//--------------------------------------------------------------------
//       CArray2D:: SetBoundaries
//
template<class T>
bool
CArray2D<T>::SetBoundaries( int MinIndex, int MaxIndex )
{
  return m_Data.SetBoundaries( MinIndex, MaxIndex );
}


template<class T>
bool
CArray2D<T>::SetBoundaries( int MinIndex1, int MaxIndex1,
			    int MinIndex2, int MaxIndex2 )
{
  bool Result = true;

  if ( m_Data.SetBoundaries( MinIndex1, MaxIndex1 ) ) {
    for ( int i = m_Data.MinIndex(); i <= m_Data.MaxIndex(); i++) {
      Result = Result && m_Data[i].SetBoundaries( MinIndex2, MaxIndex2 );
    }
  } else {
    Result = false;
  }

  return Result;
}



//--------------------------------------------------------------------
//       CArray2D:: AssureBoundaries
//
template<class T>
bool
CArray2D<T>::AssureBoundaries( int MinIndex, int MaxIndex )
{
  return m_Data.AssureBoundaries( MinIndex, MaxIndex );
} // assureSize


//--------------------------------------------------------------------
//       CArray2D:: Invalidate
//
template<class T>
void
CArray2D<T>::Invalidate()
{
  //if ( !IsValid() )
  //  return;

  //Clear();
  //count = -1;

  m_Data.Invalidate();
} // Invalidate


//--------------------------------------------------------------------
//       CArray2D:: Init
//
template<class T>
void
CArray2D<T>::Init()
{
  //m_MinIndex = 0;
  //m_MaxIndex = -1;
  //m_MinAlloc = 0;
  //m_MaxAlloc = -1;
  //m_pData = NULL;
  //m_pElements = NULL;
  //m_Increment = 8;
  m_InitToZero = false;
} // Init


//--------------------------------------------------------------------
//       CArray2D:: Clear
//
template<class T>
void
CArray2D<T>::Clear()
{
  m_Data.Clear();
} // Clear



template<class T>
void
CArray2D<T>::Display() const
{
  //for ( int j = m_MinIndex; j <= m_MaxIndex; j++ ) {
  //  cout << "Array[ " << j << " ] = " << m_pElements[j] << endl;
  //}
  //cout << flush;
  m_Data.Display();
}


template<class T>
void
CArray2D<T>::Debug() const
{
  m_Data.Debug();
}



