#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef NumericalArray_hh
#define NumericalArray_hh


// CNumericalArray
//

//========================= Definitionsabhngigkeiten ===============

#include <iostream>
#include <Types.hh>
#include <Array.hh>


//========================= Export ==================================

// zweiter Name
#define CNumericalArray  CNumericalArray1D
#define CNumericalMatrix  CNumericalArray2D

template <class T> class CNumericalArray1D;
template <class T> class CNumericalArray2D;


//--------------------------------------------------------------------
//       class CNumericalArray1D
//

template <class T>
class CNumericalArray1D : public CArray1D<T>
{

public:

  CNumericalArray1D();  // erzeugt leeres Feld
  ~CNumericalArray1D(); 

  //-- Kopierkonstruktor und Zuweisungsoperator erzeugen eine flache Kopie,
  //   d.h. bei Pointer-Arrays werden nur die Pointer, nicht die Objekte 
  //   kopiert
  //CNumericalArray1D              ( const CNumericalArray1D<T> & Array );
  //CNumericalArray1D & operator = ( const CNumericalArray1D<T> & Array );

  // compare two arrays
  // Array A is smaller than Array B if
  // A[ n ] == B[ n ] for i = MinIndex to n, but A[ n+1 ] < B[ n+1 ]
  // or A is shorter than B
  // similar for > and ==

  inline bool operator == ( const CNumericalArray1D<T> & Array ) const;
  inline bool operator < ( const CNumericalArray1D<T> & Array ) const;
  inline bool operator > ( const CNumericalArray1D<T> & Array ) const;


  inline void ClearDelete();

  // Add Value to each value in the array, ditto for Sub, Mul, Div
  inline void Add( const T Value );
  inline void Sub( const T Value );
  inline void Mul( const T Value );
  inline void Div( const T Value );

  inline void Inc( int Index );
  inline void Dec( int Index );

  inline T Sum() const;

  inline void Log();

  inline CNumericalArray operator +( const T Value );

};





//--------------------------------------------------------------------
//       class CNumericalArray2D
//

template <class T>
class CNumericalArray2D : public CArray2D<T>
{

public:

  CNumericalArray2D();  // erzeugt leeres Feld
  ~CNumericalArray2D(); 

  //-- Kopierkonstruktor und Zuweisungsoperator erzeugen eine flache Kopie,
  //   d.h. bei Pointer-Arrays werden nur die Pointer, nicht die Objekte 
  //   kopiert
  //CNumericalArray1D              ( const CNumericalArray1D<T> & Array );
  //CNumericalArray1D & operator = ( const CNumericalArray1D<T> & Array );

  CNumericalArray1D<T> & GetRow( int Index );

  // Add Value to each value in the array, ditto for Sub, Mul, Div
  //void Add( const T Value );
  //void Sub( const T Value );
  //void Mul( const T Value );
  //void Div( const T Value );

  inline void Inc( int Index1, int Index2 );
  //void Dec( int Index );
  //void Log();

  //CNumericalArray operator +( const T Value );


private:

  //CArray1D< CNumericalArray1D<T> > m_Data; //-- Array von Arrays (Zeilen)

};







//========================= Inlines ===============================
//
//--------------------------------------------------------------------
//       CNumericalArray1D inlines
//


//--------------------------------------------------------------------
//       CNumericalArray1D::CNumericalArray1D
//

template<class T>
inline
CNumericalArray1D<T>::CNumericalArray1D()
{
  this->m_InitToZero = true;
}



//--------------------------------------------------------------------
//       CNumericalArray1D::~CNumericalArray1D
//

template<class T>
inline
CNumericalArray1D<T>::~CNumericalArray1D()
{
}



//--------------------------------------------------------------------
//       CNumericalArray1D::CNumericalArray1D - Copy Constructor
//
/*
template<class T>
inline
CNumericalArray1D<T>::CNumericalArray1D( const CNumericalArray1D<T> &Array )
{
  Init();
  *this = Array;
}  // ctor
*/

//--------------------------------------------------------------------
//       CNumericalArray1D::CNumericalArray1D - Copy Constructor
//
template<class T>
void inline
CNumericalArray1D<T>::ClearDelete()
{
  for ( int i = this->MinIndex(); i <= this->MaxIndex(); i++ ) {
    delete this->Array[ i ];
    this->Array[ i ] = NULL;
  }
  this->Clear();
}




//--------------------------------------------------------------------
//       CNumericalArray2D::CNumericalArray2D
//

template<class T>
inline
CNumericalArray2D<T>::CNumericalArray2D()
{
}



//--------------------------------------------------------------------
//       CNumericalArray2D::~CNumericalArray2D
//

template<class T>
inline
CNumericalArray2D<T>::~CNumericalArray2D()
{
}




// alles inline wegen Instantiierung
#include <NumericalArray_inl.hh>

#endif
