#include <stdlib.h>
#ifndef SortedArray_hh
#define SortedArray_hh


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


//========================= Dokumentation ===========================
//
// CSortedArray realisiert ein sortiertes Array. D.h. es f"ugt 
// Elemente entsprechend einer auf diesen Elemente definierten
// Ordnungsrelation ein. F"ur die Elemente m"ussen die Operatoren
// > und == definiert sein.
// Ein CSortedArray ist immer dynamisch. Es k"onnen auch keine
// Grenzen gesetzt werden. Die Grenzen sind gegeben durch 0 und
// Anzahl der gespeicherten Elemente minus Eins.
// Schreibender Zugriff ist nur "uber die Insert-Methode m"oglich.
// Lesender Zugriff auch "uber GetAt und operator[].
//
// Beachte: z.Z. ist CSortedArray ein CSortedUniqArray, d.h. keine
// Wiederholung des Keys ist m"oglich. Sortiertes Array mit 
// Wiederholung muss noch gamacht werden.
//

//========================= Definitionsabh"angigkeiten ===============

#include <iostream>
#include <limits.h>


//#include "Bool.hh"
#include <Array.hh>


//========================= Export ==================================

// zweiter Name
#define CSortedArray  CSortedArray1D


//--------------------------------------------------------------------
//       class CArray1D
//

template <class T>
class CSortedArray1D : public CArray1D<T>
{

public:

  CSortedArray1D();  // erzeugt leeres Feld
  //~CSortedArray1D(); 

  //! sorting order; small-to-large is default
  void SortSmallToLarge();
  void SortLargeToSmall();

  // Setze maximale Groesse f"ur Array. Beim Einf"ugen eines
  // neuen Elementes in ein gef"lltes Array geht das letzte
  // Element verlohren
  void SetMaxSize( int MaxSize );

  // Element wird entsprechend seiner Gr"osse in das Array
  // eingef"ugt. D.h. f"ur Element m"ussen die Vergleichs-
  // Operatoren definiert sein. Wenn ein gleichgrosses Element
  // bereits existiert (Element = {Key, Data}), so wird es
  // "uberschrieben, ansonsten wird das Array vergr"ossert
  // und das Element neu eingef"ugt. Der Parameter bool WasNew 
  // wird in diesem Fall auf TRUE gesetzt. Der R"uckgabe-Parameter
  // ist immer die Position des eingef"ugten Elementes
  int Insert( const T & Element, bool & WasNew );

  // Testen, ob ein Element im Array vorhanden ist.
  // Wenn ja, so ist der R"uckgabewert TRUE und Index
  // gibt die Position an. Wenn nicht, dann ist der
  // R"uckgabewert false und Index ist nicht gesetzt.
  bool Find( const T & Element, int & Index ) const;

  void SortReverse( bool ReverseSort = true );

private:

  // alle Zugriffsfunkionen ausblenden, die Sortierung
  // st"oren w"urden
  bool SetMinIndex( int MinIndex );
  bool SetMaxIndex( int MaxIndex );
  bool SetBoundaries( int MinIndex, int MaxIndex );
  bool AssureBoundaries( int MinIndex, int MaxIndex );
  bool SetAt( int Index, const T & Element );
  T & operator[] ( int Index );
  bool Append( const T & Element);

  bool m_SortReverse;

  int m_MaxSize;
};


//--------------------------------------------------------------------
//       CSortedArray1D::CSortedArray1D
//
template<class T>
inline
CSortedArray1D<T>::CSortedArray1D()
{
  // Trace( 3, "CSortedArray1D Konstructor" );
  //Init();
  m_SortReverse = false;
  m_MaxSize = INT_MAX;
} // ctor


/*

//--------------------------------------------------------------------
//       CSortedArray1D::~CSortedArray1D
//
template<class T>
inline
CSortedArray1D<T>::~CSortedArray1D()
{
  // Trace( 1, "CSortedArray1D Destructor" );
  Clear();
} // dtor
*/


//--------------------------------------------------------------------
//       CSortedArray1D::SortReverse
//
template<class T>
inline void
CSortedArray1D<T>::SortReverse( bool SortReverse /*= TRUE */)
{
  m_SortReverse = SortReverse;
}



//--------------------------------------------------------------------
//       CSortedArray1D::SortSmallToLarge
//
template<class T>
inline void
CSortedArray1D<T>::SortSmallToLarge()
{
  if ( this->Size() > 0 ) {
    std::cerr << "Sorting can not changed after elements have been added to CSortedArray1D" << std::endl;
    exit( -1 );
  }

  m_SortReverse = false;
}



//--------------------------------------------------------------------
//       CSortedArray1D::SortLargeToSmall
//
template<class T>
inline void
CSortedArray1D<T>::SortLargeToSmall()
{
  if ( this->Size() > 0 ) {
    std::cerr << "Sorting can not changed after elements have been added to CSortedArray1D" << std::endl;
    exit( -1 );
  }

  m_SortReverse = true;
}



//--------------------------------------------------------------------
//       CSortedArray1D::SetMaxSize
//
template<class T>
inline void
CSortedArray1D<T>::SetMaxSize( int MaxSize )
{
  m_MaxSize = MaxSize;
}



//--------------------------------------------------------------------
//       CSortedArray1D::Insert
//
template<class T>
inline int
CSortedArray1D<T>::Insert( const T & Element, bool & WasNew /*= TRUE */)
{
  // Trace( 3, "CSortedArray1D Insert " << Element );
  // Trace( 5, "MinIndex = " << m_MinIndex << " MaxIndex = " << m_MaxIndex );

  int Left = this->m_MinIndex;
  int Right = this->m_MaxIndex;
  int Mid;
  int Pos;

  bool Found = false;
  bool Larger = false;
  bool Result = true;
  bool ElemIsLarger;

  if (Right < Left) {
    // Trace(5, "Empty Array");
    CArray1D<T>::SetAt( 0, Element );
    WasNew = true;
    return 0;
  }

  do {
    Mid = (Right + Left) / 2;
    // Trace(5, "Left = " << Left << "  Mid = " << Mid << "  Right = " << Right );

    if (Element == this->m_pElements[Mid]) {
      // Trace( 5, "Element exists: " << Element << " " 
      //	     << m_pElements[Mid] );
      Found = true;
    } else {
      ElemIsLarger = (Element > this->m_pElements[Mid]);
      if ( m_SortReverse ) {
	ElemIsLarger = !ElemIsLarger;
      }

      if (ElemIsLarger) {
	Larger = true;
	Left = Mid + 1;
      } else {
	Larger = false;
	Right = Mid - 1;
      }
    }
  } while ( (Left <= Right) && (Found == false) );


  if (Found) {
    WasNew = false;
    // Trace( 5, "Element exists already: " << Element << " " 
    //	   << m_pElements[Mid] );
    this->m_pElements[Mid] = Element;
    Pos = Mid;
  } else {
    WasNew = true;
    /*
    if ( m_SortReverse ) {
      // Trace( 5, "Sort Reverse" );
    } else {
      // Trace( 5, "Sort Straight" );
    }
    */
    if (Larger) {
      Result = CArray1D<T>::Insert( Mid + 1, Element );
      Pos = Mid + 1;
    } else {
      Result = CArray1D<T>::Insert( Mid, Element );
      Pos = Mid;
    }
    if ( this->Size() > this->m_MaxSize ) {
      this->m_MaxIndex--;
    }
  }

  return Pos;
}


//--------------------------------------------------------------------
//       CSortedArray1D::Find
//
template<class T>
inline bool
CSortedArray1D<T>::Find( const T & Element, int & Index) const
{
  // Trace( 3, "CSortedArray1D Find " << Element );
  // Trace( 3, "MinIndex = " << m_MinIndex << " MaxIndex = " << m_MaxIndex );

  int Left = this->m_MinIndex;
  int Right = this->m_MaxIndex;
  int Mid;

  bool Found = false;

  if (Right < Left) {
    // Trace(5, "Empty Array");
    return false;
  }

  do {
    Mid = (Right + Left) / 2;
    // Trace(5, "Left = " << Left << "  Mid = " << Mid << "  Right = " << Right );

    if (Element == this->m_pElements[Mid]) {
      // Trace( 5, "Element exists: " << Element << " " 
	  //	     << m_pElements[Mid] );
      Index = Mid;
      return true;
    }

    if (Element > this->m_pElements[Mid]) {
      Left = Mid + 1;
    } else {
      Right = Mid - 1;
    }

  } while ( (Left <= Right) 
	    /*  && ( (Found = (Element == m_pElements[Mid]) ) == false) */ );

  if (Found) {
    // Trace( 5, "Element exists: " << Element << " " 
	//	   << m_pElements[Mid] );
    Index = Mid;
    return true;
  } else {
    return false;
  }
}


#endif
