#include <memory.h>
#include <string.h>
//========================= Kopf ====================================
//
//
//
// 92.08.01 : Version 1.2 (Au)
//

//========================= Implementierungsah"angigkeiten ==========


#include <config.h>


#include <assert.h>
#include <limits.h>
#include <stdio.h> 
#include <stdlib.h> 
#include <ctype.h>
#include <string.h>

#include <String.hh>


//========================= Implementierung =========================


//--------------------------------------------------------------------
//       CString:: CString
//
CString::CString()
{
  //-- initialisiert das Objekt zu einem leeren CString
  //   m_pData, l m"ussen nicht initialisiert sein 

  Init();
  SetEmptyString();
} // ctor


//--------------------------------------------------------------------
//       CString:: CString()
//
CString::CString( char arg )
{
  Init();
  if ( !AssureSize( 2 ) )
    return;

  m_pData[0] = arg;
  m_pData[1] = 0x00;

  return;
} // ctor


//--------------------------------------------------------------------
//       CString:: Konstruktor
//
CString::CString( const char *arg )
{
  Init();
  Set( arg );
} // ctor


//--------------------------------------------------------------------
//       CString:: Kopier-Konstruktor
//
CString::CString( const CString &arg )
{
  Init();
  *this = arg;
} // ctor


//--------------------------------------------------------------------
//       CString:: Zuweisungsoperator
//
CString &
CString::operator = ( const CString &arg )
{
  //-- Kopie auf sich selber kann in allen Fllen ignoriert werden

  if ( this == &arg ) 
    return *this;

  //-- ist das Argument ung"ultig, wird es auch das eigene Objekt

  if ( !arg.IsValid() ) {
    Invalidate();
    return *this;
  }

  //-- Ist das Argument leer, wird eigenes Objekt nat"urlich auch leer.
  //   (Sonderfall Optimierung)

  if ( arg.IsEmptyString() ) {
    SetEmptyString();
    return *this;
  }

  Set( arg.m_pData );
  return *this;
} // op =


//--------------------------------------------------------------------
//       CString:: Zuweisungsoperator
//
CString &
CString::operator = ( const char *arg )
{
  //-- Pr"ufung auf Nullzeiger etc. wird von set selbst"andig durchgef"uhrt
  Set( arg );
  return *this;
} // op =


//--------------------------------------------------------------------
//       CString:: Konstruktor
//
CString::CString( int value, const char *fmt )
{
  Init();

  const char *f = fmt ? fmt : "%d";
  char buf[ 23 ];
  sprintf( buf, f, value );
  Set( buf );
} // ctor


//--------------------------------------------------------------------
//       CString:: Konstruktor
//
CString::CString( long value, const char *fmt )
{
  Init();

  const char *f = fmt ? fmt : "%ld";
  char buf[ 23 ];
  sprintf( buf, f, value );
  Set( buf );
} // ctor


//--------------------------------------------------------------------
//       CString:: Konstruktor
//
CString::CString( float value, const char *fmt )
{
  Init();

  const char *f = fmt ? fmt : "%f";
  char buf[ 23 ];
  sprintf( buf, f, value );
  Set( buf );
} // ctor


//--------------------------------------------------------------------
//       CString:: Konstruktor
//
CString::CString( double value, const char *fmt )
{
  Init();

  const char *f = fmt ? fmt : "%lf"; 
  char buf[ 23 ];
  sprintf( buf, f, value );
  Set( buf );
} // ctor


//--------------------------------------------------------------------
//       CString:: operator long
//
CString::operator long() const
{
  long value;
  int pos;
  bool result = ToLong( value, pos );

  return result ? value : strnaL;
} // op long


//--------------------------------------------------------------------
//       CString:: operator double
//
CString::operator double() const
{
  double value;
  int pos;
  bool result = ToDouble( value, pos );

  return result ? value : strnaD;
} // op double


//--------------------------------------------------------------------
//       CString:: IsNumeric
//
bool
CString::IsNumeric() const
{
  double value;
  int pos;
  return ToDouble( value, pos );
} // isNumeric
 

//--------------------------------------------------------------------
//       CString:: IsInteger
//
bool
CString::IsInteger() const
{
  long value;
  int pos;
  return ToLong( value, pos );
} // isInteger


//--------------------------------------------------------------------
//       CString:: ToDouble
//

bool
CString::ToDouble( double &value, int &pos ) const
{
  if ( !IsValid() )
    return FALSE;

  char *endptr;
  value = strtod( m_pData, &endptr );

  //-- endptr zeigt auf das letzte konvertierte Zeichen. 
  //   Wenn der CString eine Zahl ist, d"urfen jetzt nur noch 
  //   whitespace kommen. 

  pos = endptr - m_pData;
  while( isspace(*endptr) )
    endptr++ ;  // whitespace ignorieren

  //-- Jetzt m"ussen wir am Ende des Strs sein, sonst Fehler!
  return !(*endptr);
} // toDouble


//--------------------------------------------------------------------
//       CString:: ToLong
//
bool
CString::ToLong( long   &value, int &pos ) const
{
  if ( !IsValid() )
    return FALSE;

  char *endptr;
  value = strtol( m_pData, &endptr, 10 ); // Konvertierung zur Basis 10

  //-- endptr zeigt auf das letzte konvertierte Zeichen. 
  //   Wenn der CString eine Zahl ist, d"urfen jetzt nur noch 
  //   whitespace kommen. 

  pos = endptr - m_pData;
  while( isspace(*endptr) )
    endptr++;  // whitespace ignorieren

  //-- Jetzt m"ussen wir am Ende des Strs sein, sonst Fehler!
  return !(*endptr);
} // toLong


//--------------------------------------------------------------------
//       CString:: operator const char *
//
CString::operator const char*() const
{
  if ( IsValid() )
    return m_pData;

  return InvalidString;
} // op const char *


//--------------------------------------------------------------------
//       CString:: ToStream
//
bool
CString::ToStream( std::ostream &ostr ) const
{
  if ( IsValid() )
    ostr << m_pData;
  else 
    ostr << InvalidString;

  return ostr.good();
} // toStream


//--------------------------------------------------------------------
//       CString:: GetLength
//

int
CString::GetLength() const
{
  if ( !IsValid() ) {
    return 0;
    }
  return m_Size - 1;
} // getLength



#ifdef VALIDITYCHECK

//--------------------------------------------------------------------
//       CString:: operator []
//
char
&CString::operator [] ( int ofs )
{
  if ( checkIndex( ofs ) )
    return m_pData[ ofs ];

  static char buf;
  buf = noData;
  return buf;
} // op []


//--------------------------------------------------------------------
//       CString:: operator []
//
char
CString::operator [] ( int ofs ) const
{
  if ( CheckIndex( ofs ) )
    return m_pData[ ofs ];

  return noData;
} // op []

#endif


//--------------------------------------------------------------------
//       CString:: Compare
//

int
CString::Compare( const CString &arg ) const
{
  if ( !IsValid() || !arg.IsValid() )
    return NotValid;

  //-- wenn Vergleichsmodus Standard ist, kann strcmp/strcoll verwendet
  //   werden. F"ur alle anderen F"alle sind Hilfsstrings
  //   erforderlich

  if ( CompareMode == cmpStandard )
       return strcoll( m_pData, arg.m_pData );

  CString arg1( *this );
  CString arg2( arg );

  //-- Je nach eingeschaltetem Bit die Hilfsstrings manipulieren

  if ( CompareMode & cmpNoCase ) {
    arg1.ToUpperCase();
    arg2.ToUpperCase();
  }

  if ( CompareMode & cmpNoLeading ) {
    arg1.TrimLeadingBlanks();
    arg2.TrimLeadingBlanks();
  }

  if ( CompareMode & cmpNoTrailing ) {
    arg1.TrimTrailingBlanks();
    arg2.TrimTrailingBlanks();
  }

  //-- Trat zwischendurch ein Fehler auf (kein Speicher) ?

  if ( !arg1.IsValid()  || !arg2.IsValid()  ) 
    return NotValid;

  return strcoll( arg1.m_pData, arg2.m_pData );
} // compare


//--------------------------------------------------------------------
//       CString:: Pos
//
int
CString::Pos( const CString &arg ) const
{
  if ( !IsValid() || !arg.IsValid() )
    return NotValid;

  //-- wenn Vergleichsmodus Standard ist, kann strstr verwendet
  //   werden. F"ur alle anderen F"alle sind Hilfsstrings
  //   erforderlich

  char *q;

  if ( CompareMode == cmpStandard ) {
    q = strstr( m_pData, arg.m_pData );
    if (q)
       return q-m_pData;
    else
       return NotFound;
  }

  CString arg1( *this );
  CString arg2( arg );

  //-- Je nach eingeschaltetem Bit die Hilfsstrings manipulieren
  if ( CompareMode & cmpNoCase ) {
    arg1.ToUpperCase();
    arg2.ToUpperCase();
  }

  if ( CompareMode & cmpNoLeading ) {
    arg1.TrimLeadingBlanks();
    arg2.TrimLeadingBlanks();
  }

  if ( CompareMode & cmpNoTrailing ) {
    arg1.TrimTrailingBlanks();
    arg2.TrimTrailingBlanks();
  }

  //-- Trat zwischendurch ein Fehler auf (kein Speicher) ?
  if ( !arg1.IsValid() || !arg2.IsValid() ) 
    return NotValid;

  q = strstr( arg1.m_pData, arg2.m_pData );
  if (q)
    return q-arg1.m_pData;
  else
    return NotFound;

} // pos


//--------------------------------------------------------------------
//       CString:: SubString
//
CString
CString::SubString( int ofs, int l ) const
{
  CString result;
  result.Invalidate();

  if ( !IsValid() )
    return result;

  //-- Vereinbarung: Bei Zugriff auerhalb der Grenzen etc. wird 
  //   der leere CString geliefert.
  result.SetEmptyString();

  //-- Pr"ufung der Argumente auf Zul"assigkeit
  if ( !IsValidIndex( ofs ) )    
    //-- die Parameter spezifizieren einen Teilstring auerhalb
    //   der Grenzen. Leeren Puffer zur"uck.
    return result;

  if ( l > m_Size - ofs ) 
    //-- Das Ende des Teilstrings reicht "uber die rechte Grenze hinaus.
    //   L"ange des Teilstrings anpassen.
    l = m_Size - ofs -1;

  if ( l < 1 ) 
    return result; // Ergebnis ist hier leer

  //-- length und ofs haben jetzt g"ultige Werte. 

  if ( !result.AssureSize( l+1 ) )
    return result; // Ergebnis ist nun ung"ultig

  strncpy( result.m_pData, m_pData+ofs, l );
  result.m_pData[ l ] = 0x00;
  return result;

} // subString




int
CString::Split( CArray< CString > &ResultingStrings, const CString & SplitCharacters ) const
{
  // strtok modifies string, therefore work on copy
  char *TmpStringToBeSplitted = strdup( m_pData );
  CString TmpString;
  char *pCurrentSegment;

  int NumberOfSegments = 0;

  pCurrentSegment = strtok( TmpStringToBeSplitted, SplitCharacters );

  while ( pCurrentSegment != NULL ) {
    //cout << "  next segment " << pCurrentSegment << endl << flush;
    TmpString = CString( pCurrentSegment );
    //cout << TmpString << endl;
    ResultingStrings.Append( TmpString );
    //cout << ResultingStrings[ ResultingStrings.MaxIndex() ] << endl;
    NumberOfSegments++;
    pCurrentSegment = strtok( NULL, SplitCharacters );
  }
  //strdup returns a new'ed pointer, it needs to be freed
  free( TmpStringToBeSplitted );

  /*
  for ( int i = ResultingStrings.MinIndex(); i <= ResultingStrings.MaxIndex(); i++ ) {
    cout << "-" << ResultingStrings[ i ] << " is treated as normal word " << endl;
  }
  */

  return NumberOfSegments;
}


  
  
//--------------------------------------------------------------------
//       CString:: Append
//
bool
CString::Append( const CString &arg )
{
  if ( !IsValid() ) 
    return FALSE;

  if ( !arg.IsValid() ) {
    //-- Regel 2 f"ur Rechnen mit ung"ultigen Objekten.
    Invalidate();
    return FALSE;
  }

  //-- falls das anzuh"angende Objekt wir selber sind, m"ussen 
  //   wir vorher eine Kopie machen, da ja sonst arg durch 
  //   assurem_Size zerstrt w"urde. 
  if ( this == &arg ) {
    CString buffer( arg );
    return Append( buffer ); 
  }
    
  int origSize = m_Size;
  if ( !AssureSize( m_Size + arg.m_Size -1 ) )
    return FALSE;


  strcpy( m_pData + origSize - 1, arg.m_pData );
  return TRUE;
} // append


//--------------------------------------------------------------------
//       CString:: FillTo
//
bool
CString::FillTo( int ofs, char c )
{
  if ( !IsValid() )
    return FALSE;

  if ( ofs <= m_Size - 1 ) 
    //-- CString ist bereits lang genug. TRUE zur"uck, da kein Fehler
    return TRUE;


  int oldSize = m_Size;
  if ( !AssureSize( ofs+1 ) )
    return FALSE;

  memset( m_pData + oldSize - 1, c, ofs - oldSize + 1 );
  m_pData[ m_Size-1 ] = 0x00;
  return TRUE;
} // fillTo

//--------------------------------------------------------------------
//       CString:: Insert
//
bool
CString::Insert( int ofs, const CString &arg )
{
  if ( !IsValid() )
    return FALSE;

  if ( !IsValid() || !arg.IsValid() ) {
    Invalidate();
    return FALSE;
  }

  //-- Einf"ugen eine Stelle hinter rechtem Rand ist explizit erlaubt.
  //   IsValidOffset meldet das als Fehler -> daher vorher
  if ( ofs == m_Size-1 ) 
    return Append( arg );


  if ( !IsValidIndex( ofs ) ) {
    Invalidate();
    return FALSE;
  }

  //-- falls das einzusetzende Objekt wir selber sind, m"ussen
  //   wir zuerst eine Kopie machen. Grund: siehe append.
  if ( this == &arg ) {
    CString buffer( arg );
    return Insert( ofs, buffer );
  }

  //-- Platz machen f"ur ausreichend Zeichen
  int count = arg.m_Size-1;
  int oldm_Size = m_Size;
  if ( !AssureSize( m_Size + count ) )
    return FALSE;

  //-- da sich Quell- und Zielspeicher "uberlappen: 
  //   manuelle Schleife 

  char *src = m_pData + oldm_Size -1;
  char *tgt = src + count;

  for ( int i = oldm_Size-ofs-1; i >= 0; i-- )
    *tgt-- = *src--;
    
  memcpy( m_pData+ofs, arg.m_pData, count );
  return TRUE;
} // insert


//--------------------------------------------------------------------
//       CString:: Delete
//
bool
CString::Delete( int ofs, int l )
{
  if ( !IsValid() ) 
    return FALSE;

  if ( !IsValidIndex( ofs ) )

    //-- Lschen auerhalb der Grenzen bewirkt zwar nichts, ist aber auch
    //   nicht verboten
    return TRUE;

  if ( l >= m_Size - ofs )
    l = m_Size - ofs -1;

  //-- Ist l nun <=0, gibt es nichts zu lschen (wahrscheinlich
  //   ist eigene Instanz leer)

  if ( l <= 0 )
    return TRUE;

  assert( !IsEmptyString() );

  int copyNbr = m_Size - ofs -l;

  int i = 0;
  char *tgt = m_pData + ofs;
  char *src = tgt + l;

  for ( ; i < copyNbr; i++ ) 
    *tgt++ = *src++;

  return AssureSize( m_Size-l );
} // del


//--------------------------------------------------------------------
//   CString:: TrimTo
//
bool
CString::TrimTo( int l )
{
  if ( !IsValid() )
    return FALSE;

  if ( m_Size <= l ) 
    return TRUE;

  return Delete( l, INT_MAX );
} // trimTo


//--------------------------------------------------------------------
//       CString:: ConvertToUpperCase
//
bool
CString::ToUpperCase()
{
  if ( !IsValid() )
    return FALSE;

  for ( char *q = m_pData; *q; q++ )
    switch ((unsigned char)*q) {
       case 228  : *q = (unsigned char)196;   break;    // "a
       case 246  : *q = (unsigned char)214;   break;    // 
       case 252  : *q = (unsigned char)220;   break;    // "u

       default   : if ( islower( *q ) )
                      *q = toupper( *q );
     }

  return TRUE;
} // convertToUpperCase


//--------------------------------------------------------------------
//       CString:: ConvertToLowerCase
//
bool
CString::ToLowerCase()
{
  if ( !IsValid() )
    return FALSE;

  for ( char *q = m_pData; *q; q++ )
    switch ((unsigned char)*q) {
      case 196  : *q = (unsigned char)228;  break;    //  
      case 214  : *q = (unsigned char)246;  break;    //  
      case 220  : *q = (unsigned char)252;  break;    //  

      default   : if ( isupper( *q ) )
                    *q = tolower( *q );
    }

  return TRUE;
} // convertToLowerCase


//--------------------------------------------------------------------
//       CString:: TrimLeadingBlanks
//
bool
CString::TrimLeadingBlanks()
{
  if ( !IsValid() )
    return FALSE;

  if ( IsEmptyString() )
    return TRUE;

  //-- Z"ahlen der Leerzeichen am Anfang

  int i = 0;
  char *q = m_pData;

  for ( ; *q++==0x20; i++  ) ;

  return Delete( 0, i );
} // trimLeadingBlanks


//--------------------------------------------------------------------
//       CString:: TrimTrailingBlanks
//
bool
CString::TrimTrailingBlanks()
{
  if ( !IsValid() )
    return FALSE;

  if ( IsEmptyString() )
    return TRUE;

  //-- Z"ahlen der Leerzeichen am Ende
  char *q = m_pData+m_Size-2;
  int i = 0;

  for( ; i<=m_Size-2 && *q--==0x20; i++ );

  if ( i > 0 )
    return Delete( m_Size-i-1, i );

  return TRUE;
} // trimTrailingBlanks


//--------------------------------------------------------------------
//       CString:: Read
//
void
CString::Read( const char *prompt )
{
  SetEmptyString();

  if( prompt ) 
    printf( prompt );

  //-- Schleife zum Einlesen von Zeichen bis RETURN gedr"uckt wird

  char ch;

  while (TRUE) {
    ch = getchar();      
    if ( ch == '\n' )
       break;
    Append( ch );
  }
} // read


//--------------------------------------------------------------------
//       CString:: Assurem_Size
//
bool
CString::AssureSize( int newSize )
{
  if ( m_Size == newSize )
    return TRUE;

  if ( newSize == 1 ) {
    //-- das Objekt soll der Leerstring werden. 
    SetEmptyString();
    return TRUE;
  }

  assert( newSize > 0 );

  char *q = (char*)malloc( newSize );
  if ( !q ) {
    //-- nicht mehr gen"ugend Speicher!
    printf( "CString: kann keine %i Bytes allokieren\n", newSize );
    Invalidate();
    return FALSE;
  }

  int copyNbr = newSize > m_Size ? m_Size : newSize;
  memcpy( q, m_pData, copyNbr );

  Invalidate(); // evtl. allokierten Speicher freigeben

  m_pData = q;
  m_Size = newSize;
  return TRUE;
} // assureSize


//--------------------------------------------------------------------
//       CString:: Set
//
bool
CString::Set( const char *str_in )
{
  //-- Nullzeiger bewirkt ung"ultiges Objekt
  if ( !str_in) {
    Invalidate();
    return FALSE;
  }

  //-- set kann ein ung"ultiges Objekt g"ultig machen
  int l = strlen( str_in );
  if ( !AssureSize( l+1 ) )
    return FALSE;

  strcpy( m_pData, str_in );
  return TRUE;
} // set


//--------------------------------------------------------------------
//       CString:: Invalidate
//
void
CString::Invalidate()
{
  if ( !IsValid() )
    return;

  if ( !IsEmptyString() ) {
    assert( m_pData );
    free( m_pData );
  }

  m_pData = &NoData;
  m_Size = 0;
} // Invalidate


//--------------------------------------------------------------------
//       CString:: CheckIndex
//
bool
CString::CheckIndex( int ofs ) const
{
  if ( IsValidIndex( ofs ) )
    return TRUE;

  if ( !IsValid() ) 
    printf( "Zugriff auf ungueltiges Objekt mit offset %i \n", ofs );
  else
    printf( "Zugriff auf CString %s ausserhalb der Grenzen. Offset : %i\n",
      m_pData, ofs );

  return FALSE;
}  // checkIndex


//--------------------------------------------------------------------
//       CString statische Variablen
//
long CString::strnaL                  = 0;
float CString::strnaD                 = 0.0;

char CString::NoData                  = 0x00;

CString::TCompareMode CString::CompareMode       = CString::cmpStandard;
int CString::NotValid                 = -2;
int CString::NotFound                 = -1;
char CString::EmptyBuf                 = 0x00;

const char *CString::InvalidString    = "*** invalid ***";

//--------------------------------------------------------------------
//       operator + f"ur Strings
//
CString
operator + ( const CString &lhs, const CString &rhs )
{
  CString result( lhs );
  result.Append(rhs);
  return result;
} // op +


