#include <string.h>
#include <stdlib.h>
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// This source file defines methods of the class
//    CParamDef
//


#include <assert.h>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <sstream>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <utils.hh>

#include <paramdef.hh>

using namespace std;




//_________________________________________________________
//
// CParamDef                                           
//_________________________________________________________

CParamDef::CParamDef (char *keywordInit,
                      char flagInit,
                      types partypeInit,
                      char *partypenameInit,
                      char *descriptionInit,
                      char *errordescInit,
                      bool optInit,
                      int acceptWithoutKeywordOrder)
{
  keyword     = new char[strlen(keywordInit)+1];
  assert (keyword); 
  strcpy(keyword, keywordInit);
  flag        = flagInit;
  partype     = partypeInit;
  partypename = partypenameInit;
  description = new char[strlen(descriptionInit)+1];
  assert (description); 
  strcpy(description, descriptionInit);
  errordesc   = new char[strlen(errordescInit)+1];
  assert (errordesc); 
  strcpy(errordesc, errordescInit);
  opt         = optInit;
  acceptOrder = acceptWithoutKeywordOrder;
  
  if (opt)
    satisfied = TRUE;
  else
    satisfied = FALSE;

  paramFileName = NULL;
} // CParamDef::CParamDef
//_________________________________________________________

CParamDef::~CParamDef()
{
  delete [] keyword; delete [] description; delete [] errordesc;
}
//_________________________________________________________

int CParamDef::dealWith (char *token, char *source)
{
  if ( !getVarFromToken( token ) )
    {  
      cerr << endl << "'" << token << "' " << errordesc << endl;
      // occasion to use rcsid
      // cerr << endl << " var: " << *var;
      exit(-1);  // this avoids segmentation fault
      return 0;
    }
  else
    {
      satisfied = TRUE;
      // note that source is NOT stcpy's
      paramFileName = source;
      return 1;
    }
}

//_________________________________________________________

int CParamDef::understands (char c)
{
  return ( ((c == flag) && (flag)) ? TRUE : FALSE);
}
//_________________________________________________________


int CParamDef::understands (char *ptr) 
{
  return ( !strcmp(ptr, keyword) );
}
//_________________________________________________________


void CParamDef::writeDataType()
{
  cerr << partypename;
}

void CParamDef::usageInfo ()
// prints the param.s flag and keyword; in [] if optional
{
  if (no_keyword())
    {
      // No keyword
      if (opt)
        {
          cerr << " [";
          writeDataType();
          cerr << "]";
        }
      else
        {
          cerr << " ";
          writeDataType();
        }
    }
  else
    {
      if (flag)
        {
          if (opt)
            {
              cerr <<  " [-" << flag << " ";
              writeDataType();
              cerr << "]";
            }
          else
            {
              cerr <<  " -" << flag << " ";
              writeDataType();
            }
        }
    }
}

//_________________________________________________________


void CParamDef::writeDesc ()
{
  cerr << endl << "   ";
  if (no_keyword())
    {
      // no keyword
      writeDataType();
    }
  else
    {
      // keyword
      if (flag) 
        {
          cerr << "-" << flag << " ";
          writeDataType();
          cerr << ", ";
        }
      cerr << "--" << keyword << "=";
      writeDataType();
    }
  cerr << ":" << endl << "\t\t" << description;
  if ( optional() )
    {
      cerr <<  endl << "\t\t" << "Default: ";
      writeDefault();
    }
  else
    {
      cerr <<  endl << "\t\t" << "!! This parameter MUST be given !!";
    }
  if (no_keyword())
    {
      // no keyword
      cerr << endl << "\t\tKeyword in parameter file: " << keyword;
    }
}

bool CParamDef::myKWless_param(int nr)
{
  if (nr ==  acceptOrder) return TRUE;
  else return FALSE;
}

bool CParamDef::no_keyword()
{
  if (acceptOrder == -1) return FALSE;
  else return TRUE;
}


//_________________________________________________________
//
// CCharParamDef
//_________________________________________________________

CCharParamDef::CCharParamDef (char *keywordInit,
                              char flagInit,
                              char  *varInit,
                              char  defltInit,
                              char *partypenameInit,
                              char *descriptionInit,
                              bool optInit,
                              int acceptWithoutKeywordOrder)
  : CParamDef (keywordInit, flagInit, 
               Char, partypenameInit, descriptionInit, 
               "should be a character", 
               optInit, acceptWithoutKeywordOrder)
{
  var = varInit; deflt = defltInit; *var = deflt;
};
//_________________________________________________________


//! Print keyword=value to output stream.
void CCharParamDef::display(ostream & os) const {
  assert(keyword);
  assert(var);
  os << keyword << "=" << *var << endl;
} // param def display()

int CCharParamDef::getVarFromToken (char *token)
{
  if (!token)
    return 0;
  istringstream Token (token);

  Token >> *var;
  return ( !Token.fail() );
}

//_________________________________________________________
//
// CBoolParamDef
//_________________________________________________________

CBoolParamDef::CBoolParamDef (char *keywordInit,
                              char flagInit,
                              bool *varInit,
                              bool  defltInit,
                              char *partypenameInit,
                              char *descriptionInit,
                              bool optInit,
                              int acceptWithoutKeywordOrder)
  : CParamDef (keywordInit, flagInit, Bool, partypenameInit, descriptionInit, "", 
               optInit, acceptWithoutKeywordOrder)
{
  var = varInit; deflt = defltInit; *var = deflt;
};
//_________________________________________________________

int CBoolParamDef::dealWith (char * /*token*/, char *source)
{
  *var = TRUE;
  satisfied = TRUE;
  // note that source is NOT stcpy's
  paramFileName = source;
  return 1; 
}
//_________________________________________________________

void CBoolParamDef::usageInfo ()
{
  if (flag) cerr << flag;
}
//_________________________________________________________

void CBoolParamDef::writeDesc ()
{
  if (flag) cerr << "\n   -" << flag << ", --" << keyword << ":\n\t\t" << description;
  else cerr << "\n   --" << keyword << ":\n\t\t" << description;
}

bool CBoolParamDef::myKWless_param(int nr)
{
  if (nr == -1) return TRUE;
  else return FALSE;
}

bool CBoolParamDef::no_keyword()
{
  return FALSE;
}

//! Print keyword=value to output stream.
void CBoolParamDef::display(ostream & os) const {
  assert(keyword);
  assert(var);
  os << keyword << "=" << *var << endl;
} // param def display()

//_________________________________________________________
//
// CIntParamDef
//_________________________________________________________

CIntParamDef::CIntParamDef (char *keywordInit,
                            char flagInit,
                            int  *varInit,
                            int  defltInit,
                            char *partypenameInit,
                            char *descriptionInit,
                            bool optInit,
                            int acceptWithoutKeywordOrder)
  : CParamDef (keywordInit, flagInit, 
               Int, partypenameInit, descriptionInit, 
               "should be an integer",
               optInit, acceptWithoutKeywordOrder)
{
  var = varInit; deflt = defltInit; *var = deflt;
};
//_________________________________________________________

//! Print keyword=value to output stream.
void CIntParamDef::display(ostream & os) const {
  assert(keyword);
  assert(var);
  os << keyword << "=" << *var << endl;
} // param def display()

int CIntParamDef::getVarFromToken (char *token)
{
  if (!token)
    return 0;
  istringstream Token (token);

  Token >> *var;
  return ( !Token.fail() );
}

//_________________________________________________________
//
// CUnsignedintParamDef
//_________________________________________________________

CUnsignedintParamDef::CUnsignedintParamDef (char *keywordInit,
                                            char flagInit,
                                            unsigned int  *varInit,
                                            unsigned int  defltInit,
                                            char *partypenameInit,
                                            char *descriptionInit,
                                            bool optInit,
                                            int acceptWithoutKeywordOrder)
  : CParamDef (keywordInit, flagInit, 
               Unsignedint, partypenameInit, descriptionInit, 
               "should be an unsigned integer",
               optInit, acceptWithoutKeywordOrder)
{
  var = varInit; deflt = defltInit; *var = deflt;
};
//_________________________________________________________

//! Print keyword=value to output stream.
void CUnsignedintParamDef::display(ostream & os) const {
  assert(keyword);
  assert(var);
  os << keyword << "=" << *var << endl;
} // param def display()

int CUnsignedintParamDef::getVarFromToken (char *token)
{
  if (!token)
    return 0;

  istringstream Token (token);

  Token >> *var;
  return ( !Token.fail() );
}

//_________________________________________________________
//
// CShortintParamDef
//_________________________________________________________

CShortintParamDef::CShortintParamDef (char *keywordInit,
                                      char flagInit,
                                      short int  *varInit,
                                      short int  defltInit,
                                      char *partypenameInit,
                                      char *descriptionInit,
                                      bool optInit,
                                      int acceptWithoutKeywordOrder)
  : CParamDef (keywordInit, flagInit, 
               Shortint, partypenameInit, descriptionInit, 
               "should be a (short) integer",
               optInit, acceptWithoutKeywordOrder)
{
  var = varInit; deflt = defltInit; *var = deflt;
};
//_________________________________________________________

//! Print keyword=value to output stream.
void CShortintParamDef::display(ostream & os) const {
  assert(keyword);
  assert(var);
  os << keyword << "=" << *var << endl;
} // param def display()

int CShortintParamDef::getVarFromToken (char *token)
{
  if (!token)
    return 0;

  istringstream Token (token);

  Token >> *var;
  return ( !Token.fail() );
}

//_________________________________________________________
//
// CUnsignedshortParamDef
//_________________________________________________________

CUnsignedshortParamDef::CUnsignedshortParamDef (char *keywordInit,
                                                char flagInit,
                                                unsigned short int  *varInit,
                                                unsigned short int  defltInit,
                                                char *partypenameInit,
                                                char *descriptionInit,
                                                bool optInit,
                                                int acceptWithoutKeywordOrder)
  : CParamDef (keywordInit, flagInit, 
               Unsignedshort, partypenameInit, descriptionInit, 
               "should be an unsigned integer",
               optInit, acceptWithoutKeywordOrder)
{
  var = varInit; deflt = defltInit; *var = deflt;
};
//_________________________________________________________

//! Print keyword=value to output stream.
void CUnsignedshortParamDef::display(ostream & os) const {
  assert(keyword);
  assert(var);
  os << keyword << "=" << *var << endl;
} // param def display()

int CUnsignedshortParamDef::getVarFromToken (char *token)
{
  if (!token)
    return 0;

  istringstream Token (token);

  Token >> *var;
  return ( !Token.fail() );
}

//_________________________________________________________
//
// CLongintParamDef
//_________________________________________________________

CLongintParamDef::CLongintParamDef (char *keywordInit,
                                    char flagInit,
                                    long int  *varInit,
                                    long int  defltInit,
                                    char *partypenameInit,
                                    char *descriptionInit,
                                    bool optInit,
                                    int acceptWithoutKeywordOrder)
  : CParamDef (keywordInit, flagInit, 
               Longint, partypenameInit, descriptionInit, 
               "should be an integer",
               optInit, acceptWithoutKeywordOrder)
{
  var = varInit; deflt = defltInit; *var = deflt;
};
//_________________________________________________________

//! Print keyword=value to output stream.
void CLongintParamDef::display(ostream & os) const {
  assert(keyword);
  assert(var);
  os << keyword << "=" << *var << endl;
} // param def display()

int CLongintParamDef::getVarFromToken (char *token)
{
  if (!token)
    return 0;

  istringstream Token (token);

  Token >> *var;
  return ( !Token.fail() );
}

//_________________________________________________________
//
// CUnsignedlongParamDef
//_________________________________________________________

CUnsignedlongParamDef::CUnsignedlongParamDef (char *keywordInit,
                                              char flagInit,
                                              unsigned long int  *varInit,
                                              unsigned long int  defltInit,
                                              char *partypenameInit,
                                              char *descriptionInit,
                                              bool optInit,
                                              int acceptWithoutKeywordOrder)
  : CParamDef (keywordInit, flagInit, 
               Unsignedlong, partypenameInit, descriptionInit, 
               "should be an unsigned integer",
               optInit, acceptWithoutKeywordOrder)
{
  var = varInit; deflt = defltInit; *var = deflt;
};
//_________________________________________________________

//! Print keyword=value to output stream.
void CUnsignedlongParamDef::display(ostream & os) const {
  assert(keyword);
  assert(var);
  os << keyword << "=" << *var << endl;
} // param def display()

int CUnsignedlongParamDef::getVarFromToken (char *token)
{
  if (!token)
    return 0;

  istringstream Token (token);

  Token >> *var;
  return ( !Token.fail() );
}


//_________________________________________________________
//
// CStringParamDef
//_________________________________________________________

CStringParamDef::CStringParamDef (char *keywordInit,
                                  char flagInit,
                                  char  **varInit,
                                  char  *defltInit,
                                  char *partypenameInit,
                                  char *descriptionInit,
                                  bool optInit,
                                  int acceptWithoutKeywordOrder)
  : CParamDef (keywordInit, flagInit, 
               String, partypenameInit, descriptionInit, 
               "should be a non null string",
               optInit, acceptWithoutKeywordOrder)
{
  varToBeDeleted = FALSE;
  var = varInit; deflt = defltInit; *var = deflt;
};
//_________________________________________________________

CStringParamDef::~CStringParamDef()
{
  if (varToBeDeleted)
    delete [] *var;
}
//_________________________________________________________

//! Print keyword=value to output stream.
void CStringParamDef::display(ostream & os) const {
  assert(keyword);
  assert(var);
  if (*var != NULL)
    os << keyword << "=" << *var << endl;
  else
    os << keyword << "=" << endl;
} // param def display()

int CStringParamDef::getVarFromToken (char *token)
{
  if (!token)
    return 0;

  varToBeDeleted = TRUE;
  *var = new char[strlen (token)+1];
  //EXIT_IF (!(*var), "could not allocate memory for String parameter", -1);
  if (!(*var)){
    cerr<<"could not allocate memory for String parameter"<<endl;
    exit(-1);
  }
  strcpy (*var, token);
  return strlen (token);
}


//_________________________________________________________
//
// CDoubleParamDef
//_________________________________________________________

CDoubleParamDef::CDoubleParamDef (char *keywordInit,
                                  char flagInit,
                                  double  *varInit,
                                  double  defltInit,
                                  char *partypenameInit,
                                  char *descriptionInit,
                                  bool optInit,
                                  int acceptWithoutKeywordOrder)
  : CParamDef (keywordInit, flagInit, 
               Double, partypenameInit, descriptionInit, 
               "should be a double",
               optInit, acceptWithoutKeywordOrder)
{
  var = varInit; deflt = defltInit; *var = deflt;
};
//_________________________________________________________

//! Print keyword=value to output stream.
void CDoubleParamDef::display(ostream & os) const {
  assert(keyword);
  assert(var);
  os << keyword << "=" << *var << endl;
} // param def display()

int CDoubleParamDef::getVarFromToken (char *token)
{
  if (!token)
    return 0;

  istringstream Token (token);

  Token >> *var;
  return (! Token.fail()); // Token.eof seems to be set after reading for double
                           // or float... So Token.good is zero...
}
//_________________________________________________________
//
// CFloatParamDef
//_________________________________________________________

CFloatParamDef::CFloatParamDef (char *keywordInit,
                                char flagInit,
                                float  *varInit,
                                float  defltInit,
                                char *partypenameInit,
                                char *descriptionInit,
                                bool optInit,
                                int acceptWithoutKeywordOrder)
  : CParamDef (keywordInit, flagInit, 
               Float, partypenameInit, descriptionInit, 
               "should be a float",
               optInit, acceptWithoutKeywordOrder)
{
  var = varInit; deflt = defltInit; *var = deflt;
};
//_________________________________________________________

//! Print keyword=value to output stream.
void CFloatParamDef::display(ostream & os) const {
  assert(keyword);
  assert(var);
  os << keyword << "=" << *var << endl;
} // param def display()

int CFloatParamDef::getVarFromToken (char *token)
{
  if (!token)
    return 0;

  istringstream Token (token);

  Token >> *var;
  return ( !Token.fail() );
}


