#include <string.h>
#include <stdlib.h>
#include <stdio.h>
// _IDVocabulary.cpp: implementation of the C_IDVocabulary class.
//
//////////////////////////////////////////////////////////////////////

#include "_IDVocabulary.h"
#include <fstream>
#include <iostream>
using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

C_IDVocabulary::C_IDVocabulary(const char * fileName)
{
	this->text2id.insert(make_pair(C_String("__NULL__"),0));
	this->id2text.insert(make_pair(0, C_String("__NULL__")));
	this->maxIdInVoc = 0;

	this->loadFromFile(fileName);
}

C_IDVocabulary::~C_IDVocabulary()
{

}


IndexType C_IDVocabulary::returnId(C_String text)
{
	IndexType id;
	
	map<C_String, IndexType, ltstr>::iterator iterText2Id;
	iterText2Id = this->text2id.find(text);

	if(iterText2Id==this->text2id.end()){ //this word does not exist in the voc yet, return ID for <unk>
		id = 0;
	}
	else{
		id = iterText2Id->second;
	}
	
	return id;
}

C_String C_IDVocabulary::getText(IndexType id)
{
	map<IndexType, C_String>::iterator iterId2Text;
	iterId2Text = this->id2text.find(id);

	if(iterId2Text==this->id2text.end()){
		return C_String("<UNK>");
	}

	return iterId2Text->second;
}

IndexType C_IDVocabulary::getSize()
{
	return this->text2id.size();
}

void C_IDVocabulary::loadFromFile(const char *fileName)
{

	ifstream existingVocFile;
	existingVocFile.open(fileName);

	if(!existingVocFile){
		cout<<"Can not open existing vocabulary file "<<fileName<<endl;
		exit(0);
	}

	cout<<"Loading existing vocabulary file: "<<fileName<<endl;

	char aLine[1024];
	char * aToken;
	char delimit[] = " \t\r\n";
	char aWord[1024];
	IndexType vocId = 0;
	
	while(!existingVocFile.eof()){
		existingVocFile.getline(aLine, 1024, '\n');
		
		if(strlen(aLine)>0){	//a meaningful word, esp for the last line during reading file
			if(aLine[0]!='#'){	//this is not a comment line
				vector<C_String> tokensInLine;

				aToken = strtok(aLine, delimit);				
				while( aToken != NULL ) {	
					tokensInLine.push_back(C_String(aToken));
					aToken = strtok( NULL, delimit);
				}
				
				if(tokensInLine.size()!=2){
					cout<<"Not valid format for Vocabulary: "<<aLine<<endl;
				}
				
				vocId = atoi(tokensInLine[1].toString());

				if(vocId>this->maxIdInVoc){
					this->maxIdInVoc = vocId;
				}

				this->text2id.insert(make_pair(tokensInLine[0], vocId));
				this->id2text.insert(make_pair(vocId, C_String(aWord)));
			}
			
		}

		aWord[0]=0;	//clean the variable
		aLine[0]=0;
	}
	cout<<"Total "<<this->text2id.size()<<" word types loaded\n";
	cout<<"Max VocID="<<this->maxIdInVoc<<endl;
}

IndexType C_IDVocabulary::returnMaxID()
{
	return this->maxIdInVoc;
}
