#include <string.h>
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <fstream>
#include <assert.h>


#include <iostream>
#include <sstream>
#include <string>

#define NO_TRACE 1
#include <Trace.hh>

#include <Alignment.hh>

using namespace std;

CAlignment::CAlignment()
{
  m_pSourceText = NULL;
  m_pTargetText = NULL;
  m_pAlignment = new CArray< CNumericalArray < short > >;
  m_pAlignment->SetIncrement( 10000 );
 
}


CAlignment::~CAlignment()
{
}


bool
CAlignment::Store( const char *pFileName, int Format )  // write m_Alignment to a file
{
  short i;
  short align[32];

  CSentence *pSentenceF;
  CSentence *pSentenceE;
  CString WordE;
  TPosition FirstSentence;
  CWordIndex WordIndexE;

  m_pSourceText->FirstSentence( FirstSentence );

  ofstream OutputFile(pFileName);

  if ( !OutputFile ) {
    cout << "File " << pFileName << " could not be opened\n";
    return FALSE;
  }

  cout << m_pAlignment->MinIndex() << " " << m_pAlignment->MaxIndex() << endl;

  int SIndex, WIndex;
  for ( SIndex = m_pAlignment->MinIndex(); 
		SIndex <= m_pAlignment->MaxIndex();
		SIndex++ ) {
   
	cout << "SENT: " << SIndex << "\n";  // schreibe Satz
    
	const CNumericalArray< short > & Row = m_pAlignment->GetAt( SIndex );
	
	if ( Format == 0 ) {

	  for ( WIndex = Row.MinIndex(); 
			WIndex <= Row.MaxIndex();
			WIndex++ ) {
		OutputFile << Row[ WIndex ] << " ";
	  }
	  OutputFile << endl;

	} else if ( Format == 1 ) {  // GIZA format
	  // stores alignemts i - list of j's

	  //cout << "#Sentence pair (" << SIndex << ")\n";
	  OutputFile << "#Sentence pair (" << SIndex << ")\n";
	  pSentenceF = m_pSourceText->GetSentence( SIndex );
	  pSentenceE = m_pTargetText->GetSentence( SIndex );

	  //cout << pSentenceF->AsString( m_pSourceText->Vocabulary() ) << endl;
	  //cout << pSentenceE->AsString( m_pTargetText->Vocabulary() ) << endl;

	  OutputFile << pSentenceF->AsString( m_pSourceText->Vocabulary() ) << endl;
	  
	  //cout << "NULL ({ }) " << endl;
	  OutputFile << "NULL ({ }) ";
	  for ( i = pSentenceE->MinIndex(); i <= pSentenceE->MaxIndex(); i++ ) {
		WordIndexE = pSentenceE->GetAt( i );
		//cout << WordIndexE << " ";
		WordE = m_pTargetText->Vocabulary()->GetWord( WordIndexE );
		if ( WordE == CString( "<UNK>" ) ) {
		  WordE = "UNK";
		}
		//cout << WordE << " ({ ";
		OutputFile << WordE << " ({ ";

		for ( int j = Row.MinIndex(); 
			  j <= Row.MaxIndex();
			  j++ ) {
		  if ( Row[ j ] == i ) {
			//cout << j << " ";
			OutputFile << j << " ";
		  }
		}

		//cout << "}) " << endl;
		OutputFile << "}) ";
	  }
	  
	  cout << endl;
	  OutputFile << endl;

	} else if ( Format == 2 ) {  // EBMT format
	  // stores alignemts i - list of j's

	  //cout << "#Sentence pair (" << SIndex << ")\n";

	  pSentenceF = m_pSourceText->GetSentence( SIndex );
	  pSentenceE = m_pTargetText->GetSentence( SIndex );
	  
	  //cout << pSentenceF->AsString( m_pSourceText->Vocabulary() ) << endl;
	  //cout << pSentenceE->AsString( m_pTargetText->Vocabulary() ) << endl;
	  
	  cout << ";;;(ALIGN  " << endl;
	  OutputFile << ";;;(ALIGN ";

	  for ( WIndex = Row.MinIndex(); 
			WIndex <= Row.MaxIndex();
			WIndex++ ) {
		cout << "(" << Row[ WIndex ] << ")" << WIndex << " ";
		OutputFile << "(" << Row[ WIndex ] << ")" << WIndex << " ";
	  }
		
	  cout << ")" << endl;
	  OutputFile << ")" << endl;
	  
	  OutputFile << pSentenceF->AsString( m_pSourceText->Vocabulary() ) << endl;
	  OutputFile << pSentenceE->AsString( m_pTargetText->Vocabulary() ) << endl;
		
	} else if ( Format == 3 ) {  // EBMT reverse format 
	  // stores alignemts i - list of j's

	  //cout << "#Sentence pair (" << SIndex << ")\n";

	  pSentenceF = m_pSourceText->GetSentence( SIndex );
	  pSentenceE = m_pTargetText->GetSentence( SIndex );
	  
	  //cout << pSentenceF->AsString( m_pSourceText->Vocabulary() ) << endl;
	  //cout << pSentenceE->AsString( m_pTargetText->Vocabulary() ) << endl;
	  
	  cout << ";;;(ALIGN  " << endl;
	  OutputFile << ";;;(ALIGN ";

	  for ( i = pSentenceE->MinIndex(); i <= pSentenceE->MaxIndex(); i++ ) {

		cout << "(";
		OutputFile << "(";
		
		for ( int j = Row.MinIndex(); 
			  j <= Row.MaxIndex();
			  j++ ) {
		  if ( Row[ j ] == i ) {
			//cout << j << " ";
			cout << j << " ";
			OutputFile << j << " ";
		  }
		}

		cout << ")" << i << " ";
		OutputFile << ")" << i << " ";
				
	  }
		
	  cout << ")" << endl;
	  OutputFile << ")" << endl;
	  
	  OutputFile << pSentenceE->AsString( m_pTargetText->Vocabulary() ) << endl;
	  OutputFile << pSentenceF->AsString( m_pSourceText->Vocabulary() ) << endl;
		
	}
	
  }
  
  return TRUE;
}



bool
CAlignment::Load( const char *pFileName, int Format )
{ 
  ifstream File( pFileName, ios::in );

  if (!File) {
	cerr << "File " << pFileName << " could not be opened\n";
	return FALSE;
  }

  string Row = "";

  int NumberOfSentence = 0;
  short j, i;     

  getline( File, Row );
  while ( !File.eof() ) {

	istringstream Alignment( Row.c_str());

	j = 0;
	while ( !Alignment.eof() ) {
	  Alignment >> i;
	  j++;

	  //cout << "Alignment: " << NumberOfSentence << " " << j << " " << i << endl;

	  Set( NumberOfSentence, j, i );   // set alignment
	}
			    	    
	getline( File, Row );
	NumberOfSentence++;

  }  // end of while

  File.close();

  return TRUE;
}


/*
bool
CAlignment::Load( const char *pFileName )
{ 
  ifstream quelle(pFileName, ios ::in);  // Quelldatei

  if (!quelle) 
    {
      cerr << "File " << pFileName << " could not be opened\n";
      return FALSE;
    }

  string row="";

  int NumberOfSentence = 0 , NumberOfWord;
  short Position, j, max = -1;     

  const short maxslen = 256;  // maximum sentence length

  short val[maxslen];  // val[j]==0 : word j in actual sentence has no alignment
                       // val[j]==1 : word j in actual sentence has at least one alignment
  
  string beg="";


  for (j=0;j<maxslen;++j)
    val[j]=0;


 
  getline(quelle,row);
  while (!quelle.eof())
    {
      istringstream zeile(row.c_str());

      beg ="";
      zeile >> beg;

      if (beg!= "")
	if (!strncmp(beg.c_str(),"SENT: ",5))   // new sentence?
	  {	 	  
	    for (j=0;j<maxslen;++j)
	      {
		if (val[j] == 0)
		  m_Alignment(NumberOfSentence,j)=0;
		
		val[j]=0;
	      } 

	    max = NumberOfSentence;  
	    NumberOfSentence = -1;
	    zeile >> NumberOfSentence;	 // get number of new sentence 
	    if (NumberOfSentence == -1)
	      cerr << "Fehler in Alignmentdatei nach Satz "
		   << max 
		   << endl;

	    max = -1;
	  }
	else 
	  { 	
	    NumberOfWord = -1;
	    zeile >> NumberOfWord;	// get number of word
	    if (NumberOfWord == -1)
	      cerr << "Fehler in Alignmentdatei : Satz " 
		   << NumberOfSentence
		   << endl;
	    else
	      {
		if (NumberOfWord > max)
		  max = NumberOfWord; 
		if (val[NumberOfWord] == 0)
		  {
		    val[NumberOfWord]=1;
		    m_Alignment(NumberOfSentence,NumberOfWord) = 0; 
		  }		 
		
		Position = -1;
		zeile >> Position;      // get position of aligned word
		if (Position == -1)
		  cerr << "Fehler in Alignmentdatei : Satz " 
		       << NumberOfSentence << "   Wort "
		       << NumberOfWord
		       << endl;		
		else
		  Set(NumberOfSentence, NumberOfWord, Position);   // set alignment
	      }
	  }			    	    
      getline(quelle, row);

    }  // end of while


  for (j=0;j<maxslen;++j)    
    if (val[j] == 0)
      m_Alignment(NumberOfSentence,j)=0;
   

  quelle.close();
  return TRUE;
}
*/



void
CAlignment::Set( int NumberOfSentence, short SourcePosition, short TargetPosition, float Value )
{
  Trace( 1, "Set: " << NumberOfSentence << " " << SourcePosition << " " << TargetPosition );
  (*m_pAlignment)[ NumberOfSentence ][ SourcePosition ] = TargetPosition;
}



void
CAlignment::Get( int NumberOfSentence, short* Align)
{
  const CNumericalArray< short > &Row = m_pAlignment->GetAt( NumberOfSentence );

  for( int i = Row.MinIndex(); i <= Row.MaxIndex(); i++ ) {
	cout << i << " " << Row[ i ] << endl;
	Align[ i ] = Row[ i ];
  }
}


const CNumericalArray< short > &
CAlignment::GetSimpleAlignment( int NumberOfSentence ) const
{
  return m_pAlignment->GetAt( NumberOfSentence );

}


void
CAlignment::SetIncrement( int Increment )
{
  m_pAlignment->SetIncrement( Increment );
}



void
CAlignment::SetText( CText *pSourceText, CText *pTargetText )
{
  m_pSourceText = pSourceText;
  m_pTargetText = pTargetText;
}



/*
void 
CAlignment::Display( int NumberOfSentence , bool indexed = FALSE)
{

  assert( m_pSourceText );
  assert( m_pTargetText );

  int i, j, idx;

  int MaxSLen = 150;
  char BestPath[150][150];


  if (indexed)
    {
      idx = 1;
    }
  else    
    {
      idx = 0; 
      --NumberOfSentence;
    }


  for (j = 0; j < MaxSLen; j++) 
    for (i = 0; i < MaxSLen; i++)
      BestPath[j][i] = '.';

  int J = m_pSourceText->SentenceLength( NumberOfSentence );
  int I = m_pTargetText->SentenceLength( NumberOfSentence );
  short align[33];



  if ( (J > 0) && (I > 0) ) {
    // find the best path and store points in BestPath
    for ( j = 0; j < J; j++ ) 
      {
	Get_Array(NumberOfSentence-idx, j , align);
	i=1; 
	while (i <= align[0])
	  {
	    BestPath[j][align[i]+1] = 'o';
	    ++i;
	  }
      }
  }

  
  // now display BestPath
  CString Word;
  int maxLen=0;
  const char *pString = NULL;

  if ( I > 0 ) {
    // get length of longest e word
    for (i = I; i >= 1; i--) {    

      pString =  m_pTargetText->GetSentence( NumberOfSentence )->Word( i , m_pTargetText->Vocabulary()  );

      Trace( 5, pString );
      if ( strlen(pString) > (uint) maxLen ) 
	maxLen = strlen( pString );
    }


    for (i = I; i >= 1; i--) {
      cout.width( maxLen );
      Word = m_pTargetText->GetSentence( NumberOfSentence )->Word( i , m_pTargetText->Vocabulary() );
      cout << Word << " |   " << flush;
      
      if ( J > 0 ) {
	for (j = 0; j < (int) J; j++) {
	  cout.setf( ios::left );
	  cout.width( 3 );
	  cout << BestPath[ j ][ i ] << " ";
	}
      }
      cout << "\n";
      cout.flush();
    }
  }


  DisplaySourceWords( NumberOfSentence, maxLen );

}

*/


void
CAlignment::DisplaySourceWords( int NumberOfSentence, int maxLen )
{
  int MaxWordLength = 40;
  const char *pString = NULL;
  int MaxSLen = 150;
  char WordMatrix[ 40 ][ 150 ];
  int NoSpace;
  int i, j;
  CString Word;

  int J = m_pSourceText->SentenceLength( NumberOfSentence );

  // display line
  for ( i = 0; i < maxLen; i++ )
    cout << "-";
  cout << "-|";

  if ( J > 0 ) {
    for (j = 1; j <= J; j++) {
      cout << "----";
    }
  }
  cout << endl;

  // display french words at bottom
  //Trace( 1, "display french words at bottom" );
  char Blanks[ 150 ];

  maxLen += 4;

  for( i = 0; i < maxLen; i++ )
    Blanks[ i ] = ' ';
  Blanks[ maxLen ]='\0';
  
  for ( j = 1; j < MaxSLen; j++ )
    for ( i = 0; i < MaxWordLength; i++ )
      WordMatrix[ i ][ j ] = ' ';
  
  if ( J > 0 ) {
    for (j = 1; j <= (int) J; j++) {
      pString =  m_pSourceText->GetSentence( NumberOfSentence )->Word( j , m_pSourceText->Vocabulary() );
      cout.setf( ios::left );
      i=0;
      while ( pString[ i ] ) {
	WordMatrix[ i ][ j ] = (char) pString[ i ];
	i++;
      }
    }
  }
  
  NoSpace = 1;
  for ( i = 0; ((i < MaxWordLength) && (NoSpace == 1)); i++ ) {
    NoSpace = 0;
    cout << Blanks << " ";
    // for diagonal output an additional blank in the next iteration
    Blanks[maxLen] = ' ';
    maxLen++;
    Blanks[ maxLen ] = '\0';
    if ( J > 0 ) {
      for ( j = 1; j <= J; j++ ) {
	cout << WordMatrix[ i ][ j ] << "   ";
	if (WordMatrix[ i ][ j ] != ' ')
	  NoSpace = 1;
      }
    }
    cout << "\n";
  }
    
  cout << "\n";
  cout.flush();
}


