#include <config.h>

// CAlignment

// class to store global alignment of a bilingual corpus


#ifndef __Alignment_h__
#define __Alignment_h__


#include <NumericalArray.hh>
#include <Text.hh>


class CAlignment
{
  // Alignment is stored according to number of sentence
  // and number of word. For each word in the source sentence
  // the positions of the aligned words in the target sentence
  // are stored, togather with an aligment score.

public:
  CAlignment();
  ~CAlignment();
  
  // set to the number of sentence pairs to allocate required memory
  void SetIncrement( int Increment );

  // access to the corpus, for displaying and storing alignments with
  // actuall words
  void SetText( CText *pSourceText, CText *pTargetText );

  // set one alignment point
  void Set( int NumberOfSentence, short SourcePosition, short TargetPosition, float Value = 1.0 );

  // get complete alignment for sentence pair
  // now: simple 1:1 aligment, so only array
  // in future: matrix
  void Get( int NumberOfSentence, short* Align );

  // get complete alignment for sentence pair
  // now: simple 1:1 aligment, so only array
  // in future: matrix
  const CNumericalArray< short > & GetSimpleAlignment( int NumberOfSentence ) const;

  // get alignment for one position in sentence
  // perhaps distinguish between position in source and target sentence
  // i.e. row and column in alignment matrix
  // void GetCol( int NumberOfSentence, int SourcePosition, short* Align);
  // void GetRow( int NumberOfSentence, int SourcePosition, short* Align);

  // get alignment score for one alignment point
  // float Get( int NumberOfSentence, int SourcePosition, int TargetPosition );

  // load and store alignment, different formats possible
  bool Store( const char *pFileName, int Format = 0 );
  bool Load( const char *pFileName, int Format = 0  );

  // display alignment for a sentence pair
  void Display( int NumberOfSentence , bool indexed);

  // additional methods:
  // invert an alignment for all sentence pairs, for one sentence pair
  // union of two aligments
  // intersection of two aligments
  // combination of two alignments
  // precision and recall


private:
  
  void DisplaySourceWords( int NumberOfSentence, int maxLen );

  // alignment for total text
  // sentence - source position - target position
  CArray< CNumericalArray< short > > *m_pAlignment;

  // replace with implementation based on hash table

  // pointers to source and target text for display
  CText *m_pSourceText;
  CText *m_pTargetText;

};

#endif
