#include <stdlib.h>
#include <iostream>
using std::cout;
using std::cerr;

#include <fstream>
using std::ifstream;
using std::ofstream;

#include <FertilityModel.hh>



CFertilityModel::CFertilityModel()
{
  //m_TargetLenPr = new CArray <double>;
  //m_pInternalVoc = new CVocabulary();
}



CFertilityModel::~CFertilityModel()
{
  //delete m_pInternalVoc;
}



BOOL
CFertilityModel::Init( CVocabulary *pVoc )
{
  m_pVoc = pVoc;
  return TRUE;
}



BOOL
CFertilityModel::Load( const char* pFileName )
{
  //TODO: currently fixed to reading 5 values
  //generalize to any number of values

  //cerr << "Loading Fertility Table ... ";
  
  m_MaxFertility = 5;

  ifstream FFile( pFileName );
  
  if( !FFile ) {
    cerr << "File " << pFileName << " could not be opened" << endl;
    return FALSE;
    //exit(1);
  }

  //initialize the table with uniform probs, for all words in the voc
  //this is necessary, to handle unknown words

  // sv: should accumulated over all fertilities and take this
  // for unknown words

  m_DefaultFertilityPr = 1 / (float)m_MaxFertility;

  //for(int i=0; i<m_pVoc->Size(); i++) {
  //  for(int j=0; j<m_MaxFertility; j++ ) {
  //    m_FertilityPr.SetAt(i, j, uniprob );   //init with uniform 
  //  }
  //}

  m_FertilityPr.SetBoundaries( 0, m_pVoc->Size(), 0, m_MaxFertility );
  m_FertilityPr.SetAll( m_DefaultFertilityPr );

  //Sanjika 07/12
  m_MaxWordIndex = m_pVoc->Size();


  char *pWord = new char[256];
  double Pr[m_MaxFertility];
  CWordIndex WIndex;
  
  while( !FFile.eof() ) {

    FFile >> pWord >> Pr[0] >> Pr[1] >> Pr[2] >> Pr[3] >> Pr[4];
    //debug
    //cout << pWord << " " << Pr[0] << " " << Pr[1] << " " << Pr[2] << " " << Pr[3] << " " << Pr[4] << endl; 
  
    WIndex = m_pVoc->GetIndex( pWord );
    
    //ARV 04/25
    //sv: should be no need for that; add new words to external vocabulary
    //m_pInternalVoc->AddWord( CString(pWord) );

    //Sanjika 07/12
    //removed the internal voc
    //no new words are possible, if fertilities are obtained from same training data
    //anyway, if found, are added to the external voc
    if( WIndex == CVocabulary::EOT || WIndex == CVocabulary::UNK ){
	m_pVoc->AddWord( CString(pWord) );

	//TODO:
	//since the size of the matrix was set according to voc size,
	//m_FertilityPr should be expanded by 1 row
	//m_MaxWordIndex shold also be increased by 1
    }

    for(int i=0; i<m_MaxFertility; i++ ) {
      m_FertilityPr.SetAt(WIndex, i, Pr[i] );    
    }

    //debug
    //cout << "Word = " << m_pVoc->GetWord( WIndex ) << endl;
  }
  
  cerr << "Loading Fertility Table Done" << endl;

  return TRUE;
}





//debug
void
CFertilityModel::Display()
{
  cout <<"Fertility Table" << endl;  
  
  m_FertilityPr.Display(); 

}



double
CFertilityModel::Pr( CWordIndex WIndex, int k ) const
{
  //ARV Sanjika 04/25 (switched this to use the static variables in the vocab
  //if ( WIndex == CVocabulary::EOT || WIndex == CVocabulary::UNK){
    //Trace( 3, " CFertilityModel:: Word not found in fertility table. Using default fertility");
  //  return m_DefaultFertilityPr;
  //} else {
    //Trace( 3, " CFertilityModel:: Word found in fertility table.");
  //  return m_FertilityPr.GetAt( WIndex, k );
  //}

  //Sanjika 07/12
  //removed the internal voc
  if( m_MaxWordIndex < WIndex) { //word not in fertility table
      return m_DefaultFertilityPr;
  }
  else { 
      return m_FertilityPr.GetAt( WIndex, k );
  }

}


int
CFertilityModel::GetMaxFertility()
{
  return m_MaxFertility;
}

//CVocabulary
//CFertilityModel::GetInternalVoc()
//{
//  return *m_pInternalVoc;
//}

