#include <string.h>
// include autoconf defines
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


// CTranslationModel

// Base class for all translation models.
// Handles the biligual lexicon T and the word pair counts TCounts.
// Pointers to the to text objects; Member variables for current
// sentence pair (m_S) and the length of the sentences (m_I, m_J).
//

#include <assert.h>
#include <fstream>
using std::ifstream;
using std::ofstream;
#include <sstream>
#include <time.h>

// disable / enable trace messages for this module here
#define NO_TRACE 1
//#undef NO_TRACE
#include <Trace.hh>

#include <Lexicon.hh>
#include <math.h>

using namespace std;


//static char rcsid_tm_cc[] = "$Id: ";

//! Define this to enable dataversion 5 in plain text mode (for debugging). Otherwise binary mode is used.
#undef DATAVERSION5_PLAIN

CLexicon::CLexicon()
{
  Trace( 0, "CLexicon - Constructor" );

  m_pVocE = NULL;
  m_pVocF = NULL;

  m_LambdaSmooth = 0.1;

  m_Pr.SetIncrement( 128 );
  m_LogPr.SetIncrement( 128 );

  m_Counts.SetIncrement( 128 );

}



CLexicon::~CLexicon()
{
}



////////////////////////////////////////////////////////////////////////
// Initialization - Public Methods
////////////////////////////////////////////////////////////////////////


BOOL 
CLexicon::Init( CVocabulary *pVocF, CVocabulary *pVocE,
                CCognates *pCognates )
{
  Trace( 1, "CLexicon::Init" );

  m_pVocF = pVocF;
  m_pVocE = pVocE;

  m_VocSizeF = m_pVocF->Last();
  m_VocSizeE = m_pVocE->Last();

  m_Pr.SetBoundaries( 0, m_VocSizeE, 0, m_VocSizeF );

  // ??? already here, we need counts only in training
  // however, no memory required
  m_Counts.SetBoundaries( 0, m_VocSizeE, 0, m_VocSizeF );

  Trace( 1, "Set default prob to " <<  1.0 / m_VocSizeF );
  
  m_Pr.SetDefaultValue( 1.0 / m_VocSizeF );

  // no rows expanded
  m_IsExpanded.SetBoundaries( 0, m_VocSizeE );
  m_IsExpanded.SetAll( FALSE );

  if ( pCognates ) {
    m_pCognates = pCognates;
  }

  //TestProbabilities();

  return TRUE;
}




////////////////////////////////////////////////////////////////////////
// Methods to expand und shrink rows in the lexicon
////////////////////////////////////////////////////////////////////////


BOOL
CLexicon::ExpandAll()
{
  Trace( 1, "CLexicon::ExpandAll" );

  CArray<int> WordIndices;

  WordIndices.SetBoundaries( 0, m_VocSizeE );

  for ( int i = 0; i <= m_VocSizeE; i++ ) {
    WordIndices[ i ] = 1;
  }
  return Expand( WordIndices );
}



BOOL
CLexicon::Expand( const CArray<int> &WordIndices )
{
  CWordIndex f;
  int n;
  float Value = 0.0;

  Trace( 1, "CLexicon::Expand " );
  Trace( 1, "Pr MinIndex - MaxIndex " << WordIndices.MinIndex() << WordIndices.MaxIndex() );


  m_ExpandedPr.SetBoundaries( WordIndices.MinIndex(), WordIndices.MaxIndex() );
  m_ExpandedLogPr.SetBoundaries( WordIndices.MinIndex(), WordIndices.MaxIndex() );
  m_ExpandedCounts.SetBoundaries( WordIndices.MinIndex(), WordIndices.MaxIndex() );


  for ( int e = WordIndices.MinIndex(); e <= WordIndices.MaxIndex(); e++ ) {

    Trace( 3, "e = " << e );
    //cout << e;

    if ( WordIndices[ e ] ) {
      std::cout << " Expand " << m_pVocE->GetWord( e ) << std::endl;
      //Trace( 1, "Expand " << m_pVocE->GetWord( e ) );
      //Trace( 3, m_VocSizeF );
      //m_ExpandedPr.GetRow( e ).InitToZero();
      //m_ExpandedLogPr.GetRow( e ).InitToZero();
      //m_ExpandedCounts.GetRow( e ).InitToZero();

      m_ExpandedPr.GetRow( e ).SetBoundaries( 0, m_VocSizeF );
      m_ExpandedLogPr.GetRow( e ).SetBoundaries( 0, m_VocSizeF );
      m_ExpandedCounts.GetRow( e ).SetBoundaries( 0, m_VocSizeF );

      float DefaultValue = m_Pr.GetRow( e ).DefaultValue();
      Trace( 5, "DefaultValue " << DefaultValue );

      m_ExpandedPr.GetRow( e ).SetAll( DefaultValue );
      m_ExpandedLogPr.GetRow( e ).SetAll( log( DefaultValue ) );
      m_ExpandedCounts.GetRow( e ).SetAll( 0 );

      Trace( 5, m_ExpandedPr.GetRow( e ).GetAt( 100 ) );

      // init all values to zero
      // for ( n = 0; n <= m_VocSizeF; n++ ) {
      // 	Trace( 5, "f = " << n );
      // 	m_ExpandedPr.GetRow( e ).SetAt( n, 0.0);
      // 	m_ExpandedLogPr.GetRow( e ).SetAt( n, 0.0 );
      // 	m_ExpandedCounts.GetRow( e ).SetAt( n, 0.0 );
      // }

      // copy values from sparse arrays to full arrays
      for ( n = 0; n < m_Pr.GetRow( e ).NumberOfItems(); n++ ) {
        m_Pr.GetRow( e ).GetItem( n, f, Value );
        Trace( 5, "Position = " << n << "  f = " << f
               << "  Pr = " << Value );
        m_ExpandedPr.GetRow( e ).SetAt( f, Value );
		
        m_LogPr.GetRow( e ).GetItem( n, f, Value );
        Trace( 5, "Position = " << n << "  f = " << f
               << "  LogPr = " << Value );
        m_ExpandedLogPr.GetRow( e ).SetAt( f, Value );

        m_Counts.GetRow( e ).GetItem( n, f, Value );
        Trace( 5, "Position = " << n << "  f = " << f
               << "  Counts = " << Value );
        m_ExpandedCounts.GetRow( e ).SetAt( f, Value );
      }

      m_IsExpanded[ e ] = TRUE;
    }
  }

  /*
    for ( int e = WordIndices.MinIndex(); e <= WordIndices.MaxIndex(); e++ ) {
    if ( m_IsExpanded[ e ] )
    // TODO: maybe this should be TRACE here
    cout << e << " is expanded" << endl << flush;
    else
    cout << e << " is not expanded" << endl << flush;
    }
  */

  return TRUE;
}



BOOL
CLexicon::Expand( const char *pFileName )
{
  char String[100];

  CArray<int> WordIndices;
  WordIndices.SetBoundaries( 0, m_VocSizeE );
  WordIndices.SetAll( 0 );
  CWordIndex We;

  ifstream TFile (pFileName);

  if (!TFile) {
    cout << "WARNING:  ExpandFile " << pFileName << " not loaded.  Lexicon used as-is." << endl;
    return FALSE;
  }

  //  TFile.open(pFileName);
  //  assert(TFile);

  while (!TFile.eof()) {
    TFile >> String;
    We = m_pVocE->GetIndex( CString( String ) );
    //Trace( 5, "Expand " << We << " " << String );
    Trace( 0, "Expand " << We << " " << String );
    if ( We >= 0 )
      WordIndices[ We ] = 1;
  }

  Expand( WordIndices );

  return TRUE;
}



BOOL
CLexicon::Shrink()
{
  float Prob;
  const char *pWordE;

  Trace( 1, "CLexicon::Shrink()" );

  for ( int e = m_IsExpanded.MinIndex(); e <= m_IsExpanded.MaxIndex(); e++ ) {

    Trace( 3, "e = " << e );
    //cout << e << endl << flush;

    // TODO: What's that?
    std::cout.setf( ios::scientific );
    std::cout.precision( 10 );

    if ( m_IsExpanded[ e ] ) {
      pWordE = m_pVocE->GetWord( e ); 
      Trace( 3, "Shrink " << pWordE );

      CArray<float> & Pr = m_ExpandedPr.GetRow( e );
      float DefaultValue = m_Pr.GetRow( e ).DefaultValue();
      Trace( 5, "DefaultValue = " << DefaultValue );
      //cout <<  "DefaultValue = " << DefaultValue << endl;
      m_Pr.GetRow( e ).Clear();
      m_Pr.GetRow( e ).SetDefaultValue( DefaultValue );
      m_Pr.SetIncrement( 2000 );
      DefaultValue += 1.0e-10;

      int f;
      for ( f = Pr.MinIndex(); f <= Pr.MaxIndex(); f++ ) {
        Trace( 5, f );
        Prob = Pr.GetAt( f );
        //cout << f << " " << Prob << " " << DefaultValue << endl;
        if ( Prob <= DefaultValue ) {
          //cout << e << " " << f  << " Prob <= DefaultValue " << Prob << " " << DefaultValue << endl;
        } else {
          Trace( 5, e << " " << f  << " Prob > DefaultValue  " << Prob << " " << DefaultValue );
          //cout << e << " " << f  << " Prob > DefaultValue " << Prob << " " << DefaultValue << endl;
          // skip small values
          if ( Prob > 1.0e-5 )
            m_Pr.SetAt( e, f, Prob );
        }
      }
    } else {
      //cout << "Not expanded" << endl << flush;
    }
  }

  return TRUE;
}



// Init is for initialization for TRAINING !!! 
/*
  BOOL 
  CLexicon::Init( CLexicon &TM )
  {
  m_pTextE = TM.m_pTextE;
  m_pTextF = TM.m_pTextF;
  m_VocSizeE = TM.m_VocSizeE;
  m_VocSizeF = TM.m_VocSizeF;
  m_pT = TM.m_pT;
  m_TCounts = TM.m_TCounts;

  InitCounts ();

  return TRUE;
  }
*/


void
CLexicon::SetLambdaSmooth( float LambdaSmooth )
{
  // smoothing lexicon probabilities by p_smooth(f|e) = (1-lambda)p(f|e) + lambda*1/|VocF|
  // DefaultValue of entire matrix is 1/|VocF|
  // DefaultValue for empty rows is also 1/|VocF|
  // DefaultValue for non-empty rows is LambdaSmooth/|VocF|


  Trace( 1, "CLexicon::SetLambdaSmooth " << LambdaSmooth );

  assert( m_pVocF );
  assert( m_pVocE );

  float DefaultValue;

  for ( int Row = m_Pr.MinRow(); Row <= m_Pr.MaxRow(); Row++ ) {
    if ( !m_IsExpanded[ Row ] ) {
      CSparseArray & Pr = m_Pr.GetRow( Row );
      if ( Pr.NumberOfItems() == 0 ) {
        DefaultValue = 1.0 / m_pVocF->Size();
        Trace( 2, "no items - default = 1 / m_pVocF->Size() " << DefaultValue );
      } else {
        DefaultValue = LambdaSmooth / m_pVocF->Size();
        Trace( 2, "no items - default = LambdaSmooth / m_pVocF->Size() " << DefaultValue );
        // undo last smoothing, then calculate new smoothing
        Pr.Div( 1.0 - m_LambdaSmooth );
        Pr.Mul( 1.0 - LambdaSmooth );
      }
      Pr.SetDefaultValue( DefaultValue );
    } else {
      cout << "WARNING: Smoothing needs to be revised for expanded lexicon" << endl;
    }
  }

  m_LambdaSmooth = LambdaSmooth;

  //Renormalize();
}




void 
CLexicon::Renormalize()
{
  Trace( 0, "CLexicon::Renormalize()" );

  float S;
  int Row;
  //float DefaultValue;
  //long Index;
  //float Value;

  float SmoothValue = m_LambdaSmooth / m_VocSizeF;

  Trace( 1, "MinRow=" << m_Pr.MinRow() << "  MaxRow=" << m_Pr.MaxRow() );
  for ( Row = m_Pr.MinRow(); Row <= m_Pr.MaxRow(); Row++ ) {
    if ( !m_IsExpanded[ Row ] ) {
      CSparseArray & Pr = m_Pr.GetRow( Row );

      Trace( 5, "Row = " << Row << " not expanded: " 
             << "  DefaultValue = " << Pr.DefaultValue()
             << "  Sum = " << S 
             << "  SumOfStored = " << Pr.SumOfStoredItems() );
	   
      S = Pr.Sum( m_VocSizeF );
      Trace( 2, "Sum = " << S );

      if ( S > 0.0 ) {

        if ( S == 1.0 ) {
          Trace( 5, "  Sum already = " << S );
        } else {

          Trace( 5, "  Sum = " << S );
          //Trace( 5, "SetDefaultValue( 0.0 )" );
          Pr.Mul( 1.0 / S );
          S = Pr.Sum( m_VocSizeF );
          Trace( 5, "  Sum (Mul( 1.0 / S) = " << S );
	  
          Pr.Mul( (1.0 - m_LambdaSmooth) );
          S = Pr.Sum( m_VocSizeF );
          Trace( 5, "  Sum Mul( (1.0 - m_LambdaSmooth) )= " << S );
	  
          Pr.Add( SmoothValue );
          S = Pr.Sum( m_VocSizeF );
          Trace( 5, "  Sum Add( SmoothValue )= " << S );
	  
          Trace( 3, "  Sum final = " << S ); 
	  
        }

      } else {
        Trace( 5, "  Sum = " << S );
        Pr.SetDefaultValue( 1.0 / m_VocSizeF );
        S = Pr.Sum( m_VocSizeF );
        Trace( 5, "  Sum SetDefaultValue( 1.0 / m_VocSizeF )= " << S );
        Trace( 3, "  Sum final = " << S ); 
      }

      /*
        CSparseArray & B = m_Pr.GetRow( Row );
        //B.Display();
        Trace( 5, "Row = " << Row 
        << "  DefaultValue = " << B.DefaultValue()
        << "  Sum = " << B.Sum()
        << "  SumOfStored = " << B.SumOfStoredItems() );
	
        SumT = 0.0;
        for (Wf = 0; Wf < m_VocSizeF; Wf++) {
        Trace( 5, "m_Pr() " << m_Pr.GetAt( Row, Wf )
        << "   Row.GetAt() " << B.GetAt( Wf ) );
        SumT += m_Pr.GetAt( Row, Wf );
        }
        Trace( 1, "Sum T(f|" <<  m_pVocE->GetWord( Row ) << ") = " << SumT );
      */
    } else {

      assert( false );

      //      CArray<float> & Counts = m_ExpandedCounts.GetRow( Row );
      //CArray<float> & ExpandedPr = m_ExpandedPr.GetRow( Row );
      CSparseArray & Pr = m_Pr.GetRow( Row );

      float Sum = 0.0;
      //int f;
      //for ( f = Counts.MinIndex(); f <= Counts.MaxIndex(); f++ ) {
      //Sum += Counts.GetAt( f );
      //}

      //Sum = ExpandedPr.Sum();
      Trace( 3, "Row = " << Row << " expanded: "
             << "  Sum = " << Sum );
      if ( Sum > 0.0 ) {
        ; //ExpandedPr.Mul( 1.0 / Sum );
      } else {
        ; //ExpandedPr.SetAll( 1.0 / m_VocSizeF );
      }

      Sum = Pr.Sum();
      Trace( 3, "Row = " << Row << " dito not expanded: "
             << "  Sum = " << Sum );
      if ( Sum > 0.0 ) {
        Pr.Mul( 1.0 / Sum );
      } else {
        Pr.SetDefaultValue( 1.0 / m_VocSizeF );
        //S = Pr.Sum( SizeF );
        //Trace( 1, "  Sum = " << S );
      }

      //	for ( f = Counts.MinIndex(); f <= Counts.MaxIndex(); f++ ) {
      //if ( Counts.GetAt( f ) > 0 ) {
      //  Pr.SetAt( f , (1.0 - m_LambdaSmooth) * Counts.GetAt( f ) / Sum + SmoothValue );
      //} else {
      //  Pr.SetAt( f, SmoothValue );
      //}
      //}
      //	A.SetDefaultValue( SmoothValue );
      //} else {
      //	Pr.SetAll( 1.0 / m_VocSizeF );
      //	A.SetDefaultValue( 1.0 / m_VocSizeF );
      //}

    }
  }
}




BOOL
CLexicon::TestProbabilities() const
{
  //TraceLevel = 1;

  Trace( 0, "Test Probabilities" );
  // TODO: What's that? Maybe "if (TRACELEVEL ..." needed here
  std::cout.setf(ios::scientific);

  // test to see if probabilities add up to one
  //cout << "Test if probability sum of t(f|e) over f adds up to 1\n";
  
  CWordIndex We;
  float Sum, SumStored;
  float Diff;


  int SizeF = m_VocSizeF;

  for ( We = 0; We <= m_VocSizeE; We++ ) {
    Trace( 3, "" <<  We );
    if (!m_IsExpanded[ We ] ) {
      Trace( 2, "not expanded: ");
      const CSparseArray & A = m_Pr.GetRow( We );
      // Sum() add stored and not stored items, i.e. sum of storted
      // plus num of not stored times default
      Sum = A.Sum( SizeF );
      SumStored = A.SumOfStoredItems();
      Trace( 2, "We = " << We << " not expanded - sum of stored = " << SumStored 
             << " total = " << Sum );

    } else {
      Trace( 2, "expanded: ");
      CArray<float> & A =  CArray2D<float>(m_ExpandedPr).GetRow( We );
      CNumericalArray<float> &A1 = (CNumericalArray<float> &)A;
 
      Sum = A1.Sum();

      //CArray<float> & Pr = ((CMatrix<float>)m_ExpandedPr).GetRow( We );
      //Sum = 0.0;
      //int f;
      //for ( f = Pr.MinIndex(); f <= Pr.MaxIndex(); f++ ) {
      //Sum += Pr.GetAt( f );
      //}
      Trace( 2, "We = " << We << " expanded: "
             << "  Sum = " << Sum );
    }

    /*
      Diff = Sum - 1.0;
      Trace( 3, "Sum of Probs: " << Sum << " Diff: " << Diff );

      if ( Diff < 0.0 ) Diff = -Diff;

      if ( Diff > 1.0e-6 ) {
      if (m_IsExpanded[ We ] ) {
      Trace0( 0, "expanded: ");
      } else {
      Trace0( 0, "not expanded: ");
      }
      Trace( 0, "Sum T(f|" 
      <<  m_pVocE->GetWord( We ) << ") = " << Sum << " is not equal 1.0" );
      }
    */

  }
  
  return TRUE;
}




////////////////////////////////////////////////////////////////////////
// Methods to compute the log p(f|e) from the p(f|e)
////////////////////////////////////////////////////////////////////////
BOOL 
CLexicon::Log()
{
  Trace( 1, "Log" );

  //m_LogPr = m_Pr;

  Trace( 3, "before Clear" );
  // remember default value
  //float DefaultValue = m_Pr.DefaultValue();
  //if ( DefaultValue < 1.0e-10 ) {
  //  Trace( 0, "Wrong DefaultValue " << DefaultValue );
  //  DefaultValue = 1.0e-10;
  //}

  m_LogPr.Clear();

  Trace( 3, "after Clear; now Add Pr" );
  m_LogPr.Add( m_Pr );

  //m_LogPr.SetDefaultValue( DefaultValue );
  Trace( 3, "after Add Pr" );

  //m_Pr.Display();

  m_LogPr.Log();
  //m_PrAsLog = TRUE;
  return TRUE;
}




////////////////////////////////////////////////////////////////////////
// Display - Methods
////////////////////////////////////////////////////////////////////////

void 
CLexicon::Display() const
{
  //m_Pr.Display();

  float Value;
  cout << "\nDisplay lexicon\n\n";

  int VocSizeF = m_VocSizeF;
  int VocSizeE = m_VocSizeE;

  const char *pWordE, *pWordF;
  cout << "T: "<< "\n";
  cout << "Size of F:" << VocSizeF << "\n";
  cout << "Size of E:" << VocSizeE << "\n";
  cout.flush();
  //  cout.precision(4);

  // count how often word pairs accur in parallel sentences
  //CalculateTCounts();
  //DisplayTCounts();

  // display only those with values 10 times above uniform distribution
  float CutValue = 10.0 / (float)VocSizeF;
  CutValue = 0.0;
  cout << "Only Word Pairs with T(f_j|e_i) >= " << CutValue << " are shown \n";
  cout.flush();

  // we don't consider at the moment the <UNK> which is at index 0
  // therefore start loops with 1
  for (int i = 1; i <= VocSizeE; i++) {
    pWordE = m_pVocE->GetWord(i); 
    //cout << pWordE << " (" << m_pTextE->GetVocCount(i) << ")\n";
    cout << pWordE << endl;

    for (int j = 1; j <= VocSizeF; j++) {
      Value = m_Pr.GetAt(i,j);
      if (Value >= CutValue) {
        ostringstream String;
        pWordF = m_pVocF->GetWord(j);
        //String << "    " <<  pWordF  << " (" << m_pTextF->GetVocCount(j) 
        //       << ")                                                     ";
        String << "    " <<  pWordF
               << "                                                     ";
        cout << String.str();
        //cout << "    " << pWordF  << " (" << m_pTextF->GetVocCount(j) << ")\t";
        cout.setf(ios::left | ios::fixed);
        cout.width(6);
        cout.precision(3);
	
        //cout <<  m_pT->GetAt(i,j) << "  (" << (int)m_TCounts->GetAt(i,j) << ")\n";
        cout <<  m_Pr.GetAt(i,j) << std::endl;
        cout.flush();
      }
    }
    cout << "\n";
  }
}




////////////////////////////////////////////////////////////////////////
// Store and Load - Methods
////////////////////////////////////////////////////////////////////////

BOOL 
CLexicon::Store( const char *pFileName, int dataVersion )
{
  Trace(1,"CLexicon::Store( File = " << pFileName << ", V = " << dataVersion << " )");

  // return code
  BOOL rc = FALSE;   // should be set later, if not: error

  // open stream for output file
  ofstream file(pFileName);
  // error?
  if (!file) {
    cout << "ERROR:  CLexicon::Store(): File " << pFileName << " could not be opened for writing." << endl;
    return FALSE;
  } // endif: error
  
  // distinguish data versions
  switch (dataVersion) {
  case 4: { // Good Old Dataversion 4
    Shrink();

    //cout << "after shrink" << endl << flush;
    
    // dataversion 4: print lexicon header

    file.setf(ios::scientific);
    file.precision(6);

    file << "# Lexicon-Probabilities" << endl;
    file << "# for each e: index of e" << endl;
    file << "# for each f with p(f|e) > 0: index of f, p(f|e)" << endl;
    
    int Version = 4;
    file << "Dataversion: " << Version << endl;
    
    time_t Time = time(NULL);
    char *String;
    String = asctime(localtime(&Time));
    file << "Stored on: " << String;
    
    if ( !m_pVocE )
      file << "Vocabulary Size E: " << -1 << endl;
    else
      file << "Vocabulary Size E: " << m_VocSizeE << endl;
    
    if ( !m_pVocF )
      file << "Vocabulary Size F: " << -1 << endl;
    else
      file << "Vocabulary Size F: " << m_VocSizeF << endl;
    
    file << flush;

    // dataversion 4: print data
    rc = m_Pr.Store( file );
    
  } break; // Dataversion 4

  case 5: // Dataversion 5
    // delegate to subroutine, distinguish return codes
    switch (Store_Dataversion5(file)) {
    case IOERR_OK : // no error
      rc = TRUE;
      break; // IOERR_OK

    default:
      cout << "ERROR:  CLexicon::Store(): store_dataversion5() returned unrecognized code." << endl;
      rc = FALSE;
      break;
    } // endswitch store_dataversion5 return code

    break; // dataversion 5

  default: // unknown dataversion
    cout << "ERROR:  CLexicon::Store(): Unknown dataversion '" << dataVersion
         << " for storage of " << pFileName << " requested." << endl;
    rc = FALSE;
    break; // dataversion ?
  } // endswitch

  // cleanup
  file.close();

  return rc;
} // CLexicon::Store(...)



BOOL
CLexicon::StoreWithWords( const char *pFileName, float PruneValue ) const
{

  ofstream TFile (pFileName);

  if (!TFile) {
    cout << "ERROR:  File " << pFileName << " could not be opened for writing" << endl;
    return FALSE;
  }

  //  ofstream TFile;
  //  TFile.open(pFileName);
  //  assert(TFile);

  TFile.setf(ios::scientific);
  TFile.precision(6);

  //  TFile << "# Lexicon-Probabilities\n";
  //TFile << "# for each e: index of e\n";
  //TFile << "# for each fwith p(f|e) >= " << PruneValue << ": index of f, p(f|e)\n";

  //int TVersion = 9;
  //TFile << "Dataversion: " << TVersion << endl;

  //time_t Time = time(NULL);
  //char *String;
  //String = asctime(localtime(&Time));
  //TFile << "Stored on: " << String;

  //TFile << m_VocSizeE << "\n";
  //TFile << m_VocSizeF << "\n";
  //TFile << endl;

  float Value;
  const char *pWordE, *pWordF;
  CWordIndex e, f;
  //long int e, f;
  int n;

  for ( e = 0; e < m_VocSizeE; e++ ) {
    pWordE = m_pVocE->GetWord( e ); 
    const CSparseArray & Pr = m_Pr.GetRow( e );
  
    for ( n = 0; n < Pr.NumberOfItems(); n++ ) {
      Pr.GetItem( n, f, Value );
      Trace( 5, "Position = " << n << "  f = " << f
             << "  Pr = " << Value );
      if ( Value > PruneValue ) {
        pWordF = m_pVocF->GetWord( f ); 
        TFile <<  pWordF << " \t " << pWordE << " \t " << Value << "\n";
      }
    }
  }


  //int f;
  //for ( f = Counts.MinIndex(); f <= Counts.MaxIndex(); f++ ) {
  //Sum += Counts.GetAt( f );
  //}
  
  /*
    for ( f = 0; f < m_VocSizeF; f++ ) {
    pWordF = m_pVocF->GetWord( f ); 
    
    for ( e = 0; e < m_VocSizeE; e++ ) {
    Value = m_Pr.GetAt( e, f );

    if ( Value > PruneValue ) {
		pWordE = m_pVocE->GetWord( e ); 
		TFile <<  pWordF << " \t " << pWordE << " \t " << Value << "\n";
    }
    }

    }
  */

  TFile.close();

  return TRUE;
}



BOOL
CLexicon::StoreWithWordsGIZA( const char *pFileName, float PruneValue ) const
{
  ofstream TFile (pFileName);

  if (!TFile) {
    cout << "ERROR:  File " << pFileName << " could not be opened for writing";
    return FALSE;
  }

  //  ofstream TFile;
  //  TFile.open(pFileName);
  //  assert(TFile);

  TFile.setf(ios::scientific);
  TFile.precision(6);

  //  TFile << "# Lexicon-Probabilities\n";
  //TFile << "# for each e: index of e\n";
  //TFile << "# for each fwith p(f|e) >= " << PruneValue << ": index of f, p(f|e)\n";

  //int TVersion = 9;
  //TFile << "Dataversion: " << TVersion << endl;

  //time_t Time = time(NULL);
  //char *String;
  //String = asctime(localtime(&Time));
  //TFile << "Stored on: " << String;

  //TFile << m_VocSizeE << "\n";
  //TFile << m_VocSizeF << "\n";
  //TFile << endl;

  float Value;
  const char *pWordE, *pWordF;
  CWordIndex e, f;
  //long int e, f;
  int n;

  for ( e = 0; e < m_VocSizeE; e++ ) {
    pWordE = m_pVocE->GetWord( e ); 
    const CSparseArray & Pr = m_Pr.GetRow( e );
  
    for ( n = 0; n < Pr.NumberOfItems(); n++ ) {
      Pr.GetItem( n, f, Value );
      Trace( 5, "Position = " << n << "  f = " << f
             << "  Pr = " << Value );
      if ( Value > PruneValue ) {
        pWordF = m_pVocF->GetWord( f );
        TFile <<  pWordE << " \t " << pWordF << " \t " << Value << "\n";
      }
    }
  }

  TFile.close();

  return TRUE;
}



BOOL
CLexicon::LoadWithWords( const char *pFileName )
{

  ifstream TFile( pFileName );

  if (!TFile) {
    cout << "ERROR: File " << pFileName << " could not be opened for reading";
    return FALSE;
  }

  //  ofstream TFile;
  //  TFile.open(pFileName);
  //  assert(TFile);

  //TFile.setf(ios::scientific);
  //TFile.precision(6);

  //  TFile << "# Lexicon-Probabilities\n";
  //TFile << "# for each e: index of e\n";
  //TFile << "# for each fwith p(f|e) >= " << PruneValue << ": index of f, p(f|e)\n";

  //int TVersion = 9;
  //TFile << "Dataversion: " << TVersion << endl;

  //time_t Time = time(NULL);
  //char *String;
  //String = asctime(localtime(&Time));
  //TFile << "Stored on: " << String;

  //TFile << m_VocSizeE << "\n";
  //TFile << m_VocSizeF << "\n";
  //TFile << endl;

  float Value;
  char *pWordE = new char[ 200 ];
  char *pWordF = new char[ 200 ];
  CWordIndex e, f;

  while ( !TFile.eof() ) {

    TFile >> pWordF >> pWordE >> Value;
    //cout << pWordF << " \t " << pWordE << " \t " << Value << "\n";

    f = m_pVocF->GetIndex( pWordF ); 
    e = m_pVocE->GetIndex( pWordE ); 
    //cout << "store p( " << f << " | " << e << " ) = " << Value << endl;
    
    m_Pr.SetAt( e, f, Value );

  }
  TFile.close();


  // no rows expanded
  m_IsExpanded.SetBoundaries( m_Pr.MinRow(), m_Pr.MaxRow() ); 
  m_IsExpanded.SetAll( FALSE );


  return TRUE;
}


BOOL
CLexicon::LoadWithWordsGIZA( const char *pFileName )
{

  cout << "CLexicon::Load Giza format" << endl << flush;

  ifstream TFile( pFileName );

  if (!TFile) {
    cout << "ERROR:  File " << pFileName << " could not be opened" << endl;
    return FALSE;
  }

  float Value;
  char *pWordE = new char[ 200 ];
  char *pWordF = new char[ 200 ];
  CWordIndex e, f;

  int s = 0;

  while ( !TFile.eof() ) {
	
    TFile >> pWordE >> pWordF >> Value;
    s++;

    
    if ( (s % 1000000) == 0 ) {
      cout << s << " lexicon itmes loaded" << endl << flush;
    }

    //cout << pWordF << " \t " << pWordE << " \t " << Value << "\n";
    //}
    

    f = m_pVocF->GetIndex( pWordF ); 
    if ( f == CVocabulary::UNK) {
      // Word from internal vocab not covered by application vocab: add dynamically
      f = m_pVocF->AddWord( CString( pWordF ) );
    }

    e = m_pVocE->GetIndex( pWordE ); 
    if ( e == CVocabulary::UNK) {
      // Word from internal vocab not covered by application vocab: add dynamically
      e = m_pVocE->AddWord( CString( pWordE ) );
    }

    /*
      if ( s < 20 ) {
      cout << "store p( " << f << " | " << e << " ) = " << Value << endl;
      }
    */

    m_Pr.SetAt( e, f, Value );
    
  }

  TFile.close();
  
  // is it guaranteed that we have for each index in [MinIndex, MaxIndex] an entry
  // in the lexicon?  If a word index in between has not been seen then the
  // sparse matrix may not have a row.


  // no rows expanded
  m_IsExpanded.SetBoundaries( m_Pr.MinRow(), m_Pr.MaxRow() ); 
  m_IsExpanded.SetAll( FALSE );


  // it could be the case that we loaded an empty file
  // this should be considered an error

  if ( m_Pr.NumOfRows() < 2 ) {
    cout << "WARNING: only " << m_Pr.NumOfRows() << " target words on the lexicon, "
         << "perhaps something wrong with the file" << endl;
    return FALSE;
  }

  return TRUE;
}



/**
   @brief Load lexicon from a file.
   @param pFileName Filename of the lexicon file to be loaded (usually ending in .T)
   @return Success indicator.

   This load function can load any lexicon dataversion from 4 upwards
   (as far as known at the point of writing). Firstly, the header common
   to all lexicon data formats (known so far) is read in. That is,
   a number of comment lines starting with '#' and the line defining
   the "dataversion". The function then delegates the loading task
   to the appropiate subroutine.
*/
BOOL
CLexicon::Load(const char *pFileName)
{
  Trace(1,"CLexicon::Load(" << pFileName << ")");

  // constants
  static const int BUFFERSIZE = 256;
  static const int COMMENT_CHAR = '#';
  static const char DATAVERSION_KEYWORD[] = "Dataversion:";

  // buffers for reading
  // "String" is a dangerous variable name for a string ...
  // TODO: why not "string" here?
  char aString[BUFFERSIZE];
  char readBuffer[BUFFERSIZE];
  char dummy[BUFFERSIZE];
  int dataVersion;

  // other local vars
  int cLine = 0;         // line counter
  bool flError = false;  // error flag

  BOOL rc = FALSE; // return code

  // open input file;
  ifstream file(pFileName);

  // error on open?
  if (!file) {
    cout << "ERROR: CLexicon::Load(): File " << pFileName << " could not be opened for read"  << endl;
    return FALSE;
  } // endif: error reading input file

  // skip comment lines at beginning
  while (file.peek() == '#' && file.good()) {
    file.getline(readBuffer, BUFFERSIZE);
    ++cLine;
  } // endwhile

  // first line after comment MUST contain dataversion in the form
  // Dataversion: <VERSION>
  readBuffer[0] = '\0';  // for security ...
  file >> readBuffer; // TODO: dangerous? (if |line| > BUFFERSIZE)
  file >> dataVersion;

  // check syntax
  if (strcmp(readBuffer,DATAVERSION_KEYWORD) != 0) {
    // error
    cout << "ERROR: CLexicon::Load(): Syntax error while loading lexicon from "
         << pFileName << " (gcount() == " << file.gcount()
         << "): Expected " << DATAVERSION_KEYWORD
         << " but got " << readBuffer << "." << endl;
    cout << "  File is not a lexicon or lexicon format incompatible." << endl;
    return FALSE;
    // todo: maybe exception here
  } // endif: wrong keyword

  // assert: keyword matches

  // could not parse integer?
  if (file.fail()) {
    // error
    cout << "ERROR:  CLexicon::Load(): Syntax error while loading lexicon from "
         << pFileName << " (gcount() == " << file.gcount()
         << "): Could not parse dataversion." << endl;
    cout << "  File has been messed up or lexicon format incompatible." << endl;
    return FALSE;
  } // endif

  Trace( 1, "CLexicon::Load(): Found Dataversion " << dataVersion );

  // distinguich between dataversions
  switch (dataVersion) {
  case 0 :
  case 3 :
    cout << "ERROR:  Dataversion " << dataVersion << " not supported any more." << endl;
    rc = FALSE;
    break; // dataVersion 0

  case 4 : // Good Old Dataversion 4
    // consume newline
    file.ignore(BUFFERSIZE,'\n');
    // next line: time when stored
    file.getline(readBuffer, BUFFERSIZE);
    // TODO: a check here if line start with "Stored on"
    Trace(1,"CLexicon::Load(): Storage date: " << readBuffer);
    
    // TODO: Check keywords
    file >> dummy >> dummy >> dummy >> m_VocSizeE;
    file >> dummy >> dummy >> dummy >> m_VocSizeF;
    // could not parse integer?
    if (file.fail()) {
      // error
      cout << "ERROR: CLexicon::Load(): Syntax error while loading lexicon from "
           << pFileName << " (gcount() == " << file.gcount()
           << "): Could not parse vocabulary sizes." << endl;
      // This should never happen in Dataversion 4
      cout << "ERROR  File has been messed up or ... or ... whatever ..." << endl;
      return FALSE;
    } // endif: error
    
    Trace(1, "CLexicon::Load(): VocSizeE = " << m_VocSizeE << " VocSizeF = " << m_VocSizeF);

    // Do the actual loading
    rc = m_Pr.Load( file );
    
    break; // dataVersion 4
    
  case 5 : // Dataversion 5
    // the subroutine handles all further reading, returns error code
    switch (Load_Dataversion5(file)) {
    case IOERR_OK : // success!
      rc = TRUE;
      break; // IOERR_OK
      
    case IOERR_SYNTAX_V5 :
      // in this version, subroutine already emitted error msg (bad style :-/)
      rc = FALSE;
      break; // IOERR_OPENREAD

    case IOERR_BAD_VERSION :
      // dito
      rc = FALSE;
      break; // IOERR_BAD_VERSION
      
    default :
      cout << "ERROR:  CLexicon::Load(): Unknown return code of dataversion specific subroutine." << endl;
      rc = FALSE;
      break;
    } // endswitch: load_dataversion5 return code
    break; // Dataversion 5
  } // endswitch: distinguish dataVersions

  // File has been read, close it
  file.close();

  // no rows expanded
  m_IsExpanded.SetBoundaries( m_Pr.MinRow(), m_Pr.MaxRow() ); 
  m_IsExpanded.SetAll( FALSE );

  return rc;  
} // CLexicon::Load()




BOOL
CLexicon::StoreCountsWithWords( const char *pFileName, float PruneValue ) const
{

  ofstream TFile (pFileName);

  if (!TFile) {
    cout << "File " << pFileName << " could not be opened for writing" << endl;
    return FALSE;
  }

  //TFile.setf(ios::scientific);
  //TFile.precision(6);

  float Value;
  const char *pWordE, *pWordF;
  CWordIndex e, f;
  //long int e, f;
  int n;

  for ( e = 0; e < m_VocSizeE; e++ ) {

    //cout << e << endl;
    pWordE = m_pVocE->GetWord( e ); 
    const CSparseArray & Counts = m_Counts.GetRow( e );
  
    for ( n = 0; n < Counts.NumberOfItems(); n++ ) {

      Counts.GetItem( n, f, Value );
      //cout <<  "Position = " << n << "  f = " << f << "  Count = " << Value << endl;

      if ( Value > PruneValue ) {
        pWordF = m_pVocF->GetWord( f ); 
        //cout << "Above threshhold "  <<  pWordF << " \t " << pWordE << " \t " << (long)Value << "\n" << endl;
        TFile << pWordF << " " << pWordE << " " << (int)Value << "\n";
      } else {
        //pWordF = m_pVocF->GetWord( f ); 
        //cout << "Below threshhold "  <<  pWordF << " \t " << pWordE << " \t " << (long)Value << "\n" << endl;
      }
    }
  }


  TFile.close();

  return TRUE;
}



/**
   @brief Store word translation probabilities in a file.
   @param out Stream to output lexicon on.
   @return
   <ul>
   <li>IOERR_OK        - No error.</li>
   </ul>

   In this "dataversion", the "lexicon" contains a local, file-intern
   vocabulary, which is mapped onto the application-wide vocabulary during
   load.

   This routine is a "submodule" of Store() (as indicated by the underslash)
   and should not be called directly by the user.

   \todo Add more some file.fail() checks for loops.
*/
int CLexicon::Store_Dataversion5( ostream& out )
{
  Trace(1,"CLexicon::Store_Dataversion5()");

  // the data version field
  static const int DATAVERSION = 5;
  // version modifier
#ifdef DATAVERSION5_PLAIN
#ifdef VOCAB_STATIC_EOS
  // plain text sparse array, EOS statically 1
  assert(CVocabulary::EOS == 1);
  static const char VERSIONMOD[] = "plain_s-eos";
#else
  // plain text sparse array, EOS dynamic (last index)
  static const char VERSIONMOD[] = "plain_d-eos";
#endif
#else
#ifdef VOCAB_STATIC_EOS
  // binary sparse array, EOS statically 1
  assert(CVocabulary::EOS == 1);
  static const char VERSIONMOD[] = "binary_s-eos";
#else
  // binary sparse array, EOS dynamic (last index)
  static const char VERSIONMOD[] = "binary_d-eos";
#endif
#endif
  // a string
  CString str;
  // container for used rows and columns in sparse array.
  CSparseMatrix::CIndexContainer usedRows, usedCols;
  // an iterator of containers above.
  CSparseMatrix::CIndexContainer::iterator iter;
  // Internal vocabulary
  CVocabulary internalVocab;
  // index in internal vocab / application vocab
  CWordIndex vocIdxInt, vocIdxApp;
  // translation tables for app vocabularies to internal
  CWordIndex* vocFTransTable;
  CWordIndex* vocETransTable;
  // a value in the matrix / a probability
  float value;
  // for row traversal / number of items
  int   pos;
  // for row traversal / vocab index
  int   index;

  // make data structure more compact before write out
  Shrink();
  // set how numbers are printed
  out.setf(ios::scientific);
  out.precision(6);

  // write header
  out << "# Lexicon (word translation probs)\n";
  out << "# for each e: index of e\n";
  out << "# for each f with p(f|e) > 0: index of f, p(f|e)\n";

  out << "Dataversion: " << DATAVERSION << " " << VERSIONMOD << endl;

  time_t storeTime = time(NULL);
  string sTime = asctime(gmtime(&storeTime));
  // strip newline
  if (sTime[sTime.size()-1] == '\n')
    sTime.erase(sTime.size()-1,1);
  out << "Stored_on: " << sTime << " GMT" << endl << flush;

  // end of common header

  // Make sure the columns are up-to-date
  m_Pr.UpdateDimensions();
  // collect used column / row indices
  m_Pr.CollectUsedIndices(usedRows, usedCols);
  // todo: maybe sort columns

  // probably more than we need, but this is not memory critical
  internalVocab.Allocate(usedRows.size()+usedCols.size());

  // if this ever changes you might want to review this code
  assert(m_Pr.MinRow() >= 0);
  // min row 0 assumed, also for for later indexing
  vocETransTable = new CWordIndex[m_Pr.MaxRow()+1];
  assert(CVocabulary::UNK == 0);
  vocETransTable[CVocabulary::UNK] = CVocabulary::UNK;  // map unknown word

  // fill internal vocab: rows
  for ( iter = usedRows.begin(); iter != usedRows.end(); ++iter) {
    // get row index in sparse matrix
    vocIdxApp = *iter;    
    // only consider positive word indices
    if (vocIdxApp > 0) {
      assert(vocIdxApp != CVocabulary::UNK);
      assert(vocIdxApp <= m_Pr.MaxRow());

      // already in internal vocab?
      str = m_pVocE->GetWord(vocIdxApp);
      // Note: static </s> would return EOS = 1 here ...
      vocIdxInt = internalVocab.GetIndex(str);
      // note: vocIdxApp cannot be UNK, so str cannot be <UNK>
      if (vocIdxInt == CVocabulary::UNK ) {
        // insert new, update mapping
        vocETransTable[vocIdxApp] = internalVocab.AddWord(str);
      } else {
        // use existing index to update mapping
        // Note: at this point, this can only happen for a static EOS
#ifndef VOCAB_STATIC_EOS
        assert(0);
#endif
        assert(vocIdxApp == m_pVocE->EOS);
        assert(vocIdxInt == internalVocab.EOS);
        vocETransTable[vocIdxApp] = vocIdxInt;
      } // endif: new word
    } // endif: not EOF
  } // endfor: copy words form E to internal vocab

  // TODO: vocETransTable should be identity, shouldn't it?!

  // if this ever changes you might want to review this code
  assert(m_Pr.MinCol() >= 0);
  // min row 0 assumed, also for for later indexing
  vocFTransTable = new CWordIndex[m_Pr.MaxCol()+1];
  // map unknown word manually
  assert(CVocabulary::UNK == 0);
  vocFTransTable[CVocabulary::UNK] = CVocabulary::UNK;  // map unknown word

  // fill internal vocab: columns
  for (iter = usedCols.begin(); iter != usedCols.end(); ++iter ) {
    // get column index in sparse matrix
    vocIdxApp = *iter;
    assert(vocIdxApp >= 0);
    // only consider positive word indices; don't add unknown word again
    if (vocIdxApp > 0) {
      assert(vocIdxApp != CVocabulary::UNK);
      assert(vocIdxApp <= m_Pr.MaxCol());

      // already in internal vocab?
      str = m_pVocF->GetWord(vocIdxApp);
      // Note: </s> would return EOS = 1 here ...
      vocIdxInt = internalVocab.GetIndex(str);
      // note: vocIdxApp cannot be UNK, so str cannot be <UNK>
      if (vocIdxInt == CVocabulary::UNK )
        // insert new, update mapping
        vocFTransTable[vocIdxApp] = internalVocab.AddWord(str);
      else
        // use existing index to update mapping
        vocFTransTable[vocIdxApp] = vocIdxInt;
    } // endif: unknown word already added
  } // endfor: copy words from F to internal vocab

  // assert: we have internal vocabulary and translation table
  // for VocE: identity, for VocF: vocFTransTable

#ifndef VOCAB_STATIC_EOS
  // add sentence end (godd*mn vocabulary class won't store last element otherwise!)
  vocIdxInt = internalVocab.AddWord((string)"</s>");
#endif // EOS already in statically on index 1?

  // Store internal vocabulary
  Trace(2,"CLexicon::Store_Dataversion5(): Storing internal vocab ...");
  out << "Internal_Vocabulary_Size: " << internalVocab.Size() << endl;
  internalVocab.Store(out,true);
  //  internalVocab.Store(out,false);
  //  out << "</s>" << flush;

  // store probabilities
  Trace0(2,"CLexicon::Store_Dataversion5(): Storing data ");
  // loop over rows (target language words)
  for (iter = usedRows.begin(); iter != usedRows.end(); ++iter) {
    // get row number (vocab index of target word)
    vocIdxApp = *iter;
    assert(vocIdxApp >= 0);
    // a "." every 1000 target words
    TraceIf0(vocIdxApp%1000 == 0, 2, ".");
    // get row
    CSparseArray& row = m_Pr.GetRow(vocIdxApp);
    // line starts with internal vocab row index and number of entries 
    assert(vocIdxApp <= m_Pr.MaxRow());
#ifdef DATAVERSION5_PLAIN
    // plain text version
    if (iter != usedRows.begin())
      out << endl;
    out << vocETransTable[vocIdxApp] << " "
        << row.NumberOfItems() << " "
        << row.DefaultValue();
#else
    // binary
    vocIdxInt = vocETransTable[vocIdxApp];
    out.write(reinterpret_cast<char *>(&vocIdxInt),sizeof(vocIdxInt));
    pos = row.NumberOfItems();
    out.write(reinterpret_cast<char *>(&pos),sizeof(pos));
    value = row.DefaultValue();
    out.write(reinterpret_cast<char *>(&value),sizeof(value));
#endif
    // as we only store non-empty rows ...
    assert(row.NumberOfItems() > 0);
    // iterate over row entries
    for (pos = 0; pos < row.NumberOfItems(); ++pos) {
      row.GetItem(pos,index,value);
      vocIdxApp = index;
      // translate index and output
      assert(vocIdxApp <= m_Pr.MaxCol());
#ifdef DATAVERSION5_PLAIN
      out << endl << vocFTransTable[vocIdxApp] << " " << value;
#else
      vocIdxInt = vocFTransTable[vocIdxApp];
      out.write(reinterpret_cast<char *>(&vocIdxInt),sizeof(vocIdxInt));
      out.write(reinterpret_cast<char *>(&value),sizeof(value));
#endif
    } // endfor: loop over indices in current row
  } // endfor: loop over rows to store probs

  /* WORKAROUND
     In order to avoid buggy end-of-file detections, emit a

     idxE=0 nrOfTgtWords=0 defValue=0.0

     to mark end of lexicon. As empty lines of the matrix are not written
     out in this dataversion, this can be easily recognized by the
     nrOfTgtWords==0 field.
  */
#ifdef DATAVERSION5_PLAIN
  // plain text version
  out << endl << "0 0 0.0";
#else
  // binary
  vocIdxInt = 0;
  out.write(reinterpret_cast<char *>(&vocIdxInt),sizeof(vocIdxInt));
  pos = 0;
  out.write(reinterpret_cast<char *>(&pos),sizeof(pos));
  value = 0.0;
  out.write(reinterpret_cast<char *>(&value),sizeof(value));
#endif

  // writing done
  Trace(2," done.");

  // clean up
  delete[] vocFTransTable;
  delete[] vocETransTable;
  // no error
  return IOERR_OK;
} // CLexicon::Store_Dataversion5

/**
   @brief Load word translation probabilities.
   @param in The input stream (e.g. a file stream) from which to read the lexicon.
   @return
   <ul>
   <li>IOERR_OK        - No error.</li>
   <li>IOERR_SYNTAX_V5 - The data in the file doesn't seem to be a lexicon dataversion 5.</Li>
   </ul>

   In this "dataversion", the file contains a local, file-intern
   vocabulary which is mapped onto the application-wide vocabulary during load.

   The basic header common to all lexica (known so far) must already have been read.
   That is, some optional comment lines with leading "#", the keyword
   "Dataversion:" and the dataversion itself, that is "5". This routine starts reading
   from the stream immedeately after that.

   This method is a "subroutine" of Load() (as indicated by the underslash) and will be
   called after the header of the file being read indicates the appropriate dataversion.

   \todo Introduce some exceptions.
   \todo If BUFFERSIZE is too small, we get a problem here ...
   \todo Dynamically add vocab entries if internal vocab entry is unseen.
   \todo More and real checks instead casts
*/
int CLexicon::Load_Dataversion5(istream& in)
{
  Trace(1,"CLexicon::Load_Dataversion5()");

  // This dataversion will eventually require one common application vocab
  //  assert(m_pVocF == m_pVocE);

  // buffers
  static const int BUFFERSIZE = 1024;
  static const int DATAVERSION = 5;
  // version modifier
#ifdef DATAVERSION5_PLAIN
  // VOCAB_STATIC_EOS is set by the CVocabulary class to indicate whether CVocabulary::EOS is statically 1 or not.
  // NOTE: CVocabulary with dynamic EOS is DEPRECATED
#ifdef VOCAB_STATIC_EOS
  assert(CVocabulary::EOS == 1);
  static const char VERSIONMOD[] = "plain_s-eos";
#else
  static const char VERSIONMOD[] = "plain_d-eos";
#endif
#else
#ifdef VOCAB_STATIC_EOS
  assert(CVocabulary::EOS == 1);
  static const char VERSIONMOD[] = "binary_s-eos";
#else
  static const char VERSIONMOD[] = "binary_d-eos";
#endif
#endif
  char buffer[BUFFERSIZE];

  // internal vocabulary
  CVocabulary internalVocab;

  // A flag whether to fix lexica trained with old GraphAlign having danymic eos
  bool blFixEos = false;

  // here we are: comment line, "Dataversion:" and "5" have already been read

  // check subversion
  in >> buffer;  // version modifier

  // todo: check for line end
  if (strcmp(buffer,VERSIONMOD) != 0) {
    cout << "ERROR:  CLexicon::Load_Dataversion5(): "
         << "Wrong version modifier '" << buffer <<"'. This routine was compiled for "
         << VERSIONMOD << endl;
    // special case where version mismatch can be fixed: the d-eos subflag
    if ( strstr(VERSIONMOD,"s-eos") == VERSIONMOD + strlen(VERSIONMOD) - strlen("s-eos") &&
         strstr(buffer,"d-eos")     == buffer + strlen(buffer) -  strlen("d-eos") &&
         strlen(VERSIONMOD) == strlen(buffer) &&
         strncmp(buffer,VERSIONMOD,strlen(buffer)-5) == 0) {
      cout << "   Fixing version mismatch by shifting indices." << endl;
      blFixEos = true;
    } else {
      cout << "   Version modifier not supported, aborting loading of lexicon." << endl;
      return IOERR_BAD_VERSION; // error
    } // endif: special case of versionmod mismatch that can be fixed
  } // error occured

  // skip "Stored_on:"
  in >> buffer;
  // TODO: Check contents
  in.getline(buffer, BUFFERSIZE);
  // buffer now contains storage date
  Trace(2,"CLexicon::Load_Dataversion5(): Lexicon was stored on " << buffer);

  // dataversion 5 specific stuff

  // Internal vocab size
  int intVocabSize;
  in >> buffer;  // Skip "Internal_vocab_size:"
  in >> intVocabSize;
  if (in.fail()) {
    cout << "ERROR:  CLexicon::Load_Dataversion5(): "
         << "Could not determine internal vocab size. Fieldname / vocab size read so far was "
         << buffer << intVocabSize << endl;
    return IOERR_SYNTAX_V5; // error
  } // error occured
  // consume \n after intVocabSize
  in.get(buffer[0]);
  assert(buffer[0] == '\n');

  // read internal vocab
  Trace(1,"CLexicon::Load_Dataversion5(): Reading internal vocab. Size should be " << intVocabSize);
  internalVocab.Load(in);

  if (internalVocab.Size() != intVocabSize) {
    cout << "CLexicon::Load_Dataversion5: ERROR:";
    cout << "Internal vocab size " << internalVocab.Size();
    cout << " does not match specified size " << intVocabSize
         << endl;
    return IOERR_SYNTAX_V5; // error
  } // endif: internal vocab size mismatch

  // setup internal vocab -> application vocab
  // if we know there is only ONE application vocab, we only need one map here
#ifdef VOCAB_STATIC_EOS
  CWordIndex *int2appEMap = new CWordIndex[internalVocab.Last()+1];
  CWordIndex *int2appFMap = new CWordIndex[internalVocab.Last()+1];
  assert(CVocabulary::EOS == 1);
  int2appEMap[CVocabulary::EOS] = int2appFMap[CVocabulary::EOS] = CVocabulary::EOS;
#else
  CWordIndex *int2appEMap = new CWordIndex[internalVocab.Last()+2];
  CWordIndex *int2appFMap = new CWordIndex[internalVocab.Last()+2];
  assert(internalVocab.EOS == internalVocab.Last()+1);
  int2appEMap[internalVocab.EOS] = m_pVocE->EOS;
  int2appFMap[internalVocab.EOS] = m_pVocF->EOS;
#endif
  assert(CVocabulary::UNK == 0);
  int2appEMap[CVocabulary::UNK] = int2appFMap[CVocabulary::UNK] = CVocabulary::UNK;
  // index in internal and application vocab
  CWordIndex idxInt, idxApp;
  // the current word (yes, a string)
  CString w;
  // minimal/maximal index for target language seen so far
  CWordIndex eMinIdx(INT_MAX), eMaxIdx(INT_MIN);

  // vocab indices (without UNK, EOS) go from internalVocab.First ... internalVocab.Last
  assert(internalVocab.First() >= 0);
  // get mapping for each internal vocab word
  for (idxInt=internalVocab.First(); idxInt <= internalVocab.Last(); ++idxInt) {
    assert(idxInt != CVocabulary::UNK);
    // get current word
    w = internalVocab.GetWord(idxInt);
    // get application wide index (source)
    idxApp = m_pVocF->GetIndex(w);
    // app vocab covers internal vocab?
    if (idxApp == CVocabulary::UNK) {
      // Word from internal vocab not covered by application vocab: add dynamically
      idxApp = m_pVocF->AddWord(w);
    } // error: UNK in internal vocab
    assert(idxApp != CVocabulary::UNK);
    // store mapping (source
    int2appFMap[idxInt] = idxApp;

    // get application wide index (target)
    idxApp = m_pVocE->GetIndex(w);
    eMinIdx = min(eMinIdx,idxApp);
    eMaxIdx = max(eMaxIdx,idxApp);
    // app vocab covers internal vocab?
    if (idxApp == CVocabulary::UNK) {
      // Word from internal vocab not covered by application vocab: add dynamically
      idxApp = m_pVocE->AddWord(w);
      assert(idxApp != CVocabulary::UNK);
    } // error: UNK in internal vocab
    // store mapping (target)
    int2appEMap[idxInt] = idxApp;
  } // endfor: each internal vocab word

  // assert: internal vocab read, mappings set up

  // read data
  // here comes binary stuff, if binary mode
  Trace0(1,"CLexicon::Load_Dataversion5(): This is dataversion 5 "
         << VERSIONMOD << ". Reading data ");

  // some buffer variables
  // !! MAKE SURE TYPES MATCH WITH STORE() METHOD !!
  CWordIndex idxE, idxF;
  int nrOfSrcWords;
  float value;
  // iteration variable
  int i;
  // running number if target words
  int cTgtWords = 0;
  // end-of-lexicon: targetId=0 nrOfTgtWords=0 defValue=0.0
  bool notEOL = true;

  // do some pre-allocation
  m_Pr.SetBoundaries(min(eMinIdx,0),eMaxIdx,0,0);

  // until whole file read
  while (in && notEOL) { // good() == !fail() && !eof()
    // a "." every 1000 words
    TraceIf0((cTgtWords++)%1000 == 0, 2, ".");

    // read target word index and number of items, default value
#ifdef DATAVERSION5_PLAIN
    in >> idxE >> nrOfSrcWords >> value;
#else
    in.read(reinterpret_cast<char *>(&idxE),sizeof(idxE));    
    in.read(reinterpret_cast<char *>(&nrOfSrcWords),sizeof(nrOfSrcWords));
    in.read(reinterpret_cast<char *>(&value),sizeof(value));
#endif
    // error?
    if (in.fail())
      break;
    // value is a prob
    assert(value >= 0.0 && value <= 1.0);
    // EOL?
    if (nrOfSrcWords == 0) {
      assert(idxE == 0 && value == 0.0);
      notEOL = false;
      break;
    } // EOL found
    // get row
    assert(idxE <= internalVocab.Last());
    CSparseArray& row = m_Pr.GetRow(int2appEMap[idxE]);
    row.SetDefaultValue(value);
    // read source word / value pairs
    for (i=0; i<nrOfSrcWords; ++i) {
#ifdef DATAVERSION5_PLAIN
      in >> idxF >> value;
#else
      in.read(reinterpret_cast<char *>(&idxF),sizeof(idxF));
      in.read(reinterpret_cast<char *>(&value),sizeof(value));
#endif
      // error?
      if (in.fail())
        break;
      // value is a prob
      assert(value >= 0.0 && value <= 1.0);
      // translate indices and store
      assert(idxF <= internalVocab.Last());
      row.SetAt(int2appFMap[idxF], value);
    } // endfor: read source word/value pairs
  } // endwhile: read file until empty

  // end-of-lexicon should have been seen
  if (notEOL) {
    cout << "ERROR: CLexicon::Load_Dataversion5(): "
         << "End-of-Lexicon not found." << endl;
    return IOERR_SYNTAX_V5; // error
  } // EOL not seen

  // check if stream is still good
  if (in.fail()) {
    cout << "ERROR: CLexicon::Load_Dataversion5(): "
         << "An error occured while reading." << endl;
    return IOERR_SYNTAX_V5; // error
  } // endif: error

  Trace(2," Data read.");

  // clean up
  delete[] int2appEMap;
  delete[] int2appFMap;

  // initialize m_IsExpanded array
  m_IsExpanded.SetBoundaries( min(eMinIdx,0), eMaxIdx );
  m_IsExpanded.SetAll( FALSE );

  // update virtual matrix dimensions
  m_Pr.UpdateDimensions();

  // set some members:
  m_VocSizeE = eMaxIdx + 1;
  m_VocSizeF = m_Pr.MaxCol() + 1;

  // now we are done so far, however, for lexica that were compiled for dynamic EOS we have to fix the index shift
  if (blFixEos) {
    assert(VOCAB_STATIC_EOS);

    // This code has been copied from ConvertLexiconV5.cc
    m_Pr.InsertRow(1);
    m_Pr.InsertColumn(1);

    // The last row / column contain the probs for </s>
    // We have to move the last row / column to (new) position 1

    // move last row, if necessary
    assert(m_Pr.MaxRow() <= internalVocab.Last() + 1);
    if (m_Pr.MaxRow() == internalVocab.Last() + 1) {
      // copy last row to index 1
      m_Pr.GetRow(1) = m_Pr.GetRow(m_Pr.MaxRow());
      // delete last row
      m_Pr.SetBoundaries( m_Pr.MinRow(),  m_Pr.MaxRow()-1,
                          m_Pr.MinCol(),  m_Pr.MaxCol() );
    } // endif, move last row (</s>) to row 1

    // foreach row: move last column
    for (int i = m_Pr.MinRow(); i <= m_Pr.MaxRow(); ++i) {
      CSparseArray & row = m_Pr.GetRow(i);
      assert(row.MaxIndex() <= internalVocab.Last() + 1);
      if (row.MaxIndex() == internalVocab.Last() + 1) {
        // we found (old) </s>: move it to column 1
        row.SetAt(1,row.GetAt(row.MaxIndex()));
        row.DeleteLast();
      } // endif: move last column to index 1
    } // endfor: all rows
    
  } // endif: fix versionmod mismatch

  // success!
  return IOERR_OK;
} // CLexicon::Load_Dataversion5()

