// CTranslationModel

// Base class for all translation models.
// Handles the biligual lexicon T and the word pair counts TCounts.
// Pointers to the to text objects; Member variables for current
// sentence pair (m_S) and the length of the sentences (m_I, m_J).
//

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <iostream>
#include <fstream>
#include <sstream>
#include <assert.h>
#include <math.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define NO_TRACE 1
#include <Trace.hh>

#include <LexiconTrain.hh>
#include <Cognates.hh>

//static char rcsid_tm_cc[] = "$Id: ";
using namespace std;


// ------------------------------------------------------
// CLexiconForTraining
// ------------------------------------------------------

void
CLexiconForTraining::UpdateCountsForCognates()
{
  Trace( 0, "CLexiconForTraining::UpdateCountsForCognates" );
  
  if ( !m_pCognates )
    return;
 
  Trace( 0, "Cognates exist" );

  int Wf, We;
  
  // we don't consider at the moment the <UNK> which is at index 0
  // therefore start loops with 1
  for ( We = 1; We < m_pVocE->Size(); We++) {
    Trace( 0 , "We = " << We );
    if ( m_pCognates->IsCognateE( We ) ) {
      Wf = m_pCognates->CognateF( We );
      SetCount( Wf, We, 100000 );
      Trace( 0, "New Count for " <<  m_pVocF->GetWord( Wf ) 
             << " " << m_pVocE->GetWord(We ) 
             << " " << Count( Wf, We ) );
    }
  }
}




void 
CLexiconForTraining::DisplayCounts() const
{
  //m_Pr.Display();

  float Value;
  cout << "\nDisplay counts\n\n";

  int VocSizeF = m_pVocF->Size();
  int VocSizeE = m_pVocE->Size();

  const char *pWordE, *pWordF;
  cout << "T: "<< "\n";
  cout << "Size of F:" << VocSizeF << "\n";
  cout << "Size of E:" << VocSizeE << "\n";
  cout.flush();
  //  cout.precision(4);

  // count how often word pairs accur in parallel sentences
  //CalculateTCounts();
  //DisplayTCounts();

  // display only those with values 10 times above uniform distribution
  float CutValue = 0.01;
  cout << "Only Word Pairs with T(f_j|e_i) >= " << CutValue << " are shown \n";
  cout.flush();

  int MinIndexE = m_Counts.MinRow();
  int MaxIndexF = VocSizeF;
  int MaxIndexE = VocSizeE;
  
  /*
    int MaxIndexF = 10;
    if ( MaxIndexF > VocSizeF ) {
    MaxIndexF = VocSizeF;
    }

    int MaxIndexE = 10;
    if ( MaxIndexE > VocSizeE ) {
    MaxIndexE = VocSizeE;
    }
  */


  // we don't consider at the moment the <UNK> which is at index 0
  // therefore start loops with 1
  for (int i = MinIndexE; i <= MaxIndexE; i++) {
    pWordE = m_pVocE->GetWord(i); 
    //cout << pWordE << " (" << m_pTextE->GetVocCount(i) << ")\n";
    //cout << pWordE << endl;

    for (int j = 0; j <= MaxIndexF; j++) {
      Value = m_Counts.GetAt(i,j);
      if (Value >= CutValue) {
        ostringstream String;
        pWordF = m_pVocF->GetWord(j);
        //String << "    " <<  pWordF  << " (" << m_pTextF->GetVocCount(j) 
        //       << ")                                                     ";
        String << pWordE << "    " <<  pWordF
               << "                                                     ";
        //ARV changed, remove 40 in write
        cout<<String.str();
        //cout.write(String.str(), 40);
        //cout << "    " << pWordF  << " (" << m_pTextF->GetVocCount(j) << ")\t";
        cout.setf(ios::right | ios::fixed);
        cout.width(6);
        cout.precision(3);
	
        //cout <<  m_pT->GetAt(i,j) << "  (" << (int)m_TCounts->GetAt(i,j) << ")\n";
        cout <<  m_Counts.GetAt(i,j) << endl;
        cout.flush();
      }
    }
    //cout << "\n";
  }
}




void 
CLexiconForTraining::Renormalize()
{
  Trace( 1, "CLexiconForTraining::Renormalize()" );

  //Trace( 1, "before Clear" );
  // remember default value
  //float DefaultValue = m_Pr.DefaultValue();
  //m_Pr.Clear();
  //m_Pr.SetDefaultValue( DefaultValue );
  //Trace( 1, "after Clear; now Add Counts" );
  //m_Pr.Add( m_Counts );
  //Trace( 1, "after Add Counts" );
  //m_Pr.Display();
  //DefaultValue = m_Pr.DefaultValue();
  //Trace( 1, "DefaultValue = " << DefaultValue );

  float S;
  int Row;
  float DefaultValue;
  int Index;
  float Value;

  int SizeF = m_pVocF->Size();
  float SmoothValue = m_LambdaSmooth / m_pVocF->Size();

  Trace( 3, "MinRow=" << m_Pr.MinRow() << "  MaxRow=" << m_Pr.MaxRow() );
  for ( Row = m_Pr.MinRow(); Row <= m_Pr.MaxRow(); Row++ ) {
    if ( !m_IsExpanded[ Row ] ) {
      CSparseArray & Pr = m_Pr.GetRow( Row );
      CSparseArray & Counts = m_Counts.GetRow( Row );
      DefaultValue = Pr.DefaultValue();
      Pr.Clear();
      // we only take large values;
      S = Counts.Sum( SizeF );
      if ( S > 0.0 ) {
        Counts.Div( S );
      }
      for ( int i = 0; i < Counts.NumberOfItems(); i++) {
        Counts.GetItem( i, Index, Value );
        if ( Value > 1.0 / m_pVocF->Size() ) {
          Pr.SetAt( Index, Value );
        }
      }
      Trace( 3, "Number of items in Counts: " << Counts.NumberOfItems() << " "
             << "Number of items in Pr : " << Pr.NumberOfItems() ) ;
      //Pr.Add( Counts );
      //Pr.SetDefaultValue( DefaultValue );

      S = Pr.Sum( SizeF );
    
      Trace( 3, "Row = " << Row << " not expanded: " 
             << "  DefaultValue = " << Pr.DefaultValue()
             << "  Sum = " << S 
             << "  SumOfStored = " << Pr.SumOfStoredItems() );
	   
      //A.Display();

      if ( S > 0.0 ) {
        //Trace( 5, "SetDefaultValue( 0.0 )" );
        Pr.Mul( 1.0 / S );
        S = Pr.Sum( SizeF );
        Trace( 5, "  Sum = " << S );
 
        assert( m_LambdaSmooth < 1.0 );
        Pr.Mul( (1.0 - m_LambdaSmooth) );
        S = Pr.Sum( SizeF );
        Trace( 5, "  Sum = " << S );

        Pr.Add( SmoothValue );
        S = Pr.Sum( SizeF );
        Trace( 5, "  Sum = " << S );

        assert( SmoothValue > 0.0 );
        Pr.SetDefaultValue( SmoothValue );
        S = Pr.Sum( SizeF );
        Trace( 5, "  Sum = " << S ); 
      } else {
        Pr.SetDefaultValue( 1.0 / m_pVocF->Size() );
      }

      /*
        CSparseArray & B = m_Pr.GetRow( Row );
        //B.Display();
        Trace( 5, "Row = " << Row 
        << "  DefaultValue = " << B.DefaultValue()
        << "  Sum = " << B.Sum()
        << "  SumOfStored = " << B.SumOfStoredItems() );
	
        SumT = 0.0;
        for (Wf = 0; Wf < m_pVocF->Size(); Wf++) {
        Trace( 5, "m_Pr() " << m_Pr.GetAt( Row, Wf )
        << "   Row.GetAt() " << B.GetAt( Wf ) );
        SumT += m_Pr.GetAt( Row, Wf );
        }
        Trace( 1, "Sum T(f|" <<  m_pVocE->GetWord( Row ) << ") = " << SumT );
      */
    } else {
      CArray<float> & Counts = m_ExpandedCounts.GetRow( Row );
      CArray<float> & Pr = m_ExpandedPr.GetRow( Row );
      CSparseArray & A = m_Pr.GetRow( Row );

      float Sum = 0.0;
      int f;
      for ( f = Counts.MinIndex(); f <= Counts.MaxIndex(); f++ ) {
        Sum += Counts.GetAt( f );
      }
      Trace( 1, "Row = " << Row << " expanded: "
             << "  Sum = " << Sum );
      if ( Sum > 0.0 ) {
        for ( f = Counts.MinIndex(); f <= Counts.MaxIndex(); f++ ) {
          if ( Counts.GetAt( f ) > 0 ) {
            Pr.SetAt( f , (1.0 - m_LambdaSmooth) * Counts.GetAt( f ) / Sum + SmoothValue );
          } else {
            Pr.SetAt( f, SmoothValue );
          }
        }
        A.SetDefaultValue( SmoothValue );
      } else {
        Pr.SetAll( 1.0 / m_pVocF->Size() );
        A.SetDefaultValue( 1.0 / m_pVocF->Size() );
      }

    }
  }
}




