#include <config.h>

// ------------------------------------------------------------------
// CLexiconForTrain
//
// CLexiconForTrain is organized for quick access to all f for given e
// ------------------------------------------------------------------

// $Id: LexiconTrain.hh 1743 2006-06-27 02:31:01Z vogel $
//


#ifndef __LexiconForTrain__
#define __LexiconForTrain__


#include <Lexicon.hh>


class CLexiconForTraining : public CLexicon
{
public:
  inline void SetPr( CWordIndex f, CWordIndex e, double Value );  // p(f|e)

  // ----- set & get counts -----
  inline double Count( CWordIndex f, CWordIndex e ) const;

  inline void SetCount( CWordIndex f, CWordIndex e, double Value );
  inline void AddCount( CWordIndex f, CWordIndex e, double Value );
  inline void IncCount( CWordIndex f, CWordIndex e );
  inline void DecCount( CWordIndex f, CWordIndex e );
  void ClearCounts();

  void UpdateCountsForCognates();

  void Renormalize();

  void DisplayCounts() const;


private:

};




// ------------------------------------------------------------------
// Inline Functions for CLexiconForTraining
// ------------------------------------------------------------------


inline void
CLexiconForTraining::SetPr( CWordIndex f, CWordIndex e, double Value )
{
  if ( m_IsExpanded[ e ] ) {
    m_ExpandedLogPr.SetAt( e, f, Value );
  } else {
    std::cout << "SetPr - Not Expanded "  << f << " " << e << std::endl << std::flush;
    m_Pr.SetAt( e, f, Value );
  }
}



inline double
CLexiconForTraining::Count( CWordIndex f, CWordIndex e ) const
{
  Trace( 3, "CLexicon::Count  " << "f: " << f << " e: " << e );
  if ( m_IsExpanded[ e ] ) {
    return m_ExpandedCounts.GetAt( e, f );
  } else {
    return m_Counts.GetAt( e, f );
  }
}



inline void
CLexiconForTraining::SetCount( CWordIndex f, CWordIndex e, double Value )
{
  if ( m_IsExpanded[ e ] ) {
    m_ExpandedCounts.SetAt( e, f, Value );
  } else {
    std::cout << "SetCount - Not Expanded "  << f << " " << e << " " << Value << std::endl << std::flush;
    m_Counts.SetAt( e, f, Value );
    std::cout << "New = " << m_Counts.GetAt( e, f );
  }
}


inline void
CLexiconForTraining::AddCount( CWordIndex f, CWordIndex e, double Value )
{
  if ( m_IsExpanded[ e ] ) {
    double NewValue = Value + m_ExpandedCounts.GetAt( e, f );
    m_ExpandedCounts.SetAt( e, f, NewValue );
  } else {
    m_Counts.Add( e, f, Value );
  }
}


inline void
CLexiconForTraining::IncCount( CWordIndex f, CWordIndex e )
{
  if ( m_IsExpanded[ e ] ) {
    double NewValue = 1.0 + m_ExpandedCounts.GetAt( e, f );
    m_ExpandedCounts.SetAt( e, f, NewValue );
  } else {
    m_Counts.Inc( e, f );
  }  
}


inline void
CLexiconForTraining::DecCount( CWordIndex f, CWordIndex e )
{
  if ( m_IsExpanded[ e ] ) {
    double NewValue = m_ExpandedCounts.GetAt( e, f ) - 1.0;
    m_ExpandedCounts.SetAt( e, f, NewValue );
  } else {
    m_Counts.Dec( e, f );
  }
}


inline void
CLexiconForTraining::ClearCounts()
{
  m_ExpandedCounts.SetAll( 0 );
  m_Counts.Clear();
}




#endif
