#include <config.h>

// CTranslationModel

// Base class for all translation models.
// Handles the biligual lexicon T and the word pair counts TCounts.
// Pointers to the to text objects; Member variables for current
// sentence pair (m_S) and the length of the sentences (m_I, m_J).
//


#include <iostream>
#include <fstream>
#include <sstream>
#include <assert.h>
#include <math.h>
#include <time.h>
#include <stdio.h>
#include <string.h>

#include <TM.hh>
//#include "text.hh"
//#include "utils.hh"

using namespace std;


CTranslationModel::CTranslationModel()
{
  //m_MaxSLen = MaxSLen;  // set to default value 

  m_Iterations = 5;
  m_CurrentIteration = 0;
  m_pTextE = NULL;
  m_pTextF = NULL;
  //m_pT     = NULL;
  //m_pT2    = NULL;
  //m_TCounts = NULL;
  m_pLexicon = new CLexiconForTraining;

  m_VocSizeE = 0;
  m_VocSizeF = 0;
  m_TAsLog = FALSE;
  m_TInFile = NULL;
  m_TOutFile = NULL;
  m_Iterations = 10;
  m_DisplayLexicon = FALSE;
  m_StoreTWithWords = TRUE;

  m_pLexiconPerplexity = new CArray< float>;
  assert ( m_pLexiconPerplexity != NULL );
  m_pTotalPerplexity = new CArray<float>;
  assert ( m_pTotalPerplexity != NULL );

  m_NumOfFWords = 0; // total number of words in TextF (for calculation of per

  m_LambdaSmooth = 0.001;

  m_LambdaTrainLex = 0.99;     // interpolation factor for stat. lexicon
  m_LambdaManLex = 0.0;       // interpolation factor for manual. lexicon
  m_LambdaTrainLexCW = 0.99;   // ditto for content words
  m_LambdaManLexCW = 0.0;

  m_Verbosity = 0;
}


CTranslationModel::~CTranslationModel()
{
  //delete m_pT;
  //delete m_pT2;
  //delete m_TCounts;
  delete m_pLexicon;
  delete m_pLexiconPerplexity;
  delete m_pTotalPerplexity;
}


////////////////////////////////////////////////////////////////////////
// Initialization - Public Methods
////////////////////////////////////////////////////////////////////////



bool 
CTranslationModel::Init( CText *pTextF, CText *pTextE,
			 CCognates *pCognates )
{
  if (m_Verbosity) {
    cout << "Init translation model\n"; cout.flush();
  }
  
  InitSourceText(pTextF);
  InitTargetText(pTextE);
  //InitCounts ();
  //InitT();

  //m_pLexicon->Init( pTextF->Vocabulary(), pTextE->Vocabulary(),
  //	    pCognates );
  m_pLexicon->Init( pTextF->Vocabulary(), pTextE->Vocabulary() );

  return TRUE;
}

// Init is for initialization for TRAINING !!! 

bool 
CTranslationModel::Init(const CTranslationModel &TM)
{
  m_pTextE = TM.m_pTextE;
  m_pTextF = TM.m_pTextF;
  m_VocSizeE = TM.m_VocSizeE;
  m_VocSizeF = TM.m_VocSizeF;
  //m_pT = TM.m_pT;
  //m_TCounts = TM.m_TCounts;

  //InitCounts ();

  return TRUE;
}



void
CTranslationModel::SetTestSet( CText *pTestF, CText *pTestE )
{
  m_pTestF = pTestF;
  m_pTestE = pTestE;
}



bool
CTranslationModel::TestProbabilities()
{
  /*
  // test to see if probabilities add up to one
  cout << "Test if probability sum of t(f|e) over f adds up to 1\n";
  
  // here, only the translation probabilities are tested
  CWordIndex We;
  double SumT;

  cout.setf(ios::right | ios::fixed);
  //cout.width(6);
  cout.precision(3);
  for (We = 0; We < (CWordIndex)m_VocSizeE; We++) {
    //int NumOfUnstored = m_VocSizeF - m_pT->NumberOfItemsInRow(We);
    //double DefValue = m_pT->GetRow(We).GetDefaultValue();
    //SumT = m_pT->SumOfRow(We) + NumOfUnstored * DefValue;
    //SumT = m_pT->SumOfRow(We);
    cout << "Sum T(f|" <<  m_pTextE->GetWordFromVocabulary(We) 
      << ") = " << SumT << " \n";
  }
  */
  return TRUE;
}



void 
CTranslationModel::SetIterations( int Iterations )
{
  m_Iterations = Iterations;
}


bool 
CTranslationModel::Train()
{
  assert(FALSE);
  return FALSE;
}


bool 
CTranslationModel::NextIteration()
{
  // not allowed in base class
  assert(FALSE);
  return FALSE;
}



double 
CTranslationModel::Perplexity()
{
  return 0.0;
}




////////////////////////////////////////////////////////////////////////
// Methods to set and get Translation Probabilities
////////////////////////////////////////////////////////////////////////

void
CTranslationModel::SetLambdaT (double LambdaTrainLex, double LambdaManLex,
			       double LambdaTrainLexCW, double LambdaManLexCW)
{
  m_LambdaTrainLex = LambdaTrainLex;
  m_LambdaManLex = LambdaManLex;
  m_LambdaTrainLexCW = LambdaTrainLexCW;
  m_LambdaManLexCW = LambdaManLexCW;
}


//void
//CTranslationModel::SetT(CWordIndex f, CWordIndex e, double Value)
//{
  //m_pT->SetAt(e, f, Value);
//}


//double
//CTranslationModel::GetT(CWordIndex f, CWordIndex e) const
//{
//  double T;
  /*
  assert (m_pT != NULL);

  assert (! m_TAsLog);

  if ( m_pT2 != NULL ) {
    if (m_pTextE->IsContentWord(e)) {
//	assert (m_pT2);
	  T = m_LambdaTrainLexCW * m_pT->GetAt(e, f) 
	      + m_LambdaManLexCW * m_pT2->GetAt(e, f) + 
	      + (1 - m_LambdaTrainLexCW - m_LambdaManLexCW) / m_VocSizeF;
	  //cout << "GetT - ContentWort: " << f << " " << e << " " 
	  //<< T << endl << flush;
    } else {
	  T = m_LambdaTrainLex * m_pT->GetAt(e, f) 
	      + m_LambdaManLex * m_pT2->GetAt(e, f) + 
	      + (1 - m_LambdaTrainLex - m_LambdaManLex) / m_VocSizeF;
	  //cout << "GetT - no ContentWort: " << f << " " << e << " " 
	  //<< T << endl << flush;
    }
  }
  else {
      T = m_LambdaTrainLex * m_pT->GetAt(e, f) 
	  + ( 1 - m_LambdaTrainLex ) / m_VocSizeF;
  }
  EXIT_IF (T == 0.0,
	   "t(" << f << "| " << e << ") = 0 ",
	   -1);
  EXIT_IF (T > 1.0,
	   "t(" << f << "| " << e << ") > 1 ",
	   -2);
	   */
//  return T;
//}


//double
//CTranslationModel::GetLogT (CWordIndex f, CWordIndex e) const
//{
//  double t;
  /*

  if (m_TAsLog)
    return m_pT->GetAt(e, f);
  else
    {
      t = GetT(f, e);

      assert (t > 0.0);
      return ( log(t) );
    }
    */
//  return 1.0;
//}



////////////////////////////////////////////////////////////////////////
// Methods to compute the log p(f|e) from the p(f|e) and store in m_pT[Inv]
////////////////////////////////////////////////////////////////////////
//bool 
//CTranslationModel::ConvertTtoLog ()
//{
  /*
  double t;

  if (!m_TAsLog) {
    if (m_pT2 == NULL) {
      m_pT->Log();
    } else {
      // Achtung: so, wie das jetzt implementiert ist,
      // ist m_pT hinterher KEINE duennbesetzte Matrix mehr!! 
      for (int e = 0; e < m_pTextE->GetVocabularySize(); e++) {
	for (int f = 0; f < m_pTextF->GetVocabularySize(); f++) {
	  t = GetT(f, e);
	  assert (t > 0.0);
	  m_pT->SetAt(e, f, log(t));
	}
      }
      delete m_pT2;
      m_pT2 = NULL;
    }
  }
  
  m_TAsLog = TRUE;

  */
//  return TRUE;
//}




////////////////////////////////////////////////////////////////////////
// Init - Protected Methods
////////////////////////////////////////////////////////////////////////

//void 
//CTranslationModel::InitT()
//{
  /*
  if (m_pT == NULL) {
    m_pT = new CDblMatrix(m_VocSizeE, m_VocSizeF);
    assert(m_pT);
  }

  m_pT->SetDefaultValue(1.0 / m_VocSizeF);
  */
//}



//void 
//CTranslationModel::InitCounts()
//{
  /*
  if (m_TCounts == NULL){
    m_TCounts = new CDblMatrix(m_VocSizeE, m_VocSizeF);
    assert(m_TCounts);
  }
  */

//}

void
CTranslationModel::InitSourceText (CText *pTextF)
{
  assert(pTextF);

  m_pTextF = pTextF;
  m_VocSizeF = m_pTextF->Vocabulary()->Size();

  if (m_Verbosity) {
    cout << "size vocabular F " << m_VocSizeF << "\n"; cout.flush();
  }
}

void
CTranslationModel::InitTargetText (CText *pTextE)
{
  assert(pTextE);

  m_pTextE = pTextE;
  m_VocSizeE = m_pTextE->Vocabulary()->Size();

  if (m_Verbosity) {
    cout << "size vocabular E " << m_VocSizeE << "\n"; cout.flush();
  }
}

//void
//CTranslationModel::InitTargetVoc (const CText *pTextE)
//{
  // this will really initialize just a vocabulary in real life (probably)
//  InitTargetText (pTextE);
//}


//void 
//CTranslationModel::ClearTCounts()
//{
  //m_TCounts->Clear();
//}


////////////////////////////////////////////////////////////////////////
// Methods to set and get the maximal sentence length supported
// Cave: The setting has no effect at the moment...
////////////////////////////////////////////////////////////////////////

int
CTranslationModel::GetMaxSLen () const
{
  return m_MaxSLen;
}


void
CTranslationModel::SetMaxSLen (int /*l*/)
{
  //  if (l > MaxSLen)
  //   {
  //     cerr << "\nattempt to set max. sentence length to " << l << endl;
  //     cerr << "max. sentence length allowed: " << MaxSLen << endl;
  //     cerr.flush();
  //   }
  //  m_MaxSLen = l;
}


////////////////////////////////////////////////////////////////////////
// Method to set the verbosity
////////////////////////////////////////////////////////////////////////
void
CTranslationModel::SetVerbosity (int Verbosity)
{
  m_Verbosity = Verbosity;
}


////////////////////////////////////////////////////////////////////////
// Calculation - Methods
////////////////////////////////////////////////////////////////////////

// void 
// CTranslationModel::RenormalizeT()
// {
  //m_pT->Renormalize();
// }



// void 
// CTranslationModel::CalculateTCounts()
// {
  /*
  int Value;

  assert(m_pTextE != NULL);
  assert(m_pTextF != NULL);

  CWordIndex Se, Sf, We, Wf;
  TPosition PosSe, PosSf, PosWe, PosWf;

  ClearTCounts();

  // fetch first sentence pair
  Se = m_pTextE->FirstSentence(PosSe);
  Sf = m_pTextF->FirstSentence(PosSf);
  
  //cout << Se << " " << Sf << endl; cout.flush();

  while (Se != m_pTextE->EOT && Sf != m_pTextF->EOT) {
    
    // loop over words in sentence of source (first, F) text
    for (Wf = m_pTextF->FirstWord(PosSf, PosWf);
	 Wf != m_pTextF->EOS;
	 Wf = m_pTextF->NextWord(PosSf, PosWf) ) {
      
      // loop over words in e sentence
      for (We = m_pTextE->FirstWord(PosSe, PosWe);
	   We != m_pTextE->EOS;
	   We = m_pTextE->NextWord(PosSe, PosWe) ) {
	
	//cout << Wf << " " << We << endl; cout.flush();

	Value = (int)m_TCounts->GetAt(We,Wf);
	//cout << "m_TCounts->GetAt(We,Wf)" << m_TCounts->GetAt(We,Wf) << endl;
	//cout.flush();
	m_TCounts->SetAt(We,Wf,Value+1);

	//	m_TCounts[Wf][We]++;
	//scheint nicht mehr benutzt zu werden	m_TCountsSum[We]++;
      }
    }
    
    // fetch next sentence pair
    Se = m_pTextE->NextSentence(PosSe);
    Sf = m_pTextF->NextSentence(PosSf);
  }				// end loop sentences

  // how many word pairs are seen
  int i, j;
  int WordPairsSeen = 0;
  double DefaultValue = m_TCounts->GetDefaultValue();

  for (j = 0; j < m_VocSizeF; j++) {
    for (i = 0; i < m_VocSizeE; i++) {
      if (m_TCounts->GetAt(i,j) > DefaultValue)
	WordPairsSeen++;
    }
  }

  float Percentage = (float)(100 * WordPairsSeen) / (float)(m_VocSizeF * m_VocSizeE);

  cout.setf(ios::right | ios::fixed);
  cout.precision(2);

  cout << WordPairsSeen << " word pairs seen out of " 
    << m_VocSizeF * m_VocSizeE << ", i.e. " << Percentage << "% \n\n";

  cout.flush();
  */
// }



////////////////////////////////////////////////////////////////////////
// Display - Methods
////////////////////////////////////////////////////////////////////////

void 
CTranslationModel::DisplayPerplexities(int Iteration)
{
  if (Iteration < 0) {
    // display all perplexities
    cout.setf(ios::right | ios::fixed);
    //cout.width(6);
    cout.precision(2);
    cout << "Perplexity as Function of Iteration \n\n";
    cout << "Iteration   &  Lexicon  &   Jump   &   Total \\\\ \\hline \n";
    for (int Iteration = 0; Iteration < m_Iterations; Iteration++) {
      cout << Iteration << " & " 
	   << m_LexiconPerplexity[Iteration] << " & "  
	   << m_TotalPerplexity[Iteration] << "\n";
    }
    return;
  }

  cout <<"\n\n";
  cout.flush();
}


// void 
// CTranslationModel::DisplayT()
// {
  /*
  double Value;
  cout << "\nDisplay lexicon\n\n";

  const char *pWordE, *pWordF;
  cout << "T: "<< "\n";
  cout << "Size of F:" << m_VocSizeF << "\n";
  cout << "Size of E:" << m_VocSizeE << "\n";
  cout.flush();
  //  cout.precision(4);

  // count how often word pairs accur in parallel sentences
  CalculateTCounts();
  //DisplayTCounts();

  // display only those with values 10 times above uniform distribution
  double CutValue = 10.0 / (float)m_VocSizeF;
  cout << "Only Word Pairs with T(f_j|e_i) >= " << CutValue << " are shown \n";
  cout.flush();

  // we don't consider at the moment the <UNK> which is at index 0
  // therefore start loops with 1
  for (int i = 1; i < m_VocSizeE; i++) {
    pWordE = m_pTextE->GetWordFromVocabulary(i); 
    cout << pWordE << " (" << m_pTextE->GetVocCount(i) << ")\n";

    for (int j = 1; j < m_VocSizeF; j++) {
      Value = m_pT->GetAt(i,j);
      if (Value >= CutValue) {
	strstream String;
	pWordF = m_pTextF->GetWordFromVocabulary(j);
	String << "    " <<  pWordF  << " (" << m_pTextF->GetVocCount(j) 
	       << ")                                                     ";
	cout.write(String.str(), 40);
	//cout << "    " << pWordF  << " (" << m_pTextF->GetVocCount(j) << ")\t";
	cout.setf(ios::right | ios::fixed);
	cout.width(6);
	cout.precision(3);
	
	cout <<  m_pT->GetAt(i,j) << "  (" << (int)m_TCounts->GetAt(i,j) << ")\n";
	cout.flush();
      }
    }
    cout << "\n";
  }
  */
// }



// void 
// CTranslationModel::DisplayTCounts()
// {
  /*
  const char *pWordE, *pWordF;
  int i, j;


  cout << "TCounts: "<< "\n";
  //  cout.precision(4);
  for (i = 1; i < m_VocSizeE; i++) {
    for (j = 1; j < m_VocSizeF; j++) {
      if (m_TCounts->GetAt(i,j) > 1) {
	pWordF = m_pTextF->GetWordFromVocabulary(j); 
	pWordE = m_pTextE->GetWordFromVocabulary(i); 

	strstream String;
	String << pWordE <<  "  " << pWordF  << "                         ";
	cout.write(String.str(), 25);
	cout << m_TCounts->GetAt(i,j) << "\n";
      }
    }
    // cout << "\n";
  }
  */
// }


/*
void 
CTranslationModel::DisplaySentenceLength()
{
  // display sentence lenght frequencies
  cout << "sentence length frequencies\n";
  
  cout << "J/I ";
  for (int I = 1; I < 20; I++) {
    cout << I << " ";
  }
  cout << "\n";
  
  for (int J = 1; J < 25; J++) {
    cout << J << "   ";
    for (int I = 1; I < 25; I++) {
      cout << m_SentenceLength[J][I] << " ";
    }
    cout << "\n";
  }
}
*/



////////////////////////////////////////////////////////////////////////
// Store and Load - Methods
////////////////////////////////////////////////////////////////////////


//bool 
//CTranslationModel::StoreT(const char *pFileName)
//{
  /*
  assert (m_TAsLog == FALSE); // files should contain probabilities as such

  //if (m_StoreTWithWords)
  //  return StoreTWithWords(pFileName);

  // write m_pT to a file
  // store only pair for which probability is > DefaultValue
  // double DefaultValue = m_pT->GetDefaultValue();

  ofstream TFile (pFileName);

  if (!TFile) {
    cout << "File " << pFileName << " could not be opened\n";
    return FALSE;
  }

  //  ofstream TFile;
  //TFile.open(pFileName);
  //assert(TFile);

  TFile.setf(ios::scientific);
  TFile.precision(6);

  TFile << "# Lexicon-Probabilities\n";
  TFile << "# for each e: index of e\n";
  TFile << "# for each fwith p(f|e) >= DefaultValue(e): index of f, p(f|e)\n";

  int TVersion = 2;
  TFile << "Dataversion: " << TVersion << endl;

  time_t Time = time(NULL);
  char *String;
  String = asctime(localtime(&Time));
  TFile << "Stored on: " << String;

  TFile << m_VocSizeE << "\n";
  TFile << m_VocSizeF << "\n";

  m_pT->Store(TFile);

  TFile.close();
*/
//  return TRUE;
//}



// bool
// CTranslationModel::StoreTWithWords(const char *pFileName)
// {
  /*
  ofstream TFile (pFileName);

  if (!TFile) {
    cout << "File " << pFileName << " could not be opened\n";
    return FALSE;
  }

  //  ofstream TFile;
  //TFile.open(pFileName);
  //assert(TFile);

  TFile.setf(ios::scientific);
  TFile.precision(6);

  TFile << "# Lexicon-Probabilities\n";
  TFile << "# for each e: index of e\n";
  TFile << "# for each fwith p(f|e) >= DefaultValue(e): index of f, p(f|e)\n";

  int TVersion = 2;
  TFile << "Dataversion: " << TVersion << endl;

  time_t Time = time(NULL);
  char *String;
  String = asctime(localtime(&Time));
  TFile << "Stored on: " << String;

  TFile << m_VocSizeE << "\n";
  TFile << m_VocSizeF << "\n";
  TFile << endl;

  double Value;// DefaultValue;
  const char *pWordE, *pWordF;

  for (int i = 0; i < m_VocSizeE; i++) {
    pWordE = m_pTextE->GetWordFromVocabulary(i); 
    TFile << i << "\t# " << pWordE << endl;

    for (int j = 0; j < m_VocSizeF; j++) {
      Value = m_pT->GetAt(i,j);
      if (Value > 0.001) {
	pWordF = m_pTextF->GetWordFromVocabulary(j);
	TFile <<  j << " " << m_pT->GetAt(i,j) 
	      << "   \t# " << pWordF << endl;
      }
    }
    TFile << endl;
  }
  TFile.close();
*/
//   return TRUE;
// }



// bool 
// CTranslationModel::LoadT(const char *pFileName)
// {
  /*
  if (m_Verbosity)
    {
      cout << "Read lexicon\n"; cout.flush();
    }

  if (!m_pT) InitT ();
  m_pT->Clear();

  // read m_pT from file
  //  ifstream TFile;
  char String[100];
  char WordI[100];
  char WordJ[100];
  int i, j;
  double T;

  ifstream TFile (pFileName);

  if (!TFile) {
    cout << "File " << pFileName << " could not be opened for read\n";
    return FALSE;
  }

  //  TFile.open(pFileName);
  //  assert(TFile);

  // read comment lines
  TFile.getline(String, 100);
  TFile.getline(String, 100);
  TFile.getline(String, 100);

  int TVersion;
  TFile >> String;
  TFile >> TVersion;
  //cout << String << TVersion << endl;
  TFile.getline(String, 100);

  if (TVersion == 0) {
    // read size of the lexicon
    TFile >> m_VocSizeE;
    TFile >> m_VocSizeF;
    //cout << m_VocSizeE << " " << m_VocSizeF << endl;
    
    m_pT->SetDefaultValue(1.0e-10);
    
    // read the lexicon entries
    while(!TFile.eof()) {
      TFile >> i;
      TFile >> j;
      TFile >> T;
      // cout << i << " " << j << " " << T << endl; cout.flush();
      m_pT->SetAt(i, j, T);
    }
  }
  else if (TVersion == 3) {
    int i=0, j, i1, j1;
    double d;

    TFile.getline(String,100);
    TFile >> m_VocSizeE;
    TFile >> m_VocSizeF;
    TFile.getline(String,100);

    cout << m_VocSizeE << " " << m_VocSizeF << endl;

    while(!TFile.eof()) {
      TFile.getline(String,100);
      //cout << String << endl;
      
      if (String[0] == '\0'){
	TFile.getline (String,100);
	//cout << String << endl;
	(void) sscanf( String, "%d   \t# %s", &i1, WordI);
	i = m_pTextE->GetWordIndex(WordI);
	
	//cout << "*** " << i << " " << WordI << endl;
      } else {
	
	(void) sscanf( String, "%d %lf\t# %s", &j1, &d, WordJ);
	j = m_pTextF->GetWordIndex(WordJ);
	//cout << "*** " << j << " " << d << " " << WordJ << endl;
	//IndexJ = Text.GetWordIndex(WordJ);
	//cout << IndexJ << endl;
	 m_pT->SetAt(i, j, d);
	 //cout << i << " " << j << " " << d << endl;

      }
    }    
  } 
  else {
    // time when stored
    TFile.getline(String, 100);
    //cout << String << endl; cout.flush();
    
    TFile >> m_VocSizeE;
    TFile >> m_VocSizeF;
    m_pT->Load(TFile);
    cout << m_VocSizeE << " " << m_VocSizeF << endl;
  }

  TFile.close();
  m_pT->Renormalize();
  if (m_Verbosity)
    {
      cout << "Read lexicon done\n";
      cout.flush();
    }
  //m_pT->Print();
  */
//   return TRUE;
//   
// }



// bool 
// CTranslationModel::LoadT( const char *pFileName1, const char *pFileName2 )
// {
  /*
  CDblMatrix *pHelp;

  pHelp = m_pT;
  m_pT = m_pT2;
  LoadT( pFileName2 );
  DisplayT();
  m_pT2 = m_pT;
  m_pT = pHelp;
  LoadT( pFileName1 );
  DisplayT();       

  assert (m_pT);
  assert (m_pT2);
  */
//   return TRUE;
// }




////////////////////////////////////////////////////////////////////////
// Method to store perplexity for each sentence pair
////////////////////////////////////////////////////////////////////////

bool
CTranslationModel::StorePerplexity ( const char *pFileName ) const
{
  ofstream File (pFileName);

  if (!File) {
    cout << "File " << pFileName << " could not be opened\n";
    return FALSE;
  }

  File.setf ( ios::right | ios::fixed );
  File.precision(2);

  for ( TPosition SPos = 0; SPos < m_pLexiconPerplexity->Size (); SPos++ ) {
    File << SPos
	 << (*m_pLexiconPerplexity)[SPos] 
	 << (*m_pTotalPerplexity)[SPos];
  }

  File.close ();
  return TRUE;
}

