#ifndef TM_hh
#define TM_hh

// CTranslationModel

// Base class for all translation models.
// Handles the biligual lexicon T and the word pair counts TCounts.
// Pointers to the two text objects; Member variables for current
// sentence pair (m_S) and the length of the sentences (m_I, m_J).
//

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#define NO_TRACE 1
#include <Trace.hh>

#include <Array.hh>
#include <Types.hh>
#include <Text.hh>
#include <LexiconTrain.hh>
//#include "SparseArray.hh"
#include <param.hh>
#include <Cognates.hh>


//#define MaxSLen        150    // max length of sentences
//#define MaxVocSizeE  50000    // max size of "english" vocabulary
//#define MaxVocSizeF  50000    // max size of "french" vocabulary

//#define MaxIteration    20    // max number of iterations for which
                              // perplexity values will be stored


class CTranslationModel
{
public:
  // ----- Konstruktor & Destruktor -----
  CTranslationModel();
  virtual ~CTranslationModel();

  // ----- Initialisation -----
  // the first two are for initialisation for TRAINING
  bool Init ( CText * pSourceText, CText * pTargetText,
	      CCognates * pCognates = NULL );
  bool Init ( const CTranslationModel &TM ); // init with other translation model

  void SetTestSet( CText *pTestF, CText *pTestE );

  CLexiconForTraining * Lexicon();

  void SetIterations( int Iterations );

  // ----------
  bool Train();
  double Perplexity();
  //void RenormalizeT();

  virtual bool TestProbabilities();
  
  // ----- Set factor for interpolation of two lexica -----
  // ----- different interpolation for content words  -----
  void SetLambdaT (double LambdaTrainLex,
				      double LambdaManLex,
				      double LambdaTrainLexCW,
				      double LambdaManLexCW);


  // ----- Display Methods -----
  void DisplayPerplexities(int Iteration = -1);
  //void DisplayT();
  //void DisplayTCounts();
  //  void DisplaySentenceLength();

  // ----- Store & Load lexicon probabilities -----
  //bool StoreT (const char *pFileName);
  //bool LoadT( const char *pFileName ); 
  //bool LoadT( const char *pFileName1, const char *pFileName2 ); 

  // ----- count how often word pairs accur in parallel sentences -----
  //void CalculateTCounts();    

  // ----- Maximal sentence length treatable by this instance -----
  int GetMaxSLen() const;
  void SetMaxSLen( int MaxSentLen );
  void SetVerbosity( int Verbosity );

  // ----- Store perplexity for each sentence pair
  virtual bool StorePerplexity ( const char *pFileName ) const;


protected:
  bool NextIteration();  // perform on interation over all sentence pairs

  //void InitT();
  //void InitCounts();
  void InitSourceText( CText *pTextF );
  void InitTargetText( CText *pTextE );
  //void InitTargetVoc( const CText *pTextE );// will have parameter: CVocabulary

  //void ClearTCounts();

  //bool StoreTWithWords( const char *pFileName );


protected: // ----- VARIABLES -----
  int m_Iterations;       // number of iterations for the experiment
  int m_CurrentIteration; // number of current iterations

  CLexiconForTraining *m_pLexicon;

  // Flag indicating how contents of m_pT must be interpreted
  // True : interpret as logarithm: log p(f|e) 
  bool m_TAsLog;

  CText *m_pTextE;    // pointer to training corpus
  CText *m_pTextF;

  CText *m_pTestE;    // pointer to test corpus
  CText *m_pTestF;

  int m_VocSizeE;    // actual size of vocabulary (taken from text)
  int m_VocSizeF;

  int m_I;           // length of current target sentence
  int m_J;           // length of source target sentence
  int m_S;           // number of current sentence pair

  long m_NumOfFWords; // total number of words in TextF (for calculation of perplexity)

  TPosition m_CurrentPosSe;    // position of current sentence in text
  TPosition m_CurrentPosSf;

  // perplexity results from different interations
  CArray<double> m_LexiconPerplexity;
  CArray<double> m_TotalPerplexity;

  // perplexity of sentence pairs
  CArray<float> *m_pLexiconPerplexity;
  CArray<float> *m_pTotalPerplexity;

  char *m_TInFile;
  char *m_TOutFile;

  bool m_DisplayLexicon;
  bool m_StoreTWithWords;
  int m_MaxSLen;  // actual maximal sentence length for this model

  int m_Verbosity;     // Verbosity  -  print detailed information

  double m_LambdaSmooth;

  double m_LambdaTrainLex;     // interpolation factor for stat. lexicon
  double m_LambdaManLex;       // interpolation factor for manual. lexicon
  double m_LambdaTrainLexCW;   // ditto for content words
  double m_LambdaManLexCW;
};


/*
inline double
CTranslationModel::GetDefaultT(CWordIndex e) const
{
  //return m_pT->GetDefaultValue(e);
  return 0.0;
}


inline void
CTranslationModel::SetSumProbForUnseen( double SumProb )
{
  m_pT->SetSumProbForUnseen( SumProb );
  if (m_pT2)
    m_pT2->SetSumProbForUnseen( SumProb );
}
*/

inline CLexiconForTraining *
CTranslationModel::Lexicon()
{
  return m_pLexicon;
}

#endif


