#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <iostream>
#include <sstream>
#include <stdio.h>
#include <assert.h>
#include <math.h>

#define NO_TRACE 1
#include <Trace.hh>

#include <TM_IBM1.hh>

using namespace std;

//#include "Text.hh"


CTM_IBM1::CTM_IBM1()
{
  //cout << "CTM_IBM1::CTM_IBM1() " << endl << flush;

  // pruning of sentences
  m_SentencePrune = 0;
  m_FirstPruningIteration = 1000;
  m_BadSentenceValues.SetMaxSize( 0 );
  m_BadSentenceValues.SortReverse();

  // segmentation of sentences
  m_FirstSegmentingIteration = 1000;
  m_NumberOfSplits = 0;
  m_SplitThreshold = 0.0;
  m_RequiredSplitPoints = 2;

  m_pBoundaryWords = new CArray< CString >;
  m_pIsBoundaryWord = new CNumericalArray< int >;

  m_pSegmenter = NULL;

  //cout << "CTM_IBM1::CTM_IBM1() done " << endl << flush;
}


CTM_IBM1::~CTM_IBM1()
{
  //clog << "Destructor CTM_IBM1\n";
  //delete m_pT1;
}


void 
CTM_IBM1::SetSentencePrune( int SentencePrune )
{
  m_SentencePrune = SentencePrune;
  m_BadSentenceValues.SetMaxSize( SentencePrune );
}


void 
CTM_IBM1::SetFirstPruningIteration( int FirstPruningIteration )
{
  m_FirstPruningIteration = FirstPruningIteration;
}


void 
CTM_IBM1::SetFirstSegmentingIteration( int FirstSegmentingIteration )
{
  m_FirstSegmentingIteration = FirstSegmentingIteration;
}

void 
CTM_IBM1::SetNumberOfSplits( int NumberOfSplits )
{
  m_NumberOfSplits = NumberOfSplits;
}


void 
CTM_IBM1::SetSplitWords( const CString &SplitWords )
{
  char token[4096];
  int len = 0;
  char *delimiters(" ");
  const char *TokStr = SplitWords;
  const char *ptr = TokStr;

  Trace( 2, "SplitUpToTokens:" << TokStr );

  while ( *(ptr += strspn( ptr, delimiters )) )  { // Find first word after delimiters
    len = strcspn( ptr, delimiters );                // Find first delimiter after word
    Trace( 3, "Read from: " << ptr );
    strncpy( token, ptr, len );
    token[ len ] = '\0';
    m_pBoundaryWords->Append( CString( token ) );
    Trace( 3, token );
    ptr += len;
  }
  Trace( 3, "" );

  Trace( 3, "All split words:" );
  for ( int i = m_pBoundaryWords->MinIndex();  i <= m_pBoundaryWords->MaxIndex(); i++ ) {
    Trace( 3, i << " " << m_pBoundaryWords->GetAt( i ) );
  }
}

void 
CTM_IBM1::SetSplitThreshold( double SplitThreshold )
{
  m_SplitThreshold = SplitThreshold;
}


void 
CTM_IBM1::SetRequiredSplitPoints( int RequiredSplitPoints )
{
  m_RequiredSplitPoints = RequiredSplitPoints;
}



bool
CTM_IBM1::Train()
{
 // return TRUE;

  //m_pT->SetDefaultValue(1.0/m_VocSizeF);

  Trace( 1, "------------------------------" );
  Trace( 1,  " Train IBM1" );

  assert( m_pTextF );
  assert( m_pTextE );


  // create and init lexicon
  if ( m_pLexicon == NULL ) {
    Trace( 0, "Create new Lexicon" );
    m_pLexicon = new CLexiconForTraining;
    assert( m_pLexicon );

	// init lexicon with vocabulary
	m_pLexicon->Init( m_pTextF->Vocabulary(), m_pTextE->Vocabulary() );
  }


  // set smoothing parameter
  m_pLexicon->SetLambdaSmooth( m_LambdaSmooth );
  Trace( 1, "Test Prob: " << m_pLexicon->Pr( 3, 3 ) );

  // set array for sentence perplexities to correct size
  m_pTotalPerplexity->AssureBoundaries( 0, m_pTextF->NumberOfSentences() );


  Perplexity();
  //return TRUE;

  // iterate
  for ( m_CurrentIteration = 0; 
	m_CurrentIteration < m_Iterations; 
	m_CurrentIteration++ ) {
    Trace( 1, "\n\n== Iteration " << m_CurrentIteration << " ==\n" );

    NextIteration();
    // NextIteration calculates perplexity of the model
    // as it was before the current iteration
    //Perplexity();

    DisplayBadSentences();
    if ( m_CurrentIteration >= m_FirstPruningIteration )
      RemoveBadSentences();
  }

  // calculate perplexity after final iteration
  //double PP;
  Perplexity();
  
  //cout << "Initial perplexity measure: " << PP << "\n";

  //PerplexityForMaximum();

  // set probability for cognates
  //cout << "With Cognates" << endl;
  //InitLexicon();
  //PP = Perplexity();

  //PerplexityForMaximum();

  StoreBiText( "NewCorpus" );

  // display all perplexities
  cout.setf(ios::right | ios::fixed);
  //cout.width(6);
  cout.precision(2);
  cout << "Perplexity as Function of Iteration \n\n";
  cout << "Iteration   &  Lexicon  &   Jump   &   Total \\\\ \\hline \n";
  for ( int Iteration = 0; Iteration <= m_Iterations; Iteration++ ) {
    cout << Iteration << " & " 
	 << m_LexiconPerplexity[Iteration] << " & "  
	 << m_AlignPerplexity[Iteration] << " & "  
	 << m_TotalPerplexity[Iteration] << "\n";
  }
  

  return TRUE;
}



bool
CTM_IBM1::NextIteration()
{
  Trace(1, "CTM_IBM1::NextIteration ");
  Trace( 1, "Test Prob: " << m_pLexicon->Pr( 3, 3 ) );

  // for reestimation
  CWordIndex Se, Sf, We, Wf;
  TPosition PosSe, PosSf, PosWe, PosWf;
  double Sum, Pr;
  // for perplexity
  double PAlignSum = 0.0, PTransSum = 0.0, PTrans = 0.0, PTotal;
  int SentenceLengthE = 0, SentenceLengthF = 0;
  int TotalLengthF = 0;

  int Position;
  bool IsNew;

  m_pLexicon->ClearCounts();
  m_BadSentenceValues.RemoveAll();
  m_BadSentenceNumbers.RemoveAll();

  const CSentence *pSentenceE, *pSentenceF;

  // fetch first sentence pair
  Se = m_pTextE->FirstSentence( PosSe );
  Sf = m_pTextF->FirstSentence( PosSf );

  Trace( 1, "Sf = " << Sf << " Se " << Se );

  pSentenceE = m_pTextE->GetSentence( PosSe );
  pSentenceF = m_pTextF->GetSentence( PosSf );

  pSentenceF->Display( m_pTextF->Vocabulary() ); cout << endl;
  pSentenceE->Display( m_pTextE->Vocabulary() ); cout << endl;

  // fetch first sentence pair
  Se = m_pTextE->FirstSentence( PosSe );
  Sf = m_pTextF->FirstSentence( PosSf );
  
  uint s = 0;

  while (Se != m_pTextE->Vocabulary()->EOT && 
	 Sf != m_pTextF->Vocabulary()->EOT)     {

    if (s % 1000 == 0) {
      //Trace( 1, "Sentence pair " << s );
      cout << "Sentence pair " << s << endl << flush;
      //cout.flush();
    }
    
    // get length of sentences; if one is zero, skip
    if ( (m_pTextF->SentenceLength( PosSf ) > 0) &&
	 (m_pTextE->SentenceLength( PosSe ) > 0) ) { 

      (*m_pTotalPerplexity)[ s ] = 0.0;

      SentenceLengthF = 0;

      // display sentence of source (first, F) text
      for ( Wf = m_pTextF->FirstWord( PosSf, PosWf );
	    Wf != m_pTextF->Vocabulary()->EOS;
	    Wf = m_pTextF->NextWord( PosSf, PosWf ) ) {
	Trace0( 3, m_pTextF->Vocabulary()->GetWord( Wf ) << " " );
      }
      Trace( 3, "" );

      // display sentence of source (first, F) text
      for ( We = m_pTextE->FirstWord( PosSe, PosWe );
	    We != m_pTextE->Vocabulary()->EOS;
	    We = m_pTextE->NextWord( PosSe, PosWe ) ) {
	Trace0( 3, m_pTextE->Vocabulary()->GetWord( We ) << " " );
      }
      Trace( 3, "" );

      // loop over words in sentence of source (first, F) text
      for ( Wf = m_pTextF->FirstWord( PosSf, PosWf );
	    Wf != m_pTextF->Vocabulary()->EOS;
	    Wf = m_pTextF->NextWord( PosSf, PosWf ) ) {
      
	Sum = 0.0;
	PTrans = 0.0;
	SentenceLengthF++;
	TotalLengthF++;
	SentenceLengthE = 0;
      
	// loop over words in sentence of destination (second, e) text
	for ( We = m_pTextE->FirstWord( PosSe, PosWe );
	      We != m_pTextE->Vocabulary()->EOS;
	      We = m_pTextE->NextWord( PosSe, PosWe ) ) {	
	  Sum += m_pLexicon->Pr( Wf, We );
	  //Trace( 1, Wf << " " << We << " p = " << m_pLexicon->Pr( Wf, We ) );
	}

	//Trace( 1, "Sum = " << Sum );

	if ( Sum <= 0.0 ) {
	  Trace0( 0, m_pTextF->Vocabulary()->GetWord( Wf )<< " : " );
	  for ( We = m_pTextE->FirstWord( PosSe, PosWe );
		We != m_pTextE->Vocabulary()->EOS;
		We = m_pTextE->NextWord( PosSe, PosWe ) ) {
	    Trace0( 0, m_pTextE->Vocabulary()->GetWord( We ) << " " );
	    Trace0( 0, m_pLexicon->Pr( Wf, We ) << " " );
	    Sum += m_pLexicon->Pr( Wf, We );
	  }
	  Trace( 0, "SumProb = " << Sum );
	}

	assert( Sum > 0.0 );

	// loop over words in e sentence
	for ( We = m_pTextE->FirstWord( PosSe, PosWe );
	      We != m_pTextE->Vocabulary()->EOS;
	      We = m_pTextE->NextWord( PosSe, PosWe ) ) {
	  Trace( 3, Wf << " " << We );
	  Pr = m_pLexicon->Pr( Wf, We );
	  Trace( 3, "Wf=" << Wf << " We=" << We << " Pr=" << Pr );
	  m_pLexicon->AddCount( Wf, We, Pr / Sum );

	  //Result = m_pT1->GetAt(We,Wf) + m_pT->GetAt(We,Wf)/l;
	  //m_pT1->SetAt(We,Wf,Result);
	  //cout << "T1(" << We << "," << Wf << ") = " << m_pT1->GetAt(We,Wf) << ";  ";
	  //m_pLambda[We] += m_pT->GetAt(We,Wf) / l;
	  //m_pLambda[We] += m_pLexicon->Pr( Wf, We ) / l;
	  
	  // perplexity
	  PTrans += Pr;
	  SentenceLengthE++;
	}
	
	PTransSum += log(PTrans);
	PAlignSum += log(1.0 / (float) SentenceLengthE);
	
	(*m_pTotalPerplexity)[ s ] += log(PTrans);
      }

      
      (*m_pTotalPerplexity)[ s ] = 
	exp( -(*m_pTotalPerplexity)[ s ] / (double) SentenceLengthF );

      // insert value of sentence into list of bad sentences
      Position = m_BadSentenceValues.Insert( (*m_pTotalPerplexity)[ s ], IsNew );
      if ( IsNew ) {
	m_BadSentenceNumbers.Insert( Position, s );
      } else {
	m_BadSentenceNumbers.SetAt( Position, s );
      }
      
      if ( m_BadSentenceNumbers.MaxIndex() > m_BadSentenceValues.MaxIndex() ) {
	m_BadSentenceNumbers.SetMaxIndex( m_BadSentenceValues.MaxIndex() );
      }
      
    } else {
      //cout << "Sentence Pair " << s << " skipped" << endl << flush;
    }

    // fetch next sentence pair
    Se = m_pTextE->NextSentence( PosSe );
    Sf = m_pTextF->NextSentence( PosSf );

    s++;
  } // end loop over sentence pairs
  
  //m_pLexicon->Display();

  PTransSum = -PTransSum / TotalLengthF;
  PAlignSum = -PAlignSum / TotalLengthF;
  
  PTotal = PAlignSum + PTransSum;
  cout << m_CurrentIteration << " & " << exp(PTransSum) 
       << " & " << exp(PAlignSum) << " & " << exp(PTotal) << endl;

  m_LexiconPerplexity[m_CurrentIteration] = exp(PTransSum);
  m_AlignPerplexity[m_CurrentIteration] = exp(PAlignSum);
  m_TotalPerplexity[m_CurrentIteration] = exp(PTotal);

  // normalize the translation probabilities
  Trace( 1, "Renormalize" );
  m_pLexicon->Renormalize();
  //m_pLexicon->TestProbabilities();

  //m_pLexicon->DisplayCounts();

  return TRUE;
}



double
CTM_IBM1::Perplexity()
{
  Trace(1, "CTM_IBM1::Perplexity()");

  double Perplexity;
  
  CWordIndex Se, Sf, We, Wf;
  TPosition PosSe, PosSf, PosWe, PosWf;
  double PEta = .0, PAlign = .0, PTrans = .0, PTrans1 = .0;
  int SentenceLengthE = 0, SentenceLengthF = 0;
  int TotalLengthF = 0;

  const CSentence *pSentenceE, *pSentenceF;

  // clear the sentence length table
  //for (int i = 0; i < MaxSLen; i++)
  //  for (int j = 0; j < MaxSLen; j++)
  //    ;//      m_SentenceLength[i][j] = 0;
  
  
  // loop over all sentence pair
  Se = m_pTextE->FirstSentence(PosSe);
  Sf = m_pTextF->FirstSentence(PosSf);
  
  Trace( 1, "Sf = " << Sf << " Se " << Se );

  pSentenceE = m_pTextE->GetSentence( PosSe );
  pSentenceF = m_pTextF->GetSentence( PosSf );

  //entenceF->Display( m_pTextF->Vocabulary() );
  //entenceE->Display( m_pTextE->Vocabulary() );

  int s = 0;
  while (Se != m_pTextE->Vocabulary()->EOT && 
	 Sf != m_pTextF->Vocabulary()->EOT) {
    if (s % 1000 == 0) {
      Trace( 1, "Sentence pair " << s );
      //cout << "Sentence pair " << s << "\n";
    }
    
    // get length of sentences; if one is zero, skip
    if ( (m_pTextF->SentenceLength( PosSf ) > 0) &&
	 (m_pTextE->SentenceLength( PosSe ) > 0) ) { 

      SentenceLengthF = 0;
      
      // loop over words in sentence of first text
      for (Wf = m_pTextF->FirstWord(PosSf, PosWf);
	   Wf != m_pTextF->Vocabulary()->EOS;
	   Wf = m_pTextF->NextWord(PosSf, PosWf) ) {
	
	PTrans1 = 0.0;
	SentenceLengthF++;
	TotalLengthF++;
	
	SentenceLengthE = 0;
	// loop over words in sentence of second text
	for (We = m_pTextE->FirstWord(PosSe, PosWe);
	     We != m_pTextE->Vocabulary()->EOS;
	     We = m_pTextE->NextWord(PosSe, PosWe) ) {
	  
	  Trace(3, "Wf: " << Wf << " We: " << We);
	  PTrans1 += m_pLexicon->Pr(Wf, We);
	  // cout << m_pT->GetAt(We,Wf) << endl;
	  SentenceLengthE++;
	}
	
	PTrans += log(PTrans1);
	PAlign += log(1.0 / (float) SentenceLengthE);
      }
      
    }
    
    // fetch next sentence pair
    Se = m_pTextE->NextSentence(PosSe);
    Sf = m_pTextF->NextSentence(PosSf);
    
    s++;
    
  }				// end loop sentences
  
  PTrans = -PTrans / TotalLengthF;
  PAlign = -PAlign / TotalLengthF;
  
  Perplexity = PEta + PAlign + PTrans;
  
//  cout << "Perplexity:\n";
//  cout << "\tSentence length: " << exp(PEta) << "\n";
//  cout << "\tTranslation:     " << exp(PTrans) << "\n";
//  cout << "\tAlignment:       " << exp(PAlign) << "\n";
//  cout << "\tTotal:           " << exp(Perplexity) << "\n";

  cout << m_CurrentIteration << " & " << 
    exp(PTrans) << " & " << exp(PAlign) << " & " << exp(Perplexity) << "\n";

  m_LexiconPerplexity[m_CurrentIteration] = exp(PTrans);
  m_AlignPerplexity[m_CurrentIteration] = exp(PAlign);
  m_TotalPerplexity[m_CurrentIteration] = exp(Perplexity);

  return exp(Perplexity);

  //  return 0.0;
}


double
CTM_IBM1::PerplexityForMaximum()
{
  /*
  double Perplexity;
  
  
  CWordIndex Se, Sf, We, Wf;
  TPosition PosSe, PosSf, PosWe, PosWf;
  double PEta = .0, PAlign = .0, PTrans = .0, PTrans1 = .0, PTransMax;
  int SentenceLengthE = 0, SentenceLengthF = 0;
  int TotalLengthF = 0;
  
  // clear the sentence length table
  for (int i = 0; i < MaxSLen; i++)
    for (int j = 0; j < MaxSLen; j++)
      ;//      m_SentenceLength[i][j] = 0;
  
  
  // loop over all sentence pair
  Se = m_pTextE->FirstSentence(PosSe);
  Sf = m_pTextF->FirstSentence(PosSf);
  
  int s = 0;
  while (Se != m_pTextE->EOT && Sf != m_pTextE->EOT) {
    if (s % 10 == 0) {
      //  cout << "Sentence pair " << s << "\n";
    }
    //  cout << "Sentence pair " << s << " " << Se << " " << Sf << "\n";
    s++;
    
    SentenceLengthF = 0;
    // loop over words in sentence of first text
    for (Wf = m_pTextF->FirstWord(PosSf, PosWf);
	 Wf != m_pTextF->EOS;
	 Wf = m_pTextF->NextWord(PosSf, PosWf) ) {
      
      PTrans1 = 0.0;
      PTransMax = 0.0;

      SentenceLengthF++;
      TotalLengthF++;
      
      SentenceLengthE = 0;
      // loop over words in sentence of second text
      for (We = m_pTextE->FirstWord(PosSe, PosWe);
	   We != m_pTextE->EOS;
	   We = m_pTextE->NextWord(PosSe, PosWe) ) {
	
	SentenceLengthE++;

	PTrans1 = m_pT->GetAt(We,Wf);

	if (PTrans1 > PTransMax) {
	  PTransMax = PTrans1;
	}
      }

      PTrans += -log(PTransMax);
      PAlign += -log(1.0 / (float) SentenceLengthE);
    }
    
    // fetch next sentence pair
    Se = m_pTextE->NextSentence(PosSe);
    Sf = m_pTextF->NextSentence(PosSf);
    
  }				// end loop sentences
  
  PTrans = PTrans / TotalLengthF;
  PAlign = PAlign / TotalLengthF;
  
  Perplexity = PEta + PAlign + PTrans;
  cout << PTrans << " " << PAlign << endl;
  
  cout << "Perplexity with maximum approximation" << endl;
  cout << exp(PTrans) << " & " << exp(PAlign) << " & " << exp(Perplexity) << "\n";

  return exp(Perplexity);
  */
  return 0;
}



bool
CTM_IBM1::TestProbabilities()
{
  //CTranslationModel::TestProbabilities();
  return TRUE; 
}



bool
CTM_IBM1::StoreBiText( const char *pFileName )
{
  Trace(1, "CTM_IBM1::StoreBiText()");

  ofstream OutputFile( pFileName );

  if (!OutputFile) {
    cout << "File " << pFileName << " could not be opened\n";
    return FALSE;
  }

  
  CWordIndex Se, Sf, We, Wf;
  TPosition PosSe, PosSf, PosWe, PosWf;
  //int SentenceLengthE = 0, SentenceLengthF = 0;

  // loop over all sentence pair
  Se = m_pTextE->FirstSentence(PosSe);
  Sf = m_pTextF->FirstSentence(PosSf);
  
  int s = 0;
  while (Se != m_pTextE->Vocabulary()->EOT && 
	 Sf != m_pTextE->Vocabulary()->EOT ) {

    OutputFile << (*m_pTotalPerplexity)[ s ] << "  " << s << " : ";

    // display sentence of source (first, F) text
    for ( Wf = m_pTextF->FirstWord( PosSf, PosWf );
	  Wf != m_pTextF->Vocabulary()->EOS;
	  Wf = m_pTextF->NextWord( PosSf, PosWf ) ) {
      OutputFile << m_pTextF->Vocabulary()->GetWord( Wf ) << " ";
    }
    OutputFile << " # ";

    // display sentence of target (second, E) text
    for ( We = m_pTextE->FirstWord( PosSe, PosWe );
	  We != m_pTextE->Vocabulary()->EOS;
	  We = m_pTextE->NextWord( PosSe, PosWe ) ) {
      OutputFile << m_pTextE->Vocabulary()->GetWord( We ) << " ";
    }
    OutputFile << endl;
    
    // fetch next sentence pair
    Se = m_pTextE->NextSentence(PosSe);
    Sf = m_pTextF->NextSentence(PosSf);

    s++;
    
  } // end loop sentences

  return TRUE;
}


void
CTM_IBM1::DisplayBadSentences()
{
  Trace(1, "CTM_IBM1::DisplayBadSentences()");


  if ( m_SentencePrune <= 0 )
    return;

  cout << "Sentences with high perplexity" << endl;
  cout << "==============================" << endl;

  int i, s;

  for ( i = 0; i <= m_BadSentenceNumbers.MaxIndex(); i++ ) {
    s = m_BadSentenceNumbers[ i ];
    cout << i << " " << s << "  " 
	 << m_BadSentenceValues.GetAt( i ) << "  ";
    m_pTextF->DisplaySentence ( s );
    cout << " # ";
    m_pTextE->DisplaySentence ( s );
    cout << endl;
  }
}


void
CTM_IBM1::RemoveBadSentences()
{
  Trace(0, "CTM_IBM1::RemoveBadSentences()");

  if ( m_SentencePrune <= 0 )
    return;

  int i, s;

  for ( i = 0; i <= m_BadSentenceNumbers.MaxIndex(); i++ ) {
    s = m_BadSentenceNumbers[ i ];
    Trace(0, s << "  " << m_BadSentenceValues.GetAt( i ) );
    m_pTextF->DisplaySentence ( s );
    cout << " # ";
    m_pTextE->DisplaySentence ( s );
    cout << endl;

    m_pTextF->ClearSentence ( s );
    m_pTextE->ClearSentence ( s );

    m_pTextF->DisplaySentence ( s );
    cout << " # ";
    m_pTextE->DisplaySentence ( s );
    cout << endl;
  }
}



void 
CTM_IBM1::CheckBoundaries( const CSentence *pSentence )
{
  Trace( 0, "CTM_IBM1::CheckBoundaries" );

  if ( pSentence->Size() == 0 ) 
    return;

  int i;
  CWordIndex w;
  // loop over words in sentence
  for (i = pSentence->MinIndex();
       i < pSentence->MaxIndex();
       i++ ) {
    w = (*pSentence)[ i ];
    if ( (*m_pIsBoundaryWord)[ w ] == 1 ) {
      cout << "Boundary word at " << i << " is ";
      m_pTextF->Vocabulary()->GetWord( (*pSentence)[ i ] );
      cout << endl;
    }
    
  }
}

