#include <config.h>


#include <iostream>
#include <sstream>
#include <stdio.h>
#include <assert.h>
#include <math.h>

#include <TM_IBM1_Bi.hh>

using namespace std;

//#include "Text.hh"


#undef NO_TRACE
#include <Trace.hh>

CTM_IBM1_Bi::CTM_IBM1_Bi()
{
  cout << "CTM_IBM1_Bi::CTM_IBM1_Bi() " << endl << flush;

  // create the two lexicons
  m_pLexiconFE = new CLexiconForTraining;
  m_pLexiconEF = new CLexiconForTraining;

  m_pLexiconPerplexityEF = new CArray<float>;
  m_pTotalPerplexityEF = new CArray<float>;
  m_pLexiconPerplexityFF = new CArray<float>;
  m_pTotalPerplexityFE = new CArray<float>;

  m_MaxSentenceLength = MaxSenLen;
  m_IncMaxSentenceLength = 0;

  // pruning of sentences
  m_SentencePrune = 0;
  m_FirstPruningIteration = 1000;
  m_BadSentenceValues.SetMaxSize( 0 );
  m_BadSentenceValues.SortReverse();

  // segmentation of sentences
  m_FirstSplitIteration = 1000;
  m_NumberOfSplits = 0;
  m_SplitThreshold = 0.0;
  m_RequiredSplitPoints = 2;
  m_Lambda = 0.5;
  m_pSegmenter = NULL;
  m_pBoundaryWords = new CArray< CString >;
  m_pIsBoundaryWord = new CNumericalArray< int >;

  //cout << "CTM_IBM1::CTM_IBM1() done " << endl << flush;
}


CTM_IBM1_Bi::~CTM_IBM1_Bi()
{
  //clog << "Destructor CTM_IBM1_Bi\n";
  delete m_pLexiconFE;
  delete m_pLexiconEF;
  delete m_pLexiconPerplexityEF;
  delete m_pTotalPerplexityEF;
  delete m_pLexiconPerplexityFF;
  delete m_pTotalPerplexityFE;
}



void 
CTM_IBM1_Bi::SetIncMaxSentenceLength( int IncMaxSentenceLength )
{
  m_IncMaxSentenceLength = IncMaxSentenceLength;
  m_MaxSentenceLength = 0;
  cout << "IncMaxSentenceLength set to " << m_IncMaxSentenceLength;
}


void 
CTM_IBM1_Bi::SetSentencePrune( int SentencePrune )
{
  m_SentencePrune = SentencePrune;
  m_BadSentenceValues.SetMaxSize( SentencePrune );
}


void 
CTM_IBM1_Bi::SetFirstPruningIteration( int FirstPruningIteration )
{
  m_FirstPruningIteration = FirstPruningIteration;
}


void 
CTM_IBM1_Bi::SetFirstSplitIteration( int FirstSplitIteration )
{
  m_FirstSplitIteration = FirstSplitIteration;
}

void 
CTM_IBM1_Bi::SetNumberOfSplits( int NumberOfSplits )
{
  m_NumberOfSplits = NumberOfSplits;
}


void 
CTM_IBM1_Bi::SetSplitWords( const CString &SplitWords )
{
  char token[4096];
  int len = 0;
  char *delimiters(" ");
  const char *TokStr = SplitWords;
  const char *ptr = TokStr;

  Trace( 0, "SplitUpToTokens:" << TokStr );

  while ( *(ptr += strspn( ptr, delimiters )) )  { // Find first word after delimiters
    len = strcspn( ptr, delimiters );                // Find first delimiter after word
    Trace( 0, "Read from: " << ptr );
    strncpy( token, ptr, len );
    token[ len ] = '\0';
    m_pBoundaryWords->Append( CString( token ) );
    Trace( 0, token );
    ptr += len;
  }
  Trace( 0, "" );

  Trace( 0, "All split words:" );
  for ( int i = m_pBoundaryWords->MinIndex();  i <= m_pBoundaryWords->MaxIndex(); i++ ) {
    Trace( 0, i << " " << m_pBoundaryWords->GetAt( i ) );
  }

  // set array for word boundaries
  m_pIsBoundaryWord->SetBoundaries( 0, m_pTextF->Vocabulary()->Size() );
  // TODO: Use TRACE here
  std::cout << "Boundary words" << endl;
  CWordIndex Wb;
  for ( int n = m_pBoundaryWords->MinIndex();
	n <= m_pBoundaryWords->MaxIndex();
	n++ ) {
      // TODO: Use TRACE here
      std::cout << n << " " << (*m_pBoundaryWords)[ n ] << endl;
      Wb = m_pTextF->Vocabulary()->GetIndex( (*m_pBoundaryWords)[ n ] );
      // TODO: Use TRACE here
      std::cout << "WordIndex = " << Wb << endl;
      m_pIsBoundaryWord->SetAt( Wb, 1 );
  }
}



bool
CTM_IBM1_Bi::LoadSplitWords( const CString &FileSplitWords )
{
  cout << "Load split words from " << FileSplitWords << endl;
  char token[4096];
  int len = 0;
  char *delimiters(" ");
  const char *TokStr = "";
  const char *ptr = TokStr;

  CString Word;
  int Type;
  CWordIndex Wb;

  // set array for word boundaries
  m_pIsBoundaryWord->SetBoundaries( 0, m_pTextF->Vocabulary()->Size() );


  ifstream InStream( FileSplitWords );

  if ( !InStream.good() ) {
      cerr <<  "CTM_IBM1_Bi::ReadSplitWords: file " << FileSplitWords << " was not opened for read" << endl << flush;
      return false;
  } // endif: error opening file

  InStream >> skipws >> token >> Type;
  while ( !InStream.eof() ) {
	cout << token << " " << Type << endl;
    m_pBoundaryWords->Append( CString( token ) );
	//    m_pBoundaryTypes->Append( Type ) );
	Wb = m_pTextF->Vocabulary()->GetIndex( token );
	m_pIsBoundaryWord->SetAt( Wb, Type );
    cout << token << " WordIndex = " << Wb << " Type = " << Type << endl;

    InStream >> skipws >> token >> Type;
  }

  return true;
}



void 
CTM_IBM1_Bi::SetSplitThreshold( double SplitThreshold )
{
  m_SplitThreshold = SplitThreshold;
}


void 
CTM_IBM1_Bi::SetRequiredSplitPoints( int RequiredSplitPoints )
{
  m_RequiredSplitPoints = RequiredSplitPoints;
}


void
CTM_IBM1_Bi::SetLexiconUpdate( int LexiconFixed )
{
  m_LexiconFixed = LexiconFixed;
}


bool
CTM_IBM1_Bi::Train()
{
 // return TRUE;

  //m_pT->SetDefaultValue(1.0/m_VocSizeF);

  Trace( 1, "------------------------------" );
  Trace( 1,  " Train IBM1" );

  assert( m_pTextF );
  assert( m_pTextE );



  // create and init lexicon
  if ( m_pLexiconFE == NULL ) {
    Trace( 0, "Create new Lexicon" );
    m_pLexiconFE = new CLexiconForTraining;
    assert( m_pLexiconFE );

	// init lexicon with vocabulary
	m_pLexiconFE->Init( m_pTextF->Vocabulary(), m_pTextE->Vocabulary() );
  }


  // set smoothing parameter
  m_pLexiconFE->SetLambdaSmooth( m_LambdaSmooth );
  Trace( 1, "Test Prob: " << m_pLexiconFE->Pr( 3, 3 ) );

  // set array for sentence perplexities to correct size
  m_pTotalPerplexity->AssureBoundaries( 0, m_pTextF->NumberOfSentences() );


  //Perplexity();
  //return TRUE;

  m_MaxSentenceLength = m_IncMaxSentenceLength;

  // iterate
  for ( m_CurrentIteration = 0; m_CurrentIteration < m_Iterations; m_CurrentIteration++ ) {
    Trace( 0, "\n\n== Iteration " << m_CurrentIteration << " ==\n" );

	m_pSegmenter = new CSegmenter;

    NextIteration();
    // NextIteration calculates perplexity of the model
    // as it was before the current iteration
    //Perplexity();

    //DisplayBadSentences();
    //if ( m_CurrentIteration >= m_FirstPruningIteration )
	//RemoveBadSentences();

	MakeSplits();
	delete m_pSegmenter;

    // go in small steps till 100 words, that take everything

    //m_MaxSentenceLength *= 2;
    m_MaxSentenceLength += m_IncMaxSentenceLength;
    if ( m_MaxSentenceLength > 50 * m_IncMaxSentenceLength )
        m_MaxSentenceLength = MaxSenLen;
    if ( m_MaxSentenceLength > MaxSenLen ) 
	m_MaxSentenceLength = MaxSenLen;

  }

  // calculate perplexity after final iteration
  //double PP;
  Perplexity();
  
  //cout << "Initial perplexity measure: " << PP << "\n";

  //PerplexityForMaximum();

  // set probability for cognates
  //cout << "With Cognates" << endl;
  //InitLexicon();
  //PP = Perplexity();

  //PerplexityForMaximum();

  StoreBiText( "NewCorpus" );

  // display all perplexities
  cout.setf(ios::right | ios::fixed);
  //cout.width(6);
  cout.precision(2);
  cout << "Perplexity as Function of Iteration \n\n";
  cout << "Alignment Source -> Target          &  Alignment Target -> Source";
  cout << "n   &  Lex  &   Align   &   Total   &  Lex  &   Align   &   Total \n";
  for ( int Iteration = 0; Iteration <= m_Iterations; Iteration++ ) {
    cout << Iteration << " & " 
	 << m_LexiconPerplexityFE[Iteration] << " & "  
	 << m_AlignPerplexityFE[Iteration] << " & "  
	 << m_TotalPerplexityFE[Iteration] << "   " 
	 << m_LexiconPerplexityEF[Iteration] << " & "  
	 << m_AlignPerplexityEF[Iteration] << " & "  
	 << m_TotalPerplexityEF[Iteration] << "   " 
		 <<"\n";
  }
  

  return TRUE;
}



bool
CTM_IBM1_Bi::NextIteration()
{
  Trace(0, "CTM_IBM1_Bi::NextIteration ");
  Trace(0, "Test Prob: " << m_pLexiconFE->Pr( 3, 3 ) );

  cout << "Max Sentence Length is now: " << m_MaxSentenceLength << endl;

  // for reestimation
  int i, j;
  CWordIndex Se, Sf, We, Wf;
  TPosition PosSe, PosSf, PosWe, PosWf;
  double Sum, Pr, Count;
  // for perplexity
  double PAlignSumFE = 0.0, PTransSumFE = 0.0, PTransFE = 0.0, PTotalFE;
  double PAlignSumEF = 0.0, PTransSumEF = 0.0, PTransEF = 0.0, PTotalEF;
  int SentenceLengthE = 0, SentenceLengthF = 0;
  int TotalLengthF = 0, TotalLengthE = 0;

  int Position;
  bool IsNew;

  m_pLexiconFE->ClearCounts();
  m_pLexiconEF->ClearCounts();

  m_BadSentenceValues.RemoveAll();
  m_BadSentenceNumbers.RemoveAll();

  //m_pLexicon->Log();
  //m_pLexiconFE->Display();


  //const CSentence *pSentenceE, *pSentenceF;

  // fetch first sentence pair
  Se = m_pTextE->FirstSentence( PosSe );
  Sf = m_pTextF->FirstSentence( PosSf );

  Trace( 1, "Sf = " << Sf << " Se " << Se );

  m_pSentenceE = m_pTextE->GetSentence( PosSe );
  m_pSentenceF = m_pTextF->GetSentence( PosSf );

  m_pSentenceF->Display( m_pTextF->Vocabulary() ); cout << endl;
  m_pSentenceE->Display( m_pTextE->Vocabulary() ); cout << endl;

  // fetch first sentence pair
  Se = m_pTextE->FirstSentence( PosSe );
  Sf = m_pTextF->FirstSentence( PosSf );
  
  uint s = 0;
  uint s_aligned = 0;

  while (Se != m_pTextE->Vocabulary()->EOT && 
		 Sf != m_pTextF->Vocabulary()->EOT)     {

    if (s % 1000 == 0) {
      //Trace( 1, "Sentence pair " << s );
      cout << "Sentence pair " << s << "  Sentences aligned " << s_aligned << endl << flush;
      //cout.flush();
	}
    
	m_pSentenceE = m_pTextE->GetSentence( PosSe );
	m_pSentenceF = m_pTextF->GetSentence( PosSf );

	m_J = m_pSentenceF->Length();
	m_I = m_pSentenceE->Length();
	
    // get length of sentences; if one is zero, skip
    if ( (m_J > 0) && (m_I > 0) && (m_J < m_MaxSentenceLength) && (m_I < m_MaxSentenceLength) ) { 

	  s_aligned++;

	  // fill the two matrices for f->e and e->f
	  for ( j = m_pSentenceF->MinIndex(); j <= m_pSentenceF->MaxIndex(); j++ ) {
		Wf = (*m_pSentenceF)[ j ];
		for ( i = m_pSentenceE->MinIndex(); i <= m_pSentenceE->MaxIndex(); i++ ) {
		  We = (*m_pSentenceE)[ i ];
		  m_ProbsFE[ j ][ i ] = m_pLexiconFE->Pr( Wf, We );
		  m_ProbsEF[ j ][ i ] = m_pLexiconEF->Pr( We, Wf );
		  Trace( 3, Wf << " " << We << " p = " << m_ProbsFE[ j ][ i ] << " : " << m_ProbsEF[ i ][ j ] );
		}
	  }

      m_pTotalPerplexityFE->SetAt( s, 0.0 );
	  m_pTotalPerplexityEF->SetAt( s, 0.0 );
	  
	  TotalLengthF += m_J;
	  TotalLengthE += m_I;

	  // do alignment F->E
	  // -----------------
      // loop over words in sentence of source (first, F) text
      for ( j = m_pSentenceF->MinIndex(); j <= m_pSentenceF->MaxIndex(); j++ ) {

		Sum = 0.0;
		PTransFE = 0.0;
		
		Wf = (*m_pSentenceF)[ j ];

		// loop over words in sentence of destination (second, e) text
		for ( i = m_pSentenceE->MinIndex(); i <= m_pSentenceE->MaxIndex(); i++ ) {
		  Sum += m_ProbsFE[ j ][ i ];
		}
		assert( Sum > 0.0 );
		
		// loop over words in e sentence
		for ( i = m_pSentenceE->MinIndex(); i <= m_pSentenceE->MaxIndex(); i++ ) {

		  // update lexicon count
		  We = (*m_pSentenceE)[ i ];
		  Pr = m_ProbsFE[ j ][ i ];
		  m_pLexiconFE->AddCount( Wf, We, Pr / Sum );

		  // for calculation of perplexity
		  PTransFE += Pr;
		}
		
		PTransSumFE += log(PTransFE);
		PAlignSumFE += log(1.0 / (float) m_I);
		
		(*m_pTotalPerplexityFE)[ s ] += log(PTransFE);
      }
	  
      (*m_pTotalPerplexityFE)[ s ] = exp( -(*m_pTotalPerplexityFE)[ s ] / (double) m_J );
	  

	  // do alignment E->F
	  // -----------------
      // loop over words in sentence of target (second, e) text
      for ( i = m_pSentenceE->MinIndex(); i <= m_pSentenceE->MaxIndex(); i++ ) {

		Sum = 0.0;
		PTransEF = 0.0;
		
		We = (*m_pSentenceE)[ i ];

		// loop over words in sentence of source (first, f) text
		for ( j = m_pSentenceF->MinIndex(); j <= m_pSentenceF->MaxIndex(); j++ ) {
		  Sum += m_ProbsEF[ j ][ i ];
		}
		assert( Sum > 0.0 );
		
		// loop over words in e sentence
		for ( j = m_pSentenceF->MinIndex(); j <= m_pSentenceF->MaxIndex(); j++ ) {

		  // update lexicon count
		  Wf = (*m_pSentenceF)[ j ];
		  Pr = m_ProbsEF[ j ][ i ];
		  m_pLexiconEF->AddCount( We, Wf, Pr / Sum );

		  // for calculation of perplexity
		  PTransEF += Pr;
		}
		
		PTransSumEF += log(PTransEF);
		PAlignSumEF += log(1.0 / (float) m_J);
		
		(*m_pTotalPerplexityEF)[ s ] += log(PTransEF);
      }
      (*m_pTotalPerplexityEF)[ s ] = exp( -(*m_pTotalPerplexityEF)[ s ] / (double) m_I );
	  

      // insert value of sentence into list of bad sentences
      Position = m_BadSentenceValues.Insert( (*m_pTotalPerplexityFE)[ s ], IsNew );
      if ( IsNew ) {
		m_BadSentenceNumbers.Insert( Position, s );
      } else {
		m_BadSentenceNumbers.SetAt( Position, s );
      }
      
      if ( m_BadSentenceNumbers.MaxIndex() > m_BadSentenceValues.MaxIndex() ) {
		m_BadSentenceNumbers.SetMaxIndex( m_BadSentenceValues.MaxIndex() );
      }
      
	  if ( m_CurrentIteration + 1>= m_FirstSplitIteration ) {  // Current counted from 0
		//cout << "m_CurrentIteration + 1>= m_FirstSplitIteration" << endl;
		//cout << m_J << " " << m_MaxSentenceLength - m_IncMaxSentenceLength << " "
		//	 << m_I << " " << m_MaxSentenceLength - m_IncMaxSentenceLength;

		// we want to split only sentences which have been seen in the previous iterations
		// ie for which the lexicon probabilities are somewhat reliable
		// and also only sentences which are longer than 2 words
		//if ( (m_J <= m_MaxSentenceLength / 2) &&
		//	 (m_I <= m_MaxSentenceLength / 2) ) { 

		if ( (m_J <= m_MaxSentenceLength - m_IncMaxSentenceLength) &&
		     (m_I <= m_MaxSentenceLength - m_IncMaxSentenceLength) && 
                     (m_J >= 3) &&
		     (m_I >= 3) ) { 
		  // calculate best split point
		  FindBestSplit( s );
		}
	  }

    } else {
      //cout << "Sentence Pair " << s << " skipped" << endl << flush;
    }
	
    // fetch next sentence pair
    Se = m_pTextE->NextSentence( PosSe );
    Sf = m_pTextF->NextSentence( PosSf );
	
    s++;
  } // end loop over sentence pairs
  
  //m_pLexiconFE->Display();
  
  PTransSumFE = -PTransSumFE / TotalLengthF;
  PAlignSumFE = -PAlignSumFE / TotalLengthF;
  PTotalFE = PAlignSumFE + PTransSumFE;

  m_LexiconPerplexityFE[m_CurrentIteration] = exp(PTransSumFE);
  m_AlignPerplexityFE[m_CurrentIteration] = exp(PAlignSumFE);
  m_TotalPerplexityFE[m_CurrentIteration] = exp(PTotalFE);


  PTransSumEF = -PTransSumEF / TotalLengthE;
  PAlignSumEF = -PAlignSumEF / TotalLengthE;
  PTotalEF = PAlignSumEF + PTransSumEF;

  m_LexiconPerplexityEF[m_CurrentIteration] = exp(PTransSumEF);
  m_AlignPerplexityEF[m_CurrentIteration] = exp(PAlignSumEF);
  m_TotalPerplexityEF[m_CurrentIteration] = exp(PTotalEF);



  cout << m_CurrentIteration 
	   << " & " << exp(PTransSumFE) << " & " << exp(PAlignSumFE) << " & " << exp(PTotalFE)
	   << " & " << exp(PTransSumEF) << " & " << exp(PAlignSumEF) << " & " << exp(PTotalEF) << endl;


  // normalize the translation probabilities

  if ( m_LexiconFixed == 0 ) {
    Trace( 0, "Renormalize" );
    //m_pLexiconFE->DisplayCounts();
    m_pLexiconFE->Renormalize();
    m_pLexiconEF->Renormalize();
    //m_pLexiconFE->TestProbabilities();
  }

  return TRUE;
}



double
CTM_IBM1_Bi::Perplexity()
{
  Trace(0, "CTM_IBM1_Bi::Perplexity()");

  double Perplexity;
  
  CWordIndex Se, Sf, We, Wf;
  TPosition PosSe, PosSf, PosWe, PosWf;
  double PEta = .0, PAlign = .0, PTrans = .0, PTrans1 = .0;
  int SentenceLengthE = 0, SentenceLengthF = 0;
  int TotalLengthF = 0;

  const CSentence *pSentenceE, *pSentenceF;

  // clear the sentence length table
  //for (int i = 0; i < MaxSLen; i++)
  //  for (int j = 0; j < MaxSLen; j++)
  //    ;//      m_SentenceLength[i][j] = 0;
  
  
  // loop over all sentence pair
  Se = m_pTextE->FirstSentence(PosSe);
  Sf = m_pTextF->FirstSentence(PosSf);
  
  Trace( 1, "Sf = " << Sf << " Se " << Se );

  pSentenceE = m_pTextE->GetSentence( PosSe );
  pSentenceF = m_pTextF->GetSentence( PosSf );

  //entenceF->Display( m_pTextF->Vocabulary() );
  //entenceE->Display( m_pTextE->Vocabulary() );

  int s = 0;
  while (Se != m_pTextE->Vocabulary()->EOT && 
	 Sf != m_pTextF->Vocabulary()->EOT) {
    if (s % 1000 == 0) {
      Trace( 1, "Sentence pair " << s );
      //cout << "Sentence pair " << s << "\n";
    }
    
    // get length of sentences; if one is zero, skip
    if ( (m_pTextF->SentenceLength( PosSf ) > 0) &&
		 (m_pTextE->SentenceLength( PosSe ) > 0) ) { 

      SentenceLengthF = 0;
      
      // loop over words in sentence of first text
      for (Wf = m_pTextF->FirstWord(PosSf, PosWf);
	   Wf != m_pTextF->Vocabulary()->EOS;
	   Wf = m_pTextF->NextWord(PosSf, PosWf) ) {
	
		PTrans1 = 0.0;
		SentenceLengthF++;
		TotalLengthF++;
		
		SentenceLengthE = 0;
		// loop over words in sentence of second text
		for (We = m_pTextE->FirstWord(PosSe, PosWe);
			 We != m_pTextE->Vocabulary()->EOS;
			 We = m_pTextE->NextWord(PosSe, PosWe) ) {
		  
		  Trace(3, "Wf: " << Wf << " We: " << We);
		  PTrans1 += m_pLexiconFE->Pr(Wf, We);
		  // cout << m_pT->GetAt(We,Wf) << endl;
		  SentenceLengthE++;
		}
		
		PTrans += log(PTrans1);
		PAlign += log(1.0 / (float) SentenceLengthE);
      }
      
    }
    
    // fetch next sentence pair
    Se = m_pTextE->NextSentence(PosSe);
    Sf = m_pTextF->NextSentence(PosSf);
    
    s++;
    
  }				// end loop sentences
  
  PTrans = -PTrans / TotalLengthF;
  PAlign = -PAlign / TotalLengthF;
  
  Perplexity = PEta + PAlign + PTrans;
  
  //  cout << "Perplexity:\n";
  //  cout << "\tSentence length: " << exp(PEta) << "\n";
  //  cout << "\tTranslation:     " << exp(PTrans) << "\n";
  //  cout << "\tAlignment:       " << exp(PAlign) << "\n";
  //  cout << "\tTotal:           " << exp(Perplexity) << "\n";
  
  cout << m_CurrentIteration << " & " << 
    exp(PTrans) << " & " << exp(PAlign) << " & " << exp(Perplexity) << "\n";
  
  m_LexiconPerplexityFE[m_CurrentIteration] = exp(PTrans);
  m_AlignPerplexityFE[m_CurrentIteration] = exp(PAlign);
  m_TotalPerplexityFE[m_CurrentIteration] = exp(Perplexity);
  
  return exp(Perplexity);
  
  //  return 0.0;
}


double
CTM_IBM1_Bi::PerplexityForMaximum()
{
  /*
  double Perplexity;
  
  
  CWordIndex Se, Sf, We, Wf;
  TPosition PosSe, PosSf, PosWe, PosWf;
  double PEta = .0, PAlign = .0, PTrans = .0, PTrans1 = .0, PTransMax;
  int SentenceLengthE = 0, SentenceLengthF = 0;
  int TotalLengthF = 0;
  
  // clear the sentence length table
  for (int i = 0; i < MaxSLen; i++)
    for (int j = 0; j < MaxSLen; j++)
      ;//      m_SentenceLength[i][j] = 0;
  
  
  // loop over all sentence pair
  Se = m_pTextE->FirstSentence(PosSe);
  Sf = m_pTextF->FirstSentence(PosSf);
  
  int s = 0;
  while (Se != m_pTextE->EOT && Sf != m_pTextE->EOT) {
    if (s % 10 == 0) {
      //  cout << "Sentence pair " << s << "\n";
    }
    //  cout << "Sentence pair " << s << " " << Se << " " << Sf << "\n";
    s++;
    
    SentenceLengthF = 0;
    // loop over words in sentence of first text
    for (Wf = m_pTextF->FirstWord(PosSf, PosWf);
	 Wf != m_pTextF->EOS;
	 Wf = m_pTextF->NextWord(PosSf, PosWf) ) {
      
      PTrans1 = 0.0;
      PTransMax = 0.0;

      SentenceLengthF++;
      TotalLengthF++;
      
      SentenceLengthE = 0;
      // loop over words in sentence of second text
      for (We = m_pTextE->FirstWord(PosSe, PosWe);
	   We != m_pTextE->EOS;
	   We = m_pTextE->NextWord(PosSe, PosWe) ) {
	
	SentenceLengthE++;

	PTrans1 = m_pT->GetAt(We,Wf);

	if (PTrans1 > PTransMax) {
	  PTransMax = PTrans1;
	}
      }

      PTrans += -log(PTransMax);
      PAlign += -log(1.0 / (float) SentenceLengthE);
    }
    
    // fetch next sentence pair
    Se = m_pTextE->NextSentence(PosSe);
    Sf = m_pTextF->NextSentence(PosSf);
    
  }				// end loop sentences
  
  PTrans = PTrans / TotalLengthF;
  PAlign = PAlign / TotalLengthF;
  
  Perplexity = PEta + PAlign + PTrans;
  cout << PTrans << " " << PAlign << endl;
  
  cout << "Perplexity with maximum approximation" << endl;
  cout << exp(PTrans) << " & " << exp(PAlign) << " & " << exp(Perplexity) << "\n";

  return exp(Perplexity);
  */
  return 0;
}



bool
CTM_IBM1_Bi::TestProbabilities()
{
  //CTranslationModel::TestProbabilities();
  return TRUE; 
}



bool
CTM_IBM1_Bi::StoreBiText( const char *pFileName )
{
  Trace(1, "CTM_IBM1_Bi::StoreBiText()");

  ofstream OutputFile( pFileName );

  if (!OutputFile) {
    cout << "File " << pFileName << " could not be opened\n";
    return FALSE;
  }

  
  CWordIndex Se, Sf, We, Wf;
  TPosition PosSe, PosSf, PosWe, PosWf;
  //int SentenceLengthE = 0, SentenceLengthF = 0;

  // loop over all sentence pair
  Se = m_pTextE->FirstSentence(PosSe);
  Sf = m_pTextF->FirstSentence(PosSf);
  
  int s = 0;
  while (Se != m_pTextE->Vocabulary()->EOT && 
	 Sf != m_pTextE->Vocabulary()->EOT ) {

    OutputFile << (*m_pTotalPerplexity)[ s ] << "  " << s << " : ";

    // display sentence of source (first, F) text
    for ( Wf = m_pTextF->FirstWord( PosSf, PosWf );
	  Wf != m_pTextF->Vocabulary()->EOS;
	  Wf = m_pTextF->NextWord( PosSf, PosWf ) ) {
      OutputFile << m_pTextF->Vocabulary()->GetWord( Wf ) << " ";
    }
    OutputFile << " # ";

    // display sentence of target (second, E) text
    for ( We = m_pTextE->FirstWord( PosSe, PosWe );
	  We != m_pTextE->Vocabulary()->EOS;
	  We = m_pTextE->NextWord( PosSe, PosWe ) ) {
      OutputFile << m_pTextE->Vocabulary()->GetWord( We ) << " ";
    }
    OutputFile << endl;
    
    // fetch next sentence pair
    Se = m_pTextE->NextSentence(PosSe);
    Sf = m_pTextF->NextSentence(PosSf);

    s++;
    
  } // end loop sentences

  return TRUE;
}


void
CTM_IBM1_Bi::DisplayBadSentences()
{
  Trace(1, "CTM_IBM1_Bi::DisplayBadSentences()");


  if ( m_SentencePrune <= 0 )
    return;

  cout << "Sentences with high perplexity" << endl;
  cout << "==============================" << endl;

  int i, s;

  for ( i = 0; i <= m_BadSentenceNumbers.MaxIndex(); i++ ) {
    s = m_BadSentenceNumbers[ i ];
    cout << i << " " << s << "  " 
	 << m_BadSentenceValues.GetAt( i ) << "  ";
    m_pTextF->DisplaySentence ( s );
    cout << " # ";
    m_pTextE->DisplaySentence ( s );
    cout << endl;
  }
}


void
CTM_IBM1_Bi::RemoveBadSentences()
{
  Trace(0, "CTM_IBM1_Bi::RemoveBadSentences()");

  if ( m_SentencePrune <= 0 )
    return;

  int i, s;

  for ( i = 0; i <= m_BadSentenceNumbers.MaxIndex(); i++ ) {
    s = m_BadSentenceNumbers[ i ];
    Trace(0, s << "  " << m_BadSentenceValues.GetAt( i ) );
    m_pTextF->DisplaySentence ( s );
    cout << " # ";
    m_pTextE->DisplaySentence ( s );
    cout << endl;

    m_pTextF->ClearSentence ( s );
    m_pTextE->ClearSentence ( s );

    m_pTextF->DisplaySentence ( s );
    cout << " # ";
    m_pTextE->DisplaySentence ( s );
    cout << endl;
  }
}



void 
CTM_IBM1_Bi::CheckBoundaries( const CSentence *pSentence )
{
  Trace( 0, "CTM_IBM1_Bi::CheckBoundaries" );

  if ( pSentence->Size() == 0 ) 
    return;

  int i;
  CWordIndex w;
  // loop over words in sentence
  for (i = pSentence->MinIndex();
       i < pSentence->MaxIndex();
       i++ ) {
    w = (*pSentence)[ i ];
    if ( (*m_pIsBoundaryWord)[ w ] == 1 ) {
      cout << "Boundary word at " << i << " is ";
      m_pTextF->Vocabulary()->GetWord( (*pSentence)[ i ] );
      cout << endl;
    }
    
  }
}



void
CTM_IBM1_Bi::FindBestSplit( int SentenceNumber )
{
  // for all word pairs ( f, e ), if they are boundary words, i.e. words, 
  // where the sentence can be split, then calculate the alignment probability 
  // for the split sentence and store the best information for the best split
 
  int split_j, split_i;
  int j, i;
  double SumPr;
  double LogProbFE, LogProbLexFE, LogProbAlignFE;
  double LogProbEF, LogProbLexEF, LogProbAlignEF;
  double LogProbSplittedFE, LogProbLexSplittedFE, LogProbAlignSplittedFE;
  double LogProbSplittedEF, LogProbLexSplittedEF, LogProbAlignSplittedEF;
  double Difference, BestDifference = -10000.0;
  int Best_j =-1, Best_i = -1;
  int offset_j, offset_i;
  CWordIndex e, f;

  //cout << "Find best split point in " << endl;
  //m_pSentenceF->Display( m_pTextF->Vocabulary() ); cout << endl;
  //m_pSentenceE->Display( m_pTextE->Vocabulary() ); cout << endl;
  

  // first, renormalize probabilities in lexicon matrix
  //if ( m_Renormalize ) {
  if ( 1 ) {
	for ( j = m_pSentenceF->MinIndex(); j <= m_pSentenceF->MaxIndex(); j++ ) {
	  SumPr = 0.0;
	  for ( i = m_pSentenceE->MinIndex(); i <= m_pSentenceE->MaxIndex(); i++ ) {
		SumPr += m_ProbsFE[ j ][ i ];
	  }
	  for ( i = m_pSentenceE->MinIndex(); i <= m_pSentenceE->MaxIndex(); i++ ) {
		m_ProbsFE[ j ][ i ] /= SumPr;
	  }
	}
	
	for ( i = m_pSentenceE->MinIndex(); i <= m_pSentenceE->MaxIndex(); i++ ) {
	  SumPr = 0.0;
	  for ( j = m_pSentenceF->MinIndex(); j <= m_pSentenceF->MaxIndex(); j++ ) {
		SumPr += m_ProbsEF[ j ][ i ];
	  }
	  for ( j = m_pSentenceF->MinIndex(); j <= m_pSentenceF->MaxIndex(); j++ ) {
		m_ProbsEF[ j ][ i ] /= SumPr;
	  }
	}
  }


  /*
  for ( i = m_I; i >= 1; i-- ) {
    for ( j = 1; j <= m_J; j++ ) {
      cout.width(8);
      cout.precision(2);
      cout << m_ProbsFE[ j ][ i ] << "  ";
    }
    cout << endl;
  }
  cout << endl;

  for ( i = m_I; i >= 1; i-- ) {
    for ( j = 1; j <= m_J; j++ ) {
      cout.width(8);
      cout.precision(2);
      cout << m_ProbsEF[ j ][ i ] << "  ";
    }
    cout << endl;
  }
  cout << endl;
  cout << endl;
  */


  // calculate again the alignment probability for the sentence pair
  // is redundant, should be changed
  // for f->e
  LogProbLexFE = 0.0;
  LogProbAlignFE = 0.0;
  for ( int j = m_pSentenceF->MinIndex(); j <= m_pSentenceF->MaxIndex(); j++ ) {
	SumPr = 0.0;
	for ( int i = m_pSentenceE->MinIndex(); i <= m_pSentenceE->MaxIndex(); i++ ) {
	  SumPr += m_ProbsFE[ j ] [ i ];
	}

	LogProbLexFE -= log( SumPr );
	LogProbAlignFE -= log(1.0 / (float) m_I);
  }

  LogProbLexFE /= m_J;
  LogProbAlignFE /= m_J;
  LogProbFE = LogProbLexFE + LogProbAlignFE;


  // and for e->f
  LogProbLexEF = 0.0;
  LogProbAlignEF = 0.0;
  for ( int i = m_pSentenceE->MinIndex(); i <= m_pSentenceE->MaxIndex(); i++ ) {
	SumPr = 0.0;
	for ( int j = m_pSentenceF->MinIndex(); j <= m_pSentenceF->MaxIndex(); j++ ) {
	  SumPr += m_ProbsEF[ j ] [ i ];
	}

	LogProbLexEF -= log( SumPr );
	LogProbAlignEF -= log(1.0 / (float) m_J);
  }

  LogProbLexEF /= m_I;
  LogProbAlignEF /= m_I;
  LogProbEF = LogProbLexEF + LogProbAlignEF;

//    cout << "Unsplitted:    " << LogProbLexFE << " " << LogProbAlignFE << " " << LogProbFE << "   "
//     << LogProbLexEF << " " << LogProbAlignEF << " " << LogProbEF << endl;

  // now, find optimal split point (don't take last word as split point)
  for ( split_j = m_pSentenceF->MinIndex() + 1; split_j < m_pSentenceF->MaxIndex() - 1; split_j++ ) {
	f = (*m_pSentenceF)[ split_j ];
	
	for ( split_i = m_pSentenceE->MinIndex() + 1; split_i < m_pSentenceE->MaxIndex() - 1; split_i++ ) {
	  e = (*m_pSentenceE)[ split_i ];
	  
	  // we split only at split words
	  int FoundSplitPoints = 0;
	  if ( (*m_pIsBoundaryWord)[ f ] ) FoundSplitPoints++;
	  if ( (*m_pIsBoundaryWord)[ e ] ) FoundSplitPoints++;

	  if ( FoundSplitPoints < m_RequiredSplitPoints ) {
		//Trace( 0, (*m_pIsBoundaryWord)[ f ] << " + " << (*m_pIsBoundaryWord)[ e ] 
		//	     << " < " << m_RequiredSplitPoints );
		;
	  } else {

		// type is 1 for left associative and 2 for right associative
		offset_j = -(*m_pIsBoundaryWord)[ f ] + 1; 
		offset_i = -(*m_pIsBoundaryWord)[ e ] + 1;
		//offset_j = 0;
		//offset_i = 0;
/*
		Trace( 0, "Number of equired split points found" );
		std::cout << "Boundary word at " << split_j << " is ";
		std::cout << m_pTextF->Vocabulary()->GetWord( f );
		std::cout << " offset is " << offset_j;
		std::cout << endl;
		
		std::cout << "Boundary word at " << split_i << " is ";
		std::cout << m_pTextE->Vocabulary()->GetWord( e );
		std::cout << " offset is " << offset_i;
		std::cout << endl;
*/
/*
		cout << endl;
		cout << "Sentence length: " << m_J << " " << m_I << endl;
		cout << "Segment Points are " << split_j << " " << split_i << endl;
		cout << "First Segment " << m_pSentenceF->MinIndex() << " to " << split_j + offset_j
			<< " and " << m_pSentenceE->MinIndex() << " to " <<  split_i + offset_i << endl;
		cout << "Second Segment " << split_j + offset_j + 1 << " to " <<  m_pSentenceF->MaxIndex()
			<< " and " << split_i + offset_i + 1 << " to " << m_pSentenceE->MaxIndex() << endl;
*/
		int LengthSeg1Source = split_j + offset_j - m_pSentenceF->MinIndex() + 1;
		int LengthSeg1Target = split_i + offset_i - m_pSentenceE->MinIndex() + 1;
		int LengthSeg2Source = m_pSentenceF->MaxIndex() -( split_j + offset_j + 1) + 1;
		int LengthSeg2Target = m_pSentenceE->MaxIndex() - (split_i + offset_i + 1) + 1;
		float SentenceLengthRatio = (float)m_I / (float)m_J;
/*
		cout << "Length Source Segm1 = " << split_j + offset_j - m_pSentenceF->MinIndex() + 1 << endl;
		cout << "Length Target Segm1 = " << split_i + offset_i - m_pSentenceE->MinIndex() + 1 << endl;
		cout << "Length Source Segm2 = " << m_pSentenceF->MaxIndex() -( split_j + offset_j + 1) + 1 << endl;
		cout << "Length Target Segm2 = " << m_pSentenceE->MaxIndex() - (split_i + offset_i + 1) + 1 << endl;
		cout << "Sentence length ratio = " << SentenceLengthRatio << endl;

		cout << "Seg1 Source: " << LengthSeg1Source << " <? " << 1.0 / SentenceLengthRatio * LengthSeg1Target * 1.5 + 1 << endl;
		cout << "Seg2 Source: " << LengthSeg2Source << " <? " << 1.0 / SentenceLengthRatio * LengthSeg2Target * 1.5 + 1 << endl;
		cout << "Seg1 Target: " << LengthSeg1Target << " <? " << SentenceLengthRatio * LengthSeg1Source * 1.5 + 1 << endl;
		cout << "Seg2 Target: " << LengthSeg2Target << " <? " << SentenceLengthRatio * LengthSeg2Source * 1.5 + 1 << endl;
*/
                bool Cond1 = ( LengthSeg1Source <= 1.0 / SentenceLengthRatio * LengthSeg1Target * 1.5 + 1 );
                bool Cond2 = ( LengthSeg2Source <= 1.0 / SentenceLengthRatio * LengthSeg2Target * 1.5 + 1 );
                bool Cond3 = ( LengthSeg1Target <= SentenceLengthRatio * LengthSeg1Source * 1.5 + 1 );
                bool Cond4 = ( LengthSeg2Target <= SentenceLengthRatio * LengthSeg2Source * 1.5 + 1 );
 
		//cout << split_j << " >=? " << ( (float)m_J / (float)m_I * (float)split_i * 1.5 + 1 ) << endl;
		//cout << split_i << " >=? " << ( (float)m_I / (float)m_J * (float)split_j * 1.5 + 1 ) << endl;

		if ( ! (Cond1 && Cond2 && Cond3 && Cond4) ) {
		//if ( abs( split_i - split_j * m_I / m_J ) > 10 ) {
		  //cout << "Segment Points too far apart" << endl;
		  ;
		} else {
		  
		  // first f->e
		  // ----------
		  LogProbLexSplittedFE = 0.0;
		  LogProbAlignSplittedFE = 0.0;

		  for ( int j = m_pSentenceF->MinIndex(); j <= m_pSentenceF->MaxIndex(); j++ ) {
			SumPr = 0;
			if ( j <= split_j + offset_j ) {
			  for ( int i = m_pSentenceE->MinIndex(); i <= split_i + offset_i; i++ ) {
				SumPr += m_ProbsFE[ j ] [ i ];
			  }
			  LogProbAlignSplittedFE -= log( 1.0 / (float)(split_i + offset_i - m_pSentenceE->MinIndex() + 1) );
			} else {
			  for ( int i = split_i + offset_i + 1; i <= m_pSentenceE->MaxIndex(); i++ ) {
				SumPr += m_ProbsFE[ j ] [ i ];
			  }
			  LogProbAlignSplittedFE -= log( 1.0 / (float)(m_pSentenceE->MaxIndex() - split_i + offset_i) );
			}
			LogProbLexSplittedFE -= log( SumPr );
		  }

		  LogProbLexSplittedFE /= m_J;
		  LogProbAlignSplittedFE /= m_J;
		  LogProbSplittedFE = LogProbLexSplittedFE + LogProbAlignSplittedFE;


		  // now e->f
		  // ----------
		  LogProbLexSplittedEF = 0.0;
		  LogProbAlignSplittedEF = 0.0;

		  for ( int i = m_pSentenceE->MinIndex(); i <= m_pSentenceE->MaxIndex(); i++ ) {
			SumPr = 0;
			if ( i <= split_i + offset_i ) {
			  for ( int j = m_pSentenceF->MinIndex(); j <= split_j + offset_j; j++ ) {
				SumPr += m_ProbsEF[ j ] [ i ];
			  }
			  LogProbAlignSplittedEF -= log( 1.0 / (float)(split_j + offset_j - m_pSentenceF->MinIndex() + 1) );
			} else {
			  for ( int j = split_j + offset_j + 1; j <= m_pSentenceF->MaxIndex(); j++ ) {
				SumPr += m_ProbsEF[ j ] [ i ];
			  }
			  LogProbAlignSplittedEF -= log( 1.0 / (float)(m_pSentenceF->MaxIndex() - split_j + offset_j) );
			}
			LogProbLexSplittedEF -= log( SumPr );
		  }

		  LogProbLexSplittedEF /= m_I;
		  LogProbAlignSplittedEF /= m_I;
		  LogProbSplittedEF = LogProbLexSplittedEF + LogProbAlignSplittedEF;

		  // we are interested in the total difference
		  Difference = (1 - m_Lambda) * (LogProbFE - LogProbSplittedFE)
			+ m_Lambda * (LogProbEF - LogProbSplittedEF);

		  /*
		  cout << "Split " << split_j << " " << split_i << ": " 
			   << LogProbLexSplittedFE << " " << LogProbAlignSplittedFE << " " << LogProbSplittedFE << "   "
			   << LogProbLexSplittedEF << " " << LogProbAlignSplittedEF << " " << LogProbSplittedEF << "   "
			   << (1 - m_Lambda) * LogProbLexSplittedFE + m_Lambda * LogProbLexSplittedEF << " "
			   << (1 - m_Lambda) * LogProbAlignSplittedFE + m_Lambda * LogProbAlignSplittedEF << " " 
			   << (1 - m_Lambda) * LogProbSplittedFE + m_Lambda * LogProbSplittedEF << "   "
			   << (1 - m_Lambda) * (LogProbFE - LogProbSplittedFE) << " " << m_Lambda * (LogProbEF - LogProbSplittedEF) << "   "
			   << Difference << endl;
		  */
/*
		  Trace( 0, "Splitpoint: " << " " << split_j << " " << split_i
				 << "  Unsplitted: " << LogProbFE
				 << "  Splitted: " << (LogProbSplittedFE)
				 << "  Difference: " << (LogProbFE - LogProbSplittedFE) );
*/		  
		  // store best split points in this sentence pair
		  if ( Difference > BestDifference ) {
			BestDifference = Difference;
			Best_j = split_j + offset_j;
			Best_i = split_i + offset_i;
		  }
		  
		}
	  }
	}
  }

  // store best split point in the segmenter
  if ( Best_j >= 0 ) {

/*
                int LengthSeg1Source = Best_j + - m_pSentenceF->MinIndex() + 1;
                int LengthSeg1Target = Best_i + - m_pSentenceE->MinIndex() + 1;
                int LengthSeg2Source = m_pSentenceF->MaxIndex() -( Best_j + 1) + 1;
                int LengthSeg2Target = m_pSentenceE->MaxIndex() - (Best_i + 1) + 1;

                bool Cond1 = ( LengthSeg1Source <= LengthSeg1Target * 1.5 + 1 );
                bool Cond2 = ( LengthSeg2Source <= LengthSeg2Target * 1.5 + 1 );
                bool Cond3 = ( LengthSeg1Target <= LengthSeg1Source * 1.5 + 1 );
                bool Cond4 = ( LengthSeg2Target <= LengthSeg2Source * 1.5 + 1 );
 
		if ( ! (Cond1 && Cond2 && Cond3 && Cond4) ) {
		  cout << "LENGTH MISMATCH" << endl;
		}

		cout << "Best split point: " << Best_j << " " << Best_i << " " << -BestDifference << endl;
		cout << m_J << " " << m_I << endl;

                cout << "Length Source Segm1 = " << split_j - m_pSentenceF->MinIndex() + 1 << endl;
                cout << "Length Target Segm1 = " << split_i - m_pSentenceE->MinIndex() + 1 << endl;
                cout << "Length Source Segm2 = " << m_pSentenceF->MaxIndex() -( split_j + 1) + 1 << endl;
                cout << "Length Target Segm2 = " << m_pSentenceE->MaxIndex() - (split_i + 1) + 1 << endl;
*/

	CSplitPoint BestSplitPoint( SentenceNumber, Best_j, Best_i, -BestDifference );
	m_pSegmenter->AddSplitPoint( BestSplitPoint );
  }
}



void
CTM_IBM1_Bi::MakeSplits()
{
  cout << "MakeSplits" << endl;  
  //cout << "all split points" << endl;
  //m_pSegmenter->Display();
  m_pSegmenter->SortToValue();
  //cout << "after sorting" << endl;
  //m_pSegmenter->Display();
  // we keep only best split points
  m_pSegmenter->Resize( m_NumberOfSplits );
  //m_pSegmenter->Display();

  //cout << "sorted according to sentence number" << endl;
  m_pSegmenter->SortToPosition();
  //m_pSegmenter->Display();

  // now we split the sentences
  Trace( 0, "Now the actual splitting" );
  CSplitPoint SplitPoint;
  const vector< CSplitPoint > SplitPoints = m_pSegmenter->SplitPoints();

  cout << "Number of requested split = " << m_NumberOfSplits << endl
       << "Number of possible splits = " << SplitPoints.size() << endl;

  CSentence *pOrgSentenceF;
  CSentence *pOrgSentenceE;
  CSentence *pNewSentenceF1;
  CSentence *pNewSentenceE1;
  CSentence *pNewSentenceF2;
  CSentence *pNewSentenceE2;
  CSentence *pNewSentenceF3;
  CSentence *pNewSentenceE3;


  int SplitsDone = 0;

  for ( int n = SplitPoints.size() - 1; n >= 0; --n ) {
    SplitPoint = SplitPoints[ n ];
    //cout << SplitPoint;

    // split only if score is negative, i.e. gain in LogProbability
    if ( SplitPoint.m_Score < m_SplitThreshold ) {

	  pOrgSentenceF = m_pTextF->GetSentence( SplitPoint.m_Sentence );
      pOrgSentenceE = m_pTextE->GetSentence( SplitPoint.m_Sentence );

/*	  
      cout << "split: ";
      pOrgSentenceF->Display( m_pTextF->Vocabulary() );
      cout << " # ";
      pOrgSentenceE->Display( m_pTextE->Vocabulary() );
      cout << endl;
*/	  
      pNewSentenceF1 = new CSentence;
      pNewSentenceE1 = new CSentence;
      pNewSentenceF2 = new CSentence;
      pNewSentenceE2 = new CSentence;
      pNewSentenceF3 = new CSentence;
      pNewSentenceE3 = new CSentence;

	  /*
	  // 2-way split
      // first segment f
      for ( int j = pOrgSentenceF->MinIndex(); j <= SplitPoint.m_j; j++ ) {
		(*pNewSentenceF1)[ j ] = (*pOrgSentenceF)[ j ];
      }
      // first segment e
      for ( int i = pOrgSentenceE->MinIndex(); i <= SplitPoint.m_i; i++ ) {
		(*pNewSentenceE1)[ i ] = (*pOrgSentenceE)[ i ];
      }

	  for ( int j = SplitPoint.m_j + 1; j <= pOrgSentenceF->MaxIndex(); j++ ) {
		(*pNewSentenceF2)[ j - SplitPoint.m_j ] = (*pOrgSentenceF)[ j ];
      }
      // third segment e
      for ( int i = SplitPoint.m_i + 1; i <= pOrgSentenceE->MaxIndex(); i++ ) {
		(*pNewSentenceE2)[ i - SplitPoint.m_i ] = (*pOrgSentenceE)[ i ];
      }

      m_pTextF->PutSentence( SplitPoint.m_Sentence, *pNewSentenceF1 );
      m_pTextE->PutSentence( SplitPoint.m_Sentence, *pNewSentenceE1 );

      m_pTextF->InsertSentence( SplitPoint.m_Sentence + 1, pNewSentenceF2 );
      m_pTextE->InsertSentence( SplitPoint.m_Sentence + 1, pNewSentenceE2 );


	  cout << "into: ";
      pNewSentenceF1->Display( m_pTextF->Vocabulary() ); cout << " # ";
      pNewSentenceE1->Display( m_pTextE->Vocabulary() ); cout << endl;
      cout << "and:  ";
	  pNewSentenceF2->Display( m_pTextF->Vocabulary() ); cout<< " # "; 
	  pNewSentenceE2->Display( m_pTextE->Vocabulary() ); cout << endl;
	  */


		// 3-way split
      // first segment f
      for ( int j = pOrgSentenceF->MinIndex(); j < SplitPoint.m_j; j++ ) {
		(*pNewSentenceF1)[ j ] = (*pOrgSentenceF)[ j ];
      }
      // first segment e
      for ( int i = pOrgSentenceE->MinIndex(); i < SplitPoint.m_i; i++ ) {
		(*pNewSentenceE1)[ i ] = (*pOrgSentenceE)[ i ];
      }

      // second segment f, just the split word
	  (*pNewSentenceF2)[ 0 ] = (*pOrgSentenceF)[ SplitPoint.m_j ];
      // second segment e
	  (*pNewSentenceE2)[ 0 ] = (*pOrgSentenceE)[ SplitPoint.m_i ];

      // third segment f
	  for ( int j = SplitPoint.m_j + 1; j <= pOrgSentenceF->MaxIndex(); j++ ) {
		(*pNewSentenceF3)[ j - SplitPoint.m_j ] = (*pOrgSentenceF)[ j ];
      }
      // third segment e
      for ( int i = SplitPoint.m_i + 1; i <= pOrgSentenceE->MaxIndex(); i++ ) {
		(*pNewSentenceE3)[ i - SplitPoint.m_i ] = (*pOrgSentenceE)[ i ];
      }

      m_pTextF->PutSentence( SplitPoint.m_Sentence, *pNewSentenceF1 );
      m_pTextE->PutSentence( SplitPoint.m_Sentence, *pNewSentenceE1 );

      m_pTextF->InsertSentence( SplitPoint.m_Sentence + 1, pNewSentenceF2 );
      m_pTextE->InsertSentence( SplitPoint.m_Sentence + 1, pNewSentenceE2 );

      m_pTextF->InsertSentence( SplitPoint.m_Sentence + 2, pNewSentenceF3 );
      m_pTextE->InsertSentence( SplitPoint.m_Sentence + 2, pNewSentenceE3 );

/*
	  cout << "into: ";
      pNewSentenceF1->Display( m_pTextF->Vocabulary() ); cout << " # ";
      pNewSentenceE1->Display( m_pTextE->Vocabulary() ); cout << endl;
      cout << "and:  ";
	  pNewSentenceF2->Display( m_pTextF->Vocabulary() ); cout<< " # "; 
	  pNewSentenceE2->Display( m_pTextE->Vocabulary() ); cout << endl;
      cout << "and:  ";
	  pNewSentenceF3->Display( m_pTextF->Vocabulary() ); cout<< " # "; 
	  pNewSentenceE3->Display( m_pTextE->Vocabulary() ); cout << endl;
*/	  
      
      SplitsDone++;
    }
  }

Trace( 0, "Number of splits in iteration " << m_CurrentIteration 
		 << " actually performed: " << SplitsDone );
  cout << "Number of splits in iteration " << m_CurrentIteration 
		 << " actually performed: " << SplitsDone << endl;
  //m_pTextF->Display();
  //m_pTextE->Display();

}
