#include <config.h>



// CTM-IBM1-Bi: Translation Model IBM1, Mixture Densities without alignments
//              training in both directions


#ifndef __tm_IBM1_bi__
#define __tm_IBM1_bi__


#include <Types.hh>
#include <Text.hh>
#include <SortedArray.hh>
#include <NumericalArray.hh>
#include <TM.hh>
#include <Segmenter.hh>


#define MaxSenLen 200

class CTM_IBM1_Bi : public CTranslationModel
{
public:
  CTM_IBM1_Bi();
  ~CTM_IBM1_Bi();

  CLexiconForTraining * LexiconFE();
  CLexiconForTraining * LexiconEF();

  void SetSentencePrune( int SentencePrune );
  void SetIncMaxSentenceLength( int IncMaxSentenceLength );
  void SetFirstPruningIteration( int FirstPruningIteration );

  // settings for sentence splitting
  void SetSplitWords( const CString &SplitWords );
  bool LoadSplitWords( const CString &FileSplitWords );
  void SetFirstSplitIteration( int FirstSplitIteration );
  void SetNumberOfSplits( int NumberOfSplits );  // max splits per iteration
  void SetSplitThreshold( double SplitThreshold );
  void SetRequiredSplitPoints( int RequiredSplitPoints );  // 1 side only or both sides
  void SetInterpolationFactor( double InterpolationFactor ); // between f->e and e->f
  void SetLexiconUpdate( int LexiconFixed ); // 0 = normal update of lexicon, 1 = keep fixed

  bool Train();
  double Perplexity();
  double PerplexityForMaximum();

  virtual bool TestProbabilities();

  bool StoreBiText( const char *pFileName );

protected:
  bool NextIteration();
  //void RenormalizeT();

  void DisplayBiText();
  void DisplayBadSentences();
  void RemoveBadSentences();

  void CheckBoundaries( const CSentence *pSentence );

  void FindBestSplit( int SentenceNumber );
  void MakeSplits();


protected:
  CLexiconForTraining * m_pLexiconFE;
  CLexiconForTraining * m_pLexiconEF;
  
  //CText *m_pTextE, *m_pTextF;
  CSentence *m_pSentenceE, *m_pSentenceF;

  double m_ProbsFE[ MaxSenLen + 1 ][ MaxSenLen + 1 ];
  double m_ProbsEF[ MaxSenLen + 1 ][ MaxSenLen + 1 ];

  //double m_LambdaSmooth;

  //CArray<double> m_pLambda;
  //CDblMatrix *m_pT1;

  // perplexity results from different interations
  // AlignmentPerplexity is constant for IBM1, but should be displayed
  CArray<double> m_LexiconPerplexityFE;
  CArray<double> m_AlignPerplexityFE;
  CArray<double> m_TotalPerplexityFE;

  CArray<double> m_LexiconPerplexityEF;
  CArray<double> m_AlignPerplexityEF;
  CArray<double> m_TotalPerplexityEF;

  // perplexity of sentence pairs
  CArray<float> *m_pLexiconPerplexityEF;
  CArray<float> *m_pTotalPerplexityEF;
  CArray<float> *m_pLexiconPerplexityFF;
  CArray<float> *m_pTotalPerplexityFE;


  int m_MaxSentenceLength;
  int m_IncMaxSentenceLength;

  int m_SentencePrune;
  int m_FirstPruningIteration;
  CSortedArray<double> m_BadSentenceValues;
  CArray<int> m_BadSentenceNumbers;

  // to split sentences
  int m_FirstSplitIteration;
  CArray< CString > *m_pBoundaryWords;
  // array over wordindices 
  // 1 = word |, i.e word stays with left segment, e.g. punctuation, ) } ] ...
  // 2 = | word, i.e. word stays with right segment, e.g. and, but, because, ( { [ ...
  // 3 = | word |, i.e. word forms own segment, e.g. punctuation
  CNumericalArray< int > *m_pIsBoundaryWord;
  
  CSegmenter *m_pSegmenter;
  // number of splits in each iteration
  // default: 0
  int m_NumberOfSplits;
  // gain in perplexity required to split sentence pair
  // default: 0.0
  double m_SplitThreshold;
  // number of splitpoints required to trigger calculation of perplexity gain
  // 2: split word in both sentences, 1: in at least one sentene, 0: none required, 
  // i.e. all positions in the sentence pair near the diagonal are considered
  // default: 2
  int m_RequiredSplitPoints;

  // interpolation factor f->e and e->f
  double m_Lambda;

  // when only doing sentence splitting with existing lexicon
  // then we don't update the lexicon after each iteration
  int m_LexiconFixed;
};


inline CLexiconForTraining *
CTM_IBM1_Bi::LexiconFE()
{
  return m_pLexiconFE;
}

inline CLexiconForTraining *
CTM_IBM1_Bi::LexiconEF()
{
  return m_pLexiconEF;
}


#endif


