#include <config.h>



// CTM-IBM1-Bi: Translation Model IBM1, Mixture Densities without alignments
//              training in both directions
// aligns in both directions F->E and E->F
// used empty word

#ifndef __TM_IBM1_BI_EW__
#define __TM_IBM1_BI_EW__


#include <Types.hh>
#include <Text.hh>
#include <SortedArray.hh>
#include <NumericalArray.hh>
#include <TM.hh>
#include <Segmenter.hh>


#define MaxSenLen 200

class CTM_IBM1_Bi_EW : public CTranslationModel
{
public:
  CTM_IBM1_Bi_EW();
  ~CTM_IBM1_Bi_EW();

  CLexiconForTraining * LexiconFE();
  CLexiconForTraining * LexiconEF();

  void SetUseEmptyWord( bool UseEmptyWord );
  void SetSentencePrune( int SentencePrune );
  void SetInitialSentenceLength( int InitialSentenceLength );
  void SetIncrementSentenceLength( int IncrementSentenceLength );
  void SetFirstPruningIteration( int FirstPruningIteration );

  // settings for sentence splitting
  void SetSplitWords( const CString &SplitWords );
  bool LoadSplitWords( const CString &FileSplitWords );
  void SetFirstSplitIteration( int FirstSplitIteration );
  void SetNumberOfSplits( int NumberOfSplits );  // max splits per iteration
  void SetSplitThreshold( double SplitThreshold );
  void SetRequiredSplitPoints( int RequiredSplitPoints );  // 1 side only or both sides
  void SetSplitCorpusFileName( const CString &SplitCorpusFileNameSource, const CString &SplitCorpusFileNameTarget );
  void SetInterpolationFactor( double InterpolationFactor ); // between f->e and e->f
  void SetLexiconUpdate( int LexiconFixed ); // 0 = normal update of lexicon, 1 = keep fixed

  bool Train();

  double Perplexity( CText *pTextF, CText *pTextE,
					 CLexicon *pLexicon,
					 CArray<float> &LexiconPerplexity,
					 CArray<float> &AlignPerplexity,
					 CArray<float> &TotalPerplexity );

  double PerplexityForMaximum();

  virtual bool TestProbabilities();

  bool StoreBiText( const char *pFileName );

protected:
  bool NextIteration();

  void DisplayBiText();
  void DisplayBadSentences();
  void RemoveBadSentences();

  void CheckBoundaries( const CSentence *pSentence );

  void FillProbMatrices();
  void RenormalizeProbs();

  void AlignFE( int s, double & PTransSumFE, double &PAlignSumFE );
  void AlignEF( int s, double & PTransSumEF, double &PAlignSumEF );

  void FindBestSplit( int SentenceNumber );
  void MakeSplits();
  void MakeSplitsOld();

  void DisplayEmptyWordProbs();
  void DisplayProbMatrices();




protected:
  CLexiconForTraining * m_pLexiconFE;
  CLexiconForTraining * m_pLexiconEF;
  
  //CText *m_pTextE, *m_pTextF;
  CSentence *m_pSentenceE, *m_pSentenceF;

  double m_ProbsFE[ MaxSenLen + 1 ][ MaxSenLen + 1 ];
  double m_ProbsEF[ MaxSenLen + 1 ][ MaxSenLen + 1 ];

  // use empty word in alignment, i.e. p( f | e_0 ) for F->E alignment
  // and p( e | f_0 ) for E->F alignment
  bool m_UseEmptyWord;

  //double m_LambdaSmooth;

  //CArray<double> m_pLambda;
  //CDblMatrix *m_pT1;

  // perplexity results from different interations
  // AlignmentPerplexity is constant for IBM1, but should be displayed
  CArray<float> m_LexiconPerplexityFE;
  CArray<float> m_AlignPerplexityFE;
  CArray<float> m_TotalPerplexityFE;

  CArray<float> m_LexiconPerplexityEF;
  CArray<float> m_AlignPerplexityEF;
  CArray<float> m_TotalPerplexityEF;

  CArray<float> m_LexiconPerplexityFE_Test;
  CArray<float> m_AlignPerplexityFE_Test;
  CArray<float> m_TotalPerplexityFE_Test;

  CArray<float> m_LexiconPerplexityEF_Test;
  CArray<float> m_AlignPerplexityEF_Test;
  CArray<float> m_TotalPerplexityEF_Test;

  // perplexity of sentence pairs
  CArray<float> *m_pLexiconPerplexityEF;
  CArray<float> *m_pTotalPerplexityEF;
  CArray<float> *m_pLexiconPerplexityFF;
  CArray<float> *m_pTotalPerplexityFE;


  int m_MaxSentenceLength;
  int m_InitialSentenceLength;
  int m_IncrementSentenceLength;

  int m_SentencePrune;
  int m_FirstPruningIteration;
  CSortedArray<double> m_BadSentenceValues;
  CArray<int> m_BadSentenceNumbers;

  // to split sentences
  int m_FirstSplitIteration;

  CArray< CString > *m_pBoundaryWords;

  // array over wordindices 
  // 1 = word |, i.e word stays with left segment, e.g. punctuation, ) } ] ...
  // 2 = | word, i.e. word stays with right segment, e.g. and, but, because, ( { [ ...
  // 3 = | word |, i.e. word forms own segment, e.g. punctuation
  CNumericalArray< int > *m_pIsBoundaryWord;
  
  CSegmenter *m_pSegmenter;

  // number of splits in each iteration
  // default: 0
  int m_NumberOfSplits;

  // gain in perplexity required to split sentence pair
  // default: 0.0
  double m_SplitThreshold;

  // number of splitpoints required to trigger calculation of perplexity gain
  // 2: split word in both sentences, 1: in at least one sentene, 0: none required, 
  // i.e. all positions in the sentence pair near the diagonal are considered
  // default: 2
  int m_RequiredSplitPoints;

  // prefix for file name to store split corpus in each iteration
  CString m_SplitCorpusFileNameSource;
  CString m_SplitCorpusFileNameTarget;

  // interpolation factor f->e and e->f
  double m_Lambda;

  // when only doing sentence splitting with existing lexicon
  // then we don't update the lexicon after each iteration
  int m_LexiconFixed;
};


inline CLexiconForTraining *
CTM_IBM1_Bi_EW::LexiconFE()
{
  return m_pLexiconFE;
}

inline CLexiconForTraining *
CTM_IBM1_Bi_EW::LexiconEF()
{
  return m_pLexiconEF;
}

inline void
CTM_IBM1_Bi_EW::SetUseEmptyWord( bool UseEmptyWord )
{
  m_UseEmptyWord = UseEmptyWord;
}


#endif


