
#include <MapWordToTag.hh>


using namespace std;


CMapWordToTag::CMapWordToTag() {


}

CMapWordToTag::~CMapWordToTag() {


}

/**
   @brief Load vocabulary objects


*/
bool CMapWordToTag::Init(CVocabulary *wVoc, CVocabulary *tVoc) {
  wordVoc = wVoc;
  tagVoc = tVoc;

  return true;
}


/**
   @brief Load mapping word -> tag from input stream

   The format for a mapping file is one mapping per line. First the word and then, seperated by a space, the tag. 


*/

bool CMapWordToTag::Load(std::istream *InStream) {

  
  static const int BUFLEN = 256;
  char str[BUFLEN];
  CWordIndex wordIndex,tagIndex;
  CString *cStr;
  //error: first load vocabulary
  if((wordVoc == NULL) || (tagVoc == NULL)) {
    cerr << "First load Vocabulary by using Init()" <<endl;
    return false;
  }

  //error reading stream
  if ( !InStream->good() ) {
      cerr <<  "CVocabulary::Load(): Could not read from ifstream" << endl << flush;
      return false;
  }

  // skip comment lines
  *InStream >> str;
    while ( (str[ 0 ] == '#') && !InStream->eof() ) {
    InStream->getline( str, BUFLEN );
    *InStream >> str;
  }

   if ( InStream->eof() ) {
     cerr << "No vocabulary item read, empty file" << endl;
    return false;
  }
   int wc = 0;
   int tc = 0;
   do{
     
     //Get wordIndex or add to Vocabulary
     cStr = new CString(str);
     wordIndex = wordVoc->GetIndex(*cStr);
     //cout << "WordIndex: " << (int) wordIndex << " of word "; 
     //cStr->ToStream(cout);
     //cout << endl;
     if(wordIndex == wordVoc->UNK) {
       //cout<< " Add " << str  << " to WordVoc" << endl;
       wordIndex = wordVoc->AddWord(*cStr);
       wc++;
     }
     //cout << "Voc Size: " << wordVoc->Size() << "";
     //cout << str << "has Index " << wordVoc->GetIndex(*cStr) << " added to " << wordIndex << endl; 
     delete cStr;
     
     //Get Tag
     *InStream >> str;
     
     //Get Tag index or add to Vocabulary
     cStr = new CString(str);
     tagIndex = tagVoc->GetIndex(*cStr);     
     if(tagIndex == tagVoc->UNK) {
       tagIndex = tagVoc->AddWord(*cStr);
       //cout<< " Add " << str  << " to TagVoc" << endl;
       tc++;
     }
     delete cStr;
     
     //Store mapping
     tags[wordIndex]= tagIndex;

     //Load beginning of next line
     *InStream >> str;
     
     // skip comment lines
     while((str[ 0 ] == '#') && !InStream->eof() ) {
       InStream->getline( str, BUFLEN );
       *InStream >> str;
     }

   }while(!InStream->eof());

   cout << " Number of Words added: " << wc << endl;
   cout << " Number of Tags added: " << tc << endl;

   return true;
}


/**
   @brief Get mapping word -> tag from input stream

   if WordIndex not known return 0
*/
CWordIndex CMapWordToTag::getTag(CWordIndex WordIndex) {

  //error: first load vocabulary
  if((wordVoc == NULL) || (tagVoc == NULL)) {
    cerr << "First load Vocabulary by using Init()" <<endl;
    return false;
  }

  return tags[WordIndex];

}


