#include <string.h>
#include <stdlib.h>
// CText
// Base class for text handling
//


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


/********************************************************************/
/* CText implementation */


#include <fstream>
#include <string>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sstream>

//#define NO_TRACE 1
#include <Trace.hh>

#include <Text.hh>
#include <utils.hh>


static char rcsid_text_cc[] = "$Id:";


#define MAXstrlen 10000


CWordInfo::CWordInfo()
{
  m_Score = 0.0;
}



CWordInfo::CWordInfo( const CWordInfo * pWordInfo)
{
  m_SourceWords = pWordInfo->m_SourceWords;
  m_TargetWords = pWordInfo->m_TargetWords;

  m_Score = pWordInfo->m_Score;
}



CString
CWordInfo::SourceString() const
{
  CString Total;

  if ( m_SourceWords.Size() > 0 ) {
    int i = m_SourceWords.MinIndex();
    Total += m_SourceWords.Word( i );
    i++;
    while ( i <= m_SourceWords.MaxIndex() ) {
      Total += " ";
      Total += m_SourceWords.Word( i );
      i++;
    }
  }

  return Total;
}


CString
CWordInfo::TargetString() const
{
  CString Total;

  if ( m_TargetWords.Size() > 0 ) {
    int i = m_TargetWords.MinIndex();
    Total += m_TargetWords.Word( i );
    i++;
    while ( i <= m_TargetWords.MaxIndex() ) {
      Total += " ";
      Total += m_TargetWords.Word( i );
      i++;
    }
  }

  return Total;
}


int
CWordInfo::GetSourceWords( CArray< CString > &Words ) const
{
  Words.Clear();
  int j = 1;

  if ( m_SourceWords.Size() > 0 ) {
    int i = m_SourceWords.MinIndex();
    Words[ j ] = m_SourceWords.Word( i );
    i++;
    j++;
    while ( i <= m_SourceWords.MaxIndex() ) {
      Words[ j ] =  m_SourceWords.Word( i );
      i++;
      j++;
    }
  }

  return j;
}


int
CWordInfo::GetTargetWords( CArray< CString > &Words ) const
{
  Words.Clear();
  int j = 1;

  if ( m_TargetWords.Size() > 0 ) {
    int i = m_TargetWords.MinIndex();
    Words[ j ] = m_TargetWords.Word( i );
    i++;
    j++;
    while ( i <= m_TargetWords.MaxIndex() ) {
      Words[ j ] =  m_TargetWords.Word( i );
      i++;
      j++;
    }
  }

  return j;
}


std::istream &
operator >> ( std::istream &Stream, CWordInfo &Info )
{
  Trace( 4, "CWordInfo operator >> " );

  Trace( 4, "read SourceWords" );
  Stream >> Info.SourceWords();
  Trace( 4, "end read SourceWords" );

  // is there also a second part
  char c;
  Stream >> c;
  Trace( 4, "next token is " << c );
  if ( c == '#' ) {
    Trace( 4, "read TargetWords" );
    Stream >> Info.TargetWords();
    Trace( 4, "end read TargetWords" );
    // read # or closing bracket
    Stream >> c;
    Trace( 4, "next token is " << c );
    // is there also a third part, i.e. a score
    if ( c == '#' ) {
      Trace( 4, "read Score" );
      Stream >> Info.Score();
      Trace( 4, "end read Score; Score = " << Info.Score() );
      // read closing bracket
      Stream >> c;
      Trace( 4, "next token is " << c );
    } else if ( c == '}' ) {
      // no score, do nothing
      ;
    }
  } else if ( c == '}' ) {
    // no target words, do nothing
    ;
  } else {
    // first character of next word
    Stream.putback( c );
  }

  Trace( 4, "End CWordInfo operator >>" );

  return Stream;
}



std::ostream &
operator << ( std::ostream &Stream, const CWordInfo &Info )
{
  //std::cout << "\nCWordInfo::operator << " << std::endl;

  if ( Info.HasSourceString() ) {
    // use const-cast-away to avoid compiler warnings
    Stream << "{ " << ((CWordInfo &)Info).SourceWords();

    if ( Info.HasTargetString() ) {
      Stream << " # " << ((CWordInfo &)Info).TargetWords();

      if ( Info.HasScore() ) {
	Stream << " # " << ((CWordInfo &)Info).Score();
      }
    }
    Stream << " }";
  }

  //std::cout << "\nEnd CWordInfo::operator <<" << std::endl;
  return Stream;
}



void
CWordInfo::SwapSourceTarget()
{
  Trace( 3, "CWordInfo::SwapSourceTarget()" );

  // swaps source and target annotation only on the top level
  CSentence Tmp;
  Tmp = m_SourceWords;
  //m_SourceWords = Tmp;
  m_SourceWords = m_TargetWords;
  m_TargetWords = Tmp;
}




//--------------------------------------------------------------------
// CSentence
//--------------------------------------------------------------------

CSentence::CSentence()
{
  m_InitToZero = true;
  m_pWordInfos = NULL;
  m_pWords = NULL;
}



CSentence::CSentence( const CSentence &S )
{
  Trace( 4, "CSentence copy constructor called with " << S );

  Init();

  m_InitToZero = true;
  m_pWordInfos = NULL;
  m_pWords = NULL;

  *this = S;

}



CSentence &
CSentence::operator = ( const CSentence &Sentence )
{
  Trace( 3, "CSentence::operator = " );

  //Array.Display();

  if ( this == &Sentence ) 
    // ignore copy onto itselv
    return *this;

  if ( !Sentence.IsValid() ) {
    // if other sentence is invalid, make this sentence invalid
    Invalidate();
    return *this;
  }
   
  // delete elements
  Clear();

  // set boundaries to avoid copy operations when overflow
  if ( !AssureBoundaries( Sentence.MinIndex(), Sentence.MaxIndex() ) ) 
    // return if not sufficient memory
    return *this;

  // copy the elements
  for ( int i = Sentence.MinIndex(); i <= Sentence.MaxIndex(); i++ )
    SetAt( i, Sentence[i] );

  // array for word indices
  if ( Sentence.m_pWordInfos ) {
    const CWordInfo *pWordInfo;
    CWordInfo *pNewWordInfo;
    m_pWordInfos = new CArray< CWordInfo * >( *(Sentence.m_pWordInfos) ); // ???
    //m_pWordInfos = new CArray< CWordInfo * >;
    for ( int i = Sentence.MinIndex(); i <= Sentence.MaxIndex(); i++ ) {
      if( Sentence.m_pWordInfos->IsValidIndex( i ) ) {
        if ( Sentence.m_pWordInfos->GetAt( i ) ) {
          pWordInfo = Sentence.m_pWordInfos->GetAt( i );
          pNewWordInfo = new CWordInfo( pWordInfo );
          m_pWordInfos->SetAt( i, pNewWordInfo );
        }
      }
    }
  }
  

  // array for words as strings
  if ( Sentence.m_pWords ) {
    Trace( 5, "copy words stored as strings " 
       << Sentence.m_pWords->MinIndex() << " " << Sentence.m_pWords->MaxIndex() );
    m_pWords = new CArray< CString >;
    for ( int i = Sentence.m_pWords->MinIndex(); 
          i <= Sentence.m_pWords->MaxIndex(); i++ ) {
      Trace( 5, i << " " << (*m_pWords)[ i ] );
      (*m_pWords)[ i ] = (*(Sentence.m_pWords))[ i ];
    }
  }


  return *this;
}



CSentence::CSentence( const CString &String, CVocabulary *pVocabulary )
{
  //cout << "" << endl;
  Init();

  m_InitToZero = true;
  m_pWordInfos = NULL;
  m_pWords = NULL;

  this->ReadFromBuffer( ( const char *)String, pVocabulary );
}



CSentence::~CSentence()
{
  // free memory
  Clear();
}



void
CSentence::Append( CWordIndex WordIndex )
{
  CArray< CWordIndex >::Append( WordIndex );
}
  



void
CSentence::Append( CSentence &Sentence )
{
  Trace( 3, "CSentence::Append( CSentence )" );
  cout << "CSentence::Append( CSentence ) " << endl;

  //Array.Display();

  if ( this == &Sentence ) {
    // we currently don't allow a sentence to be appended to itself
    cout << "cannot append sentence to itself" << endl;
    assert( 0 );
    return;
  }

  if ( !Sentence.IsValid() ) {
    // if sentence is not valid, do nothing
    return;
  }

  // current MaxIndex in sentence
  int m = MaxIndex();

  // if we have an empty sentence than we are going to start at index 1
  if ( m_MinIndex > m_MaxIndex ) {
    m_MinIndex = 1;
    m_MaxIndex = 0;
}

  // set boundaries
  if ( !AssureBoundaries( MinIndex(), MaxIndex() + Sentence.Size() ) ) {
    return;
  }

  // copy the words
  CWordInfo *pWordInfo;
  CWordInfo *pNewWordInfo;

  for ( int i = Sentence.MinIndex(); i <= Sentence.MaxIndex(); i++ ) {

    // go to next position
    m++;
    // copy the word index
    cout << "copy from " << i << " to " << m << endl;
    this->SetAt( m, Sentence[ i ] );
  
    // copy the word info
    if ( Sentence.m_pWordInfos ) {
      cout << "copy also word infos" << endl;
      if ( !m_pWordInfos )
        m_pWordInfos = new CArray< CWordInfo * >( *(Sentence.m_pWordInfos) );

      if ( Sentence.m_pWordInfos->IsValidIndex( i ) ) {
        cout << "copy word info from " << i << " to " << m << endl;
        pWordInfo = Sentence.m_pWordInfos->GetAt( i );
        pNewWordInfo = new CWordInfo( pWordInfo );
        m_pWordInfos->SetAt( m, pNewWordInfo );
      }

    }
    
    // copy the word string
    if ( Sentence.m_pWords ) {
      cout << "copy also words" << endl;

      if ( !m_pWords )
        m_pWords = new CArray< CString >;
      
      if ( Sentence.m_pWords->IsValidIndex( i ) ) {
        cout << "copy word from " << i << " to " << m << endl;
        (*m_pWords)[ m ] = (*(Sentence.m_pWords))[ i ];
      }
    }
  }

}



void
CSentence::Clear()
{
  Trace( 4, "CSentence::Clear() - " << *this );

  // free memory
  delete m_pWords;
  m_pWords = NULL;

  if ( m_pWordInfos ) {
    int i;
    for ( i = m_pWordInfos->MinIndex(); i <= m_pWordInfos->MaxIndex(); i++ ) {
      if( m_pWordInfos->IsValidIndex( i ) )
        delete m_pWordInfos->GetAt( i );
    }
  }
  delete m_pWordInfos;
  m_pWordInfos = NULL;
  
  RemoveAll();
  Trace( 4, "CSentence::Clear() ended" );
}



bool
CSentence::SetWordInfo( CWordInfo *WordInfo, int Index )
{
  if ( !m_pWordInfos ) {
    //std::cout << "Create new Array for WordInfos" << std::endl;
    m_pWordInfos = new CArray< CWordInfo * >;
    if (!m_pWordInfos)
      return false;
    m_pWordInfos->InitToZero();
  }
  m_pWordInfos->SetAt( Index, WordInfo );
  
  return true;
}



int
CSentence::ReadFromBuffer( const char *pBuffer, CVocabulary *pVocabulary )
{
  //std::cout << "CSentence::ReadFromBuffer: " << pBuffer << std::endl;

  std::istringstream Stream( pBuffer );
  char str[ 10000 ];
  int Number = 0;
  CWordInfo *pWordInfo;

  while ( !Stream.eof() ) {
    str[0] = '\0';
    Stream >> str;
    if ( strlen( str ) > 0 ) {
      // if word info starts, indicated by {
      if ( strcmp( str, "{" ) == 0 ) {
		pWordInfo = new CWordInfo;
		Stream >> *pWordInfo;
		Trace(3, "Loading wordInfo: " << *pWordInfo << std::endl);
		
		if ( !SetWordInfo( pWordInfo, Number ) )
		  return 0;

      } else {
		Number++;
		SetWord( str, Number, pVocabulary );
		/*
		  Index = pVocabulary->GetIndex( str );
		  std::cout << str << " - " << "Index = " << Index << std::endl;
		  // store index
		  SetAt( Number, Index );
		  //std::cout << str << " - ";
		  if ( Index == pVocabulary->UNK ) {
		  // store string in WordInfo
		  Trace( 0, "Unknown word " << str );
		  }
		*/
      }
      //fp >> ws;
      // skip whitespaces
      while ( (Stream.peek() == 32) || (Stream.peek() == 9) ) {
		Stream.get();
      }
      // output newline
      if ( Stream.peek() == 10 ) {
		Stream.get();
		//std::cout << std::endl;
      }
    }
	
    // peek into the stream to see if EOF is coming next
    Stream.peek();
  }
  
  //std::cout << std::endl;
  //std::cout << Number << " | " << m_MinIndex << " | " << m_MaxIndex << std::endl;
  
  return Number;
}



std::istream &
operator >> ( std::istream &Stream, CSentence &Sentence )
{
  Trace( 4, "CSentence operator >>" );

  char str[ 100 ];
  int Number = 0;
  CWordInfo *pWordInfo;

  while ( !Stream.eof() ) {
    Stream >> str;
    // if we are reading part of a word annotatiion
    // we stop when encountering # or }
    if ( (strcmp( str, "#" ) == 0) || (strcmp( str, "}" ) == 0) ) {

      if ( (strcmp( str, "#" ) == 0) ) {
	Trace( 4, "# found - end reading sentence" );
	Stream.putback( '#' );
      }
      if ( (strcmp( str, "}" ) == 0) ) {
	Trace( 4, "} found - end reading sentence" );
	Stream.putback( '}' );
      }

      Trace( 4, "End CSentence operator >>" );
      return Stream;
    }

    if ( strlen( str ) > 0 ) {
      // if word info starts, indicated by {
      if ( strcmp( str, "{" ) == 0 ) {
	pWordInfo = new CWordInfo;
	Stream >> *pWordInfo;
	//std::cout << "WordInfo: " << *pWordInfo << std::endl;
	if ( !Sentence.SetWordInfo( pWordInfo, Number ) )
	  return Stream;
      } else {
	Number++;
	Sentence.SetWord( str, Number );
	/*
	Index = pVocabulary->GetIndex( str );
	std::cout << str << " - " << "Index = " << Index << std::endl;
	// store index
	SetAt( Number, Index );
	//std::cout << str << " - ";
	if ( Index == pVocabulary->UNK ) {
	  // store string in WordInfo
	  Trace( 0, "Unknown word " << str );
	}
	*/
      }
      //fp >> ws;
      // skip whitespaces
      while ( (Stream.peek() == 32) || (Stream.peek() == 9) ) {
	Stream.get();
      }
      // output newline
      if ( Stream.peek() == 10 ) {
	Stream.get();
	//std::cout << std::endl;
      }
    }

    // peek into the stream to see if EOF is coming next
    Stream.peek();
  }

  Trace( 4, "End CSentence operator >>" );

  return Stream;
}



std::ostream &
operator << ( std::ostream &OutStream, const CSentence &Sentence )
{
  //Trace( 0, "\nCSentence::operator << " );

  CString Word;
  // const-cast-away to avoid compiler warning when calling CSentence::WordInfo( int )
  CSentence &SentenceRef = (CSentence &)Sentence;

  if ( Sentence.Size() > 0 ) {

    int i = Sentence.MinIndex();
    OutStream << Sentence.Word( i );
    if ( SentenceRef.WordInfo( i ) )
      OutStream << " " << *(SentenceRef.WordInfo( i ));

    i++;

    while ( i <= Sentence.MaxIndex() ) {
      OutStream << " " << Sentence.Word( i );
      if ( SentenceRef.WordInfo( i ) )
	OutStream << " " << *(SentenceRef.WordInfo( i ));

      i++;
    }

  }

  //Trace( 0, "\nEnd CSentence::operator << " );
  return OutStream;
}




const CString &
CSentence::Word( int Index, const CVocabulary *pVocabulary) const
{
  static CString EmptyString;

  if ( m_pWords && m_pWords->IsValidIndex( Index ) ) {
    if ( m_pWords->GetAt( Index ) != CString( "" ) ) {
      return m_pWords->GetAt( Index );
    }
  }

  if ( pVocabulary ) {
    return pVocabulary->GetWord( m_pElements[ Index ] );
  }

  return EmptyString;
  
}



void
CSentence::SetWord( const char *pWord, int Index, CVocabulary *pVocabulary )
{
  // ??? delete memory of old word ?

  Trace( 5, "CSentence::SetWord " << pWord << " " << Index );

  assert ( Index >= 1 );
  assert ( m_MaxIndex < 0 ? Index <= 1 : Index <= m_MaxIndex + 1 );

  CWordIndex WordIndex;
  if ( pVocabulary ) {
    WordIndex = pVocabulary->GetIndex( pWord );
	if ( WordIndex == CVocabulary::UNK && strcmp(pWord,pVocabulary->GetWord(CVocabulary::UNK)) != 0 ) {
	  WordIndex = pVocabulary->AddWord( CString( pWord ) );
	}
    //Trace( 0, pWord << " - " << "Index = " << WordIndex );
  } else {
    //Trace( 0, pWord << " - " << "Index = " << CVocabulary::UNK );
    WordIndex = CVocabulary::UNK;
  }

  // store index
  SetAt( Index, WordIndex );
  //std::cout << str << " - ";
  if ( WordIndex == CVocabulary::UNK ) {
    // store string in WordInfo
    CString Word( pWord );
    //Trace( 0, "Store unknown word as string " << Word );

    if ( !m_pWords ) {
      m_pWords = new CArray< CString >;
    }

    m_pWords->SetAt( Index , Word );
  }

}



CString
CSentence::AsString( const CVocabulary *pVocabulary ) const
{
  Trace( 3, "CSentence::AsString" );

  //Write( std::cout, pVocabulary, true ); std::cout << std::endl;

  if ( Size() == 0 ) {
    CString String( "" );
    return String;
  }

  std::ostringstream Stream;
  Write( Stream, pVocabulary, true );
  //char *pBuffer = (Stream.str()).c_str();

  // problem with terminating 0-byte -> set one
  // ARV - seems like it would crash the program!
  //pBuffer[ Stream.str().size() ] = 0;
  //std::cout << Stream.pcount() << " " << strlen( pBuffer ) << " " << pBuffer << std::std::endl;
  // ARV this could cause problems due to the end character
  CString String( Stream.str().c_str() );
  Trace( 3, "String = " << String );

  return String;

}



void
CSentence::Display( const CVocabulary *pVocabulary, bool WithWordInfo ) const
{
  Write( std::cout, pVocabulary, WithWordInfo );
  return;

}


void
CSentence::Write( std::ostream &OutStream,
                  const CVocabulary *pVocabulary,
                  bool WithWordInfo) const
{
  //Trace( 0, "CSentence::Write" );

  CString Word;

  // write into the stream to guarantee, that the stream has a buffer
  OutStream << "";

  if ( Size() > 0 ) {
    int i = m_MinIndex;
    //OutStream << i << " ";

    TRACE( 3, i << " " << m_pElements[ i ] << " " << pVocabulary->GetWord( m_pElements[ i ] ) );

    // if word is stored as wordindex != UNK we convert the wordindex into a string
    // otherwise we write directly the ascii-string
    if ( pVocabulary != NULL && m_pElements[ i ] != pVocabulary->UNK ) {
      //OutStream << "v ";
      OutStream << pVocabulary->GetWord( m_pElements[ i ] );
    } else {
      //OutStream << "s ";
      if ( m_pWords )
        OutStream << m_pWords->GetAt( i );
      else {
        Trace( 1, "UNK" );
        OutStream << "UNK";
      }
    }

    if ( WithWordInfo ) {
      if ( m_pWordInfos && (m_pWordInfos->IsValidIndex( i ) ) && 
           (*m_pWordInfos)[ i ] ) {
        OutStream << " " << *((*m_pWordInfos)[ i ]);
      }
    }
    i++;

    while ( i <= m_MaxIndex ) {
      //OutStream<< " " << i << " ";

      TRACE( 3, i << " " << m_pElements[ i ] << " " << pVocabulary->GetWord( m_pElements[ i ] ) );

      if ( pVocabulary != NULL && m_pElements[ i ] != pVocabulary->UNK ) {
        //OutStream << "v ";
        OutStream << " " << pVocabulary->GetWord( m_pElements[ i ] );
      } else {
        //OutStream << "s ";
        if ( m_pWords )
          OutStream << " " << m_pWords->GetAt( i );
      }

      if ( WithWordInfo ) {
        if ( m_pWordInfos && (m_pWordInfos->IsValidIndex( i ) ) && 
             (*m_pWordInfos)[ i ] ) {
          OutStream << " " << *((*m_pWordInfos)[ i ]);
        }
      }
      i++;
    }
  }

}



void
CSentence::GetSourceString( CString &SourceString,
			    const CVocabulary *pVocabulary ) const
{
  //Trace( 0, "CSentence::GetSourceString" );

  CWordInfo *pWordInfo;

  if ( Size() > 0 ) {
    int i = m_MinIndex;
    //OutStream << i << " ";

    //SourceString += Word( i );
    //SourceString += CString( " " );
    pWordInfo = WordInfo( i );
    if ( pWordInfo ) {
      CSentence &Sentence = pWordInfo->SourceWords();
      Sentence.GetSourceString( SourceString );
    } else {
      SourceString += Word( i, pVocabulary );
    }
    //Trace( 1, SourceString );

    i++;

    while ( i <= m_MaxIndex ) {
      SourceString += CString( " " );

      //SourceString += Word( i );
      //SourceString += CString( " " );

      pWordInfo = WordInfo( i );
      if ( pWordInfo ) {
	CSentence &Sentence = pWordInfo->SourceWords();
	Sentence.GetSourceString( SourceString );
      } else {
	SourceString += Word( i, pVocabulary );
      }
      //Trace( 1, SourceString );
      
      i++;
    }
  }

}



void
CSentence::GetTargetString( CString &TargetString,
			    const CVocabulary *pVocabulary ) const
{
  //Trace( 0, "CSentence::GetTargetString" );

  CWordInfo *pWordInfo;

  if ( Size() > 0 ) {
    int i = m_MinIndex;
    //OutStream << i << " ";

    pWordInfo = WordInfo( i );
    if ( pWordInfo ) {
      CSentence &Sentence = pWordInfo->TargetWords();
      Sentence.GetTargetString( TargetString );
    } else {
      TargetString += Word( i, pVocabulary );
    }
    //Trace( 1, SourceString );

    i++;

    while ( i <= m_MaxIndex ) {
      TargetString += CString( " " );

      pWordInfo = WordInfo( i );
      if ( pWordInfo ) {
	CSentence &Sentence = pWordInfo->TargetWords();
	Sentence.GetTargetString( TargetString );
      } else {
	TargetString += Word( i, pVocabulary );
      }
      //Trace( 1, SourceString );
      
      i++;
    }
  }

}

/**
   @brief Calculate counts for the words in the sentence.
   @param The function result, a word ID to counter map.

   The container will <i>not</i> be cleared. This make it possible to accumulate counts over
   multiple sentences. This function directly uses the word indices stored in the CSentence object.
*/
void CSentence::GetWordCounts(CWordIdx2CountMap & result) const {
  // iterators
  CWordIdx2CountMap::iterator iterWordCount;

  // !! DO NOT RESET THE RESULT CONTAINER !!
  // This way we can acumulate counts

  // forall words
  for (int i = m_MinIndex; i <= m_MaxIndex; ++i) {
    // use CSentence's word indices directly
    iterWordCount = result.find(m_pElements[i]);
    if (iterWordCount == result.end()) {
      // new entry in ref sentence vocabulary (count = 1)
      result.insert(CWordIdx2CountMap::value_type(m_pElements[i],1));
    } else {
      // word already exists, increase counter
      ++((*iterWordCount).second);
    } // endif: entry in sentence's vocabulary already exists?
  } // endfor: collect word counts of current sentence
} // CSentence::GetWordCounts( ... )

/**
   @brief Calculate counts for the words in the sentence.
   @param The function result, a word ID to counter map.
   @param vocab The sentence words (strings) will be mapped
   onto word id's before accumulating the counts.

   The container will <i>not</i> be cleared. This make it possible to accumulate counts over
   multiple sentences. Note that the object's word id array will not be used, this function
   uses the words as a string directly.
*/
void CSentence::GetWordCounts(CWordIdx2CountMap & result, const CVocabulary & vocab) const {
  // iterators
  CWordIdx2CountMap::iterator iterWordCount;
  // word index
  CWordIndex idx;

  // !! DO NOT RESET THE RESULT CONTAINER !!
  // This way we can acumulate counts


  for (int i = m_MinIndex; i <= m_MaxIndex; ++i) {
    idx = vocab.GetIndex(Word(i));
    iterWordCount = result.find(idx);
    if (iterWordCount == result.end()) {
      // new entry in ref sentence vocabulary (count = 1)
      result.insert(CWordIdx2CountMap::value_type(idx,1));
    } else {
      // word already exists, increase counter
      ++((*iterWordCount).second);
    } // endif: entry in sentence's vocabulary already exists?
  } // endfor: collect word counts of current sentence
} // CSentence::GetWordCounts( ... )




//--------------------------------------------------------------------
// CText
//--------------------------------------------------------------------


/********************************************************************/
/* CText constructor & destructor */

CText::CText()
{
  m_TextSize = 0;
  m_pVocabulary = 0;
  m_Sentences.InitToZero();
  m_Sentences.SetIncrement( 1000 );
  m_SentenceLengthIncrement = 32;
  m_ReadOneByOne = false;
  m_pTextStream = NULL;
  m_pTextFilter = NULL;
} // CText


CText::~CText()
{
  // m_Sentences holds pointers to sentences, therefore we have
  // to delete the sentences
  Clear ();
} // ~CText

void
CText::Clear ()
{
  for ( int i = m_Sentences.MinIndex(); i <= m_Sentences.MaxIndex(); i++ ) {
    delete m_Sentences[ i ];
    m_Sentences[ i ] = NULL;
  }
  m_Sentences.RemoveAll();
}


/********************************************************************/
/* CText Load and Store Text */


// returns number of words loaded, < 0 on error
int
CText::LoadIndexedText( const char *name )
{
  // make sure that we don't have any old stuff
  Clear();

  int SentenceNumber, WordPosition;
  CWordIndex  WordIndex;
  CSentence *pSentence;

  std::ifstream InStream( name );
  if (!InStream.good()) {
      std::cerr << "CText::LoadIndexedText(): " << rcsid_text_cc
	   << "Textfile: " << name << " could not be opened for reading." << std::endl << std::flush;
      return -1;
  } // endif: error
	   
  //  Trace(2, "Textfile: " << name);

  // text file contains vocabulary, so we need a valid vocabulary object here
  assert( m_pVocabulary );

  if (!m_pVocabulary->Load( InStream )) {
      std::cerr << "CText::LoadIndexedText(): could not load Vocabulary from file "
	   << name << std::endl << std::flush;
      return -1;
  } // endif: error

  //Trace(1, 
  //"Vocsize (including UNK and empty word, but without EOS, SOS, <p>, <art>): "
  //<< m_pVocabulary->GetVocabularySize() ); 
  Trace(1, "Load Text");


  char str[ 200 ];

  // skip comment lines
  InStream.getline( str, 200 );
  while ( (str[ 0 ] == '#') && !InStream.eof()) {
    InStream.getline( str, 200 );
  }

  if ( InStream.eof() ) {
      std::cerr << "CText::LoadIndexedText(): No text read, empty file "
	   << name << std::endl << std::flush;
      return 0;
  } // endif: error. unexpected EOF

  // we have read past comment lines, so the string contains the
  // wordindex of the first word
  WordIndex = atoi( str );

  m_TextSize = 1;

  // start with the first sentence
  if ( !(pSentence = new CSentence) ) {
    std::cerr << "CText::LoadIndexedText(): Could not allocate sentence. "
	 << "Text not completely read\n" << std::flush;
    InStream.close();
    return -1;
  }

  SentenceNumber = 1;
  WordPosition = 1;

  // indexed text is ended by -1
  while ( WordIndex != -1 ) {

    //Trace(5, WordIndex << " " << m_pVocabulary->GetWord( WordIndex ) );

    // at end of sentence, store current sentence and create a new one
    if( WordIndex == m_pVocabulary->EOS ) {

      m_Sentences[ SentenceNumber ] = pSentence;
      
      TraceIf( (SentenceNumber % 1000 == 0), 1, "Sentence " << SentenceNumber );

      if ( !(pSentence = new CSentence) ) {
	std::cerr << "CText::LoadIndexedText(): Could not allocate sentence. "
	     << "text not completely read\n" <<std::flush;
	InStream.close();
	return -1;
      }

      SentenceNumber++;
      WordPosition = 1;

    } else {

      (*pSentence)[ WordPosition ] = WordIndex;
      WordPosition++;
      m_TextSize++;
    }

    InStream >> WordIndex;

    if ( InStream.eof() ) {
	std::cerr << "CText::LoadIndexedText(): unexpected eof" << std::endl;
	return -2;
    } // endif: error

  } // endwhile


  InStream.close();

  // here, we have a sentence which not used
  delete pSentence;

  Trace(3, "TextSize: " << m_TextSize);
  // any unknown word in the corpus
  //std::cout << "Number of unknown words " << voccount[0] << std::endl;
  
  //  std::cout << "Laengste: " << m_Sentences.get_maxjsize() << "\n";
  
  Trace(2, "Text loaded");

  return m_TextSize;
} // LoadIndexedText


/*
int
CText::Load( const char *name )
{
  // declaration part -----------------------------------------------------

  std::ifstream fp(name);
  char
    str [ MAXstrlen+1 ];
  int index, len;

  CSentence *pSentence;

  // implementation part --------------------------------------------------

  EXIT_IF ((!fp.good()),
	   "\n"<<  rcsid_text_cc << "\nTextfile: " 
	   << name << " could not be opened for reading\n",
	   -1);
	   
  assert( m_pVocabulary );

  Trace(1, "Load Text" << name);

  m_TextSize = 0;

  //----- Kommentar lesen -----
  for( fp.getline( str, MAXstrlen ); (str[ 0 ] == '#') && (!fp.eof());
       fp.getline( str, MAXstrlen ) )
    {
      ;
    }

  for (index = 0; !fp.eof(); index++, fp.getline( str, MAXstrlen ) ) 
    {
      TraceIf( ( index % 1000 == 0 ), 1, index );

      if ( !(pSentence = new CSentence) )
	{
	  std::cout << "\n could not allocate sentence\n" <<std::flush;
	  return 0;
	}

      if ( !m_pVocabulary->ConvertToIndices( str, pSentence, &len ) )
	{
	  std::cout << "\n could not convert: " << str << "\n" <<std::flush;
	  return 0;
	}

      m_TextSize += len;

      pSentence->Insert (0, EW); // empty word

      m_Sentences[index] = pSentence;

      TraceIf( (index % 1000 == 0), 2, ".");
    } // for m_Sentences


  fp.close();

  Trace(1, "\n M_TextSize: " << m_TextSize);
      // any unknown word in the corpus
      //std::cout << "Number of unknown words " << voccount[0] << std::endl;

      //  std::cout << "Laengste: " << m_Sentences.get_maxjsize() << "\n";

  Trace(1, "\n Text loaded");

  return m_TextSize;
} // LoadIndexedText
*/



void
CText::OpenAsciiText( const char *name )
{
  Trace( 2, "Open " << name << " for reading on sentence at a time" );
  m_FileName = CString( name );
  m_pTextStream = new std::ifstream;
  m_pTextStream->open( m_FileName );
}



// returns number of words loaded, -1 on error
int
CText::LoadAsciiText( const char *name )
{
  // make sure that we don't have any old stuff
  Clear();

  if ( m_ReadOneByOne ) {
    OpenAsciiText( name );
    return 0;   // we don't know the corpus size
  }

  // declaration part -----------------------------------------------------

  std::ifstream fp(name);
  char str [ MAXstrlen + 1 ];
  int index, len;

  CSentence *pSentence;

  // implementation part --------------------------------------------------

  if (!fp.good()) {
      std::cerr << "CText::LoadAsciiText(): " << rcsid_text_cc
	   << "Textfile: " << name << " could not be opened for reading." << std::endl <<std::flush;
      return -1;
  } // endif: error
	   
  assert( m_pVocabulary );

  Trace(1, "Load Text" << name );

  m_TextSize = 0;

  //----- Kommentar lesen -----
  for( fp.getline( str, MAXstrlen ); (str[ 0 ] == '#') && (!fp.eof());
       fp.getline( str, MAXstrlen ) ) {
    ; //std::cout << str << std::endl <<std::flush;
  }

  //for (index = 0; !fp.eof(); index++, fp.getline( str, MAXstrlen ) ) {

  index = 0;
  
  do {
    TraceIf( ( index > 0 ) && ( index % 100000 == 0 ), 0, index );
    //std::cout << str << std::endl;

    if ( !(pSentence = new CSentence) ) {
      std::cerr << "CText::LoadAsciiText(): could not allocate sentence\n" <<std::flush;
      return -1;
    } // endif: error

    pSentence->SetIncrement( m_SentenceLengthIncrement );

    // add TextFilter here
    if( m_pTextFilter != NULL ) {
      len = pSentence->ReadFromBuffer( ((std::string)(m_pTextFilter->Filter(str))).c_str(), 
                                       m_pVocabulary );
    } else {
      len = pSentence->ReadFromBuffer( str, m_pVocabulary );
    }

    m_TextSize += len;

    //pSentence->Insert (0, EW); // empty word
    
    m_Sentences[ index ] = pSentence;
    
    //TraceIf( (index % 1000 == 0), 2, ".");

    index++;
    
  } while ( !fp.eof() && fp.getline( str, MAXstrlen ) );   // for m_Sentences


  fp.close();
  
  Trace(1, "M_TextSize: " << m_TextSize);
  // any unknown word in the corpus
  //std::cout << "Number of unknown words " << voccount[0] << std::endl;
  
  //  std::cout << "Laengste: " << m_Sentences.get_maxjsize() << "\n";
  
  Trace(1, "Text loaded");
  
  return m_TextSize;
} // Load



bool
CText::AppendSentence( CSentence *pSentence )
{
  m_TextSize += pSentence->Size();

  m_Sentences.Append( pSentence );


  return TRUE;
}



int
CText::Size () const
{
  return m_TextSize;
} // Size




/********************************************************************/
/* CText Functions to get sentences or words */

/*
CWordIndex
CText::GetWordAt ( TPosition posword ) const
{
  TPosition possent;

  for( possent = 0; possent < m_Sentences.Size(); possent++ )
    {
      if( posword == 0 ) return m_pVocabulary->EOS;
      posword--; 
      if( posword <= (long)GetSentenceLength(possent) ) 
	return m_Sentences[possent]->GetAt( posword );
      posword -= GetSentenceLength(possent)+1;
      if( posword == 0  ) return m_pVocabulary->EOS;
      posword--;
    } // for possent
  return m_pVocabulary->EOT;
} // GetWordAt
*/


CWordIndex
CText::GetWordAt ( TPosition possent, int wordpos ) const
{
  //if( m_UseEmptyWord == FALSE ) wordpos++;
  assert( possent >=  m_Sentences.MinIndex() );
  assert( possent <=  m_Sentences.MaxIndex() );
  //assert( (int)word <  m_Sentences.GetAt(possent) );
  return m_Sentences[possent]->GetAt( wordpos );
} // GetWordAt






void
CText::SetWordAt ( TPosition possent, int posword, CWordIndex W )
{
  //if( m_UseEmptyWord == FALSE ) posword++;
  assert( possent <  m_Sentences.Size() );
  //assert( (int)word <  m_Sentences.GetAt(possent) );
  m_Sentences[possent]->SetAt( posword, W );
} // SetWordAt


void
CText::SetTextFilter ( CTextFilterBase* pTextFilter ) 
{
  m_pTextFilter = pTextFilter;
}


void
CText::PutSentence ( TPosition sent, const CSentence &SentenceIn )
{
  if( m_Sentences.IsValidIndex( sent ) && (m_Sentences[ sent ]) )
    {
      m_TextSize -= m_Sentences[ sent ]->Size();
      (*m_Sentences[ sent ]) = SentenceIn;
    }
  else
    {
      CSentence *pSentence;

      //ARV compatibility
      //EXIT_IF ( !(pSentence = new CSentence(SentenceIn) ), 
      //"could not allocate sentence", -1);
      pSentence = new CSentence(SentenceIn);
      if (!pSentence){
	std::cerr<<"could not allocate sentence";
	exit(-1);
      }
      m_Sentences[ sent ] = pSentence;
    }

  m_TextSize += SentenceIn.Size();
} // PutSentence


void
CText::InsertSentence ( TPosition sent, CSentence *pSentence )
{
  m_Sentences.Insert( sent, pSentence );
}



/********************************************************************/
/* CText display */

void
CText::Display( bool WithWordInfo ) const
{
  Write( std::cout, WithWordInfo );
}

void
CText::Write( std::ostream &OutStream, bool WithWordInfo ) const
{
  int Number;
  CSentence *pSentence;

  for( Number = m_Sentences.MinIndex(); 
       Number <= m_Sentences.MaxIndex(); Number++ ) {

    pSentence = m_Sentences[ Number ];
    assert( pSentence );
    pSentence->Write( OutStream, m_pVocabulary, WithWordInfo );
    OutStream << std::endl;
  }
  OutStream <<std::flush;
}




void
CText::DisplaySentence ( TPosition sent, bool WithWordInfo ) const
{
  WriteSentence( sent, std::cout, WithWordInfo );
} // DisplaySentence


void
CText::WriteSentence ( TPosition sent, std::ostream &OutStream, bool WithWordInfo ) const
{
  assert( IsValidIndex(sent) );
  
  CSentence *pSentence = m_Sentences[sent];
  assert( pSentence );
  pSentence->Write( OutStream, m_pVocabulary, WithWordInfo );
} // WriteSentence




TPosition
CText::FindSentence ( TPosition possent, CWordIndex w ) const
{
  assert( 0 );
  
  w = 0; // dummy

  // find next sentence containing the word w
  // search starts with sentence at possent

  // satz?!
  CWordIndex satz;

  assert (possent < m_Sentences.Size());

  for (satz = possent; satz < m_Sentences.Size(); satz++ ) {
    //wort = m_UseEmptyWord == TRUE? 0: 1;
    //    for (; wort < m_Sentences.getlinesize(satz); wort++ ) {
    //      if (w == m_Sentences(satz, wort)) {
    //	return satz;
    //      }
    //    } // for wort
  } // for satz
  return m_pVocabulary->EOT;
}



CWordIndex
CText::FirstSentence ( TPosition &pos )
{
  if ( m_ReadOneByOne ) {
    Trace( 2, "read first sentence only" );
 
    char Buffer[ 10001 ];
    int len;
      
    m_Sentences[ 0 ] = new CSentence;
      
    //m_TextStream.seekg( 0 );
    //cout << "Position in file is " << m_TextStream.tellg() << endl;
    // does not reposition to the beginning, therefore close and open again
    Trace( 1, "close and open again " << m_FileName );
    if ( m_pTextStream ) {
      if ( m_pTextStream->is_open() ) {
        m_pTextStream->close();
      }
      delete m_pTextStream;
    }
    m_pTextStream = new std::ifstream;
    m_pTextStream->open( m_FileName );
    if (!m_pTextStream->good()) {
      std::cerr << "Textfile: " << m_FileName 
           << " could not be opened for reading." << std::endl << std::flush;
      return -1;
    } // endif: error
    

    m_pTextStream->getline( Buffer, 10000 );
      
    if (!m_pTextStream->eof() ) {
      len = m_Sentences[ 0 ]->ReadFromBuffer( Buffer, m_pVocabulary );
      m_TextSize += len;
      pos = 0;
    } else {
      return m_pVocabulary->EOT;
    }
    
  } else {
      
    pos = m_Sentences.MinIndex();
    
    if ( pos > m_Sentences.MaxIndex() ) {
      return m_pVocabulary->EOT;
    }
  }
    
  return m_pVocabulary->EOS;
} // FirstSentence
      
    
      
CWordIndex
CText::NextSentence ( TPosition &pos )
{
  if ( m_ReadOneByOne ) {
    Trace( 2, "Read one line only" );
    // read one line and store at index 0
    
    char Buffer[ 10001 ];
     
    delete m_Sentences[ 0 ];
    m_Sentences[ 0 ] = new CSentence;
  
    m_pTextStream->getline( Buffer, 10000 );
      
    if (!m_pTextStream->eof() ) {
      int len = m_Sentences[ 0 ]->ReadFromBuffer( Buffer, m_pVocabulary );
      m_TextSize += len;
    } else {
      return m_pVocabulary->EOT;
    }
    
  } else {
    if ( pos < m_Sentences.MinIndex() ) {
      pos = m_Sentences.MinIndex();
    } else {
      pos++;
    }
  
    if ( pos > m_Sentences.MaxIndex() ) { 
      return m_pVocabulary->EOT;
    }
      
  }
    
  return m_pVocabulary->EOS;
} // NextSentence

