#include <iostream>
#include <fstream>
#include <string>

#include <TextFilterBase.hh>

CTextFilterBase::CTextFilterBase()
{
}


CTextFilterBase::~CTextFilterBase()
{
}



void
CTextFilterBase::Clear()
{
  m_Patterns.Clear();
}

/*!
 Reads instructions/patterns/regexp/etc. for subclasses of CTextFilterBase and then
 runs AddPattern( pattern ), which is a pure virtual function in the base class.
 
 \param FileName a CString with path of file with patterns.
 
 \return true if at least one pattern was loaded, false if none were loaded
 
 \note Assumes one pattern per line in file.
 
 \author Anthony D'Auria (apd@cs.cmu.edu)
 
 */
bool
CTextFilterBase::LoadPatterns( const CString & FileName )
{
  
  m_FileName = FileName;

  std::string line;
  std::ifstream textFile (((std::string)FileName).c_str());
  
  if(textFile.is_open())
  {
    while(! textFile.eof() )
    {
      getline(textFile, line);    // match
      
      // skip empty lines or lines that are commented out with '#'
      if ( line[0] != '#' && line[0] != '\0' ){
        AddPattern(CString (line.c_str()));
      }
    }
    
  } else {
    cout << "Error: TextFilterBase: unable to open file: " << FileName << endl;
    return false;
  }
  
  cout << "Patterns loaded: " << GetNumberOfPatterns() << endl;

  return ( GetNumberOfPatterns() > 0 );
}


void
CTextFilterBase::DisplayPatterns() const
{
  for ( int i = m_Patterns.MinIndex(); i <= m_Patterns.MaxIndex(); i++ ) {
	cout << m_Patterns[ i ] << endl;
  }
}
