#ifndef __TextFilterBase__
#define __TextFilterBase__

#include <String.hh>
#include <Array.hh>

//====================================================================
//  CTextFilterBase::
//
//  Base Text Filter class


class CTextFilterBase
{
public:
  CTextFilterBase();
  ~CTextFilterBase();

  //! remove all loaded regex patterns
  virtual void Clear();

  //! load regular expression patterns from file
  //! return false if invalid filename or no regexps in file
  bool LoadPatterns( const CString &FileName );

  //! set one regex pattern, delete any previously loaded patterns
  virtual void SetPattern( const CString &Pattern ) = 0;
  //virtual void SetPattern( const CString &Pattern );

  //! add one pattern to the set of all regex patterns
  virtual void AddPattern( const CString &Pattern ) = 0;

  //! apply the regex patterns to string and return the modified string
  virtual CString Filter( const CString &String ) = 0;

  virtual void DisplayPatterns() const;
  
  virtual int GetNumberOfPatterns() const = 0;

protected:
  //! file name of file which contains replace patterns
  CString m_FileName;

  //! an array to store multiple replacement patters
  CArray< CString > m_Patterns;
};





#endif
