#include <iostream>

#include <TextFilterBoostRegex.hh>


//====================================================================
//  CTextFilterBoostRegex::
//
//     Parses perl-style regexp and converts to Boost Regex
//

CTextFilterBoostRegex::CTextFilterBoostRegex()
{
  m_perlParseExpr = boost::regex("(s|tr)/(.*)/(.*)/(g?)");
}


CTextFilterBoostRegex::~CTextFilterBoostRegex()
{
}


//--------------------------------------------------------------------
//  CTextFilterBoostRegex:: AddPattern
//

void 
CTextFilterBoostRegex::AddPattern( const CString &Pattern )
{

  PerlToBoostRegex boostRegex;
  std::string regexFunc;
  std::string matchExpr;
  std::string replaceExpr;
  bool globalFunc;

  boost::cmatch perlParseSegs;         //stores matches as array
  // todo: cast (const char*)(Pattern)
  if(boost::regex_match(((std::string)Pattern).c_str(), 
                        perlParseSegs, m_perlParseExpr))
  {
    
    regexFunc   = perlParseSegs[1];
    matchExpr   = perlParseSegs[2]; // 2
    replaceExpr = perlParseSegs[3]; // 3
    globalFunc  = ((std::string)perlParseSegs[4] == "g"); // 4
                  
    if(regexFunc == "tr") 
    {
      // to lower case
      if(matchExpr == "[A-Z]" && replaceExpr == "[a-z]") {
        boostRegex.nonRegexFunc = 1;      
        m_BoostPatterns.Append(boostRegex);
      }
      
    } else if(regexFunc == "s") {
      
      boostRegex.matchExpr    = boost::regex(matchExpr);
      boostRegex.replaceExpr  = replaceExpr;
      boostRegex.global       = globalFunc;
      boostRegex.pattern      = Pattern;
      boostRegex.nonRegexFunc = 0;
      
      m_BoostPatterns.Append(boostRegex);
      
    } else {
      cout << "Unable to parse regex function out of first segment of expr: \""
      << Pattern << "\"" << endl;
      cout << "   First First seg: '" << perlParseSegs[0] << "'" << endl
        << "    First segment: '" << regexFunc << "'" << endl
        << "    matchExpr: '" << matchExpr << "'" << endl
        << "    replaceExpr: '" << replaceExpr << "'" << endl
        << "    globalFunc: '" << globalFunc << "'" << endl;
      
    }
  } else {
    if((std::string)Pattern != "") {
      cout << "Unable to parse perl regex: \"" << Pattern << "\"" << endl;
    } else {
      // read empty line
    }
  }

}



void 
CTextFilterBoostRegex::Clear()
{
  m_BoostPatterns.Clear();
}


//--------------------------------------------------------------------
//  CTextFilterBoostRegex:: SetPattern
//

void
CTextFilterBoostRegex::SetPattern( const CString &Pattern)
{
  m_BoostPatterns.Clear();
  AddPattern(Pattern);
}


//--------------------------------------------------------------------
//  CTextFilterBoostRegex:: Filter
//

CString
CTextFilterBoostRegex::Filter( const CString &String )
{
  std::string FilteredString = (std::string)String;

	// must pad end of string with " " if user 
	// wants the following regex to work correctly:
	// s/([\.,])([^0-9])/ $1 $2/g 
	// (first)(^[second]) will never match if first is at end of string.
	// This, I believe, is a Boost bug. In mteval.pl, all strings end with \n
	FilteredString += " ";

  for(int i=0; i < m_BoostPatterns.Size(); i++)
  {
    PerlToBoostRegex normalizingExpr = m_BoostPatterns[i];

    if (normalizingExpr.nonRegexFunc == 0)    // check if performing non-boost operation
    {
      if (normalizingExpr.global)
      {
        FilteredString = boost::regex_replace(FilteredString,
                                              normalizingExpr.matchExpr,
                                              normalizingExpr.replaceExpr);
      } else {
        
        FilteredString = boost::regex_replace(FilteredString,
                                              normalizingExpr.matchExpr,
                                              normalizingExpr.replaceExpr,
                                              boost::match_default 
                                              | boost::format_first_only);
      }
    } else if (normalizingExpr.nonRegexFunc == 1) {
      CString toLowerCaseStep = CString(FilteredString.c_str());
      toLowerCaseStep.ToLowerCase();
      FilteredString = (std::string)toLowerCaseStep;
    }
    
  } // end boost patterns loop

  return CString (FilteredString.c_str());
}



//--------------------------------------------------------------------
//  CTextFilterBoostRegex:: DisplayPatterns
//

void
CTextFilterBoostRegex::DisplayPatterns() const
{
  for (int i=m_BoostPatterns.MinIndex(); i <= m_BoostPatterns.MaxIndex(); i++)
  {
    if(m_BoostPatterns[i].nonRegexFunc != 0){
      cout << "External Func: " << m_BoostPatterns[i].nonRegexFunc << endl;
    } else {
      cout << "Pattern: " << m_BoostPatterns[i].pattern
      << " Search: \""    << m_BoostPatterns[i].matchExpr << "\""
      << " Replace: \""   << m_BoostPatterns[i].replaceExpr << "\""
      << " Global: "      << m_BoostPatterns[i].global << endl;
    }
  }
}

int
CTextFilterBoostRegex::GetNumberOfPatterns() const
{
  return m_BoostPatterns.Size();
}





