#ifndef __TextFilterBoostRegex__
#define __TextFilterBoostRegex__

#include <boost/regex.hpp>

#include <TextFilterBase.hh>


//====================================================================
//  PerlToBoostRegex
//    Stores segmented perl regex

struct PerlToBoostRegex
{
  CString pattern;
  
  bool global;
  bool lowerCase;
  boost::regex matchExpr;
  std::string replaceExpr;

  size_t nonRegexFunc;
};


//====================================================================
//  CTextFilterBoostRegex
//

class CTextFilterBoostRegex : public CTextFilterBase
{
public: 
  CTextFilterBoostRegex();
  virtual ~CTextFilterBoostRegex();

  //! remove all loaded regex patterns
  void Clear();

  //! load regular expression patterns from file
  //void LoadPatterns( const CString &FileName );

  //! set one regex pattern, delete any previously loaded patterns
  void SetPattern( const CString &Pattern );

  //! add one pattern to the set of all regex patterns
  void AddPattern( const CString &Pattern );

  //! apply the regex patterns to string and return the modified string
  CString Filter(const CString &String );
  
  void DisplayPatterns() const;

  int GetNumberOfPatterns() const;

protected:
  // store every parsed perl regex in array
  CArray< PerlToBoostRegex > m_BoostPatterns;
  
  // expression used to parse perl-format regex expressions
  boost::regex m_perlParseExpr;
};





#endif
