// CText - Inlines
// Base class for text handling
//
// $Id: Text_inl.hh 1742 2006-06-27 02:28:05Z vogel $ 
//


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif



//--------------------------------------------------------------------
// Inlines for CWordInfo
//--------------------------------------------------------------------


inline void
CWordInfo::AddSourceWord( const CString & Word )
{
  //cout << "CWordInfo::AddSourceWord " << Word << endl;
  ///m_SourceWords.Append( Word );
  m_SourceWords.SetWord( Word, m_SourceWords.Size() + 1 ); 
}



inline void
CWordInfo::AddTargetWord( const CString & Word )
{
  ///m_TargetWords.Append( Word );
  m_TargetWords.SetWord( Word, m_TargetWords.Size() + 1 ); 
}



inline const CString &
CWordInfo::FirstWordTarget() const
{
  return m_SourceWords.Word( m_SourceWords.MinIndex() );
}



inline const CString &
CWordInfo::LastWordTarget() const
{
  return m_TargetWords.Word( m_TargetWords.MaxIndex() );
}



inline int
CWordInfo::SourceStringLength() const
{
  return m_SourceWords.Size();
}



inline int
CWordInfo::TargetStringLength() const
{
  return m_TargetWords.Size();
}



inline CSentence & 
CWordInfo::SourceWords()
{
  return m_SourceWords;
}



inline CSentence & 
CWordInfo::TargetWords()
{
  return m_TargetWords;
}


inline double &
CWordInfo::Score()
{
  return m_Score;
}


inline bool
CWordInfo::HasSourceString() const
{
  return m_SourceWords.Size() > 0;
}



inline bool
CWordInfo::HasTargetString() const
{
  return m_TargetWords.Size() > 0;
}



inline bool
CWordInfo::HasScore() const
{
  return m_Score != 1.0e10;
}




//--------------------------------------------------------------------
// Inlines for CSentence
//--------------------------------------------------------------------

inline
CSentence::operator CWordIndex const *()
{
  return &(m_pElements[ m_MinIndex ]);
}


inline
int
CSentence::Length() const
{
  // length of sentence is size of array
  return Size();
}



inline
CWordInfo *
CSentence::WordInfo( int i) const
{
  if ( IsValidIndex( i ) )
    if( m_pWordInfos )
      if( m_pWordInfos->IsValidIndex( i ) )
	return (*m_pWordInfos)[ i ];

  return 0;
}





//--------------------------------------------------------------------
// Inlines CText
//--------------------------------------------------------------------

inline bool
CText::Allocate( int MaxNumOfSentences )
{
  m_Sentences.SetIncrement( MaxNumOfSentences );
  return true;
}


inline bool
CText::AllocateSentenceLength( int Increment )
{
  m_SentenceLengthIncrement = Increment;
  return true;
}


inline int 
CText::NumberOfSentences () const
{
  return m_Sentences.Size();
} // GetNumberOfSentences



inline CSentence *
CText::GetSentence ( TPosition possent )
{
  if ( !IsValidIndex( possent ) )
    return (CSentence *)0;

  return m_Sentences[ possent ];
} // GetSentence



inline int
CText::GetSentence ( TPosition possent, CWordIndex const *&words ) const
{
  if ( !IsValidIndex( possent ) )
    return -1;

  words = (CWordIndex const *)*(m_Sentences.GetAt( possent ));

  return SentenceLength( possent );
} // GetSentence



inline int
CText::SentenceLength ( TPosition possent ) const
{
  if ( !IsValidIndex( possent ) )
    return -1;

  return m_Sentences[ possent ]->Length();
} // GetSentenceLength



inline void
CText::ClearSentence ( TPosition possent )
{
  if ( IsValidIndex( possent ) )
    m_Sentences[ possent ]->Clear();
}



/********************************************************************/
/* CText Functions to iterate over the text */

inline void
CText::ReadOneByOne( bool Flag )
{
  m_ReadOneByOne = Flag;
}


inline int
CText::MinIndex() const
{
  return m_Sentences.MinIndex();
}


inline int
CText::MaxIndex() const
{
  return m_Sentences.MaxIndex();
}




inline CWordIndex
CText::FirstWord ( TPosition possent, TPosition &posword ) const
{
  if ( possent > m_Sentences.MaxIndex() ) {
    return m_pVocabulary->EOT;
  }

  posword = m_Sentences[ possent ]->MinIndex();
  if ( posword > m_Sentences[ possent ]->MaxIndex() ) 
    return m_pVocabulary->EOS;

  return m_Sentences[ possent ]->GetAt( posword );
} // FirstWord



inline CWordIndex
CText::NextWord ( TPosition possent, TPosition &posword ) const
{
  if ( possent > m_Sentences.MaxIndex() ) {
    return m_pVocabulary->EOT;
  }

  posword++;
  if ( posword > m_Sentences[ possent ]->MaxIndex() ) 
    return m_pVocabulary->EOS;

  return m_Sentences[ possent ]->GetAt( posword );
} // NextWord



inline bool
CText::SetVocabulary (CVocabulary * pVocabulary)
{
  m_pVocabulary = pVocabulary;

  return TRUE;
}


inline CVocabulary * 
CText::Vocabulary()
{
  return m_pVocabulary;
}


inline bool 
CText::IsValidIndex( int Index ) const
{
  return (Index >= m_Sentences.MinIndex()) && (Index <= m_Sentences.MaxIndex());
}




