#ifndef Vocabulary_hh
#error ERROR: Header was not included by Vocabulary.hh. This is an internal header. Do not attempt to include it directly. Use Vocabulary.hh instead.
#endif

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// CVocabulary - Inlines
// Base class for vocabulary handling
//

#include <stdio.h>
#include <string.h>

inline const CString &
CVocabularyItem::Word() const
{
  return m_Word;
}


inline void
CVocabularyItem::Word( const CString & Word )
{
  m_Word = Word;
}


inline const char *
CVocabularyItem::PointerToWord() const
{
  return (const char *)m_Word;
}


inline bool
CVocabularyItem::IsContentWord() const
{
  return m_IsContentWord;
}


inline void
CVocabularyItem::IsContentWord( bool IsContentWord )
{
  m_IsContentWord = IsContentWord;
}




inline bool
CVocabularyItem::IsMultiWord() const
{
  return m_IsMultiWord;
}


inline uint
CVocabularyItem::Length() const
{
  return m_pWordIndices->Size();
}

inline CWordIndex
CVocabularyItem::FirstWord()
{
  return (*m_pWordIndices)[ 1 ];
}

inline CWordIndex 
CVocabularyItem::LastWord()
{
  return (*m_pWordIndices)[ m_pWordIndices->Size() ];
}


inline void
CVocabularyItem::Display()
{
  printf("%s %i \n", (const char *)m_Word, m_IsContentWord);
}




// ********************************************************************
// CVocabulary

//! First index of a word (not UNK or EOS)
inline CWordIndex
CVocabulary::First() const
{
  //assert(CVocabulary::EOS == 1);
  if ( m_VocItems.Size() > 0 )
    return 4;
  else
    return 0;

}


//! Last Index of a word (not UNK or EOS)
inline CWordIndex
CVocabulary::Last() const
{
  return Size() + 1;
}



inline int
CVocabulary::Size() const
{
  // we do not count <UNK> and </s>
  return m_VocItems.Size() - 2;
}



// Get-Methods

inline const CVocabularyItem &
CVocabulary::GetItem( CWordIndex WordIndex ) const
{
  return *m_VocItems[ WordIndex ];
} // GetItem


inline const CString &
CVocabulary::GetWord ( CWordIndex WordIndex ) const
{
  return m_VocItems[ WordIndex ]->m_Word;
} // GetWord

inline const CString *
CVocabulary::GetWordPtr ( CWordIndex WordIndex ) const
{
  if ((m_VocItems.MinIndex()<=WordIndex) && (m_VocItems.MaxIndex()>=WordIndex)){
    return &(m_VocItems[ WordIndex ]->m_Word);
  }else{
    return NULL;
  }
} // GetWordPtr


inline bool
CVocabulary::IsContentWord( CWordIndex WordIndex ) const
{
  return m_VocItems[ WordIndex ]->m_IsContentWord;
}


inline bool
CVocabulary::IsMultiWord( CWordIndex WordIndex ) const
{
  return m_VocItems[ WordIndex ]->m_IsMultiWord;
}




inline CWordIndex
CVocabulary::FirstWord( CWordIndex WordIndex ) const
{
  return m_VocItems[ WordIndex ]->FirstWord();
}


inline CWordIndex
CVocabulary::LastWord( CWordIndex WordIndex ) const
{
  return m_VocItems[ WordIndex ]->LastWord();
}



inline void
CVocabulary::SetStringForEmptyWord( CString StringForEmptyWord )
{
  m_StringForEmptyWord = StringForEmptyWord;
}



inline CString
CVocabulary::GetStringForEmptyWord()
{
  return m_StringForEmptyWord;
}



inline CString
CVocabulary::GetStringForUNK()
{
  return "<UNK>";
}


inline CString
CVocabulary::GetStringForEOS()
{
  return "</s>";
}



