#include <stdlib.h>
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <stdlib.h>

#define NO_TRACE 1
#include <Trace.hh>

#include <CostMatrix.hh>
#include <Vocabulary.hh>

using namespace std;

int 
CCostMatrix::m_AddTL = 0;


CCostMatrix::CCostMatrix( const CVocabulary *pVoc )
{
  int VocSize;

  Trace( 5, "Called constructor of CCostMatrix" );
  m_pVoc = pVoc;

  VocSize = m_pVoc->Size();

  m_pCostForSubstitutions = new CSparseMatrix( VocSize + 1, VocSize + 1);

  m_pCostForInsertions = new CArray1D< float >;
  m_pCostForInsertions->AssureBoundaries( 0, VocSize + 1 );

  m_pCostForDeletions = new CArray1D< float >;
  m_pCostForDeletions->AssureBoundaries( 0, VocSize + 1 );


  // set default values in case no cost files are loaded
  m_pCostForSubstitutions->SetDefaultValue( 2.0 );
  m_pCostForInsertions->SetAll( 1 );
  m_pCostForDeletions->SetAll( 1 );
}


void CCostMatrix::LoadSubst(const char *FileName)
{
  ifstream fp(FileName);
  int index;
  char OrigWord[200], SubstWord[200], Word[200];
  CWordIndex WOrig, WSubst;
  int DefaultValue;
  float Cost;

  Trace(1, "Called LoadSubst");

  if (!fp) {
      cerr << "File: " << FileName << " could not be opened for read";;
      exit (-1);
  }

  // Get DefaultValue
  fp >> Word;
  if (CString(Word) == CString("DefaultValue"))
    fp >> DefaultValue;
  else {
      cerr << "Default value wasn't specified in file: " << FileName;
      exit(0);
  }

  Trace(5, "Default value is: " << DefaultValue);
    // set default values
  m_pCostForSubstitutions->SetDefaultValue((double) DefaultValue);


  for(index= 0; !fp.eof(); index++) {
    fp >> OrigWord;
    fp >> SubstWord;
    fp >> Cost;

    Trace( 1, OrigWord << " " << SubstWord << " " << Cost );

    if (fp.fail())
      continue;
    // Search word in Vocabulary
    WOrig = m_pVoc->GetIndex( OrigWord );
    WSubst = m_pVoc->GetIndex ( SubstWord );
    Trace( 1, m_pVoc->GetWord( WOrig ) << " " << m_pVoc->GetWord( WSubst ) );
    
    if ((WOrig != m_pVoc->EOT) && (WSubst != m_pVoc->EOT))
      m_pCostForSubstitutions->SetAt(WOrig, WSubst, Cost);
    else
      Trace(0, "Subst pairs " << OrigWord << " " << SubstWord
	    << " were not in vocabulary");
  }
}

void CCostMatrix::LoadIns(const char *FileName)
{
  ifstream fp(FileName);
  char Word[200];
  int DefaultValue, index;
  float Cost;
  CWordIndex W;

  Trace(1, "Called LoadIns");

  if (!fp) {
      cerr << "File: " << FileName << " could not be opened for read";
      exit (-1);
  }

  // Get DefaultValue
  fp >> Word;
  if (CString(Word) == CString("DefaultValue"))
    fp >> DefaultValue;
  else {
      cerr << "Default value wasn't specified in file: " << FileName;
      exit(0);
  }

  Trace(5, "Default value is: " << DefaultValue);

  m_pCostForInsertions->SetAll(DefaultValue);
  
  for(index= 0; !fp.eof(); index++) {
    fp >> Word;
    fp >> Cost;
    Trace( 1, Word << " " << Cost );

    if (fp.fail())
      continue;

    // Search word in Vocabulary
    W = m_pVoc->GetIndex( Word );
    Trace( 1, m_pVoc->GetWord( W ) );
    
    if (W != m_pVoc->EOT)
      m_pCostForInsertions->SetAt(W, Cost);
    else
      Trace(0, "Ins word " << Word << " wasn't in vocabulary");
  }
}

void CCostMatrix::LoadDel(const char *FileName)
{
  ifstream fp(FileName);
  char Word[200];
  int DefaultValue, index;
  float Cost;
  CWordIndex W;

  Trace(1, "Called LoadDel");
  if (!fp) {
      cerr << "File: " << FileName << " could not be opened for read";
      exit (-1);
  }


  // Get DefaultValue
  fp >> Word;
  if (CString(Word) == CString("DefaultValue"))
    fp >> DefaultValue;
  else {
      cerr << "Default value wasn't specified in file: " << FileName;
      exit(0);
  }

 Trace(5, "Default value is: " << DefaultValue);
 
  m_pCostForDeletions->SetAll(DefaultValue);

  for(index= 0; !fp.eof(); index++) {
    fp >> Word;
    fp >> Cost;

    Trace( 1, Word << " " << Cost );

    if (fp.fail())
      continue;

    // Search word in Vocabulary
    W = m_pVoc->GetIndex( Word );
    Trace( 1, m_pVoc->GetWord( W ) );
    
    if (W != m_pVoc->EOT)
      m_pCostForDeletions->SetAt(W, Cost);
    else
      Trace(0, "Del word " << Word << " wasn't in vocabulary"); 
  }
}
