#ifndef CostMatrix_hh
#define CostMatrix_hh

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <SparseArray.hh>
#include <Array.hh>
#include <Vocabulary.hh>

class CCostMatrix {
public:
  CCostMatrix( const CVocabulary *pVoc );

  float Substitution(CWordIndex WOrig, CWordIndex WSubst);
  float Deletion(CWordIndex WDel);
  float Insertion(CWordIndex WIns);

  void LoadSubst(const char *FileName);
  void LoadIns(const char *FileName);
  void LoadDel(const char *FileName);

private:
  const CVocabulary *m_pVoc;

  CSparseMatrix *m_pCostForSubstitutions;
  CArray<float> *m_pCostForInsertions;
  CArray<float> *m_pCostForDeletions;

  // setoff to trace level
  static int m_AddTL;
};


// *************** INLINES *****************
inline
float CCostMatrix::Substitution(CWordIndex WOrig, CWordIndex WSubst)
{
  Trace( 2 - m_AddTL, "Sub-Cost: " << m_pVoc->GetWord( WOrig ) << " " << m_pVoc->GetWord( WSubst ) << " "
	 << m_pCostForSubstitutions->GetAt(WOrig, WSubst) );

  return m_pCostForSubstitutions->GetAt(WOrig, WSubst);
}

inline
float CCostMatrix::Deletion(CWordIndex WDel)
{
  Trace( 2 - m_AddTL, "Del-Cost: " << m_pVoc->GetWord( WDel ) << " "
	 << m_pCostForDeletions->GetAt(WDel) );

  return m_pCostForDeletions->GetAt(WDel);
}

inline
float CCostMatrix::Insertion(CWordIndex WIns)
{
  Trace( 2 - m_AddTL, "Ins-Cost: " << m_pVoc->GetWord( WIns ) << " "
	 << m_pCostForInsertions->GetAt(WIns) );

  return m_pCostForInsertions->GetAt(WIns);
}


#endif

