#include <string.h>
#include <stdlib.h>
// Implementation file for CTransducer, CTransducerItem, CTransducerAccess

// File ID:
// static char ID_Transducer_cc[] = "$Id: Transducer.cc 1677 2006-06-14 14:45:55Z vogel $"

// Programming: S. Vogel
// Last Update: June 03, 2006

// History: See header file


/*
ToDo:

- Is ScalingFactor still needed?  Should be removed

- make sure the pre-calculation of the logs works under all conditions,
  i.e. when loading transducer, when loading translation memory, 
  and when using online phrase alignment  

- check why source and target phrase are stored as CSentence and as string

- the store function does not seem to store multiple scores - check and correct

*/



#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <string>
#include <map>
#include <fstream>
using std::ifstream;
using std::ofstream;

#define NO_TRACE 1
#include <Trace.hh>

#include <Transducer.hh>



CTransducerState CTransducer::s_NotFound = -1;

CVocabulary *CTransducerItem::s_pInputVoc  = NULL;
CVocabulary *CTransducerItem::s_pOutputVoc = NULL;


CTransducerItem::CTransducerItem()
{
  m_Number = -1;
  
  m_ScalingFactor = 1.0;
  m_Count = 0.0;
  m_ScoreSourceGivenTarget = 0.0;
  m_ScoreTargetGivenSource = 0.0;
  m_pSource = NULL;
  m_pTarget = NULL;
}



CTransducerItem::~CTransducerItem()
{
  delete m_pSource;
  delete m_pTarget;
}



void
CTransducerItem::Display() const
{
  cout << m_Label << " # " << m_Source << " # " << m_Target << " # "
       << m_Count << " ; " 
       << m_ScoreTargetGivenSource << " / " << m_ScoreSourceGivenTarget
       << endl;
}



void
CTransducerItem::DisplayReverted() const
{
  cout << m_Label << " # " << m_Target << " # " << m_Source << " # "
       << m_Count << " ; " 
       << m_ScoreSourceGivenTarget << " / " << m_ScoreTargetGivenSource
       << endl;
}




//----------------------------------------------------------------------------
CTransducerAccess::CTransducerAccess( CTransducer *pTransducer, int Direction )
{
  m_pTransducer = pTransducer;
  m_Direction = Direction;

  assert( (m_Direction == 0) || (m_Direction == 1) );

  if ( m_Direction == 0 ) {
    
    m_pInputVoc = m_pTransducer->m_pInputVoc;
    m_pOutputVoc = m_pTransducer->m_pOutputVoc;
  
	m_pTransitionTable = &(m_pTransducer->m_TransitionTableSource);
    m_pEmissionTable = &(m_pTransducer->m_EmissionTableSource);

    m_pTransducerItems = &(m_pTransducer->m_TransducerItems);

    m_NumberOfStates = m_pTransducer->m_NumberOfStatesSource;
  } else {
    m_pInputVoc = m_pTransducer->m_pOutputVoc;
    m_pOutputVoc = m_pTransducer->m_pInputVoc;
  
	m_pTransitionTable = &(m_pTransducer->m_TransitionTableTarget);
    m_pEmissionTable = &(m_pTransducer->m_EmissionTableTarget);

    m_pTransducerItems = &(m_pTransducer->m_TransducerItems);

    m_NumberOfStates = m_pTransducer->m_NumberOfStatesTarget;
  }

}



CTransducerAccess::~CTransducerAccess()
{
  //delete m_pEmissionTable;
  //delete m_pTransducerItems;
}



//CTransducerState
//CTransducerAccess::Start()
//{
//  // index of first state is 0
//  return 0;
//}



int
CTransducerAccess::Next( CTransducerState State, CWordIndex Word, 
                         CTransducerState &NextState ) const
{
  if ( m_Direction == 0 ) {
    return m_pTransducer->NextSource( State, Word, NextState );
  } else {
    return m_pTransducer->NextTarget( State, Word, NextState );
  }
}



void
CTransducerAccess::DisplayTransitionTable() const
{
  Trace( 1, "CTransducer::DisplayTransitionTable " );

  map< CStateWord, CTransducerState, CompareStateWord >::const_iterator IterState;

  for ( IterState = m_pTransitionTable->begin(); 
        IterState != m_pTransitionTable->end();
        ++IterState ) {

    cout << "State: " << IterState->first.m_State
         << " " << m_pInputVoc->GetWord( IterState->first.m_WordIndex ) 
         << " " << IterState->second << endl;
  }

  cout << endl;

}



void
CTransducerAccess::DisplayEmissionTable() const
{
  Trace( 0, "CTransducer::DisplayEmissionTable " );

  sgi::hash_map< CTransducerState , CNumericalArray< int > >::const_iterator Iter;
  
  cout << "Emission Table:" << endl;
  cout << "---------------" << endl;

  for ( Iter = m_pEmissionTable->begin(); Iter != m_pEmissionTable->end(); ++Iter ) {
    cout << "State = " << Iter->first << "  Index = " << Iter->second << endl;

    CNumericalArray< int > EmissionItems = Iter->second;
    
    for ( int i = EmissionItems.MinIndex(); i <= EmissionItems.MaxIndex(); i++ ) {
      (*m_pTransducerItems)[ EmissionItems[ i ] ]->Display();
    }
  }

  cout << endl;

}



void
CTransducerAccess::Statistics() const
{
  Trace( 1, "CTransducer::Statistics " );

  cout << "Statistics for transducer: " << m_pTransducer->Name() << endl;

  string Name = m_pTransducer->Name();
  //int NumStates = 0;  // first State has number 0
  int NumStates = m_NumberOfStates + 1;  // first State has number 0
  int NumFinalStates = m_pEmissionTable->size();
  int NumTransitions = 0;
  int NumEmissions = 0;
  int NumPatterns = m_pTransducerItems->Size();

  // transitions
  NumTransitions = m_pTransitionTable->size();
  
  // emissions
  sgi::hash_map< CTransducerState , CNumericalArray< int > >::const_iterator Iter;
  for ( Iter = m_pEmissionTable->begin(); Iter != m_pEmissionTable->end(); ++Iter ) {
    CNumericalArray< int > EmissionItems = Iter->second;
    cout << "State = " << Iter->first << "  Index = " << Iter->second 
         << "  Emissions = " << EmissionItems.Size() << endl;
    for ( int i = EmissionItems.MinIndex(); i <= EmissionItems.MaxIndex(); i++ ) {
      cout << "     "; (*m_pTransducerItems)[ EmissionItems[ i ] ]->Display();
    }
    NumEmissions += EmissionItems.Size();
  }
  

  cout << "  Number of States:      " << NumStates << endl;
  cout << "  Number of FinalStates: " << NumFinalStates << endl;
  cout << "  Number of Transitions: " << NumTransitions << endl;
  cout << "  Number of Emissions:   " << NumEmissions << endl;
  cout << "  Number of Patterns:    " << NumPatterns << endl;

  /*
    cout << Name << " \t& " 
    << NumStates << " \t& " 
    << NumTransitions << " \t& " 
    << NumFinalStates << " \t& "
    << NumEmissions << " \\\\"<< endl;
  */

}



//----------------------------------------------------------------------------
CTransducer::CTransducer()
{
  Trace( 1, "CTransducer constructor: init parameters" );

  m_Name = NULL;
  m_File = NULL;

  m_Level            = 1;
  m_Renormalize      = 0;
  m_ShouldBeIterated = false;
  m_Revert           = false; 
  m_LambdaSmooth     = 0.0;

  m_pInputVoc  = NULL;
  m_pOutputVoc = NULL;

  m_BuildMode  = FALSE;

  m_NumberOfStatesSource = 0;
  m_NumberOfStatesTarget = 0;

  m_AllocPatterns = 1000;
  //m_AllocVocSource = 1000;
  //m_AllocVocTarget = 1000;


  m_CostInsFile = NULL;
  m_CostDelFile = NULL;
  m_CostSubstFile = NULL;

  //arv 
  //m_ValuesInFileAreProbabilities = true;

  Param.AddParamDef( "Name", 'N' , &m_Name, "<none>" , 
                     "name of the transducer",
                     True );

  Param.AddParamDef( "File", 'F' , &m_File, NULL , 
                     "name of file with patterns", 
                     True );

  Param.AddParamDef( "AllocPatterns", 'A' , &m_AllocPatterns, 1000, 
                     "for how many patterns should memory be allocated",
                     True );

  Param.AddParamDef( "AllocStates", 'S' , &m_AllocStates, 1000, 
                     "for how many states should memory be allocated",
                     True );

  Param.AddParamDef( "Revert", 'R' , &m_Revert, FALSE,
                     "change source and target",
                     True );

  Param.AddParamDef( "Renormalize", 'r' , &m_Renormalize, 0,
                     "renormalize after loading",
                     True );

  Param.AddParamDef( "Iterate", 'I' , &m_ShouldBeIterated, FALSE, 
                     "transducer should be applied iteratively",
                     True );

  Param.AddParamDef( "LambdaSmooth", 'L' , &m_LambdaSmooth, 0.0, 
                     "smoothing of emmission probabilities",
                     True );

  Param.AddParamDef( "Insertions", 'i' , &m_CostInsFile, NULL,
                     "name of file that contains a list of words and their insertion cost", 
                     True );

  Param.AddParamDef( "Deletions", 'd' , &m_CostDelFile, NULL, 
                     "name of file that contains a list of words and their deletion cost", 
                     True );

  Param.AddParamDef( "Substitutions", 's' , &m_CostSubstFile, NULL , 
                     "name of file that contains pairs of words and their substitution cost", 
                     True );

  Param.AddParamDef( "MaxCostPerWord", 'p', &m_MaxCostPerWord, 0.0,
                     "Maximum allowed costs per word. 0.25 means that one error in four words is allowed",
                     True );

  Param.AddParamDef( "ValuesAreProbs", 'v', &m_ValuesInFileAreProbabilities, 1,
                     "Values in files are probabilites",
                     True );
}



CTransducer::~CTransducer()
{
  for ( int i = m_TransducerItems.MinIndex(); i <= m_TransducerItems.MaxIndex(); i++ ) {
    Trace( 5, "delete transducer item " << i );
    delete m_TransducerItems[ i ];
    m_TransducerItems[ i ] = NULL;
  }
  Trace( 1, "end of CTransducer::~CTransducer" );

  //delete m_Name;
  //delete m_File;
}



bool
CTransducer::Init( CVocabulary *pInputVocabulary, CVocabulary *pOutputVocabulary )
{
  Trace( 1, "CTransducer::Init" );

  m_pInputVoc = pInputVocabulary;
  m_pOutputVoc = pOutputVocabulary;

  CTransducerItem::s_pInputVoc = m_pInputVoc;;
  CTransducerItem::s_pOutputVoc = m_pOutputVoc;;

  return TRUE;
}



bool
CTransducer::LoadParam( const char *FileName )
{
  Trace( 1, "Load Transducer with Parameter file " << FileName);

  // first, load the parameter file
  int Result = Param.read_parameter_file( (char *)FileName );
  if ( Result < 0 ) {
    cerr << "CTransducer LoadParam with " << FileName << " failed ... exiting!!\n";
    // TODO: return false here
    exit(-1);
  }

  Trace( 0, "Init Transducer " << m_Name );
  Trace( 0, "Revert: " << m_Revert );
  Trace( 0, "Renormalize: " << m_Renormalize );
  Trace( 1, "MaxCostPerWord in Init = " << m_MaxCostPerWord );
  Trace( 0, "Smoothing factor = " << m_LambdaSmooth );

  if ( !Load( m_File ) ) {
    return FALSE;
  }

  
  Trace( 1, "Load Cost Files: " << endl << 
	 "Ins-File: " << m_CostInsFile << endl <<
	 "Del-File: " << m_CostDelFile << endl <<
	 "Sub-File: " << m_CostSubstFile << endl );

  //if ( m_CostSubstFile && m_CostDelFile && m_CostInsFile )
  //  m_PhraseTree.LoadCostFiles( m_CostSubstFile, m_CostDelFile, m_CostInsFile );
  
  return TRUE;
}



bool
CTransducer::Load( const char *FileName )
{
  cout << "Load transducer from file " << FileName << endl;

  m_TransducerItems.SetIncrement( m_AllocPatterns );
  m_File = strdup( FileName );

  ifstream fp( m_File );
  if ( !fp.good() ) {
      cerr << "CTransducer::Load(): TransducerFile: " << m_File << " could not be opened for reading" << endl;
      // TODO: return false here
      exit( -1 );
  } // endif: could not open td file
  

  char str [ 1000 ];
  str[ 0 ] = 0;
  char *strptr = str;
  int index;


  assert( m_pOutputVoc );

  Trace( 1, "Load Text: " << m_File );
  char *buffer;
  int I = 1;

  CTransducerItem *pTransducerItem;
  string Source, Target;

  for ( index = 0; !fp.eof(); index++, fp.getline( str, 1000 ) ) {
    Trace( 1, "Pattern: " << index << " " << str );
    //TraceIf( ( index % 1000 == 0 ), 1, index );

    if ( (str[ 0 ] != ';')  && strlen( str ) > 0 ) {
      Trace0( 3, I << " " );
      pTransducerItem = new CTransducerItem;
      pTransducerItem->m_Number = I;
      I++;
      buffer = strtok( strptr, "#" );
      Trace( 2, "Label: " << buffer );
      pTransducerItem->m_Label = buffer;
      Trim( pTransducerItem->m_Label );

      buffer = strtok( NULL, "#" );
      Trace( 2, "Source: " << buffer );
      Source = (const char*)(CString(buffer));
      Trim( Source );
      if ( m_Revert )
        pTransducerItem->m_Target = Source;
      else
        pTransducerItem->m_Source = Source;

      buffer = strtok( NULL, "#" );
      Trace( 2, "Target: " << buffer );
      Target = (const char*)(CString( buffer ));
      Trim( Target );
      if ( Target == "=" ) {
        //Trace( 0, "Short form for target " << str );
        //Trace( 0, "Use: " << Source );
        Target = Source;
      }
      if ( m_Revert ) {
        Trace( 3, "Revert" );
        pTransducerItem->m_Source = Target;
      } else {
        Trace( 3, "Straight" );
        pTransducerItem->m_Target = Target;
      }
	  
      //now there could be a list of scores found after the target
      //the first one will be saved into the count variable for compatibility
      //but will also be put into the newly created m_modelScores array (starting at index 1)
      //note that now we are separating based on the ' ' rather than the hash sign
      buffer = strtok( NULL, " " );
      Trace( 2, "Score: " << buffer );

      int IndexOfScore = 0;
      float Score0 = atof( buffer );
      CNumericalArray< float > &ModelScores = pTransducerItem->m_ModelScores;

      ModelScores[ IndexOfScore ] = Score0;
      pTransducerItem->m_Count = Score0;

      while ( buffer != NULL ) {
        //cout<<"Buffer: "<<buffer<<endl;
        buffer = strtok( NULL, " " );
        if ( buffer != NULL ) {
          IndexOfScore++;
          float Score = atof( buffer );
          ModelScores[ IndexOfScore ] = Score;
        }
      }

      if ( ModelScores[ IndexOfScore ] == 2.718 ) {
        cout << "WARNING: If you are using Pharaoh phrases, please remove the last model score" << endl;
      }
      
      // if scores in file were probabilites, calculate the neg log probs here
      if ( m_ValuesInFileAreProbabilities ) {
        for ( int i = ModelScores.MinIndex(); i <= ModelScores.MaxIndex(); i++ ) {
          ModelScores[ i ] = -log( ModelScores[ i ] );
        }
      }

      //pTransducerItem->Display();
      m_TransducerItems.Append( pTransducerItem );

      //keep track of how many transducer items you have loaded
      if ( index % 100000 == 0 ) {
        cout << "Loaded " <<index << " items" << endl;
      }
    }
    
  }
  

  fp.close();
  cout << "Loaded " << index << " items -- COMPLETED" << endl;

  // build transition table and emission table
  // insert each item into the tables
  Trace( 1, "Build Transition and Emission Tables" );
  BuildSourceTables();
  BuildTargetTables();

  //Renormalize();
  
  //Trace( 0, "Display Transducer" );
  //Display();
  Report();

  return TRUE;
}



void
CTransducer::ClearCounts()
{
  Trace( 0, "CTransducer::ClearCounts()" );

  for ( int i = m_TransducerItems.MinIndex(); i <= m_TransducerItems.MaxIndex(); i++ ) {
    m_TransducerItems[ i ]->m_Count = 0.0;
  }
}



// ??? same factor for all model scores, is this ever used?
void
CTransducer::SetScalingFactors( float factor )
{
  Trace( 0, "CTransducer::ClearCounts()" );

  for ( int i = m_TransducerItems.MinIndex(); i <= m_TransducerItems.MaxIndex(); i++ ) {
    //m_TransducerItems[ i ]->SetScalingFactor() = 0.0;
    m_TransducerItems[ i ]->SetScalingFactor( factor );
  }
}



void
CTransducer::Display()
{
  cout << "Transducer " << m_Name << endl;
  cout << "Label   #   Source  #  Target  #  m_Count  "
       << "; ScoreTargetGivenSource / m_ScoreSourceGivenTarget" << endl;

  for ( int i = m_TransducerItems.MinIndex(); i <= m_TransducerItems.MaxIndex(); i++ ) {
    m_TransducerItems[ i ]->Display();
  }

  cout << endl;

}



void
CTransducer::Report() const
{
  cout << "Transducer Report" << endl;
  cout << "=================" << endl;

  cout << "  Name           = " << m_Name << endl;
  cout << "  File           = " << m_File << endl;
  cout << "  Level          = " << m_Level << endl;
  cout << "  Revert         = "; if ( m_Revert ) cout << "yes"; else cout << "no"; cout << endl;
  cout << "  Renormalize    = "; if ( m_Renormalize ) cout << "yes"; else cout << "no"; cout << endl;
  cout << "  LambdaSmooth   = " << m_LambdaSmooth << endl;
  cout << "  Iterate        = "; if ( m_ShouldBeIterated ) cout << "yes"; else cout << "no"; cout << endl;
  cout << "  ValuesAreProbs = "; if ( m_ValuesInFileAreProbabilities ) cout << "yes"; else cout << "no"; cout << endl;

  cout << "  Number Items   = " << m_TransducerItems.Size() << endl;
  cout << "  Number Source  = " << m_EmissionTableSource.size() << endl;
  cout << "  Number Target  = " << m_EmissionTableTarget.size() << endl;

  /*
  cout << "  TransitionTable Size (Source/Target) = "
       << m_TransitionTableSource.size() << " "
       << m_TransitionTableTarget.size() << endl;

  cout << "  EmissionTable Size (Source/Target)   = " 
       << m_EmissionTableSource.size() << " "
       << m_EmissionTableTarget.size() << endl;
  */

}



bool
CTransducer::Store( const char *pFileName /*= NULL*/ )
{
  CString NewFileName;

  if ( pFileName == NULL ) {
    NewFileName = CString( m_File ) + CString( ".new" );
  } else {
    NewFileName = CString( pFileName );
  }

  Trace( 0, "Store transducer " << m_Name << " in " << NewFileName );

  ofstream File( NewFileName );

  if ( !File ) {
    cerr << "File " << NewFileName << " could not be opened\n";
    return FALSE;
  }

  for ( int i = m_TransducerItems.MinIndex(); i <= m_TransducerItems.MaxIndex(); i++ ) {
    if ( m_Revert ) {
      File << m_TransducerItems[ i ]->m_Label
	   << " # " << m_TransducerItems[ i ]->m_Target
	   << " # " << m_TransducerItems[ i ]->m_Source
	   << " # " << m_TransducerItems[ i ]->m_Count
	   << endl;
    } else {
      File << m_TransducerItems[ i ]->m_Label
	   << " # " << m_TransducerItems[ i ]->m_Source
	   << " # " << m_TransducerItems[ i ]->m_Target
	   << " # " << m_TransducerItems[ i ]->m_Count
	   << endl;
    }
  }

  File.close();

  return TRUE;
}



void
CTransducer::BuildSourceTables()
{

  int Pos, NextPos;
  int Length;  // length of source = number of words
  string Word;
  CWordIndex WordIndex;
  CTransducerState State, NextState;

  Trace( 1, "Build SourceTables" );
  
  m_BuildMode = TRUE;

  for ( int i = m_TransducerItems.MinIndex(); i <= m_TransducerItems.MaxIndex(); i++ ) {

    //Trace( 1, i );
    //Trace( 1, "TransitionTable Bucket count = " << m_TransitionTable.bucket_count() );
    //Trace( 1, "EmissionTable Bucket count = " << m_EmissionTable.bucket_count() );
    //m_TransducerItems[ i ]->Display();

    Pos = 0;
    Length = 0;

    //State = Start();
    State = 0;

    // add rule to the transducer:
    // split source part of the rule into words and insert it
    // thereby going to the next state
    // if (state, word) already in the table, no insertion but
    // only transition
    do {
      NextPos = m_TransducerItems[ i ]->m_Source.find( ' ', Pos );
      Length++;
      Trace( 2, "Segmenting the source of the TD: from byte position " << Pos << " to " << NextPos << " should be a word" );
      if ( NextPos > 0 ) {
        Word = m_TransducerItems[ i ]->m_Source.substr( Pos, NextPos - Pos );
        WordIndex = m_pInputVoc->GetIndex( CString( Word.c_str() ) );
        
        if ( WordIndex == m_pInputVoc->UNK &&
             Word.compare(m_pInputVoc->GetWord(m_pInputVoc->UNK)) != 0 ) {
          // new word
          WordIndex = m_pInputVoc->AddWord(Word);
        } // endif: new word
        
        assert(Word.compare(m_pInputVoc->GetWord(WordIndex)) == 0);

        Trace( 2, "Breaking the source: from byte position " << Pos << " to " << NextPos << " found " 
               << Word << " index=" << WordIndex );
        // insert
        if ( NextSource( State, WordIndex, NextState ) == 0) {
          cout << "CTransducer::BuildSourceTables: Warning: NextSource() returned 0 (whatever this means)" << endl;
        }
      } else {
        // NextPos==-1, there is only one word left. Also, this state is a final state, we can attach emission table to it
        int l = m_TransducerItems[ i ]->m_Source.size();
        Word = m_TransducerItems[ i ]->m_Source.substr( Pos, l );
        WordIndex = m_pInputVoc->GetIndex( CString( Word.c_str() ) );
        
        if ( WordIndex == m_pInputVoc->UNK &&
             Word.compare(m_pInputVoc->GetWord(m_pInputVoc->UNK)) != 0 ) {
          // new word
          WordIndex = m_pInputVoc->AddWord(Word);
        } // endif: new word
        
        assert(Word.compare(m_pInputVoc->GetWord(WordIndex)) == 0);

        Trace( 2, "Segmenting the source: from byte pos:" << Pos << " to " << l << " found word " 
               << Word << " index=" << WordIndex );
        // insert
        if ( NextSource( State, WordIndex, NextState ) == 0) {
          cout << "CTransducer::BuildSourceTables: Warning: NextSource() returned 0 (whatever this means)" << endl;
        } // endif: error (?)
        
        // this leads to a final state
        Trace( 1, "Append to emission table at state " 
               << NextState << " with transducerItem =" << i );
        m_EmissionTableSource[ NextState ].Append( i );
	  
      }
      
      Pos = NextPos + 1;
      State = NextState;
      
    } while ( NextPos >= 0 );
    
    m_TransducerItems[ i ]->m_SourceLength = Length;
    
  }
  
  m_BuildMode = FALSE;
  
  //DisplayTransitionTableSource();
  //DisplayEmissionTableSource();
  
}



CTransducerState
CTransducer::Start()
{
  // index of first state is 0
  return 0;
}



int
CTransducer::NextSource( CTransducerState State, CWordIndex Word, 
                         CTransducerState &NextState )
{

  Trace( 1, "CTransducer::NextSource " << State << " " << Word );

  // if in BuildMode we make new entry into transition table
  // otherwise: error

  CStateWord StateWord( State, Word );

  map< CStateWord, CTransducerState, CompareStateWord >::const_iterator IterState;
  IterState = m_TransitionTableSource.find( StateWord );

  if ( IterState == m_TransitionTableSource.end() ) {
    if ( m_BuildMode ) {
      m_NumberOfStatesSource++;
      NextState = m_NumberOfStatesSource;
      m_TransitionTableSource[ StateWord ] = NextState;
      Trace( 1, "Insert " << State 
             << " " << m_pInputVoc->GetWord( Word ) 
             << " " << NextState );
      
    } else {
      //cerr << "ERROR: in CTransducer::Next" << endl
      //   << "ERROR: State " << State << " is not in the transition table" << endl;
      NextState = s_NotFound;
      return 0;
    }
  } else {
    Trace( 1, "Found " << m_pInputVoc->GetWord( IterState->first.m_WordIndex ) 
           << " leading from state " << IterState->first.m_State 
           << " to state " << " " << IterState->second );
    NextState = IterState->second;
  }

  return 1;
  
}



void
CTransducer::BuildTargetTables()
{

  int Pos, NextPos;
  int Length;  // length of source = number of words
  string Word;
  CWordIndex WordIndex;
  CTransducerState State, NextState;

  Trace( 1, "Build TargetTables" );

  m_BuildMode = TRUE;

  for ( int i = m_TransducerItems.MinIndex(); i <= m_TransducerItems.MaxIndex(); i++ ) {
    //Trace( 1, i );
    //Trace( 1, "TransitionTable Bucket count = " << m_TransitionTable.bucket_count() );
    //Trace( 1, "EmissionTable Bucket count = " << m_EmissionTable.bucket_count() );
    //m_TransducerItems[ i ]->Display();

    Pos = 0;
    Length = 0;

    //State = Start();
    State = 0;

    // add rule to the transducer:
    // split source part of the rule into words and insert it
    // thereby going to the next state
    // if (state, word) already in the table, no insertion but
    // only transition
    do {
      NextPos = m_TransducerItems[ i ]->m_Target.find( ' ', Pos );
      Length++;
      Trace( 2, "from " << Pos << " to " << NextPos );
      if ( NextPos > 0 ) {
        Word = m_TransducerItems[ i ]->m_Target.substr( Pos, NextPos - Pos );
        WordIndex = m_pInputVoc->GetIndex( CString( Word.c_str() ) );

        if ( WordIndex == m_pInputVoc->UNK &&
             Word.compare(m_pInputVoc->GetWord(m_pInputVoc->UNK)) != 0 ) {
          // new word
          WordIndex = m_pInputVoc->AddWord(Word);
        } // endif: new word

        assert(Word.compare(m_pInputVoc->GetWord(WordIndex)) == 0);

        Trace( 2, "from " << Pos << " to " << NextPos << " " 
               << Word << " " << WordIndex );
        // insert
        if ( NextTarget( State, WordIndex, NextState ) == 0 ) {
          cout << "CTransducer::BuildTargetTables: Warning: NextTarget() returned 0 (whatever this means!)" << endl;
        } // endif: error (?)
        
      } else {
        int l = m_TransducerItems[ i ]->m_Target.size();
        Word = m_TransducerItems[ i ]->m_Target.substr( Pos, l );
        WordIndex = m_pInputVoc->GetIndex( CString( Word.c_str() ) );

        if ( WordIndex == m_pInputVoc->UNK &&
             Word.compare(m_pInputVoc->GetWord(m_pInputVoc->UNK)) != 0 ) {
          // new word
          WordIndex = m_pInputVoc->AddWord(Word);
        } // endif: new word

        assert(Word.compare(m_pInputVoc->GetWord(WordIndex)) == 0);

        Trace( 2, "from " << Pos << " to " << l << " " 
               << Word << " " << WordIndex );
        // insert
        if ( NextTarget( State, WordIndex, NextState ) == 0 ) {
          cout << "CTransducer::BuildTargetTables: Warning: NextTarget() returned 0 (whatever this means!)" << endl;
        } // endif: error (?)
        
        // this leads to a final state
        Trace( 1, "Append to emission table at state " 
               << NextState << " " << i );
        m_EmissionTableTarget[ NextState ].Append( i );
      }
      
      Pos = NextPos + 1;
      State = NextState;

    } while ( NextPos >= 0 );
    m_TransducerItems[ i ]->m_TargetLength = Length;

  }

  m_BuildMode = FALSE;

  //DisplayTransitionTableTarget();
  //DisplayEmissionTableTarget();

}



int
CTransducer::NextTarget( CTransducerState State, CWordIndex Word, 
                         CTransducerState &NextState )
{
  Trace( 1, "CTransducer::NextTarget " << State << " " << Word );

  // if in BuildMode we make new entry into transition table
  // otherwise: error
  
  CStateWord StateWord( State, Word );

  map< CStateWord, CTransducerState, CompareStateWord >::const_iterator IterState;
  IterState = m_TransitionTableTarget.find( StateWord );

  if ( IterState == m_TransitionTableTarget.end() ) {
    if ( m_BuildMode ) {
      m_NumberOfStatesTarget++;
      NextState = m_NumberOfStatesTarget;
      m_TransitionTableTarget[ StateWord ] = NextState;
      Trace( 1, "Insert " << State 
             << " " << m_pInputVoc->GetWord( Word ) 
             << " " << NextState );
      
    } else {
      //cerr << "ERROR: in CTransducer::Next" << endl
      //   << "ERROR: State " << State << " is not in the transition table" << endl;
      NextState = s_NotFound;
      return 0;
    }
  } else {
    Trace( 1, "Found " << IterState->first.m_State 
           << " " << m_pInputVoc->GetWord( IterState->first.m_WordIndex ) 
           << " " << IterState->second );
    NextState = IterState->second;
  }
  
  return 1;
  
}



void 
CTransducer::Trim( string &S )
{
  while( S[ 0 ] == ' ' ) {
    S.erase( 0, 1 );
  }

  while( S[ S.size() - 1 ] == ' ' ) {
    S.erase( S.size() - 1, 1 );
  }
}


void
CTransducer::Renormalize()
{
  Trace( 1, "CTransducer::Renormalize " );
  
  // first we run over source tree emissions and estimate  
  // probabilities for target given source, 
  // then we run over target tree emissions and estimate
  // probabilities for source given target
  if ( m_Renormalize ) {
    Trace( 1, "Renormalize Transducer " << m_Name );
    RenormalizeSource();
    RenormalizeTarget();
  } else {
    Trace( 1, "Don't Renormalize Transducer" << m_Name );
    CopyCountToPrSource();
    CopyCountToPrTarget();
  }

#ifdef TRACE_ENABLED
  if ( TraceLevel > 0 ) {
    Display();
  }
#endif
}



void
CTransducer::RenormalizeSource()
{
  Trace( 1, "CTransducer::RenormalizeSource " );

  const char * Label;
  float Count;

  sgi::hash_map< CTransducerState , CNumericalArray< int > >::const_iterator Iter;

  for ( Iter = m_EmissionTableSource.begin(); Iter != m_EmissionTableSource.end(); ++Iter ) {
    Trace( 2, "State = " << Iter->first << "  Index = " << Iter->second );

    CNumericalArray< int > EmissionItems = Iter->second;

    sgi::hash_map< const char *, float, sgi::hash<const char*>, EqStr > SumForLabel;
    //hash_map< const char *, float, hash< const char* >, EqStr >::const_iterator IterLabel;

    //cout << "Map size = " << SumForLabel.size() << endl;

    for ( int i = EmissionItems.MinIndex(); i <= EmissionItems.MaxIndex(); i++ ) {
      //m_TransducerItems[ EmissionItems[ i ] ]->Display();
      Label = m_TransducerItems[ EmissionItems[ i ] ]->m_Label.c_str();
      Count = SumForLabel[ Label ];
      Trace( 3, "Map size = " << SumForLabel.size() );
      Trace( 3, "Current count for " << Label << " is " << Count );
      // total sum of new counts; add 0.01 to each count for smoothing and to avoid zero sum
      if ( m_TransducerItems[ EmissionItems[ i ] ]->m_Count < 10e-6 ) {
      	SumForLabel[ Label ] +=
      	  ( m_TransducerItems[ EmissionItems[ i ] ]->m_Count + 0.01 );
      } else {
        SumForLabel[ Label ] += m_TransducerItems[ EmissionItems[ i ] ]->m_Count;
        Trace( 3, "Map size = " << SumForLabel.size() );
        Trace( 3, "Add " << m_TransducerItems[ EmissionItems[ i ] ]->m_Count );
      }
      Trace( 3, "Emission " << i << "  Label: " << Label );
      Trace( 3, "  Sum: " << SumForLabel[ Label ] );
      
      //Trace( 3, "Test map with current size = " << SumForLabel.size() );
      //for ( IterLabel = SumForLabel.begin(); IterLabel != SumForLabel.end(); ++IterLabel ) {
      //cout << "Label " << IterLabel->first << " Sum = " << IterLabel->second << endl;
      //}
    }

    //cout << "Map size = " << SumForLabel.size() << endl;
    //for ( IterLabel = SumForLabel.begin(); IterLabel != SumForLabel.end(); ++IterLabel ) {
    //  cout << "Label " << IterLabel->first << " Sum = " << IterLabel->second << endl;
    //  assert( IterLabel->second > 0 );
    //}

    //Trace( 2, "Sum of counts = " << Sum );
    // now redistribute 
    Trace( 2, "Redistribute" );
    for ( int i = EmissionItems.MinIndex(); i <= EmissionItems.MaxIndex(); i++ ) {
      //m_TransducerItems[ EmissionItems[ i ] ]->Display();
      //for ( IterLabel = SumForLabel.begin(); IterLabel != SumForLabel.end(); ++IterLabel ) {
      //cout << "Label " << IterLabel->first << " Sum = " << IterLabel->second << endl;
      //assert( IterLabel->second > 0 );
      //}
      Label = m_TransducerItems[ EmissionItems[ i ] ]->m_Label.c_str();
      Trace( 3, "Label: " << Label << "  Sum: " << SumForLabel[ Label ] );
      assert( SumForLabel[ Label ] > 0 );
      if (m_TransducerItems[ EmissionItems[ i ] ]->m_Count < 10e-6 ) {
        m_TransducerItems[ EmissionItems[ i ] ]->m_ScoreTargetGivenSource
          = 0.01 / SumForLabel[ Label ];
      } else {
        m_TransducerItems[ EmissionItems[ i ] ]->m_ScoreTargetGivenSource 
          = ( 1 - m_LambdaSmooth )
          * m_TransducerItems[ EmissionItems[ i ] ]->m_Count
          / SumForLabel[ Label ];
      }
      //m_TransducerItems[ EmissionItems[ i ] ]->Display();
    }
    
  }
  //cout << endl;
  //DisplayEmissionTable();
  
}



void
CTransducer::RenormalizeTarget()
{
  Trace( 1, "CTransducer::RenormalizeTarget " );

  const char * Label;
  float Count;

  sgi::hash_map< CTransducerState , CNumericalArray< int > >::const_iterator Iter;

  for ( Iter = m_EmissionTableTarget.begin(); Iter != m_EmissionTableTarget.end(); ++Iter ) {
    Trace( 2, "State = " << Iter->first << "  Index = " << Iter->second );
    
    CNumericalArray< int > EmissionItems = Iter->second;

    sgi::hash_map< const char *, float, sgi::hash<const char*>, EqStr > SumForLabel;
    //hash_map< const char *, float, hash< const char* >, EqStr >::const_iterator IterLabel;

    //cout << "Map size = " << SumForLabel.size() << endl;

    for ( int i = EmissionItems.MinIndex(); i <= EmissionItems.MaxIndex(); i++ ) {
      //m_TransducerItems[ EmissionItems[ i ] ]->Display();
      Label = m_TransducerItems[ EmissionItems[ i ] ]->m_Label.c_str();
      Count = SumForLabel[ Label ];
      //cout << "Map size = " << SumForLabel.size() << endl;
      //cout << "Current count for " << Label << " is " << Count << endl;
      // total sum of new counts; add 0.01 to each count for smoothing and to avoid zero sum
      if ( m_TransducerItems[ EmissionItems[ i ] ]->m_Count < 10e-6 ) {
        SumForLabel[ Label ] += 
          ( m_TransducerItems[ EmissionItems[ i ] ]->m_Count + 0.01 );
      } else {
        SumForLabel[ Label ] += m_TransducerItems[ EmissionItems[ i ] ]->m_Count;
        //cout << "Map size = " << SumForLabel.size() << endl;
        //cout << "Add " << m_TransducerItems[ EmissionItems[ i ] ]->m_Count << endl;
      }
      Trace( 3, "Emission " << i << "  Label: " << Label );
      Trace( 3, "  Sum: " << SumForLabel[ Label ] );
      
      //cout << "Test map with current size = " << SumForLabel.size() << endl;
      //for ( IterLabel = SumForLabel.begin(); IterLabel != SumForLabel.end(); ++IterLabel ) {
      //Trace( 0, "Label: " << Label << "  Sum: " << SumForLabel[ Label ] );
      //}
    }
    
    //cout << "Map size = " << SumForLabel.size() << endl;
    //for ( IterLabel = SumForLabel.begin(); IterLabel != SumForLabel.end(); ++IterLabel ) {
    //cout << "Label " << IterLabel->first << " Sum = " << IterLabel->second << endl;
    //assert( IterLabel->second > 0 );
    //}
    
    //Trace( 2, "Sum of counts = " << Sum );
    // now redistribute 
    Trace( 2, "Redistribute" );
    for ( int i = EmissionItems.MinIndex(); i <= EmissionItems.MaxIndex(); i++ ) {
      //for ( IterLabel = SumForLabel.begin(); IterLabel != SumForLabel.end(); ++IterLabel ) {
      //cout << "Label " << IterLabel->first << " Sum = " << IterLabel->second << endl;
      //assert( IterLabel->second > 0 );
      //}
      Label = m_TransducerItems[ EmissionItems[ i ] ]->m_Label.c_str();
      Trace( 3, "Label: " << Label << "  Sum: " << SumForLabel[ Label ] );
      assert( SumForLabel[ Label ] > 0 );
      if (m_TransducerItems[ EmissionItems[ i ] ]->m_Count < 10e-6 ) {
        m_TransducerItems[ EmissionItems[ i ] ]->m_ScoreSourceGivenTarget
          = 0.01 / SumForLabel[ Label ];
      } else {
        m_TransducerItems[ EmissionItems[ i ] ]->m_ScoreSourceGivenTarget
          = ( 1 - m_LambdaSmooth )
          * m_TransducerItems[ EmissionItems[ i ] ]->m_Count
          / SumForLabel[ Label ];
      }
      //m_TransducerItems[ EmissionItems[ i ] ]->Display();
    }
    
  }

  //cout << endl;
  //DisplayEmissionTable();
  
}



void
CTransducer::CopyCountToPrSource()
{
  // if transducers have scores which should be used as probabilities
  // renormalization is to be avoided.  E.g. use lexicon probabilities
  // to calculate phrase-to-phrase probabilies.

  Trace( 1, "CTransducer::CopyCountToSource " );

  sgi::hash_map< CTransducerState , CNumericalArray< int > >::const_iterator Iter;

  for ( Iter = m_EmissionTableSource.begin(); Iter != m_EmissionTableSource.end(); ++Iter ) {
    Trace( 2, "State = " << Iter->first << "  Index = " << Iter->second );

    CNumericalArray< int > EmissionItems = Iter->second;

    Trace( 2, "Copy" );
    for ( int i = EmissionItems.MinIndex(); i <= EmissionItems.MaxIndex(); i++ ) {
      //m_TransducerItems[ EmissionItems[ i ] ]->Display();
      if (m_TransducerItems[ EmissionItems[ i ] ]->m_Count < 10e-20 ) {
        m_TransducerItems[ EmissionItems[ i ] ]->m_ScoreTargetGivenSource = 10e-20;
      } else {
        m_TransducerItems[ EmissionItems[ i ] ]->m_ScoreTargetGivenSource 
          = ( 1 - m_LambdaSmooth ) * m_TransducerItems[ EmissionItems[ i ] ]->m_Count;
      }
      //m_TransducerItems[ EmissionItems[ i ] ]->Display();
    }
    
  }
  //cout << endl;
  //DisplayEmissionTable();
  
}



void
CTransducer::CopyCountToPrTarget()
{
  Trace( 1, "CTransducer::CopyCountToScoreTarget " );

  sgi::hash_map< CTransducerState , CNumericalArray< int > >::const_iterator Iter;

  for ( Iter = m_EmissionTableTarget.begin(); Iter != m_EmissionTableTarget.end(); ++Iter ) {
    Trace( 2, "State = " << Iter->first << "  Index = " << Iter->second );
    
    CNumericalArray< int > EmissionItems = Iter->second;

    Trace( 2, "Copy" );
    for ( int i = EmissionItems.MinIndex(); i <= EmissionItems.MaxIndex(); i++ ) {
      if (m_TransducerItems[ EmissionItems[ i ] ]->m_Count < 10e-20 ) {
        m_TransducerItems[ EmissionItems[ i ] ]->m_ScoreSourceGivenTarget = 10e-20;
      } else {
        m_TransducerItems[ EmissionItems[ i ] ]->m_ScoreSourceGivenTarget
          = ( 1 - m_LambdaSmooth ) * m_TransducerItems[ EmissionItems[ i ] ]->m_Count;
      }
      //m_TransducerItems[ EmissionItems[ i ] ]->Display();
    }
    
  }

  //cout << endl;
  //DisplayEmissionTable();
  
}



void
CTransducer::RenormalizePr()
{
  Trace( 1, "CTransducer::Renormalize from Prob" );

  cerr << "CTransducer::RenormalizePr is not active, check code" << endl;
  assert( 0 );

  /*
  //float Sum = 0.0;
  hash_map< const char *, float, hash< const char * >, eqstr > SumForLabel;
  string Label;

  sgi::hash_map< CTransducerState , CNumericalArray< int > >::const_iterator Iter;
  
  for ( Iter = m_EmissionTable.begin(); Iter != m_EmissionTable.end(); ++Iter ) {
    //cout << "State = " << Iter->first << "  Index = " << Iter->second << endl;

    CNumericalArray< int > EmissionItems = Iter->second;
    //Sum = 0.0;
    SumForLabel.clear();

    for ( int i = EmissionItems.MinIndex(); i <= EmissionItems.MaxIndex(); i++ ) {
      //m_TransducerItems[ EmissionItems[ i ] ]->Display();
      // total sum of new counts; add 0.1 to each count for smoothing and to avoid zero sum
      m_TransducerItems[ EmissionItems[ i ] ]->m_Score += 0.1;
      //m_TransducerItems[ EmissionItems[ i ] ]->Display();
      //Sum += m_TransducerItems[ EmissionItems[ i ] ]->m_Score;
      Label = m_TransducerItems[ EmissionItems[ i ] ]->m_Label;
      //Trace( 0, SumForLabel[ Label.c_str() ] );
      SumForLabel[ Label.c_str() ] += m_TransducerItems[ EmissionItems[ i ] ]->m_Score;
      Trace( 2, "Label: " << Label << "  Sum: " << SumForLabel[ Label.c_str() ] );
    }
    Trace( 2, "Number of Emissions = " << EmissionItems.Size() );
    //Trace( 2, "Sum of counts = " << Sum );

    // now redistribute 
    for ( int i = EmissionItems.MinIndex(); i <= EmissionItems.MaxIndex(); i++ ) {
      Label = m_TransducerItems[ EmissionItems[ i ] ]->m_Label;
      //m_TransducerItems[ EmissionItems[ i ] ]->m_Score /= Sum;
      m_TransducerItems[ EmissionItems[ i ] ]->m_Score /= SumForLabel[ Label.c_str() ];
      m_TransducerItems[ EmissionItems[ i ] ]->m_Score *= ( 1 - m_LambdaSmooth );
      //m_TransducerItems[ EmissionItems[ i ] ]->Display();
    }
    
    //cout << endl;
  }

  //cout << endl;

  //DisplayEmissionTable();

  */
}



void
CTransducer::Init()
{
  /*
  Trace( 1, "CTransducer::Init " );

  hash_map< CTransducerState , CNumericalArray< int > >::const_iterator Iter;

  for ( Iter = m_EmissionTable.begin(); Iter != m_EmissionTable.end(); ++Iter ) {
    CNumericalArray< int > EmissionItems = Iter->second;
    for ( int i = EmissionItems.MinIndex(); i <= EmissionItems.MaxIndex(); i++ ) {
      m_TransducerItems[ EmissionItems[ i ] ]->m_Score = 1;
    }
  }
  */
}




