//           PhrasalSRLInfo.cxx
//  Sat Jul 25 19:59:57 2009
//  Copyright  2009  qing
//  <qing@<host>>

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Library General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA

#include "PhrasalSRLInfo.h"
#include "SimpleLogger.h"

namespace qformat{

	 // Get the predicate information
	const CQPhrasalSRLInfo::_QPredicateInfo& 
		CQPhrasalSRLInfo::GetPredicateInfo(int index) const{
		A_FATAL(index < 0  || index >= predicates.size() );
		return predicates[index];
    }

	// Get the argument information
	const CQPhrasalSRLInfo::_QArgumentInfo& 
		CQPhrasalSRLInfo::GetArgumentInfo(int index) const{
		A_FATAL(index < 0  || index >= arguments.size() );
		return arguments[index];
	}

	CQPhrasalSRLInfo::CQPhrasalSRLInfo
		(const CQSRLSentence& sent, int start, int end){
		build_phrase_info(sent,start,end);
	}

	void CQPhrasalSRLInfo::build_phrase_info
		(const CQSRLSentence& sent, int start, int end){
		// Navigate all the predicate
		for(size_t i = 0; i< sent.GetPredicateCount(); i++){
			const CQSRLPredicate & pred = sent.GetPredicate(i);
			size_t idx = pred.GetIndex();
			if (idx >=start && idx < end){ // Provides predicate
				predicates.resize(predicates.size()+1);
				_QPredicateInfo &pinfo = predicates[predicates.size()-1];
				pinfo.Count = 1;
				pinfo.Predicate = pred.GetPredicate();
				// Navigate through all its arguments
				for(size_t j = 0 ; j < pred.GetArgumentCount(); j++){
					_QArgumentInfo* arg = add_augument_info(
					     pinfo.Predicate.c_str(),
                         pred.GetArgumentType(j).c_str(),
					     idx, pred.GetArgumentBegin(j),pred.GetArgumentEnd(j),
					     start,end);
					if (arg == NULL){ // No overlap
						pinfo.NumMissingArguments++;
						if( pred.GetArgumentBegin(j) < idx)
							pinfo.NumMissingArgumentsBefore++;
						else
							pinfo.NumMissingArgumentsAfter++;
					}else{
						if (arg->Completeness == _QArgumentInfo::CP_COMPLETE){
							pinfo.NumCompleteArguments++;
						}else if (arg->Completeness 
						          == _QArgumentInfo::CP_INCPL_BEFORE){
							pinfo.NumIncompleteArguments++;
							pinfo.NumIncompleteArugmentsBefore++;
						}else if (arg->Completeness 
						          == _QArgumentInfo::CP_INCPL_AFTER){
							pinfo.NumIncompleteArguments++;
							pinfo.NumIncompleteArugmentsAfter++;
						}else if (arg->Completeness 
						          == _QArgumentInfo::CP_INCPL_BOTH){
							pinfo.NumIncompleteArguments++;
							pinfo.NumIncompleteArugmentsAfter++;
							pinfo.NumIncompleteArugmentsBefore++;
						}else if (arg->Completeness 
						          == _QArgumentInfo::CP_COMPLETE){
							pinfo.NumCompleteArguments++;
						}
					}
					pinfo.NumArguments++;
				}
			}//if (idx >=start && idx < end)
			else{
				for(size_t j = 0 ; j < pred.GetArgumentCount(); j++){
					add_augument_info(pred.GetPredicate().c_str(),
				                  pred.GetArgumentType(j).c_str(),
				                  idx,
				                  pred.GetArgumentBegin(j),
				                  pred.GetArgumentEnd(j),
				                  start,end
				                  );
				}
			}
		}
	}
	
	CQPhrasalSRLInfo::_QArgumentInfo* 
		CQPhrasalSRLInfo::add_augument_info(const char* pred,
	       const char* label,
	       size_t predidx, size_t argstart, 
	       size_t argend, size_t pstart, size_t pend){
		int pstat;
		if(predidx < pend && predidx >= pstart){
			pstat = CQPhrasalSRLInfo::_QArgumentInfo::PL_PRESENT;
		}else if (predidx < pstart ){
			pstat = CQPhrasalSRLInfo::_QArgumentInfo::PL_MISS_BEFORE;
		}else if (predidx >= pend){
			pstat = CQPhrasalSRLInfo::_QArgumentInfo::PL_MISS_AFTER;
		}
		int cpl;
		if(argstart >= pstart && argend <= pend){
			cpl = CQPhrasalSRLInfo::_QArgumentInfo::CP_COMPLETE;
		}else if (argstart < pstart && argend > pstart && argend <=pend){
			cpl = CQPhrasalSRLInfo::_QArgumentInfo::CP_INCPL_BEFORE;
		}else if (argstart < pend && argstart>=pstart && argend > pend){
			cpl = CQPhrasalSRLInfo::_QArgumentInfo::CP_INCPL_AFTER;			
		}else if (argstart<pstart && argend > pend){
			cpl = CQPhrasalSRLInfo::_QArgumentInfo::CP_INCPL_BOTH;						
		}else{
			cpl = CQPhrasalSRLInfo::_QArgumentInfo::CP_NOT_OVLP; // no overlap	
		}
		if (cpl == CQPhrasalSRLInfo::_QArgumentInfo::CP_NOT_OVLP){ // no need to add
			return NULL;
		}else{
			arguments.resize(arguments.size()+1);
			_QArgumentInfo & arg = arguments[arguments.size()-1];
			arg.Predicate = pred;
			arg.PredicateLocation = pstat;
			arg.Completeness = cpl;
			arg.Label = label;
			arg.Count = 1;
			return &arg;
		}
    }
}





