/**
 * \file Sentence.h 
 * \author Qin Gao
 * \date  07-24-09
 * \brief The definition of qformat::CQSentence and various classes
 * CQSentence is a representation of sentence, actually an array of string
 */

/***************************************************************************
 *            Sentence.h
 *
 *  Thu Jul 23 23:24:16 2009
 *  Copyright  2009  qing
 *  <qing@<host>>
 ****************************************************************************/

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
 */

/**
 History:

 07-24-09: Created class and added functionalities (interface not fixed)
 */

/**
 The sentence format, it is simply an array of tokens (word)
*/

#ifndef _QFORMAT_SENTENCE_H__
#define _QFORMAT_SENTENCE_H__

#include "Word.h"
#include <vector>

namespace qformat{
	/*!
	 Sentence representation
	 */
	class CQSentence{
	public:
			CQSentence(){};
			virtual ~CQSentence(){};
	public:
			/*!
			 Get the length of the sentence, in words
			 */
			virtual size_t GetLength() const;

			/*!
			 Get the word in const 
			 */
			virtual const CQWord& GetWord(size_t idx) const;

			/*!
			 Add a word
			 */
			virtual void AddWord(const char* rep);
	private:
			std::vector<CQWord> words;
	};
}

#endif


