//           SwirlSRLFormat.cxx
//  Sat Jul 25 00:43:08 2009
//  Copyright  2009  qing
//  <qing@<host>>

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Library General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA

#include "Format.h"
#include "SwirlSRLFormat.h"
#include "SimpleLogger.h"

#include <string>
#include <boost/algorithm/string.hpp>
#include <IOUtils.h>

using namespace std;
using namespace boost;

namespace qformat{


/////////////////////////////////////////////////////////////////////
//	CQSRLPredicate
/////////////////////////////////////////////////////////////////////
	const std::string& CQSRLPredicate::GetArgumentType(size_t idx) const{
		A_FATAL(idx>=0 && idx < argument_types.size());
		return argument_types[idx];
	}

	size_t CQSRLPredicate::GetArgumentBegin(size_t idx) const{
		A_FATAL(idx>=0 && idx <argument_begins.size());
		return argument_begins[idx];
	}

	size_t CQSRLPredicate::GetArgumentEnd(size_t idx) const{
		A_FATAL(idx>=0 && idx <argument_ends.size());
		return argument_ends[idx];
	}

	size_t CQSRLPredicate::GetArgumentLength(size_t idx) const{
		A_FATAL(idx>=0 && idx <argument_ends.size());
		return argument_ends[idx]-argument_begins[idx];;
	}

	CQSRLPredicate::CQSRLPredicate(const char* pred,size_t idx){
		A_FATAL(pred!=NULL);		
		predicate = pred;
		index = idx;
	}

	CQSRLPredicate::CQSRLPredicate(const CQSRLPredicate& r){
		predicate = r.predicate;
		index = r.index;
		argument_types = r.argument_types;
		argument_begins = r.argument_begins;
		argument_ends = r.argument_ends;
	}

	void CQSRLPredicate::AddArgument(const char* type,size_t begin,size_t end){
		A_FATAL(type!=NULL);
		C_ERROR(begin<end);

		argument_types.push_back(string(type));
		argument_begins.push_back(begin);
		argument_ends.push_back(end);

		A_ERROR(argument_types.size()==argument_begins.size() 
		        && argument_begins.size() == argument_ends.size());

	}

	void CQSRLPredicate::RemoveArgument(size_t idx){
		P_ERROR("Not implemented");
		//A_FATAL(idx>=0 && idx <argument_begins.size());
		//argument_types.erase(idx);
		//argument_begins.erase(idx);
		//argument_ends.erase(idx);
	}

	const CQSRLPredicate& CQSRLPredicate::operator=(const CQSRLPredicate &r){
		predicate = r.predicate;
		index = r.index;
		argument_types = r.argument_types;
		argument_begins = r.argument_begins;
		argument_ends = r.argument_ends;
		return *this;
	}

///////////////////////////////////////////////////////////////////////////////
//CQSRLSentence
///////////////////////////////////////////////////////////////////////////////

	const CQSRLPredicate& CQSRLSentence::GetPredicate(size_t idx) const{
		A_FATAL(idx>=0 && idx <predicates.size());
		return predicates[idx];
	}

	CQSRLPredicate& CQSRLSentence::GetPredicate(size_t idx){
		A_FATAL(idx>=0 && idx <predicates.size());
		return predicates[idx];
	}

	CQSRLPredicate& CQSRLSentence::AddPredicate(const char* pred,size_t idx){
		A_FATAL(pred!=NULL);
		predicates.push_back(CQSRLPredicate(pred,idx));
		return predicates[predicates.size()-1];
	}
//////////////////////////////////////////////////////////////////////////////
//CQSwirlSRLFormat
//////////////////////////////////////////////////////////////////////////////

	set<TQSerialNumber> CQSwirlSRLFormat::supported_types;

	const std::set<TQSerialNumber>& CQSwirlSRLFormat::GetSupportTypes() const{
		if (supported_types.size()==0){
			supported_types.insert(TS_SwirlSRLSentence);
		}
		return CQSwirlSRLFormat::supported_types;
	}

	boost::shared_ptr<CQObject> CQSwirlSRLFormat::ReadNext(){
		return next;
	}

	bool CQSwirlSRLFormat::HasNext(){
		// do actual reading, create a new shared_ptr to the object, and 
		// store it on "next" variable
		istream & input = getStream();
		if(!input){
			P_DEBUG_VERBOSE(5,"Stream eof or not open, does not have next");
			return false;
		}
		vector<string> lines;
		// Read lines until empty line
		qutil::io::ReadLinesEndsEmpty(input,lines);
		if(!lines.size()) return false;
		// Parse, how many predicates?
		vector<vector<string> > tokens;
		tokens.resize(lines.size());

		for(int i = 0; i< lines.size() ; i++){
			split(tokens[i],lines[i],is_space(),token_compress_on);			
		}

		size_t nPredicate = tokens[0].size()-1;

		D_DEBUG(DINT,tokens.size());
    	D_DEBUG(DINT,tokens[0].size());
		
		CQSRLSentence *sentence = new CQSRLSentence(lines.size());

		size_t nRealPredicates = 0;
		// Add predicates
		for(int i = 0; i< tokens.size() ; i++){
			D_DEBUG(DINT,i);
			D_DEBUG(DSTRING,tokens[i][0].c_str());
			if (tokens[i][0] != "-") { // predicate
				CQSRLPredicate & p =
					sentence->AddPredicate(tokens[i][0].c_str(),i);
				nRealPredicates ++;
				D_DEBUG(DINT,nRealPredicates);
				C_WARN(nRealPredicates <= nPredicate);
				string cls = "";
				size_t cls_begin = -1, cls_end = -1;
				for (int j = 0 ; j < tokens.size() ; j++){
					D_DEBUG(DSTRING,tokens[j][nRealPredicates].c_str());
					if(tokens[j][nRealPredicates][0] == '('){
						C_WARN(
						       tokens[j][nRealPredicates]
						       	[tokens[j][nRealPredicates].length()-1] == '*'
						       ||
						       tokens[j][nRealPredicates]
						       	[tokens[j][nRealPredicates].length()-1] == ')'
						       );
						C_ERROR(cls=="");
						int v = 2;
						if ( tokens[j][nRealPredicates]
						       	[tokens[j][nRealPredicates].length()-1] == ')')
								   v++;
						cls = tokens[j][nRealPredicates].substr(1,
  		                       tokens[j][nRealPredicates].length()-v);
						cls_begin = j;
					}
					if (tokens[j][nRealPredicates]
					          [tokens[j][nRealPredicates].length()-1] == ')'){
						C_WARN(tokens[j][nRealPredicates][0] == '*'
						       ||tokens[j][nRealPredicates][0] == '(');
						C_ERROR(cls!="");
						cls_end = j+1;
						p.AddArgument(cls.c_str(),cls_begin,cls_end);
						cls="";
					}
				}// end for int j =0
				C_ERROR(cls=="");
			}// end if tokens[i][0]!="-"
		}// end for(int i =0)
		
		C_ERROR(nRealPredicates == nPredicate);
		// Wrap
		next = boost::shared_ptr<CQObject>(sentence);

		return true;
	}
}






















