/**
 * \file Word.h 
 * \author Qin Gao
 * \date  07-24-09
 * \brief The definition of qformat::CQWord and various classes
 * CQWord is a representation of word, it contains the string representation
 * of a word and its common operations.
 * The interface is not fixed yet
 * The CQWord representation is specially used within sentence, for those
 * word in dictionary, there will be different special representations.
 */

/***************************************************************************
 *            Word.h
 *
 *  Thu Jul 23 23:57:45 2009
 *  Copyright  2009  qing
 *  <qing@<host>>
 ****************************************************************************/

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
 */

#ifndef _QFORMAT_WORD_H_
#define _QFORMAT_WORD_H_
#include <string>

namespace qformat{

	 class CQSentence;

	 class CQWord{
	 public:
			 /// Get the string representation of the word
			 inline const char* GetString() const{return word.c_str();};
			 /// Get the sentence that the word belongs to
			 inline const CQSentence& GetSentence() const{return *c;};
			 /// Get the index the word in the sentence
			 inline size_t GetIndex() const{return idx;};

	 public:
			 ~CQWord(){};
			 /// Creating a new _QWord
			 CQWord(const char* c, const CQSentence* s, size_t index);
			 CQWord(const CQWord& rep);

	 public:
			 const CQWord& operator=(const CQWord& rep);
	 private:
			 std::string word;
			 size_t idx;
			 const CQSentence* c;
	 };

}

#endif
















