/***************************************************************************
 *            BasePhraseExtraction.h
 *
 *  Wed Jul 22 14:11:27 2009
 *  Copyright  2009  qing
 *  <qing@<host>>
 ****************************************************************************/

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
 */

#ifndef __BASE_PHRASE_EXTRACTION_
#define __BASE_PHRASE_EXTRACTION_

#include <cstdio>
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <stdlib.h>
#include <assert.h>
#include <time.h>
#include <cstring>


#ifndef _BASE_PE_IMPLEMENT_
extern std::ofstream extractFile;
extern std::ofstream extractFileInv;
extern std::ofstream extractFileOrientation;
extern int maxPhraseLength;
extern int phraseCount;
extern const char* fileNameExtract;
extern bool orientationFlag;
extern bool onlyOutputSpanInfo;
extern bool noFileLimit;
extern bool zipFiles;
extern bool properConditioning;
#endif

#define SAFE_GETLINE(_IS, _LINE, _SIZE, _DELIM) { \
                _IS.getline(_LINE, _SIZE, _DELIM); \
                if(_IS.fail() && !_IS.bad() && !_IS.eof()) _IS.clear(); \
                if (_IS.gcount() == _SIZE-1) { \
                  cerr << "Line too long! Buffer overflow. Delete lines >=" \
                    << _SIZE << " chars or raise LINE_MAX_LENGTH in phrase-extract/extract.cpp" \
                    << endl; \
                    exit(1); \
                } \
              }
#define LINE_MAX_LENGTH 60000

namespace qphrase{


	
	class SentenceAlignment {
	 public:
	  std::vector<std::string> english;
	  std::vector<std::string> foreign;
	  std::vector<int> alignedCountF;
	  std::vector<std::vector<int> > alignedToE;

	  int create( char[], char[], char[], int );
  //  void clear() { delete(alignment); };
	};
	
	class PhraseAppender{
	public:
		virtual void operator()(SentenceAlignment &, int, int, int, int) ;
		virtual void StartSentence(size_t srcSize, size_t tarSize){};
		virtual void EndSentence(){};
	};
	void extractBase( SentenceAlignment & );
	void extract( SentenceAlignment & s);
	void extract( SentenceAlignment & s, PhraseAppender& appender);
	void addPhrase( SentenceAlignment &, int, int, int, int );
	std::vector<std::string> tokenize( char [] );
	bool isAligned ( SentenceAlignment &, int, int );

}

#endif