/***************************************************************************
 *            BasePhraseScoring.h
 *
 *  Wed Jul 22 15:03:34 2009
 *  Copyright  2009  qing
 *  <qing@<host>>
 ****************************************************************************/

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
 */

#ifndef __BASE_PHRASE_SCORING_H_
#define __BASE_PHRASE_SCORING_H_

#include <sstream>
#include <cstdio>
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <stdlib.h>
#include <assert.h>
#include <time.h>
#include "AlignmentPhrase.h"
#include "tables-core.h"
#define SAFE_GETLINE(_IS, _LINE, _SIZE, _DELIM) { \
                _IS.getline(_LINE, _SIZE, _DELIM); \
                if(_IS.fail() && !_IS.bad() && !_IS.eof()) _IS.clear(); \
                if (_IS.gcount() == _SIZE-1) { \
                  cerr << "Line too long! Buffer overflow. Delete lines >=" \
                    << _SIZE << " chars or raise LINE_MAX_LENGTH in phrase-extract/scrore.cpp" \
                    << endl; \
                    exit(1); \
                } \
              }
#define LINE_MAX_LENGTH 10000

namespace qphrase{

class PhraseAlignment {
public:
  int english, foreign;
  std::vector<std::vector<size_t> > alignedToE;
  std::vector<std::vector<size_t> > alignedToF;
  
  bool create( char*, int );
  void clear();
  bool equals( const PhraseAlignment& );
};

class LexicalTable {
public:
  map< WORD_ID, map< WORD_ID, double > > ltable;
  void load( char[] );
};

std::vector<std::string> tokenize( char [] );

void processPhrasePairs(std::vector< PhraseAlignment > & );
	void outputAlignment(const qphrase::AlignmentPhrase &alignmentPhrase);
}

#ifndef __BASE_PHRASE_SCORING_IMPL_
extern std::ofstream phraseTableFile;

extern qphrase::Vocabulary vcbE;
extern qphrase::Vocabulary vcbF;
extern qphrase::LexicalTable lexTable;
extern qphrase::PhraseTable phraseTableE;
extern qphrase::PhraseTable phraseTableF;
extern bool inverseFlag;
extern int phrasePairBase; // only used for "proper" conditioning
#endif
#endif