// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Library General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA

#include "SRLPhraseExtraction.h"
#include "SimpleLogger.h"
#include "SwirlSRLFormat.h"
#include "PhrasalSRLInfo.h"

#include <fstream>

using namespace qphrase;
using namespace qformat;
using namespace std;

PhraseAppenderWithSRL::PhraseAppenderWithSRL
	(const char* inputfile,const char* outputfile, bool disableFileLimit){
	A_FATAL(inputfile!=NULL);
	A_FATAL(outputfile!=NULL);
	disablefl = disableFileLimit;
	phrasecount = 0;
//	phraseoutput.open(outputfile);
	stroutputfile = outputfile;
	if(!srlinput.OpenFile(inputfile))
		P_ERROR("Cannot open input file for SRL extraction %s",inputfile);
	/*if(!phraseoutput)
		P_ERROR("Cannot open output file for SRL extraction %s",outputfile);*/
}

void PhraseAppenderWithSRL::operator()
	(SentenceAlignment &sentence, int startE, int endE, int startF, int endF){
	if(curr == NULL){
		P_ERROR("No SRL information, fall, no extraction at all");
		return;
	}
	// Do default adding
	PhraseAppender::operator()(sentence,startE,endE,startF,endF);
	if (phrasecount % 10000000 == 0 // new file every 1e7 phrases
	    && (!disablefl || phraseCount == 0)) { // only new partial file, if file limit

    // close old file
	    if (!disablefl && phrasecount>0) {
			phraseoutput.close();
	    }
    
    // construct file name
	    char part[10];
	    if (noFileLimit)
			part[0] = '\0';
	    else
			sprintf(part,".part%04d",phrasecount/10000000);  
	    string fileNameExtractPart = stroutputfile + part;
    
    
    // open files
	    phraseoutput.open(fileNameExtractPart.c_str());
	}

	phrasecount++;
	// Now, get the SRLInfo, first output target/srouce information

	// english
	for(int ei=startE;ei<=endE;ei++) {
		phraseoutput << sentence.english[ei] << " ";
	}
	phraseoutput << "||| ";
	for(int fi=startF;fi<=endF;fi++)
		phraseoutput << sentence.foreign[fi] << " ";
	phraseoutput << "||| ";
	CQPhrasalSRLInfo info = CQPhrasalSRLInfo(*curr,startE,endE+1);
	// Format: 
	phraseoutput << info.GetProvidedPredicateCount() << " " //phrase count
			     << info.GetProvidedArgumentCount()  << " "; //argument count
	for(int i = 0; i< info.GetProvidedArgumentCount() ; i++){
		const CQPhrasalSRLInfo::_QArgumentInfo& c = info.GetArgumentInfo(i);
		phraseoutput << c.PredicateLocation << "/" << c.Completeness ;
		if (i<info.GetProvidedArgumentCount()-1){
			phraseoutput << "_";
		}
	}
	phraseoutput<<std::endl<<std::flush;
}

void PhraseAppenderWithSRL::StartSentence(size_t srcSize, size_t tarSize){
	if(srlinput.HasNext()){
		boost::shared_ptr<CQSRLSentence> ret = 
			boost::static_pointer_cast<CQSRLSentence,CQObject>(srlinput.ReadNext());
		if(ret->GetLength()!=srcSize){
			P_ERROR("Sentence length mismatch! target sentence %d and in SRL %d"
			        ,srcSize,ret->GetLength());
			return;
		}else{
			curr = ret;
		}
	}else{
		P_ERROR("SRL file ends, no more sentence can be provided");
	}
}
void PhraseAppenderWithSRL::EndSentence(){
	curr.reset();
}


