// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Library General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA

#include "Params.h"
#include "BasePhraseExtraction.h"
#include "SRLPhraseExtraction.h"
#include <stdio.h>
#include <boost/scoped_ptr.hpp>

using namespace std;

using namespace qphrase;
int main(int argc, const char* argv[])
{
	BEGIN_PARAMETERS; // Start parameter definition
	REQ_PARAM(string,EnglishCorpus,"english,e","The English corpus file. (Which you want to translate into)","Input");
	REQ_PARAM(string,FrenchCorpus,"french,f","The target corpus file. (Which you translate from)","Input");
	REQ_PARAM(string,Alignment,"align,a","The alignment file.","Input");
	DEF_PARAM(string,SRLInputfile,"","srlinput,i","The SRL file.","Input");	
	DEF_PARAM(int,MaxLength,7,"max-length,m","The max phrase length.","Configure");
	DEF_PARAM(int,SrlType,0,"srl-type,s",
	          "Specify the what kind of SRL information will be extracted, valid"
	          " types: 0: Only argument completeness info",
	          "Configure");	
	REQ_PARAM(string,ExtractFile,"extract,x","The output extraction file.","Output");
	DEF_SWITCH(bool,OnlyOutputSpanInfo,"OnlyOutputSpanInfo","Only output span information","Configure");
	DEF_SWITCH(bool,ProperConditioning,"ProperConditioning","","Configure");
	DEF_SWITCH(bool,NoFileLimit,"NoFileLimit","Don't output file parts","Configure");		
	DEF_SWITCH(bool,Orientation,"Orientation","Extract orientation file","Configure");		
	DEF_SWITCH(bool,help,"help,h","Print help screen","");

	END_PARAMETERS;
	PARSE_COMMANDLINE(argc,argv);

	if(help){
		PRINT_HELP;
		exit(-1);
	}
	PRINT_CONFIG;

	const char* fileNameE = EnglishCorpus.c_str();
	const char* fileNameF = FrenchCorpus.c_str();
	const char* fileNameA = Alignment.c_str();
  	fileNameExtract = ExtractFile.c_str();
	maxPhraseLength = MaxLength;
	onlyOutputSpanInfo = OnlyOutputSpanInfo;
	properConditioning = ProperConditioning;
	noFileLimit = NoFileLimit;
	orientationFlag = Orientation;

	ifstream eFile;
	ifstream fFile;
	ifstream aFile;
	eFile.open(fileNameE);
	fFile.open(fileNameF);
	aFile.open(fileNameA);
	istream *eFileP = &eFile;
	istream *fFileP = &fFile;
	istream *aFileP = &aFile;

	PhraseAppender* appender;

	if (SRLInputfile.length()>0){
		string srlfile = ExtractFile+".srl";
		appender = new PhraseAppenderWithSRL(SRLInputfile.c_str(),
		                                     srlfile.c_str(),noFileLimit);
	}
	else
		appender = new PhraseAppender();
  
	int i=0;
	while(true) {
		i++;
		if (i%10000 == 0) cerr << "." << flush;
		char englishString[LINE_MAX_LENGTH];
		char foreignString[LINE_MAX_LENGTH];
		char alignmentString[LINE_MAX_LENGTH];
		SAFE_GETLINE((*eFileP), englishString, LINE_MAX_LENGTH, '\n');
		if (eFileP->eof()) break;
		SAFE_GETLINE((*fFileP), foreignString, LINE_MAX_LENGTH, '\n');
		SAFE_GETLINE((*aFileP), alignmentString, LINE_MAX_LENGTH, '\n');
		SentenceAlignment sentence;
    // cout << "read in: " << englishString << " & " << foreignString << " & " << alignmentString << endl;
    //az: output src, tgt, and alingment line
	    if (onlyOutputSpanInfo) {
    	  cout << "LOG: SRC: " << foreignString << endl;
	      cout << "LOG: TGT: " << englishString << endl;
    	  cout << "LOG: ALT: " << alignmentString << endl;
    	  cout << "LOG: PHRASES_BEGIN:" << endl;
    	}
      
    	if (sentence.create( 
	        englishString, foreignString, alignmentString, i )) {
			extract(sentence, *appender);
			if (properConditioning) extractBase(sentence);
   		}
	    if (onlyOutputSpanInfo) 
			cout << "LOG: PHRASES_END:" << endl; //az: mark end of phrases
	  }

	  eFile.close();
	  fFile.close();
	  aFile.close();
	  //az: only close if we actually opened it
	  if (!onlyOutputSpanInfo) {
	    extractFile.close();
    	extractFileInv.close();
	    if (orientationFlag) extractFileOrientation.close();
	  }
}




