// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Library General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA


#include "SwirlSRLFormat.h"
#include "PhrasalSRLInfo.h"
#include "SimpleLogger.h"

using namespace qformat;
using namespace boost;
int main(int argc, const char* argv[]){
	init_logging(0);
	CQSwirlSRLFormat reader;

	reader.OpenFile("../format-samples/en.col.ra.asc.src.part.000.srl");

	int i = 0;
	while(reader.HasNext()){
		boost::shared_ptr<CQSRLSentence> ret = 
			static_pointer_cast<CQSRLSentence,CQObject>(reader.ReadNext());
		P_INFO("Read %d sentence, has %d predicates , length %d", 
		       ++i, ret->GetPredicateCount(), ret->GetLength());
		for(int j = 0 ; j< ret->GetPredicateCount() ; j++){
			CQSRLPredicate & pred = ret->GetPredicate(j);
			P_INFO("Predicate %d : %s, with %d Arguments",
			       j+1, pred.GetPredicate().c_str(), pred.GetArgumentCount());
			for(int k = 0; k<pred.GetArgumentCount(); k++){
				P_INFO("Argument %d : %s [%d-%d]",
				       k,pred.GetArgumentType(k).c_str(),
				       pred.GetArgumentBegin(k),pred.GetArgumentEnd(k)
				       );
			}
		}
		size_t pstart = 10, pend = 17;
		if(ret->GetLength ()>40){
			P_INFO("Testing extracting phrasal information, phrase %d - %d ",
			       pstart,pend);
			CQPhrasalSRLInfo pinfo(*ret,pstart,pend);
			P_INFO("Provided %d predicates and %d arguments",
			       pinfo.GetProvidedPredicateCount(),
			       pinfo.GetProvidedArgumentCount()
			       );
			for(size_t k = 0 ; k < pinfo.GetProvidedPredicateCount() ; k++){
				const CQPhrasalSRLInfo::_QPredicateInfo & pred 
					= pinfo.GetPredicateInfo(k);
				P_INFO("Predicate %d [%s] : NumArguments : %d | "
				       "NumCompleteArguments: %d | NumIncompleteArguments: %d |"
				       "NumMissingArguments: %d | "
				       "NumIncompleteArugmentsBefore: %d |"
   				       "NumIncompleteArugmentsAfter: %d |"
   				       "NumMissingArgumentsBefore: %d |"	
   				       "NumMissingArgumentsAfter: %d |"	
				       ,
				       k, pred.Predicate.c_str(),pred.NumArguments,
				       pred.NumCompleteArguments,pred.NumIncompleteArguments,
				       pred.NumMissingArguments,
				       pred.NumIncompleteArugmentsBefore,
				       pred.NumIncompleteArugmentsAfter,
				       pred.NumMissingArgumentsBefore,
				       pred.NumMissingArgumentsAfter
				       );
			}
			for(size_t k = 0 ; k < pinfo.GetProvidedArgumentCount() ; k++){
				const CQPhrasalSRLInfo::_QArgumentInfo & arg 
					= pinfo.GetArgumentInfo(k);
				P_INFO("Argument %d [%s of %s] : PredicateLocation : %d | "
				       "Completeness : %d ",
				       k,arg.Label.c_str(),arg.Predicate.c_str()
				       ,arg.PredicateLocation,
				       arg.Completeness
				       );
			}
		}
	}

	
}

